/* (C) 2009 by Deino Software */

#ifndef MPICH_NMPI_H_INCLUDED
#define MPICH_NMPI_H_INCLUDED

/* 
 * This file provides a flexible way to map MPI routines that are used
 * within the MPICH2 implementation to either the MPI or PMPI versions.
 * In normal use, it is appropriate to use PMPI, but in some cases, 
 * using the MPI routines instead is desired.
 */

/*
 * When adding names, make sure that you add them to both branches.
 * This allows the --enable-nmpi-as-mpi switch in configure to 
 * cause the internal routines to use the MPI versions; this is good for 
 * using tools that use the MPI profiling interface to collect data (e.g.,
 * for people who what to see what happens when the collective operations 
 * are implemented over MPI point-to-point.
 */

#ifdef USE_MPI_FOR_NMPI
#define NMPI_Abort(a,b)							{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Abort b ; MPIR_Nest_decr(); }
#define NMPI_Bcast(a,b)							{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Bcast b ; MPIR_Nest_decr(); }
#define NMPI_Get_count(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Get_count b ; MPIR_Nest_decr(); }
#define NMPI_Pack(a,b)							{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Pack b ; MPIR_Nest_decr(); }
#define NMPI_Pack_size(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Pack_size b ; MPIR_Nest_decr(); }
#define NMPI_Reduce(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Reduce b ; MPIR_Nest_decr(); }
#define NMPI_Reduce_scatter(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Reduce_scatter b ; MPIR_Nest_decr(); }
#define NMPI_Unpack(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Unpack b ; MPIR_Nest_decr(); }
#define NMPI_Wait(a,b)							{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Wait b ; MPIR_Nest_decr(); }
#define NMPI_Test(a,b)							{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Test b ; MPIR_Nest_decr(); }
#define NMPI_Allreduce(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Allreduce b ; MPIR_Nest_decr(); }
#define NMPI_Allgather(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Allgather b ; MPIR_Nest_decr(); }
#define NMPI_Comm_get_attr(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Comm_get_attr b ; MPIR_Nest_decr(); }
#define NMPI_Comm_set_attr(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Comm_set_attr b ; MPIR_Nest_decr(); }
#define NMPI_Comm_delete_attr(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Comm_delete_attr b ; MPIR_Nest_decr(); }
#define NMPI_Comm_create_keyval(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Comm_create_keyval b ; MPIR_Nest_decr(); }
#define NMPI_Comm_free_keyval(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Comm_free_keyval b ; MPIR_Nest_decr(); }
#define NMPI_Comm_group(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Comm_group b ; MPIR_Nest_decr(); }
#define NMPI_Comm_remote_group(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Comm_remote_group b ; MPIR_Nest_decr(); }
#define NMPI_Group_compare(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Group_compare b ; MPIR_Nest_decr(); }
#define NMPI_Group_free(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Group_free b ; MPIR_Nest_decr(); }
#define NMPI_Comm_split(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Comm_split b ; MPIR_Nest_decr(); }
#define NMPI_Comm_size(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Comm_size b ; MPIR_Nest_decr(); }
#define NMPI_Comm_rank(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Comm_rank b ; MPIR_Nest_decr(); }
#define NMPI_Alltoall(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Alltoall b ; MPIR_Nest_decr(); }
#define NMPI_Isend(a,b)							{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Isend b ; MPIR_Nest_decr(); }
#define NMPI_Irecv(a,b)							{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Irecv b ; MPIR_Nest_decr(); }
#define NMPI_Recv(a,b)							{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Recv b ; MPIR_Nest_decr(); }
#define NMPI_Send(a,b)							{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Send b ; MPIR_Nest_decr(); }
#define NMPI_Waitall(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Waitall b ; MPIR_Nest_decr(); }
#define NMPI_Sendrecv(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Sendrecv b ; MPIR_Nest_decr(); }
#define NMPI_Type_extent(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Type_extent b ; MPIR_Nest_decr(); }
#define NMPI_Comm_free(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Comm_free b ; MPIR_Nest_decr(); }
#define NMPI_Comm_dup(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Comm_dup b ; MPIR_Nest_decr(); }
#define NMPI_Type_lb(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Type_lb b ; MPIR_Nest_decr(); }
#define NMPI_Type_indexed(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Type_indexed b ; MPIR_Nest_decr(); }
#define NMPI_Type_commit(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Type_commit b ; MPIR_Nest_decr(); }
#define NMPI_Type_free(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Type_free b ; MPIR_Nest_decr(); }
#define NMPI_Type_size(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Type_size b ; MPIR_Nest_decr(); }
#define NMPI_Type_get_extent(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Type_get_extent b ; MPIR_Nest_decr(); }
#define NMPI_Cart_rank(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Cart_rank b ; MPIR_Nest_decr(); }
#define NMPI_Cart_map(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Cart_map b ; MPIR_Nest_decr(); }
#define NMPI_Graph_map(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Graph_map b ; MPIR_Nest_decr(); }
#define NMPI_Iprobe(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Iprobe b ; MPIR_Nest_decr(); }
#define NMPI_Barrier(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Barrier b ; MPIR_Nest_decr(); }
#define NMPI_Type_get_true_extent(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Type_get_true_extent b ; MPIR_Nest_decr(); }
#define NMPI_Group_translate_ranks(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Group_translate_ranks b ; MPIR_Nest_decr(); }
#define NMPI_Type_create_indexed_block(a,b) { MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Type_create_indexed_block b ; MPIR_Nest_decr(); }
#define NMPI_Wtime							MPI_Wtime
#define NMPI_Info_get_nkeys(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Info_get_nkeys b ; MPIR_Nest_decr(); }
#define NMPI_Info_get_nthkey(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Info_get_nthkey b ; MPIR_Nest_decr(); }
#define NMPI_Info_get_valuelen(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Info_get_valuelen b ; MPIR_Nest_decr(); }
#define NMPI_Info_get(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Info_get b ; MPIR_Nest_decr(); }
#define NMPI_Info_create(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Info_create b ; MPIR_Nest_decr(); }
#define NMPI_Info_set(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Info_set b ; MPIR_Nest_decr(); }
#define NMPI_Comm_get_name(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Comm_get_name b ; MPIR_Nest_decr(); }
#define NMPI_Comm_get_errhandler(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Comm_get_errhandler b ; MPIR_Nest_decr(); }
#define NMPI_Comm_set_errhandler(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Comm_set_errhandler b ; MPIR_Nest_decr(); }
#define NMPI_Comm_create_errhandler(a,b)	{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Comm_create_errhandler b ; MPIR_Nest_decr(); }
#define NMPI_Comm_call_errhandler(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Comm_call_errhandler b ; MPIR_Nest_decr(); }
#define NMPI_Open_port(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Open_port b ; MPIR_Nest_decr(); }
#define NMPI_Comm_accept(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Comm_accept b ; MPIR_Nest_decr(); }
#define NMPI_Comm_connect(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Comm_connect b ; MPIR_Nest_decr(); }
#define NMPI_Grequest_start(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Grequest_start b ; MPIR_Nest_decr(); }
#define NMPI_Grequest_complete(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Grequest_complete b ; MPIR_Nest_decr(); }
#define NMPI_Cancel(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Cancel b ; MPIR_Nest_decr(); }
#define NMPI_Test_cancelled(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Test_cancelled b ; MPIR_Nest_decr(); }
#define NMPI_Ibsend(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Ibsend b ; MPIR_Nest_decr(); }
#define NMPI_Buffer_detach(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Buffer_detach b ; MPIR_Nest_decr(); }
#define NMPI_Address(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Address b ; MPIR_Nest_decr(); }
#define NMPI_Attr_get(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Attr_get b ; MPIR_Nest_decr(); }
#define NMPI_Attr_put(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Attr_put b ; MPIR_Nest_decr(); }
#define NMPI_Comm_test_inter(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Comm_test_inter b ; MPIR_Nest_decr(); }
#define NMPI_File_read(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_File_read b ; MPIR_Nest_decr(); }
#define NMPI_File_read_at(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_File_read_at b ; MPIR_Nest_decr(); }
#define NMPI_File_read_shared(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_File_read_shared b ; MPIR_Nest_decr(); }
#define NMPI_File_write(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_File_write b ; MPIR_Nest_decr(); }
#define NMPI_File_write_at(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_File_write_at b ; MPIR_Nest_decr(); }
#define NMPI_File_write_shared(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_File_write_shared b ; MPIR_Nest_decr(); }
#define NMPI_Gather(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Gather b ; MPIR_Nest_decr(); }
#define NMPI_Gatherv(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Gatherv b ; MPIR_Nest_decr(); }
#define NMPI_Get_elements(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Get_elements b ; MPIR_Nest_decr(); }
#define NMPI_Get_processor_name(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Get_processor_name b ; MPIR_Nest_decr(); }
#define NMPI_Info_delete(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Info_delete b ; MPIR_Nest_decr(); }
#define NMPI_Info_dup(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Info_dup b ; MPIR_Nest_decr(); }
#define NMPI_Info_free(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Info_free b ; MPIR_Nest_decr(); }
#define NMPI_Keyval_create(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Keyval_create b ; MPIR_Nest_decr(); }
#define NMPI_Status_set_cancelled(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Status_set_cancelled b ; MPIR_Nest_decr(); }
#define NMPI_Status_set_elements(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Status_set_elements b ; MPIR_Nest_decr(); }
#define NMPI_Type_contiguous(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Type_contiguous b ; MPIR_Nest_decr(); }
#define NMPI_Type_get_contents(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Type_get_contents b ; MPIR_Nest_decr(); }
#define NMPI_Type_get_envelope(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Type_get_envelope b ; MPIR_Nest_decr(); }
#define NMPI_Type_hindexed(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Type_hindexed b ; MPIR_Nest_decr(); }
#define NMPI_Type_hvector(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Type_hvector b ; MPIR_Nest_decr(); }
#define NMPI_Type_struct(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Type_struct b ; MPIR_Nest_decr(); }
#define NMPI_Type_vector(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Type_vector b ; MPIR_Nest_decr(); }
#define NMPI_Type_create_resized(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Type_create_resized b ; MPIR_Nest_decr(); }
#define NMPI_Type_create_f90_integer(a,b)   { MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Type_create_f90_integer b ; MPIR_Nest_decr(); }
#define NMPI_Type_create_f90_complex(a,b)   { MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Type_create_f90_complex b ; MPIR_Nest_decr(); }
#define NMPI_Type_create_f90_real(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Type_create_f90_real b ; MPIR_Nest_decr(); }
#define NMPI_Type_create_darray(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Type_create_darray b ; MPIR_Nest_decr(); }
#define NMPI_Type_create_subarray(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Type_create_subarray b ; MPIR_Nest_decr(); }
#define NMPI_Add_error_class(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Add_error_class b ; MPIR_Nest_decr(); }
#define NMPI_Add_error_code(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Add_error_code b ; MPIR_Nest_decr(); }
#define NMPI_Add_error_string(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Add_error_string b ; MPIR_Nest_decr(); }
#define NMPI_Type_dup(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = MPI_Type_dup b ; MPIR_Nest_decr(); }
#else
#define NMPI_Abort(a,b)							{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Abort b ; MPIR_Nest_decr(); }
#define NMPI_Bcast(a,b)							{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Bcast b ; MPIR_Nest_decr(); }
#define NMPI_Get_count(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Get_count b ; MPIR_Nest_decr(); }
#define NMPI_Pack(a,b)							{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Pack b ; MPIR_Nest_decr(); }
#define NMPI_Pack_size(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Pack_size b ; MPIR_Nest_decr(); }
#define NMPI_Reduce(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Reduce b ; MPIR_Nest_decr(); }
#define NMPI_Reduce_scatter(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Reduce_scatter b ; MPIR_Nest_decr(); }
#define NMPI_Unpack(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Unpack b ; MPIR_Nest_decr(); }
#define NMPI_Wait(a,b)							{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Wait b ; MPIR_Nest_decr(); }
#define NMPI_Test(a,b)							{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Test b ; MPIR_Nest_decr(); }
#define NMPI_Allreduce(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Allreduce b ; MPIR_Nest_decr(); }
#define NMPI_Allgather(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Allgather b ; MPIR_Nest_decr(); }
#define NMPI_Comm_get_attr(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Comm_get_attr b ; MPIR_Nest_decr(); }
#define NMPI_Comm_set_attr(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Comm_set_attr b ; MPIR_Nest_decr(); }
#define NMPI_Comm_delete_attr(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Comm_delete_attr b ; MPIR_Nest_decr(); }
#define NMPI_Comm_create_keyval(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Comm_create_keyval b ; MPIR_Nest_decr(); }
#define NMPI_Comm_free_keyval(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Comm_free_keyval b ; MPIR_Nest_decr(); }
#define NMPI_Comm_group(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Comm_group b ; MPIR_Nest_decr(); }
#define NMPI_Comm_remote_group(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Comm_remote_group b ; MPIR_Nest_decr(); }
#define NMPI_Group_compare(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Group_compare b ; MPIR_Nest_decr(); }
#define NMPI_Group_free(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Group_free b ; MPIR_Nest_decr(); }
#define NMPI_Comm_split(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Comm_split b ; MPIR_Nest_decr(); }
#define NMPI_Comm_size(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Comm_size b ; MPIR_Nest_decr(); }
#define NMPI_Comm_rank(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Comm_rank b ; MPIR_Nest_decr(); }
#define NMPI_Alltoall(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Alltoall b ; MPIR_Nest_decr(); }
#define NMPI_Isend(a,b)							{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Isend b ; MPIR_Nest_decr(); }
#define NMPI_Irecv(a,b)							{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Irecv b ; MPIR_Nest_decr(); }
#define NMPI_Recv(a,b)							{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Recv b ; MPIR_Nest_decr(); }
#define NMPI_Send(a,b)							{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Send b ; MPIR_Nest_decr(); }
#define NMPI_Waitall(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Waitall b ; MPIR_Nest_decr(); }
#define NMPI_Sendrecv(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Sendrecv b ; MPIR_Nest_decr(); }
#define NMPI_Type_extent(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Type_extent b ; MPIR_Nest_decr(); }
#define NMPI_Comm_free(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Comm_free b ; MPIR_Nest_decr(); }
#define NMPI_Comm_dup(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Comm_dup b ; MPIR_Nest_decr(); }
#define NMPI_Type_lb(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Type_lb b ; MPIR_Nest_decr(); }
#define NMPI_Type_indexed(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Type_indexed b ; MPIR_Nest_decr(); }
#define NMPI_Type_commit(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Type_commit b ; MPIR_Nest_decr(); }
#define NMPI_Type_free(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Type_free b ; MPIR_Nest_decr(); }
#define NMPI_Type_size(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Type_size b ; MPIR_Nest_decr(); }
#define NMPI_Type_get_extent(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Type_get_extent b ; MPIR_Nest_decr(); }
#define NMPI_Cart_rank(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Cart_rank b ; MPIR_Nest_decr(); }
#define NMPI_Cart_map(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Cart_map b ; MPIR_Nest_decr(); }
#define NMPI_Graph_map(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Graph_map b ; MPIR_Nest_decr(); }
#define NMPI_Iprobe(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Iprobe b ; MPIR_Nest_decr(); }
#define NMPI_Barrier(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Barrier b ; MPIR_Nest_decr(); }
#define NMPI_Type_get_true_extent(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Type_get_true_extent b ; MPIR_Nest_decr(); }
#define NMPI_Group_translate_ranks(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Group_translate_ranks b ; MPIR_Nest_decr(); }
#define NMPI_Type_create_indexed_block(a,b) { MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Type_create_indexed_block b ; MPIR_Nest_decr(); }
#define NMPI_Wtime							PMPI_Wtime
#define NMPI_Info_get_nkeys(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Info_get_nkeys b ; MPIR_Nest_decr(); }
#define NMPI_Info_get_nthkey(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Info_get_nthkey b ; MPIR_Nest_decr(); }
#define NMPI_Info_get_valuelen(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Info_get_valuelen b ; MPIR_Nest_decr(); }
#define NMPI_Info_get(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Info_get b ; MPIR_Nest_decr(); }
#define NMPI_Info_create(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Info_create b ; MPIR_Nest_decr(); }
#define NMPI_Info_set(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Info_set b ; MPIR_Nest_decr(); }
#define NMPI_Comm_get_name(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Comm_get_name b ; MPIR_Nest_decr(); }
#define NMPI_Comm_get_errhandler(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Comm_get_errhandler b ; MPIR_Nest_decr(); }
#define NMPI_Comm_set_errhandler(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Comm_set_errhandler b ; MPIR_Nest_decr(); }
#define NMPI_Comm_create_errhandler(a,b)	{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Comm_create_errhandler b ; MPIR_Nest_decr(); }
#define NMPI_Comm_call_errhandler(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Comm_call_errhandler b ; MPIR_Nest_decr(); }
#define NMPI_Open_port(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Open_port b ; MPIR_Nest_decr(); }
#define NMPI_Close_port(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Close_port b ; MPIR_Nest_decr(); }
#define NMPI_Comm_accept(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Comm_accept b ; MPIR_Nest_decr(); }
#define NMPI_Comm_connect(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Comm_connect b ; MPIR_Nest_decr(); }
#define NMPI_Grequest_start(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Grequest_start b ; MPIR_Nest_decr(); }
#define NMPI_Grequest_complete(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Grequest_complete b ; MPIR_Nest_decr(); }
#define NMPI_Cancel(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Cancel b ; MPIR_Nest_decr(); }
#define NMPI_Test_cancelled(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Test_cancelled b ; MPIR_Nest_decr(); }
#define NMPI_Ibsend(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Ibsend b ; MPIR_Nest_decr(); }
#define NMPI_Buffer_detach(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Buffer_detach b ; MPIR_Nest_decr(); }
#define NMPI_Address(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Address b ; MPIR_Nest_decr(); }
#define NMPI_Attr_get(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Attr_get b ; MPIR_Nest_decr(); }
#define NMPI_Attr_put(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Attr_put b ; MPIR_Nest_decr(); }
#define NMPI_Comm_test_inter(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Comm_test_inter b ; MPIR_Nest_decr(); }
#define NMPI_File_read(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_File_read b ; MPIR_Nest_decr(); }
#define NMPI_File_read_at(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_File_read_at b ; MPIR_Nest_decr(); }
#define NMPI_File_read_shared(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_File_read_shared b ; MPIR_Nest_decr(); }
#define NMPI_File_write(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_File_write b ; MPIR_Nest_decr(); }
#define NMPI_File_write_at(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_File_write_at b ; MPIR_Nest_decr(); }
#define NMPI_File_write_shared(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_File_write_shared b ; MPIR_Nest_decr(); }
#define NMPI_Gather(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Gather b ; MPIR_Nest_decr(); }
#define NMPI_Gatherv(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Gatherv b ; MPIR_Nest_decr(); }
#define NMPI_Get_elements(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Get_elements b ; MPIR_Nest_decr(); }
#define NMPI_Get_processor_name(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Get_processor_name b ; MPIR_Nest_decr(); }
#define NMPI_Info_delete(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Info_delete b ; MPIR_Nest_decr(); }
#define NMPI_Info_dup(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Info_dup b ; MPIR_Nest_decr(); }
#define NMPI_Info_free(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Info_free b ; MPIR_Nest_decr(); }
#define NMPI_Keyval_create(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Keyval_create b ; MPIR_Nest_decr(); }
#define NMPI_Status_set_cancelled(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Status_set_cancelled b ; MPIR_Nest_decr(); }
#define NMPI_Status_set_elements(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Status_set_elements b ; MPIR_Nest_decr(); }
#define NMPI_Type_contiguous(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Type_contiguous b ; MPIR_Nest_decr(); }
#define NMPI_Type_get_contents(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Type_get_contents b ; MPIR_Nest_decr(); }
#define NMPI_Type_get_envelope(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Type_get_envelope b ; MPIR_Nest_decr(); }
#define NMPI_Type_hindexed(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Type_hindexed b ; MPIR_Nest_decr(); }
#define NMPI_Type_hvector(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Type_hvector b ; MPIR_Nest_decr(); }
#define NMPI_Type_struct(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Type_struct b ; MPIR_Nest_decr(); }
#define NMPI_Type_vector(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Type_vector b ; MPIR_Nest_decr(); }
#define NMPI_Type_create_resized(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Type_create_resized b ; MPIR_Nest_decr(); }
#define NMPI_Type_create_f90_integer(a,b)   { MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Type_create_f90_integer b ; MPIR_Nest_decr(); }
#define NMPI_Type_create_f90_complex(a,b)   { MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Type_create_f90_complex b ; MPIR_Nest_decr(); }
#define NMPI_Type_create_f90_real(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Type_create_f90_real b ; MPIR_Nest_decr(); }
#define NMPI_Type_create_darray(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Type_create_darray b ; MPIR_Nest_decr(); }
#define NMPI_Type_create_subarray(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Type_create_subarray b ; MPIR_Nest_decr(); }
#define NMPI_Add_error_class(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Add_error_class b ; MPIR_Nest_decr(); }
#define NMPI_Add_error_code(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Add_error_code b ; MPIR_Nest_decr(); }
#define NMPI_Add_error_string(a,b)			{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Add_error_string b ; MPIR_Nest_decr(); }
#define NMPI_Type_dup(a,b)					{ MPIU_THREADPRIV_DECL; MPIU_THREADPRIV_GET; MPIR_Nest_incr(); a = PMPI_Type_dup b ; MPIR_Nest_decr(); }
#endif
#endif /* MPICH_NMPI_H_INCLUDED */
