/* (C) 2009 by Deino Software */

#ifndef MPI_INCLUDED
#define MPI_INCLUDED

/* user include file for MPI programs */

/* Provide a definition so users can protect code that uses optional DeinoMPI extensions */
#define DEINO_MPI

/* Support Windows extension to specify calling convention */
#define MPI_CALL __cdecl

/* Support Windows extension to specify which functions are exported from shared (DLL) libraries */
#ifdef MPI_STATIC_LIBRARY
# define MPI_DLL_SPEC
#else
# ifdef MPI_EXPORTS
#  define MPI_DLL_SPEC __declspec(dllexport)
# else
#  define MPI_DLL_SPEC __declspec(dllimport)
# endif
#endif

/* Make sure wchar_t is available */
#ifndef _WCHAR_T_DEFINED
typedef unsigned short wchar_t;
#define _WCHAR_T_DEFINED
#endif

/* Use the W or A versions of functions that take string parameters depending on whether UNICODE is defined or not */
#if defined(UNICODE) || defined(_UNICODE)

#define MPI_Get_processor_name	  MPI_Get_processor_nameW
#define MPI_Error_string			MPI_Error_stringW
#define MPI_Init					MPI_InitW
#define MPI_Init_thread			 MPI_Init_threadW
#define MPI_Open_port			   MPI_Open_portW
#define MPI_Close_port			  MPI_Close_portW
#define MPI_Comm_accept			 MPI_Comm_acceptW
#define MPI_Comm_connect			MPI_Comm_connectW
#define MPI_Comm_spawn			  MPI_Comm_spawnW
#define MPI_Comm_spawn_multiple	 MPI_Comm_spawn_multipleW
#define MPI_Lookup_name			 MPI_Lookup_nameW
#define MPI_Publish_name			MPI_Publish_nameW
#define MPI_Unpublish_name		  MPI_Unpublish_nameW
#define MPI_Add_error_string		MPI_Add_error_stringW
#define MPI_Comm_get_name		   MPI_Comm_get_nameW
#define MPI_Comm_set_name		   MPI_Comm_set_nameW
#define MPI_Type_get_name		   MPI_Type_get_nameW
#define MPI_Type_set_name		   MPI_Type_set_nameW
#define MPI_Win_get_name			MPI_Win_get_nameW
#define MPI_Win_set_name			MPI_Win_set_nameW
#define MPI_Info_delete			 MPI_Info_deleteW
#define MPI_Info_get				MPI_Info_getW
#define MPI_Info_get_nthkey		 MPI_Info_get_nthkeyW
#define MPI_Info_get_valuelen	   MPI_Info_get_valuelenW
#define MPI_Info_set				MPI_Info_setW
#define MPI_Pack_external		   MPI_Pack_externalW
#define MPI_Pack_external_size	  MPI_Pack_external_sizeW
#define MPI_Unpack_external		 MPI_Unpack_externalW
#define MPI_File_open			   MPI_File_openW
#define MPI_File_delete			 MPI_File_deleteW
#define MPI_File_set_view		   MPI_File_set_viewW
#define MPI_File_get_view		   MPI_File_get_viewW
#define MPI_Register_datarep		MPI_Register_datarepW

#define PMPI_Get_processor_name	 PMPI_Get_processor_nameW
#define PMPI_Error_string		   PMPI_Error_stringW
#define PMPI_Init				   PMPI_InitW
#define PMPI_Init_thread			PMPI_Init_threadW
#define PMPI_Open_port			  PMPI_Open_portW
#define PMPI_Close_port			 PMPI_Close_portW
#define PMPI_Comm_accept			PMPI_Comm_acceptW
#define PMPI_Comm_connect		   PMPI_Comm_connectW
#define PMPI_Comm_spawn			 PMPI_Comm_spawnW
#define PMPI_Comm_spawn_multiple	PMPI_Comm_spawn_multipleW
#define PMPI_Lookup_name			PMPI_Lookup_nameW
#define PMPI_Publish_name		   PMPI_Publish_nameW
#define PMPI_Unpublish_name		 PMPI_Unpublish_nameW
#define PMPI_Add_error_string	   PMPI_Add_error_stringW
#define PMPI_Comm_get_name		  PMPI_Comm_get_nameW
#define PMPI_Comm_set_name		  PMPI_Comm_set_nameW
#define PMPI_Type_get_name		  PMPI_Type_get_nameW
#define PMPI_Type_set_name		  PMPI_Type_set_nameW
#define PMPI_Win_get_name		   PMPI_Win_get_nameW
#define PMPI_Win_set_name		   PMPI_Win_set_nameW
#define PMPI_Info_delete			PMPI_Info_deleteW
#define PMPI_Info_get			   PMPI_Info_getW
#define PMPI_Info_get_nthkey		PMPI_Info_get_nthkeyW
#define PMPI_Info_get_valuelen	  PMPI_Info_get_valuelenW
#define PMPI_Info_set			   PMPI_Info_setW
#define PMPI_Pack_external		  PMPI_Pack_externalW
#define PMPI_Pack_external_size	 PMPI_Pack_external_sizeW
#define PMPI_Unpack_external		PMPI_Unpack_externalW
#define PMPI_File_open			  PMPI_File_openW
#define PMPI_File_delete			PMPI_File_deleteW
#define PMPI_File_set_view		  PMPI_File_set_viewW
#define PMPI_File_get_view		  PMPI_File_get_viewW
#define PMPI_Register_datarep	   PMPI_Register_datarepW

#else

#define MPI_Get_processor_name	  MPI_Get_processor_nameA
#define MPI_Error_string			MPI_Error_stringA
#define MPI_Init					MPI_InitA
#define MPI_Init_thread			 MPI_Init_threadA
#define MPI_Open_port			   MPI_Open_portA
#define MPI_Close_port			  MPI_Close_portA
#define MPI_Comm_accept			 MPI_Comm_acceptA
#define MPI_Comm_connect			MPI_Comm_connectA
#define MPI_Comm_spawn			  MPI_Comm_spawnA
#define MPI_Comm_spawn_multiple	 MPI_Comm_spawn_multipleA
#define MPI_Lookup_name			 MPI_Lookup_nameA
#define MPI_Publish_name			MPI_Publish_nameA
#define MPI_Unpublish_name		  MPI_Unpublish_nameA
#define MPI_Add_error_string		MPI_Add_error_stringA
#define MPI_Comm_get_name		   MPI_Comm_get_nameA
#define MPI_Comm_set_name		   MPI_Comm_set_nameA
#define MPI_Type_get_name		   MPI_Type_get_nameA
#define MPI_Type_set_name		   MPI_Type_set_nameA
#define MPI_Win_get_name			MPI_Win_get_nameA
#define MPI_Win_set_name			MPI_Win_set_nameA
#define MPI_Info_delete			 MPI_Info_deleteA
#define MPI_Info_get				MPI_Info_getA
#define MPI_Info_get_nthkey		 MPI_Info_get_nthkeyA
#define MPI_Info_get_valuelen	   MPI_Info_get_valuelenA
#define MPI_Info_set				MPI_Info_setA
#define MPI_Pack_external		   MPI_Pack_externalA
#define MPI_Pack_external_size	  MPI_Pack_external_sizeA
#define MPI_Unpack_external		 MPI_Unpack_externalA
#define MPI_File_open			   MPI_File_openA
#define MPI_File_delete			 MPI_File_deleteA
#define MPI_File_set_view		   MPI_File_set_viewA
#define MPI_File_get_view		   MPI_File_get_viewA
#define MPI_Register_datarep		MPI_Register_datarepA

#define PMPI_Get_processor_name	 PMPI_Get_processor_nameA
#define PMPI_Error_string		   PMPI_Error_stringA
#define PMPI_Init				   PMPI_InitA
#define PMPI_Init_thread			PMPI_Init_threadA
#define PMPI_Open_port			  PMPI_Open_portA
#define PMPI_Close_port			 PMPI_Close_portA
#define PMPI_Comm_accept			PMPI_Comm_acceptA
#define PMPI_Comm_connect		   PMPI_Comm_connectA
#define PMPI_Comm_spawn			 PMPI_Comm_spawnA
#define PMPI_Comm_spawn_multiple	PMPI_Comm_spawn_multipleA
#define PMPI_Lookup_name			PMPI_Lookup_nameA
#define PMPI_Publish_name		   PMPI_Publish_nameA
#define PMPI_Unpublish_name		 PMPI_Unpublish_nameA
#define PMPI_Add_error_string	   PMPI_Add_error_stringA
#define PMPI_Comm_get_name		  PMPI_Comm_get_nameA
#define PMPI_Comm_set_name		  PMPI_Comm_set_nameA
#define PMPI_Type_get_name		  PMPI_Type_get_nameA
#define PMPI_Type_set_name		  PMPI_Type_set_nameA
#define PMPI_Win_get_name		   PMPI_Win_get_nameA
#define PMPI_Win_set_name		   PMPI_Win_set_nameA
#define PMPI_Info_delete			PMPI_Info_deleteA
#define PMPI_Info_get			   PMPI_Info_getA
#define PMPI_Info_get_nthkey		PMPI_Info_get_nthkeyA
#define PMPI_Info_get_valuelen	  PMPI_Info_get_valuelenA
#define PMPI_Info_set			   PMPI_Info_setA
#define PMPI_Pack_external		  PMPI_Pack_externalA
#define PMPI_Pack_external_size	 PMPI_Pack_external_sizeA
#define PMPI_Unpack_external		PMPI_Unpack_externalA
#define PMPI_File_open			  PMPI_File_openA
#define PMPI_File_delete			PMPI_File_deleteA
#define PMPI_File_set_view		  PMPI_File_set_viewA
#define PMPI_File_get_view		  PMPI_File_get_viewA
#define PMPI_Register_datarep	   PMPI_Register_datarepA

#endif

/* Keep C++ compilers from getting confused */
#if defined(__cplusplus)
extern "C" {
#endif

/* Results of the compare operations. */
#define MPI_IDENT	 0
#define MPI_CONGRUENT 1
#define MPI_SIMILAR   2
#define MPI_UNEQUAL   3

typedef int MPI_Datatype;
#define MPI_CHAR		   ((MPI_Datatype)0x4c000101)
#define MPI_SIGNED_CHAR	((MPI_Datatype)0x4c000118)
#define MPI_UNSIGNED_CHAR  ((MPI_Datatype)0x4c000102)
#define MPI_BYTE		   ((MPI_Datatype)0x4c00010d)
#define MPI_WCHAR		  ((MPI_Datatype)0x4c00020e)
#define MPI_SHORT		  ((MPI_Datatype)0x4c000203)
#define MPI_UNSIGNED_SHORT ((MPI_Datatype)0x4c000204)
#define MPI_INT			((MPI_Datatype)0x4c000405)
#define MPI_UNSIGNED	   ((MPI_Datatype)0x4c000406)
#define MPI_LONG		   ((MPI_Datatype)0x4c000407)
#define MPI_UNSIGNED_LONG  ((MPI_Datatype)0x4c000408)
#define MPI_FLOAT		  ((MPI_Datatype)0x4c00040a)
#define MPI_DOUBLE		 ((MPI_Datatype)0x4c00080b)
#define MPI_LONG_DOUBLE	((MPI_Datatype)0x4c00080c)
#define MPI_LONG_LONG_INT  ((MPI_Datatype)0x4c000809)
#define MPI_UNSIGNED_LONG_LONG ((MPI_Datatype)0x4c000819)
#define MPI_LONG_LONG	  MPI_LONG_LONG_INT

#define MPI_PACKED		 ((MPI_Datatype)0x4c00010f)
#define MPI_LB			 ((MPI_Datatype)0x4c000010)
#define MPI_UB			 ((MPI_Datatype)0x4c000011)

/* 
   The layouts for the types MPI_DOUBLE_INT etc are simply
   struct { 
	   double var;
	   int	loc;
   }
   This is documented in the man pages on the various datatypes.   
 */
#define MPI_FLOAT_INT		 ((MPI_Datatype)0x8c000000)
#define MPI_DOUBLE_INT		((MPI_Datatype)0x8c000001)
#define MPI_LONG_INT		  ((MPI_Datatype)0x8c000002)
#define MPI_SHORT_INT		 ((MPI_Datatype)0x8c000003)
#define MPI_2INT			  ((MPI_Datatype)0x4c000816)
#define MPI_LONG_DOUBLE_INT   ((MPI_Datatype)0x8c000004)

/* Fortran types */
#define MPI_COMPLEX		   ((MPI_Datatype)0x4c00081e)
#define MPI_DOUBLE_COMPLEX	((MPI_Datatype)0x4c001022)
#define MPI_LOGICAL		   ((MPI_Datatype)0x4c00041d)
#define MPI_REAL			  ((MPI_Datatype)0x4c00041c)
#define MPI_DOUBLE_PRECISION  ((MPI_Datatype)0x4c00081f)
#define MPI_INTEGER		   ((MPI_Datatype)0x4c00041b)
#define MPI_2INTEGER		  ((MPI_Datatype)0x4c000820)
#define MPI_2COMPLEX		  ((MPI_Datatype)0x4c001024)
#define MPI_2DOUBLE_COMPLEX   ((MPI_Datatype)0x4c002025)
#define MPI_2REAL			 ((MPI_Datatype)0x4c000821)
#define MPI_2DOUBLE_PRECISION ((MPI_Datatype)0x4c001023)
#define MPI_CHARACTER		 ((MPI_Datatype)0x4c00011a)

/* Size-specific types (see MPI-2, 10.2.5) */
#define MPI_REAL4			 ((MPI_Datatype)0x4c000427)
#define MPI_REAL8			 ((MPI_Datatype)0x4c000829)
#define MPI_REAL16			((MPI_Datatype)0x4c00102b)
#define MPI_COMPLEX8		  ((MPI_Datatype)0x4c000828)
#define MPI_COMPLEX16		 ((MPI_Datatype)0x4c00102a)
#define MPI_COMPLEX32		 ((MPI_Datatype)0x4c00202c)
#define MPI_INTEGER1		  ((MPI_Datatype)0x4c00012d)
#define MPI_INTEGER2		  ((MPI_Datatype)0x4c00022f)
#define MPI_INTEGER4		  ((MPI_Datatype)0x4c000430)
#define MPI_INTEGER8		  ((MPI_Datatype)0x4c000831)
#define MPI_INTEGER16		 ((MPI_Datatype)0x4c001032)

/* typeclasses */
#define MPI_TYPECLASS_REAL 1
#define MPI_TYPECLASS_INTEGER 2
#define MPI_TYPECLASS_COMPLEX 3

/* Communicators */
typedef int MPI_Comm;
#define MPI_COMM_WORLD ((MPI_Comm)0x44000000)
#define MPI_COMM_SELF  ((MPI_Comm)0x44000001)

/* Groups */
typedef int MPI_Group;
#define MPI_GROUP_EMPTY ((MPI_Group)0x48000000)

/* RMA and Windows */
typedef int MPI_Win;
#define MPI_WIN_NULL ((MPI_Win)0x20000000)

/* File and I/O */
typedef int MPI_File;
#define MPI_FILE_NULL ((MPI_File)0x50000000)

/* Collective operations */
typedef int MPI_Op;

#define MPI_MAX	 (MPI_Op)(0x58000001)
#define MPI_MIN	 (MPI_Op)(0x58000002)
#define MPI_SUM	 (MPI_Op)(0x58000003)
#define MPI_PROD	(MPI_Op)(0x58000004)
#define MPI_LAND	(MPI_Op)(0x58000005)
#define MPI_BAND	(MPI_Op)(0x58000006)
#define MPI_LOR	 (MPI_Op)(0x58000007)
#define MPI_BOR	 (MPI_Op)(0x58000008)
#define MPI_LXOR	(MPI_Op)(0x58000009)
#define MPI_BXOR	(MPI_Op)(0x5800000a)
#define MPI_MINLOC  (MPI_Op)(0x5800000b)
#define MPI_MAXLOC  (MPI_Op)(0x5800000c)
#define MPI_REPLACE (MPI_Op)(0x5800000d)

/* Permanent key values */
/* C Versions (return pointer to value),
   Fortran Versions (return integer value).
   Handled directly by the attribute value routine
   
   DO NOT CHANGE THESE.  The values encode:
   builtin kind (0x1 in bit 30-31)
   Keyval object (0x9 in bits 26-29)
   for communicator (0x1 in bits 22-25)
   
   Fortran versions of the attributes are formed by adding one to
   the C version.
 */
#define MPI_TAG_UB		   0x64400001
#define MPI_HOST			 0x64400003
#define MPI_IO			   0x64400005
#define MPI_WTIME_IS_GLOBAL  0x64400007
#define MPI_UNIVERSE_SIZE	0x64400009
#define MPI_LASTUSEDCODE	 0x6440000b
#define MPI_APPNUM		   0x6440000d

/* In addition, there are 3 predefined window attributes that are
   defined for every window */
#define MPI_WIN_BASE		 0x66000001
#define MPI_WIN_SIZE		 0x66000003
#define MPI_WIN_DISP_UNIT	0x66000005

/* Define some null objects */
#define MPI_COMM_NULL	  ((MPI_Comm)0x04000000)
#define MPI_OP_NULL		((MPI_Op)0x18000000)
#define MPI_GROUP_NULL	 ((MPI_Group)0x08000000)
#define MPI_DATATYPE_NULL  ((MPI_Datatype)0x0c000000)
#define MPI_REQUEST_NULL   ((MPI_Request)0x2c000000)
#define MPI_ERRHANDLER_NULL ((MPI_Errhandler)0x14000000)

/* These are only guesses; make sure you change them in mpif.h as well */
#define MPI_MAX_PROCESSOR_NAME 128
#define MPI_MAX_ERROR_STRING  4096
#define MPI_MAX_NAME_STRING	 63
#define MPI_MAX_PORT_NAME	  256
#define MPI_MAX_OBJECT_NAME	128

/* Pre-defined constants */
#define MPI_UNDEFINED	  (-32766)
#define MPI_UNDEFINED_RANK MPI_UNDEFINED
#define MPI_KEYVAL_INVALID 0x24000000

/* Upper bound on the overhead in bsend for each message buffer */
#define MPI_BSEND_OVERHEAD 59

/* Topology types */
typedef enum MPIR_Topo_type { MPI_GRAPH=1, MPI_CART=2 } MPIR_Topo_type;

#define MPI_BOTTOM	  (void *)0

#define MPI_PROC_NULL   (-1)
#define MPI_ANY_SOURCE 		(-2)
#define MPI_ROOT		(-3)
#define MPI_ANY_TAG	 (-1)

#define MPI_LOCK_EXCLUSIVE  234
#define MPI_LOCK_SHARED	 235

/* C functions */
typedef void (MPI_Handler_function) ( MPI_Comm *, int *, ... );
typedef int (MPI_Comm_copy_attr_function)(MPI_Comm, int, void *, void *, 
										  void *, int *);
typedef int (MPI_Comm_delete_attr_function)(MPI_Comm, int, void *, void *);
typedef int (MPI_Type_copy_attr_function)(MPI_Datatype, int, void *, void *, 
										  void *, int *);
typedef int (MPI_Type_delete_attr_function)(MPI_Datatype, int, void *, void *);
typedef int (MPI_Win_copy_attr_function)(MPI_Win, int, void *, void *, void *,
										 int *);
typedef int (MPI_Win_delete_attr_function)(MPI_Win, int, void *, void *);
typedef void (MPI_Comm_errhandler_fn)(MPI_Comm *, int *, ...);
typedef void (MPI_File_errhandler_fn)(MPI_File *, int *, ...);
typedef void (MPI_Win_errhandler_fn)(MPI_Win *, int *, ...);

/* Built in (0x1 in 30-31), errhandler (0x5 in bits 26-29, allkind (0
   in 22-25), index in the low bits */
#define MPI_ERRORS_ARE_FATAL ((MPI_Errhandler)0x54000000)
#define MPI_ERRORS_RETURN	((MPI_Errhandler)0x54000001)
	  /* #define MPIR_ERRORS_WARN	 ((MPI_Errhandler)0x54000002) */
typedef int MPI_Errhandler;

/* Make the C names for the dup function mixed case.
   This is required for systems that use all uppercase names for Fortran 
   externals.  */
/* MPI 1 names */
#define MPI_NULL_COPY_FN   ((MPI_Copy_function *)0)
#define MPI_NULL_DELETE_FN ((MPI_Delete_function *)0)
#define MPI_DUP_FN		 MPIR_Dup_fn
/* MPI 2 names */
#define MPI_COMM_NULL_COPY_FN ((MPI_Comm_copy_attr_function*)0)
#define MPI_COMM_NULL_DELETE_FN ((MPI_Comm_delete_attr_function*)0)
#define MPI_COMM_DUP_FN  ((MPI_Comm_copy_attr_function *)MPI_DUP_FN)
#define MPI_WIN_NULL_COPY_FN ((MPI_Win_copy_attr_function*)0)
#define MPI_WIN_NULL_DELETE_FN ((MPI_Win_delete_attr_function*)0)
#define MPI_WIN_DUP_FN   ((MPI_Win_copy_attr_function*)MPI_DUP_FN)
#define MPI_TYPE_NULL_COPY_FN ((MPI_Type_copy_attr_function*)0)
#define MPI_TYPE_NULL_DELETE_FN ((MPI_Type_delete_attr_function*)0)
#define MPI_TYPE_DUP_FN ((MPI_Type_copy_attr_function*)MPI_DUP_FN)

/* MPI request opjects */
typedef int MPI_Request;

/* User combination function */
typedef void (MPI_User_function) ( void *, void *, int *, MPI_Datatype * ); 

/* MPI Attribute copy and delete functions */
typedef int (MPI_Copy_function) ( MPI_Comm, int, void *, void *, void *, int * );
typedef int (MPI_Delete_function) ( MPI_Comm, int, void *, void * );

#define MPI_VERSION	2
#define MPI_SUBVERSION 0
#define MPICH_HAS_C2F  1

/* for the datatype decoders */
enum MPI_COMBINER_ENUM {
	MPI_COMBINER_NAMED			= 1,
	MPI_COMBINER_DUP			  = 2,
	MPI_COMBINER_CONTIGUOUS	   = 3, 
	MPI_COMBINER_VECTOR		   = 4,
	MPI_COMBINER_HVECTOR_INTEGER  = 5,
	MPI_COMBINER_HVECTOR		  = 6,
	MPI_COMBINER_INDEXED		  = 7,
	MPI_COMBINER_HINDEXED_INTEGER = 8, 
	MPI_COMBINER_HINDEXED		 = 9, 
	MPI_COMBINER_INDEXED_BLOCK	= 10, 
	MPI_COMBINER_STRUCT_INTEGER   = 11,
	MPI_COMBINER_STRUCT		   = 12, 
	MPI_COMBINER_SUBARRAY		 = 13,
	MPI_COMBINER_DARRAY		   = 14,
	MPI_COMBINER_F90_REAL		 = 15, 
	MPI_COMBINER_F90_COMPLEX	  = 16, 
	MPI_COMBINER_F90_INTEGER	  = 17, 
	MPI_COMBINER_RESIZED		  = 18 
};

/* for info */
typedef int MPI_Info;
#define MPI_INFO_NULL		 ((MPI_Info)0x1c000000)
#define MPI_MAX_INFO_KEY	   255
#define MPI_MAX_INFO_VAL	  1024

/* for subarray and darray constructors */
#define MPI_ORDER_C			  56
#define MPI_ORDER_FORTRAN		57
#define MPI_DISTRIBUTE_BLOCK	121
#define MPI_DISTRIBUTE_CYCLIC   122
#define MPI_DISTRIBUTE_NONE	 123
#define MPI_DISTRIBUTE_DFLT_DARG -49767

#define MPI_IN_PLACE  (void *) ~0

/* asserts for one-sided communication */
#define MPI_MODE_NOCHECK	  1024
#define MPI_MODE_NOSTORE	  2048
#define MPI_MODE_NOPUT		4096
#define MPI_MODE_NOPRECEDE	8192
#define MPI_MODE_NOSUCCEED   16384 

/* Definitions that are determined by configure. */
typedef int MPI_Aint;
typedef int MPI_Fint;
/* MPI_OFFSET_TYPEDEF is set in configure and is 
	  typedef $MPI_OFFSET MPI_Offset;
   where $MPI_OFFSET is the correct C type */
#if defined(USE_GCC) || defined(__GNUC__)
typedef long long MPI_Offset;
#else
typedef __int64 MPI_Offset;
#endif

/* The order of these elements must match that in mpif.h */
typedef struct MPI_Status {
	int count;
	int cancelled;
	int MPI_SOURCE;
	int MPI_TAG;
	int MPI_ERROR;
	
} MPI_Status;

/* Handle conversion types/functions */

/* Programs that need to convert types used in MPICH should use these */
#define MPI_Comm_c2f(comm) (MPI_Fint)(comm)
#define MPI_Comm_f2c(comm) (MPI_Comm)(comm)
#define MPI_Type_c2f(datatype) (MPI_Fint)(datatype)
#define MPI_Type_f2c(datatype) (MPI_Datatype)(datatype)
#define MPI_Group_c2f(group) (MPI_Fint)(group)
#define MPI_Group_f2c(group) (MPI_Group)(group)
#define MPI_Info_c2f(info) (MPI_Fint)(info)
#define MPI_Info_f2c(info) (MPI_Info)(info)
#define MPI_Request_f2c(request) (MPI_Request)(request)
#define MPI_Request_c2f(request) (MPI_Fint)(request)
#define MPI_Op_c2f(op) (MPI_Fint)(op)
#define MPI_Op_f2c(op) (MPI_Op)(op)
#define MPI_Errhandler_c2f(errhandler) (MPI_Fint)(errhandler)
#define MPI_Errhandler_f2c(errhandler) (MPI_Errhandler)(errhandler)
#define MPI_Win_c2f(win)   (MPI_Fint)(win)
#define MPI_Win_f2c(win)   (MPI_Win)(win)
#define MPI_File_c2f(file) (MPI_Fint)(file)
#define MPI_File_f2c(file) (MPI_File)(file)

/* PMPI versions of the handle transfer functions.  See section 4.17 */
#define PMPI_Comm_c2f(comm) (MPI_Fint)(comm)
#define PMPI_Comm_f2c(comm) (MPI_Comm)(comm)
#define PMPI_Type_c2f(datatype) (MPI_Fint)(datatype)
#define PMPI_Type_f2c(datatype) (MPI_Datatype)(datatype)
#define PMPI_Group_c2f(group) (MPI_Fint)(group)
#define PMPI_Group_f2c(group) (MPI_Group)(group)
#define PMPI_Info_c2f(info) (MPI_Fint)(info)
#define PMPI_Info_f2c(info) (MPI_Info)(info)
#define PMPI_Request_f2c(request) (MPI_Request)(request)
#define PMPI_Request_c2f(request) (MPI_Fint)(request)
#define PMPI_Op_c2f(op) (MPI_Fint)(op)
#define PMPI_Op_f2c(op) (MPI_Op)(op)
#define PMPI_Errhandler_c2f(errhandler) (MPI_Fint)(errhandler)
#define PMPI_Errhandler_f2c(errhandler) (MPI_Errhandler)(errhandler)
#define PMPI_Win_c2f(win)   (MPI_Fint)(win)
#define PMPI_Win_f2c(win)   (MPI_Win)(win)
#define PMPI_File_c2f(file) (MPI_Fint)(file)
#define PMPI_File_f2c(file) (MPI_File)(file)

#define MPI_STATUS_IGNORE (MPI_Status *)0x1
#define MPI_STATUSES_IGNORE (MPI_Status *)0x1
#define MPI_ERRCODES_IGNORE (int *)0

/* See 4.12.5 for MPI_F_STATUS(ES)_IGNORE */

extern MPI_DLL_SPEC MPI_Fint * MPI_F_STATUS_IGNORE;
extern MPI_DLL_SPEC MPI_Fint * MPI_F_STATUSES_IGNORE;
/* The annotation MPIU_DLL_SPEC to the extern statements is used 
   as a hook for systems that require C extensions to correctly construct
   DLLs, and is defined as an empty string otherwise
 */

/* The MPI standard requires that the ARGV_NULL values be the same as
   NULL (see 5.3.2) */
#if defined(UNICODE) || defined(_UNICODE)
#define MPI_ARGV_NULL (wchar_t **)0
#define MPI_ARGVS_NULL (wchar_t ***)0
#else
#define MPI_ARGV_NULL (char **)0
#define MPI_ARGVS_NULL (char ***)0
#endif

/* For supported thread levels */
#define MPI_THREAD_SINGLE 0
#define MPI_THREAD_FUNNELED 1
#define MPI_THREAD_SERIALIZED 2
#define MPI_THREAD_MULTIPLE 3

/* Typedefs for generalized requests */
typedef int (MPI_Grequest_cancel_function)(void *, int); 
typedef int (MPI_Grequest_free_function)(void *); 
typedef int (MPI_Grequest_query_function)(void *, MPI_Status *); 

/* MPI's error classes */
#define MPI_SUCCESS		  0	  /* Successful return code */
/* Communication argument parameters */
#define MPI_ERR_BUFFER	   1	  /* Invalid buffer pointer */
#define MPI_ERR_COUNT		2	  /* Invalid count argument */
#define MPI_ERR_TYPE		 3	  /* Invalid datatype argument */
#define MPI_ERR_TAG		  4	  /* Invalid tag argument */
#define MPI_ERR_COMM		 5	  /* Invalid communicator */
#define MPI_ERR_RANK		 6	  /* Invalid rank */
#define MPI_ERR_ROOT		 7	  /* Invalid root */
#define MPI_ERR_TRUNCATE	14	  /* Message truncated on receive */

/* MPI Objects (other than COMM) */
#define MPI_ERR_GROUP		8	  /* Invalid group */
#define MPI_ERR_OP		   9	  /* Invalid operation */
#define MPI_ERR_REQUEST	 19	  /* Invalid mpi_request handle */

/* Special topology argument parameters */
#define MPI_ERR_TOPOLOGY	10	  /* Invalid topology */
#define MPI_ERR_DIMS		11	  /* Invalid dimension argument */

/* All other arguments.  This is a class with many kinds */
#define MPI_ERR_ARG		 12	  /* Invalid argument */

/* Other errors that are not simply an invalid argument */
#define MPI_ERR_OTHER	   15	  /* Other error; use Error_string */

#define MPI_ERR_UNKNOWN	 13	  /* Unknown error */
#define MPI_ERR_INTERN	  16	  /* Internal error code	*/

/* Multiple completion has two special error classes */
#define MPI_ERR_IN_STATUS   17	  /* Look in status for error value */
#define MPI_ERR_PENDING	 18	  /* Pending request */

/* New MPI-2 Error classes */
#define MPI_ERR_FILE		27	  /* */
#define MPI_ERR_ACCESS	  20	  /* */
#define MPI_ERR_AMODE	   21	  /* */
#define MPI_ERR_BAD_FILE	22	  /* */
#define MPI_ERR_FILE_EXISTS 25	  /* */
#define MPI_ERR_FILE_IN_USE 26	  /* */
#define MPI_ERR_NO_SPACE	36	  /* */
#define MPI_ERR_NO_SUCH_FILE 37	 /* */
#define MPI_ERR_IO		  32	  /* */
#define MPI_ERR_READ_ONLY   40	  /* */
#define MPI_ERR_CONVERSION  23	  /* */
#define MPI_ERR_DUP_DATAREP 24	  /* */
#define MPI_ERR_UNSUPPORTED_DATAREP   43  /* */

/* MPI_ERR_INFO is NOT defined in the MPI-2 standard.  I believe that
   this is an oversight */
#define MPI_ERR_INFO		28	  /* */
#define MPI_ERR_INFO_KEY	29	  /* */
#define MPI_ERR_INFO_VALUE  30	  /* */
#define MPI_ERR_INFO_NOKEY  31	  /* */

#define MPI_ERR_NAME		33	  /* */
#define MPI_ERR_NO_MEM	  34	  /* Alloc_mem could not allocate memory */
#define MPI_ERR_NOT_SAME	35	  /* */
#define MPI_ERR_PORT		38	  /* */
#define MPI_ERR_QUOTA	   39	  /* */
#define MPI_ERR_SERVICE	 41	  /* */
#define MPI_ERR_SPAWN	   42	  /* */
#define MPI_ERR_UNSUPPORTED_OPERATION 44 /* */
#define MPI_ERR_WIN		 45	  /* */

#define MPI_ERR_BASE		46	  /* */
#define MPI_ERR_LOCKTYPE	47	  /* */
#define MPI_ERR_KEYVAL	  48	  /* Erroneous attribute key */
#define MPI_ERR_RMA_CONFLICT 49	 /* */
#define MPI_ERR_RMA_SYNC	50	  /* */ 
#define MPI_ERR_SIZE		51	  /* */
#define MPI_ERR_DISP		52	  /* */
#define MPI_ERR_ASSERT	  53	  /* */

#define MPI_ERR_LASTCODE	0x3fffffff  /* Last valid error code for a 
										   predefined error class */
#define MPICH_ERR_LAST_CLASS 53	 /* It is also helpful to know the
									   last valid class */
/* End of MPI's error classes */

/* Function type defs */
typedef int (MPI_Datarep_conversion_function)(void *, MPI_Datatype, int, 
			 void *, MPI_Offset, void *);
typedef int (MPI_Datarep_extent_function)(MPI_Datatype datatype, MPI_Aint *,
										  void *);
#define MPI_CONVERSION_FN_NULL ((MPI_Datarep_conversion_function *)0)

/* 
   For systems that may need to add additional definitions to support
   different declaration styles and options (e.g., different calling 
   conventions or DLL import/export controls).  
*/
/* --Insert Additional Definitions Here-- */

/* Include any device specific definitions */


/*
 * Normally, we provide prototypes for all MPI routines.  In a few wierd
 * cases, we need to suppress the prototypes.
 */
#ifndef MPICH_SUPPRESS_PROTOTYPES
/* We require that the C compiler support prototypes */
/* Begin Prototypes */
MPI_DLL_SPEC int MPI_CALL MPI_Send(void*, int, MPI_Datatype, int, int, MPI_Comm);
MPI_DLL_SPEC int MPI_CALL MPI_Recv(void*, int, MPI_Datatype, int, int, MPI_Comm, MPI_Status *);
MPI_DLL_SPEC int MPI_CALL MPI_Get_count(MPI_Status *, MPI_Datatype, int *);
MPI_DLL_SPEC int MPI_CALL MPI_Bsend(void*, int, MPI_Datatype, int, int, MPI_Comm);
MPI_DLL_SPEC int MPI_CALL MPI_Ssend(void*, int, MPI_Datatype, int, int, MPI_Comm);
MPI_DLL_SPEC int MPI_CALL MPI_Rsend(void*, int, MPI_Datatype, int, int, MPI_Comm);
MPI_DLL_SPEC int MPI_CALL MPI_Buffer_attach( void*, int);
MPI_DLL_SPEC int MPI_CALL MPI_Buffer_detach( void*, int *);
MPI_DLL_SPEC int MPI_CALL MPI_Isend(void*, int, MPI_Datatype, int, int, MPI_Comm, MPI_Request *);
MPI_DLL_SPEC int MPI_CALL MPI_Ibsend(void*, int, MPI_Datatype, int, int, MPI_Comm, MPI_Request *);
MPI_DLL_SPEC int MPI_CALL MPI_Issend(void*, int, MPI_Datatype, int, int, MPI_Comm, MPI_Request *);
MPI_DLL_SPEC int MPI_CALL MPI_Irsend(void*, int, MPI_Datatype, int, int, MPI_Comm, MPI_Request *);
MPI_DLL_SPEC int MPI_CALL MPI_Irecv(void*, int, MPI_Datatype, int, int, MPI_Comm, MPI_Request *);
MPI_DLL_SPEC int MPI_CALL MPI_Wait(MPI_Request *, MPI_Status *);
MPI_DLL_SPEC int MPI_CALL MPI_Test(MPI_Request *, int *, MPI_Status *);
MPI_DLL_SPEC int MPI_CALL MPI_Request_free(MPI_Request *);
MPI_DLL_SPEC int MPI_CALL MPI_Waitany(int, MPI_Request *, int *, MPI_Status *);
MPI_DLL_SPEC int MPI_CALL MPI_Testany(int, MPI_Request *, int *, int *, MPI_Status *);
MPI_DLL_SPEC int MPI_CALL MPI_Waitall(int, MPI_Request *, MPI_Status *);
MPI_DLL_SPEC int MPI_CALL MPI_Testall(int, MPI_Request *, int *, MPI_Status *);
MPI_DLL_SPEC int MPI_CALL MPI_Waitsome(int, MPI_Request *, int *, int *, MPI_Status *);
MPI_DLL_SPEC int MPI_CALL MPI_Testsome(int, MPI_Request *, int *, int *, MPI_Status *);
MPI_DLL_SPEC int MPI_CALL MPI_Iprobe(int, int, MPI_Comm, int *, MPI_Status *);
MPI_DLL_SPEC int MPI_CALL MPI_Probe(int, int, MPI_Comm, MPI_Status *);
MPI_DLL_SPEC int MPI_CALL MPI_Cancel(MPI_Request *);
MPI_DLL_SPEC int MPI_CALL MPI_Test_cancelled(MPI_Status *, int *);
MPI_DLL_SPEC int MPI_CALL MPI_Send_init(void*, int, MPI_Datatype, int, int, MPI_Comm, MPI_Request *);
MPI_DLL_SPEC int MPI_CALL MPI_Bsend_init(void*, int, MPI_Datatype, int,int, MPI_Comm, MPI_Request *);
MPI_DLL_SPEC int MPI_CALL MPI_Ssend_init(void*, int, MPI_Datatype, int,int, MPI_Comm, MPI_Request *);
MPI_DLL_SPEC int MPI_CALL MPI_Rsend_init(void*, int, MPI_Datatype, int,int, MPI_Comm, MPI_Request *);
MPI_DLL_SPEC int MPI_CALL MPI_Recv_init(void*, int, MPI_Datatype, int,int, MPI_Comm, MPI_Request *);
MPI_DLL_SPEC int MPI_CALL MPI_Start(MPI_Request *);
MPI_DLL_SPEC int MPI_CALL MPI_Startall(int, MPI_Request *);
MPI_DLL_SPEC int MPI_CALL MPI_Sendrecv(void *, int, MPI_Datatype,int, int, void *, int, MPI_Datatype, int, int, MPI_Comm, MPI_Status *);
MPI_DLL_SPEC int MPI_CALL MPI_Sendrecv_replace(void*, int, MPI_Datatype, int, int, int, int, MPI_Comm, MPI_Status *);
MPI_DLL_SPEC int MPI_CALL MPI_Type_contiguous(int, MPI_Datatype, MPI_Datatype *);
MPI_DLL_SPEC int MPI_CALL MPI_Type_vector(int, int, int, MPI_Datatype, MPI_Datatype *);
MPI_DLL_SPEC int MPI_CALL MPI_Type_hvector(int, int, MPI_Aint, MPI_Datatype, MPI_Datatype *);
MPI_DLL_SPEC int MPI_CALL MPI_Type_indexed(int, int *, int *, MPI_Datatype, MPI_Datatype *);
MPI_DLL_SPEC int MPI_CALL MPI_Type_hindexed(int, int *, MPI_Aint *, MPI_Datatype, MPI_Datatype *);
MPI_DLL_SPEC int MPI_CALL MPI_Type_struct(int, int *, MPI_Aint *, MPI_Datatype *, MPI_Datatype *);
MPI_DLL_SPEC int MPI_CALL MPI_Address(void*, MPI_Aint *);
/* We could add __attribute__((deprecated)) to routines like MPI_Type_extent */
MPI_DLL_SPEC int MPI_CALL MPI_Type_extent(MPI_Datatype, MPI_Aint *);
/* See the 1.1 version of the Standard.  The standard made an 
   unfortunate choice here, however, it is the standard.  The size returned 
   by MPI_Type_size is specified as an int, not an MPI_Aint */
MPI_DLL_SPEC int MPI_CALL MPI_Type_size(MPI_Datatype, int *);
/* MPI_Type_count was withdrawn in MPI 1.1 */
MPI_DLL_SPEC int MPI_CALL MPI_Type_lb(MPI_Datatype, MPI_Aint *);
MPI_DLL_SPEC int MPI_CALL MPI_Type_ub(MPI_Datatype, MPI_Aint *);
MPI_DLL_SPEC int MPI_CALL MPI_Type_commit(MPI_Datatype *);
MPI_DLL_SPEC int MPI_CALL MPI_Type_free(MPI_Datatype *);
MPI_DLL_SPEC int MPI_CALL MPI_Get_elements(MPI_Status *, MPI_Datatype, int *);
MPI_DLL_SPEC int MPI_CALL MPI_Pack(void*, int, MPI_Datatype, void *, int, int *,  MPI_Comm);
MPI_DLL_SPEC int MPI_CALL MPI_Unpack(void*, int, int *, void *, int, MPI_Datatype, MPI_Comm);
MPI_DLL_SPEC int MPI_CALL MPI_Pack_size(int, MPI_Datatype, MPI_Comm, int *);
MPI_DLL_SPEC int MPI_CALL MPI_Barrier(MPI_Comm );
MPI_DLL_SPEC int MPI_CALL MPI_Bcast(void*, int, MPI_Datatype, int, MPI_Comm );
MPI_DLL_SPEC int MPI_CALL MPI_Gather(void* , int, MPI_Datatype, void*, int, MPI_Datatype, int, MPI_Comm); 
MPI_DLL_SPEC int MPI_CALL MPI_Gatherv(void* , int, MPI_Datatype, void*, int *, int *, MPI_Datatype, int, MPI_Comm); 
MPI_DLL_SPEC int MPI_CALL MPI_Scatter(void* , int, MPI_Datatype, void*, int, MPI_Datatype, int, MPI_Comm);
MPI_DLL_SPEC int MPI_CALL MPI_Scatterv(void* , int *, int *,  MPI_Datatype, void*, int, MPI_Datatype, int, MPI_Comm);
MPI_DLL_SPEC int MPI_CALL MPI_Allgather(void* , int, MPI_Datatype, void*, int, MPI_Datatype, MPI_Comm);
MPI_DLL_SPEC int MPI_CALL MPI_Allgatherv(void* , int, MPI_Datatype, void*, int *, int *, MPI_Datatype, MPI_Comm);
MPI_DLL_SPEC int MPI_CALL MPI_Alltoall(void* , int, MPI_Datatype, void*, int, MPI_Datatype, MPI_Comm);
MPI_DLL_SPEC int MPI_CALL MPI_Alltoallv(void* , int *, int *, MPI_Datatype, void*, int *, int *, MPI_Datatype, MPI_Comm);
MPI_DLL_SPEC int MPI_CALL MPI_Reduce(void* , void*, int, MPI_Datatype, MPI_Op, int, MPI_Comm);
MPI_DLL_SPEC int MPI_CALL MPI_Op_create(MPI_User_function *, int, MPI_Op *);
MPI_DLL_SPEC int MPI_CALL MPI_Op_free( MPI_Op *);
MPI_DLL_SPEC int MPI_CALL MPI_Allreduce(void* , void*, int, MPI_Datatype, MPI_Op, MPI_Comm);
MPI_DLL_SPEC int MPI_CALL MPI_Reduce_scatter(void* , void*, int *, MPI_Datatype, MPI_Op, MPI_Comm);
MPI_DLL_SPEC int MPI_CALL MPI_Scan(void* , void*, int, MPI_Datatype, MPI_Op, MPI_Comm );
MPI_DLL_SPEC int MPI_CALL MPI_Group_size(MPI_Group, int *);
MPI_DLL_SPEC int MPI_CALL MPI_Group_rank(MPI_Group, int *);
MPI_DLL_SPEC int MPI_CALL MPI_Group_translate_ranks (MPI_Group, int, int *, MPI_Group, int *);
MPI_DLL_SPEC int MPI_CALL MPI_Group_compare(MPI_Group, MPI_Group, int *);
MPI_DLL_SPEC int MPI_CALL MPI_Comm_group(MPI_Comm, MPI_Group *);
MPI_DLL_SPEC int MPI_CALL MPI_Group_union(MPI_Group, MPI_Group, MPI_Group *);
MPI_DLL_SPEC int MPI_CALL MPI_Group_intersection(MPI_Group, MPI_Group, MPI_Group *);
MPI_DLL_SPEC int MPI_CALL MPI_Group_difference(MPI_Group, MPI_Group, MPI_Group *);
MPI_DLL_SPEC int MPI_CALL MPI_Group_incl(MPI_Group, int, int *, MPI_Group *);
MPI_DLL_SPEC int MPI_CALL MPI_Group_excl(MPI_Group, int, int *, MPI_Group *);
MPI_DLL_SPEC int MPI_CALL MPI_Group_range_incl(MPI_Group, int, int [][3], MPI_Group *);
MPI_DLL_SPEC int MPI_CALL MPI_Group_range_excl(MPI_Group, int, int [][3], MPI_Group *);
MPI_DLL_SPEC int MPI_CALL MPI_Group_free(MPI_Group *);
MPI_DLL_SPEC int MPI_CALL MPI_Comm_size(MPI_Comm, int *);
MPI_DLL_SPEC int MPI_CALL MPI_Comm_rank(MPI_Comm, int *);
MPI_DLL_SPEC int MPI_CALL MPI_Comm_compare(MPI_Comm, MPI_Comm, int *);
MPI_DLL_SPEC int MPI_CALL MPI_Comm_dup(MPI_Comm, MPI_Comm *);
MPI_DLL_SPEC int MPI_CALL MPI_Comm_create(MPI_Comm, MPI_Group, MPI_Comm *);
MPI_DLL_SPEC int MPI_CALL MPI_Comm_split(MPI_Comm, int, int, MPI_Comm *);
MPI_DLL_SPEC int MPI_CALL MPI_Comm_free(MPI_Comm *);
MPI_DLL_SPEC int MPI_CALL MPI_Comm_test_inter(MPI_Comm, int *);
MPI_DLL_SPEC int MPI_CALL MPI_Comm_remote_size(MPI_Comm, int *);
MPI_DLL_SPEC int MPI_CALL MPI_Comm_remote_group(MPI_Comm, MPI_Group *);
MPI_DLL_SPEC int MPI_CALL MPI_Intercomm_create(MPI_Comm, int, MPI_Comm, int, int, MPI_Comm * );
MPI_DLL_SPEC int MPI_CALL MPI_Intercomm_merge(MPI_Comm, int, MPI_Comm *);
MPI_DLL_SPEC int MPI_CALL MPI_Keyval_create(MPI_Copy_function *, MPI_Delete_function *, int *, void*);
MPI_DLL_SPEC int MPI_CALL MPI_Keyval_free(int *);
MPI_DLL_SPEC int MPI_CALL MPI_Attr_put(MPI_Comm, int, void*);
MPI_DLL_SPEC int MPI_CALL MPI_Attr_get(MPI_Comm, int, void *, int *);
MPI_DLL_SPEC int MPI_CALL MPI_Attr_delete(MPI_Comm, int);
MPI_DLL_SPEC int MPI_CALL MPI_Topo_test(MPI_Comm, int *);
MPI_DLL_SPEC int MPI_CALL MPI_Cart_create(MPI_Comm, int, int *, int *, int, MPI_Comm *);
MPI_DLL_SPEC int MPI_CALL MPI_Dims_create(int, int, int *);
MPI_DLL_SPEC int MPI_CALL MPI_Graph_create(MPI_Comm, int, int *, int *, int, MPI_Comm *);
MPI_DLL_SPEC int MPI_CALL MPI_Graphdims_get(MPI_Comm, int *, int *);
MPI_DLL_SPEC int MPI_CALL MPI_Graph_get(MPI_Comm, int, int, int *, int *);
MPI_DLL_SPEC int MPI_CALL MPI_Cartdim_get(MPI_Comm, int *);
MPI_DLL_SPEC int MPI_CALL MPI_Cart_get(MPI_Comm, int, int *, int *, int *);
MPI_DLL_SPEC int MPI_CALL MPI_Cart_rank(MPI_Comm, int *, int *);
MPI_DLL_SPEC int MPI_CALL MPI_Cart_coords(MPI_Comm, int, int, int *);
MPI_DLL_SPEC int MPI_CALL MPI_Graph_neighbors_count(MPI_Comm, int, int *);
MPI_DLL_SPEC int MPI_CALL MPI_Graph_neighbors(MPI_Comm, int, int, int *);
MPI_DLL_SPEC int MPI_CALL MPI_Cart_shift(MPI_Comm, int, int, int *, int *);
MPI_DLL_SPEC int MPI_CALL MPI_Cart_sub(MPI_Comm, int *, MPI_Comm *);
MPI_DLL_SPEC int MPI_CALL MPI_Cart_map(MPI_Comm, int, int *, int *, int *);
MPI_DLL_SPEC int MPI_CALL MPI_Graph_map(MPI_Comm, int, int *, int *, int *);
MPI_DLL_SPEC int MPI_CALL MPI_Get_processor_nameA(char *, int *);
MPI_DLL_SPEC int MPI_CALL MPI_Get_processor_nameW(wchar_t *, int *);
MPI_DLL_SPEC int MPI_CALL MPI_Get_version(int *, int *);
MPI_DLL_SPEC int MPI_CALL MPI_Errhandler_create(MPI_Handler_function *, MPI_Errhandler *);
MPI_DLL_SPEC int MPI_CALL MPI_Errhandler_set(MPI_Comm, MPI_Errhandler);
MPI_DLL_SPEC int MPI_CALL MPI_Errhandler_get(MPI_Comm, MPI_Errhandler *);
MPI_DLL_SPEC int MPI_CALL MPI_Errhandler_free(MPI_Errhandler *);
MPI_DLL_SPEC int MPI_CALL MPI_Error_stringA(int, char *, int *);
MPI_DLL_SPEC int MPI_CALL MPI_Error_stringW(int, wchar_t *, int *);
MPI_DLL_SPEC int MPI_CALL MPI_Error_class(int, int *);
MPI_DLL_SPEC double MPI_CALL MPI_Wtime(void);
MPI_DLL_SPEC double MPI_CALL MPI_Wtick(void);
#ifndef MPI_Wtime
MPI_DLL_SPEC double MPI_CALL PMPI_Wtime(void);
MPI_DLL_SPEC double MPI_CALL PMPI_Wtick(void);
#endif
MPI_DLL_SPEC int MPI_CALL MPI_InitA(int *, char ***);
MPI_DLL_SPEC int MPI_CALL MPI_InitW(int *, wchar_t ***);
MPI_DLL_SPEC int MPI_CALL MPI_Finalize(void);
MPI_DLL_SPEC int MPI_CALL MPI_Initialized(int *);
MPI_DLL_SPEC int MPI_CALL MPI_Abort(MPI_Comm, int);


/* Note that we may need to define a @PCONTROL_LIST@ depending on whether 
   stdargs are supported */
MPI_DLL_SPEC int MPI_CALL MPI_Pcontrol(const int, ...);

MPI_DLL_SPEC int MPI_CALL MPI_DUP_FN ( MPI_Comm, int, void *, void *, void *, int * );


/* MPI-2 functions */

/* Process Creation and Management */
MPI_DLL_SPEC int MPI_CALL MPI_Close_portA(char *);
MPI_DLL_SPEC int MPI_CALL MPI_Close_portW(wchar_t *);
MPI_DLL_SPEC int MPI_CALL MPI_Comm_acceptA(char *, MPI_Info, int, MPI_Comm, MPI_Comm *);
MPI_DLL_SPEC int MPI_CALL MPI_Comm_acceptW(wchar_t *, MPI_Info, int, MPI_Comm, MPI_Comm *);
MPI_DLL_SPEC int MPI_CALL MPI_Comm_connectA(char *, MPI_Info, int, MPI_Comm, MPI_Comm *);
MPI_DLL_SPEC int MPI_CALL MPI_Comm_connectW(wchar_t *, MPI_Info, int, MPI_Comm, MPI_Comm *);
MPI_DLL_SPEC int MPI_CALL MPI_Comm_disconnect(MPI_Comm *);
MPI_DLL_SPEC int MPI_CALL MPI_Comm_get_parent(MPI_Comm *);
MPI_DLL_SPEC int MPI_CALL MPI_Comm_join(int, MPI_Comm *);
MPI_DLL_SPEC int MPI_CALL MPI_Comm_spawnA(char *, char *[], int, MPI_Info, int, MPI_Comm, MPI_Comm *, int []);
MPI_DLL_SPEC int MPI_CALL MPI_Comm_spawnW(wchar_t *, wchar_t *[], int, MPI_Info, int, MPI_Comm, MPI_Comm *, int []);
MPI_DLL_SPEC int MPI_CALL MPI_Comm_spawn_multipleA(int, char *[], char **[], int [], MPI_Info [], int, MPI_Comm, MPI_Comm *, int []);
MPI_DLL_SPEC int MPI_CALL MPI_Comm_spawn_multipleW(int, wchar_t *[], wchar_t **[], int [], MPI_Info [], int, MPI_Comm, MPI_Comm *, int []);
MPI_DLL_SPEC int MPI_CALL MPI_Lookup_nameA(char *, MPI_Info, char *);
MPI_DLL_SPEC int MPI_CALL MPI_Lookup_nameW(wchar_t *, MPI_Info, wchar_t *);
MPI_DLL_SPEC int MPI_CALL MPI_Open_portA(MPI_Info, char *);
MPI_DLL_SPEC int MPI_CALL MPI_Open_portW(MPI_Info, wchar_t *);
MPI_DLL_SPEC int MPI_CALL MPI_Publish_nameA(char *, MPI_Info, char *);
MPI_DLL_SPEC int MPI_CALL MPI_Publish_nameW(wchar_t *, MPI_Info, wchar_t *);
MPI_DLL_SPEC int MPI_CALL MPI_Unpublish_nameA(char *, MPI_Info, char *);
MPI_DLL_SPEC int MPI_CALL MPI_Unpublish_nameW(wchar_t *, MPI_Info, wchar_t *);

/* One-Sided Communications */
MPI_DLL_SPEC int MPI_CALL MPI_Accumulate(void *, int, MPI_Datatype, int, MPI_Aint, int, 
				   MPI_Datatype,  MPI_Op, MPI_Win);
MPI_DLL_SPEC int MPI_CALL MPI_Get(void *, int, MPI_Datatype, int, MPI_Aint, int, MPI_Datatype, 
			MPI_Win);
MPI_DLL_SPEC int MPI_CALL MPI_Put(void *, int, MPI_Datatype, int, MPI_Aint, int, MPI_Datatype, 
			MPI_Win);
MPI_DLL_SPEC int MPI_CALL MPI_Win_complete(MPI_Win);
MPI_DLL_SPEC int MPI_CALL MPI_Win_create(void *, MPI_Aint, int, MPI_Info, MPI_Comm, MPI_Win *);
MPI_DLL_SPEC int MPI_CALL MPI_Win_fence(int, MPI_Win);
MPI_DLL_SPEC int MPI_CALL MPI_Win_free(MPI_Win *);
MPI_DLL_SPEC int MPI_CALL MPI_Win_get_group(MPI_Win, MPI_Group *);
MPI_DLL_SPEC int MPI_CALL MPI_Win_lock(int, int, int, MPI_Win);
MPI_DLL_SPEC int MPI_CALL MPI_Win_post(MPI_Group, int, MPI_Win);
MPI_DLL_SPEC int MPI_CALL MPI_Win_start(MPI_Group, int, MPI_Win);
MPI_DLL_SPEC int MPI_CALL MPI_Win_test(MPI_Win, int *);
MPI_DLL_SPEC int MPI_CALL MPI_Win_unlock(int, MPI_Win);
MPI_DLL_SPEC int MPI_CALL MPI_Win_wait(MPI_Win);
 
/* Extended Collective Operations */
MPI_DLL_SPEC int MPI_CALL MPI_Alltoallw(void *, int [], int [], MPI_Datatype [], void *, int [], 
				  int [], MPI_Datatype [], MPI_Comm);
MPI_DLL_SPEC int MPI_CALL MPI_Exscan(void *, void *, int, MPI_Datatype, MPI_Op, MPI_Comm) ;
 
/* External Interfaces */
MPI_DLL_SPEC int MPI_CALL MPI_Add_error_class(int *);
MPI_DLL_SPEC int MPI_CALL MPI_Add_error_code(int, int *);
MPI_DLL_SPEC int MPI_CALL MPI_Add_error_stringA(int, char *);
MPI_DLL_SPEC int MPI_CALL MPI_Add_error_stringW(int, wchar_t *);
MPI_DLL_SPEC int MPI_CALL MPI_Comm_call_errhandler(MPI_Comm, int);
MPI_DLL_SPEC int MPI_CALL MPI_Comm_create_keyval(MPI_Comm_copy_attr_function *, 
						   MPI_Comm_delete_attr_function *, int *, void *);
MPI_DLL_SPEC int MPI_CALL MPI_Comm_delete_attr(MPI_Comm, int);
MPI_DLL_SPEC int MPI_CALL MPI_Comm_free_keyval(int *);
MPI_DLL_SPEC int MPI_CALL MPI_Comm_get_attr(MPI_Comm, int, void *, int *);
MPI_DLL_SPEC int MPI_CALL MPI_Comm_get_nameA(MPI_Comm, char *, int *);
MPI_DLL_SPEC int MPI_CALL MPI_Comm_get_nameW(MPI_Comm, wchar_t *, int *);
MPI_DLL_SPEC int MPI_CALL MPI_Comm_set_attr(MPI_Comm, int, void *);
MPI_DLL_SPEC int MPI_CALL MPI_Comm_set_nameA(MPI_Comm, char *);
MPI_DLL_SPEC int MPI_CALL MPI_Comm_set_nameW(MPI_Comm, wchar_t *);
MPI_DLL_SPEC int MPI_CALL MPI_File_call_errhandler(MPI_File, int);
MPI_DLL_SPEC int MPI_CALL MPI_Grequest_complete(MPI_Request);
MPI_DLL_SPEC int MPI_CALL MPI_Grequest_start(MPI_Grequest_query_function *, 
					   MPI_Grequest_free_function *, 
					   MPI_Grequest_cancel_function *, void *, MPI_Request *);
MPI_DLL_SPEC int MPI_CALL MPI_Init_threadA(int *, char ***, int, int *);
MPI_DLL_SPEC int MPI_CALL MPI_Init_threadW(int *, wchar_t ***, int, int *);
MPI_DLL_SPEC int MPI_CALL MPI_Is_thread_main(int *);
MPI_DLL_SPEC int MPI_CALL MPI_Query_thread(int *);
MPI_DLL_SPEC int MPI_CALL MPI_Status_set_cancelled(MPI_Status *, int);
MPI_DLL_SPEC int MPI_CALL MPI_Status_set_elements(MPI_Status *, MPI_Datatype, int);
MPI_DLL_SPEC int MPI_CALL MPI_Type_create_keyval(MPI_Type_copy_attr_function *, 
						   MPI_Type_delete_attr_function *, int *, void *);
MPI_DLL_SPEC int MPI_CALL MPI_Type_delete_attr(MPI_Datatype, int);
MPI_DLL_SPEC int MPI_CALL MPI_Type_dup(MPI_Datatype, MPI_Datatype *);
MPI_DLL_SPEC int MPI_CALL MPI_Type_free_keyval(int *);
MPI_DLL_SPEC int MPI_CALL MPI_Type_get_attr(MPI_Datatype, int, void *, int *);
MPI_DLL_SPEC int MPI_CALL MPI_Type_get_contents(MPI_Datatype, int, int, int, int [], MPI_Aint [], 
						  MPI_Datatype []);
MPI_DLL_SPEC int MPI_CALL MPI_Type_get_envelope(MPI_Datatype, int *, int *, int *, int *);
MPI_DLL_SPEC int MPI_CALL MPI_Type_get_nameA(MPI_Datatype, char *, int *);
MPI_DLL_SPEC int MPI_CALL MPI_Type_get_nameW(MPI_Datatype, wchar_t *, int *);
MPI_DLL_SPEC int MPI_CALL MPI_Type_set_attr(MPI_Datatype, int, void *);
MPI_DLL_SPEC int MPI_CALL MPI_Type_set_nameA(MPI_Datatype, char *);
MPI_DLL_SPEC int MPI_CALL MPI_Type_set_nameW(MPI_Datatype, wchar_t *);
MPI_DLL_SPEC int MPI_CALL MPI_Type_match_size( int, int, MPI_Datatype *);
MPI_DLL_SPEC int MPI_CALL MPI_Win_call_errhandler(MPI_Win, int);
MPI_DLL_SPEC int MPI_CALL MPI_Win_create_keyval(MPI_Win_copy_attr_function *, 
						 MPI_Win_delete_attr_function *, int *, void *);
MPI_DLL_SPEC int MPI_CALL MPI_Win_delete_attr(MPI_Win, int);
MPI_DLL_SPEC int MPI_CALL MPI_Win_free_keyval(int *);
MPI_DLL_SPEC int MPI_CALL MPI_Win_get_attr(MPI_Win, int, void *, int *);
MPI_DLL_SPEC int MPI_CALL MPI_Win_get_nameA(MPI_Win, char *, int *);
MPI_DLL_SPEC int MPI_CALL MPI_Win_get_nameW(MPI_Win, wchar_t *, int *);
MPI_DLL_SPEC int MPI_CALL MPI_Win_set_attr(MPI_Win, int, void *);
MPI_DLL_SPEC int MPI_CALL MPI_Win_set_nameA(MPI_Win, char *);
MPI_DLL_SPEC int MPI_CALL MPI_Win_set_nameW(MPI_Win, wchar_t *);

/* Miscellany */
#ifdef FOO
MPI_Comm MPI_Comm_f2c(MPI_Fint);
MPI_Datatype MPI_Type_f2c(MPI_Fint);
MPI_File MPI_File_f2c(MPI_Fint);
MPI_Fint MPI_Comm_c2f(MPI_Comm);
MPI_Fint MPI_File_c2f(MPI_File);
MPI_Fint MPI_Group_c2f(MPI_Group);
MPI_Fint MPI_Info_c2f(MPI_Info);
MPI_Fint MPI_Op_c2f(MPI_Op);
MPI_Fint MPI_Request_c2f(MPI_Request);
MPI_Fint MPI_Type_c2f(MPI_Datatype);
MPI_Fint MPI_Win_c2f(MPI_Win);
MPI_Group MPI_Group_f2c(MPI_Fint);
MPI_Info MPI_Info_f2c(MPI_Fint);
MPI_Op MPI_Op_f2c(MPI_Fint);
MPI_Request MPI_Request_f2c(MPI_Fint);
MPI_Win MPI_Win_f2c(MPI_Fint);
#endif

MPI_DLL_SPEC int MPI_CALL MPI_Alloc_mem(MPI_Aint, MPI_Info info, void *baseptr);
MPI_DLL_SPEC int MPI_CALL MPI_Comm_create_errhandler(MPI_Comm_errhandler_fn *, MPI_Errhandler *);
MPI_DLL_SPEC int MPI_CALL MPI_Comm_get_errhandler(MPI_Comm, MPI_Errhandler *);
MPI_DLL_SPEC int MPI_CALL MPI_Comm_set_errhandler(MPI_Comm, MPI_Errhandler);
MPI_DLL_SPEC int MPI_CALL MPI_File_create_errhandler(MPI_File_errhandler_fn *, MPI_Errhandler *);
MPI_DLL_SPEC int MPI_CALL MPI_File_get_errhandler(MPI_File, MPI_Errhandler *);
MPI_DLL_SPEC int MPI_CALL MPI_File_set_errhandler(MPI_File, MPI_Errhandler);
MPI_DLL_SPEC int MPI_CALL MPI_Finalized(int *);
MPI_DLL_SPEC int MPI_CALL MPI_Free_mem(void *);
MPI_DLL_SPEC int MPI_CALL MPI_Get_address(void *, MPI_Aint *);
MPI_DLL_SPEC int MPI_CALL MPI_Info_create(MPI_Info *);
MPI_DLL_SPEC int MPI_CALL MPI_Info_deleteA(MPI_Info, char *);
MPI_DLL_SPEC int MPI_CALL MPI_Info_deleteW(MPI_Info, wchar_t *);
MPI_DLL_SPEC int MPI_CALL MPI_Info_dup(MPI_Info, MPI_Info *);
MPI_DLL_SPEC int MPI_CALL MPI_Info_free(MPI_Info *info);
MPI_DLL_SPEC int MPI_CALL MPI_Info_getA(MPI_Info, char *, int, char *, int *);
MPI_DLL_SPEC int MPI_CALL MPI_Info_getW(MPI_Info, wchar_t *, int, wchar_t *, int *);
MPI_DLL_SPEC int MPI_CALL MPI_Info_get_nkeys(MPI_Info, int *);
MPI_DLL_SPEC int MPI_CALL MPI_Info_get_nthkeyA(MPI_Info, int, char *);
MPI_DLL_SPEC int MPI_CALL MPI_Info_get_nthkeyW(MPI_Info, int, wchar_t *);
MPI_DLL_SPEC int MPI_CALL MPI_Info_get_valuelenA(MPI_Info, char *, int *, int *);
MPI_DLL_SPEC int MPI_CALL MPI_Info_get_valuelenW(MPI_Info, wchar_t *, int *, int *);
MPI_DLL_SPEC int MPI_CALL MPI_Info_setA(MPI_Info, char *, char *);
MPI_DLL_SPEC int MPI_CALL MPI_Info_setW(MPI_Info, wchar_t *, wchar_t *);
MPI_DLL_SPEC int MPI_CALL MPI_Pack_externalA(char *, void *, int, MPI_Datatype, void *, MPI_Aint, MPI_Aint *);
MPI_DLL_SPEC int MPI_CALL MPI_Pack_externalW(wchar_t *, void *, int, MPI_Datatype, void *, MPI_Aint, MPI_Aint *);
MPI_DLL_SPEC int MPI_CALL MPI_Pack_external_sizeA(char *, int, MPI_Datatype, MPI_Aint *); 
MPI_DLL_SPEC int MPI_CALL MPI_Pack_external_sizeW(wchar_t *, int, MPI_Datatype, MPI_Aint *);
MPI_DLL_SPEC int MPI_CALL MPI_Request_get_status(MPI_Request, int *, MPI_Status *);
MPI_DLL_SPEC int MPI_CALL MPI_Status_c2f(MPI_Status *, MPI_Fint *);
MPI_DLL_SPEC int MPI_CALL MPI_Status_f2c(MPI_Fint *, MPI_Status *);
MPI_DLL_SPEC int MPI_CALL MPI_Type_create_darray(int, int, int, int [], int [], int [], int [], int,
						   MPI_Datatype, MPI_Datatype *);
MPI_DLL_SPEC int MPI_CALL MPI_Type_create_hindexed(int, int [], MPI_Aint [], MPI_Datatype, 
							 MPI_Datatype *);
MPI_DLL_SPEC int MPI_CALL MPI_Type_create_hvector(int, int, MPI_Aint, MPI_Datatype, MPI_Datatype *);
MPI_DLL_SPEC int MPI_CALL MPI_Type_create_indexed_block(int, int, int [], MPI_Datatype, 
								  MPI_Datatype *);
MPI_DLL_SPEC int MPI_CALL MPI_Type_create_resized(MPI_Datatype, MPI_Aint, MPI_Aint, MPI_Datatype *);
MPI_DLL_SPEC int MPI_CALL MPI_Type_create_struct(int, int [], MPI_Aint [], MPI_Datatype [], 
						   MPI_Datatype *);
MPI_DLL_SPEC int MPI_CALL MPI_Type_create_subarray(int, int [], int [], int [], int, MPI_Datatype, 
							 MPI_Datatype *);
MPI_DLL_SPEC int MPI_CALL MPI_Type_get_extent(MPI_Datatype, MPI_Aint *, MPI_Aint *);
MPI_DLL_SPEC int MPI_CALL MPI_Type_get_true_extent(MPI_Datatype, MPI_Aint *, MPI_Aint *);
MPI_DLL_SPEC int MPI_CALL MPI_Unpack_externalA(char *, void *, MPI_Aint, MPI_Aint *, void *, int, MPI_Datatype);
MPI_DLL_SPEC int MPI_CALL MPI_Unpack_externalW(wchar_t *, void *, MPI_Aint, MPI_Aint *, void *, int, MPI_Datatype);
MPI_DLL_SPEC int MPI_CALL MPI_Win_create_errhandler(MPI_Win_errhandler_fn *, MPI_Errhandler *);
MPI_DLL_SPEC int MPI_CALL MPI_Win_get_errhandler(MPI_Win, MPI_Errhandler *);
MPI_DLL_SPEC int MPI_CALL MPI_Win_set_errhandler(MPI_Win, MPI_Errhandler);

/* Fortran 90-related functions.  These routines are available only if
   Fortran 90 support is enabled 
*/
MPI_DLL_SPEC int MPI_CALL MPI_Type_create_f90_integer( int, MPI_Datatype * );
MPI_DLL_SPEC int MPI_CALL MPI_Type_create_f90_real( int, int, MPI_Datatype * );
MPI_DLL_SPEC int MPI_CALL MPI_Type_create_f90_complex( int, int, MPI_Datatype * );

/* End Prototypes */
#endif /* MPICH_SUPPRESS_PROTOTYPES */



/* Here are the bindings of the profiling routines */
#if !defined(MPI_BUILD_PROFILING)
MPI_DLL_SPEC int MPI_CALL PMPI_Send(void*, int, MPI_Datatype, int, int, MPI_Comm);
MPI_DLL_SPEC int MPI_CALL PMPI_Recv(void*, int, MPI_Datatype, int, int, MPI_Comm, MPI_Status *);
MPI_DLL_SPEC int MPI_CALL PMPI_Get_count(MPI_Status *, MPI_Datatype, int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Bsend(void*, int, MPI_Datatype, int, int, MPI_Comm);
MPI_DLL_SPEC int MPI_CALL PMPI_Ssend(void*, int, MPI_Datatype, int, int, MPI_Comm);
MPI_DLL_SPEC int MPI_CALL PMPI_Rsend(void*, int, MPI_Datatype, int, int, MPI_Comm);
MPI_DLL_SPEC int MPI_CALL PMPI_Buffer_attach( void* buffer, int);
MPI_DLL_SPEC int MPI_CALL PMPI_Buffer_detach( void* buffer, int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Isend(void*, int, MPI_Datatype, int, int, MPI_Comm, MPI_Request *);
MPI_DLL_SPEC int MPI_CALL PMPI_Ibsend(void*, int, MPI_Datatype, int, int, MPI_Comm, MPI_Request *);
MPI_DLL_SPEC int MPI_CALL PMPI_Issend(void*, int, MPI_Datatype, int, int, MPI_Comm, MPI_Request *);
MPI_DLL_SPEC int MPI_CALL PMPI_Irsend(void*, int, MPI_Datatype, int, int, MPI_Comm, MPI_Request *);
MPI_DLL_SPEC int MPI_CALL PMPI_Irecv(void*, int, MPI_Datatype, int, int, MPI_Comm, MPI_Request *);
MPI_DLL_SPEC int MPI_CALL PMPI_Wait(MPI_Request *, MPI_Status *);
MPI_DLL_SPEC int MPI_CALL PMPI_Test(MPI_Request *, int *, MPI_Status *);
MPI_DLL_SPEC int MPI_CALL PMPI_Request_free(MPI_Request *);
MPI_DLL_SPEC int MPI_CALL PMPI_Waitany(int, MPI_Request *, int *, MPI_Status *);
MPI_DLL_SPEC int MPI_CALL PMPI_Testany(int, MPI_Request *, int *, int *, MPI_Status *);
MPI_DLL_SPEC int MPI_CALL PMPI_Waitall(int, MPI_Request *, MPI_Status *);
MPI_DLL_SPEC int MPI_CALL PMPI_Testall(int, MPI_Request *, int *, MPI_Status *);
MPI_DLL_SPEC int MPI_CALL PMPI_Waitsome(int, MPI_Request *, int *, int *, MPI_Status *);
MPI_DLL_SPEC int MPI_CALL PMPI_Testsome(int, MPI_Request *, int *, int *, MPI_Status *);
MPI_DLL_SPEC int MPI_CALL PMPI_Iprobe(int, int, MPI_Comm, int *, MPI_Status *);
MPI_DLL_SPEC int MPI_CALL PMPI_Probe(int, int, MPI_Comm, MPI_Status *);
MPI_DLL_SPEC int MPI_CALL PMPI_Cancel(MPI_Request *);
MPI_DLL_SPEC int MPI_CALL PMPI_Test_cancelled(MPI_Status *, int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Send_init(void*, int, MPI_Datatype, int, int, MPI_Comm, MPI_Request *);
MPI_DLL_SPEC int MPI_CALL PMPI_Bsend_init(void*, int, MPI_Datatype, int, int, MPI_Comm, MPI_Request *);
MPI_DLL_SPEC int MPI_CALL PMPI_Ssend_init(void*, int, MPI_Datatype, int, int, MPI_Comm, MPI_Request *);
MPI_DLL_SPEC int MPI_CALL PMPI_Rsend_init(void*, int, MPI_Datatype, int, int, MPI_Comm, MPI_Request *);
MPI_DLL_SPEC int MPI_CALL PMPI_Recv_init(void*, int, MPI_Datatype, int, int, MPI_Comm, MPI_Request *);
MPI_DLL_SPEC int MPI_CALL PMPI_Start(MPI_Request *);
MPI_DLL_SPEC int MPI_CALL PMPI_Startall(int, MPI_Request *);
MPI_DLL_SPEC int MPI_CALL PMPI_Sendrecv(void *, int, MPI_Datatype, int, int, void *, int, MPI_Datatype, int, int, MPI_Comm, MPI_Status *);
MPI_DLL_SPEC int MPI_CALL PMPI_Sendrecv_replace(void*, int, MPI_Datatype, int, int, int, int, MPI_Comm, MPI_Status *);
MPI_DLL_SPEC int MPI_CALL PMPI_Type_contiguous(int, MPI_Datatype, MPI_Datatype *);
MPI_DLL_SPEC int MPI_CALL PMPI_Type_vector(int, int, int, MPI_Datatype, MPI_Datatype *);
MPI_DLL_SPEC int MPI_CALL PMPI_Type_hvector(int, int, MPI_Aint, MPI_Datatype, MPI_Datatype *);
MPI_DLL_SPEC int MPI_CALL PMPI_Type_indexed(int, int *, int *, MPI_Datatype, MPI_Datatype *);
MPI_DLL_SPEC int MPI_CALL PMPI_Type_hindexed(int, int *, MPI_Aint *, MPI_Datatype, MPI_Datatype *);
MPI_DLL_SPEC int MPI_CALL PMPI_Type_struct(int, int *, MPI_Aint *, MPI_Datatype *, MPI_Datatype *);
MPI_DLL_SPEC int MPI_CALL PMPI_Address(void*, MPI_Aint *);
MPI_DLL_SPEC int MPI_CALL PMPI_Type_extent(MPI_Datatype, MPI_Aint *);
MPI_DLL_SPEC int MPI_CALL PMPI_Type_size(MPI_Datatype, int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Type_lb(MPI_Datatype, MPI_Aint *);
MPI_DLL_SPEC int MPI_CALL PMPI_Type_ub(MPI_Datatype, MPI_Aint *);
MPI_DLL_SPEC int MPI_CALL PMPI_Type_commit(MPI_Datatype *);
MPI_DLL_SPEC int MPI_CALL PMPI_Type_free(MPI_Datatype *);
MPI_DLL_SPEC int MPI_CALL PMPI_Get_elements(MPI_Status *, MPI_Datatype, int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Pack(void*, int, MPI_Datatype, void *, int, int *,  MPI_Comm);
MPI_DLL_SPEC int MPI_CALL PMPI_Unpack(void*, int, int *, void *, int, MPI_Datatype, MPI_Comm);
MPI_DLL_SPEC int MPI_CALL PMPI_Pack_size(int, MPI_Datatype, MPI_Comm, int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Barrier(MPI_Comm );
MPI_DLL_SPEC int MPI_CALL PMPI_Bcast(void* buffer, int, MPI_Datatype, int, MPI_Comm );
MPI_DLL_SPEC int MPI_CALL PMPI_Gather(void* , int, MPI_Datatype, void*, int, MPI_Datatype, int, MPI_Comm); 
MPI_DLL_SPEC int MPI_CALL PMPI_Gatherv(void* , int, MPI_Datatype, void*, int *, int *, MPI_Datatype, int, MPI_Comm); 
MPI_DLL_SPEC int MPI_CALL PMPI_Scatter(void* , int, MPI_Datatype, void*, int, MPI_Datatype, int, MPI_Comm);
MPI_DLL_SPEC int MPI_CALL PMPI_Scatterv(void* , int *, int *displs, MPI_Datatype, void*, int, MPI_Datatype, int, MPI_Comm);
MPI_DLL_SPEC int MPI_CALL PMPI_Allgather(void* , int, MPI_Datatype, void*, int, MPI_Datatype, MPI_Comm);
MPI_DLL_SPEC int MPI_CALL PMPI_Allgatherv(void* , int, MPI_Datatype, void*, int *, int *, MPI_Datatype, MPI_Comm);
MPI_DLL_SPEC int MPI_CALL PMPI_Alltoall(void* , int, MPI_Datatype, void*, int, MPI_Datatype, MPI_Comm);
MPI_DLL_SPEC int MPI_CALL PMPI_Alltoallv(void* , int *, int *, MPI_Datatype, void*, int *, int *, MPI_Datatype, MPI_Comm);
MPI_DLL_SPEC int MPI_CALL PMPI_Reduce(void* , void*, int, MPI_Datatype, MPI_Op, int, MPI_Comm);
MPI_DLL_SPEC int MPI_CALL PMPI_Op_create(MPI_User_function *, int, MPI_Op *);
MPI_DLL_SPEC int MPI_CALL PMPI_Op_free( MPI_Op *);
MPI_DLL_SPEC int MPI_CALL PMPI_Allreduce(void* , void*, int, MPI_Datatype, MPI_Op, MPI_Comm);
MPI_DLL_SPEC int MPI_CALL PMPI_Reduce_scatter(void* , void*, int *, MPI_Datatype, MPI_Op, MPI_Comm);
MPI_DLL_SPEC int MPI_CALL PMPI_Scan(void* , void*, int, MPI_Datatype, MPI_Op, MPI_Comm );
MPI_DLL_SPEC int MPI_CALL PMPI_Group_size(MPI_Group, int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Group_rank(MPI_Group, int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Group_translate_ranks (MPI_Group, int, int *, MPI_Group, int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Group_compare(MPI_Group, MPI_Group, int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Comm_group(MPI_Comm, MPI_Group *);
MPI_DLL_SPEC int MPI_CALL PMPI_Group_union(MPI_Group, MPI_Group, MPI_Group *);
MPI_DLL_SPEC int MPI_CALL PMPI_Group_intersection(MPI_Group, MPI_Group, MPI_Group *);
MPI_DLL_SPEC int MPI_CALL PMPI_Group_difference(MPI_Group, MPI_Group, MPI_Group *);
MPI_DLL_SPEC int MPI_CALL PMPI_Group_incl(MPI_Group, int, int *, MPI_Group *);
MPI_DLL_SPEC int MPI_CALL PMPI_Group_excl(MPI_Group, int, int *, MPI_Group *);
MPI_DLL_SPEC int MPI_CALL PMPI_Group_range_incl(MPI_Group, int, int [][3], MPI_Group *);
MPI_DLL_SPEC int MPI_CALL PMPI_Group_range_excl(MPI_Group, int, int [][3], MPI_Group *);
MPI_DLL_SPEC int MPI_CALL PMPI_Group_free(MPI_Group *);
MPI_DLL_SPEC int MPI_CALL PMPI_Comm_size(MPI_Comm, int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Comm_rank(MPI_Comm, int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Comm_compare(MPI_Comm, MPI_Comm, int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Comm_dup(MPI_Comm, MPI_Comm *);
MPI_DLL_SPEC int MPI_CALL PMPI_Comm_create(MPI_Comm, MPI_Group, MPI_Comm *);
MPI_DLL_SPEC int MPI_CALL PMPI_Comm_split(MPI_Comm, int, int, MPI_Comm *);
MPI_DLL_SPEC int MPI_CALL PMPI_Comm_free(MPI_Comm *);
MPI_DLL_SPEC int MPI_CALL PMPI_Comm_test_inter(MPI_Comm, int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Comm_remote_size(MPI_Comm, int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Comm_remote_group(MPI_Comm, MPI_Group *);
MPI_DLL_SPEC int MPI_CALL PMPI_Intercomm_create(MPI_Comm, int, MPI_Comm, int, int, MPI_Comm *);
MPI_DLL_SPEC int MPI_CALL PMPI_Intercomm_merge(MPI_Comm, int, MPI_Comm *);
MPI_DLL_SPEC int MPI_CALL PMPI_Keyval_create(MPI_Copy_function *, MPI_Delete_function *, int *, void*);
MPI_DLL_SPEC int MPI_CALL PMPI_Keyval_free(int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Attr_put(MPI_Comm, int, void*);
MPI_DLL_SPEC int MPI_CALL PMPI_Attr_get(MPI_Comm, int, void *, int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Attr_delete(MPI_Comm, int);
MPI_DLL_SPEC int MPI_CALL PMPI_Topo_test(MPI_Comm, int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Cart_create(MPI_Comm, int, int *, int *, int, MPI_Comm *);
MPI_DLL_SPEC int MPI_CALL PMPI_Dims_create(int, int, int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Graph_create(MPI_Comm, int, int *, int *, int, MPI_Comm *);
MPI_DLL_SPEC int MPI_CALL PMPI_Graphdims_get(MPI_Comm, int *, int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Graph_get(MPI_Comm, int, int, int *, int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Cartdim_get(MPI_Comm, int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Cart_get(MPI_Comm, int, int *, int *, int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Cart_rank(MPI_Comm, int *, int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Cart_coords(MPI_Comm, int, int, int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Graph_neighbors_count(MPI_Comm, int, int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Graph_neighbors(MPI_Comm, int, int, int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Cart_shift(MPI_Comm, int, int, int *, int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Cart_sub(MPI_Comm, int *, MPI_Comm *);
MPI_DLL_SPEC int MPI_CALL PMPI_Cart_map(MPI_Comm, int, int *, int *, int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Graph_map(MPI_Comm, int, int *, int *, int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Get_processor_nameA(char *, int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Get_processor_nameW(wchar_t *, int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Get_version(int *, int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Errhandler_create(MPI_Handler_function *, MPI_Errhandler *);
MPI_DLL_SPEC int MPI_CALL PMPI_Errhandler_set(MPI_Comm, MPI_Errhandler);
MPI_DLL_SPEC int MPI_CALL PMPI_Errhandler_get(MPI_Comm, MPI_Errhandler *);
MPI_DLL_SPEC int MPI_CALL PMPI_Errhandler_free(MPI_Errhandler *);
MPI_DLL_SPEC int MPI_CALL PMPI_Error_stringA(int, char *, int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Error_stringW(int, wchar_t *, int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Error_class(int, int *);

MPI_DLL_SPEC int MPI_CALL PMPI_Type_get_envelope(MPI_Datatype, int *, int *, int *, int *); 

/* Wtime done above */
MPI_DLL_SPEC int MPI_CALL PMPI_InitA(int *, char ***);
MPI_DLL_SPEC int MPI_CALL PMPI_InitW(int *, wchar_t ***);
MPI_DLL_SPEC int MPI_CALL PMPI_Finalize(void);
MPI_DLL_SPEC int MPI_CALL PMPI_Initialized(int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Abort(MPI_Comm, int);

MPI_DLL_SPEC int MPI_CALL PMPI_Pcontrol(const int, ...);

/* MPI-2 functions */

/* Process Creation and Management */
MPI_DLL_SPEC int MPI_CALL PMPI_Close_portA(char *);
MPI_DLL_SPEC int MPI_CALL PMPI_Close_portW(wchar_t *);
MPI_DLL_SPEC int MPI_CALL PMPI_Comm_acceptA(char *, MPI_Info, int, MPI_Comm, MPI_Comm *);
MPI_DLL_SPEC int MPI_CALL PMPI_Comm_acceptW(wchar_t *, MPI_Info, int, MPI_Comm, MPI_Comm *);
MPI_DLL_SPEC int MPI_CALL PMPI_Comm_connectA(char *, MPI_Info, int, MPI_Comm, MPI_Comm *);
MPI_DLL_SPEC int MPI_CALL PMPI_Comm_connectW(wchar_t *, MPI_Info, int, MPI_Comm, MPI_Comm *);
MPI_DLL_SPEC int MPI_CALL PMPI_Comm_disconnect(MPI_Comm *);
MPI_DLL_SPEC int MPI_CALL PMPI_Comm_get_parent(MPI_Comm *);
MPI_DLL_SPEC int MPI_CALL PMPI_Comm_join(int, MPI_Comm *);
MPI_DLL_SPEC int MPI_CALL PMPI_Comm_spawnA(char *, char *[], int, MPI_Info, int, MPI_Comm, MPI_Comm *, int []);
MPI_DLL_SPEC int MPI_CALL PMPI_Comm_spawnW(wchar_t *, wchar_t *[], int, MPI_Info, int, MPI_Comm, MPI_Comm *, int []);
MPI_DLL_SPEC int MPI_CALL PMPI_Comm_spawn_multipleA(int, char *[], char **[], int [], MPI_Info [], int, MPI_Comm, MPI_Comm *, int []);
MPI_DLL_SPEC int MPI_CALL PMPI_Comm_spawn_multipleW(int, wchar_t *[], wchar_t **[], int [], MPI_Info [], int, MPI_Comm, MPI_Comm *, int []);
MPI_DLL_SPEC int MPI_CALL PMPI_Lookup_nameA(char *, MPI_Info, char *);
MPI_DLL_SPEC int MPI_CALL PMPI_Lookup_nameW(wchar_t *, MPI_Info, wchar_t *);
MPI_DLL_SPEC int MPI_CALL PMPI_Open_portA(MPI_Info, char *);
MPI_DLL_SPEC int MPI_CALL PMPI_Open_portW(MPI_Info, wchar_t *);
MPI_DLL_SPEC int MPI_CALL PMPI_Publish_nameA(char *, MPI_Info, char *);
MPI_DLL_SPEC int MPI_CALL PMPI_Publish_nameW(wchar_t *, MPI_Info, wchar_t *);
MPI_DLL_SPEC int MPI_CALL PMPI_Unpublish_nameA(char *, MPI_Info, char *);
MPI_DLL_SPEC int MPI_CALL PMPI_Unpublish_nameW(wchar_t *, MPI_Info, wchar_t *);

/* One-Sided Communications */
MPI_DLL_SPEC int MPI_CALL PMPI_Accumulate(void *, int, MPI_Datatype, int, MPI_Aint, int, 
				   MPI_Datatype,  MPI_Op, MPI_Win);
MPI_DLL_SPEC int MPI_CALL PMPI_Get(void *, int, MPI_Datatype, int, MPI_Aint, int, MPI_Datatype, 
			MPI_Win);
MPI_DLL_SPEC int MPI_CALL PMPI_Put(void *, int, MPI_Datatype, int, MPI_Aint, int, MPI_Datatype, 
			MPI_Win);
MPI_DLL_SPEC int MPI_CALL PMPI_Win_complete(MPI_Win);
MPI_DLL_SPEC int MPI_CALL PMPI_Win_create(void *, MPI_Aint, int, MPI_Info, MPI_Comm, MPI_Win *);
MPI_DLL_SPEC int MPI_CALL PMPI_Win_fence(int, MPI_Win);
MPI_DLL_SPEC int MPI_CALL PMPI_Win_free(MPI_Win *);
MPI_DLL_SPEC int MPI_CALL PMPI_Win_get_group(MPI_Win, MPI_Group *);
MPI_DLL_SPEC int MPI_CALL PMPI_Win_lock(int, int, int, MPI_Win);
MPI_DLL_SPEC int MPI_CALL PMPI_Win_post(MPI_Group, int, MPI_Win);
MPI_DLL_SPEC int MPI_CALL PMPI_Win_start(MPI_Group, int, MPI_Win);
MPI_DLL_SPEC int MPI_CALL PMPI_Win_test(MPI_Win, int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Win_unlock(int, MPI_Win);
MPI_DLL_SPEC int MPI_CALL PMPI_Win_wait(MPI_Win);
 
/* Extended Collective Operations */
MPI_DLL_SPEC int MPI_CALL PMPI_Alltoallw(void *, int [], int [], MPI_Datatype [], void *, int [], 
				  int [], MPI_Datatype [], MPI_Comm);
MPI_DLL_SPEC int MPI_CALL PMPI_Exscan(void *, void *, int, MPI_Datatype, MPI_Op, MPI_Comm) ;
 
/* External Interfaces */
MPI_DLL_SPEC int MPI_CALL PMPI_Add_error_class(int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Add_error_code(int, int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Add_error_stringA(int, char *);
MPI_DLL_SPEC int MPI_CALL PMPI_Add_error_stringW(int, wchar_t *);
MPI_DLL_SPEC int MPI_CALL PMPI_Comm_call_errhandler(MPI_Comm, int);
MPI_DLL_SPEC int MPI_CALL PMPI_Comm_create_keyval(MPI_Comm_copy_attr_function *, 
						   MPI_Comm_delete_attr_function *, int *, void *);
MPI_DLL_SPEC int MPI_CALL PMPI_Comm_delete_attr(MPI_Comm, int);
MPI_DLL_SPEC int MPI_CALL PMPI_Comm_free_keyval(int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Comm_get_attr(MPI_Comm, int, void *, int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Comm_get_nameA(MPI_Comm, char *, int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Comm_get_nameW(MPI_Comm, wchar_t *, int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Comm_set_attr(MPI_Comm, int, void *);
MPI_DLL_SPEC int MPI_CALL PMPI_Comm_set_nameA(MPI_Comm, char *);
MPI_DLL_SPEC int MPI_CALL PMPI_Comm_set_nameW(MPI_Comm, wchar_t *);
MPI_DLL_SPEC int MPI_CALL PMPI_File_call_errhandler(MPI_File, int);
MPI_DLL_SPEC int MPI_CALL PMPI_Grequest_complete(MPI_Request);
MPI_DLL_SPEC int MPI_CALL PMPI_Grequest_start(MPI_Grequest_query_function *, 
					   MPI_Grequest_free_function *, 
					   MPI_Grequest_cancel_function *, void *, MPI_Request *);
MPI_DLL_SPEC int MPI_CALL PMPI_Init_threadA(int *, char ***, int, int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Init_threadW(int *, wchar_t ***, int, int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Is_thread_main(int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Query_thread(int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Status_set_cancelled(MPI_Status *, int);
MPI_DLL_SPEC int MPI_CALL PMPI_Status_set_elements(MPI_Status *, MPI_Datatype, int);
MPI_DLL_SPEC int MPI_CALL PMPI_Type_create_keyval(MPI_Type_copy_attr_function *, 
						   MPI_Type_delete_attr_function *, int *, void *);
MPI_DLL_SPEC int MPI_CALL PMPI_Type_delete_attr(MPI_Datatype, int);
MPI_DLL_SPEC int MPI_CALL PMPI_Type_dup(MPI_Datatype, MPI_Datatype *);
MPI_DLL_SPEC int MPI_CALL PMPI_Type_free_keyval(int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Type_get_attr(MPI_Datatype, int, void *, int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Type_get_contents(MPI_Datatype, int, int, int, int [], MPI_Aint [], 
						  MPI_Datatype []);
MPI_DLL_SPEC int MPI_CALL PMPI_Type_get_envelope(MPI_Datatype, int *, int *, int *, int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Type_get_nameA(MPI_Datatype, char *, int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Type_get_nameW(MPI_Datatype, wchar_t *, int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Type_set_attr(MPI_Datatype, int, void *);
MPI_DLL_SPEC int MPI_CALL PMPI_Type_set_nameA(MPI_Datatype, char *);
MPI_DLL_SPEC int MPI_CALL PMPI_Type_set_nameW(MPI_Datatype, wchar_t *);
MPI_DLL_SPEC int MPI_CALL PMPI_Type_match_size( int, int, MPI_Datatype *);
MPI_DLL_SPEC int MPI_CALL PMPI_Win_call_errhandler(MPI_Win, int);
MPI_DLL_SPEC int MPI_CALL PMPI_Win_create_keyval(MPI_Win_copy_attr_function *, 
						 MPI_Win_delete_attr_function *, int *, void *);
MPI_DLL_SPEC int MPI_CALL PMPI_Win_delete_attr(MPI_Win, int);
MPI_DLL_SPEC int MPI_CALL PMPI_Win_free_keyval(int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Win_get_attr(MPI_Win, int, void *, int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Win_get_nameA(MPI_Win, char *, int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Win_get_nameW(MPI_Win, wchar_t *, int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Win_set_attr(MPI_Win, int, void *);
MPI_DLL_SPEC int MPI_CALL PMPI_Win_set_nameA(MPI_Win, char *);
MPI_DLL_SPEC int MPI_CALL PMPI_Win_set_nameW(MPI_Win, wchar_t *);

/* Fortran 90-related functions.  These routines are available only if
   Fortran 90 support is enabled 
*/
MPI_DLL_SPEC int MPI_CALL PMPI_Type_create_f90_integer( int, MPI_Datatype * );
MPI_DLL_SPEC int MPI_CALL PMPI_Type_create_f90_real( int, int, MPI_Datatype * );
MPI_DLL_SPEC int MPI_CALL PMPI_Type_create_f90_complex( int, int, MPI_Datatype * );

/* Miscellany */
MPI_DLL_SPEC int MPI_CALL PMPI_Alloc_mem(MPI_Aint, MPI_Info info, void *baseptr);
MPI_DLL_SPEC int MPI_CALL PMPI_Comm_create_errhandler(MPI_Comm_errhandler_fn *, MPI_Errhandler *);
MPI_DLL_SPEC int MPI_CALL PMPI_Comm_get_errhandler(MPI_Comm, MPI_Errhandler *);
MPI_DLL_SPEC int MPI_CALL PMPI_Comm_set_errhandler(MPI_Comm, MPI_Errhandler);
MPI_DLL_SPEC int MPI_CALL PMPI_File_create_errhandler(MPI_File_errhandler_fn *, MPI_Errhandler *);
MPI_DLL_SPEC int MPI_CALL PMPI_File_get_errhandler(MPI_File, MPI_Errhandler *);
MPI_DLL_SPEC int MPI_CALL PMPI_File_set_errhandler(MPI_File, MPI_Errhandler);
MPI_DLL_SPEC int MPI_CALL PMPI_Finalized(int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Free_mem(void *);
MPI_DLL_SPEC int MPI_CALL PMPI_Get_address(void *, MPI_Aint *);
MPI_DLL_SPEC int MPI_CALL PMPI_Info_create(MPI_Info *);
MPI_DLL_SPEC int MPI_CALL PMPI_Info_deleteA(MPI_Info, char *);
MPI_DLL_SPEC int MPI_CALL PMPI_Info_deleteW(MPI_Info, wchar_t *);
MPI_DLL_SPEC int MPI_CALL PMPI_Info_dup(MPI_Info, MPI_Info *);
MPI_DLL_SPEC int MPI_CALL PMPI_Info_free(MPI_Info *info);
MPI_DLL_SPEC int MPI_CALL PMPI_Info_getA(MPI_Info, char *, int, char *, int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Info_getW(MPI_Info, wchar_t *, int, wchar_t *, int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Info_get_nkeys(MPI_Info, int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Info_get_nthkeyA(MPI_Info, int, char *);
MPI_DLL_SPEC int MPI_CALL PMPI_Info_get_nthkeyW(MPI_Info, int, wchar_t *);
MPI_DLL_SPEC int MPI_CALL PMPI_Info_get_valuelenA(MPI_Info, char *, int *, int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Info_get_valuelenW(MPI_Info, wchar_t *, int *, int *);
MPI_DLL_SPEC int MPI_CALL PMPI_Info_setA(MPI_Info, char *, char *);
MPI_DLL_SPEC int MPI_CALL PMPI_Info_setW(MPI_Info, wchar_t *, wchar_t *);
MPI_DLL_SPEC int MPI_CALL PMPI_Pack_externalA(char *, void *, int, MPI_Datatype, void *, MPI_Aint, MPI_Aint *);
MPI_DLL_SPEC int MPI_CALL PMPI_Pack_externalW(wchar_t *, void *, int, MPI_Datatype, void *, MPI_Aint, MPI_Aint *);
MPI_DLL_SPEC int MPI_CALL PMPI_Pack_external_sizeA(char *, int, MPI_Datatype, MPI_Aint *); 
MPI_DLL_SPEC int MPI_CALL PMPI_Pack_external_sizeW(wchar_t *, int, MPI_Datatype, MPI_Aint *);
MPI_DLL_SPEC int MPI_CALL PMPI_Request_get_status(MPI_Request, int *, MPI_Status *);
MPI_DLL_SPEC int MPI_CALL PMPI_Status_c2f(MPI_Status *, MPI_Fint *);
MPI_DLL_SPEC int MPI_CALL PMPI_Status_f2c(MPI_Fint *, MPI_Status *);
MPI_DLL_SPEC int MPI_CALL PMPI_Type_create_darray(int, int, int, int [], int [], int [], int [], int,
						   MPI_Datatype, MPI_Datatype *);
MPI_DLL_SPEC int MPI_CALL PMPI_Type_create_hindexed(int, int [], MPI_Aint [], MPI_Datatype, 
							 MPI_Datatype *);
MPI_DLL_SPEC int MPI_CALL PMPI_Type_create_hvector(int, int, MPI_Aint, MPI_Datatype, MPI_Datatype *);
MPI_DLL_SPEC int MPI_CALL PMPI_Type_create_indexed_block(int, int, int [], MPI_Datatype, 
								  MPI_Datatype *);
MPI_DLL_SPEC int MPI_CALL PMPI_Type_create_resized(MPI_Datatype, MPI_Aint, MPI_Aint, MPI_Datatype *);
MPI_DLL_SPEC int MPI_CALL PMPI_Type_create_struct(int, int [], MPI_Aint [], MPI_Datatype [], 
						   MPI_Datatype *);
MPI_DLL_SPEC int MPI_CALL PMPI_Type_create_subarray(int, int [], int [], int [], int, MPI_Datatype, 
							 MPI_Datatype *);
MPI_DLL_SPEC int MPI_CALL PMPI_Type_get_extent(MPI_Datatype, MPI_Aint *, MPI_Aint *);
MPI_DLL_SPEC int MPI_CALL PMPI_Type_get_true_extent(MPI_Datatype, MPI_Aint *, MPI_Aint *);
MPI_DLL_SPEC int MPI_CALL PMPI_Unpack_externalA(char *, void *, MPI_Aint, MPI_Aint *, void *, int, MPI_Datatype);
MPI_DLL_SPEC int MPI_CALL PMPI_Unpack_externalW(wchar_t *, void *, MPI_Aint, MPI_Aint *, void *, int, MPI_Datatype);
MPI_DLL_SPEC int MPI_CALL PMPI_Win_create_errhandler(MPI_Win_errhandler_fn *, MPI_Errhandler *);
MPI_DLL_SPEC int MPI_CALL PMPI_Win_get_errhandler(MPI_Win, MPI_Errhandler *);
MPI_DLL_SPEC int MPI_CALL PMPI_Win_set_errhandler(MPI_Win, MPI_Errhandler);
#endif  /* MPI_BUILD_PROFILING */
/* End of MPI bindings */

/* feature advertisement */
#define MPIIMPL_ADVERTISES_FEATURES 1
#define MPIIMPL_HAVE_MPI_INFO 1												 
#define MPIIMPL_HAVE_MPI_COMBINER_DARRAY 1									  
#define MPIIMPL_HAVE_MPI_TYPE_CREATE_DARRAY 1
#define MPIIMPL_HAVE_MPI_COMBINER_SUBARRAY 1									
#define MPIIMPL_HAVE_MPI_TYPE_CREATE_DARRAY 1
#define MPIIMPL_HAVE_MPI_COMBINER_DUP 1										 
#define MPIIMPL_HAVE_MPI_GREQUEST 1	  
#define MPIIMPL_HAVE_STATUS_SET_BYTES 1
#define MPIIMPL_HAVE_STATUS_SET_INFO 1

/* MPI I/O */
#define MPI_MODE_RDONLY			  2  /* ADIO_RDONLY */
#define MPI_MODE_RDWR				8  /* ADIO_RDWR  */
#define MPI_MODE_WRONLY			  4  /* ADIO_WRONLY  */
#define MPI_MODE_CREATE			  1  /* ADIO_CREATE */ 
#define MPI_MODE_EXCL			   64  /* ADIO_EXCL */
#define MPI_MODE_DELETE_ON_CLOSE	16  /* ADIO_DELETE_ON_CLOSE */
#define MPI_MODE_UNIQUE_OPEN		32  /* ADIO_UNIQUE_OPEN */
#define MPI_MODE_APPEND			128  /* ADIO_APPEND */
#define MPI_MODE_SEQUENTIAL		256  /* ADIO_SEQUENTIAL */

#define MPI_DISPLACEMENT_CURRENT   -54278278

#define MPI_SEEK_SET			600
#define MPI_SEEK_CUR			602
#define MPI_SEEK_END			604

#define MPI_MAX_DATAREP_STRING  128

/* MPI-IO function prototypes */

/* The compiler must support ANSI C style prototypes, otherwise 
   long long constants (e.g. 0) may get passed as ints. */

/* Section 9.2 */
/* Begin Prototypes */
MPI_DLL_SPEC int MPI_CALL MPI_File_openA(MPI_Comm, char *, int, MPI_Info, MPI_File *);
MPI_DLL_SPEC int MPI_CALL MPI_File_openW(MPI_Comm, wchar_t *, int, MPI_Info, MPI_File *);
MPI_DLL_SPEC int MPI_CALL MPI_File_close(MPI_File *);
MPI_DLL_SPEC int MPI_CALL MPI_File_deleteA(char *, MPI_Info);
MPI_DLL_SPEC int MPI_CALL MPI_File_deleteW(wchar_t *, MPI_Info);
MPI_DLL_SPEC int MPI_CALL MPI_File_set_size(MPI_File, MPI_Offset);
MPI_DLL_SPEC int MPI_CALL MPI_File_preallocate(MPI_File, MPI_Offset);
MPI_DLL_SPEC int MPI_CALL MPI_File_get_size(MPI_File, MPI_Offset *);
MPI_DLL_SPEC int MPI_CALL MPI_File_get_group(MPI_File, MPI_Group *);
MPI_DLL_SPEC int MPI_CALL MPI_File_get_amode(MPI_File, int *);
MPI_DLL_SPEC int MPI_CALL MPI_File_set_info(MPI_File, MPI_Info);
MPI_DLL_SPEC int MPI_CALL MPI_File_get_info(MPI_File, MPI_Info *);

/* Section 9.3 */
MPI_DLL_SPEC int MPI_CALL MPI_File_set_viewA(MPI_File, MPI_Offset, MPI_Datatype, MPI_Datatype, char *, MPI_Info);
MPI_DLL_SPEC int MPI_CALL MPI_File_set_viewW(MPI_File, MPI_Offset, MPI_Datatype, MPI_Datatype, wchar_t *, MPI_Info);
MPI_DLL_SPEC int MPI_CALL MPI_File_get_viewA(MPI_File, MPI_Offset *, MPI_Datatype *, MPI_Datatype *, char *);
MPI_DLL_SPEC int MPI_CALL MPI_File_get_viewW(MPI_File, MPI_Offset *, MPI_Datatype *, MPI_Datatype *, wchar_t *);

/* Section 9.4.2 */
MPI_DLL_SPEC int MPI_CALL MPI_File_read_at(MPI_File, MPI_Offset, void *,
			  int, MPI_Datatype, MPI_Status *);
MPI_DLL_SPEC int MPI_CALL MPI_File_read_at_all(MPI_File, MPI_Offset, void *,
			  int, MPI_Datatype, MPI_Status *);
MPI_DLL_SPEC int MPI_CALL MPI_File_write_at(MPI_File, MPI_Offset, void *,
			  int, MPI_Datatype, MPI_Status *);
MPI_DLL_SPEC int MPI_CALL MPI_File_write_at_all(MPI_File, MPI_Offset, void *,
			  int, MPI_Datatype, MPI_Status *);

MPI_DLL_SPEC int MPI_CALL MPI_File_iread_at(MPI_File, MPI_Offset, void *,
			  int, MPI_Datatype, MPI_Request *);
MPI_DLL_SPEC int MPI_CALL MPI_File_iwrite_at(MPI_File, MPI_Offset, void *,
			  int, MPI_Datatype, MPI_Request *);

/* Section 9.4.3 */
MPI_DLL_SPEC int MPI_CALL MPI_File_read(MPI_File, void *, int, MPI_Datatype, MPI_Status *); 
MPI_DLL_SPEC int MPI_CALL MPI_File_read_all(MPI_File, void *, int, MPI_Datatype, MPI_Status *); 
MPI_DLL_SPEC int MPI_CALL MPI_File_write(MPI_File, void *, int, MPI_Datatype, MPI_Status *);
MPI_DLL_SPEC int MPI_CALL MPI_File_write_all(MPI_File, void *, int, MPI_Datatype, MPI_Status *);

MPI_DLL_SPEC int MPI_CALL MPI_File_iread(MPI_File, void *, int, MPI_Datatype, MPI_Request *); 
MPI_DLL_SPEC int MPI_CALL MPI_File_iwrite(MPI_File, void *, int, MPI_Datatype, MPI_Request *);

MPI_DLL_SPEC int MPI_CALL MPI_File_seek(MPI_File, MPI_Offset, int);
MPI_DLL_SPEC int MPI_CALL MPI_File_get_position(MPI_File, MPI_Offset *);
MPI_DLL_SPEC int MPI_CALL MPI_File_get_byte_offset(MPI_File, MPI_Offset, MPI_Offset *);

/* Section 9.4.4 */
MPI_DLL_SPEC int MPI_CALL MPI_File_read_shared(MPI_File, void *, int, MPI_Datatype, MPI_Status *);
MPI_DLL_SPEC int MPI_CALL MPI_File_write_shared(MPI_File, void *, int, MPI_Datatype, MPI_Status *);
MPI_DLL_SPEC int MPI_CALL MPI_File_iread_shared(MPI_File, void *, int, MPI_Datatype, MPI_Request *);
MPI_DLL_SPEC int MPI_CALL MPI_File_iwrite_shared(MPI_File, void *, int, 
						   MPI_Datatype, MPI_Request *);
MPI_DLL_SPEC int MPI_CALL MPI_File_read_ordered(MPI_File, void *, int, 
						  MPI_Datatype, MPI_Status *);
MPI_DLL_SPEC int MPI_CALL MPI_File_write_ordered(MPI_File, void *, int, 
						   MPI_Datatype, MPI_Status *);
MPI_DLL_SPEC int MPI_CALL MPI_File_seek_shared(MPI_File, MPI_Offset, int);
MPI_DLL_SPEC int MPI_CALL MPI_File_get_position_shared(MPI_File, MPI_Offset *);

/* Section 9.4.5 */
MPI_DLL_SPEC int MPI_CALL MPI_File_read_at_all_begin(MPI_File, MPI_Offset, void *,
							   int, MPI_Datatype);
MPI_DLL_SPEC int MPI_CALL MPI_File_read_at_all_end(MPI_File, void *, MPI_Status *);
MPI_DLL_SPEC int MPI_CALL MPI_File_write_at_all_begin(MPI_File, MPI_Offset, void *,
								int, MPI_Datatype);
MPI_DLL_SPEC int MPI_CALL MPI_File_write_at_all_end(MPI_File, void *, MPI_Status *);
MPI_DLL_SPEC int MPI_CALL MPI_File_read_all_begin(MPI_File, void *, int, MPI_Datatype);
MPI_DLL_SPEC int MPI_CALL MPI_File_read_all_end(MPI_File, void *, MPI_Status *);
MPI_DLL_SPEC int MPI_CALL MPI_File_write_all_begin(MPI_File, void *, int, MPI_Datatype);
MPI_DLL_SPEC int MPI_CALL MPI_File_write_all_end(MPI_File, void *, MPI_Status *);
MPI_DLL_SPEC int MPI_CALL MPI_File_read_ordered_begin(MPI_File, void *, int, MPI_Datatype);
MPI_DLL_SPEC int MPI_CALL MPI_File_read_ordered_end(MPI_File, void *, MPI_Status *);
MPI_DLL_SPEC int MPI_CALL MPI_File_write_ordered_begin(MPI_File, void *, int, MPI_Datatype);
MPI_DLL_SPEC int MPI_CALL MPI_File_write_ordered_end(MPI_File, void *, MPI_Status *);

/* Section 9.5.1 */
MPI_DLL_SPEC int MPI_CALL MPI_File_get_type_extent(MPI_File, MPI_Datatype, MPI_Aint *);

/* Section 9.5.3 */
MPI_DLL_SPEC int MPI_CALL MPI_Register_datarepA(char *,
						 MPI_Datarep_conversion_function *,
						 MPI_Datarep_conversion_function *,
						 MPI_Datarep_extent_function *,
						 void *);
MPI_DLL_SPEC int MPI_CALL MPI_Register_datarepW(wchar_t *,
						 MPI_Datarep_conversion_function *,
						 MPI_Datarep_conversion_function *,
						 MPI_Datarep_extent_function *,
						 void *);

/* Section 9.6.1 */
MPI_DLL_SPEC int MPI_CALL MPI_File_set_atomicity(MPI_File, int);
MPI_DLL_SPEC int MPI_CALL MPI_File_get_atomicity(MPI_File, int *);
MPI_DLL_SPEC int MPI_CALL MPI_File_sync(MPI_File);

/**************** BINDINGS FOR THE PROFILING INTERFACE ***************/


/* Section 9.2 */
MPI_DLL_SPEC int MPI_CALL PMPI_File_openA(MPI_Comm, char *, int, MPI_Info, MPI_File *);
MPI_DLL_SPEC int MPI_CALL PMPI_File_openW(MPI_Comm, wchar_t *, int, MPI_Info, MPI_File *);
MPI_DLL_SPEC int MPI_CALL PMPI_File_close(MPI_File *);
MPI_DLL_SPEC int MPI_CALL PMPI_File_deleteA(char *, MPI_Info);
MPI_DLL_SPEC int MPI_CALL PMPI_File_deleteW(wchar_t *, MPI_Info);
MPI_DLL_SPEC int MPI_CALL PMPI_File_set_size(MPI_File, MPI_Offset);
MPI_DLL_SPEC int MPI_CALL PMPI_File_preallocate(MPI_File, MPI_Offset);
MPI_DLL_SPEC int MPI_CALL PMPI_File_get_size(MPI_File, MPI_Offset *);
MPI_DLL_SPEC int MPI_CALL PMPI_File_get_group(MPI_File, MPI_Group *);
MPI_DLL_SPEC int MPI_CALL PMPI_File_get_amode(MPI_File, int *);
MPI_DLL_SPEC int MPI_CALL PMPI_File_set_info(MPI_File, MPI_Info);
MPI_DLL_SPEC int MPI_CALL PMPI_File_get_info(MPI_File, MPI_Info *);

/* Section 9.3 */
MPI_DLL_SPEC int MPI_CALL PMPI_File_set_viewA(MPI_File, MPI_Offset, MPI_Datatype, MPI_Datatype, char *, MPI_Info);
MPI_DLL_SPEC int MPI_CALL PMPI_File_set_viewW(MPI_File, MPI_Offset, MPI_Datatype, MPI_Datatype, wchar_t *, MPI_Info);
MPI_DLL_SPEC int MPI_CALL PMPI_File_get_viewA(MPI_File, MPI_Offset *, MPI_Datatype *, MPI_Datatype *, char *);
MPI_DLL_SPEC int MPI_CALL PMPI_File_get_viewW(MPI_File, MPI_Offset *, MPI_Datatype *, MPI_Datatype *, wchar_t *);

/* Section 9.4.2 */
MPI_DLL_SPEC int MPI_CALL PMPI_File_read_at(MPI_File, MPI_Offset, void *,
			  int, MPI_Datatype, MPI_Status *);
MPI_DLL_SPEC int MPI_CALL PMPI_File_read_at_all(MPI_File, MPI_Offset, void *,
			  int, MPI_Datatype, MPI_Status *);
MPI_DLL_SPEC int MPI_CALL PMPI_File_write_at(MPI_File, MPI_Offset, void *,
			  int, MPI_Datatype, MPI_Status *);
MPI_DLL_SPEC int MPI_CALL PMPI_File_write_at_all(MPI_File, MPI_Offset, void *,
			  int, MPI_Datatype, MPI_Status *);

MPI_DLL_SPEC int MPI_CALL PMPI_File_iread_at(MPI_File, MPI_Offset, void *,
			  int, MPI_Datatype, MPI_Request *);
MPI_DLL_SPEC int MPI_CALL PMPI_File_iwrite_at(MPI_File, MPI_Offset, void *,
			  int, MPI_Datatype, MPI_Request *);

/* Section 9.4.3 */
MPI_DLL_SPEC int MPI_CALL PMPI_File_read(MPI_File, void *, int, MPI_Datatype, MPI_Status *); 
MPI_DLL_SPEC int MPI_CALL PMPI_File_read_all(MPI_File, void *, int, MPI_Datatype, MPI_Status *); 
MPI_DLL_SPEC int MPI_CALL PMPI_File_write(MPI_File, void *, int, MPI_Datatype, MPI_Status *);
MPI_DLL_SPEC int MPI_CALL PMPI_File_write_all(MPI_File, void *, int, MPI_Datatype, MPI_Status *);

MPI_DLL_SPEC int MPI_CALL PMPI_File_iread(MPI_File, void *, int, MPI_Datatype, MPI_Request *); 
MPI_DLL_SPEC int MPI_CALL PMPI_File_iwrite(MPI_File, void *, int, MPI_Datatype, MPI_Request *);

MPI_DLL_SPEC int MPI_CALL PMPI_File_seek(MPI_File, MPI_Offset, int);
MPI_DLL_SPEC int MPI_CALL PMPI_File_get_position(MPI_File, MPI_Offset *);
MPI_DLL_SPEC int MPI_CALL PMPI_File_get_byte_offset(MPI_File, MPI_Offset, MPI_Offset *);

/* Section 9.4.4 */
MPI_DLL_SPEC int MPI_CALL PMPI_File_read_shared(MPI_File, void *, int, MPI_Datatype, MPI_Status *);
MPI_DLL_SPEC int MPI_CALL PMPI_File_write_shared(MPI_File, void *, int, MPI_Datatype, MPI_Status *);
MPI_DLL_SPEC int MPI_CALL PMPI_File_iread_shared(MPI_File, void *, int, 
						   MPI_Datatype, MPI_Request *);
MPI_DLL_SPEC int MPI_CALL PMPI_File_iwrite_shared(MPI_File, void *, int, 
							MPI_Datatype, MPI_Request *);
MPI_DLL_SPEC int MPI_CALL PMPI_File_read_ordered(MPI_File, void *, int, MPI_Datatype, MPI_Status *);
MPI_DLL_SPEC int MPI_CALL PMPI_File_write_ordered(MPI_File, void *, int, MPI_Datatype, MPI_Status *);
MPI_DLL_SPEC int MPI_CALL PMPI_File_seek_shared(MPI_File, MPI_Offset, int);
MPI_DLL_SPEC int MPI_CALL PMPI_File_get_position_shared(MPI_File, MPI_Offset *);

/* Section 9.4.5 */
MPI_DLL_SPEC int MPI_CALL PMPI_File_read_at_all_begin(MPI_File, MPI_Offset, void *,
							   int, MPI_Datatype);
MPI_DLL_SPEC int MPI_CALL PMPI_File_read_at_all_end(MPI_File, void *, MPI_Status *);
MPI_DLL_SPEC int MPI_CALL PMPI_File_write_at_all_begin(MPI_File, MPI_Offset, void *,
								int, MPI_Datatype);
MPI_DLL_SPEC int MPI_CALL PMPI_File_write_at_all_end(MPI_File, void *, MPI_Status *);
MPI_DLL_SPEC int MPI_CALL PMPI_File_read_all_begin(MPI_File, void *, int, MPI_Datatype);
MPI_DLL_SPEC int MPI_CALL PMPI_File_read_all_end(MPI_File, void *, MPI_Status *);
MPI_DLL_SPEC int MPI_CALL PMPI_File_write_all_begin(MPI_File, void *, int, MPI_Datatype);
MPI_DLL_SPEC int MPI_CALL PMPI_File_write_all_end(MPI_File, void *, MPI_Status *);
MPI_DLL_SPEC int MPI_CALL PMPI_File_read_ordered_begin(MPI_File, void *, int, MPI_Datatype);
MPI_DLL_SPEC int MPI_CALL PMPI_File_read_ordered_end(MPI_File, void *, MPI_Status *);
MPI_DLL_SPEC int MPI_CALL PMPI_File_write_ordered_begin(MPI_File, void *, int, MPI_Datatype);
MPI_DLL_SPEC int MPI_CALL PMPI_File_write_ordered_end(MPI_File, void *, MPI_Status *);

/* Section 9.5.1 */
MPI_DLL_SPEC int MPI_CALL PMPI_File_get_type_extent(MPI_File, MPI_Datatype, MPI_Aint *);

/* Section 9.5.3 */
MPI_DLL_SPEC int MPI_CALL PMPI_Register_datarepA(char *,
						 MPI_Datarep_conversion_function *,
						 MPI_Datarep_conversion_function *,
						 MPI_Datarep_extent_function *,
						 void *);
MPI_DLL_SPEC int MPI_CALL PMPI_Register_datarepW(wchar_t *,
						 MPI_Datarep_conversion_function *,
						 MPI_Datarep_conversion_function *,
						 MPI_Datarep_extent_function *,
						 void *);

/* Section 9.6.1 */
MPI_DLL_SPEC int MPI_CALL PMPI_File_set_atomicity(MPI_File, int);
MPI_DLL_SPEC int MPI_CALL PMPI_File_get_atomicity(MPI_File, int *);
MPI_DLL_SPEC int MPI_CALL PMPI_File_sync(MPI_File);

#if defined(__cplusplus)
}
/* Add the C++ bindings */
/* 
   If MPICH_SKIP_MPICXX is defined, the mpicxx.h file will *not* be included.
   This is necessary, for example, when building the C++ interfaces.  It
   can also be used when you want to use a C++ compiler to compile C code,
   and do not want to load the C++ bindings.  These definitions can
   be made by the C++ compilation script
 */
#if !defined(MPICH_SKIP_MPICXX)
/* mpicxx.h contains the MPI C++ binding.  In the mpi.h.in file, this 
   include is in an autoconf variable in case the compiler is a C++ 
   compiler but MPI was built without the C++ bindings */
#include "mpicxx.h"
#endif 
#endif

#endif
