/*   Copyright (C) 2009 Deino Software. */

#include "ad_ntfs.h"

ADIOI_Fns ADIO_NTFS_operations =
{
	ADIOI_NTFS_Open, /* Open */
	ADIOI_NTFS_ReadContig, /* ReadContig */
	ADIOI_NTFS_WriteContig, /* WriteContig */
	ADIOI_GEN_ReadStridedColl, /* ReadStridedColl */
	ADIOI_GEN_WriteStridedColl, /* WriteStridedColl */
	ADIOI_GEN_SeekIndividual, /* SeekIndividual */
	ADIOI_NTFS_Fcntl, /* Fcntl */
	ADIOI_GEN_SetInfo, /* SetInfo */
	ADIOI_GEN_ReadStrided, /* ReadStrided */
	ADIOI_GEN_WriteStrided, /* WriteStrided */
	ADIOI_NTFS_Close, /* Close */
	ADIOI_NTFS_IreadContig, /* IreadContig */
	ADIOI_NTFS_IwriteContig, /* IwriteContig */
	ADIOI_NTFS_ReadDone, /* ReadDone */
	ADIOI_NTFS_WriteDone, /* WriteDone */
	ADIOI_NTFS_ReadComplete, /* ReadComplete */
	ADIOI_NTFS_WriteComplete, /* WriteComplete */
	ADIOI_FAKE_IreadStrided, /* IreadStrided */
	ADIOI_FAKE_IwriteStrided, /* IwriteStrided */
	ADIOI_NTFS_Flush, /* Flush */
	ADIOI_NTFS_Resize, /* Resize */
	ADIOI_GEN_Delete, /* Delete */
};

#undef FUNCNAME
#define FUNCNAME ADIOI_NTFS_WriteContig
int ADIOI_NTFS_WriteContig(MPID_File_ptr fd, void *buf, int count, MPI_Datatype datatype, int file_ptr_type, MPI_Offset offset, MPI_Status *status)
{
	static const wchar_t FCNAME[] = L"ADIOI_NTFS_WriteContig";
	int mpi_errno = MPI_SUCCESS;
	LONG dwTemp;
	DWORD dwNumWritten = 0;
	int err=-1, datatype_size, len;
	OVERLAPPED *pOvl = NULL;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_NTFS_WRITECONTIG);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_NTFS_WRITECONTIG);

	NMPI_Type_size(mpi_errno, (datatype, &datatype_size));
	MPIU_ERR_CHK_POP(mpi_errno);
	len = datatype_size * count;

	pOvl = (OVERLAPPED *) MPIU_Calloc(sizeof(OVERLAPPED), 1);
	MPIU_ERR_MEM_CHK_POP(mpi_errno, pOvl, sizeof(OVERLAPPED), L"pOvl");
	pOvl->hEvent = CreateEvent(NULL, TRUE, TRUE, NULL);
	if (pOvl->hEvent == NULL)
	{
		err = GetLastError();
		MPIU_ERR_SET1(mpi_errno, MPI_ERR_IO, L"**io", L"**io %s", ADIOI_NTFS_Strerror(err));
		goto fn_fail;
	}
	pOvl->Offset = DWORDLOW(offset);
	pOvl->OffsetHigh = DWORDHIGH(offset);

	if (file_ptr_type == ADIO_EXPLICIT_OFFSET)
	{
		if (fd->fp_sys_posn != offset)
		{
			dwTemp = DWORDHIGH(offset);
			if (SetFilePointer(fd->fd_sys, DWORDLOW(offset), &dwTemp, FILE_BEGIN) == INVALID_SET_FILE_POINTER)
			{
				err = GetLastError();
				if (err != NO_ERROR)
				{
					MPIU_ERR_SET1(mpi_errno, MPI_ERR_IO, L"**io", L"**io %s", ADIOI_NTFS_Strerror(err));
					goto fn_fail;
				}
			}
		}
		/*wprintf(L"WriteFile(%d bytes)\n", len);fflush(stdout);*/
		err = WriteFile(fd->fd_sys, buf, len, &dwNumWritten, pOvl);
		/* --BEGIN ERROR HANDLING-- */
		if (err == FALSE)
		{
			err = GetLastError();
			if (err != ERROR_IO_PENDING)
			{
				MPIU_ERR_SET1(mpi_errno, MPI_ERR_IO, L"**io", L"**io %s", ADIOI_NTFS_Strerror(err));
				goto fn_fail;
			}
		}
		/* --END ERROR HANDLING-- */
		err = GetOverlappedResult(fd->fd_sys, pOvl, &dwNumWritten, TRUE);
		/* --BEGIN ERROR HANDLING-- */
		if (err == FALSE)
		{
			err = GetLastError();
			MPIU_ERR_SET1(mpi_errno, MPI_ERR_IO, L"**io", L"**io %s", ADIOI_NTFS_Strerror(err));
			goto fn_fail;
		}
		/* --END ERROR HANDLING-- */
		if (!CloseHandle(pOvl->hEvent))
		{
			err = GetLastError();
			MPIU_ERR_SET1(mpi_errno, MPI_ERR_IO, L"**io", L"**io %s", ADIOI_NTFS_Strerror(err));
			pOvl->hEvent = NULL;
			goto fn_fail;
		}
		MPIU_Free(pOvl);
		pOvl = NULL;

		fd->fp_sys_posn = offset + dwNumWritten;
		/* individual file pointer not updated */		
	}
	else
	{
		/* write from curr. location of ind. file pointer */
		if (fd->fp_sys_posn != fd->fp_ind)
		{
			dwTemp = DWORDHIGH(fd->fp_ind);
			if (SetFilePointer(fd->fd_sys, DWORDLOW(fd->fp_ind), &dwTemp, FILE_BEGIN) == INVALID_SET_FILE_POINTER)
			{
				err = GetLastError();
				if (err != NO_ERROR)
				{
					MPIU_ERR_SET1(mpi_errno, MPI_ERR_IO, L"**io", L"**io %s", ADIOI_NTFS_Strerror(err));
					goto fn_fail;
				}
			}
		}
		/* Ignore the offset parameter and write from the current file position */
		pOvl->Offset = DWORDLOW(fd->fp_ind);
		pOvl->OffsetHigh = DWORDHIGH(fd->fp_ind);
		/*wprintf(L"WriteFile(%d bytes)\n", len);fflush(stdout);*/
		err = WriteFile(fd->fd_sys, buf, len, &dwNumWritten, pOvl);
		/* --BEGIN ERROR HANDLING-- */
		if (err == FALSE)
		{
			err = GetLastError();
			if (err != ERROR_IO_PENDING)
			{
				MPIU_ERR_SET1(mpi_errno, MPI_ERR_IO, L"**io", L"**io %s", ADIOI_NTFS_Strerror(err));
				goto fn_fail;
			}
		}
		/* --END ERROR HANDLING-- */
		err = GetOverlappedResult(fd->fd_sys, pOvl, &dwNumWritten, TRUE);
		/* --BEGIN ERROR HANDLING-- */
		if (err == FALSE)
		{
			err = GetLastError();
			MPIU_ERR_SET1(mpi_errno, MPI_ERR_IO, L"**io", L"**io %s", ADIOI_NTFS_Strerror(err));
			goto fn_fail;
		}
		/* --END ERROR HANDLING-- */
		if (!CloseHandle(pOvl->hEvent))
		{
			err = GetLastError();
			MPIU_ERR_SET1(mpi_errno, MPI_ERR_IO, L"**io", L"**io %s", ADIOI_NTFS_Strerror(err));
			pOvl->hEvent = NULL;
			goto fn_fail;
		}
		MPIU_Free(pOvl);
		pOvl = NULL;

		fd->fp_ind = fd->fp_ind + dwNumWritten;
		fd->fp_sys_posn = fd->fp_ind;
	}

#ifdef HAVE_STATUS_SET_BYTES
	if (err != FALSE)
	{
		MPIR_Status_set_bytes(status, datatype, dwNumWritten);
	}
#endif

	/* --BEGIN ERROR HANDLING-- */
	if (err == FALSE)
	{
		err = GetLastError();
		MPIU_ERR_SET1(mpi_errno, MPI_ERR_IO, L"**io", L"**io %s", ADIOI_NTFS_Strerror(err));
		goto fn_fail;
	}
	/* --END ERROR HANDLING-- */

fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_NTFS_WRITECONTIG);
	return mpi_errno;

fn_fail:
	if (pOvl != NULL)
	{
		if (pOvl->hEvent != NULL)
		{
			CloseHandle(pOvl->hEvent);
		}
		MPIU_Free(pOvl);
	}
	goto fn_exit;
}

#undef FUNCNAME
#define FUNCNAME ADIOI_NTFS_ReadComplete
int ADIOI_NTFS_ReadComplete(ADIO_Request *request, MPI_Status *status)
{
	static const wchar_t FCNAME[] = L"ADIOI_NTFS_ReadComplete";
	int mpi_errno = MPI_SUCCESS;
	DWORD ret_val;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_NTFS_READCOMPLETE);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_NTFS_READCOMPLETE);

	if (*request == NULL)
	{
		goto fn_exit;
	}
	
	if ((*request)->queued)
	{
		ret_val = GetOverlappedResult((*request)->fd, (*request)->handle, &(*request)->nbytes, TRUE);

		if (!ret_val)
		{
			(*request)->nbytes = (DWORD)-1;
		}

		/* --BEGIN ERROR HANDLING-- */
		if (ret_val == FALSE)
		{
			ret_val = GetLastError();
			MPIU_ERR_SET1(mpi_errno, MPI_ERR_IO, L"**io", L"**io %s", ADIOI_NTFS_Strerror(ret_val));
			goto fn_fail;
		}
		/* --END ERROR HANDLING-- */
	} /* if ((*request)->queued) ... */

#ifdef HAVE_STATUS_SET_BYTES
	if ((*request)->nbytes != -1)
	{
		MPIR_Status_set_bytes(status, (*request)->datatype, (*request)->nbytes);
	}
#endif

	if ((*request)->queued != -1)
	{
		/* queued = -1 is an internal hack used when the request must
		   be completed, but the request object should not be
		   freed. This is used in ADIOI_Complete_async, because the user
		   will call MPI_Wait later, which would require status to
		   be filled. Ugly but works. queued = -1 should be used only
		   in ADIOI_Complete_async. 
		   This should not affect the user in any way. */

		/* if request is still queued in the system, it is also there
		   on ADIOI_Async_list. Delete it from there. */
		if ((*request)->queued) ADIOI_Del_req_from_list(request);

		(*request)->fd->async_count--;
		if ((*request)->handle)
		{
			if ((((OVERLAPPED*)((*request)->handle))->hEvent) != NULL)
			{
				if (!CloseHandle(((OVERLAPPED*)((*request)->handle))->hEvent))
				{
					ret_val = GetLastError();
					MPIU_ERR_SET1(mpi_errno, MPI_ERR_IO, L"**io", L"**io %s", ADIOI_NTFS_Strerror(ret_val));
					MPIU_Free((*request)->handle);
					goto fn_fail;
				}
			}
			MPIU_Free((*request)->handle);
		}
		ADIOI_Free_request((ADIOI_Req_node *) (*request));
		*request = NULL;
	}

fn_exit:
fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_NTFS_READCOMPLETE);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME ADIOI_NTFS_WriteComplete
int ADIOI_NTFS_WriteComplete(ADIO_Request *request, MPI_Status *status)
{
	static const wchar_t FCNAME[] = L"ADIOI_NTFS_WriteComplete";
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_NTFS_WRITECOMPLETE);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_NTFS_WRITECOMPLETE);

	mpi_errno = ADIOI_NTFS_ReadComplete(request, status);
	MPIU_ERR_CHK_POP(mpi_errno);

fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_NTFS_WRITECOMPLETE);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME ADIOI_NTFS_Resize
int ADIOI_NTFS_Resize(MPID_File_ptr fd, MPI_Offset size)
{
	static const wchar_t FCNAME[] = L"ADIOI_NTFS_Resize";
	int mpi_errno = MPI_SUCCESS;
	LONG dwTemp;
	DWORD err;
	BOOL result;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_NTFS_RESIZE);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_NTFS_RESIZE);

	dwTemp = DWORDHIGH(size);
	err = SetFilePointer(fd->fd_sys, DWORDLOW(size), &dwTemp, FILE_BEGIN);
	/* --BEGIN ERROR HANDLING-- */
	if (err == INVALID_SET_FILE_POINTER)
	{
		err = GetLastError();
		if (err != NO_ERROR)
		{
			MPIU_ERR_SET1(mpi_errno, MPI_ERR_IO, L"**io", L"**io %s", ADIOI_NTFS_Strerror(err));
			goto fn_fail;
		}
	}
	/*wprintf(L"setting file length to %d\n", size);fflush(stdout);*/
	/* --END ERROR HANDLING-- */
	result = SetEndOfFile(fd->fd_sys);
	/* --BEGIN ERROR HANDLING-- */
	if (result == FALSE)
	{
		err = GetLastError();
		MPIU_ERR_SET1(mpi_errno, MPI_ERR_IO, L"**io", L"**io %s", ADIOI_NTFS_Strerror(err));
		goto fn_fail;
	}
	/* --END ERROR HANDLING-- */

fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_NTFS_RESIZE);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME ADIOI_NTFS_ReadContig
int ADIOI_NTFS_ReadContig(MPID_File_ptr fd, void *buf, int count, MPI_Datatype datatype, int file_ptr_type, MPI_Offset offset, MPI_Status *status)
{
	static const wchar_t FCNAME[] = L"ADIOI_NTFS_ReadContig";
	int mpi_errno = MPI_SUCCESS;
	LONG dwTemp;
	DWORD dwNumRead = 0;
	int err=-1, datatype_size, len;
	OVERLAPPED *pOvl = NULL;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_NTFS_READCONTIG);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_NTFS_READCONTIG);

	NMPI_Type_size(mpi_errno, (datatype, &datatype_size));
	MPIU_ERR_CHK_POP(mpi_errno);
	len = datatype_size * count;

	pOvl = (OVERLAPPED *) MPIU_Calloc(sizeof(OVERLAPPED), 1);
	MPIU_ERR_MEM_CHK_POP(mpi_errno, pOvl, sizeof(OVERLAPPED), L"pOvl");
	pOvl->hEvent = CreateEvent(NULL, TRUE, TRUE, NULL);
	if (pOvl->hEvent == NULL)
	{
		err = GetLastError();
		MPIU_ERR_SET1(mpi_errno, MPI_ERR_IO, L"**io", L"**io %s", ADIOI_NTFS_Strerror(err));
		goto fn_fail;
	}
	pOvl->Offset = DWORDLOW(offset);
	pOvl->OffsetHigh = DWORDHIGH(offset);

	if (file_ptr_type == ADIO_EXPLICIT_OFFSET)
	{
		if (fd->fp_sys_posn != offset)
		{
			dwTemp = DWORDHIGH(offset);
			if (SetFilePointer(fd->fd_sys, DWORDLOW(offset), &dwTemp, FILE_BEGIN) == INVALID_SET_FILE_POINTER)
			{
				err = GetLastError();
				if (err != NO_ERROR)
				{
					MPIU_ERR_SET1(mpi_errno, MPI_ERR_IO, L"**io", L"**io %s", ADIOI_NTFS_Strerror(err));
					goto fn_fail;
				}
			}
		}
		/*
		{
			ADIO_Fcntl_t fcntl_struct;
			int error_code;
			ADIO_Fcntl(fd, ADIO_FCNTL_GET_FSIZE, &fcntl_struct, &error_code);
			wprintf(L"File size b: %d\n", fcntl_struct.fsize);
		}
		wprintf(L"ReadFile(%d bytes)\n", len);fflush(stdout);
		*/
		err = ReadFile(fd->fd_sys, buf, len, &dwNumRead, pOvl);
		/* --BEGIN ERROR HANDLING-- */
		if (err == FALSE)
		{
			err = GetLastError();
			switch (err)
			{
			case ERROR_IO_PENDING:
				break;
			case ERROR_HANDLE_EOF:
				/*wprintf(L"EOF error\n");fflush(stdout);*/
				SetEvent(pOvl->hEvent);
				break;
			default:
				MPIU_ERR_SET1(mpi_errno, MPI_ERR_IO,
					L"**io",
					L"**io %s", ADIOI_NTFS_Strerror(err));
				goto fn_fail;
			}
		}
		/* --END ERROR HANDLING-- */
		err = GetOverlappedResult(fd->fd_sys, pOvl, &dwNumRead, TRUE);
		/* --BEGIN ERROR HANDLING-- */
		if (err == FALSE)
		{
			err = GetLastError();
			if (err != ERROR_HANDLE_EOF) /* Ignore EOF errors */
			{
				MPIU_ERR_SET1(mpi_errno, MPI_ERR_IO, L"**io", L"**io %s", ADIOI_NTFS_Strerror(err));
				goto fn_fail;
			}
		}
		/* --END ERROR HANDLING-- */
		if (!CloseHandle(pOvl->hEvent))
		{
			err = GetLastError();
			MPIU_ERR_SET1(mpi_errno, MPI_ERR_IO, L"**io", L"**io %s", ADIOI_NTFS_Strerror(err));
			goto fn_fail;
		}
		MPIU_Free(pOvl);
		pOvl = NULL;

		fd->fp_sys_posn = offset + (MPI_Offset)dwNumRead;
		/* individual file pointer not updated */		
	}
	else
	{
		/* read from curr. location of ind. file pointer */
		if (fd->fp_sys_posn != fd->fp_ind)
		{
			dwTemp = DWORDHIGH(fd->fp_ind);
			if (SetFilePointer(fd->fd_sys, DWORDLOW(fd->fp_ind), &dwTemp, FILE_BEGIN) == INVALID_SET_FILE_POINTER)
			{
				err = GetLastError();
				if (err != NO_ERROR)
				{
					MPIU_ERR_SET1(mpi_errno, MPI_ERR_IO, L"**io", L"**io %s", ADIOI_NTFS_Strerror(err));
					goto fn_fail;
				}
			}
		}
		/*
		{
			ADIO_Fcntl_t fcntl_struct;
			int error_code;
			ADIO_Fcntl(fd, ADIO_FCNTL_GET_FSIZE, &fcntl_struct, &error_code);
			wprintf(L"File size c: %d\n", fcntl_struct.fsize);
		}
		wprintf(L"ReadFile(%d bytes)\n", len);fflush(stdout);
		*/
		/* Ignore the offset parameter and read from the current file position */
		pOvl->Offset = DWORDLOW(fd->fp_ind);
		pOvl->OffsetHigh = DWORDHIGH(fd->fp_ind);
		err = ReadFile(fd->fd_sys, buf, len, &dwNumRead, pOvl);
		/* --BEGIN ERROR HANDLING-- */
		if (err == FALSE)
		{
			err = GetLastError();
			switch (err)
			{
			case ERROR_IO_PENDING:
				break;
			case ERROR_HANDLE_EOF:
				/*wprintf(L"EOF error\n");fflush(stdout);*/
				SetEvent(pOvl->hEvent);
				break;
			default:
				MPIU_ERR_SET1(mpi_errno, MPI_ERR_IO, L"**io", L"**io %s", ADIOI_NTFS_Strerror(err));
				goto fn_fail;
			}
		}
		/* --END ERROR HANDLING-- */
		err = GetOverlappedResult(fd->fd_sys, pOvl, &dwNumRead, TRUE);
		/* --BEGIN ERROR HANDLING-- */
		if (err == FALSE)
		{
			err = GetLastError();
			if (err != ERROR_HANDLE_EOF) /* Ignore EOF errors */
			{
				MPIU_ERR_SET1(mpi_errno, MPI_ERR_IO, L"**io", L"**io %s", ADIOI_NTFS_Strerror(err));
				goto fn_fail;
			}
		}
		/* --END ERROR HANDLING-- */
		if (!CloseHandle(pOvl->hEvent))
		{
			err = GetLastError();
			pOvl->hEvent = NULL;
			MPIU_ERR_SET1(mpi_errno, MPI_ERR_IO, L"**io", L"**io %s", ADIOI_NTFS_Strerror(err));
			goto fn_fail;
		}
		MPIU_Free(pOvl);
		pOvl = NULL;

		fd->fp_ind = fd->fp_ind + (MPI_Offset)dwNumRead; 
		fd->fp_sys_posn = fd->fp_ind;
	}		 

#ifdef HAVE_STATUS_SET_BYTES
	if (err != FALSE)
	{
		MPIR_Status_set_bytes(status, datatype, dwNumRead);
	}
#endif

	/* --BEGIN ERROR HANDLING-- */
	if (err == FALSE)
	{
		err = GetLastError();
		MPIU_ERR_SET1(mpi_errno, MPI_ERR_IO, L"**io", L"**io %s", ADIOI_NTFS_Strerror(err));
		goto fn_fail;
	}
	/* --END ERROR HANDLING-- */

fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_NTFS_READCONTIG);
	return mpi_errno;

fn_fail:
	if (pOvl != NULL)
	{
		if (pOvl->hEvent != NULL)
		{
			CloseHandle(pOvl->hEvent);
		}
		MPIU_Free(pOvl);
	}
	goto fn_exit;
}

#undef FUNCNAME
#define FUNCNAME ADIOI_NTFS_Open
int ADIOI_NTFS_Open(MPID_File_ptr fd)
{
	static const wchar_t FCNAME[] = L"ADIOI_NTFS_Open";
	int mpi_errno = MPI_SUCCESS;
	int err;
	int cmode, amode, attrib;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_NTFS_OPEN);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_NTFS_OPEN);

	amode = 0;
	cmode = OPEN_EXISTING;
#ifdef USE_WIN_THREADED_IO
	attrib = FILE_FLAG_OVERLAPPED;
#else
	attrib = FILE_ATTRIBUTE_NORMAL;
#endif

	if (fd->access_mode & ADIO_CREATE)
	{
		cmode = OPEN_ALWAYS;
	}
	if (fd->access_mode & ADIO_EXCL)
	{
		cmode = CREATE_NEW;
	}

	if (fd->access_mode & ADIO_RDONLY)
	{
		amode = GENERIC_READ;
	}
	if (fd->access_mode & ADIO_WRONLY)
	{
		amode = GENERIC_WRITE;
	}
	if (fd->access_mode & ADIO_RDWR)
	{
		amode = GENERIC_READ | GENERIC_WRITE;
	}

	if (fd->access_mode & ADIO_DELETE_ON_CLOSE)
	{
		attrib = attrib | FILE_FLAG_DELETE_ON_CLOSE;
	}
	if (fd->access_mode & ADIO_SEQUENTIAL)
	{
		attrib = attrib | FILE_FLAG_SEQUENTIAL_SCAN;
	}
	else
	{
		attrib = attrib | FILE_FLAG_RANDOM_ACCESS;
	}

	fd->fd_sys = CreateFile(fd->filename, 
		amode,
		FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE, 
		NULL, 
		cmode, 
		attrib, 
		NULL);
	fd->fd_direct = -1;

	if ((fd->fd_sys != INVALID_HANDLE_VALUE) && (fd->access_mode & ADIO_APPEND))
	{
		fd->fp_ind = fd->fp_sys_posn = SetFilePointer(fd->fd_sys, 0, NULL, FILE_END);
		if (fd->fp_ind == INVALID_SET_FILE_POINTER)
		{
			err = GetLastError();
			if (err != NO_ERROR)
			{
				MPIU_ERR_SET1(mpi_errno, MPI_ERR_IO, L"**io", L"**io %s", ADIOI_NTFS_Strerror(err));
				goto fn_fail;
			}
		}
	}

	/* --BEGIN ERROR HANDLING-- */
	if (fd->fd_sys == INVALID_HANDLE_VALUE)
	{
		err = GetLastError();
		MPIU_ERR_SET1(mpi_errno, MPI_ERR_IO, L"**io", L"**io %s", ADIOI_NTFS_Strerror(err));
		goto fn_fail;
	}
	/* --END ERROR HANDLING-- */

fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_NTFS_OPEN);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME ADIOI_NTFS_IwriteContig
int ADIOI_NTFS_IwriteContig(MPID_File_ptr fd, void *buf, int count, MPI_Datatype datatype, int file_ptr_type, MPI_Offset offset, ADIO_Request *request)
{
	static const wchar_t FCNAME[] = L"ADIOI_NTFS_IwriteContig";
	int mpi_errno = MPI_SUCCESS;
	int len, typesize;
	int err;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_NTFS_IWRITECONTIG);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_NTFS_IWRITECONTIG);

	mpi_errno = ADIOI_Malloc_request(request);
	MPIU_ERR_MEM_CHK_POP(mpi_errno, *request, sizeof(ADIO_Request_t), L"ADIO_Request");
	(*request)->optype = ADIOI_WRITE;
	(*request)->fd = fd;
	(*request)->datatype = datatype;

	NMPI_Type_size(mpi_errno, (datatype, &typesize));
	MPIU_ERR_CHK_POP(mpi_errno);
	len = count * typesize;

	if (file_ptr_type == ADIO_INDIVIDUAL)
	{
		offset = fd->fp_ind;
	}
	err = ADIOI_NTFS_aio(fd, buf, len, offset, 1, &((*request)->handle));
	if (file_ptr_type == ADIO_INDIVIDUAL)
	{
		fd->fp_ind += len;
	}

	(*request)->queued = 1;
	ADIOI_Add_req_to_list(request);

	/* --BEGIN ERROR HANDLING-- */
	if (err != MPI_SUCCESS)
	{
		MPIU_ERR_SET(mpi_errno, MPI_ERR_IO, L"**io");
		goto fn_fail;
	}
	/* --END ERROR HANDLING-- */

	fd->fp_sys_posn = -1;   /* set it to null. */
	fd->async_count++;

fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_NTFS_IWRITECONTIG);
	return mpi_errno;
}


/* This function is for implementation convenience. It is not user-visible.
 * If wr==1 write, wr==0 read.
 *
 * Returns MPI_SUCCESS on success, mpi_errno on failure.
 */
#undef FUNCNAME
#define FUNCNAME ADIOI_NTFS_aio
int ADIOI_NTFS_aio(MPID_File_ptr fd, void *buf, int len, MPI_Offset offset,
				   int wr, void *handle)
{
	static const wchar_t FCNAME[] = L"ADIOI_NTFS_aio";
	static DWORD dwNumWritten, dwNumRead;
	BOOL ret_val = FALSE;
	FDTYPE fd_sys;
	int mpi_errno = MPI_SUCCESS;
	OVERLAPPED *pOvl;
	DWORD err;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_NTFS_AIO);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_NTFS_AIO);

	fd_sys = fd->fd_sys;

	pOvl = (OVERLAPPED *) MPIU_Calloc(sizeof(OVERLAPPED), 1);
	MPIU_ERR_MEM_CHK_POP(mpi_errno, pOvl, sizeof(OVERLAPPED), L"pOvl");
	pOvl->hEvent = CreateEvent(NULL, TRUE, TRUE, NULL);
	if (pOvl->hEvent == NULL)
	{
		err = GetLastError();
		MPIU_ERR_SET1(mpi_errno, MPI_ERR_IO, L"**io", L"**io %s", ADIOI_NTFS_Strerror(err));
		MPIU_Free(pOvl);
		goto fn_fail;
	}
	pOvl->Offset = DWORDLOW(offset);
	pOvl->OffsetHigh = DWORDHIGH(offset);

	if (wr)
	{
		/*wprintf(L"WriteFile(%d bytes)\n", len);fflush(stdout);*/
		ret_val = WriteFile(fd_sys, buf, len, &dwNumWritten, pOvl);
	}
	else
	{
		/*
		{
			ADIO_Fcntl_t fcntl_struct;
			int error_code;
			ADIO_Fcntl(fd, ADIO_FCNTL_GET_FSIZE, &fcntl_struct, &error_code);
			wprintf(L"File size a: %d\n", fcntl_struct.fsize);
		}
		wprintf(L"ReadFile(%d bytes)\n", len);fflush(stdout);
		*/
		ret_val = ReadFile(fd_sys, buf, len, &dwNumRead, pOvl);
	}

	/* --BEGIN ERROR HANDLING-- */
	if (ret_val == FALSE) 
	{
		mpi_errno = GetLastError();
		if (mpi_errno != ERROR_IO_PENDING)
		{
			MPIU_ERR_SET1(mpi_errno, MPI_ERR_IO, L"**io", L"**io %s", ADIOI_NTFS_Strerror(mpi_errno));
			goto fn_fail;
		}
		mpi_errno = MPI_SUCCESS;
	}
	/* --END ERROR HANDLING-- */

	*((OVERLAPPED **) handle) = pOvl;

fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_NTFS_AIO);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME ADIOI_NTFS_Strerror
const wchar_t * ADIOI_NTFS_Strerror(int error)
{
	/* obviously not thread safe to store a message like this */
	static wchar_t msg[1024];
	HLOCAL str;
	int num_characters;
	wchar_t *tok;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_NTFS_STRERROR);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_NTFS_STRERROR);

	num_characters = FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_ALLOCATE_BUFFER, 0, error, MAKELANGID( LANG_NEUTRAL, SUBLANG_DEFAULT ), (LPTSTR) &str, 0,0);
	if (num_characters == 0)
	{
		msg[0] = L'\0';
	}
	else
	{
		if (num_characters > 1021)
		{
			memcpy(msg, str, 1021 * sizeof(wchar_t));
			msg[1021] = L'\0';
		}
		else
		{
			memcpy(msg, str, (num_characters+1)*sizeof(wchar_t));
		}
		LocalFree(str);
		wcstok_s(msg, L"\r\n", &tok);
	}

	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_NTFS_STRERROR);
	return msg;
}

#undef FUNCNAME
#define FUNCNAME ADIOI_NTFS_IreadContig
int ADIOI_NTFS_IreadContig(MPID_File_ptr fd, void *buf, int count, MPI_Datatype datatype, int file_ptr_type, MPI_Offset offset, ADIO_Request *request)
{
	static const wchar_t FCNAME[] = L"ADIOI_NTFS_IreadContig";
	int mpi_errno = MPI_SUCCESS;
	int len, typesize;
	int err;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_NTFS_IREADCONTIG);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_NTFS_IREADCONTIG);

	mpi_errno = ADIOI_Malloc_request(request);
	MPIU_ERR_MEM_CHK_POP(mpi_errno, *request, sizeof(ADIO_Request_t), L"ADIO_Request");
	(*request)->optype = ADIOI_READ;
	(*request)->fd = fd;
	(*request)->datatype = datatype;

	NMPI_Type_size(mpi_errno, (datatype, &typesize));
	MPIU_ERR_CHK_POP(mpi_errno);
	len = count * typesize;

	if (file_ptr_type == ADIO_INDIVIDUAL)
	{
		offset = fd->fp_ind;
	}
	err = ADIOI_NTFS_aio(fd, buf, len, offset, 0, &((*request)->handle));
	if (file_ptr_type == ADIO_INDIVIDUAL)
	{
		fd->fp_ind += len;
	}

	(*request)->queued = 1;
	ADIOI_Add_req_to_list(request);

	/* --BEGIN ERROR HANDLING-- */
	if (err != MPI_SUCCESS)
	{
		MPIU_ERR_SET(mpi_errno, MPI_ERR_IO, L"**io");
		goto fn_fail;
	}
	/* --END ERROR HANDLING-- */

	fd->fp_sys_posn = -1;   /* set it to null. */
	fd->async_count++;

fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_NTFS_IREADCONTIG);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME ADIOI_NTFS_Flush
int ADIOI_NTFS_Flush(MPID_File_ptr fd)
{
	static const wchar_t FCNAME[] = L"ADIOI_NTFS_Flush";
	int mpi_errno = MPI_SUCCESS;
	int err;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_NTFS_FLUSH);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_NTFS_FLUSH);

	err = (fd->access_mode & ADIO_RDONLY) ? TRUE : FlushFileBuffers(fd->fd_sys);

	/* --BEGIN ERROR HANDLING-- */
	if (err == FALSE)
	{
		err = GetLastError();
		MPIU_ERR_SET1(mpi_errno, MPI_ERR_IO, L"**io", L"**io %s", ADIOI_NTFS_Strerror(err));
	}
	/* --END ERROR HANDLING-- */

	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_NTFS_FLUSH);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME ADIOI_NTFS_Fcntl
int ADIOI_NTFS_Fcntl(MPID_File_ptr fd, int flag, ADIO_Fcntl_t *fcntl_struct)
{
	static const wchar_t FCNAME[] = L"ADIOI_NTFS_Fcntl";
	int mpi_errno = MPI_SUCCESS;
	DWORD err;
	LONG dwTemp;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_NTFS_FCNTL);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_NTFS_FCNTL);

	switch(flag)
	{
	case ADIO_FCNTL_GET_FSIZE:
		fcntl_struct->fsize = SetFilePointer(fd->fd_sys, 0, 0, FILE_END);
		if (fd->fp_sys_posn != -1) 
		{
			dwTemp = DWORDHIGH(fd->fp_sys_posn);
			if (SetFilePointer(fd->fd_sys, DWORDLOW(fd->fp_sys_posn), &dwTemp, FILE_BEGIN) == INVALID_SET_FILE_POINTER)
			{
				err = GetLastError();
				if (err != NO_ERROR)
				{
					MPIU_ERR_SET1(mpi_errno, MPI_ERR_IO, L"**io", L"**io %s", ADIOI_NTFS_Strerror(err));
					goto fn_fail;
				}
			}
		}
		/* --BEGIN ERROR HANDLING-- */
		if (fcntl_struct->fsize == INVALID_SET_FILE_POINTER)
		{
			dwTemp = GetLastError();
			MPIU_ERR_SET1(mpi_errno, MPI_ERR_IO, L"**io", L"**io %s", ADIOI_NTFS_Strerror(dwTemp));
			goto fn_fail;
		}
		/* --END ERROR HANDLING-- */
		break;

	case ADIO_FCNTL_SET_DISKSPACE:
		mpi_errno = ADIOI_GEN_Prealloc(fd, fcntl_struct->diskspace);
		MPIU_ERR_CHK_POP(mpi_errno);
		break;

	case ADIO_FCNTL_SET_ATOMICITY:
		fd->atomicity = (fcntl_struct->atomicity == 0) ? 0 : 1;
		/*
		fd->atomicity = 0;
		*error_code = MPI_ERR_UNSUPPORTED_OPERATION;
		*/
		break;

	default:
		/* --BEGIN ERROR HANDLING-- */
		MPIU_ERR_SET1(mpi_errno, MPI_ERR_ARG, L"**flag", L"**flag %d", flag);
		/* --END ERROR HANDLING-- */
	}

fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_NTFS_FCNTL);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME ADIOI_NTFS_ReadDone
int ADIOI_NTFS_ReadDone(ADIO_Request *request, MPI_Status *status, int *flag)
{
	static const wchar_t FCNAME[] = L"ADIOI_NTFS_ReadDone";
	int mpi_errno = MPI_SUCCESS;
	DWORD ret_val;
	int done = 0;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_NTFS_READDONE);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_NTFS_READDONE);

	if (*request == NULL)
	{
		*flag = 1;
		goto fn_exit;
	}

	if ((*request)->queued) 
	{
		(*request)->nbytes = 0;
		ret_val = GetOverlappedResult((*request)->fd, (*request)->handle, &(*request)->nbytes, FALSE);

		if (!ret_val)
		{
			/* --BEGIN ERROR HANDLING-- */
			ret_val = GetLastError();
			if (ret_val == ERROR_IO_INCOMPLETE)
			{
				done = 0;
			}
			else
			{
				MPIU_ERR_SET1(mpi_errno, MPI_ERR_IO, L"**io", L"**io %s", ADIOI_NTFS_Strerror(ret_val));
				goto fn_fail;
			}
			/* --END ERROR HANDLING-- */
		}
		else 
		{
			done = 1;
		}
	}
	else
	{
		done = 1;
	}
#ifdef HAVE_STATUS_SET_BYTES
	if (done && ((*request)->nbytes != -1))
	{
		MPIR_Status_set_bytes(status, (*request)->datatype, (*request)->nbytes);
	}
#endif
	
	if (done) 
	{
		/* if request is still queued in the system, it is also there
		   on ADIOI_Async_list. Delete it from there. */
		if ((*request)->queued) ADIOI_Del_req_from_list(request);
		
		(*request)->fd->async_count--;
		if ((*request)->handle) 
		{
			if (!CloseHandle(((OVERLAPPED*)((*request)->handle))->hEvent))
			{
				ret_val = GetLastError();
				MPIU_ERR_SET1(mpi_errno, MPI_ERR_IO, L"**io", L"**io %s", ADIOI_NTFS_Strerror(ret_val));
				/*goto fn_fail;*/ /* Don't skip the following free code */
			}
			MPIU_Free((*request)->handle);
		}
		ADIOI_Free_request((ADIOI_Req_node *) (*request));
		*request = NULL;
	}
	*flag = done;

fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_NTFS_READDONE);
	return mpi_errno;
fn_fail:
	goto fn_exit;
}


#undef FUNCNAME
#define FUNCNAME ADIOI_NTFS_WriteDone
int ADIOI_NTFS_WriteDone(ADIO_Request *request, MPI_Status *status, int *flag)
{
	static const wchar_t FCNAME[] = L"ADIOI_NTFS_WriteDone";
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_NTFS_WRITEDONE);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_NTFS_WRITEDONE);

	mpi_errno = ADIOI_NTFS_ReadDone(request, status, flag);
	MPIU_ERR_CHK_POP(mpi_errno);

fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_NTFS_WRITEDONE);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME ADIOI_NTFS_Close
int ADIOI_NTFS_Close(MPID_File_ptr fd)
{
	static const wchar_t FCNAME[] = L"ADIOI_NTFS_Close";
	int mpi_errno = MPI_SUCCESS;
	int err;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_NTFS_CLOSE);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_NTFS_CLOSE);

	err = CloseHandle(fd->fd_sys);
	/* --BEGIN ERROR HANDLING-- */
	if (err == FALSE)
	{
		err = GetLastError();
		MPIU_ERR_SET1(mpi_errno, MPI_ERR_IO, L"**io", L"**io %s", ADIOI_NTFS_Strerror(err));
	}
	/* --END ERROR HANDLING-- */
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_NTFS_CLOSE);
	return mpi_errno;
}
