/*   Copyright (C) 2009 Deino Software. */

#include "mpiimpl.h"

#undef FUNCNAME
#define FUNCNAME ADIOI_Malloc_async_node
int ADIOI_Malloc_async_node(ADIOI_Async_node **node_ptr)
{
/* returns a pointer to a new node that can be added to ADIOI_Async_list.
   To reduce the number of system calls, mallocs NUM nodes at a time
   and maintains list of available nodes. Supplies a node from this
   list if available, else mallocs a new set of NUM and provides one
   from that set. Is NUM=100 a good number? */

#define NUM 100

	static const wchar_t FCNAME[] = L"ADIOI_Malloc_async_node";
	int mpi_errno = MPI_SUCCESS;
	ADIOI_Async_node *curr, *ptr;
	int i;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_MALLOC_ASYNC_NODE);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_MALLOC_ASYNC_NODE);

	if (!ADIOI_Async_avail_head)
	{
		ADIOI_Async_avail_head = (ADIOI_Async_node *)MPIU_Malloc(NUM*sizeof(ADIOI_Async_node));
		MPIU_ERR_MEM_CHK_POP(mpi_errno, ADIOI_Async_avail_head, NUM*sizeof(ADIOI_Async_node), L"ADIOI_Async_avail_head");
		curr = ADIOI_Async_avail_head;
		for (i=1; i<NUM; i++)
		{
			curr->next = ADIOI_Async_avail_head+i;
			curr = curr->next;
		}
		curr->next = NULL;
		ADIOI_Async_avail_tail = curr;

		/* keep track of malloced area that needs to be freed later */
		if (!ADIOI_Malloc_async_tail)
		{
			ADIOI_Malloc_async_tail = (ADIOI_Malloc_async *)MPIU_Malloc(sizeof(ADIOI_Malloc_async));
			MPIU_ERR_MEM_CHK_POP(mpi_errno, ADIOI_Malloc_async_tail, sizeof(ADIOI_Malloc_async), L"ADIOI_Malloc_async_tail");
			ADIOI_Malloc_async_head = ADIOI_Malloc_async_tail;
			ADIOI_Malloc_async_head->ptr = ADIOI_Async_avail_head;
			ADIOI_Malloc_async_head->next = NULL;
		}
		else
		{
			ADIOI_Malloc_async_tail->next = (ADIOI_Malloc_async *)MPIU_Malloc(sizeof(ADIOI_Malloc_async));
			MPIU_ERR_MEM_CHK_POP(mpi_errno, ADIOI_Malloc_async_tail->next, sizeof(ADIOI_Malloc_async), L"ADIOI_Malloc_async_tail->next");
			ADIOI_Malloc_async_tail = ADIOI_Malloc_async_tail->next;
			ADIOI_Malloc_async_tail->ptr = ADIOI_Async_avail_head;
			ADIOI_Malloc_async_tail->next = NULL;
		}
	}

	ptr = ADIOI_Async_avail_head;
	ADIOI_Async_avail_head = ADIOI_Async_avail_head->next;
	if (!ADIOI_Async_avail_head)
	{
		ADIOI_Async_avail_tail = NULL;
	}

	*node_ptr = ptr;

fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_MALLOC_ASYNC_NODE);
	return mpi_errno;
fn_fail:
	goto fn_exit;
}


#undef FUNCNAME
#define FUNCNAME ADIOI_Free_async_node
int ADIOI_Free_async_node(ADIOI_Async_node *node)
{
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_FREE_ASYNC_NODE);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_FREE_ASYNC_NODE);
	/* moves this node to available pool. does not actually free it. */

	if (!ADIOI_Async_avail_tail)
	{
		ADIOI_Async_avail_head = ADIOI_Async_avail_tail = node;
	}
	else
	{
		ADIOI_Async_avail_tail->next = node;
		ADIOI_Async_avail_tail = node;
	}
	node->next = NULL;
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_FREE_ASYNC_NODE);
	return MPI_SUCCESS;
}


#undef FUNCNAME
#define FUNCNAME ADIOI_Add_req_to_list
int ADIOI_Add_req_to_list(ADIO_Request *request)
{
	static const wchar_t FCNAME[] = L"ADIOI_Add_req_to_list";
	int mpi_errno = MPI_SUCCESS;
	ADIOI_Async_node *curr;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_ADD_REQ_TO_LIST);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_ADD_REQ_TO_LIST);
	/* add request to list of outstanding requests */

	if (!ADIOI_Async_list_head)
	{
		mpi_errno = ADIOI_Malloc_async_node(&ADIOI_Async_list_head);
		MPIU_ERR_CHK_POP(mpi_errno);
		ADIOI_Async_list_head->request = request;
		ADIOI_Async_list_head->prev = ADIOI_Async_list_head->next = NULL;
		ADIOI_Async_list_tail = ADIOI_Async_list_head;
		(*request)->ptr_in_async_list = ADIOI_Async_list_head;
	}
	else
	{
		curr = ADIOI_Async_list_tail;
		mpi_errno = ADIOI_Malloc_async_node(&curr->next);
		MPIU_ERR_CHK_POP(mpi_errno);
		ADIOI_Async_list_tail = curr->next;
		ADIOI_Async_list_tail->request = request;
		ADIOI_Async_list_tail->prev = curr;
		ADIOI_Async_list_tail->next = NULL;
		(*request)->ptr_in_async_list = ADIOI_Async_list_tail;
	}

fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_ADD_REQ_TO_LIST);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME ADIOI_Complete_async
int ADIOI_Complete_async()
{
	/* complete all outstanding async I/O operations so that new ones can be
	   initiated. Remove them all from async_list. */
	static const wchar_t FCNAME[] = L"ADIOI_Complete_async";
	int mpi_errno = MPI_SUCCESS;
	MPI_Status status;
	ADIO_Request *request;
	ADIOI_Async_node *tmp;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_COMPLETE_ASYNC);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_COMPLETE_ASYNC);

	while (ADIOI_Async_list_head)
	{
		request = ADIOI_Async_list_head->request;
		(*request)->queued = -1; /* ugly internal hack that prevents
				  ADIOI_xxxComplete from freeing the request object. 
				  This is required, because the user will call MPI_Wait
				  later, which would require status to be filled. */
		switch ((*request)->optype)
		{
		case ADIOI_READ:
/*			mpi_errno = (*((*request)->fd->fns->ADIOI_xxx_ReadComplete))(request,
													&status);*/
			mpi_errno = ADIO_ReadComplete(request, &status);
			break;
		case ADIOI_WRITE:
/*			mpi_errno = (*((*request)->fd->fns->ADIOI_xxx_WriteComplete))(request,
													 &status);*/
			mpi_errno = ADIO_WriteComplete(request, &status);
			break;
		default:
			/* --BEGIN ERROR HANDLING-- */
			MPIU_ERR_SET1(mpi_errno, MPI_ERR_INTERN, L"**fail", L"**fail %d", (*request)->optype);
			MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_COMPLETE_ASYNC);
			return mpi_errno;
			/* --END ERROR HANDLING-- */
		}
		(*request)->queued = 0;  /* dequeued, but request object not
									freed */

		tmp = ADIOI_Async_list_head;
		ADIOI_Async_list_head = ADIOI_Async_list_head->next;
		ADIOI_Free_async_node(tmp);
	}
	ADIOI_Async_list_tail = NULL;
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_COMPLETE_ASYNC);
	return mpi_errno;
}


#undef FUNCNAME
#define FUNCNAME ADIOI_Del_req_from_list
int ADIOI_Del_req_from_list(ADIO_Request *request)
{
/* Delete a request that has already been completed from the async
   list and move it to the list of available nodes. Typically called
   from within an ADIO_Test/ADIO_Wait. */ 
	int mpi_errno = MPI_SUCCESS;
	ADIOI_Async_node *curr, *prev, *next;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_DEL_REQ_FROM_LIST);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_DEL_REQ_FROM_LIST);

	curr = (*request)->ptr_in_async_list;
	prev = curr->prev;

	if (prev)
	{
		prev->next = curr->next;
	}
	else
	{
		ADIOI_Async_list_head = curr->next;
	}

	next = curr->next;
	if (next)
	{
		next->prev = prev;
	}
	else
	{
		ADIOI_Async_list_tail = prev;
	}

	mpi_errno = ADIOI_Free_async_node(curr);

	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_DEL_REQ_FROM_LIST);
	return mpi_errno;
}
