/*   Copyright (C) 2009 Deino Software. */

#include "mpiimpl.h"

/* Returns the absolute byte position in the file corresponding to 
   L"offset" etypes relative to the current view. */

#undef FUNCNAME
#define FUNCNAME ADIOI_Get_byte_offset
int ADIOI_Get_byte_offset(MPID_File_ptr fd, MPI_Offset offset, MPI_Offset *disp)
{
	static const wchar_t FCNAME[] = L"ADIOI_Get_byte_offset";
	int mpi_errno = MPI_SUCCESS;
	ADIOI_Flatlist_node *flat_file;
	int i, sum, n_etypes_in_filetype, size_in_filetype;
	int n_filetypes, etype_in_filetype;
	MPI_Offset abs_off_in_filetype=0;
	int filetype_size, etype_size, filetype_is_contig;
	MPI_Aint filetype_extent;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_GET_BYTE_OFFSET);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_GET_BYTE_OFFSET);

	mpi_errno = ADIOI_Datatype_iscontig(fd->filetype, &filetype_is_contig);
	MPIU_ERR_CHK_POP(mpi_errno);
	etype_size = fd->etype_size;

	if (filetype_is_contig)
	{
		*disp = fd->disp + etype_size * offset;
	}
	else
	{
		/* filetype already flattened in ADIO_Open */
		flat_file = ADIOI_Flatlist;
		while (flat_file->type != fd->filetype)
		{
			flat_file = flat_file->next;
		}

		NMPI_Type_size(mpi_errno, (fd->filetype, &filetype_size));
		MPIU_ERR_CHK_POP(mpi_errno);
		n_etypes_in_filetype = filetype_size/etype_size;
		n_filetypes = (int) (offset / n_etypes_in_filetype);
		etype_in_filetype = (int) (offset % n_etypes_in_filetype);
		size_in_filetype = etype_in_filetype * etype_size;
 
		sum = 0;
		for (i=0; i<flat_file->count; i++)
		{
			sum += flat_file->blocklens[i];
			if (sum > size_in_filetype)
			{
				abs_off_in_filetype = flat_file->indices[i] +
					size_in_filetype - (sum - flat_file->blocklens[i]);
				break;
			}
		}

		/* abs. offset in bytes in the file */
		NMPI_Type_extent(mpi_errno, (fd->filetype, &filetype_extent));
		MPIU_ERR_CHK_POP(mpi_errno);
		*disp = fd->disp + (MPI_Offset) n_filetypes*filetype_extent + abs_off_in_filetype;
	}

fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_GET_BYTE_OFFSET);
	return mpi_errno;
}
