/*   Copyright (C) 2009 Deino Software. */

/* cb_config_list.c
 *
 * The important, externally used functions from this file are:
 * ADIOI_cb_bcast_rank_map()
 * ADIOI_cb_gather_name_array()
 * ADIOI_cb_config_list_parse()
 * ADIOI_cb_copy_name_array()
 * ADIOI_cb_delete_name_array()
 *
 * Prototypes for these are in adio/include/adio_cb_config_list.h
 */

/*
#include "adio.h"
#include "mpi.h"
*/
#include "mpiimpl.h"
/*#include "adio_cb_config_list.h"*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

/* token types */
#define AGG_WILDCARD 1
#define AGG_STRING   2
#define AGG_COMMA	3
#define AGG_COLON	4
#define AGG_ERROR   -1
#define AGG_EOS	  0

#undef CB_CONFIG_LIST_DEBUG

/* a couple of globals keep things simple */
static int cb_config_list_keyval = MPI_KEYVAL_INVALID;
static wchar_t *yylval;
static wchar_t *token_ptr;

/* internal stuff */
static int get_max_procs(int cb_nodes);
static int match_procs(wchar_t *name, int max_per_proc, wchar_t *procnames[], 
					   wchar_t used_procnames[],
					   int nr_procnames, int ranks[], int nr_ranks, 
					   int *nr_ranks_allocated);
static int match_this_proc(wchar_t *name, int cur_proc, int max_matches,
						   wchar_t *procnames[], wchar_t used_procnames[],
						   int nr_procnames, int ranks[], 
						   int nr_ranks, int nr_ranks_allocated);
static int find_name(wchar_t *name, wchar_t *procnames[], wchar_t used_procnames[], 
					 int nr_procnames, int start_ind);
static int cb_config_list_lex(void);


/* ADIOI_cb_bcast_rank_map() - broadcast the rank array
 *
 * Parameters:
 * fd - MPID_File_ptr for which update is occurring.  cb_nodes and ranklist
 * parameters must be up-to-date on rank 0 of the fd->comm.
 *
 * should probably be a void fn.
 */
#undef FUNCNAME
#define FUNCNAME ADIOI_cb_bcast_rank_map
int ADIOI_cb_bcast_rank_map(MPID_File_ptr fd)
{
	static const wchar_t FCNAME[] = L"ADIOI_cb_bcast_rank_map";
	int mpi_errno = MPI_SUCCESS;
	int my_rank;
	wchar_t *value = NULL;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_CB_BCAST_RANK_MAP);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_CB_BCAST_RANK_MAP);

	NMPI_Bcast(mpi_errno, (&(fd->hints->cb_nodes), 1, MPI_INT, 0, fd->comm));
	MPIU_ERR_CHK_POP(mpi_errno);
	if (fd->hints->cb_nodes > 0)
	{
		NMPI_Comm_rank(mpi_errno, (fd->comm, &my_rank));
		MPIU_ERR_CHK_POP(mpi_errno);
		if (my_rank != 0)
		{
			fd->hints->ranklist = MPIU_Malloc(fd->hints->cb_nodes*sizeof(int));
			MPIU_ERR_MEM_CHK_POP(mpi_errno, fd->hints->ranklist, fd->hints->cb_nodes * sizeof(int), L"fd->hints->ranklist");
		}
		NMPI_Bcast(mpi_errno, (fd->hints->ranklist, fd->hints->cb_nodes, MPI_INT, 0, fd->comm));
		MPIU_ERR_CHK_POP(mpi_errno);
	}
	/* TEMPORARY -- REMOVE WHEN NO LONGER UPDATING INFO FOR
	 * FS-INDEP. */
	value = (wchar_t *) MPIU_Malloc((MPI_MAX_INFO_VAL+1)*sizeof(wchar_t));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, value, (MPI_MAX_INFO_VAL+1) * sizeof(wchar_t), L"value");
	MPIU_Snprintf(value, MPI_MAX_INFO_VAL+1, _TRUNCATE, L"%d", fd->hints->cb_nodes);
	NMPI_Info_set(mpi_errno, (fd->info, L"cb_nodes", value));
	MPIU_ERR_CHK_POP(mpi_errno);
	MPIU_Free(value);
	value = NULL;

fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_CB_BCAST_RANK_MAP);
	return mpi_errno;

fn_fail:
	if (value != NULL)
	{
		MPIU_Free(value);
	}
	goto fn_exit;
}

/* ADIOI_cb_gather_name_array() - gather a list of processor names from all processes
 *						  in a communicator and store them on rank 0.
 *
 * This is a collective call on the communicator(s) passed in.
 *
 * Obtains a rank-ordered list of processor names from the processes in
 * "dupcomm".
 *
 * Returns 0 on success, -1 on failure.
 *
 * NOTE: Needs some work to cleanly handle out of memory cases!  
 */
#undef FUNCNAME
#define FUNCNAME ADIOI_cb_gather_name_array
int ADIOI_cb_gather_name_array(MPI_Comm comm,
							   MPI_Comm dupcomm,
							   ADIO_cb_name_array *arrayp)
{
	static const wchar_t FCNAME[] = L"ADIOI_cb_gather_name_array";
	int mpi_errno = MPI_SUCCESS;
	wchar_t my_procname[MPI_MAX_PROCESSOR_NAME], **procname = NULL;
	int *procname_len = NULL, my_procname_len, *disp = NULL, i;
	int commsize, commrank, found;
	ADIO_cb_name_array array = NULL;
	int alloc_size;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_CB_GATHER_NAME_ARRAY);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_CB_GATHER_NAME_ARRAY);

	if (cb_config_list_keyval == MPI_KEYVAL_INVALID)
	{
		NMPI_Keyval_create(mpi_errno, ((MPI_Copy_function *) ADIOI_cb_copy_name_array, 
						  (MPI_Delete_function *) ADIOI_cb_delete_name_array,
						  &cb_config_list_keyval, NULL));
		MPIU_ERR_CHK_POP(mpi_errno);
	}
	else
	{
		NMPI_Attr_get(mpi_errno, (comm, cb_config_list_keyval, (void *) &array, &found));
		MPIU_ERR_CHK_POP(mpi_errno);
		if (found)
		{
			*arrayp = array;
			goto fn_exit;
		}
	}

	NMPI_Comm_size(mpi_errno, (dupcomm, &commsize));
	MPIU_ERR_CHK_POP(mpi_errno);
	NMPI_Comm_rank(mpi_errno, (dupcomm, &commrank));
	MPIU_ERR_CHK_POP(mpi_errno);

	NMPI_Get_processor_name(mpi_errno, (my_procname, &my_procname_len));
	MPIU_ERR_CHK_POP(mpi_errno);

	/* allocate space for everything */
	array = (ADIO_cb_name_array) MPIU_Malloc(sizeof(*array));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, array, sizeof(*array), L"array");

	array->refct = 2; /* we're going to associate this with two comms */

	if (commrank == 0)
	{
		/* process 0 keeps the real list */
		array->namect = commsize;

		array->names = (wchar_t **) MPIU_Malloc(sizeof(wchar_t *) * commsize);
		MPIU_ERR_MEM_CHK_POP(mpi_errno, array->names, sizeof(wchar_t *) * commsize, L"array->names");
		procname = array->names; /* simpler to read */

		procname_len = (int *) MPIU_Malloc(commsize * sizeof(int));
		MPIU_ERR_MEM_CHK_POP(mpi_errno, procname_len, commsize * sizeof(int), L"procname_len");
	}
	else
	{
		/* everyone else just keeps an empty list as a placeholder */
		array->namect = 0;
		array->names = NULL;
	}
	/* gather lengths first */
	NMPI_Gather(mpi_errno, (&my_procname_len, 1, MPI_INT, 
			   procname_len, 1, MPI_INT, 0, dupcomm));
	MPIU_ERR_CHK_POP(mpi_errno);

	if (commrank == 0)
	{
#ifdef CB_CONFIG_LIST_DEBUG
		for (i=0; i < commsize; i++) {
			FPRINTF(stderr, L"len[%d] = %d\n", i, procname_len[i]);
		}
#endif

		alloc_size = 0;
		for (i=0; i < commsize; i++)
		{
			/* add one to the lengths because we need to count the
			 * terminator, and we are going to use this list of lengths
			 * again in the gatherv.  
			 */
			/*
			procname_len[i]++;
			procname[i] = MPIU_Malloc(procname_len[i] * sizeof(wchar_t));
			MPIU_ERR_MEM_CHK_POP(mpi_errno, procname[i], procname_len[i], L"procname[i]");
			*/
			alloc_size += ++procname_len[i];
		}
		
		procname[0] = MPIU_Malloc(alloc_size * sizeof(wchar_t));
		MPIU_ERR_MEM_CHK_POP(mpi_errno, procname[0], alloc_size*sizeof(wchar_t), L"procname[0]");

		for (i=1; i<commsize; i++)
		{
			procname[i] = procname[i-1] + procname_len[i-1];
		}

		/* create our list of displacements for the gatherv.  we're going
		 * to do everything relative to the start of the region allocated
		 * for procname[0]
		 *
		 * I suppose it is theoretically possible that the distance between 
		 * malloc'd regions could be more than will fit in an int.  We don't
		 * cover that case.
		 */
		disp = MPIU_Malloc(commsize * sizeof(int));
		MPIU_ERR_MEM_CHK_POP(mpi_errno, disp, commsize * sizeof(int), L"disp");
		disp[0] = 0;
		for (i=1; i < commsize; i++)
		{
			disp[i] = (int) (procname[i] - procname[0]); /*/ sizeof(wchar_t);*/ /* Are displacements in bytes or MPI_Datatype units? */
		}
	}

	/* now gather strings */
	if (commrank == 0)
	{
		NMPI_Gatherv(mpi_errno, (my_procname, my_procname_len + 1, MPI_WCHAR, 
					procname[0], procname_len, disp, MPI_WCHAR,
					0, dupcomm));
		MPIU_ERR_CHK_POP(mpi_errno);
	}
	else
	{
		/* if we didn't do this, we would need to allocate procname[]
		 * on all processes...which seems a little silly.
		 */
		NMPI_Gatherv(mpi_errno, (my_procname, my_procname_len + 1, MPI_WCHAR, 
					NULL, NULL, NULL, MPI_WCHAR, 0, dupcomm));
		MPIU_ERR_CHK_POP(mpi_errno);
	}

	if (commrank == 0)
	{
		/* no longer need the displacements or lengths */
		MPIU_Free(disp);
		MPIU_Free(procname_len);
		disp = NULL;
		procname_len = NULL;

#ifdef CB_CONFIG_LIST_DEBUG
		for (i=0; i < commsize; i++) {
			FPRINTF(stderr, L"name[%d] = %s\n", i, procname[i]);
		}
#endif
	}

	/* store the attribute; we want to store SOMETHING on all processes
	 * so that they can all tell if we have gone through this procedure 
	 * or not for the given communicator.
	 *
	 * specifically we put it on both the original comm, so we can find
	 * it next time an open is performed on this same comm, and on the
	 * dupcomm, so we can use it in I/O operations.
	 */
	NMPI_Attr_put(mpi_errno, (comm, cb_config_list_keyval, array));
	MPIU_ERR_CHK_POP(mpi_errno);
	NMPI_Attr_put(mpi_errno, (dupcomm, cb_config_list_keyval, array));
	MPIU_ERR_CHK_POP(mpi_errno);
	*arrayp = array;

fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_CB_GATHER_NAME_ARRAY);
	return mpi_errno;
fn_fail:
	goto fn_exit;
}


/* ADIOI_cb_config_list_parse() - parse the cb_config_list and build the 
 * ranklist
 *
 * Parameters:
 * (pretty self explanatory)
 *
 * Returns number of ranks allocated in parsing, -1 on error.
 */
#undef FUNCNAME
#define FUNCNAME ADIOI_cb_config_list_parse
int ADIOI_cb_config_list_parse(wchar_t *config_list, 
						 ADIO_cb_name_array array,
						 int ranklist[], 
						 int cb_nodes,
						 int *rank_ptr)
{
	static const wchar_t FCNAME[] = L"ADIOI_cb_config_list_parse";
	int mpi_errno = MPI_SUCCESS;
	int token, max_procs, cur_rank = 0, nr_procnames;
	wchar_t *cur_procname = NULL, *cur_procname_p, **procnames;
	wchar_t *used_procnames = NULL;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_CB_CONFIG_LIST_PARSE);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_CB_CONFIG_LIST_PARSE);

	nr_procnames = array->namect;
	procnames = array->names;

	/* nothing big goes on the stack */
	/* we use info val here and for yylval because we know the string
	 * cannot be any bigger than this.
	 */
	cur_procname = MPIU_Malloc((MPI_MAX_INFO_VAL+1) * sizeof(wchar_t));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, cur_procname, (MPI_MAX_INFO_VAL+1) * sizeof(wchar_t), L"cur_procname");

	yylval = MPIU_Malloc((MPI_MAX_INFO_VAL+1) * sizeof(wchar_t));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, yylval, (MPI_MAX_INFO_VAL+1) * sizeof(wchar_t), L"yylval");

	token_ptr = config_list;

	/* right away let's make sure cb_nodes isn't too big */
	if (cb_nodes > nr_procnames) cb_nodes = nr_procnames;

	/* used_procnames is used as a mask so that we don't have to destroy
	 * our procnames array
	 */
	used_procnames = MPIU_Malloc(array->namect * sizeof(wchar_t));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, used_procnames, array->namect * sizeof(wchar_t), L"used_procnames");
	memset(used_procnames, 0, array->namect * sizeof(wchar_t));

	/* optimization for "*:*"; arguably this could be done before we
	 * build the list of processor names...but that would make things
	 * messy.
	 */
	if (wcscmp(config_list, L"*:*") == 0)
	{
		for (cur_rank = 0; cur_rank < cb_nodes; cur_rank++)
		{
			ranklist[cur_rank] = cur_rank;
		}
		MPIU_Free(cur_procname);
		cur_procname = NULL;
		MPIU_Free(yylval);
		yylval = NULL;
			MPIU_Free(used_procnames);
		used_procnames = NULL;
		*rank_ptr = cb_nodes;
		goto fn_exit;
	}

	while (cur_rank < cb_nodes)
	{
		token = cb_config_list_lex();

		if (token == AGG_EOS)
		{
			MPIU_Free(cur_procname);
			cur_procname = NULL;
			MPIU_Free(yylval);
			yylval = NULL;
				MPIU_Free(used_procnames);
			used_procnames = NULL;
			*rank_ptr = cur_rank;
			goto fn_exit;
		}

		if (token != AGG_WILDCARD && token != AGG_STRING)
		{
			/* maybe ignore and try to keep going? */
			/*FPRINTF(stderr, L"error parsing config list\n");*/
			MPIU_Free(cur_procname);
			cur_procname = NULL;
			MPIU_Free(yylval);
			yylval = NULL;
				MPIU_Free(used_procnames);
			used_procnames = NULL;
			*rank_ptr = cur_rank;
			goto fn_exit;
		}

		if (token == AGG_WILDCARD)
		{
			cur_procname_p = NULL;
		}
		else
		{
			/* AGG_STRING is the only remaining case */
			/* save procname (for now) */
			MPIU_Wcsncpy(cur_procname, yylval, MPI_MAX_INFO_VAL+1);
			cur_procname_p = cur_procname;
		}

		/* after we have saved the current procname, we can grab max_procs */
		max_procs = get_max_procs(cb_nodes);

#ifdef CB_CONFIG_LIST_DEBUG
		if (token == AGG_WILDCARD)
		{
			FPRINTF(stderr, L"looking for *:%d\n", max_procs);
		}
		else
		{
			FPRINTF(stderr, L"looking for %s:%d\n", cur_procname, max_procs);
		}
#endif

		/* do the matching for this piece of the cb_config_list */
		match_procs(cur_procname_p, max_procs, procnames, used_procnames,
					nr_procnames, ranklist, cb_nodes, &cur_rank);
	}
	MPIU_Free(cur_procname);
	MPIU_Free(yylval);
	yylval = NULL;
	MPIU_Free(used_procnames);
	*rank_ptr = cur_rank;

fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_CB_CONFIG_LIST_PARSE);
	return mpi_errno;

fn_fail:
	if (cur_procname != NULL)
	{
		MPIU_Free(cur_procname);
	}
	if (yylval != NULL)
	{
		MPIU_Free(yylval);
		yylval = NULL;
	}
	if (used_procnames != NULL)
	{
		MPIU_Free(used_procnames);
	}
	goto fn_exit;
}

/* ADIOI_cb_copy_name_array() - attribute copy routine */
#undef FUNCNAME
#define FUNCNAME ADIOI_cb_copy_name_array
int ADIOI_cb_copy_name_array(MPI_Comm comm, 
					   int *keyval, 
					   void *extra, 
					   void *attr_in,
					   void **attr_out, 
					   int *flag)
{
	ADIO_cb_name_array array;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_CB_COPY_NAME_ARRAY);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_CB_COPY_NAME_ARRAY);

	MPIU_UNREFERENCED_ARG(comm);
	MPIU_UNREFERENCED_ARG(keyval);
	MPIU_UNREFERENCED_ARG(extra);

	array = (ADIO_cb_name_array) attr_in;
	array->refct++;

	*attr_out = attr_in;
	*flag = 1; /* make a copy in the new communicator */

	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_CB_COPY_NAME_ARRAY);
	return MPI_SUCCESS;
}

/* ADIOI_cb_delete_name_array() - attribute destructor */
#undef FUNCNAME
#define FUNCNAME ADIOI_cb_delete_name_array
int ADIOI_cb_delete_name_array(MPI_Comm comm, 
						 int *keyval, 
						 void *attr_val, 
						 void *extra)
{
	/*int i;*/
	ADIO_cb_name_array array;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_CB_DELETE_NAME_ARRAY);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_CB_DELETE_NAME_ARRAY);

	MPIU_UNREFERENCED_ARG(comm);
	MPIU_UNREFERENCED_ARG(keyval);
	MPIU_UNREFERENCED_ARG(extra);

	array = (ADIO_cb_name_array) attr_val;
	array->refct--;

	if (array->refct <= 0)
	{
		/* time to free the structures (names, array of ptrs to names, struct) 
		 */
		/*
		for (i=0; i < array->namect; i++)
		{
			MPIU_Free(array->names[i]);
		}
		*/
			/* Note that array->names[i], where i > 0, 
			 * are just pointers into the allocated region array->names[0]
			 */
		if (array->namect) MPIU_Free(array->names[0]);
		if (array->names != NULL) MPIU_Free(array->names);
		MPIU_Free(array);
	}

	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_CB_DELETE_NAME_ARRAY);
	return MPI_SUCCESS;
}

/* match_procs() - given a name (or NULL for wildcard) and a max. number
 *				 of aggregator processes (per processor name), this 
 *				 matches in the procnames[] array and puts appropriate
 *				 ranks in the ranks array.
 *
 * Parameters:
 * name - processor name (or NULL for wildcard)
 * max_per_proc - maximum # of processes to use for aggregation from a
 *				single processor
 * procnames - array of processor names
 * nr_procnames - length of procnames array
 * ranks - array of process ranks
 * nr_ranks - length of process ranks array (also max. # of aggregators)
 * nr_ranks_allocated - # of array entries which have been filled in,
 *					  which is also the index to the first empty entry
 *
 * Returns number of matches.
 */
#undef FUNCNAME
#define FUNCNAME match_procs
static int match_procs(wchar_t *name, 
					   int max_per_proc, 
					   wchar_t *procnames[],
					   wchar_t used_procnames[],
					   int nr_procnames,
					   int ranks[],
					   int nr_ranks,
					   int *nr_ranks_allocated)
{
	int wildcard_proc, cur_proc, old_nr_allocated, ret;
	MPIDI_STATE_DECL(MPID_STATE_MATCH_PROCS);

	MPIDI_FUNC_ENTER(MPID_STATE_MATCH_PROCS);

	/* save this so we can report on progress */
	old_nr_allocated = *nr_ranks_allocated;

	if (name == NULL)
	{
		/* wildcard case */

		/* optimize for *:0 case */
		if (max_per_proc == 0)
		{
			/* loop through procnames and mark them all as used */
			for (cur_proc = 0; cur_proc < nr_procnames; cur_proc++)
			{
				used_procnames[cur_proc] = 1;
			}
			MPIDI_FUNC_EXIT(MPID_STATE_MATCH_PROCS);
			return 0;
		}

		/* the plan here is to start at the beginning of the procnames
		 * array looking for processor names to apply the wildcard to.
		 *
		 * we set wildcard_proc to 0 here but do the search inside the
		 * while loop so that we aren't starting our search from the
		 * beginning of the procnames array each time.
		 */
		wildcard_proc = 0;

		while (nr_ranks - *nr_ranks_allocated > 0)
		{
			/* find a name */
			while ((wildcard_proc < nr_procnames) && 
				   (used_procnames[wildcard_proc] != 0)) 
			{
				wildcard_proc++;
			}

			if (wildcard_proc == nr_procnames)
			{
				/* we have used up the entire procnames list */
				MPIDI_FUNC_EXIT(MPID_STATE_MATCH_PROCS);
				return *nr_ranks_allocated - old_nr_allocated;
			}

#ifdef CB_CONFIG_LIST_DEBUG
			FPRINTF(stderr, L"performing wildcard match (*:%d) starting with %s (%d)\n", 
				   max_per_proc, procnames[wildcard_proc], wildcard_proc);
#endif

			cur_proc = wildcard_proc;

#ifdef CB_CONFIG_LIST_DEBUG
			FPRINTF(stderr, L"  assigning name %s (%d) to rank %d in mapping\n",
				   procnames[cur_proc], cur_proc, *nr_ranks_allocated);
#endif

			/* alloc max_per_proc from this host; cur_proc points to
			 * the first one.  We want to save this name for use in
			 * our while loop.
			 */
			ranks[*nr_ranks_allocated] = cur_proc;
			*nr_ranks_allocated = *nr_ranks_allocated + 1;			
			cur_proc++;

			/* so, to accomplish this we use the match_this_proc() to
			 * alloc max_per_proc-1.  we increment cur_proc so that the
			 * procnames[] entry doesn't get trashed.  then AFTER the call
			 * we clean up the first instance of the name.
			 */
			ret = match_this_proc(procnames[wildcard_proc],  cur_proc,
								  max_per_proc-1, procnames, used_procnames,
								  nr_procnames,
								  ranks, nr_ranks, *nr_ranks_allocated);
			if (ret > 0) *nr_ranks_allocated = *nr_ranks_allocated + ret;
	
			/* clean up and point wildcard_proc to the next entry, since
			 * we know that this one is NULL now.
			 */
			used_procnames[wildcard_proc] = 1;
			wildcard_proc++;
		}
	}
	else {
		/* specific host was specified; this one is easy */
#ifdef CB_CONFIG_LIST_DEBUG
		FPRINTF(stderr, L"performing name match (%s:%d)\n", name, max_per_proc);
#endif

		ret = match_this_proc(name, 0, max_per_proc, procnames, used_procnames,
							  nr_procnames, ranks, nr_ranks,
							  *nr_ranks_allocated);
		if (ret > 0) *nr_ranks_allocated = *nr_ranks_allocated + ret;
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MATCH_PROCS);
	return *nr_ranks_allocated - old_nr_allocated;
}

/* match_this_proc() - find each instance of processor name "name" in 
 *					 the "procnames" array, starting with index "cur_proc"
 *					 and add the first "max_matches" into the "ranks"
 *					 array.  remove all instances of "name" from
 *					 "procnames".
 *
 * Parameters:
 * name - processor name to match
 * cur_proc - index into procnames[] at which to start matching
 * procnames - array of processor names
 * used_procnames - array of values indicating if a given procname has
 *				  been allocated or removed already
 * nr_procnames - length of procnames array
 * ranks - array of processor ranks
 * nr_ranks - length of ranks array
 * nr_ranks_allocated - number of ranks already filled in, or the next
 *					  entry to fill in (equivalent)
 *
 * Returns number of ranks filled in (allocated).
 */
#undef FUNCNAME
#define FUNCNAME match_this_proc
static int match_this_proc(wchar_t *name,
						   int cur_proc,
						   int max_matches,
						   wchar_t *procnames[],
						   wchar_t used_procnames[],
						   int nr_procnames, 
						   int ranks[], 
						   int nr_ranks,
						   int nr_ranks_allocated)
{
	int ranks_remaining, nr_to_alloc, old_nr_allocated;
	MPIDI_STATE_DECL(MPID_STATE_MATCH_THIS_PROC);

	MPIDI_FUNC_ENTER(MPID_STATE_MATCH_THIS_PROC);

	old_nr_allocated = nr_ranks_allocated;

	/* calculate how many ranks we want to allocate */
	ranks_remaining = nr_ranks - nr_ranks_allocated;
	nr_to_alloc = (max_matches < ranks_remaining) ? 
		max_matches : ranks_remaining;

	while (nr_to_alloc > 0)
	{
		cur_proc = find_name(name, procnames, used_procnames, nr_procnames, 
							 cur_proc);
		if (cur_proc < 0)
		{
			/* didn't find it */
			MPIDI_FUNC_EXIT(MPID_STATE_MATCH_THIS_PROC);
			return nr_ranks_allocated - old_nr_allocated;
		}

		/* need bounds checking on ranks */
#ifdef CB_CONFIG_LIST_DEBUG
		FPRINTF(stderr, L"  assigning name %s (%d) to rank %d in mapping\n",
			   procnames[cur_proc], cur_proc, nr_ranks_allocated);
#endif

		ranks[nr_ranks_allocated] = cur_proc;
		nr_ranks_allocated++;
		used_procnames[cur_proc] = 1;
			
		cur_proc++;
		nr_to_alloc--;
	}
		
	/* take all other instances of this host out of the list */
	while (cur_proc >= 0)
	{
		cur_proc = find_name(name, procnames, used_procnames, nr_procnames, cur_proc);
		if (cur_proc >= 0)
		{
#ifdef CB_CONFIG_LIST_DEBUG
			FPRINTF(stderr, L"  taking name %s (%d) out of procnames\n",
				   procnames[cur_proc], cur_proc);
#endif
			used_procnames[cur_proc] = 1;
			cur_proc++;
		}
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MATCH_THIS_PROC);
	return nr_ranks_allocated - old_nr_allocated;
}
  

/* find_name() - finds the first entry in procnames[] which matches name,
 *			   starting at index start_ind
 *
 * Returns an index [0..nr_procnames-1] on success, -1 if not found.
 */
#undef FUNCNAME
#define FUNCNAME find_name
static int find_name(wchar_t *name, 
					 wchar_t *procnames[], 
					 wchar_t used_procnames[],
					 int nr_procnames, 
					 int start_ind)
{
	int i;
	MPIDI_STATE_DECL(MPID_STATE_FIND_NAME);

	MPIDI_FUNC_ENTER(MPID_STATE_FIND_NAME);

	for (i=start_ind; i < nr_procnames; i++)
	{
		if (!used_procnames[i] && !wcscmp(name, procnames[i])) break;
	}

	MPIDI_FUNC_EXIT(MPID_STATE_FIND_NAME);
	if (i < nr_procnames) return i;
	else return -1;
}

/* get_max_procs() - grab the maximum number of processes to use out of
 *				   the cb_config_list string
 *
 * Parameters:
 * cb_nodes - cb_nodes value.  this is returned when a "*" is encountered
 *			as the max_procs value.
 *
 * Returns # of processes, or -1 on error.
 */
#undef FUNCNAME
#define FUNCNAME get_max_procs
static int get_max_procs(int cb_nodes)
{
	int token, max_procs = -1;
	wchar_t *errptr;
	MPIDI_STATE_DECL(MPID_STATE_GET_MAX_PROCS);

	MPIDI_FUNC_ENTER(MPID_STATE_GET_MAX_PROCS);

	token = cb_config_list_lex();

	switch (token)
	{
	case AGG_EOS:
	case AGG_COMMA:
		MPIDI_FUNC_EXIT(MPID_STATE_GET_MAX_PROCS);
		return 1;
	case AGG_COLON:
		token = cb_config_list_lex();
		if (token != AGG_WILDCARD && token != AGG_STRING)
		{
			MPIDI_FUNC_EXIT(MPID_STATE_GET_MAX_PROCS);
			return -1;
		}
		if (token == AGG_WILDCARD)
		{
			max_procs = cb_nodes;
		}
		else if (token == AGG_STRING)
		{
			max_procs = wcstol(yylval, &errptr, 10);
			if (*errptr != L'\0') {
				/* some garbage value; default to 1 */
				max_procs = 1;
			}
		}
		/* strip off next comma (if there is one) */
		token = cb_config_list_lex();
		if (token != AGG_COMMA && token != AGG_EOS)
		{
			MPIDI_FUNC_EXIT(MPID_STATE_GET_MAX_PROCS);
			return -1;
		}
		
		/* return max_procs */
		MPIDI_FUNC_EXIT(MPID_STATE_GET_MAX_PROCS);
		if (max_procs < 0) return -1;
		else return max_procs;
	}
	MPIDI_FUNC_EXIT(MPID_STATE_GET_MAX_PROCS);
	return -1;
}


/* cb_config_list_lex() - lexical analyzer for cb_config_list language
 *
 * Returns a token of types defined at top of this file.
 */
#undef FUNCNAME
#define FUNCNAME cb_config_list_lex
static int cb_config_list_lex(void)
{
	int slen;
	MPIDI_STATE_DECL(MPID_STATE_CB_CONFIG_LIST_LEX);

	MPIDI_FUNC_ENTER(MPID_STATE_CB_CONFIG_LIST_LEX);

	if (*token_ptr == L'\0') return AGG_EOS;

	slen = (int)wcscspn(token_ptr, L":,");

	if (*token_ptr == L':')
	{
		token_ptr++;
		MPIDI_FUNC_EXIT(MPID_STATE_CB_CONFIG_LIST_LEX);
		return AGG_COLON;
	}
	if (*token_ptr == L',')
	{
		token_ptr++;
		MPIDI_FUNC_EXIT(MPID_STATE_CB_CONFIG_LIST_LEX);
		return AGG_COMMA;
	}

	if (*token_ptr == L'*')
	{
		/* make sure that we don't have characters after the '*' */
		if (slen == 1)
		{
			token_ptr++;
			MPIDI_FUNC_EXIT(MPID_STATE_CB_CONFIG_LIST_LEX);
			return AGG_WILDCARD;
		}
		else
		{
			MPIDI_FUNC_EXIT(MPID_STATE_CB_CONFIG_LIST_LEX);
			return AGG_ERROR;
		}
	}

	/* last case: some kind of string.  for now we copy the string. */

	/* it would be a good idea to look at the string and make sure that
	 * it doesn't have any illegal characters in it.  in particular we
	 * should ensure that no one tries to use wildcards with strings 
	 * (e.g. "ccn*").
	 */
	MPIU_Wcsncpy(yylval, token_ptr, slen);
	yylval[slen] = L'\0';
	token_ptr += slen;

	MPIDI_FUNC_EXIT(MPID_STATE_CB_CONFIG_LIST_LEX);
	return AGG_STRING;
}
