/*   Copyright (C) 2009 Deino Software. */

#include "mpiimpl.h"

#ifdef HAVE_WEAK_SYMBOLS

#if defined(HAVE_PRAGMA_WEAK)
#pragma weak MPI_File_close = PMPI_File_close
#elif defined(HAVE_PRAGMA_HP_SEC_DEF)
#pragma _HP_SECONDARY_DEF PMPI_File_close MPI_File_close
#elif defined(HAVE_PRAGMA_CRI_DUP)
#pragma _CRI duplicate MPI_File_close as PMPI_File_close
/* end of weak pragmas */
#endif

#endif

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build
   the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#define MPI_File_close PMPI_File_close
#endif

#undef FUNCNAME
#define FUNCNAME MPI_File_close

/*@
	MPI_File_close - Closes a file

Input Parameters:
. fh - file handle (handle)

.N fortran
@*/
MPI_DLL_SPEC int MPI_CALL MPI_File_close(MPI_File *file)
{
	int mpi_errno = MPI_SUCCESS;
	MPID_File *file_ptr = NULL;
	static const wchar_t FCNAME[] = L"MPI_File_close";
	MPIU_THREADPRIV_DECL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_FILE_CLOSE);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPIU_THREAD_SINGLE_CS_ENTER("io");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_FILE_CLOSE);
	MPIU_THREADPRIV_GET;

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_ARGNULL(file, L"file", mpi_errno);
			if (mpi_errno) goto fn_fail;
			MPIR_ERRTEST_FILE(*file, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* Get handles to MPI objects. */
	MPID_File_get_ptr( *file, file_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate file_ptr */
			MPID_File_valid_ptr( file_ptr, mpi_errno );
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_file_close %p", file); }

	MPIU_Free(file_ptr->shared_fp_fname);

	/* need a barrier because the file containing the shared file
	   pointer is opened with COMM_SELF. We don't want it to be
	   deleted while others are still accessing it. */ 
	NMPI_Barrier(mpi_errno, (file_ptr->comm));
	MPIU_ERR_CHK_POP(mpi_errno);
	if (file_ptr->shared_fp_fd != NULL)
	{
		mpi_errno = ADIO_Close(file_ptr->shared_fp_fd);
		MPIU_ERR_CHK_POP(mpi_errno);
	}

	mpi_errno = ADIO_Close(file_ptr);
	MPIU_ERR_CHK_POP(mpi_errno);
	MPIU_Handle_obj_free( &MPID_File_mem, file_ptr);
	*file = MPI_FILE_NULL;

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_FILE_CLOSE);
	MPIU_THREAD_SINGLE_CS_EXIT("io");
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_file_close",
			L"**mpi_file_close %p", file);
	}
#   endif
	mpi_errno = MPIR_Err_return_file(file_ptr, FCNAME, mpi_errno);
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}
