/*   Copyright (C) 2009 Deino Software. */

#include "mpiimpl.h"

/* return the current end of file in etype units relative to the 
   current view */

#undef FUNCNAME
#define FUNCNAME ADIOI_Get_eof_offset
int ADIOI_Get_eof_offset(MPID_File_ptr fd, MPI_Offset *eof_offset)
{
	static const wchar_t FCNAME[] = L"ADIOI_Get_eof_offset";
	int mpi_errno = MPI_SUCCESS;
	int filetype_is_contig, etype_size, filetype_size;
	MPI_Offset fsize, disp, sum=0, size_in_file;
	int n_filetypes, flag, i, rem;
	ADIO_Fcntl_t *fcntl_struct;
	MPI_Aint filetype_extent;
	ADIOI_Flatlist_node *flat_file;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_GET_EOF_OFFSET);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_GET_EOF_OFFSET);

	/* find the eof in bytes */
	fcntl_struct = (ADIO_Fcntl_t *) MPIU_Malloc(sizeof(ADIO_Fcntl_t));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, fcntl_struct, sizeof(ADIO_Fcntl_t), L"fcntl_struct");
	mpi_errno = ADIO_Fcntl(fd, ADIO_FCNTL_GET_FSIZE, fcntl_struct);
	MPIU_ERR_CHK_POP(mpi_errno);
	fsize = fcntl_struct->fsize;
	MPIU_Free(fcntl_struct);
		
	/* Find the offset in etype units corresponding to eof.
	   The eof could lie in a hole in the current view, or in the 
	   middle of an etype. In that case the offset will be the offset
	   corresponding to the start of the next etype in the current view.*/

	mpi_errno = ADIOI_Datatype_iscontig(fd->filetype, &filetype_is_contig);
	MPIU_ERR_CHK_POP(mpi_errno);
	etype_size = fd->etype_size;

	if (filetype_is_contig) 
	{
		*eof_offset = (fsize - fd->disp + etype_size - 1)/etype_size;
		/* ceiling division in case fsize is not a multiple of etype_size;*/
	}
	else
	{
		/* filetype already flattened in ADIO_Open */
		flat_file = ADIOI_Flatlist;
		while (flat_file->type != fd->filetype) 
			flat_file = flat_file->next;

		NMPI_Type_size(mpi_errno, (fd->filetype, &filetype_size));
		MPIU_ERR_CHK_POP(mpi_errno);
		NMPI_Type_extent(mpi_errno, (fd->filetype, &filetype_extent));
		MPIU_ERR_CHK_POP(mpi_errno);

		disp = fd->disp;
		n_filetypes = -1;
		flag = 0;
		while (!flag)
		{
			sum = 0;
			n_filetypes++;
			for (i=0; i<flat_file->count; i++)
			{
				sum += flat_file->blocklens[i];
				if (disp + flat_file->indices[i] + 
					(MPI_Offset) n_filetypes*filetype_extent + 
					   flat_file->blocklens[i] >= fsize)
				{
					if (disp + flat_file->indices[i] + (MPI_Offset) n_filetypes*filetype_extent >= fsize)
					{
						sum -= flat_file->blocklens[i];
					}
					else
					{
						rem = (int) (disp + flat_file->indices[i] + 
								(MPI_Offset) n_filetypes*filetype_extent
								+ flat_file->blocklens[i] - fsize);
						sum -= rem;
					}
					flag = 1;
					break;
				}
			}
		}
		size_in_file = (MPI_Offset) n_filetypes*filetype_size + sum;
		*eof_offset = (size_in_file+etype_size-1)/etype_size; /* ceiling division */
	}

fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_GET_EOF_OFFSET);
	return mpi_errno;
fn_fail:
	goto fn_exit;
}
