/*   Copyright (C) 2009 Deino Software. */

#include "mpiimpl.h"

int ADIOI_Optimize_flattened(ADIOI_Flatlist_node *flat_type);
int ADIOI_Flatten_copy_type(ADIOI_Flatlist_node *flat,
							 int old_type_start,
							 int old_type_end,
							 int new_type_start,
							 MPI_Offset offset_adjustment);

/* flatten datatype and add it to Flatlist */
#undef FUNCNAME
#define FUNCNAME ADIOI_Flatten_datatype
int ADIOI_Flatten_datatype(MPI_Datatype datatype)
{
	static const wchar_t FCNAME[] = L"ADIOI_Flatten_datatype";
	int mpi_errno = MPI_SUCCESS;
#ifdef HAVE_MPIR_TYPE_FLATTEN
	MPI_Aint flatten_idx;
#endif
	int curr_index=0, is_contig;
	ADIOI_Flatlist_node *flat, *prev=0;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_FLATTEN_DATATYPE);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_FLATTEN_DATATYPE);

	/* check if necessary to flatten. */
 
	/* is it entirely contiguous? */
	mpi_errno = ADIOI_Datatype_iscontig(datatype, &is_contig);
	MPIU_ERR_CHK_POP(mpi_errno);
	if (is_contig)
	{
		goto fn_exit;
	}

	/* has it already been flattened? */
	flat = ADIOI_Flatlist;
	while (flat)
	{
		if (flat->type == datatype)
		{
			goto fn_exit;
		}
		else
		{
			prev = flat;
			flat = flat->next;
		}
	}

	/* flatten and add to the list */
	flat = prev;
	flat->next = (ADIOI_Flatlist_node *)MPIU_Malloc(sizeof(ADIOI_Flatlist_node));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, flat->next, sizeof(ADIOI_Flatlist_node), L"flat->next");
	flat = flat->next;

	flat->type = datatype;
	flat->next = NULL;
	flat->blocklens = NULL;
	flat->indices = NULL;

	mpi_errno = ADIOI_Count_contiguous_blocks(datatype, &curr_index, &flat->count);
	MPIU_ERR_CHK_POP(mpi_errno);
#if 0
	FPRINTF(stderr, L"cur_idx = %d\n", curr_index);
#endif
/*	FPRINTF(stderr, L"%d\n", flat->count);*/

	if (flat->count)
	{
		flat->blocklens = (int *) MPIU_Malloc(flat->count * sizeof(int));
		MPIU_ERR_MEM_CHK_POP(mpi_errno, flat->blocklens, flat->count * sizeof(int), L"flat->blocklens");
		flat->indices = (MPI_Offset *) MPIU_Malloc(flat->count * sizeof(MPI_Offset));
		MPIU_ERR_MEM_CHK_POP(mpi_errno, flat->indices, flat->count * sizeof(MPI_Offset), L"flat->indices");
	}
		
	curr_index = 0;
#ifdef HAVE_MPIR_TYPE_FLATTEN
	flatten_idx = (MPI_Aint) flat->count;
	MPIR_Type_flatten(datatype, flat->indices, flat->blocklens, &flatten_idx);
#else
	mpi_errno = ADIOI_Flatten(datatype, flat, 0, &curr_index);
	MPIU_ERR_CHK_POP(mpi_errno);

	mpi_errno = ADIOI_Optimize_flattened(flat);
	MPIU_ERR_CHK_POP(mpi_errno);
#endif
/* debug */
#if 0
	{
		int i;
		FPRINTF(stderr, L"blens: ");
		for (i=0; i<flat->count; i++) 
			FPRINTF(stderr, L"%d ", flat->blocklens[i]);
		FPRINTF(stderr, L"\n\n");
		FPRINTF(stderr, L"indices: ");
		for (i=0; i<flat->count; i++) 
			FPRINTF(stderr, L"%ld ", (long) flat->indices[i]);
		FPRINTF(stderr, L"\n\n");
	}
#endif

fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_FLATTEN_DATATYPE);
	return mpi_errno;
fn_fail:
	goto fn_exit;
}

/* ADIOI_Flatten()
 *
 * Assumption: input datatype is not a basic!!!!
 */
#undef FUNCNAME
#define FUNCNAME ADIOI_Flatten
int ADIOI_Flatten(MPI_Datatype datatype, ADIOI_Flatlist_node *flat, 
				  MPI_Offset st_offset, int *curr_index)  
{
	static const wchar_t FCNAME[] = L"ADIOI_Flatten";
	int mpi_errno = MPI_SUCCESS;
	int i, j, k, m, n, num, basic_num, prev_index;
	int top_count, combiner, old_combiner, old_is_contig;
	int old_size, nints, nadds, ntypes, old_nints, old_nadds, old_ntypes;
	MPI_Aint old_extent;
	int *ints;
	MPI_Aint *adds;
	MPI_Datatype *types;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_FLATTEN);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_FLATTEN);

	NMPI_Type_get_envelope(mpi_errno, (datatype, &nints, &nadds, &ntypes, &combiner));
	MPIU_ERR_CHK_POP(mpi_errno);
	ints = (int *) MPIU_Malloc((nints+1)*sizeof(int));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, ints, (nints+1)*sizeof(int), L"ints");
	adds = (MPI_Aint *) MPIU_Malloc((nadds+1)*sizeof(MPI_Aint));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, adds, (nadds+1) * sizeof(MPI_Aint), L"adds");
	types = (MPI_Datatype *) MPIU_Malloc((ntypes+1)*sizeof(MPI_Datatype));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, types, (ntypes+1) * sizeof(MPI_Datatype), L"types");
	NMPI_Type_get_contents(mpi_errno, (datatype, nints, nadds, ntypes, ints, adds, types));
	MPIU_ERR_CHK_POP(mpi_errno);

	switch (combiner)
	{
#ifdef MPIIMPL_HAVE_MPI_COMBINER_DUP
	case MPI_COMBINER_DUP:
		NMPI_Type_get_envelope(mpi_errno, (types[0], &old_nints, &old_nadds,
							  &old_ntypes, &old_combiner));
		MPIU_ERR_CHK_POP(mpi_errno);
		mpi_errno = ADIOI_Datatype_iscontig(types[0], &old_is_contig);
		MPIU_ERR_CHK_POP(mpi_errno);
		if ((old_combiner != MPI_COMBINER_NAMED) && (!old_is_contig))
		{
			mpi_errno = ADIOI_Flatten(types[0], flat, st_offset, curr_index);
			MPIU_ERR_CHK_POP(mpi_errno);
		}
		break;
#endif
#ifdef MPIIMPL_HAVE_MPI_COMBINER_SUBARRAY
	case MPI_COMBINER_SUBARRAY:
		{
			int dims = ints[0];
			MPI_Datatype stype;

			mpi_errno = ADIO_Type_create_subarray(dims,
									  &ints[1],		/* sizes */
									  &ints[dims+1],   /* subsizes */
									  &ints[2*dims+1], /* starts */
									  ints[3*dims+1],  /* order */
									  types[0],		/* type */
									  &stype);
			MPIU_ERR_CHK_POP(mpi_errno);
			mpi_errno = ADIOI_Flatten(stype, flat, st_offset, curr_index);
			MPIU_ERR_CHK_POP(mpi_errno);
			NMPI_Type_free(mpi_errno, (&stype));
			MPIU_ERR_CHK_POP(mpi_errno);
		}
		break;
#endif
#ifdef MPIIMPL_HAVE_MPI_COMBINER_DARRAY
	case MPI_COMBINER_DARRAY:
		{
			int dims = ints[2];
			MPI_Datatype dtype;

			mpi_errno = ADIO_Type_create_darray(ints[0],		 /* size */
									ints[1],		 /* rank */
									dims,
									&ints[3],		/* gsizes */
									&ints[dims+3],   /* distribs */
									&ints[2*dims+3], /* dargs */
									&ints[3*dims+3], /* psizes */
									ints[4*dims+3],  /* order */
									types[0],
									&dtype);
			MPIU_ERR_CHK_POP(mpi_errno);
			mpi_errno = ADIOI_Flatten(dtype, flat, st_offset, curr_index);
			MPIU_ERR_CHK_POP(mpi_errno);
			NMPI_Type_free(mpi_errno, (&dtype));
			MPIU_ERR_CHK_POP(mpi_errno);
		}
		break;
#endif
	case MPI_COMBINER_CONTIGUOUS:
		top_count = ints[0];
		NMPI_Type_get_envelope(mpi_errno, (types[0], &old_nints, &old_nadds,
							  &old_ntypes, &old_combiner));
		MPIU_ERR_CHK_POP(mpi_errno);
		mpi_errno = ADIOI_Datatype_iscontig(types[0], &old_is_contig);
		MPIU_ERR_CHK_POP(mpi_errno);

		prev_index = *curr_index;
		if ((old_combiner != MPI_COMBINER_NAMED) && (!old_is_contig))
		{
			mpi_errno = ADIOI_Flatten(types[0], flat, st_offset, curr_index);
			MPIU_ERR_CHK_POP(mpi_errno);
		}

		if (prev_index == *curr_index)
		{
			/* simplest case, made up of basic or contiguous types */
			j = *curr_index;
			flat->indices[j] = st_offset;
			NMPI_Type_size(mpi_errno, (types[0], &old_size));
			MPIU_ERR_CHK_POP(mpi_errno);
			flat->blocklens[j] = top_count * old_size;
			(*curr_index)++;
		}
		else
		{
			/* made up of noncontiguous derived types */
			j = *curr_index;
			num = *curr_index - prev_index;

			/* The noncontiguous types have to be replicated count times */
			NMPI_Type_extent(mpi_errno, (types[0], &old_extent));
			MPIU_ERR_CHK_POP(mpi_errno);
			for (m=1; m<top_count; m++)
			{
				for (i=0; i<num; i++)
				{
					flat->indices[j] = flat->indices[j-num] + old_extent;
					flat->blocklens[j] = flat->blocklens[j-num];
					j++;
				}
			}
			*curr_index = j;
		}
		break;

	case MPI_COMBINER_VECTOR: 
		top_count = ints[0];
		NMPI_Type_get_envelope(mpi_errno, (types[0], &old_nints, &old_nadds,
							  &old_ntypes, &old_combiner));
		MPIU_ERR_CHK_POP(mpi_errno);
		mpi_errno = ADIOI_Datatype_iscontig(types[0], &old_is_contig);
		MPIU_ERR_CHK_POP(mpi_errno);

		prev_index = *curr_index;
		if ((old_combiner != MPI_COMBINER_NAMED) && (!old_is_contig))
		{
			mpi_errno = ADIOI_Flatten(types[0], flat, st_offset, curr_index);
			MPIU_ERR_CHK_POP(mpi_errno);
		}

		if (prev_index == *curr_index)
		{
			/* simplest case, vector of basic or contiguous types */
			j = *curr_index;
			flat->indices[j] = st_offset;
			NMPI_Type_size(mpi_errno, (types[0], &old_size));
			MPIU_ERR_CHK_POP(mpi_errno);
			flat->blocklens[j] = ints[1] * old_size;
			for (i=j+1; i<j+top_count; i++)
			{
				//flat->indices[i] = flat->indices[i-1] + ints[2]*old_size;
				flat->indices[i] = flat->indices[i-1] + (unsigned) ints[2] * (unsigned) old_size;
				flat->blocklens[i] = flat->blocklens[j];
			}
			*curr_index = i;
		}
		else
		{
			/* vector of noncontiguous derived types */

			j = *curr_index;
			num = *curr_index - prev_index;

			/* The noncontiguous types have to be replicated blocklen times
			   and then strided. Replicate the first one. */
			NMPI_Type_extent(mpi_errno, (types[0], &old_extent));
			MPIU_ERR_CHK_POP(mpi_errno);
			for (m=1; m<ints[1]; m++)
			{
				for (i=0; i<num; i++)
				{
					flat->indices[j] = flat->indices[j-num] + old_extent;
					flat->blocklens[j] = flat->blocklens[j-num];
					j++;
				}
			}
			*curr_index = j;

			/* Now repeat with strides. */
			num = *curr_index - prev_index;
			for (i=1; i<top_count; i++)
			{
 				for (m=0; m<num; m++)
				{
				   flat->indices[j] =  flat->indices[j-num] + ints[2]
					   *old_extent;
				   flat->blocklens[j] = flat->blocklens[j-num];
				   j++;
				}
			}
			*curr_index = j;
		}
		break;

	case MPI_COMBINER_HVECTOR: 
		top_count = ints[0];
		NMPI_Type_get_envelope(mpi_errno, (types[0], &old_nints, &old_nadds,
							  &old_ntypes, &old_combiner));
		MPIU_ERR_CHK_POP(mpi_errno);
		mpi_errno = ADIOI_Datatype_iscontig(types[0], &old_is_contig);
		MPIU_ERR_CHK_POP(mpi_errno);

		prev_index = *curr_index;
		if ((old_combiner != MPI_COMBINER_NAMED) && (!old_is_contig))
		{
			mpi_errno = ADIOI_Flatten(types[0], flat, st_offset, curr_index);
			MPIU_ERR_CHK_POP(mpi_errno);
		}

		if (prev_index == *curr_index)
		{
			/* simplest case, vector of basic or contiguous types */
			j = *curr_index;
			flat->indices[j] = st_offset;
			NMPI_Type_size(mpi_errno, (types[0], &old_size));
			MPIU_ERR_CHK_POP(mpi_errno);
			flat->blocklens[j] = ints[1] * old_size;
			for (i=j+1; i<j+top_count; i++)
			{
				flat->indices[i] = flat->indices[i-1] + adds[0];
				flat->blocklens[i] = flat->blocklens[j];
			}
			*curr_index = i;
		}
		else
		{
			/* vector of noncontiguous derived types */

			j = *curr_index;
			num = *curr_index - prev_index;

			/* The noncontiguous types have to be replicated blocklen times
			   and then strided. Replicate the first one. */
			NMPI_Type_extent(mpi_errno, (types[0], &old_extent));
			MPIU_ERR_CHK_POP(mpi_errno);
			for (m=1; m<ints[1]; m++)
			{
				for (i=0; i<num; i++)
				{
					flat->indices[j] = flat->indices[j-num] + old_extent;
					flat->blocklens[j] = flat->blocklens[j-num];
					j++;
				}
			}
			*curr_index = j;

			/* Now repeat with strides. */
			num = *curr_index - prev_index;
			for (i=1; i<top_count; i++)
			{
 				for (m=0; m<num; m++)
				{
				   flat->indices[j] =  flat->indices[j-num] + adds[0];
				   flat->blocklens[j] = flat->blocklens[j-num];
				   j++;
				}
			}
			*curr_index = j;
		}
		break;

	case MPI_COMBINER_INDEXED: 
		top_count = ints[0];
		NMPI_Type_get_envelope(mpi_errno, (types[0], &old_nints, &old_nadds,
							  &old_ntypes, &old_combiner));
		MPIU_ERR_CHK_POP(mpi_errno);
		mpi_errno = ADIOI_Datatype_iscontig(types[0], &old_is_contig);
		MPIU_ERR_CHK_POP(mpi_errno);
		NMPI_Type_extent(mpi_errno, (types[0], &old_extent));
		MPIU_ERR_CHK_POP(mpi_errno);

		prev_index = *curr_index;
		if ((old_combiner != MPI_COMBINER_NAMED) && (!old_is_contig))
		{
			mpi_errno = ADIOI_Flatten(types[0], flat,
						 st_offset+ints[top_count+1]*old_extent, curr_index);
		}

		if (prev_index == *curr_index)
		{
			/* simplest case, indexed type made up of basic or contiguous types */
			j = *curr_index;
			for (i=j; i<j+top_count; i++)
			{
				flat->indices[i] = st_offset + ints[top_count+1+i-j]*old_extent;
				flat->blocklens[i] = (int) (ints[1+i-j]*old_extent);
			}
			*curr_index = i;
		}
		else
		{
			/* indexed type made up of noncontiguous derived types */

			j = *curr_index;
			num = *curr_index - prev_index;
			basic_num = num;

			/* The noncontiguous types have to be replicated blocklens[i] times
			   and then strided. Replicate the first one. */
			for (m=1; m<ints[1]; m++)
			{
				for (i=0; i<num; i++)
				{
					flat->indices[j] = flat->indices[j-num] + old_extent;
					flat->blocklens[j] = flat->blocklens[j-num];
					j++;
				}
			}
			*curr_index = j;

			/* Now repeat with strides. */
			for (i=1; i<top_count; i++)
			{
				num = *curr_index - prev_index;
				prev_index = *curr_index;
				for (m=0; m<basic_num; m++)
				{
					flat->indices[j] = flat->indices[j-num] + 
						(ints[top_count+1+i]-ints[top_count+i])*old_extent;
					flat->blocklens[j] = flat->blocklens[j-num];
					j++;
				}
				*curr_index = j;
				for (m=1; m<ints[1+i]; m++)
				{
					for (k=0; k<basic_num; k++)
					{
						flat->indices[j] = flat->indices[j-basic_num] + old_extent;
						flat->blocklens[j] = flat->blocklens[j-basic_num];
						j++;
					}
				}
				*curr_index = j;
			}
		}
		break;

	case MPI_COMBINER_HINDEXED: 
		top_count = ints[0];
		NMPI_Type_get_envelope(mpi_errno, (types[0], &old_nints, &old_nadds,
							  &old_ntypes, &old_combiner));
		MPIU_ERR_CHK_POP(mpi_errno);
		mpi_errno = ADIOI_Datatype_iscontig(types[0], &old_is_contig);
		MPIU_ERR_CHK_POP(mpi_errno);

		prev_index = *curr_index;
		if ((old_combiner != MPI_COMBINER_NAMED) && (!old_is_contig))
		{
			mpi_errno = ADIOI_Flatten(types[0], flat, st_offset+adds[0], curr_index);
			MPIU_ERR_CHK_POP(mpi_errno);
		}

		if (prev_index == *curr_index)
		{
			/* simplest case, indexed type made up of basic or contiguous types */
			j = *curr_index;
			NMPI_Type_size(mpi_errno, (types[0], &old_size));
			MPIU_ERR_CHK_POP(mpi_errno);
			for (i=j; i<j+top_count; i++)
			{
				flat->indices[i] = st_offset + adds[i-j];
				flat->blocklens[i] = ints[1+i-j]*old_size;
			}
			*curr_index = i;
		}
		else
		{
			/* indexed type made up of noncontiguous derived types */

			j = *curr_index;
			num = *curr_index - prev_index;
			basic_num = num;

			/* The noncontiguous types have to be replicated blocklens[i] times
			   and then strided. Replicate the first one. */
			NMPI_Type_extent(mpi_errno, (types[0], &old_extent));
			MPIU_ERR_CHK_POP(mpi_errno);
			for (m=1; m<ints[1]; m++)
			{
				for (i=0; i<num; i++)
				{
					flat->indices[j] = flat->indices[j-num] + old_extent;
					flat->blocklens[j] = flat->blocklens[j-num];
					j++;
				}
			}
			*curr_index = j;

			/* Now repeat with strides. */
			for (i=1; i<top_count; i++)
			{
				num = *curr_index - prev_index;
				prev_index = *curr_index;
				for (m=0; m<basic_num; m++)
				{
					flat->indices[j] = flat->indices[j-num] + adds[i] - adds[i-1];
					flat->blocklens[j] = flat->blocklens[j-num];
					j++;
				}
				*curr_index = j;
				for (m=1; m<ints[1+i]; m++)
				{
					for (k=0; k<basic_num; k++)
					{
						flat->indices[j] = flat->indices[j-basic_num] + old_extent;
						flat->blocklens[j] = flat->blocklens[j-basic_num];
					j++;
					}
				}
				*curr_index = j;
			}
		}
		break;

	case MPI_COMBINER_STRUCT: 
		top_count = ints[0];
		for (n=0; n<top_count; n++)
		{
			NMPI_Type_get_envelope(mpi_errno, (types[n], &old_nints, &old_nadds,
								  &old_ntypes, &old_combiner));
			MPIU_ERR_CHK_POP(mpi_errno);
			mpi_errno = ADIOI_Datatype_iscontig(types[n], &old_is_contig);
			MPIU_ERR_CHK_POP(mpi_errno);

			prev_index = *curr_index;
			if ((old_combiner != MPI_COMBINER_NAMED) && (!old_is_contig))
			{
				mpi_errno = ADIOI_Flatten(types[n], flat, st_offset+adds[n], curr_index);
				MPIU_ERR_CHK_POP(mpi_errno);
			}

			if (prev_index == *curr_index)
			{
				/* simplest case, current type is basic or contiguous types */
				j = *curr_index;
				flat->indices[j] = st_offset + adds[n];
				NMPI_Type_size(mpi_errno, (types[n], &old_size));
				MPIU_ERR_CHK_POP(mpi_errno);
				flat->blocklens[j] = ints[1+n] * old_size;
				(*curr_index)++;
			}
			else
			{
				/* current type made up of noncontiguous derived types */

				j = *curr_index;
				num = *curr_index - prev_index;

				/* The current type has to be replicated blocklens[n] times */
				NMPI_Type_extent(mpi_errno, (types[n], &old_extent));
				MPIU_ERR_CHK_POP(mpi_errno);
				for (m=1; m<ints[1+n]; m++)
				{
					for (i=0; i<num; i++)
					{
						flat->indices[j] = flat->indices[j-num] + old_extent;
						flat->blocklens[j] = flat->blocklens[j-num];
						j++;
					}
				}
				*curr_index = j;
			}
		}
 		break;

	default:
		/*FPRINTF(stderr, L"Error: Unsupported datatype passed to ADIOI_Flatten\n");*/
		MPIU_ERR_SET(mpi_errno, MPI_ERR_OTHER, L"**arg");
		goto fn_fail;
	}

	for (i=0; i<ntypes; i++)
	{
 		NMPI_Type_get_envelope(mpi_errno, (types[i], &old_nints, &old_nadds, &old_ntypes, &old_combiner));
		MPIU_ERR_CHK_POP(mpi_errno);
 		if (old_combiner != MPI_COMBINER_NAMED)
		{
			NMPI_Type_free(mpi_errno, (types+i));
			MPIU_ERR_CHK_POP(mpi_errno);
		}
	}

	MPIU_Free(ints);
	MPIU_Free(adds);
	MPIU_Free(types);

fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_FLATTEN);
	return mpi_errno;
fn_fail:
	goto fn_exit;
}

/********************************************************/

/* ADIOI_Count_contiguous_blocks
 *
 * Returns number of contiguous blocks in type, and also updates
 * curr_index to reflect the space for the additional blocks.
 *
 * ASSUMES THAT TYPE IS NOT A BASIC!!!
 */
#ifdef HAVE_MPIR_TYPE_GET_CONTIG_BLOCKS

#undef FUNCNAME
#define FUNCNAME ADIOI_Count_contiguous_blocks
int ADIOI_Count_contiguous_blocks(MPI_Datatype datatype, int *curr_index, int *contig_blocks_ptr)
{
	static const wchar_t FCNAME[] = L"ADIOI_Count_contiguous_blocks";
	int mpi_errno = MPI_SUCCESS;
	int blks;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_COUNT_CONTIGUOUS_BLOCKS);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_COUNT_CONTIGUOUS_BLOCKS);

	mpi_errno = MPIR_Type_get_contig_blocks(datatype, &blks);
	MPIU_ERR_CHK_POP(mpi_errno);
	*curr_index = blks;
	*contig_blocks_ptr = blks;

fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_COUNT_CONTIGUOUS_BLOCKS);
	return mpi_errno;
fn_fail:
	goto fn_exit;
}

#else

#undef FUNCNAME
#define FUNCNAME ADIOI_Count_contiguous_blocks
int ADIOI_Count_contiguous_blocks(MPI_Datatype datatype, int *curr_index, int *contig_blocks_ptr)
{
	static const wchar_t FCNAME[] = L"ADIOI_Count_contiguous_blocks";
	int mpi_errno = MPI_SUCCESS;
	int count=0, i, n, num, basic_num, prev_index;
	int top_count, combiner, old_combiner, old_is_contig;
	int nints, nadds, ntypes, old_nints, old_nadds, old_ntypes;
	int *ints;
	MPI_Aint *adds;
	MPI_Datatype *types;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_COUNT_CONTIGUOUS_BLOCKS);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_COUNT_CONTIGUOUS_BLOCKS);

	NMPI_Type_get_envelope(mpi_errno, (datatype, &nints, &nadds, &ntypes, &combiner));
	MPIU_ERR_CHK_POP(mpi_errno);
	ints = (int *) MPIU_Malloc((nints+1)*sizeof(int));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, ints, (nints+1)*sizeof(int), L"ints");
	adds = (MPI_Aint *) MPIU_Malloc((nadds+1)*sizeof(MPI_Aint));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, adds, (nadds+1)*sizeof(MPI_Aint), L"adds");
	types = (MPI_Datatype *) MPIU_Malloc((ntypes+1)*sizeof(MPI_Datatype));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, types, (ntypes+1)*sizeof(MPI_Datatype), L"types");
	NMPI_Type_get_contents(mpi_errno, (datatype, nints, nadds, ntypes, ints, adds, types));

	switch (combiner) {
#ifdef MPIIMPL_HAVE_MPI_COMBINER_DUP
	case MPI_COMBINER_DUP:
		NMPI_Type_get_envelope(mpi_errno, (types[0], &old_nints, &old_nadds,
							  &old_ntypes, &old_combiner));
		mpi_errno = ADIOI_Datatype_iscontig(types[0], &old_is_contig);
		MPIU_ERR_CHK_POP(mpi_errno);
		if ((old_combiner != MPI_COMBINER_NAMED) && (!old_is_contig))
		{
			mpi_errno = ADIOI_Count_contiguous_blocks(types[0], curr_index, &count);
			MPIU_ERR_CHK_POP(mpi_errno);
		}
		else
		{
			count = 1;
			(*curr_index)++;
		}
		break;
#endif
#ifdef MPIIMPL_HAVE_MPI_COMBINER_SUBARRAY
	case MPI_COMBINER_SUBARRAY:
		{
			int dims = ints[0];
			MPI_Datatype stype;

			mpi_errno = ADIO_Type_create_subarray(dims,
									  &ints[1],		/* sizes */
									  &ints[dims+1],   /* subsizes */
									  &ints[2*dims+1], /* starts */
									  ints[3*dims+1],  /* order */
									  types[0],		/* type */
									  &stype);
			MPIU_ERR_CHK_POP(mpi_errno);
			mpi_errno = ADIOI_Count_contiguous_blocks(stype, curr_index, &count);
			MPIU_ERR_CHK_POP(mpi_errno);
			/* curr_index will have already been updated; just pass
			 * count back up.
			 */
			NMPI_Type_free(mpi_errno, (&stype));
			MPIU_ERR_CHK_POP(mpi_errno);
		}
		break;
#endif
#ifdef MPIIMPL_HAVE_MPI_COMBINER_DARRAY
	case MPI_COMBINER_DARRAY:
		{
			int dims = ints[2];
			MPI_Datatype dtype;

			mpi_errno = ADIO_Type_create_darray(ints[0],		 /* size */
									ints[1],		 /* rank */
									dims,
									&ints[3],		/* gsizes */
									&ints[dims+3],   /* distribs */
									&ints[2*dims+3], /* dargs */
									&ints[3*dims+3], /* psizes */
									ints[4*dims+3],  /* order */
									types[0],
									&dtype);
			MPIU_ERR_CHK_POP(mpi_errno);
			mpi_errno = ADIOI_Count_contiguous_blocks(dtype, curr_index, &count);
			MPIU_ERR_CHK_POP(mpi_errno);
			/* curr_index will have already been updated; just pass
			 * count back up.
			 */
			NMPI_Type_free(mpi_errno, (&dtype));
			MPIU_ERR_CHK_POP(mpi_errno);
		}
		break;
#endif
	case MPI_COMBINER_CONTIGUOUS:
		top_count = ints[0];
		NMPI_Type_get_envelope(mpi_errno, (types[0], &old_nints, &old_nadds,
							  &old_ntypes, &old_combiner));
		MPIU_ERR_CHK_POP(mpi_errno);
		mpi_errno = ADIOI_Datatype_iscontig(types[0], &old_is_contig);
		MPIU_ERR_CHK_POP(mpi_errno);

		prev_index = *curr_index;
		if ((old_combiner != MPI_COMBINER_NAMED) && (!old_is_contig))
		{
			mpi_errno = ADIOI_Count_contiguous_blocks(types[0], curr_index, &count);
			MPIU_ERR_CHK_POP(mpi_errno);
		}
		else
		{
			count = 1;
		}

		if (prev_index == *curr_index)
		{
			/* simplest case, made up of basic or contiguous types */
			(*curr_index)++;
		}
		else
		{
			/* made up of noncontiguous derived types */
			num = *curr_index - prev_index;
			count *= top_count;
			*curr_index += (top_count - 1)*num;
		}
		break;

	case MPI_COMBINER_VECTOR:
	case MPI_COMBINER_HVECTOR:
		top_count = ints[0];
		NMPI_Type_get_envelope(mpi_errno, (types[0], &old_nints, &old_nadds,
							  &old_ntypes, &old_combiner));
		MPIU_ERR_CHK_POP(mpi_errno);
		mpi_errno = ADIOI_Datatype_iscontig(types[0], &old_is_contig);
		MPIU_ERR_CHK_POP(mpi_errno);

		prev_index = *curr_index;
		if ((old_combiner != MPI_COMBINER_NAMED) && (!old_is_contig))
		{
			mpi_errno = ADIOI_Count_contiguous_blocks(types[0], curr_index, &count);
			MPIU_ERR_CHK_POP(mpi_errno);
		}
		else
		{
			count = 1;
		}

		if (prev_index == *curr_index)
		{
			/* simplest case, vector of basic or contiguous types */
			count = top_count;
			*curr_index += count;
		}
		else
		{
			/* vector of noncontiguous derived types */
			num = *curr_index - prev_index;

			/* The noncontiguous types have to be replicated blocklen times
			   and then strided. */
			count *= ints[1] * top_count;

			/* First one */
			*curr_index += (ints[1] - 1)*num;

			/* Now repeat with strides. */
			num = *curr_index - prev_index;
			*curr_index += (top_count - 1)*num;
		}
		break;

	case MPI_COMBINER_INDEXED: 
	case MPI_COMBINER_HINDEXED:
		top_count = ints[0];
		NMPI_Type_get_envelope(mpi_errno, (types[0], &old_nints, &old_nadds,
							  &old_ntypes, &old_combiner));
		MPIU_ERR_CHK_POP(mpi_errno);
		mpi_errno = ADIOI_Datatype_iscontig(types[0], &old_is_contig);
		MPIU_ERR_CHK_POP(mpi_errno);

		prev_index = *curr_index;
		if ((old_combiner != MPI_COMBINER_NAMED) && (!old_is_contig))
		{
			mpi_errno = ADIOI_Count_contiguous_blocks(types[0], curr_index, &count);
			MPIU_ERR_CHK_POP(mpi_errno);
		}
		else
		{
			count = 1;
		}

		if (prev_index == *curr_index)
		{
			/* simplest case, indexed type made up of basic or contiguous types */
			count = top_count;
			*curr_index += count;
		}
		else
		{
			/* indexed type made up of noncontiguous derived types */
			basic_num = *curr_index - prev_index;

			/* The noncontiguous types have to be replicated blocklens[i] times
			   and then strided. */
			*curr_index += (ints[1]-1) * basic_num;
			count *= ints[1];

			/* Now repeat with strides. */
			for (i=1; i<top_count; i++)
			{
				count += ints[1+i] * basic_num;
				*curr_index += ints[1+i] * basic_num;
			}
		}
		break;

	case MPI_COMBINER_STRUCT: 
		top_count = ints[0];
		count = 0;
		for (n=0; n<top_count; n++)
		{
			NMPI_Type_get_envelope(mpi_errno, (types[n], &old_nints, &old_nadds,
								  &old_ntypes, &old_combiner));
			MPIU_ERR_CHK_POP(mpi_errno);
			mpi_errno = ADIOI_Datatype_iscontig(types[n], &old_is_contig);
			MPIU_ERR_CHK_POP(mpi_errno);

			prev_index = *curr_index;
			if ((old_combiner != MPI_COMBINER_NAMED) && (!old_is_contig))
			{
				int temp_count;
				mpi_errno = ADIOI_Count_contiguous_blocks(types[n], curr_index, &temp_count);
				MPIU_ERR_CHK_POP(mpi_errno);
				count += temp_count;
			}

			if (prev_index == *curr_index)
			{
				/* simplest case, current type is basic or contiguous types */
				count++;
				(*curr_index)++;
			}
			else
			{
				/* current type made up of noncontiguous derived types */
				/* The current type has to be replicated blocklens[n] times */

				num = *curr_index - prev_index;
				count += (ints[1+n]-1)*num;
				(*curr_index) += (ints[1+n]-1)*num;
			}
		}
		break;
	default:
		MPIU_ERR_SET(mpi_errno, MPI_ERR_OTHER, L"**arg");
		goto fn_fail;
	}


	for (i=0; i<ntypes; i++)
	{
 		NMPI_Type_get_envelope(mpi_errno, (types[i], &old_nints, &old_nadds, &old_ntypes,
 							  &old_combiner));
		MPIU_ERR_CHK_POP(mpi_errno);
 		if (old_combiner != MPI_COMBINER_NAMED)
		{
			NMPI_Type_free(mpi_errno, (types+i));
			MPIU_ERR_CHK_POP(mpi_errno);
		}
	}

	MPIU_Free(ints);
	MPIU_Free(adds);
	MPIU_Free(types);
	*contig_blocks_ptr = count;

fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_COUNT_CONTIGUOUS_BLOCKS);
	return mpi_errno;
fn_fail:
	goto fn_exit;
}
#endif /* HAVE_MPIR_TYPE_GET_CONTIG_BLOCKS */



/****************************************************************/

/* ADIOI_Optimize_flattened()
 *
 * Scans the blocks of a flattened type and merges adjacent blocks 
 * together, resulting in a shorter blocklist (and thus fewer
 * contiguous operations).
 *
 * Q: IS IT SAFE TO REMOVE THE 0-LENGTH BLOCKS TOO?
 */
#undef FUNCNAME
#define FUNCNAME ADIOI_Optimize_flattened
int ADIOI_Optimize_flattened(ADIOI_Flatlist_node *flat_type)
{
	static const wchar_t FCNAME[] = L"ADIOI_Optimize_flattened";
	int mpi_errno = MPI_SUCCESS;
	int i, j, opt_blocks;
	int *opt_blocklens;
	MPI_Offset *opt_indices;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_OPTIMIZE_FLATTENED);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_OPTIMIZE_FLATTENED);

	opt_blocks = 1;
	
	/* save number of noncontiguous blocks in opt_blocks */
	for (i=0; i < (flat_type->count - 1); i++)
	{
		if ((flat_type->indices[i] + flat_type->blocklens[i] !=
			 flat_type->indices[i + 1]))
			opt_blocks++;
	}

	/* if we can't reduce the number of blocks, quit now */
	if (opt_blocks == flat_type->count)
	{
		goto fn_exit;
	}

	opt_blocklens = (int *) MPIU_Malloc(opt_blocks * sizeof(int));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, opt_blocklens, opt_blocks * sizeof(int), L"opt_blocklens");
	opt_indices = (MPI_Offset *)MPIU_Malloc(opt_blocks*sizeof(MPI_Offset));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, opt_indices, opt_blocks*sizeof(MPI_Offset), L"opt_indices");

	/* fill in new blocklists */
	opt_blocklens[0] = flat_type->blocklens[0];
	opt_indices[0] = flat_type->indices[0];
	j = 0;
	for (i=0; i < (flat_type->count - 1); i++)
	{
		if ((flat_type->indices[i] + flat_type->blocklens[i] == flat_type->indices[i + 1]))
		{
			opt_blocklens[j] += flat_type->blocklens[i + 1];
		}
		else
		{
			j++;
			opt_indices[j] = flat_type->indices[i + 1];
			opt_blocklens[j] = flat_type->blocklens[i + 1];
		} 
	}
	flat_type->count = opt_blocks;
	MPIU_Free(flat_type->blocklens);
	MPIU_Free(flat_type->indices);
	flat_type->blocklens = opt_blocklens;
	flat_type->indices = opt_indices;

fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_OPTIMIZE_FLATTENED);
	return mpi_errno;
fn_fail:
	goto fn_exit;
}

#undef FUNCNAME
#define FUNCNAME ADIOI_Delete_flattened
int ADIOI_Delete_flattened(MPI_Datatype datatype)
{
	/*static const wchar_t FCNAME[] = L"ADIOI_Delete_flattened";*/
	int mpi_errno = MPI_SUCCESS;
	ADIOI_Flatlist_node *flat, *prev;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_DELETE_FLATTENED);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_DELETE_FLATTENED);

	prev = flat = ADIOI_Flatlist;
	while (flat && (flat->type != datatype))
	{
		prev = flat;
		flat = flat->next;
	}
	if (flat)
	{
		prev->next = flat->next;
		if (flat->blocklens) MPIU_Free(flat->blocklens);
		if (flat->indices) MPIU_Free(flat->indices);
		MPIU_Free(flat);
	}

	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_DELETE_FLATTENED);
	return mpi_errno;
}

/* ADIOI_Flatten_copy_type()
 * flat - pointer to flatlist node holding offset and lengths
 * start - starting index of src type in arrays
 * end - one larger than ending index of src type (makes loop clean)
 * offset_adjustment - amount to add to "indices" (offset) component
 *					 of each off/len pair copied
 */
#undef FUNCNAME
#define FUNCNAME ADIOI_Flatten_copy_type
int ADIOI_Flatten_copy_type(ADIOI_Flatlist_node *flat,
							 int old_type_start,
							 int old_type_end,
							 int new_type_start,
							 MPI_Offset offset_adjustment)
{
	int mpi_errno = MPI_SUCCESS;
	int i, out_index = new_type_start;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_FLATTEN_COPY_TYPE);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_FLATTEN_COPY_TYPE);

	for (i=old_type_start; i < old_type_end; i++)
	{
		flat->indices[out_index]   = flat->indices[i] + offset_adjustment;
		flat->blocklens[out_index] = flat->blocklens[i];
		out_index++;
	}

	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_FLATTEN_COPY_TYPE);
	return mpi_errno;
}
