/*   Copyright (C) 2009 Deino Software. */

#include "mpiimpl.h"

#ifdef HAVE_WEAK_SYMBOLS

#if defined(HAVE_PRAGMA_WEAK)
#pragma weak MPI_File_sync = PMPI_File_sync
#elif defined(HAVE_PRAGMA_HP_SEC_DEF)
#pragma _HP_SECONDARY_DEF PMPI_File_sync MPI_File_sync
#elif defined(HAVE_PRAGMA_CRI_DUP)
#pragma _CRI duplicate MPI_File_sync as PMPI_File_sync
/* end of weak pragmas */
#endif

#endif

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build
   the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#define MPI_File_sync PMPI_File_sync
#endif

#undef FUNCNAME
#define FUNCNAME MPI_File_sync

/*@
	MPI_File_sync - Causes all previous writes to be transferred
					to the storage device

Input Parameters:
. fh - file handle (handle)

.N fortran
@*/
MPI_DLL_SPEC int MPI_CALL MPI_File_sync(MPI_File file)
{
	int mpi_errno = MPI_SUCCESS;
	MPID_File *file_ptr = NULL;
	static const wchar_t FCNAME[] = L"MPI_File_sync";
	MPIU_THREADPRIV_DECL;
	MPID_MPI_STATE_DECL(MPID_STATE_MPI_FILE_SYNC);

	MPIR_ERRTEST_INITIALIZED_ORDIE();

	MPIU_THREAD_SINGLE_CS_ENTER("io");
	MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_FILE_SYNC);
	MPIU_THREADPRIV_GET;

	/* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			MPIR_ERRTEST_FILE(file, mpi_errno);
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* Get handles to MPI objects. */
	MPID_File_get_ptr( file, file_ptr );

	/* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
	{
		MPID_BEGIN_ERROR_CHECKS;
		{
			/* Validate file_ptr */
			MPID_File_valid_ptr( file_ptr, mpi_errno );
			if (mpi_errno) goto fn_fail;
		}
		MPID_END_ERROR_CHECKS;
	}
#   endif /* HAVE_ERROR_CHECKING */

	/* --BEGIN ERROR HANDLING-- */
	if (file_ptr->cookie != ADIOI_FILE_COOKIE)
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE,
										  FCNAME, __LINE__, MPI_ERR_ARG,
										  L"**iobadfh", 0);
		goto fn_fail;
	}
	/* --END ERROR HANDLING-- */

	if (MPIR_Process.save_call_history) { MPIU_Save(L"**mpi_file_sync %F", file); }

	MPIR_TEST_FILE_DEFERRED(file_ptr, mpi_errno);

	mpi_errno = ADIO_Flush(file_ptr);
	MPIU_ERR_CHK_POP(mpi_errno);

fn_exit:
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_FILE_SYNC);
	MPIU_THREAD_SINGLE_CS_EXIT("io");
	return mpi_errno;
	
  fn_fail:
	/* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
	{
		mpi_errno = MPIR_Err_create_code(
			mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mpi_file_sync",
			L"**mpi_file_sync %F", file);
	}
#   endif
	mpi_errno = MPIR_Err_return_file(file_ptr, FCNAME, mpi_errno);
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}
