/*   Copyright (C) 2009 Deino Software. */

#include "mpiimpl.h"

#undef FUNCNAME
#define FUNCNAME ADIOI_Datatype_iscontig
int ADIOI_Datatype_iscontig(MPI_Datatype datatype, int *flag)
{
	static const wchar_t FCNAME[] = L"ADIOI_Datatype_iscontig";
	int mpi_errno = MPI_SUCCESS;
	MPID_Datatype *datatype_ptr;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_DATATYPE_ISCONTIG);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_DATATYPE_ISCONTIG);

	if (HANDLE_GET_KIND(datatype) == HANDLE_KIND_BUILTIN)
	{
		*flag = 1;
	}
	else
	{
		MPID_Datatype_get_ptr(datatype, datatype_ptr);
		MPID_Datatype_valid_ptr(datatype_ptr, mpi_errno);
		MPIU_ERR_CHK_POP(mpi_errno);
		*flag = datatype_ptr->is_contig;
	}

	/* if it is MPICH2 and the datatype is reported as contigous,
	   check if the true_lb is non-zero, and if so, mark the 
	   datatype as noncontiguous */
	if (*flag)
	{
		MPI_Aint true_extent, true_lb;

		NMPI_Type_get_true_extent(mpi_errno, (datatype, &true_lb, &true_extent));
		MPIU_ERR_CHK_POP(mpi_errno);

		if (true_lb > 0)
			*flag = 0;
	}

fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_DATATYPE_ISCONTIG);
	return mpi_errno;
}
