/*   Copyright (C) 2009 Deino Software. */

#include "ad_ntfs.h"

#undef FUNCNAME
#define FUNCNAME ADIOI_Set_lock
int ADIOI_Set_lock(FDTYPE fd, int cmd, int type, MPI_Offset offset, int whence, MPI_Offset len)
{
	static const wchar_t FCNAME[] = L"ADIOI_Set_lock";
	int ret_val, mpi_errno = MPI_SUCCESS;
	OVERLAPPED Overlapped;
	DWORD dwFlags;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_SET_LOCK);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_SET_LOCK);

	MPIU_UNREFERENCED_ARG(whence);

	if (len == 0)
	{
		goto fn_exit;
	}

	dwFlags = type;

	Overlapped.hEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
	if (Overlapped.hEvent == NULL)
	{
		ret_val = GetLastError();
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__,
			MPI_ERR_IO, L"**io", L"**io %s", ADIOI_NTFS_Strerror(ret_val));
		goto fn_fail;
	}
#ifdef HAVE_INT64
	Overlapped.Offset = ( (DWORD) ( offset & (__int64) 0xFFFFFFFF ) );
	Overlapped.OffsetHigh = ( (DWORD) ( (offset >> 32) & (__int64) 0xFFFFFFFF ) );

	if (cmd == ADIOI_LOCK_CMD)
	{
		/*wprintf(L"locking %d\n", (int)fd);fflush(stdout);*/
		ret_val = LockFileEx(fd, dwFlags, 0, 
		( (DWORD) ( len & (__int64) 0xFFFFFFFF ) ), 
		( (DWORD) ( (len >> 32) & (__int64) 0xFFFFFFFF ) ), 
		&Overlapped);
	}
	else
	{
		/*wprintf(L"unlocking %d\n", (int)fd);fflush(stdout);*/
		ret_val = UnlockFileEx(fd, 0, 
		( (DWORD) ( len & (__int64) 0xFFFFFFFF ) ), 
		( (DWORD) ( (len >> 32) & (__int64) 0xFFFFFFFF ) ), 
		&Overlapped);
	}
#else
	Overlapped.Offset = offset;
	Overlapped.OffsetHigh = 0;

	if (cmd == ADIOI_LOCK_CMD)
	{
		/*wprintf(L"locking %d\n", (int)fd);fflush(stdout);*/
		ret_val = LockFileEx(fd, dwFlags, 0, len, 0, &Overlapped);
	}
	else
	{
		/*wprintf(L"unlocking %d\n", (int)fd);fflush(stdout);*/
		ret_val = UnlockFileEx(fd, 0, len, 0, &Overlapped);
	}
#endif

	if (!ret_val)
	{
		/*
		FPRINTF(stderr, L"File locking failed in ADIOI_Set_lock.\n");
		MPI_Abort(MPI_COMM_WORLD, 1);
		*/
		ret_val = GetLastError();
		if (ret_val == ERROR_IO_PENDING)
		{
			DWORD dummy;
			ret_val = GetOverlappedResult(fd, &Overlapped, &dummy, TRUE);
			if (ret_val)
			{
				CloseHandle(Overlapped.hEvent);
				goto fn_exit;
			}
			ret_val = GetLastError();
		}
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__,
			MPI_ERR_IO, L"**io", L"**io %s", ADIOI_NTFS_Strerror(ret_val));
	}
	CloseHandle(Overlapped.hEvent);

fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_SET_LOCK);
	return mpi_errno;
fn_fail:
	goto fn_exit;
}
