/*   Copyright (C) 2009 Deino Software. */

#include <string.h>
#include "mpiimpl.h"

#ifdef HAVE_WEAK_SYMBOLS
/* Include mapping from MPI->PMPI */
#define MPIO_BUILD_PROFILING
#include "mpioprof.h"
#endif

#undef FUNCNAME
#define FUNCNAME MPIU_Greq_query_fn
int MPIU_Greq_query_fn(void *extra_state, MPI_Status *status)
{
	int old_errno, mpi_errno = MPI_SUCCESS;
	MPIU_THREADPRIV_DECL;
	MPIDI_STATE_DECL(MPID_STATE_MPIU_GREQ_QUERY_FN);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIU_GREQ_QUERY_FN);

	/* canL't touch user's MPI_ERROR, so hold it for a moment */
	old_errno = status->MPI_ERROR;

	/* get the status from the blocking operation */
	memcpy(status, extra_state, sizeof(MPI_Status));

	/* restore MPI_ERROR to whatever it had when we got it */
	status->MPI_ERROR = old_errno;

	/* and let Test|Wait know we weren't canceled */
	MPIU_THREADPRIV_GET;
	NMPI_Status_set_cancelled(mpi_errno, (status, 0));

	/* the MPI_Status structure is a convienent place to stash the return
	* code of the blocking operation */

	MPIDI_FUNC_EXIT(MPID_STATE_MPIU_GREQ_QUERY_FN);
	return ((MPI_Status*)extra_state)->MPI_ERROR;
}

#undef FUNCNAME
#define FUNCNAME MPIU_Greq_free_fn
int MPIU_Greq_free_fn(void *extra_state)
{
	MPIDI_STATE_DECL(MPID_STATE_MPIU_GREQ_FREE_FN);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIU_GREQ_FREE_FN);
	MPIU_Free(extra_state);
	MPIDI_FUNC_EXIT(MPID_STATE_MPIU_GREQ_FREE_FN);
	return MPI_SUCCESS;
}

#undef FUNCNAME
#define FUNCNAME MPIU_Greq_cancel_fn
int MPIU_Greq_cancel_fn(void *extra_state, int complete)
{
	MPIDI_STATE_DECL(MPID_STATE_MPIU_GREQ_CANCEL_FN);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIU_GREQ_CANCEL_FN);
	MPIU_UNREFERENCED_ARG(extra_state);
	MPIU_UNREFERENCED_ARG(complete);

	/* can't cancel */

	MPIDI_FUNC_EXIT(MPID_STATE_MPIU_GREQ_CANCEL_FN);
	return MPI_SUCCESS;
}
