/*   Copyright (C) 2009 Deino Software. */

#include "mpiimpl.h"

ADIOI_Req_node *ADIOI_Req_avail_head = NULL, *ADIOI_Req_avail_tail = NULL;
	/* list of available (already malloced) request objects */
ADIOI_Malloc_req *ADIOI_Malloc_req_head = NULL, *ADIOI_Malloc_req_tail = NULL;
	/* list of malloced areas for requests, which must be freed in ADIO_End */

#undef FUNCNAME
#define FUNCNAME ADIOI_Malloc_request
int ADIOI_Malloc_request(ADIO_Request_t **request_ptr)
{
/* returns a pointer to a new request object.
   To reduce the number of system calls, mallocs NUM requests at a time
   and maintains list of available requests. Supplies an object from this
   list if available, else mallocs a new set of NUM and provides one
   from that set. Is NUM=100 a good number? */

#define NUM 100

	static const wchar_t FCNAME[] = L"ADIOI_Malloc_request";
	int mpi_errno = MPI_SUCCESS;
	ADIOI_Req_node *curr, *ptr;
	int i;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_MALLOC_REQUEST);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_MALLOC_REQUEST);

	if (!ADIOI_Req_avail_head)
	{
		ADIOI_Req_avail_head = (ADIOI_Req_node *)MPIU_Malloc(NUM*sizeof(ADIOI_Req_node));
		MPIU_ERR_MEM_CHK_POP(mpi_errno, ADIOI_Req_avail_head, NUM*sizeof(ADIOI_Req_node), L"ADIOI_Req_avail_head");
		curr = ADIOI_Req_avail_head;
		for (i=1; i<NUM; i++)
		{
			curr->next = ADIOI_Req_avail_head+i;
			curr = curr->next;
		}
		curr->next = NULL;
		ADIOI_Req_avail_tail = curr;

		/* keep track of malloced area that needs to be freed later */
		if (!ADIOI_Malloc_req_tail)
		{
			ADIOI_Malloc_req_tail = (ADIOI_Malloc_req *)MPIU_Malloc(sizeof(ADIOI_Malloc_req));
			MPIU_ERR_MEM_CHK_POP(mpi_errno, ADIOI_Malloc_req_tail, sizeof(ADIOI_Malloc_req), L"ADIOI_Malloc_req_tail");
			ADIOI_Malloc_req_head = ADIOI_Malloc_req_tail;
			ADIOI_Malloc_req_head->ptr = ADIOI_Req_avail_head;
			ADIOI_Malloc_req_head->next = NULL;
		}
		else
		{
			ADIOI_Malloc_req_tail->next = (ADIOI_Malloc_req *)MPIU_Malloc(sizeof(ADIOI_Malloc_req));
			MPIU_ERR_MEM_CHK_POP(mpi_errno, ADIOI_Malloc_req_tail->next, sizeof(ADIOI_Malloc_req), L"ADIOI_Malloc_req_tail->next");
			ADIOI_Malloc_req_tail = ADIOI_Malloc_req_tail->next;
			ADIOI_Malloc_req_tail->ptr = ADIOI_Req_avail_head;
			ADIOI_Malloc_req_tail->next = NULL;
		}
	}

	ptr = ADIOI_Req_avail_head;
	ADIOI_Req_avail_head = ADIOI_Req_avail_head->next;
	if (!ADIOI_Req_avail_head) ADIOI_Req_avail_tail = NULL;
	
	(ptr->reqd).cookie = ADIOI_REQ_COOKIE;
	*request_ptr = &(ptr->reqd);

fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_MALLOC_REQUEST);
	return mpi_errno;
}


#undef FUNCNAME
#define FUNCNAME ADIOI_Free_request
int ADIOI_Free_request(ADIOI_Req_node *node)
{
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_FREE_REQUEST);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_FREE_REQUEST);
/* This function could be called as ADIOI_Free_request(ADIO_Request request), 
   because request would be a pointer to the first element of ADIOI_Req_node.*/

/* moves this node to available pool. does not actually free it. */

	(node->reqd).cookie = 0;

	if (!ADIOI_Req_avail_tail)
		ADIOI_Req_avail_head = ADIOI_Req_avail_tail = node;
	else
	{
		ADIOI_Req_avail_tail->next = node;
		ADIOI_Req_avail_tail = node;
	}
	node->next = NULL;

	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_FREE_REQUEST);
	return MPI_SUCCESS;
}
