/*   Copyright (C) 2009 Deino Software. */

#include "mpiimpl.h"

/* The following function selects the name of the file to be used to 
   store the shared file pointer. The shared-file-pointer file is a 
   hidden file in the same directory as the real file being accessed.
   If the real file is /tmp/thakur/testfile, the shared-file-pointer
   file will be /tmp/thakur/.testfile.shfp.xxxx, where xxxx is
   a random number. This file is created only if the shared
   file pointer functions are used and is deleted when the real
   file is closed. */

#undef FUNCNAME
#define FUNCNAME ADIOI_Shfp_fname
int ADIOI_Shfp_fname(MPID_File_ptr fd, int rank)
{
#define MAX_FILE_NAME_LENGTH 1024
#define MAX_FILE_NAME_LENGTH_BYTES MAX_FILE_NAME_LENGTH * sizeof(wchar_t)
	static const wchar_t FCNAME[] = L"ADIOI_Shfp_fname";
	int mpi_errno = MPI_SUCCESS;
	double tm;
	int i, len;
	wchar_t *slash, *ptr, tmp[128];
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_SHFP_FNAME);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_SHFP_FNAME);

	fd->shared_fp_fname = (wchar_t *) MPIU_Malloc(MAX_FILE_NAME_LENGTH_BYTES);
	MPIU_ERR_MEM_CHK_POP(mpi_errno, fd->shared_fp_fname, MAX_FILE_NAME_LENGTH_BYTES, L"fd->shared_fp_fname");

	if (!rank)
	{
		tm = NMPI_Wtime();
		while (tm > 1000000000.0) tm -= 1000000000.0;
		i = (int) tm;
		tm = tm - (double) i;
		tm *= 1000000.0;
		i = (int) tm;
		
		MPIU_Wcsncpy(fd->shared_fp_fname, fd->filename, MAX_FILE_NAME_LENGTH);
		
		slash = wcsrchr(fd->filename, L'\\');

		if (!slash)
		{
			MPIU_Wcsncpy(fd->shared_fp_fname, L".", 2);
			MPIU_Wcsncpy(fd->shared_fp_fname + 1, fd->filename, 255);
		}
		else
		{
			ptr = slash;
			slash = wcsrchr(fd->shared_fp_fname, L'\\');
			MPIU_Wcsncpy(slash + 1, L".", 2);
			len = (int)(MAX_FILE_NAME_LENGTH - (slash+2 - fd->shared_fp_fname));
			MPIU_Wcsncpy(slash + 2, ptr + 1, len);
		}
			
		MPIU_Snprintf(tmp, 128, _TRUNCATE, L".shfp.%d", i);
		MPIU_Wcsnapp(fd->shared_fp_fname, tmp, MAX_FILE_NAME_LENGTH);
		
		len = (int)wcslen(fd->shared_fp_fname);
		NMPI_Bcast(mpi_errno, (&len, 1, MPI_INT, 0, fd->comm));
		MPIU_ERR_CHK_POP(mpi_errno);
		NMPI_Bcast(mpi_errno, (fd->shared_fp_fname, len+1, MPI_WCHAR, 0, fd->comm));
		MPIU_ERR_CHK_POP(mpi_errno);
	}
	else
	{
		NMPI_Bcast(mpi_errno, (&len, 1, MPI_INT, 0, fd->comm));
		MPIU_ERR_CHK_POP(mpi_errno);
		NMPI_Bcast(mpi_errno, (fd->shared_fp_fname, len+1, MPI_WCHAR, 0, fd->comm));
		MPIU_ERR_CHK_POP(mpi_errno);
	}

fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_SHFP_FNAME);
	return mpi_errno;
}
