/*  (C) 2006 by Deino Software. */

#include "mpidi_ch3_impl.h"
#include "pmi.h"
#include "mpidu_sock.h"
#include "ch3usock.h"
#include <string.h>

static int MPIDI_CH3I_Progress_init(void);
static int sockMPIDI_CH3I_Progress_finalize(void);
static int MPIDI_CH3I_Progress_handle_sock_event(MPIDU_Sock_event_t * event);
static inline int connection_post_sendq_req(MPIDI_CH3I_Connection_t * conn);
static inline int connection_post_recv_pkt(MPIDI_CH3I_Connection_t * conn);
static int adjust_iov(MPID_IOV ** iovp, int * countp, MPIU_Size_t nb);


#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_Finalize
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int sockMPIDI_CH3_Finalize( void )
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_FINALIZE);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_FINALIZE);
	MPIU_DBG_MSG(CH3_CHANNEL,VERBOSE,L"Entering/exiting MPIDI_CH3_Finalize");
	mpi_errno = sockMPIDI_CH3I_Progress_finalize();
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_FINALIZE);
	return mpi_errno;
}

#ifdef USE_SIMPLE_MEMORY_CHECKING

#define MODCHAR 0xAB

typedef struct malloc_node
{
	void *ptr;
	unsigned char *real_ptr;
	size_t length;
	char *file;
	int line;
	struct malloc_node *next;
} malloc_node;
static malloc_node *list = NULL;

void * MPIU_MallocX(size_t size, int line, const char *file)
{
	int i;
	malloc_node *node = (malloc_node*)malloc(sizeof(malloc_node));
	node->real_ptr = malloc(size+32);
	node->length = size;
	node->file = strdup(file);
	node->line = line;
	node->next = list;
	list = node;
	for (i=0; i<16; i++)
	{
		node->real_ptr[i] = (unsigned char)i;
		node->real_ptr[size + 16 + i] = (unsigned char)i;
	}
	node->real_ptr[15] = MODCHAR;
	node->real_ptr[size+16] = MODCHAR;
	node->ptr = &node->real_ptr[16];
	return node->ptr;
}

void * MPIU_CallocX(size_t a, size_t b, int line, const char *file)
{
	unsigned char * ptr = MPIU_MallocX(a * b, line, file);
	memset(ptr, 0, a*b);
	return ptr;
}

wchar_t * MPIU_StrdupX(const wchar_t *str, int line, const char *file)
{
	size_t length = wcslen(str) + 1;
	wchar_t *dup = MPIU_MallocX(length * sizeof(wchar_t), line, file);
	wcscpy(dup, str);
	return dup;
}

void MPIU_Free(void *ptr)
{
	malloc_node *iter, *trailer;
	iter = trailer = list;
	while (iter != NULL)
	{
		if (iter->ptr == ptr)
		{
			if (iter->real_ptr[15] != MODCHAR)
			{
				printf("Buffer overrun before the beginning of the pointer.\n");
				fflush(stdout);
			}
			if (iter->real_ptr[iter->length + 16] != MODCHAR)
			{
				printf("Buffer overrun.\n");
				fflush(stdout);
			}
			free(iter->real_ptr);
			free(iter->file);
			trailer->next = iter->next;
			if (list == iter)
				list = iter->next;
			free(iter);
			return;
		}
		if (trailer != iter)
			trailer = trailer->next;
		iter = iter->next;
	}
	printf("freeing pointer that was already freed or never allocated.\n");fflush(stdout);
	abort();
}

void MPIU_PrintCurrentMemoryUsage()
{
	int n = 0;
	size_t size = 0;
	malloc_node *iter;

	iter = list;
	while (iter != NULL)
	{
		size += iter->length;
		iter = iter->next;
	}
	wprintf(L"MPIU_MEMORY:\n");
	iter = list;
	while (iter != NULL)
	{
		n++;
		if (n < 10)
		{
			printf("%s[%d] = %d\n", iter->file, iter->line, iter->length);
		}
		else
		{
			wprintf(L"...\n");
			break;
		}
		iter = iter->next;
	}
	wprintf(L"Total: %d bytes\n", size);
	/*
	iter = list;
	while (iter != NULL)
	{
		n++;
		size += iter->length;
		iter = iter->next;
	}
	if (n > 20)
		wprintf(L"MPIU_MEMORY: (..., ");
	else
		wprintf(L"MPIU_MEMORY: (");
	iter = list;
	while (iter != NULL)
	{
		n--;
		if (n < 20)
		{
			wprintf(L"%d, ", iter->length);
		}
		iter = iter->next;
	}
	wprintf(L") %d bytes\n", size);
	*/
	/*wprintf(L"MPIU_MEMORY: %d bytes\n", size);*/
	fflush(stdout);
}

#endif /* USE_SIMPLE_MEMORY_CHECKING */

/*
 *  MPIDI_CH3_Init  - makes socket specific initializations.  Most of this 
 *					functionality is in the MPIDI_CH3U_Init_sock upcall 
 *					because the same tasks need to be done for the ssh 
 *					(sock + shm) channel.  
 */

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_Init
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int sockMPIDI_CH3_Init(int has_parent, MPIDI_PG_t * pg_p, int pg_rank )
{
	int mpi_errno = MPI_SUCCESS;
	wchar_t *publish_bc_orig = NULL;
	wchar_t *bc_val = NULL;
	int val_max_remaining;
	MPIDI_STATE_DECL(MPID_STATE_MPID_CH3_INIT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPID_CH3_INIT);

	MPIDI_CH3I_Eager_max_msg_size = (256 * 1024);

	mpi_errno = MPIDI_CH3I_Progress_init();
	MPIU_ERR_CHK_POP(mpi_errno);

	/* Initialize the business card */
	mpi_errno = MPIDI_CH3I_BCInit( &bc_val, &val_max_remaining );
	MPIU_ERR_CHK_POP(mpi_errno);
	publish_bc_orig = bc_val;

	/* initialize aspects specific to sockets  */
	mpi_errno = MPIDI_CH3U_Init_sock(has_parent, pg_p, pg_rank, &bc_val, &val_max_remaining);
	MPIU_ERR_CHK_POP(mpi_errno);

	/*wprintf(L"publishing business card: %s\n", publish_bc_orig);fflush(stdout);*/
	/* Set the connection information in our process group (publish the business card ) */
	mpi_errno = MPIDI_PG_SetConnInfo( pg_rank, (const wchar_t *)publish_bc_orig );
	MPIU_ERR_CHK_POP(mpi_errno);

	/* Free the business card now that it is published
	 (note that publish_bc_orig is the head of bc_val ) */
	mpi_errno = MPIDI_CH3I_BCFree( publish_bc_orig );
	MPIU_ERR_CHK_POP(mpi_errno);

 fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPID_CH3_INIT);
	return mpi_errno;
 fn_fail:
	if (publish_bc_orig != NULL) {
		MPIU_Free(publish_bc_orig);
	}		   
	goto fn_exit;
}

/* This function simply tells the CH3 device to use the defaults for the MPI Port functions */
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_PortFnsInit
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int sockMPIDI_CH3_PortFnsInit( MPIDI_PortFns *portFns ) 
{
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_PORTFNSINIT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_PORTFNSINIT);
	MPIU_UNREFERENCED_ARG(portFns);
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_PORTFNSINIT);
	return MPI_SUCCESS;
}

/* Perform the channel-specific vc initialization */
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_VC_Init
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int sockMPIDI_CH3_VC_Init( MPIDI_VC_t *vc )
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_VC_INIT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_VC_INIT);
	vc->ch.sendq_head		 = NULL;
	vc->ch.sendq_tail		 = NULL;
	vc->ch.state			  = MPIDI_CH3I_VC_STATE_UNCONNECTED;
	mpi_errno = MPIDI_VC_InitSock( vc );
	MPIU_ERR_CHK_POP(mpi_errno);
	MPIU_DBG_MSG_P(CH3_CONNECT,TYPICAL,"vc=%p: Setting state (ch) to VC_STATE_UNCONNECTED (Initialization)", vc );
fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_VC_INIT);
	return mpi_errno;
}

/* Select the routine that uses sockets to connect two communicators using a socket */
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_Connect_to_root
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int sockMPIDI_CH3_Connect_to_root(const wchar_t * port_name, MPIDI_VC_t ** new_vc)
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_CONNECT_TO_ROOT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_CONNECT_TO_ROOT);
	mpi_errno = MPIDI_CH3I_Connect_to_root_sock( port_name, new_vc );
	MPIU_ERR_CHK_POP(mpi_errno);
fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_CONNECT_TO_ROOT);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME update_request_isend
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static void update_request_isend(MPID_Request * sreq, void * hdr, MPIDI_msg_sz_t hdr_sz, MPIU_Size_t nb)
{
	MPIDI_STATE_DECL(MPID_STATE_UPDATE_REQUEST);

	MPIDI_FUNC_ENTER(MPID_STATE_UPDATE_REQUEST);
	MPIU_Assert(hdr_sz == sizeof(MPIDI_CH3_Pkt_t));
	sreq->ch.pkt = *(MPIDI_CH3_Pkt_t *) hdr;
	sreq->dev.iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)((char *) &sreq->ch.pkt + nb);
	sreq->dev.iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)(hdr_sz - nb);
	sreq->dev.iov_count = 1;
	MPIDI_FUNC_EXIT(MPID_STATE_UPDATE_REQUEST);
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_iSend
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int sockMPIDI_CH3_iSend(MPIDI_VC_t * vc, MPID_Request * sreq, void * hdr, MPIDI_msg_sz_t hdr_sz)
{
	int mpi_errno = MPI_SUCCESS;
	int complete;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_ISEND);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_ISEND);

#ifdef MPICH_DBG_OUTPUT
	/* --BEGIN ERROR HANDLING-- */
	if (hdr_sz > sizeof(MPIDI_CH3_Pkt_t))
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**arg", 0);
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISEND);
		return mpi_errno;
	}
	/* --END ERROR HANDLING-- */
#endif

	/* The sock channel uses a fixed length header, the size of which is the maximum of all possible packet headers */
	hdr_sz = sizeof(MPIDI_CH3_Pkt_t);
	MPIU_DBG_STMT(CH3_CHANNEL,VERBOSE,
				  MPIDI_DBG_Print_packet((MPIDI_CH3_Pkt_t*)hdr));

	if (vc->ch.state == MPIDI_CH3I_VC_STATE_CONNECTED) /* MT */
	{
		/* Connection already formed.  If send queue is empty attempt to send data, queuing any unsent data. */
		if (MPIDI_CH3I_SendQ_empty(vc)) /* MT */
		{
			MPIU_Size_t nb;
			int rc;

			MPIU_DBG_MSG(CH3_CHANNEL,VERBOSE,
						 L"send queue empty, attempting to write");
			MPIU_DBG_PKT(vc->ch.conn,hdr,"isend");
			/* MT: need some signalling to lock down our right to use the channel, thus insuring that the progress engine does
			   also try to write */
			rc = MPIDU_Sock_write(vc->ch.sock, hdr, (MPIU_Size_t)hdr_sz, &nb);
			if (rc == MPI_SUCCESS)
			{
				MPIU_DBG_MSG_D(CH3_CHANNEL,VERBOSE,
							   L"wrote %ld bytes", (unsigned long) nb);
				
				if (nb == (MPIU_Size_t)hdr_sz)
				{
					MPIU_DBG_MSG_D(CH3_CHANNEL,VERBOSE,
		  L"write complete %d bytes, calling MPIDI_CH3U_Handle_send_req()", nb);
					MPIDI_CH3U_Handle_send_req(vc, sreq, &complete);
					if (!complete)
					{
						MPIDI_CH3I_SendQ_enqueue_head(vc, sreq);
						MPIU_DBG_MSG_FMT(CH3_CHANNEL,VERBOSE,(MPIU_DBG_FDEST,
					L"posting writev, vc=0x%p, sreq=0x%08x", vc, sreq->handle));
						vc->ch.conn->send_active = sreq;
						mpi_errno = MPIDU_Sock_post_writev(vc->ch.conn->sock, sreq->dev.iov, sreq->dev.iov_count, NULL);
						/* --BEGIN ERROR HANDLING-- */
						if (mpi_errno != MPI_SUCCESS)
						{
							mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER,
															 L"**ch3|sock|postwrite", L"ch3|sock|postwrite %p %p %p",
															 sreq, vc->ch.conn, vc);
						}
						/* --END ERROR HANDLING-- */
					}
				}
				else
				{
					MPIU_DBG_MSG_D(CH3_CHANNEL,VERBOSE,
					 L"partial write of %d bytes, request enqueued at head", nb);
					update_request_isend(sreq, hdr, hdr_sz, nb);
					MPIDI_CH3I_SendQ_enqueue_head(vc, sreq);
					MPIU_DBG_MSG_FMT(CH3_CHANNEL,VERBOSE,
	 (MPIU_DBG_FDEST,L"posting write, vc=0x%p, sreq=0x%08x", vc, sreq->handle));
					vc->ch.conn->send_active = sreq;
					mpi_errno = MPIDU_Sock_post_write(vc->ch.conn->sock, sreq->dev.iov[0].MPID_IOV_BUF,
													  sreq->dev.iov[0].MPID_IOV_LEN, sreq->dev.iov[0].MPID_IOV_LEN, NULL);
					/* --BEGIN ERROR HANDLING-- */
					if (mpi_errno != MPI_SUCCESS)
					{
						mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER,
														 L"**ch3|sock|postwrite", L"ch3|sock|postwrite %p %p %p",
														 sreq, vc->ch.conn, vc);
					}
					/* --END ERROR HANDLING-- */
				}
			}
			/* --BEGIN ERROR HANDLING-- */
			else if (MPIR_ERR_GET_CLASS(rc) == MPIDU_SOCK_ERR_NOMEM)
			{
				MPIU_DBG_MSG(CH3_CHANNEL,TYPICAL,
							 L"MPIDU_Sock_write failed, out of memory");
				sreq->status.MPI_ERROR = MPIR_ERR_MEMALLOCFAILED;
			}
			else
			{
				MPIU_DBG_MSG_D(CH3_CHANNEL,TYPICAL,
							   L"MPIDU_Sock_write failed, rc=%d", rc);
				/* Connection just failed. Mark the request complete and return an error. */
				MPIU_DBG_VCCHSTATECHANGE(vc,VC_STATE_FAILED);
				/* FIXME: Shouldn't the vc->state also change? */
				vc->ch.state = MPIDI_CH3I_VC_STATE_FAILED;
				/* TODO: Create an appropriate error message based on the return value (rc) */
				sreq->status.MPI_ERROR = MPI_ERR_INTERN;
				 /* MT -CH3U_Request_complete() performs write barrier */
				MPIDI_CH3U_Request_complete(sreq);
			}
			/* --END ERROR HANDLING-- */
		}
		else
		{
			MPIU_DBG_MSG(CH3_CHANNEL,VERBOSE,L"send queue not empty, enqueuing");
			update_request_isend(sreq, hdr, hdr_sz, 0);
			MPIDI_CH3I_SendQ_enqueue(vc, sreq);
		}
	}
	else if (vc->ch.state == MPIDI_CH3I_VC_STATE_UNCONNECTED) /* MT */
	{
		/* Form a new connection, queuing the data so it can be sent later. */
		MPIU_DBG_VCUSE(vc, L"unconnected.  enqueuing request");
		update_request_isend(sreq, hdr, hdr_sz, 0);
		MPIDI_CH3I_SendQ_enqueue(vc, sreq);
		mpi_errno = MPIDI_CH3I_VC_post_sockconnect(vc);
		/* --BEGIN ERROR HANDLING-- */
		if (mpi_errno != MPI_SUCCESS)
		{
			mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", 0);
		}
		/* --END ERROR HANDLING-- */
	}
	else if (vc->ch.state != MPIDI_CH3I_VC_STATE_FAILED)
	{
		/* Unable to send data at the moment, so queue it for later */
		MPIU_DBG_VCUSE(vc, L"still connecting. Enqueuing request");
		update_request_isend(sreq, hdr, hdr_sz, 0);
		MPIDI_CH3I_SendQ_enqueue(vc, sreq);
	}
	/* --BEGIN ERROR HANDLING-- */
	else
	{
		/* Connection failed.  Mark the request complete and return an error. */
		/* TODO: Create an appropriate error message */
		sreq->status.MPI_ERROR = MPI_ERR_INTERN;
		/* MT - CH3U_Request_complete() performs write barrier */
		MPIDI_CH3U_Request_complete(sreq);
	}
	/* --END ERROR HANDLING-- */

	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISEND);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME update_request_isendv
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static void update_request_isendv(MPID_Request * sreq, MPID_IOV * iov, int iov_count, int iov_offset, MPIU_Size_t nb)
{
	int i;
	MPIDI_STATE_DECL(MPID_STATE_UPDATE_REQUEST);

	MPIDI_FUNC_ENTER(MPID_STATE_UPDATE_REQUEST);
	
	for (i = 0; i < iov_count; i++)
	{
		sreq->dev.iov[i] = iov[i];
	}
	if (iov_offset == 0)
	{
		MPIU_Assert(iov[0].MPID_IOV_LEN == sizeof(MPIDI_CH3_Pkt_t));
		sreq->ch.pkt = *(MPIDI_CH3_Pkt_t *) iov[0].MPID_IOV_BUF;
		sreq->dev.iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST) &sreq->ch.pkt;
	}
	sreq->dev.iov[iov_offset].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)((char *) sreq->dev.iov[iov_offset].MPID_IOV_BUF + nb);
	sreq->dev.iov[iov_offset].MPID_IOV_LEN -= (MPID_IOV_LEN_CAST)nb;
	sreq->dev.iov_count = iov_count;

	MPIDI_FUNC_EXIT(MPID_STATE_UPDATE_REQUEST);
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_iSendv
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int sockMPIDI_CH3_iSendv(MPIDI_VC_t * vc, MPID_Request * sreq, MPID_IOV * iov, int n_iov)
{
	int mpi_errno = MPI_SUCCESS;
	int complete;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_ISENDV);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_ISENDV);
#ifdef MPICH_DBG_OUTPUT
	/* --BEGIN ERROR HANDLING-- */
	if (n_iov > MPID_IOV_LIMIT)
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**arg", 0);
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISENDV);
		return mpi_errno;
	}
	if (iov[0].MPID_IOV_LEN > (MPID_IOV_LEN_CAST)sizeof(MPIDI_CH3_Pkt_t))
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**arg", 0);
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISENDV);
		return mpi_errno;
	}
	/* --END ERROR HANDLING-- */
#endif

	/* The sock channel uses a fixed length header, the size of which is the maximum of all possible packet headers */
	iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)sizeof(MPIDI_CH3_Pkt_t);
	MPIU_DBG_STMT(CH3_CHANNEL,VERBOSE,
		 MPIDI_DBG_Print_packet((MPIDI_CH3_Pkt_t *)iov[0].MPID_IOV_BUF));


	if (vc->ch.state == MPIDI_CH3I_VC_STATE_CONNECTED) /* MT */
	{
		/* Connection already formed.  If send queue is empty attempt to send data, queuing any unsent data. */
		if (MPIDI_CH3I_SendQ_empty(vc)) /* MT */
		{
			MPIU_Size_t nb;
			int rc;

			MPIU_DBG_MSG(CH3_CHANNEL,VERBOSE,
						 L"send queue empty, attempting to write");
			
			MPIU_DBG_PKT(vc->ch.conn,(MPIDI_CH3_Pkt_t*)iov[0].MPID_IOV_BUF,
						 "isendv");
			/* MT - need some signalling to lock down our right to use the channel, thus insuring that the progress engine does
			   also try to write */

			/* FIXME: the current code only agressively writes the first IOV.  Eventually it should be changed to agressively write
			   as much as possible.  Ideally, the code would be shared between the send routines and the progress engine. */
			rc = MPIDU_Sock_writev(vc->ch.sock, iov, n_iov, &nb);
			if (rc == MPI_SUCCESS)
			{
				int offset = 0;

				MPIU_DBG_MSG_D(CH3_CHANNEL,VERBOSE,
							   L"wrote %ld bytes", (unsigned long) nb);
				
				while (offset < n_iov)
				{
					if ((MPIU_Size_t)iov[offset].MPID_IOV_LEN <= nb)
					{
						nb -= (MPIU_Size_t)iov[offset].MPID_IOV_LEN;
						offset++;
					}
					else
					{
						MPIU_DBG_MSG(CH3_CHANNEL,VERBOSE,
							 L"partial write, request enqueued at head");
						update_request_isendv(sreq, iov, n_iov, offset, nb);
						MPIDI_CH3I_SendQ_enqueue_head(vc, sreq);
						MPIU_DBG_MSG_FMT(CH3_CHANNEL,VERBOSE,
	(MPIU_DBG_FDEST,L"posting writev, vc=0x%p, sreq=0x%08x", vc, sreq->handle));
						vc->ch.conn->send_active = sreq;
						mpi_errno = MPIDU_Sock_post_writev(vc->ch.conn->sock, sreq->dev.iov + offset,
														   sreq->dev.iov_count - offset, NULL);
						/* --BEGIN ERROR HANDLING-- */
						if (mpi_errno != MPI_SUCCESS)
						{
							mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER,
															 L"**ch3|sock|postwrite", L"ch3|sock|postwrite %p %p %p",
															 sreq, vc->ch.conn, vc);
						}
						/* --END ERROR HANDLING-- */

						break;
					}

				}
				if (offset == n_iov)
				{
					MPIU_DBG_MSG(CH3_CHANNEL,VERBOSE,
					  L"write complete, calling MPIDI_CH3U_Handle_send_req()");
					MPIDI_CH3U_Handle_send_req(vc, sreq, &complete);
					if (!complete)
					{
						MPIDI_CH3I_SendQ_enqueue_head(vc, sreq);
						MPIU_DBG_MSG_FMT(CH3_CHANNEL,VERBOSE,
	(MPIU_DBG_FDEST,L"posting writev, vc=0x%p, sreq=0x%08x", vc, sreq->handle));
						vc->ch.conn->send_active = sreq;
						mpi_errno = MPIDU_Sock_post_writev(vc->ch.conn->sock, sreq->dev.iov, sreq->dev.iov_count, NULL);
						/* --BEGIN ERROR HANDLING-- */
						if (mpi_errno != MPI_SUCCESS)
						{
							mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER,
															 L"**ch3|sock|postwrite", L"ch3|sock|postwrite %p %p %p",
															 sreq, vc->ch.conn, vc);
						}
						/* --END ERROR HANDLING-- */
					}
				}
			}
			/* --BEGIN ERROR HANDLING-- */
			else if (MPIR_ERR_GET_CLASS(rc) == MPIDU_SOCK_ERR_NOMEM)
			{
				MPIU_DBG_MSG(CH3_CHANNEL,TYPICAL,
							 L"MPIDU_Sock_writev failed, out of memory");
				sreq->status.MPI_ERROR = MPIR_ERR_MEMALLOCFAILED;
			}
			else
			{
				MPIU_DBG_MSG_D(CH3_CHANNEL,TYPICAL,
							   L"MPIDU_Sock_writev failed, rc=%d", rc);
				/* Connection just failed.  Mark the request complete and 
				   return an error. */
				MPIU_DBG_VCCHSTATECHANGE(vc,VC_STATE_FAILED);
				/* FIXME: Shouldn't the vc->state also change? */

				vc->ch.state = MPIDI_CH3I_VC_STATE_FAILED;
				/* TODO: Create an appropriate error message based on the return value (rc) */
				sreq->status.MPI_ERROR = MPI_ERR_INTERN;
				 /* MT - CH3U_Request_complete performs write barrier */
				MPIDI_CH3U_Request_complete(sreq);
			}
			/* --END ERROR HANDLING-- */
		}
		else
		{
			MPIU_DBG_MSG(CH3_CHANNEL,VERBOSE,L"send queue not empty, enqueuing");
			update_request_isendv(sreq, iov, n_iov, 0, 0);
			MPIDI_CH3I_SendQ_enqueue(vc, sreq);
		}
	}
	else if (vc->ch.state == MPIDI_CH3I_VC_STATE_UNCONNECTED)
	{
		/* Form a new connection, queuing the data so it can be sent later. */
		MPIU_DBG_VCUSE(vc, L"unconnected.  Enqueuing request");
		update_request_isendv(sreq, iov, n_iov, 0, 0);
		MPIDI_CH3I_SendQ_enqueue(vc, sreq);
		mpi_errno = MPIDI_CH3I_VC_post_sockconnect(vc);
		/* --BEGIN ERROR HANDLING-- */
		if (mpi_errno != MPI_SUCCESS)
		{
			mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", 0);
		}
		/* --END ERROR HANDLING-- */
	}
	else if (vc->ch.state != MPIDI_CH3I_VC_STATE_FAILED)
	{
		/* Unable to send data at the moment, so queue it for later */
		MPIU_DBG_VCUSE(vc, L"still connecting.  enqueuing request");
		update_request_isendv(sreq, iov, n_iov, 0, 0);
		MPIDI_CH3I_SendQ_enqueue(vc, sreq);
	}
	/* --BEGIN ERROR HANDLING-- */
	else
	{
		MPIU_DBG_VCUSE(vc, L"connection failed");
		/* Connection failed.  Mark the request complete and return an error. */
		/* TODO: Create an appropriate error message */
		sreq->status.MPI_ERROR = MPI_ERR_INTERN;
		/* MT - CH3U_Request_complete performs write barrier */
		MPIDI_CH3U_Request_complete(sreq);
	}
	/* --END ERROR HANDLING-- */

	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISENDV);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME create_request_istartmsg
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static MPID_Request * create_request_istartmsg(void * hdr, MPIDI_msg_sz_t hdr_sz, MPIU_Size_t nb)
{
	MPID_Request * sreq;
	MPIDI_STATE_DECL(MPID_STATE_CREATE_REQUEST);

	MPIDI_FUNC_ENTER(MPID_STATE_CREATE_REQUEST);

	sreq = MPID_Request_create();
	/*MPIU_Assert(sreq != NULL);*/
	/* --BEGIN ERROR HANDLING-- */
	if (sreq == NULL)
		return NULL;
	/* --END ERROR HANDLING-- */
	MPIU_Object_set_ref(sreq, 2);
	sreq->kind = MPID_REQUEST_SEND;
	MPIU_Assert(hdr_sz == sizeof(MPIDI_CH3_Pkt_t));
	sreq->ch.pkt = *(MPIDI_CH3_Pkt_t *) hdr;
	sreq->dev.iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)((char *) &sreq->ch.pkt + nb);
	sreq->dev.iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)(hdr_sz - nb);
	sreq->dev.iov_count = 1;
	sreq->dev.ca = MPIDI_CH3_CA_COMPLETE;
	
	MPIDI_FUNC_EXIT(MPID_STATE_CREATE_REQUEST);
	return sreq;
}

/*
 * MPIDI_CH3_iStartMsg() attempts to send the message immediately.  If the entire message is successfully sent, then NULL is
 * returned.  Otherwise a request is allocated, the header is copied into the request, and a pointer to the request is returned.
 * An error condition also results in a request be allocated and the errror being returned in the status field of the request.
 */
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_iStartMsg
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int sockMPIDI_CH3_iStartMsg(MPIDI_VC_t * vc, void * hdr, MPIDI_msg_sz_t hdr_sz, MPID_Request ** sreq_ptr)
{
	MPID_Request * sreq = NULL;
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_ISTARTMSG);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_ISTARTMSG);
	
#ifdef MPICH_DBG_OUTPUT
	/* --BEGIN ERROR HANDLING-- */
	if (hdr_sz > sizeof(MPIDI_CH3_Pkt_t))
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**arg", 0);
		goto fn_exit;
	}
	/* --END ERROR HANDLING-- */
#endif

	/* The SOCK channel uses a fixed length header, the size of which is the maximum of all possible packet headers */
	hdr_sz = sizeof(MPIDI_CH3_Pkt_t);
	MPIU_DBG_STMT(CH3_CHANNEL,VERBOSE,
				  MPIDI_DBG_Print_packet((MPIDI_CH3_Pkt_t*)hdr));

	if (vc->ch.state == MPIDI_CH3I_VC_STATE_CONNECTED) /* MT */
	{
		/* Connection already formed.  If send queue is empty attempt to send data, queuing any unsent data. */
		if (MPIDI_CH3I_SendQ_empty(vc)) /* MT */
		{
			MPIU_Size_t nb;
			int rc;

			MPIU_DBG_MSG(CH3_CHANNEL,VERBOSE,
						 L"send queue empty, attempting to write");
			
			MPIU_DBG_PKT(vc->ch.conn,hdr,"istartmsg");
			/* MT - need some signalling to lock down our right to use the channel, thus insuring that the progress engine does
			   not also try to write */
			rc = MPIDU_Sock_write(vc->ch.sock, hdr, (MPIU_Size_t)hdr_sz, &nb);
			if (rc == MPI_SUCCESS)
			{
				MPIU_DBG_MSG_D(CH3_CHANNEL,VERBOSE,
							   L"wrote %ld bytes", (unsigned long) nb);
				
				if (nb == (MPIU_Size_t)hdr_sz)
				{ 
					MPIU_DBG_MSG_D(CH3_CHANNEL,VERBOSE,
								   L"entire write complete, %d bytes", nb);
					/* done.  get us out of here as quickly as possible. */
				}
				else
				{
					MPIU_DBG_MSG_D(CH3_CHANNEL,VERBOSE,
					L"partial write of %d bytes, request enqueued at head", nb);
					sreq = create_request_istartmsg(hdr, hdr_sz, nb);
					/* --BEGIN ERROR HANDLING-- */
					if (sreq == NULL)
					{
						mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER,
														 L"**nomem", 0);
						goto fn_exit;
					}
					/* --END ERROR HANDLING-- */
					MPIDI_CH3I_SendQ_enqueue_head(vc, sreq);
					MPIU_DBG_MSG_FMT(CH3_CHANNEL,VERBOSE,
	 (MPIU_DBG_FDEST,L"posting write, vc=0x%p, sreq=0x%08x", vc, sreq->handle));
					vc->ch.conn->send_active = sreq;
					mpi_errno = MPIDU_Sock_post_write(vc->ch.conn->sock, sreq->dev.iov[0].MPID_IOV_BUF,
													  (MPIU_Size_t)sreq->dev.iov[0].MPID_IOV_LEN, (MPIU_Size_t)sreq->dev.iov[0].MPID_IOV_LEN, NULL);
					/* --BEGIN ERROR HANDLING-- */
					if (mpi_errno != MPI_SUCCESS)
					{
						mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER,
														 L"**ch3|sock|postwrite", L"ch3|sock|postwrite %p %p %p",
														 sreq, vc->ch.conn, vc);
						goto fn_exit;
					}
					/* --END ERROR HANDLING-- */
				}
			}
			/* --BEGIN ERROR HANDLING-- */
			else
			{
				MPIU_DBG_MSG_D(CH3_CHANNEL,TYPICAL,
							   L"ERROR - MPIDU_Sock_write failed, rc=%d", rc);
				sreq = MPID_Request_create();
				if (sreq == NULL)
				{
					mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
					goto fn_exit;
				}
				sreq->kind = MPID_REQUEST_SEND;
				sreq->cc = 0;
				/* TODO: Create an appropriate error message based on the return value */
				sreq->status.MPI_ERROR = MPI_ERR_INTERN;
			}
			/* --END ERROR HANDLING-- */
		}
		else
		{
			MPIU_DBG_MSG(CH3_CHANNEL,VERBOSE,
						 L"send in progress, request enqueued");
			sreq = create_request_istartmsg(hdr, hdr_sz, 0);
			/* --BEGIN ERROR HANDLING-- */
			if (sreq == NULL)
			{
				mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
				goto fn_exit;
			}
			/* --END ERROR HANDLING-- */
			MPIDI_CH3I_SendQ_enqueue(vc, sreq);
		}
	}
	else if (vc->ch.state == MPIDI_CH3I_VC_STATE_UNCONNECTED) /* MT */
	{
		MPIU_DBG_VCUSE(vc, L"unconnected.  posting connect and enqueuing request");
		
		/* queue the data so it can be sent after the connection is formed */
		sreq = create_request_istartmsg(hdr, hdr_sz, 0);
		/* --BEGIN ERROR HANDLING-- */
		if (sreq == NULL)
		{
			mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
			goto fn_exit;
		}
		/* --END ERROR HANDLING-- */
		MPIDI_CH3I_SendQ_enqueue(vc, sreq);

		/* Form a new connection */
		MPIDI_CH3I_VC_post_sockconnect(vc);
	}
	else if (vc->ch.state != MPIDI_CH3I_VC_STATE_FAILED)
	{
		/* Unable to send data at the moment, so queue it for later */
		MPIU_DBG_VCUSE(vc, L"forming connection, request enqueued");
		sreq = create_request_istartmsg(hdr, hdr_sz, 0);
		/* --BEGIN ERROR HANDLING-- */
		if (sreq == NULL)
		{
			mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
			goto fn_exit;
		}
		/* --END ERROR HANDLING-- */
		MPIDI_CH3I_SendQ_enqueue(vc, sreq);
	}
	/* --BEGIN ERROR HANDLING-- */
	else
	{
		/* Connection failed, so allocate a request and return an error. */
		MPIU_DBG_VCUSE(vc, L"ERROR - connection failed");
		sreq = MPID_Request_create();
		if (sreq == NULL)
		{
			mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
			goto fn_exit;
		}
		sreq->kind = MPID_REQUEST_SEND;
		sreq->cc = 0;
		/* TODO: Create an appropriate error message */
		sreq->status.MPI_ERROR = MPI_ERR_INTERN;
	}
	/* --END ERROR HANDLING-- */

  fn_exit:
	*sreq_ptr = sreq;
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISTARTMSG);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME create_request_istartmsgv
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static MPID_Request * create_request_istartmsgv(MPID_IOV * iov, int iov_count, int iov_offset, MPIU_Size_t nb)
{
	MPID_Request * sreq;
	int i;
	MPIDI_STATE_DECL(MPID_STATE_CREATE_REQUEST);
	/*MPIDI_STATE_DECL(MPID_STATE_MEMCPY);*/

	MPIDI_FUNC_ENTER(MPID_STATE_CREATE_REQUEST);
	
	sreq = MPID_Request_create();
	/* --BEGIN ERROR HANDLING-- */
	if (sreq == NULL)
		return NULL;
	/* --END ERROR HANDLING-- */
	MPIU_Object_set_ref(sreq, 2);
	sreq->kind = MPID_REQUEST_SEND;
	
	/*
	MPIDI_FUNC_ENTER(MPID_STATE_MEMCPY);
	memcpy(sreq->dev.iov, iov, iov_count * sizeof(MPID_IOV));
	MPIDI_FUNC_EXIT(MPID_STATE_MEMCPY);
	*/
	for (i = 0; i < iov_count; i++)
	{
		sreq->dev.iov[i] = iov[i];
	}
	if (iov_offset == 0)
	{
		/*
		MPIDI_FUNC_ENTER(MPID_STATE_MEMCPY);
		memcpy(&sreq->ch.pkt, iov[0].MPID_IOV_BUF, iov[0].MPID_IOV_LEN);
		MPIDI_FUNC_EXIT(MPID_STATE_MEMCPY);
		*/
		MPIU_Assert(iov[0].MPID_IOV_LEN == sizeof(MPIDI_CH3_Pkt_t));
		sreq->ch.pkt = *(MPIDI_CH3_Pkt_t *) iov[0].MPID_IOV_BUF;
		sreq->dev.iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST) &sreq->ch.pkt;
	}
	sreq->dev.iov[iov_offset].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)((char *) sreq->dev.iov[iov_offset].MPID_IOV_BUF + nb);
	sreq->dev.iov[iov_offset].MPID_IOV_LEN -= (MPID_IOV_LEN_CAST)nb;
	sreq->dev.iov_count = iov_count;
	sreq->dev.ca = MPIDI_CH3_CA_COMPLETE;

	MPIDI_FUNC_EXIT(MPID_STATE_CREATE_REQUEST);
	return sreq;
}

/*
 * MPIDI_CH3_iStartMsgv() attempts to send the message immediately.  If the entire message is successfully sent, then NULL is
 * returned.  Otherwise a request is allocated, the iovec and the first buffer pointed to by the iovec (which is assumed to be a
 * MPIDI_CH3_Pkt_t) are copied into the request, and a pointer to the request is returned.  An error condition also results in a
 * request be allocated and the errror being returned in the status field of the request.
 */

/* XXX - What do we do if MPID_Request_create() returns NULL???  If MPIDI_CH3_iStartMsgv() returns NULL, the calling code
   assumes the request completely successfully, but the reality is that we couldn't allocate the memory for a request.  This
   seems like a flaw in the CH3 API. */

/* NOTE - The completion action associated with a request created by CH3_iStartMsgv() is always MPIDI_CH3_CA_COMPLETE.  This
   implies that CH3_iStartMsgv() can only be used when the entire message can be described by a single iovec of size
   MPID_IOV_LIMIT. */
	
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_iStartMsgv
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int sockMPIDI_CH3_iStartMsgv(MPIDI_VC_t * vc, MPID_IOV * iov, int n_iov, MPID_Request ** sreq_ptr)
{
	MPID_Request * sreq = NULL;
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_ISTARTMSGV);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_ISTARTMSGV);

#ifdef MPICH_DBG_OUTPUT
	/* --BEGIN ERROR HANDLING-- */
	if (n_iov > MPID_IOV_LIMIT)
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**arg", 0);
		goto fn_exit;
	}
	if (iov[0].MPID_IOV_LEN > (MPID_IOV_LEN_CAST)sizeof(MPIDI_CH3_Pkt_t))
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**arg", 0);
		goto fn_exit;
	}
	/* --END ERROR HANDLING-- */
#endif

	/* The SOCK channel uses a fixed length header, the size of which is the maximum of all possible packet headers */
	iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)sizeof(MPIDI_CH3_Pkt_t);
	MPIU_DBG_STMT(CH3_CHANNEL,VERBOSE,
		   MPIDI_DBG_Print_packet((MPIDI_CH3_Pkt_t*)iov[0].MPID_IOV_BUF));
	
	if (vc->ch.state == MPIDI_CH3I_VC_STATE_CONNECTED) /* MT */
	{
		/* Connection already formed.  If send queue is empty attempt to send data, queuing any unsent data. */
		if (MPIDI_CH3I_SendQ_empty(vc)) /* MT */
		{
			int rc;
			MPIU_Size_t nb;

			MPIU_DBG_MSG(CH3_CHANNEL,VERBOSE, L"send queue empty, attempting to write");
			MPIU_DBG_PKT(vc->ch.conn, (MPIDI_CH3_Pkt_t*)iov[0].MPID_IOV_BUF, L"isend");
			
			/* MT - need some signalling to lock down our right to use the channel, thus insuring that the progress engine does
			   also try to write */
			rc = MPIDU_Sock_writev(vc->ch.sock, iov, n_iov, &nb);
			if (rc == MPI_SUCCESS)
			{
				int offset = 0;
	
				MPIU_DBG_MSG_D(CH3_CHANNEL,VERBOSE, L"wrote %ld bytes", (unsigned long) nb);
				
				while (offset < n_iov)
				{
					if (nb >= (MPIU_Size_t)iov[offset].MPID_IOV_LEN)
					{
						nb -= (MPIU_Size_t)iov[offset].MPID_IOV_LEN;
						offset++;
					}
					else
					{
						MPIU_DBG_MSG(CH3_CHANNEL,VERBOSE, L"partial write, request enqueued at head");
						sreq = create_request_istartmsgv(iov, n_iov, offset, nb);
						/* --BEGIN ERROR HANDLING-- */
						if (sreq == NULL)
						{
							mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
							goto fn_exit;
						}
						/* --END ERROR HANDLING-- */
						MPIDI_CH3I_SendQ_enqueue_head(vc, sreq);
						MPIU_DBG_MSG_FMT(CH3_CHANNEL,VERBOSE, (MPIU_DBG_FDEST,L"posting writev, vc=0x%p, sreq=0x%08x", vc, sreq->handle));
						vc->ch.conn->send_active = sreq;
						mpi_errno = MPIDU_Sock_post_writev(vc->ch.conn->sock, sreq->dev.iov + offset, sreq->dev.iov_count - offset, NULL);
						/* --BEGIN ERROR HANDLING-- */
						if (mpi_errno != MPI_SUCCESS)
						{
							mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER,
															 L"**ch3|sock|postwrite", L"ch3|sock|postwrite %p %p %p",
															 sreq, vc->ch.conn, vc);
						}
						/* --END ERROR HANDLING-- */
						break;
					}
				}

				if (offset == n_iov)
				{
					MPIU_DBG_MSG(CH3_CHANNEL,VERBOSE,L"entire write complete");
				}
			}
			/* --BEGIN ERROR HANDLING-- */
			else
			{
				MPIU_DBG_MSG_D(CH3_CHANNEL,TYPICAL, L"ERROR - MPIDU_Sock_writev failed, rc=%d", rc);
				sreq = MPID_Request_create();
				if (sreq == NULL)
				{
					mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
					goto fn_exit;
				}
				sreq->kind = MPID_REQUEST_SEND;
				sreq->cc = 0;
				/* TODO: Create an appropriate error message based on the return value */
				sreq->status.MPI_ERROR = MPI_ERR_INTERN;
			}
			/* --END ERROR HANDLING-- */
		}
		else
		{
			MPIU_DBG_MSG(CH3_CHANNEL,VERBOSE, L"send in progress, request enqueued");
			sreq = create_request_istartmsgv(iov, n_iov, 0, 0);
			/* --BEGIN ERROR HANDLING-- */
			if (sreq == NULL)
			{
				mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
				goto fn_exit;
			}
			/* --END ERROR HANDLING-- */
			MPIDI_CH3I_SendQ_enqueue(vc, sreq);
		}
	}
	else if (vc->ch.state == MPIDI_CH3I_VC_STATE_UNCONNECTED)
	{
		MPIU_DBG_VCUSE(vc, L"unconnected.  posting connect and enqueuing request");
		
		/* queue the data so it can be sent after the connection is formed */
		sreq = create_request_istartmsgv(iov, n_iov, 0, 0);
		/* --BEGIN ERROR HANDLING-- */
		if (sreq == NULL)
		{
			mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
			goto fn_exit;
		}
		/* --END ERROR HANDLING-- */
		MPIDI_CH3I_SendQ_enqueue(vc, sreq);
		
		/* Form a new connection */
		MPIDI_CH3I_VC_post_sockconnect(vc);
	}
	else if (vc->ch.state != MPIDI_CH3I_VC_STATE_FAILED)
	{
		/* Unable to send data at the moment, so queue it for later */
		MPIU_DBG_VCUSE(vc, L"forming connection, request enqueued");
		sreq = create_request_istartmsgv(iov, n_iov, 0, 0);
		/* --BEGIN ERROR HANDLING-- */
		if (sreq == NULL)
		{
			mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
			goto fn_exit;
		}
		/* --END ERROR HANDLING-- */
		MPIDI_CH3I_SendQ_enqueue(vc, sreq);
	}
	/* --BEGIN ERROR HANDLING-- */
	else
	{
		/* Connection failed, so allocate a request and return an error. */
		MPIU_DBG_VCUSE(vc, L"ERROR - connection failed");
		sreq = MPID_Request_create();
		if (sreq == NULL)
		{
			mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
			goto fn_exit;
		}
		sreq->kind = MPID_REQUEST_SEND;
		sreq->cc = 0;
		/* TODO: Create an appropriate error message */
		sreq->status.MPI_ERROR = MPI_ERR_INTERN;
	}
	/* --END ERROR HANDLING-- */

  fn_exit:
	*sreq_ptr = sreq;
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISTARTMSGV);
	return mpi_errno;
}

/* FIXME: No-one defines this.  Should we remove it and all uses? */
#undef USE_CH3I_PROGRESS_DELAY_QUEUE

/* FIXME: Move thread stuff into some set of abstractions in order to remove ifdefs */
#ifdef MPICH_IS_THREADED
	volatile int MPIDI_CH3I_progress_blocked = FALSE;
	volatile int MPIDI_CH3I_progress_wakeup_signalled = FALSE;

#   if (USE_THREAD_IMPL == MPICH_THREAD_IMPL_GLOBAL_MUTEX)
#	   if defined(USE_CH3I_PROGRESS_DELAY_QUEUE)
			struct MPIDI_CH3I_Progress_delay_queue_elem
			{
				unsigned int count;
				volatile int flag;
				MPID_Thread_cond_t cond;
				struct MPIDI_CH3I_Progress_delay_queue_elem * next;
			};

			static struct MPIDI_CH3I_Progress_delay_queue_elem * MPIDI_CH3I_Progress_delay_queue_head = NULL;
			static struct MPIDI_CH3I_Progress_delay_queue_elem * MPIDI_CH3I_Progress_delay_queue_tail = NULL;
#	   else
			/* This value must be static so that it isn't an uninitialized
			   common symbol */
			static MPID_Thread_cond_t MPIDI_CH3I_progress_completion_cond;
#	   endif
#   endif
#endif

#ifdef MPICH_IS_THREADED
	static int MPIDI_CH3I_Progress_delay(unsigned int completion_count);
	static int MPIDI_CH3I_Progress_continue(unsigned int completion_count);
#endif

#undef FUNCNAME
#define FUNCNAME sockMPIDI_CH3_Progress_test
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int sockMPIDI_CH3_Progress_test(void)
{
	MPIDU_Sock_event_t event;
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_PROGRESS_TEST);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_PROGRESS_TEST);

#   ifdef MPICH_IS_THREADED
	{
		/* We don't bother testing whether threads are enabled in the 
		   runtime-checking case because this simple test will always be false
		   if threads are not enabled. */
		if (MPIDI_CH3I_progress_blocked == TRUE) 
		{
			/*
			 * Another thread is already blocking in the progress engine.  
			 * We are not going to block waiting for progress, so we
			 * simply return.  It might make sense to yield before * returning,
			 * giving the PE thread a change to make progress.
			 *
			 * MT: Another thread is already blocking in poll.  Right now, 
			 * calls to the progress routines are effectively
			 * serialized by the device.  The only way another thread may 
			 * enter this function is if MPIDU_Sock_wait() blocks.  If
			 * this changes, a flag other than MPIDI_CH3I_Progress_blocked 
			 * may be required to determine if another thread is in
			 * the progress engine.
			 */
			
			goto fn_exit;
		}
	}
#   endif
	
	mpi_errno = MPIDU_Sock_wait(MPIDI_CH3I_sock_set, 0, &event);

	if (mpi_errno == MPI_SUCCESS)
	{
		mpi_errno = MPIDI_CH3I_Progress_handle_sock_event(&event);
		if (mpi_errno != MPI_SUCCESS)
		{
			MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, L"**ch3|sock|handle_sock_event");
		}
	}
	else if (MPIR_ERR_GET_CLASS(mpi_errno) == MPIDU_SOCK_ERR_TIMEOUT)
	{
		mpi_errno = MPI_SUCCESS;
		goto fn_exit;
	}
	else
	{
		MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, L"**progress_sock_wait");
	}

  fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_PROGRESS_TEST);
	return mpi_errno;
 fn_fail:
	goto fn_exit;
}
/* end sockMPIDI_CH3_Progress_test() */


#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_Progress_wait
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int sockMPIDI_CH3_Progress_wait(MPID_Progress_state * progress_state)
{
	MPIDU_Sock_event_t event;
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_PROGRESS_WAIT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_PROGRESS_WAIT);

	/*
	 * MT: the following code will be needed if progress can occur between 
	 * MPIDI_CH3_Progress_start() and
	 * MPIDI_CH3_Progress_wait(), or iterations of MPIDI_CH3_Progress_wait().
	 *
	 * This is presently not possible, and thus the code is commented out.
	 */
#   if 0
	/* FIXME: Was (USE_THREAD_IMPL == MPICH_THREAD_IMPL_NOT_IMPLEMENTED),
	   which really meant not-using-global-mutex-thread model .  This
	   was true for the single threaded case, but was probably not intended
	   for that case*/
	{
		if (progress_state->ch.completion_count != MPIDI_CH3I_progress_completion_count)
		{
			goto fn_exit;
		}
	}
#   endif
		
#   ifdef MPICH_IS_THREADED
	MPIU_THREAD_CHECK_BEGIN;
	{
		if (MPIDI_CH3I_progress_blocked == TRUE) 
		{
			/*
			 * Another thread is already blocking in the progress engine.
			 *
			 * MT: Another thread is already blocking in poll.  Right now, 
			 * calls to MPIDI_CH3_Progress_wait() are effectively
			 * serialized by the device.  The only way another thread may 
			 * enter this function is if MPIDU_Sock_wait() blocks.  If
			 * this changes, a flag other than MPIDI_CH3I_Progress_blocked 
			 * may be required to determine if another thread is in
			 * the progress engine.
			 */
			MPIDI_CH3I_Progress_delay(MPIDI_CH3I_progress_completion_count);
				
			goto fn_exit;
		}
	}
	MPIU_THREAD_CHECK_END;
#   endif
	
	do
	{
#	   if MPICH_IS_THREADED
		MPIU_THREAD_CHECK_BEGIN;
		{
			MPIDI_CH3I_progress_blocked = TRUE;
		}
		MPIU_THREAD_CHECK_END;
		mpi_errno = MPIDU_Sock_wait(MPIDI_CH3I_sock_set, 
									MPIDU_SOCK_INFINITE_TIME, &event);
		MPIU_THREAD_CHECK_BEGIN;
		{
			MPIDI_CH3I_progress_blocked = FALSE;
			MPIDI_CH3I_progress_wakeup_signalled = FALSE;
		}
		MPIU_THREAD_CHECK_END;
#	   else
		mpi_errno = MPIDU_Sock_wait(MPIDI_CH3I_sock_set, 
									MPIDU_SOCK_INFINITE_TIME, &event);
#		endif

		/* --BEGIN ERROR HANDLING-- */
		if (mpi_errno != MPI_SUCCESS)
		{
			MPIU_Assert(MPIR_ERR_GET_CLASS(mpi_errno) != MPIDU_SOCK_ERR_TIMEOUT);
			MPIU_ERR_SET(mpi_errno,MPI_ERR_OTHER,L"**progress_sock_wait");
			goto fn_fail;
		}
		/* --END ERROR HANDLING-- */

		mpi_errno = MPIDI_CH3I_Progress_handle_sock_event(&event);
		if (mpi_errno != MPI_SUCCESS)
		{
			MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, L"**ch3|sock|handle_sock_event");
		}
	}
	while (progress_state->ch.completion_count == MPIDI_CH3I_progress_completion_count);

	/*
	 * We could continue to call MPIU_Sock_wait in a non-blocking fashion 
	 * and process any other events; however, this would not
	 * give the application a chance to post new receives, and thus could 
	 * result in an increased number of unexpected messages
	 * that would need to be buffered.
	 */
	
#   if MPICH_IS_THREADED
	{
		/*
		 * Awaken any threads which are waiting for the progress that just 
		 * occurred
		 */
		MPIDI_CH3I_Progress_continue(MPIDI_CH3I_progress_completion_count);
	}
#   endif
	
 fn_exit:
	/*
	 * Reset the progress state so it is fresh for the next iteration
	 */
	progress_state->ch.completion_count = MPIDI_CH3I_progress_completion_count;
	
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_PROGRESS_WAIT);
	return mpi_errno;
 fn_fail:
	goto fn_exit;
}
/* end MPIDI_CH3_Progress_wait() */


#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_Connection_terminate
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int sockMPIDI_CH3_Connection_terminate(MPIDI_VC_t * vc)
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_CONNECTION_TERMINATE);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_CONNECTION_TERMINATE);

	MPIU_DBG_CONNSTATECHANGE(vc,vc->ch.conn, CONN_STATE_CLOSING);
	vc->ch.conn->state = CONN_STATE_CLOSING;
	mpi_errno = MPIDU_Sock_post_close(vc->ch.sock);
	if (mpi_errno != MPI_SUCCESS)
	{
		MPIU_ERR_POP(mpi_errno);
	}

  fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_CONNECTION_TERMINATE);
	return mpi_errno;
 fn_fail:
	goto fn_exit;
}
/* end MPIDI_CH3_Connection_terminate() */


#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_Progress_init
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int MPIDI_CH3I_Progress_init(void)
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_PROGRESS_INIT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_PROGRESS_INIT);

	/* FIXME: Should this be within a check_begin/end block? */
#   if (USE_THREAD_IMPL == MPICH_THREAD_IMPL_GLOBAL_MUTEX && !defined(USE_CH3I_PROGRESS_DELAY_QUEUE))
	{
		MPID_Thread_cond_create(&MPIDI_CH3I_progress_completion_cond, NULL);
	}
#   endif
		
	mpi_errno = MPIDU_Sock_init();
	if (mpi_errno != MPI_SUCCESS)
	{
		MPIU_ERR_POP(mpi_errno);
	}
	
	/* create sock set */
	mpi_errno = MPIDU_Sock_create_set(&MPIDI_CH3I_sock_set);
	if (mpi_errno != MPI_SUCCESS)
	{
		MPIU_ERR_POP(mpi_errno);
	}
	
	/* establish non-blocking listener */
	mpi_errno = MPIDU_CH3I_SetupListener( MPIDI_CH3I_sock_set );
	if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }
	
  fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_PROGRESS_INIT);
	return mpi_errno;
 fn_fail:
	goto fn_exit;
}
/* end MIPDI_CH3I_Progress_init() */

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_Progress_finalize
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int sockMPIDI_CH3I_Progress_finalize(void)
{
	int mpi_errno;
	/*MPID_Progress_state progress_state;*/
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_PROGRESS_FINALIZE);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_PROGRESS_FINALIZE);

	/* Shut down the listener */
	mpi_errno = MPIDU_CH3I_ShutdownListener();
	if (mpi_errno != MPI_SUCCESS) { MPIU_ERR_POP(mpi_errno); }
	
	/* FIXME: Cleanly shutdown other socks and free connection structures. 
	   (close protocol?) */


	/*
	 * MT: in a multi-threaded environment, finalize() should signal any 
	 * thread(s) blocking on MPIDU_Sock_wait() and wait for
	 * those * threads to complete before destroying the progress engine 
	 * data structures.
	 */

	mpi_errno = MPIDU_Sock_destroy_set(MPIDI_CH3I_sock_set);
	MPIU_ERR_CHK_POP(mpi_errno);
	mpi_errno = MPIDU_Sock_finalize();
	MPIU_ERR_CHK_POP(mpi_errno);

	/* FIXME: Should this be within a check_begin/end block? */
#   if (USE_THREAD_IMPL == MPICH_THREAD_IMPL_GLOBAL_MUTEX && !defined(USE_CH3I_PROGRESS_DELAY_QUEUE))
	{
		MPID_Thread_cond_destroy(&MPIDI_CH3I_progress_completion_cond, NULL);
	}
#   endif
	
  fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_PROGRESS_FINALIZE);
	return mpi_errno;
 fn_fail:
	goto fn_exit;
}
/* end MPIDI_CH3I_Progress_finalize() */

#ifdef MPICH_IS_THREADED
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_Progress_wakeup
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
void MPIDI_CH3I_Progress_wakeup(void)
{
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_PROGRESS_WAKEUP);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_PROGRESS_WAKEUP);
	MPIDU_Sock_wakeup(MPIDI_CH3I_sock_set);
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_PROGRESS_WAKEUP);
}
#endif

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_Get_business_card
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int sockMPIDI_CH3I_Get_business_card(wchar_t *value, int length)
{
	int mpi_errno;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_GET_BUSINESS_CARD);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_GET_BUSINESS_CARD);
	mpi_errno = MPIDI_CH3U_Get_business_card_sock(&value, &length);
	MPIU_ERR_CHK_POP(mpi_errno);
fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_GET_BUSINESS_CARD);
	return mpi_errno;
}


#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_Progress_handle_sock_event
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int MPIDI_CH3I_Progress_handle_sock_event(MPIDU_Sock_event_t * event)
{
	int complete;
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_PROGRESS_HANDLE_SOCK_EVENT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_PROGRESS_HANDLE_SOCK_EVENT);

	MPIU_DBG_MSG_D(CH3,VERBOSE,L"Socket event of type %d", event->op_type );

	switch (event->op_type)
	{
		case MPIDU_SOCK_OP_READ:
		{
			MPIDI_CH3I_Connection_t * conn = (MPIDI_CH3I_Connection_t *) event->user_ptr;
				
			MPID_Request * rreq = conn->recv_active;

			/* --BEGIN ERROR HANDLING-- */
			if (event->error != MPI_SUCCESS)
			{
				/* FIXME: the following should be handled by the close protocol */
				if (MPIR_ERR_GET_CLASS(event->error) != MPIDU_SOCK_ERR_CONN_CLOSED) {
					/*printf("error received interrupting a posted receive operation.\n");fflush(stdout);*/
					if ((conn->vc != NULL) && ((conn->vc->state == MPIDI_VC_STATE_LOCAL_CLOSE) || (conn->vc->state == MPIDI_VC_STATE_CLOSE_ACKED)))
					{
						/* Handle a broken connection the same as an ack */
						conn->vc->state = MPIDI_VC_STATE_CLOSE_ACKED;
						sockMPIDI_CH3_Connection_terminate(conn->vc);
					}
					else
					{
						mpi_errno = event->error;
						MPIU_ERR_POP(mpi_errno);
					}
				}
				/*
				else
				{
					if (conn->vc != NULL)
						printf("OP_READ:MPIDU_SOCK_ERR_CONN_CLOSED: %p\n", conn->vc->pg);
					else
						printf("OP_READ:MPIDU_SOCK_ERR_CONN_CLOSED\n");
					fflush(stdout);
				}
				*/
				break;
			}
			/* --END ERROR HANDLING-- */
				
			if (conn->state == CONN_STATE_CONNECTED)
			{
				if (conn->recv_active == NULL)
				{
					MPIU_Assert(conn->pkt.type < MPIDI_CH3_PKT_END_CH3);
						
					mpi_errno = MPIDI_CH3U_Handle_recv_pkt(conn->vc, &conn->pkt, &rreq);
					if (mpi_errno != MPI_SUCCESS)
					{
						MPIU_ERR_POP(mpi_errno);
					}

					if (rreq == NULL)
					{
						if (conn->state != CONN_STATE_CLOSING)
						{
							/* conn->recv_active = NULL;  -- already set to NULL */
							mpi_errno = connection_post_recv_pkt(conn);
							if (mpi_errno != MPI_SUCCESS)
							{
								MPIU_ERR_POP(mpi_errno);
							}
						}
					}
					else
					{
						for(;;)
						{
							MPID_IOV * iovp;
							MPIU_Size_t nb;
								
							iovp = rreq->dev.iov;
							
							mpi_errno = MPIDU_Sock_readv(conn->sock, iovp, rreq->dev.iov_count, &nb);
							/* --BEGIN ERROR HANDLING-- */
							if (mpi_errno != MPI_SUCCESS)
							{
								mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER,
																 L"**ch3|sock|immedread", L"ch3|sock|immedread %p %p %p",
																 rreq, conn, conn->vc);
								goto fn_fail;
							}
							/* --END ERROR HANDLING-- */

							MPIU_DBG_MSG_FMT(CH3_CHANNEL,VERBOSE, (MPIU_DBG_FDEST,L"immediate readv, vc=%p nb=%d, rreq=0x%08x", conn->vc, nb, rreq->handle));

							if (nb > 0 && adjust_iov(&iovp, &rreq->dev.iov_count, nb))
							{
								mpi_errno = MPIDI_CH3U_Handle_recv_req(conn->vc, rreq, &complete);
								if (mpi_errno != MPI_SUCCESS) {
									MPIU_ERR_POP(mpi_errno);
								}

								if (complete)
								{
									/* conn->recv_active = NULL; -- already set to NULL */
									mpi_errno = connection_post_recv_pkt(conn);
									if (mpi_errno != MPI_SUCCESS) {
										MPIU_ERR_POP(mpi_errno);
									}

									break;
								}
							}
							else
							{
								MPIU_DBG_MSG_FMT(CH3_CHANNEL,VERBOSE, (MPIU_DBG_FDEST,L"posting readv, vc=%p, rreq=0x%08x", conn->vc, rreq->handle));
								conn->recv_active = rreq;
								mpi_errno = MPIDU_Sock_post_readv(conn->sock, iovp, rreq->dev.iov_count, NULL);
								/* --BEGIN ERROR HANDLING-- */
								if (mpi_errno != MPI_SUCCESS)
								{
									mpi_errno = MPIR_Err_create_code(
										mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|sock|postread",
										L"ch3|sock|postread %p %p %p", rreq, conn, conn->vc);
									goto fn_fail;
								}
								/* --END ERROR HANDLING-- */

								break;
							}
						}
					}
				}
				else /* incoming data */
				{
					mpi_errno = MPIDI_CH3U_Handle_recv_req(conn->vc, rreq, &complete);
					if (mpi_errno != MPI_SUCCESS)
					{
						MPIU_ERR_POP(mpi_errno);
					}
						
					if (complete)
					{
						conn->recv_active = NULL;
						mpi_errno = connection_post_recv_pkt(conn);
						if (mpi_errno != MPI_SUCCESS)
						{
							MPIU_ERR_POP(mpi_errno);
						}
					}
					else /* more data to be read */
					{
						for(;;)
						{
							MPID_IOV * iovp;
							MPIU_Size_t nb;
								
							iovp = rreq->dev.iov;
							
							mpi_errno = MPIDU_Sock_readv(conn->sock, iovp, rreq->dev.iov_count, &nb);
							/* --BEGIN ERROR HANDLING-- */
							if (mpi_errno != MPI_SUCCESS)
							{
								mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER,
																 L"**ch3|sock|immedread", L"ch3|sock|immedread %p %p %p",
																 rreq, conn, conn->vc);
								goto fn_fail;
							}
							/* --END ERROR HANDLING-- */

							MPIU_DBG_MSG_FMT(CH3_CHANNEL,VERBOSE, (MPIU_DBG_FDEST,L"immediate readv, vc=%p nb=%d, rreq=0x%08x", conn->vc, rreq->handle, nb));

							if (nb > 0 && adjust_iov(&iovp, &rreq->dev.iov_count, nb))
							{
								mpi_errno = MPIDI_CH3U_Handle_recv_req(conn->vc, rreq, &complete);
								if (mpi_errno != MPI_SUCCESS)
								{
									MPIU_ERR_POP(mpi_errno);
								}

								if (complete)
								{
									conn->recv_active = NULL;
									mpi_errno = connection_post_recv_pkt(conn);
									if (mpi_errno != MPI_SUCCESS)
									{
										MPIU_ERR_POP(mpi_errno);
									}

									break;
								}
							}
							else
							{
								MPIU_DBG_MSG_FMT(CH3_CHANNEL,VERBOSE, (MPIU_DBG_FDEST,L"posting readv, vc=%p, rreq=0x%08x", conn->vc, rreq->handle));
								/* conn->recv_active = rreq;  -- already set to current request */
								mpi_errno = MPIDU_Sock_post_readv(conn->sock, iovp, rreq->dev.iov_count, NULL);
								/* --BEGIN ERROR HANDLING-- */
								if (mpi_errno != MPI_SUCCESS)
								{
									mpi_errno = MPIR_Err_create_code(
										mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|sock|postread",
										L"ch3|sock|postread %p %p %p", rreq, conn, conn->vc);
									goto fn_fail;
								}
								/* --END ERROR HANDLING-- */

								break;
							}
						}
					}
				}
			}
			else if (conn->state == CONN_STATE_OPEN_LRECV_DATA)
			{
				mpi_errno = MPIDI_CH3_Sockconn_handle_connopen_event( conn );
				if (mpi_errno) { MPIU_ERR_POP( mpi_errno ); }
			}
			else /* Handling some internal connection establishment or tear down packet */
			{ 
				mpi_errno = MPIDI_CH3_Sockconn_handle_conn_event( conn );
				if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }
			}
			break;
		}

		/* END OF SOCK_OP_READ */

		case MPIDU_SOCK_OP_WRITE:
		{
			MPIDI_CH3I_Connection_t * conn = (MPIDI_CH3I_Connection_t *) event->user_ptr;
			/* --BEGIN ERROR HANDLING-- */
			if (event->error != MPI_SUCCESS)
			{
				mpi_errno = event->error;
				MPIU_ERR_POP(mpi_errno);
			}
			/* --END ERROR HANDLING-- */
				
			if (conn->send_active)
			{
				MPID_Request * sreq = conn->send_active;

				mpi_errno = MPIDI_CH3U_Handle_send_req(conn->vc, sreq, &complete);
				if (mpi_errno != MPI_SUCCESS)
				{
					MPIU_ERR_POP(mpi_errno);
				}
					
				if (complete)
				{
					MPIDI_CH3I_SendQ_dequeue(conn->vc);
					mpi_errno = connection_post_sendq_req(conn);
					if (mpi_errno != MPI_SUCCESS)
					{
						MPIU_ERR_POP(mpi_errno);
					}
				}
				else /* more data to send */
				{
					for(;;)
					{
						MPID_IOV * iovp;
						MPIU_Size_t nb;
								
						iovp = sreq->dev.iov;
							
						mpi_errno = MPIDU_Sock_writev(conn->sock, iovp, sreq->dev.iov_count, &nb);
						/* --BEGIN ERROR HANDLING-- */
						if (mpi_errno != MPI_SUCCESS)
						{
							mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER,
															 L"**ch3|sock|immedwrite", L"ch3|sock|immedwrite %p %p %p",
															 sreq, conn, conn->vc);
							goto fn_fail;
						}
						/* --END ERROR HANDLING-- */

						MPIU_DBG_MSG_FMT(CH3_CHANNEL,VERBOSE, (MPIU_DBG_FDEST,L"immediate writev, vc=%p, sreq=0x%08x, nb=%d", conn->vc, sreq->handle, nb));

						if (nb > 0 && adjust_iov(&iovp, &sreq->dev.iov_count, nb))
						{
							mpi_errno = MPIDI_CH3U_Handle_send_req(conn->vc, sreq, &complete);
							if (mpi_errno != MPI_SUCCESS)
							{
								MPIU_ERR_POP(mpi_errno);
							}

							if (complete)
							{
								MPIDI_CH3I_SendQ_dequeue(conn->vc);
								mpi_errno = connection_post_sendq_req(conn);
								if (mpi_errno != MPI_SUCCESS)
								{
									MPIU_ERR_POP(mpi_errno);
								}
								break;
							}
						}
						else
						{
							MPIU_DBG_MSG_FMT(CH3_CHANNEL,VERBOSE, (MPIU_DBG_FDEST, L"posting writev, vc=%p, conn=%p, sreq=0x%08x", conn->vc, conn, sreq->handle));
							mpi_errno = MPIDU_Sock_post_writev(conn->sock, iovp, sreq->dev.iov_count, NULL);
							/* --BEGIN ERROR HANDLING-- */
							if (mpi_errno != MPI_SUCCESS)
							{
								mpi_errno = MPIR_Err_create_code(
									mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|sock|postwrite",
									L"ch3|sock|postwrite %p %p %p", sreq, conn, conn->vc);
								goto fn_fail;
							}
							/* --END ERROR HANDLING-- */

							break;
						}
					}
				}
			}
			else /* finished writing internal packet header */
			{
				mpi_errno = MPIDI_CH3_Sockconn_handle_connwrite( conn );
				if (mpi_errno) { MPIU_ERR_POP( mpi_errno ); }
			}
			break;
		}
		/* END OF SOCK_OP_WRITE */

		case MPIDU_SOCK_OP_ACCEPT:
		{
			mpi_errno = MPIDI_CH3_Sockconn_handle_accept_event();
			if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }
			break;
		}
			
		case MPIDU_SOCK_OP_CONNECT:
		{
			mpi_errno = MPIDI_CH3_Sockconn_handle_connect_event( (MPIDI_CH3I_Connection_t *) event->user_ptr, event->error );
			if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }
			break;
		}
			
		case MPIDU_SOCK_OP_CLOSE:
		{
			mpi_errno = MPIDI_CH3_Sockconn_handle_close_event( (MPIDI_CH3I_Connection_t *) event->user_ptr );
			if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }
			break;
		}

		case MPIDU_SOCK_OP_WAKEUP:
		{
			MPIDI_CH3_Progress_signal_completion();
			/* MPIDI_CH3I_progress_completion_count++; */
			break;
		}
	}

 fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_PROGRESS_HANDLE_SOCK_EVENT);
	return mpi_errno;
 fn_fail:
	goto fn_exit;
}
/* end MPIDI_CH3I_Progress_handle_sock_event() */


#ifdef MPICH_IS_THREADED

/* Note that this routine is only called if threads are enabled; 
   it does not need to check whether runtime threads are enabled */
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_Progress_delay
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int MPIDI_CH3I_Progress_delay(unsigned int completion_count)
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_PROGRESS_DELAY);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_PROGRESS_DELAY);
	
#   if (USE_THREAD_IMPL == MPICH_THREAD_IMPL_GLOBAL_MONITOR)
	{
#		error This is so not right.  But what is the correct technique?
		
		if (MPIU_Monitor_closet_get_occupancy_count(MPIR_Process.global_closet) > 0)
		{
			MPIU_Monitor_continue(MPIR_Process.global_monitor, MPIR_Process.global_closet);
			MPIU_Monitor_enter(MPIR_Process.global_monitor);
			if (completion_count != MPIDI_CH3I_progress_completion_count)
			{
				goto impl_exit;
			}
		}
					
		MPIU_Monitor_delay(MPIR_Process.global_monitor, MPIR_Process.global_closet);

	  impl_exit:
		{
		}
	}
#   elif (USE_THREAD_IMPL == MPICH_THREAD_IMPL_GLOBAL_MUTEX)
	{
#		if defined(USE_CH3I_PROGRESS_DELAY_QUEUE)
		{
			int rc;
			struct MPIDI_CH3I_Progress_delay_queue_elem dq_elem;
		
			dq_elem.count = completion_count;
			dq_elem.flag = FALSE;
	
			dq_elem.next = NULL;
			MPIDI_CH3I_Progress_delay_queue_tail->next = &dq_elem;
			MPIDI_CH3I_Progress_delay_queue_tail = &dq_elem;
			if (MPIDI_CH3I_Progress_delay_queue_head == NULL)
			{
				MPIDI_CH3I_Progress_delay_queue_head = &dq_elem;
			}

			rc = MPID_Thread_cond_create(&dq_elem.cond, NULL);
			/* --BEGIN ERROR HANDLING-- */
			if (rc != 0)
			{ 
				mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", NULL);
				goto impl_exit;
			}
			/* --END ERROR HANDLING-- */
	
			do
			{
				MPID_Thread_cond_wait(&dq_elem.cond, &MPIR_Process.global_mutex);
			}
			while(dq_elem.flag == FALSE);
			
			MPID_Thread_cond_destroy(&dq_elem.cond, NULL);
			
		  impl_exit:
			{
			}
		}
#		else
		{ 
			while (completion_count == MPIDI_CH3I_progress_completion_count)
			{
				MPID_Thread_cond_wait(&MPIDI_CH3I_progress_completion_cond, &MPIR_Process.global_mutex);
			}
		}
#	   endif
	}
#   endif

	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_PROGRESS_DELAY);
	return mpi_errno;
}
/* end MPIDI_CH3I_Progress_delay() */


#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_Progress_continue
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int MPIDI_CH3I_Progress_continue(unsigned int completion_count)
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_PROGRESS_CONTINUE);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_PROGRESS_CONTINUE);

#   if (USE_THREAD_IMPL == MPICH_THREAD_IMPL_GLOBAL_MONITOR)
	{
#		error This is so not right.  But what is the correct technique?
		if (MPIU_Monitor_closet_get_occupancy(MPIR_Process.global_closet) > 0)
		{
			MPIU_Monitor_continue(MPIR_Process.global_monitor, MPIR_Process.global_closet);
		}
		else
		{ 
			MPIU_Monitor_exit(MPIR_Process.global_monitor);
		}
	}
#   elif (USE_THREAD_IMPL == MPICH_THREAD_IMPL_GLOBAL_MUTEX)
	{
#		if defined(USE_CH3I_PROGRESS_DELAY_QUEUE)
		{
			struct MPIDI_CH3I_Progress_delay_queue_elem * dq_elem;
			
			dq_elem = MPIDI_CH3I_Progress_delay_queue_head;
			while(dq_elem != NULL && dq_elem->count != completion_count)
			{
				dq_elem->flag = TRUE;
				MPID_Thread_cond_signal(&dq_elem->cond);
				dq_elem = dq_elem->next;
			}
			MPIDI_CH3I_Progress_delay_queue_head = dq_elem;
		}
#		else
		{
			MPID_Thread_cond_broadcast(&MPIDI_CH3I_progress_completion_cond);
		}
#		endif
	}
#   endif

	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_PROGRESS_CONTINUE);
	return mpi_errno;
}
/* end MPIDI_CH3I_Progress_continue() */

#endif /* (USE_THREAD_IMPL == MPICH_THREAD_IMPL_GLOBAL) */

#undef FUNCNAME
#define FUNCNAME connection_post_sendq_req
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static inline int connection_post_sendq_req(MPIDI_CH3I_Connection_t * conn)
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_CONNECTION_POST_SENDQ_REQ);

	MPIDI_FUNC_ENTER(MPID_STATE_CONNECTION_POST_SENDQ_REQ);
	/* post send of next request on the send queue */
	conn->send_active = MPIDI_CH3I_SendQ_head(conn->vc); /* MT */
	if (conn->send_active != NULL)
	{
		MPIU_DBG_MSG_P(CH3,TYPICAL,"conn=%p: Posting message from connection send queue", conn );
		mpi_errno = MPIDU_Sock_post_writev(conn->sock, conn->send_active->dev.iov, conn->send_active->dev.iov_count, NULL);
		/* --BEGIN ERROR HANDLING-- */
		if (mpi_errno != MPI_SUCCESS)
		{
			mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", NULL);
		}
		/* --END ERROR HANDLING-- */
	}
	
	MPIDI_FUNC_EXIT(MPID_STATE_CONNECTION_POST_SENDQ_REQ);
	return mpi_errno;
}



#undef FUNCNAME
#define FUNCNAME connection_post_recv_pkt
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static inline int connection_post_recv_pkt(MPIDI_CH3I_Connection_t * conn)
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_CONNECTION_POST_RECV_PKT);

	MPIDI_FUNC_ENTER(MPID_STATE_CONNECTION_POST_RECV_PKT);

	mpi_errno = MPIDU_Sock_post_read(conn->sock, &conn->pkt, sizeof(conn->pkt), sizeof(conn->pkt), NULL);
	if (mpi_errno != MPI_SUCCESS) {
		MPIU_ERR_SET(mpi_errno,MPI_ERR_OTHER, L"**fail");
	}
	
	MPIDI_FUNC_EXIT(MPID_STATE_CONNECTION_POST_RECV_PKT);
	return mpi_errno;
}


#undef FUNCNAME
#define FUNCNAME adjust_iov
#undef FCNAME
#define FCNAME MPIU_QUOTE(FUNCNAME)
static int adjust_iov(MPID_IOV ** iovp, int * countp, MPIU_Size_t nb)
{
	MPID_IOV * const iov = *iovp;
	const int count = *countp;
	int offset = 0;
	MPIDI_STATE_DECL(MPID_STATE_ADJUST_IOV);

	MPIDI_FUNC_ENTER(MPID_STATE_ADJUST_IOV);
	
	while (offset < count)
	{
		if (iov[offset].MPID_IOV_LEN <= (MPID_IOV_LEN_CAST)nb)
		{
			nb -= (MPIU_Size_t)iov[offset].MPID_IOV_LEN;
			offset++;
		}
		else
		{
			iov[offset].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)((char *) iov[offset].MPID_IOV_BUF + nb);
			iov[offset].MPID_IOV_LEN -= (MPID_IOV_LEN_CAST)nb;
			break;
		}
	}

	*iovp += offset;
	*countp -= offset;

	MPIDI_FUNC_EXIT(MPID_STATE_ADJUST_IOV);
	return (*countp == 0);
}
/* end adjust_iov() */

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_iStartRndvMsg
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int sockMPIDI_CH3_iStartRndvMsg(MPIDI_VC_t * vc, MPID_Request * sreq, MPIDI_CH3_Pkt_t * rts_full_pkt)
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_CH3_Pkt_rndv_req_to_send_t * const rts_pkt = &rts_full_pkt->rndv_req_to_send;
	MPID_Request * rts_sreq;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_ISTARTRNDVMSG);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_ISTARTRNDVMSG);

	mpi_errno = sockMPIDI_CH3_iStartMsg(vc, rts_pkt, sizeof(*rts_pkt), &rts_sreq);
	/* --BEGIN ERROR HANDLING-- */
	if (mpi_errno != MPI_SUCCESS)
	{
		MPIU_Object_set_ref(sreq, 0);
		MPIDI_CH3_Request_destroy(sreq);
		sreq = NULL;
		mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|rtspkt", 0);
		goto fn_exit;
	}
	/* --END ERROR HANDLING-- */
	if (rts_sreq != NULL)
	{
		/* --BEGIN ERROR HANDLING-- */
		if (rts_sreq->status.MPI_ERROR != MPI_SUCCESS)
		{
			MPIU_Object_set_ref(sreq, 0);
			MPIDI_CH3_Request_destroy(sreq);
			mpi_errno = MPIR_Err_create_code(rts_sreq->status.MPI_ERROR, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|rtspkt", 0);
			MPID_Request_release(rts_sreq);
			goto fn_exit;
		}
		/* --END ERROR HANDLING-- */
		MPID_Request_release(rts_sreq);
	}

fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISTARTRNDVMSG);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_iStartRndvTransfer
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int sockMPIDI_CH3_iStartRndvTransfer(MPIDI_VC_t * vc, MPID_Request * rreq)
{
	int mpi_errno = MPI_SUCCESS;
	MPID_Request * cts_req;
	MPIDI_CH3_Pkt_t upkt;
	MPIDI_CH3_Pkt_rndv_clr_to_send_t * cts_pkt = &upkt.rndv_clr_to_send;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_ISTARTRNDVTRANSFER);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_ISTARTRNDVTRANSFER);

	MPIDI_DBG_PRINTF((15, FCNAME, L"rndv RTS in the request, sending rndv CTS"));

	MPIDI_Pkt_init(cts_pkt, MPIDI_CH3_PKT_RNDV_CLR_TO_SEND);
	cts_pkt->sender_req_id = rreq->dev.sender_req_id;
	cts_pkt->receiver_req_id = rreq->handle;
	mpi_errno = sockMPIDI_CH3_iStartMsg(vc, cts_pkt, sizeof(*cts_pkt), &cts_req);
	/* --BEGIN ERROR HANDLING-- */
	if (mpi_errno != MPI_SUCCESS)
	{
		MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, L"**ch3|ctspkt");
	}
	/* --END ERROR HANDLING-- */
	if (cts_req != NULL)
	{
		/* FIXME: Ideally we could specify that a req not be returned.  This would avoid our having to decrement the reference count on a req we don't want/need. */
		/* --BEGIN ERROR HANDLING-- */
		if (cts_req->status.MPI_ERROR != MPI_SUCCESS)
		{
			MPIU_Object_set_ref(rreq, 0);
			MPIDI_CH3_Request_destroy(rreq);
			mpi_errno = MPIR_Err_create_code(cts_req->status.MPI_ERROR, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|ctspkt", 0);
			MPID_Request_release(cts_req);
			goto fn_fail;
		}
		/* --END ERROR HANDLING-- */
		MPID_Request_release(cts_req);
	}

fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISTARTRNDVTRANSFER);
	return mpi_errno;
}

int sockMPIDI_CH3_Init_fns(MPIDI_CH3_Functions_t *function_table)
{
	function_table->Init = sockMPIDI_CH3_Init;
	function_table->PortFnsInit = sockMPIDI_CH3_PortFnsInit;
	function_table->RMAFnsInit = MPIDI_CH3U_RMAFnsInit;
	function_table->Finalize = sockMPIDI_CH3_Finalize;
	function_table->VC_Init = sockMPIDI_CH3_VC_Init;
	function_table->PG_Init = NULL;
	function_table->iStartMsg = sockMPIDI_CH3_iStartMsg;
	function_table->iStartMsgv = sockMPIDI_CH3_iStartMsgv;
	function_table->iSend = sockMPIDI_CH3_iSend;
	function_table->iSendv = sockMPIDI_CH3_iSendv;
	function_table->Progress_wait = sockMPIDI_CH3_Progress_wait;
	function_table->Progress_test = sockMPIDI_CH3_Progress_test;
	function_table->Connection_terminate = sockMPIDI_CH3_Connection_terminate;
	function_table->Connect_to_root = sockMPIDI_CH3_Connect_to_root;
	function_table->Complete_unidirectional_connection_connect = MPIDI_CH3_Complete_unidirectional_connection_connect;
	function_table->Complete_unidirectional_connection_accept = MPIDI_CH3_Complete_unidirectional_connection_accept;
	function_table->Cleanup_after_connection = MPIDI_CH3_Cleanup_after_connection;
	function_table->iStartRndvMsg = sockMPIDI_CH3_iStartRndvMsg;
	function_table->iStartRndvTransfer = sockMPIDI_CH3_iStartRndvTransfer;
	function_table->Get_business_card = sockMPIDI_CH3I_Get_business_card;
	function_table->VC_NeedsSendClose = MPIDI_CH3U_VC_NeedsSendClose;
	function_table->VC_SendClose = MPIDI_CH3U_VC_SendClose;
	return MPI_SUCCESS;
}
