/*  (C) 2009 by Deino Software. */
#include "mpidi_ch3_impl.h"
#include "pmi.h"
#include <stdio.h>

/* local prototypes */
static int sshmMPIDI_CH3I_Progress(int is_blocking, MPID_Progress_state *state);
static int sshmMPIDI_CH3_Connect_to_root(const wchar_t * port_name, MPIDI_VC_t ** new_vc);
static int MPIDI_CH3I_VC_post_shmconnect(MPIDI_VC_t * vc);

/* global variables */
static MPIDI_CH3I_Alloc_mem_list_t *MPIDI_CH3I_Alloc_mem_list_head = NULL;

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_Complete_unidirectional_connection
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int sshmMPIDI_CH3_Complete_unidirectional_connection_connect( MPIDI_VC_t *connect_vc )
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_VC_t *accept_vc = NULL;
	MPID_Progress_state progress_state;
	int port_name_tag;
	wchar_t connector_port[MPI_MAX_PORT_NAME];
	int num_written = 0;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_COMPLETE_UNIDIRECTIONAL_CONNECTION);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_COMPLETE_UNIDIRECTIONAL_CONNECTION);

	/* If the VC creates non-duplex connections then the acceptor will
	 * need to connect back to form the other half of the connection.
	 * This code accepts the return connection. 
	 */

	/* Open a port on this side and send it to the acceptor so it can connect back */
	mpi_errno = MPID_Open_port(NULL, connector_port);
	if (mpi_errno != MPI_SUCCESS)
	{
		MPIU_ERR_POP(mpi_errno);
	}
	mpi_errno = MPIDI_CH3I_SHM_write(connect_vc, connector_port, MPI_MAX_PORT_NAME*sizeof(wchar_t), &num_written);
	if (mpi_errno != MPI_SUCCESS || num_written != MPI_MAX_PORT_NAME)
	{
		MPIU_ERR_POP(mpi_errno);
	}

	/* extract the tag from the port_name */
	mpi_errno = MPIDI_GetTagFromPort( connector_port, &port_name_tag);
	if (mpi_errno != MPIU_STR_SUCCESS)
	{
		MPIU_ERR_POP(mpi_errno);
	}

	/* dequeue the accept queue to see if a connection with the
	   root on the connect side has been formed in the progress
	   engine (the connection is returned in the form of a vc). If
	   not, poke the progress engine. */

	MPID_Progress_start(&progress_state);
	for(;;)
	{
		MPIDI_CH3I_Acceptq_dequeue(&accept_vc, port_name_tag);
		if (accept_vc != NULL)
		{
			break;
		}

		mpi_errno = MPID_Progress_wait(&progress_state);
		/* --BEGIN ERROR HANDLING-- */
		if (mpi_errno)
		{
			MPID_Progress_end(&progress_state);
			MPIU_ERR_POP(mpi_errno);
		}
		/* --END ERROR HANDLING-- */
	}
	MPID_Progress_end(&progress_state);

	/* merge the newly accepted VC with the connect VC */
	connect_vc->ch.shm_read_queue_info = accept_vc->ch.shm_read_queue_info;
	connect_vc->ch.read_shmq = accept_vc->ch.read_shmq;
	connect_vc->ch.shm_reading_pkt = TRUE;

	/* remove the accept_vc from the reading list and add the connect_vc */
	MPIDI_CH3I_SHM_Remove_vc_references(accept_vc);
	MPIDI_CH3I_SHM_Add_to_reader_list(connect_vc);
	MPIU_Free(accept_vc);

fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_COMPLETE_UNIDIRECTIONAL_CONNECTION);
	return mpi_errno;
fn_fail:
	goto fn_exit;
}

/* This routine is the "other side" of MPIDI_CH3_Complete_unidirectional_connection */
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_Complete_unidirectional_connection2
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int sshmMPIDI_CH3_Complete_unidirectional_connection_accept( MPIDI_VC_t *new_vc )
{
	int mpi_errno = MPI_SUCCESS;
	wchar_t connector_port[MPI_MAX_PORT_NAME];
	int num_read = 0;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_COMPLETE_UNIDIRECTIONAL_CONNECTION2);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_COMPLETE_UNIDIRECTIONAL_CONNECTION2);

	num_read = 0;
	while (num_read == 0)
	{
		mpi_errno = MPIDI_CH3I_SHM_read(new_vc, connector_port, MPI_MAX_PORT_NAME*sizeof(wchar_t), &num_read);
		if (mpi_errno != MPI_SUCCESS)
		{
			MPIU_ERR_POP(mpi_errno);
		}
	}
	/* --BEGIN ERROR HANDLING-- */
	if (num_read != MPI_MAX_PORT_NAME)
	{
		mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", L"**fail %d", num_read);
		goto fn_fail;
	}
	/* --END ERROR HANDLING-- */

	mpi_errno = sshmMPIDI_CH3_Connect_to_root(connector_port, &new_vc);
	if (mpi_errno != MPI_SUCCESS)
	{
		MPIU_ERR_POP(mpi_errno);
	}

	/* restore this vc to the readers list */
	MPIDI_CH3I_SHM_Add_to_reader_list(new_vc);

fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_COMPLETE_UNIDIRECTIONAL_CONNECTION2);
	return mpi_errno;
fn_fail:
	goto fn_exit;
}

/* This routine is used to free any channel-specific resources that were allocated when building a connection */
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_Cleanup_after_connection
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int sshmMPIDI_CH3_Cleanup_after_connection( MPIDI_VC_t *new_vc )
{
	MPIDI_VC_t *iter, *trailer;
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_CLEANUP_AFTER_CONNECTION);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_CLEANUP_AFTER_CONNECTION);

	/* remove the new_vc from the reading list */
	iter = trailer = MPIDI_CH3I_Process.shm_reading_list;
	while (iter != NULL)
	{
		if (iter == new_vc)
		{
			/* First free the resources associated with this VC */
			MPIDI_CH3I_SHM_Release_memblock(&new_vc->ch.shm_read_queue_info);
			
			if (trailer != iter)
			{
				/* remove the new_vc from the list */
				trailer->ch.shm_next_reader = iter->ch.shm_next_reader;
			}
			else
			{
				/* remove the new_vc from the head of the list */
				MPIDI_CH3I_Process.shm_reading_list = MPIDI_CH3I_Process.shm_reading_list->ch.shm_next_reader;
			}
		}
		if (trailer != iter)
			trailer = trailer->ch.shm_next_reader;
		iter = iter->ch.shm_next_reader;
	}
	/* remove the new_vc from the writing list */
	iter = trailer = MPIDI_CH3I_Process.shm_writing_list;
	while (iter != NULL)
	{
		if (iter == new_vc)
		{
			/* First free the resources associated with this VC */
			MPIDI_CH3I_SHM_Release_memblock(&new_vc->ch.shm_write_queue_info);
			
			if (trailer != iter)
			{
				/* remove the new_vc from the list */
				trailer->ch.shm_next_writer = iter->ch.shm_next_writer;
			}
			else
			{
				/* remove the new_vc from the head of the list */
				MPIDI_CH3I_Process.shm_writing_list = MPIDI_CH3I_Process.shm_writing_list->ch.shm_next_writer;
			}
		}
		if (trailer != iter)
			trailer = trailer->ch.shm_next_writer;
		iter = iter->ch.shm_next_writer;
	}

	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_CLEANUP_AFTER_CONNECTION);
	return mpi_errno;
}

/* FIXME: This doesn't need to be in a separate file.  And is there a 
   better way to specify which business card routines to use ? */
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_Get_business_card
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int sshmMPIDI_CH3I_Get_business_card(wchar_t *value, int length)
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_GET_BUSINESS_CARD);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_GET_BUSINESS_CARD);
	mpi_errno = MPIDI_CH3U_Get_business_card_sshm(&value, &length);
	MPIU_ERR_CHK_POP(mpi_errno);
fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_GET_BUSINESS_CARD);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_Finalize
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int sshmMPIDI_CH3_Finalize(void)
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_FINALIZE);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_FINALIZE);

	mpi_errno = MPIDI_CH3U_Finalize_sshm();
	MPIDI_CH3_Cleanup_mem();

	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_FINALIZE);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_Init
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int sshmMPIDI_CH3_Init(int has_parent, MPIDI_PG_t * pg_p, int pg_rank )
{
	int mpi_errno;
	wchar_t *publish_bc_orig = NULL;
	wchar_t *bc_val = NULL;
	int val_max_remaining;
	MPIDI_STATE_DECL(MPID_STATE_MPID_CH3_INIT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPID_CH3_INIT);

	MPIDI_CH3I_Eager_max_msg_size = 128000;

	/* Initialize the business card */
	mpi_errno = MPIDI_CH3I_BCInit( &bc_val, &val_max_remaining );
	if (mpi_errno) MPIU_ERR_POP(mpi_errno);
	publish_bc_orig = bc_val;

	mpi_errno = MPIDI_CH3U_Init_sshm(has_parent, pg_p, pg_rank, &bc_val, &val_max_remaining);
	if (mpi_errno != MPI_SUCCESS) MPIU_ERR_POP(mpi_errno);

	/* Set the connection information in our process group (publish the business card ) */
	MPIDI_PG_SetConnInfo( pg_rank, publish_bc_orig );

	/* Free the business card now that it is published
	 (note that publish_bc_orig is the head of bc_val ) */
	MPIDI_CH3I_BCFree( publish_bc_orig );

 fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPID_CH3_INIT);
	return mpi_errno;
 fn_fail:
	if (publish_bc_orig != NULL)
	{
		MPIU_Free(publish_bc_orig);
	}		   
	goto fn_exit;
}

/* This function simply tells the CH3 device to use the defaults for the 
   MPI Port functions.  This should be ok here, since we want to 
   use the socket routines to perform all connect/accept actions
   after MPID_Init returns (see the shm_unlink discussion) */
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_PortFnsInit
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int sshmMPIDI_CH3_PortFnsInit(MPIDI_PortFns *portFns) 
{
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_PORTFNSINIT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_PORTFNSINIT);
	MPIU_UNREFERENCED_ARG(portFns);
	/*
	portFns->OpenPort	= 0;
	portFns->ClosePort   = 0;
	portFns->CommAccept  = 0;
	portFns->CommConnect = 0;
	*/
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_PORTFNSINIT);
	return MPI_SUCCESS;
}

/* Perform the channel-specific vc initialization */
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_VC_Init
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int sshmMPIDI_CH3_VC_Init( MPIDI_VC_t *vc )
{
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_VC_INIT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_VC_INIT);
	vc->ch.sendq_head			= NULL;
	vc->ch.sendq_tail			= NULL;
	vc->ch.recv_active			= NULL;
	vc->ch.send_active			= NULL;
	vc->ch.req					= NULL;
	vc->ch.read_shmq			= NULL;
	vc->ch.write_shmq			= NULL;
	vc->ch.shm					= NULL;
	vc->ch.shm_state			= 0;
	vc->ch.state				= MPIDI_CH3I_VC_STATE_UNCONNECTED;
	vc->ch.shm_read_connected	= 0;
	vc->ch.shm_reading_pkt		= 0;
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_VC_INIT);
	return 0;
}

static int sshmMPIDI_CH3_PG_Init( MPIDI_PG_t *pg )
{
	MPIDI_PG_Add_ref(pg);
	return MPI_SUCCESS;
}

/* Select the routine that uses shared memory to connect two communicators using a socket */
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_Connect_to_root
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int sshmMPIDI_CH3_Connect_to_root(const wchar_t * port_name, MPIDI_VC_t ** new_vc)
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_CONNECT_TO_ROOT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_CONNECT_TO_ROOT);
	mpi_errno = MPIDI_CH3I_Connect_to_root_sshm( port_name, new_vc );
	MPIU_ERR_CHK_POP(mpi_errno);
fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_CONNECT_TO_ROOT);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME sshmMPIDI_CH3_RMAFnsInit
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int sshmMPIDI_CH3_RMAFnsInit( MPIDI_RMAFns *RMAFns ) 
{ 
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_RMAFNSINIT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_RMAFNSINIT);
	RMAFns->Win_create = MPIDI_CH3_Win_create;
	RMAFns->Win_free = MPIDI_CH3_Win_free;
	RMAFns->Put = MPIDI_CH3_Put;
	RMAFns->Get = MPIDI_CH3_Get;
	RMAFns->Accumulate = MPIDI_CH3_Accumulate;
	RMAFns->Win_fence = MPIDI_CH3_Win_fence;
	RMAFns->Win_post = MPIDI_CH3_Win_post;
	RMAFns->Win_start = MPIDI_CH3_Win_start;
	RMAFns->Win_complete = MPIDI_CH3_Win_complete;
	RMAFns->Win_wait = MPIDI_CH3_Win_wait;
	RMAFns->Win_lock = MPIDI_CH3_Win_lock;
	RMAFns->Win_unlock = MPIDI_CH3_Win_unlock;
	RMAFns->Alloc_mem = MPIDI_CH3_Alloc_mem;
	RMAFns->Free_mem = MPIDI_CH3_Free_mem;

	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_RMAFNSINIT);
	return MPI_SUCCESS;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_iSend
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int sshmMPIDI_CH3_iSend(MPIDI_VC_t * vc, MPID_Request * sreq, void * pkt, MPIDI_msg_sz_t pkt_sz)
{
	int mpi_errno = MPI_SUCCESS;
	int complete;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_ISEND);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_ISEND);

	MPIDI_DBG_PRINTF((50, FCNAME, L"entering"));
#ifdef MPICH_DBG_OUTPUT
	if (pkt_sz > sizeof(MPIDI_CH3_Pkt_t))
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**arg", 0);
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISEND);
		return mpi_errno;
	}
#endif

	/* The sock channel uses a fixed length header, the size of which is the maximum of all possible packet headers */
	pkt_sz = sizeof(MPIDI_CH3_Pkt_t);
	MPIDI_DBG_Print_packet((MPIDI_CH3_Pkt_t*)pkt);
	
	if (vc->ch.state == MPIDI_CH3I_VC_STATE_CONNECTED) /* MT */
	{
		/* Connection already formed.  If send queue is empty attempt to send data, queuing any unsent data. */
		if (MPIDI_CH3I_SendQ_empty(vc)) /* MT */
		{
			int nb;

			MPIDI_DBG_PRINTF((55, FCNAME, L"send queue empty, attempting to write"));
			
			/* MT: need some signalling to lock down our right to use the channel, thus insuring that the progress engine does
			   also try to write */
			mpi_errno = MPIDI_CH3I_SHM_write(vc, pkt, (int)pkt_sz, &nb);
			if (mpi_errno == MPI_SUCCESS)
			{
				MPIDI_DBG_PRINTF((55, FCNAME, L"wrote %d bytes", nb));

				if (nb == pkt_sz)
				{ 
					MPIDI_DBG_PRINTF((55, FCNAME, L"write complete %d bytes, calling MPIDI_CH3U_Handle_send_req()", nb));
					/* ssm version: */
					MPIDI_CH3U_Handle_send_req(vc, sreq, &complete);
					if (!complete)
					{
						sreq->ch.iov_offset = 0;
						MPIDI_CH3I_SendQ_enqueue_head(vc, sreq);
						vc->ch.send_active = sreq;
					}
					/*
					MPIDI_CH3I_SendQ_enqueue_head(vc, sreq);
					MPIDI_CH3U_Handle_send_req(vc, sreq);
					if (sreq->dev.iov_count == 0)
					{
						if (MPIDI_CH3I_SendQ_head(vc) == sreq)
						{
							MPIDI_CH3I_SendQ_dequeue(vc);
						}
					}
					*/
				}
				else
				{
					MPIDI_DBG_PRINTF((55, FCNAME, L"partial write of %d bytes, request enqueued at head", nb));
					sreq->ch.pkt = *(MPIDI_CH3_Pkt_t *) pkt;
					sreq->dev.iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)((char *) &sreq->ch.pkt + nb);
					sreq->dev.iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)(pkt_sz - nb);
					sreq->dev.iov_count = 1;
					sreq->ch.iov_offset = 0;
					MPIDI_CH3I_SendQ_enqueue_head(vc, sreq);
					vc->ch.send_active = sreq;
				}
			}
			else
			{
				MPIDI_DBG_PRINTF((55, FCNAME, L"MPIDI_CH3I_SHM_write failed"));
				/* Connection just failed. Mark the request complete and return an error. */
				vc->ch.state = MPIDI_CH3I_VC_STATE_FAILED;
				/* TODO: Create an appropriate error message based on the return value (rc) */
				sreq->status.MPI_ERROR = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ssmwrite", 0);
				 /* MT -CH3U_Request_complete() performs write barrier */
				MPIDI_CH3U_Request_complete(sreq);
			}
		}
		else
		{
			MPIDI_DBG_PRINTF((55, FCNAME, L"send queue not empty, enqueuing"));
			sreq->ch.pkt = *(MPIDI_CH3_Pkt_t *) pkt;
			sreq->dev.iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)((char *) &sreq->ch.pkt);
			sreq->dev.iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)(pkt_sz);
			sreq->dev.iov_count = 1;
			sreq->ch.iov_offset = 0;
			MPIDI_CH3I_SendQ_enqueue(vc, sreq);
		}
	}
	else if (vc->ch.state == MPIDI_CH3I_VC_STATE_UNCONNECTED) /* MT */
	{
		/* Form a new connection, queuing the data so it can be sent later. */
		MPIDI_DBG_PRINTF((55, FCNAME, L"unconnected.  enqueuing request"));
		sreq->ch.pkt = *(MPIDI_CH3_Pkt_t *) pkt;
		sreq->dev.iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)((char *) &sreq->ch.pkt);
		sreq->dev.iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)(pkt_sz);
		sreq->dev.iov_count = 1;
		sreq->ch.iov_offset = 0;
		MPIDI_CH3I_SendQ_enqueue(vc, sreq);
		mpi_errno = MPIDI_CH3I_VC_post_shmconnect(vc);
		if (mpi_errno != MPI_SUCCESS)
		{
			mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", 0);
		}
	}
	else if (vc->ch.state != MPIDI_CH3I_VC_STATE_FAILED)
	{
		/* Unable to send data at the moment, so queue it for later */
		MPIDI_DBG_PRINTF((55, FCNAME, L"still connecting.  enqueuing request"));
		sreq->ch.pkt = *(MPIDI_CH3_Pkt_t *) pkt;
		sreq->dev.iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)((char *) &sreq->ch.pkt);
		sreq->dev.iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)(pkt_sz);
		sreq->dev.iov_count = 1;
		sreq->ch.iov_offset = 0;
		MPIDI_CH3I_SendQ_enqueue(vc, sreq);
	}
	else
	{
		/* Connection failed.  Mark the request complete and return an error. */
		/* TODO: Create an appropriate error message */
		sreq->status.MPI_ERROR = MPI_ERR_INTERN;
		/* MT - CH3U_Request_complete() performs write barrier */
		MPIDI_CH3U_Request_complete(sreq);
	}
	
	MPIDI_DBG_PRINTF((50, FCNAME, L"exiting"));
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISEND);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_iSendv
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int sshmMPIDI_CH3_iSendv(MPIDI_VC_t * vc, MPID_Request * sreq, MPID_IOV * iov, int n_iov)
{
	int mpi_errno = MPI_SUCCESS;
	int complete;
	int i;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_ISENDV);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_ISENDV);
	MPIDI_DBG_PRINTF((50, FCNAME, L"entering"));
#ifdef MPICH_DBG_OUTPUT
	if (n_iov > MPID_IOV_LIMIT)
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**arg", 0);
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISENDV);
		return mpi_errno;
	}
	if (iov[0].MPID_IOV_LEN > (MPID_IOV_LEN_CAST)sizeof(MPIDI_CH3_Pkt_t))
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**arg", 0);
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISENDV);
		return mpi_errno;
	}
#endif

	/* The mm channel uses a fixed length header, the size of which is the maximum of all possible packet headers */
	iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)sizeof(MPIDI_CH3_Pkt_t);
	MPIDI_DBG_Print_packet((MPIDI_CH3_Pkt_t*)iov[0].MPID_IOV_BUF);

	if (vc->ch.state == MPIDI_CH3I_VC_STATE_CONNECTED) /* MT */
	{
		/* Connection already formed.  If send queue is empty attempt to send data, queuing any unsent data. */
		if (MPIDI_CH3I_SendQ_empty(vc)) /* MT */
		{
			int nb;

			MPIDI_DBG_PRINTF((55, FCNAME, L"send queue empty, attempting to write"));
			
			/* MT - need some signalling to lock down our right to use the channel, thus insuring that the progress engine does also try to write */

			/* FIXME: the current code only agressively writes the first IOV.  Eventually it should be changed to agressively write
			   as much as possible.  Ideally, the code would be shared between the send routines and the progress engine. */
			mpi_errno = MPIDI_CH3I_SHM_writev(vc, iov, n_iov, &nb);
			if (mpi_errno == MPI_SUCCESS)
			{
				int offset = 0;

				MPIDI_DBG_PRINTF((55, FCNAME, L"wrote %d bytes", nb));
				
				while (offset < n_iov)
				{
					if ((int)iov[offset].MPID_IOV_LEN <= nb)
					{
						nb -= (int)iov[offset].MPID_IOV_LEN;
						offset++;
					}
					else
					{
						MPIDI_DBG_PRINTF((55, FCNAME, L"partial write, request enqueued at head"));
						for (i = 0; i < n_iov; i++)
						{
							sreq->dev.iov[i] = iov[i];
						}
						if (offset == 0)
						{
							sreq->ch.pkt = *(MPIDI_CH3_Pkt_t *) iov[0].MPID_IOV_BUF;
							sreq->dev.iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST) &sreq->ch.pkt;
						}
						sreq->dev.iov[offset].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)((char *) sreq->dev.iov[offset].MPID_IOV_BUF + nb);
						sreq->dev.iov[offset].MPID_IOV_LEN -= (MPID_IOV_LEN_CAST)nb;
						sreq->ch.iov_offset = offset;
						sreq->dev.iov_count = n_iov;
						MPIDI_CH3I_SendQ_enqueue_head(vc, sreq);
						vc->ch.send_active = sreq;
						break;
					}
				}
				if (offset == n_iov)
				{
					MPIDI_DBG_PRINTF((55, FCNAME, L"write complete, calling MPIDI_CH3U_Handle_send_req()"));
					/* ssm version: */
					MPIDI_CH3U_Handle_send_req(vc, sreq, &complete);
					if (!complete)
					{
						sreq->ch.iov_offset = 0;
						MPIDI_CH3I_SendQ_enqueue_head(vc, sreq);
						vc->ch.send_active = sreq;
					}
					/*
					MPIDI_CH3I_SendQ_enqueue_head(vc, sreq);
					MPIDI_CH3U_Handle_send_req(vc, sreq);
					if (sreq->dev.iov_count == 0)
					{
						if (MPIDI_CH3I_SendQ_head(vc) == sreq)
						{
							MPIDI_CH3I_SendQ_dequeue(vc);
						}
					}
					*/
				}
			}
			else
			{
				/* Connection just failed.  Mark the request complete and return an error. */
				vc->ch.state = MPIDI_CH3I_VC_STATE_FAILED;
				/* TODO: Create an appropriate error message based on the value of errno */
				sreq->status.MPI_ERROR = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ssmwritev", 0);
				/* MT - CH3U_Request_complete performs write barrier */
				MPIDI_CH3U_Request_complete(sreq);
			}
		}
		else
		{
			MPIDI_DBG_PRINTF((55, FCNAME, L"send queue not empty, enqueuing"));
			for (i = 0; i < n_iov; i++)
			{
				sreq->dev.iov[i] = iov[i];
			}
			sreq->ch.pkt = *(MPIDI_CH3_Pkt_t *) iov[0].MPID_IOV_BUF;
			sreq->dev.iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)((char *) sreq->dev.iov[0].MPID_IOV_BUF);
			sreq->ch.iov_offset = 0;
			sreq->dev.iov_count = n_iov;
			MPIDI_CH3I_SendQ_enqueue(vc, sreq);
		}
	}
	else if (vc->ch.state == MPIDI_CH3I_VC_STATE_UNCONNECTED)
	{
		/* Form a new connection, queuing the data so it can be sent later. */
		MPIDI_DBG_PRINTF((55, FCNAME, L"unconnected.  enqueuing request"));
		/*MPIDI_CH3I_VC_post_shmconnect(vc);*/
		for (i = 0; i < n_iov; i++)
		{
			sreq->dev.iov[i] = iov[i];
		}
		sreq->ch.pkt = *(MPIDI_CH3_Pkt_t *) iov[0].MPID_IOV_BUF;
		sreq->dev.iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)((char *) sreq->dev.iov[0].MPID_IOV_BUF);
		sreq->ch.iov_offset = 0;
		sreq->dev.iov_count = n_iov;
		MPIDI_CH3I_SendQ_enqueue(vc, sreq);
		mpi_errno = MPIDI_CH3I_VC_post_shmconnect(vc);
		if (mpi_errno != MPI_SUCCESS)
		{
			mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", 0);
		}
	}
	else if (vc->ch.state != MPIDI_CH3I_VC_STATE_FAILED)
	{
		/* Unable to send data at the moment, so queue it for later */
		MPIDI_DBG_PRINTF((55, FCNAME, L"still connecting.  enqueuing request"));
		for (i = 0; i < n_iov; i++)
		{
			sreq->dev.iov[i] = iov[i];
		}
		sreq->ch.pkt = *(MPIDI_CH3_Pkt_t *) iov[0].MPID_IOV_BUF;
		sreq->dev.iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)((char *) sreq->dev.iov[0].MPID_IOV_BUF);
		sreq->ch.iov_offset = 0;
		sreq->dev.iov_count = n_iov;
		MPIDI_CH3I_SendQ_enqueue(vc, sreq);
	}
	else
	{
		/* Connection failed.  Mark the request complete and return an error. */
		/* TODO: Create an appropriate error message */
		sreq->status.MPI_ERROR = MPI_ERR_INTERN;
		/* MT - CH3U_Request_complete performs write barrier */
		MPIDI_CH3U_Request_complete(sreq);
	}
	
	MPIDI_DBG_PRINTF((50, FCNAME, L"exiting"));
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISENDV);
	return mpi_errno;
}

/*
 * sshmMPIDI_CH3_iStartMsg() attempts to send the message immediately.  If the entire message is successfully sent, then NULL is
 * returned.  Otherwise a request is allocated, the header is copied into the request, and a pointer to the request is returned.
 * An error condition also results in a request be allocated and the errror being returned in the status field of the request.
 */
#undef FUNCNAME
#define FUNCNAME sshmMPIDI_CH3_iStartMsg
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int sshmMPIDI_CH3_iStartMsg(MPIDI_VC_t * vc, void * pkt, MPIDI_msg_sz_t pkt_sz, MPID_Request **sreq_ptr)
{
	int mpi_errno = MPI_SUCCESS;
	MPID_Request * sreq = NULL;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_ISTARTMSG);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_ISTARTMSG);
	
	MPIDI_DBG_PRINTF((50, FCNAME, L"entering"));
#ifdef MPICH_DBG_OUTPUT
	if (pkt_sz > sizeof(MPIDI_CH3_Pkt_t))
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**arg", 0);
		goto fn_exit;
	}
#endif

	/* The MM channel uses a fixed length header, the size of which is the maximum of all possible packet headers */
	pkt_sz = sizeof(MPIDI_CH3_Pkt_t);
	MPIDI_DBG_Print_packet((MPIDI_CH3_Pkt_t*)pkt);
	
	if (vc->ch.state == MPIDI_CH3I_VC_STATE_CONNECTED) /* MT */
	{
		/* Connection already formed.  If send queue is empty attempt to send data, queuing any unsent data. */
		if (MPIDI_CH3I_SendQ_empty(vc)) /* MT */
		{
			int nb;

			MPIDI_DBG_PRINTF((55, FCNAME, L"send queue empty, attempting to write"));

			/* MT - need some signalling to lock down our right to use the channel, thus insuring that the progress engine does not also try to write */

			mpi_errno = MPIDI_CH3I_SHM_write(vc, pkt, (int)pkt_sz, &nb);
			if (mpi_errno == MPI_SUCCESS)
			{
				MPIDI_DBG_PRINTF((55, FCNAME, L"wrote %d bytes", nb));

				if (nb == pkt_sz)
				{ 
					MPIDI_DBG_PRINTF((55, FCNAME, L"entire write complete, %d bytes", nb));
					/* done.  get us out of here as quickly as possible. */
				}
				else
				{
					MPIDI_DBG_PRINTF((55, FCNAME, L"partial write of %d bytes, request enqueued at head", nb));
					sreq = MPID_Request_create();
					if (sreq == NULL)
					{
						mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
						MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISTARTMSG);
						return mpi_errno;
					}
					MPIU_Object_set_ref(sreq, 2);
					sreq->kind = MPID_REQUEST_SEND;
					sreq->ch.pkt = *(MPIDI_CH3_Pkt_t *) pkt;
					sreq->dev.iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)((char *) &sreq->ch.pkt + nb);
					sreq->dev.iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)(pkt_sz - nb);
					sreq->dev.iov_count = 1;
					sreq->ch.iov_offset = 0;
					sreq->dev.ca = MPIDI_CH3_CA_COMPLETE;
					MPIDI_CH3I_SendQ_enqueue_head(vc, sreq);
					vc->ch.send_active = sreq;
				}
			}
			else
			{
				sreq = MPID_Request_create();
				if (sreq == NULL)
				{
					mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
					goto fn_exit;
				}
				sreq->kind = MPID_REQUEST_SEND;
				sreq->cc = 0;
				/* TODO: Create an appropriate error message based on the return value */
				sreq->status.MPI_ERROR = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ssmwrite", 0);
			}
		}
		else
		{
			MPIDI_DBG_PRINTF((55, FCNAME, L"send in progress, request enqueued"));
			sreq = MPID_Request_create();
			if (sreq == NULL)
			{
				mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
				MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISTARTMSG);
				return mpi_errno;
			}
			MPIU_Object_set_ref(sreq, 2);
			sreq->kind = MPID_REQUEST_SEND;
			sreq->ch.pkt = *(MPIDI_CH3_Pkt_t *) pkt;
			sreq->dev.iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)((char *) &sreq->ch.pkt);
			sreq->dev.iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)(pkt_sz);
			sreq->dev.iov_count = 1;
			sreq->ch.iov_offset = 0;
			sreq->dev.ca = MPIDI_CH3_CA_COMPLETE;
			MPIDI_CH3I_SendQ_enqueue(vc, sreq);
		}
	}
	else if (vc->ch.state == MPIDI_CH3I_VC_STATE_UNCONNECTED) /* MT */
	{
		MPIDI_DBG_PRINTF((55, FCNAME, L"unconnected.  posting connect and enqueuing request"));

		/* queue the data so it can be sent after the connection is formed */
		sreq = MPID_Request_create();
		if (sreq == NULL)
		{
			mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
			MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISTARTMSG);
			return mpi_errno;
		}
		MPIU_Object_set_ref(sreq, 2);
		sreq->kind = MPID_REQUEST_SEND;
		sreq->ch.pkt = *(MPIDI_CH3_Pkt_t *) pkt;
		sreq->dev.iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)((char *) &sreq->ch.pkt);
		sreq->dev.iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)(pkt_sz);
		sreq->dev.iov_count = 1;
		sreq->ch.iov_offset = 0;
		sreq->dev.ca = MPIDI_CH3_CA_COMPLETE;
		MPIDI_CH3I_SendQ_enqueue(vc, sreq);

		MPIDI_CH3I_VC_post_shmconnect(vc);
	}
	else if (vc->ch.state != MPIDI_CH3I_VC_STATE_FAILED)
	{
		/* Unable to send data at the moment, so queue it for later */
		MPIDI_DBG_PRINTF((55, FCNAME, L"forming connection, request enqueued"));
		sreq = MPID_Request_create();
		if (sreq == NULL)
		{
			mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
			MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISTARTMSG);
			return mpi_errno;
		}
		MPIU_Object_set_ref(sreq, 2);
		sreq->kind = MPID_REQUEST_SEND;
		sreq->ch.pkt = *(MPIDI_CH3_Pkt_t *) pkt;
		sreq->dev.iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)((char *) &sreq->ch.pkt);
		sreq->dev.iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)(pkt_sz);
		sreq->dev.iov_count = 1;
		sreq->ch.iov_offset = 0;
		sreq->dev.ca = MPIDI_CH3_CA_COMPLETE;
		MPIDI_CH3I_SendQ_enqueue(vc, sreq);
	}
	else
	{
		/* Connection failed, so allocate a request and return an error. */
		MPIDI_DBG_PRINTF((55, FCNAME, L"ERROR - connection failed"));
		sreq = MPID_Request_create();
		if (sreq == NULL)
		{
			mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
			goto fn_exit;
		}
		sreq->kind = MPID_REQUEST_SEND;
		sreq->cc = 0;
		/* TODO: Create an appropriate error message */
		sreq->status.MPI_ERROR = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_INTERN, L"**connfailed", 0);
	}

fn_exit:
	*sreq_ptr = sreq;

	MPIDI_DBG_PRINTF((50, FCNAME, L"exiting"));
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISTARTMSG);
	return mpi_errno;
}

static MPID_Request * create_request(MPID_IOV * iov, int count, int offset, int nb)
{
	MPID_Request * sreq;
	/*int i;*/
	MPIDI_STATE_DECL(MPID_STATE_CREATE_REQUEST);
	MPIDI_STATE_DECL(MPID_STATE_MEMCPY);
	MPIDI_FUNC_ENTER(MPID_STATE_CREATE_REQUEST);
	sreq = MPID_Request_create();
	if (sreq == NULL)
	{
		MPIDI_FUNC_EXIT(MPID_STATE_CREATE_REQUEST);
		return NULL;
	}
	MPIU_Object_set_ref(sreq, 2);
	sreq->kind = MPID_REQUEST_SEND;
	MPIDI_FUNC_ENTER(MPID_STATE_MEMCPY);
	memcpy(sreq->dev.iov, iov, count * sizeof(MPID_IOV));
	MPIDI_FUNC_EXIT(MPID_STATE_MEMCPY);
	/*for (i = 0; i < count; i++) { sreq->dev.iov[i] = iov[i]; }*/
	if (offset == 0)
	{
		/* memcpy(&sreq->ch.pkt, iov[0].MPID_IOV_BUF, iov[0].MPID_IOV_LEN); */
		/*MPIU_Assert(iov[0].MPID_IOV_LEN == sizeof(MPIDI_CH3_Pkt_t));*/
		sreq->ch.pkt = *(MPIDI_CH3_Pkt_t *) iov[0].MPID_IOV_BUF;
		sreq->dev.iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST) &sreq->ch.pkt;
	}
	sreq->dev.iov[offset].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)((char *) sreq->dev.iov[offset].MPID_IOV_BUF + nb);
	sreq->dev.iov[offset].MPID_IOV_LEN -= (MPID_IOV_LEN_CAST)nb;
	sreq->ch.iov_offset = offset;
	sreq->dev.iov_count = count;
	sreq->dev.ca = MPIDI_CH3_CA_COMPLETE;
	MPIDI_FUNC_EXIT(MPID_STATE_CREATE_REQUEST);
	return sreq;
}

/*
 * sshmMPIDI_CH3_iStartMsgv() attempts to send the message immediately.  If the entire message is successfully sent, then NULL is
 * returned.  Otherwise a request is allocated, the iovec and the first buffer pointed to by the iovec (which is assumed to be a
 * MPIDI_CH3_Pkt_t) are copied into the request, and a pointer to the request is returned.  An error condition also results in a
 * request be allocated and the errror being returned in the status field of the request.
 */

/* XXX - What do we do if MPID_Request_create() returns NULL???  If sshmMPIDI_CH3_iStartMsgv() returns NULL, the calling code
   assumes the request completely successfully, but the reality is that we couldn't allocate the memory for a request.  This
   seems like a flaw in the CH3 API. */

/* NOTE - The completion action associated with a request created by CH3_iStartMsgv() is alway MPIDI_CH3_CA_COMPLETE.  This
   implies that CH3_iStartMsgv() can only be used when the entire message can be described by a single iovec of size
   MPID_IOV_LIMIT. */
	
#undef FUNCNAME
#define FUNCNAME sshmMPIDI_CH3_iStartMsgv
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int sshmMPIDI_CH3_iStartMsgv(MPIDI_VC_t * vc, MPID_IOV * iov, int n_iov, MPID_Request **sreq_ptr)
{
	int mpi_errno = MPI_SUCCESS;
	MPID_Request * sreq = NULL;
	MPIDI_STATE_DECL(MPID_STATE_MEMCPY);
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_ISTARTMSGV);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_ISTARTMSGV);

	MPIDI_DBG_PRINTF((50, FCNAME, L"entering"));
#ifdef MPICH_DBG_OUTPUT
	if (n_iov > MPID_IOV_LIMIT)
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**arg", 0);
		goto fn_exit;
	}
	if (iov[0].MPID_IOV_LEN > sizeof(MPIDI_CH3_Pkt_t))
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**arg", 0);
		goto fn_exit;
	}
#endif

	/* The SOCK channel uses a fixed length header, the size of which is the maximum of all possible packet headers */
	iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)sizeof(MPIDI_CH3_Pkt_t);
	MPIDI_DBG_Print_packet((MPIDI_CH3_Pkt_t*)iov[0].MPID_IOV_BUF);

	if (vc->ch.state == MPIDI_CH3I_VC_STATE_CONNECTED) /* MT */
	{
		/* Connection already formed.  If send queue is empty attempt to send data, queuing any unsent data. */
		if (MPIDI_CH3I_SendQ_empty(vc)) /* MT */
		{
			int nb;

			MPIDI_DBG_PRINTF((55, FCNAME, L"send queue empty, attempting to write"));

			/* MT - need some signalling to lock down our right to use the channel, thus insuring that the progress engine does also try to write */
			mpi_errno = MPIDI_CH3I_SHM_writev(vc, iov, n_iov, &nb);
			if (mpi_errno == MPI_SUCCESS)
			{
				int offset = 0;
	
				MPIDI_DBG_PRINTF((55, FCNAME, L"wrote %d bytes", nb));
				
				while (offset < n_iov)
				{
					if (nb >= (int)iov[offset].MPID_IOV_LEN)
					{
						nb -= (int)iov[offset].MPID_IOV_LEN;
						offset++;
					}
					else
					{
						MPIDI_DBG_PRINTF((55, FCNAME, L"partial write, request enqueued at head"));
						sreq = create_request(iov, n_iov, offset, nb);
						if (sreq == NULL)
						{
							mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
							goto fn_exit;
						}
						MPIDI_CH3I_SendQ_enqueue_head(vc, sreq);
						vc->ch.send_active = sreq;
						break;
					}
				}

				if (offset == n_iov)
				{
					MPIDI_DBG_PRINTF((55, FCNAME, L"entire write complete"));
				}
			}
			else
			{
				MPIDI_DBG_PRINTF((55, FCNAME, L"ERROR - MPIDI_CH3I_SHM_writev failed"));
				sreq = MPID_Request_create();
				if (sreq == NULL)
				{
					mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
					goto fn_exit;
				}
				sreq->kind = MPID_REQUEST_SEND;
				sreq->cc = 0;
				/* TODO: Create an appropriate error message based on the return value */
				sreq->status.MPI_ERROR = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ssmwrite", 0);
			}
		}
		else
		{
			MPIDI_DBG_PRINTF((55, FCNAME, L"send in progress, request enqueued"));
			sreq = MPID_Request_create();
			if (sreq == NULL)
			{
				mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
				goto fn_exit;
			}
			MPIU_Object_set_ref(sreq, 2);
			sreq->kind = MPID_REQUEST_SEND;
			MPIDI_FUNC_ENTER(MPID_STATE_MEMCPY);
			memcpy(sreq->dev.iov, iov, n_iov * sizeof(MPID_IOV));
			MPIDI_FUNC_EXIT(MPID_STATE_MEMCPY);
			sreq->ch.pkt = *(MPIDI_CH3_Pkt_t *) iov[0].MPID_IOV_BUF;
			sreq->dev.iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)((char *) sreq->dev.iov[0].MPID_IOV_BUF);
			sreq->ch.iov_offset = 0;
			sreq->dev.iov_count = n_iov;
			sreq->dev.ca = MPIDI_CH3_CA_COMPLETE;
			MPIDI_CH3I_SendQ_enqueue(vc, sreq);
		}
	}
	else if (vc->ch.state == MPIDI_CH3I_VC_STATE_UNCONNECTED)
	{
		MPIDI_DBG_PRINTF((55, FCNAME, L"unconnected.  posting connect and enqueuing request"));
		
		/* Form a new connection */
		/*MPIDI_CH3I_VC_post_shmconnect(vc);*/
		
		/* queue the data so it can be sent after the connection is formed */
		sreq = MPID_Request_create();
		if (sreq == NULL)
		{
			mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
			goto fn_exit;
		}
		MPIU_Object_set_ref(sreq, 2);
		sreq->kind = MPID_REQUEST_SEND;
		MPIDI_FUNC_ENTER(MPID_STATE_MEMCPY);
		memcpy(sreq->dev.iov, iov, n_iov * sizeof(MPID_IOV));
		MPIDI_FUNC_EXIT(MPID_STATE_MEMCPY);
		sreq->ch.pkt = *(MPIDI_CH3_Pkt_t *) iov[0].MPID_IOV_BUF;
		sreq->dev.iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)((char *) sreq->dev.iov[0].MPID_IOV_BUF);
		sreq->ch.iov_offset = 0;
		sreq->dev.iov_count = n_iov;
		sreq->dev.ca = MPIDI_CH3_CA_COMPLETE;
		MPIDI_CH3I_SendQ_enqueue(vc, sreq);

		MPIDI_CH3I_VC_post_shmconnect(vc);
	}
	else if (vc->ch.state != MPIDI_CH3I_VC_STATE_FAILED)
	{
		/* Unable to send data at the moment, so queue it for later */
		MPIDI_DBG_PRINTF((55, FCNAME, L"forming connection, request enqueued"));
		sreq = MPID_Request_create();
		if (sreq == NULL)
		{
			mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
			goto fn_exit;
		}
		MPIU_Object_set_ref(sreq, 2);
		sreq->kind = MPID_REQUEST_SEND;
		MPIDI_FUNC_ENTER(MPID_STATE_MEMCPY);
		memcpy(sreq->dev.iov, iov, n_iov * sizeof(MPID_IOV));
		MPIDI_FUNC_EXIT(MPID_STATE_MEMCPY);
		sreq->ch.pkt = *(MPIDI_CH3_Pkt_t *) iov[0].MPID_IOV_BUF;
		sreq->dev.iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)((char *) sreq->dev.iov[0].MPID_IOV_BUF);
		sreq->ch.iov_offset = 0;
		sreq->dev.iov_count = n_iov;
		sreq->dev.ca = MPIDI_CH3_CA_COMPLETE;
		MPIDI_CH3I_SendQ_enqueue(vc, sreq);
	}
	else
	{
		/* Connection failed, so allocate a request and return an error. */
		MPIDI_DBG_PRINTF((55, FCNAME, L"ERROR - connection failed"));
		sreq = MPID_Request_create();
		if (sreq == NULL)
		{
			mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
			goto fn_exit;
		}
		sreq->kind = MPID_REQUEST_SEND;
		sreq->cc = 0;
		/* TODO: Create an appropriate error message */
		sreq->status.MPI_ERROR = MPI_ERR_INTERN;
	}

fn_exit:
	*sreq_ptr = sreq;

	MPIDI_DBG_PRINTF((50, FCNAME, L"exiting"));
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISTARTMSGV);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_iStartRndvMsg
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int sshmMPIDI_CH3_iStartRndvMsg(MPIDI_VC_t * vc, MPID_Request * sreq, MPIDI_CH3_Pkt_t * rts_pkt)
{
	int mpi_errno = MPI_SUCCESS;
	MPID_Request * rts_sreq;
#ifdef MPICH_DBG_OUTPUT
	int i;
#endif
#ifdef USE_SHM_RDMA_GET
	MPIDI_CH3_Pkt_t pkt;
#endif
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_ISTARTRNDVMSG);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_ISTARTRNDVMSG);

#ifdef USE_SHM_RDMA_GET
	MPIDI_Pkt_init(&pkt.rts_iov, MPIDI_CH3_PKT_RTS_IOV);
	pkt.rts_iov.sreq = sreq->handle;
	pkt.rts_iov.iov_len = sreq->dev.iov_count;

	sreq->dev.rdma_iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)&pkt;
	sreq->dev.rdma_iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)sizeof(MPIDI_CH3_Pkt_t);
	sreq->dev.rdma_iov[1].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)sreq->dev.iov;
	sreq->dev.rdma_iov[1].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)(sreq->dev.iov_count * sizeof(MPID_IOV));
	/* Save a copy of the rts packet in case it is not completely written by the iStartMsgv call */
	/*sreq->dev.rdma_iov[2].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)rts_pkt;*/
	sreq->ch.pkt = *rts_pkt;
	sreq->dev.rdma_iov[2].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)&sreq->ch.pkt;
	sreq->dev.rdma_iov[2].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)sizeof(MPIDI_CH3_Pkt_t);

#ifdef MPICH_DBG_OUTPUT
	for (i=0; i<sreq->dev.iov_count; i++)
	{
		MPIU_DBG_PRINTF((L"iStartRndvMsg: send buf[%d] = %p, len = %d\n", i, sreq->dev.iov[i].MPID_IOV_BUF, sreq->dev.iov[i].MPID_IOV_LEN));
	}
#endif
	mpi_errno = sshmMPIDI_CH3_iStartMsgv(vc, sreq->dev.rdma_iov, 3, &rts_sreq);
	/* --BEGIN ERROR HANDLING-- */
	if (mpi_errno != MPI_SUCCESS)
	{
		MPIU_Object_set_ref(sreq, 0);
		MPIDI_CH3_Request_destroy(sreq);
		sreq = NULL;
		mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|rtspkt", 0);
		goto fn_exit;
	}
	/* --END ERROR HANDLING-- */
	if (rts_sreq != NULL)
	{
		/* The sender doesn't need to know when the message has been sent.  So release the request immediately */
		MPID_Request_release(rts_sreq);
	}
#else
#ifdef MPICH_DBG_OUTPUT
	for (i=0; i<sreq->dev.iov_count; i++)
	{
		MPIU_DBG_PRINTF((L"iStartRndvMsg: send buf[%d] = %p, len = %d\n", i, sreq->dev.iov[i].MPID_IOV_BUF, sreq->dev.iov[i].MPID_IOV_LEN));
	}
#endif
	mpi_errno = sshmMPIDI_CH3_iStartMsg(vc, rts_pkt, sizeof(*rts_pkt), &rts_sreq);
	/* --BEGIN ERROR HANDLING-- */
	if (mpi_errno != MPI_SUCCESS)
	{
		MPIU_Object_set_ref(sreq, 0);
		MPIDI_CH3_Request_destroy(sreq);
		sreq = NULL;
		mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|rtspkt", 0);
		goto fn_exit;
	}
	/* --END ERROR HANDLING-- */
	if (rts_sreq != NULL)
	{
		/* The sender doesn't need to know when the packet has been sent.  So release the request immediately */
		MPID_Request_release(rts_sreq);
	}
#endif

fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISTARTRNDVMSG);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME sshmMPIDI_CH3_iStartRndvTransfer
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int sshmMPIDI_CH3_iStartRndvTransfer(MPIDI_VC_t * vc, MPID_Request * rreq)
{
	int mpi_errno = MPI_SUCCESS;
#ifndef USE_SHM_RDMA_GET
	MPIDI_CH3_Pkt_t pkt;
	MPID_Request *request_ptr;
#endif
#ifdef MPICH_DBG_OUTPUT
	int i;
#endif
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_ISTARTRNDVTRANSFER);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_ISTARTRNDVTRANSFER);

#ifdef USE_SHM_RDMA_GET

#ifdef MPICH_DBG_OUTPUT
	for (i=0; i<rreq->dev.iov_count; i++)
	{
		MPIU_DBG_PRINTF((L"iStartRndvTransfer: recv buf[%d] = %p, len = %d\n", i, rreq->dev.iov[i].MPID_IOV_BUF, rreq->dev.iov[i].MPID_IOV_LEN));
		MPIU_DBG_PRINTF((L"iStartRndvTransfer: send buf[%d] = %p, len = %d\n", i, rreq->dev.rdma_iov[i].MPID_IOV_BUF, rreq->dev.rdma_iov[i].MPID_IOV_LEN));
	}
#endif
	mpi_errno = MPIDI_CH3I_SHM_rdma_readv(vc, rreq);
	/* --BEGIN ERROR HANDLING-- */
	if (mpi_errno != MPI_SUCCESS)
	{
		mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", 0);
		goto fn_exit;
	}
	/* --END ERROR HANDLING-- */

#else

	MPIDI_Pkt_init(&pkt.cts_iov, MPIDI_CH3_PKT_CTS_IOV);
	pkt.cts_iov.sreq = rreq->dev.sender_req_id;
	pkt.cts_iov.rreq = rreq->handle;
	pkt.cts_iov.iov_len = rreq->dev.iov_count;

	rreq->dev.rdma_iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)&pkt;
	rreq->dev.rdma_iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)sizeof(MPIDI_CH3_Pkt_t);
	rreq->dev.rdma_iov[1].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)rreq->dev.iov;
	rreq->dev.rdma_iov[1].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)(rreq->dev.iov_count * sizeof(MPID_IOV));

#ifdef MPICH_DBG_OUTPUT
	for (i=0; i<rreq->dev.iov_count; i++)
	{
		MPIU_DBG_PRINTF((L"iStartRndvTransfer: recv buf[%d] = %p, len = %d\n", i, rreq->dev.iov[i].MPID_IOV_BUF, rreq->dev.iov[i].MPID_IOV_LEN));
	}
#endif
	mpi_errno = sshmMPIDI_CH3_iStartMsgv(vc, rreq->dev.rdma_iov, 2, &request_ptr);
	/* --BEGIN ERROR HANDLING-- */
	if (mpi_errno != MPI_SUCCESS)
	{
		MPIU_Object_set_ref(rreq, 0);
		MPIDI_CH3_Request_destroy(rreq);
		rreq = NULL;
		mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|ctspkt", 0);
		goto fn_exit;
	}
	/* --END ERROR HANDLING-- */
	if (request_ptr != NULL)
	{
		/* The sender doesn't need to know when the message has been sent.  So release the request immediately */
		MPID_Request_release(request_ptr);
	}

#endif

fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISTARTRNDVTRANSFER);
	return mpi_errno;
}

/*
 * MPIDI_CH3_alloc_mem()
 */
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_Alloc_mem
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
void *MPIDI_CH3_Alloc_mem(size_t size, MPID_Info *info_ptr)
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_CH3I_Shmem_block_request_result *shm_struct;
	void *ap=NULL;
	MPIDI_CH3I_Alloc_mem_list_t *new_ptr;
	
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_ALLOC_MEM);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_ALLOC_MEM);

	MPIU_UNREFERENCED_ARG(info_ptr);

	shm_struct = (MPIDI_CH3I_Shmem_block_request_result *) MPIU_Malloc(sizeof(MPIDI_CH3I_Shmem_block_request_result));
	/* --BEGIN ERROR HANDLING-- */
	if (shm_struct == NULL)
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
		goto fn_exit;
	}
	/* --END ERROR HANDLING-- */

	mpi_errno = MPIDI_CH3I_SHM_Get_memblock((int)size, shm_struct);

	if (mpi_errno == MPI_SUCCESS)
	{
		/* got shared memory. keep track of it by adding it to the list. */

		new_ptr = (MPIDI_CH3I_Alloc_mem_list_t *) MPIU_Malloc(sizeof(MPIDI_CH3I_Alloc_mem_list_t));
		/* --BEGIN ERROR HANDLING-- */
		if (new_ptr == NULL)
		{
			mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
			goto fn_exit;
		}
		/* --END ERROR HANDLING-- */

		new_ptr->shm_struct = shm_struct;

		new_ptr->next = MPIDI_CH3I_Alloc_mem_list_head;
		MPIDI_CH3I_Alloc_mem_list_head = new_ptr;

		ap = shm_struct->addr;
	}
	else
	{
		/* no shared memory. do a regular malloc and return. */
		ap = MPIU_Malloc(size);

		MPIU_Free(shm_struct);
	}

 fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ALLOC_MEM);
	return ap;
}


/*
 * MPIDI_CH3_Free_mem()
 */
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_Free_mem
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3_Free_mem(void *addr)
{
	int mpi_errno = MPI_SUCCESS, found;
	MPIDI_CH3I_Alloc_mem_list_t *prev_ptr, *curr_ptr;

	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_FREE_MEM);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_FREE_MEM);

	/* check whether it was allocated as shared memory */

	prev_ptr = curr_ptr = MPIDI_CH3I_Alloc_mem_list_head;
	found = 0;
	while (curr_ptr != NULL)
	{
		if (addr == curr_ptr->shm_struct->addr)
		{
			found = 1;

			/* deallocate shared memory */
			mpi_errno = MPIDI_CH3I_SHM_Unlink_and_detach_mem(curr_ptr->shm_struct);
			if (mpi_errno != MPI_SUCCESS) goto fn_exit;
			
			/* remove this entry from the list */

			if (curr_ptr == MPIDI_CH3I_Alloc_mem_list_head)
				MPIDI_CH3I_Alloc_mem_list_head = curr_ptr->next;
			else 
				prev_ptr->next = curr_ptr->next;

			MPIU_Free(curr_ptr->shm_struct);
			MPIU_Free(curr_ptr);

			break;
		}
		prev_ptr = curr_ptr;
		curr_ptr = curr_ptr->next;
	}

	if (!found) /* it wasn't shared memory. call MPIU_Free. */
		MPIU_Free(addr);

 fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_FREE_MEM);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_Cleanup_mem
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
void MPIDI_CH3_Cleanup_mem()
{
	MPIDI_CH3I_Alloc_mem_list_t *curr_ptr, *next_ptr;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_CLEANUP_MEM);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_CLEANUP_MEM);
	curr_ptr = MPIDI_CH3I_Alloc_mem_list_head;
	while (curr_ptr != NULL)
	{
		/* deallocate shared memory */
		MPIDI_CH3I_SHM_Unlink_and_detach_mem(curr_ptr->shm_struct);  /* brad : func is sshm only */
			
		next_ptr = curr_ptr->next;

		MPIU_Free(curr_ptr->shm_struct);
		MPIU_Free(curr_ptr);

		curr_ptr = next_ptr;
	}
	MPIDI_CH3I_Alloc_mem_list_head = NULL;	
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_CLEANUP_MEM);
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_Get_mem_list_head
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
MPIDI_CH3I_Alloc_mem_list_t * MPIDI_CH3I_Get_mem_list_head()
{
	MPIDI_CH3I_Alloc_mem_list_t * ret_val = NULL;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_GET_MEM_LIST_HEAD);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_GET_MEM_LIST_HEAD);
	ret_val = MPIDI_CH3I_Alloc_mem_list_head;
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_GET_MEM_LIST_HEAD);
	return ret_val;
}

#undef FUNCNAME
#define FUNCNAME sshmMPIDI_CH3_Progress_test
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int sshmMPIDI_CH3_Progress_test(void)
{
	return sshmMPIDI_CH3I_Progress(FALSE, NULL);
}

#undef FUNCNAME
#define FUNCNAME sshmMPIDI_CH3_Progress_wait
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int sshmMPIDI_CH3_Progress_wait(MPID_Progress_state *state)
{
	return sshmMPIDI_CH3I_Progress(TRUE, state);
}

#undef FUNCNAME
#define FUNCNAME sshmMPIDI_CH3I_Progress
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int sshmMPIDI_CH3I_Progress(int is_blocking, MPID_Progress_state *state)
{
	int mpi_errno = MPI_SUCCESS;
	int rc;
#ifdef MPICH_DBG_OUTPUT
	int count;
#endif
	int bShmProgressMade;
	unsigned completions = MPIDI_CH3I_progress_completion_count;
	MPIDI_CH3I_Shmem_queue_info info;
	int num_bytes;
	MPIDI_VC_t *vc_ptr;
	static int msg_queue_count = 0;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_PROGRESS);
	MPIDI_STATE_DECL(MPID_STATE_MPIDU_YIELD);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_PROGRESS);

	MPIU_UNREFERENCED_ARG(state);

#ifdef MPICH_DBG_OUTPUT
	if (is_blocking)
	{
		MPIDI_DBG_PRINTF((50, FCNAME, L"entering, blocking=%s", is_blocking ? L"true" : L"false"));
	}
#endif
	do
	{
		/* make progress on the shared memory queues */

		bShmProgressMade = FALSE;
		if (MPIDI_CH3I_Process.shm_reading_list)
		{
			rc = MPIDI_CH3I_SHM_read_progress(MPIDI_CH3I_Process.shm_reading_list, 0, &vc_ptr, &num_bytes);
			if (rc == MPI_SUCCESS)
			{
				MPIDI_DBG_PRINTF((50, FCNAME, L"MPIDI_CH3I_SHM_read_progress reported %d bytes read", num_bytes));
				mpi_errno = MPIDI_CH3I_Handle_shm_read(vc_ptr, num_bytes);
				if (mpi_errno != MPI_SUCCESS)
				{
					mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**progress", 0);
					goto fn_exit;
				}
				bShmProgressMade = TRUE;
			}
			else
			{
				if (rc != SHM_WAIT_TIMEOUT)
				{
					/*MPIDI_err_printf(L"MPIDI_CH3_Progress", L"MPIDI_CH3I_SHM_read_progress returned error %d\n", rc);*/
					mpi_errno = MPIR_Err_create_code(rc, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**progress", 0);
					goto fn_exit;
				}
			}
		}

		if (MPIDI_CH3I_Process.shm_writing_list)
		{
			vc_ptr = MPIDI_CH3I_Process.shm_writing_list;
			while (vc_ptr)
			{
				if (vc_ptr->ch.send_active != NULL)
				{
					mpi_errno = MPIDI_CH3I_SHM_write_progress(vc_ptr);
					if (mpi_errno != MPI_SUCCESS)
					{
						mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**progress", 0);
						goto fn_exit;
					}
					/*bShmProgressMade = TRUE;*/
				}
				vc_ptr = vc_ptr->ch.shm_next_writer;
			}
		}

		if (!bShmProgressMade)
		{
			MPIDI_FUNC_ENTER(MPID_STATE_MPIDU_YIELD);
			MPIDU_Yield(); /* always yield for now */
			MPIDI_FUNC_EXIT(MPID_STATE_MPIDU_YIELD);
		}

		if ((msg_queue_count++ % MPIDI_CH3I_MSGQ_ITERATIONS) == 0)
		{
			/* check for new shmem queue connection requests */
			rc = MPIDI_CH3I_BootstrapQ_recv_msg(MPIDI_Process.my_pg->ch.bootstrapQ, &info, sizeof(info), &num_bytes, FALSE);
			if (rc != MPI_SUCCESS)
			{
				mpi_errno = MPIR_Err_create_code(rc, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**boot_recv", 0);
				goto fn_exit;
			}
#ifdef MPICH_DBG_OUTPUT
			/*MPIU_Assert(num_bytes == 0 || num_bytes == sizeof(info));*/
			if (num_bytes != 0 && num_bytes != sizeof(info))
			{
				mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**bootqmsg", L"**bootqmsg %d", num_bytes);
				goto fn_exit;
			}
#endif
			if (num_bytes)
			{
				MPIDI_PG_t *pg;

				mpi_errno = MPIDI_PG_Find(info.pg_id, &pg);
				if (mpi_errno != MPI_SUCCESS)
				{
					mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", 0);
					goto fn_exit;
				}
				if (pg == NULL)
				{
					/* FIXME: Check for connect/accept case. */
					/* For now assume unmatched VC's are for MPI_Comm_connect */

					vc_ptr = (MPIDI_VC_t *) MPIU_Malloc(sizeof(MPIDI_VC_t));
					/* --BEGIN ERROR HANDLING-- */
					if (vc_ptr == NULL)
					{
						mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", NULL);
						goto fn_exit;
					}
					/* --END ERROR HANDLING-- */

					MPIDI_VC_Init(vc_ptr, NULL, 0);
					vc_ptr->ch.sendq_head = NULL;
					vc_ptr->ch.sendq_tail = NULL;
					vc_ptr->ch.state = MPIDI_CH3I_VC_STATE_UNCONNECTED;
					vc_ptr->ch.recv_active = NULL;
					vc_ptr->ch.send_active = NULL;
					vc_ptr->ch.req = NULL;
					vc_ptr->ch.read_shmq = NULL;
					vc_ptr->ch.write_shmq = NULL;
					vc_ptr->ch.shm = NULL;
					vc_ptr->ch.shm_state = 0;
					vc_ptr->ch.shm_next_reader = NULL;
					vc_ptr->ch.shm_next_writer = NULL;
					vc_ptr->ch.shm_read_connected = 0;

					/*
					mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", 0);
					goto fn_exit;
					*/
				}
				else
				{
					MPIDI_PG_Get_vc(pg, info.pg_rank, &vc_ptr);
				}
				rc = MPIDI_CH3I_SHM_Attach_to_memblock(&info.info, &vc_ptr->ch.shm_read_queue_info);
				if (rc != MPI_SUCCESS)
				{
					mpi_errno = MPIR_Err_create_code(rc, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**attach_to_mem", L"**attach_to_mem %d", vc_ptr->ch.shm_read_queue_info.error);
					goto fn_exit;
				}
				MPIU_DBG_PRINTF((L"attached to queue from process %d\n", info.pg_rank));
				/*MPIU_DBG_PRINTF((L"Opening process[%d]: %d\n", i, pSharedProcess[i].nPid));*/
				vc_ptr->ch.hSharedProcessHandle =
					OpenProcess(STANDARD_RIGHTS_REQUIRED | PROCESS_VM_READ | PROCESS_VM_WRITE | PROCESS_VM_OPERATION, 
					FALSE, info.pid);
				if (vc_ptr->ch.hSharedProcessHandle == NULL)
				{
					int err = GetLastError();
					mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**OpenProcess", L"**OpenProcess %d %d", info.pg_rank, err);
					return mpi_errno;
				}
				/*vc_ptr->ch.state = MPIDI_CH3I_VC_STATE_CONNECTED;*/ /* we are read connected but not write connected */
				vc_ptr->ch.shm_read_connected = 1;
				vc_ptr->ch.read_shmq = vc_ptr->ch.shm_read_queue_info.addr;/*info.info.addr;*/
				MPIU_DBG_PRINTF((L"read_shmq = %p\n", vc_ptr->ch.read_shmq));
				vc_ptr->ch.shm_reading_pkt = TRUE;
				/* add this VC to the global list to be shm_waited on */
				MPIDI_CH3I_SHM_Add_to_reader_list(vc_ptr);
			}
		}
	}
	while (completions == MPIDI_CH3I_progress_completion_count && is_blocking);

fn_exit:
#ifdef MPICH_DBG_OUTPUT
	count = MPIDI_CH3I_progress_completion_count - completions;
	if (is_blocking)
	{
		MPIDI_DBG_PRINTF((50, FCNAME, L"exiting, count=%d", count));
	}
	else
	{
		if (count > 0)
		{
			MPIDI_DBG_PRINTF((50, FCNAME, L"exiting (non-blocking), count=%d", count));
		}
	}
#endif
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_PROGRESS);
	return mpi_errno;
}

#if !defined(MPIDI_CH3_Progress_poke)
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_Progress_poke
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3_Progress_poke()
{
	int mpi_errno;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_PROGRESS_POKE);
	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_PROGRESS_POKE);
	mpi_errno = MPIDI_CH3.Progress_test();
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_PROGRESS_POKE);
	return mpi_errno;
}
#endif

#if !defined(MPIDI_CH3_Progress_start)
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_Progress_start
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
void MPIDI_CH3_Progress_start(MPID_Progress_state *state)
{
	/* MT - This function is empty for the single-threaded implementation */
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_PROGRESS_START);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_PROGRESS_START);
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_PROGRESS_START);
}
#endif

#if !defined(MPIDI_CH3_Progress_end)
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_Progress_end
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
void MPIDI_CH3_Progress_end(MPID_Progress_state *state)
{
	/* MT: This function is empty for the single-threaded implementation */
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_PROGRESS_END);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_PROGRESS_END);
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_PROGRESS_END);
}
#endif

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_Connection_terminate
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int sshmMPIDI_CH3_Connection_terminate(MPIDI_VC_t * vc)
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_CONNECTION_TERMINATE);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_CONNECTION_TERMINATE);

	/* There is no post_close for shm connections so handle them as closed immediately. */
	MPIDI_CH3I_SHM_Remove_vc_references(vc);
	mpi_errno = MPIDI_CH3U_Handle_connection(vc, MPIDI_VC_EVENT_TERMINATED);
	if (mpi_errno != MPI_SUCCESS)
	{
		MPIU_ERR_SET(mpi_errno,MPI_ERR_OTHER, L"**fail");
	}

	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_CONNECTION_TERMINATE);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_VC_post_shmconnect
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int MPIDI_CH3I_VC_post_shmconnect(MPIDI_VC_t * vc)
{
	wchar_t val[MPIDI_MAX_KVS_VALUE_LEN];
	int mpi_errno = MPI_SUCCESS;
	int connected;
	MPIDI_VC_t *iter;
	int count = 0;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_VC_POST_CONNECT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_VC_POST_CONNECT);

	MPIDI_DBG_PRINTF((60, FCNAME, L"entering"));

	if (vc->ch.state != MPIDI_CH3I_VC_STATE_UNCONNECTED)
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**vc_state", L"**vc_state %d", vc->ch.state);
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_VC_POST_CONNECT);
		return mpi_errno;
	}

	vc->ch.state = MPIDI_CH3I_VC_STATE_CONNECTING;

	mpi_errno = MPIDI_PG_GetConnString( vc->pg, vc->pg_rank, val, MPIDI_MAX_KVS_VALUE_LEN);
	if (mpi_errno != MPI_SUCCESS)
	{
		MPIU_ERR_SET(mpi_errno,MPI_ERR_OTHER, L"**fail");
		return mpi_errno;
	}

	MPIU_DBG_PRINTF((L"%d: %s\n", vc->pg_rank, val));

	/* attempt to connect through shared memory */
	connected = FALSE;
	MPIU_DBG_PRINTF((L"business card: <%d> = <%s>\n", vc->pg_rank, val));
	mpi_errno = MPIDI_CH3I_Shm_connect(vc, val, &connected);
	if (mpi_errno != MPI_SUCCESS)
	{
		mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**post_connect", L"**post_connect %s", L"MPIDI_CH3I_Shm_connect");
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_VC_POST_CONNECT);
		return mpi_errno;
	}
	if (!connected)
	{
		mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", L"**fail %s", L"unable to establish a shared memory queue connection");
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_VC_POST_CONNECT);
		return mpi_errno;
	}

	MPIDI_CH3I_SHM_Add_to_writer_list(vc);

	/* If there are more shm connections than cpus, reduce the spin count to one. */
	/* This does not take into account connections between other processes on the same machine. */
	iter = MPIDI_CH3I_Process.shm_writing_list;
	while (iter)
	{
		count++;
		iter = iter->ch.shm_next_writer;
	}
	if (count >= MPIDI_CH3I_Process.num_cpus)
		MPIDI_Process.my_pg->ch.nShmWaitSpinCount = 1;

	vc->ch.state = MPIDI_CH3I_VC_STATE_CONNECTED;
	vc->ch.shm_reading_pkt = TRUE;
	vc->ch.send_active = MPIDI_CH3I_SendQ_head(vc); /* MT */

	MPIDI_DBG_PRINTF((60, FCNAME, L"exiting"));
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_VC_POST_CONNECT);
	return mpi_errno;
}

/*
 * MPIDI_CH3_Win_create()
 */
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_Win_create
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3_Win_create(void *base, MPI_Aint size, int disp_unit, MPID_Info *info, MPID_Comm *comm_ptr, MPID_Win **win_ptr, MPIDI_RMAFns *RMAFns)
{
	int mpi_errno=MPI_SUCCESS, i, comm_size, rank, found, result;
	void *offset=0;
	MPIDI_CH3I_Alloc_mem_list_t *curr_ptr;
	MPIDU_Process_lock_t *locks_base_addr;
	int *shared_lock_state_baseaddr;
	volatile char *pscw_sync_addr;
	MPIU_CHKPMEM_DECL(4);
		
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_WIN_CREATE);
	MPIDI_STATE_DECL(MPID_STATE_MEMCPY);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_WIN_CREATE);

	/* We first call the generic MPIDI_Win_create */
	mpi_errno = MPIDI_Win_create(base, size, disp_unit, info, comm_ptr, win_ptr, RMAFns);
	if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }

	/* Now we do the channel-specific stuff */

	/* All processes first check whether their base address refers to an address in 
	   shared memory. If everyone's address is in shared memory, we set 
	   MPIDI_Use_optimized_rma=1 to indicate that shared memory optimizations are possible.
	   If even one process's win_base is not in shared memory, we revert to the generic 
	   implementation of RMA in CH3 by setting MPIDI_Use_optimized_rma=0. */

	/* For the special case where win_base is NULL, we treat it as if
	 * it is found in shared memory, so as not to disable
	   optimizations. For example, where the window is allocated in
	   shared memory on one process and others call win_create with NULL. */

	comm_size = comm_ptr->local_size;
	rank = comm_ptr->rank;

	curr_ptr = MPIDI_CH3I_Get_mem_list_head();
	if (base == NULL)
	{
		found = 1;
	}
	else
	{
		found = 0;
		while (curr_ptr != NULL)
		{
			if ((curr_ptr->shm_struct->addr <= base) && 
				(base < (void *) ((char *) curr_ptr->shm_struct->addr + curr_ptr->shm_struct->size)))
			{
				found = 1;
				offset = (void *) ((char *) curr_ptr->shm_struct->addr - (char *) base);
				break;
			}
			curr_ptr = curr_ptr->next;
		}
	}

	NMPI_Allreduce(mpi_errno, (&found, &result, 1, MPI_INT, MPI_BAND, comm_ptr->handle));
	if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }

	if (result == 0)
	{
		/* not all windows are in shared memory. can't be optimized. Reset functions pointers for the RMA functions to the default. */
		(*win_ptr)->shm_structs = NULL;
		(*win_ptr)->locks = NULL;

		RMAFns->Win_free = MPIDI_Win_free;
		RMAFns->Put = MPIDI_Put;
		RMAFns->Get = MPIDI_Get;
		RMAFns->Accumulate = MPIDI_Accumulate;
		RMAFns->Win_fence = MPIDI_Win_fence;
		RMAFns->Win_post = MPIDI_Win_post ;
		RMAFns->Win_start = MPIDI_Win_start;
		RMAFns->Win_complete = MPIDI_Win_complete ;
		RMAFns->Win_wait = MPIDI_Win_wait;
		RMAFns->Win_lock = MPIDI_Win_lock;
		RMAFns->Win_unlock = MPIDI_Win_unlock;
		/* leave win_create, alloc_mem, and free_mem as they are (set to the channel-specific version)  */
	}
	else
	{
		/* all windows in shared memory. can be optimized */
		
		/* again set the channel-specific version of the RMA functions because they may 
		   have been reset to the default in an earlier call to win_create */
		sshmMPIDI_CH3_RMAFnsInit( RMAFns );

		/* allocate memory for the shm_structs */
		MPIU_CHKPMEM_MALLOC((*win_ptr)->shm_structs, MPIDI_CH3I_Shmem_block_request_result *, 
							comm_size * sizeof(MPIDI_CH3I_Shmem_block_request_result), 
							mpi_errno, L"(*win_ptr)->shm_structs");

		/* allocate memory for the offsets from base of shared memory */
		MPIU_CHKPMEM_MALLOC((*win_ptr)->offsets, void **, comm_size * sizeof(void *), mpi_errno, L"(*win_ptr)->offsets");

		if (base != NULL)
		{
			/* copy this process's shmem struct into right location in array of shmem structs */
			MPIDI_FUNC_ENTER(MPID_STATE_MEMCPY);
			memcpy(&((*win_ptr)->shm_structs[rank]), curr_ptr->shm_struct, 
				   sizeof(MPIDI_CH3I_Shmem_block_request_result));
			MPIDI_FUNC_EXIT(MPID_STATE_MEMCPY);
			
			/* copy this process's offset into right location in array of offsets */
			(*win_ptr)->offsets[rank] = offset;
		}
		else
		{
			(*win_ptr)->shm_structs[rank].addr = NULL;
			(*win_ptr)->shm_structs[rank].size = 0;

			(*win_ptr)->offsets[rank] = 0;
		}

		/* collect everyone's shm_structs and offsets */
		NMPI_Allgather(mpi_errno, (MPI_IN_PLACE, 0, MPI_DATATYPE_NULL,
								   (*win_ptr)->shm_structs, 
								   sizeof(MPIDI_CH3I_Shmem_block_request_result), 
								   MPI_BYTE, comm_ptr->handle));
		if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }
		
		NMPI_Allgather(mpi_errno, (MPI_IN_PLACE, 0, MPI_DATATYPE_NULL,
								   (*win_ptr)->offsets, sizeof(void *), 
								   MPI_BYTE, comm_ptr->handle));
		if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }


		/* each process now attaches to the shared memory segments
		   (windows) of other processes (if they are non-zero), so that direct RMA is
		   possible. The newly mapped addresses are stored in the addr
		   field in the shmem struct. */ 

		for (i=0; i<comm_size; i++)
		{
			if ((i != rank) && ((*win_ptr)->shm_structs[i].size != 0))
			{
				mpi_errno = MPIDI_CH3I_SHM_Attach_notunlink_mem( &((*win_ptr)->shm_structs[i]) );
				if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }
			}
		}

		/* allocate memory for the locks and the shared lock
		 * state. locks are needed for accumulate operations and for
		 *  passive target RMA. */
		MPIU_CHKPMEM_MALLOC((*win_ptr)->locks, MPIDI_CH3I_Shmem_block_request_result *, 
							sizeof(MPIDI_CH3I_Shmem_block_request_result), 
							mpi_errno, L"(*win_ptr)->locks");


		/* Rank 0 allocates shared memory for an array of locks, one for each process, and 
		   for an array of shared lock states and initializes the
		   locks and the shared lock state. */  
		if (rank == 0)
		{
			mpi_errno = MPIDI_CH3I_SHM_Get_memblock(comm_size * sizeof(MPIDU_Process_lock_t)
											   + comm_size * sizeof(int), 
											   (*win_ptr)->locks);
			MPIU_ERR_CHKANDJUMP(mpi_errno,mpi_errno,MPI_ERR_OTHER,L"**nomem");

			locks_base_addr = (*win_ptr)->locks->addr;
			for (i=0; i<comm_size; i++)
				MPIDU_Process_lock_init(&locks_base_addr[i]);

			shared_lock_state_baseaddr = (int *) ((char *) (*win_ptr)->locks->addr + 
										 comm_size * sizeof(MPIDU_Process_lock_t));
			/* initialize shared lock state of all processes to 0 */
			for (i=0; i<comm_size; i++)
				shared_lock_state_baseaddr[i] = 0;
		}

		/* rank 0 broadcasts the locks struct to others so that they can attach to the shared 
		   memory containing the locks */
		NMPI_Bcast(mpi_errno, ((*win_ptr)->locks, 
							   sizeof(MPIDI_CH3I_Shmem_block_request_result), MPI_BYTE,
							   0, comm_ptr->handle));
		if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }

		/* Processes other than rank 0 attach to the shared memory containing the lock structure. */
		if (rank != 0)
		{
			mpi_errno = MPIDI_CH3I_SHM_Attach_notunlink_mem( (*win_ptr)->locks );
			if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }
		}

		/* allocate memory for the shm struct needed to allocate shared memory
		   for synchronizing post-start-complete-wait. */
		MPIU_CHKPMEM_MALLOC((*win_ptr)->pscw_shm_structs, MPIDI_CH3I_Shmem_block_request_result *, 
							comm_size * sizeof(MPIDI_CH3I_Shmem_block_request_result), 
							mpi_errno, L"(*win_ptr)->pscw_shm_structs");

		
		mpi_errno = MPIDI_CH3I_SHM_Get_memblock(2 * comm_size, &(*win_ptr)->pscw_shm_structs[rank]);
		if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }

		/* initialize it */
		pscw_sync_addr = (*win_ptr)->pscw_shm_structs[rank].addr;
		for (i=0; i<2*comm_size; i++)
		{
			pscw_sync_addr[i] = L'0';
		}

		/* collect everyone's pscw_shm_structs*/
		NMPI_Allgather(mpi_errno, (MPI_IN_PLACE, 0, MPI_DATATYPE_NULL,
								   (*win_ptr)->pscw_shm_structs, 
								   sizeof(MPIDI_CH3I_Shmem_block_request_result), 
								   MPI_BYTE, comm_ptr->handle));
		if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }
		

		/* each process now attaches to the shared memory for pscw sync
		   of other processes. */ 
		for (i=0; i<comm_size; i++)
		{
			if (i != rank)
			{
				mpi_errno = MPIDI_CH3I_SHM_Attach_notunlink_mem( &((*win_ptr)->pscw_shm_structs[i]) );
				if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }
			}
		}
	}

	(*win_ptr)->access_epoch_grp_ptr = NULL;
	(*win_ptr)->access_epoch_grp_ranks_in_win = NULL;
	(*win_ptr)->exposure_epoch_grp_ptr = NULL;
	(*win_ptr)->exposure_epoch_grp_ranks_in_win = NULL;
	(*win_ptr)->pt_rma_excl_lock = 0;
		
 fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_WIN_CREATE);
	return mpi_errno;
	/* --BEGIN ERROR HANDLING-- */
 fn_fail:
	MPIU_CHKPMEM_REAP();
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}



/*
 * MPIDI_CH3_Win_free()
 */
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_Win_free
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3_Win_free(MPID_Win **win_ptr)
{
	int mpi_errno = MPI_SUCCESS, comm_size, rank, i;
	MPID_Comm *comm_ptr;
	MPIDU_Process_lock_t *locks_base_addr;
	
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_WIN_FREE);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_WIN_FREE);

	MPID_Comm_get_ptr( (*win_ptr)->comm, comm_ptr );
	comm_size = comm_ptr->local_size;
	rank = comm_ptr->rank;

	/* barrier needed so that all passive target rmas directed toward this process are over */
	NMPI_Barrier(mpi_errno, ((*win_ptr)->comm));
	if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }

	NMPI_Comm_free(mpi_errno, (&((*win_ptr)->comm)));
	if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }

	MPIU_Free((*win_ptr)->base_addrs);
	MPIU_Free((*win_ptr)->disp_units);
	MPIU_Free((*win_ptr)->all_win_handles);
	MPIU_Free((*win_ptr)->pt_rma_puts_accs);
	
	for (i=0; i<comm_size; i++)
	{
		if ((i != rank) && ((*win_ptr)->shm_structs[i].size != 0))
		{
			mpi_errno = MPIDI_CH3I_SHM_Release_memblock( &((*win_ptr)->shm_structs[i]) );
			if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }
		}
	}

	if (rank != 0)
	{
		mpi_errno = MPIDI_CH3I_SHM_Release_memblock( (*win_ptr)->locks );
	}
	else
	{
		locks_base_addr = (*win_ptr)->locks->addr;
		for (i=0; i<comm_size; i++)
		{
			MPIDU_Process_lock_free(&locks_base_addr[i]);
		}
		mpi_errno = MPIDI_CH3I_SHM_Unlink_and_detach_mem( (*win_ptr)->locks );
	}
	if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }

	for (i=0; i<comm_size; i++)
	{
		if (i != rank)
		{
			mpi_errno = MPIDI_CH3I_SHM_Release_memblock( &((*win_ptr)->pscw_shm_structs[i]) );
		}
		else
		{
			mpi_errno = MPIDI_CH3I_SHM_Unlink_and_detach_mem( &((*win_ptr)->pscw_shm_structs[i]) );
		}
		if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }
	}

	MPIU_Free((*win_ptr)->locks);
	MPIU_Free((*win_ptr)->shm_structs);
	MPIU_Free((*win_ptr)->offsets);
	MPIU_Free((*win_ptr)->pscw_shm_structs);

	/* check whether refcount needs to be decremented here as in group_free */
	MPIU_Handle_obj_free( &MPID_Win_mem, *win_ptr );

 fn_exit:	
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_WIN_FREE);
	return mpi_errno;
	/* --BEGIN ERROR HANDLING-- */
 fn_fail:
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/*
 * MPIDI_CH3_Put()
 */
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_Put
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3_Put(void *origin_addr, int origin_count, MPI_Datatype
			origin_datatype, int target_rank, MPI_Aint target_disp,
			int target_count, MPI_Datatype target_datatype, MPID_Win *win_ptr)
{
	int mpi_errno = MPI_SUCCESS;
	void *target_addr;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_PUT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_PUT);

	target_addr = (char *) win_ptr->shm_structs[target_rank].addr + 
				  MPIU_PtrToAint(win_ptr->offsets[target_rank]) +
				  win_ptr->disp_units[target_rank] * target_disp;

	mpi_errno = MPIR_Localcopy (origin_addr, origin_count, origin_datatype, 
								target_addr, target_count, target_datatype);
	/* --BEGIN ERROR HANDLING-- */
	if (mpi_errno)
	{
		mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", 0);
	}
	/* --END ERROR HANDLING-- */

	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_PUT);
	return mpi_errno;
}

/*
 * MPIDI_CH3_Get()
 */
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_Get
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3_Get(void *origin_addr, int origin_count, MPI_Datatype
			origin_datatype, int target_rank, MPI_Aint target_disp,
			int target_count, MPI_Datatype target_datatype, MPID_Win *win_ptr)
{
	int mpi_errno = MPI_SUCCESS;
	void *target_addr;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_GET);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_GET);

	target_addr = (char *) win_ptr->shm_structs[target_rank].addr + 
				  MPIU_PtrToAint(win_ptr->offsets[target_rank]) +
				  win_ptr->disp_units[target_rank] * target_disp;

	mpi_errno = MPIR_Localcopy (target_addr, target_count, target_datatype,
								origin_addr, origin_count, origin_datatype);
	/* --BEGIN ERROR HANDLING-- */
	if (mpi_errno)
	{
		mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", 0);
	}
	/* --END ERROR HANDLING-- */


	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_GET);
	return mpi_errno;
}

/*
 * MPIDI_CH3_Accumulate()
 */
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_Accumulate
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3_Accumulate(void *origin_addr, int origin_count, MPI_Datatype
					origin_datatype, int target_rank, MPI_Aint target_disp,
					int target_count, MPI_Datatype target_datatype, MPI_Op op,
					MPID_Win *win_ptr)
{
	int mpi_errno = MPI_SUCCESS;
	void *target_addr, *source_addr, *tmp_buf=NULL;
	MPI_Aint true_lb, true_extent, extent;
	MPI_User_function *uop;
	MPIDU_Process_lock_t *locks_base_addr;
	MPIU_CHKLMEM_DECL(2);
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_ACCUMULATE);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_ACCUMULATE);

	target_addr = (char *) win_ptr->shm_structs[target_rank].addr + 
				  MPIU_PtrToAint(win_ptr->offsets[target_rank]) +
				  win_ptr->disp_units[target_rank] * target_disp;
	source_addr = origin_addr;

	locks_base_addr = win_ptr->locks->addr;

	if (op == MPI_REPLACE)
	{
		/* simply do a memcpy */

		/* all accumulate operations need to be done atomically. If the process does 
		   not already have an exclusive lock, acquire it */

		if (win_ptr->pt_rma_excl_lock == 0)
			MPIDU_Process_lock( &locks_base_addr[target_rank] );

		mpi_errno = MPIR_Localcopy (origin_addr, origin_count, origin_datatype, 
									target_addr, target_count, target_datatype);

		if (win_ptr->pt_rma_excl_lock == 0)
			MPIDU_Process_unlock( &locks_base_addr[target_rank] );

		if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }
	}
	else
	{
		MPIU_ERR_CHKANDJUMP1((HANDLE_GET_KIND(op) != HANDLE_KIND_BUILTIN), mpi_errno, MPI_ERR_OP, L"**opnotpredefined", L"**opnotpredefined %d", op );

		/* get the function by indexing into the op table */
		uop = MPIR_Op_table[op%16 - 1];
	
		if ((HANDLE_GET_KIND(target_datatype) == HANDLE_KIND_BUILTIN) &&
			(HANDLE_GET_KIND(origin_datatype) == HANDLE_KIND_BUILTIN))
		{
				/* basic datatype on origin and target */
			if (win_ptr->pt_rma_excl_lock == 0)
				MPIDU_Process_lock( &locks_base_addr[target_rank] );

			(*uop)(origin_addr, target_addr, &origin_count, &origin_datatype);

			if (win_ptr->pt_rma_excl_lock == 0)
				MPIDU_Process_unlock( &locks_base_addr[target_rank] );
		}
		else
		{
			if (origin_datatype != target_datatype)
			{
				/* allocate a tmp buffer of extent equal to extent of target buf */

				NMPI_Type_get_true_extent(mpi_errno, (target_datatype, 
													  &true_lb, &true_extent));
				if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }

				MPID_Datatype_get_extent_macro(target_datatype, extent); 

				MPIU_CHKLMEM_MALLOC(tmp_buf, void *, target_count * (MPIR_MAX(extent,true_extent)), mpi_errno, L"temporary buffer");

				/* adjust for potential negative lower bound in datatype */
				tmp_buf = (void *)((char*)tmp_buf - true_lb);

				/*  copy origin buffer into tmp_buf with same datatype as target. */
				mpi_errno = MPIR_Localcopy (origin_addr, origin_count, origin_datatype, 
											tmp_buf, target_count, target_datatype);
				if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }

				source_addr = tmp_buf;
			}
			
			if (HANDLE_GET_KIND(target_datatype) == HANDLE_KIND_BUILTIN)
			{
				/* basic datatype on target. call uop. */
				if (win_ptr->pt_rma_excl_lock == 0)
					MPIDU_Process_lock( &locks_base_addr[target_rank] );

				(*uop)(source_addr, target_addr, &target_count, &target_datatype);

				if (win_ptr->pt_rma_excl_lock == 0)
					MPIDU_Process_unlock( &locks_base_addr[target_rank] );
			}
			else
			{
				/* derived datatype on target */
				MPID_Segment *segp;
				DLOOP_VECTOR *dloop_vec;
				MPI_Aint first, last;
				int vec_len, i, type_size, count;
				MPI_Datatype type;
				MPID_Datatype *dtp;
				
				segp = MPID_Segment_alloc();
				MPIU_ERR_CHKANDJUMP((!segp), mpi_errno, MPI_ERR_OTHER, L"**nomem"); 

				MPID_Segment_init(NULL, target_count, target_datatype, segp, 0);
				first = 0;
				last  = SEGMENT_IGNORE_LAST;
				
				MPID_Datatype_get_ptr(target_datatype, dtp);
				vec_len = dtp->n_contig_blocks * target_count + 1; 
				/* +1 needed because Rob says so */
				MPIU_CHKLMEM_MALLOC(dloop_vec, DLOOP_VECTOR *, vec_len * sizeof(DLOOP_VECTOR), mpi_errno, L"dloop vector");
				
				MPID_Segment_pack_vector(segp, first, &last, dloop_vec, &vec_len);
				
				type = dtp->eltype;
				type_size = MPID_Datatype_get_basic_size(type);

				if (win_ptr->pt_rma_excl_lock == 0)
					MPIDU_Process_lock( &locks_base_addr[target_rank] );
				for (i=0; i<vec_len; i++)
				{
					count = (dloop_vec[i].DLOOP_VECTOR_LEN)/type_size;
					(*uop)((char *)source_addr + MPIU_PtrToAint( dloop_vec[i].DLOOP_VECTOR_BUF ),
						   (char *)target_addr + MPIU_PtrToAint( dloop_vec[i].DLOOP_VECTOR_BUF ),
						   &count, &type);
				}
				if (win_ptr->pt_rma_excl_lock == 0)
					MPIDU_Process_unlock( &locks_base_addr[target_rank] );
				
				MPID_Segment_free(segp);
			}
		}
	}

 fn_exit:
	MPIU_CHKLMEM_FREEALL();
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ACCUMULATE);
	return mpi_errno;
	/* --BEGIN ERROR HANDLING-- */
 fn_fail:
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_Win_fence
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3_Win_fence(int assert, MPID_Win *win_ptr)
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_WIN_FENCE);

	MPIDI_RMA_FUNC_ENTER(MPID_STATE_MPIDI_CH3_WIN_FENCE);

	MPIU_UNREFERENCED_ARG(assert);

	NMPI_Barrier(mpi_errno, (win_ptr->comm));

	if (mpi_errno != MPI_SUCCESS)
	{
		MPIU_ERR_SET(mpi_errno,MPI_ERR_OTHER,L"**fail");
	}

	MPIDI_RMA_FUNC_EXIT(MPID_STATE_MPIDI_CH3_WIN_FENCE);
	return mpi_errno;
}


#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_Win_post
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3_Win_post(MPID_Group *group_ptr, int assert, MPID_Win *win_ptr)
{
	int mpi_errno = MPI_SUCCESS;
	MPI_Group win_grp, grp;
	int i, grp_size, *ranks_in_grp, *ranks_in_win, dst, rank;
	volatile char *pscw_sync_addr;
	MPIU_CHKLMEM_DECL(1);
	MPIU_CHKPMEM_DECL(1);
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_WIN_POST);

	MPIDI_RMA_FUNC_ENTER(MPID_STATE_MPIDI_CH3_WIN_POST);

	/* First translate the ranks of the processes in
	   group_ptr to ranks in win_ptr->comm. Save the translated ranks
	   and group_ptr in the win object because they will be needed in 
	   end_epoch. */
		
	grp_size = group_ptr->size;
	
	MPIU_CHKLMEM_MALLOC(ranks_in_grp, int *, grp_size * sizeof(int), mpi_errno, L"ranks_in_grp");
	MPIU_CHKPMEM_MALLOC(ranks_in_win, int *, grp_size * sizeof(int), mpi_errno, L"ranks_in_win");
	
	for (i=0; i<grp_size; i++)
	{
		ranks_in_grp[i] = i;
	}
	
	grp = group_ptr->handle;
	
	NMPI_Comm_group(mpi_errno, (win_ptr->comm, &win_grp));
	if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }
	
	NMPI_Group_translate_ranks(mpi_errno, (grp, grp_size, ranks_in_grp, win_grp, ranks_in_win));
	if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }

	if ((assert & MPI_MODE_NOCHECK) == 0)
	{
		/* MPI_MODE_NOCHECK not specified. We need to notify the 
		   source processes that Post has been called. */
		NMPI_Comm_rank(mpi_errno, (win_ptr->comm, &rank));
		for (i=0; i<grp_size; i++)
		{
			dst = ranks_in_win[i];
			if (dst != rank) {
				/* Write a '1' in the sync array of the origin process at the location
				   indexed by the rank of this process */
				pscw_sync_addr = win_ptr->pscw_shm_structs[dst].addr;
				pscw_sync_addr[rank] = L'1';
			}
		}
	}	

	/* save the ranks_in_win and group_ptr in win object */
	win_ptr->exposure_epoch_grp_ranks_in_win = ranks_in_win;
	win_ptr->exposure_epoch_grp_ptr = group_ptr;
	
	MPIU_Object_add_ref( group_ptr );
	
 fn_exit:
	MPIU_CHKLMEM_FREEALL();
	MPIDI_RMA_FUNC_EXIT(MPID_STATE_MPIDI_CH3_WIN_POST);
	return mpi_errno;
	/* --BEGIN ERROR HANDLING-- */
 fn_fail:
	MPIU_CHKPMEM_REAP();
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_Win_start
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3_Win_start(MPID_Group *group_ptr, int assert, MPID_Win *win_ptr)
{
	int mpi_errno=MPI_SUCCESS;
	MPI_Group win_grp, grp;
	int i, grp_size, *ranks_in_grp, *ranks_in_win, src, rank;
	volatile char *pscw_sync_addr;
	MPIU_CHKLMEM_DECL(1);
	MPIU_CHKPMEM_DECL(1);
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_WIN_START);

	MPIDI_RMA_FUNC_ENTER(MPID_STATE_MPIDI_CH3_WIN_START);

	/* First translate the ranks of the processes in
	   group_ptr to ranks in win_ptr->comm. Save the translated ranks
	   and group_ptr in the win object because they will be needed in 
	   end_epoch. */
		
	grp_size = group_ptr->size;
	
	MPIU_CHKLMEM_MALLOC(ranks_in_grp, int *, grp_size * sizeof(int), mpi_errno, L"ranks_in_grp");
	MPIU_CHKPMEM_MALLOC(ranks_in_win, int *, grp_size * sizeof(int), mpi_errno, L"ranks_in_win");

	for (i=0; i<grp_size; i++)
	{
		ranks_in_grp[i] = i;
	}
	
	grp = group_ptr->handle;
	
	NMPI_Comm_group(mpi_errno, (win_ptr->comm, &win_grp));
	if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }
	
	NMPI_Group_translate_ranks(mpi_errno, (grp, grp_size, ranks_in_grp, win_grp, ranks_in_win));
	if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }
	
	if ((assert & MPI_MODE_NOCHECK) == 0)
	{
		/* MPI_MODE_NOCHECK not specified. Synchronization is necessary. */
		NMPI_Comm_rank(mpi_errno, (win_ptr->comm, &rank));
		for (i=0; i<grp_size; i++)
		{
			src = ranks_in_win[i];
			if (src != rank)
			{
				/* Wait until a '1' is written in the sync array by the target 
				   at the location indexed by the rank of the target */
				pscw_sync_addr = win_ptr->pscw_shm_structs[rank].addr;
				while (pscw_sync_addr[src] == L'0');

				/* reset it back to 0 */
				pscw_sync_addr[src] = L'0';
			}
		}
	}	

	/* save the ranks_in_win and group_ptr in win object */
	win_ptr->access_epoch_grp_ranks_in_win = ranks_in_win;
	win_ptr->access_epoch_grp_ptr = group_ptr;

	MPIU_Object_add_ref( group_ptr );

 fn_exit:
	MPIU_CHKLMEM_FREEALL();
	MPIDI_RMA_FUNC_EXIT(MPID_STATE_MPIDI_CH3_WIN_START);
	return mpi_errno;
	/* --BEGIN ERROR HANDLING-- */
 fn_fail:
	MPIU_CHKPMEM_REAP();
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}



#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_Win_complete
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3_Win_complete(MPID_Win *win_ptr)
{
	int mpi_errno = MPI_SUCCESS;
	int i, dst, *ranks_in_win, grp_size, rank, comm_size;
	volatile char *pscw_sync_addr;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_WIN_COMPLETE);

	MPIDI_RMA_FUNC_ENTER(MPID_STATE_MPIDI_CH3_WIN_COMPLETE);

	NMPI_Comm_rank(mpi_errno, (win_ptr->comm, &rank));
	NMPI_Comm_size(mpi_errno, (win_ptr->comm, &comm_size));

	/* Send a sync message to each target process */
	grp_size = win_ptr->access_epoch_grp_ptr->size;
	ranks_in_win = win_ptr->access_epoch_grp_ranks_in_win;
	for (i=0; i<grp_size; i++)
	{
		dst = ranks_in_win[i];
		if (dst != rank)
		{
			/* Write a '1' in the sync array of the target process at the location indexed by the rank of this process */
			pscw_sync_addr = (char *) win_ptr->pscw_shm_structs[dst].addr + comm_size;
			pscw_sync_addr[rank] = L'1';
		}
	}

	MPIU_Free(win_ptr->access_epoch_grp_ranks_in_win);
	MPIR_Group_release(win_ptr->access_epoch_grp_ptr);

	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_WIN_COMPLETE);
	return mpi_errno;
}



#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_Win_wait
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3_Win_wait(MPID_Win *win_ptr)
{
	int mpi_errno = MPI_SUCCESS;
	int i, src, *ranks_in_win, grp_size, rank, comm_size;
	volatile char *pscw_sync_addr;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_WIN_WAIT);

	MPIDI_RMA_FUNC_ENTER(MPID_STATE_MPIDI_CH3_WIN_WAIT);

	NMPI_Comm_rank(mpi_errno, (win_ptr->comm, &rank));
	NMPI_Comm_size(mpi_errno, (win_ptr->comm, &comm_size));

	grp_size = win_ptr->exposure_epoch_grp_ptr->size;
	ranks_in_win = win_ptr->exposure_epoch_grp_ranks_in_win;

	for (i=0; i<grp_size; i++)
	{
		src = ranks_in_win[i];
		if (src != rank)
		{
			/* Wait until a '1' is written in the sync array by the origin
			   at the location indexed by the rank of the origin */
			pscw_sync_addr = (char *) win_ptr->pscw_shm_structs[rank].addr + comm_size;
			while (pscw_sync_addr[src] == L'0');

			/* reset it to 0 */
			pscw_sync_addr[src] = L'0';
		}
	}

	MPIU_Free(win_ptr->exposure_epoch_grp_ranks_in_win);
	MPIR_Group_release(win_ptr->exposure_epoch_grp_ptr);

	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_WIN_WAIT);
	return mpi_errno;
}



/*
 * MPIDI_CH3_Win_lock()
 */
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_Win_lock
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3_Win_lock(int lock_type, int dest, int assert, MPID_Win *win_ptr)
{
	int mpi_errno = MPI_SUCCESS, comm_size;
	volatile int *shared_lock_state_baseaddr;
	MPIDU_Process_lock_t *locks_base_addr;
	MPID_Comm *comm_ptr;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_WIN_LOCK);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_WIN_LOCK);

	MPIU_UNREFERENCED_ARG(assert);
	MPID_Comm_get_ptr( win_ptr->comm, comm_ptr );
	comm_size = comm_ptr->local_size;

	locks_base_addr = win_ptr->locks->addr;
	shared_lock_state_baseaddr = (volatile int *) ((char *) win_ptr->locks->addr + comm_size * sizeof(MPIDU_Process_lock_t));

	if (lock_type == MPI_LOCK_SHARED)
	{
		/* acquire lock. increment shared lock state. release lock */

		MPIDU_Process_lock( &locks_base_addr[dest] );
		shared_lock_state_baseaddr[dest]++;
		MPIDU_Process_unlock( &locks_base_addr[dest] );
	}
	else
	{
		/* exclusive lock. acquire lock. check shared lock state. if non-zero (shared lock exists), unlock and try again. else retain lock. */

		for (;;)
		{
			MPIDU_Process_lock( &locks_base_addr[dest] );

			if (shared_lock_state_baseaddr[dest]) {
				MPIDU_Process_unlock( &locks_base_addr[dest] );
			}
			else
			{
				break;
			}
		}

		/* mark this as a passive target excl lock epoch so that a following accumulate does not need to acquire a lock */
		win_ptr->pt_rma_excl_lock = 1;
	}


	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_WIN_LOCK);
	return mpi_errno;
}


/*
 * MPIDI_CH3_Win_unlock()
 */
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_Win_unlock
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3_Win_unlock(int dest, MPID_Win *win_ptr)
{
	int mpi_errno = MPI_SUCCESS, comm_size;
	volatile int *shared_lock_state_baseaddr;
	MPIDU_Process_lock_t *locks_base_addr;
	MPID_Comm *comm_ptr;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_WIN_UNLOCK);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_WIN_UNLOCK);

	MPID_Comm_get_ptr( win_ptr->comm, comm_ptr );
	comm_size = comm_ptr->local_size;

	locks_base_addr = win_ptr->locks->addr;
	shared_lock_state_baseaddr = (volatile int *) ((char *) win_ptr->locks->addr + comm_size * sizeof(MPIDU_Process_lock_t));

	if (win_ptr->pt_rma_excl_lock == 0)
	{
		/* it's a shared lock. acquire lock. decrement shared lock state. release lock */

		MPIDU_Process_lock( &locks_base_addr[dest] );
		shared_lock_state_baseaddr[dest]--;
		MPIDU_Process_unlock( &locks_base_addr[dest] );
	}
	else
	{
		/* exclusive lock. just release it */

		MPIDU_Process_unlock( &locks_base_addr[dest] );
		win_ptr->pt_rma_excl_lock = 0;
	}

	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_WIN_UNLOCK);
	return mpi_errno;
}

MPIU_BOOL sshmMPIDI_CH3U_VC_NeedsSendClose( MPIDI_VC_t *vc )
{
	if (vc->state == MPIDI_VC_STATE_ACTIVE || 
		vc->state == MPIDI_VC_STATE_REMOTE_CLOSE
		/* sshm queues are uni-directional.  A VC that is connected 
		* in the read direction is marked MPIDI_VC_STATE_INACTIVE
		* so that a connection will be formed on the first write.  
		* Since the other side is marked MPIDI_VC_STATE_ACTIVE for 
		* writing we need to initiate the close protocol on the read side 
		* even if the write state is MPIDI_VC_STATE_INACTIVE. */
		|| ((vc->state == MPIDI_VC_STATE_INACTIVE) && vc->ch.shm_read_connected)
		)
	{
		return MPIU_TRUE;
	}
	return MPIU_FALSE;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_VC_SendClose
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
/*@
  sshmMPIDI_CH3U_VC_SendClose - Initiate a close on a virtual connection
  
  Input Parameters:
+ vc - Virtual connection to close
- i  - rank of virtual connection within a process group (used for debugging)

  Notes:
  The current state of this connection must be either 'MPIDI_VC_STATE_ACTIVE' 
  or 'MPIDI_VC_STATE_REMOTE_CLOSE'.  
  @*/
int sshmMPIDI_CH3_VC_SendClose( MPIDI_VC_t *vc, int rank )
{
	MPIDI_CH3_Pkt_t upkt;
	MPIDI_CH3_Pkt_close_t * close_pkt = &upkt.close;
	MPID_Request * sreq;
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3U_VC_SENDCLOSE);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3U_VC_SENDCLOSE);

	MPIU_UNREFERENCED_ARG(rank);

	/* FIXME: Remove this IFDEF */
	MPIU_Assert( vc->state == MPIDI_VC_STATE_ACTIVE || 
				 vc->state == MPIDI_VC_STATE_REMOTE_CLOSE 
				 /* sshm queues are uni-directional.  A VC that is connected 
				 * in the read direction is marked MPIDI_VC_STATE_INACTIVE
				 * so that a connection will be formed on the first write.  
				 * Since the other side is marked MPIDI_VC_STATE_ACTIVE for writing 
				 * we need to initiate the close protocol on the read side 
				 * even if the write state is MPIDI_VC_STATE_INACTIVE. */
				 || ((vc->state == MPIDI_VC_STATE_INACTIVE) && vc->ch.shm_read_connected) );

	MPIDI_Pkt_init(close_pkt, MPIDI_CH3_PKT_CLOSE);
	close_pkt->ack = (vc->state == MPIDI_VC_STATE_ACTIVE) ? FALSE : TRUE;
	
	/* MT: this is not thread safe */
	MPIDI_Outstanding_close_ops += 1;
	/*wprintf(L"close_ops = %d\n", MPIDI_Outstanding_close_ops);fflush(stdout);*/
	MPIU_DBG_MSG_FMT(CH3_CONNECT,VERBOSE,(MPIU_DBG_FDEST, L"sending close(%s) on vc (pg=%p) %p to rank %d, ops = %d", close_pkt->ack ? L"TRUE" : L"FALSE", vc->pg, vc, rank, MPIDI_Outstanding_close_ops));
					

	/*
	 * A close packet acknowledging this close request could be
	 * received during iStartMsg, therefore the state must
	 * be changed before the close packet is sent.
	 */
	if (vc->state == MPIDI_VC_STATE_ACTIVE)
	{
		MPIU_DBG_VCSTATECHANGE(vc, VC_STATE_LOCAL_CLOSE);
		vc->state = MPIDI_VC_STATE_LOCAL_CLOSE;
	}
	else
	{
		MPIU_Assert( vc->state == MPIDI_VC_STATE_REMOTE_CLOSE );
		MPIU_DBG_VCSTATECHANGE(vc, VC_STATE_CLOSE_ACKED);
		vc->state = MPIDI_VC_STATE_CLOSE_ACKED;
	}
				
	mpi_errno = MPIDI_CH3.iStartMsg(vc, close_pkt, sizeof(*close_pkt), &sreq);
	if (mpi_errno != MPI_SUCCESS)
	{
		MPIU_ERR_SET(mpi_errno,MPI_ERR_OTHER, L"**ch3|send_close_ack");
	}
	
	if (sreq != NULL)
	{
		MPID_Request_release(sreq);
	}

	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3U_VC_SENDCLOSE);
	return mpi_errno;
}

int sshmMPIDI_CH3_Init_fns(MPIDI_CH3_Functions_t *function_table)
{
	function_table->Init = sshmMPIDI_CH3_Init;
	function_table->PortFnsInit = sshmMPIDI_CH3_PortFnsInit;
	function_table->RMAFnsInit = sshmMPIDI_CH3_RMAFnsInit;
	function_table->Finalize = sshmMPIDI_CH3_Finalize;
	function_table->VC_Init = sshmMPIDI_CH3_VC_Init;
	function_table->PG_Init = sshmMPIDI_CH3_PG_Init;
	function_table->iStartMsg = sshmMPIDI_CH3_iStartMsg;
	function_table->iStartMsgv = sshmMPIDI_CH3_iStartMsgv;
	function_table->iSend = sshmMPIDI_CH3_iSend;
	function_table->iSendv = sshmMPIDI_CH3_iSendv;
	function_table->Progress_wait = sshmMPIDI_CH3_Progress_wait;
	function_table->Progress_test = sshmMPIDI_CH3_Progress_test;
	function_table->Connection_terminate = sshmMPIDI_CH3_Connection_terminate;
	function_table->Connect_to_root = sshmMPIDI_CH3_Connect_to_root;
	function_table->Complete_unidirectional_connection_connect = sshmMPIDI_CH3_Complete_unidirectional_connection_connect;
	function_table->Complete_unidirectional_connection_accept = sshmMPIDI_CH3_Complete_unidirectional_connection_accept;
	function_table->Cleanup_after_connection = sshmMPIDI_CH3_Cleanup_after_connection;
	function_table->iStartRndvMsg = sshmMPIDI_CH3_iStartRndvMsg;
	function_table->iStartRndvTransfer = sshmMPIDI_CH3_iStartRndvTransfer;
	function_table->Get_business_card = sshmMPIDI_CH3I_Get_business_card;
	function_table->VC_NeedsSendClose = sshmMPIDI_CH3U_VC_NeedsSendClose;
	function_table->VC_SendClose = sshmMPIDI_CH3_VC_SendClose;
	return MPI_SUCCESS;
}
