/*  (C) 2009 by Deino Software. */

#include "mpidi_ch3_impl.h"
#include "pmi.h"
#include <stdio.h>

/* prototypes */
static int MPIDI_SHM_InitRWProc( DWORD, HANDLE * );
static int ssmMPIDI_CH3I_Progress_init();
static int ssmMPIDI_CH3I_Progress_finalize(void);
static int ssmMPIDI_CH3I_VC_post_connect(MPIDI_VC_t * vc);
static int ssmMPIDI_CH3I_Progress_handle_sock_event(MPIDU_Sock_event_t * event);
static void ssmMPIDI_CH3I_Connection_free(MPIDI_CH3I_Connection_t * conn);

static int (*ssmMPIDI_CH3_Progress_wait)(MPID_Progress_state *state);
static int fixed_spinMPIDI_CH3I_Progress(int is_blocking, MPID_Progress_state *state);
static int fixed_activeMPIDI_CH3I_Progress(int is_blocking, MPID_Progress_state *state);

int MPIDI_CH3I_shm_read_active = 0;
int MPIDI_CH3I_shm_write_active = 0;
int MPIDI_CH3I_sock_read_active = 0;
int MPIDI_CH3I_sock_write_active = 0;
int MPIDI_CH3I_active_flag = 0;

/* Initialize for reading and writing to the designated process */
static int MPIDI_SHM_InitRWProc( DWORD pid, HANDLE *phandle )
{
	int mpi_errno = MPI_SUCCESS;
	*phandle =
		OpenProcess(STANDARD_RIGHTS_REQUIRED | PROCESS_VM_READ | 
					PROCESS_VM_WRITE | PROCESS_VM_OPERATION, 
					FALSE, pid);
	if (*phandle == NULL) {
		int err = GetLastError();
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**OpenProcess", L"**OpenProcess %d", err);
	}
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_Finalize
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int ssmMPIDI_CH3_Finalize(void)
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_FINALIZE);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_FINALIZE);
	
	mpi_errno = ssmMPIDI_CH3I_Progress_finalize();
	MPIU_ERR_CHK_POP(mpi_errno);

	mpi_errno = MPIDI_CH3U_Finalize_sshm();
	MPIU_ERR_CHK_POP(mpi_errno);

fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_FINALIZE);
	return mpi_errno;
}

#ifdef USE_MPIU_DBG_PRINT_VC

/* VC state printing debugging functions */

#undef FUNCNAME
#define FUNCNAME VCState2
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static const wchar_t * VCState2(MPIDI_VC_State_t state)
{
	static wchar_t unknown_state[32];

	switch (state)
	{
	case MPIDI_VC_STATE_INACTIVE:
		return L"INACTIVE";
	case MPIDI_VC_STATE_ACTIVE:
		return L"ACTIVE";
	case MPIDI_VC_STATE_LOCAL_CLOSE:
		return L"LOCAL_CLOSE";
	case MPIDI_VC_STATE_REMOTE_CLOSE:
		return L"REMOTE_CLOSE";
	case MPIDI_VC_STATE_CLOSE_ACKED:
		return L"CLOSE_ACKED";
	}
	MPIU_Snprintf(unknown_state, 32, _TRUNCATE, L"STATE_%d", state);
	return unknown_state;
}

#undef FUNCNAME
#define FUNCNAME VCState
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static const wchar_t * VCState(MPIDI_VC_t *vc)
{
	return VCState2(vc->state);
}

#undef FUNCNAME
#define FUNCNAME MPIU_DBG_PrintVC
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
void MPIU_DBG_PrintVC(MPIDI_VC_t *vc)
{
	wprintf(L"vc.state : %s\n", VCState(vc));
	wprintf(L"vc.pg_rank : %d\n", vc->pg_rank);
	wprintf(L"vc.ref_count: %d\n", vc->ref_count);
	wprintf(L"vc.lpid : %d\n", vc->lpid);
	if (vc->pg == NULL)
	{
		wprintf(L"vc.pg : NULL\n");
	}
	else
	{
		wprintf(L"vc.pg->id : %s\n", vc->pg->id);
		wprintf(L"vc.pg->size : %d\n", vc->pg->size);
		wprintf(L"vc.pg->ref_count : %d\n", vc->pg->ref_count);
	}
	fflush(stdout);
}

#undef FUNCNAME
#define FUNCNAME MPIU_DBG_PrintVCState2
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
void MPIU_DBG_PrintVCState2(MPIDI_VC_t *vc, MPIDI_VC_State_t new_state)
{
	wprintf(L"[%s%d]vc%d.state = %s->%s (%s)\n", MPIU_DBG_parent_str, MPIR_Process.comm_world->rank, vc->pg_rank, VCState(vc), VCState2(new_state), (vc->pg && vc->pg_rank >= 0) ? vc->pg->id : L"?");
	fflush(stdout);
}

#undef FUNCNAME
#define FUNCNAME MPIU_DBG_PrintVCState
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
void MPIU_DBG_PrintVCState(MPIDI_VC_t *vc)
{
	wprintf(L"[%s%d]vc%d.state = %s (%s)\n", MPIU_DBG_parent_str, MPIR_Process.comm_world->rank, vc->pg_rank, VCState(vc), (vc->pg && vc->pg_rank >= 0) ? vc->pg->id : L"?");
	fflush(stdout);
}

#endif /* USE_MPIU_DBG_PRINT_VC */

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_Init
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int ssmMPIDI_CH3_Init(int has_parent, MPIDI_PG_t *pg_p, int pg_rank )
{
	int mpi_errno = MPI_SUCCESS;
	int pg_size;
	int p;
	wchar_t *publish_bc_orig = NULL;
	wchar_t *bc_val = NULL;
	int val_max_remaining;
	MPIDI_STATE_DECL(MPID_STATE_MPID_CH3_INIT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPID_CH3_INIT);

	MPIDI_CH3I_Eager_max_msg_size = 128000;

	mpi_errno = ssmMPIDI_CH3I_Progress_init();
	if (mpi_errno != MPI_SUCCESS) MPIU_ERR_POP(mpi_errno);

	/* Initialize the business card */
	mpi_errno = MPIDI_CH3I_BCInit( &bc_val, &val_max_remaining );
	if (mpi_errno) MPIU_ERR_POP(mpi_errno);
	publish_bc_orig = bc_val;

	/* initialize aspects specific to sockets.  do NOT publish business card yet  */
	mpi_errno = MPIDI_CH3U_Init_sock(has_parent, pg_p, pg_rank, &bc_val, &val_max_remaining);
	if (mpi_errno != MPI_SUCCESS) MPIU_ERR_POP(mpi_errno);

	/* initialize aspects specific to sshm.  now publish business card   */
	mpi_errno = MPIDI_CH3U_Init_sshm(has_parent, pg_p, pg_rank, &bc_val, &val_max_remaining);
	if (mpi_errno != MPI_SUCCESS) MPIU_ERR_POP(mpi_errno);

	/* Set the connection information in our process group (publish the business card ) */
	MPIDI_PG_SetConnInfo( pg_rank, publish_bc_orig );

	/* Free the business card now that it is published (note that publish_bc_orig is the head of bc_val ) */
	MPIDI_CH3I_BCFree( publish_bc_orig );

	mpi_errno = PMI_Get_size(&pg_size);
	if (mpi_errno != 0)
	{
		MPIU_ERR_SETANDJUMP2(mpi_errno,MPI_ERR_OTHER, L"**pmi_get_size", L"**pmi_get_size %d %s", mpi_errno, PMI_Get_last_error_msgW());
	}

	/* Allocate and initialize the VC table associated with this process group (and thus COMM_WORLD) */
	/* FIXME: This doesn't allocate and only inits one field.  Is this
	   now part of the channel-specific hook for channel-specific VC info? */
	for (p = 0; p < pg_size; p++)
	{
		pg_p->vct[p].ch.bShm = FALSE;
	}

fn_exit:

	MPIDI_FUNC_EXIT(MPID_STATE_MPID_CH3_INIT);
	return mpi_errno;

fn_fail:
	goto fn_exit;
}

/* This function simply tells the CH3 device to use the defaults for the 
   MPI Port functions.  This should be ok here, since we want to 
   use the socket routines to perform all connect/accept actions
   after MPID_Init returns (see the shm_unlink discussion) */
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_PortFnsInit
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int ssmMPIDI_CH3_PortFnsInit( MPIDI_PortFns *a ) 
{ 
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_PORTFNSINIT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_PORTFNSINIT);
	MPIU_UNREFERENCED_ARG(a);
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_PORTFNSINIT);
	return MPI_SUCCESS;
}


/* Perform the channel-specific vc initialization */
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_VC_Init
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int ssmMPIDI_CH3_VC_Init( MPIDI_VC_t *vc )
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_VC_INIT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_VC_INIT);
	vc->ch.sendq_head		= NULL;
	vc->ch.sendq_tail		= NULL;
	vc->ch.state			= MPIDI_CH3I_VC_STATE_UNCONNECTED;
	mpi_errno = MPIDI_VC_InitSock( vc );
	MPIU_ERR_CHK_POP(mpi_errno);
	mpi_errno = MPIDI_VC_InitShm( vc );
	MPIU_ERR_CHK_POP(mpi_errno);
	/* This variable is used when sock and sshm are combined */
	vc->ch.bShm				= FALSE;
fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_VC_INIT);
	return mpi_errno;
}

/* Select the routine that uses sockets to connect two communicators using a socket */
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_Connect_to_root
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int ssmMPIDI_CH3_Connect_to_root(const wchar_t * port_name, MPIDI_VC_t ** new_vc)
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_CONNECT_TO_ROOT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_CONNECT_TO_ROOT);
	mpi_errno = MPIDI_CH3I_Connect_to_root_sock( port_name, new_vc );
	MPIU_ERR_CHK_POP(mpi_errno);
fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_CONNECT_TO_ROOT);
	return mpi_errno;
}

/*static void update_request(MPID_Request * sreq, void * pkt, MPIDI_msg_sz_t pkt_sz, int nb)*/
#undef update_request
#define update_request(sreq, pkt, pkt_sz, nb) \
{ \
	MPIDI_STATE_DECL(MPID_STATE_UPDATE_REQUEST); \
	MPIDI_FUNC_ENTER(MPID_STATE_UPDATE_REQUEST); \
	/*MPIU_Assert(pkt_sz == sizeof(MPIDI_CH3_Pkt_t));*/ \
	sreq->ch.pkt = *(MPIDI_CH3_Pkt_t *) pkt; \
	sreq->dev.iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)((char *) &sreq->ch.pkt + nb); \
	sreq->dev.iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)(pkt_sz - nb); \
	sreq->dev.iov_count = 1; \
	sreq->ch.iov_offset = 0; \
	MPIDI_FUNC_EXIT(MPID_STATE_UPDATE_REQUEST); \
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_iSend
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int ssmMPIDI_CH3_iSend(MPIDI_VC_t * vc, MPID_Request * sreq, void * pkt, MPIDI_msg_sz_t pkt_sz)
{
	int mpi_errno = MPI_SUCCESS;
	int complete;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_ISEND);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_ISEND);

	MPIDI_DBG_PRINTF((50, FCNAME, L"entering"));
#ifdef MPICH_DBG_OUTPUT
	if (pkt_sz > sizeof(MPIDI_CH3_Pkt_t))
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**arg", 0);
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISEND);
		return mpi_errno;
	}
#endif

	/* The sock channel uses a fixed length header, the size of which is the maximum of all possible packet headers */
	pkt_sz = sizeof(MPIDI_CH3_Pkt_t);
	MPIDI_DBG_Print_packet((MPIDI_CH3_Pkt_t*)pkt);
	
	if (vc->ch.state == MPIDI_CH3I_VC_STATE_CONNECTED) /* MT */
	{
		/* Connection already formed.  If send queue is empty attempt to send data, queuing any unsent data. */
		if (MPIDI_CH3I_SendQ_empty(vc)) /* MT */
		{
			int nb;
			MPIU_Size_t snb;

			MPIDI_DBG_PRINTF((55, FCNAME, L"send queue empty, attempting to write"));

			/* MT: need some signalling to lock down our right to use the channel, thus insuring that the progress engine does also try to write */
			if (vc->ch.bShm)
			{
				mpi_errno = MPIDI_CH3I_SHM_write(vc, pkt, (int)pkt_sz, &nb);
			}
			else
			{
				mpi_errno = MPIDU_Sock_write(vc->ch.sock, pkt, (MPIU_Size_t)pkt_sz, &snb);
				nb = (int)snb;
			}
			if (mpi_errno == MPI_SUCCESS)
			{
				MPIDI_DBG_PRINTF((55, FCNAME, L"wrote %d bytes", nb));

				if (nb == pkt_sz)
				{ 
					MPIDI_DBG_PRINTF((55, FCNAME, L"write complete %d bytes, calling MPIDI_CH3U_Handle_send_req()", nb));
					MPIDI_CH3U_Handle_send_req(vc, sreq, &complete);
					if (!complete)
					{
						sreq->ch.iov_offset = 0;
						MPIDI_CH3I_SendQ_enqueue_head(vc, sreq);
						if (vc->ch.bShm)
						{
							vc->ch.send_active = sreq;
						}
						else
						{
							vc->ch.conn->send_active = sreq;
							mpi_errno = MPIDU_Sock_post_writev(vc->ch.conn->sock, sreq->dev.iov, sreq->dev.iov_count, NULL);
							if (mpi_errno != MPI_SUCCESS)
							{
								mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER,
									L"**ch3|sock|postwrite", L"ch3|sock|postwrite %p %p %p",
									sreq, vc->ch.conn, vc);
							}
						}
					}
				}
				else
				{
					MPIDI_DBG_PRINTF((55, FCNAME, L"partial write of %d bytes, request enqueued at head", nb));
					update_request(sreq, pkt, pkt_sz, nb);
					MPIDI_CH3I_SendQ_enqueue_head(vc, sreq);
					if (vc->ch.bShm)
					{
						vc->ch.send_active = sreq;
					}
					else
					{
						vc->ch.conn->send_active = sreq;
						mpi_errno = MPIDU_Sock_post_write(vc->ch.conn->sock, sreq->dev.iov[0].MPID_IOV_BUF,
							(MPIU_Size_t)sreq->dev.iov[0].MPID_IOV_LEN, (MPIU_Size_t)sreq->dev.iov[0].MPID_IOV_LEN, NULL);
						if (mpi_errno != MPI_SUCCESS)
						{
							mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER,
								L"**ch3|sock|postwrite", L"ch3|sock|postwrite %p %p %p",
								sreq, vc->ch.conn, vc);
						}
					}
				}
			}
			else
			{
				vc->ch.state = MPIDI_CH3I_VC_STATE_FAILED;
				/* TODO: Create an appropriate error message based on the return value (rc) */
				sreq->status.MPI_ERROR = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ssmwrite", 0);
				 /* MT -CH3U_Request_complete() performs write barrier */
				MPIDI_CH3U_Request_complete(sreq);
			}
		}
		else
		{
			MPIDI_DBG_PRINTF((55, FCNAME, L"send queue not empty, enqueuing"));
			update_request(sreq, pkt, pkt_sz, 0);
			MPIDI_CH3I_SendQ_enqueue(vc, sreq);
		}
	}
	else if (vc->ch.state == MPIDI_CH3I_VC_STATE_UNCONNECTED) /* MT */
	{
		/* Form a new connection, queuing the data so it can be sent later. */
		MPIDI_DBG_PRINTF((55, FCNAME, L"unconnected.  enqueuing request"));
		update_request(sreq, pkt, pkt_sz, 0);
		MPIDI_CH3I_SendQ_enqueue(vc, sreq);
		mpi_errno = ssmMPIDI_CH3I_VC_post_connect(vc);
		if (mpi_errno != MPI_SUCCESS)
		{
			mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", 0);
		}
	}
	else if (vc->ch.state != MPIDI_CH3I_VC_STATE_FAILED)
	{
		/* Unable to send data at the moment, so queue it for later */
		MPIDI_DBG_PRINTF((55, FCNAME, L"still connecting.  enqueuing request"));
		update_request(sreq, pkt, pkt_sz, 0);
		MPIDI_CH3I_SendQ_enqueue(vc, sreq);
	}
	else
	{
		/* Connection failed.  Mark the request complete and return an error. */
		/* TODO: Create an appropriate error message */
		/* FIXME: For fault tolerance, retry the operation before marking the
		   connection as failed */
		sreq->status.MPI_ERROR = MPI_ERR_INTERN;
		/* MT - CH3U_Request_complete() performs write barrier */
		MPIDI_CH3U_Request_complete(sreq);
	}
	
	MPIDI_DBG_PRINTF((50, FCNAME, L"exiting"));
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISEND);
	return mpi_errno;
}

/*static void update_request(MPID_Request * sreq, MPID_IOV * iov, int count, int offset, int nb)*/
#undef update_request
#define update_request(sreq, iov, count, offset, nb) \
{ \
	int i; \
	MPIDI_STATE_DECL(MPID_STATE_UPDATE_REQUEST); \
	MPIDI_FUNC_ENTER(MPID_STATE_UPDATE_REQUEST); \
	for (i = 0; i < count; i++) \
	{ \
		sreq->dev.iov[i] = iov[i]; \
	} \
	if (offset == 0) \
	{ \
		/*MPIU_Assert(iov[0].MPID_IOV_LEN == sizeof(MPIDI_CH3_Pkt_t));*/ \
		sreq->ch.pkt = *(MPIDI_CH3_Pkt_t *) iov[0].MPID_IOV_BUF; \
		sreq->dev.iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST) &sreq->ch.pkt; \
	} \
	sreq->dev.iov[offset].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)((char *) sreq->dev.iov[offset].MPID_IOV_BUF + nb); \
	sreq->dev.iov[offset].MPID_IOV_LEN -= (MPID_IOV_LEN_CAST)nb; \
	sreq->dev.iov_count = count; \
	sreq->ch.iov_offset = offset; \
	MPIDI_FUNC_EXIT(MPID_STATE_UPDATE_REQUEST); \
}

#undef update_request_zeros
#define update_request_zeros(sreq, iov, count) \
{ \
	int i; \
	MPIDI_STATE_DECL(MPID_STATE_UPDATE_REQUEST); \
	MPIDI_FUNC_ENTER(MPID_STATE_UPDATE_REQUEST); \
	for (i = 0; i < count; i++) \
	{ \
		sreq->dev.iov[i] = iov[i]; \
	} \
	sreq->ch.pkt = *(MPIDI_CH3_Pkt_t *) iov[0].MPID_IOV_BUF; \
	sreq->dev.iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST) &sreq->ch.pkt; \
	sreq->dev.iov_count = count; \
	sreq->ch.iov_offset = 0; \
	MPIDI_FUNC_EXIT(MPID_STATE_UPDATE_REQUEST); \
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_iSendv
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int ssmMPIDI_CH3_iSendv(MPIDI_VC_t * vc, MPID_Request * sreq, MPID_IOV * iov, int n_iov)
{
	int mpi_errno = MPI_SUCCESS;
	int complete;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_ISENDV);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_ISENDV);
	MPIDI_DBG_PRINTF((50, FCNAME, L"entering"));
#ifdef MPICH_DBG_OUTPUT
	if (n_iov > MPID_IOV_LIMIT)
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**arg", 0);
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISENDV);
		return mpi_errno;
	}
	if (iov[0].MPID_IOV_LEN > (MPID_IOV_LEN_CAST)sizeof(MPIDI_CH3_Pkt_t))
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**arg", 0);
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISENDV);
		return mpi_errno;
	}
#endif

	/* The mm channel uses a fixed length header, the size of which is the maximum of all possible packet headers */
	iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)sizeof(MPIDI_CH3_Pkt_t);
	MPIDI_DBG_Print_packet((MPIDI_CH3_Pkt_t*)iov[0].MPID_IOV_BUF);

	if (vc->ch.state == MPIDI_CH3I_VC_STATE_CONNECTED) /* MT */
	{
		/* Connection already formed.  If send queue is empty attempt to send data, queuing any unsent data. */
		if (MPIDI_CH3I_SendQ_empty(vc)) /* MT */
		{
			int nb;
			MPIU_Size_t snb;

			MPIDI_DBG_PRINTF((55, FCNAME, L"send queue empty, attempting to write"));
			
			/* MT - need some signalling to lock down our right to use the channel, thus insuring that the progress engine does also try to write */

			/* FIXME: the current code only agressively writes the first IOV.  Eventually it should be changed to agressively write
			   as much as possible.  Ideally, the code would be shared between the send routines and the progress engine. */
			if (vc->ch.bShm)
			{
				mpi_errno = MPIDI_CH3I_SHM_writev(vc, iov, n_iov, &nb);
			}
			else
			{
				mpi_errno = MPIDU_Sock_writev(vc->ch.sock, iov, n_iov, &snb);
				nb = (int)snb;
			}
			if (mpi_errno == MPI_SUCCESS)
			{
				int offset = 0;

				MPIDI_DBG_PRINTF((55, FCNAME, L"wrote %d bytes", nb));
				
				while (offset < n_iov)
				{
					if ((int)iov[offset].MPID_IOV_LEN <= nb)
					{
						nb -= (int)iov[offset].MPID_IOV_LEN;
						offset++;
					}
					else
					{
						MPIDI_DBG_PRINTF((55, FCNAME, L"partial write, request enqueued at head"));
						update_request(sreq, iov, n_iov, offset, nb);
						MPIDI_CH3I_SendQ_enqueue_head(vc, sreq);
						if (vc->ch.bShm)
						{
							vc->ch.send_active = sreq;
						}
						else
						{
							/*MPIDI_CH3I_SSM_VC_post_write(vc, sreq);*/
							MPIDI_DBG_PRINTF((55, FCNAME, L"posting writev, vc=0x%p, sreq=0x%08x", vc, sreq->handle));
							vc->ch.conn->send_active = sreq;
							mpi_errno = MPIDU_Sock_post_writev(vc->ch.conn->sock, sreq->dev.iov + offset, sreq->dev.iov_count - offset, NULL);
							if (mpi_errno != MPI_SUCCESS)
							{
								mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|sock|postwrite", L"ch3|sock|postwrite %p %p %p", sreq, vc->ch.conn, vc);
							}
						}
						break;
					}

				}
				if (offset == n_iov)
				{
					MPIDI_DBG_PRINTF((55, FCNAME, L"write complete, calling MPIDI_CH3U_Handle_send_req()"));
					MPIDI_CH3U_Handle_send_req(vc, sreq, &complete);
					if (!complete)
					{
						sreq->ch.iov_offset = 0;
						MPIDI_CH3I_SendQ_enqueue_head(vc, sreq);
						if (vc->ch.bShm)
						{
							vc->ch.send_active = sreq;
						}
						else
						{
							MPIDI_DBG_PRINTF((55, FCNAME, L"posting writev, vc=0x%p, sreq=0x%08x", vc, sreq->handle));
							vc->ch.conn->send_active = sreq;
							mpi_errno = MPIDU_Sock_post_writev(vc->ch.conn->sock, sreq->dev.iov, sreq->dev.iov_count, NULL);
							if (mpi_errno != MPI_SUCCESS)
							{
								mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER,
									L"**ch3|sock|postwrite", L"ch3|sock|postwrite %p %p %p",
									sreq, vc->ch.conn, vc);
							}
						}
					}
					/*
					MPIDI_CH3I_SendQ_enqueue_head(vc, sreq);
					MPIDI_CH3U_Handle_send_req(vc, sreq);
					if (sreq->dev.iov_count == 0)
					{
						if (MPIDI_CH3I_SendQ_head(vc) == sreq)
						{
							MPIDI_CH3I_SendQ_dequeue(vc);
						}
					}
					*/
				}
			}
			else
			{
				vc->ch.state = MPIDI_CH3I_VC_STATE_FAILED;
				/* TODO: Create an appropriate error message based on the value of errno */
				sreq->status.MPI_ERROR = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ssmwritev", 0);
				/* MT - CH3U_Request_complete performs write barrier */
				MPIDI_CH3U_Request_complete(sreq);
			}
		}
		else
		{
			MPIDI_DBG_PRINTF((55, FCNAME, L"send queue not empty, enqueuing"));
			update_request_zeros(sreq, iov, n_iov);
			MPIDI_CH3I_SendQ_enqueue(vc, sreq);
		}
	}
	else if (vc->ch.state == MPIDI_CH3I_VC_STATE_UNCONNECTED)
	{
		/* Form a new connection, queuing the data so it can be sent later. */
		MPIDI_DBG_PRINTF((55, FCNAME, L"unconnected.  enqueuing request"));
		/*ssmMPIDI_CH3I_VC_post_connect(vc);*/
		update_request_zeros(sreq, iov, n_iov);
		MPIDI_CH3I_SendQ_enqueue(vc, sreq);
		mpi_errno = ssmMPIDI_CH3I_VC_post_connect(vc);
		if (mpi_errno != MPI_SUCCESS)
		{
			mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", 0);
		}
	}
	else if (vc->ch.state != MPIDI_CH3I_VC_STATE_FAILED)
	{
		/* Unable to send data at the moment, so queue it for later */
		MPIDI_DBG_PRINTF((55, FCNAME, L"still connecting.  enqueuing request"));
		update_request_zeros(sreq, iov, n_iov);
		MPIDI_CH3I_SendQ_enqueue(vc, sreq);
	}
	else
	{
		/* Connection failed.  Mark the request complete and return an error. */
		/* TODO: Create an appropriate error message */
		sreq->status.MPI_ERROR = MPI_ERR_INTERN;
		/* MT - CH3U_Request_complete performs write barrier */
		MPIDI_CH3U_Request_complete(sreq);
	}
	
	MPIDI_DBG_PRINTF((50, FCNAME, L"exiting"));
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISENDV);
	return mpi_errno;
}

/*static MPID_Request * create_request(void * pkt, MPIDI_msg_sz_t pkt_sz, int nb)*/
#undef create_request
#define create_request(sreq, pkt, pkt_sz, nb) \
{ \
	/*MPID_Request * sreq;*/ \
	MPIDI_STATE_DECL(MPID_STATE_CREATE_REQUEST); \
	MPIDI_FUNC_ENTER(MPID_STATE_CREATE_REQUEST); \
	sreq = MPID_Request_create(); \
	/*if (sreq == NULL) return NULL;*/ \
	if (sreq == NULL) \
	{ \
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0); \
		MPIDI_FUNC_EXIT(MPID_STATE_CREATE_REQUEST); \
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISTARTMSG); \
		return mpi_errno; \
	} \
	MPIU_Object_set_ref(sreq, 2); \
	sreq->kind = MPID_REQUEST_SEND; \
	/*MPIU_Assert(pkt_sz == sizeof(MPIDI_CH3_Pkt_t));*/ \
	sreq->ch.pkt = *(MPIDI_CH3_Pkt_t *) pkt; \
	sreq->dev.iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)((char *) &sreq->ch.pkt + nb); \
	sreq->dev.iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)(pkt_sz - nb); \
	sreq->dev.iov_count = 1; \
	sreq->ch.iov_offset = 0; \
	sreq->dev.ca = MPIDI_CH3_CA_COMPLETE; \
	MPIDI_FUNC_EXIT(MPID_STATE_CREATE_REQUEST); \
	/*return sreq;*/ \
}

/*
 * MPIDI_CH3_iStartMsg() attempts to send the message immediately.  If the entire message is successfully sent, then NULL is
 * returned.  Otherwise a request is allocated, the header is copied into the request, and a pointer to the request is returned.
 * An error condition also results in a request be allocated and the errror being returned in the status field of the request.
 */
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_iStartMsg
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int ssmMPIDI_CH3_iStartMsg(MPIDI_VC_t * vc, void * pkt, MPIDI_msg_sz_t pkt_sz, MPID_Request **sreq_ptr)
{
	int mpi_errno = MPI_SUCCESS;
	MPID_Request * sreq = NULL;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_ISTARTMSG);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_ISTARTMSG);
	
	MPIDI_DBG_PRINTF((50, FCNAME, L"entering"));
#ifdef MPICH_DBG_OUTPUT
	if (pkt_sz > sizeof(MPIDI_CH3_Pkt_t))
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**arg", 0);
		goto fn_exit;
	}
#endif

	/* The MM channel uses a fixed length header, the size of which is the maximum of all possible packet headers */
	pkt_sz = sizeof(MPIDI_CH3_Pkt_t);
	MPIDI_DBG_Print_packet((MPIDI_CH3_Pkt_t*)pkt);
	
	if (vc->ch.state == MPIDI_CH3I_VC_STATE_CONNECTED) /* MT */
	{
		/* Connection already formed.  If send queue is empty attempt to send data, queuing any unsent data. */
		if (MPIDI_CH3I_SendQ_empty(vc)) /* MT */
		{
			int nb;
			MPIU_Size_t snb;

			/* MT - need some signalling to lock down our right to use the channel, thus insuring that the progress engine does not also try to write */
			if (vc->ch.bShm)
			{
				mpi_errno = MPIDI_CH3I_SHM_write(vc, pkt, (int)pkt_sz, &nb);
			}
			else
			{
				mpi_errno = MPIDU_Sock_write(vc->ch.sock, pkt, (MPIU_Size_t)pkt_sz, &snb);
				nb = (int)snb;
			}
			if (mpi_errno == MPI_SUCCESS)
			{
				MPIDI_DBG_PRINTF((55, FCNAME, L"wrote %d bytes", nb));

				if (nb == pkt_sz)
				{ 
					MPIDI_DBG_PRINTF((55, FCNAME, L"entire write complete, %d bytes", nb));
					/* done.  get us out of here as quickly as possible. */
				}
				else
				{
					MPIDI_DBG_PRINTF((55, FCNAME, L"partial write of %d bytes, request enqueued at head", nb));
					create_request(sreq, pkt, pkt_sz, nb);
					/*sreq = create_request(pkt, pkt_sz, nb);
					if (sreq == NULL)
					{
					mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
					MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISTARTMSG);
					return mpi_errno;
					}
					*/
					MPIDI_CH3I_SendQ_enqueue_head(vc, sreq);
					if (vc->ch.bShm)
					{
						vc->ch.send_active = sreq;
					}
					else
					{
						/*MPIDI_CH3I_SSM_VC_post_write(vc, sreq);*/
						MPIDI_DBG_PRINTF((55, FCNAME, L"posting write, vc=0x%p, sreq=0x%08x", vc, sreq->handle));
						vc->ch.conn->send_active = sreq;
						mpi_errno = MPIDU_Sock_post_write(vc->ch.conn->sock, sreq->dev.iov[0].MPID_IOV_BUF,
							(MPIU_Size_t)sreq->dev.iov[0].MPID_IOV_LEN, (MPIU_Size_t)sreq->dev.iov[0].MPID_IOV_LEN, NULL);
						if (mpi_errno != MPI_SUCCESS)
						{
							mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|sock|postwrite", L"ch3|sock|postwrite %p %p %p", sreq, vc->ch.conn, vc);
							goto fn_exit;
						}
					}
				}
			}
			else
			{
				sreq = MPID_Request_create();
				if (sreq == NULL)
				{
					mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
					goto fn_exit;
				}
				sreq->kind = MPID_REQUEST_SEND;
				sreq->cc = 0;
				/* TODO: Create an appropriate error message based on the return value */
				sreq->status.MPI_ERROR = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ssmwrite", 0);
			}
		}
		else
		{
			MPIDI_DBG_PRINTF((55, FCNAME, L"send in progress, request enqueued"));
			create_request(sreq, pkt, pkt_sz, 0);
			/*sreq = create_request(pkt, pkt_sz, 0);
			if (sreq == NULL)
			{
				mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
				MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISTARTMSG);
				return mpi_errno;
			}
			*/
			MPIDI_CH3I_SendQ_enqueue(vc, sreq);
		}
	}
	else if (vc->ch.state == MPIDI_CH3I_VC_STATE_UNCONNECTED) /* MT */
	{
		MPIDI_DBG_PRINTF((55, FCNAME, L"unconnected.  posting connect and enqueuing request"));
		
		/* Form a new connection */
		/*ssmMPIDI_CH3I_VC_post_connect(vc);*/

		/* queue the data so it can be sent after the connection is formed */
		create_request(sreq, pkt, pkt_sz, 0);
		/*sreq = create_request(pkt, pkt_sz, 0);
		if (sreq == NULL)
		{
			mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
			MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISTARTMSG);
			return mpi_errno;
		}
		*/
		MPIDI_CH3I_SendQ_enqueue(vc, sreq);

		ssmMPIDI_CH3I_VC_post_connect(vc);
	}
	else if (vc->ch.state != MPIDI_CH3I_VC_STATE_FAILED)
	{
		/* Unable to send data at the moment, so queue it for later */
		MPIDI_DBG_PRINTF((55, FCNAME, L"forming connection, request enqueued"));
		create_request(sreq, pkt, pkt_sz, 0);
		/*sreq = create_request(pkt, pkt_sz, 0);
		if (sreq == NULL)
		{
			mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
			MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISTARTMSG);
			return mpi_errno;
		}
		*/
		MPIDI_CH3I_SendQ_enqueue(vc, sreq);
	}
	else
	{
		/* Connection failed, so allocate a request and return an error. */
		MPIDI_DBG_PRINTF((55, FCNAME, L"ERROR - connection failed"));
		sreq = MPID_Request_create();
		if (sreq == NULL)
		{
			mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
			goto fn_exit;
		}
		sreq->kind = MPID_REQUEST_SEND;
		sreq->cc = 0;
		/* TODO: Create an appropriate error message */
		sreq->status.MPI_ERROR = MPI_ERR_INTERN;
	}

fn_exit:
	*sreq_ptr = sreq;

	MPIDI_DBG_PRINTF((50, FCNAME, L"exiting"));
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISTARTMSG);
	return mpi_errno;
}

/*static MPID_Request * create_request(MPID_IOV * iov, int count, int offset, int nb)*/
#undef create_request
#define create_request(sreq, iov, count, offset, nb) \
{ \
	/*MPID_Request * sreq;*/ \
	/*int i;*/ \
	MPIDI_STATE_DECL(MPID_STATE_CREATE_REQUEST); \
	MPIDI_STATE_DECL(MPID_STATE_MEMCPY); \
	MPIDI_FUNC_ENTER(MPID_STATE_CREATE_REQUEST); \
	sreq = MPID_Request_create(); \
	if (sreq == NULL) \
	{ \
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0); \
		MPIDI_FUNC_EXIT(MPID_STATE_CREATE_REQUEST); \
		goto fn_exit; \
	} \
	MPIU_Object_set_ref(sreq, 2); \
	sreq->kind = MPID_REQUEST_SEND; \
	MPIDI_FUNC_ENTER(MPID_STATE_MEMCPY); \
	memcpy(sreq->dev.iov, iov, count * sizeof(MPID_IOV)); \
	MPIDI_FUNC_EXIT(MPID_STATE_MEMCPY); \
	/*for (i = 0; i < count; i++) { sreq->dev.iov[i] = iov[i]; }*/ \
	if (offset == 0) \
	{ \
		/* memcpy(&sreq->ch.pkt, iov[0].MPID_IOV_BUF, iov[0].MPID_IOV_LEN); */ \
		/*MPIU_Assert(iov[0].MPID_IOV_LEN == sizeof(MPIDI_CH3_Pkt_t));*/ \
		sreq->ch.pkt = *(MPIDI_CH3_Pkt_t *) iov[0].MPID_IOV_BUF; \
		sreq->dev.iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST) &sreq->ch.pkt; \
	} \
	sreq->dev.iov[offset].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)((char *) sreq->dev.iov[offset].MPID_IOV_BUF + nb); \
	sreq->dev.iov[offset].MPID_IOV_LEN -= (MPID_IOV_LEN_CAST)nb; \
	sreq->dev.iov_count = count; \
	sreq->ch.iov_offset = offset; \
	sreq->dev.ca = MPIDI_CH3_CA_COMPLETE; \
	MPIDI_FUNC_EXIT(MPID_STATE_CREATE_REQUEST); \
	/*return sreq;*/ \
}

#undef create_request_zeros
#define create_request_zeros(sreq, iov, count) \
{ \
	/*MPID_Request * sreq;*/ \
	/*int i;*/ \
	MPIDI_STATE_DECL(MPID_STATE_CREATE_REQUEST); \
	MPIDI_STATE_DECL(MPID_STATE_MEMCPY); \
	MPIDI_FUNC_ENTER(MPID_STATE_CREATE_REQUEST); \
	sreq = MPID_Request_create(); \
	if (sreq == NULL) \
	{ \
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0); \
		MPIDI_FUNC_EXIT(MPID_STATE_CREATE_REQUEST); \
		goto fn_exit; \
	} \
	MPIU_Object_set_ref(sreq, 2); \
	sreq->kind = MPID_REQUEST_SEND; \
	MPIDI_FUNC_ENTER(MPID_STATE_MEMCPY); \
	memcpy(sreq->dev.iov, iov, count * sizeof(MPID_IOV)); \
	MPIDI_FUNC_EXIT(MPID_STATE_MEMCPY); \
	/*for (i = 0; i < count; i++) { sreq->dev.iov[i] = iov[i]; }*/ \
	sreq->ch.pkt = *(MPIDI_CH3_Pkt_t *) iov[0].MPID_IOV_BUF; \
	sreq->dev.iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST) &sreq->ch.pkt; \
	sreq->dev.iov_count = count; \
	sreq->ch.iov_offset = 0; \
	sreq->dev.ca = MPIDI_CH3_CA_COMPLETE; \
	MPIDI_FUNC_EXIT(MPID_STATE_CREATE_REQUEST); \
	/*return sreq;*/ \
}

/*
 * MPIDI_CH3_iStartMsgv() attempts to send the message immediately.  If the entire message is successfully sent, then NULL is
 * returned.  Otherwise a request is allocated, the iovec and the first buffer pointed to by the iovec (which is assumed to be a
 * MPIDI_CH3_Pkt_t) are copied into the request, and a pointer to the request is returned.  An error condition also results in a
 * request be allocated and the errror being returned in the status field of the request.
 */

/* XXX - What do we do if MPID_Request_create() returns NULL???  If MPIDI_CH3_iStartMsgv() returns NULL, the calling code
   assumes the request completely successfully, but the reality is that we couldn't allocate the memory for a request.  This
   seems like a flaw in the CH3 API. */

/* NOTE - The completion action associated with a request created by CH3_iStartMsgv() is alway MPIDI_CH3_CA_COMPLETE.  This
   implies that CH3_iStartMsgv() can only be used when the entire message can be described by a single iovec of size
   MPID_IOV_LIMIT. */
	
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_iStartMsgv
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int ssmMPIDI_CH3_iStartMsgv(MPIDI_VC_t * vc, MPID_IOV * iov, int n_iov, MPID_Request **sreq_ptr)
{
	int mpi_errno = MPI_SUCCESS;
	MPID_Request * sreq = NULL;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_ISTARTMSGV);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_ISTARTMSGV);

	MPIDI_DBG_PRINTF((50, FCNAME, L"entering"));
#ifdef MPICH_DBG_OUTPUT
	if (n_iov > MPID_IOV_LIMIT)
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**arg", 0);
		goto fn_exit;
	}
	if (iov[0].MPID_IOV_LEN > (MPID_IOV_LEN_CAST)sizeof(MPIDI_CH3_Pkt_t))
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**arg", 0);
		goto fn_exit;
	}
#endif

	/* The SOCK channel uses a fixed length header, the size of which is the maximum of all possible packet headers */
	iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)sizeof(MPIDI_CH3_Pkt_t);
	MPIDI_DBG_Print_packet((MPIDI_CH3_Pkt_t*)iov[0].MPID_IOV_BUF);

	if (vc->ch.state == MPIDI_CH3I_VC_STATE_CONNECTED) /* MT */
	{
		/* Connection already formed.  If send queue is empty attempt to send data, queuing any unsent data. */
		if (MPIDI_CH3I_SendQ_empty(vc)) /* MT */
		{
			int nb;
			MPIU_Size_t snb;

			/* MT - need some signalling to lock down our right to use the channel, thus insuring that the progress engine does also try to write */
			if (vc->ch.bShm)
			{
				mpi_errno = MPIDI_CH3I_SHM_writev(vc, iov, n_iov, &nb);
			}
			else
			{
				mpi_errno = MPIDU_Sock_writev(vc->ch.sock, iov, n_iov, &snb);
				nb = (int)snb;
			}
			if (mpi_errno == MPI_SUCCESS)
			{
				int offset = 0;
	
				MPIDI_DBG_PRINTF((55, FCNAME, L"wrote %d bytes", nb));
				
				while (offset < n_iov)
				{
					if (nb >= (int)iov[offset].MPID_IOV_LEN)
					{
						nb -= (int)iov[offset].MPID_IOV_LEN;
						offset++;
					}
					else
					{
						MPIDI_DBG_PRINTF((55, FCNAME, L"partial write, request enqueued at head"));
						create_request(sreq, iov, n_iov, offset, nb);
						/*sreq = create_request(iov, n_iov, offset, nb);
						if (sreq == NULL)
						{
							mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
							return mpi_errno;
						}
						*/
						MPIDI_CH3I_SendQ_enqueue_head(vc, sreq);
						if (vc->ch.bShm)
						{
							vc->ch.send_active = sreq;
						}
						else
						{
							/*MPIDI_CH3I_SSM_VC_post_write(vc, sreq);*/
							MPIDI_DBG_PRINTF((55, FCNAME, L"posting writev, vc=0x%p, sreq=0x%08x", vc, sreq->handle));
							vc->ch.conn->send_active = sreq;
							mpi_errno = MPIDU_Sock_post_writev(vc->ch.conn->sock, sreq->dev.iov + offset, sreq->dev.iov_count - offset, NULL);
							if (mpi_errno != MPI_SUCCESS)
							{
								mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|sock|postwrite", L"ch3|sock|postwrite %p %p %p", sreq, vc->ch.conn, vc);
							}
						}
						break;
					}
				}

				if (offset == n_iov)
				{
					MPIDI_DBG_PRINTF((55, FCNAME, L"entire write complete"));
				}
			}
			else
			{
				MPIDI_DBG_PRINTF((55, FCNAME, L"ERROR - MPIDI_CH3I_SHM_writev failed"));
				sreq = MPID_Request_create();
				if (sreq == NULL)
				{
					mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
					goto fn_exit;
				}
				sreq->kind = MPID_REQUEST_SEND;
				sreq->cc = 0;
				/* TODO: Create an appropriate error message based on the return value */
				sreq->status.MPI_ERROR = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ssmwrite", 0);
			}
		}
		else
		{
			MPIDI_DBG_PRINTF((55, FCNAME, L"send in progress, request enqueued"));
			create_request_zeros(sreq, iov, n_iov);
			/*
			sreq = create_request(iov, n_iov, 0, 0);
			if (sreq == NULL)
			{
				mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
				return mpi_errno;
			}
			*/
			MPIDI_CH3I_SendQ_enqueue(vc, sreq);
		}
	}
	else if (vc->ch.state == MPIDI_CH3I_VC_STATE_UNCONNECTED)
	{
		MPIDI_DBG_PRINTF((55, FCNAME, L"unconnected.  posting connect and enqueuing request"));
		
		/* Form a new connection */
		/*ssmMPIDI_CH3I_VC_post_connect(vc);*/
		
		/* queue the data so it can be sent after the connection is formed */
		create_request_zeros(sreq, iov, n_iov);
		/*
		sreq = create_request(iov, n_iov, 0, 0);
		if (sreq == NULL)
		{
			mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
			return mpi_errno;
		}
		*/
		MPIDI_CH3I_SendQ_enqueue(vc, sreq);

		mpi_errno = ssmMPIDI_CH3I_VC_post_connect(vc);
		if (mpi_errno != MPI_SUCCESS)
		{
			mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", 0);
			goto fn_exit;
		}
	}
	else if (vc->ch.state != MPIDI_CH3I_VC_STATE_FAILED)
	{
		/* Unable to send data at the moment, so queue it for later */
		MPIDI_DBG_PRINTF((55, FCNAME, L"forming connection, request enqueued"));
		create_request_zeros(sreq, iov, n_iov);
		/*
		sreq = create_request(iov, n_iov, 0, 0);
		if (sreq == NULL)
		{
			mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
			return mpi_errno;
		}
		*/
		MPIDI_CH3I_SendQ_enqueue(vc, sreq);
	}
	else
	{
		/* Connection failed, so allocate a request and return an error. */
		MPIDI_DBG_PRINTF((55, FCNAME, L"ERROR - connection failed"));
		sreq = MPID_Request_create();
		if (sreq == NULL)
		{
			mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
			goto fn_exit;
		}
		sreq->kind = MPID_REQUEST_SEND;
		sreq->cc = 0;
		/* TODO: Create an appropriate error message */
		sreq->status.MPI_ERROR = MPI_ERR_INTERN;
	}

fn_exit:
	*sreq_ptr = sreq;

	MPIDI_DBG_PRINTF((50, FCNAME, L"exiting"));
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISTARTMSGV);
	return mpi_errno;
}


#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_iStartRndvMsgShm
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int MPIDI_CH3_iStartRndvMsgShm(MPIDI_VC_t * vc, MPID_Request * sreq, MPIDI_CH3_Pkt_t * rts_pkt)
{
	int mpi_errno = MPI_SUCCESS;
	MPID_Request * rts_sreq;
#ifdef MPICH_DBG_OUTPUT
	int i;
#endif
#ifdef USE_SHM_RDMA_GET
	MPIDI_CH3_Pkt_t pkt;
#endif
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_ISTARTRNDVMSGSHM);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_ISTARTRNDVMSGSHM);

#ifdef USE_SHM_RDMA_GET
	MPIDI_Pkt_init(&pkt.rts_iov, MPIDI_CH3_PKT_RTS_IOV);
	pkt.rts_iov.sreq = sreq->handle;
	pkt.rts_iov.iov_len = sreq->dev.iov_count;

	sreq->dev.rdma_iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)&pkt;
	sreq->dev.rdma_iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)sizeof(MPIDI_CH3_Pkt_t);
	sreq->dev.rdma_iov[1].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)sreq->dev.iov;
	sreq->dev.rdma_iov[1].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)(sreq->dev.iov_count * sizeof(MPID_IOV));
	/* Save a copy of the rts packet in case it is not completely written by the iStartMsgv call */
	/*sreq->dev.rdma_iov[2].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)rts_pkt;*/
	sreq->ch.pkt = *rts_pkt;
	sreq->dev.rdma_iov[2].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)&sreq->ch.pkt;
	sreq->dev.rdma_iov[2].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)sizeof(MPIDI_CH3_Pkt_t);

#ifdef MPICH_DBG_OUTPUT
	for (i=0; i<sreq->dev.iov_count; i++)
	{
		MPIU_DBG_PRINTF((L"iStartRndvMsg: send buf[%d] = %p, len = %d\n", i, sreq->dev.iov[i].MPID_IOV_BUF, sreq->dev.iov[i].MPID_IOV_LEN));
	}
#endif
	mpi_errno = ssmMPIDI_CH3_iStartMsgv(vc, sreq->dev.rdma_iov, 3, &rts_sreq);
	/* --BEGIN ERROR HANDLING-- */
	if (mpi_errno != MPI_SUCCESS)
	{
		MPIU_Object_set_ref(sreq, 0);
		MPIDI_CH3_Request_destroy(sreq);
		sreq = NULL;
		mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|rtspkt", 0);
		goto fn_exit;
	}
	/* --END ERROR HANDLING-- */
	if (rts_sreq != NULL)
	{
		/* The sender doesn't need to know when the message has been sent.  So release the request immediately */
		MPID_Request_release(rts_sreq);
	}
#else
#ifdef MPICH_DBG_OUTPUT
	for (i=0; i<sreq->dev.iov_count; i++)
	{
		MPIU_DBG_PRINTF((L"iStartRndvMsg: send buf[%d] = %p, len = %d\n", i, sreq->dev.iov[i].MPID_IOV_BUF, sreq->dev.iov[i].MPID_IOV_LEN));
	}
#endif
	mpi_errno = MPIDI_CH3.iStartMsg(vc, rts_pkt, sizeof(*rts_pkt), &rts_sreq);
	/* --BEGIN ERROR HANDLING-- */
	if (mpi_errno != MPI_SUCCESS)
	{
		MPIU_Object_set_ref(sreq, 0);
		MPIDI_CH3_Request_destroy(sreq);
		sreq = NULL;
		mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|rtspkt", 0);
		goto fn_exit;
	}
	/* --END ERROR HANDLING-- */
	if (rts_sreq != NULL)
	{
		/* The sender doesn't need to know when the packet has been sent.  So release the request immediately */
		MPID_Request_release(rts_sreq);
	}
#endif

fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISTARTRNDVMSGSHM);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_iStartRndvMsg
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int ssmMPIDI_CH3_iStartRndvMsg(MPIDI_VC_t * vc, MPID_Request * sreq, MPIDI_CH3_Pkt_t * rts_full_pkt)
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_CH3_Pkt_rndv_req_to_send_t * const rts_pkt = &rts_full_pkt->rndv_req_to_send;
	MPID_Request * rts_sreq;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_ISTARTRNDVMSG);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_ISTARTRNDVMSG);

	if (vc->ch.bShm)
	{
		mpi_errno = MPIDI_CH3_iStartRndvMsgShm(vc, sreq, rts_full_pkt);
		/* --BEGIN ERROR HANDLING-- */
		if (mpi_errno != MPI_SUCCESS)
		{
			mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", 0);
			goto fn_exit;
		}
		/* --END ERROR HANDLING-- */
	}
	else
	{
		mpi_errno = ssmMPIDI_CH3_iStartMsg(vc, rts_pkt, sizeof(*rts_pkt), &rts_sreq);
		/* --BEGIN ERROR HANDLING-- */
		if (mpi_errno != MPI_SUCCESS)
		{
			MPIU_Object_set_ref(sreq, 0);
			MPIDI_CH3_Request_destroy(sreq);
			sreq = NULL;
			mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|rtspkt", 0);
			goto fn_exit;
		}
		/* --END ERROR HANDLING-- */
		if (rts_sreq != NULL)
		{
			/* --BEGIN ERROR HANDLING-- */
			if (rts_sreq->status.MPI_ERROR != MPI_SUCCESS)
			{
				MPIU_Object_set_ref(sreq, 0);
				MPIDI_CH3_Request_destroy(sreq);
				mpi_errno = MPIR_Err_create_code(rts_sreq->status.MPI_ERROR, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|rtspkt", 0);
				MPID_Request_release(rts_sreq);
				goto fn_exit;
			}
			/* --END ERROR HANDLING-- */
			MPID_Request_release(rts_sreq);
		}
	}

fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISTARTRNDVMSG);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_iStartRndvTransferShm
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int MPIDI_CH3_iStartRndvTransferShm(MPIDI_VC_t * vc, MPID_Request * rreq)
{
	int mpi_errno = MPI_SUCCESS;
#ifndef USE_SHM_RDMA_GET
	MPIDI_CH3_Pkt_t pkt;
	MPID_Request *request_ptr;
#endif
#ifdef MPICH_DBG_OUTPUT
	int i;
#endif
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_ISTARTRNDVTRANSFERSHM);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_ISTARTRNDVTRANSFERSHM);

#ifdef USE_SHM_RDMA_GET

#ifdef MPICH_DBG_OUTPUT
	for (i=0; i<rreq->dev.iov_count; i++)
	{
		MPIU_DBG_PRINTF((L"iStartRndvTransfer: recv buf[%d] = %p, len = %d\n", i, rreq->dev.iov[i].MPID_IOV_BUF, rreq->dev.iov[i].MPID_IOV_LEN));
		MPIU_DBG_PRINTF((L"iStartRndvTransfer: send buf[%d] = %p, len = %d\n", i, rreq->dev.rdma_iov[i].MPID_IOV_BUF, rreq->dev.rdma_iov[i].MPID_IOV_LEN));
	}
#endif
	mpi_errno = MPIDI_CH3I_SHM_rdma_readv(vc, rreq);
	/* --BEGIN ERROR HANDLING-- */
	if (mpi_errno != MPI_SUCCESS)
	{
		mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", 0);
		goto fn_exit;
	}
	/* --END ERROR HANDLING-- */

#else

	MPIDI_Pkt_init(&pkt.cts_iov, MPIDI_CH3_PKT_CTS_IOV);
	pkt.cts_iov.sreq = rreq->dev.sender_req_id;
	pkt.cts_iov.rreq = rreq->handle;
	pkt.cts_iov.iov_len = rreq->dev.iov_count;

	rreq->dev.rdma_iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)&pkt;
	rreq->dev.rdma_iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)sizeof(MPIDI_CH3_Pkt_t);
	rreq->dev.rdma_iov[1].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)rreq->dev.iov;
	rreq->dev.rdma_iov[1].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)(rreq->dev.iov_count * sizeof(MPID_IOV));

#ifdef MPICH_DBG_OUTPUT
	for (i=0; i<rreq->dev.iov_count; i++)
	{
		MPIU_DBG_PRINTF((L"iStartRndvTransfer: recv buf[%d] = %p, len = %d\n", i, rreq->dev.iov[i].MPID_IOV_BUF, rreq->dev.iov[i].MPID_IOV_LEN));
	}
#endif
	mpi_errno = MPIDI_CH3.iStartMsgv(vc, rreq->dev.rdma_iov, 2, &request_ptr);
	/* --BEGIN ERROR HANDLING-- */
	if (mpi_errno != MPI_SUCCESS)
	{
		MPIU_Object_set_ref(rreq, 0);
		MPIDI_CH3_Request_destroy(rreq);
		rreq = NULL;
		mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|ctspkt", 0);
		goto fn_exit;
	}
	/* --END ERROR HANDLING-- */
	if (request_ptr != NULL)
	{
		/* The sender doesn't need to know when the message has been sent.  So release the request immediately */
		MPID_Request_release(request_ptr);
	}

#endif

fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISTARTRNDVTRANSFERSHM);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_iStartRndvTransfer
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int ssmMPIDI_CH3_iStartRndvTransfer(MPIDI_VC_t * vc, MPID_Request * rreq)
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_ISTARTRNDVTRANSFER);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_ISTARTRNDVTRANSFER);

	if (vc->ch.bShm)
	{
		mpi_errno = MPIDI_CH3_iStartRndvTransferShm(vc, rreq);
		/* --BEGIN ERROR HANDLING-- */
		if (mpi_errno != MPI_SUCCESS)
		{
			mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", 0);
			goto fn_fail;
		}
		/* --END ERROR HANDLING-- */
	}
	else
	{
		MPID_Request * cts_req;
		MPIDI_CH3_Pkt_t upkt;
		MPIDI_CH3_Pkt_rndv_clr_to_send_t * cts_pkt = &upkt.rndv_clr_to_send;

		MPIDI_DBG_PRINTF((15, FCNAME, L"rndv RTS in the request, sending rndv CTS"));

		MPIDI_Pkt_init(cts_pkt, MPIDI_CH3_PKT_RNDV_CLR_TO_SEND);
		cts_pkt->sender_req_id = rreq->dev.sender_req_id;
		cts_pkt->receiver_req_id = rreq->handle;
		mpi_errno = ssmMPIDI_CH3_iStartMsg(vc, cts_pkt, sizeof(*cts_pkt), &cts_req);
		/* --BEGIN ERROR HANDLING-- */
		if (mpi_errno != MPI_SUCCESS)
		{
			MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, L"**ch3|ctspkt");
		}
		/* --END ERROR HANDLING-- */
		if (cts_req != NULL)
		{
			/* FIXME: Ideally we could specify that a req not be returned.  This would avoid our having to decrement the
					  reference count on a req we don't want/need. */
			/* --BEGIN ERROR HANDLING-- */
			if (cts_req->status.MPI_ERROR != MPI_SUCCESS)
			{
				MPIU_Object_set_ref(rreq, 0);
				MPIDI_CH3_Request_destroy(rreq);
				mpi_errno = MPIR_Err_create_code(cts_req->status.MPI_ERROR, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|ctspkt", 0);
				MPID_Request_release(cts_req);
				goto fn_fail;
			}
			/* --END ERROR HANDLING-- */
			MPID_Request_release(cts_req);
		}
	}

fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_ISTARTRNDVTRANSFER);
	return mpi_errno;
}


/****************************************/
/*									  */
/*   Fixed spin waits progress engine   */
/*									  */
/****************************************/

static int fixed_spinMPIDI_CH3_Progress_wait(MPID_Progress_state *state)
{
	return fixed_spinMPIDI_CH3I_Progress(TRUE, state);
}

static int fixed_spinMPIDI_CH3_Progress_test(void)
{
	return fixed_spinMPIDI_CH3I_Progress(FALSE, NULL);
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_Progress
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int fixed_spinMPIDI_CH3I_Progress(int is_blocking, MPID_Progress_state *state)
{
	int mpi_errno = MPI_SUCCESS;
	int rc;
#ifdef MPICH_DBG_OUTPUT
	int count;
#endif
	int bShmProgressMade;
	MPIDU_Sock_event_t event;
	unsigned completions = MPIDI_CH3I_progress_completion_count;
	MPIDI_CH3I_Shmem_queue_info info;
	int num_bytes;
	MPIDI_VC_t *vc_ptr;
	static int spin_count = 1;
	static int msg_queue_count = 0;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_PROGRESS);
#ifdef USE_SLEEP_YIELD
	MPIDI_STATE_DECL(MPID_STATE_MPIDU_SLEEP_YIELD);
#else
	MPIDI_STATE_DECL(MPID_STATE_MPIDU_YIELD);
#endif

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_PROGRESS);

	MPIU_UNREFERENCED_ARG(state);

#ifdef MPICH_DBG_OUTPUT
	if (is_blocking)
	{
		MPIDI_DBG_PRINTF((50, FCNAME, L"entering, blocking=%s", is_blocking ? L"true" : L"false"));
	}
#endif
	do
	{
		/* make progress on the shared memory queues */

		bShmProgressMade = FALSE;
		if (MPIDI_CH3I_Process.shm_reading_list)
		{
			rc = MPIDI_CH3I_SHM_read_progress(MPIDI_CH3I_Process.shm_reading_list, 0, &vc_ptr, &num_bytes);
			if (rc == MPI_SUCCESS)
			{
				MPIDI_DBG_PRINTF((50, FCNAME, L"MPIDI_CH3I_SHM_read_progress reported %d bytes read", num_bytes));
				mpi_errno = MPIDI_CH3I_Handle_shm_read(vc_ptr, num_bytes);
				if (mpi_errno != MPI_SUCCESS)
				{
					mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**progress", 0);
					goto fn_exit;
				}
				bShmProgressMade = TRUE;
			}
			else
			{
				if (rc != SHM_WAIT_TIMEOUT)
				{
					/*MPIDI_err_printf(L"MPIDI_CH3_Progress", L"MPIDI_CH3I_SHM_read_progress returned error %d\n", rc);*/
					mpi_errno = MPIR_Err_create_code(rc, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**shm_read_progress", 0);
					goto fn_exit;
				}
			}
		}

		if (MPIDI_CH3I_Process.shm_writing_list)
		{
			vc_ptr = MPIDI_CH3I_Process.shm_writing_list;
			while (vc_ptr)
			{
				if (vc_ptr->ch.send_active != NULL)
				{
					rc = MPIDI_CH3I_SHM_write_progress(vc_ptr);
					if (rc == MPI_SUCCESS)
					{
						bShmProgressMade = TRUE;
					}
					else if (rc != SHM_WAIT_TIMEOUT)
					{
						mpi_errno = MPIR_Err_create_code(rc, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**progress", 0);
						goto fn_exit;
					}
				}
				vc_ptr = vc_ptr->ch.shm_next_writer;
			}
		}

#if 0
		if (bShmProgressMade) /* This variable is not determined correctly */
		{
			spin_count = 1;
#ifdef USE_SLEEP_YIELD
			MPIDI_Sleep_yield_count = 0;
#endif
			continue;
		}
#endif

		if (spin_count >= MPIDI_Process.my_pg->ch.nShmWaitSpinCount)
		{
#ifdef USE_SLEEP_YIELD
			if (spin_count >= MPIDI_Process.my_pg->ch.nShmWaitYieldCount)
			{
				MPIDI_FUNC_ENTER(MPID_STATE_MPIDU_SLEEP_YIELD);
				MPIDU_Sleep_yield();
				MPIDI_FUNC_EXIT(MPID_STATE_MPIDU_SLEEP_YIELD);
			}
			else
			{
				MPIDI_FUNC_ENTER(MPID_STATE_MPIDU_YIELD);
				MPIDU_Yield();
				MPIDI_FUNC_EXIT(MPID_STATE_MPIDU_YIELD);
			}
#else
			MPIDI_FUNC_ENTER(MPID_STATE_MPIDU_YIELD);
			MPIDU_Yield();
			MPIDI_FUNC_EXIT(MPID_STATE_MPIDU_YIELD);
			spin_count = 1;
#endif
		}
		/* FIXME: Why do the other yields have MPIDI_FUNC_ENTER/EXIT 
		   macros around them but this one does not? */
		else MPIDU_Yield(); /* always yield for now */
		spin_count++;

		if (spin_count > (MPIDI_Process.my_pg->ch.nShmWaitSpinCount >> 1) )
		{
			/* make progress on the sockets */

			mpi_errno = MPIDU_Sock_wait(MPIDI_CH3I_sock_set, 0, &event);
			if (mpi_errno == MPI_SUCCESS)
			{
				mpi_errno = ssmMPIDI_CH3I_Progress_handle_sock_event(&event);
				if (mpi_errno != MPI_SUCCESS)
				{
					mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**progress_handle_sock_op", 0);
					goto fn_exit;
				}
			}
			else
			{
				if (MPIR_ERR_GET_CLASS(mpi_errno) != MPIDU_SOCK_ERR_TIMEOUT)
				{
					mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**progress_sock_wait", 0);
					goto fn_exit;
				}
				mpi_errno = MPI_SUCCESS;
				/* FIXME: Why do the other yields have MPIDI_FUNC_ENTER/EXIT 
				   macros around them but this one does not? */
				MPIDU_Yield();
			}
		}

		if (((msg_queue_count++ % MPIDI_CH3I_MSGQ_ITERATIONS) == 0) || !is_blocking)
		{
			/* check for new shmem queue connection requests */
			rc = MPIDI_CH3I_BootstrapQ_recv_msg(MPIDI_Process.my_pg->ch.bootstrapQ, &info, sizeof(info), &num_bytes, FALSE);
			if (rc != MPI_SUCCESS)
			{
				mpi_errno = MPIR_Err_create_code(rc, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**boot_recv", 0);
				goto fn_exit;
			}
#ifdef MPICH_DBG_OUTPUT
			/*MPIU_Assert(num_bytes == 0 || num_bytes == sizeof(info));*/
			if (num_bytes != 0 && num_bytes != sizeof(info))
			{
				mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**bootqmsg", L"**bootqmsg %d", num_bytes);
				goto fn_exit;
			}
#endif
			if (num_bytes)
			{
				MPIDI_PG_t *pg;

				MPIDI_PG_Find(info.pg_id, &pg);
				MPIDI_PG_Get_vc(pg, info.pg_rank, &vc_ptr);
				/*
				wprintf(L"attaching to shared memory queue:\nVC.rank %d\nVC.pg_id <%s>\nPG.id <%s>\n",
					vc_ptr->pg_rank, vc_ptr->pg->id, pg->id);
				fflush(stdout);
				*/
				/*vc_ptr = &MPIDI_Process.my_pg->ch.vc_table[info.pg_rank];*/
				rc = MPIDI_CH3I_SHM_Attach_to_memblock(&info.info, &vc_ptr->ch.shm_read_queue_info);
				if (rc != MPI_SUCCESS)
				{
					mpi_errno = MPIR_Err_create_code(rc, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**attach_to_mem", L"**attach_to_mem %d", vc_ptr->ch.shm_read_queue_info.error);
					goto fn_exit;
				}
				MPIU_DBG_PRINTF((L"attached to queue from process %d\n", info.pg_rank));
				mpi_errno = MPIDI_SHM_InitRWProc( info.pid, &vc_ptr->ch.hSharedProcessHandle );
				if (mpi_errno)
				{
					mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**init_rwproc", L"**init_rwproc %d", info.pg_rank);
					return mpi_errno;
				}
				/*vc_ptr->ch.state = MPIDI_CH3I_VC_STATE_CONNECTED;*/ /* we are read connected but not write connected */
				vc_ptr->ch.shm_read_connected = 1;
				vc_ptr->ch.bShm = TRUE;
				vc_ptr->ch.read_shmq = vc_ptr->ch.shm_read_queue_info.addr;/*info.info.addr;*/
				MPIU_DBG_PRINTF((L"read_shmq = %p\n", vc_ptr->ch.read_shmq));
				vc_ptr->ch.shm_reading_pkt = TRUE;
				/* add this VC to the global list to be shm_waited on */
				MPIDI_CH3I_SHM_Add_to_reader_list(vc_ptr);
			}
		}
	}
	while (completions == MPIDI_CH3I_progress_completion_count && is_blocking);

fn_exit:
#ifdef MPICH_DBG_OUTPUT
	count = MPIDI_CH3I_progress_completion_count - completions;
	if (is_blocking)
	{
		MPIDI_DBG_PRINTF((50, FCNAME, L"exiting, count=%d", count));
	}
	else
	{
		if (count > 0)
		{
			MPIDI_DBG_PRINTF((50, FCNAME, L"exiting (non-blocking), count=%d", count));
		}
	}
#endif
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_PROGRESS);
	return mpi_errno;
}


#ifdef USE_ADAPTIVE_PROGRESS_ENGINE
/********************************/
/*								*/
/*   Adaptive progress engine   */
/*								*/
/********************************/

#define MPIDI_CH3I_UPDATE_ITERATIONS	10
#define MPID_SINGLE_ACTIVE_FACTOR	  100

#undef FUNCNAME
#define FUNCNAME adaptiveMPIDI_CH3I_Message_queue_progress
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int adaptiveMPIDI_CH3I_Message_queue_progress()
{
	MPIDI_CH3I_Shmem_queue_info info;
	int num_bytes;
	MPIDI_VC_t *vc_ptr;
	int mpi_errno;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_MESSAGE_QUEUE_PROGRESS);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_MESSAGE_QUEUE_PROGRESS);

	/* check for new shmem queue connection requests */
	/*wprintf(L"<%dR>", MPIR_Process.comm_world->rank);fflush(stdout);*/
	mpi_errno = MPIDI_CH3I_BootstrapQ_recv_msg(
		MPIDI_Process.my_pg->ch.bootstrapQ, &info, sizeof(info), 
		&num_bytes, FALSE);
	if (mpi_errno != MPI_SUCCESS)
	{
		mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**boot_recv", 0);
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_MESSAGE_QUEUE_PROGRESS);
		return mpi_errno;
	}
#ifdef MPICH_DBG_OUTPUT
	/*MPIU_Assert(num_bytes == 0 || num_bytes == sizeof(info));*/
	if (num_bytes != 0 && num_bytes != sizeof(info))
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**bootqmsg", L"**bootqmsg %d", num_bytes);
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_MESSAGE_QUEUE_PROGRESS);
		return mpi_errno;
	}
#endif
	if (num_bytes)
	{
		MPIDI_PG_t *pg;

		MPIDI_PG_Find(info.pg_id, &pg);
		MPIDI_PG_Get_vc(pg, info.pg_rank, &vc_ptr);
		/*vc_ptr = &MPIDI_Process.my_pg->ch.vc_table[info.pg_rank];*/
		/*wprintf(L"rank %d attaching to shm(%s) from rank %d\n", MPIR_Process.comm_world->rank, info.info.name, info.pg_rank);fflush(stdout);*/
		mpi_errno = MPIDI_CH3I_SHM_Attach_to_memblock(
			&info.info, &vc_ptr->ch.shm_read_queue_info);
		if (mpi_errno != MPI_SUCCESS)
		{
			mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**MPIDI_CH3I_SHM_Attach_to_memblock", L"**MPIDI_CH3I_SHM_Attach_to_memblock %d", vc_ptr->ch.shm_read_queue_info.error);
			MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_MESSAGE_QUEUE_PROGRESS);
			return mpi_errno;
		}
		MPIU_DBG_PRINTF((L"attached to queue from process %d\n", info.pg_rank));
		mpi_errno = MPIDI_SHM_InitRWProc( info.pid, &vc_ptr->ch.hSharedProcessHandle );
		if (mpi_errno)
		{
			mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**init_rwproc", L"**init_rwproc %d", info.pg_rank);
			return mpi_errno;
		}
		/*vc_ptr->ch.state = MPIDI_CH3I_VC_STATE_CONNECTED;*/ /* we are read connected but not write connected */
		vc_ptr->ch.shm_read_connected = 1;
		vc_ptr->ch.bShm = TRUE;
		vc_ptr->ch.read_shmq = vc_ptr->ch.shm_read_queue_info.addr;
		MPIU_DBG_PRINTF((L"read_shmq = %p\n", vc_ptr->ch.read_shmq));
		vc_ptr->ch.shm_reading_pkt = TRUE;
		/* add this VC to the global list to be shm_waited on */
		/*wprintf(L"vc added to reading list.\n");fflush(stdout);*/
		MPIDI_CH3I_SHM_Add_to_reader_list(vc_ptr);
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_MESSAGE_QUEUE_PROGRESS);
	return MPI_SUCCESS;
}

#undef FUNCNAME
#define FUNCNAME adaptiveMPIDI_CH3_Progress_test
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int adaptiveMPIDI_CH3_Progress_test(void)
{
	/* This function has a problem that is #if 0'd out.
	 * The commented out code causes test to only probe the message queue for connection attempts
	 * every MPIDI_CH3I_MSGQ_ITERATIONS iterations.  This can delay shm connection formation for 
	 * codes that call test infrequently.
	 * But the uncommented code also has the problem that the overhead of checking the message queue
	 * is incurred with every call to test.
	 */
	int mpi_errno = MPI_SUCCESS;
	int rc;
	MPIDU_Sock_event_t event;
	int num_bytes;
	MPIDI_VC_t *vc_ptr;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_PROGRESS_TEST);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_PROGRESS_TEST);

	if (MPIDI_CH3I_Process.shm_reading_list)
	{
		rc = MPIDI_CH3I_SHM_read_progress( MPIDI_CH3I_Process.shm_reading_list, 0, &vc_ptr, &num_bytes );
		if (rc == MPI_SUCCESS)
		{
			MPIDI_DBG_PRINTF((50, FCNAME, L"MPIDI_CH3I_SHM_read_progress reported %d bytes read", num_bytes));
			mpi_errno = MPIDI_CH3I_Handle_shm_read(vc_ptr, num_bytes);
			if (mpi_errno != MPI_SUCCESS)
			{
				mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**progress_test", 0);
				goto fn_exit;
			}
		}
	}

	if (MPIDI_CH3I_Process.shm_writing_list)
	{
		vc_ptr = MPIDI_CH3I_Process.shm_writing_list;
		while (vc_ptr)
		{
			if (vc_ptr->ch.send_active != NULL)
			{
				rc = MPIDI_CH3I_SHM_write_progress(vc_ptr);
				if (rc != MPI_SUCCESS && rc != SHM_WAIT_TIMEOUT)
				{
					mpi_errno = MPIR_Err_create_code(rc, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**progress", 0);
					goto fn_exit;
				}
			}
			vc_ptr = vc_ptr->ch.shm_next_writer;
		}
	}

	mpi_errno = MPIDU_Sock_wait(MPIDI_CH3I_sock_set, 0, &event);
	if (mpi_errno == MPI_SUCCESS)
	{
		mpi_errno = ssmMPIDI_CH3I_Progress_handle_sock_event(&event);
		if (mpi_errno != MPI_SUCCESS)
		{
			mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**handle_sock_op", 0);
			return mpi_errno;
		}
		/*active = active | MPID_CH3I_SOCK_BIT;*/
		MPIDI_CH3I_active_flag |= MPID_CH3I_SOCK_BIT;
	}
	else
	{
		if (MPIR_ERR_GET_CLASS(mpi_errno) != MPIDU_SOCK_ERR_TIMEOUT)
		{
			mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**test_sock_wait", 0);
			goto fn_exit;
		}
		mpi_errno = MPI_SUCCESS;
	}

	mpi_errno = adaptiveMPIDI_CH3I_Message_queue_progress();
	if (mpi_errno != MPI_SUCCESS)
	{
		mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**mqp_failure", 0);
	}

fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_PROGRESS_TEST);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_Progress_wait
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int adaptiveMPIDI_CH3_Progress_wait(MPID_Progress_state *state)
{
	int mpi_errno = MPI_SUCCESS;
	int rc;
	MPIDU_Sock_event_t event;
	unsigned completions = MPIDI_CH3I_progress_completion_count;
	int num_bytes;
	MPIDI_VC_t *vc_ptr;

	/*static int active = 0;*/
	static int updateIter = 0;
	static int msgqIter = MPIDI_CH3I_MSGQ_ITERATIONS;
	MPID_CPU_Tick_t start, end;

	static MPID_CPU_Tick_t shmTicks = 0;
	static int shmIter = 0;
	static int shmReps = 1;
	static int shmTotalReps = 0;
	static int spin_count = 1;

	static MPID_CPU_Tick_t sockTicks = 0;
	static int sockIter = 0;
	static int sockReps = 1;
	static int sockTotalReps = 0;

	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_PROGRESS_WAIT);
	/*MPIDI_STATE_DECL(MPID_STATE_MPIDU_YIELD);*/

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_PROGRESS_WAIT);

	MPIU_UNREFERENCED_ARG(state);

	MPIDI_DBG_PRINTF((50, FCNAME, L"entering"));
	do
	{
		/* make progress on the shared memory queues */

		if (sockIter != 0)
			goto skip_shm_loop;

		MPID_CPU_TICK(&start);
		for (; shmIter<shmReps; shmIter++, shmTotalReps++)
		{
			if (MPIDI_CH3I_Process.shm_reading_list)
			{
				rc = MPIDI_CH3I_SHM_read_progress( MPIDI_CH3I_Process.shm_reading_list, 0, &vc_ptr, &num_bytes );
				if (rc == MPI_SUCCESS)
				{
					MPIDI_DBG_PRINTF((50, FCNAME, L"MPIDI_CH3I_SHM_read_progress reported %d bytes read", num_bytes));
					mpi_errno = MPIDI_CH3I_Handle_shm_read(vc_ptr, num_bytes);
					if (mpi_errno != MPI_SUCCESS)
					{
						mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**progress", 0);
						goto fn_exit;
					}
				}
				else
				{
					if (rc != SHM_WAIT_TIMEOUT)
					{
						/*MPIDI_err_printf(L"MPIDI_CH3_Progress", L"MPIDI_CH3I_SHM_read_progress returned error %d\n", rc);*/
						mpi_errno = MPIR_Err_create_code(rc, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**progress", 0);
						goto fn_exit;
					}
					/*
					if (rc == SHM_WAIT_TIMEOUT_WHILE_ACTIVE)
					{
						active = active | MPID_CH3I_SHM_BIT;
					}
					*/
				}
				if (MPIDI_CH3I_shm_read_active)
					spin_count = 1;
				if (completions != MPIDI_CH3I_progress_completion_count)
				{
					MPIDI_CH3I_active_flag |= MPID_CH3I_SHM_BIT;
					/*
					active = active | MPID_CH3I_SHM_BIT;
					spin_count = 1;
					*/
					shmIter++;
					shmTotalReps++;
					goto after_shm_loop;
				}
			}

			if (MPIDI_CH3I_Process.shm_writing_list)
			{
				vc_ptr = MPIDI_CH3I_Process.shm_writing_list;
				while (vc_ptr)
				{
					if (vc_ptr->ch.send_active != NULL)
					{
						rc = MPIDI_CH3I_SHM_write_progress(vc_ptr);
						if (rc == MPI_SUCCESS)
						{
							/*active = active | MPID_CH3I_SHM_BIT;*/
							if (completions != MPIDI_CH3I_progress_completion_count)
							{
								MPIDI_CH3I_active_flag |= MPID_CH3I_SHM_BIT;
								spin_count = 1;
								shmIter++;
								shmTotalReps++;
								goto after_shm_loop;
							}
						}
						else if (rc != SHM_WAIT_TIMEOUT)
						{
							mpi_errno = MPIR_Err_create_code(rc, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**progress", 0);
							goto fn_exit;
						}
					}
					vc_ptr = vc_ptr->ch.shm_next_writer;
				}
			}
/*
			if (spin_count++ >= MPIDI_Process.my_pg->ch.nShmWaitSpinCount)
			{
				MPIDI_FUNC_ENTER(MPID_STATE_MPIDU_YIELD);
				MPIDU_Yield();
				MPIDI_FUNC_EXIT(MPID_STATE_MPIDU_YIELD);
				spin_count = 1;
			}
*/
			/* FIXME: Why do the other yields have MPIDI_FUNC_ENTER/EXIT macros around them but this one does not? */
			MPIDU_Yield();
		}
after_shm_loop:
		if (shmIter == shmReps)
		{
			shmIter = 0;
		}
		MPID_CPU_TICK(&end);
		shmTicks += end - start;

		if (shmIter != 0)
			goto skip_sock_loop;

skip_shm_loop:
		MPID_CPU_TICK(&start);
		for (; sockIter<sockReps; sockIter++, sockTotalReps++)
		{
			/* make progress on the sockets */

			mpi_errno = MPIDU_Sock_wait(MPIDI_CH3I_sock_set, 0, &event);
			if (mpi_errno == MPI_SUCCESS)
			{
				mpi_errno = ssmMPIDI_CH3I_Progress_handle_sock_event(&event);
				if (mpi_errno != MPI_SUCCESS)
				{
					mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**progress_sock_wait", 0);
					goto fn_exit;
				}
				/*active = active | MPID_CH3I_SOCK_BIT;*/
				MPIDI_CH3I_active_flag |= MPID_CH3I_SOCK_BIT;
			}
			else
			{
				if (MPIR_ERR_GET_CLASS(mpi_errno) != MPIDU_SOCK_ERR_TIMEOUT)
				{
					mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**progress_sock_wait", 0);
					goto fn_exit;
				}
				/* comment out this line to test the error functions */
				mpi_errno = MPI_SUCCESS;
				/* FIXME: Why do the other yields have MPIDI_FUNC_ENTER/EXIT macros around them but this one does not? */
				MPIDU_Yield();
			}
			if (completions != MPIDI_CH3I_progress_completion_count)
			{
				sockIter++;
				sockTotalReps++;
				goto after_sock_loop;
			}
		}
after_sock_loop:
		if (sockIter == sockReps)
		{ 
			sockIter = 0;
			updateIter++;
		}
		MPID_CPU_TICK(&end);
		sockTicks += end - start;

skip_sock_loop:
		if (updateIter == MPIDI_CH3I_UPDATE_ITERATIONS)
		{
			updateIter = 0;
			switch (MPIDI_CH3I_active_flag)
			{
			case MPID_CH3I_SHM_BIT:
				/* only shared memory has been active */
				/* give shm MPID_SINGLE_ACTIVE_FACTOR cycles for every 1 sock cycle */
				shmReps = (int) (
					( sockTicks * (MPID_CPU_Tick_t)MPID_SINGLE_ACTIVE_FACTOR * (MPID_CPU_Tick_t)shmTotalReps ) 
					/ (MPID_CPU_Tick_t)sockTotalReps 
					/ shmTicks);
				if (shmReps < 1)
					shmReps = 1;
				if (shmReps > 256)
				{
					shmReps = 256;
				}
				sockReps = 1;
				/*MPIU_DBG_PRINTF((L"(SHM_BIT: shmReps = %d, sockReps = %d)", shmReps, sockReps));*/
				break;
			case MPID_CH3I_SOCK_BIT:
				/* only sockets have been active */
				/* give sock MPID_SINGLE_ACTIVE_FACTOR cycles for every 1 shm cycle */
				sockReps = (int) ( 
					( shmTicks * (MPID_CPU_Tick_t)MPID_SINGLE_ACTIVE_FACTOR * (MPID_CPU_Tick_t)sockTotalReps )
					/ (MPID_CPU_Tick_t)shmTotalReps
					/ sockTicks );
				if (sockReps < 1)
					sockReps = 1;
				if (sockReps > 100)
				{
					sockReps = 100;
				}
				shmReps = 1;
				/*MPIU_DBG_PRINTF((L"(SOCK_BIT: shmReps = %d, sockReps = %d)", shmReps, sockReps));*/
				break;
			case MPID_CH3I_SHM_BIT + MPID_CH3I_SOCK_BIT:
				/* both channels have been active */
				/* give each channel 50% of the spin cycles */
				shmReps = (int)
					(
					 (
					  (MPID_CPU_Tick_t)((shmTicks + sockTicks) >> 1)
					  * (MPID_CPU_Tick_t)shmTotalReps
					 ) / shmTicks
					) / MPIDI_CH3I_UPDATE_ITERATIONS;
				if (shmReps < 1)
					shmReps = 1;
				if (shmReps > 256)
				{
					shmReps = 256;
				}
				sockReps = (int)
					(
					 (
					  (MPID_CPU_Tick_t)((shmTicks + sockTicks) >> 1)
					  * (MPID_CPU_Tick_t)sockTotalReps
					 ) / sockTicks
					) / MPIDI_CH3I_UPDATE_ITERATIONS;
				if (sockReps < 1)
					sockReps = 1;
				if (sockReps > 100)
				{
					sockReps = 100;
				}
				/*MPIU_DBG_PRINTF((L"(BOTH_BITS: shmReps = %d, sockReps = %d)", shmReps, sockReps));*/
				break;
			default:
				/* neither channel has been active */
				/* leave the state in its previous state */
				/*wprintf(L".");*/
				break;
			}
			/*
			if (shmReps > 500 || sockReps > 100)
			{
				wprintf(L"[%d] SHMREPS=%d, SOCKREPS=%d\n", MPIR_Process.comm_world->rank, shmReps, sockReps);
			}
			*/
			MPIDI_CH3I_active_flag = 0;
			shmTotalReps = 0;
			shmTicks = 0;
			shmIter = 0;
			sockTotalReps = 0;
			sockTicks = 0;
			sockIter = 0;
		}

		if (msgqIter++ == MPIDI_CH3I_MSGQ_ITERATIONS)
		{
			msgqIter = 0;
			/*wprintf(L"[%d] calling message queue progress\n", MPIR_Process.comm_world->rank);fflush(stdout);*/
			/* wprintf(L"%d",2*MPIR_Process.comm_world->rank); */
			mpi_errno = adaptiveMPIDI_CH3I_Message_queue_progress();
			if (mpi_errno != MPI_SUCCESS)
			{
				mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER,
												 L"**mqp_failure", 0);
				goto fn_exit;
			}
			/* wprintf(L"%d",2*MPIR_Process.comm_world->rank+1); */
		}
	}
	while (completions == MPIDI_CH3I_progress_completion_count);

fn_exit:
	MPIDI_DBG_PRINTF((50, FCNAME, L"exiting, count=%d", MPIDI_CH3I_progress_completion_count - completions));
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_PROGRESS_WAIT);
	return mpi_errno;
}

#endif

#ifdef USE_FIXED_ACTIVE_PROGRESS_ENGINE

/**********************************************************/
/*														  */
/*   Fixed active progress engine						  */
/*														  */
/*   Like the adaptive engine but active reps are fixed   */
/*   instead of calculated from CPU time.				  */
/*														  */
/**********************************************************/

#define MPIDI_CH3I_UPDATE_ITERATIONS	10
#define MPID_SINGLE_ACTIVE_FACTOR	  100

#undef FUNCNAME
#define FUNCNAME fixed_activeMPIDI_CH3I_Message_queue_progress
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int fixed_activeMPIDI_CH3I_Message_queue_progress()
{
	MPIDI_CH3I_Shmem_queue_info info;
	int num_bytes;
	MPIDI_VC_t *vc_ptr;
	int rc, mpi_errno;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_MESSAGE_QUEUE_PROGRESS);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_MESSAGE_QUEUE_PROGRESS);

	/* check for new shmem queue connection requests */
	rc = MPIDI_CH3I_BootstrapQ_recv_msg(
		MPIDI_Process.my_pg->ch.bootstrapQ, &info, sizeof(info), 
		&num_bytes, FALSE);
	if (rc != MPI_SUCCESS)
	{
		mpi_errno = MPIR_Err_create_code(rc, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**boot_recv", 0);
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_MESSAGE_QUEUE_PROGRESS);
		return mpi_errno;
	}
#ifdef MPICH_DBG_OUTPUT
	/*MPIU_Assert(num_bytes == 0 || num_bytes == sizeof(info));*/
	if (num_bytes != 0 && num_bytes != sizeof(info))
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**bootqmsg", L"**bootqmsg %d", num_bytes);
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_MESSAGE_QUEUE_PROGRESS);
		return mpi_errno;
	}
#endif
	if (num_bytes)
	{
		MPIDI_PG_t *pg;

		MPIDI_PG_Find(info.pg_id, &pg);
		MPIDI_PG_Get_vc(pg, info.pg_rank, &vc_ptr);
		/*vc_ptr = &MPIDI_Process.my_pg->ch.vc_table[info.pg_rank];*/
		rc = MPIDI_CH3I_SHM_Attach_to_memblock(
			&info.info, &vc_ptr->ch.shm_read_queue_info);
		if (rc != MPI_SUCCESS)
		{
			mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**MPIDI_CH3I_SHM_Attach_to_memblock", L"**MPIDI_CH3I_SHM_Attach_to_memblock %d", vc_ptr->ch.shm_read_queue_info.error); /*L"MPIDI_CH3I_SHM_Attach_to_memblock failed, error %d\n", vc_ptr->ch.shm_read_queue_info.error);*/
			MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_MESSAGE_QUEUE_PROGRESS);
			return mpi_errno;
		}
		MPIU_DBG_PRINTF((L"attached to queue from process %d\n", info.pg_rank));
		mpi_errno = MPIDI_SHM_InitRWProc( info.pid, &vc_ptr->ch.hSharedProcessHandle );
		if (mpi_errno)
		{
			mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**init_rwproc", L"**init_rwproc %d", info.pg_rank);
			return mpi_errno;
		}
		/*vc_ptr->ch.state = MPIDI_CH3I_VC_STATE_CONNECTED;*/ /* we are read connected but not write connected */
		vc_ptr->ch.shm_read_connected = 1;
		vc_ptr->ch.bShm = TRUE;
		vc_ptr->ch.read_shmq = vc_ptr->ch.shm_read_queue_info.addr;
		MPIU_DBG_PRINTF((L"read_shmq = %p\n", vc_ptr->ch.read_shmq));
		vc_ptr->ch.shm_reading_pkt = TRUE;
		/* add this VC to the global list to be shm_waited on */
		MPIDI_CH3I_SHM_Add_to_reader_list(vc_ptr);
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_MESSAGE_QUEUE_PROGRESS);
	return MPI_SUCCESS;
}


static int fixed_activeMPIDI_CH3_Progress_wait(MPID_Progress_state *state)
{
	return fixed_activeMPIDI_CH3I_Progress(TRUE, state);
}

static int fixed_activeMPIDI_CH3_Progress_test(void)
{
	return fixed_activeMPIDI_CH3I_Progress(FALSE, NULL);
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_Progress
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int fixed_activeMPIDI_CH3I_Progress(int is_blocking, MPID_Progress_state *state)
{
	int mpi_errno = MPI_SUCCESS;
	int rc;
	MPIDU_Sock_event_t event;
	unsigned completions = MPIDI_CH3I_progress_completion_count;
	int num_bytes;
	MPIDI_VC_t *vc_ptr;

	/*static int active = 0;*/
	static int updateIter = 0;
	static int msgqIter = 0;
	MPID_CPU_Tick_t start, end;

	static MPID_CPU_Tick_t shmTicks = 0;
	static int shmIter = 0;
	static int shmReps = 1;
	static int shmTotalReps = 0;
	static int spin_count = 1;

	static MPID_CPU_Tick_t sockTicks = 0;
	static int sockIter = 0;
	static int sockReps = 1;
	static int sockTotalReps = 0;

	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_PROGRESS);
	/*MPIDI_STATE_DECL(MPID_STATE_MPIDU_YIELD);*/

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_PROGRESS);

	MPIU_UNREFERENCED_ARG(state);

#ifdef MPICH_DBG_OUTPUT
	if (is_blocking)
	{
		MPIDI_DBG_PRINTF((50, FCNAME, L"entering, blocking=%s", is_blocking ? L"true" : L"false"));
	}
#endif
	do
	{
		/* make progress on the shared memory queues */

		if (sockIter != 0)
			goto skip_shm_loop;

		MPID_CPU_TICK(&start);
		for (; shmIter<shmReps; shmIter++, shmTotalReps++)
		{
			if (MPIDI_CH3I_Process.shm_reading_list)
			{
				rc = MPIDI_CH3I_SHM_read_progress(
					MPIDI_CH3I_Process.shm_reading_list,
					0, &vc_ptr, &num_bytes);
				if (rc == MPI_SUCCESS)
				{
					MPIDI_DBG_PRINTF((50, FCNAME, L"MPIDI_CH3I_SHM_read_progress reported %d bytes read", num_bytes));
					mpi_errno = MPIDI_CH3I_Handle_shm_read(vc_ptr, num_bytes);
					if (mpi_errno != MPI_SUCCESS)
					{
						mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**progress", 0);
						goto fn_exit;
					}
				}
				else
				{
					if (rc != SHM_WAIT_TIMEOUT)
					{
						MPIDI_err_printf(L"MPIDI_CH3_Progress", L"MPIDI_CH3I_SHM_read_progress returned error %d\n", rc);
						mpi_errno = MPIR_Err_create_code(rc, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**progress", 0);
						goto fn_exit;
					}
					/*
					if (rc == SHM_WAIT_TIMEOUT_WHILE_ACTIVE)
					{
						active = active | MPID_CH3I_SHM_BIT;
					}
					*/
				}
				if (MPIDI_CH3I_shm_read_active)
					spin_count = 1;
				if (completions != MPIDI_CH3I_progress_completion_count)
				{
					MPIDI_CH3I_active_flag |= MPID_CH3I_SHM_BIT;
					/*
					active = active | MPID_CH3I_SHM_BIT;
					spin_count = 1;
					*/
					shmIter++;
					shmTotalReps++;
					goto after_shm_loop;
				}
			}

			if (MPIDI_CH3I_Process.shm_writing_list)
			{
				vc_ptr = MPIDI_CH3I_Process.shm_writing_list;
				while (vc_ptr)
				{
					if (vc_ptr->ch.send_active != NULL)
					{
						rc = MPIDI_CH3I_SHM_write_progress(vc_ptr);
						if (rc == MPI_SUCCESS)
						{
							/*active = active | MPID_CH3I_SHM_BIT;*/
							if (completions != MPIDI_CH3I_progress_completion_count)
							{
								MPIDI_CH3I_active_flag |= MPID_CH3I_SHM_BIT;
								spin_count = 1;
								shmIter++;
								shmTotalReps++;
								goto after_shm_loop;
							}
						}
						else if (rc != SHM_WAIT_TIMEOUT)
						{
							mpi_errno = MPIR_Err_create_code(rc, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**progress", 0);
							goto fn_exit;
						}
					}
					vc_ptr = vc_ptr->ch.shm_next_writer;
				}
			}
			/*
			if (spin_count++ >= MPIDI_Process.my_pg->ch.nShmWaitSpinCount)
			{
				MPIDI_FUNC_ENTER(MPID_STATE_MPIDU_YIELD);
				MPIDU_Yield();
				MPIDI_FUNC_EXIT(MPID_STATE_MPIDU_YIELD);
				spin_count = 1;
			}
			*/
			/* FIXME: Why do the other yields have MPIDI_FUNC_ENTER/EXIT 
			   macros around them but this one does not? */
			MPIDU_Yield();
		}
after_shm_loop:
		if (shmIter == shmReps)
		{
			shmIter = 0;
		}
		MPID_CPU_TICK(&end);
		shmTicks += end - start;

		if (shmIter != 0)
			goto skip_sock_loop;

skip_shm_loop:
		MPID_CPU_TICK(&start);
		for (; sockIter<sockReps; sockIter++, sockTotalReps++)
		{
			/* make progress on the sockets */

			mpi_errno = MPIDU_Sock_wait(MPIDI_CH3I_sock_set, 0, &event);
			if (mpi_errno == MPI_SUCCESS)
			{
				mpi_errno = ssmMPIDI_CH3I_Progress_handle_sock_event(&event);
				if (mpi_errno != MPI_SUCCESS)
				{
					mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**handle_sock_op", 0);
					goto fn_exit;
				}
				/*active = active | MPID_CH3I_SOCK_BIT;*/
				MPIDI_CH3I_active_flag |= MPID_CH3I_SOCK_BIT;
			}
			else
			{
				if (MPIR_ERR_GET_CLASS(mpi_errno) != MPIDU_SOCK_ERR_TIMEOUT)
				{
					mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**handle_sock_op", 0);
					goto fn_exit;
				}
				mpi_errno = MPI_SUCCESS;
				/* FIXME: Why do the other yields have MPIDI_FUNC_ENTER/EXIT 
				   macros around them but this one does not? */
				MPIDU_Yield();
			}
			if (completions != MPIDI_CH3I_progress_completion_count)
			{
				sockIter++;
				sockTotalReps++;
				goto after_sock_loop;
			}
		}
after_sock_loop:
		if (sockIter == sockReps)
		{ 
			sockIter = 0;
			updateIter++;
		}
		MPID_CPU_TICK(&end);
		sockTicks += end - start;

skip_sock_loop:
		if (updateIter == MPIDI_CH3I_UPDATE_ITERATIONS)
		{
			updateIter = 0;
			switch (MPIDI_CH3I_active_flag)
			{
			case MPID_CH3I_SHM_BIT:
				/* only shared memory has been active */
				/* give shm MPID_SINGLE_ACTIVE_FACTOR cycles for every 1 sock cycle */
				shmReps = (int) (
					( sockTicks * (MPID_CPU_Tick_t)MPID_SINGLE_ACTIVE_FACTOR * (MPID_CPU_Tick_t)shmTotalReps ) 
					/ (MPID_CPU_Tick_t)sockTotalReps 
					/ shmTicks);
				if (shmReps < 1)
					shmReps = 1;
				sockReps = 1;
				/*MPIU_DBG_PRINTF((L"(SHM_BIT: shmReps = %d, sockReps = %d)", shmReps, sockReps));*/
				break;
			case MPID_CH3I_SOCK_BIT:
				/* only sockets have been active */
				/* give sock MPID_SINGLE_ACTIVE_FACTOR cycles for every 1 shm cycle */
				sockReps = (int) ( 
					( shmTicks * (MPID_CPU_Tick_t)MPID_SINGLE_ACTIVE_FACTOR * (MPID_CPU_Tick_t)sockTotalReps )
					/ (MPID_CPU_Tick_t)shmTotalReps
					/ sockTicks );
				if (sockReps < 1)
					sockReps = 1;
				shmReps = 1;
				/*MPIU_DBG_PRINTF((L"(SOCK_BIT: shmReps = %d, sockReps = %d)", shmReps, sockReps));*/
				break;
			case MPID_CH3I_SHM_BIT + MPID_CH3I_SOCK_BIT:
				/* both channels have been active */
				/* give each channel 50% of the spin cycles */
				shmReps = (int)
					(
					 (
					  (MPID_CPU_Tick_t)((shmTicks + sockTicks) >> 1)
					  * (MPID_CPU_Tick_t)shmTotalReps
					 ) / shmTicks
					) / MPIDI_CH3I_UPDATE_ITERATIONS;
				if (shmReps < 1)
					shmReps = 1;
				sockReps = (int)
					(
					 (
					  (MPID_CPU_Tick_t)((shmTicks + sockTicks) >> 1)
					  * (MPID_CPU_Tick_t)sockTotalReps
					 ) / sockTicks
					) / MPIDI_CH3I_UPDATE_ITERATIONS;
				if (sockReps < 1)
					sockReps = 1;
				/*MPIU_DBG_PRINTF((L"(BOTH_BITS: shmReps = %d, sockReps = %d)", shmReps, sockReps));*/
				break;
			default:
				/* neither channel has been active */
				/* leave the state in its previous state */
				/*wprintf(L".");*/
				break;
			}
			MPIDI_CH3I_active_flag = 0;
			shmTotalReps = 0;
			shmTicks = 0;
			shmIter = 0;
			sockTotalReps = 0;
			sockTicks = 0;
			sockIter = 0;
		}

		if (msgqIter++ == MPIDI_CH3I_MSGQ_ITERATIONS)
		{
			msgqIter = 0;
			mpi_errno = fixed_activeMPIDI_CH3I_Message_queue_progress();
			if (mpi_errno != MPI_SUCCESS)
			{
				mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER,
												 L"**mqp_failure", 0);
				goto fn_exit;
			}
		}
	}
	while (completions == MPIDI_CH3I_progress_completion_count && is_blocking);

fn_exit:
#ifdef MPICH_DBG_OUTPUT
	if (is_blocking)
	{
		MPIDI_DBG_PRINTF((50, FCNAME, L"exiting, count=%d", MPIDI_CH3I_progress_completion_count - completions));
	}
	else
	{
		if (MPIDI_CH3I_progress_completion_count - completions > 0)
		{
			MPIDI_DBG_PRINTF((50, FCNAME, L"exiting (non-blocking), count=%d", MPIDI_CH3I_progress_completion_count - completions));
		}
	}
#endif
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_PROGRESS);
	/*return MPIDI_CH3I_progress_completion_count - completions;*/
	return mpi_errno;
}
#endif

#if !defined(MPIDI_CH3_Progress_poke)
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_Progress_poke
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3_Progress_poke()
{
	int mpi_errno;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_PROGRESS_POKE);
	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_PROGRESS_POKE);
	mpi_errno = MPIDI_CH3.Progress_test();
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_PROGRESS_POKE);
	return mpi_errno;
}
#endif

#if !defined(MPIDI_CH3_Progress_start)
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_Progress_start
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
void MPIDI_CH3_Progress_start(MPID_Progress_state *state)
{
	/* MT - This function is empty for the single-threaded implementation */
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_PROGRESS_START);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_PROGRESS_START);
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_PROGRESS_START);
}
#endif

#if !defined(MPIDI_CH3_Progress_end)
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_Progress_end
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
void MPIDI_CH3_Progress_end(MPID_Progress_state *state)
{
	/* MT: This function is empty for the single-threaded implementation */
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_PROGRESS_END);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_PROGRESS_END);
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_PROGRESS_END);
}
#endif

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_Progress_init
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int ssmMPIDI_CH3I_Progress_init()
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_PROGRESS_INIT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_PROGRESS_INIT);

	mpi_errno = MPIDU_Sock_init();
	/* --BEGIN ERROR HANDLING-- */
	if (mpi_errno != MPI_SUCCESS)
	{
		mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**progress_init", 0);
		goto fn_exit;
	}
	/* --END ERROR HANDLING-- */

	/* create sock set */
	mpi_errno = MPIDU_Sock_create_set(&MPIDI_CH3I_sock_set);
	/* --BEGIN ERROR HANDLING-- */
	if (mpi_errno != MPI_SUCCESS)
	{
		mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**progress_init", 0);
		goto fn_exit;
	}
	/* --END ERROR HANDLING-- */

	/* establish non-blocking listener */
	mpi_errno = MPIDU_CH3I_SetupListener( MPIDI_CH3I_sock_set );
	if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }

 fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_PROGRESS_INIT);
	return mpi_errno;

 fn_fail:
	goto fn_exit;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_Progress_finalize
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int ssmMPIDI_CH3I_Progress_finalize(void)
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_PROGRESS_FINALIZE);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_PROGRESS_FINALIZE);
	MPIDI_DBG_PRINTF((60, FCNAME, L"entering"));

	/* Shut down the listener */
	mpi_errno = MPIDU_CH3I_ShutdownListener();
	MPIU_ERR_CHK_POP(mpi_errno);

	/* FIXME: Cleanly shutdown other socks and MPIU_Free connection 
	   structures. (close protocol?) */

	mpi_errno = MPIDU_Sock_destroy_set(MPIDI_CH3I_sock_set);
	MPIU_ERR_CHK_POP(mpi_errno);
	mpi_errno = MPIDU_Sock_finalize();
	MPIU_ERR_CHK_POP(mpi_errno);

fn_fail:
	MPIDI_DBG_PRINTF((60, FCNAME, L"exiting"));
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_PROGRESS_FINALIZE);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_Connection_terminate
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int ssmMPIDI_CH3_Connection_terminate(MPIDI_VC_t * vc)
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_CONNECTION_TERMINATE);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_CONNECTION_TERMINATE);

	if (vc->ch.bShm)
	{
		/* There is no post_close for shm connections so handle them as closed immediately. */
		MPIDI_CH3I_SHM_Remove_vc_references(vc);
		MPIDI_CH3U_Handle_connection(vc, MPIDI_VC_EVENT_TERMINATED);
	}
	else
	{
		vc->ch.conn->state = CONN_STATE_CLOSING;
		mpi_errno = MPIDU_Sock_post_close(vc->ch.sock);
		if (mpi_errno != MPI_SUCCESS)
		{
			MPIU_ERR_SET(mpi_errno,MPI_ERR_OTHER, L"**fail");
			goto fn_exit;
		}
	}

  fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_CONNECTION_TERMINATE);
	return mpi_errno;
}


#undef FUNCNAME
#define FUNCNAME ssmMPIDI_CH3I_VC_post_connect
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int ssmMPIDI_CH3I_VC_post_connect(MPIDI_VC_t * vc)
{
	int mpi_errno = MPI_SUCCESS;
	wchar_t val[MPIDI_MAX_KVS_VALUE_LEN];
	wchar_t host_description[MAX_HOST_DESCRIPTION_LEN];
	int port;
	MPIDU_Sock_ifaddr_t ifaddr;
	int hasIfaddr = 0;
	MPIDI_CH3I_Connection_t * conn;
	int connected;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_VC_POST_CONNECT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_VC_POST_CONNECT);

	MPIDI_DBG_PRINTF((60, FCNAME, L"entering"));

	if (vc->ch.state != MPIDI_CH3I_VC_STATE_UNCONNECTED)
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**vc_state", L"**vc_state %d", vc->ch.state);
		goto fn_fail;
	}

	vc->ch.state = MPIDI_CH3I_VC_STATE_CONNECTING;

	/* get the business card */
	mpi_errno = MPIDI_PG_GetConnString( vc->pg, vc->pg_rank, val, sizeof(val));
	if (mpi_errno)
	{
		MPIU_ERR_POP(mpi_errno);
	}

	/* attempt to connect through shared memory */
/*	wprintf( L"trying to connect through shared memory...\n"); fflush(stdout); */
	connected = FALSE;
/*	MPIU_DBG_PRINTF((L"business card: <%s> = <%s>\n", key, val)); */
	mpi_errno = MPIDI_CH3I_Shm_connect(vc, val, &connected);
/*	wprintf( L"After attempt to connect, flag = %d and mpi_errno = %d\n", connected, mpi_errno ); fflush(stdout); */
	if (mpi_errno != MPI_SUCCESS) 
	{
		mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**post_connect", L"**post_connect %s", L"MPIDI_CH3I_Shm_connect");
		goto fn_fail;
	}
	if (connected)
	{
		MPIDI_VC_t *iter;
		int count = 0;

		/*MPIU_DBG_PRINTF((L"shmem connected\n"));*/
		MPIDI_CH3I_SHM_Add_to_writer_list(vc);

		/* If there are more shm connections than cpus, reduce the spin count to one. */
		/* This does not take into account connections between other processes on the same machine. */
		iter = MPIDI_CH3I_Process.shm_writing_list;
		while (iter)
		{
			count++;
			iter = iter->ch.shm_next_writer;
		}
		if (count >= MPIDI_CH3I_Process.num_cpus)
		{
			MPIDI_Process.my_pg->ch.nShmWaitSpinCount = 1;
		}

		vc->ch.state = MPIDI_CH3I_VC_STATE_CONNECTED;
		vc->ch.bShm = TRUE;
		vc->ch.shm_reading_pkt = TRUE;
		vc->ch.send_active = MPIDI_CH3I_SendQ_head(vc); /* MT */
		goto fn_exit;
	}

/*	wprintf( L"Attempting to connect through socket\n" );fflush(stdout); */
	MPIU_DBG_MSG_FMT(CH3_CONNECT, TYPICAL, (MPIU_DBG_FDEST, L"vc=%p: Attempting to connect with business card %s", vc, val) );
	/* attempt to connect through sockets */
	mpi_errno = MPIDU_Sock_get_conninfo_from_bc( val, host_description,
												 sizeof(host_description),
												 &port, &ifaddr, &hasIfaddr );
	if (mpi_errno)
	{
		MPIU_ERR_POP(mpi_errno);
	}

/*	wprintf (L"Allocating connection\n" );fflush(stdout);*/
	mpi_errno = MPIDI_CH3I_Connection_alloc(&conn);
	if (mpi_errno == MPI_SUCCESS)
	{
/*		wprintf( L"Posting socket connection\n" );fflush(stdout);*/
		/* FIXME: This is a hack to allow Windows to continue to use
		   the host description string instead of the interface address
		   bytes when posting a socket connection.  This should be fixed 
		   by changing the Sock_post_connect to only accept interface
		   address.  See also channels/sock/ch3_progress.c */
		{
			mpi_errno = MPIDU_Sock_post_connect(MPIDI_CH3I_sock_set, conn, host_description, port, &conn->sock);
		}
		if (mpi_errno == MPI_SUCCESS)
		{
			vc->ch.sock = conn->sock;
			vc->ch.conn = conn;
			conn->vc = vc;
			conn->state = CONN_STATE_CONNECTING;
			conn->send_active = NULL;
			conn->recv_active = NULL;
		}
		else
		{
			mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|sock|postconnect", L"**ch3|sock|postconnect %d %d %s", MPIR_Process.comm_world->rank, vc->pg_rank, val);

			vc->ch.state = MPIDI_CH3I_VC_STATE_FAILED;
			if (vc->ch.conn == conn) vc->ch.conn = 0;
			ssmMPIDI_CH3I_Connection_free(conn);
		}
	}
	else
	{
		MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, L"**ch3|sock|connalloc");
	}

 fn_exit:
/*	wprintf(L"Exiting with %d\n", mpi_errno );fflush(stdout);*/
	MPIDI_DBG_PRINTF((60, FCNAME, L"exiting"));
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_VC_POST_CONNECT);
	return mpi_errno;
 fn_fail:
	goto fn_exit;
}

#undef FUNCNAME
#define FUNCNAME ssmMPIDI_CH3I_Connection_free
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
void ssmMPIDI_CH3I_Connection_free(MPIDI_CH3I_Connection_t * conn)
{
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_CONNECTION_FREE);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_CONNECTION_FREE);

	MPIU_Free(conn->pg_id);
	MPIU_Free(conn);

	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_CONNECTION_FREE);
}


static inline void connection_post_send_pkt_and_pgid(MPIDI_CH3I_Connection_t * conn);
static inline int connection_post_recv_pkt(MPIDI_CH3I_Connection_t * conn);
static inline int connection_post_send_pkt(MPIDI_CH3I_Connection_t * conn);
static inline int connection_post_sendq_req(MPIDI_CH3I_Connection_t * conn);
static int adjust_iov(MPID_IOV ** iovp, int * countp, MPIU_Size_t nb);

#undef FUNCNAME
#define FUNCNAME connection_post_sendq_req
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static inline int connection_post_sendq_req(MPIDI_CH3I_Connection_t * conn)
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_CONNECTION_POST_SENDQ_REQ);

	MPIDI_FUNC_ENTER(MPID_STATE_CONNECTION_POST_SENDQ_REQ);
	/* post send of next request on the send queue */
	conn->send_active = MPIDI_CH3I_SendQ_head(conn->vc); /* MT */
	if (conn->send_active != NULL)
	{
		mpi_errno = MPIDU_Sock_post_writev(conn->sock, conn->send_active->dev.iov, conn->send_active->dev.iov_count, NULL);
		if (mpi_errno != MPI_SUCCESS)
		{
			mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", NULL);
		}
	}
	
	MPIDI_FUNC_EXIT(MPID_STATE_CONNECTION_POST_SENDQ_REQ);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME connection_post_send_pkt
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static inline int connection_post_send_pkt(MPIDI_CH3I_Connection_t * conn)
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_CONNECTION_POST_SEND_PKT);

	MPIDI_FUNC_ENTER(MPID_STATE_CONNECTION_POST_SEND_PKT);
	
	mpi_errno = MPIDU_Sock_post_write(conn->sock, &conn->pkt, sizeof(conn->pkt), sizeof(conn->pkt), NULL);
	/* --BEGIN ERROR HANDLING-- */
	if (mpi_errno != MPI_SUCCESS)
	{
		mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", NULL);
	}
	/* --END ERROR HANDLING-- */

	MPIDI_FUNC_EXIT(MPID_STATE_CONNECTION_POST_SEND_PKT);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME connection_post_recv_pkt
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static inline int connection_post_recv_pkt(MPIDI_CH3I_Connection_t * conn)
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_CONNECTION_POST_RECV_PKT);

	MPIDI_FUNC_ENTER(MPID_STATE_CONNECTION_POST_RECV_PKT);

	mpi_errno = MPIDU_Sock_post_read(conn->sock, &conn->pkt, sizeof(conn->pkt), sizeof(conn->pkt), NULL);
	/* --BEGIN ERROR HANDLING-- */
	if (mpi_errno != MPI_SUCCESS)
	{
		mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", NULL);
	}
	/* --END ERROR HANDLING-- */

	MPIDI_FUNC_EXIT(MPID_STATE_CONNECTION_POST_RECV_PKT);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME adjust_iov
#undef FCNAME
#define FCNAME MPIU_QUOTE(FUNCNAME)
static int adjust_iov(MPID_IOV ** iovp, int * countp, MPIU_Size_t nb)
{
	MPID_IOV * const iov = *iovp;
	const int count = *countp;
	int offset = 0;
	MPIDI_STATE_DECL(MPID_STATE_ADJUST_IOV);

	MPIDI_FUNC_ENTER(MPID_STATE_ADJUST_IOV);

	while (offset < count)
	{
		if (iov[offset].MPID_IOV_LEN <= (MPID_IOV_LEN_CAST)nb)
		{
			nb -= (MPIU_Size_t)iov[offset].MPID_IOV_LEN;
			offset++;
		}
		else
		{
			iov[offset].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)((char *) iov[offset].MPID_IOV_BUF + nb);
			iov[offset].MPID_IOV_LEN -= (MPID_IOV_LEN_CAST)nb;
			break;
		}
	}

	*iovp += offset;
	*countp -= offset;

	MPIDI_FUNC_EXIT(MPID_STATE_ADJUST_IOV);
	return (*countp == 0);
}

#undef FUNCNAME
#define FUNCNAME connection_post_send_pkt_and_pgid
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static inline void connection_post_send_pkt_and_pgid(MPIDI_CH3I_Connection_t * conn)
{
	int mpi_errno;
	MPIDI_STATE_DECL(MPID_STATE_CONNECTION_POST_SEND_PKT_AND_PGID);

	MPIDI_FUNC_ENTER(MPID_STATE_CONNECTION_POST_SEND_PKT_AND_PGID);
	
	conn->iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST) &conn->pkt;
	conn->iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST) sizeof(conn->pkt);

	conn->iov[1].MPID_IOV_BUF = (MPID_IOV_BUF_CAST) MPIDI_Process.my_pg->id;
	conn->iov[1].MPID_IOV_LEN = (MPID_IOV_LEN_CAST) ((wcslen(MPIDI_Process.my_pg->id) + 1) * sizeof(wchar_t));

	mpi_errno = MPIDU_Sock_post_writev(conn->sock, conn->iov, 2, NULL);
	/* --BEGIN ERROR HANDLING-- */
	if (mpi_errno != MPI_SUCCESS)
	{
		mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", NULL);
	}
	/* --END ERROR HANDLING-- */

	MPIDI_FUNC_EXIT(MPID_STATE_CONNECTION_POST_SEND_PKT_AND_PGID);
}


#undef FUNCNAME
#define FUNCNAME ssmMPIDI_CH3I_Progress_handle_sock_event
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int ssmMPIDI_CH3I_Progress_handle_sock_event(MPIDU_Sock_event_t * event)
{
	int complete;
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_PROGRESS_HANDLE_SOCK_EVENT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_PROGRESS_HANDLE_SOCK_EVENT);

	switch (event->op_type)
	{
		case MPIDU_SOCK_OP_READ:
		{
			MPIDI_CH3I_Connection_t * conn = (MPIDI_CH3I_Connection_t *) event->user_ptr;
				
			MPID_Request * rreq = conn->recv_active;

			/* --BEGIN ERROR HANDLING-- */
			if (event->error != MPI_SUCCESS)
			{
				/* FIXME: the following should be handled by the close protocol */
				/*if (!shutting_down || MPIR_ERR_GET_CLASS(event->error) != MPIDU_SOCK_ERR_CONN_CLOSED)*/
				{
					mpi_errno = MPIR_Err_create_code(event->error, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", NULL);
					goto fn_exit;
				}
					
				break;
			}
			/* --END ERROR HANDLING-- */

			if (conn->state == CONN_STATE_CONNECTED)
			{
				if (conn->recv_active == NULL)
				{
					MPIU_Assert(conn->pkt.type < MPIDI_CH3_PKT_END_CH3);
						
					mpi_errno = MPIDI_CH3U_Handle_recv_pkt(conn->vc, &conn->pkt, &rreq);
					/* --BEGIN ERROR HANDLING-- */
					if (mpi_errno != MPI_SUCCESS)
					{
						mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER,
														 L"**fail", NULL);
						goto fn_exit;
					}
					/* --END ERROR HANDLING-- */

					if (rreq == NULL)
					{
						if (conn->state != CONN_STATE_CLOSING)
						{
							/* conn->recv_active = NULL;  -- already set to NULL */
							mpi_errno = connection_post_recv_pkt(conn);
							/* --BEGIN ERROR HANDLING-- */
							if (mpi_errno != MPI_SUCCESS)
							{
								mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_INTERN,
									L"**fail", NULL);
								goto fn_exit;
							}
							/* --END ERROR HANDLING-- */
						}
					}
					else
					{
						for(;;)
						{
							MPID_IOV * iovp;
							MPIU_Size_t nb;
								
							iovp = rreq->dev.iov;
							
							mpi_errno = MPIDU_Sock_readv(conn->sock, iovp, rreq->dev.iov_count, &nb);
							/* --BEGIN ERROR HANDLING-- */
							if (mpi_errno != MPI_SUCCESS)
							{
								mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|sock|immedread", L"ch3|sock|immedread %p %p %p", rreq, conn, conn->vc);
								goto fn_exit;
							}
							/* --END ERROR HANDLING-- */

							MPIDI_DBG_PRINTF((55, FCNAME, L"immediate readv, vc=0x%p nb=%d, rreq=0x%08x", conn->vc, rreq->handle, nb));

							if (nb > 0 && adjust_iov(&iovp, &rreq->dev.iov_count, nb))
							{
								mpi_errno = MPIDI_CH3U_Handle_recv_req(conn->vc, rreq, &complete);
								/* --BEGIN ERROR HANDLING-- */
								if (mpi_errno != MPI_SUCCESS)
								{
									mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", NULL);
									goto fn_exit;
								}
								/* --END ERROR HANDLING-- */

								if (complete)
								{
									/* conn->recv_active = NULL; -- already set to NULL */
									mpi_errno = connection_post_recv_pkt(conn);
									/* --BEGIN ERROR HANDLING-- */
									if (mpi_errno != MPI_SUCCESS)
									{
										mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_INTERN, L"**fail", NULL);
										goto fn_exit;
									}
									/* --END ERROR HANDLING-- */

									break;
								}
							}
							else
							{
								MPIDI_DBG_PRINTF((55, FCNAME, L"posting readv, vc=0x%p, rreq=0x%08x", conn->vc, rreq->handle));
								conn->recv_active = rreq;
								mpi_errno = MPIDU_Sock_post_readv(conn->sock, iovp, rreq->dev.iov_count, NULL);
								/* --BEGIN ERROR HANDLING-- */
								if (mpi_errno != MPI_SUCCESS)
								{
									mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|sock|postread", L"ch3|sock|postread %p %p %p", rreq, conn, conn->vc);
									goto fn_exit;
								}
								/* --END ERROR HANDLING-- */

								break;
							}
						}
					}
				}
				else /* incoming data */
				{
					mpi_errno = MPIDI_CH3U_Handle_recv_req(conn->vc, rreq, &complete);
					/* --BEGIN ERROR HANDLING-- */
					if (mpi_errno != MPI_SUCCESS)
					{
						mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", NULL);
						goto fn_exit;
					}
					/* --END ERROR HANDLING-- */

					if (complete)
					{
						conn->recv_active = NULL;
						mpi_errno = connection_post_recv_pkt(conn);
						/* --BEGIN ERROR HANDLING-- */
						if (mpi_errno != MPI_SUCCESS)
						{
							mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_INTERN, L"**fail", NULL);
							goto fn_exit;
						}
						/* --END ERROR HANDLING-- */
					}
					else /* more data to be read */
					{
						for(;;)
						{
							MPID_IOV * iovp;
							MPIU_Size_t nb;
								
							iovp = rreq->dev.iov;
							
							mpi_errno = MPIDU_Sock_readv(conn->sock, iovp, rreq->dev.iov_count, &nb);
							/* --BEGIN ERROR HANDLING-- */
							if (mpi_errno != MPI_SUCCESS)
							{
								mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|sock|immedread", L"ch3|sock|immedread %p %p %p", rreq, conn, conn->vc);
								goto fn_exit;
							}
							/* --END ERROR HANDLING-- */

							MPIDI_DBG_PRINTF((55, FCNAME, L"immediate readv, vc=0x%p nb=%d, rreq=0x%08x", conn->vc, rreq->handle, nb));
								
							if (nb > 0 && adjust_iov(&iovp, &rreq->dev.iov_count, nb))
							{
								mpi_errno = MPIDI_CH3U_Handle_recv_req(conn->vc, rreq, &complete);
								/* --BEGIN ERROR HANDLING-- */
								if (mpi_errno != MPI_SUCCESS)
								{
									mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", NULL);
									goto fn_exit;
								}
								/* --END ERROR HANDLING-- */

								if (complete)
								{
									conn->recv_active = NULL;
									mpi_errno = connection_post_recv_pkt(conn);
									/* --BEGIN ERROR HANDLING-- */
									if (mpi_errno != MPI_SUCCESS)
									{
										mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_INTERN, L"**fail", NULL);
										goto fn_exit;
									}
									/* --END ERROR HANDLING-- */

									break;
								}
							}
							else
							{
								MPIDI_DBG_PRINTF((55, FCNAME, L"posting readv, vc=0x%p, rreq=0x%08x", conn->vc, rreq->handle));
								/* conn->recv_active = rreq;  -- already set to current request */
								mpi_errno = MPIDU_Sock_post_readv(conn->sock, iovp, rreq->dev.iov_count, NULL);
								/* --BEGIN ERROR HANDLING-- */
								if (mpi_errno != MPI_SUCCESS)
								{
									mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|sock|postread", L"ch3|sock|postread %p %p %p", rreq, conn, conn->vc);
									goto fn_exit;
								}
								/* --END ERROR HANDLING-- */

								break;
							}
						}
					}
				}
			}
			else if (conn->state == CONN_STATE_OPEN_LRECV_DATA)
			{
				MPIDI_PG_t * pg;
				int pg_rank;
				MPIDI_VC_t * vc;

				/* Look up pg based on conn->pg_id */
				mpi_errno = MPIDI_PG_Find(conn->pg_id, &pg);
				/* --BEGIN ERROR HANDLING-- */
				if (pg == NULL)
				{
					mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**pglookup", L"**pglookup %s", conn->pg_id);
					goto fn_exit;
				}
				/* --END ERROR HANDLING-- */

				pg_rank = conn->pkt.sc_open_req.pg_rank;
				MPIDI_PG_Get_vc(pg, pg_rank, &vc);
				MPIU_Assert(vc->pg_rank == pg_rank);
					
				if (vc->ch.conn == NULL)
				{
					/* no head-to-head connects, accept the
					   connection */
					vc->ch.state = MPIDI_CH3I_VC_STATE_CONNECTING;
					vc->ch.sock = conn->sock;
					vc->ch.conn = conn;
					conn->vc = vc;
						
					MPIDI_Pkt_init(&conn->pkt, MPIDI_CH3I_PKT_SC_OPEN_RESP);
					conn->pkt.sc_open_resp.ack = TRUE;
				}
				else
				{
					/* head to head situation */
					if (pg == MPIDI_Process.my_pg)
					{
						/* the other process is in the same comm_world; just compare the ranks */
						if (MPIR_Process.comm_world->rank < pg_rank)
						{
							/* accept connection */
							vc->ch.state = MPIDI_CH3I_VC_STATE_CONNECTING;
							vc->ch.sock = conn->sock;
							vc->ch.conn = conn;
							conn->vc = vc;
								
							MPIDI_Pkt_init(&conn->pkt, MPIDI_CH3I_PKT_SC_OPEN_RESP);
							conn->pkt.sc_open_resp.ack = TRUE;
						}
						else
						{
							/* refuse connection */
							MPIDI_Pkt_init(&conn->pkt, MPIDI_CH3I_PKT_SC_OPEN_RESP);
							conn->pkt.sc_open_resp.ack = FALSE;
						}
					}
					else
					{ 
						/* the two processes are in different comm_worlds; compare their unique pg_ids. */
						if (wcscmp(MPIDI_Process.my_pg->id, pg->id) < 0)
						{
							/* accept connection */
							vc->ch.state = MPIDI_CH3I_VC_STATE_CONNECTING;
							vc->ch.sock = conn->sock;
							vc->ch.conn = conn;
							conn->vc = vc;
								
							MPIDI_Pkt_init(&conn->pkt, MPIDI_CH3I_PKT_SC_OPEN_RESP);
							conn->pkt.sc_open_resp.ack = TRUE;
						}
						else
						{
							/* refuse connection */
							MPIDI_Pkt_init(&conn->pkt, MPIDI_CH3I_PKT_SC_OPEN_RESP);
							conn->pkt.sc_open_resp.ack = FALSE;
						}
					}
				}
					
				conn->state = CONN_STATE_OPEN_LSEND;
				mpi_errno = connection_post_send_pkt(conn);
				/* --BEGIN ERROR HANDLING-- */
				if (mpi_errno != MPI_SUCCESS)
				{
					mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_INTERN, L"**ch3|sock|open_lrecv_data", NULL);
					goto fn_exit;
				}
				/* --END ERROR HANDLING-- */
			}
			else /* Handling some internal connection establishment or tear down packet */
			{ 
				if (conn->pkt.type == MPIDI_CH3I_PKT_SC_OPEN_REQ)
				{
					conn->state = CONN_STATE_OPEN_LRECV_DATA;
					mpi_errno = MPIDU_Sock_post_read(conn->sock, conn->pg_id, conn->pkt.sc_open_req.pg_id_len, conn->pkt.sc_open_req.pg_id_len, NULL);   
					/* --BEGIN ERROR HANDLING-- */
					if (mpi_errno != MPI_SUCCESS)
					{
						mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", NULL);
						goto fn_exit;
					}
					/* --END ERROR HANDLING-- */
				}
				else if (conn->pkt.type == MPIDI_CH3I_PKT_SC_CONN_ACCEPT)
				{
					MPIDI_VC_t *vc;
					int port_name_tag;

					vc = (MPIDI_VC_t *) MPIU_Malloc(sizeof(MPIDI_VC_t));
					/* --BEGIN ERROR HANDLING-- */
					if (vc == NULL)
					{
						mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", NULL);
						goto fn_exit;
					}
					/* --END ERROR HANDLING-- */
					/* FIXME - where does this vc get freed? */

					/* Initialize the device fields */
					MPIDI_VC_Init(vc, NULL, 0);
					/* Initialize the sock fields */
					vc->ch.sendq_head = NULL;
					vc->ch.sendq_tail = NULL;
					vc->ch.state = MPIDI_CH3I_VC_STATE_CONNECTING;
					vc->ch.sock = conn->sock;
					vc->ch.conn = conn;
					conn->vc = vc;
					/* Initialize the shm fields */
					vc->ch.recv_active = NULL;
					vc->ch.send_active = NULL;
					vc->ch.req = NULL;
					vc->ch.read_shmq = NULL;
					vc->ch.write_shmq = NULL;
					vc->ch.shm = NULL;
					vc->ch.shm_state = 0;
					vc->ch.shm_next_reader = NULL;
					vc->ch.shm_next_writer = NULL;
					vc->ch.bShm = FALSE;
					vc->ch.shm_read_connected = 0;

					port_name_tag = conn->pkt.sc_conn_accept.port_name_tag;
					
					MPIDI_Pkt_init(&conn->pkt, MPIDI_CH3I_PKT_SC_OPEN_RESP);
					conn->pkt.sc_open_resp.ack = TRUE;
						
					conn->state = CONN_STATE_OPEN_LSEND;
					mpi_errno = connection_post_send_pkt(conn);
					/* --BEGIN ERROR HANDLING-- */
					if (mpi_errno != MPI_SUCCESS)
					{
						mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_INTERN, L"**ch3|sock|scconnaccept", NULL);
						goto fn_exit;
					}
					/* --END ERROR HANDLING-- */

					/* ENQUEUE vc */
					MPIDI_CH3I_Acceptq_enqueue(vc, port_name_tag);

				}
				else if (conn->pkt.type == MPIDI_CH3I_PKT_SC_OPEN_RESP)
				{
					if (conn->pkt.sc_open_resp.ack)
					{
						conn->state = CONN_STATE_CONNECTED;
						conn->vc->ch.state = MPIDI_CH3I_VC_STATE_CONNECTED;
						MPIU_Assert(conn->vc->ch.conn == conn);
						MPIU_Assert(conn->vc->ch.sock == conn->sock);
							
						mpi_errno = connection_post_recv_pkt(conn);
						/* --BEGIN ERROR HANDLING-- */
						if (mpi_errno != MPI_SUCCESS)
						{
							mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_INTERN, L"**fail", NULL);
							goto fn_exit;
						}
						/* --END ERROR HANDLING-- */
						mpi_errno = connection_post_sendq_req(conn);
						/* --BEGIN ERROR HANDLING-- */
						if (mpi_errno != MPI_SUCCESS)
						{
							mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_INTERN, L"**ch3|sock|scopenresp", NULL);
							goto fn_exit;
						}
						/* --END ERROR HANDLING-- */
					}
					else
					{
						conn->vc = NULL;
						conn->state = CONN_STATE_CLOSING;
						MPIDU_Sock_post_close(conn->sock);
					}
				}
				/* --BEGIN ERROR HANDLING-- */
				else
				{
					MPIDI_DBG_Print_packet(&conn->pkt);
					mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_INTERN, L"**ch3|sock|badpacket", L"**ch3|sock|badpacket %d", conn->pkt.type);
					goto fn_exit;
				}
				/* --END ERROR HANDLING-- */
			}

			break;
		}
			
		case MPIDU_SOCK_OP_WRITE:
		{
			MPIDI_CH3I_Connection_t * conn = (MPIDI_CH3I_Connection_t *) event->user_ptr;

			/* --BEGIN ERROR HANDLING-- */
			if (event->error != MPI_SUCCESS)
			{
				mpi_errno = MPIR_Err_create_code(event->error, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", NULL);
				goto fn_exit;
			}
			/* --END ERROR HANDLING-- */
				
			if (conn->send_active)
			{
				MPID_Request * sreq = conn->send_active;

				mpi_errno = MPIDI_CH3U_Handle_send_req(conn->vc, sreq, &complete);
				/* --BEGIN ERROR HANDLING-- */
				if (mpi_errno != MPI_SUCCESS)
				{
					mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", NULL);
					goto fn_exit;
				}
				/* --END ERROR HANDLING-- */

				if (complete)
				{
					MPIDI_CH3I_SendQ_dequeue(conn->vc);
					mpi_errno = connection_post_sendq_req(conn);
					/* --BEGIN ERROR HANDLING-- */
					if (mpi_errno != MPI_SUCCESS)
					{
						mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_INTERN, L"**fail", NULL);
						goto fn_exit;
					}
					/* --END ERROR HANDLING-- */
				}
				else /* more data to send */
				{
					for(;;)
					{
						MPID_IOV * iovp;
						MPIU_Size_t nb;
								
						iovp = sreq->dev.iov;
							
						mpi_errno = MPIDU_Sock_writev(conn->sock, iovp, sreq->dev.iov_count, &nb);
						/* --BEGIN ERROR HANDLING-- */
						if (mpi_errno != MPI_SUCCESS)
						{
							mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|sock|immedwrite", L"ch3|sock|immedwrite %p %p %p", sreq, conn, conn->vc);
							goto fn_exit;
						}
						/* --END ERROR HANDLING-- */

						MPIDI_DBG_PRINTF((55, FCNAME, L"immediate writev, vc=0x%p, sreq=0x%08x, nb=%d", conn->vc, sreq->handle, nb));

						if (nb > 0 && adjust_iov(&iovp, &sreq->dev.iov_count, nb))
						{
							mpi_errno = MPIDI_CH3U_Handle_send_req(conn->vc, sreq, &complete);
							/* --BEGIN ERROR HANDLING-- */
							if (mpi_errno != MPI_SUCCESS)
							{
								mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", NULL);
								goto fn_exit;
							}
							/* --END ERROR HANDLING-- */

							if (complete)
							{
								MPIDI_CH3I_SendQ_dequeue(conn->vc);
								mpi_errno = connection_post_sendq_req(conn);
								/* --BEGIN ERROR HANDLING-- */
								if (mpi_errno != MPI_SUCCESS)
								{
									mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_INTERN, L"**fail", NULL);
									goto fn_exit;
								}
								/* --END ERROR HANDLING-- */

								break;
							}
						}
						else
						{
							MPIDI_DBG_PRINTF((55, FCNAME, L"posting writev, vc=0x%p, sreq=0x%08x", conn->vc, sreq->handle));
							mpi_errno = MPIDU_Sock_post_writev(conn->sock, iovp, sreq->dev.iov_count, NULL);
							/* --BEGIN ERROR HANDLING-- */
							if (mpi_errno != MPI_SUCCESS)
							{
								mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|sock|postwrite", L"ch3|sock|postwrite %p %p %p", sreq, conn, conn->vc);
								goto fn_exit;
							}
							/* --END ERROR HANDLING-- */

							break;
						}
					}
				}
			}
			else /* finished writing internal packet header */
			{
				if (conn->state == CONN_STATE_OPEN_CSEND)
				{
					/* finished sending open request packet */
					/* post receive for open response packet */
					conn->state = CONN_STATE_OPEN_CRECV;
					mpi_errno = connection_post_recv_pkt(conn);
					/* --BEGIN ERROR HANDLING-- */
					if (mpi_errno != MPI_SUCCESS)
					{
						mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_INTERN, L"**fail", NULL);
						goto fn_exit;
					}
					/* --END ERROR HANDLING-- */
				}
				else if (conn->state == CONN_STATE_OPEN_LSEND)
				{
					/* finished sending open response packet */
					if (conn->pkt.sc_open_resp.ack == TRUE)
					{ 
						/* post receive for packet header */
						conn->state = CONN_STATE_CONNECTED;
						conn->vc->ch.state = MPIDI_CH3I_VC_STATE_CONNECTED;
						mpi_errno = connection_post_recv_pkt(conn);
						/* --BEGIN ERROR HANDLING-- */
						if (mpi_errno != MPI_SUCCESS)
						{
							mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_INTERN, L"**fail", NULL);
							goto fn_exit;
						}
						/* --END ERROR HANDLING-- */

						mpi_errno = connection_post_sendq_req(conn);
						/* --BEGIN ERROR HANDLING-- */
						if (mpi_errno != MPI_SUCCESS)
						{
							mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_INTERN, L"**ch3|sock|openlsend", NULL);
							goto fn_exit;
						}
						/* --END ERROR HANDLING-- */
					}
					else
					{
						/* head-to-head connections - close this connection */
						conn->state = CONN_STATE_CLOSING;
						mpi_errno = MPIDU_Sock_post_close(conn->sock);
						/* --BEGIN ERROR HANDLING-- */
						if (mpi_errno != MPI_SUCCESS)
						{
							mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**sock_post_close", NULL);
							goto fn_exit;
						}
						/* --END ERROR HANDLING-- */
					}
				}
			}

			break;
		}

		case MPIDU_SOCK_OP_ACCEPT:
		{
			mpi_errno = MPIDI_CH3_Sockconn_handle_accept_event();
			if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }
			break;
		}

		case MPIDU_SOCK_OP_CONNECT:
		{
			MPIDI_CH3I_Connection_t * conn = (MPIDI_CH3I_Connection_t *) event->user_ptr;

			/* --BEGIN ERROR HANDLING-- */
			if (event->error != MPI_SUCCESS)
			{
				mpi_errno = MPIR_Err_create_code(event->error, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|sock|connfailed", L"**ch3|sock|connfailed %s %d", conn->vc->pg->id, conn->vc->pg_rank);
				goto fn_exit;
			}
			/* --END ERROR HANDLING-- */

			if (conn->state == CONN_STATE_CONNECTING)
			{
				conn->state = CONN_STATE_OPEN_CSEND;
				MPIDI_Pkt_init(&conn->pkt, MPIDI_CH3I_PKT_SC_OPEN_REQ);
				conn->pkt.sc_open_req.pg_id_len = (int) (wcslen(MPIDI_Process.my_pg->id) + 1)*sizeof(wchar_t);
				conn->pkt.sc_open_req.pg_rank = MPIR_Process.comm_world->rank;

				connection_post_send_pkt_and_pgid(conn);
			}
			else
			{
				/* CONN_STATE_CONNECT_ACCEPT */
				   int port_name_tag;
	   
				conn->state = CONN_STATE_OPEN_CSEND;

				/* pkt contains port name tag. In memory debugging mode, MPIDI_Pkt_init resets the packet contents. Therefore,
				   save the port name tag and then add it back. */
				port_name_tag = conn->pkt.sc_conn_accept.port_name_tag;
				
				MPIDI_Pkt_init(&conn->pkt, MPIDI_CH3I_PKT_SC_CONN_ACCEPT);

				conn->pkt.sc_conn_accept.port_name_tag = port_name_tag;
				
				mpi_errno = connection_post_send_pkt(conn);
				/* --BEGIN ERROR HANDLING-- */
				if (mpi_errno != MPI_SUCCESS)
				{
					mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_INTERN, L"**ch3|sock|scconnaccept", NULL);
					goto fn_exit;
				}
				/* --END ERROR HANDLING-- */
			}

			break;
		}

		case MPIDU_SOCK_OP_CLOSE:
		{
			mpi_errno = MPIDI_CH3_Sockconn_handle_close_event( (MPIDI_CH3I_Connection_t *) event->user_ptr );
			if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }
			break;
		}

		case MPIDU_SOCK_OP_WAKEUP:
		{
			MPIDI_CH3_Progress_signal_completion();
			/* MPIDI_CH3I_progress_completion_count++; */
			break;
		}
	}

  fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_PROGRESS_HANDLE_SOCK_EVENT);
	return mpi_errno;

 fn_fail:
	goto fn_exit;
}

/* shmem functions */
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_Get_business_card
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int ssmMPIDI_CH3I_Get_business_card(wchar_t *value, int length)
{
	int mpi_errno;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_GET_BUSINESS_CARD);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_GET_BUSINESS_CARD);

	mpi_errno = MPIDI_CH3U_Get_business_card_sock(&value, &length);
	if (mpi_errno != MPI_SUCCESS) {
		MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER,L"**buscard");
	}

	mpi_errno = MPIDI_CH3U_Get_business_card_sshm(&value, &length);
	if (mpi_errno != MPI_SUCCESS) {
		MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER,L"**buscard");
	}

 fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_GET_BUSINESS_CARD);
	return mpi_errno;
}

int ssmMPIDI_CH3_Init_fns(MPIDI_CH3_Functions_t *function_table)
{
	const wchar_t *env_var;

	function_table->Init = ssmMPIDI_CH3_Init;
	function_table->PortFnsInit = ssmMPIDI_CH3_PortFnsInit;
	function_table->RMAFnsInit = MPIDI_CH3U_RMAFnsInit;
	function_table->Finalize = ssmMPIDI_CH3_Finalize;
	function_table->VC_Init = ssmMPIDI_CH3_VC_Init;
	function_table->PG_Init = NULL;
	function_table->iStartMsg = ssmMPIDI_CH3_iStartMsg;
	function_table->iStartMsgv = ssmMPIDI_CH3_iStartMsgv;
	function_table->iSend = ssmMPIDI_CH3_iSend;
	function_table->iSendv = ssmMPIDI_CH3_iSendv;
	function_table->Connection_terminate = ssmMPIDI_CH3_Connection_terminate;
	function_table->Connect_to_root = ssmMPIDI_CH3_Connect_to_root;
	function_table->Complete_unidirectional_connection_connect = MPIDI_CH3_Complete_unidirectional_connection_connect;
	function_table->Complete_unidirectional_connection_accept = MPIDI_CH3_Complete_unidirectional_connection_accept;
	function_table->Cleanup_after_connection = MPIDI_CH3_Cleanup_after_connection;
	function_table->iStartRndvMsg = ssmMPIDI_CH3_iStartRndvMsg;
	function_table->iStartRndvTransfer = ssmMPIDI_CH3_iStartRndvTransfer;
	function_table->Get_business_card = ssmMPIDI_CH3I_Get_business_card;
	function_table->VC_NeedsSendClose = MPIDI_CH3U_VC_NeedsSendClose;
	function_table->VC_SendClose = MPIDI_CH3U_VC_SendClose;

	env_var = MPIU_GetStaticEnv(L"SSM_Progress");
	if (env_var == NULL)
	{
		/* Default progress engine */
		function_table->Progress_wait = fixed_spinMPIDI_CH3_Progress_wait;
		function_table->Progress_test = fixed_spinMPIDI_CH3_Progress_test;
	}
	else
	{
		if (_wcsicmp(env_var, L"fixed_spin") == 0)
		{
			function_table->Progress_wait = fixed_spinMPIDI_CH3_Progress_wait;
			function_table->Progress_test = fixed_spinMPIDI_CH3_Progress_test;
		}
		else if (_wcsicmp(env_var, L"adaptive") == 0)
		{
#ifdef USE_ADAPTIVE_PROGRESS_ENGINE
			function_table->Progress_wait = adaptiveMPIDI_CH3_Progress_wait;
			function_table->Progress_test = adaptiveMPIDI_CH3_Progress_test;
#else
			/* Default progress engine */
			function_table->Progress_wait = fixed_spinMPIDI_CH3_Progress_wait;
			function_table->Progress_test = fixed_spinMPIDI_CH3_Progress_test;
#endif
		}
		else if (_wcsicmp(env_var, L"fixed_active") == 0)
		{
#ifdef USE_FIXED_ACTIVE_PROGRESS_ENGINE
			function_table->Progress_wait = fixed_activeMPIDI_CH3_Progress_wait;
			function_table->Progress_test = fixed_activeMPIDI_CH3_Progress_test;
#else
			/* Default progress engine */
			function_table->Progress_wait = fixed_spinMPIDI_CH3_Progress_wait;
			function_table->Progress_test = fixed_spinMPIDI_CH3_Progress_test;
#endif
		}
		else
		{
			/* Default progress engine */
			function_table->Progress_wait = fixed_spinMPIDI_CH3_Progress_wait;
			function_table->Progress_test = fixed_spinMPIDI_CH3_Progress_test;
		}
	}
	return MPI_SUCCESS;
}
