/*  (C) 2009 by Deino Software. */
#if !defined(MPICH_MPIDI_CH3_IMPL_H_INCLUDED)
#define MPICH_MPIDI_CH3_IMPL_H_INCLUDED

/* Includes */
#include "mpidi_ch3_conf.h"
#include "mpidimpl.h"
#include "ch3usock.h"
#include "mpidu_process_locks.h"

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_ASSERT_H
#include <assert.h>
#endif
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif
#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif
#ifdef HAVE_PROCESS_H
#include <process.h>
#endif
#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif

/* Definitions */
/* This cycle counter is the read time stamp (rdtsc) instruction with Microsoft asm */
#ifdef HAVE_MASM_AND_X86
#define MPID_CPU_TICK(var_ptr) \
{ \
	register int *f1 = (int*)var_ptr; \
	__asm cpuid \
	__asm rdtsc \
	__asm mov ecx, f1 \
	__asm mov [ecx], eax \
	__asm mov [ecx + TYPE int], edx \
}
typedef unsigned __int64 MPID_CPU_Tick_t;

#define USE_ADAPTIVE_PROGRESS_ENGINE
#define USE_FIXED_ACTIVE_PROGRESS_ENGINE

#define MPID_WRITE_BARRIER()
#define MPID_READ_BARRIER() __asm { __asm _emit 0x0f __asm _emit 0xae __asm _emit 0xe8 }
#define MPID_READ_WRITE_BARRIER()

#else

#undef USE_ADAPTIVE_PROGRESS_ENGINE
#undef USE_FIXED_ACTIVE_PROGRESS_ENGINE
#define MPID_WRITE_BARRIER()
#define MPID_READ_BARRIER()
#define MPID_READ_WRITE_BARRIER()
#endif

#define MPIDI_CH3I_NUM_PACKETS			16
#define MPIDI_CH3I_PACKET_SIZE			(16*1024)
#define MPIDI_CH3I_PKT_AVAILABLE		0
#define MPIDI_CH3I_PKT_USED				1
#define MPIDI_CH3I_PKT_EMPTY			0
#define MPIDI_CH3I_PKT_FILLED			1
#define MPIDI_CH3I_SPIN_COUNT_DEFAULT   250
#define MPIDI_CH3I_YIELD_COUNT_DEFAULT	5000
#define MPIDI_CH3I_MSGQ_ITERATIONS		250
#define sshmMPIDI_SHM_EAGER_LIMIT		(128*1024)
#define sshmMPIDI_SHM_RNDV_LIMIT		(15*1024)
#define MPIDI_SHM_EAGER_LIMIT			10240
#define MPIDI_SHM_RNDV_LIMIT			10240
#define MPID_SHMEM_PER_PROCESS			1048576

#define MPIDU_MAX_SHM_BLOCK_SIZE ((unsigned int)2*1024*1024*1024)

#define SHM_READING_BIT	 0x0008
#define SHM_WRITING_BIT	 0x0010

#ifndef MAXHOSTNAMELEN
#define MAXHOSTNAMELEN 256
#endif

#define MPIDI_CH3I_SHM_HOST_KEY			L"shm_host"
#define MPIDI_CH3I_SHM_QUEUE_KEY		L"shm_queue"
#define MPIDI_CH3I_SHM_QUEUE_NAME_KEY	L"shm_name"
#define MPIDI_CH3I_SHM_PID_KEY			L"shm_pid"

#define MPIDI_BOOTSTRAP_NAME_LEN 100
#define BOOTSTRAP_MAX_NUM_MSGS 2048
#define BOOTSTRAP_MAX_MSG_SIZE sizeof(MPIDI_CH3I_Shmem_queue_info)

#define SHM_FAIL		-1

#define MPID_CH3I_SHM_BIT	0x01
#define MPID_CH3I_SOCK_BIT	0x02

/* USE_FIXED_ACTIVE_PROGRESS
#define MPIDI_INC_WRITE_ACTIVE() MPIDI_CH3I_shm_write_active++; MPIDI_CH3I_active_flag |= MPID_CH3I_SHM_BIT
#define MPIDI_DEC_WRITE_ACTIVE() MPIDI_CH3I_shm_write_active--
*/
#define MPIDI_INC_WRITE_ACTIVE()
#define MPIDI_DEC_WRITE_ACTIVE()

/* Structures */

/* This structure uses the avail field to signal that the data is available for reading.
   The code fills the data and then sets the avail field.
   This assumes that declaring avail to be volatile causes the compiler to insert a
   write barrier when the avail location is written to.
   */
typedef struct MPIDI_CH3I_SHM_Packet_t
{
	volatile int avail;
	/*char pad_avail[60];*/ /* keep the avail flag on a separate cache line */ /* This sounds good in theory but in practice degrades performance. Analysis needed. */
	int num_bytes;
	int offset;
	/* insert stuff here to align data? */
	int pad;/*char pad_data[56];*/ /* cache align the data */
	char data[MPIDI_CH3I_PACKET_SIZE];
	/* insert stuff here to pad the packet up to an aligned boundary? */
} MPIDI_CH3I_SHM_Packet_t;

typedef struct MPIDI_CH3I_SHM_Queue_t
{
	int head_index;
	int tail_index;
	/* insert stuff here to align the packets? */
	/*char pad[56];*/ /* cache align the data */
	MPIDI_CH3I_SHM_Packet_t packet[MPIDI_CH3I_NUM_PACKETS];
} MPIDI_CH3I_SHM_Queue_t;

typedef struct MPIDI_CH3I_Process_s
{
	MPIDI_VC_t *shm_reading_list, *shm_writing_list;
	int num_cpus;
	MPIDI_VC_t *unex_finished_list, *vc;
}
MPIDI_CH3I_Process_t;

typedef struct MPIDI_CH3I_Shmem_queue_info
{
	int pg_rank, pid;
	wchar_t pg_id[100];
	int type;
	MPIDI_CH3I_Shmem_block_request_result info;
} MPIDI_CH3I_Shmem_queue_info;

typedef enum shm_wait_e
{
SHM_WAIT_TIMEOUT = 10101010,
SHM_WAIT_READ,
SHM_WAIT_WRITE,
SHM_WAIT_WAKEUP
} shm_wait_t;

/* Global Variables */

extern MPIDI_CH3I_Process_t MPIDI_CH3I_Process;
extern int MPIDI_CH3I_active_flag;
extern int MPIDI_CH3I_shm_read_active, MPIDI_CH3I_shm_write_active;
extern int MPIDI_CH3I_sock_read_active, MPIDI_CH3I_sock_write_active;
extern volatile unsigned int MPIDI_CH3I_progress_completion_count;
extern MPIDU_Sock_set_t MPIDI_CH3I_sock_set;
/* Count the number of outstanding close requests */
extern volatile int MPIDI_Outstanding_close_ops;

/* Interface Macros */

/* Socket connection definitions */

/*#define USE_VALIDATING_QUEUE_MACROS*/
#ifdef USE_VALIDATING_QUEUE_MACROS

#define MPIDI_CH3I_SendQ_enqueue(vc, req)								\
{																		\
	MPID_Request *iter;														\
	iter = vc->ch.sendq_head;												\
	while (iter) {														\
		if (iter == req) {												\
			/*abort();*/ break; /*wprintf(L"Error: enqueueing request twice.\n");fflush(stdout);*/\
		}																\
		iter = iter->dev.next;												\
	}																		\
	if (!iter) {														\
	/* MT - not thread safe! */												\
	MPIDI_DBG_PRINTF((50, FCNAME, L"SendQ_enqueue vc=%p req=0x%08x", vc, req->handle));\
	req->dev.next = NULL;												\
	if (vc->ch.sendq_tail != NULL)										\
	{																		\
		vc->ch.sendq_tail->dev.next = req;								\
	}																		\
	else																\
	{																		\
		/* increment number of active writes when posting to an empty queue */\
		MPIDI_INC_WRITE_ACTIVE();										\
		vc->ch.sendq_head = req;										\
	}																		\
	vc->ch.sendq_tail = req;												\
	}																		\
}

#define MPIDI_CH3I_SendQ_enqueue_head(vc, req)								\
{																		\
	MPID_Request *iter;														\
	iter = vc->ch.sendq_head;												\
	while (iter) {														\
		if (iter == req) {												\
			/*abort();*/ break; /*wprintf(L"Error: head enqueueing request twice.\n");fflush(stdout);*/\
		}																\
		iter = iter->dev.next;												\
	}																		\
	/* MT - not thread safe! */												\
	if (!iter) {														\
	MPIDI_DBG_PRINTF((50, FCNAME, L"SendQ_enqueue_head vc=%p req=0x%08x", vc, req->handle));\
	req->dev.next = vc->ch.sendq_head; /*sendq_tail;*/						\
	if (vc->ch.sendq_tail == NULL)										\
	{																		\
		/* increment number of active writes when posting to an empty queue */\
		MPIDI_INC_WRITE_ACTIVE();										\
		vc->ch.sendq_tail = req;										\
	}																		\
	vc->ch.sendq_head = req;												\
	}																		\
}

#else

#define MPIDI_CH3I_SendQ_enqueue(vc, req)								\
{																		\
	/* MT - not thread safe! */												\
	MPIDI_DBG_PRINTF((50, FCNAME, L"SendQ_enqueue vc=%p req=0x%08x", vc, req->handle));		\
	req->dev.next = NULL;												\
	if (vc->ch.sendq_tail != NULL)										\
	{																		\
		vc->ch.sendq_tail->dev.next = req;								\
	}																		\
	else																\
	{																		\
		/* increment number of active writes when posting to an empty queue */\
		MPIDI_INC_WRITE_ACTIVE();										\
		vc->ch.sendq_head = req;										\
	}																		\
	vc->ch.sendq_tail = req;												\
}

#define MPIDI_CH3I_SendQ_enqueue_head(vc, req)								\
{																		\
	/* MT - not thread safe! */												\
	MPIDI_DBG_PRINTF((50, FCNAME, L"SendQ_enqueue_head vc=%p req=0x%08x", vc, req->handle));\
	req->dev.next = vc->ch.sendq_head;										\
	if (vc->ch.sendq_tail == NULL)										\
	{																		\
		/* increment number of active writes when posting to an empty queue */\
		MPIDI_INC_WRITE_ACTIVE();										\
		vc->ch.sendq_tail = req;										\
	}																		\
	vc->ch.sendq_head = req;												\
}

#endif

#define MPIDI_CH3I_SendQ_dequeue(vc)										\
{																		\
	/* MT - not thread safe! */												\
	MPIDI_DBG_PRINTF((50, FCNAME, L"SendQ_dequeue vc=%p req=0x%08x", vc, vc->ch.sendq_head->handle));\
	vc->ch.sendq_head = vc->ch.sendq_head->dev.next;						\
	if (vc->ch.sendq_head == NULL)										\
	{																		\
		/* decrement number of active writes when a queue becomes empty */\
		MPIDI_DEC_WRITE_ACTIVE();										\
		vc->ch.sendq_tail = NULL;										\
	}																		\
}

#define MPIDI_CH3I_SendQ_head(vc) (vc->ch.sendq_head)

#define MPIDI_CH3I_SendQ_empty(vc) (vc->ch.sendq_head == NULL)

/* End of socket connection-related macros */

/* Function Prototypes */

void MPIDI_CH3U_Generate_shm_string(wchar_t *str);

int MPIDI_CH3U_RMAFnsInit( MPIDI_RMAFns *a );

int MPIDI_CH3I_VC_post_connect(MPIDI_VC_t *);
int MPIDI_CH3I_Connection_alloc(MPIDI_CH3I_Connection_t **);
int MPIDI_CH3I_Request_adjust_iov(MPID_Request *, MPIDI_msg_sz_t);
int MPIDI_CH3I_Shm_connect(MPIDI_VC_t *, const wchar_t *, int *);
int MPIDI_CH3I_SSM_VC_post_read(MPIDI_VC_t *, MPID_Request *);
int MPIDI_CH3I_SSM_VC_post_write(MPIDI_VC_t *, MPID_Request *);
int MPIDI_CH3I_Handle_shm_read(MPIDI_VC_t *vc, int nb);
int MPIDI_CH3I_SHM_write_progress(MPIDI_VC_t * vc);

int MPIDI_CH3I_SHM_post_read(MPIDI_VC_t *vc, void *buf, int len, int (*read_progress_update)(int, void*));
int MPIDI_CH3I_SHM_post_readv(MPIDI_VC_t *vc, MPID_IOV *iov, int n, int (*read_progress_update)(int, void*));
int MPIDI_CH3I_SHM_write(MPIDI_VC_t *vc, void *buf, int len, int *num_bytes_ptr);
int MPIDI_CH3I_SHM_writev(MPIDI_VC_t *vc, MPID_IOV *iov, int n, int *num_bytes_ptr);
int MPIDI_CH3I_SHM_rdma_readv(MPIDI_VC_t *vc, MPID_Request *rreq);
int MPIDI_CH3I_SHM_rdma_writev(MPIDI_VC_t *vc, MPID_Request *sreq);

int MPIDI_CH3I_BootstrapQ_create_unique_name(wchar_t *name, int length);
int MPIDI_CH3I_BootstrapQ_create_named(MPIDI_CH3I_BootstrapQ *queue_ptr, const wchar_t *name, const int initialize);
int MPIDI_CH3I_BootstrapQ_create(MPIDI_CH3I_BootstrapQ *queue_ptr);
int MPIDI_CH3I_BootstrapQ_tostring(MPIDI_CH3I_BootstrapQ queue, wchar_t *name, int length);
int MPIDI_CH3I_BootstrapQ_destroy(MPIDI_CH3I_BootstrapQ queue);
int MPIDI_CH3I_BootstrapQ_unlink(MPIDI_CH3I_BootstrapQ queue);
int MPIDI_CH3I_BootstrapQ_attach(wchar_t *name, MPIDI_CH3I_BootstrapQ * queue_ptr);
int MPIDI_CH3I_BootstrapQ_detach(MPIDI_CH3I_BootstrapQ queue);
int MPIDI_CH3I_BootstrapQ_send_msg(MPIDI_CH3I_BootstrapQ queue, void *buffer, int length);
int MPIDI_CH3I_BootstrapQ_recv_msg(MPIDI_CH3I_BootstrapQ queue, void *buffer, int length, int *num_bytes_ptr, BOOL blocking);

int MPIDI_CH3I_SHM_Get_mem(MPIDI_PG_t *pg, int nTotalSize, int nRank, int nNproc, BOOL bUseShm);
int MPIDI_CH3I_SHM_Release_mem(MPIDI_PG_t *pg, BOOL bUseShm);
int MPIDI_CH3I_SHM_Unlink_and_detach_mem(MPIDI_CH3I_Shmem_block_request_result *p);
int MPIDI_CH3I_SHM_Attach_notunlink_mem(MPIDI_CH3I_Shmem_block_request_result *pInput);
int MPIDI_CH3I_SHM_Get_memblock(int size, MPIDI_CH3I_Shmem_block_request_result *pOutput);
int MPIDI_CH3I_SHM_Get_memblock_named(int size, MPIDI_CH3I_Shmem_block_request_result *pInOutput);
int MPIDI_CH3I_SHM_Attach_to_memblock(MPIDI_CH3I_Shmem_block_request_result *pInput, MPIDI_CH3I_Shmem_block_request_result *pOutput);
int MPIDI_CH3I_SHM_Unlink_memblock(MPIDI_CH3I_Shmem_block_request_result *p);
int MPIDI_CH3I_SHM_Release_memblock(MPIDI_CH3I_Shmem_block_request_result *p);
int MPIDI_CH3I_SHM_read_progress(MPIDI_VC_t *vc, int millisecond_timeout, MPIDI_VC_t **vc_pptr, int *num_bytes_ptr);
int MPIDI_CH3I_SHM_post_read(MPIDI_VC_t *vc, void *buf, int len, int (*read_progress_update)(int, void*));
int MPIDI_CH3I_SHM_post_readv(MPIDI_VC_t *vc, MPID_IOV *iov, int n, int (*read_progress_update)(int, void*));
int MPIDI_CH3I_SHM_read(MPIDI_VC_t * vc, void *buf, int len, int *num_bytes_ptr);
int MPIDI_CH3I_SHM_write(MPIDI_VC_t *vc, void *buf, int len, int *num_bytes_ptr);
int MPIDI_CH3I_SHM_writev(MPIDI_VC_t *vc, MPID_IOV *iov, int n, int *num_bytes_ptr);
int MPIDI_CH3I_SHM_rdma_readv(MPIDI_VC_t *vc, MPID_Request *rreq);
int MPIDI_CH3I_SHM_rdma_writev(MPIDI_VC_t *vc, MPID_Request *sreq);
void MPIDI_CH3I_SHM_Remove_vc_references(MPIDI_VC_t *vc);
void MPIDI_CH3I_SHM_Add_to_reader_list(MPIDI_VC_t *vc);
void MPIDI_CH3I_SHM_Add_to_writer_list(MPIDI_VC_t *vc);
void MPIDI_CH3I_SHM_Remove_vc_write_references(MPIDI_VC_t *vc);

MPIDI_CH3I_Alloc_mem_list_t * MPIDI_CH3I_Get_mem_list_head(void);

int MPIDI_CH3_Complete_unidirectional_connection_connect( MPIDI_VC_t *connect_vc );
int MPIDI_CH3_Complete_unidirectional_connection_accept( MPIDI_VC_t *new_vc );
int MPIDI_CH3_Cleanup_after_connection( MPIDI_VC_t *new_vc );

#endif /* !defined(MPICH_MPIDI_CH3_IMPL_H_INCLUDED) */
