/*  (C) 2009 by Deino Software. */
#include "mpidi_ch3_impl.h"
#include <winsock2.h>
#include <windows.h>
#include <time.h>

typedef struct bootstrap_msg
{
	int length;
	char buffer[1024];
	struct bootstrap_msg *next;
} bootstrap_msg;

typedef struct MPIDI_CH3I_BootstrapQ_struct
{
	wchar_t name[MPIDI_BOOTSTRAP_NAME_LEN];
	bootstrap_msg *msg_list;
	HWND hWnd;
	HANDLE hMsgThread;
	HANDLE hReadyEvent;
	HANDLE hMutex;
	HANDLE hMessageArrivedEvent;
	int error;
	struct MPIDI_CH3I_BootstrapQ_struct *next;
} MPIDI_CH3I_BootstrapQ_struct;

static MPIDI_CH3I_BootstrapQ_struct * g_queue_list = NULL;

static MPIDI_CH3I_BootstrapQ s_queue = NULL;

#undef FUNCNAME
#define FUNCNAME GetNextBootstrapMsg
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int GetNextBootstrapMsg(MPIDI_CH3I_BootstrapQ queue, bootstrap_msg ** msg_ptr, BOOL blocking)
{
	MPIDI_STATE_DECL(MPID_STATE_GET_NEXT_BOOTSTRAP_MSG);
	MPIDI_FUNC_ENTER(MPID_STATE_GET_NEXT_BOOTSTRAP_MSG);
	while (WaitForSingleObject(queue->hMessageArrivedEvent, blocking ? INFINITE : 0) == WAIT_OBJECT_0)
	{
		WaitForSingleObject(queue->hMutex, INFINITE);
		if (queue->msg_list)
		{
			*msg_ptr = queue->msg_list;
			queue->msg_list = queue->msg_list->next;
			if (queue->msg_list == NULL)
				ResetEvent(queue->hMessageArrivedEvent);
			ReleaseMutex(queue->hMutex);
			MPIDI_FUNC_EXIT(MPID_STATE_GET_NEXT_BOOTSTRAP_MSG);
			return MPI_SUCCESS;
		}
		ReleaseMutex(queue->hMutex);
	}
	*msg_ptr = NULL;
	MPIDI_FUNC_EXIT(MPID_STATE_GET_NEXT_BOOTSTRAP_MSG);
	return MPI_SUCCESS;
}

#undef FUNCNAME
#define FUNCNAME BootstrapQWndProc
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
LRESULT CALLBACK BootstrapQWndProc(
	HWND hwnd,		/* handle to window */
	UINT uMsg,		/* message identifier */
	WPARAM wParam,	/* first message parameter */
	LPARAM lParam)	/* second message parameter */
{
	PCOPYDATASTRUCT p;
	bootstrap_msg *bsmsg_ptr, *msg_iter;
	/*MPIDI_STATE_DECL(MPID_STATE_BOOTSTRAPQWNDPROC);
	MPIDI_FUNC_ENTER(MPID_STATE_BOOTSTRAPQWNDPROC);*/
	switch (uMsg)
	{
	case WM_DESTROY:
		PostQuitMessage(0);
		break;
	case WM_COPYDATA:
		/*wprintf(L"WM_COPYDATA received\n");fflush(stdout);*/
		p = (PCOPYDATASTRUCT) lParam;
		if (WaitForSingleObject(s_queue->hMutex, INFINITE) == WAIT_OBJECT_0)
		{
			bsmsg_ptr = MPIU_Malloc(sizeof(bootstrap_msg));
			if (bsmsg_ptr != NULL)
			{
				memcpy(bsmsg_ptr->buffer, p->lpData, p->cbData);
				bsmsg_ptr->length = p->cbData;
				bsmsg_ptr->next = NULL;
				msg_iter = s_queue->msg_list;
				if (msg_iter == NULL)
					s_queue->msg_list = bsmsg_ptr;
				else
				{
					while (msg_iter->next != NULL)
						msg_iter = msg_iter->next;
					msg_iter->next = bsmsg_ptr;
				}
				/*wprintf(L"bootstrap message received: %d bytes\n", bsmsg_ptr->length);fflush(stdout);*/
				SetEvent(s_queue->hMessageArrivedEvent);
			}
			else
			{
				MPIU_Error_printf(L"MPIU_Malloc failed\n");
			}
			ReleaseMutex(s_queue->hMutex);
		}
		else
		{
			MPIU_Error_printf(L"Error waiting for s_queue mutex, error %d\n", GetLastError());
		}
		/*MPIDI_FUNC_EXIT(MPID_STATE_BOOTSTRAPQWNDPROC);*/
		return 101;
	default:
		/*MPIDI_FUNC_EXIT(MPID_STATE_BOOTSTRAPQWNDPROC);*/
		return DefWindowProc(hwnd, uMsg, wParam, lParam);
	}
	/*MPIDI_FUNC_EXIT(MPID_STATE_BOOTSTRAPQWNDPROC);*/
	return 0;
}

void MessageQueueThreadFn(MPIDI_CH3I_BootstrapQ_struct *queue)
{
	MSG msg;
	BOOL bRet;
	WNDCLASS wc;
	UUID guid;

	UuidCreate(&guid);
	MPIU_Snprintf(queue->name, MPIDI_BOOTSTRAP_NAME_LEN, _TRUNCATE, L"%08lX-%04X-%04x-%02X%02X-%02X%02X%02X%02X%02X%02X",
		guid.Data1, guid.Data2, guid.Data3,
		guid.Data4[0], guid.Data4[1], guid.Data4[2], guid.Data4[3],
		guid.Data4[4], guid.Data4[5], guid.Data4[6], guid.Data4[7]);

	wc.style = 0;
	wc.lpfnWndProc = (WNDPROC) BootstrapQWndProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = NULL;
	wc.hIcon = LoadIcon((HINSTANCE) NULL, IDI_APPLICATION);
	wc.hCursor = LoadCursor((HINSTANCE) NULL, IDC_ARROW);
	wc.hbrBackground = GetStockObject(WHITE_BRUSH);
	wc.lpszMenuName =  L"MainMenu";
	wc.lpszClassName = queue->name;
	
	if (!RegisterClass(&wc))
	{
		queue->error = GetLastError();
		SetEvent(queue->hReadyEvent);
		return;
	}

	/* Create the hidden window. */
 
	queue->hWnd = CreateWindow(
		queue->name, /* window class */
		queue->name, /* window name */
		WS_OVERLAPPEDWINDOW,
		CW_USEDEFAULT,
		CW_USEDEFAULT,
		CW_USEDEFAULT,
		CW_USEDEFAULT,
		(HWND) HWND_MESSAGE, 
		(HMENU) NULL,
		0,
		(LPVOID) NULL); 
 
	if (!queue->hWnd)
	{
		queue->error = GetLastError();
		SetEvent(queue->hReadyEvent);
		return;
	}
 
	ShowWindow(queue->hWnd, SW_HIDE); 
 
	/* initialize the synchronization objects */
	queue->msg_list = NULL;
	queue->hMessageArrivedEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
	queue->hMutex = CreateMutex(NULL, FALSE, NULL);

	if (s_queue != NULL)
	{
		MPIU_Error_printf(L"Error: more than one message queue created\n");
		return;
	}
	s_queue = queue;

	/*wprintf(L"signalling queue is ready\n");fflush(stdout);*/
	/* signal that the queue is ready */
	SetEvent(queue->hReadyEvent);

	/* Start handling messages. */
	while( (bRet = GetMessage( &msg, NULL, 0, 0 )) != 0)
	{ 
		if (bRet == -1)
		{
			MPIU_Error_printf(L"MessageQueueThreadFn window received a WM_QUIT message, exiting...\n");
			return;
		}
		else
		{
			TranslateMessage(&msg);
			DispatchMessage(&msg);
			/*wprintf(L"window message: %d\n", msg.message);fflush(stdout);*/
		}
	} 
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_BootstrapQ_create_unique_name
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3I_BootstrapQ_create_unique_name(wchar_t *name, int length)
{
	UUID guid;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_BOOTSTRAPQ_CREATE_UNIQUE_NAME);
	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_BOOTSTRAPQ_CREATE_UNIQUE_NAME);
	if (length < 40)
	{
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_BOOTSTRAPQ_CREATE_UNIQUE_NAME);
		return -1;
	}
	UuidCreate(&guid);
	MPIU_Snprintf(name, length, _TRUNCATE, L"%08lX-%04X-%04x-%02X%02X-%02X%02X%02X%02X%02X%02X",
		guid.Data1, guid.Data2, guid.Data3,
		guid.Data4[0], guid.Data4[1], guid.Data4[2], guid.Data4[3],
		guid.Data4[4], guid.Data4[5], guid.Data4[6], guid.Data4[7]);
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_BOOTSTRAPQ_CREATE_UNIQUE_NAME);
	return MPI_SUCCESS;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_BootstrapQ_create_named
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3I_BootstrapQ_create_named(MPIDI_CH3I_BootstrapQ *queue_ptr, const wchar_t *name, const int initialize)
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_BOOTSTRAPQ_CREATE_NAMED);
	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_BOOTSTRAPQ_CREATE_NAMED);

	MPIU_UNREFERENCED_ARG(queue_ptr);
	MPIU_UNREFERENCED_ARG(name);
	MPIU_UNREFERENCED_ARG(initialize);
	mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**notimpl", 0);
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_BOOTSTRAPQ_CREATE_NAMED);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_BootstrapQ_create
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3I_BootstrapQ_create(MPIDI_CH3I_BootstrapQ *queue_ptr)
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_CH3I_BootstrapQ_struct *queue;
	DWORD dwThreadID;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_BOOTSTRAPQ_CREATE);
	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_BOOTSTRAPQ_CREATE);

	/* allocate a queue structure and add it to the global list */
	queue = (MPIDI_CH3I_BootstrapQ_struct*)MPIU_Malloc(sizeof(MPIDI_CH3I_BootstrapQ_struct));
	if (queue == NULL)
	{
		mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_BOOTSTRAPQ_CREATE);
		return mpi_errno;
	}
	queue->next = g_queue_list;
	g_queue_list = queue;

	queue->hReadyEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
	queue->hMsgThread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)MessageQueueThreadFn, queue, 0, &dwThreadID);
	if (queue->hMsgThread == NULL)
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**CreateThread", L"**CreateThread %d", GetLastError());
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_BOOTSTRAPQ_CREATE);
		return mpi_errno;
	}
	if (WaitForSingleObject(queue->hReadyEvent, 60000) != WAIT_OBJECT_0)
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**winwait", 0);
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_BOOTSTRAPQ_CREATE);
		return mpi_errno;
	}
	CloseHandle(queue->hReadyEvent);
	queue->hReadyEvent = NULL;

	*queue_ptr = queue;

	MPIU_DBG_PRINTF((L"Created bootstrap queue: %s\n", queue->name));
	/*wprintf(L"Created bootstrap queue: %s\n", queue->name); fflush(stdout);*/

	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_BOOTSTRAPQ_CREATE);
	return MPI_SUCCESS;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_BootstrapQ_tostring
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3I_BootstrapQ_tostring(MPIDI_CH3I_BootstrapQ queue, wchar_t *name, int length)
{
	int mpi_errno;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_BOOTSTRAPQ_TOSTRING);
	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_BOOTSTRAPQ_TOSTRING);
	
	if (queue == NULL)
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", 0);
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_BOOTSTRAPQ_TOSTRING);
		return mpi_errno;
	}

	/*wprintf(L"[%d] queue->name = %s\n", MPIR_Process.comm_world->rank, queue->name);fflush(stdout);*/
	mpi_errno = MPIU_Wcsncpy(name, queue->name, length);
	if (mpi_errno)
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**arg", 0);
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_BOOTSTRAPQ_TOSTRING);
		return mpi_errno;
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_BOOTSTRAPQ_TOSTRING);
	return MPI_SUCCESS;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_BootstrapQ_unlink
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3I_BootstrapQ_unlink(MPIDI_CH3I_BootstrapQ queue)
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_BOOTSTRAPQ_UNLINK);
	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_BOOTSTRAPQ_UNLINK);
	MPIU_UNREFERENCED_ARG(queue);
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_BOOTSTRAPQ_UNLINK);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_BootstrapQ_destroy
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3I_BootstrapQ_destroy(MPIDI_CH3I_BootstrapQ queue)
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_BOOTSTRAPQ_DESTROY);
	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_BOOTSTRAPQ_DESTROY);
	PostMessage(queue->hWnd, WM_DESTROY, 0, 0);
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_BOOTSTRAPQ_DESTROY);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_BootstrapQ_attach
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3I_BootstrapQ_attach(wchar_t *name, MPIDI_CH3I_BootstrapQ * queue_ptr)
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_CH3I_BootstrapQ_struct *iter = g_queue_list;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_BOOTSTRAPQ_ATTACH);
	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_BOOTSTRAPQ_ATTACH);
	while (iter)
	{
		if (wcscmp(iter->name, name) == 0)
		{
			*queue_ptr = iter;
			MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_BOOTSTRAPQ_ATTACH);
			return MPI_SUCCESS;
		}
		iter = iter->next;
	}
	iter = (MPIDI_CH3I_BootstrapQ_struct*)MPIU_Malloc(sizeof(MPIDI_CH3I_BootstrapQ_struct));
	if (iter == NULL)
	{
		mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_BOOTSTRAPQ_ATTACH);
		return mpi_errno;
	}
	MPIU_Wcsncpy(iter->name, name, MPIDI_BOOTSTRAP_NAME_LEN);
	/*wprintf(L"looking for window %s\n", name);fflush(stdout);*/
	iter->hWnd = FindWindowEx(HWND_MESSAGE, NULL, name, name);
	/*if (iter->hWnd != NULL) { wprintf(L"FindWindowEx found the window\n"); fflush(stdout); }*/
	if (iter->hWnd == NULL)
	{
		iter->hWnd = FindWindow(name, name);
		/*if (iter->hWnd != NULL) { wprintf(L"FindWindow found the window\n"); fflush(stdout); }*/
	}
	if (iter->hWnd == NULL)
	{
		mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**FindWindowEx", L"**FindWindowEx %d", GetLastError());
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_BOOTSTRAPQ_ATTACH);
		return mpi_errno;
	}
	iter->hMutex = CreateMutex(NULL, FALSE, NULL);
	iter->hReadyEvent = NULL;
	iter->msg_list = NULL;
	iter->hMsgThread = NULL;
	iter->hMessageArrivedEvent = NULL;
	iter->next = g_queue_list;
	/*iter->next = NULL;*/
	*queue_ptr = iter;

	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_BOOTSTRAPQ_ATTACH);
	return MPI_SUCCESS;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_BootstrapQ_detach
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3I_BootstrapQ_detach(MPIDI_CH3I_BootstrapQ queue)
{
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_BOOTSTRAPQ_DETACH);
	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_BOOTSTRAPQ_DETACH);
	/* remove the queue from the global list? */
	MPIU_UNREFERENCED_ARG(queue);
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_BOOTSTRAPQ_DETACH);
	return MPI_SUCCESS;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_BootstrapQ_send_msg
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3I_BootstrapQ_send_msg(MPIDI_CH3I_BootstrapQ queue, void *buffer, int length)
{
	int mpi_errno = MPI_SUCCESS;
	COPYDATASTRUCT data;
	LRESULT rc;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_BOOTSTRAPQ_SEND_MSG);
	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_BOOTSTRAPQ_SEND_MSG);
	data.dwData = 0; /* immediate data field */
	data.cbData = length;
	data.lpData = buffer;
	rc = SendMessage(queue->hWnd, WM_COPYDATA, 0, (LPARAM)&data);
	/*wprintf(L"SendMessage returned %d\n", rc);fflush(stdout);*/

	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_BOOTSTRAPQ_SEND_MSG);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_BootstrapQ_recv_msg
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3I_BootstrapQ_recv_msg(MPIDI_CH3I_BootstrapQ queue, void *buffer, int length, int *num_bytes_ptr, BOOL blocking)
{
	int mpi_errno = MPI_SUCCESS;
	bootstrap_msg * msg;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_BOOTSTRAPQ_RECV_MSG);
	MPIDI_STATE_DECL(MPID_STATE_MEMCPY);
	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_BOOTSTRAPQ_RECV_MSG);

	mpi_errno = GetNextBootstrapMsg(queue, &msg, blocking);
	if (mpi_errno != MPI_SUCCESS)
	{
		mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nextbootmsg", 0);
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_BOOTSTRAPQ_RECV_MSG);
		return mpi_errno;
	}
	if (msg)
	{
		MPIDI_FUNC_ENTER(MPID_STATE_MEMCPY);
		memcpy(buffer, msg->buffer, min(length, msg->length));
		MPIDI_FUNC_EXIT(MPID_STATE_MEMCPY);
		*num_bytes_ptr = min(length, msg->length);
		MPIU_Free(msg);
		/*wprintf(L"Received %d byte bootstrap queue message.\r\n", *num_bytes_ptr);fflush(stdout);*/
	}
	else
	{
		*num_bytes_ptr = 0;
	}

	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_BOOTSTRAPQ_RECV_MSG);
	return mpi_errno;
}
