/*   Copyright (C) 2009 Deino Software. */

#include <mpiimpl.h>
#include <mpid_dataloop.h>
#include <stdlib.h>

/*@
  MPID_Type_commit
 
  Input Parameters:
. datatype_p - pointer to MPI datatype

  Output Parameters:

  Return Value:
  0 on success, -1 on failure.
@*/
#undef FUNCNAME
#define FUNCNAME MPID_Type_commit
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPID_Type_commit(MPI_Datatype *datatype_p)
{
	int		   mpi_errno=MPI_SUCCESS;
	MPI_Aint	  first, last;
	MPID_Datatype *datatype_ptr;
	MPID_Segment *segp;
	MPIDI_STATE_DECL(MPID_STATE_MPID_TYPE_COMMIT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPID_TYPE_COMMIT);

	MPIU_Assert(HANDLE_GET_KIND(*datatype_p) != HANDLE_KIND_BUILTIN);

	MPID_Datatype_get_ptr(*datatype_p, datatype_ptr);

	if (datatype_ptr->is_committed == 0) {
		datatype_ptr->is_committed = 1;

		/* determine number of contiguous blocks in the type */
		segp = MPID_Segment_alloc();
		/* --BEGIN ERROR HANDLING-- */
		if (!segp)
		{
			mpi_errno = MPIR_Err_create_code(MPI_SUCCESS,
											 MPIR_ERR_RECOVERABLE,
											 L"MPID_Type_commit",
											 __LINE__,
											 MPI_ERR_OTHER,
											 L"**nomem",
											 0);
			MPIDI_FUNC_EXIT(MPID_STATE_MPID_TYPE_COMMIT);
			return mpi_errno;
		}
		/* --END ERROR HANDLING-- */
		MPID_Segment_init(0, 1, *datatype_p, segp, 0); /* first 0 is bufptr,
														* 1 is count
														* last 0 is homogeneous
														*/

		first = 0;
		last  = SEGMENT_IGNORE_LAST;

		MPID_Segment_free(segp);

		MPIU_DBG_PRINTF((L"# contig blocks = %d\n",
						 (int) datatype_ptr->n_contig_blocks));

#ifdef USE_MEMORY_TRACING
		if ( ((unsigned int)datatype_ptr->hetero_dloop & 0xFFFFFFFF ) == 0xefefefef) {
			mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_FATAL,
											  __FILE__, __LINE__, 
											  MPI_ERR_OTHER, "**fail",
											  "**fail %s", "hetero_dloop not set" );
		}
#endif
#if 0
		MPIDI_Dataloop_dot_printf(datatype_ptr->dataloop, 0, 1);
#endif
	}

	MPIDI_FUNC_EXIT(MPID_STATE_MPID_TYPE_COMMIT);
	return mpi_errno;
}
