/*   Copyright (C) 2009 Deino Software. */

#include <mpiimpl.h>
#include <mpid_dataloop.h>
#include <stdlib.h>

/*@
  MPIR_Type_get_contig_blocks
 
  Input Parameters:
. type - MPI Datatype (must have been committed)

  Output Parameters:
. nr_blocks_p - pointer to int in which to store the number of contiguous blocks in the type


  Return Value:
  0 on success, -1 on failure.
@*/

#undef FUNCNAME
#define FUNCNAME MPIR_Type_get_contig_blocks
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIR_Type_get_contig_blocks(MPI_Datatype type,
								int *nr_blocks_p)
{
	MPID_Datatype *datatype_ptr;
	MPIDI_STATE_DECL(MPID_STATE_MPIR_TYPE_GET_CONTIG_BLOCKS);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIR_TYPE_GET_CONTIG_BLOCKS);

	if (HANDLE_GET_KIND(type) == HANDLE_KIND_BUILTIN) {
		*nr_blocks_p = 1;
		MPIDI_FUNC_EXIT(MPID_STATE_MPIR_TYPE_GET_CONTIG_BLOCKS);
		return 0;
	}

	MPID_Datatype_get_ptr(type, datatype_ptr);
	MPIU_Assert(datatype_ptr->is_committed);

	*nr_blocks_p = datatype_ptr->n_contig_blocks;
	MPIDI_FUNC_EXIT(MPID_STATE_MPIR_TYPE_GET_CONTIG_BLOCKS);
	return 0;
}
