/*   Copyright (C) 2009 Deino Software. */

#include "mpidu_process_locks.h"

#ifdef USE_SLEEP_YIELD
int MPIDI_Sleep_yield_count = 0;
/* FIXME - add the appropriate backoff values here */
#ifdef HAVE_WIN32_SLEEP
int MPIDI_Sleep_yield_counts[10] = { 1, 2, 3, 4, 5, 6, 7, 8, 9, 1000 };
#elif defined(HAVE_USLEEP)
int MPIDI_Sleep_yield_counts[10] = { 1, 2, 3, 4, 5, 6, 7, 8, 9, 1000 };
#elif defined(HAVE_SELECT)
int MPIDI_Sleep_yield_counts[10] = { 1, 2, 3, 4, 5, 6, 7, 8, 9, 1000 };
#elif defined(HAVE_SLEEP)
int MPIDI_Sleep_yield_counts[10] = { 1, 2, 3, 4, 5, 6, 7, 8, 9, 1000 };
#else
#error *** No function available to implement sleep_yield ***
#endif
#endif

#include <errno.h>
#include <winsock2.h>
#include <windows.h>

int g_nLockSpinCount = 100;

#undef FUNCNAME
#define FUNCNAME MPIDU_Process_lock_init
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
/*@
   MPIDU_Process_lock_init - 

   Parameters:
+  MPIDU_Process_lock_t *lock

   Notes:
@*/
void MPIDU_Process_lock_init( MPIDU_Process_lock_t *lock )
{
	MPIDI_STATE_DECL(MPID_STATE_MPIDU_PROCESS_LOCK_INIT);
	MPIDI_FUNC_ENTER(MPID_STATE_MPIDU_PROCESS_LOCK_INIT);
	*lock = CreateMutex(NULL, FALSE, NULL);
	if (*lock == NULL)
	{
		MPIU_Error_printf(L"error in mutex_init: %d\n", GetLastError());
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDU_PROCESS_LOCK_INIT);
}

static void PrintWaitFailure()
{
	wchar_t error_msg[1024];
	_wcserror_s(error_msg, 1024, GetLastError());
	MPIU_Error_printf(L"error in mutex_lock: %s\n", error_msg);
}

#undef FUNCNAME
#define FUNCNAME MPIDU_Process_lock
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
/*@
   MPIDU_Process_lock - 

   Parameters:
+  MPIDU_Process_lock_t *lock

   Notes:
@*/
void MPIDU_Process_lock( MPIDU_Process_lock_t *lock )
{
	DWORD dwRetVal;
	MPIDI_STATE_DECL(MPID_STATE_MPIDU_PROCESS_LOCK);
	MPIDI_FUNC_ENTER(MPID_STATE_MPIDU_PROCESS_LOCK);
	/*wprintf(L"nt lock %x\n", lock);fflush(stdout);*/
	dwRetVal = WaitForSingleObject(*lock, INFINITE);
	if (dwRetVal != WAIT_OBJECT_0)
	{
		if (dwRetVal == WAIT_FAILED)
		{
			PrintWaitFailure();
		}
		else
			MPIU_Error_printf(L"error in mutex_lock: %d\n", GetLastError());
	}
	/*wprintf(L"lock: Handle = %u\n", (unsigned long)*lock);*/
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDU_PROCESS_LOCK);
}

#undef FUNCNAME
#define FUNCNAME MPIDU_Process_unlock
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
/*@
   MPIDU_Process_unlock - 

   Parameters:
+  MPIDU_Process_lock_t *lock

   Notes:
@*/
void MPIDU_Process_unlock( MPIDU_Process_lock_t *lock )
{
	MPIDI_STATE_DECL(MPID_STATE_MPIDU_PROCESS_UNLOCK);
	MPIDI_FUNC_ENTER(MPID_STATE_MPIDU_PROCESS_UNLOCK);
	if (!ReleaseMutex(*lock))
	{
		MPIU_Error_printf(L"error in mutex_unlock: %d\n", GetLastError());
		MPIU_Error_printf(L"Handle = %p\n", *lock);
	}
	/*wprintf(L"unlock: Handle = %u\n", (unsigned long)*lock);*/
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDU_PROCESS_UNLOCK);
}

#undef FUNCNAME
#define FUNCNAME MPIDU_Process_lock_busy_wait
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
/*@
   MPIDU_Process_lock_busy_wait - 

   Parameters:
+  MPIDU_Process_lock_t *lock

   Notes:
@*/
void MPIDU_Process_lock_busy_wait( MPIDU_Process_lock_t *lock )
{
	MPIDI_STATE_DECL(MPID_STATE_MPIDU_PROCESS_LOCK_BUSY_WAIT);
	MPIDI_FUNC_ENTER(MPID_STATE_MPIDU_PROCESS_LOCK_BUSY_WAIT);
	MPIDU_Process_lock(lock);
	MPIDU_Process_unlock(lock);
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDU_PROCESS_LOCK_BUSY_WAIT);
}

#undef FUNCNAME
#define FUNCNAME MPIDU_Process_lock_free
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
/*@
   MPIDU_Process_lock_free - 

   Parameters:
+  MPIDU_Process_lock_t *lock

   Notes:
@*/
void MPIDU_Process_lock_free( MPIDU_Process_lock_t *lock )
{
	MPIDI_STATE_DECL(MPID_STATE_MPIDU_PROCESS_LOCK_FREE);
	MPIDI_FUNC_ENTER(MPID_STATE_MPIDU_PROCESS_LOCK_FREE);
	/*wprintf(L"Free_lock: Handle = %u\n", (unsigned long)*lock);*/
	CloseHandle(*lock);
	*lock = NULL;
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDU_PROCESS_LOCK_FREE);
}
