/*  (C) 2009 by Deino Software. */
#ifndef MPIDU_PROCESS_LOCKS_H
#define MPIDU_PROCESS_LOCKS_H

#include "mpiimpl.h"
#include "mpid_locksconf.h"
#include <errno.h>
#include <winsock2.h>
#include <windows.h>
#include <stdio.h>

#ifdef HAVE_ICC_AND_IA64
#define HAVE__INTERLOCKEDEXCHANGE 1
#include <ia64intrin.h>
#define _InterlockedExchange(ptr,x) _InterlockedExchange(ptr,x)
#endif

extern int g_nLockSpinCount;

/* Define MPIDU_Yield() */
#ifdef HAVE_YIELD
#define MPIDU_Yield() yield()
#elif defined(HAVE_WIN32_SLEEP)
#define MPIDU_Yield() Sleep(0)
#elif defined (HAVE_SELECT)
#define MPIDU_Yield() { struct timeval t; t.tv_sec = 0; t.tv_usec = 0; select(0,0,0,0,&t); }
#elif defined (HAVE_SLEEP)
#define MPIDU_Yield() sleep(0)
#else
#error *** No yield function specified ***
#endif

#ifdef USE_SLEEP_YIELD
/* Define MPIDU_Sleep_yield() */
extern int MPIDI_Sleep_yield_count;
extern int MPIDI_Sleep_yield_counts[10];
static inline void MPIDU_Sleep_yield()
{
#ifdef HAVE_WIN32_SLEEP
	Sleep(MPIDI_Sleep_yield_counts[MPIDI_Sleep_yield_count]);
#elif defined(HAVE_SELECT)
	struct timeval t;
	t.tv_sec = 0;
	t.tv_usec = MPIDI_Sleep_yield_counts[MPIDI_Sleep_yield_count];
	select(0,0,0,0,&t);
#elif defined(HAVE_SLEEP)
	sleep(MPIDI_Sleep_yield_counts[MPIDI_Sleep_yield_count]);
#else
#error *** No function available to implement sleep_yield ***
#endif

	if (MPIDI_Sleep_yield_count < 9)
		MPIDI_Sleep_yield_count++;
}
#endif

typedef volatile long MPIDU_Process_busy_lock_t;
typedef HANDLE MPIDU_Process_lock_t;

void MPIDU_Process_lock_init( MPIDU_Process_lock_t *lock );
void MPIDU_Process_lock( MPIDU_Process_lock_t *lock );
void MPIDU_Process_unlock( MPIDU_Process_lock_t *lock );
void MPIDU_Process_lock_free( MPIDU_Process_lock_t *lock );
void MPIDU_Process_lock_busy_wait( MPIDU_Process_lock_t *lock );

#undef FUNCNAME
#define FUNCNAME MPIDU_Process_busy_lock_init
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static inline void MPIDU_Process_busy_lock_init( MPIDU_Process_busy_lock_t *lock )
{
	MPIDI_STATE_DECL(MPID_STATE_MPIDU_PROCESS_LOCK_INIT);
	MPIDI_FUNC_ENTER(MPID_STATE_MPIDU_PROCESS_LOCK_INIT);
	*(lock) = 0;
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDU_PROCESS_LOCK_INIT);
}

#undef FUNCNAME
#define FUNCNAME MPIDU_Process_busy_lock
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static inline void MPIDU_Process_busy_lock( MPIDU_Process_busy_lock_t *lock )
{
	int i;
	MPIDI_STATE_DECL(MPID_STATE_MPIDU_PROCESS_LOCK);
	MPIDI_FUNC_ENTER(MPID_STATE_MPIDU_PROCESS_LOCK);
	for (;;)
	{
		for (i=0; i<g_nLockSpinCount; i++)
		{
			if (*lock == 0)
			{
#ifdef HAVE_INTERLOCKEDEXCHANGE
				if (InterlockedExchange((LPLONG)lock, 1) == 0)
				{
					MPIDI_FUNC_EXIT(MPID_STATE_MPIDU_PROCESS_LOCK);
					return;
				}
#elif defined(HAVE__INTERLOCKEDEXCHANGE)
				/* The Intel compiler complains if the lock is cast to
				 * volatile void * (the type of lock is probably
				 * volatile long *).  The void * works for the Intel 
				 * compiler. */
				if (_InterlockedExchange((void *)lock, 1) == 0)
				{
					MPIDI_FUNC_EXIT(MPID_STATE_MPIDU_PROCESS_LOCK);
					return;
				}
#else
#error *** No atomic memory operation specified to implement busy locks ***
#endif
			}
		}
		MPIDU_Yield();
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDU_PROCESS_LOCK);
}

#undef FUNCNAME
#define FUNCNAME MPIDU_Process_busy_unlock
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static inline void MPIDU_Process_busy_unlock( MPIDU_Process_busy_lock_t *lock )
{
	MPIDI_STATE_DECL(MPID_STATE_MPIDU_PROCESS_UNLOCK);
	MPIDI_FUNC_ENTER(MPID_STATE_MPIDU_PROCESS_UNLOCK);
	*(lock) = 0;
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDU_PROCESS_UNLOCK);
}

#undef FUNCNAME
#define FUNCNAME MPIDU_Process_busy_lock_busy_wait
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static inline void MPIDU_Process_busy_lock_busy_wait( MPIDU_Process_busy_lock_t *lock )
{
	int i;
	MPIDI_STATE_DECL(MPID_STATE_MPIDU_PROCESS_LOCK_BUSY_WAIT);
	MPIDI_FUNC_ENTER(MPID_STATE_MPIDU_PROCESS_LOCK_BUSY_WAIT);
	for (;;)
	{
		for (i=0; i<g_nLockSpinCount; i++)
			if (!*lock)
			{
				MPIDI_FUNC_EXIT(MPID_STATE_MPIDU_PROCESS_LOCK_BUSY_WAIT);
				return;
			}
		MPIDU_Yield();
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDU_PROCESS_LOCK_BUSY_WAIT);
}

#undef FUNCNAME
#define FUNCNAME MPIDU_Process_busy_lock_free
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static inline void MPIDU_Process_busy_lock_free( MPIDU_Process_busy_lock_t *lock )
{
	MPIDI_STATE_DECL(MPID_STATE_MPIDU_PROCESS_LOCK_FREE);
	MPIDI_FUNC_ENTER(MPID_STATE_MPIDU_PROCESS_LOCK_FREE);
	MPIU_UNREFERENCED_ARG(lock);
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDU_PROCESS_LOCK_FREE);
}

#undef FUNCNAME
#define FUNCNAME MPIDU_Compare_swap
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
/*@
   MPIDU_Compare_swap - 

   Parameters:
+  void **dest
.  void *new_val
.  void *compare_val
.  MPIDU_Process_lock_t *lock
-  void **original_val

   Notes:
@*/
static inline int MPIDU_Compare_swap( void **dest, void *new_val, void *compare_val, MPIDU_Process_lock_t *lock, void **original_val )
{
	/* dest = pointer to value to be checked (address size)
	   new_val = value to set dest to if *dest == compare_val
	   original_val = value of dest prior to this operation */

	MPIDI_STATE_DECL(MPID_STATE_MPIDU_COMPARE_SWAP);
	MPIDI_FUNC_ENTER(MPID_STATE_MPIDU_COMPARE_SWAP);
	MPIU_UNREFERENCED_ARG(lock);
	/**original_val = InterlockedCompareExchange(dest, new_val, compare_val);*/
	/**original_val = (void*)InterlockedCompareExchange((LONG*)dest, (LONG)new_val, (LONG)compare_val);*/
	*original_val = (void*)InterlockedCompareExchangePointer(dest, new_val, compare_val);

	MPIDI_FUNC_EXIT(MPID_STATE_MPIDU_COMPARE_SWAP);
	return 0;
}

#endif
