using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Threading;
using System.Diagnostics;
using System.ComponentModel;

namespace DeinoMPI
{
    internal class DumpOutputThread
    {
        public Mpiexec mpiexec;
        public StreamReader output;

        public DumpOutputThread(Mpiexec mpiexec, StreamReader output)
        {
            this.mpiexec = mpiexec;
            this.output = output;
        }

        public void Run()
        {
            try
            {
                string line = output.ReadLine();
                while (line != null)
                {
                    //mpiexec.WriteLine("DeinoPM: {0}", line);
                    line = output.ReadLine();
                }
            }
            catch (Exception e)
            {
                if (mpiexec != null)
                    mpiexec.WriteLine("Exception caught reading output: {0}", e.Message);
            }
        }
    }

    public class DeinoPMUtil
    {
        public static int StartLocalPM(Mpiexec mpiexec, SecurityKeys keys)
        {
            Process p = new Process();
            ProcessStartInfo pinfo = new ProcessStartInfo();
            pinfo.ErrorDialog = false;
            pinfo.CreateNoWindow = true;
            pinfo.FileName = Path.GetDirectoryName(System.Reflection.Assembly.GetEntryAssembly().Location) + @"\DeinoPM.exe";
            pinfo.LoadUserProfile = true;
            pinfo.RedirectStandardOutput = true;
            pinfo.RedirectStandardInput = true;
            pinfo.RedirectStandardError = true;
            string single_arg_encoding = Convert.ToBase64String(Encoding.Unicode.GetBytes(keys.PublicKey));
            pinfo.Arguments = "/public_key";
            pinfo.UseShellExecute = false;
            p.StartInfo = pinfo;

            string port_str = "";
            Random r = null;
            bool start_result = true;
            string start_error = "";
            for (int i = 0; i < 5; i++)
            {
                try
                {
                    p.Start();
                    p.StandardInput.WriteLine(single_arg_encoding);
                    port_str = p.StandardOutput.ReadLine();
                    break;
                }
                catch (Win32Exception e)
                {
                    if ((e.NativeErrorCode == 1723) && (i < 4))
                    {
                        // RPC Server busy error so wait a second and try again
                        if (r == null)
                            r = new Random();
                        int time = (int)(r.NextDouble() * 300.0);
                        //verify_mpiexec.WriteLine("Sleeping {0} milliseconds", time);
                        System.Threading.Thread.Sleep(time);
                    }
                    else
                    {
                        start_error = String.Format("Starting '{0}' failed: {1}, error #{2}", pinfo.FileName, e.Message, e.NativeErrorCode);
                        start_result = false;
                    }
                }
                catch (Exception e)
                {
                    start_error = e.Message;
                    start_result = false;
                }
                if (!start_result)
                    break;
            }
            if (!start_result)
            {
                throw new Exception(start_error);
            }

            int port = Convert.ToInt32(port_str);
            //verify_mpiexec.WriteLine("DeinoPM listening on port: {0}", port);
            DumpOutputThread othread_data = new DumpOutputThread(mpiexec, p.StandardOutput);
            ThreadStart ts = new ThreadStart(othread_data.Run);
            Thread othread = new Thread(ts);
            othread.Start();

            DumpOutputThread ethread_data = new DumpOutputThread(mpiexec, p.StandardError);
            ThreadStart ts2 = new ThreadStart(ethread_data.Run);
            Thread ethread = new Thread(ts2);
            ethread.Start();

            return port;
        }
    }

    public class VerifiedNode
    {
        public enum VerificationState { INDETERMINATE, VERIFICATION_PENDING, GOOD, OK, BAD }
        public VerificationState state = VerificationState.INDETERMINATE;
        public LaunchInfoMessage message = null;
        public LaunchNode node = null;
        public string error_msg = null;

        public string GetLaunchInfoString()
        {
            StringBuilder sb = new StringBuilder();
            sb.AppendFormat("__{0}__ {1}\n", node.iproc, node.hostname);
            if (message != null)
            {
                if (message.good_exe_path != "")
                {
                    sb.AppendFormat("Command line: {0}\n", message.good_exe_path);
                }
                if (message.bad_exe_path != "")
                {
                    sb.AppendFormat("ERROR: Path not found: {0}\n", message.bad_exe_path);
                }
                if (message.good_wdir != "")
                {
                    sb.AppendFormat("Working directory: {0}\n", message.good_wdir);
                }
                if (message.bad_wdir != "")
                {
                    sb.AppendFormat("WARNING: Working directory not found: {0}\n", message.bad_wdir);
                }
                if (message.missing_dlls != "")
                {
                    sb.AppendFormat("ERROR: Executable unable not start due to missing dlls: {0}\n", message.missing_dlls);
                }
                if (message.missing_delayed_dlls != "")
                {
                    sb.AppendFormat("WARNING: Executable may not be able to run due to missing delay loaded dlls: {0}\n", message.missing_delayed_dlls);
                }
            }
            else if (error_msg != null)
            {
                sb.AppendFormat("{0}\n", error_msg);
            }
            else
            {
                sb.Append("Information pending ...\n");
            }
            return sb.ToString();
        }
    }
}
