using System;
using System.Collections.Generic;
using System.Text;
using System.Runtime.InteropServices;
using System.Collections;

namespace DeinoMPI
{
    public class DriveMapping
    {
        class SavedMapping
        {
            public char drive_letter;
            public string mapping;
        }
        private ArrayList mappings;

        // 
        // Import Declarations 
        //
        /*
        [DllImport("mpr.dll")]
        private static extern int WNetGetUniversalName(string lpLocalPath,
            int dwInfoLevel, ref UNIVERSAL_NAME_INFO lpBuffer, ref int lpBufferSize);
        */
        [DllImport("mpr.dll")]
        private static extern int WNetGetUniversalName(string lpLocalPath,
            int dwInfoLevel, ref REMOTE_NAME_INFO lpBuffer, ref int lpBufferSize);

        [DllImport("mpr", CharSet = CharSet.Auto)]
        protected static extern int WNetGetUniversalName(string lpLocalPath,
            int dwInfoLevel, IntPtr lpBuffer, ref int lpBufferSize);

        [DllImport("mpr.dll")]
        private static extern int WNetAddConnection2(ref NETRESOURCE pstNetRes, string psPassword, string psUsername, int piFlags);

        [DllImport("mpr", CharSet = CharSet.Auto)]
        protected static extern int WNetAddConnection2(IntPtr pstNetRes, string psPassword, string psUsername, int piFlags);

        [DllImport("mpr.dll")]
        private static extern int WNetCancelConnection2(string psName, int piFlags, int pfForce);

        /*
        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Auto)]
        private struct UNIVERSAL_NAME_INFO
        {
            [MarshalAs(UnmanagedType.LPTStr)]
            public string lpUniversalName;
        }
        */

        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Auto)]
        private struct REMOTE_NAME_INFO
        {
            [MarshalAs(UnmanagedType.LPTStr)]
            public string lpUniversalName;
            [MarshalAs(UnmanagedType.LPTStr)]
            public string lpConnectionName;
            [MarshalAs(UnmanagedType.LPTStr)]
            public string lpRemainingPath;
        }

        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Auto)]
        private struct NETRESOURCE
        {
            public int iScope;
            public int iType;
            public int iDisplayType;
            public int iUsage;
            [MarshalAs(UnmanagedType.LPTStr)]
            public string sLocalName;
            [MarshalAs(UnmanagedType.LPTStr)]
            public string sRemoteName;
            [MarshalAs(UnmanagedType.LPTStr)]
            public string sComment;
            [MarshalAs(UnmanagedType.LPTStr)]
            public string sProvider;
        }

        // 
        // Constants 
        // 
        protected const int NO_ERROR                    = 0;
        protected const int ERROR_MORE_DATA             = 234;
        protected const int ERROR_NOT_CONNECTED         = 2250;
        protected const int UNIVERSAL_NAME_INFO_LEVEL   = 0x00000001;
        protected const int REMOTE_NAME_INFO_LEVEL      = 0x00000002;
        protected const int RESOURCETYPE_DISK           = 0x00000001;
        protected const int CONNECT_INTERACTIVE         = 0x00000008;
        protected const int CONNECT_PROMPT              = 0x00000010;
        protected const int CONNECT_UPDATE_PROFILE      = 0x00000001;
        protected const int CONNECT_REDIRECT            = 0x00000080;
        protected const int CONNECT_COMMANDLINE         = 0x00000800;
        protected const int CONNECT_CMD_SAVECRED        = 0x00001000;

        /// 
        /// Converts a Mapped Path to its equivalent UNC Path
        /// 
        public static string ExeToUNC(string exe)
        {
            REMOTE_NAME_INFO uni = new REMOTE_NAME_INFO();
            int bufferSize = Marshal.SizeOf(uni);

            int returnValue = WNetGetUniversalName(exe, REMOTE_NAME_INFO_LEVEL, ref uni, ref bufferSize);

            if (ERROR_MORE_DATA == returnValue)
            {
                IntPtr pBuffer = Marshal.AllocHGlobal(bufferSize); ;
                try
                {
                    returnValue = WNetGetUniversalName(exe, REMOTE_NAME_INFO_LEVEL, pBuffer, ref bufferSize);

                    if (NO_ERROR == returnValue)
                    {
                        uni = (REMOTE_NAME_INFO)Marshal.PtrToStructure(pBuffer, typeof(REMOTE_NAME_INFO));
                    }
                }
                finally
                {
                    Marshal.FreeHGlobal(pBuffer);
                }
            }
            if (returnValue == NO_ERROR)
            {
                //Console.WriteLine("Universal name : {0}", uni.lpUniversalName);
                //Console.WriteLine("Connection name: {0}", uni.lpConnectionName);
                //Console.WriteLine("Remaining path : {0}", uni.lpRemainingPath);
                return uni.lpUniversalName;
            }
            return exe;
        }

        public DriveMapping()
        {
            mappings = new ArrayList();
        }

        private bool AlreadyMapped(char drive_letter)
        {
            return false;
        }

        /// <summary>
        /// Map a network drive
        /// </summary>
        /// <param name="drive_letter">Drive letter to be mapped</param>
        /// <param name="mapping">UNC share name to be mapped</param>
        /// <param name="username">Username for authentication</param>
        /// <param name="password">Password to go with the username</param>
        public void MapDrive(char drive_letter, string mapping, string username, string password)
        {
            drive_letter = Char.ToUpper(drive_letter);
            if (AlreadyMapped(drive_letter))
            {
                return;
            }

            /*
            NETRESOURCE res = new NETRESOURCE();
            res.iScope = 0; // 2;
            res.iType = RESOURCETYPE_DISK;
            res.iDisplayType = 0; // 3;
            res.iUsage = 0; // 1;
            res.sRemoteName = mapping;
            res.sLocalName = String.Format("{0}:", drive_letter);
            res.sProvider = null;
            res.sComment = null;

            Console.WriteLine("WNetAddConnection2({0}, {1})", res.sLocalName, res.sRemoteName);
            int result = WNetAddConnection2(ref res, username, password, CONNECT_REDIRECT);
            */
            int result = DeinoLP.DriveMapping.MapDrive(drive_letter, mapping, username, password);
            if (result == NO_ERROR)
            {
                // save the drive in a list to be unmapped
                SavedMapping sm = new SavedMapping();
                sm.drive_letter = drive_letter;
                sm.mapping = mapping;
                mappings.Add(sm);
            }
            else
            {
                throw new System.ComponentModel.Win32Exception(result);
            }
        }

        /// <summary>
        /// Map a network drive
        /// </summary>
        /// <param name="drive_letter">Drive letter to be mapped</param>
        /// <param name="mapping">UNC share name to be mapped</param>
        public void MapDrive(char drive_letter, string mapping)
        {
            MapDrive(drive_letter, mapping, null, null);
        }

        /// <summary>
        /// Unmap a network drive
        /// </summary>
        /// <param name="drive_letter">Drive letter to be unmapped</param>
        public void UnMapDrive(char drive_letter)
        {
            drive_letter = Char.ToUpper(drive_letter);
            foreach (SavedMapping map in mappings)
            {
                if (map.drive_letter == drive_letter)
                {
                    //int result = WNetCancelConnection2(String.Format("{0}:", drive_letter), CONNECT_UPDATE_PROFILE, 1);
                    int result = DeinoLP.DriveMapping.UnMapDrive(drive_letter);
                    if (result == NO_ERROR)
                    {
                        return;
                    }
                    else
                    {
                        throw new System.ComponentModel.Win32Exception(result);
                    }
                }
            }
        }

        /// <summary>
        /// Unmap all mapped drives mapped by the instance of this class
        /// </summary>
        public void UnmapDrives()
        {
            foreach (SavedMapping map in mappings)
            {
                UnMapDrive(map.drive_letter);
            }
        }
    }
}
