using System;
using System.Collections.Generic;
using System.Text;
using System.Xml;

namespace DeinoMPI
{
    public class HostInfo
    {
        public enum ARCH { WIN32, WIN64 };

        private string _host, _version, _os;
        private ARCH _arch;
        private string _arch_string;
        private string _cpu;
        private string _cpu_name;
        private long _memory;
        private long _free_disk;
        private int _num_cpus;
        private string _nic_speeds;
        private string _ipaddresses;
        private bool _valid;

        public HostInfo()
        {
            _valid = false;
        }

        public HostInfo(HostInfo info)
        {
            _valid = false;
            if (info != null)
            {
                _host = info.Host;
                _version = info.Version;
                _os = info.OS;
                _arch = info.Arch;
                _arch_string = info.ArchString;
                _cpu = info.CPU;
                _cpu_name = info.CPUName;
                _memory = info.Memory;
                _free_disk = info.FreeDisk;
                _num_cpus = info.NumCPUs;
                _nic_speeds = info.NicSpeeds;
                _ipaddresses = info.IPAddresses;
                _valid = info.Valid;
            }
        }

        public string Host { get { return _host; } set { _host = value; } }
        public string Version { get { return _version; } set { _version = value; } }
        public string OS { get { return _os; } set { _os = value; } }
        public ARCH Arch { get { return _arch; } set { _arch = value; } }
        public string ArchString { get { return _arch_string; } set { _arch_string = value; } }
        public string CPU { get { return _cpu; } set { _cpu = value; } }
        public string CPUName { get { return _cpu_name; } set { _cpu_name = value; } }
        public long Memory { get { return _memory; } set { _memory = value; } }
        public long FreeDisk { get { return _free_disk; } set { _free_disk = value; } }
        public int NumCPUs { get { return _num_cpus; } set { _num_cpus = value; } }
        public string NicSpeeds { get { return _nic_speeds; } set { _nic_speeds = value; } }
        public string IPAddresses { get { return _ipaddresses; } set { _ipaddresses = value; } }
        public bool Valid { get { return _valid; } set { _valid = value; } }

        public void Write(XmlTextWriter xwriter)
        {
            xwriter.WriteElementString("Host", Host);
            xwriter.WriteElementString("Version", Version);
            xwriter.WriteElementString("OS", OS);
            xwriter.WriteElementString("Arch", (Arch == HostInfo.ARCH.WIN32) ? "Win32" : "Win64");
            xwriter.WriteElementString("ArchString", ArchString);
            xwriter.WriteElementString("CPU", CPU);
            xwriter.WriteElementString("CPUName", CPUName);
            xwriter.WriteElementString("Memory", XmlConvert.ToString(Memory));
            xwriter.WriteElementString("FreeDisk", XmlConvert.ToString(FreeDisk));
            xwriter.WriteElementString("NumCPUs", XmlConvert.ToString(NumCPUs));
            xwriter.WriteElementString("NicSpeeds", NicSpeeds);
            xwriter.WriteElementString("IPAddresses", IPAddresses);
            xwriter.WriteElementString("Valid", XmlConvert.ToString(Valid));
        }

        public void Read(XmlTextReader xreader)
        {
            Host = xreader.ReadElementString("Host");
            Version = xreader.ReadElementString("Version");
            OS = xreader.ReadElementString("OS");
            string arch = xreader.ReadElementString("Arch");
            Arch = (arch == "Win32") ? HostInfo.ARCH.WIN32 : HostInfo.ARCH.WIN64;
            ArchString = xreader.ReadElementString("ArchString");
            CPU = xreader.ReadElementString("CPU");
            CPUName = xreader.ReadElementString("CPUName");
            Memory = XmlConvert.ToInt64(xreader.ReadElementString("Memory"));
            FreeDisk = XmlConvert.ToInt64(xreader.ReadElementString("FreeDisk"));
            NumCPUs = XmlConvert.ToInt32(xreader.ReadElementString("NumCPUs"));
            NicSpeeds = xreader.ReadElementString("NicSpeeds");
            IPAddresses = xreader.ReadElementString("IPAddresses");
            Valid = XmlConvert.ToBoolean(xreader.ReadElementString("Valid"));
        }
    }
}
