using System;
using System.Collections.Generic;
using System.Text;
using System.Collections;
using System.Threading;
using System.IO;

namespace DeinoMPI
{
    public class InputHandling
    {
        private Queue _input_queue;
        private ManualResetEvent _input_queue_event;
        private TextReader _stream;
        private ArrayList _clients;
        private Thread _thread;
        private bool _running;

        public bool IsRunning { get { return _running; } }

        public InputHandling()
        {
            _running = false;
            _input_queue = null;
            _input_queue_event = null;
            _stream = null;
            _thread = null;
            _clients = new ArrayList();
        }

        private void input_thread_fn()
        {
            try
            {
                char[] buffer = new char[4096];
                int num_read;
                InputOutputMessage input_message = new InputOutputMessage();
                input_message.io_type = InputOutputMessage.IO_Type.INPUT;

                if (_input_queue != null)
                {
                    // Read from the input queue populated by keystroke events in the output text box
                    while (true)
                    {
                        _input_queue_event.WaitOne();
                        _input_queue_event.Reset();
                        // An empty queue signals exit
                        if (_input_queue.Count == 0)
                            return;
                        while (_input_queue.Count > 0)
                        {
                            object o = _input_queue.Dequeue();
                            int length = 0;
                            if (o is char)
                            {
                                buffer[0] = (char)o;
                                length = 1;
                            }
                            else
                            {
                                string str = (string)o;
                                int i;
                                while (str.Length > 4096)
                                {
                                    for (i = 0; i < 4096; i++)
                                    {
                                        buffer[i] = str[i];
                                    }
                                    input_message.SetBuffer(buffer, 0, 4096);
                                    foreach (ClientSession client in _clients)
                                    {
                                        client.SendMessage(input_message);
                                    }
                                    str = str.Substring(4096);
                                }
                                for (i = 0; i < str.Length; i++)
                                    buffer[i] = str[i];
                                length = str.Length;
                            }
                            if (length > 0)
                            {
                                input_message.SetBuffer(buffer, 0, length);
                                foreach (ClientSession client in _clients)
                                {
                                    client.SendMessage(input_message);
                                }
                            }
                        }
                    }
                }
                else
                {
                    // Read from the provided stream which is usually Standard Input
                    num_read = _stream.ReadBlock(buffer, 0, 1);
                    while (num_read > 0)
                    {
                        input_message.SetBuffer(buffer, 0, num_read);
                        foreach (ClientSession client in _clients)
                        {
                            client.SendMessage(input_message);
                        }
                        num_read = _stream.ReadBlock(buffer, 0, 1);
                    }
                }
            }
            catch
            {
                _running = false;
            }
        }

        public void Start(TextReader stream)
        {
            Stop();
            _stream = stream;
            ThreadStart ts = new ThreadStart(input_thread_fn);
            _thread = new Thread(ts);
            _thread.Start();
            _running = true;
        }

        public void Start()
        {
            Stop();
            _input_queue = Queue.Synchronized(new Queue());
            _input_queue_event = new ManualResetEvent(false);
            ThreadStart ts = new ThreadStart(input_thread_fn);
            _thread = new Thread(ts);
            _thread.Start();
            _running = true;
        }

        public void EnqueueInput(string str)
        {
            if (_input_queue != null)
                _input_queue.Enqueue(str);
            if (_input_queue_event != null)
                _input_queue_event.Set();
        }

        public void Stop()
        {
            try
            {
                if (_thread != null)
                {
                    if (_input_queue != null)
                    {
                        _input_queue.Clear();
                        _input_queue_event.Set();
                        if (!_thread.Join(5000))
                            _thread.Abort();
                    }
                    else
                    {
                        _thread.Abort();
                        // Apparently you can't Abort and Join the thread while it is blocked in a stream.ReadBlock function call.
                        //_thread.Join();
                    }
                    _thread = null;
                    _running = false;
                }
            }
            catch
            {
                _running = false;
            }
        }

        public void AddClient(ClientSession client)
        {
            _clients.Add(client);
        }

        public void RemoveClient(ClientSession client)
        {
            _clients.Remove(client);
        }
    }
}
