using System;
using System.Collections.Generic;
using System.Text;
using System.Collections;

namespace DeinoMPI
{
    public class MpiexecCmdArgs
    {
        enum ArgType { INT, STRING, BOOLEAN, SIMPLE, NOT_ARG, UNKNOWN };
        class ArgDescription
        {
            public string []name;
            public ArgType type;
            public int num_values;
            public string[] example_values;
            public string help_string;

            #region Constructors
            public ArgDescription(string name, string help_string)
            {
                this.name = new string[1] { name };
                type = ArgType.INT;
                num_values = 1;
                this.help_string = help_string;
                example_values = new string[1] { "number" };
            }

            public ArgDescription(string [] names, string help_string)
            {
                this.name = (string[])names.Clone();
                type = ArgType.INT;
                num_values = 1;
                this.help_string = help_string;
                example_values = new string[1] { "number" };
            }

            public ArgDescription(string name, string example_value, string help_string)
            {
                this.name = new string[1] { name };
                type = ArgType.INT;
                num_values = 1;
                this.help_string = help_string;
                example_values = new string[1] { example_value };
            }

            public ArgDescription(string []names, string example_value, string help_string)
            {
                this.name = (string[])names.Clone();
                type = ArgType.INT;
                num_values = 1;
                this.help_string = help_string;
                example_values = new string[1] { example_value };
            }

            public ArgDescription(string name, ArgType type, string help_string)
            {
                this.name = new string[1] { name };
                this.type = type;
                if (type == ArgType.BOOLEAN)
                    num_values = 0;
                else
                    num_values = 1;
                this.help_string = help_string;
                example_values = new string[1];
                switch (type)
                {
                    case ArgType.BOOLEAN:
                        example_values[0] = "";
                        break;
                    case ArgType.INT:
                        example_values[0] = "number";
                        break;
                    case ArgType.SIMPLE:
                        example_values[0] = "";
                        break;
                    case ArgType.STRING:
                        example_values[0] = "string";
                        break;
                }
            }

            public ArgDescription(string []names, ArgType type, string help_string)
            {
                this.name = (string[])names.Clone();
                this.type = type;
                if (type == ArgType.BOOLEAN)
                    num_values = 0;
                else
                    num_values = 1;
                this.help_string = help_string;
                example_values = new string[1];
                switch (type)
                {
                    case ArgType.BOOLEAN:
                        example_values[0] = "";
                        break;
                    case ArgType.INT:
                        example_values[0] = "number";
                        break;
                    case ArgType.SIMPLE:
                        example_values[0] = "";
                        break;
                    case ArgType.STRING:
                        example_values[0] = "string";
                        break;
                }
            }

            public ArgDescription(string name, ArgType type, string example_value, string help_string)
            {
                this.name = new string[1] { name };
                this.type = type;
                if (type == ArgType.BOOLEAN)
                    num_values = 0;
                else
                    num_values = 1;
                this.help_string = help_string;
                example_values = new string[1] { example_value };
            }

            public ArgDescription(string []names, ArgType type, string example_value, string help_string)
            {
                this.name = (string[])names.Clone();
                this.type = type;
                if (type == ArgType.BOOLEAN)
                    num_values = 0;
                else
                    num_values = 1;
                this.help_string = help_string;
                example_values = new string[1] { example_value };
            }

            public ArgDescription(string name, ArgType type, int num_values, string help_string)
            {
                this.name = new string[1] { name };
                this.type = type;
                this.num_values = num_values;
                this.help_string = help_string;
                example_values = new string[num_values];
                for (int i = 0; i < num_values; i++)
                {
                    switch (type)
                    {
                        case ArgType.BOOLEAN:
                            example_values[i] = "";
                            break;
                        case ArgType.INT:
                            example_values[i] = String.Format("num{0}", i+1);
                            break;
                        case ArgType.SIMPLE:
                            example_values[i] = "";
                            break;
                        case ArgType.STRING:
                            example_values[i] = String.Format("str{0}", i+1);
                            break;
                    }
                }
            }

            public ArgDescription(string []names, ArgType type, int num_values, string help_string)
            {
                this.name = (string[])names.Clone();
                this.type = type;
                this.num_values = num_values;
                this.help_string = help_string;
                example_values = new string[num_values];
                for (int i = 0; i < num_values; i++)
                {
                    switch (type)
                    {
                        case ArgType.BOOLEAN:
                            example_values[i] = "";
                            break;
                        case ArgType.INT:
                            example_values[i] = String.Format("num{0}", i + 1);
                            break;
                        case ArgType.SIMPLE:
                            example_values[i] = "";
                            break;
                        case ArgType.STRING:
                            example_values[i] = String.Format("str{0}", i + 1);
                            break;
                    }
                }
            }

            public ArgDescription(string name, ArgType type, int num_values, string[] example_value, string help_string)
            {
                this.name = new string[1] { name };
                this.type = type;
                this.num_values = num_values;
                this.help_string = help_string;
                example_values = new string[num_values];
                for (int i = 0; i < num_values; i++)
                {
                    if (i < example_value.Length)
                        example_values[i] = example_value[i];
                    else
                    {
                        switch (type)
                        {
                            case ArgType.BOOLEAN:
                                example_values[i] = "";
                                break;
                            case ArgType.INT:
                                example_values[i] = String.Format("num{0}", i + 1);
                                break;
                            case ArgType.SIMPLE:
                                example_values[i] = "";
                                break;
                            case ArgType.STRING:
                                example_values[i] = String.Format("str{0}", i + 1);
                                break;
                        }
                    }
                }
            }

            public ArgDescription(string []names, ArgType type, int num_values, string[] example_value, string help_string)
            {
                this.name = (string[])names.Clone();
                this.type = type;
                this.num_values = num_values;
                this.help_string = help_string;
                example_values = new string[num_values];
                for (int i = 0; i < num_values; i++)
                {
                    if (i < example_value.Length)
                        example_values[i] = example_value[i];
                    else
                    {
                        switch (type)
                        {
                            case ArgType.BOOLEAN:
                                example_values[i] = "";
                                break;
                            case ArgType.INT:
                                example_values[i] = String.Format("num{0}", i + 1);
                                break;
                            case ArgType.SIMPLE:
                                example_values[i] = "";
                                break;
                            case ArgType.STRING:
                                example_values[i] = String.Format("str{0}", i + 1);
                                break;
                        }
                    }
                }
            }
            #endregion

            private void PrintHelpString()
            {
                if (help_string.Length > 0)
                {
                    string[] lines = help_string.Split('\n');
                    foreach (string line in lines)
                    {
                        Console.WriteLine("  {0}", line);
                    }
                }
            }

            public void PrintDescription()
            {
                int i;
                switch (type)
                {
                    case ArgType.STRING:
                        for (i = 0; i < name.Length; i++)
                        {
                            if (num_values == 1)
                            {
                                //Console.WriteLine("-{0} <string> or /{0}:<string>", name);
                                Console.WriteLine("-{0} <{1}>", name[i], example_values[0]);
                                Console.WriteLine("/{0}:<{1}>", name[i], example_values[0]);
                            }
                            else
                            {
                                Console.Write("-{0} ", name[i]);
                                for (int j = 0; j < num_values; j++)
                                {
                                    Console.Write("<{0}> ", example_values[j]);
                                }
                                Console.WriteLine();
                            }
                        }
                        PrintHelpString();
                        break;
                    case ArgType.SIMPLE:
                        for (i = 0; i < name.Length; i++)
                        {
                            //Console.WriteLine("-{0} or /{0}", name);
                            Console.WriteLine("-{0}", name[i]);
                            Console.WriteLine("/{0}", name[i]);
                        }
                        PrintHelpString();
                        break;
                    case ArgType.INT:
                        for (i = 0; i < name.Length; i++)
                        {
                            //Console.WriteLine("-{0} <number> or /{0}:<number>", name);
                            Console.WriteLine("-{0} <{1}>", name[i], example_values[0]);
                            Console.WriteLine("/{0}:<{1}>", name[i], example_values[0]);
                        }
                        PrintHelpString();
                        break;
                    case ArgType.BOOLEAN:
                        for (i = 0; i < name.Length; i++)
                        {
                            //Console.WriteLine("-{0} or /{0}[+-]", name);
                            Console.WriteLine("-{0}", name[i]);
                            Console.WriteLine("/{0}[+-]", name[i]);
                        }
                        PrintHelpString();
                        break;
                    default:
                        break;
                }
            }
        }
        class ArgInstance
        {
            public string name;
            public ArgType type;
            public string[] str;
            public int[] n;
            public bool[] b;
            public bool first_value_set;

            public ArgInstance()
            {
                name = null;
                type = ArgType.NOT_ARG;
                str = null;
                n = null;
                b = null;
                first_value_set = false;
            }

            public ArgInstance(string str, ArrayList arg_descriptions)
            {
                string first_str = null;
                int index;
                if (str.Length > 1 && str[0] == '-' && str[1] == '-')
                {
                    name = str.Remove(0, 2);
                }
                else if (str.Length > 0 && str[0] == '-')
                {
                    name = str.Remove(0, 1);
                }
                else if (str.Length > 0 && str[0] == '/')
                {
                    name = str.Remove(0, 1);
                    if (name.Contains(":"))
                    {
                        index = name.IndexOf(':');
                        first_str = name.Remove(0, index + 1);
                        name = name.Substring(0, index);
                    }
                    else if (name.EndsWith("+"))
                    {
                        name = name.Substring(0, name.Length - 1);
                        first_str = "+";
                    }
                    else if (name.EndsWith("-"))
                    {
                        name = name.Substring(0, name.Length - 1);
                        first_str = "-";
                    }
                }
                else
                {
                    name = str;
                    type = ArgType.NOT_ARG;
                    return;
                }
                foreach (ArgDescription arg in arg_descriptions)
                {
                    foreach (string arg_name in arg.name)
                    {
                        if (arg_name == name)
                        {
                            type = arg.type;
                            switch (arg.type)
                            {
                                case ArgType.BOOLEAN:
                                    b = new bool[arg.num_values > 0 ? arg.num_values : 1];
                                    if (first_str != null)
                                    {
                                        if (first_str == "+")
                                        {
                                            b[0] = true;
                                            first_value_set = true;
                                        }
                                        else if (first_str == "-")
                                        {
                                            b[0] = false;
                                            first_value_set = true;
                                        }
                                    }
                                    return;
                                case ArgType.INT:
                                    n = new int[arg.num_values > 0 ? arg.num_values : 1];
                                    if (first_str != null)
                                    {
                                        n[0] = Convert.ToInt32(first_str);
                                        first_value_set = true;
                                    }
                                    return;
                                case ArgType.SIMPLE:
                                    return;
                                case ArgType.STRING:
                                    this.str = new string[arg.num_values > 0 ? arg.num_values : 1];
                                    if (first_str != null)
                                    {
                                        this.str[0] = first_str;
                                        first_value_set = true;
                                    }
                                    return;
                                default:
                                    break;
                            }
                        }
                    }
                }
                type = ArgType.UNKNOWN;
                if (first_str != null)
                {
                    this.str = new string[1] { first_str };
                }
            }

            public void Print()
            {
                switch (type)
                {
                    case ArgType.BOOLEAN:
                        Console.WriteLine("arg name: {0}", name);
                        Console.Write("arg value: ");
                        foreach (bool val in b)
                        {
                            Console.Write("{0},", val);
                        }
                        Console.WriteLine();
                        break;
                    case ArgType.INT:
                        Console.WriteLine("arg name: {0}", name);
                        Console.Write("arg value: ");
                        foreach (int val in n)
                        {
                            Console.Write("{0},", val);
                        }
                        Console.WriteLine();
                        break;
                    case ArgType.SIMPLE:
                        Console.WriteLine("arg name: {0}", name);
                        break;
                    case ArgType.STRING:
                        Console.WriteLine("arg name: {0}", name);
                        Console.Write("arg value: ");
                        foreach (string val in str)
                        {
                            Console.Write("{0},", val);
                        }
                        Console.WriteLine();
                        break;
                    case ArgType.UNKNOWN:
                        Console.WriteLine("unknown arg");
                        Console.WriteLine("arg name: {0}", name);
                        Console.Write("arg value: ");
                        foreach (string val in str)
                        {
                            Console.Write("{0},", val);
                        }
                        Console.WriteLine();
                        break;
                    case ArgType.NOT_ARG:
                        Console.WriteLine("not an arg");
                        Console.WriteLine(name);
                        break;
                    default:
                        break;
                }
            }
        }
        class ArgBlock
        {
            public ArrayList mpiexec_args;
            public string command;
            public string[] cmd_args;

            public ArgBlock()
            {
                mpiexec_args = new ArrayList();
                command = null;
                cmd_args = null;
            }
        }
        ArrayList arg_descriptions;
        ArrayList arg_blocks;

        public MpiexecCmdArgs()
        {
            arg_blocks = new ArrayList();
            arg_descriptions = new ArrayList();
            arg_descriptions.Add(new ArgDescription(new string[2] { "n", "np" }, "x", "launch x number of processes"));
            arg_descriptions.Add(new ArgDescription("localonly", "x", "launch x number of processes (launched on the local machine only)"));
            arg_descriptions.Add(new ArgDescription("localonly", ArgType.BOOLEAN, "launch processes on the local machine only"));
            arg_descriptions.Add(new ArgDescription("machinefile", ArgType.STRING, "filename", "use a file to list the names of machines to launch on"));
            arg_descriptions.Add(new ArgDescription("host", ArgType.STRING, "hostname", "launch processes on the specified host"));
            //hosts n host1 host2 ... hostn
            //hosts n host1 m1 host2 m2 ... hostn mn
            arg_descriptions.Add(new ArgDescription("map", ArgType.STRING, @"drive:\\host\share", "map a network drive on all the nodes\nthis mapping will be removed when the processes exit"));
            arg_descriptions.Add(new ArgDescription(new string[2] { "wdir", "dir" }, ArgType.STRING, @"drive:\my\working\directory", "use the specified working directory when launching processes"));
            //arg_descriptions.Add(new ArgDescription("wdir", ArgType.STRING, "/my/working/directory", "use the specified working directory when launching processes"));
            arg_descriptions.Add(new ArgDescription("env", ArgType.STRING, 2, new string[2] { "var", "val" }, "set the environment variable to the specified value"));
            arg_descriptions.Add(new ArgDescription(new string[2] { "logon", "login" }, ArgType.BOOLEAN, "prompt for a user account and password"));
            arg_descriptions.Add(new ArgDescription("pwdfile", ArgType.STRING, "filename", "read the account and password from the specified file\nput the account on the first line and the password on the second"));
            arg_descriptions.Add(new ArgDescription("nompi", ArgType.BOOLEAN, "launch non-MPI processes"));
            arg_descriptions.Add(new ArgDescription("nopopup_debug", ArgType.BOOLEAN, "disable the system popup dialog if the process crashes"));
            arg_descriptions.Add(new ArgDescription("dbg", ArgType.BOOLEAN, "catch unhandled exceptions"));
            arg_descriptions.Add(new ArgDescription(new string[2] { "nodbg", "dbgoff" }, ArgType.BOOLEAN, "don't catch unhandled exceptions"));
            arg_descriptions.Add(new ArgDescription("exitcodes", ArgType.BOOLEAN, "print the process exit code when each process exits"));
            arg_descriptions.Add(new ArgDescription("noprompt", ArgType.BOOLEAN, "prevent mpiexec from prompting for user credentials"));
            arg_descriptions.Add(new ArgDescription("priority", ArgType.STRING, "class[:level]", "set the process startup priority and optionally level\nclass = 0,1,2,3,4   = idle, below, normal, above, high\nlevel = 0,1,2,3,4,5 = idle, lowest, below, normal, above, highest\nthe default is -priority 2:3"));
            arg_descriptions.Add(new ArgDescription("localroot", ArgType.BOOLEAN, "launch the root process directly from mpiexec bypassing the process manager\n(This allows the root process to create windows and be debugged)"));
            arg_descriptions.Add(new ArgDescription(new string[2] { "port", "p" }, "specify the port where the process manager is listening"));
            arg_descriptions.Add(new ArgDescription("phrase", ArgType.STRING, "specify the passphrase used to authenticate with the process manager")); // obsolete
            arg_descriptions.Add(new ArgDescription("smpdfile", ArgType.STRING, "specify the file where the smpd options are stored")); // obsolete
            arg_descriptions.Add(new ArgDescription("path", ArgType.STRING, "search path for executables, ; separated"));
            arg_descriptions.Add(new ArgDescription("register", ArgType.BOOLEAN, "encrypt a user name and password"));
            arg_descriptions.Add(new ArgDescription("user", "specify the user slot"));
            arg_descriptions.Add(new ArgDescription("remove", ArgType.SIMPLE, "delete the encrypted user name and password"));
            arg_descriptions.Add(new ArgDescription("validate", ArgType.SIMPLE, "validate the encrypted user credentials")); //validate [-user n] [-host hostname]
            arg_descriptions.Add(new ArgDescription("timeout", "seconds", "timeout for the job"));
            arg_descriptions.Add(new ArgDescription("plaintext", ArgType.SIMPLE, "don't encrypt data on the wire"));
            arg_descriptions.Add(new ArgDescription("delegate", ArgType.BOOLEAN, "use passwordless delegation to launch processes"));
            arg_descriptions.Add(new ArgDescription("impersonate", ArgType.BOOLEAN, "use passwordless authentication to launch processes"));
            arg_descriptions.Add(new ArgDescription("add_job", ArgType.STRING, 2, new string[2] { "job_name", @"domain\user" }, "add a job key for the specified user on the local or specified host")); //add_job <job_name> <domain\user> -host <hostname>
            arg_descriptions.Add(new ArgDescription("remove_job", ArgType.STRING, "name", "remove a job key from the local or specified host")); //remove_job <name> -host <hostname>
            arg_descriptions.Add(new ArgDescription("associate_job", ArgType.STRING, "name", "associate the current user's token with the specified job on the local or specified host")); //associate_job <name> -host <hostname>
            arg_descriptions.Add(new ArgDescription("job", ArgType.STRING, "launch the processes in the context of the specified job"));
            arg_descriptions.Add(new ArgDescription("whoami", ArgType.SIMPLE, "print the current user name"));
            arg_descriptions.Add(new ArgDescription("l", ArgType.BOOLEAN, "prefix output with the process number\n(This option is a lowercase L not the number one)"));
            arg_descriptions.Add(new ArgDescription("input", ArgType.STRING, "filename", "redirect standard input from the specified file"));
            arg_descriptions.Add(new ArgDescription("inputto", ArgType.STRING, "range of ranks", "redirect standard input to the specified ranks\nexamples: 1 or 1..10 or 1..5,7 or all"));
            arg_descriptions.Add(new ArgDescription("output", ArgType.STRING, "filename", "redirect standard output to the specified file"));
            arg_descriptions.Add(new ArgDescription("error", ArgType.STRING, "filename", "redirect error output to the specified file"));
            arg_descriptions.Add(new ArgDescription("alloutput", ArgType.STRING, "filename", "redirect both standard output and error output to the specified file"));
        }

        public void PrintOptions()
        {
            Console.WriteLine();
            Console.WriteLine("Usage:");
            Console.WriteLine("mpiexec -n <maxprocs> [options] executable [args ...]");
            Console.WriteLine("mpiexec [options] executable [args ...] : [options] exe [args] : ...");
            Console.WriteLine("mpiexec -configfile <configfile>");
            Console.WriteLine();
            Console.WriteLine("options:");
            Console.WriteLine();
            Console.WriteLine("standard:");
            Console.WriteLine("-n <maxprocs>");
            Console.WriteLine("-wdir <working directory>");
            Console.WriteLine("-configfile <filename>");
            Console.WriteLine("       each line contains a complete set of mpiexec options");
            Console.WriteLine("       including the executable and arguments");
            Console.WriteLine("-host <hostname>");
            Console.WriteLine("-path <search path for executable, ; separated>");
            Console.WriteLine();
            Console.WriteLine("extensions:");
            Console.WriteLine("-env <variable value>");
            Console.WriteLine("-hosts <n host1 host2 ... hostn>");
            Console.WriteLine("-hosts <n host1 m1 host2 m2 ... hostn mn>");
            Console.WriteLine("-machinefile <filename> - one host per line, #commented");
            Console.WriteLine("-localonly <numprocs>");
            Console.WriteLine("-exitcodes - print the exit codes of processes as they exit");
            Console.WriteLine("-genvlist <list of env var names a,b,c,...> - pass current values of these vars");
            Console.WriteLine("-g<local arg name> - global version of local options");
            Console.WriteLine("  genv, gwdir, ghost, gpath, gmap");
            //Console.WriteLine("-file <filename> - old mpich1 job configuration file");
            Console.WriteLine();
            Console.WriteLine("All options to mpiexec:");
            foreach (ArgDescription d in arg_descriptions)
            {
                d.PrintDescription();
            }
            Console.WriteLine();
            Console.WriteLine("examples:");
            Console.WriteLine("mpiexec -n 4 cpi");
            Console.WriteLine("mpiexec -n 1 -host foo master : -n 8 worker");
        }

        public void ParseCommandLine(string[] args)
        {
            arg_blocks.Clear();
            {
                ArgBlock block = new ArgBlock();
                IEnumerator iter = args.GetEnumerator();
                while (iter.MoveNext())
                {
                    ArgInstance arg = new ArgInstance((string)iter.Current, arg_descriptions);
                    if (arg.type != ArgType.NOT_ARG)
                    {
                        switch (arg.type)
                        {
                            case ArgType.BOOLEAN:
                                if (arg.b.Length > 0)
                                {
                                    int i = 0;
                                    if (arg.first_value_set)
                                        i++;
                                    for (; i < arg.b.Length; i++)
                                    {
                                        if (iter.MoveNext())
                                        {
                                            arg.b[i] = ((string)iter.Current == "true" || (string)iter.Current == "+");
                                        }
                                    }
                                }
                                break;
                            case ArgType.INT:
                                if (arg.n.Length > 0)
                                {
                                    int i = 0;
                                    if (arg.first_value_set)
                                        i++;
                                    for (; i < arg.n.Length; i++)
                                    {
                                        if (iter.MoveNext())
                                        {
                                            arg.n[i] = Convert.ToInt32(iter.Current);
                                        }
                                    }
                                }
                                break;
                            case ArgType.STRING:
                                if (arg.str.Length > 0)
                                {
                                    int i = 0;
                                    if (arg.first_value_set)
                                        i++;
                                    for (; i < arg.str.Length; i++)
                                    {
                                        if (iter.MoveNext())
                                        {
                                            arg.str[i] = (string)iter.Current;
                                        }
                                    }
                                }
                                break;
                        }
                        block.mpiexec_args.Add(arg);
                    }
                    else
                    {
                        block.command = arg.name;
                        ArrayList cmd_args = new ArrayList();
                        bool new_block_started = false;
                        while (iter.MoveNext())
                        {
                            if ((string)iter.Current == ":")
                            {
                                //block.cmd_args = (string[])cmd_args.ToArray();
                                block.cmd_args = new string[cmd_args.Count];
                                Array.Copy(cmd_args.ToArray(), block.cmd_args, cmd_args.Count);
                                arg_blocks.Add(block);
                                block = new ArgBlock();
                                new_block_started = true;
                                break;
                            }
                            else
                            {
                                cmd_args.Add((string)iter.Current);
                            }
                        }
                        if (!new_block_started)
                        {
                            //block.cmd_args = (string[])cmd_args.ToArray();
                            block.cmd_args = new string[cmd_args.Count];
                            Array.Copy(cmd_args.ToArray(), block.cmd_args, cmd_args.Count);
                            break;
                        }
                    }
                }
                arg_blocks.Add(block);
            }
            foreach (ArgBlock block in arg_blocks)
            {
                Console.WriteLine("Block:");
                Console.WriteLine("command: {0}", block.command);
                foreach (string arg in block.cmd_args)
                {
                    Console.WriteLine("cmd_arg: {0}", arg);
                }
                foreach (ArgInstance arg in block.mpiexec_args)
                {
                    arg.Print();
                }
            }
        }
    }
}
