using System;
using System.IO;
using System.Text;
using System.Security.Cryptography;

namespace DeinoMPI
{
    class PasswordEncryption
    {
        // Random data that must be the same for encryption and decryption.  Should we use the password bytes instead?
        private static byte[] initVectorBytes = new byte[16] { 32, 182, 244, 23, 1, 0, 113, 87, 3, 203, 155, 57, 77, 7, 199, 2 };
        private const int key_byte_size = 32; // 32 == 256 bits

        static public string Encrypt(string text, string password)
        {
            byte[] saltValueBytes = UnicodeEncoding.Unicode.GetBytes(password);
            Array.Reverse(saltValueBytes);
            byte[] plainTextBytes = UnicodeEncoding.Unicode.GetBytes(text);

            // First, we must create a password, from which the key will be derived.
            PasswordDeriveBytes derived_password = new PasswordDeriveBytes(password, saltValueBytes, "SHA1", 4);

            // Use the derived_password to generate pseudo-random bytes for the encryption key.
            byte[] keyBytes = derived_password.GetBytes(key_byte_size);

            RijndaelManaged symmetricKey = new RijndaelManaged();
            symmetricKey.Mode = CipherMode.CBC;

            // Generate encryptor from the existing key bytes and initialization vector.
            // Key size will be defined based on the number of the key bytes.
            ICryptoTransform encryptor = symmetricKey.CreateEncryptor(keyBytes, initVectorBytes);
            MemoryStream memoryStream = new MemoryStream();
            CryptoStream cryptoStream = new CryptoStream(memoryStream, encryptor, CryptoStreamMode.Write);

            // Encrypt the plaintext bytes into the memory stream.
            cryptoStream.Write(plainTextBytes, 0, plainTextBytes.Length);
            cryptoStream.FlushFinalBlock();

            // Get the encrypted bytes stored in the memory stream.
            byte[] cipherTextBytes = memoryStream.ToArray();

            memoryStream.Close();
            cryptoStream.Close();

            // Convert encrypted data into a base64-encoded string.
            string cipherText = Convert.ToBase64String(cipherTextBytes);

            return cipherText;
        }

        static public string Decrypt(string cipher_text, string password)
        {
            byte[] saltValueBytes = UnicodeEncoding.Unicode.GetBytes(password);
            Array.Reverse(saltValueBytes);
            byte[] cipherTextBytes = Convert.FromBase64String(cipher_text);

            // First, we must create a password, from which the key will be derived.
            PasswordDeriveBytes derived_password = new PasswordDeriveBytes(password, saltValueBytes, "SHA1", 4);

            // Use the derived_password to generate pseudo-random bytes for the encryption key.
            byte[] keyBytes = derived_password.GetBytes(key_byte_size);

            RijndaelManaged symmetricKey = new RijndaelManaged();
            symmetricKey.Mode = CipherMode.CBC;

            // Generate decryptor from the existing key bytes and initialization vector.
            ICryptoTransform decryptor = symmetricKey.CreateDecryptor(keyBytes, initVectorBytes);

            MemoryStream memoryStream = new MemoryStream(cipherTextBytes);
            CryptoStream cryptoStream = new CryptoStream(memoryStream, decryptor, CryptoStreamMode.Read);

            // Allocate a buffer long enough to hold the ciphertext.
            // Since the plaintext is never longer than the ciphertext use the ciphertext length.
            byte[] plainTextBytes = new byte[cipherTextBytes.Length];

            int decryptedByteCount = cryptoStream.Read(plainTextBytes, 0, plainTextBytes.Length);

            memoryStream.Close();
            cryptoStream.Close();

            // Convert decrypted data into a string. 
            string plainText = Encoding.Unicode.GetString(plainTextBytes, 0, decryptedByteCount);

            return plainText;
        }
    }
}
