using System;
using System.Collections.Generic;
using System.Text;
using System.Collections;
using Microsoft.Win32;

namespace DeinoMPI
{
    public class PublicKeyStore
    {
        const string PublicKeyStoreRegistryKey = @"Software\Deino\PublicKeyStore";

        static public bool AutoAcceptPublicKeys()
        {
            try
            {
                RegistryKey reg_key;
                reg_key = Registry.LocalMachine.OpenSubKey(PublicKeyStoreRegistryKey, false);
                if (reg_key != null)
                {
                    bool auto_accept = true; // default value is true
                    auto_accept = (((string)reg_key.GetValue("auto_accept_public_keys", auto_accept.ToString())) == true.ToString());
                    reg_key.Close();
                    return auto_accept;
                }
            }
            catch (Exception)
            {
            }
            return true; // default value is true
        }

        static public void SetAutoAcceptPublicKeys(bool setting)
        {
            try
            {
                RegistryKey reg_key;
                reg_key = Registry.LocalMachine.CreateSubKey(PublicKeyStoreRegistryKey);
                if (reg_key != null)
                {
                    reg_key.SetValue("auto_accept_public_keys", setting);
                    reg_key.Close();
                }
            }
            catch (Exception)
            {
            }
        }

        static public SecurityKeys GetPublicKey(string key_hash)
        {
            try
            {
                RegistryKey reg_key;
                reg_key = Registry.LocalMachine.OpenSubKey(PublicKeyStoreRegistryKey, false);
                if (reg_key != null)
                {
                    object o;
                    o = reg_key.GetValue(key_hash, null);
                    if (o != null)
                    {
                        try
                        {
                            SecurityKeys key = new SecurityKeys((string)o);
                            reg_key.Close();
                            return key;
                        }
                        catch
                        {
                        }
                    }
                    reg_key.Close();
                }
            }
            catch (Exception)
            {
            }
            return null;
        }

        static public string[] GetPublicKeyHashes()
        {
            try
            {
                RegistryKey reg_key;
                reg_key = Registry.LocalMachine.OpenSubKey(PublicKeyStoreRegistryKey, false);
                if (reg_key != null)
                {
                    string[] hashes = reg_key.GetValueNames();
                    reg_key.Close();
                    if (hashes != null && hashes.Length > 0)
                    {
                        bool found = false;
                        foreach (string entry in hashes)
                        {
                            if (entry == "auto_accept_public_keys")
                                found = true;
                        }
                        if (found)
                        {
                            string[] temp = new string[hashes.Length - 1];
                            int i = 0;
                            foreach (string entry in hashes)
                            {
                                if (entry != "auto_accept_public_keys")
                                {
                                    temp[i++] = entry;
                                }
                            }
                            hashes = temp;
                        }
                        return hashes;
                    }
                }
            }
            catch (Exception)
            {
            }
            return null;
        }

        static public bool RemovePublicKey(SecurityKeys key)
        {
            try
            {
                RegistryKey reg_key;
                reg_key = Registry.LocalMachine.OpenSubKey(PublicKeyStoreRegistryKey, true);
                if (reg_key != null)
                {
                    reg_key.DeleteValue(key.Hash, false);
                    reg_key.Close();
                }
            }
            catch (Exception)
            {
                return false;
            }
            return true;
        }

        static public bool RemovePublicKey(string key_hash)
        {
            try
            {
                RegistryKey reg_key;
                reg_key = Registry.LocalMachine.OpenSubKey(PublicKeyStoreRegistryKey, true);
                if (reg_key != null)
                {
                    reg_key.DeleteValue(key_hash, false);
                    reg_key.Close();
                }
            }
            catch (Exception)
            {
                return false;
            }
            return true;
        }

        static public bool ClearAllKeys()
        {
            try
            {
                RegistryKey reg_key;
                reg_key = Registry.LocalMachine.OpenSubKey(PublicKeyStoreRegistryKey, true);
                if (reg_key != null)
                {
                    string[] hashes = reg_key.GetValueNames();
                    if (hashes != null && hashes.Length > 0)
                    {
                        foreach (string entry in hashes)
                        {
                            if (entry != "auto_accept_public_keys")
                            {
                                reg_key.DeleteValue(entry);
                            }
                        }
                    }
                    reg_key.Close();
                }
            }
            catch (Exception)
            {
                return false;
            }
            return true;
        }

        static public bool SavePublicKey(SecurityKeys key)
        {
            try
            {
                RegistryKey reg_key;
                reg_key = Registry.LocalMachine.CreateSubKey(PublicKeyStoreRegistryKey);
                if (reg_key != null)
                {
                    reg_key.SetValue(key.Hash, key.PublicKey);
                    reg_key.Close();
                    return true;
                }
            }
            catch (Exception)
            {
            }
            return false;
        }
    }
}
