using System;
using System.Collections.Generic;
using System.Text;
using System.Xml;
using System.IO;
using System.Windows.Forms;

namespace DeinoMPI
{
    public class SecureCredential
    {
        private string password;
        private string account;
        private SecurityKeys keys;

        #region Getters and Setters
        public string Account
        {
            get
            {
                return account;
            }
            set
            {
                account = value;
            }
        }

        public string Password
        {
            get
            {
                return password;
            }
            set
            {
                password = value;
            }
        }

        public string SecureXml
        {
            get
            {
                return ToXml();
            }
        }
        #endregion

        public SecureCredential(SecurityKeys keys)
        {
            this.keys = keys;
        }

        public SecureCredential(SecurityKeys keys, string xml)
        {
            try
            {
                this.keys = keys;
                FromXml(xml);
            }
            catch
            {
                keys = null;
                account = null;
                password = null;
            }
        }

        static public SecureCredential FromSecureXml(SecurityKeys keys, string xml)
        {
            SecureCredential cred = new SecureCredential(keys, xml);
            if (cred.keys != null && cred.account != null)
                return cred;
            return null;
        }

        private string ToXml()
        {
            try
            {
                StringBuilder sb = new StringBuilder();
                using (XmlWriter wr = XmlWriter.Create(sb))
                {
                    wr.WriteStartDocument();
                    wr.WriteStartElement("SecureCredential");
                    wr.WriteElementString("Account", account);
                    wr.WriteElementString("Password", keys.Encrypt(password));
                    wr.WriteEndElement(); // SecureCredential
                    wr.WriteEndDocument();
                    wr.Close();
                }
                return sb.ToString();
            }
            catch //(Exception e)
            {
                //MessageBox.Show("ToXml failed: " + e.Message);
                return null;
            }
        }

        private void FromXml(string str)
        {
            MemoryStream ms = new MemoryStream(UnicodeEncoding.Unicode.GetBytes(str));
            StreamReader sr = new StreamReader(ms, new UnicodeEncoding());
            XmlReaderSettings settings = new XmlReaderSettings();
            settings.IgnoreComments = true;
            settings.IgnoreWhitespace = true;
            using (XmlReader rd = XmlReader.Create(sr, settings))
            {
                rd.ReadStartElement("SecureCredential");
                account = rd.ReadElementString("Account");
                password = keys.Decrypt(rd.ReadElementString("Password"));
                rd.ReadEndElement(); // SecureCredential
                rd.Close();
            }
        }
    }
}
