using System;
using System.Collections.Generic;
using System.Text;
using System.Security.Cryptography;
using System.IO;
using System.Xml;
using Microsoft.Win32;

namespace DeinoMPI
{
    public class SecurityKeys
    {
        private /*System.Security.SecureString*/string password;
        private string public_key;
        private /*System.Security.SecureString*/string private_key;
        private string encrypted_private_key;
        private RSACryptoServiceProvider rsa;
        private const int rsa_key_size = 2048;
        public enum EncryptionMethod
        {
            PROTECT_DATA, /// Encrypt using the ProtectedData.Protect/Unprotect functions
            SYMMETRIC_KEY, /// Encrypt using a symmetric key derived from the password
            NONE /// Do not encrypt
        }
        private EncryptionMethod method = EncryptionMethod.PROTECT_DATA;

        #region Getters and Setters
        public EncryptionMethod Method
        {
            get
            {
                return method;
            }
            set
            {
                method = value;
            }
        }

        public string PublicKey
        {
            get
            {
                StringBuilder sb = new StringBuilder();
                WritePublicKey(sb);
                return sb.ToString();
            }
        }

        public /*System.Security.SecureString*/string PrivateKey
        {
            get
            {
                StringBuilder sb = new StringBuilder();
                WritePrivateKey(sb);
                return sb.ToString();
            }
        }

        public /*System.Security.SecureString*/string Password
        {
            get
            {
                return password;
            }
            set
            {
                password = value;
            }
        }

        public string Hash
        {
            get
            {
                MD5 md5 = MD5.Create();
                byte[] hash_bytes = md5.ComputeHash(UnicodeEncoding.Unicode.GetBytes(public_key));
                StringBuilder sb = new StringBuilder();
                foreach (byte b in hash_bytes)
                {
                    sb.AppendFormat("{0:X}", b);
                }
                return sb.ToString();
            }
        }

        private bool UseEncryption
        {
            get
            {
                if (method == EncryptionMethod.NONE)
                    return false;
                if (password != null && password.Length > 0)
                    return true;
                if (method == EncryptionMethod.PROTECT_DATA)
                    return true;
                return false;
            }
        }
        #endregion

        #region Constructors
        public SecurityKeys()
        {
        }

        public SecurityKeys(Stream s)
        {
            ReadStream(s, true);
        }

        public SecurityKeys(Stream s, bool get_password)
        {
            ReadStream(s, get_password);
        }

        public SecurityKeys(Stream s, /*System.Security.SecureString*/string password)
        {
            this.password = password;
            ReadStream(s, false);
        }

        public SecurityKeys(string private_key, /*System.Security.SecureString*/string password)
        {
            this.password = password;
            try
            {
                using (MemoryStream ms = new MemoryStream(Encoding.Unicode.GetBytes(private_key)))
                {
                    using (StreamReader sr = new StreamReader(ms, new UnicodeEncoding()))
                    {
                        Read(sr);
                        sr.Close();
                    }
                    ms.Close();
                }
            }
            catch (Exception e)
            {
                Console.WriteLine("Unable to parse the private key, error: {0}", e.Message);
            }
        }

        public static SecurityKeys Create(string private_key, string password)
        {
            SecurityKeys keys = new SecurityKeys();
            keys.Password = password;
            try
            {
                using (MemoryStream ms = new MemoryStream(Encoding.Unicode.GetBytes(private_key)))
                {
                    using (StreamReader sr = new StreamReader(ms, new UnicodeEncoding()))
                    {
                        keys.Read(sr);
                        sr.Close();
                    }
                    ms.Close();
                }
            }
            catch //(Exception e)
            {
                keys = null;
            }
            return keys;
        }

        public static bool RequiresPassword(string private_key)
        {
            SecurityKeys keys = new SecurityKeys();
            keys.Password = null;
            try
            {
                using (MemoryStream ms = new MemoryStream(Encoding.Unicode.GetBytes(private_key)))
                {
                    using (StreamReader sr = new StreamReader(ms, new UnicodeEncoding()))
                    {
                        keys.Read(sr);
                        sr.Close();
                    }
                    ms.Close();
                }
            }
            catch //(Exception e)
            {
                keys = null;
            }
            if (keys != null)
                return false;
            return true;
        }

        public static SecurityKeys Create(string public_key)
        {
            SecurityKeys keys = new SecurityKeys();
            try
            {
                using (MemoryStream ms = new MemoryStream(Encoding.Unicode.GetBytes(public_key)))
                {
                    using (StreamReader sr = new StreamReader(ms, new UnicodeEncoding()))
                    {
                        keys.Read(sr);
                        sr.Close();
                    }
                    ms.Close();
                }
            }
            catch //(Exception e)
            {
                keys = null;
            }
            return keys;
        }

        public SecurityKeys(string public_key)
        {
            try
            {
                using (MemoryStream ms = new MemoryStream(Encoding.Unicode.GetBytes(public_key)))
                {
                    using (StreamReader sr = new StreamReader(ms, new UnicodeEncoding()))
                    {
                        Read(sr);
                        sr.Close();
                    }
                    ms.Close();
                }
            }
            catch (Exception e)
            {
                Console.WriteLine("Unable to parse the public key, error: {0}", e.Message);
            }
        }
        #endregion

        public void CreateKeys()
        {
            rsa = new RSACryptoServiceProvider(rsa_key_size);
            public_key = rsa.ToXmlString(false);
            private_key = rsa.ToXmlString(true);
            if (UseEncryption)
            {
                encrypted_private_key = EncryptPrivateKey(private_key);
            }
            else
            {
                encrypted_private_key = null;
            }
        }

        public string Encrypt(string text)
        {
            byte[] text_bytes = UnicodeEncoding.Unicode.GetBytes(text);
            byte[] cipher_bytes;
            try
            {
                cipher_bytes = rsa.Encrypt(text_bytes, true);
            }
            catch
            {
                cipher_bytes = rsa.Encrypt(text_bytes, false);
            }
            string cipher_text = Convert.ToBase64String(cipher_bytes);
            return cipher_text;
        }

        public string Decrypt(string cipher_text)
        {
            byte[] cipher_bytes = Convert.FromBase64String(cipher_text);
            byte[] text_bytes;
            try
            {
                text_bytes = rsa.Decrypt(cipher_bytes, true);
            }
            catch
            {
                text_bytes = rsa.Decrypt(cipher_bytes, false);
            }
            string text = UnicodeEncoding.Unicode.GetString(text_bytes);
            return text;
        }

        public void GetPassword(bool from_console)
        {
            GetPassword(from_console, true);
        }

        private void GetPassword(bool from_console, bool creating_keys)
        {
            if (from_console)
            {
                //SecureString entered_password = new SecureString();
                StringBuilder entered_password = new StringBuilder();
                do
                {
                    if (creating_keys)
                    {
                        Console.Out.WriteLine("Please enter a passphrase to protect the private key:");
                        //Console.Out.WriteLine("(Hit enter to select no password for the private key.)");
                    }
                    else
                    {
                        Console.Out.WriteLine("Please enter the passphrase used to protect the private key:");
                    }
                    try
                    {
                        // get the first character of the password
                        ConsoleKeyInfo nextKey = Console.ReadKey(true);

                        while (nextKey.Key != ConsoleKey.Enter)
                        {
                            if (nextKey.Key == ConsoleKey.Backspace)
                            {
                                if (entered_password.Length > 0)
                                {
                                    //entered_password.RemoveAt(entered_password.Length - 1);
                                    entered_password.Remove(entered_password.Length - 1, 1);

                                    // erase the last * as well
                                    Console.Write(nextKey.KeyChar);
                                    Console.Write(" ");
                                    Console.Write(nextKey.KeyChar);
                                }
                            }
                            else
                            {
                                //entered_password.AppendChar(nextKey.KeyChar);
                                entered_password.Append(nextKey.KeyChar);
                                Console.Write("*");
                            }

                            nextKey = Console.ReadKey(true);
                        }
                    }
                    catch (InvalidOperationException)
                    {
                        // The console has been redirected so read a line from the redirected input
                        entered_password = new StringBuilder(Console.In.ReadLine());
                    }

                    Console.WriteLine();
                    if (entered_password.Length == 0)
                    {
                        if (creating_keys)
                        {
                            Console.Out.WriteLine("Are you sure you want no password for the private key (yes,no)? [no]");
                            string answer = Console.In.ReadLine();
                            answer = answer.Trim().ToLower();
                            if (answer == "yes")
                            {
                                password = null;
                                return;
                            }
                        }
                        else
                        {
                            password = null;
                            return;
                        }
                    }
                } while (entered_password.Length == 0);

                // lock the password down
                //entered_password.MakeReadOnly();
                password = entered_password.ToString();
            }
            else
            {
                // bring up a dialog to input the password
            }
        }

        public void GetEncryptionMethod(bool from_console)
        {
            if (from_console)
            {
                string entered_method;
                for (int max_tries = 0; max_tries < 5; max_tries++)
                {
                    Console.Out.WriteLine("Options for encrypting the private key:");
                    Console.Out.WriteLine(" protect   - Use the ProtectedData.Protect/Unprotect methods to encrypt.");
                    if (password != null && password.Length > 0)
                        Console.Out.WriteLine(" symmetric - Use a symmetric key to encrypt.");
                    Console.Out.WriteLine(" none      - Do not encrypt.");
                    Console.Out.WriteLine("Please enter a method to encrypt the private key: [protect]");
                    entered_method = Console.In.ReadLine();
                    if (entered_method == null)
                        entered_method = "";
                    else
                        entered_method = entered_method.Trim().ToLower();
                    if (entered_method.Length == 0)
                    {
                        method = EncryptionMethod.PROTECT_DATA;
                        return;
                    }
                    if ((entered_method == "protect") || (entered_method == "protected"))
                    {
                        method = EncryptionMethod.PROTECT_DATA;
                        return;
                    }
                    if (password != null && password.Length > 0 && entered_method == "symmetric")
                    {
                        method = EncryptionMethod.SYMMETRIC_KEY;
                        return;
                    }
                    if (entered_method == "none")
                    {
                        method = EncryptionMethod.NONE;
                        return;
                    }
                }
                Console.Out.WriteLine("Exceeded maximum number of attempts, defaulting to Protect");
                method = EncryptionMethod.PROTECT_DATA;
            }
            else
            {
                // bring up a dialog to input the password
            }
        }

        public byte[] GetKeyExchange(byte[] key)
        {
            RSAPKCS1KeyExchangeFormatter formatter = new RSAPKCS1KeyExchangeFormatter(rsa);
            byte[] key_exchange = formatter.CreateKeyExchange(key);
            return key_exchange;
        }

        public byte[] DecryptKeyExchange(byte[] exchange_blob)
        {
            RSAPKCS1KeyExchangeDeformatter deformatter = new RSAPKCS1KeyExchangeDeformatter(rsa);
            byte[] key = deformatter.DecryptKeyExchange(exchange_blob);
            return key;
        }

        public bool FromKeyFile(string filename)
        {
            try
            {
                using (StreamReader sr = new StreamReader(filename))
                {
                    if (password == null)
                        Read(sr, true);
                    else
                        Read(sr);
                    sr.Close();
                    return true;
                }
            }
            catch
            {
            }
            return false;
        }

        public bool ToKeyFile(string filename, bool include_private_key)
        {
            try
            {
                if (File.Exists(filename))
                {
                    FileAttributes a = File.GetAttributes(filename);
                    File.SetAttributes(filename, a & ~FileAttributes.ReadOnly);
                    File.Delete(filename);
                }
                using (StreamWriter sw = new StreamWriter(filename))
                {
                    if (include_private_key)
                        WritePrivateKey(sw);
                    else
                        WritePublicKey(sw);
                    sw.Close();
                    FileAttributes a = File.GetAttributes(filename);
                    File.SetAttributes(filename, a | FileAttributes.ReadOnly);
                    return true;
                }
            }
            catch (Exception e)
            {
                Console.WriteLine("Unable to save the keys to the file, {0}, error: {1}", filename, e.Message);
            }
            return false;
        }

        private bool ReadRegistryPrivateKey()
        {
            try
            {
                RegistryKey key;
                key = Registry.CurrentUser.OpenSubKey(@"Software\Deino\SecurityKeys", false);
                if (key != null)
                {
                    object o;
                    o = key.GetValue("private_key", null);
                    if (o != null)
                    {
                        try
                        {
                            MemoryStream ms = new MemoryStream(UnicodeEncoding.Unicode.GetBytes((string)o));
                            StreamReader sr = new StreamReader(ms, new UnicodeEncoding());
                            Read(sr, true);
                            key.Close();
                            return true;
                        }
                        catch
                        {
                        }
                    }
                    key.Close();
                }
            }
            catch (Exception)
            {
            }
            return false;
        }

        private bool ReadRegistryPublicKey()
        {
            try
            {
                RegistryKey key;
                key = Registry.CurrentUser.OpenSubKey(@"Software\Deino\SecurityKeys", false);
                if (key != null)
                {
                    object o;
                    o = key.GetValue("public_key", null);
                    if (o != null)
                    {
                        try
                        {
                            MemoryStream ms = new MemoryStream(UnicodeEncoding.Unicode.GetBytes((string)o));
                            StreamReader sr = new StreamReader(ms, new UnicodeEncoding());
                            Read(sr, true);
                            key.Close();
                            return true;
                        }
                        catch
                        {
                        }
                    }
                    key.Close();
                }
            }
            catch (Exception)
            {
            }
            return false;
        }

        public bool FromRegistryCurrentUser()
        {
            if (ReadRegistryPrivateKey())
                return true;
            if (ReadRegistryPublicKey())
                return true;
            return false;
        }

        public bool FromRegistryCurrentUserPublicKey()
        {
            return ReadRegistryPublicKey();
        }

        public bool FromRegistryCurrentUserPrivateKey()
        {
            return ReadRegistryPrivateKey();
        }

        public bool ToRegistryCurrentUser(bool include_private_key)
        {
            try
            {
                StringBuilder pub_sb = new StringBuilder();
                StringBuilder priv_sb = new StringBuilder();
                if (include_private_key)
                {
                    WritePrivateKey(priv_sb);
                }
                WritePublicKey(pub_sb);
                RegistryKey key;
                key = Registry.CurrentUser.CreateSubKey(@"Software\Deino\SecurityKeys");
                if (key != null)
                {
                    bool public_key_found = false;
                    bool private_key_found = false;
                    string[] names = key.GetValueNames();
                    if (names != null)
                    {
                        foreach (string value_name in names)
                        {
                            if (value_name == "public_key")
                                public_key_found = true;
                            if (value_name == "private_key")
                                private_key_found = true;
                        }
                    }
                    if (public_key_found)
                        key.DeleteValue("public_key");
                    if (private_key_found)
                        key.DeleteValue("private_key");
                    if (include_private_key)
                        key.SetValue("private_key", priv_sb.ToString());
                    key.SetValue("public_key", pub_sb.ToString());
                    key.Close();
                    return true;
                }
            }
            catch (Exception)
            {
            }
            return false;
        }

        private string DecryptPrivateKey(string key)
        {
            if (method == EncryptionMethod.PROTECT_DATA)
            {
                byte[] encrypted_data, password_data = null, decrypted_data;
                encrypted_data = Convert.FromBase64String(key);
                if (password != null && password.Length > 0)
                    password_data = UnicodeEncoding.Unicode.GetBytes(password);
                decrypted_data = ProtectedData.Unprotect(encrypted_data, password_data, DataProtectionScope.CurrentUser);
                return UnicodeEncoding.Unicode.GetString(decrypted_data);
            }
            return PasswordEncryption.Decrypt(key, password);
        }

        private string EncryptPrivateKey(string key)
        {
            if (method == EncryptionMethod.PROTECT_DATA)
            {
                byte[] key_data, password_data = null, encrypted_data;
                key_data = UnicodeEncoding.Unicode.GetBytes(key);
                if (password != null && password.Length > 0)
                    password_data = UnicodeEncoding.Unicode.GetBytes(password);
                encrypted_data = ProtectedData.Protect(key_data, password_data, DataProtectionScope.CurrentUser);
                return Convert.ToBase64String(encrypted_data);
            }
            return PasswordEncryption.Encrypt(key, password);
        }

        #region Read/Write keys
        private void ReadStream(Stream s, bool get_password)
        {
            try
            {
                using (StreamReader sr = new StreamReader(s))
                {
                    Read(sr, get_password);
                    sr.Close();
                }
            }
            catch (Exception e)
            {
                Console.WriteLine("Unable to parse the keys, error: {0}", e.Message);
            }
        }

        private void Read(StreamReader sr)
        {
            Read(sr, false);
        }

        private void Read(StreamReader sr, bool get_password)
        {
            XmlReaderSettings settings = new XmlReaderSettings();
            settings.IgnoreComments = true;
            settings.IgnoreWhitespace = true;
            using (XmlReader rd = XmlReader.Create(sr, settings))
            {
                rd.ReadStartElement("SecurityKeys");
                if (rd.IsStartElement())
                {
                    switch (rd.Name)
                    {
                        case "public_key":
                            string xml_public_key = rd.ReadElementString("public_key");
                            rsa = new RSACryptoServiceProvider();
                            rsa.FromXmlString(xml_public_key);
                            public_key = rsa.ToXmlString(false);
                            break;
                        case "private_key":
                            bool decrypt = false;
                            if (rd.HasAttributes)
                            {
                                string is_encrypted = rd.GetAttribute("encrypted");
                                if (is_encrypted != null)
                                {
                                    is_encrypted = is_encrypted.ToLower();
                                    if (is_encrypted == "yes")
                                    {
                                        decrypt = true;
                                    }

                                    string method_str = rd.GetAttribute("method");
                                    if (method_str == "protect")
                                    {
                                        method = EncryptionMethod.PROTECT_DATA;
                                        string has_password = rd.GetAttribute("password");
                                        if (has_password == "no")
                                            get_password = false;
                                    }
                                    if (method_str == "symmetric")
                                    {
                                        method = EncryptionMethod.SYMMETRIC_KEY;
                                    }
                                }
                            }
                            else
                            {
                                method = EncryptionMethod.NONE;
                            }
                            string xml_private_key = rd.ReadElementString("private_key");
                            if (decrypt)
                            {
                                if (get_password)
                                    GetPassword(true, false);
                                xml_private_key = DecryptPrivateKey(xml_private_key);
                            }
                            rsa = new RSACryptoServiceProvider();
                            rsa.FromXmlString(xml_private_key);
                            public_key = rsa.ToXmlString(false);
                            private_key = rsa.ToXmlString(true);
                            if (decrypt)
                                encrypted_private_key = EncryptPrivateKey(private_key);
                            break;
                    }
                }
                rd.ReadEndElement();
                rd.Close();
            }
        }

        private void WritePublicKeyStream(Stream s)
        {
            using (StreamWriter sw = new StreamWriter(s))
            {
                WritePublicKey(sw);
                sw.Close();
            }
        }

        private void WritePublicKey(StreamWriter sw)
        {
            StringBuilder sb = new StringBuilder();
            WritePublicKey(sb);
            sw.Write(sb.ToString());
        }

        private void WritePublicKey(StringBuilder sb)
        {
            using (XmlWriter wr = XmlWriter.Create(sb))
            {
                wr.WriteStartDocument();
                wr.WriteStartElement("SecurityKeys");
                wr.WriteElementString("public_key", public_key);
                wr.WriteEndElement(); // SecurityKeys
                wr.WriteEndDocument();
                wr.Close();
            }
        }

        private void WritePrivateKeyStream(Stream s)
        {
            using (StreamWriter sw = new StreamWriter(s))
            {
                WritePrivateKey(sw);
                sw.Close();
            }
        }

        private void WritePrivateKey(StreamWriter sw)
        {
            StringBuilder sb = new StringBuilder();
            WritePrivateKey(sb);
            sw.Write(sb.ToString());
        }

        private void WritePrivateKey(StringBuilder sb)
        {
            using (XmlWriter wr = XmlWriter.Create(sb))
            {
                wr.WriteStartDocument();
                wr.WriteStartElement("SecurityKeys");
                if (UseEncryption)
                {
                    wr.WriteStartElement("private_key");
                    wr.WriteAttributeString("encrypted", "yes");
                    if (method == EncryptionMethod.PROTECT_DATA)
                    {
                        wr.WriteAttributeString("method", "protect");
                        if (password != null && password.Length > 0)
                            wr.WriteAttributeString("password", "yes");
                        else
                            wr.WriteAttributeString("password", "no");
                    }
                    else
                    {
                        wr.WriteAttributeString("method", "symmetric");
                    }
                    wr.WriteString(encrypted_private_key);
                    wr.WriteEndElement(); // private_key
                }
                else
                {
                    wr.WriteElementString("private_key", private_key);
                }
                wr.WriteEndElement(); // SecurityKeys
                wr.WriteEndDocument();
                wr.Close();
            }
        }
        #endregion
    }
}
