using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using DeinoMPI;
using System.Collections;
using Microsoft.Win32;
using System.IO;
using System.Threading;
using System.Runtime.InteropServices;
using System.Net.NetworkInformation;
using System.Net;
using System.Xml;
using System.Diagnostics;
using System.Text.RegularExpressions;

namespace DeinoMPIwin
{
    public partial class DeinoMPIwin : Form
    {
        Mpiexec mpiexec, stager_mpiexec, verify_mpiexec;
        ProcessGroup stager_pg;
        CredentialStore store;
        const string SettingsRegistryKey = @"Software\Deino\DeinoMPI";
        const string DeinoMPIwebPage = "http://mpi.deino.net";
        const string CredentialRequiredMessage = "In order to retrieve host information an account must be selected from the Credential store first.";
        ArrayList extra_controls;
        int hidden_Y, visisble_Y;
        bool removable_media_loaded = false;
        bool web_tab_opened = false;
        bool cluster_tab_opened = false;
        public ArrayList job_threads;
        Thread watcher_thread;
        public delegate void AppendTextDelegate(string str);
        public delegate void PrintExitcodesDelegate();
        public delegate void EnableExecuteDelegate();
        private StringBuilder current_input;
        private Color output_background_color = Color.FromKnownColor(KnownColor.Window);
        bool domain_populated = false;
        string onload_job_filename = null;
        bool onload_execute_job = false;
        private ListViewColumnSorter lvwColumnSorter;

        public void PrintExitcodes()
        {
            if (mpiexec != null)
                mpiexec.PrintExitcodes();
        }

        public void AppendText(string str)
        {
            //output_richTextBox.AppendText(str.Replace("\n", "\r\n"));
            output_richTextBox.AppendText(str.Replace("\r", ""));
        }

        private Thread wait_for_stager_thread = null;
        private void WaitForStager()
        {
            try
            {
                stager_mpiexec.input.EnqueueInput("\n");
                stager_mpiexec.WaitForAllExit();
                stager_mpiexec = null;
                stager_pg = null;
            }
            catch
            {
            }
            EnableExecuteDelegate eb = new EnableExecuteDelegate(EnableExecute);
            Invoke(eb);
        }

        public void EnableExecute()
        {
            if (wait_for_stager_thread != null)
            {
                //wait_for_stager_thread.Join(); // Can't join because the thread won't exit until this call returns
                wait_for_stager_thread = null;
            }
            else if (stager_mpiexec != null)
            {
                ThreadStart ts = new ThreadStart(WaitForStager);
                wait_for_stager_thread = new Thread(ts);
                wait_for_stager_thread.Start();
                return;
            }

            execute_button.Enabled = true;
            break_button.Enabled = false;
            show_msg_queues_button.Enabled = false;
            output_richTextBox.ReadOnly = true;
            //output_richTextBox.BackColor = Color.FromArgb(255, 200, 200);
            output_richTextBox.BackColor = output_background_color;
            verify_job_button.Enabled = true;
        }

        void DisableExecute()
        {
            execute_button.Enabled = false;
            break_button.Enabled = true;
            show_msg_queues_button.Enabled = true;
            output_richTextBox.ReadOnly = false;
            output_background_color = output_richTextBox.BackColor;
            output_richTextBox.BackColor = Color.FromArgb(200, 255, 200);
            verify_job_button.Enabled = false;
        }

        public DeinoMPIwin(string []args)
        {
            InitializeComponent();

            mpiexec = null;
            stager_mpiexec = null;
            verify_mpiexec = null;
            stager_pg = null;
            store = null;
            job_threads = new ArrayList();
            current_input = new StringBuilder();
            version_delegate = new UpdateVersionDelegate(UpdateVersion);
            info_delegate = new UpdateInfoDelegate(UpdateInfo);
            verified_delegate = new UpdateVerifiedMessageDelegate(UpdateVerifyListboxItem);
            verified_error_delegate = new UpdateVerifiedErrorDelegate(UpdateVerifyListboxItemError);
            verified_state_delegate = new UpdateVerifiedStateDelegate(UpdateVerifyListboxItemState);
            get_credential_storeAccount_delegate = new CredentialStoreAccountDelegate(GetCredentialStoreAccount);
            lvwColumnSorter = new ListViewColumnSorter();
            hosts_listView.ListViewItemSorter = lvwColumnSorter;

            // add all controls here that change visibility when the extra options are expanded or contracted
            extra_controls = new ArrayList();
            extra_controls.Add(options_tabControl);
            extra_controls.Add(hosts_textBox);
            extra_controls.Add(hosts_label);
            extra_controls.Add(env_vars_textBox);
            extra_controls.Add(env_label);
            extra_controls.Add(wdir_textBox);
            extra_controls.Add(wdir_browse_button);
            extra_controls.Add(wdir_label);
            extra_controls.Add(channel_comboBox);
            extra_controls.Add(channel_label);
            extra_controls.Add(show_msg_queues_button);
            extra_controls.Add(load_job_button);
            extra_controls.Add(save_job_button);
            extra_controls.Add(exitcodes_checkBox);
            extra_controls.Add(log_checkBox);
            extra_controls.Add(jumpshot_button);
            extra_controls.Add(localroot_checkBox);
            extra_controls.Add(localonly_checkBox);
            extra_controls.Add(map_label);
            extra_controls.Add(map_textBox);
            extra_controls.Add(use_configfile_checkBox);
            extra_controls.Add(configfile_textBox);
            extra_controls.Add(configfile_browse_button);
            extra_controls.Add(configfile_label);
            extra_controls.Add(use_history_checkBox);
            extra_controls.Add(use_smp_optimizations_checkBox);
            extra_controls.Add(use_staging_checkBox);
            extra_controls.Add(staging_src_dir_textBox);
            extra_controls.Add(staging_src_dir_browse_button);
            extra_controls.Add(staging_src_label);
            extra_controls.Add(staging_recursive_checkBox);
            extra_controls.Add(staging_wildcards_textBox);
            extra_controls.Add(staging_wildcards_label);
            extra_controls.Add(staging_dest_dir_textBox);
            extra_controls.Add(staging_dest_label);
            extra_controls.Add(staging_pull_checkBox);

            this.Text = String.Format("DeinoMPI {0}", ServerSession.Version);

            // Check the command line parameters
            if (args != null)
            {
                // The command line can contain a job file name and the /execute flag if the job should be started immediately
                foreach (string arg in args)
                {
                    if (arg.ToLower() == "/execute")
                    {
                        onload_execute_job = true;
                    }
                    else
                    {
                        onload_job_filename = arg;
                    }
                }
            }
        }

        private bool GetLastSettings()
        {
            try
            {
                RegistryKey reg_key;
                reg_key = Registry.CurrentUser.OpenSubKey(SettingsRegistryKey, false);
                if (reg_key != null)
                {
                    app_comboBox.Text = (string)reg_key.GetValue("application", "");
                    numericUpDown.Value = (decimal)Convert.ToDecimal(reg_key.GetValue("nproc", "1"));
                    hosts_textBox.Text = (string)reg_key.GetValue("hosts", "");
                    env_vars_textBox.Text = (string)reg_key.GetValue("environment", "");
                    wdir_textBox.Text = (string)reg_key.GetValue("wdir", "");
                    channel_comboBox.Text = (string)reg_key.GetValue("channel", "sock");
                    credential_comboBox.Text = (string)reg_key.GetValue("account", "");
                    extra_options_checkBox.Checked = (bool)Convert.ToBoolean(reg_key.GetValue("expanded", "false"));
                    exitcodes_checkBox.Checked = (bool)Convert.ToBoolean(reg_key.GetValue("exitcodes", "false"));
                    log_checkBox.Checked = (bool)Convert.ToBoolean(reg_key.GetValue("log", "false"));
                    jumpshot_button.Enabled = log_checkBox.Checked;
                    localroot_checkBox.Checked = (bool)Convert.ToBoolean(reg_key.GetValue("localroot", "false"));
                    localonly_checkBox.Checked = (bool)Convert.ToBoolean(reg_key.GetValue("localonly", "false"));
                    if (localroot_checkBox.Checked && localonly_checkBox.Checked)
                        localroot_checkBox.Checked = false;
                    map_textBox.Text = (string)reg_key.GetValue("map", "");
                    staging_src_dir_textBox.Text = (string)reg_key.GetValue("staging_src", "");
                    staging_dest_dir_textBox.Text = (string)reg_key.GetValue("staging_dest", "");
                    staging_recursive_checkBox.Checked = (bool)Convert.ToBoolean(reg_key.GetValue("staging_subdirs", "false"));
                    staging_pull_checkBox.Checked = (bool)Convert.ToBoolean(reg_key.GetValue("staging_pull", "true"));
                    staging_wildcards_textBox.Text = (string)reg_key.GetValue("staging_wildcards", "");
                    use_staging_checkBox.Checked = (bool)Convert.ToBoolean(reg_key.GetValue("use_staging", "false"));
                    configfile_textBox.Text = (string)reg_key.GetValue("cfg", "");
                    use_configfile_checkBox.Checked = (bool)Convert.ToBoolean(reg_key.GetValue("use_cfg", "false"));
                    string wstate = (string)reg_key.GetValue("window_state", "Normal");
                    switch (wstate)
                    {
                        case "Normal":
                            WindowState = FormWindowState.Normal;
                            break;
                        case "Maximized":
                            WindowState = FormWindowState.Maximized;
                            break;
                        case "Minimized":
                            WindowState = FormWindowState.Minimized;
                            break;
                    }
                    if (WindowState == FormWindowState.Normal)
                    {
                        int x = -1, y = -1;
                        int width = -1, height = -1;
                        string[] names = reg_key.GetValueNames();
                        if (names != null)
                        {
                            foreach (string name in names)
                            {
                                if (name == "X")
                                    x = Convert.ToInt32(reg_key.GetValue("X", "-1"));
                                if (name == "Y")
                                    y = Convert.ToInt32(reg_key.GetValue("Y", "-1"));
                                if (name == "width")
                                    width = Convert.ToInt32(reg_key.GetValue("width", "-1"));
                                if (name == "height")
                                    height = Convert.ToInt32(reg_key.GetValue("height", "-1"));
                            }
                        }
                        if (x != -1 && y != -1)
                        {
                            Location = new Point(x, y);
                        }
                        if (width != -1 && height != -1)
                        {
                            Size = new Size(width, height);
                        }
                    }
                    reg_key.Close();
                }
                else
                {
                    return false;
                }
            }
            catch
            {
            }
            return true;
        }

        private void SaveSettings()
        {
            try
            {
                RegistryKey reg_key;
                reg_key = Registry.CurrentUser.CreateSubKey(SettingsRegistryKey);
                if (reg_key != null)
                {
                    reg_key.SetValue("application", app_comboBox.Text);
                    reg_key.SetValue("nproc", numericUpDown.Value);
                    reg_key.SetValue("hosts", hosts_textBox.Text);
                    reg_key.SetValue("environment", env_vars_textBox.Text);
                    reg_key.SetValue("wdir", wdir_textBox.Text);
                    reg_key.SetValue("channel", channel_comboBox.Text);
                    reg_key.SetValue("account", credential_comboBox.Text);
                    reg_key.SetValue("expanded", extra_options_checkBox.Checked);
                    switch (WindowState)
                    {
                        case FormWindowState.Normal:
                            reg_key.SetValue("window_state", "Normal");
                            break;
                        case FormWindowState.Maximized:
                            reg_key.SetValue("window_state", "Maximized");
                            break;
                        case FormWindowState.Minimized:
                            reg_key.SetValue("window_state", "Minimized");
                            break;
                    }
                    if (WindowState == FormWindowState.Normal)
                    {
                        reg_key.SetValue("X", Location.X);
                        reg_key.SetValue("Y", Location.Y);
                        reg_key.SetValue("width", Size.Width);
                        reg_key.SetValue("height", Size.Height);
                    }
                    reg_key.SetValue("exitcodes", exitcodes_checkBox.Checked);
                    reg_key.SetValue("log", log_checkBox.Checked);
                    reg_key.SetValue("localroot", localroot_checkBox.Checked);
                    reg_key.SetValue("localonly", localonly_checkBox.Checked);
                    reg_key.SetValue("map", map_textBox.Text);
                    reg_key.SetValue("cfg", configfile_textBox.Text);
                    reg_key.SetValue("use_cfg", use_configfile_checkBox.Checked);
                    reg_key.SetValue("use_staging", use_staging_checkBox.Checked);
                    reg_key.SetValue("staging_src", staging_src_dir_textBox.Text);
                    reg_key.SetValue("staging_dest", staging_dest_dir_textBox.Text);
                    reg_key.SetValue("staging_subdirs", staging_recursive_checkBox.Checked);
                    reg_key.SetValue("staging_pull", staging_pull_checkBox.Checked);
                    reg_key.SetValue("staging_wildcards", staging_wildcards_textBox.Text);
                    reg_key.Close();
                }
            }
            catch
            {
            }
        }

        private void HideExtraOptions()
        {
            // do nothing if already hidden
            if (output_richTextBox.Location.Y == hidden_Y)
                return;

            // hide the extra controls and labels
            foreach (Control c in extra_controls)
            {
                c.Hide();
            }

            // reposition the output box
            int height = output_richTextBox.Height + (visisble_Y - hidden_Y);
            output_richTextBox.Location = new Point(output_richTextBox.Location.X, hidden_Y);
            output_richTextBox.Height = height;
        }

        private void ShowExtraOptions()
        {
            // do nothing if already visible
            if (output_richTextBox.Location.Y == visisble_Y)
                return;

            // reposition the output box
            int height = output_richTextBox.Height - (visisble_Y - hidden_Y);
            output_richTextBox.Location = new Point(output_richTextBox.Location.X, visisble_Y);
            output_richTextBox.Height = height;

            // show the extra controls and labels
            foreach (Control c in extra_controls)
            {
                c.Show();
            }
        }

        private void EnableExtraOptions()
        {
            foreach (Control c in extra_controls)
            {
                if (c != show_msg_queues_button && c != load_job_button && c != save_job_button)
                    c.Enabled = true;
            }
            hosts_label.Enabled = !localonly_checkBox.Checked;
            hosts_textBox.Enabled = !localonly_checkBox.Checked;
            jumpshot_button.Enabled = log_checkBox.Checked;
            configfile_textBox.Enabled = false;
            configfile_browse_button.Enabled = false;
            configfile_label.Enabled = false;
            //configfile_help_button.Enabled = false;
            app_comboBox.Enabled = true;
            browse_application_button.Enabled = true;
            numericUpDown.Enabled = true;
            num_procs_label.Enabled = true;
            staging_src_dir_textBox.Enabled = use_staging_checkBox.Checked;
            staging_src_dir_browse_button.Enabled = use_staging_checkBox.Checked;
            staging_src_label.Enabled = use_staging_checkBox.Checked;
            staging_recursive_checkBox.Enabled = use_staging_checkBox.Checked;
            staging_wildcards_textBox.Enabled = use_staging_checkBox.Checked;
            staging_wildcards_label.Enabled = use_staging_checkBox.Checked;
            staging_dest_dir_textBox.Enabled = use_staging_checkBox.Checked;
            staging_dest_label.Enabled = use_staging_checkBox.Checked;
            staging_pull_checkBox.Enabled = use_staging_checkBox.Checked;
        }

        private void DisableExtraOptions()
        {
            foreach (Control c in extra_controls)
            {
                if (c != show_msg_queues_button && c != load_job_button && c != save_job_button && c != options_tabControl)
                    c.Enabled = false;
            }
            use_configfile_checkBox.Enabled = true;
            configfile_textBox.Enabled = true;
            configfile_browse_button.Enabled = true;
            configfile_label.Enabled = true;
            app_comboBox.Enabled = false;
            browse_application_button.Enabled = false;
            numericUpDown.Enabled = false;
            num_procs_label.Enabled = false;
        }

        private void DeinoMPIwin_Load(object sender, EventArgs e)
        {
            //hidden_Y = hosts_textBox.Location.Y; // This requires that the host text box be just below the "more options" check box.
            hidden_Y = extra_options_checkBox.Location.Y + extra_options_checkBox.Height + 5;
            visisble_Y = output_richTextBox.Location.Y;
            HideExtraOptions();
            break_button.Enabled = false;
            show_msg_queues_button.Enabled = false;
            output_richTextBox.ReadOnly = true;
            bool first_time = !GetLastSettings();
            new_keys_button.Enabled = enable_create_keys_checkBox.Checked;

            if (first_time && !CredentialStore.Exists())
            {
                Color lback_color, cback_color;
                // save the old background color
                lback_color = location_groupBox.BackColor;
                cback_color = create_store_groupBox.BackColor;
                // set the new background to an obvious color
                location_groupBox.BackColor = Color.Aquamarine;
                create_store_groupBox.BackColor = Color.Aquamarine;
                // show the create credentials group
                enable_create_checkBox.Checked = true;
                SetCreateCredentialsEnabled(enable_create_checkBox.Checked);
                // select the keys page
                tabControl.SelectedTab = keys_tabPage;
                // Inform the user that they need to create a credential store or select an existing one
                CreateCredentailStoreWarning dlg = new CreateCredentailStoreWarning();
                dlg.ShowDialog();
                location_groupBox.BackColor = lback_color;
                create_store_groupBox.BackColor = cback_color;
            }
            else
            {
                SetCreateCredentialsEnabled(enable_create_checkBox.Checked);
            }

            if (onload_job_filename != null)
            {
                LoadJobFromFile(onload_job_filename);
                if (onload_execute_job)
                {
                    execute_button_Click(null, null);
                }
            }
        }

        private void extra_options_checkBox_CheckedChanged(object sender, EventArgs e)
        {
            if (extra_options_checkBox.Checked)
            {
                ShowExtraOptions();
            }
            else
            {
                HideExtraOptions();
            }
        }

        private void browse_application_button_Click(object sender, EventArgs e)
        {
            string[] args = Mpiexec.ParseStrings(app_comboBox.Text);
            OpenFileDialog dlg = new OpenFileDialog();
            dlg.Filter = "Executable files (*.exe)|*.exe|All files (*.*)|*.*";
            if (args != null)
            {
                if (File.Exists(args[0]))
                    dlg.FileName = args[0]; //app_comboBox.Text;
            }
            if (dlg.ShowDialog() == DialogResult.OK)
            {
                if (dlg.FileName.IndexOf(' ') != -1)
                {
                    app_comboBox.Text = "\"" + dlg.FileName + "\"";
                }
                else
                {
                    app_comboBox.Text = dlg.FileName;
                }
            }
        }

        private ArrayList GenerateLaunchNodes(bool update_mpiexec_tab, ref Mpiexec mpiexec, ref string error_str)
        {
            ArrayList nodes;

            if (use_configfile_checkBox.Checked)
            {
                if (!File.Exists(configfile_textBox.Text.Trim()))
                {
                    if (update_mpiexec_tab)
                    {
                        MessageBox.Show("Please enter a valid configuration file before executing the job.");
                        configfile_textBox.Focus();
                    }
                    else
                    {
                        error_str = "Please enter a valid configuration file before executing the job.";
                    }
                    return null;
                }
            }
            else
            {
                if (app_comboBox.Text.Trim().Length < 1)
                {
                    if (update_mpiexec_tab)
                    {
                        MessageBox.Show("Please enter the path to an executable before executing a job.");
                        app_comboBox.Focus();
                    }
                    else
                    {
                        error_str = "Please enter the path to an executable before executing a job.";
                    }
                    return null;
                }
            }

            if (update_mpiexec_tab)
            {
                // disable the execute button so it cannot be pushed twice
                DisableExecute();
                output_richTextBox.Clear();
            }

            if (store == null)
            {
                if (CredentialStore.RequiresPassword())
                {
                    if (update_mpiexec_tab)
                    {
                        MessageBox.Show("Please log in to the Credential Store before launching a job.");
                        EnableExecute();
                    }
                    else
                    {
                        error_str = "Please log in to the Credential Store before launching a job.";
                    }
                    return null;
                }
                else
                {
                    store = new CredentialStore("");
                    store_password_textBox.Text = "";
                    store_password_textBox.Enabled = false;
                    login_button.Enabled = false;
                    LoadDialogFromStore();
                }
            }
            else
            {
                if (!store.UseRegistry)
                {
                    if (!File.Exists(store.FileName))
                    {
                        string str = String.Format("The Credential store is not available: {0}\r\nThis can be a result of the store being on a removable media device that is not plugged in\r\nor it could be that the store has been deleted.\r\nPlease select a valid Credential store or create a new one and then try again.", store.FileName);
                        if (update_mpiexec_tab)
                        {
                            output_richTextBox.Text = str;
                            EnableExecute();
                        }
                        else
                        {
                            error_str = str;
                        }
                        return null;
                    }
                }
            }

            if (use_configfile_checkBox.Checked)
            {
                try
                {
                    mpiexec = new Mpiexec();
                    mpiexec.TotalNumProcs = 0; // Reset the total number of processes
                    string[] args = new string[2] { "-configfile", configfile_textBox.Text.Trim() };
                    nodes = mpiexec.ParseArgs(args);
                    if (nodes == null)
                    {
                        if (update_mpiexec_tab)
                            EnableExecute();
                        return null;
                    }
                    int result = 0;
                    if (mpiexec.global_options.localonly)
                    {
                        result = mpiexec.GetStore(nodes);
                        foreach (LaunchNode node in nodes)
                        {
                            node.launch_local = true;
                        }
                    }
                    else
                    {
                        result = mpiexec.GetStoreAndCredential(nodes);
                    }
                    if (result != 0)
                    {
                        if (update_mpiexec_tab)
                            EnableExecute();
                        return null;
                    }
                    if (update_mpiexec_tab)
                    {
                        LaunchNode tnode = (LaunchNode)nodes[0];
                        app_comboBox.Text = tnode.exe;
                        numericUpDown.Value = tnode.nproc;
                    }
                }
                catch (Exception x)
                {
                    string str = String.Format("Unable to parse the configuration file:\r\n{0}\r\nError:\r\n{1}", configfile_textBox.Text.Trim(), x.Message);
                    if (update_mpiexec_tab)
                    {
                        output_richTextBox.Text = str;
                        EnableExecute();
                    }
                    else
                    {
                        error_str = str;
                    }
                    return null;
                }
            }
            else
            {
                SecureCredential credential = null;
                if (!localonly_checkBox.Checked)
                {
                    credential = store.GetCredential(credential_comboBox.Text);
                    if (credential == null)
                    {
                        if (credential_comboBox.Text.Trim().Length == 0)
                        {
                            if (credential_comboBox.Items.Count > 0)
                            {
                                if (update_mpiexec_tab)
                                {
                                    MessageBox.Show("Please enter an account from the Credential store before launching a job.");
                                    credential_comboBox.Focus();
                                    EnableExecute();
                                }
                                else
                                {
                                    error_str = "Please enter an account from the Credential store before launching a job.";
                                }
                                return null;
                            }
                            else
                            {
                                UsernameAndPassword dlg = new UsernameAndPassword();
                                if (dlg.ShowDialog() == DialogResult.OK)
                                {
                                    credential = new SecureCredential(store.Keys);
                                    credential.Account = dlg.UserName;
                                    credential.Password = dlg.Password;
                                }
                                else
                                {
                                    if (update_mpiexec_tab)
                                        EnableExecute();
                                    return null;
                                }
                            }
                        }
                        else
                        {
                            string str = String.Format("The account '{0}' does not exist in the Credential store", credential_comboBox.Text);
                            if (update_mpiexec_tab)
                                MessageBox.Show(str);
                            else
                                error_str = str;
                        }
                        if (credential == null)
                        {
                            if (update_mpiexec_tab)
                                EnableExecute();
                            return null;
                        }
                    }
                }

                if (update_mpiexec_tab)
                {
                    // Add the command to the application drop down if it has not already been added
                    bool found = false;
                    foreach (string str in app_comboBox.Items)
                    {
                        if (str == app_comboBox.Text)
                            found = true;
                    }
                    if (!found)
                    {
                        app_comboBox.Items.Add(app_comboBox.Text);
                    }
                }

                // Create the launch structures based on the dialog data
                mpiexec = new Mpiexec();
                mpiexec.global_options.exitcodes = exitcodes_checkBox.Checked;
                mpiexec.global_options.localonly = localonly_checkBox.Checked;
                mpiexec.global_options.localroot = localroot_checkBox.Checked;
                mpiexec.TotalNumProcs = 0; // Reset the total number of processes
                //mpiexec.global_options.Clear();
                string env_vars;
                if (channel_comboBox.Text.Trim().Length > 0)
                    env_vars = String.Format("DeinoMPI_CHANNEL={0} {1}", channel_comboBox.Text, env_vars_textBox.Text);
                else
                    env_vars = env_vars_textBox.Text;
                if (log_checkBox.Checked)
                {
                    if (env_vars.Length > 0)
                        env_vars = String.Format("MPI_WRAP_DLL_NAME=mpe {0}", env_vars);
                    else
                        env_vars = "MPI_WRAP_DLL_NAME=mpe";
                }
                if (map_textBox.Text.Trim().Length > 0)
                {
                    string[] tokens = map_textBox.Text.Trim().Split(';');
                    if (tokens != null)
                    {
                        foreach (string mapping in tokens)
                        {
                            string[] parts = mapping.Split(':');
                            if (parts.Length > 1)
                            {
                                mpiexec.global_options.map[parts[0]] = parts[1];
                            }
                        }
                    }
                }
                try
                {
                    nodes = mpiexec.CreateLaunchNodes(app_comboBox.Text, (int)numericUpDown.Value, hosts_textBox.Text, env_vars/*env_vars_textBox.Text*/, wdir_textBox.Text);
                    if (nodes == null)
                    {
                        if (update_mpiexec_tab)
                        {
                            MessageBox.Show("The information entered in the dialog was not sufficient to describe a valid job.");
                            EnableExecute();
                        }
                        else
                        {
                            error_str = "The information entered in the dialog was not sufficient to describe a valid job.";
                        }
                        return null;
                    }
                }
                catch (Exception x)
                {
                    if (update_mpiexec_tab)
                    {
                        output_richTextBox.Text = x.Message;
                        EnableExecute();
                    }
                    else
                    {
                        error_str = x.Message;
                    }
                    return null;
                }
                // Add the selected user credential
                foreach (LaunchNode node in nodes)
                {
                    node.credential = credential;
                    node.private_key = store.Keys;
                }
            }
            return nodes;
        }

        private bool GenerateLaunchNodes(ref ArrayList nodes, ref Mpiexec mpiexec)
        {
            string ignored_parameter = null;
            nodes = GenerateLaunchNodes(true, ref mpiexec, ref ignored_parameter);
            if (nodes == null)
                return false;
            return true;
        }

        private ArrayList GenerateLaunchNodes(ref string error_str)
        {
            return GenerateLaunchNodes(false, ref verify_mpiexec, ref error_str);
        }

        private bool GenerateStagingNodes(ref ArrayList nodes, Mpiexec mpiexec, ArrayList launch_nodes)
        {
            string src_dir, dest_dir, wildcards;
            bool use_one_node = false;

            if (launch_nodes == null || launch_nodes.Count == 0)
            {
                nodes = null;
                return true;
            }

            // get and validate the source directory
            if (staging_src_dir_textBox.Text.Trim() == "")
            {
                MessageBox.Show("Please enter a valid path in the source directory field or uncheck the Use Directory Staging check box.");
                EnableExecute();
                staging_src_dir_textBox.Focus();
                return false;
            }
            if (!Directory.Exists(staging_src_dir_textBox.Text.Trim()))
            {
                MessageBox.Show(String.Format("The specified directory does not exist: {0}", staging_src_dir_textBox.Text));
                EnableExecute();
                staging_src_dir_textBox.Focus();
                return false;
            }
            src_dir = Path.GetFullPath(staging_src_dir_textBox.Text.Trim());

            // get and validate the destination directory
            dest_dir = staging_dest_dir_textBox.Text.Trim();
            if (staging_dest_dir_textBox.Text.IndexOfAny(Path.GetInvalidPathChars()) >= 0)
            {
                MessageBox.Show(String.Format("The specified destination directory contains invalid path characters: {0}", staging_dest_dir_textBox.Text));
                EnableExecute();
                staging_dest_dir_textBox.Focus();
                return false;
            }
            if (dest_dir.Length > 0)
            {
                if (dest_dir[0] == '.')
                {
                    MessageBox.Show(String.Format("The destination directory must be an absolute path, relative paths are not valid: {0}", staging_dest_dir_textBox.Text));
                    EnableExecute();
                    staging_dest_dir_textBox.Focus();
                    return false;
                }
                if (dest_dir.StartsWith(@"\\"))
                {
                    if (dest_dir.ToUpper() == src_dir.ToUpper())
                    {
                        // The source and destination are on a network path and they are the same so staging will have no effect
                        nodes = null;
                        return true;
                    }
                    /*
                    if (MessageBox.Show(String.Format("Are you sure you want to copy files to a network directory?\r\n{0}", dest_dir), "Warning", MessageBoxButtons.YesNo) == DialogResult.No)
                    {
                        EnableExecute();
                        staging_dest_dir_textBox.Focus();
                        return false;
                    }
                    */
                    // Use only one node since multiple nodes will only clobber the files in the shared directory
                    use_one_node = true;
                }
            }
            else
            {
                dest_dir = src_dir;
                if (dest_dir.StartsWith(@"\\"))
                {
                    // The source and destination are on a network path and they are the same so staging will have no effect
                    nodes = null;
                    return true;
                }
            }

            // FIXME: Validate the wildcard string?
            wildcards = staging_wildcards_textBox.Text.Trim();
            if (wildcards.Length == 0)
                wildcards = "*.*";

            ArrayList args = new ArrayList();
            args.Add(src_dir);
            args.Add(String.Format("/dest_dir:{0}", dest_dir));
            if (staging_recursive_checkBox.Checked)
                args.Add("/recursive");
            args.Add("/ignore_src");
            if (!staging_pull_checkBox.Checked)
                args.Add("/nopull");
            if (wildcards != "*.*")
                args.Add(String.Format("/wc:{0}", wildcards));

            LaunchNode src_node = new LaunchNode(mpiexec);
            src_node.appnum = 0;
            src_node.args = (ArrayList)args.Clone();
            src_node.clique = "0";
            src_node.credential = null;
            src_node.dir = null;
            src_node.env = null;
            src_node.exe = "deino_stager.exe";
            src_node.hostname = Environment.MachineName;
            src_node.iproc = 0;
            src_node.launch_local = true;
            src_node.map = null;
            src_node.nproc = -1; // to be determined
            src_node.path = "%DEINO_MPI_DIR%";
            src_node.phrase = null;
            src_node.pmi = null;
            src_node.port = -1;
            src_node.private_key = ((LaunchNode)launch_nodes[0]).private_key;
            src_node.secure_user = null;
            src_node.spawned = false;
            nodes = new ArrayList();
            nodes.Add(src_node);
            if (use_one_node)
            {
                src_node.nproc = 2;
                LaunchNode node = ((LaunchNode)launch_nodes[0]).Clone();
                node.appnum = 0;
                node.args = (ArrayList)args.Clone();
                node.clique = "1";
                node.exe = "deino_stager.exe";
                node.iproc = 1;
                node.nproc = 2;
                node.path = "%DEINO_MPI_DIR%";
                node.env = null;
                nodes.Add(node);
            }
            else
            {
                int i, nproc = ((LaunchNode)launch_nodes[0]).nproc;
                int[] node_root_ranks = new int[nproc];
                for (i = 0; i < nproc; i++)
                    node_root_ranks[i] = -1;
                bool all_ranks_covered = false;
                int lowest_not_covered = 0;
                while (!all_ranks_covered)
                {
                    all_ranks_covered = true;
                    for (i = lowest_not_covered; i < nproc; i++)
                    {
                        if (node_root_ranks[i] == -1)
                        {
                            lowest_not_covered = i;
                            all_ranks_covered = false;
                            int[] clique_ranks = ProcessGroup.ParseCliqueInts(((LaunchNode)launch_nodes[i]).clique);
                            if (clique_ranks != null)
                            {
                                foreach (int clique_rank in clique_ranks)
                                {
                                    node_root_ranks[clique_rank] = clique_ranks[0]; // This assumes that the first clique rank is the lowest
                                }
                            }
                            break;
                        }
                    }
                }
                Hashtable node_ranks_hash = new Hashtable();
                for (i = 0; i < nproc; i++)
                {
                    node_ranks_hash[node_root_ranks[i]] = node_root_ranks[i];
                }
                int[] node_ranks = new int[node_ranks_hash.Count];
                i = 0;
                foreach (int node_rank in node_ranks_hash.Keys)
                    node_ranks[i++] = node_rank;
                Array.Sort(node_ranks);

                src_node.nproc = node_ranks.Length + 1;
                for (i = 0; i < node_ranks.Length; i++)
                {
                    LaunchNode node = ((LaunchNode)launch_nodes[node_ranks[i]]).Clone();
                    node.appnum = 0;
                    node.args = (ArrayList)args.Clone();
                    node.clique = String.Format("{0}", i + 1);
                    node.exe = "deino_stager.exe";
                    node.iproc = i + 1;
                    node.nproc = node_ranks.Length + 1;
                    node.path = "%DEINO_MPI_DIR%";
                    node.env = null;
                    nodes.Add(node);
                }
            }

            //MessageBox.Show(String.Format("Staging:\r\n Source: {0}{1}{2}\r\n Destination: {3}", src_dir, src_dir.EndsWith("\\") ? "" : "\\", wildcards, dest_dir));
            return true;
        }

        private void execute_button_Click(object sender, EventArgs e)
        {
            ArrayList nodes = null, staging_nodes = null;

            // Wait for all groups in the previous job to exit
            if (mpiexec != null)
            {
                mpiexec.WaitForAllExit();
            }

            // Generate a list of launch nodes and a new Mpiexec space
            if (!GenerateLaunchNodes(ref nodes, ref mpiexec))
                return;

            // Generate a list of staging nodes from the list of launch nodes
            bool use_staging = false;
            if (use_configfile_checkBox.Checked)
            {
                if (mpiexec != null && mpiexec.global_options.stage_src != null)
                {
                    stager_mpiexec = new Mpiexec();
                    stager_mpiexec.global_options.use_output_queue = true;
                    string error_msg = null;
                    if (!stager_mpiexec.GenerateStagingNodes(ref staging_nodes, ref error_msg, nodes, mpiexec.global_options))
                    {
                        if (error_msg != null)
                        {
                            MessageBox.Show(error_msg, "Error parsing the directory staging options");
                        }
                        stager_mpiexec = null;
                        EnableExecute();
                        return;
                    }
                    use_staging = true;
                }
            }
            else if (use_staging_checkBox.Checked)
            {
                stager_mpiexec = new Mpiexec();
                stager_mpiexec.global_options.use_output_queue = true;
                if (!GenerateStagingNodes(ref staging_nodes, stager_mpiexec, nodes))
                {
                    stager_mpiexec = null;
                    EnableExecute();
                    return;
                }
                use_staging = true;
            }
            if (use_staging)
            {
                output_richTextBox.Text = "Staging directories";
                /*
                foreach (LaunchNode node in staging_nodes)
                {
                    output_richTextBox.AppendText(node.ToString());
                }
                */
                stager_pg = new ProcessGroup(stager_mpiexec);
                Cursor.Current = Cursors.WaitCursor;
                if (stager_pg.StartGroup(staging_nodes))
                {
                    while (true)
                    {
                        if (stager_mpiexec.global_options.output_queue.Count > 0)
                        {
                            object o = stager_mpiexec.global_options.output_queue.Dequeue();
                            if (o is Thread)
                            {
                                ((Thread)o).Join();
                                // The process exited without any output
                                stager_mpiexec.Kill();
                                stager_pg = null;
                                stager_mpiexec = null;
                                MessageBox.Show("Directory staging failed", "WARNING");
                                break;
                            }
                            else
                            {
                                if (o is string)
                                {
                                    string str = (string)o;
                                    str = str.TrimEnd(new char[] { '\r', '\n', ' ' });
                                    if (str != "READY")
                                    {
                                        stager_mpiexec.Kill();
                                        stager_pg = null;
                                        stager_mpiexec = null;
                                        MessageBox.Show(String.Format("Directory staging failed: {0}", str), "WARNING");
                                    }
                                }
                                break;
                            }
                        }
                        else
                        {
                            Thread.Sleep(100);
                        }
                    }
                }
                else
                {
                    stager_pg.Kill();
                }
                Cursor.Current = Cursors.Default;
                output_richTextBox.Clear();
            }
            else
            {
                stager_mpiexec = null;
                stager_pg = null;
            }

            // Tell mpiexec to use the output queue so the output can be forwarded to the output text box
            // and start the first process group
            mpiexec.global_options.use_output_queue = true;
            ProcessGroup process_group = new ProcessGroup(mpiexec);
            if (!process_group.StartGroup(nodes))
            {
                /*
                string[] errors = new string[nodes.Count];
                process_group.GetStartErrors(errors);
                foreach (string str in errors)
                {
                    mpiexec.WriteLine(str);
                }
                */
                process_group.Kill();
            }

            // Start a thread to collect output from the mpiexec queue and forward it to the output text box
            OutputQueueWatcher watcher = new OutputQueueWatcher();
            watcher.mpiexec = mpiexec;
            watcher.nproc = ((LaunchNode)nodes[0]).nproc;
            watcher.win = this;
            watcher.queue = mpiexec.global_options.output_queue;
            ThreadStart ts = new ThreadStart(watcher.Run);
            watcher_thread = new Thread(ts);
            watcher_thread.Start();

            // Move the focus to the output box so the user can start entering input if so desired
            output_richTextBox.Focus();
        }

        private void break_button_Click(object sender, EventArgs e)
        {
            if (wait_for_stager_thread != null)
            {
                // Accessing wait_for_stager_thread and stager_mpiexec can cause null reference exceptions
                // because the wait_for_stager_thread can exit and set the values to null at any time.  So
                // use try/catch blocks to protect access to these objects

                // Try to kill the directory staging job
                try
                {
                    stager_mpiexec.Kill();
                    if (!wait_for_stager_thread.Join(10000)) // Wait 10 seconds for the staging job to be killed
                    {
                        wait_for_stager_thread.Abort();
                        wait_for_stager_thread.Join(2000);
                        wait_for_stager_thread = null;
                    }
                    stager_mpiexec = null;
                    stager_pg = null;
                    EnableExecute();
                    return;
                }
                catch
                {
                }
                // If the staging cannot be stopped then abort the waiting thread
                try
                {
                    wait_for_stager_thread.Abort();
                    wait_for_stager_thread.Join(2000);
                    wait_for_stager_thread = null;
                }
                catch
                {
                }
                stager_mpiexec = null;
                stager_pg = null;
                EnableExecute();
                return;
            }
            if (mpiexec != null)
                mpiexec.Kill();
        }

        private void wdir_browse_button_Click(object sender, EventArgs e)
        {
            FolderBrowserDialog dlg = new FolderBrowserDialog();
            if (wdir_textBox.Text.Length > 0)
            {
                dlg.SelectedPath = wdir_textBox.Text;
            }
            else
            {
                dlg.SelectedPath = Environment.CurrentDirectory;
            }
            if (dlg.ShowDialog() == DialogResult.OK)
            {
                wdir_textBox.Text = dlg.SelectedPath;
            }
        }

        private void LoadJobFromFile(string filename)
        {
            StreamReader stream;
            using (stream = new StreamReader(filename))
            {
                string line;
                int index;

                // reset all the values before loading new ones
                app_comboBox.Text = "";
                numericUpDown.Value = 1;
                wdir_textBox.Text = "";
                hosts_textBox.Text = "";
                configfile_textBox.Text = "";
                use_configfile_checkBox.Checked = false;
                env_vars_textBox.Text = "";
                map_textBox.Text = "";
                channel_comboBox.Text = "";
                log_checkBox.Checked = false;
                exitcodes_checkBox.Checked = false;
                /*credential_comboBox.Text = "";*/
                localroot_checkBox.Checked = false;
                localonly_checkBox.Checked = false;
                use_staging_checkBox.Checked = false;
                staging_src_dir_textBox.Text = "";
                staging_dest_dir_textBox.Text = "";
                staging_recursive_checkBox.Checked = false;
                staging_pull_checkBox.Checked = true;
                staging_wildcards_textBox.Text = "";

                line = stream.ReadLine();
                while (line != null)
                {
                    index = line.IndexOf(" ");
                    if (index > 0)
                    {
                        string key, val;
                        key = line.Substring(0, index);
                        val = line.Substring(index + 1);
                        switch (key)
                        {
                            case "exe":
                                app_comboBox.Text = val;
                                break;
                            case "n":
                                numericUpDown.Value = Convert.ToInt32(val);
                                break;
                            case "wdir":
                                wdir_textBox.Text = val;
                                break;
                            case "hosts":
                                hosts_textBox.Text = val;
                                break;
                            case "cfg":
                                configfile_textBox.Text = val;
                                break;
                            case "env":
                                env_vars_textBox.Text = val;
                                break;
                            case "map":
                                map_textBox.Text = val;
                                break;
                            case "channel":
                                channel_comboBox.Text = val;
                                break;
                            case "log":
                                if (val == "yes")
                                    log_checkBox.Checked = true;
                                else
                                    log_checkBox.Checked = false;
                                break;
                            case "localonly":
                                if (val == "yes")
                                    localonly_checkBox.Checked = true;
                                else
                                    localonly_checkBox.Checked = false;
                                break;
                            case "localroot":
                                if (val == "yes")
                                    localroot_checkBox.Checked = true;
                                else
                                    localroot_checkBox.Checked = false;
                                break;
                            case "exit_codes":
                                if (val == "yes")
                                    exitcodes_checkBox.Checked = true;
                                else
                                    exitcodes_checkBox.Checked = false;
                                break;
                            case "credential":
                                credential_comboBox.Text = val;
                                break;
                            case "use_cfg":
                                if (val == "yes")
                                    use_configfile_checkBox.Checked = true;
                                else
                                    use_configfile_checkBox.Checked = false;
                                break;
                            case "use_staging":
                                if (val == "yes")
                                    use_staging_checkBox.Checked = true;
                                else
                                    use_staging_checkBox.Checked = false;
                                break;
                            case "staging_subdirs":
                                if (val == "yes")
                                    staging_recursive_checkBox.Checked = true;
                                else
                                    staging_recursive_checkBox.Checked = false;
                                break;
                            case "staging_pull_files":
                                if (val == "yes")
                                    staging_pull_checkBox.Checked = true;
                                else
                                    staging_pull_checkBox.Checked = false;
                                break;
                            case "staging_src":
                                staging_src_dir_textBox.Text = val;
                                break;
                            case "staging_dest":
                                staging_dest_dir_textBox.Text = val;
                                break;
                            case "staging_wildcards":
                                staging_wildcards_textBox.Text = val;
                                break;
                        }
                    }
                    line = stream.ReadLine();
                }
                stream.Close();
            }
        }

        private void SaveJobToFile(string filename)
        {
            StreamWriter stream;
            using (stream = new StreamWriter(filename))
            {
                stream.WriteLine("exe {0}", app_comboBox.Text.Trim());
                stream.WriteLine("n {0}", numericUpDown.Value);
                stream.WriteLine("wdir {0}", wdir_textBox.Text.Trim());
                stream.WriteLine("hosts {0}", hosts_textBox.Text.Trim());
                stream.WriteLine("cfg {0}", configfile_textBox.Text.Trim());
                stream.WriteLine("env {0}", env_vars_textBox.Text.Trim());
                stream.WriteLine("map {0}", map_textBox.Text.Trim());
                stream.WriteLine("channel {0}", channel_comboBox.Text.Trim());
                stream.WriteLine(log_checkBox.Checked ? "log yes" : "log no");
                stream.WriteLine(localroot_checkBox.Checked ? "localroot yes" : "localroot no");
                stream.WriteLine(localonly_checkBox.Checked ? "localonly yes" : "localonly no");
                stream.WriteLine(exitcodes_checkBox.Checked ? "exit_codes yes" : "exit_codes no");
                stream.WriteLine("credential {0}", credential_comboBox.Text);
                stream.WriteLine(use_configfile_checkBox.Checked ? "use_cfg yes" : "use_cfg no");
                stream.WriteLine(use_staging_checkBox.Checked ? "use_staging yes" : "use_staging no");
                stream.WriteLine("staging_src {0}", staging_src_dir_textBox.Text);
                stream.WriteLine("staging_dest {0}", staging_dest_dir_textBox.Text);
                stream.WriteLine("staging_wildcards {0}", staging_wildcards_textBox.Text);
                stream.WriteLine(staging_recursive_checkBox.Checked ? "staging_subdirs yes" : "staging_subdirs no");
                stream.WriteLine(staging_pull_checkBox.Checked ? "staging_pull_files yes" : "staging_pull_files no");
                stream.Close();
            }
        }

        private void load_job_button_Click(object sender, EventArgs e)
        {
            OpenFileDialog dlg = new OpenFileDialog();
            dlg.Filter = "Job files (*.txt *.mpi)|*.txt;*.mpi|All files (*.*)|*.*";
            if (dlg.ShowDialog() == DialogResult.OK)
            {
                LoadJobFromFile(dlg.FileName);
            }
        }

        private void save_job_button_Click(object sender, EventArgs e)
        {
            SaveFileDialog dlg = new SaveFileDialog();
            dlg.Filter = "Job files (*.txt *.mpi)|*.txt;*.mpi|All files (*.*)|*.*";
            if (dlg.ShowDialog() == DialogResult.OK)
            {
                SaveJobToFile(dlg.FileName);
            }
        }

        private void show_msg_queues_button_Click(object sender, EventArgs e)
        {
            show_msg_queues_button.Enabled = false;
            foreach (ProcessGroup pg in mpiexec.process_groups)
            {
                pg.print_queues_event.Set();
            }
            show_msg_queues_button.Enabled = true;
            output_richTextBox.Focus();
        }

        private void PopulateCredentialComboBox()
        {
            try
            {
                if (store == null)
                {
                    string password = null;
                    if (CredentialStore.RequiresPassword())
                    {
                        CredentialsPassword dlg = new CredentialsPassword();
                        if (dlg.ShowDialog() == DialogResult.OK)
                        {
                            password = dlg.Password;
                        }
                        else
                        {
                            return;
                        }
                    }
                    store = new CredentialStore(password);
                }
                foreach (SecureCredential cred in store)
                {
                    bool found = false;
                    foreach (string item in credential_comboBox.Items)
                    {
                        if (item == cred.Account)
                        {
                            found = true;
                            break;
                        }
                    }
                    if (!found)
                    {
                        credential_comboBox.Items.Add(cred.Account);
                    }
                }
                ArrayList items_to_remove = new ArrayList();
                foreach (string item in credential_comboBox.Items)
                {
                    bool found = false;
                    foreach (SecureCredential cred in store)
                    {
                        if (item == cred.Account)
                        {
                            found = true;
                            break;
                        }
                    }
                    if (!found)
                        items_to_remove.Add(item);
                }
                foreach (string item in items_to_remove)
                {
                    credential_comboBox.Items.Remove(item);
                }
                if (credential_comboBox.Items.Count == 0)
                {
                    credential_comboBox.Text = "";
                }
            }
            catch { }
        }

        private void DeinoMPIwin_FormClosing(object sender, FormClosingEventArgs e)
        {
            SaveSettings();
        }

        private void credential_comboBox_MouseDown(object sender, MouseEventArgs e)
        {
            PopulateCredentialComboBox();
        }

        private void login_button_Click(object sender, EventArgs e)
        {
            try
            {
                if (!CredentialStore.LocationReady())
                {
                    MessageBox.Show("Please insert the removable media and select Login again.", "Drive not ready");
                    return;
                }
                store = new CredentialStore(store_password_textBox.Text);
                LoadDialogFromStore();
                store_password_textBox.Enabled = false;
                login_button.Enabled = false;
                credentials_listBox.Focus();
            }
            catch (Exception ex)
            {
                MessageBox.Show(String.Format("Unable to open the Credential store with the provided password:\r\n{0}", ex.Message), "Error");
                store_password_textBox.Focus();
            }
        }

        private void select_location_button_Click(object sender, EventArgs e)
        {
            SelectLocation dlg = new SelectLocation();
            if (dlg.ShowDialog() == DialogResult.OK)
            {
                if (dlg.UseRegistry)
                {
                    if (CredentialStore.RegistryStoreRequiresPassword())
                    {
                        CredentialsPassword pwd_dlg = new CredentialsPassword();
                        if (pwd_dlg.ShowDialog() == DialogResult.OK)
                        {
                            try
                            {
                                store = new CredentialStore(pwd_dlg.Password);
                            }
                            catch (Exception ex)
                            {
                                MessageBox.Show(String.Format("Unable to load the Credential store from the registry:\r\n{0}", ex.Message), "Error");
                                return;
                            }
                        }
                        else
                        {
                            return;
                        }
                    }
                    else
                    {
                        try
                        {
                            store = new CredentialStore("", true);
                        }
                        catch (Exception ex)
                        {
                            MessageBox.Show(String.Format("Unable to load the Credential store from the registry:\r\n{0}", ex.Message), "Error");
                            return;
                        }
                    }
                    location_textBox.Text = "Registry";
                }
                else
                {
                    if (CredentialStore.RequiresPassword(dlg.FileName))
                    {
                        CredentialsPassword pwd_dlg = new CredentialsPassword();
                        if (pwd_dlg.ShowDialog() == DialogResult.OK)
                        {
                            try
                            {
                                store = new CredentialStore(dlg.FileName, pwd_dlg.Password);
                            }
                            catch (Exception ex)
                            {
                                MessageBox.Show(String.Format("Unable to load the Credential store from the file:\r\n{0}", ex.Message), "Error");
                                return;
                            }
                        }
                        else
                        {
                            return;
                        }
                    }
                    else
                    {
                        try
                        {
                            store = new CredentialStore(dlg.FileName, "");
                        }
                        catch (Exception ex)
                        {
                            MessageBox.Show(String.Format("Unable to load the Credential store from the file:\r\n{0}", ex.Message), "Error");
                            return;
                        }
                    }
                    location_textBox.Text = dlg.FileName;
                }
                credential_comboBox.Text = "";
                LoadDialogFromStore();
            }
        }

        private void add_cred_button_Click(object sender, EventArgs e)
        {
            if (store != null)
            {
                if (account_textBox.Text.Length > 0)
                {
                    if (password_textBox.Text.Length == 0)
                    {
                        MessageBox.Show("DeinoMPI is unable to use accounts with blank passwords, please enter another account.", "Notice");
                        return;
                    }
                    SecureCredential cred = new SecureCredential(store.Keys);
                    cred.Account = account_textBox.Text;
                    cred.Password = password_textBox.Text;
                    try
                    {
                        store.AddCredential(cred);
                    }
                    catch (Exception ex)
                    {
                        MessageBox.Show(String.Format("Unable to save the credential.\r\n{0}", ex.Message), "Error");
                    }
                    account_textBox.Text = "";
                    password_textBox.Text = "";
                    LoadDialogFromStore();
                    if (credential_comboBox.Text.Trim() == "")
                        credential_comboBox.Text = cred.Account;
                    account_textBox.Focus();
                }
                else
                {
                    MessageBox.Show("Please enter the account name and password to add to the Credential store.", "Notice");
                    account_textBox.Focus();
                    return;
                }
            }
            else
            {
                MessageBox.Show("You must be logged into a Credential store in order to add accounts.", "Notice");
            }
        }

        private void remove_cred_button_Click(object sender, EventArgs e)
        {
            if (store != null)
            {
                if (credentials_listBox.SelectedItem != null)
                {
                    try
                    {
                        store.RemoveCredential((string)credentials_listBox.SelectedItem);
                    }
                    catch (Exception ex)
                    {
                        MessageBox.Show(String.Format("Unable to remove the credential from the store.\r\n{0}", ex.Message), "Error");
                    }
                    LoadDialogFromStore();
                    if (credentials_listBox.Items.Count == 0)
                    {
                        credential_comboBox.Text = "";
                    }
                }
            }
            else
            {
                MessageBox.Show("You must be logged into a Credential store in order to remove accounts.", "Notice");
            }
        }

        private void new_keys_button_Click(object sender, EventArgs e)
        {
            if (store == null)
            {
                MessageBox.Show("You must be logged into a Credential store before you can change the keys.", "Error");
                login_button.Focus();
                return;
            }

            try
            {
                string filename = null;
                bool use_registry = store.UseRegistry;
                if (!use_registry)
                    filename = store.FileName;
                Hashtable old_credentials = new Hashtable();
                foreach (SecureCredential cred in store)
                    old_credentials[cred.Account] = cred.Password;

                SecurityKeys keys = new SecurityKeys();
                keys.Password = store.Keys.Password;
                keys.Method = store.Keys.Method;
                keys.CreateKeys();

                if (use_registry)
                    store = new CredentialStore(keys);
                else
                    store = new CredentialStore(keys, filename);
                foreach (string key in old_credentials.Keys)
                {
                    SecureCredential cred = new SecureCredential(keys);
                    cred.Account = key;
                    cred.Password = (string)old_credentials[key];
                    store.AddCredential(cred);
                }
                pubkey_hash_textBox.Text = store.Keys.Hash;
            }
            catch (Exception ex)
            {
                MessageBox.Show(String.Format("Unable to create new keys for the Credential store:\r\n{0}", ex.Message), "Error");
                pubkey_hash_textBox.Text = "";
                store = null;
                return;
            }
        }

        private void export_keys_button_Click(object sender, EventArgs e)
        {
            if (store == null)
            {
                MessageBox.Show("You must log in to a Credential store before attempting to export the public key.", "Notice");
                return;
            }

            SaveFileDialog dlg = new SaveFileDialog();
            dlg.Filter = "SecurityKey files (*.secure_key)|*.secure_key|All files (*.*)|*.*";
            if (dlg.ShowDialog() == DialogResult.OK)
            {
                if (store.Keys.ToKeyFile(dlg.FileName, false))
                {
                    MessageBox.Show(String.Format("Public key successfully exported to file:\r\n{0}", dlg.FileName), "Success");
                }
                else
                {
                    MessageBox.Show("Unable to export the public key to a file.", "Error");
                }
            }
        }

        private void SetCreateCredentialsEnabled(bool val)
        {
            if (val)
                create_store_groupBox.Show();
            else
                create_store_groupBox.Hide();
        }

        private void enable_create_checkBox_CheckedChanged(object sender, EventArgs e)
        {
            SetCreateCredentialsEnabled(enable_create_checkBox.Checked);
            if (!removable_media_loaded && enable_create_checkBox.Checked)
            {
                bool found = false;
                removable_media_comboBox.Items.Clear();
                foreach (DriveInfo drive in System.IO.DriveInfo.GetDrives())
                {
                    if (drive.DriveType == DriveType.Removable)
                    {
                        string str;
                        if (drive.IsReady)
                            str = String.Format("{0} {1}", drive.Name, drive.VolumeLabel);
                        else
                            str = drive.Name;
                        removable_media_comboBox.Items.Add(str);
                        found = true;
                    }
                }
                if (!found)
                {
                    if (removable_radioButton.Checked)
                    {
                        removable_radioButton.Checked = false;
                        registry_radioButton.Checked = true;
                    }
                    removable_radioButton.Enabled = false;
                    removable_media_comboBox.Enabled = false;
                }
                removable_media_loaded = true;
            }
        }

        private void help_button_Click(object sender, EventArgs e)
        {
            CreateCredentailStoreWarning dlg = new CreateCredentailStoreWarning();
            dlg.Text = "Credential store creation";
            dlg.ShowDialog();
        }

        private void hd_browse_button_Click(object sender, EventArgs e)
        {
            FolderBrowserDialog dlg = new FolderBrowserDialog();
            if (hd_location_textBox.Text.Length > 0)
            {
                dlg.SelectedPath = hd_location_textBox.Text;
            }
            else
            {
                dlg.SelectedPath = Environment.CurrentDirectory;
            }
            if (dlg.ShowDialog() == DialogResult.OK)
            {
                hd_location_textBox.Text = dlg.SelectedPath;
                hard_drive_radioButton.Checked = true;
            }

        }

        private void LoadDialogFromStore()
        {
            if (store == null)
                return;
            credentials_listBox.Items.Clear();
            foreach (SecureCredential cred in store)
            {
                if (credential_comboBox.Text == "")
                    credential_comboBox.Text = cred.Account;
                credentials_listBox.Items.Add(cred.Account);
            }
            remove_cred_button.Enabled = credentials_listBox.Items.Count > 0;
            pubkey_hash_textBox.Text = store.Keys.Hash;
            if (store.UseRegistry)
                location_textBox.Text = "Registry";
            else
                location_textBox.Text = store.FileName;
        }

        private void tabControl_Selecting(object sender, TabControlCancelEventArgs e)
        {
            if (e.TabPage == null)
                return;
            
            if (e.TabPage == keys_tabPage)
            {
                if (store == null)
                {
                    if (CredentialStore.RequiresPassword())
                    {
                        string str = CredentialStore.GetLocation();
                        if (str != null)
                            location_textBox.Text = str;
                        store_password_textBox.Focus();
                    }
                    else
                    {
                        store = new CredentialStore("");
                        store_password_textBox.Text = "";
                        store_password_textBox.Enabled = false;
                        login_button.Enabled = false;
                        LoadDialogFromStore();
                    }
                }
            }

            if (!web_tab_opened && e.TabPage == web_tabPage)
            {
                webBrowser.Navigate(DeinoMPIwebPage);
                web_tab_opened = true;
            }

            if (!cluster_tab_opened && e.TabPage == cluster_tabPage)
            {
                /*
                PopulateDomainList();
                domain_populated = true;
                */
                GetHosts();
                cluster_tab_opened = true;
            }
        }

        private void enable_create_keys_checkBox_CheckedChanged(object sender, EventArgs e)
        {
            new_keys_button.Enabled = enable_create_keys_checkBox.Checked;
        }

        private void no_encryption_radioButton_CheckedChanged(object sender, EventArgs e)
        {
            if (no_encryption_radioButton.Checked)
            {
                use_password_radioButton.Enabled = false;
                new_password_textBox.Enabled = false;
                no_password_radioButton.Enabled = false;
            }
            else
            {
                use_password_radioButton.Enabled = true;
                new_password_textBox.Enabled = true;
                no_password_radioButton.Enabled = true;
            }
        }

        private void no_password_radioButton_CheckedChanged(object sender, EventArgs e)
        {
            if (no_password_radioButton.Checked)
            {
                if (symmetric_radioButton.Checked)
                {
                    protect_radioButton.Checked = true;
                }
                symmetric_radioButton.Enabled = false;
            }
            else
            {
                symmetric_radioButton.Enabled = true;
            }
        }

        private void create_credential_store_button_Click(object sender, EventArgs e)
        {
            CredentialStoreInfo dlg = new CredentialStoreInfo();

            // get the location
            string location = "";
            if (removable_radioButton.Checked)
            {
                if (removable_media_comboBox.Text == "")
                {
                    MessageBox.Show("Please enter the removable media location.", "Incomplete information");
                    removable_media_comboBox.Focus();
                    return;
                }
                location = removable_media_comboBox.Text[0].ToString() + @":\CredentialStore.xml";
            }
            if (registry_radioButton.Checked)
            {
                location = "Registry";
            }
            if (hard_drive_radioButton.Checked)
            {
                location = hd_location_textBox.Text.Trim();
                if (location == "")
                {
                    MessageBox.Show("Please enter the path where you want to save the Credential Store.", "Incomplete information");
                    hd_location_textBox.Focus();
                    return;
                }
                if (Directory.Exists(location))
                {
                    if (location.EndsWith("\\"))
                        location = location + "CredentialStore.xml";
                    else
                        location = location + "\\CredentialStore.xml";
                }
            }
            if (location == "Registry")
            {
                dlg.location_textBox.Text = "Windows Registry";
                RegistryKey key;
                key = Registry.CurrentUser.OpenSubKey(@"Software\Deino\CredentialStore");
                if (key != null)
                {
                    if (key.GetValue("SecurityKeys", null) != null)
                        dlg.location_textBox.Text = "Registry - this will override an existing store";
                    key.Close();
                }
            }
            else
                dlg.location_textBox.Text = location;

            // get the encryption method
            if (protect_radioButton.Checked)
            {
                dlg.encryption_textBox.Text = "Windows ProtectData API";
            }
            if (symmetric_radioButton.Checked)
            {
                dlg.encryption_textBox.Text = "symmetric key derived from the password";
            }
            if (no_encryption_radioButton.Checked)
            {
                dlg.encryption_textBox.Text = "No encryption";
            }

            // get the password
            if (use_password_radioButton.Checked && !no_encryption_radioButton.Checked)
            {
                if (new_password_textBox.Text == "")
                {
                    MessageBox.Show("Please enter a password to protect the private key", "Incomplete information");
                    new_password_textBox.Focus();
                    return;
                }
                dlg.password_label.Text = "Password protected";
            }
            if (no_password_radioButton.Checked)
            {
                dlg.password_label.Text = "No password";
            }

            // verify the action with the user
            if (dlg.ShowDialog() == DialogResult.OK)
            {
                // create the new store
                SecurityKeys keys = new SecurityKeys();
                try
                {
                    if (protect_radioButton.Checked)
                    {
                        keys.Method = SecurityKeys.EncryptionMethod.PROTECT_DATA;
                        if (use_password_radioButton.Checked)
                            keys.Password = new_password_textBox.Text;

                    }
                    if (symmetric_radioButton.Checked)
                    {
                        keys.Method = SecurityKeys.EncryptionMethod.SYMMETRIC_KEY;
                        keys.Password = new_password_textBox.Text;
                    }
                    if (no_encryption_radioButton.Checked)
                    {
                        keys.Method = SecurityKeys.EncryptionMethod.NONE;
                    }
                    keys.CreateKeys();
                }
                catch (Exception ex)
                {
                    MessageBox.Show(String.Format("Unable to create the SecurityKeys for the new Credential store:\r\n{0}", ex.Message), "Error");
                    return;
                }
                try
                {
                    if (location == "Registry")
                    {
                        store = new CredentialStore(keys);
                    }
                    else
                    {
                        store = new CredentialStore(keys, location);
                    }
                    LoadDialogFromStore();
                    store_password_textBox.Enabled = false;
                    login_button.Enabled = false;
                    credentials_listBox.Focus();
                    enable_create_checkBox.Checked = false;
                }
                catch (Exception ex)
                {
                    MessageBox.Show(String.Format("Unable to create the Credential store:\r\n{0}", ex.Message), "Error");
                }
            }
        }

        private void MoveCursorToEndOfOutput()
        {
            output_richTextBox.Select(output_richTextBox.Text.Length, 0);
        }

        private void output_richTextBox_KeyDown(object sender, KeyEventArgs e)
        {
            if (e.KeyCode == Keys.Back)
            {
                if (current_input.Length < 1)
                    e.SuppressKeyPress = true;
            }
            else if (e.Modifiers == Keys.None)
            {
                MoveCursorToEndOfOutput();
            }
        }

        private void output_richTextBox_KeyPress(object sender, KeyPressEventArgs e)
        {
            switch (e.KeyChar)
            {
                case '\b':
                    if (current_input.Length > 0)
                        current_input.Remove(current_input.Length - 1, 1);
                    //else e.Handled = true; // marking it handled does not prevent the backspace from erasing characters.
                    break;
                case '\r': // The '\r' but not '\n' characters are created when the user hits enter
                    current_input.Append(e.KeyChar);
                    current_input.Append('\n');
                    mpiexec.input.EnqueueInput(current_input.ToString());
                    current_input = new StringBuilder();
                    break;
                case '\x16':
                    // Ctrl-V pressed
                    // Grab the text from the clipboard
                    if (Clipboard.ContainsText())
                    {
                        MoveCursorToEndOfOutput();
                        string text = Clipboard.GetText();
                        //MessageBox.Show(String.Format("pasting '{0}'", text));
                        current_input.Append(text);
                        if (text.Contains("\n") || text.Contains("\r"))
                        {
                            if (mpiexec != null)
                                mpiexec.input.EnqueueInput(current_input.ToString());
                            current_input = new StringBuilder();
                        }
                    }
                    break;
                default:
                    //MessageBox.Show(String.Format("Key {0} pressed", Convert.ToInt32(e.KeyChar)));
                    if (Char.IsControl(e.KeyChar))
                    {
                        //MessageBox.Show(String.Format("Key {0} pressed", Convert.ToInt32(e.KeyChar)));
                        //MessageBox.Show(String.Format("Skipping key {0}", Char.GetNumericValue(e.KeyChar)));
                    }
                    else
                    {
                        current_input.Append(e.KeyChar);
                    }
                    break;
            }
        }

        private void DeinoMPIwin_FormClosed(object sender, FormClosedEventArgs e)
        {
            // This is a hack to kill the extra threads and make sure the application exits.
            // FIXME: The right answer is to go through the code and find out which thread is not exiting and create an exit path for it.
            System.Environment.Exit(0);
        }

        private void high_security_label_Click(object sender, EventArgs e)
        {
            use_password_radioButton.Checked = true;
            protect_radioButton.Checked = true;
            removable_radioButton.Checked = true;
        }

        private void convenient_security_label_Click(object sender, EventArgs e)
        {
            no_password_radioButton.Checked = true;
            protect_radioButton.Checked = true;
            registry_radioButton.Checked = true;
        }

        private void webBrowser_Navigated(object sender, WebBrowserNavigatedEventArgs e)
        {
            url_textBox.Text = webBrowser.Url.ToString();
        }

        private void url_textBox_KeyUp(object sender, KeyEventArgs e)
        {
            if (e.KeyCode == Keys.Enter)
            {
                webBrowser.Navigate(url_textBox.Text);
            }
        }

        private void back_button_Click(object sender, EventArgs e)
        {
            webBrowser.GoBack();
        }

        private void forward_button_Click(object sender, EventArgs e)
        {
            webBrowser.GoForward();
        }

        internal void PopulateDomainList()
        {
            ComputerBrowser ce;
            string domainName = "";
            string currentDomain = System.Environment.UserDomainName;
            int index = -1;

            Cursor.Current = Cursors.WaitCursor;

            // browse for domains
            ce = new ComputerBrowser(ComputerBrowser.ServerType.SV_TYPE_DOMAIN_ENUM);
            for (int i = 0; i < ce.Length; i++)
            {
                domainName = ce[i].Name;
                // Add the name to the dropdown list if it already doesn't exist
                bool found = false;
                foreach (string str in domain_comboBox.Items)
                {
                    if (str.ToLower() == domainName.ToLower())
                        found = true;
                }
                if (!found)
                {
                    domain_comboBox.Items.Add(domainName);
                }
                // Save the index of the current machine's domain so it will be selected by default
                if (domainName.ToLower() == currentDomain.ToLower())
                {
                    index = i;
                }
            }
            if (index == -1)
            {
                index = domain_comboBox.Items.Add(currentDomain);
            }
            domain_comboBox.SelectedIndex = index;
            Cursor.Current = Cursors.Default;
        }

        private void domain_comboBox_DropDown(object sender, EventArgs e)
        {
            if (!domain_populated)
            {
                PopulateDomainList();
                domain_populated = true;
            }
        }

        internal string[] GetHostNames()
        {
            ComputerBrowser ce;
            string[] results = null;

            Cursor.Current = Cursors.WaitCursor;

            ce = new ComputerBrowser(domain_comboBox.Text);
            int numServer = ce.Length;

            if (ce.LastError.Length == 0)
            {
                IEnumerator enumerator = ce.GetEnumerator();

                // add the domain text to the dropdown list if it isn't already there
                bool found = false;
                foreach (string str in domain_comboBox.Items)
                {
                    if (str == domain_comboBox.Text)
                    {
                        found = true;
                    }
                }
                if (!found)
                {
                    domain_comboBox.Items.Add(domain_comboBox.Text);
                }

                results = new string[numServer];
                int i = 0;
                while (enumerator.MoveNext())
                {
                    results[i] = ce[i].Name;
                    i++;
                }
            }
            else
            {
                //output_textBox.Text = "Error \"" + ce.LastError + "\"";
            }

            Cursor.Current = Cursors.Default;
            return results;
        }

        private void GetHosts()
        {
            string[] hosts;
            if (domain_comboBox.Text.ToLower() == Environment.MachineName.ToLower() || domain_comboBox.Text == "")
            {
                hosts = new string[1] { Environment.MachineName };
            }
            else
            {
                hosts = GetHostNames();
            }
            if (hosts != null)
            {
                //hosts_listView.Items.Clear();
                foreach (string host in hosts)
                {
                    AddHost(host);
                }
            }
        }

        private void get_hosts_button_Click(object sender, EventArgs e)
        {
            GetHosts();
        }

        private void view_comboBox_SelectionChangeCommitted(object sender, EventArgs e)
        {
            switch (view_comboBox.Text)
            {
                case "Large icons":
                    hosts_listView.View = View.LargeIcon;
                    break;
                case "Small icons":
                    hosts_listView.View = View.SmallIcon;
                    break;
                case "List":
                    hosts_listView.View = View.List;
                    break;
                case "Tiles":
                    hosts_listView.View = View.Tile;
                    break;
                case "Details":
                    hosts_listView.View = View.Details;
                    break;
            }
        }

        private bool CheckCredentialBox()
        {
            if ((credential_comboBox.Text.Trim() == "") || (credential_comboBox.Text == null))
            {
                MessageBox.Show("Scanning hosts requires that you select an account from the Credential store that is valid on all the hosts.\r\nIf the Credential store is empty you must add an account before you can scan hosts.");
                return false;
            }
            return true;
        }

        private void scan_hosts_button_Click(object sender, EventArgs e)
        {
            if (hosts_listView.Items.Count < 1)
                return;
            if (scan_host_backgroundWorker.IsBusy)
                return;
            if (!CheckCredentialBox())
                return;
            scan_hosts_button.Enabled = false;
            scanToolStripMenuItem.Enabled = false;
            int i;
            int count = 0;
            for (i = 0; i < hosts_listView.Items.Count; i++)
            {
                //if (hosts_listView.Items[i].Tag == null && (hosts_listView.Items[i].ImageIndex != 4))
                if (hosts_listView.Items[i].ImageIndex == 0)
                    count++;
            }
            string[] hosts = new string[count];
            for (i = 0; i < hosts_listView.Items.Count; i++)
            {
                //if (hosts_listView.Items[i].Tag == null && (hosts_listView.Items[i].ImageIndex != 4))
                if (hosts_listView.Items[i].ImageIndex == 0)
                {
                    hosts[--count] = hosts_listView.Items[i].Text;
                }
            }
            scan_host_backgroundWorker.RunWorkerAsync(hosts);
        }

        private delegate string CredentialStoreAccountDelegate();
        private CredentialStoreAccountDelegate get_credential_storeAccount_delegate;
        private string GetCredentialStoreAccount()
        {
            return credential_comboBox.Text;
        }

        private HostInfo GetHostInfo(string host, bool bPrintErrors)
        {
            ClientSession client;
            if (store == null)
            {
                if (CredentialStore.RequiresPassword())
                {
                    if (bPrintErrors)
                        hostinfo_richTextBox.Text = CredentialRequiredMessage;
                    return null;
                }
                else
                {
                    store = new CredentialStore("");
                    if (bPrintErrors)
                    {
                        store_password_textBox.Text = "";
                        store_password_textBox.Enabled = false;
                        login_button.Enabled = false;
                        LoadDialogFromStore();
                    }
                }
            }
            string account = (string)Invoke(get_credential_storeAccount_delegate);
            SecureCredential credential = store.GetCredential(account/*credential_comboBox.Text*/);
            if (credential == null)
            {
                if (bPrintErrors)
                    hostinfo_richTextBox.Text = CredentialRequiredMessage;
                return null;
            }
            try
            {
                if (mpiexec != null)
                    client = new ClientSession(host, -1, mpiexec.global_options.phrase, store.Keys, credential);
                else
                    client = new ClientSession(host, -1, null, store.Keys, credential);
            }
            catch (Exception e)
            {
                if (bPrintErrors)
                {
                    StringBuilder sb = new StringBuilder();
                    sb.Append(e.Message);
                    foreach (string key in e.Data.Keys)
                    {
                        sb.AppendFormat("\r\n{0}={1}", key, e.Data[key]);
                    }
                    hostinfo_richTextBox.Text = sb.ToString();
                }
                return null;
            }
            InfoRequestMessage message = new InfoRequestMessage("host_info");
            client.SendMessage(message);
            HostInfoMessage hmessage = (HostInfoMessage)client.NextMessage();
            QuitMessage quit_message = new QuitMessage();
            client.SendMessage(quit_message);
            // get a response to the quit message
            DeinoMPI.Message message2;
            message2 = client.NextMessage();
            try
            {
                // Send the last leg of the quit handshake
                quit_message = new QuitMessage();
                client.SendMessage(quit_message);
            }
            catch
            {
                // Ignore errors here since we are guaranteed to be in the last leg of the quit handshake.
            }
            return hmessage.Info;
        }

        private string NiceNumberString(long n)
        {
            double d;
            if (n >= 1024 * 1024 * 1024)
            {
                d = (double)n / (double)(1024 * 1024 * 1024);
                if (d > 10)
                    return String.Format("{0} Gigabytes", Math.Floor(d));
                return String.Format("{0:#.##} Gigabytes", d);
            }
            else if (n >= 1024 * 1024)
            {
                d = (double)n / (double)(1024 * 1024);
                if (d > 10)
                    return String.Format("{0} Megabytes", Math.Floor(d));
                return String.Format("{0:#.##} Megabytes", d);
            }
            else if (n >= 1024)
            {
                d = (double)n / (double)1024.0;
                if (d > 10)
                    return String.Format("{0} Kilobytes", Math.Floor(d));
                return String.Format("{0:#.##} Kilobytes", d);
            }
            return String.Format("{0} bytes", n);
        }

        private string NiceNumberStringMetric(long n)
        {
            double d;
            if (n > 100000000)
            {
                d = (double)n / (double)(1000000000);
                if (d > 10)
                    return String.Format("{0} Gigabit", Math.Floor(d));
                return String.Format("{0:#.##} Gigabit", d);
            }
            else if (n >= 1000000)
            {
                d = (double)n / (double)(1000000);
                if (d > 10)
                    return String.Format("{0} Megabit", Math.Floor(d));
                return String.Format("{0:#.##} Megabit", d);
            }
            else if (n >= 1000)
            {
                d = (double)n / (double)1000.0;
                if (d > 10)
                    return String.Format("{0} Kilobit", Math.Floor(d));
                return String.Format("{0:#.##} Kilobit", d);
            }
            return String.Format("{0} baud", n);
        }

        void UpdateHostInfoString(HostInfo info)
        {
            if (info == null)
                return;
            if (!info.Valid)
            {
                hostinfo_richTextBox.Text = "DeinoMPI not installed";
                return;
            }
            StringBuilder sb = new StringBuilder();
            //sb.AppendLine(info.Host);
            sb.AppendLine(info.OS);
            sb.AppendFormat("{0} - {1}\n", (info.Arch == HostInfo.ARCH.WIN32) ? "Win32" : "Win64", info.ArchString);
            sb.AppendFormat("CPU: {0}\n", info.CPU);
            sb.AppendFormat("CPU name: {0}\n", info.CPUName);
            sb.AppendFormat("Number of CPUs: {0}\n", info.NumCPUs);
            sb.AppendFormat("Physical memory: {0}\n", NiceNumberString(info.Memory));
            sb.AppendFormat("Free disk space: {0}\n", NiceNumberString(info.FreeDisk));
            foreach (string ip in info.IPAddresses.Split(' '))
            {
                sb.AppendFormat("IP: {0}\n", ip);
            }
            /*
            string nic_speed;
            if (info.NicSpeed >= 1000000000)
                nic_speed = "Gigabit";
            else if (info.NicSpeed >= 100000000)
                nic_speed = "100Mbit";
            else if (info.NicSpeed >= 10000000)
                nic_speed = "10Mbit";
            else
                nic_speed = String.Format("{0}Kbit", info.NicSpeed / 1000);
            sb.AppendFormat("Network interface - {0}\n", nic_speed);
            */
            foreach (string nic_speed in info.NicSpeeds.Split(' '))
            {
                try
                {
                    sb.AppendFormat("NIC speed: {0}\n", NiceNumberStringMetric(Convert.ToInt64(nic_speed)));
                }
                catch
                {
                }
            }
            sb.AppendLine();
            sb.AppendFormat("DeinoMPI: {0}\n", info.Version);
            hostinfo_richTextBox.Text = sb.ToString();
        }

        private void hosts_listView_Click(object sender, EventArgs e)
        {
            if (hosts_listView.SelectedItems.Count == 1)
            {
                ListViewItem item = hosts_listView.SelectedItems[0];
                hostname_label.Text = item.Text;
                if (item.Tag == null)
                {
                    if (item.ImageIndex == 1)
                    {
                        // machine does not have DeinoMPI installed
                        HostInfo info = new HostInfo();
                        info.Host = item.Text;
                        info.Valid = false;
                        item.Tag = info;
                    }
                    else if (item.ImageIndex == 2)
                    {
                        // machine does have DeinoMPI installed
                        HostInfo info = GetHostInfo(item.Text, true);
                        SetSubItems(item, info);
                        item.Tag = info;
                    }
                    else if (item.ImageIndex == 4)
                    {
                        // machine has an old version of DeinoMPI installed
                        hostinfo_richTextBox.Text = String.Format("The version of DeinoMPI installed does not match the version installed locally.  In order to launch jobs the versions must match.\n\nDeinoMPI: {0}",
                            item.SubItems[1].Text);
                    }
                    else
                    {
                        // it has not been determined if the machine has DeinoMPI installed
                        hostinfo_richTextBox.Text = "";
                        if (!scan_host_backgroundWorker.IsBusy)
                        {
                            Cursor.Current = Cursors.WaitCursor;
                            hostinfo_richTextBox.Text = "";
                            HostInfo info = GetHostInfo(item.Text, true);
                            if (info == null)
                            {
                                if (hostinfo_richTextBox.Text != CredentialRequiredMessage)
                                {
                                    UpdateInfo(item.Text, null);
                                }
                            }
                            else
                            {
                                UpdateInfo(item.Text, info);
                            }
                            //UpdateInfo(item.Text, GetHostInfo(item.Text, true));
                            Cursor.Current = Cursors.Default;
                        }
                    }
                }
                if (item.Tag != null)
                {
                    UpdateHostInfoString((HostInfo)item.Tag);
                }
            }
        }

        private void hosts_listView_ItemSelectionChanged(object sender, ListViewItemSelectionChangedEventArgs e)
        {
            if (hosts_listView.SelectedItems.Count > 1)
            {
                hostname_label.Text = "Hosts";
                StringBuilder sb = new StringBuilder();
                foreach (ListViewItem item in hosts_listView.SelectedItems)
                {
                    sb.AppendLine(item.Text);
                }
                hostinfo_richTextBox.Text = sb.ToString();
            }
        }

        private delegate void UpdateInfoDelegate(string host, HostInfo info);
        private UpdateInfoDelegate info_delegate;
        private void UpdateInfo(string host, HostInfo info)
        {
            foreach (ListViewItem item in hosts_listView.Items)
            {
                if (item.Text == host)
                {
                    item.Tag = new HostInfo(info);
                    if (info == null || info.Version == null || info.Version == "")
                    {
                        ClearSubItems(item);
                        item.SubItems[1].Text = "";
                        item.ImageIndex = 1;
                    }
                    else
                    {
                        SetSubItems(item, info);
                        item.ImageIndex = 2;
                    }
                }
            }
        }

        private delegate void UpdateVersionDelegate(string host, string version);
        private UpdateVersionDelegate version_delegate;
        private void UpdateVersion(string host, string version)
        {
            foreach (ListViewItem item in hosts_listView.Items)
            {
                if (item.Text == host)
                {
                    if (version == null || version == "")
                    {
                        ClearSubItems(item);
                        item.SubItems[1].Text = "";
                        item.ImageIndex = 1;
                    }
                    else if (version == "?")
                    {
                        ClearSubItems(item);
                        item.SubItems[1].Text = version;
                        item.ImageIndex = 3;
                    }
                    else
                    {
                        ClearSubItems(item);
                        item.SubItems[1].Text = version;
                        item.ImageIndex = 4;
                    }
                }
            }
        }

        Queue version_hosts_queue = Queue.Synchronized(new Queue());
        private void version_thread(object o)
        {
            try
            {
                string host = (string)version_hosts_queue.Dequeue();
                while (host != null)
                {
                    Invoke(version_delegate, new object[] { host, "?" });
                    bool success = false;
                    try
                    {
                        HostInfo info = GetHostInfo(host, false);
                        if (info != null)
                        {
                            Invoke(info_delegate, new object[] { host, info });
                            success = true;
                        }
                    }
                    catch
                    {
                    }
                    if (!success)
                    {
                        try
                        {
                            ClientSession client;
                            string version = null;
                            client = new ClientSession(host, -1, ref version);
                            Invoke(version_delegate, new object[] { host, version });
                        }
                        catch
                        {
                            Invoke(version_delegate, new object[] { host, null });
                        }
                    }
                    host = (string)version_hosts_queue.Dequeue();
                }
            }
            catch
            {
            }
        }

        private void scan_host_backgroundWorker_DoWork(object sender, DoWorkEventArgs e)
        {
            try
            {
                int i, num_threads;
                Thread[] threads;
                string[] hosts = (string[])e.Argument;
                if (hosts == null)
                    return;
                version_hosts_queue.Clear();
                foreach (string host in hosts)
                {
                    version_hosts_queue.Enqueue(host);
                }
                num_threads = (hosts.Length > 32) ? 32 : hosts.Length;
                threads = new Thread[num_threads];
                for (i=0; i<num_threads; i++)
                {
                    ParameterizedThreadStart ts = new ParameterizedThreadStart(version_thread);
                    threads[i] = new Thread(ts);
                    threads[i].Start();
                }
                for (i = 0; i < num_threads; i++)
                    threads[i].Join();
            }
            catch
            {
            }
        }

        private void hosts_listView_ColumnClick(object sender, ColumnClickEventArgs e)
        {
            if (e.Column == lvwColumnSorter.SortColumn)
            {
                // Reverse the current sort direction for this column.
                if (lvwColumnSorter.Order == SortOrder.Ascending)
                {
                    lvwColumnSorter.Order = SortOrder.Descending;
                }
                else
                {
                    lvwColumnSorter.Order = SortOrder.Ascending;
                }
            }
            else
            {
                // Set the column number that is to be sorted; default to ascending.
                lvwColumnSorter.SortColumn = e.Column;
                lvwColumnSorter.Order = SortOrder.Ascending;
            }

            // Perform the sort with these new sort options.
            hosts_listView.Sort();
        }

        private void AddHost(string host)
        {
            if (host == null || host == "")
                return;
            bool found = false;
            foreach (ListViewItem item in hosts_listView.Items)
            {
                if (item.Text.ToLower() == host.ToLower())
                {
                    found = true;
                    break;
                }
            }
            if (!found)
            {
                if (!host.Contains(".")) // grass -> GRASS but grass.comcast.net stays the same
                    host = host.ToUpper();
                ListViewItem item = new ListViewItem(host, 0);
                item.SubItems.AddRange(new string[] { "", "", "", "", "", "", "", "", "", "", "", "" });
                hosts_listView.Items.Add(item);
            }
        }

        private void one_host_button_Click(object sender, EventArgs e)
        {
            AddHost(one_host_textBox.Text);
            one_host_textBox.Text = "";
            one_host_textBox.Focus();
        }

        private void reset_hosts_button_Click(object sender, EventArgs e)
        {
            foreach (ListViewItem item in hosts_listView.Items)
            {
                item.ImageIndex = 0;
                item.Tag = null;
            }
        }

        private void clear_hosts_button_Click(object sender, EventArgs e)
        {
            hosts_listView.Items.Clear();
        }

        private void sendTompiexecTabToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (hosts_listView.SelectedItems.Count > 0)
            {
                string[] hosts = new string[hosts_listView.SelectedItems.Count];
                for (int i = 0; i < hosts_listView.SelectedItems.Count; i++)
                    hosts[i] = hosts_listView.SelectedItems[i].Text;
                hosts_textBox.Text = String.Join(" ", hosts);
                if (!extra_options_checkBox.Checked)
                {
                    extra_options_checkBox.Checked = true;
                }
                tabControl.SelectedTab = mpiexec_tabPage;
                hosts_textBox.Focus();
                execute_button.Focus();
            }
        }

        private void scanToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (hosts_listView.SelectedItems.Count < 1)
                return;
            if (scan_host_backgroundWorker.IsBusy)
                return;
            if (!CheckCredentialBox())
                return;
            scan_hosts_button.Enabled = false;
            scanToolStripMenuItem.Enabled = false;
            int i;
            int count = 0;
            count = hosts_listView.SelectedItems.Count;
            string[] hosts = new string[count];
            for (i = 0; i < hosts_listView.SelectedItems.Count; i++)
            {
                hosts[--count] = hosts_listView.SelectedItems[i].Text;
            }
            scan_host_backgroundWorker.RunWorkerAsync(hosts);
        }

        private void removeToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (hosts_listView.SelectedItems.Count < 1)
                return;
            ArrayList items = new ArrayList();
            foreach (ListViewItem item in hosts_listView.SelectedItems)
            {
                items.Add(item);
            }
            foreach (ListViewItem item in items)
            {
                hosts_listView.Items.Remove(item);
            }
        }

        private void resetToolStripMenuItem_Click(object sender, EventArgs e)
        {
            foreach (ListViewItem item in hosts_listView.SelectedItems)
            {
                item.ImageIndex = 0;
                item.Tag = null;
            }
        }

        private void scan_host_backgroundWorker_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e)
        {
            scan_hosts_button.Enabled = true;
            scanToolStripMenuItem.Enabled = true;
            if (hosts_listView.SelectedItems.Count == 1)
            {
                if (hosts_listView.SelectedItems[0].Tag != null)
                {
                    UpdateHostInfoString((HostInfo)hosts_listView.SelectedItems[0].Tag);
                }
            }
        }

        private void save_list_button_Click(object sender, EventArgs e)
        {
            SaveFileDialog dlg = new SaveFileDialog();
            dlg.Filter = "XML files (*.xml)|*.xml|All files (*.*)|*.*";
            if (dlg.ShowDialog() == DialogResult.OK)
            {
                using (XmlTextWriter xwriter = new XmlTextWriter(dlg.FileName, Encoding.UTF8))
                {
                    xwriter.Formatting = Formatting.Indented;
                    xwriter.WriteStartDocument();
                    xwriter.WriteStartElement("DeinoMPI_HostList");
                    foreach (ListViewItem item in hosts_listView.Items)
                    {
                        switch (item.ImageIndex)
                        {
                            case 0:
                                // unknown
                                xwriter.WriteStartElement("Unknown");
                                xwriter.WriteElementString("Host", item.Text);
                                xwriter.WriteEndElement();
                                break;
                            case 1:
                                // no DeinoMPI
                                xwriter.WriteStartElement("No");
                                xwriter.WriteElementString("Host", item.Text);
                                xwriter.WriteEndElement();
                                break;
                            case 2:
                                // yes DeinoMPI
                                xwriter.WriteStartElement("DeinoMPI");
                                ((HostInfo)item.Tag).Write(xwriter);
                                xwriter.WriteEndElement();
                                break;
                            case 3:
                                // querying
                                xwriter.WriteStartElement("Unknown");
                                xwriter.WriteElementString("Host", item.Text);
                                xwriter.WriteEndElement();
                                break;
                            case 4:
                                // old version
                                xwriter.WriteStartElement("oldDeinoMPI");
                                xwriter.WriteElementString("Host", item.Text);
                                xwriter.WriteElementString("Version", item.SubItems[1].Text);
                                xwriter.WriteEndElement();
                                break;
                        }
                    }
                    xwriter.WriteEndElement();
                    xwriter.WriteEndDocument();
                    xwriter.Close();
                }
            }
        }

        private void SetSubItems(ListViewItem item, HostInfo info)
        {
            item.SubItems[1].Text = info.Version;
            item.SubItems[2].Text = (info.Arch == HostInfo.ARCH.WIN32) ? "Win32" : "Win64";
            item.SubItems[3].Text = info.ArchString;
            item.SubItems[4].Text = info.NumCPUs.ToString();
            item.SubItems[5].Text = info.Memory.ToString();
            long max_speed = 0;
            string[] tokens = info.NicSpeeds.Split(' ');
            foreach (string str in tokens)
            {
                long speed = Convert.ToInt64(str);
                max_speed = (speed > max_speed) ? speed : max_speed;
            }
            item.SubItems[6].Text = (max_speed >= 1000000000) ? "yes" : "no";
            tokens = info.IPAddresses.Split(' ');
            if (tokens != null && tokens.Length > 0)
            {
                item.SubItems[7].Text = tokens[0];
                if (tokens.Length > 1)
                    item.SubItems[8].Text = tokens[1];
                else
                    item.SubItems[8].Text = tokens[0];
            }
            else
            {
                item.SubItems[7].Text = "";
                item.SubItems[8].Text = "";
            }
            item.SubItems[9].Text = info.FreeDisk.ToString();
            item.SubItems[10].Text = info.CPU;
            item.SubItems[11].Text = info.CPUName;
            item.SubItems[12].Text = info.OS;
        }

        private void ClearSubItems(ListViewItem item)
        {
            item.SubItems[1].Text = "";
            item.SubItems[2].Text = "";
            item.SubItems[3].Text = "";
            item.SubItems[4].Text = "";
            item.SubItems[5].Text = "";
            item.SubItems[6].Text = "";
            item.SubItems[7].Text = "";
            item.SubItems[8].Text = "";
            item.SubItems[9].Text = "";
            item.SubItems[10].Text = "";
            item.SubItems[11].Text = "";
            item.SubItems[12].Text = "";
        }

        private void load_list_button_Click(object sender, EventArgs e)
        {
            OpenFileDialog dlg = new OpenFileDialog();
            dlg.Filter = "XML files (*.xml)|*.xml|All files (*.*)|*.*";
            if (dlg.ShowDialog() == DialogResult.OK)
            {
                try
                {
                    using (StreamReader sr = new StreamReader(dlg.FileName))
                    {
                        ArrayList new_items = new ArrayList();
                        XmlTextReader xreader = new XmlTextReader(sr);
                        xreader.ReadStartElement("DeinoMPI_HostList");
                        while (xreader.IsStartElement())
                        {
                            string host;
                            ListViewItem item;
                            switch (xreader.Name)
                            {
                                case "Unknown":
                                    xreader.ReadStartElement("Unknown");
                                    host = xreader.ReadElementString("Host");
                                    item = new ListViewItem(host);
                                    item.SubItems.AddRange(new string[] { "", "", "", "", "", "", "", "", "", "", "", "" });
                                    item.ImageIndex = 0;
                                    new_items.Add(item);
                                    break;
                                case "No":
                                    xreader.ReadStartElement("No");
                                    host = xreader.ReadElementString("Host");
                                    item = new ListViewItem(host);
                                    item.SubItems.AddRange(new string[] { "", "", "", "", "", "", "", "", "", "", "", "" });
                                    item.ImageIndex = 1;
                                    new_items.Add(item);
                                    break;
                                case "DeinoMPI":
                                    xreader.ReadStartElement("DeinoMPI");
                                    HostInfo info = new HostInfo();
                                    info.Read(xreader);
                                    if (info.Host.Contains("."))
                                        item = new ListViewItem(info.Host);
                                    else
                                        item = new ListViewItem(info.Host.ToUpper());
                                    item.SubItems.AddRange(new string[] { "", "", "", "", "", "", "", "", "", "", "", "" });
                                    SetSubItems(item, info);
                                    item.Tag = info;
                                    item.ImageIndex = 2;
                                    new_items.Add(item);
                                    break;
                                case "oldDeinoMPI":
                                    xreader.ReadStartElement("oldDeinoMPI");
                                    host = xreader.ReadElementString("Host");
                                    item = new ListViewItem(host);
                                    string version = xreader.ReadElementString("Version");
                                    item.SubItems.Add(version);
                                    item.SubItems.AddRange(new string[] { "", "", "", "", "", "", "", "", "", "", "" });
                                    item.ImageIndex = 4;
                                    new_items.Add(item);
                                    break;
                            }
                            xreader.ReadEndElement();
                        }
                        xreader.ReadEndElement();
                        xreader.Close();
                        hosts_listView.Items.Clear();
                        foreach (ListViewItem item in new_items)
                        {
                            hosts_listView.Items.Add(item);
                        }
                    }
                }
                catch (Exception ex)
                {
                    MessageBox.Show(String.Format("Unable to parse the selected file: {0}", ex.Message));
                }
            }
        }

        private void cluster_hosts_contextMenuStrip_Opening(object sender, CancelEventArgs e)
        {
            if (hosts_listView.SelectedItems.Count < 1)
            {
                scanToolStripMenuItem.Enabled = false;
                resetToolStripMenuItem.Enabled = false;
                removeToolStripMenuItem.Enabled = false;
                sendTompiexecTabToolStripMenuItem.Enabled = false;
            }
            else
            {
                if (scan_host_backgroundWorker.IsBusy)
                {
                    scanToolStripMenuItem.Enabled = false;
                    resetToolStripMenuItem.Enabled = false;
                    removeToolStripMenuItem.Enabled = false;
                }
                else
                {
                    scanToolStripMenuItem.Enabled = true;
                    resetToolStripMenuItem.Enabled = true;
                    removeToolStripMenuItem.Enabled = true;
                }
                sendTompiexecTabToolStripMenuItem.Enabled = true;
            }
        }

        private string get_clog2()
        {
            int index, last;
            string app = app_comboBox.Text;
            try
            {
                if (app[0] == '\"')
                {
                    // remove the quotes to be restored at the end
                    app = app.Replace("\"", "");
                }
                index = app.LastIndexOf('\\');
                if (index == -1)
                {
                    index = 0;
                }
                last = index;
                // find the index of .exe starting from the last \
                index = app.IndexOf(".exe", last);
                if (index == -1)
                {
                    // no .exe so find the first space after the last \
                    // This assumes the executable does not have spaces in its name
                    index = app.IndexOf(' ', last);
                    if (index == -1)
                    {
                        // no .exe and no spaces found so take the entire string
                        index = app.Length;
                    }
                }
                app = app.Substring(0, index) + ".exe.clog2";
                app = app.Trim();
                if (!File.Exists(app) && wdir_textBox.Text.Length > 0)
                {
                    // file not found from the application combo box so try finding it in the working directory
                    string app2 = app_comboBox.Text;
                    if (app2[0] == '\"')
                    {
                        // remove the quotes to be restored at the end
                        app2 = app2.Replace("\"", "");
                    }
                    index = app2.LastIndexOf('\\');
                    if (index != -1)
                    {
                        // remove the path
                        app2 = app2.Remove(0, index);
                    }
                    // find the index of .exe
                    index = app2.IndexOf(".exe");
                    if (index != -1)
                    {
                        // remove any arguments after the executable name
                        app2 = app2.Substring(0, index + 4);
                    }
                    else
                    {
                        // no .exe so find the first space
                        // This assumes the executable does not have spaces in its name
                        index = app.IndexOf(' ');
                        if (index != -1)
                        {
                            app2 = app2.Substring(0, index) + ".exe";
                        }
                        else
                        {
                            // no .exe and no arguments so append .exe
                            app2 += ".exe";
                        }
                    }
                    // app2 should now contain only the executable name
                    // Add the wdir path and clog2 extension
                    if (wdir_textBox.Text.EndsWith("\\"))
                        app2 = wdir_textBox.Text + app2 + ".clog2";
                    else
                        app2 = wdir_textBox.Text + "\\" + app2 + ".clog2";
                    if (File.Exists(app2))
                    {
                        app = app2;
                        FileStream f = File.OpenRead(app2);
                        app = f.Name;
                        f.Close();
                    }
                }
                if (app.IndexOf(' ') != -1)
                {
                    app = "\"" + app + "\"";
                }
            }
            catch (Exception)
            {
                app = "";
            }
            return app;
        }

        private void jumpshot_button_Click(object sender, EventArgs e)
        {
            string jumpshot_filename = Path.GetDirectoryName(System.Reflection.Assembly.GetEntryAssembly().Location) + @"\jumpshot.jar";
            if (!File.Exists(jumpshot_filename))
            {
                jumpshot_filename = System.Environment.CurrentDirectory + @"\jumpshot.jar";
                if (!File.Exists(jumpshot_filename))
                    jumpshot_filename = "jumpshot.jar";
            }
            string args = "-Xms32m -Xmx256m -jar \"" + jumpshot_filename + "\" " + get_clog2();
            //MessageBox.Show("javaw.exe " + args);
            //output_richTextBox.AppendText("javaw.exe " + args + "\r\n");
            try
            {
                Process p;
                p = Process.Start("javaw.exe", args);
            }
            catch (Exception x)
            {
                MessageBox.Show("Unable to start: javaw.exe " + args + "\r\n" + x.Message, "Error");
            }
        }

        private void log_checkBox_CheckedChanged(object sender, EventArgs e)
        {
            jumpshot_button.Enabled = log_checkBox.Checked;
        }

        private void localroot_checkBox_CheckedChanged(object sender, EventArgs e)
        {
            if (localroot_checkBox.Checked)
                localonly_checkBox.Checked = false;
        }

        private void localonly_checkBox_CheckedChanged(object sender, EventArgs e)
        {
            hosts_label.Enabled = !localonly_checkBox.Checked;
            hosts_textBox.Enabled = !localonly_checkBox.Checked;
            if (localonly_checkBox.Checked)
                localroot_checkBox.Checked = false;
        }

        private void use_configfile_checkBox_CheckedChanged(object sender, EventArgs e)
        {
            if (use_configfile_checkBox.Checked)
            {
                DisableExtraOptions();
            }
            else
            {
                EnableExtraOptions();
            }
        }

        private void configfile_browse_button_Click(object sender, EventArgs e)
        {
            OpenFileDialog dlg = new OpenFileDialog();
            dlg.Filter = "Configuration files (*.txt *.mpi *.cfg)|*.txt;*.mpi;*.cfg|All files (*.*)|*.*";
            if (configfile_textBox.Text.Length > 0)
            {
                if (File.Exists(configfile_textBox.Text))
                    dlg.FileName = configfile_textBox.Text;
            }
            if (dlg.ShowDialog() == DialogResult.OK)
            {
                configfile_textBox.Text = dlg.FileName;
            }
        }

        private void configfile_help_button_Click(object sender, EventArgs e)
        {
            ConfigfileHelp dlg = new ConfigfileHelp();
            dlg.ShowDialog();
        }

        private void use_history_checkBox_CheckedChanged(object sender, EventArgs e)
        {
            Regex rex = new Regex(@"DeinoMPI_USE_CALL_HISTORY\s*=\s*""?1""?", RegexOptions.IgnoreCase);
            if (use_history_checkBox.Checked)
            {
                // Add DeinoMPI_USE_CALL_HISTORY=1
                Match match = rex.Match(env_vars_textBox.Text);
                if (!match.Success)
                {
                    env_vars_textBox.Text = (env_vars_textBox.Text.Trim() + " DeinoMPI_USE_CALL_HISTORY=1").Trim();
                }
                this.toolTip.SetToolTip(this.use_history_checkBox, "Remove MPI history");
            }
            else
            {
                // Remove DeinoMPI_USE_CALL_HISTORY=1
                env_vars_textBox.Text = rex.Replace(env_vars_textBox.Text, "").Trim();
                this.toolTip.SetToolTip(this.use_history_checkBox, "Add MPI history");
            }
        }

        private void use_smp_optimizations_checkBox_CheckedChanged(object sender, EventArgs e)
        {
            Regex rex = new Regex(@"DeinoMPI_USE_SMP_OPTIMIZATIONS\s*=\s*""?1""?", RegexOptions.IgnoreCase);
            if (use_smp_optimizations_checkBox.Checked)
            {
                // Add DeinoMPI_USE_SMP_OPTIMIZATIONS=1
                Match match = rex.Match(env_vars_textBox.Text);
                if (!match.Success)
                {
                    env_vars_textBox.Text = (env_vars_textBox.Text.Trim() + " DeinoMPI_USE_SMP_OPTIMIZATIONS=1").Trim();
                }
                this.toolTip.SetToolTip(this.use_smp_optimizations_checkBox, "Remove SMP optimizations");
            }
            else
            {
                // Remove DeinoMPI_USE_SMP_OPTIMIZATIONS=1
                env_vars_textBox.Text = rex.Replace(env_vars_textBox.Text, "").Trim();
                this.toolTip.SetToolTip(this.use_smp_optimizations_checkBox, "Add SMP optimizations");
            }
        }

        private void verify_job_button_Click(object sender, EventArgs e)
        {
            if (verify_backgroundWorker.IsBusy)
                return;

            string error_str = null;
            verified_hosts_listBox.Items.Clear();
            verified_launch_info_richTextBox.Text = "";
            ArrayList nodes = GenerateLaunchNodes(ref error_str);
            if (error_str != null && error_str.Trim() != "")
            {
                verified_launch_info_richTextBox.Text = error_str;
                return;
            }
            if (nodes == null)
            {
                verified_launch_info_richTextBox.Text = "Empty job: The data in the mpiexec tab did not generate a valid job description.";
                return;
            }
            VerifiedNode vnode;
            foreach (LaunchNode node in nodes)
            {
                vnode = new VerifiedNode();
                vnode.node = node;
                verified_hosts_listBox.Items.Add(vnode);
            }
            vnode = new VerifiedNode();
            verified_hosts_listBox.Items.Add(vnode); // add an empty node to the end to represent the entire job
            verify_backgroundWorker.RunWorkerAsync(nodes);
        }

        private void verified_hosts_listBox_DrawItem(object sender, DrawItemEventArgs e)
        {
            VerifiedNode node;
            if (e.Index == verified_hosts_listBox.Items.Count - 1)
                node = null;
            else
                node = (VerifiedNode)verified_hosts_listBox.Items[e.Index];
            if ((e.State & DrawItemState.Selected) == DrawItemState.Selected)
            {
                e.Graphics.FillRectangle(Brushes.CornflowerBlue, e.Bounds);
            }
            else
            {
                if (node != null)
                {
                    switch (node.state)
                    {
                        case VerifiedNode.VerificationState.GOOD:
                            SolidBrush green_brush = new SolidBrush(Color.FromArgb(200, 255, 200));
                            e.Graphics.FillRectangle(green_brush, e.Bounds);
                            break;
                        case VerifiedNode.VerificationState.OK:
                            SolidBrush turq_brush = new SolidBrush(Color.FromArgb(204, 204, 51));
                            e.Graphics.FillRectangle(turq_brush, e.Bounds);
                            break;
                        case VerifiedNode.VerificationState.BAD:
                            e.Graphics.FillRectangle(Brushes.Red, e.Bounds);
                            break;
                        case VerifiedNode.VerificationState.INDETERMINATE:
                            e.DrawBackground();
                            break;
                        case VerifiedNode.VerificationState.VERIFICATION_PENDING:
                            e.Graphics.FillRectangle(Brushes.LightGray, e.Bounds);
                            break;
                    }
                }
                else
                    e.DrawBackground();
            }

            Font font;
            Brush black_brush;
            font = verified_hosts_listBox.Font;
            black_brush = new SolidBrush(Color.Black);
            string node_str;
            if (node != null)
            {
                if (verify_mpiexec.global_options.localonly || node.node.launch_local)
                    node_str = String.Format("[{0}] {1}\n", node.node.iproc, Environment.MachineName);
                else
                    node_str = String.Format("[{0}] {1}\n", node.node.iproc, node.node.hostname);
            }
            else
                node_str = "Entire Job";
            if (e.Index >= 0)
                e.Graphics.DrawString(node_str, font, black_brush, e.Bounds.Left, e.Bounds.Top);

            e.DrawFocusRectangle();

            if ((e.Index >= 0) && (e.Index == verified_hosts_listBox.SelectedIndex))
            {
                if (node == null)
                {
                    verified_launch_info_richTextBox.Clear();
                    for (int i = 0; i < verified_hosts_listBox.Items.Count - 1; i++)
                    {
                        node = (VerifiedNode)verified_hosts_listBox.Items[i];
                        verified_launch_info_richTextBox.AppendText(node.GetLaunchInfoString());
                    }
                }
                else
                {
                    verified_launch_info_richTextBox.Text = node.GetLaunchInfoString();
                }
            }
        }

        private delegate void UpdateVerifiedMessageDelegate(int index, LaunchInfoMessage message);
        private UpdateVerifiedMessageDelegate verified_delegate;
        void UpdateVerifyListboxItem(int index, LaunchInfoMessage message)
        {
            VerifiedNode vnode = (VerifiedNode)verified_hosts_listBox.Items[index];
            vnode.message = message;
            if (message.bad_exe_path != "" || message.missing_dlls != "")
                vnode.state = VerifiedNode.VerificationState.BAD;
            else if (message.bad_wdir != "" || message.missing_delayed_dlls != "")
                vnode.state = VerifiedNode.VerificationState.OK;
            else
                vnode.state = VerifiedNode.VerificationState.GOOD;
            verified_hosts_listBox.Invalidate();
        }

        private delegate void UpdateVerifiedErrorDelegate(int index, string error_msg);
        private UpdateVerifiedErrorDelegate verified_error_delegate;
        void UpdateVerifyListboxItemError(int index, string error_msg)
        {
            VerifiedNode vnode = (VerifiedNode)verified_hosts_listBox.Items[index];
            vnode.error_msg = error_msg;
            vnode.state = VerifiedNode.VerificationState.BAD;
            verified_hosts_listBox.Invalidate();
        }

        private delegate void UpdateVerifiedStateDelegate(int index, VerifiedNode.VerificationState state);
        private UpdateVerifiedStateDelegate verified_state_delegate;
        void UpdateVerifyListboxItemState(int index, VerifiedNode.VerificationState state)
        {
            VerifiedNode vnode = (VerifiedNode)verified_hosts_listBox.Items[index];
            vnode.state = state;
            verified_hosts_listBox.Invalidate();
        }

        // code in verify_thread must be synchronized with deinopm_thread_fn and MpiexecProcess.verify_thread
        private void verify_thread(object o)
        {
            LaunchNode node = null;
            try
            {
                ClientSession client;
                node = (LaunchNode)o;
                Invoke(verified_state_delegate, new object[] { node.iproc, VerifiedNode.VerificationState.VERIFICATION_PENDING });
                try
                {
                    if (verify_mpiexec.global_options.localonly || node.launch_local)
                    {
                        int port = DeinoPMUtil.StartLocalPM(verify_mpiexec, node.private_key);
                        client = new ClientSession(port, node.private_key);
                    }
                    else
                    {
                        client = new ClientSession(node.hostname, -1, verify_mpiexec.global_options.phrase, store.Keys, node.credential);
                    }
                }
                catch (Exception e)
                {
                    StringBuilder sb = new StringBuilder("ERROR: ");
                    sb.Append(e.Message);
                    foreach (string key in e.Data.Keys)
                    {
                        sb.AppendFormat("\r\n{0}={1}", key, e.Data[key]);
                    }
                    Invoke(verified_error_delegate, new object[] { node.iproc, sb.ToString() });
                    return;
                }
                PMI pmi;
                pmi = node.pmi;
                LaunchMessage launch = new LaunchMessage();

                // create a launch message from the launch node

                // add the process fields
                launch.path = node.path;
                foreach (string drive in node.map.Keys)
                {
                    launch.map[drive] = node.map[drive];
                }
                launch.dir = node.dir;
                launch.exe = node.exe;
                foreach (string arg in node.args)
                {
                    launch.args.Add(arg);
                }
                foreach (string key in node.env.Keys)
                {
                    launch.env[key] = node.env[key];
                }
                launch.priority_class = node.priority_class;
                launch.priority_thread = node.priority_thread;

                // add the pmi variables
                if (node.spawned)
                    launch.env["PMI_SPAWN"] = "1";
                launch.env["PMI_RANK"] = node.iproc.ToString();
                launch.env["PMI_SIZE"] = node.nproc.ToString();
                launch.env["PMI_CLIQUE"] = node.clique;
                launch.env["PMI_APPNUM"] = node.appnum.ToString();
                if (pmi != null)
                {
                    launch.env["PMI_DOMAIN"] = pmi.domain;
                    launch.env["PMI_KVS"] = pmi.id;
                }
                else
                {
                    launch.env["PMI_DOMAIN"] = "any_domain";
                    launch.env["PMI_KVS"] = "any_id";
                }
                // SMPD variables
                launch.env["PMI_HOST"] = System.Net.Dns.GetHostName();
                launch.env["PMI_PORT"] = "-1";
                launch.env["PMI_SMPD_ID"] = node.iproc.ToString();
                launch.env["PMI_SMPD_KEY"] = node.iproc.ToString();
                // DeinoPMI variables
                launch.env["PMI_KEY"] = node.iproc.ToString();

                InfoRequestMessage message = new InfoRequestMessage("launch_info");
                client.SendMessage(message);
                client.SendMessage(launch);
                LaunchInfoMessage lmessage = (LaunchInfoMessage)client.NextMessage();
                QuitMessage quit_message = new QuitMessage();
                client.SendMessage(quit_message);
                // get a response to the quit message
                DeinoMPI.Message message2;
                message2 = client.NextMessage();
                try
                {
                    // Send the last leg of the quit handshake
                    quit_message = new QuitMessage();
                    client.SendMessage(quit_message);
                }
                catch
                {
                    // Ignore errors here since we are guaranteed to be in the last leg of the quit handshake.
                }
                Invoke(verified_delegate, new object[] { node.iproc, lmessage });
            }
            catch (Exception ex)
            {
                StringBuilder sb = new StringBuilder("ERROR: ");
                sb.Append(ex.Message);
                foreach (string key in ex.Data.Keys)
                {
                    sb.AppendFormat("\r\n{0}={1}", key, ex.Data[key]);
                }
                if (node != null)
                {
                    Invoke(verified_error_delegate, new object[] { node.iproc, sb.ToString() });
                }
                else
                {
                    Invoke(verified_error_delegate, new object[] { 0, sb.ToString() });
                }
            }
        }

        private void verify_backgroundWorker_DoWork(object sender, DoWorkEventArgs e)
        {
            try
            {
                int i, num_threads;
                Thread[] threads;
                ArrayList nodes = (ArrayList)e.Argument;
                if (nodes == null)
                    return;
                num_threads = nodes.Count;
                threads = new Thread[num_threads];
                for (i = 0; i < num_threads; i++)
                {
                    ParameterizedThreadStart ts = new ParameterizedThreadStart(verify_thread);
                    threads[i] = new Thread(ts);
                    threads[i].Start(nodes[i]);
                }
                for (i = 0; i < num_threads; i++)
                    threads[i].Join();
            }
            catch
            {
            }
        }

        private void staging_dir_browse_button_Click(object sender, EventArgs e)
        {
            FolderBrowserDialog dlg = new FolderBrowserDialog();
            if (staging_src_dir_textBox.Text.Length > 0)
            {
                dlg.SelectedPath = staging_src_dir_textBox.Text;
            }
            else
            {
                dlg.SelectedPath = Environment.CurrentDirectory;
            }
            if (dlg.ShowDialog() == DialogResult.OK)
            {
                staging_src_dir_textBox.Text = dlg.SelectedPath;
            }
        }

        private void use_staging_checkBox_CheckedChanged(object sender, EventArgs e)
        {
            if (use_staging_checkBox.Checked)
            {
                staging_src_dir_textBox.Enabled = true;
                staging_src_dir_browse_button.Enabled = true;
                staging_src_label.Enabled = true;
                staging_recursive_checkBox.Enabled = true;
                staging_dest_dir_textBox.Enabled = true;
                staging_dest_label.Enabled = true;
                staging_pull_checkBox.Enabled = true;
                staging_wildcards_textBox.Enabled = true;
                staging_wildcards_label.Enabled = true;
                staging_warning_label.Visible = true;
            }
            else
            {
                staging_src_dir_textBox.Enabled = false;
                staging_src_dir_browse_button.Enabled = false;
                staging_src_label.Enabled = false;
                staging_recursive_checkBox.Enabled = false;
                staging_dest_dir_textBox.Enabled = false;
                staging_dest_label.Enabled = false;
                staging_pull_checkBox.Enabled = false;
                staging_wildcards_textBox.Enabled = false;
                staging_wildcards_label.Enabled = false;
                staging_warning_label.Visible = false;
            }
        }

        private void staging_help_button_Click(object sender, EventArgs e)
        {
            StagingHelp dlg = new StagingHelp();
            dlg.ShowDialog();
        }

        private void options_tabControl_SelectedIndexChanged(object sender, EventArgs e)
        {

        }
    }

    public class OutputQueueWatcher
    {
        public Mpiexec mpiexec;
        public DeinoMPIwin win;
        public Queue queue;
        public int nproc;

        public void Run()
        {
            int num_exited = 0;
            while (num_exited < mpiexec.TotalNumProcs)
            {
                if (queue.Count > 0)
                {
                    object o = queue.Dequeue();
                    if (o is Thread)
                    {
                        ((Thread)o).Join();
                        num_exited++;
                    }
                    else
                    {
                        object[] pList = { o };
                        // put the string in the edit box in a thread safe way
                        DeinoMPIwin.AppendTextDelegate ap = new DeinoMPIwin.AppendTextDelegate(win.AppendText);
                        win.output_richTextBox.Invoke(ap, pList);
                    }
                }
                else
                {
                    Thread.Sleep(100);
                }
            }
            if (mpiexec.global_options.exitcodes)
            {
                DeinoMPIwin.PrintExitcodesDelegate pe = new DeinoMPIwin.PrintExitcodesDelegate(win.PrintExitcodes);
                win.output_richTextBox.Invoke(pe);
                /* We cannot assume there will be mpiexec.TotalNumProcs number of exitcode messages because some of them may already have been outputed in the above loop while waiting for the first process group to exit.
                // If we assume that there is one line of exitcode output for each process then this loop should work:
                for (int i = 0; i < mpiexec.TotalNumProcs; i++)
                {
                    while (queue.Count < 1)
                        Thread.Sleep(100);
                    object o = queue.Dequeue();
                    object[] pList = { o };
                    // put the string in the edit box in a thread safe way
                    DeinoMPIwin.AppendTextDelegate ap = new DeinoMPIwin.AppendTextDelegate(win.AppendText);
                    win.output_richTextBox.Invoke(ap, pList);
                }
                */
                while (queue.Count > 0)
                {
                    object o = queue.Dequeue();
                    object[] pList = { o };
                    // put the string in the edit box in a thread safe way
                    DeinoMPIwin.AppendTextDelegate ap = new DeinoMPIwin.AppendTextDelegate(win.AppendText);
                    win.output_richTextBox.Invoke(ap, pList);
                }
                // Let's wait a short time to allow spawned threads to print their exit codes.
                // This in no way guarantees all exit code messages will be printed but it can be a bandaid until a synchronized method of printing exit codes is implemented.
                Thread.Sleep(250);
                while (queue.Count > 0)
                {
                    object o = queue.Dequeue();
                    object[] pList = { o };
                    // put the string in the edit box in a thread safe way
                    DeinoMPIwin.AppendTextDelegate ap = new DeinoMPIwin.AppendTextDelegate(win.AppendText);
                    win.output_richTextBox.Invoke(ap, pList);
                }
            }
            DeinoMPIwin.EnableExecuteDelegate eb = new DeinoMPIwin.EnableExecuteDelegate(win.EnableExecute);
            win.execute_button.Invoke(eb);
        }
    }

    public class ComputerBrowser : IEnumerable, IDisposable
    {
        #region "Server type enumeration"
        [FlagsAttribute]
        public enum ServerType : uint
        {
            SV_TYPE_WORKSTATION = 0x00000001, // All workstations
            SV_TYPE_SERVER = 0x00000002, // All computers that have the server service running
            SV_TYPE_SQLSERVER = 0x00000004, // Any server running Microsoft SQL Server
            SV_TYPE_DOMAIN_CTRL = 0x00000008, // Primary domain controller
            SV_TYPE_DOMAIN_BAKCTRL = 0x00000010, // Backup domain controller
            SV_TYPE_TIME_SOURCE = 0x00000020, // Server running the Timesource service
            SV_TYPE_AFP = 0x00000040, // Apple File Protocol servers
            SV_TYPE_NOVELL = 0x00000080, // Novell servers
            SV_TYPE_DOMAIN_MEMBER = 0x00000100, // LAN Manager 2.x domain member
            SV_TYPE_PRINTQ_SERVER = 0x00000200, // Server sharing print queue
            SV_TYPE_DIALIN_SERVER = 0x00000400, // Server running dial-in service
            SV_TYPE_XENIX_SERVER = 0x00000800, // Xenix server
            SV_TYPE_NT = 0x00001000, // Windows NT workstation or server
            SV_TYPE_WFW = 0x00002000, // Server running Windows for Workgroups
            SV_TYPE_SERVER_MFPN = 0x00004000, // Microsoft File and Print for NetWare
            SV_TYPE_SERVER_NT = 0x00008000, // Server that is not a domain controller
            SV_TYPE_POTENTIAL_BROWSER = 0x00010000, // Server that can run the browser service
            SV_TYPE_BACKUP_BROWSER = 0x00020000, // Server running a browser service as backup
            SV_TYPE_MASTER_BROWSER = 0x00040000, // Server running the master browser service
            SV_TYPE_DOMAIN_MASTER = 0x00080000, // Server running the domain master browser
            SV_TYPE_WINDOWS = 0x00400000, // Windows 95 or later
            SV_TYPE_DFS = 0x00800000, // Root of a DFS tree
            SV_TYPE_TERMINALSERVER = 0x02000000, // Terminal Server
            SV_TYPE_CLUSTER_NT = 0x01000000, // Server clusters available in the domain
            SV_TYPE_CLUSTER_VS_NT = 0x04000000, // Cluster virtual servers available in the domain
            SV_TYPE_DCE = 0x10000000, // IBM DSS (Directory and Security Services) or equivalent
            SV_TYPE_ALTERNATE_XPORT = 0x20000000, // Return list for alternate transport
            SV_TYPE_LOCAL_LIST_ONLY = 0x40000000, // Return local list only
            SV_TYPE_DOMAIN_ENUM = 0x80000000  // Lists available domains
        }
        #endregion

        // Server information structure
        [StructLayoutAttribute(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
        internal struct SERVER_INFO_101
        {
            public int sv101_platform_id;
            public string sv101_name;
            public int sv101_version_major;
            public int sv101_version_minor;
            public int sv101_type;
            public string sv101_comment;
        }

        // enumerates network computers
        [DllImport("Netapi32", CharSet = CharSet.Unicode)]
        private static extern int NetServerEnum(
            string servername,		// must be null
            int level,				// 100 or 101
            out IntPtr bufptr,		// pointer to buffer receiving data
            int prefmaxlen,			// max length of returned data
            out int entriesread,	// num entries read
            out int totalentries,	// total servers + workstations
            uint servertype,		// server type filter
            [MarshalAs(UnmanagedType.LPWStr)]
			string domain,			// domain to enumerate
            IntPtr resume_handle);

        // Frees buffer created by NetServerEnum
        [DllImport("netapi32.dll")]
        private extern static int NetApiBufferFree(IntPtr buf);

        // Constants
        private const int ERROR_ACCESS_DENIED = 5;
        private const int ERROR_MORE_DATA = 234;
        private const int ERROR_NO_SERVERS = 6118;

        private NetworkComputers[] _computers;
        private string _lastError = "";

        public ComputerBrowser(ServerType serverType, string domain)
            : this(UInt32.Parse(Enum.Format(typeof(ServerType), serverType, "x"), System.Globalization.NumberStyles.HexNumber), domain)
        {
        }
        public ComputerBrowser(ServerType serverType)
            : this(UInt32.Parse(Enum.Format(typeof(ServerType), serverType, "x"), System.Globalization.NumberStyles.HexNumber), null)
        {
        }
        public ComputerBrowser(string domainName)
            : this(0xFFFFFFFF, domainName)
        {
        }

        /// <summary>
        /// Enumerate the hosts of type serverType in the specified domain
        /// </summary>
        /// <param name="serverType">Server type filter</param>
        /// <param name="domain">The domain to search for computers in</param>
        public ComputerBrowser(uint serverType, string domainName)
        {
            int entriesread;  // number of entries actually read
            int totalentries; // total visible servers and workstations
            int result;		  // result of the call to NetServerEnum

            // Pointer to buffer that receives the data
            IntPtr pBuf = IntPtr.Zero;
            Type svType = typeof(SERVER_INFO_101);

            // structure containing info about the server
            SERVER_INFO_101 si;

            try
            {
                result = NetServerEnum(null, 101, out pBuf, -1, out entriesread, out totalentries, serverType, domainName, IntPtr.Zero);

                // Successful?
                if (result != 0)
                {
                    switch (result)
                    {
                        case ERROR_MORE_DATA:
                            _lastError = "More data is available";
                            break;
                        case ERROR_ACCESS_DENIED:
                            _lastError = "Access was denied";
                            break;
                        case ERROR_NO_SERVERS:
                            _lastError = "No servers available for this domain";
                            break;
                        default:
                            _lastError = "Unknown error code " + result;
                            break;
                    }
                    return;
                }
                else
                {
                    _computers = new NetworkComputers[entriesread];

                    int tmp = (int)pBuf;
                    for (int i = 0; i < entriesread; i++)
                    {
                        // fill our struct
                        si = (SERVER_INFO_101)Marshal.PtrToStructure((IntPtr)tmp, svType);
                        _computers[i] = new NetworkComputers(si);

                        // next struct
                        tmp += Marshal.SizeOf(svType);
                    }
                }
            }
            finally
            {
                // free the buffer
                NetApiBufferFree(pBuf);
                pBuf = IntPtr.Zero;
            }
        }

        /// <summary>
        /// Total computers in the collection
        /// </summary>
        public int Length
        {
            get
            {
                if (_computers != null)
                {
                    return _computers.Length;
                }
                else
                {
                    return 0;
                }
            }
        }

        /// <summary>
        /// Last error message
        /// </summary>
        public string LastError
        {
            get { return _lastError; }
        }

        /// <summary>
        /// Obtains the enumerator for ComputerEnumerator class
        /// </summary>
        /// <returns>IEnumerator</returns>
        public IEnumerator GetEnumerator()
        {
            return new ComputerEnumerator(_computers);
        }

        // cleanup
        /*
        public void Dispose()
        {
            _computers = null;
        }
        */

        public NetworkComputers this[int item]
        {
            get
            {
                return _computers[item];
            }
        }

        // holds computer info.
        public struct NetworkComputers
        {
            ComputerBrowser.SERVER_INFO_101 _computerinfo;
            internal NetworkComputers(ComputerBrowser.SERVER_INFO_101 info)
            {
                _computerinfo = info;
            }

            /// <summary>
            /// Name of computer
            /// </summary>
            public string Name
            {
                get { return _computerinfo.sv101_name; }
            }
            /// <summary>
            /// Server comment
            /// </summary>
            public string Comment
            {
                get { return _computerinfo.sv101_comment; }
            }
            /// <summary>
            /// Major version number of OS
            /// </summary>
            public int VersionMajor
            {
                get { return _computerinfo.sv101_version_major; }
            }
            /// <summary>
            /// Minor version number of OS
            /// </summary>
            public int VersionMinor
            {
                get { return _computerinfo.sv101_version_minor; }
            }
        }

        /// <summary>
        /// Enumerates the collection of computers
        /// </summary>
        public class ComputerEnumerator : IEnumerator
        {
            private NetworkComputers[] aryComputers;
            private int indexer;

            internal ComputerEnumerator(NetworkComputers[] networkComputers)
            {
                aryComputers = networkComputers;
                indexer = -1;
            }

            public void Reset()
            {
                indexer = -1;
            }

            public object Current
            {
                get
                {
                    return aryComputers[indexer];
                }
            }

            public bool MoveNext()
            {
                if (aryComputers == null)
                {
                    return false;
                }
                if (indexer < aryComputers.Length)
                {
                    indexer++;
                }
                return (!(indexer == aryComputers.Length));
            }
        }

        #region IDisposable Members

        void IDisposable.Dispose()
        {
            _computers = null;
        }

        #endregion
    }

    /// <summary>
    /// This class is an implementation of the 'IComparer' interface.
    /// </summary>
    public class ListViewColumnSorter : IComparer
    {
        /// <summary>
        /// Specifies the column to be sorted
        /// </summary>
        private int ColumnToSort;
        /// <summary>
        /// Specifies the order in which to sort (i.e. 'Ascending').
        /// </summary>
        private SortOrder OrderOfSort;
        /// <summary>
        /// Case insensitive comparer object
        /// </summary>
        private CaseInsensitiveComparer ObjectCompare;

        /// <summary>
        /// Class constructor.  Initializes various elements
        /// </summary>
        public ListViewColumnSorter()
        {
            // Initialize the column to '0'
            ColumnToSort = 0;

            // Initialize the sort order to 'none'
            OrderOfSort = SortOrder.None;

            // Initialize the CaseInsensitiveComparer object
            ObjectCompare = new CaseInsensitiveComparer();
        }

        /// <summary>
        /// This method is inherited from the IComparer interface.  It compares the two objects passed using a case insensitive comparison.
        /// </summary>
        /// <param name="x">First object to be compared</param>
        /// <param name="y">Second object to be compared</param>
        /// <returns>The result of the comparison. "0" if equal, negative if 'x' is less than 'y' and positive if 'x' is greater than 'y'</returns>
        public int Compare(object x, object y)
        {
            int compareResult;
            ListViewItem listviewX, listviewY;

            // Cast the objects to be compared to ListViewItem objects
            listviewX = (ListViewItem)x;
            listviewY = (ListViewItem)y;

            // Compare the two items
            compareResult = ObjectCompare.Compare(listviewX.SubItems[ColumnToSort].Text, listviewY.SubItems[ColumnToSort].Text);

            // Calculate correct return value based on object comparison
            if (OrderOfSort == SortOrder.Ascending)
            {
                // Ascending sort is selected, return normal result of compare operation
                return compareResult;
            }
            else if (OrderOfSort == SortOrder.Descending)
            {
                // Descending sort is selected, return negative result of compare operation
                return (-compareResult);
            }
            else
            {
                // Return '0' to indicate they are equal
                return 0;
            }
        }

        /// <summary>
        /// Gets or sets the number of the column to which to apply the sorting operation (Defaults to '0').
        /// </summary>
        public int SortColumn
        {
            set
            {
                ColumnToSort = value;
            }
            get
            {
                return ColumnToSort;
            }
        }

        /// <summary>
        /// Gets or sets the order of sorting to apply (for example, 'Ascending' or 'Descending').
        /// </summary>
        public SortOrder Order
        {
            set
            {
                OrderOfSort = value;
            }
            get
            {
                return OrderOfSort;
            }
        }

    }

}
