using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Diagnostics;
using System.ServiceProcess;
using System.Text;
using DeinoMPI;
using System.Runtime.InteropServices;
using System.Configuration.Install;

namespace DeinoPM
{
    public partial class DeinoPM : ServiceBase
    {
        public static DeinoPM_Shell dash = new DeinoPM_Shell();

        [Flags]
        public enum ErrorModes
        {
            Default = 0x0,
            FailCriticalErrors = 0x1,
            NoGpFaultErrorBox = 0x2,
            NoAlignmentFaultExcept = 0x4,
            NoOpenFileErrorBox = 0x8000
        }
        [DllImport("kernel32.dll")]
        private static extern int SetErrorMode(int newMode);
        public int ChangeErrorMode(ErrorModes mode)
        {
            return SetErrorMode((int)mode);
        }

        public DeinoPM()
        {
            InitializeComponent();
        }

        public void OnDebugStart(string[] args)
        {
            DeinoPM_Shell.RedirectOutput = false;
            DeinoPM_Shell.DebugMode = true;
            dash.OnStart(args);
            dash.WaitForStop();
        }

        protected override void OnStart(string[] args)
        {
            dash.OnStart(args);
        }

        public void OnDebugStop()
        {
            DeinoPM_Shell.RedirectOutput = false;
            dash.OnStop();
        }

        protected override void OnStop()
        {
            dash.OnStop();
        }

        internal void RunManager(string public_key)
        {
            // The manager cannot print messages to the console until after it writes the port number back to the calling process */
            /*Console.WriteLine("calling ChangeErrorMode");*/
            try
            {
                ChangeErrorMode(ErrorModes.FailCriticalErrors /*| ErrorModes.NoAlignmentFaultExcept*/ | ErrorModes.NoGpFaultErrorBox | ErrorModes.NoOpenFileErrorBox);
            }
            catch //(Exception e)
            {
            }
            string[] args = new string[1];
            args[0] = public_key;
            DeinoPM_Shell.IsManager = true;
            dash.OnStart(args);
            dash.WaitForStop();
        }

        internal void RunManager(string read_handle, string write_handle)
        {
            try
            {
                ChangeErrorMode(ErrorModes.FailCriticalErrors /*| ErrorModes.NoAlignmentFaultExcept*/ | ErrorModes.NoGpFaultErrorBox | ErrorModes.NoOpenFileErrorBox);
            }
            catch //(Exception e)
            {
            }
            string[] args = new string[2];
            args[0] = read_handle;
            args[1] = write_handle;
            DeinoPM_Shell.IsManager = true;
            dash.OnStart(args);
            dash.WaitForStop();
        }
    }

    [RunInstaller(true)]
    public class WindowsServiceInstaller : Installer
    {
        /// <summary>
        /// Public Constructor for WindowsServiceInstaller.
        /// </summary>
        public WindowsServiceInstaller()
        {
            ServiceProcessInstaller serviceProcessInstaller = new ServiceProcessInstaller();
            ServiceInstaller serviceInstaller = new ServiceInstaller();

            //# Service Account Information
            serviceProcessInstaller.Account = ServiceAccount.LocalSystem;
            //serviceProcessInstaller.Account = ServiceAccount.NetworkService;
            serviceProcessInstaller.Username = null;
            serviceProcessInstaller.Password = null;

            //# Service Information
            serviceInstaller.DisplayName = "DeinoMPI process manager service";
            serviceInstaller.StartType = ServiceStartMode.Automatic;

            //# This must be identical to the WindowsService.ServiceBase name
            //# set in the constructor of WindowsService.cs
            serviceInstaller.ServiceName = "DeinoPM";

            this.Installers.Add(serviceProcessInstaller);
            this.Installers.Add(serviceInstaller);
        }
    }
}
