#include <stdio.h>
#include <ctype.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <stdlib.h>
#include <windows.h>
#include "pmi.h"

using namespace DeinoMPI;

// FIXME: include mpi.h to get these definitions
#define MPI_ERR_NAME        33
#define MPI_MAX_PORT_NAME  256

#define PMI_MAX_KEY_LEN          256
#define PMI_MAX_VALUE_LEN        8192
#define PMI_MAX_KVS_NAME_LENGTH  100

#define PMI_INITIALIZED 0
#define PMI_FINALIZED   1

#define PMI_TRUE        1
#define PMI_FALSE       0

public ref class global
{
public:
	static PMI_Client ^client = nullptr;
	static ActivePMI_Client ^active = nullptr;
};

public class error_string_node
{
public:
	int error;
	wchar_t *error_str;
	error_string_node *next;
};

static error_string_node * pmi_errors = NULL;
static int pmi_error = 1; /* Don't start with 0 because 0 equals success */
static wchar_t pmi_last_error_msg[8192] = L"";
static PMI_Abort_callback_t abort_callback_function = NULL;
static PMI_Print_queues_fn_t print_queues_function = NULL;
static PMI_Print_call_history_fn_t print_call_history_function = NULL;
static HANDLE active_thread = NULL;

static void CopyString(System::String ^str, wchar_t dest[], int length)
{
	int len;
	if (str == nullptr)
	{
		if (length > 0)
			dest[0] = L'\0';
		return;
	}
	len = min(length-1, str->Length);
	for (int i=0; i<len; i++)
	{
		dest[i] = (*str)[i];//str->Chars[i];//(*str)[i];
	}
	dest[len] = L'\0';
}

DWORD ActiveThread(void *p)
{
	bool active = true;
	ActiveMessage ^message, ^result;
	System::String ^abort_callback_str = gcnew System::String(L"abort_callback");
	System::String ^queues_str = gcnew System::String(L"queues");
	System::String ^history_str = gcnew System::String(L"history");
	//System::String ^quit_str = gcnew System::String(L"quit");
	System::String ^exit_str = gcnew System::String(L"exit");
	wchar_t queue_state[8192];

	global::active->Accept();
	while (active)
	{
		message = (ActiveMessage^)global::active->session->NextMessage();
		result = gcnew ActiveMessage();
		result->active_type = gcnew System::String(L"server");
		//result->name = gcnew System::String(L"result");
		result->name = message->name;
		result->data[L"success"] = gcnew System::String(L"true");
		if (System::String::Compare(message->name, abort_callback_str) == 0)
		{
			int exit_code = System::Convert::ToInt32(message->data["exit_code"]);
			if (abort_callback_function != NULL)
				abort_callback_function(exit_code);
			// return result message with success = true
			global::active->session->SendMessage(result);
		}
		else if (System::String::Compare(message->name, queues_str) == 0)
		{
			//printf("returning queues empty active message\n");fflush(stdout);
			// return queues message with success = true and data = queues string
			result->name = queues_str;
			if (print_queues_function != NULL)
			{
				queue_state[0] = L'\0';
				print_queues_function(queue_state, 8192);
				queue_state[8191] = L'\0';
				result->data[L"data"] = gcnew System::String(queue_state);
				//wprintf(L"print queues returning <%s>\n", queue_state);fflush(stdout);
			}
			else
			{
				result->data[L"data"] = gcnew System::String(L"Queues empty");
			}
			// print queue results are sent back to the root through the regular pmi connection
			global::client->SendMessage(result);
		}
		else if (System::String::Compare(message->name, history_str) == 0)
		{
			// return history message with success = true and data = MPI call history string
			result->name = history_str;
			if (print_call_history_function != NULL)
			{
				/* Use the queue_state buffer to save memory */
				queue_state[0] = L'\0';
				print_call_history_function(queue_state, 8192);
				queue_state[8191] = L'\0';
				result->data[L"data"] = gcnew System::String(queue_state);
				//wprintf(L"print queues returning <%s>\n", queue_state);fflush(stdout);
			}
			else
			{
				result->data[L"data"] = gcnew System::String(L"");
			}
			// print queue results are sent back to the root through the regular pmi connection
			global::client->SendMessage(result);
		}
		/*
		else if (System::String::Compare(message->name, quit_str) == 0)
		{
		// return result message with success = true
		global::active->session->SendMessage(result);
		active = false;
		}
		*/
		else if (System::String::Compare(message->name, exit_str) == 0)
		{
			//System::Console::WriteLine(message->data["exit_code"]);
			int exit_code = System::Convert::ToInt32(message->data["exit_code"]);
			//printf("exiting with code %d\n", exit_code);fflush(stdout);
			ExitProcess((DWORD)exit_code);
		}
		else
		{
			// return result message with success = false
			result->data[L"success"] = gcnew System::String(L"false");
			global::active->session->SendMessage(result);
		}
	}
	return 0;
}

PMI_API int PMI_CALL PMI_Initialized(PMI_BOOL *initialized)
{
	if (initialized == NULL)
		return PMI_ERR_INVALID_ARG;
	if (global::client == nullptr)
		*initialized = PMI_FALSE;
	else
	{
		*initialized = global::client->Initialized ? PMI_TRUE : PMI_FALSE;
	}
	return PMI_SUCCESS;
}

PMI_API int PMI_CALL PMI_Init(int *spawned)
{
	if (spawned == NULL)
		return PMI_ERR_INVALID_ARG;

	try
	{
		// I wish this would work for the entire application but the scope is the dll.
		// If the main executable module wants this setting it has to be set there or inherited from the process that started this process.
		SetErrorMode(SEM_FAILCRITICALERRORS | SEM_NOGPFAULTERRORBOX | SEM_NOOPENFILEERRORBOX);

		//System::Console::WriteLine("PMI_Init");
		if (global::client == nullptr)
		{
			global::active = gcnew ActivePMI_Client();
			global::active->Listen();
			active_thread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)ActiveThread, NULL, 0, NULL);
			global::client = gcnew PMI_Client();
		}
		bool ret_val = global::client->Init(global::active);
		*spawned = ret_val ? 1 : 0;
	} catch (System::Exception ^ex)
	{
		CopyString(ex->Message, pmi_last_error_msg, 8192);
		return PMI_FAIL;
	}
	/*
	printf("PMI_RANK=%s PMI_SIZE=%s PMI_KVS=%s PMI_SMPD_ID=%s PMI_SMPD_FD=%s PMI_SMPD_KEY=%s\n PMI_SPAWN=%s",
	getenv("PMI_RANK"), getenv("PMI_SIZE"), getenv("PMI_KVS"), getenv("PMI_SMPD_ID"),
	getenv("PMI_SMPD_FD"), getenv("PMI_SMPD_KEY"), getenv("PMI_SPAWN"));
	fflush(stdout);
	*/

	/*printf("PMI_Init returning success.\n");fflush(stdout);*/
	return PMI_SUCCESS;
}

PMI_API int PMI_CALL PMI_Finalize()
{
	if (global::client == nullptr)
		return PMI_ERR_INIT;
	try
	{
		global::client->pmiFinalize();
	} catch (System::Exception ^ex)
	{
		CopyString(ex->Message, pmi_last_error_msg, 8192);
		return PMI_FAIL;
	}
	/*printf("PMI_Finalize success.\n");fflush(stdout);*/
	return PMI_SUCCESS;
}

PMI_API int PMI_CALL PMI_AbortW(int exit_code, const wchar_t error_msg[])
{
	if (global::client == nullptr)
		return PMI_ERR_INIT;
	try
	{
		//printf("calling PMI_Client::Abort\n");
		/* flush any output before aborting */
		fflush(stdout);
		fflush(stderr);
		if (error_msg != NULL)
			global::client->Abort(exit_code, gcnew System::String(error_msg));
		else
			global::client->Abort(exit_code, nullptr);
	}
	catch (System::Exception ^ex)
	{
		CopyString(ex->Message, pmi_last_error_msg, 8192);
	}
	catch (...)
	{
		wprintf(L"PMI_Abort failed.\n");fflush(stdout);
	}
	// Wait for the active thread to handle the exit command
	WaitForSingleObject(active_thread, 30000);
	// If the active_thread does not handle the exit command then exit here.
	ExitProcess(exit_code);
	return PMI_FAIL;
}

PMI_API int PMI_CALL PMI_Get_size(int *size)
{
	if (global::client == nullptr)
		return PMI_ERR_INIT;
	if (size == NULL)
		return PMI_ERR_INVALID_ARG;

	try
	{
		*size = global::client->size;
	} catch (System::Exception ^ex)
	{
		CopyString(ex->Message, pmi_last_error_msg, 8192);
		return PMI_FAIL;
	}

	return PMI_SUCCESS;
}

PMI_API int PMI_CALL PMI_Get_rank(int *rank)
{
	if (global::client == nullptr)
		return PMI_ERR_INIT;
	if (rank == NULL)
		return PMI_ERR_INVALID_ARG;

	try
	{
		*rank = global::client->rank;
	} catch (System::Exception ^ex)
	{
		CopyString(ex->Message, pmi_last_error_msg, 8192);
		return PMI_FAIL;
	}

	return PMI_SUCCESS;
}

PMI_API int PMI_CALL PMI_Get_universe_size(int *size)
{
	if (global::client == nullptr)
		return PMI_ERR_INIT;
	if (size == NULL)
		return PMI_ERR_INVALID_ARG;

	try
	{
		*size = global::client->universe_size;
	} catch (System::Exception ^ex)
	{
		CopyString(ex->Message, pmi_last_error_msg, 8192);
		return PMI_FAIL;
	}

	return PMI_SUCCESS;
}

PMI_API int PMI_CALL PMI_Get_appnum(int *appnum)
{
	if (global::client == nullptr)
		return PMI_ERR_INIT;
	if (appnum == NULL)
		return PMI_ERR_INVALID_ARG;

	try
	{
		*appnum = global::client->appnum;
	} catch (System::Exception ^ex)
	{
		CopyString(ex->Message, pmi_last_error_msg, 8192);
		return PMI_FAIL;
	}

	return PMI_SUCCESS;
}

PMI_API int PMI_CALL PMI_Get_clique_size( int *size )
{
	if (global::client == nullptr)
		return PMI_ERR_INIT;
	if (size == NULL)
		return PMI_ERR_INVALID_ARG;

	try
	{
		*size = global::client->clique_size;
	} catch (System::Exception ^ex)
	{
		CopyString(ex->Message, pmi_last_error_msg, 8192);
		return PMI_FAIL;
	}

	return PMI_SUCCESS;
}

PMI_API int PMI_CALL PMI_Get_clique_ranks( int ranks[], int length )
{
	int i;

	if (global::client == nullptr)
		return PMI_ERR_INIT;
	if (ranks == NULL)
		return PMI_ERR_INVALID_ARG;
	if (length < global::client->clique_size)
		return PMI_ERR_INVALID_LENGTH;

	try
	{
		for (i=0; i<global::client->clique_size; i++)
		{
			ranks[i] = global::client->clique_ranks[i];
		}
	} catch (System::Exception ^ex)
	{
		CopyString(ex->Message, pmi_last_error_msg, 8192);
		return PMI_FAIL;
	}
	return PMI_SUCCESS;
}

PMI_API int PMI_CALL PMI_Get_idW( wchar_t id_str[], int length )
{
	if (global::client == nullptr)
		return PMI_ERR_INIT;
	if (id_str == NULL)
		return PMI_ERR_INVALID_ARG;
	if (length < PMI_MAX_KVS_NAME_LENGTH)
		return PMI_ERR_INVALID_LENGTH;

	try
	{
		CopyString(global::client->id, id_str, length);
	} catch (System::Exception ^ex)
	{
		CopyString(ex->Message, pmi_last_error_msg, 8192);
		return PMI_FAIL;
	}

	return PMI_SUCCESS;
}

PMI_API int PMI_CALL PMI_Get_id_length_max(int *maxlen)
{
	if (global::client == nullptr)
		return PMI_ERR_INIT;
	try
	{
		*maxlen = global::client->id_length_max;
	} catch (System::Exception ^ex)
	{
		CopyString(ex->Message, pmi_last_error_msg, 8192);
		return PMI_FAIL;
	}
	return PMI_SUCCESS;
}

PMI_API int PMI_CALL PMI_Get_kvs_domain_idW(wchar_t id_str[], int length)
{
	if (global::client == nullptr)
		return PMI_ERR_INIT;
	if (id_str == NULL)
		return PMI_ERR_INVALID_ARG;
	if (length < PMI_MAX_KVS_NAME_LENGTH)
		return PMI_ERR_INVALID_LENGTH;

	try
	{
		CopyString(global::client->domain_id, id_str, length);
	} catch (System::Exception ^ex)
	{
		CopyString(ex->Message, pmi_last_error_msg, 8192);
		return PMI_FAIL;
	}

	return PMI_SUCCESS;
}

PMI_API int PMI_CALL PMI_Barrier()
{
	if (global::client == nullptr)
		return PMI_ERR_INIT;

	try
	{
		global::client->Barrier();
	} catch (System::Exception ^ex)
	{
		CopyString(ex->Message, pmi_last_error_msg, 8192);
		return PMI_FAIL;
	}
	/*printf("PMI_Barrier success.\n");fflush(stdout);*/
	return PMI_SUCCESS;
}

PMI_API int PMI_CALL PMI_KVS_Get_my_nameW(wchar_t kvsname[], int length)
{
	if (global::client == nullptr)
		return PMI_ERR_INIT;
	if (kvsname == NULL)
		return PMI_ERR_INVALID_ARG;
	if (length < PMI_MAX_KVS_NAME_LENGTH)
		return PMI_ERR_INVALID_LENGTH;

	try
	{
		CopyString(global::client->my_name, kvsname, length);
	} catch (System::Exception ^ex)
	{
		CopyString(ex->Message, pmi_last_error_msg, 8192);
		return PMI_FAIL;
	}

	/*
	wprintf(L"my kvs name is %s\n", kvsname);fflush(stdout);
	*/

	return PMI_SUCCESS;
}

PMI_API int PMI_CALL PMI_KVS_Get_name_length_max(int *maxlen)
{
	if (global::client == nullptr)
		return PMI_ERR_INIT;
	if (maxlen == NULL)
		return PMI_ERR_INVALID_ARG;

	try
	{
		*maxlen = global::client->name_length_max;
	} catch (System::Exception ^ex)
	{
		CopyString(ex->Message, pmi_last_error_msg, 8192);
		return PMI_FAIL;
	}

	return PMI_SUCCESS;
}

PMI_API int PMI_CALL PMI_KVS_Get_key_length_max(int *maxlen)
{
	if (global::client == nullptr)
		return PMI_ERR_INIT;
	if (maxlen == NULL)
		return PMI_ERR_INVALID_ARG;

	try
	{
		*maxlen = global::client->key_length_max;
	} catch (System::Exception ^ex)
	{
		CopyString(ex->Message, pmi_last_error_msg, 8192);
		return PMI_FAIL;
	}

	return PMI_SUCCESS;
}

PMI_API int PMI_CALL PMI_KVS_Get_value_length_max(int *maxlen)
{
	if (global::client == nullptr)
		return PMI_ERR_INIT;
	if (maxlen == NULL)
		return PMI_ERR_INVALID_ARG;

	try
	{
		*maxlen = global::client->value_length_max;
	} catch (System::Exception ^ex)
	{
		CopyString(ex->Message, pmi_last_error_msg, 8192);
		return PMI_FAIL;
	}

	return PMI_SUCCESS;
}

PMI_API int PMI_CALL PMI_KVS_CreateW(wchar_t kvsname[], int length)
{
	if (global::client == nullptr)
		return PMI_ERR_INIT;
	if (kvsname == NULL)
		return PMI_ERR_INVALID_ARG;
	if (length < global::client->name_length_max)
		return PMI_ERR_INVALID_LENGTH;

	try
	{
		CopyString(global::client->KVS_Create(), kvsname, length);
	} catch (System::Exception ^ex)
	{
		CopyString(ex->Message, pmi_last_error_msg, 8192);
		return PMI_FAIL;
	}

	/*wprintf(L"PMI_KVS_Create success.\n");fflush(stdout);*/
	return PMI_SUCCESS;
}

PMI_API int PMI_CALL PMI_KVS_DestroyW(const wchar_t kvsname[])
{
	if (global::client == nullptr)
		return PMI_ERR_INIT;
	if (kvsname == NULL)
		return PMI_ERR_INVALID_ARG;

	try
	{
		global::client->KVS_Destroy(gcnew System::String(kvsname));
	} catch (System::Exception ^ex)
	{
		CopyString(ex->Message, pmi_last_error_msg, 8192);
		return PMI_FAIL;
	}

	return PMI_SUCCESS;
}

PMI_API int PMI_CALL PMI_KVS_PutW(const wchar_t kvsname[], const wchar_t key[], const wchar_t value[])
{
	if (global::client == nullptr)
		return PMI_ERR_INIT;
	if (kvsname == NULL)
		return PMI_ERR_INVALID_ARG;
	if (key == NULL)
		return PMI_ERR_INVALID_KEY;
	if (value == NULL)
		return PMI_ERR_INVALID_VAL;

	/*wprintf(L"putting <%s><%s><%s>\n", kvsname, key, value);fflush(stdout);*/

	try
	{
		global::client->KVS_Put(gcnew System::String(kvsname), gcnew System::String(key), gcnew System::String(value));
	} catch (System::Exception ^ex)
	{
		CopyString(ex->Message, pmi_last_error_msg, 8192);
		//System::Console::WriteLine(gcnew System::String(L"KVS_Put failed: {0}"), e->Message);
		return PMI_FAIL;
	}
	/*wprintf(L"PMI_KVS_Put success.\n");fflush(stdout);*/
	return PMI_SUCCESS;
}

PMI_API int PMI_CALL PMI_KVS_CommitW(const wchar_t kvsname[])
{
	if (global::client == nullptr)
		return PMI_ERR_INIT;
	if (kvsname == NULL)
		return PMI_ERR_INVALID_ARG;

	try
	{
		global::client->KVS_Commit(gcnew System::String(kvsname));
	} catch (System::Exception ^ex)
	{
		CopyString(ex->Message, pmi_last_error_msg, 8192);
		return PMI_FAIL;
	}

	return PMI_SUCCESS;
}

PMI_API int PMI_CALL PMI_KVS_GetW(const wchar_t kvsname[], const wchar_t key[], wchar_t value[], int length)
{
	System::String ^str;

	if (global::client == nullptr)
		return PMI_ERR_INIT;
	if (kvsname == NULL)
		return PMI_ERR_INVALID_ARG;
	if (key == NULL)
		return PMI_ERR_INVALID_KEY;
	if (value == NULL)
		return PMI_ERR_INVALID_VAL;

	try
	{
		str = global::client->KVS_Get(gcnew System::String(kvsname), gcnew System::String(key));
		CopyString(str, value, length);
	} catch (System::Exception ^ex)
	{
		CopyString(ex->Message, pmi_last_error_msg, 8192);
		return PMI_FAIL;
	}

	/*
	wprintf(L"PMI_KVS_Get success.\n");fflush(stdout);
	wprintf(L"get <%s><%s><%s>\n", kvsname, key, value);
	fflush(stdout);
	*/
	return PMI_SUCCESS;
}

PMI_API int PMI_CALL PMI_KVS_Iter_firstW(const wchar_t kvsname[], wchar_t key[], int key_len, wchar_t value[], int val_len)
{
	System::String ^key_str, ^val_str;

	if (global::client == nullptr)
		return PMI_ERR_INIT;
	if (kvsname == NULL)
		return PMI_ERR_INVALID_ARG;
	if (key == NULL)
		return PMI_ERR_INVALID_KEY;
	if (key_len < global::client->key_length_max)
		return PMI_ERR_INVALID_KEY_LENGTH;
	if (value == NULL)
		return PMI_ERR_INVALID_VAL;
	if (val_len < global::client->value_length_max)
		return PMI_ERR_INVALID_VAL_LENGTH;

	try
	{
		//key_str = gcnew System::String();
		//val_str = gcnew System::String();
		global::client->KVS_Iter_first(gcnew System::String(kvsname), key_str, val_str);
		if (key_str != nullptr && val_str != nullptr)
		{
			CopyString(key_str, key, key_len);
			CopyString(val_str, value, val_len);
		}
		else
		{
			key[0] = L'\0';
			value[0] = L'\0';
		}
	} catch (System::Exception ^ex)
	{
		CopyString(ex->Message, pmi_last_error_msg, 8192);
		return PMI_FAIL;
	}

	return PMI_SUCCESS;
}

PMI_API int PMI_CALL PMI_KVS_Iter_nextW(const wchar_t kvsname[], wchar_t key[], int key_len, wchar_t value[], int val_len)
{
	System::String ^key_str, ^val_str;

	if (global::client == nullptr)
		return PMI_ERR_INIT;
	if (kvsname == NULL)
		return PMI_ERR_INVALID_ARG;
	if (key == NULL)
		return PMI_ERR_INVALID_KEY;
	if (key_len < global::client->key_length_max)
		return PMI_ERR_INVALID_KEY_LENGTH;
	if (value == NULL)
		return PMI_ERR_INVALID_VAL;
	if (val_len < global::client->value_length_max)
		return PMI_ERR_INVALID_VAL_LENGTH;

	try
	{
		//key_str = gcnew System::String();
		//val_str = gcnew System::String();
		global::client->KVS_Iter_next(gcnew System::String(kvsname), key_str, val_str);
		if (key_str != nullptr && val_str != nullptr)
		{
			CopyString(key_str, key, key_len);
			CopyString(val_str, value, val_len);
		}
		else
		{
			key[0] = L'\0';
			value[0] = L'\0';
		}
	} catch (System::Exception ^ex)
	{
		CopyString(ex->Message, pmi_last_error_msg, 8192);
		return PMI_FAIL;
	}

	return PMI_SUCCESS;
}

static int SaveErrorString(System::String^ str)
{
	error_string_node * node = new error_string_node();
	node->error = pmi_error++;
	node->error_str = new wchar_t[str->Length + 1];
	if (node->error_str != NULL)
		CopyString(str, node->error_str, str->Length + 1);
	node->next = pmi_errors;
	pmi_errors = node;
	return node->error;
}

PMI_API int PMI_CALL PMI_Get_error_stringW(int pmi_error, wchar_t *str, int max_length)
{
	error_string_node *iter = pmi_errors;
	while (iter != NULL)
	{
		if (iter->error == pmi_error)
		{
			if (iter->error_str != NULL)
			{
				size_t len = wcslen(iter->error_str);
				if (len > (size_t)max_length - 1)
					len = max_length - 1;
				for (size_t i=0; i<len; i++)
					str[i] = iter->error_str[i];
				str[len] = L'\0';
			}
			break;
		}
		iter = iter->next;
	}
	return PMI_SUCCESS;
}

PMI_API int PMI_CALL PMI_Free_error_string(int pmi_error)
{
	error_string_node *iter, *trailer;
	iter = trailer = pmi_errors;
	while (iter != NULL)
	{
		if (iter->error == pmi_error)
		{
			if (trailer != iter)
				trailer->next = iter->next;
			if (iter == pmi_errors)
				pmi_errors = iter->next;
			delete iter;
			return PMI_SUCCESS;
		}
		if (trailer != iter)
			trailer = trailer->next;
		iter = iter->next;
	}
	return PMI_SUCCESS;
}

PMI_API int PMI_CALL PMI_Spawn_multipleW(int count,
										 const wchar_t * cmds[],
										 const wchar_t ** argvs[],
										 const int maxprocs[],
										 const int cinfo_keyval_sizes[],
										 const PMI_keyvalW_t * info_keyval_vectors[],
										 int preput_keyval_size,
										 const PMI_keyvalW_t preput_keyval_vector[],
										 int errors[],
										 int quiet)
{
	//System::Console::WriteLine("PMI_Spawn_mulitple");

	if (global::client == nullptr)
		return PMI_ERR_INIT;
	if (count < 1 || cmds == NULL || maxprocs == NULL || preput_keyval_size < 0)
		return PMI_ERR_INVALID_ARG;

	/*
	wprintf(L"cmds = %p, argvs = %p, maxprocs = %p, sizes = %p, keyvals = %p\n", cmds, argvs, maxprocs, cinfo_keyval_sizes, info_keyval_vectors);fflush(stdout);
	System::Console::WriteLine("allocating data");
	*/

	array<System::String^> ^gcmds = gcnew array<System::String^>(count);
	array<array<System::String^>^ > ^gargvs;
	gargvs = gcnew array<array<System::String^>^ >(count);
	array<int> ^gmaxprocs = gcnew array<int>(count);
	array<array<DeinoMPI::PMI_Keyval^>^ > ^ginfos;
	ginfos = gcnew array<array<DeinoMPI::PMI_Keyval^>^>(count);
	array<DeinoMPI::PMI_Keyval^> ^gpreputs = nullptr;
	array<System::String^> ^gerrors;// = gcnew array<System::String^>(count);
	int total_num_processes = 0;
	bool gquiet = (quiet != 0);
	for (int i=0; i<count; i++)
	{
		total_num_processes += maxprocs[i];
		if (argvs != NULL && argvs[i] != NULL)
		{
			int cur_count = 0;
			const wchar_t **iter = argvs[i];
			while (*iter != NULL)
			{
				cur_count++;
				iter++;
			}
			gargvs[i] = gcnew array<System::String^>(cur_count);
			for (int j=0; j<cur_count; j++)
			{
				gargvs[i][j] = gcnew System::String(argvs[i][j]);
			}
		}
		else
		{
			gargvs[i] = nullptr;
		}
		if (cinfo_keyval_sizes != NULL && *cinfo_keyval_sizes != NULL && cinfo_keyval_sizes[i] > 0)
		{
			ginfos[i] = gcnew array<DeinoMPI::PMI_Keyval^>(cinfo_keyval_sizes[i]);
			for (int j=0; j<cinfo_keyval_sizes[i]; j++)
			{
				PMI_Keyval ^kv = gcnew PMI_Keyval();
				//wprintf(L"adding key=%s, val=%s\n", info_keyval_vectors[i][j].key, info_keyval_vectors[i][j].val);fflush(stdout);
				kv->key = gcnew System::String(info_keyval_vectors[i][j].key);
				kv->val = gcnew System::String(info_keyval_vectors[i][j].val);
				ginfos[i][j] = kv;
			}
		}
		else
		{
			ginfos[i] = nullptr;
		}
		gcmds[i] = gcnew System::String(cmds[i]);
		gmaxprocs[i] = maxprocs[i];
		//gerrors[i] = nullptr;
	}
	gerrors = gcnew array<System::String^>(total_num_processes);
	for (int i=0; i<total_num_processes; i++)
		gerrors[i] = nullptr;
	if (preput_keyval_size > 0)
	{
		gpreputs = gcnew array<DeinoMPI::PMI_Keyval^>(preput_keyval_size);
		for (int i=0; i<preput_keyval_size; i++)
		{
			/*
			PMI_Keyval ^kv = gcnew PMI_Keyval();
			wprintf(L"adding key=%s, val=%s\n", preput_keyval_vector[i].key, preput_keyval_vector[i].val);fflush(stdout);
			kv->key = gcnew System::String(preput_keyval_vector[i].key);
			kv->val = gcnew System::String(preput_keyval_vector[i].val);
			gpreputs[i] = kv;
			*/
			//wprintf(L"adding key=%s, val=%s\n", preput_keyval_vector[i].key, preput_keyval_vector[i].val);fflush(stdout);
			gpreputs[i] = gcnew PMI_Keyval();
			gpreputs[i]->key = gcnew System::String(preput_keyval_vector[i].key);
			gpreputs[i]->val = gcnew System::String(preput_keyval_vector[i].val);
			//System::Console::WriteLine(L"adding preput key={0}, val={1}", gpreputs[i]->key, gpreputs[i]->val);
		}
	}
	try
	{
		//wprintf(L"calling PMI_Client::Spawn_multiple\n");fflush(stdout);
		global::client->Spawn_multiple(gcmds, gargvs, gmaxprocs, ginfos, gpreputs, gerrors, gquiet);
		//wprintf(L"PMI_Client::Spawn_multiple returned\n");fflush(stdout);
		if (errors != NULL)
		{
			for (int i=0; i<total_num_processes; i++)
			{
				if (gerrors[i] == nullptr)
					errors[i] = 0;
				else
				{
					errors[i] = SaveErrorString(gerrors[i]);
				}
			}
		}
	} catch (System::Exception ^ex)
	{
		CopyString(ex->Message, pmi_last_error_msg, 8192);
		return PMI_FAIL;
	}
	return PMI_SUCCESS;
}

PMI_API int PMI_CALL PMI_Parse_option(int num_args, wchar_t *args[], int *num_parsed, PMI_keyval_t **keyvalp, int *size)
{
	if (num_args < 1)
		return PMI_ERR_INVALID_NUM_ARGS;
	if (args == NULL)
		return PMI_ERR_INVALID_ARGS;
	if (num_parsed == NULL)
		return PMI_ERR_INVALID_NUM_PARSED;
	if (keyvalp == NULL)
		return PMI_ERR_INVALID_KEYVALP;
	if (size == NULL)
		return PMI_ERR_INVALID_SIZE;
	*num_parsed = 0;
	*keyvalp = NULL;
	*size = 0;
	return PMI_SUCCESS;
}

PMI_API int PMI_CALL PMI_Args_to_keyval(int *argcp, wchar_t *((*argvp)[]), PMI_keyval_t **keyvalp, int *size)
{
	if (argcp == NULL || argvp == NULL || keyvalp == NULL || size == NULL)
		return PMI_ERR_INVALID_ARG;
	return PMI_SUCCESS;
}

PMI_API int PMI_CALL PMI_Free_keyvals(PMI_keyval_t keyvalp[], int size)
{
	if (keyvalp == NULL || size < 0)
		return PMI_ERR_INVALID_ARG;
	if (size == 0)
		return PMI_SUCCESS;
	/* free stuff */
	return PMI_SUCCESS;
}

PMI_API int PMI_CALL PMI_Publish_nameW( const wchar_t service_name[], const wchar_t port[] )
{
	if (global::client == nullptr)
		return PMI_ERR_INIT;
	if (service_name == NULL || port == NULL)
		return PMI_ERR_INVALID_ARG;

	try
	{
		global::client->Publish_name(gcnew System::String(service_name), gcnew System::String(port));
	} catch (System::Exception ^ex)
	{
		CopyString(ex->Message, pmi_last_error_msg, 8192);
		return PMI_FAIL;
	}

	return PMI_SUCCESS;
}

PMI_API int PMI_CALL PMI_Unpublish_nameW( const wchar_t service_name[] )
{
	if (global::client == nullptr)
		return PMI_ERR_INIT;
	if (service_name == NULL)
		return PMI_ERR_INVALID_ARG;

	try
	{
		global::client->Unpublish_name(gcnew System::String(service_name));
	} catch (System::Exception ^ex)
	{
		CopyString(ex->Message, pmi_last_error_msg, 8192);
		return PMI_FAIL;
	}

	return PMI_SUCCESS;
}

PMI_API int PMI_CALL PMI_Lookup_nameW( const wchar_t service_name[], wchar_t port[] )
{
	System::String ^port_str;

	if (global::client == nullptr)
		return PMI_ERR_INIT;
	if (service_name == NULL || port == NULL)
		return PMI_ERR_INVALID_ARG;

	try
	{
		port_str = global::client->Lookup_name(gcnew System::String(service_name));
		CopyString(port_str, port, MPI_MAX_PORT_NAME);
		if (port[0] == L'\0')
		{
			return MPI_ERR_NAME;
		}
	} catch (System::Exception ^ex)
	{
		CopyString(ex->Message, pmi_last_error_msg, 8192);
		return PMI_FAIL;
	}
	return PMI_SUCCESS;
}

PMI_API int PMI_CALL PMI_Register_abort_hook(PMI_Abort_callback_t function)
{
	abort_callback_function = function;
	return PMI_SUCCESS;
}

PMI_API int PMI_CALL PMI_Register_print_queues_fn(PMI_Print_queues_fn_t function)
{
	print_queues_function = function;
	return PMI_SUCCESS;
}

PMI_API int PMI_CALL PMI_Print_queues(wchar_t *str, int max_length)
{
	if (global::client == nullptr)
		return PMI_ERR_INIT;
	if (str == NULL || max_length < 1)
		return PMI_ERR_INVALID_ARG;
	if (print_queues_function == NULL)
		return PMI_ERR_INVALID_ARG; // FIXME: add a new error code for unregistered functions

	print_queues_function(str, max_length);
	// check the result
	for (int i=0; i<max_length; i++)
	{
		if (str[i] == L'\0')
			return PMI_SUCCESS;
	}
	// Returned string was not NULL terminated.
	str[max_length-1] = L'\0';
	return PMI_FAIL;
}

PMI_API int PMI_CALL PMI_Register_print_call_history_fn(PMI_Print_call_history_fn_t function)
{
	print_call_history_function = function;
	return PMI_SUCCESS;
}

PMI_API int PMI_CALL PMI_Print_call_history(wchar_t *str, int max_length)
{
	if (global::client == nullptr)
		return PMI_ERR_INIT;
	if (str == NULL || max_length < 1)
		return PMI_ERR_INVALID_ARG;
	if (print_call_history_function == NULL)
		return PMI_ERR_INVALID_ARG; // FIXME: add a new error code for unregistered functions

	print_call_history_function(str, max_length);
	// check the result
	for (int i=0; i<max_length; i++)
	{
		if (str[i] == L'\0')
			return PMI_SUCCESS;
	}
	// Returned string was not NULL terminated.
	str[max_length-1] = L'\0';
	return PMI_FAIL;
}

PMI_API const wchar_t * PMI_CALL PMI_Get_last_error_msgW()
{
	return pmi_last_error_msg;
}
