using System;
using System.Collections.Generic;
using System.Text;
using DeinoMPI;
using System.IO;
using System.Collections;
using System.Threading;

namespace create_credential_store
{
    class create_credential_store
    {
        static void SaveStoreToHardDrive(SecurityKeys keys)
        {
            string path = "";
            for (; ; )
            {
                Console.WriteLine();
                Console.WriteLine("Enter the path to the folder where you want to save the store:");
                path = Console.ReadLine().Trim();
                if (Directory.Exists(path))
                    break;
                if (path == "")
                {
                    Console.WriteLine("The path cannot be empty. (Enter . for the current directory)");
                }
                else
                {
                    Console.WriteLine("That directory does not exist, would you like to create it (yes,no)? [no]");
                    if (Console.ReadLine().Trim().ToLower() == "yes")
                    {
                        try
                        {
                            Directory.CreateDirectory(path);
                            break;
                        }
                        catch (Exception e)
                        {
                            Console.WriteLine("Unable to create the directory:");
                            Console.WriteLine(e.Message);
                        }
                    }
                }
            }
            if (path.Length > 0 && !path.EndsWith("\\"))
                path += "\\";
            string filename = path + "CredentialStore.xml";
            try
            {
                CredentialStore store = new CredentialStore(keys, filename);
                if (store != null)
                {
                    Console.WriteLine();
                    Console.WriteLine("Credential store written to file: {0}", filename);
                }
            }
            catch (Exception e)
            {
                Console.WriteLine();
                Console.WriteLine("Failed to write the Credential store to file {0}", filename);
                Console.WriteLine("Error: {0}", e.Message);
            }
        }

        static bool SaveStoreToRemovableMedia(SecurityKeys keys, DriveInfo[] drives)
        {
            int i;
            string drives_str = "";
            ArrayList drive_letters = new ArrayList();
            Console.WriteLine();
            for (i = 0; i < drives.Length; i++)
            {
                if (drives[i].IsReady)
                    Console.WriteLine("({0}) {1} {2}", drives[i].Name[0], drives[i].Name, drives[i].VolumeLabel);
                else
                    Console.WriteLine("({0}) {1}", drives[i].Name[0], drives[i].Name);
                if (drives_str.Length > 0)
                    drives_str = drives_str + "," + drives[i].Name[0];
                else
                    drives_str = drives[i].Name[0].ToString();
                drive_letters.Add(drives[i].Name[0].ToString().ToLower());
            }
            string selected_drive = null;
            for (int max_tries = 0; max_tries < 5; max_tries++)
            {
                Console.WriteLine("Enter the drive letter where the store will be created ({0}): [{1}]", drives_str, drives[0].Name[0]);
                Console.WriteLine("Or enter \"cancel\" to return without saving the store.");
                selected_drive = Console.ReadLine();
                if (selected_drive == null)
                    selected_drive = "";
                else
                    selected_drive = selected_drive.Trim().ToLower();
                if (selected_drive == "cancel")
                    return false;
                bool found = false;
                foreach (string drive_letter in drive_letters)
                {
                    if (drive_letter == selected_drive)
                        found = true;
                }
                if (found)
                    break;
                if (max_tries == 4)
                    return false; // too many attempts
            }
            string filename = selected_drive + ":\\CredentialStore.xml";
            try
            {
                CredentialStore store = new CredentialStore(keys, filename);
                if (store != null)
                {
                    Console.WriteLine();
                    Console.WriteLine("Credential store written to file: {0}", filename);
                    return true;
                }
            }
            catch (Exception e)
            {
                Console.WriteLine();
                Console.WriteLine("Failed to write the Credential store to file {0}", filename);
                Console.WriteLine("Error: {0}", e.Message);
            }
            return false;
        }

        static bool SaveStoreToRegistry(SecurityKeys keys)
        {
            try
            {
                CredentialStore store = new CredentialStore(keys);
                if (store != null)
                {
                    Console.WriteLine();
                    Console.WriteLine("Credential store written to the Windows Registry.");
                    return true;
                }
            }
            catch (Exception e)
            {
                Console.WriteLine();
                Console.WriteLine("Failed to write the Credential store to the Windows Registry.");
                Console.WriteLine("Error: {0}", e.Message);
            }
            return false;
        }

        static void CreateDefaultCredentialStore()
        {
            try
            {
                SecurityKeys keys = new SecurityKeys();
                keys.Password = null;
                keys.Method = SecurityKeys.EncryptionMethod.PROTECT_DATA;
                keys.CreateKeys();
                SaveStoreToRegistry(keys);
            }
            catch (Exception e)
            {
                Console.WriteLine("Unable to create a Credential Store with the default settings: Error, {0}", e.Message);
            }
        }

        static void CreateCredentialStoreFromResponseFile(string filename)
        {
            try
            {
                StreamReader sr;
                using (sr = new StreamReader(filename))
                {
                    string password = sr.ReadLine().TrimEnd('\r', '\n');
                    string format = sr.ReadLine().TrimEnd('\r', '\n').ToLower();
                    if (format != "protect" && format != "symmetric" && format != "none")
                    {
                        Console.WriteLine("ERROR: Invalid format line in the response file.");
                        Console.WriteLine("Valid values are protect, symmetric, or none.");
                        Console.WriteLine("Invalid line = {0}", format);
                        return;
                    }
                    bool use_registry = true;
                    string path = sr.ReadLine().TrimEnd('\r', '\n');
                    if (path.ToLower() == "registry")
                    {
                        use_registry = true;
                    }
                    else
                    {
                        use_registry = false;
                        if (!Directory.Exists(path))
                        {
                            Console.WriteLine("ERROR: Invalid path in the response file.  The following path does not exist:");
                            Console.WriteLine(path);
                            return;
                        }
                    }
                    sr.Close();
                    SecurityKeys keys = new SecurityKeys();
                    keys.Password = (password != "") ? password : null;
                    switch (format)
                    {
                        case "protect":
                            keys.Method = SecurityKeys.EncryptionMethod.PROTECT_DATA;
                            break;
                        case "symmetric":
                            keys.Method = SecurityKeys.EncryptionMethod.SYMMETRIC_KEY;
                            break;
                        case "none":
                            keys.Method = SecurityKeys.EncryptionMethod.NONE;
                            break;
                    }
                    keys.CreateKeys();
                    if (use_registry)
                    {
                        SaveStoreToRegistry(keys);
                    }
                    else
                    {
                        if (path.Length > 0 && !path.EndsWith("\\"))
                            path += "\\";
                        string store_filename = path + "CredentialStore.xml";
                        try
                        {
                            CredentialStore store = new CredentialStore(keys, store_filename);
                            if (store != null)
                            {
                                Console.WriteLine();
                                Console.WriteLine("Credential store written to file: {0}", store_filename);
                            }
                        }
                        catch (Exception e)
                        {
                            Console.WriteLine();
                            Console.WriteLine("Failed to write the Credential store to file {0}", store_filename);
                            Console.WriteLine("Error: {0}", e.Message);
                        }
                    }
                }
            }
            catch (Exception e)
            {
                Console.WriteLine("Unable to create a Credential Store from the file [{0}]: Error, {1}", filename, e.Message);
            }
        }

        static void Main(string[] args)
        {
            if (args.Length > 0)
            {
                if (String.Compare(args[0], "--help", true) == 0 ||
                    String.Compare(args[0], "-help", true) == 0 ||
                    String.Compare(args[0], "/help", true) == 0 ||
                    String.Compare(args[0], "/?", true) == 0 ||
                    String.Compare(args[0], "-?", true) == 0)
                {
                    Console.WriteLine("Usage:");
                    Console.WriteLine(" create_credential_store.exe");
                    Console.WriteLine("    Interactive mode.");
                    Console.WriteLine(" create_credential_store.exe -default");
                    Console.WriteLine("    Create a store with the default settings:");
                    Console.WriteLine("    no password, protect, registry");
                    Console.WriteLine(" create_credential_store.exe filename");
                    Console.WriteLine("    Create a store for the current user from a response file.");
                    Console.WriteLine("    File format:");
                    Console.WriteLine("    [password]");
                    Console.WriteLine("    'protect' or 'symmetric' or 'none'");
                    Console.WriteLine("    [path] or 'registry'");
                    Console.WriteLine("    Example:");
                    Console.WriteLine("   >foobar");
                    Console.WriteLine("   >protect");
                    Console.WriteLine(@"   >C:\me\mydata");
                    return;
                }
                if (args[0] == "-default")
                    CreateDefaultCredentialStore();
                else
                    CreateCredentialStoreFromResponseFile(args[0]);
                return;
            }
            SecurityKeys keys = new SecurityKeys();
            Console.WriteLine();
            Console.WriteLine("This program creates a Credential Store for the current user.");
            Console.WriteLine("The credential store contains a public and private security key");
            Console.WriteLine("and zero or more secure user credentials.");
            Console.WriteLine("Deino MPI uses this store to establish secure connections between machines");
            Console.WriteLine("and store sensitive information like user credentials.");
            Console.WriteLine();
            Console.Write("Do you want to create a new store (yes, no)? [yes] ");
            string answer = Console.ReadLine().Trim().ToLower();
            if (answer != "yes" && answer != "")
            {
                Console.WriteLine("Goodbye.");
                return;
            }
            Console.WriteLine();
            keys.GetPassword(true);
            Console.WriteLine();
            keys.Method = SecurityKeys.EncryptionMethod.PROTECT_DATA;
            keys.GetEncryptionMethod(true);
            keys.CreateKeys();
            ArrayList removable_drives = new ArrayList();
            foreach (DriveInfo drive in System.IO.DriveInfo.GetDrives())
            {
                if (drive.DriveType == DriveType.Removable)
                    removable_drives.Add(drive);
            }
            int max_tries = 0;
        try_again:
            string location = "";
            for (; ; )
            {
                Console.WriteLine("Available locations to create the store:");
                if (removable_drives.Count > 0)
                {
                    Console.WriteLine("(1) Removable media");
                    Console.WriteLine("(2) Windows Registry");
                    Console.WriteLine("(3) Hard drive");
                    Console.Write("Where would you like to create the store (1,2,3): [1] ");
                }
                else
                {
                    Console.WriteLine("(1) Windows Registry");
                    Console.WriteLine("(2) Hard drive");
                    Console.Write("Where would you like to create the store (1,2): [1] ");
                }
                location = Console.ReadLine();
                if (location == null)
                    location = "";
                else
                    location = location.Trim();
                if (location == "")
                {
                    location = "1";
                    break;
                }
                if (location == "1")
                    break;
                if (location == "2")
                    break;
                if (location == "3" && removable_drives.Count > 0)
                    break;
            }
            if (removable_drives.Count > 0)
            {
                if (location == "1")
                {
                    DriveInfo[] drives = new DriveInfo[removable_drives.Count];
                    Array.Copy(removable_drives.ToArray(), drives, drives.Length);
                    if (!SaveStoreToRemovableMedia(keys, drives))
                    {
                        max_tries++;
                        if (max_tries > 4)
                            return; // give up after 5 tries
                        goto try_again;
                    }
                }
                else if (location == "2")
                {
                    SaveStoreToRegistry(keys);
                }
                else if (location == "3")
                {
                    SaveStoreToHardDrive(keys);
                }
            }
            else
            {
                if (location == "1")
                {
                    SaveStoreToRegistry(keys);
                }
                else if (location == "2")
                {
                    SaveStoreToHardDrive(keys);
                }
            }
        }
    }
}
