#pragma once
#include <mpi.h>
#include <windows.h>

class FileTreeFileNode
{
private:
    wchar_t *filename;
    wchar_t *name_part;
    FILETIME write_time;
public:
    bool traversed;
    FileTreeFileNode(void);
    ~FileTreeFileNode(void);
    void SetFileName(const wchar_t *filename);
    void SetWriteTime(FILETIME write_time);
    const wchar_t *GetFullFileName();
    const wchar_t *GetFileName();
    FILETIME GetWriteTime();
    bool Different(FileTreeFileNode *file);
    FileTreeFileNode *next;
    void ResetTraversed();
    void Print();
};

class FileTreeDirNode
{
private:
    FileTreeDirNode *parent, *children, *sibling;
    FileTreeFileNode *files;
    wchar_t *full_path, *dir_name;
public:
    bool traversed;
    FileTreeDirNode(void);
    FileTreeDirNode(const wchar_t *current_dir, const wchar_t *sub_dir);
    ~FileTreeDirNode(void);
    void ResetTraversed();
    void AddChild(wchar_t *dir, wchar_t *wild_cards);
    FileTreeDirNode* CreateChild(const wchar_t *dir);
    FileTreeDirNode* CreateSibling(const wchar_t *dir);
    void Populate(wchar_t *wild_cards);
    void AddFile(const wchar_t *filename, const FILETIME last_write);
    bool Empty();
    void PruneEmptyDirectories();
    FileTreeDirNode *GetParent();
    FileTreeDirNode *GetFirstChild();
    FileTreeDirNode *GetNextSibling();
    FileTreeFileNode *GetFirstFile();
    const wchar_t *GetDirName();
    const wchar_t *GetFullPath();
    void Print();
};

class FileTreeFileNodeList
{
public:
    FileTreeFileNodeList();
    FileTreeFileNode *file;
    FileTreeFileNodeList *next;
};

class FileTree
{
private:
    wchar_t *src_dir, *dest_dir;
    FileTreeDirNode *root;
    bool recursive;
    int src_rank;
    void PopulateTree(wchar_t *src_dir, wchar_t *wild_cards);
    FileTreeFileNodeList *iter_list;
public:
    FileTree(void);
    FileTree(wchar_t *src_dir, bool recursive);
    ~FileTree(void);

    void SetSrcRank(int comm_world_rank);
    void SetSrcDir(wchar_t *dir);
    void SetDestDir(wchar_t *dir);
    const wchar_t *GetDestDir();
    void CreateDestDir();

    void PopulateTreeFromSrc(wchar_t *wild_cards);
    void PopulateTreeFromDest(wchar_t *wild_cards);

    void SendTree();
    void SendTree(FileTree *dest_tree);
    void ReceiveTree();

    FileTree* GetChangedFiles(FileTree *tree);

    FileTreeFileNode *GetFirstFile();
    FileTreeFileNode *GetNextFile();
    void Print();
};
