/* -*- Mode: C; c-basic-offset:4 ; -*- */
/*
 *  (C) 2006 by Deino Software.
 *  (C) 2001 by Argonne National Laboratory.
 *	  See COPYRIGHT in top-level directory.
 */

#include <stdio.h>
#include <stdlib.h>
#include "smpd.h"
#include "mpi.h"
#include "smpd_service.h"
#ifdef HAVE_SIGNAL_H
#include <signal.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif

#undef FCNAME
#define FCNAME L"main"
int wmain(int argc, wchar_t* argv[])
{
	int result;
	SERVICE_TABLE_ENTRY dispatchTable[] =
	{
		{ TEXT(SMPD_SERVICE_NAME), (LPSERVICE_MAIN_FUNCTION)smpd_service_main },
		{ NULL, NULL }
	};

	smpd_enter_fn(FCNAME);

	/* initialization */
	result = PMPI_Init(&argc, &argv);
	if (result != MPI_SUCCESS)
	{
		smpd_err_printf(L"MPI_Init failed,\n error: %d\n", result);
		smpd_exit_fn(FCNAME);
		return result;
	}

	/* Initialization now done by MPI_Init above.
	result = MPIDU_Sock_init();
	if (result != MPI_SUCCESS)
	{
		smpd_err_printf(L"MPIDU_Sock_init failed,\nsock error: %s\n", get_sock_error_string(result));
		smpd_exit_fn(FCNAME);
		return result;
	}
	*/

	/* If PMI_Init uses smpd then smpd_init_process will be called but if another process manager is chosen then init needs to be called here: */
	result = smpd_init_process();
	if (result != SMPD_SUCCESS)
	{
		smpd_err_printf(L"smpd_init_process failed.\n");
		smpd_exit_fn(FCNAME);
		return result;
	}

	/* parse the command line */
	result = smpd_parse_command_args(&argc, &argv);
	if (result != SMPD_SUCCESS)
	{
		smpd_err_printf(L"Unable to parse the command arguments.\n");
		smpd_exit_fn(FCNAME);
		return result;
	}

	if (smpd_process.bService)
	{
		/*
		wprintf( L"\nStartServiceCtrlDispatcher being called.\n" );
		wprintf( L"This may take several seconds.  Please wait.\n" );
		fflush(stdout);
		*/

		/* If StartServiceCtrlDispatcher returns true the service has exited */
		result = StartServiceCtrlDispatcher(dispatchTable);
		if (result)
		{
			smpd_exit_fn(FCNAME);
			smpd_exit(0);
		}

		result = GetLastError();
		if (result != ERROR_FAILED_SERVICE_CONTROLLER_CONNECT)
		{
			smpd_add_error_to_message_log(TEXT("StartServiceCtrlDispatcher failed."));
			smpd_exit_fn(FCNAME);
			smpd_exit(0);
		}
		smpd_print_options();
		smpd_exit_fn(FCNAME);
		smpd_exit(0);
	}
	/*
	wprintf(L"\nRunning smpd from the console, not as a service.\n");
	fflush(stdout);
	smpd_process.bService = SMPD_FALSE;
	*/

	if (smpd_process.passphrase[0] == L'\0')
		smpd_get_smpd_data(L"phrase", smpd_process.passphrase, SMPD_PASSPHRASE_MAX_LENGTH);
	if (smpd_process.passphrase[0] == L'\0')
	{
		if (smpd_process.noprompt)
		{
			wprintf(L"Error: No smpd passphrase specified through the registry or .smpd file, exiting.\n");
			smpd_exit_fn(FCNAME);
			return -1;
		}
		wprintf(L"Please specify an authentication passphrase for this smpd: ");
		fflush(stdout);
		smpd_get_password(smpd_process.passphrase, SMPD_PASSPHRASE_MAX_LENGTH);
	}

	result = smpd_entry_point();

	smpd_finalize_printf();

	smpd_exit(result);
	smpd_exit_fn(FCNAME);
	return result;
}

#undef FCNAME
#define FCNAME L"smpd_entry_point"
int smpd_entry_point()
{
	int result;
	MPIDU_Sock_set_t set;
	MPIDU_Sock_t listener;
	SMPD_BOOL print_port = SMPD_FALSE;

	/* This function is called by main or by smpd_service_main in the case of a Windows service */

	smpd_enter_fn(FCNAME);

	/* prevent the os from bringing up debug message boxes if this process crashes */
	if (smpd_process.bService)
	{
		SetErrorMode(SEM_FAILCRITICALERRORS | SEM_NOGPFAULTERRORBOX | SEM_NOOPENFILEERRORBOX);
		if (!smpd_report_status_to_sc_mgr(SERVICE_RUNNING, NO_ERROR, 0))
		{
			result = GetLastError();
			smpd_err_printf(L"Unable to report that the service has started, error: %d\n", result);
			smpd_exit_fn(FCNAME);
			return result;
		}
		smpd_clear_process_registry();
	}

	/* This process is the root_smpd.  All sessions are child processes of this process. */
	smpd_process.id = 0;
	smpd_process.root_smpd = SMPD_TRUE;

	if (smpd_process.pszExe[0] != L'\0')
	{
		smpd_set_smpd_data(L"binary", smpd_process.pszExe);
	}

	result = MPIDU_Sock_create_set(&set);
	if (result != MPI_SUCCESS)
	{
		smpd_err_printf(L"MPIDU_Sock_create_set failed,\nsock error: %s\n", get_sock_error_string(result));
		smpd_exit_fn(FCNAME);
		return result;
	}
	smpd_process.set = set;
	smpd_dbg_printf(L"created a set for the listener: %d\n", MPIDU_Sock_get_sock_set_id(set));
	if (smpd_process.port == 0)
		print_port = SMPD_TRUE;
	result = MPIDU_Sock_listen(set, NULL, &smpd_process.port, &listener); 
	if (result != MPI_SUCCESS)
	{
		/* If another smpd is running and listening on this port, tell it to shutdown or restart? */
		smpd_err_printf(L"MPIDU_Sock_listen failed,\nsock error: %s\n", get_sock_error_string(result));
		smpd_exit_fn(FCNAME);
		return result;
	}
	smpd_dbg_printf(L"smpd listening on port %d\n", smpd_process.port);
	if (print_port)
	{
		wprintf(L"%d\n", smpd_process.port);
		fflush(stdout);
	}

	result = smpd_create_context(SMPD_CONTEXT_LISTENER, set, listener, -1, &smpd_process.listener_context);
	if (result != SMPD_SUCCESS)
	{
		smpd_err_printf(L"unable to create a context for the smpd listener.\n");
		smpd_exit_fn(FCNAME);
		return result;
	}
	result = MPIDU_Sock_set_user_ptr(listener, smpd_process.listener_context);
	if (result != MPI_SUCCESS)
	{
		smpd_err_printf(L"MPIDU_Sock_set_user_ptr failed,\nsock error: %s\n", get_sock_error_string(result));
		smpd_exit_fn(FCNAME);
		return result;
	}
	smpd_process.listener_context->state = SMPD_SMPD_LISTENING;

	if (smpd_process.root_smpd)
		smpd_insert_into_dynamic_hosts();

	result = smpd_enter_at_state(set, SMPD_SMPD_LISTENING);
	if (result != SMPD_SUCCESS)
	{
		smpd_err_printf(L"state machine failed.\n");
	}

	if (smpd_process.root_smpd)
		smpd_remove_from_dynamic_hosts();

	result = MPIDU_Sock_destroy_set(set);
	if (result != MPI_SUCCESS)
	{
		smpd_err_printf(L"unable to destroy the set, error:\n%s\n",
			get_sock_error_string(result));
	}

	/*
	result = MPIDU_Sock_finalize();
	if (result != MPI_SUCCESS)
	{
		smpd_err_printf(L"MPIDU_Sock_finalize failed,\nsock error: %s\n", get_sock_error_string(result));
	}
	*/

	smpd_exit_fn(FCNAME);
	return 0;
}

