/*   Copyright (C) 2009 Deino Software. */

#include "smpd.h"
#include <windows.h>
#include <iads.h>
#include "smpd_service.h"
#include <ntdsapi.h>
#include <dsgetdc.h>
#include <lm.h>
#define SECURITY_WIN32
#include <security.h>
#include <adshlp.h>
#include <atlbase.h>
#include <adserr.h>

static HRESULT GetGCSearch(IDirectorySearch **ppDS);
static int smpd_build_spn_list();

typedef struct smpd_host_spn_node_t
{
	wchar_t host[SMPD_MAX_NAME_LENGTH];
	wchar_t dnshost[SMPD_MAX_NAME_LENGTH];
	wchar_t spn[SMPD_MAX_NAME_LENGTH];
	struct smpd_host_spn_node_t *next;
} smpd_host_spn_node_t;
static smpd_host_spn_node_t *spn_list = NULL;

#undef FCNAME
#define FCNAME L"smpd_register_spn"
int smpd_register_spn(const wchar_t *dc, const wchar_t *dn, const wchar_t *dh)
{
	DWORD len;
	wchar_t err_msg[256];
	LPWSTR *spns;
	HANDLE ds;
	DWORD result;
	wchar_t domain_controller[SMPD_MAX_HOST_LENGTH] = L"";
	wchar_t domain_name[SMPD_MAX_HOST_LENGTH] = L"";
	wchar_t domain_host[SMPD_MAX_HOST_LENGTH] = L"";
	wchar_t host[SMPD_MAX_HOST_LENGTH] = L"";
	int really = 0;
	const wchar_t *really_env;
	PDOMAIN_CONTROLLER_INFO pInfo;

	result = DsGetDcName(NULL/*local computer*/, NULL, NULL, NULL,
		/*DS_IS_FLAT_NAME | DS_RETURN_DNS_NAME | DS_DIRECTORY_SERVICE_REQUIRED, */
		DS_DIRECTORY_SERVICE_REQUIRED | DS_KDC_REQUIRED,
		&pInfo);
	if (result == ERROR_SUCCESS)
	{
		wcscpy_s(domain_controller, SMPD_MAX_HOST_LENGTH, pInfo->DomainControllerName);
		wcscpy_s(domain_name, SMPD_MAX_HOST_LENGTH, pInfo->DomainName);
		NetApiBufferFree(pInfo);
	}

	if (dc && *dc != L'\0')
	{
		wcscpy_s(domain_controller, SMPD_MAX_HOST_LENGTH, dc);
	}
	if (dn && *dn != L'\0')
	{
		wcscpy_s(domain_name, SMPD_MAX_HOST_LENGTH, dn);
	}
	if (dh && *dh != L'\0')
	{
		wcscpy_s(domain_host, SMPD_MAX_HOST_LENGTH, dh);
	}
	if (domain_host[0] == L'\0')
	{
		smpd_get_hostname(host, SMPD_MAX_HOST_LENGTH);
		if (domain_name[0] != L'\0')
		{
			swprintf_s(domain_host, SMPD_MAX_HOST_LENGTH, L"%s\\%s", domain_name, host);
		}
		else
		{
			wcscpy_s(domain_host, SMPD_MAX_HOST_LENGTH, host);
		}
	}

	wprintf(L"DsBind(%s, %s, ...)\n", domain_controller[0] == L'\0' ? NULL : domain_controller, domain_name[0] == L'\0' ? NULL : domain_name);
	result = DsBind(
		domain_controller[0] == L'\0' ? NULL : domain_controller,
		domain_name[0] == L'\0' ? NULL : domain_name, &ds);
	if (result != ERROR_SUCCESS)
	{
		smpd_translate_win_error(result, err_msg, 256, NULL);
		smpd_err_printf(L"DsBind failed: %s\n", err_msg);
		return SMPD_FAIL;
	}

	really_env = MPIU_GetStaticEnv(L"really");
	if (really_env)
		really = 1;

#if 1
	len = 1;
	/*result = DsGetSpn(DS_SPN_SERVICE, SMPD_SERVICE_NAME, SMPD_SERVICE_NAME, 0, 0, NULL, NULL, &len, &spns);*/
	result = DsGetSpn(DS_SPN_DNS_HOST, SMPD_SERVICE_NAMEW, NULL, SMPD_LISTENER_PORT, 0, NULL, NULL, &len, &spns);
	if (result != ERROR_SUCCESS)
	{
		smpd_translate_win_error(result, err_msg, 256, NULL);
		smpd_err_printf(L"DsGetSpn failed: %s\n", err_msg);
		return SMPD_FAIL;
	}
	if (really)
	{
		wprintf(L"registering: %s\n", spns[0]);
		len = SMPD_MAX_HOST_LENGTH;
		GetComputerObjectName(NameFullyQualifiedDN, domain_host, &len);
		wprintf(L"on account: %s\n", domain_host);
		result = DsWriteAccountSpn(ds, DS_SPN_ADD_SPN_OP, domain_host, 1, (LPCWSTR*)spns);
		if (result != ERROR_SUCCESS)
		{
			DsFreeSpnArray(1, spns);
			smpd_translate_win_error(result, err_msg, 256, NULL);
			smpd_err_printf(L"DsWriteAccountSpn failed: %s\n", err_msg);
			return SMPD_FAIL;
		}
	}
	else
	{
		wprintf(L"would register '%s' on %s\n", spns[0], domain_host);
	}
	DsFreeSpnArray(1, spns);
#else
	if (really)
	{
		result = DsServerRegisterSpn(DS_SPN_ADD_SPN_OP, SMPD_SERVICE_NAME, domain_host);
		if (result != ERROR_SUCCESS)
		{
			smpd_translate_win_error(result, err_msg, 256, NULL);
			smpd_err_printf(L"DsServerRegisterSpn failed: %s\n", err_msg);
			return SMPD_FAIL;
		}
	}
	else
	{
		wprintf(L"would register '%s' on %s\n", SMPD_SERVICE_NAME, domain_host);
	}
#endif
	result = DsUnBind(&ds);
	return SMPD_SUCCESS;
}

static int smpd_build_spn_list()
{
	HRESULT hr;
	IDirectoryObject *pSCP = NULL;
	ADS_ATTR_INFO *pPropEntries = NULL;
	IDirectorySearch *pSearch = NULL;
	ADS_SEARCH_HANDLE hSearch = NULL;
	LPWSTR pszDN;				  /* distinguished name of SCP. */
	LPWSTR pszServiceDNSName;	  /* service DNS name. */
	LPWSTR pszClass;			   /* name of service class. */
	USHORT usPort;				 /* service port. */
	WCHAR pszSearchString[SMPD_MAX_NAME_LENGTH];
	wchar_t temp_str[SMPD_MAX_NAME_LENGTH];
	wchar_t temp_str2[SMPD_MAX_NAME_LENGTH];
	smpd_host_spn_node_t *iter;
	double t1, t2;
	static int initialized = 0;

	if (initialized)
	{
		return SMPD_SUCCESS;
	}
	initialized = 1;

	t1 = PMPI_Wtime();

	CoInitialize(NULL);

	/* Get an IDirectorySearch pointer for the Global Catalog.  */
	hr = GetGCSearch(&pSearch);
	if (FAILED(hr) || pSearch == NULL) 
	{
		smpd_err_printf(L"GetGC failed 0x%x\n", hr);
		goto Cleanup;
	}

	/* Set up a deep search.
	  Thousands of objects are not expected in this example, therefore
	  query for 1000 rows per page.*/
	ADS_SEARCHPREF_INFO SearchPref[2];
	DWORD dwPref = sizeof(SearchPref)/sizeof(ADS_SEARCHPREF_INFO);
	SearchPref[0].dwSearchPref =	ADS_SEARCHPREF_SEARCH_SCOPE;
	SearchPref[0].vValue.dwType =   ADSTYPE_INTEGER;
	SearchPref[0].vValue.Integer =  ADS_SCOPE_SUBTREE;

	SearchPref[1].dwSearchPref =	ADS_SEARCHPREF_PAGESIZE;
	SearchPref[1].vValue.dwType =   ADSTYPE_INTEGER;
	SearchPref[1].vValue.Integer =  1000;

	hr = pSearch->SetSearchPreference(SearchPref, dwPref);
	if (FAILED(hr))
	{
		smpd_err_printf(L"Failed to set search prefs: hr:0x%x\n", hr);
		goto Cleanup;
	}

	/* Execute the search. From the GC get the distinguished name 
	  of the SCP. Use the DN to bind to the SCP and get the other 
	  properties. */
	LPWSTR rgszDN[] = {L"distinguishedName"};

	/* Search for a match of the product GUID. */
	swprintf_s(pszSearchString, SMPD_MAX_NAME_LENGTH, L"keywords=%s", SMPD_SERVICE_VENDOR_GUIDW);
	hr = pSearch->ExecuteSearch(pszSearchString, rgszDN, 1, &hSearch);
	/*hr = pSearch->ExecuteSearch(L"keywords=f5e3c95b-7a1f-49c1-b518-25cbf0796936", rgszDN, 1, &hSearch);*/
	if (FAILED(hr))
	{
		smpd_err_printf(L"ExecuteSearch failed: hr:0x%x\n", hr);
		goto Cleanup;
	}

	/* Loop through the results. Each row should be an instance of the 
	  service identified by the product GUID.
	  Add logic to select from multiple service instances. */
	while (SUCCEEDED(hr = pSearch->GetNextRow(hSearch)))
	{
		if (hr == S_ADS_NOMORE_ROWS)
		{
			DWORD dwError = ERROR_SUCCESS;
			WCHAR szError[512];
			WCHAR szProvider[512];

			ADsGetLastError(&dwError, szError, 512, szProvider, 512);
			if (ERROR_MORE_DATA == dwError)
			{
				continue;
			}
			goto Cleanup;
		}

		ADS_SEARCH_COLUMN Col;

		hr = pSearch->GetColumn(hSearch, L"distinguishedName", &Col);
		pszDN = AllocADsStr(Col.pADsValues->CaseIgnoreString);
		pSearch->FreeColumn(&Col);

		/* Bind to the DN to get the other properties. */
		LPWSTR lpszLDAPPrefix = L"LDAP://";
		DWORD dwSCPPathLength = (DWORD)(wcslen(lpszLDAPPrefix) + wcslen(pszDN) + 1);
		LPWSTR pwszSCPPath = (LPWSTR)malloc(sizeof(WCHAR) * dwSCPPathLength);
		if (pwszSCPPath)
		{
			wcscpy_s(pwszSCPPath, dwSCPPathLength, lpszLDAPPrefix);
			wcscat_s(pwszSCPPath, dwSCPPathLength, pszDN);
		}	   
		else
		{
			smpd_err_printf(L"Failed to allocate a buffer\n");
			goto Cleanup;
		}			   
		/*wprintf(L"pszDN = %s\n", pszDN);*/
		/*FreeADsStr(pszDN);*/

		hr = ADsGetObject(pwszSCPPath, IID_IDirectoryObject, (void**)&pSCP);
		free(pwszSCPPath);

		if (SUCCEEDED(hr)) 
		{
			/* Properties to retrieve from the SCP object. */
			LPWSTR rgszAttribs[]=
			{
				{L"serviceClassName"},
				{L"serviceDNSName"},
				/*{L"serviceDNSNameType"},*/
				{L"serviceBindingInformation"}
			};

			DWORD dwAttrs = sizeof(rgszAttribs)/sizeof(LPWSTR);
			DWORD dwNumAttrGot;
			hr = pSCP->GetObjectAttributes(rgszAttribs, dwAttrs, &pPropEntries, &dwNumAttrGot);
			if (FAILED(hr)) 
			{
				smpd_err_printf(L"GetObjectAttributes Failed. hr:0x%x\n", hr);
				goto Cleanup;
			}

			pszServiceDNSName = NULL;
			pszClass = NULL;
			iter = (smpd_host_spn_node_t*)malloc(sizeof(smpd_host_spn_node_t));
			if (iter == NULL)
			{
				smpd_err_printf(L"Unable to allocate memory to store an SPN entry.\n");
				goto Cleanup;
			}
			iter->next = NULL;
			iter->host[0] = L'\0';
			iter->spn[0] = L'\0';
			iter->dnshost[0] = L'\0';

			/* Loop through the entries returned by GetObjectAttributes 
			and save the values in the appropriate buffers.  */
			for (int i = 0; i < (LONG)dwAttrs; i++) 
			{
				if ((wcscmp(L"serviceDNSName", pPropEntries[i].pszAttrName) == 0) &&
					(pPropEntries[i].dwADsType == ADSTYPE_CASE_IGNORE_STRING)) 
				{
					pszServiceDNSName = AllocADsStr(pPropEntries[i].pADsValues->CaseIgnoreString);
					/*wprintf(L"pszServiceDNSName = %s\n", pszServiceDNSName);*/
				}

				/*
				if ((wcscmp(L"serviceDNSNameType", pPropEntries[i].pszAttrName) == 0) &&
				(pPropEntries[i].dwADsType == ADSTYPE_CASE_IGNORE_STRING)) 
				{
				pszServiceDNSNameType = AllocADsStr(pPropEntries[i].pADsValues->CaseIgnoreString);
				wprintf(L"pszServiceDNSNameType = %s\n", pszServiceDNSNameType);
				}
				*/

				if ((wcscmp(L"serviceClassName", pPropEntries[i].pszAttrName) == 0) &&
					(pPropEntries[i].dwADsType == ADSTYPE_CASE_IGNORE_STRING)) 
				{
					pszClass = AllocADsStr(pPropEntries[i].pADsValues->CaseIgnoreString);
					/*wprintf(L"pszClass = %s\n", pszClass);*/
				}

				if ((wcscmp(L"serviceBindingInformation", pPropEntries[i].pszAttrName) == 0) &&
					(pPropEntries[i].dwADsType == ADSTYPE_CASE_IGNORE_STRING)) 
				{
					usPort=(USHORT)_wtoi(pPropEntries[i].pADsValues->CaseIgnoreString);
					/*wprintf(L"usPort = %d\n", usPort);*/
				}
			}

			/*wcstombs(iter->dnshost, pszServiceDNSName, SMPD_MAX_NAME_LENGTH);*/
			MPIU_Wcsncpy(iter->dnshost, pszServiceDNSName, SMPD_MAX_NAME_LENGTH);
			/*wcstombs(temp_str, pszClass, SMPD_MAX_NAME_LENGTH);*/
			MPIU_Wcsncpy(temp_str, pszClass, SMPD_MAX_NAME_LENGTH);
			/*MPIU_Snprintf(iter->spn, SMPD_MAX_NAME_LENGTH, _TRUNCATE, L"%s/%s:%d", temp_str, iter->dnshost, usPort);*/
			/*wcstombs(temp_str2, pszDN, SMPD_MAX_NAME_LENGTH);*/
			MPIU_Wcsncpy(temp_str2, pszDN, SMPD_MAX_NAME_LENGTH);
			MPIU_Snprintf(iter->spn, SMPD_MAX_NAME_LENGTH, _TRUNCATE, L"%s/%s/%s", temp_str, iter->dnshost, temp_str2);
			MPIU_Wcsncpy(iter->host, iter->dnshost, SMPD_MAX_NAME_LENGTH);
			wchar_t *tok;
			wcstok_s(iter->host, L".", &tok);
			iter->next = spn_list;
			spn_list = iter;
			if (pszServiceDNSName != NULL)
			{
				FreeADsStr(pszServiceDNSName);
			}
			if (pszClass != NULL)
			{
				FreeADsStr(pszClass);
			}
		}
		FreeADsStr(pszDN);
	}

Cleanup:
	/*
	iter = spn_list;
	while (iter != NULL)
	{
		wprintf(L"host   : %s\n", iter->host);
		wprintf(L"dnshost: %s\n", iter->dnshost);
		wprintf(L"spn	: %s\n", iter->spn);
		iter = iter->next;
	}
	fflush(stdout);
	*/
	if (pSCP)
	{
		pSCP->Release();
		pSCP = NULL;
	}

	if (pPropEntries)
	{
		FreeADsMem(pPropEntries);
		pPropEntries = NULL;
	}

	if (pSearch)
	{
		if (hSearch)
		{
			pSearch->CloseSearchHandle(hSearch);
			hSearch = NULL;
		}

		pSearch->Release();
		pSearch = NULL;
	}
	CoUninitialize();

	t2 = PMPI_Wtime();
	smpd_dbg_printf(L"build_spn_list took %0.6f seconds\n", t2-t1);

	return SMPD_SUCCESS;
}

static HRESULT GetGCSearch(IDirectorySearch **ppDS)
{
	HRESULT hr;
	IEnumVARIANT *pEnum = NULL;
	IADsContainer *pCont = NULL;
	IDispatch *pDisp = NULL;
	VARIANT var;
	ULONG lFetch;

	*ppDS = NULL;

	/* Bind to the GC: namespace container object. The true GC DN 
	  is a single immediate child of the GC: namespace, which must 
	  be obtained using enumeration. */
	hr = ADsOpenObject(L"GC:", NULL, NULL, ADS_SECURE_AUTHENTICATION, /* Use Secure Authentication. */
		IID_IADsContainer,
		(void**)&pCont);
	if (FAILED(hr)) 
	{
		smpd_err_printf(L"ADsOpenObject failed: 0x%x\n", hr);
		goto cleanup;
	} 

	/* Get an enumeration interface for the GC container.  */
	hr = ADsBuildEnumerator(pCont, &pEnum);
	if (FAILED(hr)) 
	{
		smpd_err_printf(L"ADsBuildEnumerator failed: 0x%x\n", hr);
		goto cleanup;
	} 

	/* Now enumerate. There is only one child of the GC: object. */
	hr = ADsEnumerateNext(pEnum, 1, &var, &lFetch);
	if (FAILED(hr)) 
	{
		smpd_err_printf(L"ADsEnumerateNext failed: 0x%x\n", hr);
		goto cleanup;
	} 

	if ((hr == S_OK) && (lFetch == 1))
	{
		pDisp = V_DISPATCH(&var);
		hr = pDisp->QueryInterface(IID_IDirectorySearch, (void**)ppDS); 
	}

cleanup:
	if (pEnum)
	{
		ADsFreeEnumerator(pEnum);
		pEnum = NULL;
	}

	if (pCont)
	{
		pCont->Release();
		pCont = NULL;
	}

	if (pDisp)
	{
		pDisp->Release();
		pDisp = NULL;
	}

	return hr;
}


#undef FCNAME
#define FCNAME L"smpd_lookup_spn"
int smpd_lookup_spn(wchar_t *target, int length, const wchar_t * host, int port)
{
	int result;
	wchar_t err_msg[256];
	ULONG len = length/*SMPD_MAX_NAME_LENGTH*/;
	const wchar_t *env;
	smpd_host_spn_node_t *iter;

	env = MPIU_GetStaticEnv(L"DeinoMPI_SPN");
	if (env)
	{
		MPIU_Wcsncpy(target, env, SMPD_MAX_NAME_LENGTH);
		return SMPD_SUCCESS;
	}

	smpd_build_spn_list();
	iter = spn_list;
	while (iter != NULL)
	{
		if (_wcsicmp(iter->host, host) == 0)
		{
			MPIU_Wcsncpy(target, iter->spn, SMPD_MAX_NAME_LENGTH);
			return SMPD_SUCCESS;
		}
		if (_wcsicmp(iter->dnshost, host) == 0)
		{
			MPIU_Wcsncpy(target, iter->spn, SMPD_MAX_NAME_LENGTH);
			return SMPD_SUCCESS;
		}
		iter = iter->next;
	}

	result = DsMakeSpn(SMPD_SERVICE_NAMEW, NULL, host, (USHORT)port, NULL, &len, target);
	if (result != ERROR_SUCCESS)
	{
		smpd_translate_win_error(result, err_msg, 255, NULL);
		smpd_err_printf(L"DsMakeSpn(%s, %s, %d) failed: %s\n", SMPD_SERVICE_NAME, host, port, err_msg);
		return SMPD_FAIL;
	}


	/*result = DsMakeSpn(SMPD_SERVICE_NAME, SMPD_SERVICE_NAME, NULL, 0, NULL, &len, target);*/
	/*
	wchar_t **spns;
	result = DsGetSpn(DS_SPN_DNS_HOST, SMPD_SERVICE_NAME, NULL, port, 1, &host, NULL, &len, &spns);
	if (result != ERROR_SUCCESS)
	{
		smpd_translate_win_error(result, err_msg, 255, NULL);
		smpd_err_printf(L"DsGetSpn failed: %s\n", err_msg);
		return SMPD_FAIL;
	}
	MPIU_Wcsncpy(target, spns[0], SMPD_MAX_NAME_LENGTH);
	DsFreeSpnArray(1, spns);
	*/
	/*MPIU_Snprintf(target, SMPD_MAX_NAME_LENGTH, _TRUNCATE, L"%s/%s:%d", SMPD_SERVICE_NAME, host, port);*/
	return SMPD_SUCCESS;
}
