/*   Copyright (C) 2009 Deino Software. */

#include <stdio.h>
#include <stdlib.h>
#include "mpiexec.h"
#include "smpd.h"

#undef FCNAME
#define FCNAME L"smpd_get_cmd_state_string"
const wchar_t * smpd_get_cmd_state_string(smpd_command_state_t state)
{
	const wchar_t *result;
	static wchar_t unknown_str[100];

	smpd_enter_fn(FCNAME);

	switch (state)
	{
	case SMPD_CMD_INVALID:
		result = L"SMPD_CMD_INVALID";
		break;
	case SMPD_CMD_READING_HDR:
		result = L"SMPD_CMD_READING_HDR";
		break;
	case SMPD_CMD_READING_CMD:
		result = L"SMPD_CMD_READING_CMD";
		break;
	case SMPD_CMD_WRITING_CMD:
		result = L"SMPD_CMD_WRITING_CMD";
		break;
	case SMPD_CMD_READY:
		result = L"SMPD_CMD_READY";
		break;
	case SMPD_CMD_HANDLED:
		result = L"SMPD_CMD_HANDLED";
		break;
	default:
		swprintf_s(unknown_str, 100, L"unknown state %d", state);
		result = unknown_str;
		break;
	}
	smpd_exit_fn(FCNAME);
	return result;
}

#undef FCNAME
#define FCNAME L"smpd_command_to_string"
SMPD_BOOL smpd_command_to_string(wchar_t **str_pptr, int *len_ptr, int indent, smpd_command_t *cmd_ptr)
{
	wchar_t indent_str[SMPD_MAX_TO_STRING_INDENT+1];

	smpd_enter_fn(FCNAME);

	if (*len_ptr < 1)
	{
		smpd_exit_fn(FCNAME);
		return SMPD_FALSE;
	}

	if (indent > SMPD_MAX_TO_STRING_INDENT)
		indent = SMPD_MAX_TO_STRING_INDENT;

	memset(indent_str, MPIU_STR_SEPAR_CHAR, indent);
	indent_str[indent] = L'\0';

	smpd_snprintf_update(str_pptr, len_ptr, L"%sstate: %s\n", indent_str, smpd_get_cmd_state_string(cmd_ptr->state));
	if (*len_ptr < 1) { smpd_exit_fn(FCNAME); return SMPD_FALSE; }
	smpd_snprintf_update(str_pptr, len_ptr, L"%scmd_str: %s\n", indent_str, cmd_ptr->cmd_str);
	if (*len_ptr < 1) { smpd_exit_fn(FCNAME); return SMPD_FALSE; }
	smpd_snprintf_update(str_pptr, len_ptr, L"%ssrc: %d\n", indent_str, cmd_ptr->src);
	if (*len_ptr < 1) { smpd_exit_fn(FCNAME); return SMPD_FALSE; }
	smpd_snprintf_update(str_pptr, len_ptr, L"%sdest: %d\n", indent_str, cmd_ptr->dest);
	if (*len_ptr < 1) { smpd_exit_fn(FCNAME); return SMPD_FALSE; }
	smpd_snprintf_update(str_pptr, len_ptr, L"%stag: %d\n", indent_str, cmd_ptr->tag);
	if (*len_ptr < 1) { smpd_exit_fn(FCNAME); return SMPD_FALSE; }
	smpd_snprintf_update(str_pptr, len_ptr, L"%swait: %s\n", indent_str, cmd_ptr->wait ? L"TRUE" : L"FALSE");
	if (*len_ptr < 1) { smpd_exit_fn(FCNAME); return SMPD_FALSE; }
	smpd_snprintf_update(str_pptr, len_ptr, L"%scmd_hdr_str: %s\n", indent_str, cmd_ptr->cmd_hdr_str);
	if (*len_ptr < 1) { smpd_exit_fn(FCNAME); return SMPD_FALSE; }
	smpd_snprintf_update(str_pptr, len_ptr, L"%slength: %d\n", indent_str, cmd_ptr->length);
	if (*len_ptr < 1) { smpd_exit_fn(FCNAME); return SMPD_FALSE; }
	smpd_snprintf_update(str_pptr, len_ptr, L"%scmd: %s\n", indent_str, cmd_ptr->cmd);
	if (*len_ptr < 1) { smpd_exit_fn(FCNAME); return SMPD_FALSE; }
	smpd_snprintf_update(str_pptr, len_ptr, L"%sfreed: %d\n", indent_str, cmd_ptr->freed);
	if (*len_ptr < 1) { smpd_exit_fn(FCNAME); return SMPD_FALSE; }
	smpd_snprintf_update(str_pptr, len_ptr, L"%siov[0].buf: %p\n", indent_str, cmd_ptr->iov[0].MPID_IOV_BUF);
	if (*len_ptr < 1) { smpd_exit_fn(FCNAME); return SMPD_FALSE; }
	smpd_snprintf_update(str_pptr, len_ptr, L"%siov[0].len: %d\n", indent_str, cmd_ptr->iov[0].MPID_IOV_LEN);
	if (*len_ptr < 1) { smpd_exit_fn(FCNAME); return SMPD_FALSE; }
	smpd_snprintf_update(str_pptr, len_ptr, L"%siov[1].buf: %p\n", indent_str, cmd_ptr->iov[1].MPID_IOV_BUF);
	if (*len_ptr < 1) { smpd_exit_fn(FCNAME); return SMPD_FALSE; }
	smpd_snprintf_update(str_pptr, len_ptr, L"%siov[1].len: %d\n", indent_str, cmd_ptr->iov[1].MPID_IOV_LEN);
	if (*len_ptr < 1) { smpd_exit_fn(FCNAME); return SMPD_FALSE; }
	smpd_snprintf_update(str_pptr, len_ptr, L"%sstdin_read_offset: %d\n", indent_str, cmd_ptr->stdin_read_offset);
	if (*len_ptr < 1) { smpd_exit_fn(FCNAME); return SMPD_FALSE; }
	smpd_snprintf_update(str_pptr, len_ptr, L"%snext: %p\n", indent_str, cmd_ptr->next);
	if (*len_ptr < 1) { smpd_exit_fn(FCNAME); return SMPD_FALSE; } /* this misses the case of an exact fit */

	smpd_exit_fn(FCNAME);
	return SMPD_TRUE;
}

#undef FCNAME
#define FCNAME L"smpd_command_destination"
int smpd_command_destination(int dest, smpd_context_t **dest_context)
{
	int src, level_bit, sub_tree_mask;

	smpd_enter_fn(FCNAME);

	/* get the source */
	src = smpd_process.id;

	/*smpd_dbg_printf(L"determining destination context for %d -> %d\n", src, dest);*/

	/* determine the route and return the context */
	if (src == dest)
	{
		*dest_context = NULL;
		smpd_dbg_printf(L"%d -> %d : returning NULL context\n", src, dest);
		smpd_exit_fn(FCNAME);
		return SMPD_SUCCESS;
	}
	if (src == 1 && dest == 0 && smpd_process.map0to1)
	{
		*dest_context = NULL;
		smpd_dbg_printf(L"%d -> %d : returning NULL context\n", src, dest);
		smpd_exit_fn(FCNAME);
		return SMPD_SUCCESS;
	}

	if (src == 0)
	{
		/* this assumes that the root uses the left context for it's only child. */
		if (smpd_process.left_context == NULL)
		{
			smpd_exit_fn(FCNAME);
			return SMPD_FAIL;
		}
		*dest_context = smpd_process.left_context;
		smpd_dbg_printf(L"%d -> %d : returning left_context\n", src, dest);
		smpd_exit_fn(FCNAME);
		return SMPD_SUCCESS;
	}

	if (dest < src)
	{
		if (smpd_process.parent_context == NULL)
		{
			smpd_exit_fn(FCNAME);
			return SMPD_FAIL;
		}
		*dest_context = smpd_process.parent_context;
		smpd_dbg_printf(L"%d -> %d : returning parent_context: %d < %d\n", src, dest, dest, src);
		smpd_exit_fn(FCNAME);
		return SMPD_SUCCESS;
	}

	level_bit = 0x1 << smpd_process.level;
	sub_tree_mask = (level_bit << 1) - 1;

	if (( src ^ level_bit ) == ( dest & sub_tree_mask ))
	{
		if (smpd_process.left_context == NULL)
		{
			smpd_exit_fn(FCNAME);
			return SMPD_FAIL;
		}
		*dest_context = smpd_process.left_context;
		smpd_dbg_printf(L"%d -> %d : returning left_context\n", src, dest);
		smpd_exit_fn(FCNAME);
		return SMPD_SUCCESS;
	}

	if ( src == ( dest & sub_tree_mask ) )
	{
		if (smpd_process.right_context == NULL)
		{
			smpd_exit_fn(FCNAME);
			return SMPD_FAIL;
		}
		*dest_context = smpd_process.right_context;
		smpd_dbg_printf(L"%d -> %d : returning right_context\n", src, dest);
		smpd_exit_fn(FCNAME);
		return SMPD_SUCCESS;
	}

	if (smpd_process.parent_context == NULL)
	{
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	*dest_context = smpd_process.parent_context;
	smpd_dbg_printf(L"%d -> %d : returning parent_context: fall through\n", src, dest);
	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}

#undef FCNAME
#define FCNAME L"smpd_init_command"
int smpd_init_command(smpd_command_t *cmd)
{
	smpd_enter_fn(FCNAME);

	if (cmd == NULL)
	{
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}

	cmd->cmd_hdr_str[0] = L'\0';
	cmd->cmd_str[0] = L'\0';
	cmd->cmd[0] = L'\0';
	cmd->dest = -1;
	cmd->src = -1;
	cmd->tag = -1;
	cmd->next = NULL;
	cmd->length = 0;
	cmd->wait = SMPD_FALSE;
	cmd->state = SMPD_CMD_INVALID;
	cmd->stdin_read_offset = 0;
	cmd->freed = 0;
	cmd->context = NULL;

	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}

#undef FCNAME
#define FCNAME L"smpd_parse_command"
int smpd_parse_command(smpd_command_t *cmd_ptr)
{
	smpd_enter_fn(FCNAME);

	/* get the source */
	if (MPIU_Str_get_int_arg(cmd_ptr->cmd, L"src", &cmd_ptr->src) != MPIU_STR_SUCCESS)
	{
		smpd_err_printf(L"no src flag in the command.\n");
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	if (cmd_ptr->src < 0)
	{
		smpd_err_printf(L"invalid command src: %d\n", cmd_ptr->src);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}

	/* get the destination */
	if (MPIU_Str_get_int_arg(cmd_ptr->cmd, L"dest", &cmd_ptr->dest) != MPIU_STR_SUCCESS)
	{
		smpd_err_printf(L"no dest flag in the command.\n");
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	if (cmd_ptr->dest < 0)
	{
		smpd_err_printf(L"invalid command dest: %d\n", cmd_ptr->dest);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}

	/* get the command string */
	if (MPIU_Str_get_string_arg(cmd_ptr->cmd, L"cmd", cmd_ptr->cmd_str, SMPD_MAX_CMD_STR_LENGTH) != MPIU_STR_SUCCESS)
	{
		smpd_err_printf(L"no cmd string in the command.\n");
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}

	/* get the tag */
	/* A command does not have to have a tag so don't check for failure */
	MPIU_Str_get_int_arg(cmd_ptr->cmd, L"tag", &cmd_ptr->tag);

	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}

#undef FCNAME
#define FCNAME L"smpd_create_command"
int smpd_create_command(wchar_t *cmd, int src, int dest, int want_reply, smpd_command_t **cmd_pptr)
{
	smpd_command_t *cmd_ptr;
	wchar_t *str;
	int len;
	int result;

	smpd_enter_fn(FCNAME);

	cmd_ptr = (smpd_command_t*)malloc(sizeof(smpd_command_t));
	if (cmd_ptr == NULL)
	{
		smpd_err_printf(L"unable to allocate memory for a command.\n");
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	memset(cmd_ptr, 0, sizeof(smpd_command_t));
	smpd_init_command(cmd_ptr);
	cmd_ptr->src = src;
	cmd_ptr->dest = dest;
	cmd_ptr->tag = smpd_process.cur_tag++;

	if (wcslen(cmd) >= SMPD_MAX_CMD_STR_LENGTH)
	{
		smpd_err_printf(L"command string too long: %s\n", cmd);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	wcscpy_s(cmd_ptr->cmd_str, SMPD_MAX_CMD_STR_LENGTH, cmd);

	str = cmd_ptr->cmd;
	len = SMPD_MAX_CMD_LENGTH;
	result = MPIU_Str_add_string_arg(&str, &len, L"cmd", cmd);
	if (result != MPIU_STR_SUCCESS)
	{
		smpd_err_printf(L"unable to create the command.\n");
		smpd_free_command(cmd_ptr);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	result = MPIU_Str_add_int_arg(&str, &len, L"src", src);
	if (result != MPIU_STR_SUCCESS)
	{
		smpd_err_printf(L"unable to add the src to the command.\n");
		smpd_free_command(cmd_ptr);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	result = MPIU_Str_add_int_arg(&str, &len, L"dest", dest);
	if (result != MPIU_STR_SUCCESS)
	{
		smpd_err_printf(L"unable to add the dest to the command.\n");
		smpd_free_command(cmd_ptr);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	result = MPIU_Str_add_int_arg(&str, &len, L"tag", cmd_ptr->tag);
	if (result != MPIU_STR_SUCCESS)
	{
		smpd_err_printf(L"unable to add the tag to the command.\n");
		smpd_free_command(cmd_ptr);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	if (want_reply)
	{
		cmd_ptr->wait = SMPD_TRUE;
	}

	*cmd_pptr = cmd_ptr;
	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}

#undef FCNAME
#define FCNAME L"smpd_create_command_copy"
int smpd_create_command_copy(smpd_command_t *src_ptr, smpd_command_t **cmd_pptr)
{
	smpd_command_t *cmd_ptr;

	smpd_enter_fn(FCNAME);

	cmd_ptr = (smpd_command_t*)malloc(sizeof(smpd_command_t));
	if (cmd_ptr == NULL)
	{
		smpd_err_printf(L"unable to allocate memory for a command.\n");
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	
	*cmd_ptr = *src_ptr;
	*cmd_pptr = cmd_ptr;

	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}

#undef FCNAME
#define FCNAME L"smpd_free_command"
int smpd_free_command(smpd_command_t *cmd_ptr)
{
	smpd_enter_fn(FCNAME);
	if (cmd_ptr)
	{
		/* this check isn't full-proof because random data might match SMPD_FREE_COOKIE */
		if (cmd_ptr->freed == SMPD_FREE_COOKIE)
		{
			smpd_err_printf(L"attempt to free a command more than once.\n");
			smpd_exit_fn(FCNAME);
			return SMPD_FAIL;
		}
		/* erase the contents to help track down use of freed structures */
		smpd_init_command(cmd_ptr);
		cmd_ptr->freed = SMPD_FREE_COOKIE;
		free(cmd_ptr); /* unfortunately, free probably erases the data making the cookie check ineffective */
	}
	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}

#undef FCNAME
#define FCNAME L"smpd_create_context"
int smpd_create_context(smpd_context_type_t type, MPIDU_Sock_set_t set, MPIDU_Sock_t sock, int id, smpd_context_t **context_pptr)
{
	int result;
	smpd_context_t *context;
	
	smpd_enter_fn(FCNAME);
	context = (smpd_context_t*)malloc(sizeof(smpd_context_t));
	if (context == NULL)
	{
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	memset(context, 0, sizeof(smpd_context_t));
	result = smpd_init_context(context, type, set, sock, id);
	if (result != SMPD_SUCCESS)
	{
		*context_pptr = NULL;
		free(context);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	
	/* add the context to the global list */
	context->next = smpd_process.context_list;
	smpd_process.context_list = context;

	*context_pptr = context;
	smpd_exit_fn(FCNAME);
	return result;
}

/*#define DEBUG_SMPD_FREE_CONTEXT*/

#ifdef DEBUG_SMPD_FREE_CONTEXT
typedef struct cfree_t
{
	smpd_context_t *context;
	struct cfree_t *next;
} cfree_t;
static cfree_t *free_list = NULL;
#endif

#undef FCNAME
#define FCNAME L"smpd_free_context"
int smpd_free_context(smpd_context_t *context)
{
	SMPD_BOOL found = SMPD_FALSE;
	smpd_context_t *iter, *trailer;

	smpd_enter_fn(FCNAME);
	if (context)
	{
#ifdef DEBUG_SMPD_FREE_CONTEXT
		/* check debugging free list */
		cfree_t *citer = free_list;
		while (citer)
		{
			if (citer->context == context)
			{
				smpd_err_printf(L"%s context freed twice.\n", smpd_get_context_str(context));
			}
			citer = citer->next;
		}
#endif

		/* remove the context from the global list */
		iter = trailer = smpd_process.context_list;
		while (iter)
		{
			if (iter == context)
			{
				if (iter == smpd_process.context_list)
					smpd_process.context_list = smpd_process.context_list->next;
				else
					trailer->next = iter->next;
				found = SMPD_TRUE;
				break;
			}
			if (trailer != iter)
				trailer = trailer->next;
			iter = iter->next;
		}

		if (!found)
		{
#ifdef DEBUG_SMPD_FREE_CONTEXT
			smpd_dbg_printf(L"freeing a %s context not in the global list - this should be impossible.\n",
				smpd_get_context_str(context));
#else
			smpd_dbg_printf(L"freeing a context not in the global list - this should be impossible.\n");
			smpd_exit_fn(FCNAME);
			return SMPD_SUCCESS;
#endif
		}

		smpd_dbg_printf(L"freeing %s context.\n", smpd_get_context_str(context));

		/* this check isn't full-proof because random data might match SMPD_CONTEXT_FREED */
		/* This also doesn't work because free clobbers the memory instead of leaving it alone */
		if (context->type == SMPD_CONTEXT_FREED)
		{
			smpd_err_printf(L"attempt to free context more than once.\n");
			smpd_exit_fn(FCNAME);
			return SMPD_FAIL;
		}

		/* remove any references to this context in the process structure */
		if (context->process)
		{
			switch (context->type)
			{
			case SMPD_CONTEXT_STDIN:
				context->process->in = NULL;
				break;
			case SMPD_CONTEXT_STDOUT:
			case SMPD_CONTEXT_STDOUT_RSH:
				context->process->out = NULL;
				break;
			case SMPD_CONTEXT_STDERR:
			case SMPD_CONTEXT_STDERR_RSH:
				context->process->err = NULL;
				break;
			case SMPD_CONTEXT_PMI:
				context->process->pmi = NULL;
				break;
			}
		}

		/* erase the contents to help track down use of freed structures */
		/* This doesn't work because free clobbers the memory instead of leaving it alone */
		memset(context, 0, sizeof(smpd_context_t));
		smpd_init_context(context, SMPD_CONTEXT_FREED, MPIDU_SOCK_INVALID_SET, MPIDU_SOCK_INVALID_SOCK, -1);

#ifdef DEBUG_SMPD_FREE_CONTEXT
		/* add to debugging free list */
		citer = (cfree_t*)malloc(sizeof(cfree_t));
		citer->context = context;
		citer->next = free_list;
		free_list = citer;
#else
		free(context);
#endif
	}
	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}

#undef FCNAME
#define FCNAME L"smpd_add_command_arg"
int smpd_add_command_arg(smpd_command_t *cmd_ptr, wchar_t *param, wchar_t *value)
{
	wchar_t *str;
	int len;
	int result;
	int cmd_length;

	smpd_enter_fn(FCNAME);

	cmd_length = (int)wcslen(cmd_ptr->cmd);
	if (cmd_length > SMPD_MAX_CMD_LENGTH)
	{
		smpd_err_printf(L"invalid cmd string length: %d\n", cmd_length);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}

	len = (int)(SMPD_MAX_CMD_LENGTH - cmd_length);
	str = &cmd_ptr->cmd[cmd_length];

	/* make sure there is a space after the last parameter in the command */
	if (cmd_length > 0)
	{
		if (cmd_ptr->cmd[cmd_length-1] != MPIU_STR_SEPAR_CHAR)
		{
			if (len < 2)
			{
				smpd_err_printf(L"unable to add the command parameter: %s=%s\n", param, value);
				smpd_exit_fn(FCNAME);
				return SMPD_FAIL;
			}
			cmd_ptr->cmd[cmd_length] = MPIU_STR_SEPAR_CHAR;
			len--;
			str++;
		}
	}

	result = MPIU_Str_add_string_arg(&str, &len, param, value);
	if (result != MPIU_STR_SUCCESS)
	{
		smpd_err_printf(L"unable to add the command parameter: %s=%s\n", param, value);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}

#undef FCNAME
#define FCNAME L"smpd_add_command_int_arg"
int smpd_add_command_int_arg(smpd_command_t *cmd_ptr, wchar_t *param, int value)
{
	wchar_t *str;
	int len;
	int result;
	int cmd_length;

	smpd_enter_fn(FCNAME);

	cmd_length = (int)wcslen(cmd_ptr->cmd);
	if (cmd_length > SMPD_MAX_CMD_LENGTH)
	{
		smpd_err_printf(L"invalid cmd string length: %d\n", cmd_length);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}

	len = (int)(SMPD_MAX_CMD_LENGTH - cmd_length);
	str = &cmd_ptr->cmd[cmd_length];

	/* make sure there is a space after the last parameter in the command */
	if (cmd_length > 0)
	{
		if (cmd_ptr->cmd[cmd_length-1] != MPIU_STR_SEPAR_CHAR)
		{
			if (len < 2)
			{
				smpd_err_printf(L"unable to add the command parameter: %s=%d\n", param, value);
				smpd_exit_fn(FCNAME);
				return SMPD_FAIL;
			}
			cmd_ptr->cmd[cmd_length] = MPIU_STR_SEPAR_CHAR;
			len--;
			str++;
		}
	}

	result = MPIU_Str_add_int_arg(&str, &len, param, value);
	if (result != MPIU_STR_SUCCESS)
	{
		smpd_err_printf(L"unable to add the command parameter: %s=%d\n", param, value);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}

#undef FCNAME
#define FCNAME L"smpd_add_command_binary_arg"
int smpd_add_command_binary_arg(smpd_command_t *cmd_ptr, wchar_t *param, void *buffer, int length)
{
	wchar_t *str;
	int len;
	int result;
	int cmd_length;
	int saved_length;

	smpd_enter_fn(FCNAME);

	cmd_length = (int)wcslen(cmd_ptr->cmd);
	if (cmd_length > SMPD_MAX_CMD_LENGTH)
	{
		smpd_err_printf(L"invalid cmd string length: %d\n", cmd_length);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}

	len = (int)(SMPD_MAX_CMD_LENGTH - cmd_length);
	str = &cmd_ptr->cmd[cmd_length];

	/* make sure there is a space after the last parameter in the command */
	if (cmd_length > 0)
	{
		if (cmd_ptr->cmd[cmd_length-1] != MPIU_STR_SEPAR_CHAR)
		{
			if (len < 2)
			{
				smpd_err_printf(L"unable to add the command parameter: %s=%d byte buffer\n", param, length);
				smpd_exit_fn(FCNAME);
				return SMPD_FAIL;
			}
			cmd_ptr->cmd[cmd_length] = MPIU_STR_SEPAR_CHAR;
			len--;
			str++;
		}
	}

	saved_length = len;
	result = MPIU_Str_add_binary_arg(&str, &len, param, buffer, length);
	if (result != MPIU_STR_SUCCESS)
	{
		smpd_err_printf(L"unable to add the command parameter: %s=%d byte buffer won't fit in %d character length string\n", param, length, saved_length);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}

#undef FCNAME
#define FCNAME L"smpd_forward_command"
int smpd_forward_command(smpd_context_t *src, smpd_context_t *dest)
{
	int result;
	smpd_command_t *cmd;

	smpd_enter_fn(FCNAME);

	result = smpd_create_command_copy(&src->read_cmd, &cmd);
	if (result != SMPD_SUCCESS)
	{
		smpd_err_printf(L"unable to create a copy of the command to forward.\n");
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}

	smpd_dbg_printf(L"posting write of forwarded command: \"%s\"\n", cmd->cmd);
	result = smpd_post_write_command(dest, cmd);
	if (result != SMPD_SUCCESS)
	{
		smpd_err_printf(L"unable to post a write of a forwarded command.\n");
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}

#undef FCNAME
#define FCNAME L"smpd_post_read_command"
int smpd_post_read_command(smpd_context_t *context)
{
	int result;

	smpd_enter_fn(FCNAME);

	/* post a read for the next command header */
	smpd_dbg_printf(L"posting a read for a command header on the %s context, sock %d\n", smpd_get_context_str(context), MPIDU_Sock_get_sock_id(context->sock));
	context->read_state = SMPD_READING_CMD_HEADER;
	context->read_cmd.state = SMPD_CMD_READING_HDR;
	result = MPIDU_Sock_post_read(context->sock, context->read_cmd.cmd_hdr_str, SMPD_CMD_HDR_LENGTH * sizeof(wchar_t), SMPD_CMD_HDR_LENGTH * sizeof(wchar_t), NULL);
	if (result != MPI_SUCCESS)
	{
		smpd_err_printf(L"unable to post a read for the next command header,\nsock error: %s\n", get_sock_error_string(result));
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}

	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}

#undef FCNAME
#define FCNAME L"smpd_post_write_command"
int smpd_post_write_command(smpd_context_t *context, smpd_command_t *cmd)
{
	int result;
	smpd_command_t *iter;

	smpd_enter_fn(FCNAME);

	if (context == NULL)
	{
		smpd_dbg_printf(L"unable to post a write of command '%s' on a NULL context", cmd->cmd);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}

	smpd_package_command(cmd);
	/*smpd_dbg_printf(L"command after packaging: \"%s\"\n", cmd->cmd);*/
	cmd->next = NULL;
	cmd->state = SMPD_CMD_WRITING_CMD;
	context->write_state = SMPD_WRITING_CMD;

	if (!context->write_list)
	{
		context->write_list = cmd;
	}
	else
	{
		smpd_dbg_printf(L"enqueueing write at the end of the list.\n");
		iter = context->write_list;
		while (iter->next)
			iter = iter->next;
		iter->next = cmd;
		smpd_exit_fn(FCNAME);
		return SMPD_SUCCESS;
	}

	cmd->iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)cmd->cmd_hdr_str;
	cmd->iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)(SMPD_CMD_HDR_LENGTH * sizeof(wchar_t));
	cmd->iov[1].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)cmd->cmd;
	cmd->iov[1].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)(cmd->length * sizeof(wchar_t));
	/*smpd_dbg_printf(L"command at this moment: \"%s\"\n", cmd->cmd);*/
	smpd_dbg_printf(L"smpd_post_write_command on the %s context sock %d: %d bytes for command: \"%s\"\n",
		smpd_get_context_str(context), MPIDU_Sock_get_sock_id(context->sock),
		cmd->iov[0].MPID_IOV_LEN + cmd->iov[1].MPID_IOV_LEN,
		cmd->cmd);
	result = MPIDU_Sock_post_writev(context->sock, cmd->iov, 2, NULL);
	if (result != MPI_SUCCESS)
	{
		smpd_err_printf(L"unable to post a write for the next command,\nsock error: %s\n", get_sock_error_string(result));
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}

	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}

#undef FCNAME
#define FCNAME L"smpd_package_command"
int smpd_package_command(smpd_command_t *cmd)
{
	int length;

	smpd_enter_fn(FCNAME);

	/* create the command header - for now it is simply the length of the command string */
	length = (int)wcslen(cmd->cmd) + 1;
	if (length > SMPD_MAX_CMD_LENGTH)
	{
		smpd_err_printf(L"unable to package invalid command of length %d\n", length);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	_snwprintf_s(cmd->cmd_hdr_str, SMPD_CMD_HDR_LENGTH, _TRUNCATE, L"%d", length);
	cmd->length = length;

	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}
