/*   Copyright (C) 2009 Deino Software. */

#include <stdio.h>
#include "smpd.h"
#include <stdlib.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_STDARG_H
#include <stdarg.h>
#endif
#ifdef HAVE_TERMIOS_H
#include <termios.h>
#endif
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_SIGACTION
#ifdef HAVE_SIGNAL_H
#include <signal.h>
#endif
#ifdef HAVE_SYS_WAIT_H
#include <sys/wait.h>
#endif
#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif
#endif

smpd_global_t smpd_process = 
	{ SMPD_IDLE,		/* state				  */
	  -1,			   /* id					 */
	  -1,			   /* parent_id			  */
	  -1,			   /* level				  */
	  NULL,			 /* left_context		   */
	  NULL,			 /* right_context		  */
	  NULL,			 /* parent_context		 */
	  NULL,			 /* context_list		   */
	  NULL,			 /* listener_context	   */
	  NULL,			 /* process_list		   */
	  SMPD_FALSE,	   /* closing				*/
	  SMPD_FALSE,	   /* root_smpd			  */
	  MPIDU_SOCK_INVALID_SET, /* set					*/
	  L"",			   /* host				   */
	  L"",			   /* pszExe				 */
	  SMPD_FALSE,	   /* bService			   */
	  SMPD_TRUE,		/* bNoTTY				 */
	  SMPD_FALSE,	   /* bPasswordProtect	   */
	  L"",			   /* SMPDPassword		   */
	  L"",			   /* passphrase			 */
	  SMPD_FALSE,	   /* logon				  */
	  L"",			   /* UserAccount			*/
	  L"",			   /* UserPassword		   */
	  0,				/* cur_tag				*/
	  SMPD_DBG_STATE_ERROUT, /* dbg_state		 */
	  /*NULL,*/			 /* dbg_fout			   */
	  L"",			   /* dbg_filename		   */
	  SMPD_DBG_FILE_SIZE,/*dbg_file_size		  */
	  SMPD_FALSE,	   /* have_dbs			   */
	  L"",			   /* kvs_name			   */
	  L"",			   /* domain_name			*/
	  NULL,			 /* hCloseStdinThreadEvent */
	  NULL,			 /* hStdinThread		   */
#ifdef USE_PTHREAD_STDIN_REDIRECTION
	  0,				/* stdin_thread		   */
	  0,				/* stdin_read			 */
	  0,				/* stdin_write			*/
#endif
	  SMPD_FALSE,	   /* do_console			 */
	  SMPD_LISTENER_PORT, /* smpd port			*/
	  L"",			   /* console_host		   */
	  NULL,			 /* host_list			  */
	  NULL,			 /* launch_list			*/
	  SMPD_TRUE,		/* credentials_prompt	 */
	  SMPD_TRUE,		/* do_multi_color_output  */
	  SMPD_FALSE,	   /* no_mpi				 */
	  SMPD_FALSE,	   /* output_exit_codes	  */
	  SMPD_FALSE,	   /* local_root			 */
	  SMPD_FALSE,	   /* use_iproot			 */
	  SMPD_FALSE,	   /* use_process_session	*/
	  0,				/* nproc				  */
	  0,				/* nproc_launched		 */
	  0,				/* nproc_exited		   */
	  SMPD_FALSE,	   /* verbose				*/
	  /*SMPD_FALSE,*/	   /* shutdown			   */
	  /*SMPD_FALSE,*/	   /* restart				*/
	  /*SMPD_FALSE,*/	   /* validate			   */
	  /*SMPD_FALSE,*/	   /* do_status			  */
	  SMPD_CMD_NONE,	/* builtin_cmd			*/
	  FALSE,			/* bOutputInitialized	 */
	  NULL,			 /* hOutputMutex		   */
	  NULL,			 /* hLaunchProcessMutex	*/
#ifdef USE_WIN_MUTEX_PROTECT
	  NULL,			 /* hDBSMutext			 */
#endif
	  NULL,			 /* pDatabase			  */
	  NULL,			 /* pDatabaseIter		  */
	  /*0,*/				/* nNextAvailableDBSID	*/
	  0,				/* nInitDBSRefCount	   */
	  NULL,			 /* barrier_list		   */
	  {				 /* ssStatus					*/
		  SERVICE_WIN32_OWN_PROCESS, /* dwServiceType  */
		  SERVICE_STOPPED,		   /* dwCurrentState */
		  SERVICE_ACCEPT_STOP,   /* dwControlsAccepted */
		  NO_ERROR,	 /* dwWin32ExitCode			 */
		  0,			/* dwServiceSpecificExitCode   */
		  0,			/* dwCheckPoint				*/
		  3000,		 /* dwWaitHint				  */
	  },
	  NULL,			 /* sshStatusHandle		 */
	  NULL,			 /* hBombDiffuseEvent	   */
	  NULL,			 /* hBombThread			 */
	  SMPD_FALSE,	   /* service_stop			*/
	  SMPD_FALSE,	   /* noprompt				*/
	  L"",			   /* smpd_filename		   */
	  SMPD_FALSE,	   /* stdin_toall			 */
	  SMPD_FALSE,	   /* stdin_redirecting	   */
	  NULL,			 /* default_host_list	   */
	  NULL,			 /* cur_default_host		*/
	  0,				/* cur_default_iproc	   */
	  NULL,			 /* hSMPDDataMutex		  */
	  L"",			   /* printf_buffer		   */
	  SMPD_SUCCESS,	 /* state_machine_ret_val   */
	  SMPD_FALSE,	   /* exit_on_done			*/
	  0,				/* tree_parent			 */
	  1,				/* tree_id				 */
	  NULL,			 /* s_host_list			 */
	  NULL,			 /* s_cur_host			  */
	  0,				/* s_cur_count			 */
	  SMPD_FALSE,	   /* use_inherited_handles   */
	  NULL,			 /* pg_list				 */
	  SMPD_FALSE,	   /* use_abort_exit_code	 */
	  0,				/* abort_exit_code		 */
	  SMPD_TRUE,		/* verbose_abort_output	*/
	  0,				/* mpiexec_exit_code	   */
	  SMPD_FALSE,	   /* map0to1				 */
	  SMPD_FALSE,	   /* rsh_mpiexec			 */
	  SMPD_FALSE,	   /* mpiexec_inorder_launch  */
	  SMPD_FALSE,	   /* mpiexec_run_local	   */
	  NULL,			 /* timeout_thread		  */
	  -1,			   /* timeout				 */
	  MPIDU_SOCK_INVALID_SOCK, /* timeout_sock	 */
	  MPIDU_SOCK_INVALID_SOCK, /* mpiexec_abort_sock */
	  SMPD_TRUE,		/* use_pmi_server		  */
	  NULL,			 /* mpiexec_argv0		   */
	  L"dummy",		  /* encrypt_prefix		  */
	  SMPD_FALSE,	   /* plaintext			   */
	  SMPD_FALSE,	   /* use_sspi				*/
	  SMPD_FALSE,	   /* use_delegation		  */
	  SMPD_FALSE,	   /* use_sspi_job_key		*/
	  NULL,			 /* sec_fn				  */
	  NULL,			 /* sspi_context_list	   */
	  L"",			   /* job_key				 */
	  L"",			   /* job_key_account		 */
	  L"",			   /* job_key_password		*/
	  L"",			   /* key					 */
	  L"",			   /* val					 */
	  SMPD_FALSE,	   /* do_console_returns	  */
	  L"",			   /* env_channel			 */
	  L"",			   /* env_dll				 */
	  L"",			   /* env_wrap_dll			*/
	  NULL,			 /* delayed_spawn_queue	 */
	  SMPD_FALSE,	   /* spawning				*/
	  0,				/* user_index			  */
	  SMPD_FALSE,	   /* prefix_output		   */
	  SMPD_ABORT_CALLBACK_TIMEOUT, /* abort_callback_timeout */
	  NULL,			 /* stdin_filename		  */
	  NULL,			 /* stdout_filename		 */
	  NULL,			 /* stderr_filename		 */
	  0,				/* num_stdin_ranks		 */
	  NULL			  /* stdin_ranks			 */
	};

#undef FCNAME
#define FCNAME L"smpd_post_abort_command"
int smpd_post_abort_command(wchar_t *fmt, ...)
{
	int result;
	wchar_t error_str[2048] = L"";
	smpd_command_t *cmd_ptr;
	smpd_context_t *context;
	va_list list;

	smpd_enter_fn(FCNAME);

	va_start(list, fmt);
	_vsnwprintf_s(error_str, 2048, _TRUNCATE, fmt, list);
	va_end(list);

	result = smpd_create_command(L"abort", smpd_process.id, 0, SMPD_FALSE, &cmd_ptr);
	if (result != SMPD_SUCCESS)
	{
		smpd_err_printf(L"unable to create an abort command.\n");
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	result = smpd_add_command_arg(cmd_ptr, L"error", error_str);
	if (result != SMPD_SUCCESS)
	{
		smpd_err_printf(L"Unable to add the error string to the abort command.\n");
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	smpd_command_destination(0, &context);
	if (context == NULL)
	{
		if (smpd_process.left_context == NULL)
		{
			wprintf(L"Aborting: %s\n", error_str);
			fflush(stdout);
			smpd_exit_fn(FCNAME);
			smpd_exit(-1);
		}

		smpd_process.closing = SMPD_TRUE;
		result = smpd_create_command(L"close", 0, 1, SMPD_FALSE, &cmd_ptr);
		if (result != SMPD_SUCCESS)
		{
			smpd_err_printf(L"unable to create the close command to tear down the job tree.\n");
			smpd_exit_fn(FCNAME);
			return SMPD_FAIL;
		}
		result = smpd_post_write_command(smpd_process.left_context, cmd_ptr);
		if (result != SMPD_SUCCESS)
		{
			smpd_err_printf(L"unable to post a write of the close command to tear down the job tree as part of the abort process.\n");
			smpd_exit_fn(FCNAME);
			return SMPD_FAIL;
		}
	}
	else
	{
		smpd_dbg_printf(L"sending abort command to %s context: \"%s\"\n", smpd_get_context_str(context), cmd_ptr->cmd);
		result = smpd_post_write_command(context, cmd_ptr);
		if (result != SMPD_SUCCESS)
		{
			smpd_err_printf(L"unable to post a write of the abort command to the %s context.\n", smpd_get_context_str(context));
			smpd_exit_fn(FCNAME);
			return SMPD_FAIL;
		}
	}
	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}

#ifdef HAVE_SIGACTION
void smpd_child_handler(int code)
{
	int status;
	int pid;

	if (smpd_process.root_smpd && code == SIGCHLD)
	{
		/*pid = waitpid(-1, &status, WNOHANG);*/
		pid = waitpid(-1, &status, 0);
		if (pid < 0)
		{
			fwprintf(stderr, L"waitpid failed, error %d\n", errno);
		}
		/*
		else
		{
			wprintf(L"process %d exited with code: %d\n", pid, WIFEXITED(status) ? WEXITSTATUS(status) : -1);
			fflush(stdout);
		}
		*/
	}
}
#endif

#undef FCNAME
#define FCNAME L"smpd_make_socket_loop"
int smpd_make_socket_loop(SOCKET *pRead, SOCKET *pWrite)
{
	SOCKET sock;
	wchar_t host[100];
	int port;
	int len;
	/*LINGER linger;*/
	BOOL b;
	SOCKADDR_IN sockAddr;
	int error;

	smpd_enter_fn(FCNAME);

	/* Create a listener */

	/* create the socket */
	sock = WSASocket(PF_INET, SOCK_STREAM, 0, NULL, 0, WSA_FLAG_OVERLAPPED);
	if (sock == INVALID_SOCKET)
	{
		*pRead = INVALID_SOCKET;
		*pWrite = INVALID_SOCKET;
		smpd_exit_fn(FCNAME);
		return WSAGetLastError();
	}

	memset(&sockAddr,0,sizeof(sockAddr));
	
	sockAddr.sin_family = AF_INET;
	sockAddr.sin_addr.s_addr = INADDR_ANY;
	sockAddr.sin_port = htons((unsigned short)ADDR_ANY);
	
	if (bind(sock, (SOCKADDR*)&sockAddr, sizeof(sockAddr)) == SOCKET_ERROR)
	{
		error = WSAGetLastError();
		smpd_err_printf(L"bind failed: error %d\n", error);
		*pRead = INVALID_SOCKET;
		*pWrite = INVALID_SOCKET;
		smpd_exit_fn(FCNAME);
		return error;
	}
	
	/* listen */
	listen(sock, 2);

	/* get the host and port where we're listening */
	len = sizeof(sockAddr);
	getsockname(sock, (struct sockaddr*)&sockAddr, &len);
	port = ntohs(sockAddr.sin_port);
	smpd_get_hostname(host, 100);

	/* Connect to myself */

	/* create the socket */
	*pWrite = WSASocket(PF_INET, SOCK_STREAM, 0, NULL, 0, WSA_FLAG_OVERLAPPED);
	if (*pWrite == INVALID_SOCKET)
	{
		error = WSAGetLastError();
		smpd_err_printf(L"WSASocket failed, error %d\n", error);
		if (closesocket(sock) == SOCKET_ERROR)
		{
			smpd_err_printf(L"closesocket failed, sock %d, error %d\n", sock, WSAGetLastError());
		}
		*pRead = INVALID_SOCKET;
		*pWrite = INVALID_SOCKET;
		smpd_exit_fn(FCNAME);
		return error;
	}

	/* set the nodelay option */
	/*
	b = TRUE;
	setsockopt(*pWrite, IPPROTO_TCP, TCP_NODELAY, (char*)&b, sizeof(BOOL));
	*/

	/* Set the linger on close option */
	/*
	linger.l_onoff = 1 ;
	linger.l_linger = 60;
	setsockopt(*pWrite, SOL_SOCKET, SO_LINGER, (char*)&linger, sizeof(linger));
	*/

	sockAddr.sin_addr.s_addr = inet_addr("127.0.0.1");

	/* connect to myself */
	if (connect(*pWrite, (SOCKADDR*)&sockAddr, sizeof(sockAddr)) == SOCKET_ERROR)
	{
		error = WSAGetLastError();
		if (closesocket(*pWrite) ==  SOCKET_ERROR)
		{
			smpd_err_printf(L"closesocket failed, sock %d, error %d\n", *pWrite, WSAGetLastError());
		}
		if (closesocket(sock) == SOCKET_ERROR)
		{
			smpd_err_printf(L"closesocket failed, sock %d, error %d\n", sock, WSAGetLastError());
		}
		*pRead = INVALID_SOCKET;
		*pWrite = INVALID_SOCKET;
		smpd_exit_fn(FCNAME);
		return error;
	}

	/* Accept the connection from myself */
	len = sizeof(sockAddr);
	*pRead = accept(sock, (SOCKADDR*)&sockAddr, &len);

	/* set the nodelay option */
	b = TRUE;
	setsockopt(*pWrite, IPPROTO_TCP, TCP_NODELAY, (char*)&b, sizeof(BOOL));
	/* set the nodelay option */
	b = TRUE;
	setsockopt(*pRead, IPPROTO_TCP, TCP_NODELAY, (char*)&b, sizeof(BOOL));

	if (closesocket(sock) == SOCKET_ERROR)
	{
		smpd_err_printf(L"closesocket failed, sock %d, error %d\n", sock, WSAGetLastError());
	}
	smpd_exit_fn(FCNAME);
	return 0;
}

#undef FCNAME
#define FCNAME L"smpd_make_socket_loop_choose"
int smpd_make_socket_loop_choose(SOCKET *pRead, int read_overlapped, SOCKET *pWrite, int write_overlapped)
{
	SOCKET sock;
	wchar_t host[100];
	int port;
	int len;
	/*LINGER linger;*/
	BOOL b;
	SOCKADDR_IN sockAddr;
	int error;
	DWORD flag;

	smpd_enter_fn(FCNAME);

	/* Create a listener */

	/* create the socket */
	flag = read_overlapped ? WSA_FLAG_OVERLAPPED : 0;
	sock = WSASocket(PF_INET, SOCK_STREAM, 0, NULL, 0, flag);
	if (sock == INVALID_SOCKET)
	{
		*pRead = INVALID_SOCKET;
		*pWrite = INVALID_SOCKET;
		smpd_exit_fn(FCNAME);
		return WSAGetLastError();
	}

	memset(&sockAddr,0,sizeof(sockAddr));
	
	sockAddr.sin_family = AF_INET;
	sockAddr.sin_addr.s_addr = INADDR_ANY;
	sockAddr.sin_port = htons((unsigned short)ADDR_ANY);

	if (bind(sock, (SOCKADDR*)&sockAddr, sizeof(sockAddr)) == SOCKET_ERROR)
	{
		error = WSAGetLastError();
		smpd_err_printf(L"bind failed: error %d\n", error);
		*pRead = INVALID_SOCKET;
		*pWrite = INVALID_SOCKET;
		smpd_exit_fn(FCNAME);
		return error;
	}
	
	/* listen */
	listen(sock, 2);

	/* get the host and port where we're listening */
	len = sizeof(sockAddr);
	getsockname(sock, (struct sockaddr*)&sockAddr, &len);
	port = ntohs(sockAddr.sin_port);
	smpd_get_hostname(host, 100);

	/* Connect to myself */

	/* create the socket */
	flag = write_overlapped ? WSA_FLAG_OVERLAPPED : 0;
	*pWrite = WSASocket(PF_INET, SOCK_STREAM, 0, NULL, 0, flag);
	if (*pWrite == INVALID_SOCKET)
	{
		error = WSAGetLastError();
		smpd_err_printf(L"WSASocket failed, error %d\n", error);
		if (closesocket(sock) == SOCKET_ERROR)
		{
			smpd_err_printf(L"closesocket failed, sock %d, error %d\n", sock, WSAGetLastError());
		}
		*pRead = INVALID_SOCKET;
		*pWrite = INVALID_SOCKET;
		smpd_exit_fn(FCNAME);
		return error;
	}

	/* set the nodelay option */
	b = TRUE;
	setsockopt(*pWrite, IPPROTO_TCP, TCP_NODELAY, (char*)&b, sizeof(BOOL));

	/* Set the linger on close option */
	/*
	linger.l_onoff = 1 ;
	linger.l_linger = 60;
	setsockopt(*pWrite, SOL_SOCKET, SO_LINGER, (char*)&linger, sizeof(linger));
	*/

	sockAddr.sin_addr.s_addr = inet_addr("127.0.0.1");

	/* connect to myself */
	if (connect(*pWrite, (SOCKADDR*)&sockAddr, sizeof(sockAddr)) == SOCKET_ERROR)
	{
		error = WSAGetLastError();
		if (closesocket(*pWrite) == SOCKET_ERROR)
		{
			smpd_err_printf(L"closesocket failed, sock %d, error %d\n", *pWrite, WSAGetLastError());
		}
		if (closesocket(sock) == SOCKET_ERROR)
		{
			smpd_err_printf(L"closesocket failed, sock %d, error %d\n", sock, WSAGetLastError());
		}
		*pRead = INVALID_SOCKET;
		*pWrite = INVALID_SOCKET;
		smpd_exit_fn(FCNAME);
		return error;
	}

	/* Accept the connection from myself */
	len = sizeof(sockAddr);
	*pRead = accept(sock, (SOCKADDR*)&sockAddr, &len);

	/* set the nodelay option */
	b = TRUE;
	setsockopt(*pRead, IPPROTO_TCP, TCP_NODELAY, (char*)&b, sizeof(BOOL));

	if (closesocket(sock) == SOCKET_ERROR)
	{
		smpd_err_printf(L"closesocket failed, sock %d, error %d\n", sock, WSAGetLastError());
	}
	smpd_exit_fn(FCNAME);
	return 0;
}

#undef FCNAME
#define FCNAME L"smpd_init_process"
int smpd_init_process(void)
{
	HMODULE hModule;
#ifdef HAVE_SIGACTION
	struct sigaction act;
#endif

	smpd_enter_fn(FCNAME);

	/* initialize the debugging output print engine */
	smpd_init_printf();

	/* tree data */
	smpd_process.parent_id = -1;
	smpd_process.id = -1;
	smpd_process.level = -1;
	smpd_process.left_context = NULL;
	smpd_process.right_context = NULL;
	smpd_process.parent_context = NULL;
	smpd_process.set = MPIDU_SOCK_INVALID_SET;

	/* local data */
	hModule = GetModuleHandle(NULL);
	if (!GetModuleFileName(hModule, smpd_process.pszExe, SMPD_MAX_EXE_LENGTH)) 
		smpd_process.pszExe[0] = L'\0';
	wcscpy_s(smpd_process.SMPDPassword, 100, SMPD_DEFAULT_PASSWORD);
	smpd_process.bPasswordProtect = SMPD_FALSE;
	smpd_process.bService = SMPD_FALSE;
	smpd_get_hostname(smpd_process.host, SMPD_MAX_HOST_LENGTH);
	smpd_process.UserAccount[0] = L'\0';
	smpd_process.UserPassword[0] = L'\0';
	smpd_process.closing = SMPD_FALSE;
	smpd_process.root_smpd = SMPD_FALSE;

	srand(smpd_getpid());

#ifdef HAVE_SIGACTION
	memset(&act, 0, sizeof(act));
	act.sa_handler = smpd_child_handler;
#ifdef SA_NODEFER
	act.sa_flags = SA_NOCLDSTOP | SA_NODEFER;
#else
	act.sa_flags = SA_NOCLDSTOP;
#endif
	sigaction(SIGCHLD, &act, NULL);
#endif

	smpd_process.hBombDiffuseEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
	smpd_process.hLaunchProcessMutex = CreateMutex(NULL, FALSE, NULL);

	smpd_get_smpd_data(L"phrase", smpd_process.passphrase, SMPD_PASSPHRASE_MAX_LENGTH);

	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}

#undef FCNAME
#define FCNAME L"smpd_init_context"
int smpd_init_context(smpd_context_t *context, smpd_context_type_t type, MPIDU_Sock_set_t set, MPIDU_Sock_t sock, int id)
{
	int result;

	smpd_enter_fn(FCNAME);
	context->type = type;
	context->target = SMPD_TARGET_UNDETERMINED;
	context->access = SMPD_ACCESS_USER_PROCESS;/*SMPD_ACCESS_NONE;*/
	context->host[0] = L'\0';
	context->id = id;
	context->rank = 0;
	context->write_list = NULL;
	context->wait_list = NULL;
	smpd_init_command(&context->read_cmd);
	context->next = NULL;
	context->set = set;
	context->sock = sock;
	context->state = SMPD_IDLE;
	context->read_state = SMPD_IDLE;
	context->write_state = SMPD_IDLE;
	context->account[0] = L'\0';
	context->domain[0] = L'\0';
	context->full_domain[0] = L'\0';
	context->connect_return_id = -1;
	context->connect_return_tag = -1;
	context->connect_to = NULL;
	context->spawn_context = NULL;
	context->cred_request[0] = L'\0';
	context->password[0] = L'\0';
	context->encrypted_password[0] = L'\0';
	context->port_str[0] = L'\0';
	context->pszChallengeResponse[0] = L'\0';
	context->pszCrypt[0] = L'\0';
	context->pwd_request[0] = L'\0';
	context->session[0] = L'\0';
	context->session_header[0] = L'\0';
	context->smpd_pwd[0] = L'\0';
	context->wait.hProcess = NULL;
	context->wait.hThread = NULL;
	context->process = NULL;
	context->sspi_header[0] = L'\0';
	context->sspi_context = NULL;
	context->sspi_type = SMPD_SSPI_DELEGATE;
	context->sspi_job_key[0] = L'\0';
	context->first_output_stderr = SMPD_TRUE;
	context->first_output_stdout = SMPD_TRUE;

	if (sock != MPIDU_SOCK_INVALID_SOCK)
	{
		result = MPIDU_Sock_set_user_ptr(sock, context);
		if (result != MPI_SUCCESS)
		{
			smpd_err_printf(L"unable to set the sock user ptr while initializing context,\nsock error: %s\n",
				get_sock_error_string(result));
		}
	}

	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}

#undef FCNAME
#define FCNAME L"smpd_create_sspi_client_context"
int smpd_create_sspi_client_context(smpd_sspi_client_context_t **new_context)
{
	smpd_sspi_client_context_t *context;

	smpd_enter_fn(FCNAME);

	context = (smpd_sspi_client_context_t *)malloc(sizeof(smpd_sspi_client_context_t));
	if (context == NULL)
	{
		*new_context = NULL;
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	context->buffer = NULL;
	context->buffer_length = 0;
	context->out_buffer = NULL;
	context->out_buffer_length = 0;
	context->max_buffer_size = 0;
	SecInvalidateHandle(&context->credential);
	SecInvalidateHandle(&context->context);
	memset(&context->expiration_time, 0, sizeof(TimeStamp));
	context->user_handle = INVALID_HANDLE_VALUE;
	context->job = INVALID_HANDLE_VALUE;
	context->flags = 0;
	context->close_handle = SMPD_TRUE;
	/* FIXME: this insertion needs to be thread safe */
	if (smpd_process.sspi_context_list == NULL)
	{
		context->id = 0;
	}
	else
	{
		context->id = smpd_process.sspi_context_list->id + 1;
	}
	context->next = smpd_process.sspi_context_list;
	smpd_process.sspi_context_list = context;
	*new_context = context;
	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}

#undef FCNAME
#define FCNAME L"smpd_free_sspi_client_context"
int smpd_free_sspi_client_context(smpd_sspi_client_context_t **context)
{
	smpd_sspi_client_context_t *iter, *trailer;

	smpd_enter_fn(FCNAME);

	trailer = iter = smpd_process.sspi_context_list;
	while (iter)
	{
		if (iter == *context)
		{
			if (trailer != iter)
			{
				trailer->next = iter->next;
			}
			else
			{
				smpd_process.sspi_context_list = iter->next;
			}
			break;
		}
		if (trailer != iter)
			trailer = trailer->next;
		iter = iter->next;
	}
	if (iter == NULL)
	{
		smpd_dbg_printf(L"freeing a sspi_client_context not in the global list\n");
	}
	/* FIXME: cleanup sspi structures */
	free(*context);
	*context = NULL;
	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}

#undef FCNAME
#define FCNAME L"smpd_generate_session_header"
int smpd_generate_session_header(wchar_t *str, int session_id)
{
	wchar_t * str_orig;
	int result;
	int len;

	smpd_enter_fn(FCNAME);

	str_orig = str;
	*str = L'\0';
	len = SMPD_MAX_SESSION_HEADER_LENGTH;

	/* add header fields */
	result = MPIU_Str_add_int_arg(&str, &len, L"id", session_id);
	if (result != MPIU_STR_SUCCESS)
	{
		smpd_err_printf(L"unable to create session header, adding session id failed.\n");
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	result = MPIU_Str_add_int_arg(&str, &len, L"parent", smpd_process.id);
	if (result != MPIU_STR_SUCCESS)
	{
		smpd_err_printf(L"unable to create session header, adding parent id failed.\n");
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	result = MPIU_Str_add_int_arg(&str, &len, L"level", smpd_process.level + 1);
	if (result != MPIU_STR_SUCCESS)
	{
		smpd_err_printf(L"unable to create session header, adding session level failed.\n");
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}

	/* remove the trailing space */
	str--;
	*str = L'\0';

	smpd_dbg_printf(L"session header: (%s)\n", str_orig);
	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}

#undef FCNAME
#define FCNAME L"smpd_get_password"
void smpd_get_password(wchar_t *password, size_t maxlen)
{
	HANDLE hStdin;
	DWORD dwMode;
	size_t len;

	smpd_enter_fn(FCNAME);

	hStdin = GetStdHandle(STD_INPUT_HANDLE);
	if (!GetConsoleMode(hStdin, &dwMode))
		dwMode = ENABLE_LINE_INPUT | ENABLE_PROCESSED_INPUT | ENABLE_MOUSE_INPUT;
	SetConsoleMode(hStdin, dwMode & ~ENABLE_ECHO_INPUT);
	*password = L'\0';
	fgetws(password, (int)maxlen, stdin);
	SetConsoleMode(hStdin, dwMode);

	fwprintf(stderr, L"\n");

	while ((len = wcslen(password)) > 0)
	{
		if (password[len-1] == L'\r' || password[len-1] == L'\n')
			password[len-1] = L'\0';
		else
			break;
	}
	smpd_exit_fn(FCNAME);
}

#undef FCNAME
#define FCNAME L"smpd_get_account_and_password"
void smpd_get_account_and_password(wchar_t *account, size_t maxlen1, wchar_t *password, size_t maxlen2)
{
	size_t len;
	wchar_t default_username[100] = L"";
	ULONG default_len = 100;

	smpd_enter_fn(FCNAME);

	if (!GetUserNameEx(NameSamCompatible, default_username, &default_len))
	{
		default_username[0] = L'\0';
	}

	do
	{
		do
		{
			if (default_username[0] != L'\0')
			{
				fwprintf(stderr, L"account (domain\\user) [%s]: ", default_username);
			}
			else
			{
				fwprintf(stderr, L"account (domain\\user): ");
			}
			fflush(stderr);
			*account = L'\0';
			fgetws(account, 100, stdin);
			while (wcslen(account))
			{
				len = wcslen(account);
				if (account[len-1] == L'\r' || account[len-1] == L'\n')
					account[len-1] = L'\0';
				else
					break;
			}
			if ((wcslen(account) == 0) && (default_username[0] != L'\0'))
			{
				wcscpy_s(account, maxlen1, default_username);
			}
		} 
		while (wcslen(account) == 0);

		fwprintf(stderr, L"password: ");
		fflush(stderr);

		smpd_get_password(password, maxlen2);
		if (wcslen(password) == 0)
		{
			fwprintf(stderr, L"MPICH2 is unable to manage jobs using credentials with a blank password.\nPlease enter another account.\n");
		}
	}
	while (wcslen(password) == 0);

	smpd_exit_fn(FCNAME);
}

#undef FCNAME
#define FCNAME L"smpd_get_credentials_from_parent"
int smpd_get_credentials_from_parent(MPIDU_Sock_set_t set, MPIDU_Sock_t sock)
{
	smpd_enter_fn(FCNAME);
	SMPD_UNREFERENCED_ARG(set);
	SMPD_UNREFERENCED_ARG(sock);
	smpd_exit_fn(FCNAME);
	return SMPD_FAIL;
}

#undef FCNAME
#define FCNAME L"smpd_get_smpd_password_from_parent"
int smpd_get_smpd_password_from_parent(MPIDU_Sock_set_t set, MPIDU_Sock_t sock)
{
	smpd_enter_fn(FCNAME);
	SMPD_UNREFERENCED_ARG(set);
	SMPD_UNREFERENCED_ARG(sock);
	smpd_exit_fn(FCNAME);
	return SMPD_FAIL;
}

#undef FCNAME
#define FCNAME L"smpd_get_default_hosts"
int smpd_get_default_hosts()
{
	wchar_t hosts[8192];
	wchar_t *host, *tok;
	wchar_t *ncpu;
	smpd_host_node_t *cur_host, *iter;
	DWORD len;

	smpd_enter_fn(FCNAME);

	if (smpd_process.default_host_list != NULL && smpd_process.cur_default_host != NULL)
	{
		smpd_dbg_printf(L"default list already populated, returning success.\n");
		smpd_exit_fn(FCNAME);
		return SMPD_SUCCESS;
	}

	if (smpd_get_smpd_data(L"hosts", hosts, 8192) != SMPD_SUCCESS)
	{
		len = 8192;
		/*if (GetComputerName(hosts, &len))*/
		if (GetComputerNameEx(ComputerNameDnsFullyQualified, hosts, &len))
		{
			smpd_process.default_host_list = (smpd_host_node_t*)malloc(sizeof(smpd_host_node_t));
			if (smpd_process.default_host_list == NULL)
			{
				smpd_exit_fn(FCNAME);
				return SMPD_FAIL;
			}
			wcscpy_s(smpd_process.default_host_list->host, SMPD_MAX_HOST_LENGTH, hosts);
			smpd_process.default_host_list->alt_host[0] = L'\0';
			smpd_process.default_host_list->nproc = 1;
			smpd_process.default_host_list->connected = SMPD_FALSE;
			smpd_process.default_host_list->connect_cmd_tag = -1;
			smpd_process.default_host_list->next = smpd_process.default_host_list;
			smpd_process.default_host_list->left = NULL;
			smpd_process.default_host_list->right = NULL;
			smpd_process.cur_default_host = smpd_process.default_host_list;
			smpd_exit_fn(FCNAME);
			return SMPD_SUCCESS;
		}
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}

	/* FIXME: Insert code here to parse a compressed host string */
	/* For now, just use a space separated list of host names */

	host = wcstok_s(hosts, L" \t\r\n", &tok);
	while (host)
	{
		cur_host = (smpd_host_node_t*)malloc(sizeof(smpd_host_node_t));
		if (cur_host != NULL)
		{
			/*wprintf(L"default host: %s\n", host);*/
			wcscpy_s(cur_host->host, SMPD_MAX_HOST_LENGTH, host);
			cur_host->alt_host[0] = L'\0';
			cur_host->nproc = 1;
			ncpu = wcsstr(cur_host->host, L":");
			if (ncpu)
			{
				*ncpu = L'\0';
				ncpu++;
				cur_host->nproc = _wtoi(ncpu);
				if (cur_host->nproc < 1)
					cur_host->nproc = 1;
			}
			cur_host->connected = SMPD_FALSE;
			cur_host->connect_cmd_tag = -1;
			cur_host->next = NULL;
			cur_host->left = NULL;
			cur_host->right = NULL;
			if (smpd_process.default_host_list == NULL)
			{
				smpd_process.default_host_list = cur_host;
			}
			else
			{
				iter = smpd_process.default_host_list;
				while (iter->next)
					iter = iter->next;
				iter->next = cur_host;
			}
		}
		host = wcstok_s(NULL, L" \t\r\n", &tok);
	}
	if (smpd_process.default_host_list)
	{
		/* make the default list into a ring */
		iter = smpd_process.default_host_list;
		while (iter->next)
			iter = iter->next;
		iter->next = smpd_process.default_host_list;
		/* point the cur_default_host to the first node in the ring */
		smpd_process.cur_default_host = smpd_process.default_host_list;
	}

	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}

#undef FCNAME
#define FCNAME L"smpd_add_host_to_default_list"
int smpd_add_host_to_default_list(const wchar_t *hostname)
{
	int result;
	smpd_enter_fn(FCNAME);
	result = smpd_add_extended_host_to_default_list(hostname, NULL, 1);
	smpd_exit_fn(FCNAME);
	return result;
}

#undef FCNAME
#define FCNAME L"smpd_add_extended_host_to_default_list"
int smpd_add_extended_host_to_default_list(const wchar_t *hostname, const wchar_t *alt_hostname, const int num_cpus)
{
	smpd_host_node_t *iter;

	smpd_enter_fn(FCNAME);

	iter = smpd_process.default_host_list;
	if (iter == NULL)
	{
		smpd_process.default_host_list = (smpd_host_node_t*)malloc(sizeof(smpd_host_node_t));
		if (smpd_process.default_host_list == NULL)
		{
			smpd_exit_fn(FCNAME);
			return SMPD_FAIL;
		}
		wcscpy_s(smpd_process.default_host_list->host, SMPD_MAX_HOST_LENGTH, hostname);
		smpd_process.default_host_list->alt_host[0] = L'\0';
		if (alt_hostname != NULL)
		{
			wcscpy_s(smpd_process.default_host_list->alt_host, SMPD_MAX_HOST_LENGTH, alt_hostname);
		}
		smpd_process.default_host_list->nproc = num_cpus;
		smpd_process.default_host_list->connected = SMPD_FALSE;
		smpd_process.default_host_list->connect_cmd_tag = -1;
		smpd_process.default_host_list->next = smpd_process.default_host_list;
		smpd_process.default_host_list->left = NULL;
		smpd_process.default_host_list->right = NULL;
		smpd_process.cur_default_host = smpd_process.default_host_list;
		smpd_exit_fn(FCNAME);
		return SMPD_SUCCESS;
	}

	if (wcscmp(iter->host, hostname) == 0)
	{
		smpd_exit_fn(FCNAME);
		return SMPD_SUCCESS;
	}
	while (iter->next != NULL && iter->next != smpd_process.default_host_list)
	{
		iter = iter->next;
		if (wcscmp(iter->host, hostname) == 0)
		{
			smpd_exit_fn(FCNAME);
			return SMPD_SUCCESS;
		}
	}

	iter->next = (smpd_host_node_t*)malloc(sizeof(smpd_host_node_t));
	if (iter->next == NULL)
	{
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	iter = iter->next;
	wcscpy_s(iter->host, SMPD_MAX_HOST_LENGTH, hostname);
	iter->alt_host[0] = L'\0';
	if (alt_hostname != NULL)
	{
		wcscpy_s(iter->alt_host, SMPD_MAX_HOST_LENGTH, alt_hostname);
	}
	iter->nproc = num_cpus;
	iter->connected = SMPD_FALSE;
	iter->connect_cmd_tag = -1;
	iter->next = smpd_process.default_host_list;
	iter->left = NULL;
	iter->right = NULL;

	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}

#undef FCNAME
#define FCNAME L"smpd_insert_into_dynamic_hosts"
int smpd_insert_into_dynamic_hosts(void)
{
	smpd_enter_fn(FCNAME);
	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}

#undef FCNAME
#define FCNAME L"smpd_remove_from_dynamic_hosts"
int smpd_remove_from_dynamic_hosts(void)
{
	smpd_enter_fn(FCNAME);
	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}
