/*   Copyright (C) 2009 Deino Software. */

#include "smpd.h"

static void get_uuid(wchar_t *str, size_t max_len)
{
	UUID guid;
	UuidCreate(&guid);
	swprintf_s(str, max_len, L"%08lX-%04X-%04x-%02X%02X-%02X%02X%02X%02X%02X%02X",
		guid.Data1, guid.Data2, guid.Data3,
		guid.Data4[0], guid.Data4[1], guid.Data4[2], guid.Data4[3],
		guid.Data4[4], guid.Data4[5], guid.Data4[6], guid.Data4[7]);
}

#undef FCNAME
#define FCNAME L"smpd_dbs_init"
int smpd_dbs_init()
{
	smpd_enter_fn(FCNAME);
#ifdef USE_WIN_MUTEX_PROTECT
	if (smpd_process.nInitDBSRefCount == 0)
	{
		smpd_process.hDBSMutex = CreateMutex(NULL, FALSE, NULL);
	}
#endif
	smpd_process.nInitDBSRefCount++;
	if (smpd_process.domain_name[0] == L'\0')
	{
		/*get_uuid(smpd_process.domain_name, SMPD_MAX_DBS_NAME_LEN);*/
		/* create a database for the domain to be used for global operations like name publishing */
		smpd_dbs_create(smpd_process.domain_name, SMPD_MAX_DBS_NAME_LEN);
	}
	smpd_exit_fn(FCNAME);
	return SMPD_DBS_SUCCESS;
}

#undef FCNAME
#define FCNAME L"smpd_dbs_finalize"
int smpd_dbs_finalize()
{
	smpd_database_node_t *pNode, *pNext;
	smpd_database_element_t *pElement;

	smpd_enter_fn(FCNAME);

	smpd_process.nInitDBSRefCount--;

	if (smpd_process.nInitDBSRefCount == 0)
	{
#ifdef USE_WIN_MUTEX_PROTECT
		WaitForSingleObject(smpd_process.hDBSMutex, INFINITE);
#endif

		pNode = smpd_process.pDatabase;
		while (pNode)
		{
			pNext = pNode->pNext;

			while (pNode->pData)
			{
				pElement = pNode->pData;
				pNode->pData = pNode->pData->pNext;
				free(pElement);
			}
			free(pNode);

			pNode = pNext;
		}

		smpd_process.pDatabase = NULL;
		smpd_process.pDatabaseIter = NULL;

#ifdef USE_WIN_MUTEX_PROTECT
		ReleaseMutex(smpd_process.hDBSMutex);
		CloseHandle(smpd_process.hDBSMutex);
		smpd_process.hDBSMutex = NULL;
#endif
	}

	smpd_exit_fn(FCNAME);
	return SMPD_DBS_SUCCESS;
}

#undef FCNAME
#define FCNAME L"smpd_dbs_create"
int smpd_dbs_create(wchar_t *name, size_t maxlen)
{
	/*wchar_t guid_str[100];*/
	smpd_database_node_t *pNode, *pNodeTest;

	smpd_enter_fn(FCNAME);

#ifdef USE_WIN_MUTEX_PROTECT
	/* Lock */
	WaitForSingleObject(smpd_process.hDBSMutex, INFINITE);
#endif

	pNode = smpd_process.pDatabase;
	if (pNode)
	{
		while (pNode->pNext)
			pNode = pNode->pNext;
		pNode->pNext = (smpd_database_node_t*)malloc(sizeof(smpd_database_node_t));
		pNode = pNode->pNext;
	}
	else
	{
		smpd_process.pDatabase = (smpd_database_node_t*)malloc(sizeof(smpd_database_node_t));
		pNode = smpd_process.pDatabase;
	}
	pNode->pNext = NULL;
	pNode->pData = NULL;
	pNode->pIter = NULL;
	do
	{
		/*
		swprintf_s(pNode->pszName, SMPD_MAX_DBS_NAME_LEN, L"%d", smpd_process.nNextAvailableDBSID);
		smpd_process.nNextAvailableDBSID++;
		*/
		get_uuid(pNode->pszName, SMPD_MAX_DBS_NAME_LEN);
		if (pNode->pszName[0] == L'\0')
		{
#ifdef USE_WIN_MUTEX_PROTECT
			/* Unlock */
			ReleaseMutex(smpd_process.hDBSMutex);
#endif
			smpd_exit_fn(FCNAME);
			return SMPD_DBS_FAIL;
		}
		pNodeTest = smpd_process.pDatabase;
		while (wcscmp(pNodeTest->pszName, pNode->pszName) != 0)
			pNodeTest = pNodeTest->pNext;
	} while (pNodeTest != pNode);
	wcscpy_s(name, maxlen, pNode->pszName);

#ifdef USE_WIN_MUTEX_PROTECT
	/* Unlock */
	ReleaseMutex(smpd_process.hDBSMutex);
#endif

	/* put a unique id in the kvs database */
	/*
	get_uuid(guid_str, 100);
	smpd_dbs_put(name, PMI_KVS_ID_KEY, guid_str);
	*/

	smpd_exit_fn(FCNAME);
	return SMPD_DBS_SUCCESS;
}

#undef FCNAME
#define FCNAME L"smpd_dbs_create_name_in"
int smpd_dbs_create_name_in(wchar_t *name)
{
	/*wchar_t guid_str[100];*/
	smpd_database_node_t *pNode;

	smpd_enter_fn(FCNAME);

	if (wcslen(name) < 1 || wcslen(name) > SMPD_MAX_DBS_NAME_LEN)
	{
		smpd_exit_fn(FCNAME);
		return SMPD_DBS_FAIL;
	}

#ifdef USE_WIN_MUTEX_PROTECT
	/* Lock */
	WaitForSingleObject(smpd_process.hDBSMutex, INFINITE);
#endif

	/* Check if the name already exists */
	pNode = smpd_process.pDatabase;
	while (pNode)
	{
		if (wcscmp(pNode->pszName, name) == 0)
		{
#ifdef USE_WIN_MUTEX_PROTECT
			/* Unlock */
			ReleaseMutex(smpd_process.hDBSMutex);
#endif
			/*return SMPD_DBS_FAIL;*/
			/* Empty database? */
			smpd_exit_fn(FCNAME);
			return SMPD_DBS_SUCCESS;
		}
		pNode = pNode->pNext;
	}

	pNode = smpd_process.pDatabase;
	if (pNode)
	{
		while (pNode->pNext)
			pNode = pNode->pNext;
		pNode->pNext = (smpd_database_node_t*)malloc(sizeof(smpd_database_node_t));
		pNode = pNode->pNext;
	}
	else
	{
		smpd_process.pDatabase = (smpd_database_node_t*)malloc(sizeof(smpd_database_node_t));
		pNode = smpd_process.pDatabase;
	}
	pNode->pNext = NULL;
	pNode->pData = NULL;
	pNode->pIter = NULL;
	wcscpy_s(pNode->pszName, SMPD_MAX_DBS_NAME_LEN, name);
	
#ifdef USE_WIN_MUTEX_PROTECT
	/* Unlock */
	ReleaseMutex(smpd_process.hDBSMutex);
#endif

	/* put a unique id in the kvs database */
	/*
	get_uuid(guid_str, 100);
	smpd_dbs_put(name, PMI_KVS_ID_KEY, guid_str);
	*/

	smpd_exit_fn(FCNAME);
	return SMPD_DBS_SUCCESS;
}

#undef FCNAME
#define FCNAME L"smpd_dbs_get"
int smpd_dbs_get(const wchar_t *name, const wchar_t *key, wchar_t *value, size_t maxlen)
{
	smpd_database_node_t *pNode;
	smpd_database_element_t *pElement;

	smpd_enter_fn(FCNAME);

#ifdef USE_WIN_MUTEX_PROTECT
	WaitForSingleObject(smpd_process.hDBSMutex, INFINITE);
#endif

	pNode = smpd_process.pDatabase;
	while (pNode)
	{
		if (wcscmp(pNode->pszName, name) == 0)
		{
			pElement = pNode->pData;
			while (pElement)
			{
				if (wcscmp(pElement->pszKey, key) == 0)
				{
					wcscpy_s(value, maxlen, pElement->pszValue);
#ifdef USE_WIN_MUTEX_PROTECT
					ReleaseMutex(smpd_process.hDBSMutex);
#endif
					smpd_exit_fn(FCNAME);
					return SMPD_DBS_SUCCESS;
				}
				pElement = pElement->pNext;
			}
		}
		pNode = pNode->pNext;
	}

#ifdef USE_WIN_MUTEX_PROTECT
	ReleaseMutex(smpd_process.hDBSMutex);
#endif

	smpd_exit_fn(FCNAME);
	return SMPD_DBS_FAIL;
}

#undef FCNAME
#define FCNAME L"smpd_dbs_put"
int smpd_dbs_put(const wchar_t *name, const wchar_t *key, const wchar_t *value)
{
	smpd_database_node_t *pNode;
	smpd_database_element_t *pElement;

	smpd_enter_fn(FCNAME);

#ifdef USE_WIN_MUTEX_PROTECT
	WaitForSingleObject(smpd_process.hDBSMutex, INFINITE);
#endif

	pNode = smpd_process.pDatabase;
	while (pNode)
	{
		if (wcscmp(pNode->pszName, name) == 0)
		{
			pElement = pNode->pData;
			while (pElement)
			{
				if (wcscmp(pElement->pszKey, key) == 0)
				{
					wcscpy_s(pElement->pszValue, SMPD_MAX_DBS_VALUE_LEN, value);
#ifdef USE_WIN_MUTEX_PROTECT
					ReleaseMutex(smpd_process.hDBSMutex);
#endif
					smpd_exit_fn(FCNAME);
					return SMPD_DBS_SUCCESS;
				}
				pElement = pElement->pNext;
			}
			pElement = (smpd_database_element_t*)malloc(sizeof(smpd_database_element_t));
			pElement->pNext = pNode->pData;
			wcscpy_s(pElement->pszKey, SMPD_MAX_DBS_KEY_LEN, key);
			wcscpy_s(pElement->pszValue, SMPD_MAX_DBS_VALUE_LEN, value);
			pNode->pData = pElement;
#ifdef USE_WIN_MUTEX_PROTECT
			ReleaseMutex(smpd_process.hDBSMutex);
#endif
			smpd_exit_fn(FCNAME);
			return SMPD_DBS_SUCCESS;
		}
		pNode = pNode->pNext;
	}

#ifdef USE_WIN_MUTEX_PROTECT
	ReleaseMutex(smpd_process.hDBSMutex);
#endif

	smpd_exit_fn(FCNAME);
	return SMPD_DBS_FAIL;
}

#undef FCNAME
#define FCNAME L"smpd_dbs_delete"
int smpd_dbs_delete(const wchar_t *name, const wchar_t *key)
{
	smpd_database_node_t *pNode;
	smpd_database_element_t *pElement, *pElementTrailer;

	smpd_enter_fn(FCNAME);

#ifdef USE_WIN_MUTEX_PROTECT
	WaitForSingleObject(smpd_process.hDBSMutex, INFINITE);
#endif

	pNode = smpd_process.pDatabase;
	while (pNode)
	{
		if (wcscmp(pNode->pszName, name) == 0)
		{
			pElementTrailer = pElement = pNode->pData;
			while (pElement)
			{
				if (wcscmp(pElement->pszKey, key) == 0)
				{
					if (pElementTrailer != pElement)
						pElementTrailer->pNext = pElement->pNext;
					else
						pNode->pData = pElement->pNext;
					free(pElement);
#ifdef USE_WIN_MUTEX_PROTECT
					ReleaseMutex(smpd_process.hDBSMutex);
#endif
					smpd_exit_fn(FCNAME);
					return SMPD_DBS_SUCCESS;
				}
				pElementTrailer = pElement;
				pElement = pElement->pNext;
			}
#ifdef USE_WIN_MUTEX_PROTECT
			ReleaseMutex(smpd_process.hDBSMutex);
#endif
			smpd_exit_fn(FCNAME);
			return SMPD_DBS_FAIL;
		}
		pNode = pNode->pNext;
	}

#ifdef USE_WIN_MUTEX_PROTECT
	ReleaseMutex(smpd_process.hDBSMutex);
#endif

	smpd_exit_fn(FCNAME);
	return SMPD_DBS_FAIL;
}

#undef FCNAME
#define FCNAME L"smpd_dbs_destroy"
int smpd_dbs_destroy(const wchar_t *name)
{
	smpd_database_node_t *pNode, *pNodeTrailer;
	smpd_database_element_t *pElement;

	smpd_enter_fn(FCNAME);

#ifdef USE_WIN_MUTEX_PROTECT
	WaitForSingleObject(smpd_process.hDBSMutex, INFINITE);
#endif

	pNodeTrailer = pNode = smpd_process.pDatabase;
	while (pNode)
	{
		if (wcscmp(pNode->pszName, name) == 0)
		{
			while (pNode->pData)
			{
				pElement = pNode->pData;
				pNode->pData = pNode->pData->pNext;
				free(pElement);
			}
			if (pNodeTrailer == pNode)
				smpd_process.pDatabase = smpd_process.pDatabase->pNext;
			else
				pNodeTrailer->pNext = pNode->pNext;
			free(pNode);
#ifdef USE_WIN_MUTEX_PROTECT
			ReleaseMutex(smpd_process.hDBSMutex);
#endif
			smpd_exit_fn(FCNAME);
			return SMPD_DBS_SUCCESS;
		}
		pNodeTrailer = pNode;
		pNode = pNode->pNext;
	}

#ifdef USE_WIN_MUTEX_PROTECT
	ReleaseMutex(smpd_process.hDBSMutex);
#endif

	smpd_exit_fn(FCNAME);
	return SMPD_DBS_FAIL;
}

#undef FCNAME
#define FCNAME L"smpd_dbs_first"
int smpd_dbs_first(const wchar_t *name, wchar_t *key, size_t maxlen1, wchar_t *value, size_t maxlen2)
{
	smpd_database_node_t *pNode;

	smpd_enter_fn(FCNAME);

#ifdef USE_WIN_MUTEX_PROTECT
	WaitForSingleObject(smpd_process.hDBSMutex, INFINITE);
#endif

	if (key != NULL)
		key[0] = L'\0';
	pNode = smpd_process.pDatabase;
	while (pNode)
	{
		if (wcscmp(pNode->pszName, name) == 0)
		{
			if (key != NULL)
			{
				if (pNode->pData)
				{
					wcscpy_s(key, maxlen1, pNode->pData->pszKey);
					wcscpy_s(value, maxlen2, pNode->pData->pszValue);
					pNode->pIter = pNode->pData->pNext;
				}
				else
					key[0] = L'\0';
			}
			else
			{
				pNode->pIter = pNode->pData;
			}
#ifdef USE_WIN_MUTEX_PROTECT
			ReleaseMutex(smpd_process.hDBSMutex);
#endif
			smpd_exit_fn(FCNAME);
			return SMPD_DBS_SUCCESS;
		}
		pNode = pNode->pNext;
	}

#ifdef USE_WIN_MUTEX_PROTECT
	ReleaseMutex(smpd_process.hDBSMutex);
#endif

	smpd_exit_fn(FCNAME);
	return SMPD_DBS_FAIL;
}

#undef FCNAME
#define FCNAME L"smpd_dbs_next"
int smpd_dbs_next(const wchar_t *name, wchar_t *key, size_t maxlen1, wchar_t *value, size_t maxlen2)
{
	smpd_database_node_t *pNode;

	smpd_enter_fn(FCNAME);

#ifdef USE_WIN_MUTEX_PROTECT
	WaitForSingleObject(smpd_process.hDBSMutex, INFINITE);
#endif

	key[0] = L'\0';
	pNode = smpd_process.pDatabase;
	while (pNode)
	{
		if (wcscmp(pNode->pszName, name) == 0)
		{
			if (pNode->pIter)
			{
				wcscpy_s(key, maxlen1, pNode->pIter->pszKey);
				wcscpy_s(value, maxlen2, pNode->pIter->pszValue);
				pNode->pIter = pNode->pIter->pNext;
			}
			else
				key[0] = L'\0';
#ifdef USE_WIN_MUTEX_PROTECT
			ReleaseMutex(smpd_process.hDBSMutex);
#endif
			smpd_exit_fn(FCNAME);
			return SMPD_DBS_SUCCESS;
		}
		pNode = pNode->pNext;
	}

#ifdef USE_WIN_MUTEX_PROTECT
	ReleaseMutex(smpd_process.hDBSMutex);
#endif

	smpd_exit_fn(FCNAME);
	return SMPD_DBS_FAIL;
}

#undef FCNAME
#define FCNAME L"smpd_dbs_firstdb"
int smpd_dbs_firstdb(wchar_t *name, size_t maxlen)
{
	smpd_enter_fn(FCNAME);

#ifdef USE_WIN_MUTEX_PROTECT
	WaitForSingleObject(smpd_process.hDBSMutex, INFINITE);
#endif

	smpd_process.pDatabaseIter = smpd_process.pDatabase;
	if (name != NULL)
	{
		if (smpd_process.pDatabaseIter)
			wcscpy_s(name, maxlen, smpd_process.pDatabaseIter->pszName);
		else
			name[0] = L'\0';
	}

#ifdef USE_WIN_MUTEX_PROTECT
	ReleaseMutex(smpd_process.hDBSMutex);
#endif

	smpd_exit_fn(FCNAME);
	return SMPD_DBS_SUCCESS;
}

#undef FCNAME
#define FCNAME L"smpd_dbs_nextdb"
int smpd_dbs_nextdb(wchar_t *name, size_t maxlen)
{
	smpd_enter_fn(FCNAME);

#ifdef USE_WIN_MUTEX_PROTECT
	WaitForSingleObject(smpd_process.hDBSMutex, INFINITE);
#endif

	if (smpd_process.pDatabaseIter == NULL)
		name[0] = L'\0';
	else
	{
		smpd_process.pDatabaseIter = smpd_process.pDatabaseIter->pNext;
		if (smpd_process.pDatabaseIter)
			wcscpy_s(name, maxlen, smpd_process.pDatabaseIter->pszName);
		else
			name[0] = L'\0';
	}

#ifdef USE_WIN_MUTEX_PROTECT
	ReleaseMutex(smpd_process.hDBSMutex);
#endif

	smpd_exit_fn(FCNAME);
	return SMPD_DBS_SUCCESS;
}
