/*   Copyright (C) 2009 Deino Software. */

#include "smpd.h"

typedef struct smpd_job_key_list_t
{
	wchar_t key[SMPD_MAX_NAME_LENGTH];
	wchar_t username[SMPD_MAX_ACCOUNT_LENGTH];
	wchar_t domain[SMPD_MAX_ACCOUNT_LENGTH];
	wchar_t full_domain[SMPD_MAX_ACCOUNT_LENGTH];
	HANDLE user_handle;
	HANDLE job;
	struct smpd_job_key_list_t *next;
} smpd_job_key_list_t;

static smpd_job_key_list_t *list = NULL;

#if 0 /* check for the key and username */
#undef FCNAME
#define FCNAME L"job_key_exists"
static SMPD_BOOL job_key_exists(const wchar_t *key, const wchar_t *username, const wchar_t *domain, const wchar_t *full_domain)
{
	smpd_job_key_list_t *iter;
	wchar_t account_str[SMPD_MAX_ACCOUNT_LENGTH];
	wchar_t domain_str[SMPD_MAX_ACCOUNT_LENGTH];

	smpd_enter_fn(FCNAME);

	if (domain == NULL && full_domain == NULL)
	{
		smpd_parse_account_domain(username, account_str, domain_str);
		if (domain_str[0] != L'\0')
		{
			username = account_str;
			domain = domain_str;
		}
	}
	iter = list;
	while (iter)
	{
		if (wcscmp(iter->key, key) == 0)
		{
			/* key matches */
			if (wcscmp(iter->username, username) == 0)
			{
				/* username matches */
				if ((domain != NULL && _wcsicmp(domain, iter->domain) == 0) ||				/* domain matches */
					(full_domain != NULL && _wcsicmp(full_domain, iter->full_domain) == 0) || /* full domain name matches */
					(iter->domain[0] == L'\0' && iter->full_domain[0] == L'\0') ||			 /* no domain to match */
					(domain == NULL && full_domain == NULL))								 /* no domain to match */
				{
					/* domain matches */
					smpd_exit_fn(FCNAME);
					return SMPD_TRUE;
				}
			}
		}
		iter = iter->next;
	}

	smpd_exit_fn(FCNAME);
	return SMPD_FALSE;
}
#endif

#undef FCNAME
#define FCNAME L"job_key_exists"
static SMPD_BOOL job_key_exists(const wchar_t *key)
{
	smpd_job_key_list_t *iter;

	smpd_enter_fn(FCNAME);

	iter = list;
	while (iter)
	{
		if (wcscmp(iter->key, key) == 0)
		{
			/* key matches */
			smpd_exit_fn(FCNAME);
			return SMPD_TRUE;
		}
		iter = iter->next;
	}

	smpd_exit_fn(FCNAME);
	return SMPD_FALSE;
}

#undef FCNAME
#define FCNAME L"smpd_add_job_key"
int smpd_add_job_key(const wchar_t *key, const wchar_t *username, const wchar_t *domain, const wchar_t *full_domain)
{
	int error;
	smpd_job_key_list_t *node;

	smpd_enter_fn(FCNAME);

	if (job_key_exists(key/*, username, domain, full_domain*/))
	{
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}

	node = (smpd_job_key_list_t*)malloc(sizeof(smpd_job_key_list_t));
	if (node == NULL)
	{
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	wcscpy_s(node->key, SMPD_MAX_NAME_LENGTH, key);
	node->domain[0] = L'\0';
	node->full_domain[0] = L'\0';
	if (domain == NULL && full_domain == NULL)
	{
		smpd_parse_account_domain(username, node->username, SMPD_MAX_ACCOUNT_LENGTH, node->domain, SMPD_MAX_ACCOUNT_LENGTH);
		node->full_domain[0] = L'\0';
	}
	else
	{
		wcscpy_s(node->username, SMPD_MAX_ACCOUNT_LENGTH, username);
	}
	if (domain != NULL)
		wcscpy_s(node->domain, SMPD_MAX_ACCOUNT_LENGTH, domain);
	if (full_domain != NULL)
		wcscpy_s(node->full_domain, SMPD_MAX_ACCOUNT_LENGTH, full_domain);
	node->user_handle = INVALID_HANDLE_VALUE;
	node->job = CreateJobObject(NULL, NULL);
	if (node->job == NULL)
	{
		error = GetLastError();
		smpd_err_printf(L"CreateJobObject failed: %d\n", error);
		free(node);
		smpd_exit_fn(FCNAME);
		return error;
	}
	node->next = list;
	list = node;

	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}

#undef FCNAME
#define FCNAME L"smpd_add_job_key"
int smpd_add_job_key_and_handle(const wchar_t *key, const wchar_t *username, const wchar_t *domain, const wchar_t *full_domain, HANDLE hUser)
{
	int error;
	smpd_job_key_list_t *node;

	smpd_enter_fn(FCNAME);

	if (job_key_exists(key/*, username, domain, full_domain*/))
	{
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}

	node = (smpd_job_key_list_t*)malloc(sizeof(smpd_job_key_list_t));
	if (node == NULL)
	{
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	wcscpy_s(node->key, SMPD_MAX_NAME_LENGTH, key);
	node->domain[0] = L'\0';
	node->full_domain[0] = L'\0';
	if (domain == NULL && full_domain == NULL)
	{
		smpd_parse_account_domain(username, node->username, SMPD_MAX_ACCOUNT_LENGTH, node->domain, SMPD_MAX_ACCOUNT_LENGTH);
		node->full_domain[0] = L'\0';
	}
	else
	{
		wcscpy_s(node->username, SMPD_MAX_ACCOUNT_LENGTH, username);
	}
	if (domain != NULL)
		wcscpy_s(node->domain, SMPD_MAX_ACCOUNT_LENGTH, domain);
	if (full_domain != NULL)
		wcscpy_s(node->full_domain, SMPD_MAX_ACCOUNT_LENGTH, full_domain);
	node->user_handle = hUser;
	node->job = CreateJobObject(NULL, NULL);
	if (node->job == NULL)
	{
		error = GetLastError();
		smpd_err_printf(L"CreateJobObject failed: %d\n", error);
		free(node);
		smpd_exit_fn(FCNAME);
		return error;
	}
	node->next = list;
	list = node;

	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}

#undef FCNAME
#define FCNAME L"smpd_remove_job_key"
int smpd_remove_job_key(const wchar_t *key)
{
	smpd_job_key_list_t *iter, *trailer;

	smpd_enter_fn(FCNAME);

	iter = trailer = list;
	while (iter)
	{
		if (wcscmp(iter->key, key) == 0)
		{
			if (iter == list)
			{
				list = list->next;
			}
			else
			{
				trailer->next = iter->next;
			}
			if (iter->user_handle != INVALID_HANDLE_VALUE)
				CloseHandle(iter->user_handle);
			if (iter->job != NULL && iter->job != INVALID_HANDLE_VALUE)
			{
				TerminateJobObject(iter->job, (UINT)-1);
				CloseHandle(iter->job);
			}
			free(iter);
			smpd_exit_fn(FCNAME);
			return SMPD_SUCCESS;
		}
		if (trailer != iter)
			trailer = trailer->next;
		iter = iter->next;
	}

	smpd_exit_fn(FCNAME);
	return SMPD_FAIL;
}

#undef FCNAME
#define FCNAME L"smpd_associate_job_key"
int smpd_associate_job_key(const wchar_t *key, const wchar_t *username, const wchar_t *domain, const wchar_t *full_domain, HANDLE user_handle)
{
	smpd_job_key_list_t *iter;

	smpd_enter_fn(FCNAME);

	iter = list;
	while (iter)
	{
		if (wcscmp(iter->key, key) == 0)
		{
			/* key matches */
			if (wcscmp(iter->username, username) == 0)
			{
				/* username matches */
				if ((domain != NULL && _wcsicmp(domain, iter->domain) == 0) ||				/* domain matches */
					(full_domain != NULL && _wcsicmp(full_domain, iter->full_domain) == 0) || /* full domain name matches */
					(iter->domain[0] == L'\0' && iter->full_domain[0] == L'\0'))			   /* no domain to match */
				{
					/* domain matches */
					if (iter->user_handle == INVALID_HANDLE_VALUE)
					{
						/* handle not already set */
						smpd_dbg_printf(L"associated user token: %s,%p\n", username, user_handle);
						iter->user_handle = user_handle;
						smpd_exit_fn(FCNAME);
						return SMPD_SUCCESS;
					}
				}
			}
		}
		iter = iter->next;
	}

	smpd_exit_fn(FCNAME);
	return SMPD_FAIL;
}

#undef FCNAME
#define FCNAME L"smpd_lookup_job_key"
int smpd_lookup_job_key(const wchar_t *key, const wchar_t *username, HANDLE *user_handle, HANDLE *job_handle)
{
	smpd_job_key_list_t *iter;
	wchar_t account[SMPD_MAX_ACCOUNT_LENGTH];
	wchar_t domain[SMPD_MAX_ACCOUNT_LENGTH];

	smpd_enter_fn(FCNAME);

	smpd_parse_account_domain(username, account, SMPD_MAX_ACCOUNT_LENGTH, domain, SMPD_MAX_ACCOUNT_LENGTH);

	iter = list;
	while (iter)
	{
		if (wcscmp(iter->key, key) == 0)
		{
			/* key matches */
			if (_wcsicmp(iter->username, account) == 0)
			{
				/* username matches */
				if ((_wcsicmp(domain, iter->domain) == 0) ||
					(_wcsicmp(domain, iter->full_domain) == 0) ||
					(iter->domain[0] == L'\0' && iter->full_domain[0] == L'\0'))
				{
					/* domain matches */
					if (iter->user_handle != INVALID_HANDLE_VALUE)
					{
						*user_handle = iter->user_handle;
						*job_handle = iter->job;
						smpd_exit_fn(FCNAME);
						return SMPD_SUCCESS;
					}
				}
			}
		}
		iter = iter->next;
	}

	smpd_exit_fn(FCNAME);
	return SMPD_FAIL;
}
