/*   Copyright (C) 2009 Deino Software. */

#include "smpd.h"
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif
#ifdef HAVE_SYS_STAT_H
/* This is needed for umask */
#include <sys/stat.h>
#endif
#include <stdlib.h>
#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#ifdef HAVE_SIGNAL_H
#include <signal.h>
#endif

#define MAX_ERROR_LENGTH 512

#undef FCNAME
#define FCNAME L"smpd_clear_process_registry"
int smpd_clear_process_registry()
{
	HKEY tkey;
	DWORD dwLen, result;
	int i;
	DWORD dwNumSubKeys, dwMaxSubKeyLen;
	wchar_t pid_str[256];
	wchar_t err_msg[MAX_ERROR_LENGTH] = L"";

	smpd_enter_fn(FCNAME);

	result = RegOpenKeyEx(HKEY_LOCAL_MACHINE, SMPD_REGISTRY_KEY L"\\process", 0, KEY_ALL_ACCESS, &tkey);
	if (result != ERROR_SUCCESS)
	{
		if (result != ERROR_PATH_NOT_FOUND)
		{
			smpd_err_printf(L"Unable to open the " SMPD_REGISTRY_KEY L"\\process registry key, error %d\n", result);
			smpd_exit_fn(FCNAME);
			return SMPD_FAIL;
		}
		return SMPD_SUCCESS;
	}

	result = RegQueryInfoKey(tkey, NULL, NULL, NULL, &dwNumSubKeys, &dwMaxSubKeyLen, NULL, NULL, NULL, NULL, NULL, NULL);
	if (result != ERROR_SUCCESS)
	{
		RegCloseKey(tkey);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	if (dwMaxSubKeyLen > 256)
	{
		smpd_err_printf(L"Error: Invalid process subkeys, max length is too large: %d\n", dwMaxSubKeyLen);
		RegCloseKey(tkey);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	if (dwNumSubKeys == 0)
	{
		result = RegCloseKey(tkey);
		if (result != ERROR_SUCCESS)
		{
			smpd_translate_win_error(result, err_msg, MAX_ERROR_LENGTH, NULL);
			smpd_err_printf(L"Error: RegCloseKey(HKEY_LOCAL_MACHINE\\" SMPD_REGISTRY_KEY L"\\process) failed, %s\n", err_msg);
			smpd_exit_fn(FCNAME);
			return SMPD_FAIL;
		}
		result = RegDeleteKey(HKEY_LOCAL_MACHINE, SMPD_REGISTRY_KEY L"\\process");
		if (result != ERROR_SUCCESS)
		{
			smpd_translate_win_error(result, err_msg, MAX_ERROR_LENGTH, NULL);
			smpd_err_printf(L"Error: Unable to remove the HKEY_LOCAL_MACHINE\\" SMPD_REGISTRY_KEY L"\\process registry key, %s\n", err_msg);
			smpd_exit_fn(FCNAME);
			return SMPD_FAIL;
		}
		smpd_exit_fn(FCNAME);
		return SMPD_SUCCESS;
	}
	/* count backwards so keys can be removed */
	for (i=dwNumSubKeys-1; i>=0; i--)
	{
		dwLen = 256;
		result = RegEnumKeyEx(tkey, i, pid_str, &dwLen, NULL, NULL, NULL, NULL);
		if (result != ERROR_SUCCESS)
		{
			smpd_err_printf(L"Error: Unable to enumerate the %d subkey in the " SMPD_REGISTRY_KEY L"\\process registry key\n", i);
			RegCloseKey(tkey);
			smpd_exit_fn(FCNAME);
			return SMPD_FAIL;
		}
		result = RegDeleteKey(tkey, pid_str);
		if (result != ERROR_SUCCESS)
		{
			smpd_translate_win_error(result, err_msg, MAX_ERROR_LENGTH, NULL);
			smpd_err_printf(L"Error: RegDeleteKey(HKEY_LOCAL_MACHINE\\" SMPD_REGISTRY_KEY L"\\process\\%s) failed, %s\n", pid_str, err_msg);
			smpd_exit_fn(FCNAME);
			return SMPD_FAIL;
		}
	}
	result = RegCloseKey(tkey);
	if (result != ERROR_SUCCESS)
	{
		smpd_translate_win_error(result, err_msg, MAX_ERROR_LENGTH, NULL);
		smpd_err_printf(L"Error: RegCloseKey(HKEY_LOCAL_MACHINE\\" SMPD_REGISTRY_KEY L") failed, %s\n", err_msg);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	result = RegDeleteKey(HKEY_LOCAL_MACHINE, SMPD_REGISTRY_KEY L"\\process");
	if (result != ERROR_SUCCESS)
	{
		smpd_translate_win_error(result, err_msg, MAX_ERROR_LENGTH, NULL);
		smpd_err_printf(L"Error: Unable to remove the HKEY_LOCAL_MACHINE\\" SMPD_REGISTRY_KEY L"\\process registry key, %s\n", err_msg);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}

#undef FCNAME
#define FCNAME L"smpd_validate_process_registry"
int smpd_validate_process_registry()
{
	int error;
	HKEY tkey;
	DWORD dwLen, result;
	int i;
	DWORD dwNumSubKeys, dwMaxSubKeyLen;
	wchar_t pid_str[100];
	int pid;
	HANDLE hTemp;
	wchar_t err_msg[MAX_ERROR_LENGTH] = L"";

	smpd_enter_fn(FCNAME);

	result = RegOpenKeyEx(HKEY_LOCAL_MACHINE, SMPD_REGISTRY_KEY L"\\process", 0, KEY_ALL_ACCESS, &tkey);
	if (result != ERROR_SUCCESS)
	{
		if (result != ERROR_PATH_NOT_FOUND)
		{
			smpd_translate_win_error(result, err_msg, MAX_ERROR_LENGTH, NULL);
			smpd_err_printf(L"Unable to open the smpd\\process registry key, error %d, %s\n", result, err_msg);
			smpd_exit_fn(FCNAME);
			return SMPD_FAIL;
		}
		return SMPD_SUCCESS;
	}

	result = RegQueryInfoKey(tkey, NULL, NULL, NULL, &dwNumSubKeys, &dwMaxSubKeyLen, NULL, NULL, NULL, NULL, NULL, NULL);
	if (result != ERROR_SUCCESS)
	{
		RegCloseKey(tkey);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	if (dwMaxSubKeyLen > 100)
	{
		smpd_err_printf(L"Error: Invalid process subkeys, max length is too large: %d\n", dwMaxSubKeyLen);
		RegCloseKey(tkey);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	if (dwNumSubKeys == 0)
	{
		RegCloseKey(tkey);
		smpd_exit_fn(FCNAME);
		return SMPD_SUCCESS;
	}
	/* count backwards so keys can be removed */
	for (i=dwNumSubKeys-1; i>=0; i--)
	{
		dwLen = 100;
		result = RegEnumKeyEx(tkey, i, pid_str, &dwLen, NULL, NULL, NULL, NULL);
		if (result != ERROR_SUCCESS)
		{
			smpd_translate_win_error(result, err_msg, MAX_ERROR_LENGTH, NULL);
			smpd_err_printf(L"Error: Unable to enumerate the %d subkey in the smpd\\process registry key, %s\n", i, err_msg);
			RegCloseKey(tkey);
			smpd_exit_fn(FCNAME);
			return SMPD_FAIL;
		}
		pid = _wtoi(pid_str);
		wprintf(L"pid = %d\n", pid);fflush(stdout);
		hTemp = OpenProcess(PROCESS_QUERY_INFORMATION, FALSE, pid);
		if (hTemp == NULL)
		{
			error = GetLastError();
			if (error == ERROR_INVALID_PARAMETER)
			{
				RegDeleteKey(tkey, pid_str);
			}
			/*
			else
			{
				wprintf(L"error = %d\n", error);
			}
			*/
		}
		else
		{
			CloseHandle(hTemp);
		}
	}
	result = RegCloseKey(tkey);
	if (result != ERROR_SUCCESS)
	{
		smpd_translate_win_error(result, err_msg, MAX_ERROR_LENGTH, NULL);
		smpd_err_printf(L"Error: RegCloseKey(HKEY_LOCAL_MACHINE\\" SMPD_REGISTRY_KEY L"\\process) failed, %s\n", err_msg);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}

#undef FCNAME
#define FCNAME L"smpd_process_to_registry"
int smpd_process_to_registry(smpd_process_t *process, wchar_t *actual_exe)
{
	HKEY tkey;
	DWORD len, result;
	wchar_t name[1024];
	wchar_t err_msg[MAX_ERROR_LENGTH] = L"";

	smpd_enter_fn(FCNAME);

	if (process == NULL)
	{
		smpd_dbg_printf(L"NULL process passed to smpd_process_to_registry.\n");
		return SMPD_FAIL;
	}

	len = _snwprintf_s(name, 1024, _TRUNCATE, SMPD_REGISTRY_KEY L"\\process\\%d", process->pid);
	if (len < 0 || len > 1023)
	{
		smpd_dbg_printf(L"unable to create a string of the registry key.\n");
		return SMPD_FAIL;
	}

	result = RegCreateKeyEx(HKEY_LOCAL_MACHINE, name,
		0, NULL, REG_OPTION_VOLATILE, KEY_ALL_ACCESS, NULL, &tkey, NULL);
	if (result != ERROR_SUCCESS)
	{
		smpd_translate_win_error(result, err_msg, MAX_ERROR_LENGTH, NULL);
		smpd_err_printf(L"Unable to open the HKEY_LOCAL_MACHINE\\%s registry key, error %d, %s\n", name, result, err_msg);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}

	len = (DWORD)(wcslen(actual_exe)+1)*sizeof(wchar_t);
	result = RegSetValueEx(tkey, L"exe", 0, REG_SZ, (const BYTE *)actual_exe, len);
	if (result != ERROR_SUCCESS)
	{
		smpd_translate_win_error(result, err_msg, MAX_ERROR_LENGTH, NULL);
		smpd_err_printf(L"Unable to write the process registry value 'exe:%s', error %d, %s\n", process->exe, result, err_msg);
		RegCloseKey(tkey);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}

	result = RegCloseKey(tkey);
	if (result != ERROR_SUCCESS)
	{
		smpd_translate_win_error(result, err_msg, MAX_ERROR_LENGTH, NULL);
		smpd_err_printf(L"Error: RegCloseKey(HKEY_LOCAL_MACHINE\\" SMPD_REGISTRY_KEY L"\\process\\%d) failed, %s\n", process->pid, err_msg);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}

#undef FCNAME
#define FCNAME L"smpd_process_from_registry"
int smpd_process_from_registry(smpd_process_t *process)
{
	DWORD len, result;
	wchar_t name[1024];
	wchar_t err_msg[MAX_ERROR_LENGTH] = L"";

	smpd_enter_fn(FCNAME);

	if (process == NULL)
	{
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}

	len = _snwprintf_s(name, 1024, _TRUNCATE, SMPD_REGISTRY_KEY L"\\process\\%d", process->pid);
	if (len < 0 || len > 1023)
	{
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}

	result = RegDeleteKey(HKEY_LOCAL_MACHINE, name);
	if (result != ERROR_SUCCESS)
	{
		if (result == ERROR_FILE_NOT_FOUND || result == ERROR_PATH_NOT_FOUND)
		{
			/* Key already deleted, return success */
			smpd_exit_fn(FCNAME);
			return SMPD_SUCCESS;
		}
		smpd_translate_win_error(result, err_msg, MAX_ERROR_LENGTH, NULL);
		smpd_err_printf(L"Unable to delete the HKEY_LOCAL_MACHINE\\%s registry key, error %d\n", name, result, err_msg);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}

	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}

#undef FCNAME
#define FCNAME L"smpd_get_user_handle"
int smpd_get_user_handle(wchar_t *account, wchar_t *domain, wchar_t *password, HANDLE *handle_ptr)
{
	HANDLE hUser;
	int error;
	int num_tries = 3;

	smpd_enter_fn(FCNAME);

	if (domain)
	{
		smpd_dbg_printf(L"LogonUser(%s\\%s)\n", domain, account);
	}
	else
	{
		smpd_dbg_printf(L"LogonUser(%s)\n", account);
	}

	/* logon the user */
	while (!LogonUser(
		account,
		domain,
		password,
		LOGON32_LOGON_INTERACTIVE,
		LOGON32_PROVIDER_DEFAULT,
		&hUser))
	{
		error = GetLastError();
		if (error == ERROR_NO_LOGON_SERVERS)
		{
			if (num_tries)
				Sleep(250);
			else
			{
				*handle_ptr = INVALID_HANDLE_VALUE;
				smpd_exit_fn(FCNAME);
				return error;
			}
			num_tries--;
		}
		else
		{
			*handle_ptr = INVALID_HANDLE_VALUE;
			smpd_exit_fn(FCNAME);
			return error;
		}
	}

	*handle_ptr = hUser;
	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}

#undef FCNAME
#define FCNAME L"smpd_get_user_name"
int smpd_get_user_name(wchar_t *account, size_t maxlen_account, wchar_t *domain, size_t maxlen_domain, wchar_t *full_domain, size_t maxlen_full_domain)
{
	DWORD len;
	wchar_t name[SMPD_MAX_ACCOUNT_LENGTH];
	wchar_t *separator;
	size_t i;

	*account = L'\0';
	if (domain != NULL)
		*domain = L'\0';
	if (full_domain != NULL)
		*full_domain = L'\0';

	len = 100;
	if (GetUserNameEx(NameSamCompatible, name, &len))
	{
		for (i=0; i<wcslen(name); i++)
		{
			name[i] = (wchar_t)(towlower(name[i]));
		}
		separator = wcschr(name, L'\\');
		if (separator)
		{
			*separator = L'\0';
			separator++;
		}
		if (domain != NULL)
			wcscpy_s(domain, maxlen_domain, name);
		wcscpy_s(account, maxlen_account, separator);
	}

	if (full_domain != NULL)
	{
		len = 100;
		if (GetUserNameEx(NameDnsDomain, name, &len))
		{
			for (i=0; i<wcslen(name); i++)
			{
				name[i] = (wchar_t)(towlower(name[i]));
			}
			separator = wcschr(name, L'\\');
			if (separator)
			{
				*separator = L'\0';
				wcscpy_s(full_domain, maxlen_full_domain, name);
			}
		}
	}

	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}

#undef FCNAME
#define FCNAME L"SetEnvironmentVariables"
static void SetEnvironmentVariables(wchar_t *bEnv)
{
	wchar_t name[SMPD_MAX_ENV_LENGTH], equals[3], value[SMPD_MAX_ENV_LENGTH];

	smpd_enter_fn(FCNAME);
	for (;;)
	{
		name[0] = L'\0';
		equals[0] = L'\0';
		value[0] = L'\0';
		if (MPIU_Str_get_string(&bEnv, name, SMPD_MAX_ENV_LENGTH) != MPIU_STR_SUCCESS)
			break;
		if (name[0] == L'\0')
			break;
		if (MPIU_Str_get_string(&bEnv, equals, 3) != MPIU_STR_SUCCESS)
			break;
		if (equals[0] == L'\0')
			break;
		if (MPIU_Str_get_string(&bEnv, value, SMPD_MAX_ENV_LENGTH) != MPIU_STR_SUCCESS)
			break;
		smpd_dbg_printf(L"setting environment variable: <%s> = <%s>\n", name, value);
		if (!SetEnvironmentVariable(name, value))
		{
			smpd_err_printf(L"Unable to set environment variable <%s> = <%s>: Error %d\n", name, value, GetLastError());
		}
	}
	smpd_exit_fn(FCNAME);
}

#undef FCNAME
#define FCNAME L"RemoveEnvironmentVariables"
static void RemoveEnvironmentVariables(wchar_t *bEnv)
{
	wchar_t name[SMPD_MAX_ENV_LENGTH], equals[3], value[SMPD_MAX_ENV_LENGTH];

	smpd_enter_fn(FCNAME);
	for (;;)
	{
		name[0] = L'\0';
		equals[0] = L'\0';
		value[0] = L'\0';
		if (MPIU_Str_get_string(&bEnv, name, SMPD_MAX_ENV_LENGTH) != MPIU_STR_SUCCESS)
			break;
		if (name[0] == L'\0')
			break;
		if (MPIU_Str_get_string(&bEnv, equals, 3) != MPIU_STR_SUCCESS)
			break;
		if (equals[0] == L'\0')
			break;
		if (MPIU_Str_get_string(&bEnv, value, SMPD_MAX_ENV_LENGTH) != MPIU_STR_SUCCESS)
			break;
		/*smpd_dbg_printf(L"removing environment variable <%s>\n", name);*/
		if (!SetEnvironmentVariable(name, NULL))
		{
			smpd_err_printf(L"Unable to remove environment variable <%s>: Error %d\n", name, GetLastError());
		}
	}
	smpd_exit_fn(FCNAME);
}

#undef FCNAME
#define FCNAME L"smpd_priority_class_to_win_class"
int smpd_priority_class_to_win_class(int *priorityClass)
{
	smpd_enter_fn(FCNAME);
	switch (*priorityClass)
	{
	case 0:
		*priorityClass = IDLE_PRIORITY_CLASS;
		break;
	case 1:
		*priorityClass = BELOW_NORMAL_PRIORITY_CLASS;
		break;
	case 2:
		*priorityClass = NORMAL_PRIORITY_CLASS;
		break;
	case 3:
		*priorityClass = ABOVE_NORMAL_PRIORITY_CLASS;
		break;
	case 4:
		*priorityClass = HIGH_PRIORITY_CLASS;
		break;
	default:
		*priorityClass = NORMAL_PRIORITY_CLASS;
		break;
	}
	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}

#undef FCNAME
#define FCNAME L"smpd_priority_to_win_priority"
int smpd_priority_to_win_priority(int *priority)
{
	smpd_enter_fn(FCNAME);
	switch (*priority)
	{
	case 0:
		*priority = THREAD_PRIORITY_IDLE;
		break;
	case 1:
		*priority = THREAD_PRIORITY_LOWEST;
		break;
	case 2:
		*priority = THREAD_PRIORITY_BELOW_NORMAL;
		break;
	case 3:
		*priority = THREAD_PRIORITY_NORMAL;
		break;
	case 4:
		*priority = THREAD_PRIORITY_ABOVE_NORMAL;
		break;
	case 5:
		*priority = THREAD_PRIORITY_HIGHEST;
		break;
	default:
		*priority = THREAD_PRIORITY_NORMAL;
		break;
	}
	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}

/* Windows code */

typedef struct smpd_piothread_arg_t
{
	HANDLE hIn;
	SOCKET hOut;
	int pid;
} smpd_piothread_arg_t;

typedef struct smpd_pinthread_arg_t
{
	SOCKET hIn;
	HANDLE hOut;
	int pid;
} smpd_pinthread_arg_t;

#undef FCNAME
#define FCNAME L"smpd_easy_send"
static int smpd_easy_send(SOCKET sock, char *buffer, int length)
{
	int error;
	int num_sent, num_left;

	smpd_exit_fn(FCNAME);
	num_left = length;
	while (num_left)
	{
		while ((num_sent = send(sock, buffer, num_left, 0)) == SOCKET_ERROR)
		{
			error = WSAGetLastError();
			if (error == WSAEWOULDBLOCK)
			{
				Sleep(0);
				continue;
			}
			if (error == WSAENOBUFS)
			{
				/* If there is no buffer space available then split the buffer in half and send each piece separately.*/
				if (smpd_easy_send(sock, buffer, num_left/2) == SOCKET_ERROR)
				{
					smpd_exit_fn(FCNAME);
					return SOCKET_ERROR;
				}
				if (smpd_easy_send(sock, buffer+(num_left/2), num_left - (num_left/2)) == SOCKET_ERROR)
				{
					smpd_exit_fn(FCNAME);
					return SOCKET_ERROR;
				}
				smpd_exit_fn(FCNAME);
				return length;
			}
			WSASetLastError(error);
			smpd_exit_fn(FCNAME);
			return SOCKET_ERROR;
		}
		num_left = num_left - num_sent;
		buffer = buffer + num_sent;
	}
	smpd_exit_fn(FCNAME);
	return length;
}

int smpd_piothread(smpd_piothread_arg_t *p)
{
	char buffer[8192];
	DWORD num_read;
	HANDLE hIn;
	SOCKET hOut;
	DWORD error;
	char bogus_char;
	int pid;
	double t1, t2;

	hIn = p->hIn;
	hOut = p->hOut;
	pid = p->pid;
	free(p);
	p = NULL;

	smpd_dbg_printf(L"*** entering smpd_piothread pid:%d sock:%d ***\n", pid, hOut);
	for (;;)
	{
		num_read = 0;
		if (!ReadFile(hIn, buffer, 8192, &num_read, NULL))
		{
			error = GetLastError();
			/* If there was an error but some bytes were read, send those bytes before exiting */
			if (num_read > 0)
			{
				if (smpd_easy_send(hOut, buffer, num_read) == SOCKET_ERROR)
				{
					smpd_dbg_printf(L"smpd_easy_send of %d bytes failed.\n", num_read);
					break;
				}
			}
			smpd_dbg_printf(L"ReadFile failed, error %d\n", error);
			break;
		}
		if (num_read < 1)
		{
			smpd_dbg_printf(L"ReadFile returned %d bytes\n", num_read);
			break;
		}
		/*smpd_dbg_printf(L"*** smpd_piothread read %d bytes ***\n", num_read);*/
		if (smpd_easy_send(hOut, buffer, num_read) == SOCKET_ERROR)
		{
			smpd_dbg_printf(L"smpd_easy_send of %d bytes failed.\n", num_read);
			break;
		}
		/*smpd_dbg_printf(L"*** smpd_piothread wrote %d bytes ***\n", num_read);*/
	}
	smpd_dbg_printf(L"*** smpd_piothread finishing pid:%d ***\n", pid);
	/*
	FlushFileBuffers((HANDLE)hOut);
	if (shutdown(hOut, SD_BOTH) == SOCKET_ERROR)
	{
		smpd_err_printf(L"shutdown failed, sock %d, error %d\n", hOut, WSAGetLastError());
	}
	if (closesocket(hOut) == SOCKET_ERROR)
	{
		smpd_err_printf(L"closesocket failed, sock %d, error %d\n", hOut, WSAGetLastError());
	}
	*/
	if (shutdown(hOut, SD_SEND) == SOCKET_ERROR)
	{
		smpd_err_printf(L"shutdown failed, sock %d, error %d\n", hOut, WSAGetLastError());
	}
	t1 = PMPI_Wtime();
	recv(hOut, &bogus_char, 1, 0);
	if (closesocket(hOut) == SOCKET_ERROR)
	{
		smpd_err_printf(L"closesocket failed, sock %d, error %d\n", hOut, WSAGetLastError());
	}
	t2 = PMPI_Wtime();
	smpd_dbg_printf(L"closing output socket took %.3f seconds\n", t2-t1);
	CloseHandle(hIn);
	/*smpd_dbg_printf(L"*** exiting smpd_piothread ***\n");*/
	return 0;
}

/* one line at a time version */
#define ISODD(a) ((int)(a) & 0x1)
int smpd_pinthread(smpd_pinthread_arg_t *p)
{
	char str[SMPD_MAX_CMD_LENGTH];
	wchar_t *wchar;
	int index;
	DWORD num_written;
	int num_read;
	SOCKET hIn;
	HANDLE hOut;
	int pid;
	/*int i;*/
	/*
	wchar_t bogus_char;
	double t1, t2;
	*/

	hIn = p->hIn;
	hOut = p->hOut;
	pid = p->pid;
	free(p);
	p = NULL;

	smpd_dbg_printf(L"*** entering smpd_pinthread pid:%d sock:%d ***\n", pid, hIn);
	index = 0;
	for (;;)
	{
		num_read = recv(hIn, &str[index], 1, 0);
		/*num_read = recv(hIn, (char*)&str[index], 2, 0);*/
		if (num_read == SOCKET_ERROR || num_read == 0)
		{
			if (num_read == SOCKET_ERROR && WSAGetLastError() == WSAEWOULDBLOCK)
			{
				u_long optval = (u_long)TRUE;
				ioctlsocket(hIn, FIONBIO, &optval);
				continue;
			}
			if (index > 0)
			{
				/* write any buffered data before exiting */
				if (!WriteFile(hOut, str, index, &num_written, NULL))
				{
					smpd_dbg_printf(L"WriteFile failed, error %d\n", GetLastError());
					break;
				}
			}
			if (num_read != 0)
			{
				smpd_dbg_printf(L"recv from stdin socket failed, error %d.\n", WSAGetLastError());
			}
			break;
		}

		/* FIXME: How do you know if the input is UNICODE wchar_t or MBCS(ASCII) char characters? */
		if (ISODD(index))
		{
			wchar = (wchar_t*)&str[index-1];
		}
		else
		{
			wchar = NULL;
		}
		if (str[index] == L'\n' || (index == SMPD_MAX_CMD_LENGTH-1) || (wchar != NULL && *wchar == L'\n'))
		{
			smpd_dbg_printf(L"writing %d bytes to the process's stdin\n", index+1);
			if (!WriteFile(hOut, str, index+1, &num_written, NULL))
			{
				smpd_dbg_printf(L"WriteFile failed, error %d\n", GetLastError());
				break;
			}
			/*
			smpd_dbg_printf(L"wrote: ");
			for (i=0; i<=index; i++)
			{
				smpd_dbg_printf(L"(%d)'%c'", (int)str[i], str[i]);
			}
			smpd_dbg_printf(L"\n");
			*/
			index = 0;
		}
		else
		{
			/*smpd_dbg_printf(L"read character(%d)'%c'\n", (int)str[index], str[index]);*/
			smpd_dbg_printf(L"read character(%d)\n", (int)str[index]);
			index++;
		}
	}
	smpd_dbg_printf(L"*** smpd_pinthread finishing pid:%d ***\n", pid);
	FlushFileBuffers(hOut);
	if (shutdown(hIn, SD_BOTH) == SOCKET_ERROR)
	{
		smpd_err_printf(L"shutdown failed, sock %d, error %d\n", hIn, WSAGetLastError());
	}
	if (closesocket(hIn) == SOCKET_ERROR)
	{
		smpd_err_printf(L"closesocket failed, sock %d, error %d\n", hIn, WSAGetLastError());
	}
	/*
	if (shutdown(hIn, SD_SEND) == SOCKET_ERROR)
	{
		smpd_err_printf(L"shutdown failed, sock %d, error %d\n", hIn, WSAGetLastError());
	}
	t1 = PMPI_Wtime();
	recv(hIn, &bogus_char, 1, 0);
	if (closesocket(hIn) == SOCKET_ERROR)
	{
		smpd_err_printf(L"closesocket failed, sock %d, error %d\n", hIn, WSAGetLastError());
	}
	t2 = PMPI_Wtime();
	smpd_dbg_printf(L"closing input socket took %.3f seconds\n", t2-t1);
	*/
	CloseHandle(hOut);
	/*smpd_dbg_printf(L"*** exiting smpd_pinthread ***\n");*/
	return 0;
}
#if 0
/* 1 byte at a time version */
int smpd_pinthread(smpd_pinthread_arg_t *p)
{
	wchar_t ch;
	DWORD num_written;
	SOCKET hIn;
	HANDLE hOut;
	int num_read;

	hIn = p->hIn;
	hOut = p->hOut;
	free(p);
	p = NULL;

	smpd_dbg_printf(L"*** entering smpd_pinthread ***\n");
	for (;;)
	{
		num_read = recv(hIn, &ch, 1, 0);
		if (num_read == SOCKET_ERROR)
		{
			if (num_read == SOCKET_ERROR && WSAGetLastError() == WSAEWOULDBLOCK)
			{
				int optval = TRUE;
				ioctlsocket(hIn, FIONBIO, &optval);
				continue;
			}
			smpd_dbg_printf(L"recv from stdin socket failed, error %d.\n", WSAGetLastError());
			break;
		}
		if (num_read == 0)
		{
			break;
		}
		if (!WriteFile(hOut, &ch, 1, &num_written, NULL))
		{
			smpd_dbg_printf(L"WriteFile failed, error %d\n", GetLastError());
			break;
		}
	}
	smpd_dbg_printf(L"*** smpd_pinthread finishing ***\n");
	FlushFileBuffers(hOut);
	if (closesocket(hIn) == SOCKET_ERROR)
	{
		smpd_err_printf(L"closesocket failed, sock %d, error %d\n", hIn, WSAGetLastError());
	}
	CloseHandle(hOut);
	/*smpd_dbg_printf(L"*** exiting smpd_pinthread ***\n");*/
	return 0;
}
#endif

typedef struct HandleNode
{
	HANDLE handle;
	struct HandleNode *pNext;
} HandleNode;

static void AddHandleToList(HandleNode **pList, HANDLE h)
{
	HandleNode *p;

	p = (HandleNode*)MPIU_Malloc(sizeof(HandleNode));
	if (p != NULL)
	{
		p->handle = h;
		p->pNext = *pList;
		*pList = p;
	}
}

static void CloseHandleList(HandleNode *pList)
{
	HandleNode *p;
	while (pList)
	{
		p = pList;
		pList = pList->pNext;
		CloseHandle(p->handle);
		MPIU_Free(p);
	}
}

static DWORD DebugWaitForProcess(smpd_process_t *process)
{
	DEBUG_EVENT DebugEv;
	DWORD dwContinueStatus = DBG_CONTINUE;
	SMPD_BOOL bExit = SMPD_FALSE;
	HandleNode *pList = NULL;

	process->bAborted = SMPD_FALSE;
	if (!SetEvent(process->hDebugReady))
	{
		smpd_err_printf(L"unable to set the debug event.  Error %d\n", GetLastError());
		return (DWORD)-1;
	}

	for(;;) 
	{
		/* Wait for a debugging event to occur. The second parameter indicates 
		 * that the function does not return until a debugging event occurs. 
		 */

		if (!WaitForDebugEvent(&DebugEv, INFINITE))
		{
			smpd_err_printf(L"WaitForDebugEvent failed: Error %d\n", GetLastError());
			return (DWORD)-1;
		}

		/* Process the debugging event code. */

		switch (DebugEv.dwDebugEventCode)
		{
		case EXCEPTION_DEBUG_EVENT: 
			/*
			// Process the exception code. When handling 
			// exceptions, remember to set the continuation 
			// status parameter (dwContinueStatus). This value 
			// is used by the ContinueDebugEvent function.
			*/
			/*process->bAborted = (DebugEv.u.Exception.ExceptionRecord.ExceptionFlags == EXCEPTION_NONCONTINUABLE);*/
			dwContinueStatus = DBG_EXCEPTION_NOT_HANDLED;
			/* only abort on last chance exceptions to allow the user process to catch the exception */
			if (!DebugEv.u.Exception.dwFirstChance)
			{
				process->bAborted = SMPD_TRUE;
			}
			switch (DebugEv.u.Exception.ExceptionRecord.ExceptionCode)
			{
			case EXCEPTION_ACCESS_VIOLATION: 
				/*
				// First chance: Pass this on to the system. 
				// Last chance: Display an appropriate error.
				*/
				if (!DebugEv.u.Exception.dwFirstChance)
				{
					if (DebugEv.u.Exception.ExceptionRecord.NumberParameters == 2)
					{
						if (DebugEv.u.Exception.ExceptionRecord.ExceptionInformation[0] == 1)
						{
							/* write error */
							MPIU_Snprintf(process->dbgError, SMPD_MAX_ERROR_LEN, _TRUNCATE, L"Fatal error: EXCEPTION_ACCESS_VIOLATION: instruction address: 0x%p, invalid write to 0x%x",
								DebugEv.u.Exception.ExceptionRecord.ExceptionAddress,
								DebugEv.u.Exception.ExceptionRecord.ExceptionInformation[1]);
							smpd_dbg_printf(L"%s\n", process->dbgError);
						}
						else
						{
							/* read error */
							MPIU_Snprintf(process->dbgError, SMPD_MAX_ERROR_LEN, _TRUNCATE, L"Fatal error: EXCEPTION_ACCESS_VIOLATION: instruction address: 0x%p, invalid read from 0x%x",
								DebugEv.u.Exception.ExceptionRecord.ExceptionAddress,
								DebugEv.u.Exception.ExceptionRecord.ExceptionInformation[1]);
							smpd_dbg_printf(L"%s\n", process->dbgError);
						}
					}
					else
					{
						MPIU_Snprintf(process->dbgError, SMPD_MAX_ERROR_LEN, _TRUNCATE, L"Fatal error: EXCEPTION_ACCESS_VIOLATION: instruction address: 0x%p",
							DebugEv.u.Exception.ExceptionRecord.ExceptionAddress);
						smpd_dbg_printf(L"%s\n", process->dbgError);
					}
				}
				break;
			case EXCEPTION_BREAKPOINT:
				/*
				// First chance: Display the current 
				// instruction and register values. 
				//MPIU_Snprintf(process->dbgError, SMPD_MAX_ERROR_LEN, _TRUNCATE, L"EXCEPTION_BREAKPOINT: address: 0x%x", DebugEv.u.Exception.ExceptionRecord.ExceptionAddress);
				// I don't know what this exception is for but it is thrown for all processes good and bad, so just ignore it.
				*/
				process->bAborted = SMPD_FALSE;
				dwContinueStatus = DBG_CONTINUE;
				break;
			case EXCEPTION_DATATYPE_MISALIGNMENT: 
				/*
				// First chance: Pass this on to the system. 
				// Last chance: Display an appropriate error.
				*/
				MPIU_Snprintf(process->dbgError, SMPD_MAX_ERROR_LEN, _TRUNCATE, L"Fatal error: EXCEPTION_DATATYPE_MISALIGNMENT");
				smpd_dbg_printf(L"%s\n", process->dbgError);
				break;
			case EXCEPTION_SINGLE_STEP:
				/*
				// First chance: Update the display of the 
				// current instruction and register values.
				*/
				MPIU_Snprintf(process->dbgError, SMPD_MAX_ERROR_LEN, _TRUNCATE, L"Fatal error: EXCEPTION_SINGLE_STEP");
				smpd_dbg_printf(L"%s\n", process->dbgError);
				break;
			case DBG_CONTROL_C:
				/*
				// First chance: Pass this on to the system. 
				// Last chance: Display an appropriate error.
				*/
				MPIU_Snprintf(process->dbgError, SMPD_MAX_ERROR_LEN, _TRUNCATE, L"Fatal error: DBG_CONTROL_C");
				smpd_dbg_printf(L"%s\n", process->dbgError);
				break;
			case EXCEPTION_ARRAY_BOUNDS_EXCEEDED:
				MPIU_Snprintf(process->dbgError, SMPD_MAX_ERROR_LEN, _TRUNCATE, L"Fatal error: The thread tried to access an array element that is out of bounds and the underlying hardware supports bounds checking.");
				smpd_dbg_printf(L"%s\n", process->dbgError);
				break;
			case EXCEPTION_FLT_DENORMAL_OPERAND:
				MPIU_Snprintf(process->dbgError, SMPD_MAX_ERROR_LEN, _TRUNCATE, L"Fatal error: One of the operands in a floating-point operation is denormal. A denormal value is one that is too small to represent as a standard floating-point value.");
				smpd_dbg_printf(L"%s\n", process->dbgError);
				break;
			case EXCEPTION_FLT_DIVIDE_BY_ZERO:
				MPIU_Snprintf(process->dbgError, SMPD_MAX_ERROR_LEN, _TRUNCATE, L"Fatal error: The thread tried to divide a floating-point value by a floating-point divisor of zero.");
				smpd_dbg_printf(L"%s\n", process->dbgError);
				break;
			case EXCEPTION_FLT_INEXACT_RESULT:
				MPIU_Snprintf(process->dbgError, SMPD_MAX_ERROR_LEN, _TRUNCATE, L"Fatal error: The result of a floating-point operation cannot be represented exactly as a decimal fraction.");
				smpd_dbg_printf(L"%s\n", process->dbgError);
				break;
			case EXCEPTION_FLT_INVALID_OPERATION:
				MPIU_Snprintf(process->dbgError, SMPD_MAX_ERROR_LEN, _TRUNCATE, L"Fatal error: This exception represents any floating-point exception not included in this list.");
				smpd_dbg_printf(L"%s\n", process->dbgError);
				break;
			case EXCEPTION_FLT_OVERFLOW:
				MPIU_Snprintf(process->dbgError, SMPD_MAX_ERROR_LEN, _TRUNCATE, L"Fatal error: The exponent of a floating-point operation is greater than the magnitude allowed by the corresponding type.");
				smpd_dbg_printf(L"%s\n", process->dbgError);
				break;
			case EXCEPTION_FLT_STACK_CHECK:
				MPIU_Snprintf(process->dbgError, SMPD_MAX_ERROR_LEN, _TRUNCATE, L"Fatal error: The stack overflowed or underflowed as the result of a floating-point operation.");
				smpd_dbg_printf(L"%s\n", process->dbgError);
				break;
			case EXCEPTION_FLT_UNDERFLOW:
				MPIU_Snprintf(process->dbgError, SMPD_MAX_ERROR_LEN, _TRUNCATE, L"Fatal error: The exponent of a floating-point operation is less than the magnitude allowed by the corresponding type.");
				smpd_dbg_printf(L"%s\n", process->dbgError);
				break;
			case EXCEPTION_ILLEGAL_INSTRUCTION:
				MPIU_Snprintf(process->dbgError, SMPD_MAX_ERROR_LEN, _TRUNCATE, L"Fatal error: The thread tried to execute an invalid instruction.");
				smpd_dbg_printf(L"%s\n", process->dbgError);
				break;
			case EXCEPTION_IN_PAGE_ERROR:
				MPIU_Snprintf(process->dbgError, SMPD_MAX_ERROR_LEN, _TRUNCATE, L"Fatal error: The thread tried to access a page that was not present, and the system was unable to load the page. For example, this exception might occur if a network connection is lost while running a program over the network.");
				smpd_dbg_printf(L"%s\n", process->dbgError);
				break;
			case EXCEPTION_INT_DIVIDE_BY_ZERO:
				MPIU_Snprintf(process->dbgError, SMPD_MAX_ERROR_LEN, _TRUNCATE, L"Fatal error: The thread tried to divide an integer value by an integer divisor of zero.");
				smpd_dbg_printf(L"%s\n", process->dbgError);
				break;
			case EXCEPTION_INT_OVERFLOW:
				MPIU_Snprintf(process->dbgError, SMPD_MAX_ERROR_LEN, _TRUNCATE, L"Fatal error: The result of an integer operation caused a carry out of the most significant bit of the result.");
				smpd_dbg_printf(L"%s\n", process->dbgError);
				break;
			case EXCEPTION_INVALID_DISPOSITION:
				MPIU_Snprintf(process->dbgError, SMPD_MAX_ERROR_LEN, _TRUNCATE, L"Fatal error: An exception handler returned an invalid disposition to the exception dispatcher. Programmers using a high-level language such as C should never encounter this exception.");
				smpd_dbg_printf(L"%s\n", process->dbgError);
				break;
			case EXCEPTION_NONCONTINUABLE_EXCEPTION:
				MPIU_Snprintf(process->dbgError, SMPD_MAX_ERROR_LEN, _TRUNCATE, L"Fatal error: The thread tried to continue execution after a noncontinuable exception occurred.");
				smpd_dbg_printf(L"%s\n", process->dbgError);
				break;
			case EXCEPTION_PRIV_INSTRUCTION:
				MPIU_Snprintf(process->dbgError, SMPD_MAX_ERROR_LEN, _TRUNCATE, L"Fatal error: The thread tried to execute an instruction whose operation is not allowed in the current machine mode.");
				smpd_dbg_printf(L"%s\n", process->dbgError);
				break;
			case EXCEPTION_STACK_OVERFLOW:
				MPIU_Snprintf(process->dbgError, SMPD_MAX_ERROR_LEN, _TRUNCATE, L"Fatal error: The thread used up its stack.");
				smpd_dbg_printf(L"%s\n", process->dbgError);
				break;
			default:
				MPIU_Snprintf(process->dbgError, SMPD_MAX_ERROR_LEN, _TRUNCATE, L"Fatal error: EXCEPTION_DEBUG_EVENT %d", DebugEv.u.Exception.ExceptionRecord.ExceptionCode);
				smpd_dbg_printf(L"%s\n", process->dbgError);
				process->bAborted = SMPD_FALSE;
				dwContinueStatus = DBG_CONTINUE;
				break;
			}
			break;
		case CREATE_THREAD_DEBUG_EVENT:
			/*
			// As needed, examine or change the thread's registers 
			// with the GetThreadContext and SetThreadContext functions; 
			// and suspend and resume thread execution with the 
			// SuspendThread and ResumeThread functions.
			*/
			/*MPIU_Snprintf(process->dbgError, SMPD_MAX_ERROR_LEN, _TRUNCATE, L"CREATE_THREAD_DEBUG_EVENT");*/
			if (DebugEv.u.CreateThread.hThread != NULL)
				AddHandleToList(&pList, DebugEv.u.CreateThread.hThread);
			break;
		case CREATE_PROCESS_DEBUG_EVENT:
			/*
			// As needed, examine or change the registers of the 
			// process's initial thread with the GetThreadContext and 
			// SetThreadContext functions; read from and write to the 
			// process's virtual memory with the ReadProcessMemory and 
			// WriteProcessMemory functions; and suspend and resume 
			// thread execution with the SuspendThread and ResumeThread 
			// functions.
			*/
			/*MPIU_Snprintf(process->dbgError, SMPD_MAX_ERROR_LEN, _TRUNCATE, L"CREATE_PROCESS_DEBUG_EVENT");*/
			if (DebugEv.u.CreateProcessInfo.hFile != NULL)
				AddHandleToList(&pList, DebugEv.u.CreateProcessInfo.hFile);
			if (DebugEv.u.CreateProcessInfo.hProcess != NULL)
				AddHandleToList(&pList, DebugEv.u.CreateProcessInfo.hProcess);
			if (DebugEv.u.CreateProcessInfo.hThread != NULL)
				AddHandleToList(&pList, DebugEv.u.CreateProcessInfo.hThread);
			break;
		case EXIT_THREAD_DEBUG_EVENT:
			/* Display the thread's exit code. */
			/*MPIU_Snprintf(process->dbgError, SMPD_MAX_ERROR_LEN, _TRUNCATE, L"EXIT_THREAD_DEBUG_EVENT: exit code: %d", DebugEv.u.ExitThread.dwExitCode);*/
			break;
		case EXIT_PROCESS_DEBUG_EVENT: 
			/* Display the process's exit code. */
			/*MPIU_Snprintf(process->dbgError, SMPD_MAX_ERROR_LEN, _TRUNCATE, L"EXIT_PROCESS_DEBUG_EVENT: exit code: %d", DebugEv.u.ExitProcess.dwExitCode);*/
			bExit = SMPD_TRUE;
			break;
		case LOAD_DLL_DEBUG_EVENT:
			/*
			// Read the debugging information included in the newly 
			// loaded DLL.
			*/
			/*MPIU_Snprintf(process->dbgError, SMPD_MAX_ERROR_LEN, _TRUNCATE, L"LOAD_DLL_DEBUG_EVENT");*/
			if (DebugEv.u.LoadDll.hFile != NULL)
				AddHandleToList(&pList, DebugEv.u.LoadDll.hFile);
			break;
		case UNLOAD_DLL_DEBUG_EVENT:
			/* Display a message that the DLL has been unloaded. */
			/*MPIU_Snprintf(process->dbgError, SMPD_MAX_ERROR_LEN, _TRUNCATE, L"UNLOAD_DLL_DEBUG_EVENT");*/
			break;
		case OUTPUT_DEBUG_STRING_EVENT: 
			/* Display the output debugging string. */
			/*MPIU_Snprintf(process->dbgError, SMPD_MAX_ERROR_LEN, _TRUNCATE, L"OUTPUT_DEBUG_STRING_EVENT");*/
			break;
		default:
			MPIU_Snprintf(process->dbgError, SMPD_MAX_ERROR_LEN, _TRUNCATE, L"Fatal error: Unknown event %d", DebugEv.dwDebugEventCode);
			smpd_dbg_printf(L"%s\n", process->dbgError);
			break;
		}

		/* Callback thread doesn't run for some reason.  I guess you can't create a thread in a process that has an unhandled exception */
		/*
		if (process->bAborted == SMPD_TRUE && process->abort_fn != NULL)
		{
			smpd_call_abort_hook(process, DebugEv.dwDebugEventCode);
		}
		*/

		/* Resume executing the thread that reported the debugging event. */

		ContinueDebugEvent(DebugEv.dwProcessId, DebugEv.dwThreadId, dwContinueStatus);

		if (bExit)
			break;
	}

	CloseHandleList(pList);
	return 0;
}

/* static variables to pass to the debug thread so it can launch the process */
typedef struct debug_thread_params
{
	wchar_t *actual_exe;
	DWORD launch_flag;
	void *pEnv;
	STARTUPINFO *saInfo;
	PROCESS_INFORMATION *psInfo;
	int priority;
	SMPD_BOOL bSuccess;
} debug_thread_params;

static debug_thread_params params;

static DWORD DebugLaunchProcess(smpd_process_t *process)
{
	if (CreateProcess(
		NULL,
		params.actual_exe,
		NULL, NULL, TRUE,
		params.launch_flag,
		params.pEnv,
		NULL,
		params.saInfo, params.psInfo))
	{
		SetThreadPriority(params.psInfo->hThread, params.priority);
		process->pid = params.psInfo->dwProcessId;
		DebugWaitForProcess(process);
	}
	else
	{
		int nError = GetLastError();
		smpd_err_printf(L"CreateProcess('%s') failed, error %d\n", process->exe, nError);
		smpd_translate_win_error(nError, process->err_msg, SMPD_MAX_ERROR_LEN, L"CreateProcess(%s) on '%s' failed, error %d - ",
			process->exe, smpd_process.host, nError);
		params.psInfo->hProcess = INVALID_HANDLE_VALUE;
		params.bSuccess = FALSE;
		SetEvent(process->hDebugReady);
	}
	return 0;
}

#undef FCNAME
#define FCNAME L"smpd_launch_process"
int smpd_launch_process(smpd_process_t *process, int priorityClass, int priority, int dbg, MPIDU_Sock_set_t set)
{
	HANDLE hStdin = INVALID_HANDLE_VALUE, hStdout = INVALID_HANDLE_VALUE, hStderr = INVALID_HANDLE_VALUE;
	SOCKET hSockStdinR = INVALID_SOCKET, hSockStdinW = INVALID_SOCKET;
	SOCKET hSockStdoutR = INVALID_SOCKET, hSockStdoutW = INVALID_SOCKET;
	SOCKET hSockStderrR = INVALID_SOCKET, hSockStderrW = INVALID_SOCKET;
	SOCKET hSockPmiR = INVALID_SOCKET, hSockPmiW = INVALID_SOCKET;
	HANDLE hPipeStdinR = NULL, hPipeStdinW = NULL;
	HANDLE hPipeStdoutR = NULL, hPipeStdoutW = NULL;
	HANDLE hPipeStderrR = NULL, hPipeStderrW = NULL;
	HANDLE hIn = INVALID_HANDLE_VALUE, hOut = INVALID_HANDLE_VALUE, hErr = INVALID_HANDLE_VALUE;
	STARTUPINFO saInfo;
	PROCESS_INFORMATION psInfo = { 0 };
	void *pEnv=NULL;
	wchar_t tSavedPath[MAX_PATH] = L".";
	DWORD launch_flag = 0;
	int nError = 0, result = 0;
	/*unsigned long blocking_flag = 0;*/
	MPIDU_Sock_t sock_in = MPIDU_SOCK_INVALID_SOCK, sock_out = MPIDU_SOCK_INVALID_SOCK, sock_err = MPIDU_SOCK_INVALID_SOCK, sock_pmi = MPIDU_SOCK_INVALID_SOCK;
	SECURITY_ATTRIBUTES saAttr;
	wchar_t str[8192], sock_str[20];
	BOOL bSuccess = TRUE;
	wchar_t *actual_exe, exe_data[SMPD_MAX_EXE_LENGTH];
	wchar_t *args;
	wchar_t temp_exe[SMPD_MAX_EXE_LENGTH];
	MPIDU_Sock_t sock_pmi_listener;
	smpd_context_t *listener_context;
	int listener_port = 0;
	wchar_t host_description[256];
	wchar_t err_msg[MAX_ERROR_LENGTH] = L"";
	DWORD resume_count;

	smpd_enter_fn(FCNAME);

	/* Initialize the psInfo structure in case there is an error an we jump to CLEANUP before psInfo is set */
	psInfo.hProcess = INVALID_HANDLE_VALUE;

	/* resolve the executable name */
	if (process->path[0] != L'\0')
	{
		args = process->exe;
		result = MPIU_Str_get_string(&args, temp_exe, SMPD_MAX_EXE_LENGTH);
		if (result != MPIU_STR_SUCCESS)
		{
		}
		smpd_dbg_printf(L"searching for '%s' in '%s'\n", temp_exe, process->path);
		if (smpd_search_path(process->path, temp_exe, SMPD_MAX_EXE_LENGTH, exe_data))
		{
			if (args != NULL)
			{
				wcsncat_s(exe_data, SMPD_MAX_EXE_LENGTH - wcslen(exe_data), L" ", _TRUNCATE);
				wcsncat_s(exe_data, SMPD_MAX_EXE_LENGTH - wcslen(exe_data), args, _TRUNCATE);
				exe_data[SMPD_MAX_EXE_LENGTH-1] = L'\0';
			}
			actual_exe = exe_data;
		}
		else
		{
			actual_exe = process->exe;
		}
	}
	else
	{
		actual_exe = process->exe;
	}

	smpd_priority_class_to_win_class(&priorityClass);
	smpd_priority_to_win_priority(&priority);

	/* Save stdin, stdout, and stderr */
	result = WaitForSingleObject(smpd_process.hLaunchProcessMutex, INFINITE);
	if (result == WAIT_FAILED)
	{
		result = GetLastError();
		smpd_translate_win_error(result, err_msg, MAX_ERROR_LENGTH, NULL);
		smpd_err_printf(L"Error waiting for smpd_process.hLaunchProcessMutex %d, %s\n", result, err_msg);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	hStdin = GetStdHandle(STD_INPUT_HANDLE);
	hStdout = GetStdHandle(STD_OUTPUT_HANDLE);
	hStderr = GetStdHandle(STD_ERROR_HANDLE);
	if (hStdin == INVALID_HANDLE_VALUE || hStdout == INVALID_HANDLE_VALUE  || hStderr == INVALID_HANDLE_VALUE)
	{
		nError = GetLastError(); /* This will only be correct if stderr failed */
		ReleaseMutex(smpd_process.hLaunchProcessMutex);
		smpd_translate_win_error(nError, err_msg, MAX_ERROR_LENGTH, NULL);
		smpd_err_printf(L"GetStdHandle failed, error %d, %s\n", nError, err_msg);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;;
	}

	/* Create sockets for stdin, stdout, and stderr */
	nError = smpd_make_socket_loop_choose(&hSockStdinR, SMPD_FALSE, &hSockStdinW, SMPD_TRUE);
	if (nError)
	{
		smpd_err_printf(L"smpd_make_socket_loop failed, error %d\n", nError);
		goto CLEANUP;
	}
	nError = smpd_make_socket_loop_choose(&hSockStdoutR, SMPD_TRUE, &hSockStdoutW, SMPD_FALSE);
	if (nError)
	{
		smpd_err_printf(L"smpd_make_socket_loop failed, error %d\n", nError);
		goto CLEANUP;
	}
	nError = smpd_make_socket_loop_choose(&hSockStderrR, SMPD_TRUE, &hSockStderrW, SMPD_FALSE);
	if (nError)
	{
		smpd_err_printf(L"smpd_make_socket_loop failed, error %d\n", nError);
		goto CLEANUP;
	}
	if (process->pmi != NULL)
	{
		if (smpd_process.use_inherited_handles)
		{
			nError = smpd_make_socket_loop(&hSockPmiR, &hSockPmiW);
			if (nError)
			{
				smpd_err_printf(L"smpd_make_socket_loop failed, error %d\n", nError);
				goto CLEANUP;
			}
		}
		else
		{
			nError = MPIDU_Sock_listen(set, NULL, &listener_port, &sock_pmi_listener); 
			if (nError != MPI_SUCCESS)
			{
				smpd_err_printf(L"MPIDU_Sock_listen failed,\nsock error: %s\n", get_sock_error_string(nError));
				goto CLEANUP;
			}
			smpd_dbg_printf(L"pmi listening on port %d\n", listener_port);

			nError = smpd_create_context(SMPD_CONTEXT_PMI_LISTENER, set, sock_pmi_listener, -1, &listener_context);
			if (nError != SMPD_SUCCESS)
			{
				smpd_err_printf(L"unable to create a context for the pmi listener.\n");
				goto CLEANUP;
			}
			nError = MPIDU_Sock_set_user_ptr(sock_pmi_listener, listener_context);
			if (nError != MPI_SUCCESS)
			{
				smpd_err_printf(L"MPIDU_Sock_set_user_ptr failed,\nsock error: %s\n", get_sock_error_string(nError));
				goto CLEANUP;
			}
			listener_context->state = SMPD_PMI_LISTENING;
			nError = MPIDU_Sock_get_host_description(host_description, 256);
			if (nError != MPI_SUCCESS)
			{
				smpd_err_printf(L"MPIDU_Sock_get_host_description failed,\nsock error: %s\n", get_sock_error_string(nError));
				goto CLEANUP;
			}
			/* save the listener sock in the slot reserved for the client sock so we can match the listener
			to the process structure when the client sock is accepted */
			process->pmi->sock = sock_pmi_listener;
		}
	}

	saAttr.nLength = sizeof(SECURITY_ATTRIBUTES);
	saAttr.lpSecurityDescriptor = NULL;
	saAttr.bInheritHandle = TRUE;

	/* Create the pipes for stdout, stderr */
	if (!CreatePipe(&hPipeStdinR, &hPipeStdinW, &saAttr, 0))
	{
		smpd_err_printf(L"CreatePipe(stdin) failed, error %d\n", GetLastError());
		goto CLEANUP;
	}
	if (!CreatePipe(&hPipeStdoutR, &hPipeStdoutW, &saAttr, 0))
	{
		smpd_err_printf(L"CreatePipe(stdout) failed, error %d\n", GetLastError());
		goto CLEANUP;
	}
	if (!CreatePipe(&hPipeStderrR, &hPipeStderrW, &saAttr, 0))
	{
		smpd_err_printf(L"CreatePipe(stderr) failed, error %d\n", GetLastError());
		goto CLEANUP;
	}

	/* Make the ends of the pipes that this process will use not inheritable */
	/*
	if (!DuplicateHandle(GetCurrentProcess(), (HANDLE)hSockStdinW, GetCurrentProcess(), &hIn, 
		0, FALSE, DUPLICATE_CLOSE_SOURCE | DUPLICATE_SAME_ACCESS))
	{
		nError = GetLastError();
		smpd_err_printf(L"DuplicateHandle failed, error %d\n", nError);
		goto CLEANUP;
	}
	*/
	if (!DuplicateHandle(GetCurrentProcess(), (HANDLE)hPipeStdinW, GetCurrentProcess(), &hIn, 
		0, FALSE, DUPLICATE_CLOSE_SOURCE | DUPLICATE_SAME_ACCESS))
	{
		nError = GetLastError();
		smpd_err_printf(L"DuplicateHandle failed, error %d\n", nError);
		goto CLEANUP;
	}
	/*
	if (!DuplicateHandle(GetCurrentProcess(), (HANDLE)hSockStdoutR, GetCurrentProcess(), &hOut, 
		0, FALSE, DUPLICATE_CLOSE_SOURCE | DUPLICATE_SAME_ACCESS))
	{
		nError = GetLastError();
		smpd_err_printf(L"DuplicateHandle failed, error %d\n", nError);
		goto CLEANUP;
	}
	if (!DuplicateHandle(GetCurrentProcess(), (HANDLE)hSockStderrR, GetCurrentProcess(), &hErr, 
		0, FALSE, DUPLICATE_CLOSE_SOURCE | DUPLICATE_SAME_ACCESS))
	{
		nError = GetLastError();
		smpd_err_printf(L"DuplicateHandle failed, error %d\n", nError);
		goto CLEANUP;
	}
	*/
	if (!DuplicateHandle(GetCurrentProcess(), (HANDLE)hPipeStdoutR, GetCurrentProcess(), &hOut, 
		0, FALSE, DUPLICATE_CLOSE_SOURCE | DUPLICATE_SAME_ACCESS))
	{
		nError = GetLastError();
		smpd_err_printf(L"DuplicateHandle failed, error %d\n", nError);
		goto CLEANUP;
	}
	if (!DuplicateHandle(GetCurrentProcess(), (HANDLE)hPipeStderrR, GetCurrentProcess(), &hErr, 
		0, FALSE, DUPLICATE_CLOSE_SOURCE | DUPLICATE_SAME_ACCESS))
	{
		nError = GetLastError();
		smpd_err_printf(L"DuplicateHandle failed, error %d\n", nError);
		goto CLEANUP;
	}
	if (process->pmi != NULL && smpd_process.use_inherited_handles)
	{
		if (!DuplicateHandle(GetCurrentProcess(), (HANDLE)hSockPmiR, GetCurrentProcess(), (LPHANDLE)&hSockPmiR, 
			0, FALSE, DUPLICATE_CLOSE_SOURCE | DUPLICATE_SAME_ACCESS))
		{
			nError = GetLastError();
			smpd_err_printf(L"DuplicateHandle failed, error %d\n", nError);
			goto CLEANUP;
		}
	}

	/* prevent the socket loops from being inherited */
	if (!DuplicateHandle(GetCurrentProcess(), (HANDLE)hSockStdinR, GetCurrentProcess(), (LPHANDLE)&hSockStdinR, 
		0, FALSE, DUPLICATE_CLOSE_SOURCE | DUPLICATE_SAME_ACCESS))
	{
		nError = GetLastError();
		smpd_err_printf(L"DuplicateHandle failed, error %d\n", nError);
		goto CLEANUP;
	}
	if (!DuplicateHandle(GetCurrentProcess(), (HANDLE)hSockStdoutR, GetCurrentProcess(), (LPHANDLE)&hSockStdoutR, 
		0, FALSE, DUPLICATE_CLOSE_SOURCE | DUPLICATE_SAME_ACCESS))
	{
		nError = GetLastError();
		smpd_err_printf(L"DuplicateHandle failed, error %d\n", nError);
		goto CLEANUP;
	}
	if (!DuplicateHandle(GetCurrentProcess(), (HANDLE)hSockStderrR, GetCurrentProcess(), (LPHANDLE)&hSockStderrR, 
		0, FALSE, DUPLICATE_CLOSE_SOURCE | DUPLICATE_SAME_ACCESS))
	{
		nError = GetLastError();
		smpd_err_printf(L"DuplicateHandle failed, error %d\n", nError);
		goto CLEANUP;
	}
	if (!DuplicateHandle(GetCurrentProcess(), (HANDLE)hSockStdinW, GetCurrentProcess(), (LPHANDLE)&hSockStdinW, 
		0, FALSE, DUPLICATE_CLOSE_SOURCE | DUPLICATE_SAME_ACCESS))
	{
		nError = GetLastError();
		smpd_err_printf(L"DuplicateHandle failed, error %d\n", nError);
		goto CLEANUP;
	}
	if (!DuplicateHandle(GetCurrentProcess(), (HANDLE)hSockStdoutW, GetCurrentProcess(), (LPHANDLE)&hSockStdoutW, 
		0, FALSE, DUPLICATE_CLOSE_SOURCE | DUPLICATE_SAME_ACCESS))
	{
		nError = GetLastError();
		smpd_err_printf(L"DuplicateHandle failed, error %d\n", nError);
		goto CLEANUP;
	}
	if (!DuplicateHandle(GetCurrentProcess(), (HANDLE)hSockStderrW, GetCurrentProcess(), (LPHANDLE)&hSockStderrW, 
		0, FALSE, DUPLICATE_CLOSE_SOURCE | DUPLICATE_SAME_ACCESS))
	{
		nError = GetLastError();
		smpd_err_printf(L"DuplicateHandle failed, error %d\n", nError);
		goto CLEANUP;
	}

	/* make the ends used by the spawned process blocking */
	/*
	blocking_flag = 0;
	ioctlsocket(hSockStdinR, FIONBIO, &blocking_flag);
	*/
	/*
	blocking_flag = 0;
	ioctlsocket(hSockStdoutW, FIONBIO, &blocking_flag);
	blocking_flag = 0;
	ioctlsocket(hSockStderrW, FIONBIO, &blocking_flag);
	*/

	/* Set stdin, stdout, and stderr to the ends of the pipe the created process will use */
	/*
	if (!SetStdHandle(STD_INPUT_HANDLE, (HANDLE)hSockStdinR))
	{
		nError = GetLastError();
		smpd_err_printf(L"SetStdHandle failed, error %d\n", nError);
		goto CLEANUP;
	}
	*/
	if (!SetStdHandle(STD_INPUT_HANDLE, (HANDLE)hPipeStdinR))
	{
		nError = GetLastError();
		smpd_err_printf(L"SetStdHandle failed, error %d\n", nError);
		goto CLEANUP;
	}
	/*
	if (!SetStdHandle(STD_OUTPUT_HANDLE, (HANDLE)hSockStdoutW))
	{
		nError = GetLastError();
		smpd_err_printf(L"SetStdHandle failed, error %d\n", nError);
		goto RESTORE_CLEANUP;
	}
	if (!SetStdHandle(STD_ERROR_HANDLE, (HANDLE)hSockStderrW))
	{
		nError = GetLastError();
		smpd_err_printf(L"SetStdHandle failed, error %d\n", nError);
		goto RESTORE_CLEANUP;
	}
	*/
	if (!SetStdHandle(STD_OUTPUT_HANDLE, (HANDLE)hPipeStdoutW))
	{
		nError = GetLastError();
		smpd_err_printf(L"SetStdHandle failed, error %d\n", nError);
		goto RESTORE_CLEANUP;
	}
	if (!SetStdHandle(STD_ERROR_HANDLE, (HANDLE)hPipeStderrW))
	{
		nError = GetLastError();
		smpd_err_printf(L"SetStdHandle failed, error %d\n", nError);
		goto RESTORE_CLEANUP;
	}

	/* Create the process */
	memset(&saInfo, 0, sizeof(STARTUPINFO));
	saInfo.cb = sizeof(STARTUPINFO);
	/*saInfo.hStdInput = (HANDLE)hSockStdinR;*/
	saInfo.hStdInput = (HANDLE)hPipeStdinR;
	/*
	saInfo.hStdOutput = (HANDLE)hSockStdoutW;
	saInfo.hStdError = (HANDLE)hSockStderrW;
	*/
	saInfo.hStdOutput = (HANDLE)hPipeStdoutW;
	saInfo.hStdError = (HANDLE)hPipeStderrW;
	saInfo.dwFlags = STARTF_USESTDHANDLES;

	SetEnvironmentVariables(process->env);
	if (process->pmi != NULL)
	{
		swprintf_s(str, 8192, L"%d", process->rank);
		smpd_dbg_printf(L"env: PMI_RANK=%s\n", str);
		if (!SetEnvironmentVariable(L"PMI_RANK", str))
		{
			smpd_err_printf(L"Unable to set environment variable <%s> = <%s>: Error %d\n", L"PMI_RANK", str, GetLastError());
		}
		swprintf_s(str, 8192, L"%d", process->nproc);
		smpd_dbg_printf(L"env: PMI_SIZE=%s\n", str);
		if (!SetEnvironmentVariable(L"PMI_SIZE", str))
		{
			smpd_err_printf(L"Unable to set environment variable <%s> = <%s>: Error %d\n", L"PMI_SIZE", str, GetLastError());
		}
		swprintf_s(str, 8192, L"%s", process->kvs_name);
		smpd_dbg_printf(L"env: PMI_KVS=%s\n", str);
		if (!SetEnvironmentVariable(L"PMI_KVS", str))
		{
			smpd_err_printf(L"Unable to set environment variable <%s> = <%s>: Error %d\n", L"PMI_KVS", str, GetLastError());
		}
		swprintf_s(str, 8192, L"%s", process->domain_name);
		smpd_dbg_printf(L"env: PMI_DOMAIN=%s\n", str);
		if (!SetEnvironmentVariable(L"PMI_DOMAIN", str))
		{
			smpd_err_printf(L"Unable to set environment variable <%s> = <%s>: Error %d\n", L"PMI_DOMAIN", str, GetLastError());
		}
		if (smpd_process.use_inherited_handles)
		{
			smpd_encode_handle(sock_str, 20, (HANDLE)hSockPmiW);
			swprintf_s(str, 8192, L"%s", sock_str);
			smpd_dbg_printf(L"env: PMI_SMPD_FD=%s\n", str);
			if (!SetEnvironmentVariable(L"PMI_SMPD_FD", str))
			{
				smpd_err_printf(L"Unable to set environment variable <%s> = <%s>: Error %d\n", L"PMI_SMPD_FD", str, GetLastError());
			}
		}
		else
		{
			smpd_dbg_printf(L"env: PMI_HOST=%s\n", host_description);
			if (!SetEnvironmentVariable(L"PMI_HOST", host_description))
			{
				smpd_err_printf(L"Unable to set environment variable <%s> = <%s>: Error %d\n", L"PMI_HOST", str, GetLastError());
			}
			swprintf_s(str, 8192, L"%d", listener_port);
			smpd_dbg_printf(L"env: PMI_PORT=%s\n", str);
			if (!SetEnvironmentVariable(L"PMI_PORT", str))
			{
				smpd_err_printf(L"Unable to set environment variable <%s> = <%s>: Error %d\n", L"PMI_PORT", str, GetLastError());
			}
		}
		swprintf_s(str, 8192, L"%d", smpd_process.id);
		smpd_dbg_printf(L"env: PMI_SMPD_ID=%s\n", str);
		if (!SetEnvironmentVariable(L"PMI_SMPD_ID", str))
		{
			smpd_err_printf(L"Unable to set environment variable <%s> = <%s>: Error %d\n", L"PMI_SMPD_ID", str, GetLastError());
		}
		swprintf_s(str, 8192, L"%d", process->id);
		smpd_dbg_printf(L"env: PMI_SMPD_KEY=%s\n", str);
		if (!SetEnvironmentVariable(L"PMI_SMPD_KEY", str))
		{
			smpd_err_printf(L"Unable to set environment variable <%s> = <%s>: Error %d\n", L"PMI_SMPD_KEY", str, GetLastError());
		}
		if (process->clique[0] != L'\0')
		{
			swprintf_s(str, 8192, L"%s", process->clique);
			smpd_dbg_printf(L"env: PMI_CLIQUE=%s\n", str);
			if (!SetEnvironmentVariable(L"PMI_CLIQUE", str))
			{
				smpd_err_printf(L"Unable to set environment variable <%s> = <%s>: Error %d\n", L"PMI_CLIQUE", str, GetLastError());
			}
		}
		swprintf_s(str, 8192, L"%d", process->spawned);
		smpd_dbg_printf(L"env: PMI_SPAWN=%s\n", str);
		if (!SetEnvironmentVariable(L"PMI_SPAWN", str))
		{
			smpd_err_printf(L"Unable to set environment variable <%s> = <%s>: Error %d\n", L"PMI_SPAWN", str, GetLastError());
		}
		swprintf_s(str, 8192, L"%d", process->appnum);
		smpd_dbg_printf(L"env: PMI_APPNUM=%s\n", str);
		if (!SetEnvironmentVariable(L"PMI_APPNUM", str))
		{
			smpd_err_printf(L"Unable to set environment variable <%s> = <%s>: Error %d\n", L"PMI_APPNUM", str, GetLastError());
		}
	}
	pEnv = GetEnvironmentStrings();

	GetCurrentDirectory(MAX_PATH, tSavedPath);
	SetCurrentDirectory(process->dir);

	launch_flag = 
		CREATE_SUSPENDED | /*CREATE_NEW_CONSOLE*/ /*DETACHED_PROCESS*/ CREATE_NO_WINDOW | priorityClass | CREATE_UNICODE_ENVIRONMENT;
	if (dbg)
		launch_flag = launch_flag | DEBUG_PROCESS;

	smpd_dbg_printf(L"CreateProcess(%s)\n", actual_exe);
	psInfo.hProcess = INVALID_HANDLE_VALUE;
	if (dbg)
	{
		smpd_dbg_printf(L"using DebugLaunchProcess.\n");

		params.actual_exe = actual_exe;
		params.launch_flag = launch_flag;
		params.pEnv = pEnv;
		params.saInfo = &saInfo;
		params.psInfo = &psInfo;
		params.priority = priority;
		params.bSuccess = SMPD_TRUE;

		process->hDebugReady = CreateEvent(NULL, TRUE, FALSE, NULL);
		if (process->hDebugReady != NULL)
		{
			process->hDebugThread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)DebugLaunchProcess, process, 0, NULL);
			if (process->hDebugThread != NULL)
			{
				DWORD result;
				result = WaitForSingleObject(process->hDebugReady, SMPD_SHORT_TIMEOUT);
				if (result != WAIT_OBJECT_0)
				{
					switch (result)
					{
					case WAIT_ABANDONED:
					case WAIT_TIMEOUT:
						break;
					case WAIT_FAILED:
						smpd_err_printf(L"unable to wait for the debug thread to set the ready event.  Error %d\n", GetLastError());
						break;
					}
				}
				/* FIXME: This isn't right because if we are going to use the event again the thread could have set it during the time between
				 * the return of the WaitForSingleObject call and the following call to ResetEvent.  Maybe the thread could wait until the event
				 * has been reset before setting it again.
				 */
				if (!ResetEvent(process->hDebugReady))
				{
					smpd_err_printf(L"unable to reset the debug event.  Error %d\n", GetLastError());
				}
			}
			else
			{
				smpd_err_printf(L"unable to create the debug thread: Error %d\n", GetLastError());
			}
		}
		else
		{
			smpd_err_printf(L"unable to create an event for the debug thread: Error %d\n", GetLastError());
		}

		bSuccess = params.bSuccess;
	}
	else
	{
		if (CreateProcess(
			NULL,
			actual_exe,
			NULL, NULL, TRUE,
			launch_flag,
			pEnv,
			NULL,
			&saInfo, &psInfo))
		{
			SetThreadPriority(psInfo.hThread, priority);
			process->pid = psInfo.dwProcessId;
		}
		else
		{
			nError = GetLastError();
			smpd_err_printf(L"CreateProcess('%s') failed, error %d\n", process->exe, nError);
			smpd_translate_win_error(nError, process->err_msg, SMPD_MAX_ERROR_LEN, L"CreateProcess(%s) on '%s' failed, error %d - ",
				process->exe, smpd_process.host, nError);
			psInfo.hProcess = INVALID_HANDLE_VALUE;
			bSuccess = FALSE;
		}
	}

	FreeEnvironmentStrings((TCHAR*)pEnv);
	SetCurrentDirectory(tSavedPath);
	RemoveEnvironmentVariables(process->env);
	if (process->pmi != NULL)
	{
		SetEnvironmentVariable(L"PMI_RANK", NULL);
		SetEnvironmentVariable(L"PMI_SIZE", NULL);
		SetEnvironmentVariable(L"PMI_KVS", NULL);
		SetEnvironmentVariable(L"PMI_DOMAIN", NULL);
		if (smpd_process.use_inherited_handles)
		{
			SetEnvironmentVariable(L"PMI_SMPD_FD", NULL);
		}
		else
		{
			SetEnvironmentVariable(L"PMI_HOST", NULL);
			SetEnvironmentVariable(L"PMI_PORT", NULL);
		}
		SetEnvironmentVariable(L"PMI_SMPD_ID", NULL);
		SetEnvironmentVariable(L"PMI_SMPD_KEY", NULL);
		SetEnvironmentVariable(L"PMI_SPAWN", NULL);
	}

	if (bSuccess)
	{
		/* make sock structures out of the sockets */
		/*
		nError = MPIDU_Sock_native_to_sock(set, hIn, NULL, &sock_in);
		if (nError != MPI_SUCCESS)
		{
			smpd_err_printf(L"MPIDU_Sock_native_to_sock failed, error %s\n", get_sock_error_string(nError));
		}
		*/
		/*wprintf(L"native sock for stdin\n");fflush(stdout);*/
		nError = MPIDU_Sock_native_to_sock(set, (MPIDU_SOCK_NATIVE_FD)hSockStdinW, NULL, &sock_in);
		if (nError != MPI_SUCCESS)
		{
			smpd_err_printf(L"MPIDU_Sock_native_to_sock failed, error %s\n", get_sock_error_string(nError));
		}
		/*wprintf(L"native sock for stdout\n");fflush(stdout);*/
		nError = MPIDU_Sock_native_to_sock(set, (MPIDU_SOCK_NATIVE_FD)hSockStdoutR, NULL, &sock_out);
		if (nError != MPI_SUCCESS)
		{
			smpd_err_printf(L"MPIDU_Sock_native_to_sock failed, error %s\n", get_sock_error_string(nError));
		}
		/*wprintf(L"native sock for stderr\n");fflush(stdout);*/
		nError = MPIDU_Sock_native_to_sock(set, (MPIDU_SOCK_NATIVE_FD)hSockStderrR, NULL, &sock_err);
		if (nError != MPI_SUCCESS)
		{
			smpd_err_printf(L"MPIDU_Sock_native_to_sock failed, error %s\n", get_sock_error_string(nError));
		}
		if (process->pmi != NULL && smpd_process.use_inherited_handles)
		{
			/*wprintf(L"native sock for pmi\n");fflush(stdout);*/
			nError = MPIDU_Sock_native_to_sock(set, (MPIDU_SOCK_NATIVE_FD)hSockPmiR, NULL, &sock_pmi);
			if (nError != MPI_SUCCESS)
			{
				smpd_err_printf(L"MPIDU_Sock_native_to_sock failed, error %s\n", get_sock_error_string(nError));
			}
		}

		process->in->sock = sock_in;
		process->out->sock = sock_out;
		process->err->sock = sock_err;
		if (process->pmi != NULL && smpd_process.use_inherited_handles)
			process->pmi->sock = sock_pmi;
		process->pid = process->in->id = process->out->id = process->err->id = psInfo.dwProcessId;
		MPIDU_Sock_set_user_ptr(sock_in, process->in);
		MPIDU_Sock_set_user_ptr(sock_out, process->out);
		MPIDU_Sock_set_user_ptr(sock_err, process->err);
		if (process->pmi != NULL && smpd_process.use_inherited_handles)
			MPIDU_Sock_set_user_ptr(sock_pmi, process->pmi);
	}
	else
	{
		/* close all the sockets and handles allocated in this function */
		/*CloseHandle(hIn);*/
		CloseHandle((HANDLE)hSockStdinW);
		CloseHandle((HANDLE)hSockStdoutR);
		CloseHandle((HANDLE)hSockStderrR);
		if (process->pmi != NULL && smpd_process.use_inherited_handles)
			CloseHandle((HANDLE)hSockPmiR);
	}

RESTORE_CLEANUP:
	/* Restore stdin, stdout, stderr */
	SetStdHandle(STD_INPUT_HANDLE, hStdin);
	SetStdHandle(STD_OUTPUT_HANDLE, hStdout);
	SetStdHandle(STD_ERROR_HANDLE, hStderr);

CLEANUP:
	ReleaseMutex(smpd_process.hLaunchProcessMutex);
	/*CloseHandle((HANDLE)hSockStdinR);*/
	CloseHandle((HANDLE)hPipeStdinR);
	/*
	CloseHandle((HANDLE)hSockStdoutW);
	CloseHandle((HANDLE)hSockStderrW);
	*/
	CloseHandle((HANDLE)hPipeStdoutW);
	CloseHandle((HANDLE)hPipeStderrW);
	if (process->pmi != NULL && smpd_process.use_inherited_handles)
		CloseHandle((HANDLE)hSockPmiW);

	if (psInfo.hProcess != INVALID_HANDLE_VALUE)
	{
		HANDLE hThread;
		smpd_piothread_arg_t *arg_ptr;
		smpd_pinthread_arg_t *in_arg_ptr;

		in_arg_ptr = (smpd_pinthread_arg_t*)malloc(sizeof(smpd_pinthread_arg_t));
		in_arg_ptr->hIn = hSockStdinR;
		in_arg_ptr->hOut = hPipeStdinW;
		in_arg_ptr->pid = psInfo.dwProcessId;
		hThread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)smpd_pinthread, in_arg_ptr, 0, NULL);
		CloseHandle(hThread);
		arg_ptr = (smpd_piothread_arg_t*)malloc(sizeof(smpd_piothread_arg_t));
		arg_ptr->hIn = hOut;
		arg_ptr->hOut = hSockStdoutW;
		arg_ptr->pid = psInfo.dwProcessId;
		hThread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)smpd_piothread, arg_ptr, 0, NULL);
		CloseHandle(hThread);
		arg_ptr = (smpd_piothread_arg_t*)malloc(sizeof(smpd_piothread_arg_t));
		arg_ptr->hIn = hErr;
		arg_ptr->hOut = hSockStderrW;
		arg_ptr->pid = psInfo.dwProcessId;
		hThread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)smpd_piothread, arg_ptr, 0, NULL);
		CloseHandle(hThread);

		if (process->pmi != NULL && smpd_process.use_inherited_handles)
			process->context_refcount = 3;
		else
			process->context_refcount = 2;
		process->out->read_state = SMPD_READING_STDOUT;
		result = MPIDU_Sock_post_read(sock_out, process->out->read_cmd.cmd, 1, 1, NULL);
		if (result != MPI_SUCCESS)
		{
			smpd_err_printf(L"posting first read from stdout context failed, sock error: %s\n",
				get_sock_error_string(result));
			smpd_exit_fn(FCNAME);
			return SMPD_FAIL;
		}
		process->err->read_state = SMPD_READING_STDERR;
		result = MPIDU_Sock_post_read(sock_err, process->err->read_cmd.cmd, 1, 1, NULL);
		if (result != MPI_SUCCESS)
		{
			smpd_err_printf(L"posting first read from stderr context failed, sock error: %s\n",
				get_sock_error_string(result));
			smpd_exit_fn(FCNAME);
			return SMPD_FAIL;
		}
		if (process->pmi != NULL && smpd_process.use_inherited_handles)
		{
			result = smpd_post_read_command(process->pmi);
			if (result != SMPD_SUCCESS)
			{
				smpd_err_printf(L"unable to post a read of the first command on the pmi control channel.\n");
				smpd_exit_fn(FCNAME);
				return SMPD_FAIL;
			}
		}
		process->wait.hProcess = process->in->wait.hProcess = process->out->wait.hProcess = process->err->wait.hProcess = psInfo.hProcess;
		process->wait.hThread = process->in->wait.hThread = process->out->wait.hThread = process->err->wait.hThread = psInfo.hThread;
		smpd_process_to_registry(process, actual_exe);
		resume_count = ResumeThread(psInfo.hThread);
		if (resume_count == (DWORD)-1)
		{
			smpd_err_printf(L"unable to resume the main thread after CreateProcess.  Error %d\n", GetLastError());
			smpd_exit_fn(FCNAME);
			return SMPD_FAIL;
		}
		smpd_dbg_printf(L"Resume count = %d after ResumeThread.\n", resume_count);
		smpd_exit_fn(FCNAME);
		return SMPD_SUCCESS;
	}

	smpd_exit_fn(FCNAME);
	return SMPD_FAIL;
}

#undef FCNAME
#define FCNAME L"smpd_parse_account_domain"
void smpd_parse_account_domain(const wchar_t *domain_account, wchar_t *account, size_t maxlen1, wchar_t *domain, size_t maxlen2)
{
	const wchar_t *pCh;
	wchar_t *pCh2;

	smpd_enter_fn(FCNAME);

	if ((wcschr(domain_account, L'\\') == NULL) && (wcschr(domain_account, L'@') != NULL))
	{
		pCh = domain_account;
		pCh2 = account;
		while ((*pCh != L'@') && (*pCh != L'\0'))
		{
			*pCh2 = *pCh;
			pCh++;
			pCh2++;
		}
		*pCh2 = L'\0';
		if (*pCh == L'@')
		{
			pCh++;
			wcscpy_s(domain, maxlen2, pCh);
		}
		else
		{
			domain[0] = L'\0';
		}
	}
	else
	{
		pCh = domain_account;
		pCh2 = domain;
		while ((*pCh != L'\\') && (*pCh != L'\0'))
		{
			*pCh2 = *pCh;
			pCh++;
			pCh2++;
		}
		if (*pCh == L'\\')
		{
			pCh++;
			wcscpy_s(account, maxlen1, pCh);
			*pCh2 = L'\0';
		}
		else
		{
			wcscpy_s(account, maxlen1, domain_account);
			domain[0] = L'\0';
		}
	}

	smpd_exit_fn(FCNAME);
}

#undef FCNAME
#define FCNAME L"smpd_wait_process"
int smpd_wait_process(smpd_pwait_t wait, int *exit_code_ptr)
{
	int result;
	DWORD exit_code;

	smpd_enter_fn(FCNAME);

	if (wait.hProcess == INVALID_HANDLE_VALUE || wait.hProcess == NULL)
	{
		smpd_dbg_printf(L"No process to wait for.\n");
		*exit_code_ptr = -1;
		smpd_exit_fn(FCNAME);
		return SMPD_SUCCESS;
	}
	if (WaitForSingleObject(wait.hProcess, INFINITE) != WAIT_OBJECT_0)
	{
		smpd_err_printf(L"WaitForSingleObject failed, error %d\n", GetLastError());
		*exit_code_ptr = -1;
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	result = GetExitCodeProcess(wait.hProcess, &exit_code);
	if (!result)
	{
		smpd_err_printf(L"GetExitCodeProcess failed, error %d\n", GetLastError());
		*exit_code_ptr = -1;
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	CloseHandle(wait.hProcess);
	CloseHandle(wait.hThread);

	*exit_code_ptr = exit_code;

	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}

#undef FCNAME
#define FCNAME L"smpd_suspend_process"
int smpd_suspend_process(smpd_process_t *process)
{
	int result = SMPD_SUCCESS;
	smpd_enter_fn(FCNAME);

	if (SuspendThread(process->wait.hThread) == -1)
	{
		result = GetLastError();
		smpd_err_printf(L"SuspendThread failed with error %d for process %d:%s:'%s'\n",
			result, process->rank, process->kvs_name, process->exe);
	}

	smpd_exit_fn(FCNAME);
	return result;
}

static BOOL SafeTerminateProcess(HANDLE hProcess, UINT uExitCode)
{
	DWORD dwTID, dwCode, dwErr = 0;
	HANDLE hProcessDup = INVALID_HANDLE_VALUE;
	HANDLE hRT = NULL;
	HINSTANCE hKernel = GetModuleHandle(L"Kernel32");
	BOOL bSuccess = FALSE;

	BOOL bDup = DuplicateHandle(GetCurrentProcess(),
		hProcess,
		GetCurrentProcess(),
		&hProcessDup,
		PROCESS_ALL_ACCESS,
		FALSE,
		0);

	if (GetExitCodeProcess((bDup) ? hProcessDup : hProcess, &dwCode) &&
		(dwCode == STILL_ACTIVE))
	{
		FARPROC pfnExitProc;

		pfnExitProc = GetProcAddress(hKernel, "ExitProcess");

		if (pfnExitProc)
		{
			hRT = CreateRemoteThread((bDup) ? hProcessDup : hProcess,
				NULL,
				0,
				/*
				This relies on the probability that Kernel32.dll is mapped to the same place on all processes
				If it gets relocated, this function will produce spurious results
				*/
				(LPTHREAD_START_ROUTINE)pfnExitProc,
				UintToPtr(uExitCode)/*(LPVOID)uExitCode*/, 0, &dwTID);
		}
		
		if (hRT == NULL)
			dwErr = GetLastError();
	}
	else
	{
		dwErr = ERROR_PROCESS_ABORTED;
	}

	if (hRT)
	{
		if (WaitForSingleObject((bDup) ? hProcessDup : hProcess, 30000) == WAIT_OBJECT_0)
			bSuccess = TRUE;
		else
		{
			dwErr = ERROR_TIMEOUT;
			bSuccess = FALSE;
		}
		CloseHandle(hRT);
	}

	if (bDup)
		CloseHandle(hProcessDup);

	if (!bSuccess)
		SetLastError(dwErr);

	return bSuccess;
}

#undef FCNAME
#define FCNAME L"smpd_call_abort_hook"
int smpd_call_abort_hook(smpd_process_t *process, int exit_code)
{
	int result = SMPD_FAIL;
	HANDLE hProcess;
	HANDLE hThread;

	hProcess = OpenProcess(PROCESS_CREATE_THREAD | PROCESS_QUERY_INFORMATION | PROCESS_VM_OPERATION | PROCESS_VM_WRITE | PROCESS_VM_READ,
		FALSE, process->abort_pid);
	if (hProcess != NULL)
	{
		hThread = CreateRemoteThread(hProcess, NULL, 0, (LPTHREAD_START_ROUTINE)(process->abort_fn), (LPVOID)MPIU_IntToPtr(exit_code), 0, NULL);
		if (hThread != NULL)
		{
			WaitForSingleObject(hThread, smpd_process.abort_callback_timeout * 1000);
			CloseHandle(hThread);
			result = SMPD_SUCCESS;
		}
		CloseHandle(hProcess);
	}
	return result;
}

#undef FCNAME
#define FCNAME L"smpd_kill_process"
int smpd_kill_process(smpd_process_t *process, int exit_code)
{
	smpd_enter_fn(FCNAME);

	smpd_process_from_registry(process);
	if (!SafeTerminateProcess(process->wait.hProcess, exit_code))
	{
		if (GetLastError() != ERROR_PROCESS_ABORTED)
		{
			TerminateProcess(process->wait.hProcess, exit_code);
		}
	}

	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}

#undef FCNAME
#define FCNAME L"smpd_kill_all_processes"
int smpd_kill_all_processes(void)
{
	smpd_process_t *iter;

	smpd_enter_fn(FCNAME);

	if (smpd_process.local_root)
	{
		/* the mpiexec process should not kill the smpd manager that it created for the -localroot option */
		smpd_exit_fn(FCNAME);
		return SMPD_SUCCESS;
	}

	if (smpd_process.rsh_mpiexec)
	{
		int i;
		int count = 0;
		iter = smpd_process.process_list;
		while (iter)
		{
			count++;
			iter = iter->next;
		}
		if (count > 0)
		{
			smpd_pwait_t *wait_array;
			wait_array = (smpd_pwait_t*)malloc(sizeof(smpd_pwait_t) * count);
			for (i=0, iter = smpd_process.process_list; i<count; i++)
			{
				wait_array[i] = iter->wait;
				iter = iter->next;
			}
			for (i=0; i<count; i++)
			{
				if (!SafeTerminateProcess(wait_array[i].hProcess, 123))
				{
					if (GetLastError() != ERROR_PROCESS_ABORTED)
					{
						TerminateProcess(wait_array[i].hProcess, 255);
					}
				}
			}
		}
	}
	else
	{
		iter = smpd_process.process_list;
		while (iter)
		{
			/*DWORD dwProcessId;*/
			smpd_process_from_registry(iter);
			/* For some reason break signals don't work on processes created by smpd_launch_process
			wprintf(L"ctrl-c process: %s\n", iter->exe);fflush(stdout);
			dwProcessId = GetProcessId(iter->wait.hProcess);
			GenerateConsoleCtrlEvent(CTRL_C_EVENT, dwProcessId);
			if (WaitForSingleObject(iter->wait.hProcess, 1000) != WAIT_OBJECT_0)
			{
			wprintf(L"breaking process: %s\n", iter->exe);fflush(stdout);
			GenerateConsoleCtrlEvent(CTRL_BREAK_EVENT, dwProcessId);
			if (WaitForSingleObject(iter->wait.hProcess, 1000) != WAIT_OBJECT_0)
			{
			*/
			if (!SafeTerminateProcess(iter->wait.hProcess, 123))
			{
				if (GetLastError() != ERROR_PROCESS_ABORTED)
				{
					TerminateProcess(iter->wait.hProcess, 255);
				}
			}
			/*
			}
			}
			*/
			iter = iter->next;
		}
	}

	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}

#undef FCNAME
#define FCNAME L"smpd_exit"
int smpd_exit(int exitcode)
{
	smpd_enter_fn(FCNAME);
	smpd_kill_all_processes();
	smpd_finalize_drive_maps();
	smpd_finalize_printf();
	PMPI_Finalize();
	/* If we're exiting due to a user abort, use the exit code supplied by the abort call */
	if (smpd_process.use_abort_exit_code)
		exitcode = smpd_process.abort_exit_code;
	if (smpd_process.hCloseStdinThreadEvent)
	{
		CloseHandle(smpd_process.hCloseStdinThreadEvent);
	}
	/* This is necessary because exit() can deadlock flushing file buffers while the stdin thread is running */
	ExitProcess(exitcode);
}
