/*   Copyright (C) 2009 Deino Software. */

#include "smpd.h"

static SMPD_BOOL UnmapDrive(wchar_t *pszDrive, wchar_t *pszError, int maxerrlength);
static SMPD_BOOL MapDrive(wchar_t *pszDrive, wchar_t *pszShare, wchar_t *pszAccount, wchar_t *pszPassword, wchar_t *pszError, int maxerrlength);

typedef struct DriveMapStruct
{
	int nRefCount;
	wchar_t pszDrive[10];
	wchar_t pszShare[MAX_PATH];
	SMPD_BOOL bUnmap;
	struct DriveMapStruct *pNext;
} DriveMapStruct;

DriveMapStruct * allocate_DriveMapStruct()
{
	DriveMapStruct *p;
	p = (DriveMapStruct*)malloc(sizeof(DriveMapStruct));
	if (p == NULL)
		return NULL;
	p->nRefCount = 1;
	p->pszDrive[0] = L'\0';
	p->pNext = NULL;
	p->bUnmap = SMPD_TRUE;
	return p;
}

static DriveMapStruct *g_pDriveList = NULL;

static SMPD_BOOL AlreadyMapped(wchar_t *pszDrive, wchar_t *pszShare, SMPD_BOOL *pMatched)
{
	DriveMapStruct *p;
	if (g_pDriveList == NULL)
		return SMPD_FALSE;
	p = g_pDriveList;
	while (p)
	{
		if (pszDrive[0] == p->pszDrive[0])
		{
			if ((_wcsicmp(pszShare, p->pszShare) == 0))
			{
				p->nRefCount++;
				*pMatched = SMPD_TRUE;
			}
			else
				*pMatched = SMPD_FALSE;
			return SMPD_TRUE;
		}
		p = p->pNext;
	}
	return SMPD_FALSE;
}

#if 1
static SMPD_BOOL CompareHosts(wchar_t *pwszHost1, wchar_t *pwszHost2)
{
	unsigned long ip1, ip2;
	struct hostent *pH;
	char pszHost1[100], pszHost2[100];
	size_t num_converted;

	pszHost1[0] = '\0';
	pszHost2[0] = '\0';
	wcstombs_s(&num_converted, pszHost1, 100, pwszHost1, _TRUNCATE);
	wcstombs_s(&num_converted, pszHost2, 100, pwszHost2, _TRUNCATE);

	pH = gethostbyname(pszHost1);
	if (pH == NULL)
		return SMPD_FALSE;

	ip1 = *(unsigned long*)(pH->h_addr_list[0]);

	pH = gethostbyname(pszHost2);
	if (pH == NULL)
		return SMPD_FALSE;

	ip2 = *(unsigned long*)(pH->h_addr_list[0]);

	return (ip1 == ip2);
}
#else
static SMPD_BOOL CompareHosts(wchar_t *pszHost1, wchar_t *pszHost2)
{
	SMPD_BOOL result;
	PADDRINFOW iter1, iter2;

	if (GetAddrInfo(pszHost1, NULL, NULL, &iter1))
		return SMPD_FALSE;

	if (GetAddrInfo(pszHost2, NULL, NULL, &iter2))
	{
		FreeAddrInfo(iter1);
		return SMPD_FALSE;
	}

	if (_wcsicmp(iter1->ai_canonname, iter2->ai_canonname) == 0)
		result = SMPD_TRUE;
	else
		result = SMPD_FALSE;
	FreeAddrInfo(iter1);
	FreeAddrInfo(iter2);
	return result;
}
#endif

static BOOL EnumerateDisksFunc(LPNETRESOURCE lpnr, DWORD dwScope, DWORD dwType, wchar_t *pszDrive, wchar_t *pszShare, SMPD_BOOL *pbFound, SMPD_BOOL *pbMatched)
{
	DWORD dwResult, dwResultEnum;
	HANDLE hEnum;
	DWORD cbBuffer = 16384;	  /* 16K is a good size */
	DWORD cEntries = (DWORD)-1;		 /* enumerate all possible entries */
	LPNETRESOURCE lpnrLocal;	 /* pointer to enumerated structures */
	DWORD i;

	dwResult = WNetOpenEnum(
		dwScope,
		dwType,
		0,		/* enumerate all resources */
		lpnr,	 /* NULL first time the function is called */
		&hEnum);  /* handle to the resource */

	if (dwResult != NO_ERROR)
		return FALSE;

	lpnrLocal = (LPNETRESOURCE) GlobalAlloc(GPTR, cbBuffer);

	do
	{  
		ZeroMemory(lpnrLocal, cbBuffer);
		dwResultEnum = WNetEnumResource(
			hEnum,	  /* resource handle */
			&cEntries,  /* defined locally as -1 */
			lpnrLocal,  /* LPNETRESOURCE */
			&cbBuffer); /* buffer size */
		/* If the call succeeds, loop through the structures. */
		if (dwResultEnum == NO_ERROR)
		{
			for(i = 0; i < cEntries; i++)
			{
				if (lpnrLocal[i].lpLocalName && lpnrLocal[i].lpRemoteName)
				{
					if (toupper(*lpnrLocal[i].lpLocalName) == *pszDrive)
					{
						*pbFound = SMPD_TRUE;
						if ((_wcsicmp(lpnrLocal[i].lpRemoteName, pszShare) == 0))
						{
							*pbMatched = SMPD_TRUE;
						}
						else
						{
							wchar_t *pPath1, *pPath2;
							pPath1 = wcsstr(&lpnrLocal[i].lpRemoteName[2], L"\\");
							if (pPath1 != NULL)
							{
								pPath1++; /* advance over the \ character */
								pPath2 = wcsstr(&pszShare[2], L"\\");
								if (pPath2 != NULL)
								{
									pPath2++; /* advance over the \ character */
									if (_wcsicmp(pPath1, pPath2) == 0)
									{
										wchar_t pszHost1[50], pszHost2[50];
										int nLength1, nLength2;
										nLength1 = (int)(pPath1 - &lpnrLocal[i].lpRemoteName[2] - 1);
										nLength2 = (int)(pPath2 - &pszShare[2] - 1);
										MPIU_Wcsncpy(pszHost1, &lpnrLocal[i].lpRemoteName[2], nLength1);
										MPIU_Wcsncpy(pszHost2, &pszShare[2], nLength2);
										pszHost1[nLength1] = L'\0';
										pszHost2[nLength2] = L'\0';
										if (CompareHosts(pszHost1, pszHost2))
											*pbMatched = SMPD_TRUE;
									}
								}
							}
						}
					}
				}

				/* If the NETRESOURCE structure represents a container resource,  */
				/*  call the EnumerateDisksFunc function recursively. */
				if (RESOURCEUSAGE_CONTAINER == (lpnrLocal[i].dwUsage & RESOURCEUSAGE_CONTAINER))
					EnumerateDisksFunc(&lpnrLocal[i], dwScope, dwType, pszDrive, pszShare, pbFound, pbMatched);
			}
		}
		else if (dwResultEnum != ERROR_NO_MORE_ITEMS)
		{
			break;
		}
	} while(dwResultEnum != ERROR_NO_MORE_ITEMS);

	GlobalFree((HGLOBAL)lpnrLocal);

	dwResult = WNetCloseEnum(hEnum);

	if (dwResult != NO_ERROR)
		return FALSE;

	return TRUE;
}

static SMPD_BOOL MatchesExistingMapping(wchar_t *pszDrive, wchar_t *pszShare)
{
	SMPD_BOOL bFound = SMPD_FALSE;
	SMPD_BOOL bMatched = SMPD_FALSE;
	wchar_t ch;

	if (pszDrive == NULL || pszShare == NULL)
		return SMPD_FALSE;

	ch = (wchar_t)(toupper(*pszDrive));
	EnumerateDisksFunc(NULL, RESOURCE_CONNECTED, RESOURCETYPE_DISK, &ch, pszShare, &bFound, &bMatched);
	if (bMatched)
		return SMPD_TRUE;
	EnumerateDisksFunc(NULL, RESOURCE_REMEMBERED, RESOURCETYPE_DISK, &ch, pszShare, &bFound, &bMatched);
	if (bMatched)
		return SMPD_TRUE;

	/* If it was not found, assume that it matches */
	return !bFound;
}

static void RemoveDriveStruct(wchar_t *pszDrive)
{
	DriveMapStruct *p, *pTrailer;

	pTrailer = p = g_pDriveList;
	while (p)
	{
		if (p->pszDrive[0] == pszDrive[0])
		{
			p->nRefCount--;
			if (p->nRefCount == 0)
			{
				if (pTrailer != p)
					pTrailer->pNext = p->pNext;
				if (g_pDriveList == p)
					g_pDriveList = g_pDriveList->pNext;
				free(p);
			}
			return;
		}
		if (pTrailer != p)
			pTrailer = pTrailer->pNext;
		p = p->pNext;
	}
}

#undef FCNAME
#define FCNAME L"smpd_finalize_drive_maps"
void smpd_finalize_drive_maps()
{
	wchar_t err_msg[256];
	wchar_t temp[MAX_PATH+2];

	smpd_enter_fn(FCNAME);
	while (g_pDriveList)
	{
		_snwprintf_s(temp, MAX_PATH+2, _TRUNCATE, L"%c:%s", g_pDriveList->pszDrive, g_pDriveList->pszShare);
		if (!UnmapDrive(temp, err_msg, 256))
			break;
	}
	while (g_pDriveList)
	{
		RemoveDriveStruct(g_pDriveList->pszDrive);
	}
	smpd_exit_fn(FCNAME);
}

static SMPD_BOOL ParseDriveShareAccountPassword(wchar_t *str, wchar_t *pszDrive, size_t maxlen_drive, wchar_t *pszShare, size_t maxlen_share, wchar_t *pszAccount, size_t maxlen_account, wchar_t *pszPassword, size_t maxlen_password)
{
	MPIU_UNREFERENCED_ARG(maxlen_account);
	MPIU_UNREFERENCED_ARG(maxlen_drive);

	pszDrive[0] = str[0];
	pszDrive[1] = L':';
	pszDrive[2] = L'\0';
	while (*str != L'\\')
		str++;
	if (wcsstr(str, L":"))
	{
		while (*str != L':')
			*pszShare++ = *str++;
		*pszShare = L'\0';
		str++;
		if (!wcsstr(str, L":"))
			return SMPD_FALSE;
		while (*str != L':')
			*pszAccount++ = *str++;
		*pszAccount = L'\0';
		str++;
		wcscpy_s(pszPassword, maxlen_password, str);
	}
	else
	{
		wcscpy_s(pszShare, maxlen_share, str);
		*pszAccount = L'\0';
	}
	return SMPD_TRUE;
}

#undef FCNAME
#define FCNAME L"smpd_map_user_drives"
SMPD_BOOL smpd_map_user_drives(wchar_t *pszMap, wchar_t *pszAccount, wchar_t *pszPassword, wchar_t *pszError, int maxerrlength)
{
	wchar_t pszDrive[3];
	wchar_t pszShare[MAX_PATH];
	wchar_t ipszAccount[SMPD_MAX_ACCOUNT_LENGTH];
	wchar_t ipszPassword[SMPD_MAX_PASSWORD_LENGTH];
	wchar_t *token;
	wchar_t *temp = _wcsdup(pszMap);
	wchar_t *tok;

	smpd_enter_fn(FCNAME);

	token = wcstok_s(temp, L";\n", &tok);
	if (token == NULL)
	{
		smpd_exit_fn(FCNAME);
		return SMPD_TRUE;
	}
	while (token != NULL)
	{
		ipszAccount[0] = L'\0';
		if (ParseDriveShareAccountPassword(token, pszDrive, 3, pszShare, MAX_PATH, ipszAccount, SMPD_MAX_ACCOUNT_LENGTH, ipszPassword, SMPD_MAX_PASSWORD_LENGTH))
		{
			if (ipszAccount[0]  != L'\0')
			{
				if (!MapDrive(pszDrive, pszShare, ipszAccount, ipszPassword, pszError, maxerrlength))
				{
					free(temp);
					smpd_err_printf(L"MapUserDrives: MapDrive(%s, %s, %s, ... ) failed, %s\n", pszDrive, pszShare, ipszAccount, pszError);
					smpd_exit_fn(FCNAME);
					return SMPD_FALSE;
				}
			}
			else
			{
				if (!MapDrive(pszDrive, pszShare, pszAccount, pszPassword, pszError, maxerrlength))
				{
					free(temp);
					smpd_err_printf(L"MapUserDrives: MapDrive(%s, %s, %s, ... ) failed, %s\n", pszDrive, pszShare, pszAccount, pszError);
					smpd_exit_fn(FCNAME);
					return SMPD_FALSE;
				}
			}
		}
		token = wcstok_s(NULL, L";\n", &tok);
	}
	free(temp);

	smpd_exit_fn(FCNAME);
	return SMPD_TRUE;
}

#undef FCNAME
#define FCNAME L"smpd_unmap_user_drives"
SMPD_BOOL smpd_unmap_user_drives(wchar_t *pszMap)
{
	wchar_t pszError[256];
	wchar_t pszDrive[3];
	wchar_t pszShare[MAX_PATH];
	wchar_t pszAccount[SMPD_MAX_ACCOUNT_LENGTH];
	wchar_t pszPassword[SMPD_MAX_PASSWORD_LENGTH];
	wchar_t *temp = _wcsdup(pszMap);
	wchar_t *token;
	wchar_t *tok;

	smpd_enter_fn(FCNAME);

	token = wcstok_s(temp, L";\n", &tok);
	if (token == NULL)
	{
		free(temp);
		smpd_exit_fn(FCNAME);
		return SMPD_TRUE;
	}
	while (token != NULL)
	{
		if (ParseDriveShareAccountPassword(token, pszDrive, 3, pszShare, MAX_PATH, pszAccount, SMPD_MAX_ACCOUNT_LENGTH, pszPassword, SMPD_MAX_PASSWORD_LENGTH))
		{
			if (!UnmapDrive(pszDrive, pszError, 256))
			{
				free(temp);
				smpd_exit_fn(FCNAME);
				return SMPD_FALSE;
			}
		}
		token = wcstok_s(NULL, L";\n", &tok);
	}
	free(temp);

	smpd_exit_fn(FCNAME);
	return SMPD_FALSE;
}

static SMPD_BOOL MapDrive(wchar_t *pszDrive, wchar_t *pszShare, wchar_t *pszAccount, wchar_t *pszPassword, wchar_t *pszError, int maxerrlength)
{
	wchar_t pszDriveLetter[3];
	DWORD dwResult;
	wchar_t pszName[1024];
	wchar_t pszProvider[256];
	NETRESOURCE net;
	SMPD_BOOL bMatched;

	if (pszDrive == NULL)
	{
		wcscpy_s(pszError, maxerrlength, L"Invalid drive string");
		return SMPD_FALSE;
	}
	pszDriveLetter[0] = pszDrive[0];
	pszDriveLetter[1] = L':';
	pszDriveLetter[2] = L'\0';

	memset(&net, 0, sizeof(NETRESOURCE));
	net.lpLocalName = pszDriveLetter;
	net.lpRemoteName = pszShare;
	net.dwType = RESOURCETYPE_DISK;
	net.lpProvider = NULL;

	if (AlreadyMapped(pszDriveLetter, pszShare, &bMatched))
	{
		if (bMatched)
			return SMPD_TRUE;
		swprintf_s(pszError, maxerrlength, L"Drive %s already mapped.", pszDrive);
		smpd_err_printf(L"MapDrive failed, drive is already mapped\n");
		return SMPD_FALSE;
	}

	if (pszAccount != NULL)
	{
		if (*pszAccount == L'\0')
		{
			/* Change empty username to NULL pointer to work with WNetAddConnection2. */
			pszAccount = NULL;
			pszPassword = NULL;
		}
	}

	dwResult = WNetAddConnection2(&net, pszPassword, pszAccount, CONNECT_REDIRECT);

	if (dwResult == NO_ERROR)
	{
		DriveMapStruct *p = allocate_DriveMapStruct();
		if (p == NULL)
		{
			smpd_err_printf(L"unable to allocate a drive map structure.\n");
			return SMPD_FALSE;
		}
		wcscpy_s(p->pszDrive, 10, pszDriveLetter);
		MPIU_Wcsncpy(p->pszShare, pszShare, MAX_PATH);
		p->pNext = g_pDriveList;
		g_pDriveList = p;
		return SMPD_TRUE;
	}

	switch (dwResult)
	{
	case ERROR_ACCESS_DENIED:
		wcscpy_s(pszError, maxerrlength, L"Access to the network resource was denied.");
		break;
	case ERROR_ALREADY_ASSIGNED:
		if (MatchesExistingMapping(pszDriveLetter, pszShare))
		{
			DriveMapStruct *p = allocate_DriveMapStruct();
			wcscpy_s(p->pszDrive, 10, pszDriveLetter);
			MPIU_Wcsncpy(p->pszShare, pszShare, MAX_PATH);
			p->bUnmap = SMPD_FALSE; /* don't unmap this drive since it was mapped outside mpd */
			p->pNext = g_pDriveList;
			g_pDriveList = p;
			return SMPD_TRUE;
		}
		else
		{
			swprintf_s(pszError, maxerrlength, L"The local device '%s' is already connected to a network resource.", pszDriveLetter);
		}
		break;
	case ERROR_BAD_DEV_TYPE:
		wcscpy_s(pszError, maxerrlength, L"The type of local device and the type of network resource do not match.");
		break;
	case ERROR_BAD_DEVICE:
		swprintf_s(pszError, maxerrlength, L"The value '%s' is invalid.", pszDriveLetter);
		break;
	case ERROR_BAD_NET_NAME:
		swprintf_s(pszError, maxerrlength, L"The value '%s' is not acceptable to any network resource provider because the resource name is invalid, or because the named resource cannot be located.", pszShare);
		break;
	case ERROR_BAD_PROFILE:
		wcscpy_s(pszError, maxerrlength, L"The user profile is in an incorrect format.");
		break;
	case ERROR_BAD_PROVIDER:
		wcscpy_s(pszError, maxerrlength, L"The value specified by the lpProvider member does not match any provider.");
		break;
	case ERROR_BUSY:
		wcscpy_s(pszError, maxerrlength, L"The router or provider is busy, possibly initializing. The caller should retry.");
		break;
	case ERROR_CANCELLED:
		wcscpy_s(pszError, maxerrlength, L"The attempt to make the connection was canceled by the user through a dialog box from one of the network resource providers, or by a called resource.");
		break;
	case ERROR_CANNOT_OPEN_PROFILE:
		wcscpy_s(pszError, maxerrlength, L"The system is unable to open the user profile to process persistent connections.");
		break;
	case ERROR_DEVICE_ALREADY_REMEMBERED:
		if (MatchesExistingMapping(pszDriveLetter, pszShare))
		{
			DriveMapStruct *p = allocate_DriveMapStruct();
			wcscpy_s(p->pszDrive, 10, pszDriveLetter);
			MPIU_Wcsncpy(p->pszShare, pszShare, MAX_PATH);
			p->bUnmap = SMPD_FALSE; /* don't unmap this drive since it was mapped outside mpd */
			p->pNext = g_pDriveList;
			g_pDriveList = p;
			return SMPD_TRUE;
		}
		else
		{
			swprintf_s(pszError, maxerrlength, L"An entry for the device '%s' is already in the user profile.", pszDriveLetter);
		}
		break;
	case ERROR_EXTENDED_ERROR:
		if (WNetGetLastError(&dwResult, pszName, 1024, pszProvider, 256) == NO_ERROR)
		{
			swprintf_s(pszError, maxerrlength, L"'%s' returned this error: %d, %s", pszProvider, dwResult, pszName);
		}
		else
		{
			wcscpy_s(pszError, maxerrlength, L"A network-specific error occurred.");
		}
		break;
	case ERROR_INVALID_PASSWORD:
		wcscpy_s(pszError, maxerrlength, L"The specified password is invalid.");
		break;
	case ERROR_NO_NET_OR_BAD_PATH:
		wcscpy_s(pszError, maxerrlength, L"The operation could not be completed, either because a network component is not started, or because the specified resource name is not recognized.");
		break;
	case ERROR_NO_NETWORK:
		wcscpy_s(pszError, maxerrlength, L"The network is unavailable.");
		break;
	default:
		smpd_translate_win_error(dwResult, pszError, maxerrlength, NULL);
		smpd_err_printf(L"MapDrive: unknown error %d\n", dwResult);
		break;
	}

	smpd_err_printf(L"MapDrive failed, error: %s\n", pszError);
	return SMPD_FALSE;
}

static SMPD_BOOL UnmapDrive(wchar_t *pszDrive, wchar_t *pszError, int maxerrlength)
{
	wchar_t pszName[1024];
	wchar_t pszProvider[256];
	wchar_t pszDriveLetter[3];
	DWORD dwResult;
	DriveMapStruct *p = g_pDriveList;

	if (pszDrive == NULL)
	{
		return SMPD_FALSE;
	}
	pszDriveLetter[0] = pszDrive[0];
	pszDriveLetter[1] = L':';
	pszDriveLetter[2] = L'\0';

	while (p)
	{
		if (p->pszDrive[0] == pszDrive[0])
		{
			if (p->nRefCount > 1)
			{
				p->nRefCount--;
				return SMPD_TRUE;
			}
			break;
		}
		p = p->pNext;
	}

	if (p->bUnmap)
	{
		dwResult = WNetCancelConnection2(pszDriveLetter, 
			CONNECT_UPDATE_PROFILE, /* This option makes sure that the connection is not re-established at the next user logon */
			TRUE);
	}
	else
	{
		dwResult = NO_ERROR;
	}

	if (dwResult == NO_ERROR)
	{
		RemoveDriveStruct(pszDriveLetter);
		return SMPD_TRUE;
	}

	switch (dwResult)
	{
	case ERROR_BAD_PROFILE:
		wcscpy_s(pszError, maxerrlength, L"The user profile is in an incorrect format.");
		break;
	case ERROR_CANNOT_OPEN_PROFILE:
		wcscpy_s(pszError, maxerrlength, L"The system is unable to open the user profile to process persistent connections.");
		break;
	case ERROR_DEVICE_IN_USE:
		wcscpy_s(pszError, maxerrlength, L"The device is in use by an active process and cannot be disconnected.");
		break;
	case ERROR_EXTENDED_ERROR:
		if (WNetGetLastError(&dwResult, pszName, 1024, pszProvider, 256) == NO_ERROR)
		{
			swprintf_s(pszError, maxerrlength, L"'%s' returned this error: %d, %s", pszProvider, dwResult, pszName);
		}
		else
		{
			wcscpy_s(pszError, maxerrlength, L"A network-specific error occurred.");
		}
		break;
	case ERROR_NOT_CONNECTED:
		swprintf_s(pszError, maxerrlength, L"'%s' is not a redirected device, or the system is not currently connected to '%s'.", pszDriveLetter, pszDriveLetter);
		break;
	case ERROR_OPEN_FILES:
		wcscpy_s(pszError, maxerrlength, L"There are open files, the drive cannot be disconnected.");
		break;
	default:
		smpd_translate_win_error(dwResult, pszError, maxerrlength, NULL);
		break;
	}

	return SMPD_FALSE;
}
