/*   Copyright (C) 2009 Deino Software. */

#include "smpd.h"
#include <stdio.h>
#ifdef HAVE_STDARG_H
#include <stdarg.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#define SMPD_MAX_INDENT 20
static wchar_t indent[SMPD_MAX_INDENT+1] = L"";
static int cur_indent = 0;

void smpd_translate_win_error(int error, wchar_t *msg, int maxlen, wchar_t *prepend, ...)
{
	HLOCAL str;
	int num_characters;
	int len;
	va_list list;

	num_characters = FormatMessage(
		FORMAT_MESSAGE_FROM_SYSTEM |
		FORMAT_MESSAGE_ALLOCATE_BUFFER,
		0,
		error,
		MAKELANGID( LANG_NEUTRAL, SUBLANG_DEFAULT ),
		(LPTSTR) &str,
		0,0);
	if (prepend == NULL)
		memcpy(msg, str, min(maxlen, num_characters+1)*sizeof(wchar_t));
	else
	{
		va_start(list, prepend);
		len = _vsnwprintf_s(msg, maxlen, _TRUNCATE, prepend, list);
		va_end(list);
		msg += len;
		maxlen -= len;
		_snwprintf_s(msg, maxlen, _TRUNCATE, L"%s", str);
	}
	LocalFree(str);
}

wchar_t * get_sock_error_string(int error)
{
	static wchar_t str[1024];

	if (error == MPIDU_SOCK_SUCCESS)
		return L"operation completed successfully";

	str[0] = L'\0';
	MPIR_Err_get_string(error, str, 1024, MPIDU_Sock_get_error_class_string);

	return str;
}

wchar_t * smpd_get_context_str(smpd_context_t *context)
{
	if (context == NULL)
		return L"null";
	switch (context->type)
	{
	case SMPD_CONTEXT_INVALID:
		return L"invalid";
	case SMPD_CONTEXT_STDIN:
		return L"stdin";
	case SMPD_CONTEXT_MPIEXEC_STDIN:
		return L"mpi_stdin";
	case SMPD_CONTEXT_MPIEXEC_STDIN_RSH:
		return L"mpi_stdin_rsh";
	case SMPD_CONTEXT_STDOUT:
		return L"stdout";
	case SMPD_CONTEXT_STDOUT_RSH:
		return L"stdout_rsh";
	case SMPD_CONTEXT_STDERR:
		return L"stderr";
	case SMPD_CONTEXT_STDERR_RSH:
		return L"stderr_rsh";
	case SMPD_CONTEXT_PARENT:
		return L"parent";
	case SMPD_CONTEXT_LEFT_CHILD:
		return L"left";
	case SMPD_CONTEXT_RIGHT_CHILD:
		return L"right";
	case SMPD_CONTEXT_CHILD:
		return L"child";
	case SMPD_CONTEXT_LISTENER:
		return L"listener";
	case SMPD_CONTEXT_SMPD:
		return L"smpd";
	case SMPD_CONTEXT_PMI:
		return L"pmi";
	case SMPD_CONTEXT_TIMEOUT:
		return L"timeout";
	case SMPD_CONTEXT_MPIEXEC_ABORT:
		return L"SMPD_CONTEXT_MPIEXEC_ABORT";
	case SMPD_CONTEXT_UNDETERMINED:
		return L"undetermined";
	case SMPD_CONTEXT_FREED:
		return L"freed";
	}
	return L"unknown";
}

int smpd_init_printf(void)
{
#define MAX_ENV_STR_LEN 8192
	wchar_t envstr[MAX_ENV_STR_LEN];
	static int initialized = 0;
	wchar_t str[1024];
	int len;

	if (initialized)
		return 0;
	initialized = 1;

	smpd_process.dbg_state = SMPD_DBG_STATE_ERROUT;

	if (GetEnvironmentVariable(L"SMPD_DBG_OUTPUT", envstr, MAX_ENV_STR_LEN) > 0)
	{
		if (wcsstr(envstr, L"stdout"))
			smpd_process.dbg_state |= SMPD_DBG_STATE_STDOUT;
		if (wcsstr(envstr, L"log"))
			smpd_process.dbg_state |= SMPD_DBG_STATE_LOGFILE;
		if (wcsstr(envstr, L"rank"))
			smpd_process.dbg_state |= SMPD_DBG_STATE_PREPEND_RANK;
		if (wcsstr(envstr, L"trace"))
			smpd_process.dbg_state |= SMPD_DBG_STATE_TRACE;
		if (wcsstr(envstr, L"all"))
			smpd_process.dbg_state = SMPD_DBG_STATE_ALL;
	}

	if (smpd_option_on(L"log"))
		smpd_process.dbg_state |= SMPD_DBG_STATE_LOGFILE;
	if (smpd_option_on(L"prepend_rank"))
		smpd_process.dbg_state |= SMPD_DBG_STATE_PREPEND_RANK;
	if (smpd_option_on(L"trace"))
		smpd_process.dbg_state |= SMPD_DBG_STATE_TRACE;

	if (smpd_process.dbg_state & SMPD_DBG_STATE_LOGFILE)
	{
		if (GetEnvironmentVariable(L"SMPD_DBG_LOG_FILENAME", envstr, MAX_ENV_STR_LEN) > 0)
		{
			wcscpy_s(smpd_process.dbg_filename, SMPD_MAX_FILENAME, envstr);
		}
		else
		{
			if (smpd_get_smpd_data(L"logfile", smpd_process.dbg_filename, SMPD_MAX_FILENAME) != SMPD_SUCCESS)
			{
				smpd_process.dbg_state ^= SMPD_DBG_STATE_LOGFILE;
			}
		}
		if (smpd_get_smpd_data(L"max_logfile_size", str, 1024) == SMPD_SUCCESS)
		{
			smpd_process.dbg_file_size = _wtoi(str);
		}
		if (GetEnvironmentVariable(L"SMPD_MAX_LOG_FILE_SIZE", envstr, MAX_ENV_STR_LEN) > 0)
		{
			len = _wtoi(envstr);
			if (len > 0)
			{
				smpd_process.dbg_file_size = _wtoi(envstr);
			}
		}
	}

	if (!smpd_process.bOutputInitialized)
	{
		smpd_process.hOutputMutex = CreateMutex(NULL, FALSE, SMPD_OUTPUT_MUTEX_NAME);
		smpd_process.bOutputInitialized = TRUE;
	}
	return SMPD_SUCCESS;
}

int smpd_finalize_printf()
{
	fflush(stdout);
	fflush(stderr);
	return SMPD_SUCCESS;
}

void smpd_clean_output(wchar_t *str)
{
	wchar_t *temp_str;

	temp_str = wcsstr(str, L"password");
	while (temp_str != NULL)
	{
		smpd_hide_string_arg(temp_str, L"password");
		temp_str = wcsstr(temp_str+1, L"password");
	}

	temp_str = wcsstr(str, L"pwd");
	while (temp_str != NULL)
	{
		smpd_hide_string_arg(temp_str, L"pwd");
		temp_str = wcsstr(temp_str+1, L"pwd");
	}

	temp_str = wcsstr(str, L"phrase");
	while (temp_str != NULL)
	{
		smpd_hide_string_arg(temp_str, L"phrase");
		temp_str = wcsstr(temp_str+1, L"phrase");
	}
}

static int read_file(FILE *fin, void *buffer, int length)
{
	int num_read, total = 0;
	wchar_t *buf = buffer;

	while (length > 0)
	{
		num_read = (int)fread(buf, sizeof(wchar_t), (size_t)length, fin);
		if (num_read < 1)
		{
			return total;
		}
		length = length - num_read;
		buf = buf + num_read;
		total = total + num_read;
	}
	return total;
}

static int write_file(FILE *fout, void *buffer, int length)
{
	int num_written, total = 0;
	wchar_t *buf = buffer;

	while (length > 0)
	{
		num_written = (int)fwrite(buf, sizeof(wchar_t), (size_t)length, fout);
		if (num_written < 1)
		{
			return total;
		}
		length = length - num_written;
		buf = buf + num_written;
		total = total + num_written;
	}
	return total;
}

void smpd_trim_logfile()
{
	/* Do nothing until I can figure out why the trim code isn't working */
}

#if 0
void smpd_trim_logfile_new()
{
	static int count = 0;
	FILE *fout;
	long size;
	int num_read, num_written;
	wchar_t *buffer, *buffer_orig;
	wchar_t copy_cmd[4096];
	int docopy = 0;
	static copy_number = 0;
	int a, b;

	/* check every 1000 iterations */
	if (count++ % 1000)
		return;

	fout = _wfopen(smpd_process.dbg_filename, L"rb");
	if (fout == NULL)
	{
		/*wprintf(L"unable to open the smpd log file\n");*/
		return;
	}
	setvbuf(fout, NULL, _IONBF, 0);
	fseek(fout, 0, SEEK_END);
	size = ftell(fout);
	if (size > smpd_process.dbg_file_size)
	{
		if (copy_number == 0)
		{
			srand(smpd_getpid());
			copy_number = rand();
		}
		swprintf(copy_cmd, L"copy %s %s.%d.old", smpd_process.dbg_filename, smpd_process.dbg_filename, copy_number);
		system(copy_cmd);
		buffer = buffer_orig = malloc(size);
		if (buffer != NULL)
		{
			fseek(fout, 0, SEEK_SET);
			num_read = read_file(fout, buffer, size);
			if (num_read == size)
			{
				fclose(fout);
				fout = _wfopen(smpd_process.dbg_filename, L"rb");
				if (fout == NULL)
				{
				}
				setvbuf(fout, NULL, _IONBF, 0);
				fseek(fout, -4, SEEK_END);
				fread(&a, 1, 4, fout);
				fclose(fout);
				if (a != *(int*)&buffer[size-4])
				{
					FILE *ferr = _wfopen(L"c:\\temp\\smpd.err", L"a+");
					fwprintf(ferr, L"last 4 bytes not equal, a: %x != %x", a, *(int*)&buffer[size-4]);
					fclose(ferr);
				}
				fout = _wfopen(smpd_process.dbg_filename, L"wb");
				if (fout != NULL)
				{
					buffer = buffer + size - (smpd_process.dbg_file_size / 2);
					size = smpd_process.dbg_file_size / 2;
					num_written = write_file(fout, buffer, size);
					docopy = 1;
					if (num_written != size)
					{
						FILE *ferr = _wfopen(L"c:\\temp\\smpd.err", L"a+");
						fwprintf(ferr, L"wrote %d instead of %d bytes to the smpd log file.\n", num_written, size);
						fclose(ferr);
					}
					/*
					else
					{
						wprintf(L"wrote %d bytes to the smpd log file.\n", num_written);
						fflush(stdout);
					}
					*/
					fclose(fout);
					fout = _wfopen(smpd_process.dbg_filename, L"rb");
					if (fout == NULL)
					{
					}
					setvbuf(fout, NULL, _IONBF, 0);
					fseek(fout, -4, SEEK_END);
					fread(&b, 1, 4, fout);
					fclose(fout);
					if (b != *(int*)&buffer[size-4])
					{
						FILE *ferr = _wfopen(L"c:\\temp\\smpd.err", L"a+");
						fwprintf(ferr, L"last 4 bytes not equal, b: %x != %x", b, *(int*)&buffer[size-4]);
						fclose(ferr);
					}
				}
				/*
				else
				{
					wprintf(L"unable to truncate the smpd log file.\n");
					fflush(stdout);
				}
				*/
			}
			/*
			else
			{
				wprintf(L"read %d instead of %d bytes from the smpd log file.\n", num_read, size);
				fflush(stdout);
			}
			*/
			free(buffer_orig);
		}
		/*
		else
		{
			wprintf(L"malloc failed to allocate %d bytes.\n", smpd_process.dbg_file_size / 2);
			fflush(stdout);
		}
		*/
	}
	/*
	else
	{
		wprintf(L"smpd file size: %d\n", size);
		fflush(stdout);
	}
	*/
	fclose(fout);
	if (docopy)
	{
		swprintf(copy_cmd, L"copy %s %s.%d.new", smpd_process.dbg_filename, smpd_process.dbg_filename, copy_number);
		system(copy_cmd);
		copy_number++;
	}
}

void smpd_trim_logfile_old()
{
	static int count = 0;
	FILE *fout;
	long size;
	int num_read, num_written;
	int new_size;
	wchar_t *buffer;
	wchar_t copy_cmd[4096];
	int docopy = 0;
	static copy_number = 0;
	int a, b;

	/* check every 1000 iterations */
	if (count++ % 1000)
		return;

	fout = _wfopen(smpd_process.dbg_filename, L"rb");
	if (fout == NULL)
	{
		/*wprintf(L"unable to open the smpd log file\n");*/
		return;
	}
	fseek(fout, 0, SEEK_END);
	size = ftell(fout);
	if (size > smpd_process.dbg_file_size)
	{
		if (copy_number == 0)
		{
			srand(smpd_getpid());
			copy_number = rand();
		}
		swprintf(copy_cmd, L"copy %s %s.%d.old", smpd_process.dbg_filename, smpd_process.dbg_filename, copy_number);
		system(copy_cmd);
		new_size = smpd_process.dbg_file_size / 2;
		buffer = malloc(new_size);
		if (buffer != NULL)
		{
			fseek(fout, - (long)new_size, SEEK_END);
			num_read = read_file(fout, buffer, new_size);
			if (num_read == new_size)
			{
				fclose(fout);
				fout = _wfopen(smpd_process.dbg_filename, L"rb");
				fseek(fout, -4, SEEK_END);
				fread(&a, 1, 4, fout);
				fclose(fout);
				if (a != *(int*)&buffer[new_size-4])
				{
					FILE *ferr = _wfopen(L"c:\\temp\\smpd.err", L"a+");
					fwprintf(ferr, L"last 4 bytes not equal, a: %x != %x", a, *(int*)&buffer[new_size-4]);
					fclose(ferr);
				}
				fout = _wfopen(smpd_process.dbg_filename, L"wb");
				if (fout != NULL)
				{
					num_written = write_file(fout, buffer, new_size);
					docopy = 1;
					if (num_written != new_size)
					{
						FILE *ferr = _wfopen(L"c:\\temp\\smpd.err", L"a+");
						fwprintf(ferr, L"wrote %d instead of %d bytes to the smpd log file.\n", num_written, new_size);
						fclose(ferr);
					}
					/*
					else
					{
						wprintf(L"wrote %d bytes to the smpd log file.\n", num_written);
						fflush(stdout);
					}
					*/
					fclose(fout);
					fout = _wfopen(smpd_process.dbg_filename, L"rb");
					fseek(fout, -4, SEEK_END);
					fread(&b, 1, 4, fout);
					fclose(fout);
					if (b != *(int*)&buffer[new_size-4])
					{
						FILE *ferr = _wfopen(L"c:\\temp\\smpd.err", L"a+");
						fwprintf(ferr, L"last 4 bytes not equal, b: %x != %x", b, *(int*)&buffer[new_size-4]);
						fclose(ferr);
					}
				}
				/*
				else
				{
					wprintf(L"unable to truncate the smpd log file.\n");
					fflush(stdout);
				}
				*/
			}
			/*
			else
			{
				wprintf(L"read %d instead of %d bytes from the smpd log file.\n", num_read, new_size);
				fflush(stdout);
			}
			*/
			free(buffer);
		}
		/*
		else
		{
			wprintf(L"malloc failed to allocate %d bytes.\n", smpd_process.dbg_file_size / 2);
			fflush(stdout);
		}
		*/
	}
	/*
	else
	{
		wprintf(L"smpd file size: %d\n", size);
		fflush(stdout);
	}
	*/
	fclose(fout);
	if (docopy)
	{
		swprintf(copy_cmd, L"copy %s %s.%d.new", smpd_process.dbg_filename, smpd_process.dbg_filename, copy_number);
		system(copy_cmd);
		copy_number++;
	}
}
#endif

int smpd_err_printf(wchar_t *str, ...)
{
	va_list list;
	wchar_t *indent_str;
	wchar_t *cur_str;
	int num_bytes;

	if (smpd_process.id == -1)
		smpd_init_printf();

	if (!(smpd_process.dbg_state & (SMPD_DBG_STATE_ERROUT | SMPD_DBG_STATE_LOGFILE)))
		return 0;

	if (!smpd_process.bOutputInitialized)
	{
		smpd_process.hOutputMutex = CreateMutex(NULL, FALSE, SMPD_OUTPUT_MUTEX_NAME);
		smpd_process.bOutputInitialized = TRUE;
	}
	WaitForSingleObject(smpd_process.hOutputMutex, INFINITE);

	/* write the formatted string to a global buffer */

	if (smpd_process.dbg_state & SMPD_DBG_STATE_TRACE)
		indent_str = indent;
	else
		indent_str = L"";

	num_bytes = 0;
	if (smpd_process.dbg_state & SMPD_DBG_STATE_PREPEND_RANK)
	{
		/* prepend output with the process tree node id */
		num_bytes = _snwprintf_s(smpd_process.printf_buffer, SMPD_MAX_DBG_PRINTF_LENGTH, _TRUNCATE, L"[%02d:%d]%sERROR:", smpd_process.id, smpd_getpid(), indent_str);
	}
	else
	{
		num_bytes = _snwprintf_s(smpd_process.printf_buffer, SMPD_MAX_DBG_PRINTF_LENGTH, _TRUNCATE, L"%s", indent_str);
	}
	cur_str = &smpd_process.printf_buffer[num_bytes];

	va_start(list, str);
	num_bytes += _vsnwprintf_s(cur_str, SMPD_MAX_DBG_PRINTF_LENGTH - num_bytes, _TRUNCATE, str, list);
	va_end(list);

	/* strip protected fields - passwords, etc */
	smpd_clean_output(smpd_process.printf_buffer);

	if (smpd_process.dbg_state & SMPD_DBG_STATE_ERROUT)
	{
		/* use stdout instead of stderr so that ordering will be consistent with dbg messages */
		/*
		HANDLE hOutput = GetStdHandle(STD_OUTPUT_HANDLE);
		smpd_write_handle(hOutput, smpd_process.printf_buffer, (DWORD)((wcslen(smpd_process.printf_buffer) + 1) * sizeof(wchar_t)));
		FlushFileBuffers(hOutput);
		*/
		if (smpd_process.stdout_filename != NULL)
		{
			errno_t error;
			FILE *fout;
			error = _wfopen_s(&fout, smpd_process.stdout_filename, L"a+");
			if (fout != NULL)
			{
				fwprintf(fout, L"%s", smpd_process.printf_buffer);
				fflush(fout);
				fclose(fout);
			}
			else
			{
				wprintf(L"%s", smpd_process.printf_buffer);
				fflush(stdout);
			}
		}
		else
		{
			wprintf(L"%s", smpd_process.printf_buffer);
			fflush(stdout);
		}
	}
	if ((smpd_process.dbg_state & SMPD_DBG_STATE_LOGFILE) && (smpd_process.dbg_filename[0] != L'\0'))
	{
		errno_t error;
		FILE *fout;
		smpd_trim_logfile();
		error = _wfopen_s(&fout, smpd_process.dbg_filename, L"a+");
		if (error != 0 || fout == NULL)
		{
			smpd_process.dbg_state ^= SMPD_DBG_STATE_LOGFILE;
		}
		else
		{
			setvbuf(fout, NULL, _IONBF, 0);
			fwprintf(fout, L"%s", smpd_process.printf_buffer);
			fclose(fout);
		}
	}
	

	ReleaseMutex(smpd_process.hOutputMutex);
	return num_bytes;
}

int smpd_dbg_printf(wchar_t *str, ...)
{
	va_list list;
	wchar_t *indent_str;
	wchar_t *cur_str;
	int num_bytes;

	if (smpd_process.id == -1)
		smpd_init_printf();

	if (!(smpd_process.dbg_state & (SMPD_DBG_STATE_STDOUT | SMPD_DBG_STATE_LOGFILE)))
		return 0;

	if (!smpd_process.bOutputInitialized)
	{
		smpd_process.hOutputMutex = CreateMutex(NULL, FALSE, SMPD_OUTPUT_MUTEX_NAME);
		smpd_process.bOutputInitialized = TRUE;
	}
	WaitForSingleObject(smpd_process.hOutputMutex, INFINITE);

	/* write the formatted string to a global buffer */

	if (smpd_process.dbg_state & SMPD_DBG_STATE_TRACE)
		indent_str = indent;
	else
		indent_str = L"";

	num_bytes = 0;
	if (smpd_process.dbg_state & SMPD_DBG_STATE_PREPEND_RANK)
	{
		/* prepend output with the process tree node id */
		num_bytes = _snwprintf_s(smpd_process.printf_buffer, SMPD_MAX_DBG_PRINTF_LENGTH, _TRUNCATE, L"[%02d:%d]%s", smpd_process.id, smpd_getpid(), indent_str);
	}
	else
	{
		num_bytes = _snwprintf_s(smpd_process.printf_buffer, SMPD_MAX_DBG_PRINTF_LENGTH, _TRUNCATE, L"%s", indent_str);
	}
	cur_str = &smpd_process.printf_buffer[num_bytes];

	va_start(list, str);
	num_bytes += _vsnwprintf_s(cur_str, SMPD_MAX_DBG_PRINTF_LENGTH - num_bytes, _TRUNCATE, str, list);
	va_end(list);

	/* strip protected fields - passwords, etc */
	smpd_clean_output(smpd_process.printf_buffer);

	if (smpd_process.dbg_state & SMPD_DBG_STATE_STDOUT)
	{
		/*
		HANDLE hOutput = GetStdHandle(STD_OUTPUT_HANDLE);
		smpd_write_handle(hOutput, smpd_process.printf_buffer, (DWORD)((wcslen(smpd_process.printf_buffer) + 1) * sizeof(wchar_t)));
		FlushFileBuffers(hOutput);
		*/
		if (smpd_process.stdout_filename != NULL)
		{
			errno_t error;
			FILE *fout;
			error = _wfopen_s(&fout, smpd_process.stdout_filename, L"a+");
			if (error == 0 && fout != NULL)
			{
				fwprintf(fout, L"%s", smpd_process.printf_buffer);
				fflush(fout);
				fclose(fout);
			}
			else
			{
				wprintf(L"%s", smpd_process.printf_buffer);
				fflush(stdout);
			}
		}
		else
		{
			wprintf(L"%s", smpd_process.printf_buffer);
			fflush(stdout);
		}
	}
	if ((smpd_process.dbg_state & SMPD_DBG_STATE_LOGFILE) && (smpd_process.dbg_filename[0] != L'\0'))
	{
		errno_t error;
		FILE *fout = NULL;
		smpd_trim_logfile();
		error = _wfopen_s(&fout, smpd_process.dbg_filename, L"a+");
		if (error != 0 || fout == NULL)
		{
			/*smpd_process.dbg_state ^= SMPD_DBG_STATE_LOGFILE;*/
		}
		else
		{
			setvbuf(fout, NULL, _IONBF, 0);
			fwprintf(fout, L"%s", smpd_process.printf_buffer);
			fclose(fout);
		}
	}
	
	ReleaseMutex(smpd_process.hOutputMutex);
	return num_bytes;
}

int smpd_enter_fn(wchar_t *fcname)
{
	if (smpd_process.dbg_state & SMPD_DBG_STATE_TRACE)
	{
		smpd_dbg_printf(L"\\%s\n", fcname);
	}
	if (cur_indent >= 0 && cur_indent < SMPD_MAX_INDENT)
	{
		indent[cur_indent] = L'.';
		indent[cur_indent+1] = L'\0';
	}
	cur_indent++;
	return SMPD_SUCCESS;
}

int smpd_exit_fn(wchar_t *fcname)
{
	if (cur_indent > 0 && cur_indent < SMPD_MAX_INDENT)
	{
		indent[cur_indent-1] = L'\0';
	}
	cur_indent--;
	if (smpd_process.dbg_state & SMPD_DBG_STATE_TRACE)
	{
		smpd_dbg_printf(L"/%s\n", fcname);
	}
	return SMPD_SUCCESS;
}

SMPD_BOOL smpd_snprintf_update(wchar_t **str_pptr, int *len_ptr, wchar_t *str_format, ...)
{
	va_list list;
	int n;

	va_start(list, str_format);
	n = _vsnwprintf_s(*str_pptr, *len_ptr, _TRUNCATE, str_format, list);
	va_end(list);

	if (n < 0)
	{
		(*str_pptr)[(*len_ptr)-1] = L'\0';
		*len_ptr = 0;
		return SMPD_FALSE;
	}

	(*str_pptr )= &(*str_pptr)[n];
	*len_ptr = (*len_ptr) - n;

	return SMPD_TRUE;
}
