/*   Copyright (C) 2009 Deino Software. */

#include "smpd.h"

static wchar_t hexchar(int x)
{
	switch (x)
	{
	case 0: return L'0';
	case 1: return L'1';
	case 2: return L'2';
	case 3: return L'3';
	case 4: return L'4';
	case 5: return L'5';
	case 6: return L'6';
	case 7: return L'7';
	case 8: return L'8';
	case 9: return L'9';
	case 10: return L'A';
	case 11: return L'B';
	case 12: return L'C';
	case 13: return L'D';
	case 14: return L'E';
	case 15: return L'F';
	}
	return L'0';
}

static char charhex(wchar_t c)
{
	switch (c)
	{
	case L'0': return 0;
	case L'1': return 1;
	case L'2': return 2;
	case L'3': return 3;
	case L'4': return 4;
	case L'5': return 5;
	case L'6': return 6;
	case L'7': return 7;
	case L'8': return 8;
	case L'9': return 9;
	case L'A': return 10;
	case L'B': return 11;
	case L'C': return 12;
	case L'D': return 13;
	case L'E': return 14;
	case L'F': return 15;
	case L'a': return 10;
	case L'b': return 11;
	case L'c': return 12;
	case L'd': return 13;
	case L'e': return 14;
	case L'f': return 15;
	}
	return 0;
}

static void char_to_hex(char ch, wchar_t *hex)
{
	*hex = hexchar((ch>>4) & 0xF);
	hex++;
	*hex = hexchar(ch & 0xF);
}

static char hex_to_char(const wchar_t *hex)
{
	char ch;
	ch = charhex(*hex) << 4;
	hex++;
	ch = ch | charhex(*hex);
	return (char)ch;
}

int smpd_encode_buffer(wchar_t *dest, int dest_length, const char *src, int src_length, int *num_encoded)
{
	int n = 0;
	while (src_length && (dest_length > 2))
	{
		char_to_hex(*src, dest);
		src++;
		dest += 2;
		src_length--;
		dest_length -= 2;
		n++;
	}
	*dest = L'\0';
	*num_encoded = n;
	return SMPD_SUCCESS;
}

int smpd_decode_buffer(const wchar_t *str, char *dest, int length, int *num_decoded)
{
	int n = 0;

	if (length < 1)
	{
		return SMPD_FAIL;
	}

	while (*str != L'\0' && length)
	{
		*dest = hex_to_char(str);
		str += 2;
		dest++;
		length--;
		n++;
	}
	*num_decoded = n;
	return SMPD_SUCCESS;
}

#if 0
int smpd_encode_buffer(wchar_t *dest, int dest_length, const wchar_t *src, int src_length, int *num_encoded)
{
	int num_used;
	int n = 0;
	unsigned wchar_t ch;
	while (src_length && dest_length)
	{
		ch = *src;
		num_used = _snwprintf_s(dest, dest_length, _TRUNCATE, L"%02X", (unsigned int)ch);
		if (num_used < 0)
		{
			*num_encoded = n;
			/* FIXME: why does this return success on truncation? */
			return SMPD_SUCCESS;
		}
		/*smpd_dbg_printf(L" %c = %c%c\n", ch, dest[0], dest[1]);*/
		dest += num_used;
		dest_length -= num_used;
		src++;
		n++;
		src_length--;
	}
	*num_encoded = n;
	return SMPD_SUCCESS;
}

int smpd_decode_buffer(const wchar_t *str, wchar_t *dest, int length, int *num_decoded)
{
	wchar_t hex[3];
	int value;
	int n = 0;
	wchar_t *udest = dest;

	if (length < 1)
	{
		return SMPD_FAIL;
	}

	hex[2] = L'\0';
	while (*str != L'\0' && length)
	{
		hex[0] = *str;
		str++;
		hex[1] = *str;
		str++;
		swscanf_s(hex, L"%X", &value);
		*udest = (wchar_t)value;
		/*smpd_dbg_printf(L" %s = %c\n", hex, *dest);*/
		udest++;
		length--;
		n++;
	}
	/* FIXME: Error on truncation?
	if (*str != L'\0')
	{
		*num_decoded = n;
		return SMPD_FAIL;
	}
	*/
	*num_decoded = n;
	return SMPD_SUCCESS;
}
#endif

#undef FCNAME
#define FCNAME L"smpd_read"
int smpd_read(MPIDU_Sock_t sock, void *buf, MPIDU_Sock_size_t len)
{
	int result;
	MPIU_Size_t num_read;

	smpd_enter_fn(FCNAME);

	smpd_dbg_printf(L"reading %d bytes from sock %d\n", len, MPIDU_Sock_get_sock_id(sock));

	while (len)
	{
		/* aggressively read */
		result = MPIDU_Sock_read(sock, buf, len, &num_read);
		if (result != MPI_SUCCESS)
		{
			smpd_err_printf(L"Unable to read %d bytes,\nsock error: %s\n", len, get_sock_error_string(result));
			smpd_exit_fn(FCNAME);
			return SMPD_FAIL;
		}
		if (num_read == len)
		{
			smpd_exit_fn(FCNAME);
			return SMPD_SUCCESS;
		}

		if (num_read == 0)
		{
			Sleep(1);
		}
		else
		{
			buf = (char*)buf + num_read;
			len -= num_read;
		}
	}

	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}

#undef FCNAME
#define FCNAME L"smpd_write"
int smpd_write(MPIDU_Sock_t sock, void *buf, MPIDU_Sock_size_t len)
{
	int result;
	MPIU_Size_t num_written;

	smpd_enter_fn(FCNAME);

	smpd_dbg_printf(L"writing %d bytes to sock %d\n", len, MPIDU_Sock_get_sock_id(sock));

	while (len)
	{
		/* aggressively write */
		result = MPIDU_Sock_write(sock, buf, len, &num_written);
		if (result != MPI_SUCCESS)
		{
			smpd_err_printf(L"Unable to write %d bytes,\nsock error: %s\n",
							len, get_sock_error_string(result));
			smpd_exit_fn(FCNAME);
			return SMPD_FAIL;
		}
		if (num_written == len)
		{
			smpd_exit_fn(FCNAME);
			return SMPD_SUCCESS;
		}

		if (num_written == 0)
		{
			smpd_dbg_printf(L"0 bytes written.\n");
			Sleep(1);
		}
		else
		{
			smpd_dbg_printf(L"wrote %d bytes\n", num_written);
			buf = (char*)buf + num_written;
			len -= num_written;
		}
	}

	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}

#undef FCNAME
#define FCNAME L"smpd_write_string"
int smpd_write_string(MPIDU_Sock_t sock, wchar_t *str)
{
	int result;
	MPIU_Size_t len, num_written;
	LPBYTE buffer;

	smpd_enter_fn(FCNAME);

	smpd_dbg_printf(L"writing string on sock %d: \"%s\"\n", MPIDU_Sock_get_sock_id(sock), str);

	len = (MPIU_Size_t)((wcslen(str)+1)*sizeof(wchar_t));
	buffer = (LPBYTE)str;

	while (len)
	{
		/* aggressively write string */
		result = MPIDU_Sock_write(sock, buffer, len, &num_written);
		if (result != MPI_SUCCESS)
		{
			smpd_err_printf(L"Unable to write string of length %d,\nsock error: %s\n", len, get_sock_error_string(result));
			smpd_exit_fn(FCNAME);
			return SMPD_FAIL;
		}
		if (num_written == len)
		{
			smpd_exit_fn(FCNAME);
			return SMPD_SUCCESS;
		}

		/* post a write for whatever is left of the string */
		buffer += num_written;
		len -= num_written;
	}

	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}

static int read2(MPIDU_Sock_t sock, wchar_t *ch)
{
	MPIU_Size_t num_bytes;
	int result;
	LPBYTE buffer;

	buffer = (LPBYTE)ch;
	result = MPIDU_Sock_read(sock, buffer, 2, &num_bytes);
	if (result == MPI_SUCCESS)
	{
		if (num_bytes == 0)
			return 0;
		if (num_bytes == 1)
		{
			buffer++;
			result = MPIDU_Sock_read(sock, buffer, 1, &num_bytes);
			if (result == MPI_SUCCESS)
			{
				if (num_bytes == 0)
					return 1;
				return 2;
			}
			return 1;
		}
		return 2;
	}
	smpd_err_printf(L"Unable to read a string,\nsock error: %s\n", get_sock_error_string(result));
	return -1;
}

static int read_string(MPIDU_Sock_t sock, wchar_t *str, int maxlen)
{
	wchar_t ch;
	int result;
	int total = 0;

	if (maxlen < 1)
		return 0;
	result = read2(sock, &ch);
	while (result == 2)
	{
		total++;
		*str = ch;
		str++;
		if (ch == L'\0' || total >= maxlen)
			return total;
		result = read2(sock, &ch);
	}
	if (total > 0)
		return total;
	return -1;
}

static int chew_up_string(MPIDU_Sock_t sock)
{
	wchar_t ch;
	int result;

	result = smpd_read_string(sock, &ch, 1);
	while (result == SMPD_SUCCESS)
	{
		if (ch == L'\0')
			return SMPD_SUCCESS;
		result = smpd_read_string(sock, &ch, 1);
	}
	smpd_err_printf(L"Unable to read a string,\nsock error: %s\n", get_sock_error_string(result));
	return SMPD_FAIL;
}

#undef FCNAME
#define FCNAME L"smpd_read_string"
int smpd_read_string(MPIDU_Sock_t sock, wchar_t *str, int maxlen)
{
	int num_bytes;
	wchar_t *str_orig;

	smpd_enter_fn(FCNAME);

	str_orig = str;

	if (maxlen == 0)
	{
		smpd_dbg_printf(L"zero length read string request on sock %d\n", MPIDU_Sock_get_sock_id(sock));
		smpd_exit_fn(FCNAME);
		return SMPD_SUCCESS;
	}

	for (;;)
	{
		num_bytes = read_string(sock, str, maxlen);
		if (num_bytes == -1)
		{
			smpd_exit_fn(FCNAME);
			return SMPD_FAIL;
		}
		if (num_bytes > 0 && str[num_bytes-1] == L'\0')
		{
			smpd_dbg_printf(L"received string on sock %d: \"%s\"\n", MPIDU_Sock_get_sock_id(sock), str_orig);
			smpd_exit_fn(FCNAME);
			return SMPD_SUCCESS;
		}
		if (num_bytes == maxlen)
		{
			/* received truncated string */
			str[num_bytes-1] = L'\0';
			chew_up_string(sock);
			smpd_dbg_printf(L"received truncated string on sock %d: \"%s\"\n", MPIDU_Sock_get_sock_id(sock), str_orig);
			smpd_exit_fn(FCNAME);
			return SMPD_SUCCESS;
		}
		str += num_bytes;
		maxlen -= num_bytes;
	}
	/*
	smpd_exit_fn(FCNAME);
	return SMPD_FAIL;
	*/
}

#undef FCNAME
#define FCNAME L"smpd_write_handle"
int smpd_write_handle(HANDLE hOut, void *buffer, DWORD len)
{
	DWORD error;
	DWORD num_written = 0;

	/* This function cannot have enter/exit calls since it is used to implement those calls. infinite recursion */
	/*smpd_enter_fn(FCNAME);*/

	while (len > 0)
	{
		num_written = 0;
		if (!WriteFile(hOut, buffer, len, &num_written, NULL))
		{
			error = GetLastError();
			smpd_err_printf(L"unable to write %d bytes to a file handle, error %d\n", len, error);
			/*smpd_exit_fn(FCNAME);*/
			return SMPD_FAIL;
		}

		if (num_written < 0)
		{
			/* Is this possible? */
			num_written = 0;
		}

		len -= num_written;
		buffer = (BYTE*)buffer + num_written;
	}
	/*smpd_exit_fn(FCNAME);*/
	return SMPD_SUCCESS;
}
