/*   Copyright (C) 2009 Deino Software. */

#include "smpd.h"
#include <wincrypt.h>
#include <stdio.h>
#include <stdlib.h>

#define MAX_ERR_LENGTH 512

#undef FCNAME
#define FCNAME L"smpd_get_all_key_names"
static int smpd_get_all_key_names(smpd_data_t **data)
{
	HKEY tkey;
	DWORD result;
	LONG enum_result;
	wchar_t name[SMPD_MAX_NAME_LENGTH];
	DWORD name_length, index;
	smpd_data_t *list, *item;
	wchar_t err_msg[MAX_ERR_LENGTH];

	smpd_enter_fn(FCNAME);

	if (data == NULL)
	{
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}

	result = RegOpenKeyEx(HKEY_CURRENT_USER, MPICH_REGISTRY_KEY,
		0, 
		KEY_READ,
		&tkey);
	if (result != ERROR_SUCCESS)
	{
		/* No key therefore no settings */
		*data = NULL;
		smpd_exit_fn(FCNAME);
		return SMPD_SUCCESS;
	}

	list = NULL;
	index = 0;
	name_length = SMPD_MAX_NAME_LENGTH;
	enum_result = RegEnumValue(tkey, index, name, &name_length, NULL, NULL, NULL, NULL);
	while (enum_result == ERROR_SUCCESS)
	{
		item = (smpd_data_t*)malloc(sizeof(smpd_data_t));
		if (item == NULL)
		{
			*data = NULL;
			RegCloseKey(tkey);
			smpd_exit_fn(FCNAME);
			return SMPD_FAIL;
		}
		memcpy(item->name, name, SMPD_MAX_NAME_LENGTH*sizeof(wchar_t));
		item->value[0] = L'\0';
		item->next = list;
		list = item;
		index++;
		name_length = SMPD_MAX_NAME_LENGTH;
		enum_result = RegEnumValue(tkey, index, name, &name_length, NULL, NULL, NULL, NULL);
	}
	result = RegCloseKey(tkey);
	if (result != ERROR_SUCCESS)
	{
		smpd_translate_win_error(result, err_msg, MAX_ERR_LENGTH, L"Unable to close the HKEY_CURRENT_USER\\" MPICH_REGISTRY_KEY L" registry key, error %d: ", result);
		smpd_err_printf(L"%s\n", err_msg);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}

	*data = list;
	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}

#undef FCNAME
#define FCNAME L"smpd_delete_current_password_registry_entry"
SMPD_BOOL smpd_delete_current_password_registry_entry(int index)
{
	int nError;
	HKEY hRegKey = NULL;
	DWORD dwNumValues;
	wchar_t key_name[256] = L"smpdAccount";
	wchar_t key_pwd[256] = L"smpdPassword";

	smpd_enter_fn(FCNAME);

	if (index > 0)
	{
		swprintf_s(key_name, 256, L"smpdAccount%d", index);
		swprintf_s(key_pwd, 256, L"smpdPassword%d", index);
	}

	if (index >= 0)
	{
		nError = RegOpenKeyEx(HKEY_CURRENT_USER, MPICH_REGISTRY_KEY, 0, KEY_ALL_ACCESS, &hRegKey);
		if (nError != ERROR_SUCCESS && nError != ERROR_PATH_NOT_FOUND && nError != ERROR_FILE_NOT_FOUND)
		{
			smpd_err_printf(L"DeleteCurrentPasswordRegistryEntry:RegOpenKeyEx(...) failed, error: %d\n", nError);
			smpd_exit_fn(FCNAME);
			return SMPD_FALSE;
		}

		nError = RegDeleteValue(hRegKey, key_pwd);
		if (nError != ERROR_SUCCESS && nError != ERROR_FILE_NOT_FOUND)
		{
			smpd_err_printf(L"DeleteCurrentPasswordRegistryEntry:RegDeleteValue(password) failed, error: %d\n", nError);
			RegCloseKey(hRegKey);
			smpd_exit_fn(FCNAME);
			return SMPD_FALSE;
		}

		nError = RegDeleteValue(hRegKey, key_name);
		if (nError != ERROR_SUCCESS && nError != ERROR_FILE_NOT_FOUND)
		{
			smpd_err_printf(L"DeleteCurrentPasswordRegistryEntry:RegDeleteValue(account) failed, error: %d\n", nError);
			RegCloseKey(hRegKey);
			smpd_exit_fn(FCNAME);
			return SMPD_FALSE;
		}
	}
	else
	{
		/* find all keys named smpdAccount%d and smpdPassword%d and delete them */
		smpd_data_t *data;

		if (smpd_get_all_key_names(&data) == SMPD_SUCCESS)
		{
			smpd_data_t *iter = data;

			nError = RegOpenKeyEx(HKEY_CURRENT_USER, MPICH_REGISTRY_KEY, 0, KEY_ALL_ACCESS, &hRegKey);
			if (nError != ERROR_SUCCESS && nError != ERROR_PATH_NOT_FOUND && nError != ERROR_FILE_NOT_FOUND)
			{
				smpd_err_printf(L"DeleteCurrentPasswordRegistryEntry:RegOpenKeyEx(...) failed, error: %d\n", nError);
				smpd_exit_fn(FCNAME);
				return SMPD_FALSE;
			}

			while (iter != NULL)
			{
				if ((wcsncmp(iter->name, L"smpdAccount", 11) == 0) || (wcsncmp(iter->name, L"smpdPassword", 12) == 0))
				{
					nError = RegDeleteValue(hRegKey, iter->name);
					if (nError != ERROR_SUCCESS && nError != ERROR_FILE_NOT_FOUND)
					{
						smpd_err_printf(L"DeleteCurrentPasswordRegistryEntry:RegDeleteValue(%s) failed, error: %d\n", iter->name, nError);
						RegCloseKey(hRegKey);
						smpd_exit_fn(FCNAME);
						return SMPD_FALSE;
					}
				}
				iter = iter->next;
			}
			while (data != NULL)
			{
				iter = data;
				data = data->next;
				free(iter);
			}
		}
	}

	if (RegQueryInfoKey(hRegKey, NULL, NULL, NULL, NULL, NULL, NULL, &dwNumValues, NULL, NULL, NULL, NULL) != ERROR_SUCCESS)
	{
		RegCloseKey(hRegKey);
		RegDeleteKey(HKEY_CURRENT_USER, MPICH_REGISTRY_KEY);
	}
	else
	{
		RegCloseKey(hRegKey);
		if (dwNumValues == 0)
			RegDeleteKey(HKEY_CURRENT_USER, MPICH_REGISTRY_KEY);
	}

	smpd_exit_fn(FCNAME);
	return SMPD_TRUE;
}

#undef FCNAME
#define FCNAME L"smpd_save_password_to_registry"
SMPD_BOOL smpd_save_password_to_registry(int index, const wchar_t *szAccount, const wchar_t *szPassword, SMPD_BOOL persistent)
{
	int nError;
	SMPD_BOOL bResult = SMPD_TRUE;
	HKEY hRegKey = NULL;
	DATA_BLOB password_blob, blob;
	wchar_t key_name[256] = L"smpdAccount";
	wchar_t key_pwd[256] = L"smpdPassword";
	wchar_t err_msg[MAX_ERR_LENGTH];

	smpd_enter_fn(FCNAME);

	if (persistent)
	{
		if (RegCreateKeyEx(HKEY_CURRENT_USER, MPICH_REGISTRY_KEY,
			0, 
			NULL, 
			REG_OPTION_NON_VOLATILE,
			KEY_ALL_ACCESS, 
			NULL,
			&hRegKey, 
			NULL) != ERROR_SUCCESS) 
		{
			nError = GetLastError();
			smpd_err_printf(L"SavePasswordToRegistry:RegCreateKeyEx(...) failed, error: %d\n", nError);
			smpd_exit_fn(FCNAME);
			return SMPD_FALSE;
		}
	}
	else
	{
		/* delete the key if it exists */
		RegDeleteKey(HKEY_CURRENT_USER, MPICH_REGISTRY_KEY);
		/* create the key with the volatile option */
		if (RegCreateKeyEx(HKEY_CURRENT_USER, MPICH_REGISTRY_KEY,
			0, 
			NULL, 
			REG_OPTION_VOLATILE,
			KEY_ALL_ACCESS, 
			NULL,
			&hRegKey, 
			NULL) != ERROR_SUCCESS) 
		{
			nError = GetLastError();
			smpd_err_printf(L"SavePasswordToRegistry:RegDeleteKey(...) failed, error: %d\n", nError);
			smpd_exit_fn(FCNAME);
			return SMPD_FALSE;
		}
	}

	if (index > 0)
	{
		swprintf_s(key_name, 256, L"smpdAccount%d", index);
		swprintf_s(key_pwd, 256, L"smpdPassword%d", index);
	}

	/* Store the account name*/
	if (RegSetValueEx(hRegKey, key_name, 0, REG_SZ, (BYTE*)szAccount, (DWORD)(wcslen(szAccount)+1)*sizeof(wchar_t)) != ERROR_SUCCESS)
	{
		nError = GetLastError();
		smpd_err_printf(L"SavePasswordToRegistry:RegSetValueEx(...) failed, error: %d\n", nError);
		RegCloseKey(hRegKey);
		smpd_exit_fn(FCNAME);
		return SMPD_FALSE;
	}

	password_blob.cbData = (DWORD)(wcslen(szPassword)+1)*sizeof(wchar_t); /* store the NULL termination */
	password_blob.pbData = (BYTE*)szPassword;
	if (CryptProtectData(&password_blob, L"MPICH2 User Credentials", NULL, NULL, NULL, CRYPTPROTECT_UI_FORBIDDEN, &blob))
	{
		/* Write data to registry.*/
		if (RegSetValueEx(hRegKey, key_pwd, 0, REG_BINARY, blob.pbData, blob.cbData) != ERROR_SUCCESS)
		{
			nError = GetLastError();
			smpd_err_printf(L"SavePasswordToRegistry:RegSetValueEx(...) failed, error: %d\n", nError);
			bResult = SMPD_FALSE;
		}
		LocalFree(blob.pbData);
	}
	else
	{
		nError = GetLastError();
		smpd_err_printf(L"SavePasswordToRegistry:CryptProtectData(...) failed, error: %d\n", nError);
		bResult = SMPD_FALSE;
	}

	nError = RegCloseKey(hRegKey);
	if (nError != ERROR_SUCCESS)
	{
		smpd_translate_win_error(nError, err_msg, MAX_ERR_LENGTH, L"Unable to close the HKEY_CURRENT_USER\\" MPICH_REGISTRY_KEY L" registry key, error %d: ", nError);
		smpd_err_printf(L"%s\n", err_msg);
		smpd_exit_fn(FCNAME);
		return SMPD_FALSE;
	}

	smpd_exit_fn(FCNAME);
	return bResult;
}

#undef FCNAME
#define FCNAME L"smpd_read_password_from_registry"
SMPD_BOOL smpd_read_password_from_registry(int index, wchar_t *szAccount, size_t maxlen1, wchar_t *szPassword, size_t maxlen2) 
{
	int nError;
	SMPD_BOOL bResult = SMPD_TRUE;
	HKEY hRegKey = NULL;
	DWORD dwType;
	DATA_BLOB password_blob, blob;
	wchar_t key_name[256] = L"smpdAccount";
	wchar_t key_pwd[256] = L"smpdPassword";
	wchar_t err_msg[MAX_ERR_LENGTH];

	smpd_enter_fn(FCNAME);

	MPIU_UNREFERENCED_ARG(maxlen1);

	if (index > 0)
	{
		swprintf_s(key_name, 256, L"smpdAccount%d", index);
		swprintf_s(key_pwd, 256, L"smpdPassword%d", index);
	}

	if (RegOpenKeyEx(HKEY_CURRENT_USER, MPICH_REGISTRY_KEY, 0, KEY_QUERY_VALUE, &hRegKey) == ERROR_SUCCESS) 
	{
		DWORD dwLength = SMPD_MAX_PASSWORD_LENGTH;
		*szAccount = L'\0';
		if (RegQueryValueEx(hRegKey, key_name, NULL, NULL, (BYTE*)szAccount, &dwLength) != ERROR_SUCCESS)
		{
			nError = GetLastError();
			/*smpd_err_printf(L"ReadPasswordFromRegistry:RegQueryValueEx(...) failed, error: %d\n", nError);*/
			RegCloseKey(hRegKey);
			smpd_exit_fn(FCNAME);
			return SMPD_FALSE;
		}
		if (wcslen(szAccount) < 1)
		{
			smpd_exit_fn(FCNAME);
			return SMPD_FALSE;
		}

		dwType = REG_BINARY;
		if (RegQueryValueEx(hRegKey, key_pwd, NULL, &dwType, NULL, &dwLength) == ERROR_SUCCESS)
		{
			blob.cbData = dwLength;
			blob.pbData = MPIU_Malloc(dwLength);
			if (blob.pbData == NULL)
			{
				smpd_err_printf(L"Unable to allocate %d bytes of memory for decrypting the password from the registry.\n", dwLength);
				smpd_exit_fn(FCNAME);
				return SMPD_FALSE;
			}
			if (RegQueryValueEx(hRegKey, key_pwd, NULL, &dwType, (BYTE*)blob.pbData, &dwLength) == ERROR_SUCCESS)
			{
				if (CryptUnprotectData(&blob, NULL, NULL, NULL, NULL, CRYPTPROTECT_UI_FORBIDDEN, &password_blob))
				{
					wcscpy_s(szPassword, maxlen2, (const wchar_t *)(password_blob.pbData));
					LocalFree(password_blob.pbData);
				}
				else
				{
					nError = GetLastError();
					/*smpd_err_printf(L"ReadPasswordFromRegistry:CryptUnprotectData(...) failed, error: %d\n", nError);*/
					bResult = SMPD_FALSE;
				}
			}
			else
			{
				nError = GetLastError();
				/*smpd_err_printf(L"ReadPasswordFromRegistry:RegQueryValueEx(...) failed, error: %d\n", nError);*/
				bResult = SMPD_FALSE;
			}
			MPIU_Free(blob.pbData);
		}
		else
		{
			nError = GetLastError();
			/*smpd_err_printf(L"ReadPasswordFromRegistry:RegQueryValueEx(...) failed, error: %d\n", nError);*/
			bResult = SMPD_FALSE;
		}
		nError = RegCloseKey(hRegKey);
		if (nError != ERROR_SUCCESS)
		{
			smpd_translate_win_error(nError, err_msg, MAX_ERR_LENGTH, L"Unable to close the HKEY_CURRENT_USER\\" MPICH_REGISTRY_KEY L" registry key, error %d: ", nError);
			smpd_err_printf(L"%s\n", err_msg);
			smpd_exit_fn(FCNAME);
			return SMPD_FALSE;
		}
	}
	else
	{
		nError = GetLastError();
		/*smpd_err_printf(L"ReadPasswordFromRegistry:RegOpenKeyEx(...) failed, error: %d\n", nError);*/
		bResult = SMPD_FALSE;
	}

	smpd_exit_fn(FCNAME);
	return bResult;
}

#undef FCNAME
#define FCNAME L"smpd_cache_password"
int smpd_cache_password(const wchar_t *account, const wchar_t *password)
{
	int nError;
	wchar_t szEncodedPassword[SMPD_MAX_PASSWORD_LENGTH*2+1];
	HKEY hRegKey = NULL;
	int num_bytes;
	DWORD len;
	wchar_t err_msg[MAX_ERR_LENGTH];

	smpd_enter_fn(FCNAME);

	if (smpd_option_on(L"nocache"))
	{
		smpd_exit_fn(FCNAME);
		return SMPD_SUCCESS;
	}

	RegDeleteKey(HKEY_CURRENT_USER, SMPD_REGISTRY_CACHE_KEY);
	if (RegCreateKeyEx(HKEY_CURRENT_USER, SMPD_REGISTRY_CACHE_KEY,
		0, 
		NULL, 
		REG_OPTION_VOLATILE,
		KEY_ALL_ACCESS, 
		NULL,
		&hRegKey, 
		NULL) != ERROR_SUCCESS) 
	{
		nError = GetLastError();
		/*smpd_err_printf(L"CachePassword:RegDeleteKey(...) failed, error: %d\n", nError);*/
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}

	/* Store the account name*/
	len = (DWORD)(wcslen(account)+1)*sizeof(wchar_t);
	nError = RegSetValueEx(hRegKey, L"smpda", 0, REG_SZ, (BYTE*)account, len);
	if (nError != ERROR_SUCCESS)
	{
		/*smpd_err_printf(L"CachePassword:RegSetValueEx(%s) failed, error: %d\n", g_pszAccount, nError);*/
		RegCloseKey(hRegKey);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}

	/* encode the password*/
	smpd_encode_buffer(szEncodedPassword, SMPD_MAX_PASSWORD_LENGTH*2, (const char *)password, (int)((wcslen(password)+1)*sizeof(wchar_t)), &num_bytes);
	szEncodedPassword[num_bytes*2] = L'\0';
	/*smpd_dbg_printf(L"szEncodedPassword = '%s'\n", szEncodedPassword);*/

	/* Store the encoded password*/
	nError = RegSetValueEx(hRegKey, L"smpdp", 0, REG_SZ, (BYTE*)szEncodedPassword, (num_bytes*2+1)*sizeof(wchar_t));
	if (nError != ERROR_SUCCESS)
	{
		/*smpd_err_printf(L"CachePassword:RegSetValueEx(...) failed, error: %d\n", nError);*/
		RegCloseKey(hRegKey);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}

	nError = RegCloseKey(hRegKey);
	if (nError != ERROR_SUCCESS)
	{
		smpd_translate_win_error(nError, err_msg, MAX_ERR_LENGTH, L"Unable to close the HKEY_CURRENT_USER\\" SMPD_REGISTRY_CACHE_KEY L" registry key, error %d: ", nError);
		smpd_err_printf(L"%s\n", err_msg);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}

	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}

#undef FCNAME
#define FCNAME L"smpd_get_cached_password"
SMPD_BOOL smpd_get_cached_password(wchar_t *account, size_t maxlen1, wchar_t *password, size_t maxlen2)
{
	int nError;
	wchar_t szAccount[SMPD_MAX_ACCOUNT_LENGTH];
	wchar_t szPassword[SMPD_MAX_PASSWORD_LENGTH*2];
	HKEY hRegKey = NULL;
	DWORD dwLength;
	int num_bytes;
	wchar_t err_msg[MAX_ERR_LENGTH];

	smpd_enter_fn(FCNAME);

	if (RegOpenKeyEx(HKEY_CURRENT_USER, SMPD_REGISTRY_CACHE_KEY, 0, KEY_QUERY_VALUE, &hRegKey) == ERROR_SUCCESS) 
	{
		*szAccount = L'\0';
		dwLength = SMPD_MAX_ACCOUNT_LENGTH;
		if ((nError = RegQueryValueEx(
			hRegKey, 
			L"smpda", NULL, 
			NULL, 
			(BYTE*)szAccount, 
			&dwLength))!=ERROR_SUCCESS)
		{
			/*smpd_err_printf(L"ReadPasswordFromRegistry:RegQueryValueEx(...) failed, error: %d\n", nError);*/
			RegCloseKey(hRegKey);
			smpd_exit_fn(FCNAME);
			return SMPD_FALSE;
		}
		if (wcslen(szAccount) < 1)
		{
			smpd_exit_fn(FCNAME);
			return SMPD_FALSE;
		}

		*szPassword = L'\0';
		dwLength = (SMPD_MAX_PASSWORD_LENGTH*2)*sizeof(wchar_t);
		if ((nError = RegQueryValueEx(
			hRegKey, 
			L"smpdp", NULL, 
			NULL, 
			(BYTE*)szPassword, 
			&dwLength))!=ERROR_SUCCESS)
		{
			/*smpd_err_printf(L"ReadPasswordFromRegistry:RegQueryValueEx(...) failed, error: %d\n", nError);*/
			RegCloseKey(hRegKey);
			smpd_exit_fn(FCNAME);
			return SMPD_FALSE;
		}

		nError = RegCloseKey(hRegKey);
		if (nError != ERROR_SUCCESS)
		{
			smpd_translate_win_error(nError, err_msg, MAX_ERR_LENGTH, L"Unable to close the HKEY_CURRENT_USER\\" SMPD_REGISTRY_CACHE_KEY L" registry key, error %d: ", nError);
			smpd_err_printf(L"%s\n", err_msg);
			smpd_exit_fn(FCNAME);
			return SMPD_FALSE;
		}

		wcscpy_s(account, maxlen1, szAccount);
		smpd_decode_buffer(szPassword, (char *)password, (int)maxlen2/*SMPD_MAX_PASSWORD_LENGTH*/, &num_bytes);
		smpd_exit_fn(FCNAME);
		return SMPD_TRUE;
	}

	smpd_exit_fn(FCNAME);
	return SMPD_FALSE;
}

#undef FCNAME
#define FCNAME L"smpd_delete_cached_password"
int smpd_delete_cached_password()
{
	smpd_enter_fn(FCNAME);
	RegDeleteKey(HKEY_CURRENT_USER, SMPD_REGISTRY_CACHE_KEY);
	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}
