/*   Copyright (C) 2009 Deino Software. */

#include "smpd.h"
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#ifdef HAVE_SIGNAL_H
#include <signal.h>
#endif
#ifdef HAVE_STDLIB_H
#include <stdlib.h> /* MPIU_GetStaticEnv */
#endif
#ifdef HAVE_STRING_H
#include <string.h> /* wcsrchr */
#endif
#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#undef FCNAME
#define FCNAME L"smpd_get_full_path_name"
SMPD_BOOL smpd_get_full_path_name(const wchar_t *exe, int maxlen, wchar_t *exe_path, wchar_t **namepart)
{
	DWORD dwResult;
	DWORD dwLength;
	int len;
	wchar_t buffer[SMPD_MAX_EXE_LENGTH];
	wchar_t info_buffer[sizeof(REMOTE_NAME_INFO) + SMPD_MAX_EXE_LENGTH];
	REMOTE_NAME_INFO *info = (REMOTE_NAME_INFO*)info_buffer;
	wchar_t *filename;
	wchar_t temp_name[SMPD_MAX_EXE_LENGTH];
	wchar_t *temp_exe = NULL;

	smpd_enter_fn(FCNAME);

	/* handle the common case of unix programmers specifying executable like this: ./foo */
	if (wcslen(exe) > 2)
	{
		if (exe[0] == L'.' && exe[1] == L'/')
		{
			temp_exe = _wcsdup(exe);
			temp_exe[1] = L'\\';
			exe = temp_exe;
		}
		smpd_dbg_printf(L"fixing up exe name: '%s' -> '%s'\n", exe, temp_exe);
	}

	/* make a full path out of the name provided */
	len = GetFullPathName(exe, maxlen, exe_path, namepart);
	if (temp_exe != NULL)
	{
		free(temp_exe);
		temp_exe = NULL;
	}
	if (len == 0 || len > maxlen)
	{
		smpd_err_printf(L"buffer provided too short for path: %d provided, %d needed\n", maxlen, len);
		smpd_exit_fn(FCNAME);
		return SMPD_FALSE;
	}
	*(*namepart - 1) = L'\0'; /* separate the path from the executable */
	
	/* Verify file exists.  If it doesn't search the path for exe */
	if ((len = SearchPath(exe_path, *namepart, NULL, SMPD_MAX_EXE_LENGTH, buffer, &filename)) == 0)
	{
		if ((len = SearchPath(exe_path, *namepart, L".exe", SMPD_MAX_EXE_LENGTH, buffer, &filename)) == 0)
		{
			/* search the default path for an exact match */
			if ((len = SearchPath(NULL, *namepart, NULL, SMPD_MAX_EXE_LENGTH, buffer, &filename)) == 0)
			{
				/* search the default path for a match + .exe */
				if ((len = SearchPath(NULL, *namepart, L".exe", SMPD_MAX_EXE_LENGTH, buffer, &filename)) == 0)
				{
					smpd_dbg_printf(L"path not found. leaving as is in case the path exists on the remote machine.\n");
					smpd_exit_fn(FCNAME);
					return SMPD_TRUE;
				}
			}
			if (len > SMPD_MAX_EXE_LENGTH || len > maxlen)
			{
				smpd_err_printf(L"buffer provided too short for path: %d provided, %d needed\n", maxlen, len);
				smpd_exit_fn(FCNAME);
				return SMPD_FALSE;
			}
			*(filename - 1) = L'\0'; /* separate the file name */
			/* copy the path */
			wcscpy_s(exe_path, maxlen, buffer);
			*namepart = &exe_path[wcslen(exe_path)+1];
			/* copy the filename */
			wcscpy_s(*namepart, maxlen - wcslen(exe_path) - 1, filename);
		}
	}
	if (len > maxlen)
	{
		smpd_err_printf(L"buffer provided too short for path: %d provided, %d needed\n", maxlen, len);
		smpd_exit_fn(FCNAME);
		return SMPD_FALSE;
	}

	/* save the filename */
	wcscpy_s(temp_name, SMPD_MAX_EXE_LENGTH, *namepart);

	/* convert the path to its UNC equivalent to avoid need to map a drive */
	dwLength = sizeof(REMOTE_NAME_INFO)+SMPD_MAX_EXE_LENGTH;
	info->lpConnectionName = NULL;
	info->lpRemainingPath = NULL;
	info->lpUniversalName = NULL;
	dwResult = WNetGetUniversalName(exe_path, REMOTE_NAME_INFO_LEVEL, info, &dwLength);
	if (dwResult == NO_ERROR)
	{
		if ((int)(wcslen(info->lpUniversalName) + wcslen(temp_name) + 2) > maxlen)
		{
			smpd_err_printf(L"buffer provided too short for path: %d provided, %d needed\n",
				maxlen, wcslen(info->lpUniversalName) + wcslen(temp_name) + 2);
			smpd_exit_fn(FCNAME);
			return SMPD_FALSE;
		}
		wcscpy_s(exe_path, maxlen, info->lpUniversalName);
		*namepart = &exe_path[wcslen(exe_path)+1];
		wcscpy_s(*namepart, maxlen - wcslen(exe_path) - 1, temp_name);
	}

	smpd_exit_fn(FCNAME);
	return SMPD_TRUE;
}

#undef FCNAME
#define FCNAME L"smpd_search_path"
SMPD_BOOL smpd_search_path(const wchar_t *smpd_path, const wchar_t *exe, int maxlen, wchar_t *str)
{
	wchar_t *filepart;
	wchar_t *path_spec, *smpd_path2 = NULL, exe_path[SMPD_MAX_EXE_LENGTH];
	size_t len_pre, len_spec, len_post;
	HMODULE hModule;

	smpd_enter_fn(FCNAME);

	path_spec = wcsstr(smpd_path, SMPD_PATH_SPEC);
	if (path_spec != NULL)
	{
		if (smpd_process.pszExe[0] == L'\0')
		{
			hModule = GetModuleHandle(NULL);
			if (!GetModuleFileName(hModule, smpd_process.pszExe, SMPD_MAX_EXE_LENGTH)) 
			{
				wcscpy_s(smpd_process.pszExe, SMPD_MAX_EXE_LENGTH, L".\\smpd.exe");
			}
		}
		smpd_path2 = wcsrchr(smpd_process.pszExe, L'\\');
		if (smpd_path2 == NULL)
		{
			exe_path[0] = L'\0';
		}
		else
		{
			memcpy(exe_path, smpd_process.pszExe, (smpd_path2 - smpd_process.pszExe)*sizeof(wchar_t));
			exe_path[smpd_path2 - smpd_process.pszExe] = L'\0';
		}
		len_pre = path_spec - smpd_path;
		len_spec = wcslen(exe_path);
		len_post = wcslen(path_spec + wcslen(SMPD_PATH_SPEC));
		smpd_path2 = (wchar_t *)malloc((len_pre + len_spec + len_post + 1) * sizeof(wchar_t));
		if (len_pre)
		{
			memcpy(smpd_path2, smpd_path, len_pre*sizeof(wchar_t));
		}
		memcpy(smpd_path2 + len_pre, exe_path, len_spec*sizeof(wchar_t));
		wcscpy_s(smpd_path2 + len_pre + len_spec, len_post + 1, path_spec + wcslen(SMPD_PATH_SPEC));
		smpd_dbg_printf(L"changed path(%s) to (%s)\n", smpd_path, smpd_path2);
		smpd_path = smpd_path2;
	}

	/* search for exactly what's specified */
	if (SearchPath(smpd_path, exe, NULL, maxlen, str, &filepart) == 0)
	{
		/* search for file + .exe */
		if (SearchPath(smpd_path, exe, L".exe", maxlen, str, &filepart) == 0)
		{
			/* search the default path */
			if (SearchPath(NULL, exe, NULL, maxlen, str, &filepart) == 0)
			{
				/* search the default path + .exe */
				if (SearchPath(NULL, exe, L".exe", maxlen, str, &filepart) == 0)
				{
					if (smpd_path2 != NULL)
					{
						free(smpd_path2);
					}
					smpd_exit_fn(FCNAME);
					return SMPD_FALSE;
				}
			}
		}
	}
	if (smpd_path2 != NULL)
	{
		free(smpd_path2);
	}
	smpd_exit_fn(FCNAME);
	return SMPD_TRUE;
}

#undef FCNAME
#define FCNAME L"smpd_create_process_struct"
int smpd_create_process_struct(int rank, smpd_process_t **process_ptr)
{
	int result;
	smpd_process_t *p;
	static int cur_id = 0;

	smpd_enter_fn(FCNAME);

	p = (smpd_process_t*)malloc(sizeof(smpd_process_t));
	if (p == NULL)
	{
		*process_ptr = NULL;
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	p->id = cur_id++; /* MT - If smpd is to be thread safe, this will have to be changed */
	p->rank = rank;
	p->nproc = 1;
	p->kvs_name[0] = L'\0';
	p->domain_name[0] = L'\0';
	p->exe[0] = L'\0';
	p->env[0] = L'\0';
	p->dir[0] = L'\0';
	p->path[0] = L'\0';
	p->clique[0] = L'\0';
	p->err_msg[0] = L'\0';
	p->stdin_write_list = NULL;
	result = smpd_create_context(SMPD_CONTEXT_STDIN, smpd_process.set, MPIDU_SOCK_INVALID_SOCK, -1, &p->in);
	if (result != SMPD_SUCCESS)
	{
		free(p);
		*process_ptr = NULL;
		smpd_err_printf(L"unable to create stdin context.\n");
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	result = smpd_create_context(SMPD_CONTEXT_STDOUT, smpd_process.set, MPIDU_SOCK_INVALID_SOCK, -1, &p->out);
	if (result != SMPD_SUCCESS)
	{
		free(p);
		*process_ptr = NULL;
		smpd_err_printf(L"unable to create stdout context.\n");
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	result = smpd_create_context(SMPD_CONTEXT_STDERR, smpd_process.set, MPIDU_SOCK_INVALID_SOCK, -1, &p->err);
	if (result != SMPD_SUCCESS)
	{
		free(p);
		*process_ptr = NULL;
		smpd_err_printf(L"unable to create stderr context.\n");
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	result = smpd_create_context(SMPD_CONTEXT_PMI, smpd_process.set, MPIDU_SOCK_INVALID_SOCK, -1, &p->pmi);
	if (result != SMPD_SUCCESS)
	{
		free(p);
		*process_ptr = NULL;
		smpd_err_printf(L"unable to create pmi context.\n");
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	p->in->rank = rank;
	p->out->rank = rank;
	p->err->rank = rank;
	p->num_valid_contexts = 3;
	p->context_refcount = 0;
	p->exitcode = 0;
	p->in->process = p;
	p->out->process = p;
	p->err->process = p;
	p->pmi->process = p;
	p->next = NULL;
	p->spawned = 0;
	p->local_process = SMPD_TRUE;
	p->map_list = NULL;
	p->appnum = 0;
	p->abort_fn = NULL;
	p->abort_pid = -1;
	p->hDebugReady = NULL;
	p->hDebugThread = NULL;
	p->dbgError[0] = L'\0';
	p->bAborted = SMPD_FALSE;

	*process_ptr = p;

	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}

#undef FCNAME
#define FCNAME L"smpd_process_to_string"
SMPD_BOOL smpd_process_to_string(wchar_t **str_pptr, int *len_ptr, int indent, smpd_process_t *process)
{
	wchar_t indent_str[SMPD_MAX_TO_STRING_INDENT+1];

	smpd_enter_fn(FCNAME);

	if (*len_ptr < 1)
	{
		smpd_exit_fn(FCNAME);
		return SMPD_FALSE;
	}

	if (indent > SMPD_MAX_TO_STRING_INDENT)
	{
		indent = SMPD_MAX_TO_STRING_INDENT;
	}

	memset(indent_str, L' ', indent);
	indent_str[indent] = L'\0';

	smpd_snprintf_update(str_pptr, len_ptr, L"%sid: %d\n", indent_str, process->id);
	if (*len_ptr < 1) { smpd_exit_fn(FCNAME); return SMPD_FALSE; }
	smpd_snprintf_update(str_pptr, len_ptr, L"%srank: %d\n", indent_str, process->rank);
	if (*len_ptr < 1) { smpd_exit_fn(FCNAME); return SMPD_FALSE; }
	smpd_snprintf_update(str_pptr, len_ptr, L"%sexe: %s\n", indent_str, process->exe);
	if (*len_ptr < 1) { smpd_exit_fn(FCNAME); return SMPD_FALSE; }
	smpd_snprintf_update(str_pptr, len_ptr, L"%sdir: %s\n", indent_str, process->dir);
	if (*len_ptr < 1) { smpd_exit_fn(FCNAME); return SMPD_FALSE; }
	smpd_snprintf_update(str_pptr, len_ptr, L"%senv: %s\n", indent_str, process->env);
	if (*len_ptr < 1) { smpd_exit_fn(FCNAME); return SMPD_FALSE; }
	smpd_snprintf_update(str_pptr, len_ptr, L"%spath: %s\n", indent_str, process->path);
	if (*len_ptr < 1) { smpd_exit_fn(FCNAME); return SMPD_FALSE; }
	smpd_snprintf_update(str_pptr, len_ptr, L"%spid: %d\n", indent_str, process->pid);
	if (*len_ptr < 1) { smpd_exit_fn(FCNAME); return SMPD_FALSE; }
	smpd_snprintf_update(str_pptr, len_ptr, L"%sexitcode: %d\n", indent_str, process->exitcode);
	if (*len_ptr < 1) { smpd_exit_fn(FCNAME); return SMPD_FALSE; }
	smpd_snprintf_update(str_pptr, len_ptr, L"%scontext_refcount: %s\n", indent_str, process->context_refcount);
	if (*len_ptr < 1) { smpd_exit_fn(FCNAME); return SMPD_FALSE; }
	smpd_snprintf_update(str_pptr, len_ptr, L"%serr_msg: %s\n", indent_str, process->err_msg);
	if (*len_ptr < 1) { smpd_exit_fn(FCNAME); return SMPD_FALSE; }
	smpd_snprintf_update(str_pptr, len_ptr, L"%snum_valid_contexts: %d\n", indent_str, process->num_valid_contexts);
	if (*len_ptr < 1) { smpd_exit_fn(FCNAME); return SMPD_FALSE; }
	smpd_snprintf_update(str_pptr, len_ptr, L"%s in: %p\n", indent_str, process->in);
	if (*len_ptr < 1) { smpd_exit_fn(FCNAME); return SMPD_FALSE; }
	smpd_snprintf_update(str_pptr, len_ptr, L"%s out: %p\n", indent_str, process->out);
	if (*len_ptr < 1) { smpd_exit_fn(FCNAME); return SMPD_FALSE; }
	smpd_snprintf_update(str_pptr, len_ptr, L"%s err: %p\n", indent_str, process->err);
	if (*len_ptr < 1) { smpd_exit_fn(FCNAME); return SMPD_FALSE; }
	smpd_snprintf_update(str_pptr, len_ptr, L"%s pmi: %p\n", indent_str, process->pmi);
	if (*len_ptr < 1) { smpd_exit_fn(FCNAME); return SMPD_FALSE; }
	smpd_snprintf_update(str_pptr, len_ptr, L"%skvs_name: %s\n", indent_str, process->kvs_name);
	if (*len_ptr < 1) { smpd_exit_fn(FCNAME); return SMPD_FALSE; }
	smpd_snprintf_update(str_pptr, len_ptr, L"%snproc: %d\n", indent_str, process->nproc);
	if (*len_ptr < 1) { smpd_exit_fn(FCNAME); return SMPD_FALSE; }
	smpd_snprintf_update(str_pptr, len_ptr, L"%swait: %p:%p\n", indent_str, process->wait.hProcess, process->wait.hThread);
	if (*len_ptr < 1) { smpd_exit_fn(FCNAME); return SMPD_FALSE; }
	smpd_snprintf_update(str_pptr, len_ptr, L"%snext: %p\n", indent_str, process->next);
	if (*len_ptr < 1) { smpd_exit_fn(FCNAME); return SMPD_FALSE; } /* this misses the case of an exact fit */

	smpd_exit_fn(FCNAME);
	return SMPD_TRUE;
}

#undef FCNAME
#define FCNAME L"smpd_free_process_struct"
int smpd_free_process_struct(smpd_process_t *process)
{
	smpd_enter_fn(FCNAME);
	if (process == NULL)
	{
		smpd_dbg_printf(L"smpd_free_process_struct passed NULL process pointer.\n");
		smpd_exit_fn(FCNAME);
		return SMPD_SUCCESS;
	}
	if (process->in)
		smpd_free_context(process->in);
	process->in = NULL;
	if (process->out)
		smpd_free_context(process->out);
	process->out = NULL;
	if (process->err)
		smpd_free_context(process->err);
	process->err = NULL;
	if (process->pmi)
		smpd_free_context(process->pmi);
	process->pmi = NULL;
	process->dir[0] = L'\0';
	process->env[0] = L'\0';
	process->exe[0] = L'\0';
	process->path[0] = L'\0';
	process->pid = -1;
	process->rank = -1;
	process->next = NULL;
	free(process);
	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}

#undef FCNAME
#define FCNAME L"smpd_interpret_session_header"
int smpd_interpret_session_header(wchar_t *str)
{
	wchar_t temp_str[100];

	smpd_enter_fn(FCNAME);

	smpd_dbg_printf(L"interpreting session header: \"%s\"\n", str);

	/* get my id */
	if (MPIU_Str_get_string_arg(str, L"id", temp_str, 100) == MPIU_STR_SUCCESS)
	{
		smpd_dbg_printf(L" id = %s\n", temp_str);
		smpd_process.id = _wtoi(temp_str);
		if (smpd_process.id < 0)
		{
			smpd_err_printf(L"invalid id passed in session header: %d\n", smpd_process.id);
			smpd_process.id = 0;
		}
	}

	/* get my parent's id */
	if (MPIU_Str_get_string_arg(str, L"parent", temp_str, 100) == MPIU_STR_SUCCESS)
	{
		smpd_dbg_printf(L" parent = %s\n", temp_str);
		smpd_process.parent_id = _wtoi(temp_str);
		if (smpd_process.parent_id < 0)
		{
			smpd_err_printf(L"invalid parent id passed in session header: %d\n", smpd_process.parent_id);
			smpd_process.parent_id = -1;
		}
	}

	/* get my level */
	if (MPIU_Str_get_string_arg(str, L"level", temp_str, 100) == MPIU_STR_SUCCESS)
	{
		smpd_dbg_printf(L" level = %s\n", temp_str);
		smpd_process.level = _wtoi(temp_str);
		if (smpd_process.level < 0)
		{
			smpd_err_printf(L"invalid session level passed in session header: %d\n", smpd_process.level);
			smpd_process.level = 0;
		}
	}

	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}
