/*   Copyright (C) 2009 Deino Software. */

#include "smpd.h"
#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif

wchar_t *smpd_encode_handle(wchar_t *str, size_t maxlen, HANDLE h)
{
	swprintf_s(str, maxlen, L"%p", h);
	return str;
}

HANDLE smpd_decode_handle(wchar_t *str)
{
	HANDLE p;
	swscanf_s(str, L"%p", &p);
	return p;
}

#undef FCNAME
#define FCNAME L"smpd_start_win_mgr"
int smpd_start_win_mgr(smpd_context_t *context, SMPD_BOOL use_context_user_handle)
{
	int result;
	wchar_t dbg_str[20];
	wchar_t read_handle_str[20], write_handle_str[20];
	wchar_t domainaccount[SMPD_MAX_ACCOUNT_LENGTH], account[SMPD_MAX_ACCOUNT_LENGTH], domain[SMPD_MAX_ACCOUNT_LENGTH];
	wchar_t *pszDomain;
	wchar_t password[SMPD_MAX_PASSWORD_LENGTH];
	HANDLE user_handle = INVALID_HANDLE_VALUE;
	HANDLE job = INVALID_HANDLE_VALUE;
	int num_tries;
	wchar_t cmd[8192];
	PROCESS_INFORMATION pInfo;
	STARTUPINFO sInfo;
	SECURITY_ATTRIBUTES security;
	HANDLE hRead, hWrite, hReadRemote, hWriteRemote;
	DWORD num_read, num_written;

	smpd_enter_fn(FCNAME);

	/* initialize to null so in the sspi case garbage isn't sent to the manager */
	domainaccount[0] = L'\0';
	account[0] = L'\0';
	domain[0] = L'\0';
	password[0] = L'\0';

	/* start the manager */
	security.bInheritHandle = TRUE;
	security.lpSecurityDescriptor = NULL;
	security.nLength = sizeof(security);
	/* create a pipe to send the listening port information through */
	if (!CreatePipe(&hRead, &hWriteRemote, &security, 0))
	{
		smpd_err_printf(L"CreatePipe failed, error %d\n", GetLastError());
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	/* prevent the local read end of the pipe from being inherited */
	if (!DuplicateHandle(GetCurrentProcess(), hRead, GetCurrentProcess(), &hRead, 0, FALSE, DUPLICATE_CLOSE_SOURCE | DUPLICATE_SAME_ACCESS))
	{
		smpd_err_printf(L"Unable to duplicate the read end of the pipe, error %d\n", GetLastError());
		CloseHandle(hRead);
		CloseHandle(hWriteRemote);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	/* create a pipe to send the account information through */
	if (!CreatePipe(&hReadRemote, &hWrite, &security, 0))
	{
		smpd_err_printf(L"CreatePipe failed, error %d\n", GetLastError());
		CloseHandle(hRead);
		CloseHandle(hWriteRemote);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	/* prevent the local write end of the pipe from being inherited */
	if (!DuplicateHandle(GetCurrentProcess(), hWrite, GetCurrentProcess(), &hWrite, 0, FALSE, DUPLICATE_CLOSE_SOURCE | DUPLICATE_SAME_ACCESS))
	{
		smpd_err_printf(L"Unable to duplicate the read end of the pipe, error %d\n", GetLastError());
		CloseHandle(hRead);
		CloseHandle(hWrite);
		CloseHandle(hReadRemote);
		CloseHandle(hWriteRemote);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	/* encode the command line */
	if (smpd_process.dbg_state == SMPD_DBG_STATE_ALL)
		wcscpy_s(dbg_str, 20, L"-d");
	else if (smpd_process.dbg_state)
		_snwprintf_s(dbg_str, 20, _TRUNCATE, L"-d %d", smpd_process.dbg_state);
	else
		dbg_str[0] = L'\0';
	if (smpd_process.port != SMPD_LISTENER_PORT)
	{
		_snwprintf_s(cmd, 8192, _TRUNCATE, L"\"%s\" -p %d %s -mgr -read %s -write %s", smpd_process.pszExe, smpd_process.port, dbg_str,
			smpd_encode_handle(read_handle_str, 20, hReadRemote), 
			smpd_encode_handle(write_handle_str, 20, hWriteRemote));
	}
	else
	{
		_snwprintf_s(cmd, 8192, _TRUNCATE, L"\"%s\" %s -mgr -read %s -write %s", smpd_process.pszExe, dbg_str,
			smpd_encode_handle(read_handle_str, 20, hReadRemote), 
			smpd_encode_handle(write_handle_str, 20, hWriteRemote));
	}
	if (context->connect_to)
	{
		smpd_dbg_printf(L"starting command:%d: %s\n", context->connect_to->id, cmd);
	}
	else
	{
		smpd_dbg_printf(L"starting command: %s\n", cmd);
	}
	GetStartupInfo(&sInfo);
	if (smpd_process.bService)
	{
		if (use_context_user_handle)
		{
			if (context->sspi_context == NULL)
			{
				smpd_err_printf(L"use_context_user_handle set to TRUE but sspi_context == NULL\n");
				CloseHandle(hRead);
				CloseHandle(hWrite);
				CloseHandle(hReadRemote);
				CloseHandle(hWriteRemote);
				smpd_exit_fn(FCNAME);
				return SMPD_ERR_INVALID_USER;
			}
			user_handle = context->sspi_context->user_handle;
			job = context->sspi_context->job;
		}
		else
		{
			wcscpy_s(domainaccount, SMPD_MAX_ACCOUNT_LENGTH, context->account);
			wcscpy_s(password, SMPD_MAX_PASSWORD_LENGTH, context->password);
			smpd_parse_account_domain(domainaccount, account, SMPD_MAX_ACCOUNT_LENGTH, domain, SMPD_MAX_ACCOUNT_LENGTH);
			if (wcslen(domain) < 1)
				pszDomain = NULL;
			else
				pszDomain = domain;

			result = smpd_get_user_handle(account, pszDomain, password, &user_handle);
			if (user_handle == INVALID_HANDLE_VALUE)
			{
				smpd_err_printf(L"smpd_get_user_handle failed, error %d.\n", result);
				CloseHandle(hRead);
				CloseHandle(hWrite);
				CloseHandle(hReadRemote);
				CloseHandle(hWriteRemote);
				smpd_exit_fn(FCNAME);
				return SMPD_ERR_INVALID_USER;
			}
		}

		result = SMPD_SUCCESS;
		if (!ImpersonateLoggedOnUser(user_handle))
		{
			result = GetLastError();
			smpd_err_printf(L"ImpersonateLoggedOnUser failed, error %d\n", result);
			CloseHandle(hRead);
			CloseHandle(hWrite);
			CloseHandle(hReadRemote);
			CloseHandle(hWriteRemote);
			smpd_exit_fn(FCNAME);
			return SMPD_ERR_INVALID_USER;
		}
	}

	num_tries = 4;
	do
	{
		if (smpd_process.bService)
		{
			smpd_dbg_printf(L"CreateProcessAsUser\n");
			result = CreateProcessAsUser(
				user_handle,
				NULL, cmd, NULL, NULL, TRUE,
				CREATE_SUSPENDED,
				NULL, NULL, &sInfo, &pInfo);
		}
		else
		{
			smpd_dbg_printf(L"CreateProcess\n");
			result = CreateProcess(
				NULL, cmd, NULL, NULL, TRUE,
				CREATE_SUSPENDED,
				NULL, NULL, &sInfo, &pInfo);
		}

		if (result)
		{
			result = SMPD_SUCCESS;
			num_tries = 0;
		}
		else
		{
			result = GetLastError();
			if (result == ERROR_REQ_NOT_ACCEP)
			{
				Sleep(1000);
				num_tries--;
				if (num_tries == 0)
				{
					smpd_err_printf(L"%s failed, error %d - ERROR_REQ_NOT_ACCEP\n", smpd_process.bService ? L"CreateProcessAsUser" : L"CreateProcess", result);
				}
			}
			else
			{
				smpd_err_printf(L"%s failed, error %d\n", smpd_process.bService ? L"CreateProcessAsUser" : L"CreateProcess", result);
				num_tries = 0;
			}
		}
	} while (num_tries);

	if (job != INVALID_HANDLE_VALUE)
	{
		smpd_dbg_printf(L"assinging smpd manager to job\n");
		if (!AssignProcessToJobObject(job, pInfo.hProcess))
		{
			smpd_err_printf(L"AssignProcessToJobObject failed: %d\n", GetLastError());
			TerminateProcess(pInfo.hProcess, (UINT)-1);
			result = SMPD_FAIL;
		}
		else
		{
			ResumeThread(pInfo.hThread);
		}
	}
	else
	{
		ResumeThread(pInfo.hThread);
	}

	if (smpd_process.bService)
	{
		RevertToSelf();
		if (use_context_user_handle && context->sspi_context != NULL)
		{
			if (context->sspi_context->close_handle)
			{
				CloseHandle(context->sspi_context->user_handle);
				context->sspi_context->user_handle = INVALID_HANDLE_VALUE;
			}
		}
		else
		{
			CloseHandle(user_handle);
		}
	}
	if (result != SMPD_SUCCESS)
	{
		CloseHandle(hRead);
		CloseHandle(hWrite);
		CloseHandle(hReadRemote);
		CloseHandle(hWriteRemote);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}

	CloseHandle(pInfo.hThread);
	CloseHandle(pInfo.hProcess);
	CloseHandle(hReadRemote);
	CloseHandle(hWriteRemote);

	smpd_dbg_printf(L"smpd reading the port string from the manager\n");
	/* read the listener port from the pipe to the manager */
	if (!ReadFile(hRead, context->port_str, 20*sizeof(wchar_t), &num_read, NULL))
	{
		smpd_err_printf(L"ReadFile() failed, error %d\n", GetLastError());
		CloseHandle(hRead);
		CloseHandle(hWrite);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	CloseHandle(hRead);
	if (num_read != 20*sizeof(wchar_t))
	{
		smpd_err_printf(L"parital port string read, %d bytes of %d\n", num_read, 20*sizeof(wchar_t));
		CloseHandle(hWrite);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	/* send the account and password to the manager */
	smpd_dbg_printf(L"smpd sending the account to the manager\n");
	if (!WriteFile(hWrite, domainaccount, SMPD_MAX_ACCOUNT_LENGTH*sizeof(wchar_t), &num_written, NULL))
	{
		smpd_err_printf(L"WriteFile('%s') failed to write the account, error %d\n", domainaccount, GetLastError());
		CloseHandle(hWrite);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	if (num_written != SMPD_MAX_ACCOUNT_LENGTH*sizeof(wchar_t))
	{
		smpd_err_printf(L"parital account string written, %d bytes of %d\n", num_written, SMPD_MAX_ACCOUNT_LENGTH*sizeof(wchar_t));
		CloseHandle(hWrite);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	smpd_dbg_printf(L"smpd sending the password to the manager\n");
	if (!WriteFile(hWrite, password, SMPD_MAX_PASSWORD_LENGTH*sizeof(wchar_t), &num_written, NULL))
	{
		smpd_err_printf(L"WriteFile() failed to write the password, error %d\n", GetLastError());
		CloseHandle(hWrite);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	if (num_written != SMPD_MAX_PASSWORD_LENGTH*sizeof(wchar_t))
	{
		smpd_err_printf(L"parital password string written, %d bytes of %d\n", num_written, SMPD_MAX_PASSWORD_LENGTH*sizeof(wchar_t));
		CloseHandle(hWrite);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	smpd_dbg_printf(L"smpd sending the smpd passphrase to the manager\n");
	if (!WriteFile(hWrite, smpd_process.passphrase, SMPD_PASSPHRASE_MAX_LENGTH*sizeof(wchar_t), &num_written, NULL))
	{
		smpd_err_printf(L"WriteFile() failed to write the passphrase, error %d\n", GetLastError());
		CloseHandle(hWrite);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	if (num_written != SMPD_PASSPHRASE_MAX_LENGTH*sizeof(wchar_t))
	{
		smpd_err_printf(L"parital passphrase string written, %d bytes of %d\n", num_written, SMPD_PASSPHRASE_MAX_LENGTH*sizeof(wchar_t));
		CloseHandle(hWrite);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	smpd_dbg_printf(L"closing the pipe to the manager\n");
	CloseHandle(hWrite);

	return SMPD_SUCCESS;
}
