/*   Copyright (C) 2009 Deino Software. */

#include "smpd.h"
#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#include <stdlib.h>

#undef FCNAME
#define FCNAME L"smpd_is_affirmative"
SMPD_BOOL smpd_is_affirmative(const wchar_t *str)
{
	const wchar_t *iter = str;

	smpd_enter_fn(FCNAME);
	if (*iter != L'y' && *iter != L'Y')
	{
		smpd_exit_fn(FCNAME);
		return SMPD_FALSE;
	}
	iter++;
	if (*iter == L'\0')
	{
		smpd_exit_fn(FCNAME);
		return SMPD_TRUE;
	}
	if (*iter == L'\n')
	{
		iter++;
		if (*iter == L'\0')
		{
			smpd_exit_fn(FCNAME);
			return SMPD_TRUE;
		}
		smpd_exit_fn(FCNAME);
		return SMPD_FALSE;
	}
	if (*iter != L'e' && *iter != L'E')
	{
		smpd_exit_fn(FCNAME);
		return SMPD_FALSE;
	}
	iter++;
	if (*iter != L's' && *iter != L's')
	{
		smpd_exit_fn(FCNAME);
		return SMPD_FALSE;
	}
	iter++;
	if (*iter == L'\n')
	{
		iter++;
		if (*iter == L'\0')
		{
			smpd_exit_fn(FCNAME);
			return SMPD_TRUE;
		}
		smpd_exit_fn(FCNAME);
		return SMPD_FALSE;
	}
	if (*iter == L'\0')
	{
		smpd_exit_fn(FCNAME);
		return SMPD_TRUE;
	}
	smpd_exit_fn(FCNAME);
	return SMPD_FALSE;

	/*
	if (wcscmp(str, L"yes\n") == 0 ||
		wcscmp(str, L"Yes\n") == 0 ||
		wcscmp(str, L"YES\n") == 0 ||
		wcscmp(str, L"Y\n") == 0 ||
		wcscmp(str, L"y\n") == 0 ||
		wcscmp(str, L"yes") == 0 ||
		wcscmp(str, L"Yes") == 0 ||
		wcscmp(str, L"YES") == 0 ||
		wcscmp(str, L"Y") == 0 ||
		wcscmp(str, L"y") == 0)
		return SMPD_TRUE;
	return SMPD_FALSE;
	*/
}

#undef FCNAME
#define FCNAME L"smpd_option_on"
SMPD_BOOL smpd_option_on(const wchar_t *option)
{
	wchar_t val[SMPD_MAX_VALUE_LENGTH];

	smpd_enter_fn(FCNAME);

	if (smpd_get_smpd_data(option, val, SMPD_MAX_VALUE_LENGTH) == SMPD_SUCCESS)
	{
		if (smpd_is_affirmative(val) || (wcscmp(val, L"1") == 0))
		{
			smpd_exit_fn(FCNAME);
			return SMPD_TRUE;
		}
	}

	smpd_exit_fn(FCNAME);
	return SMPD_FALSE;
}

#undef FCNAME
#define FCNAME L"smpd_delete_user_data"
int smpd_delete_user_data(const wchar_t *key)
{
	HKEY tkey;
	DWORD result;
	wchar_t err_msg[512];

	smpd_enter_fn(FCNAME);

	if (key == NULL)
	{
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}

	result = RegCreateKeyEx(HKEY_CURRENT_USER, SMPD_REGISTRY_KEY,
		0, NULL, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &tkey, NULL);
	if (result != ERROR_SUCCESS)
	{
		smpd_translate_win_error(result, err_msg, 512, L"Unable to open the HKEY_CURRENT_USER\\" SMPD_REGISTRY_KEY L" registry key, error %d: ", result);
		smpd_err_printf(L"%s\n", err_msg);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}

	result = RegDeleteValue(tkey, key);
	if (result != ERROR_SUCCESS)
	{
		smpd_translate_win_error(result, err_msg, 512, L"Unable to delete the user smpd registry value '%s', error %d: ", key, result);
		smpd_err_printf(L"%s\n", err_msg);
		RegCloseKey(tkey);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}

	result = RegCloseKey(tkey);
	if (result != ERROR_SUCCESS)
	{
		smpd_translate_win_error(result, err_msg, 512, L"Unable to close the HKEY_CURRENT_USER\\" SMPD_REGISTRY_KEY L" registry key, error %d: ", result);
		smpd_err_printf(L"%s\n", err_msg);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}

	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}

#undef FCNAME
#define FCNAME L"smpd_delete_smpd_data"
int smpd_delete_smpd_data(const wchar_t *key)
{
	HKEY tkey;
	DWORD result;
	wchar_t err_msg[512];

	smpd_enter_fn(FCNAME);

	if (key == NULL)
	{
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}

	result = RegCreateKeyEx(HKEY_LOCAL_MACHINE, SMPD_REGISTRY_KEY,
		0, NULL, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &tkey, NULL);
	if (result != ERROR_SUCCESS)
	{
		smpd_translate_win_error(result, err_msg, 512, L"Unable to open the HKEY_CURRENT_USER\\" SMPD_REGISTRY_KEY L" registry key, error %d: ", result);
		smpd_err_printf(L"%s\n", err_msg);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}

	result = RegDeleteValue(tkey, key);
	if (result != ERROR_SUCCESS)
	{
		if (result != ERROR_FILE_NOT_FOUND && result != ERROR_PATH_NOT_FOUND)
		{
			smpd_translate_win_error(result, err_msg, 512, L"Unable to delete the smpd registry value '%s', error %d: ", key, result);
			smpd_err_printf(L"%s\n", err_msg);
			RegCloseKey(tkey);
			smpd_exit_fn(FCNAME);
			return SMPD_FAIL;
		}
	}

	result = RegCloseKey(tkey);
	if (result != ERROR_SUCCESS)
	{
		smpd_translate_win_error(result, err_msg, 512, L"Unable to close the HKEY_CURRENT_USER\\" SMPD_REGISTRY_KEY L" registry key, error %d: ", result);
		smpd_err_printf(L"%s\n", err_msg);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}

	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}

#undef FCNAME
#define FCNAME L"smpd_set_user_data"
int smpd_set_user_data(const wchar_t *key, const wchar_t *value)
{
	HKEY tkey;
	DWORD len, result;
	wchar_t err_msg[512];

	smpd_enter_fn(FCNAME);

	if (key == NULL || value == NULL)
	{
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}

	result = RegCreateKeyEx(HKEY_CURRENT_USER, SMPD_REGISTRY_KEY,
		0, NULL, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &tkey, NULL);
	if (result != ERROR_SUCCESS)
	{
		smpd_translate_win_error(result, err_msg, 512, L"Unable to open the HKEY_CURRENT_USER\\" SMPD_REGISTRY_KEY L" registry key, error %d: ", result);
		smpd_err_printf(L"%s\n", err_msg);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}

	len = (DWORD)(wcslen(value)+1)*sizeof(wchar_t);
	result = RegSetValueEx(tkey, key, 0, REG_SZ, (const BYTE *)value, len);
	if (result != ERROR_SUCCESS)
	{
		smpd_translate_win_error(result, err_msg, 512, L"Unable to write the user smpd registry value '%s:%s', error %d\n", key, value, result);
		smpd_err_printf(L"%s\n", err_msg);
		RegCloseKey(tkey);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}

	result = RegCloseKey(tkey);
	if (result != ERROR_SUCCESS)
	{
		smpd_translate_win_error(result, err_msg, 512, L"Unable to close the HKEY_CURRENT_USER\\" SMPD_REGISTRY_KEY L" registry key, error %d: ", result);
		smpd_err_printf(L"%s\n", err_msg);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}

	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}

#undef FCNAME
#define FCNAME L"smpd_set_smpd_data"
int smpd_set_smpd_data(const wchar_t *key, const wchar_t *value)
{
	HKEY tkey;
	DWORD len, result;
	wchar_t err_msg[512];

	smpd_enter_fn(FCNAME);

	if (key == NULL || value == NULL)
	{
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}

	result = RegCreateKeyEx(HKEY_LOCAL_MACHINE, SMPD_REGISTRY_KEY,
		0, NULL, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &tkey, NULL);
	if (result != ERROR_SUCCESS)
	{
		smpd_translate_win_error(result, err_msg, 512, L"Unable to open the HKEY_LOCAL_MACHINE\\" SMPD_REGISTRY_KEY L" registry key, error %d\n", result);
		smpd_err_printf(L"%s\n", err_msg);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}

	len = (DWORD)(wcslen(value)+1)*sizeof(wchar_t);
	result = RegSetValueEx(tkey, key, 0, REG_SZ, (const BYTE *)value, len);
	if (result != ERROR_SUCCESS)
	{
		smpd_translate_win_error(result, err_msg, 512, L"Unable to write the smpd registry value '%s:%s', error %d\n", key, value, result);
		smpd_err_printf(L"%s\n", err_msg);
		RegCloseKey(tkey);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}

	result = RegCloseKey(tkey);
	if (result != ERROR_SUCCESS)
	{
		smpd_translate_win_error(result, err_msg, 512, L"Unable to close the HKEY_LOCAL_MACHINE\\" SMPD_REGISTRY_KEY L" registry key, error %d: ", result);
		smpd_err_printf(L"%s\n", err_msg);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}

	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}

#undef FCNAME
#define FCNAME L"smpd_get_user_data_default"
int smpd_get_user_data_default(const wchar_t *key, wchar_t *value, int value_len)
{
	smpd_enter_fn(FCNAME);
	/* FIXME: function not implemented */
	SMPD_UNREFERENCED_ARG(key);
	SMPD_UNREFERENCED_ARG(value);
	SMPD_UNREFERENCED_ARG(value_len);
	smpd_exit_fn(FCNAME);
	return SMPD_FAIL;
}

#undef FCNAME
#define FCNAME L"smpd_get_user_data"
int smpd_get_user_data(const wchar_t *key, wchar_t *value, int value_len)
{
	HKEY tkey;
	DWORD len, result;
	wchar_t err_msg[512];

	smpd_enter_fn(FCNAME);

	result = RegOpenKeyEx(HKEY_CURRENT_USER, SMPD_REGISTRY_KEY, 0, KEY_READ, &tkey);
	if (result != ERROR_SUCCESS)
	{
		smpd_translate_win_error(result, err_msg, 512, L"Unable to open the HKEY_CURRENT_USER\\" SMPD_REGISTRY_KEY L" registry key, error %d\n", result);
		smpd_dbg_printf(L"%s\n", err_msg);
		result = smpd_get_user_data_default(key, value, value_len);
		smpd_exit_fn(FCNAME);
		return result;
	}

	len = value_len;
	result = RegQueryValueEx(tkey, key, 0, NULL, (LPBYTE)value, &len);
	if (result != ERROR_SUCCESS)
	{
		smpd_translate_win_error(result, err_msg, 512, L"Unable to read the smpd registry key '%s', error %d\n", key, result);
		smpd_dbg_printf(L"%s\n", err_msg);
		RegCloseKey(tkey);
		result = smpd_get_user_data_default(key, value, value_len);
		smpd_exit_fn(FCNAME);
		return result;
	}

	result = RegCloseKey(tkey);
	if (result != ERROR_SUCCESS)
	{
		smpd_translate_win_error(result, err_msg, 512, L"Unable to close the HKEY_CURRENT_USER\\" SMPD_REGISTRY_KEY L" registry key, error %d: ", result);
		smpd_err_printf(L"%s\n", err_msg);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}

	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}

#undef FCNAME
#define FCNAME L"smpd_get_smpd_data_default"
int smpd_get_smpd_data_default(const wchar_t *key, wchar_t *value, int value_len)
{
	smpd_enter_fn(FCNAME);
	if (wcscmp(key, L"phrase") == 0)
	{
		MPIU_Wcsncpy(value, SMPD_DEFAULT_PASSPHRASE, value_len);
		value[value_len-1] = L'\0';
		smpd_exit_fn(FCNAME);
		return SMPD_SUCCESS;
	}
	if (wcscmp(key, L"log") == 0)
	{
		MPIU_Wcsncpy(value, L"no", value_len);
		value[value_len-1] = L'\0';
	}
	else if (wcscmp(key, L"prepend_rank") == 0)
	{
		MPIU_Wcsncpy(value, L"yes", value_len);
		value[value_len-1] = L'\0';
	}
	else if (wcscmp(key, L"trace") == 0)
	{
		MPIU_Wcsncpy(value, L"yes", value_len);
		value[value_len-1] = L'\0';
	}
	else if (wcscmp(key, L"noprompt") == 0)
	{
		MPIU_Wcsncpy(value, L"no", value_len);
		value[value_len-1] = L'\0';
	}
	/*
	else if (wcscmp(key, L"hosts") == 0)
	{
		if (smpd_get_hostname(value, value_len) != 0)
		{
			smpd_exit_fn(FCNAME);
			return SMPD_FAIL;
		}
	}
	*/
	else
	{
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}

#undef FCNAME
#define FCNAME L"smpd_get_smpd_data_from_environment"
static SMPD_BOOL smpd_get_smpd_data_from_environment(const wchar_t *key, wchar_t *value, int value_len)
{
	wchar_t *env_option, *iter;
	const wchar_t *env;
	size_t length;

	smpd_enter_fn(FCNAME);

	/* Check to see if the option has been set in the environment */
	length = wcslen(key) + wcslen(SMPD_ENV_OPTION_PREFIX) + 1;
	env_option = (wchar_t*)malloc(length * sizeof(wchar_t));
	if (env_option != NULL)
	{
		wcscpy_s(env_option, length, SMPD_ENV_OPTION_PREFIX);
		iter = &env_option[wcslen(SMPD_ENV_OPTION_PREFIX)];
		while (*key != L'\0')
		{
			*iter = (wchar_t)(toupper(*key));
			key++;
			iter++;
		}
		*iter = L'\0';
		env = MPIU_GetStaticEnv(env_option);
		free(env_option);
		if (env != NULL)
		{
			if ((int)wcslen(env) < value_len)
			{
				wcscpy_s(value, value_len, env);
				smpd_exit_fn(FCNAME);
				return SMPD_TRUE;
			}
		}
	}
	smpd_exit_fn(FCNAME);
	return SMPD_FALSE;
}

#undef FCNAME
#define FCNAME L"smpd_get_smpd_data"
int smpd_get_smpd_data(const wchar_t *key, wchar_t *value, int value_len)
{
	HKEY tkey;
	DWORD len, result;
	wchar_t err_msg[512];

	smpd_enter_fn(FCNAME);

	if (smpd_get_smpd_data_from_environment(key, value, value_len) == SMPD_TRUE)
	{
		smpd_exit_fn(FCNAME);
		return SMPD_SUCCESS;
	}

	result = RegOpenKeyEx(HKEY_LOCAL_MACHINE, SMPD_REGISTRY_KEY, 0, KEY_READ, &tkey);
	if (result != ERROR_SUCCESS)
	{
		if (smpd_get_smpd_data_default(key, value, value_len) != SMPD_SUCCESS)
		{
			smpd_dbg_printf(L"Unable to get the data for the key '%s'\n", key);
			smpd_exit_fn(FCNAME);
			return SMPD_FAIL;
		}
		smpd_exit_fn(FCNAME);
		return SMPD_SUCCESS;
	}

	len = value_len;
	result = RegQueryValueEx(tkey, key, 0, NULL, (LPBYTE)value, &len);
	if (result != ERROR_SUCCESS)
	{
		RegCloseKey(tkey);
		if (smpd_get_smpd_data_default(key, value, value_len) != SMPD_SUCCESS)
		{
			smpd_dbg_printf(L"Unable to get the data for the key '%s'\n", key);
			smpd_exit_fn(FCNAME);
			return SMPD_FAIL;
		}
		smpd_exit_fn(FCNAME);
		return SMPD_SUCCESS;
	}

	result = RegCloseKey(tkey);
	if (result != ERROR_SUCCESS)
	{
		smpd_translate_win_error(result, err_msg, 512, L"Unable to close the HKEY_LOCAL_MACHINE\\" SMPD_REGISTRY_KEY L" registry key, error %d: ", result);
		smpd_err_printf(L"%s\n", err_msg);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}

	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}

#undef FCNAME
#define FCNAME L"smpd_get_all_smpd_data"
int smpd_get_all_smpd_data(smpd_data_t **data)
{
	HKEY tkey;
	DWORD result;
	LONG enum_result;
	wchar_t name[SMPD_MAX_NAME_LENGTH], value[SMPD_MAX_VALUE_LENGTH];
	DWORD name_length, value_length, index;
	smpd_data_t *list, *item;
	wchar_t err_msg[512];

	smpd_enter_fn(FCNAME);

	if (data == NULL)
	{
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}

	result = RegOpenKeyEx(HKEY_LOCAL_MACHINE, SMPD_REGISTRY_KEY, 0, KEY_READ, &tkey);
	if (result != ERROR_SUCCESS)
	{
		/* No key therefore no settings */
		/* No access to the key, therefore no soup for you */
		*data = NULL;
		smpd_exit_fn(FCNAME);
		return SMPD_SUCCESS;
	}

	list = NULL;
	index = 0;
	name_length = SMPD_MAX_NAME_LENGTH;
	value_length = SMPD_MAX_VALUE_LENGTH;
	enum_result = RegEnumValue(tkey, index, name, &name_length, NULL, NULL, (LPBYTE)value, &value_length);
	while (enum_result == ERROR_SUCCESS)
	{
		item = (smpd_data_t*)malloc(sizeof(smpd_data_t));
		if (item == NULL)
		{
			*data = NULL;
			result = RegCloseKey(tkey);
			smpd_exit_fn(FCNAME);
			return SMPD_FAIL;
		}
		memcpy(item->name, name, SMPD_MAX_NAME_LENGTH*sizeof(wchar_t));
		memcpy(item->value, value, SMPD_MAX_VALUE_LENGTH*sizeof(wchar_t));
		item->next = list;
		list = item;
		index++;
		name_length = SMPD_MAX_NAME_LENGTH;
		value_length = SMPD_MAX_VALUE_LENGTH;
		enum_result = RegEnumValue(tkey, index, name, &name_length, NULL, NULL, (LPBYTE)value, &value_length);
	}

	result = RegCloseKey(tkey);
	if (result != ERROR_SUCCESS)
	{
		smpd_translate_win_error(result, err_msg, 512, L"Unable to close the HKEY_LOCAL_MACHINE\\" SMPD_REGISTRY_KEY L" registry key, error %d: ", result);
		smpd_err_printf(L"%s\n", err_msg);
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}

	*data = list;
	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}

#undef FCNAME
#define FCNAME L"smpd_lock_smpd_data"
int smpd_lock_smpd_data(void)
{
	smpd_enter_fn(FCNAME);
	if (smpd_process.hSMPDDataMutex == NULL)
	{
		smpd_process.hSMPDDataMutex = CreateMutex(NULL, FALSE, SMPD_DATA_MUTEX_NAME);
		if (smpd_process.hSMPDDataMutex == NULL)
		{
			smpd_exit_fn(FCNAME);
			return SMPD_FAIL;
		}
	}
	if (WaitForSingleObject(smpd_process.hSMPDDataMutex, SMPD_SHORT_TIMEOUT*1000) != WAIT_OBJECT_0)
	{
		smpd_err_printf(L"lock_smpd_data failed\n");
		smpd_exit_fn(FCNAME);
		return SMPD_FAIL;
	}
	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}

#undef FCNAME
#define FCNAME L"smpd_unlock_smpd_data"
int smpd_unlock_smpd_data(void)
{
	smpd_enter_fn(FCNAME);
	if (!ReleaseMutex(smpd_process.hSMPDDataMutex))
	{
		int result;
		wchar_t err_msg[512];
		result = GetLastError();
		smpd_translate_win_error(result, err_msg, 512, L"Unable to release the SMPD data mutex, error %d: ", result);
		smpd_err_printf(L"%s\n", err_msg);
		return SMPD_FAIL;
	}
	smpd_exit_fn(FCNAME);
	return SMPD_SUCCESS;
}
