/*   Copyright (C) 2009 Deino Software. */

#include "ipmi.h"
#ifdef HAVE_CTYPE_H
#include <ctype.h>
#endif

/* pmiimpl.h */

static int root_smpd(void *p);

/* Define to prevent an smpd root thread or process from being created when there is only one process. */
/* Currently, defining this prevents the use of the spawn command. */
/*#define SINGLE_PROCESS_OPTIMIZATION*/

#define PMI_MAX_KEY_LEN		  256
#define PMI_MAX_VALUE_LEN		8192
#define PMI_MAX_KVS_NAME_LENGTH  100

#define PMI_INITIALIZED 0
#define PMI_FINALIZED   1

#define PMI_TRUE		1
#define PMI_FALSE	   0

typedef struct pmi_process_t
{
	int rpmi;
	HANDLE hRootThread;
	HANDLE hRootThreadReadyEvent;
	wchar_t root_host[100];
	int root_port;
	int local_kvs;
	wchar_t kvs_name[PMI_MAX_KVS_NAME_LENGTH];
	wchar_t domain_name[PMI_MAX_KVS_NAME_LENGTH];
	MPIDU_Sock_t sock;
	MPIDU_Sock_set_t set;
	int iproc;
	int nproc;
	int init_finalized;
	int smpd_id;
	MPIDU_SOCK_NATIVE_FD smpd_fd;
	int smpd_key;
	smpd_context_t *context;
	int clique_size;
	int *clique_ranks;
	wchar_t host[100];
	int port;
	int appnum;
} pmi_process_t;

/* global variables */
static pmi_process_t pmi_process =
{
	PMI_FALSE,		   /* rpmi		   */
	NULL,				/* root thread	*/
	NULL,				/* hRootThreadReadyEvent */
	L"",				  /* root host	  */
	0,				   /* root port	  */
	PMI_FALSE,		   /* local_kvs	  */
	L"",				  /* kvs_name	   */
	L"",				  /* domain_name	*/
	MPIDU_SOCK_INVALID_SOCK,  /* sock		   */
	MPIDU_SOCK_INVALID_SET,   /* set			*/
	-1,				  /* iproc		  */
	-1,				  /* nproc		  */
	PMI_FINALIZED,	   /* init_finalized */
	-1,				  /* smpd_id		*/
	0,				   /* smpd_fd		*/
	0,				   /* smpd_key	   */
	NULL,				/* context		*/
	0,				   /* clique_size	*/
	NULL,				/* clique_ranks   */
	L"",				  /* host		   */
	-1,				  /* port		   */
	0					/* appnum		 */
};

static FILE * get_output_file()
{
	FILE *fout = stdout;
	if (smpd_process.stdout_filename != NULL)
	{
		errno_t error = _wfopen_s(&fout, smpd_process.stdout_filename, L"a+");
		if (error != 0)
		{
			fout = stdout;
		}
	}
	return fout;
}

static void close_output_file(FILE *fout)
{
	if (fout != NULL && fout != stdout && fout != stderr)
		fclose(fout);
}

static int silence = 0;
static wchar_t pmi_last_error_msg[8192] = L"";
static char pmi_last_error_msgA[8192] = "";
static int pmi_err_printf(wchar_t *str, ...)
{
	int n=0;
	FILE *fout = get_output_file();
	va_list list;

	if (!silence)
	{
		fwprintf(fout, L"[%d] ", pmi_process.iproc);
		va_start(list, str);
		n = vfwprintf(fout, str, list);
		vswprintf_s(pmi_last_error_msg, 8192, str, list);
		va_end(list);
		//fwprintf(fout, L"%s", pmi_last_error_msg);

		fflush(fout);
	}

	close_output_file(fout);

	return n;
}

const char * iPMI_Get_last_error_msg()
{
	size_t num_converted;
	errno_t error;

	pmi_last_error_msgA[0] = '\0';
	error = wcstombs_s(&num_converted, pmi_last_error_msgA, 8192, pmi_last_error_msg, _TRUNCATE);

	return pmi_last_error_msgA;
}

const wchar_t * iPMI_Get_last_error_msgW()
{
	return pmi_last_error_msg;
}

int iPMI_Get_error_string(int pmi_error, char *str, int max_length)
{
	MPIU_UNREFERENCED_ARG(pmi_error);
	if (str != NULL && max_length > 0)
		str[0] = '\0';
	return PMI_SUCCESS;
}

int iPMI_Get_error_stringW(int pmi_error, wchar_t *str, int max_length)
{
	MPIU_UNREFERENCED_ARG(pmi_error);
	if (str != NULL && max_length > 0)
		str[0] = L'\0';
	return PMI_SUCCESS;
}

int iPMI_Free_error_string(int pmi_error)
{
	MPIU_UNREFERENCED_ARG(pmi_error);
	return PMI_SUCCESS;
}

static int pmi_mpi_err_printf(int mpi_errno, wchar_t *fmt, ... )
{
	int n;
	FILE *fout = get_output_file();
	va_list list;

	/* convert the error code to a string */
	fwprintf(fout, L"mpi_errno: %d\n", mpi_errno);

	fwprintf(fout, L"[%d] ", pmi_process.iproc);
	va_start(list, fmt);
	n = vfwprintf(fout, fmt, list);
	//vswprintf(pmi_last_error_msg, fmt, list);
	vswprintf_s(pmi_last_error_msg, 8192, fmt, list);
	va_end(list);

	fflush(fout);
	close_output_file(fout);

	MPIR_Err_return_comm(NULL, L"", mpi_errno);

	return n;
}

static int convert_to_wcs(const char *str, wchar_t **wstr_ptr)
{
	size_t len;
	wchar_t *wstr;

	len = strlen(str) + 1;
	len = len * sizeof(wchar_t);
	wstr = MPIU_Malloc(len);
	if (wstr == NULL)
		return PMI_ERR_NOMEM;

	*wstr_ptr = wstr;
	return PMI_SUCCESS;
}

static int pmi_create_post_command(const wchar_t *command, const wchar_t *name, const wchar_t *key, const wchar_t *value)
{
	int result;
	smpd_command_t *cmd_ptr;
	int dest = 1;
	int add_id = 0;

	if (!pmi_process.rpmi)
	{
		if (wcscmp(command, L"done") == 0)
		{
			/* done commands go to the immediate smpd, not the root */
			dest = pmi_process.smpd_id;
		}
	}
	if ((wcscmp(command, L"init") == 0) || (wcscmp(command, L"finalize") == 0))
	{
		add_id = 1;
		dest = 0;
	}

	result = smpd_create_command((wchar_t*)command, pmi_process.smpd_id, dest, SMPD_TRUE, &cmd_ptr);
	if (result != SMPD_SUCCESS)
	{
		pmi_err_printf(L"unable to create a %s command.\n", command);
		return PMI_FAIL;
	}
	result = smpd_add_command_int_arg(cmd_ptr, L"ctx_key", pmi_process.smpd_key);
	if (result != SMPD_SUCCESS)
	{
		pmi_err_printf(L"unable to add the key to the %s command.\n", command);
		return PMI_FAIL;
	}

	if (name != NULL)
	{
		result = smpd_add_command_arg(cmd_ptr, L"name", (wchar_t*)name);
		if (result != SMPD_SUCCESS)
		{
			pmi_err_printf(L"unable to add the kvs name('%s') to the %s command.\n", name, command);
			return PMI_FAIL;
		}
	}

	if (key != NULL)
	{
		result = smpd_add_command_arg(cmd_ptr, L"key", (wchar_t*)key);
		if (result != SMPD_SUCCESS)
		{
			pmi_err_printf(L"unable to add the key('%s') to the %s command.\n", key, command);
			return PMI_FAIL;
		}
	}

	if (value != NULL)
	{
		result = smpd_add_command_arg(cmd_ptr, L"value", (wchar_t*)value);
		if (result != SMPD_SUCCESS)
		{
			pmi_err_printf(L"unable to add the value('%s') to the %s command.\n", value, command);
			return PMI_FAIL;
		}
	}

	if (add_id)
	{
		result = smpd_add_command_int_arg(cmd_ptr, L"node_id", pmi_process.smpd_id);
		if (result != SMPD_SUCCESS)
		{
			pmi_err_printf(L"unable to add the node_id(%d) to the %s command.\n", pmi_process.smpd_id, command);
			return PMI_FAIL;
		}
	}

	/* post the write of the command */
	/*
	wprintf(L"posting write of dbs command to %s context, sock %d: '%s'\n",
		smpd_get_context_str(pmi_process.context), MPIDU_Sock_getid(pmi_process.context->sock), cmd_ptr->cmd);
	fflush(stdout);
	*/
	result = smpd_post_write_command(pmi_process.context, cmd_ptr);
	if (result != SMPD_SUCCESS)
	{
		pmi_err_printf(L"unable to post a write of the %s command.\n", command);
		return PMI_FAIL;
	}
	if (wcscmp(command, L"done"))
	{
		/* and post a read for the result if it is not a done command */
		result = smpd_post_read_command(pmi_process.context);
		if (result != SMPD_SUCCESS)
		{
			pmi_err_printf(L"unable to post a read of the next command on the pmi context.\n");
			return PMI_FAIL;
		}
	}

	/* let the state machine send the command and receive the result */
	result = smpd_enter_at_state(pmi_process.set, SMPD_WRITING_CMD);
	if (result != SMPD_SUCCESS)
	{
		pmi_err_printf(L"the state machine logic failed to get the result of the %s command.\n", command);
		return PMI_FAIL;
	}
	return PMI_SUCCESS;
}

int iPMI_Initialized(PMI_BOOL *initialized)
{
	if (initialized == NULL)
		return PMI_ERR_INVALID_ARG;
	if (pmi_process.init_finalized == PMI_INITIALIZED)
	{
		*initialized = PMI_TRUE;
	}
	else
	{
		*initialized = PMI_FALSE;
	}
	return PMI_SUCCESS;
}

static int parse_clique(const wchar_t *str_orig)
{
	int count, i;
	wchar_t *str, *token;
	wchar_t *tok;
	int first, last;

	/* count clique */
	count = 0;
	str = _wcsdup(str_orig);
	if (str == NULL)
		return PMI_FAIL;
	token = wcstok_s(str, L",", &tok);
	while (token)
	{
		first = _wtoi(token);
		while (isdigit(*token))
			token++;
		if (*token == L'\0')
			count++;
		else
		{
			if (*token == L'.')
			{
				token++;
				token++;
				last = _wtoi(token);
				count += last - first + 1;
			}
			else
			{
				pmi_err_printf(L"unexpected clique token: '%s'\n", token);
				free(str);
				return PMI_FAIL;
			}
		}
		token = wcstok_s(NULL, L",", &tok);
	}
	free(str);

	/* allocate array */
	pmi_process.clique_ranks = (int*)malloc(count * sizeof(int));
	if (pmi_process.clique_ranks == NULL)
		return PMI_FAIL;
	pmi_process.clique_size = count;
	
	/* populate array */
	count = 0;
	str = _wcsdup(str_orig);
	if (str == NULL)
		return PMI_FAIL;
	token = wcstok_s(str, L",", &tok);
	while (token)
	{
		first = _wtoi(token);
		while (isdigit(*token))
			token++;
		if (*token == L'\0')
		{
			pmi_process.clique_ranks[count] = first;
			count++;
		}
		else
		{
			if (*token == L'.')
			{
				token++;
				token++;
				last = _wtoi(token);
				for (i=first; i<=last; i++)
				{
					pmi_process.clique_ranks[count] = i;
					count++;
				}
			}
			else
			{
				pmi_err_printf(L"unexpected clique token: '%s'\n", token);
				free(str);
				return PMI_FAIL;
			}
		}
		token = wcstok_s(NULL, L",", &tok);
	}
	free(str);

	/*
	wprintf(L"clique: %d [", pmi_process.iproc);
	for (i=0; i<pmi_process.clique_size; i++)
	{
		wprintf(L"%d,", pmi_process.clique_ranks[i]);
	}
	wprintf(L"]\n");
	fflush(stdout);
	*/
	return PMI_SUCCESS;
}

static int uPMI_ConnectToHost(wchar_t *host, int port, smpd_state_t state)
{
	int result;
	wchar_t error_msg[MPI_MAX_ERROR_STRING];
	int len;

	/*wprintf(L"posting a connect to %s:%d\n", host, port);fflush(stdout);*/
	result = smpd_create_context(SMPD_CONTEXT_PMI, pmi_process.set, MPIDU_SOCK_INVALID_SOCK/*pmi_process.sock*/, smpd_process.id, &pmi_process.context);
	if (result != SMPD_SUCCESS)
	{
		pmi_err_printf(L"PMI_ConnectToHost failed: unable to create a context to connect to %s:%d with.\n", host, port);
		return PMI_FAIL;
	}

	result = MPIDU_Sock_post_connect(pmi_process.set, pmi_process.context, host, port, &pmi_process.sock);
	if (result != MPI_SUCCESS)
	{
		wprintf(L"MPIDU_Sock_post_connect failed.\n");fflush(stdout);
		len = MPI_MAX_ERROR_STRING;
		PMPI_Error_string(result, error_msg, &len);
		pmi_err_printf(L"PMI_ConnectToHost failed: unable to post a connect to %s:%d, error: %s\n", host, port, error_msg);
		wprintf(L"uPMI_ConnectToHost returning PMI_FAIL\n");fflush(stdout);
		return PMI_FAIL;
	}

	pmi_process.context->sock = pmi_process.sock;
	pmi_process.context->state = state;

	result = smpd_enter_at_state(pmi_process.set, state);
	if (result != MPI_SUCCESS)
	{
		pmi_mpi_err_printf(result, L"PMI_ConnectToHost failed: unable to connect to %s:%d.\n", host, port);
		return PMI_FAIL;
	}

	if (state == SMPD_CONNECTING_RPMI)
	{
		/* remote pmi processes receive their smpd_key when they connect to the smpd pmi server */
		pmi_process.smpd_key = _wtoi(pmi_process.context->session);
	}

	return SMPD_SUCCESS;
}

static int rPMI_Init(int *spawned)
{
#define MAX_ENV_LENGTH 8192
	wchar_t p[MAX_ENV_LENGTH];
	int result;
	wchar_t rank_str[100], size_str[100];
	wchar_t str[1024];

	if (spawned == NULL)
		return PMI_ERR_INVALID_ARG;

	/* initialize to defaults */
	smpd_process.id = 1;
	pmi_process.smpd_id = 1;
	pmi_process.rpmi = PMI_TRUE;
	pmi_process.iproc = 0;
	pmi_process.nproc = 1;

	if (GetEnvironmentVariable(L"PMI_ROOT_HOST", p, MAX_ENV_LENGTH) < 1)
	{
		pmi_err_printf(L"unable to initialize the rPMI library: no PMI_ROOT_HOST specified.\n");
		return PMI_FAIL;
	}
	MPIU_Wcsncpy(pmi_process.root_host, p, 100);

	if (GetEnvironmentVariable(L"PMI_ROOT_PORT", p, MAX_ENV_LENGTH) < 1)
	{
		/* set to default port? */
		pmi_err_printf(L"unable to initialize the rPMI library: no PMI_ROOT_PORT specified.\n");
		return PMI_FAIL;
	}
	pmi_process.root_port = _wtoi(p);
	if (pmi_process.root_port < 1)
	{
		pmi_err_printf(L"invalid root port specified: %s\n", p);
		return PMI_FAIL;
	}
	smpd_process.port = pmi_process.root_port;
	wcscpy_s(smpd_process.host, SMPD_MAX_HOST_LENGTH, pmi_process.root_host);

	if (GetEnvironmentVariable(L"PMI_SPAWN", p, MAX_ENV_LENGTH) > 0)
	{
		*spawned = _wtoi(p);
	}
	else
	{
		*spawned = 0;
	}

	if (GetEnvironmentVariable(L"PMI_KVS", p, MAX_ENV_LENGTH) > 0)
	{
		/* use specified kvs name */
		MPIU_Wcsncpy(pmi_process.kvs_name, p, PMI_MAX_KVS_NAME_LENGTH);
		MPIU_Wcsncpy(smpd_process.kvs_name, p, PMI_MAX_KVS_NAME_LENGTH);
	}
	else
	{
		/* use default kvs name */
		MPIU_Wcsncpy(pmi_process.kvs_name, L"default_mpich_kvs_name", PMI_MAX_KVS_NAME_LENGTH);
		MPIU_Wcsncpy(smpd_process.kvs_name, L"default_mpich_kvs_name", PMI_MAX_KVS_NAME_LENGTH);
	}

	if (GetEnvironmentVariable(L"PMI_DOMAIN", p, MAX_ENV_LENGTH) > 0)
	{
		MPIU_Wcsncpy(pmi_process.domain_name, p, PMI_MAX_KVS_NAME_LENGTH);
		MPIU_Wcsncpy(smpd_process.domain_name, p, PMI_MAX_KVS_NAME_LENGTH);
	}
	else
	{
		MPIU_Wcsncpy(pmi_process.domain_name, L"mpich2", PMI_MAX_KVS_NAME_LENGTH);
		MPIU_Wcsncpy(smpd_process.domain_name, L"mpich2", PMI_MAX_KVS_NAME_LENGTH);
	}

	if (GetEnvironmentVariable(L"PMI_RANK", p, MAX_ENV_LENGTH) > 0)
	{
		pmi_process.iproc = _wtoi(p);
		if (pmi_process.iproc < 0)
		{
			pmi_err_printf(L"invalid rank %d\n", pmi_process.iproc);
			return PMI_FAIL;
		}
	}

	if (GetEnvironmentVariable(L"PMI_SIZE", p, MAX_ENV_LENGTH) > 0)
	{
		pmi_process.nproc = _wtoi(p);
		if (pmi_process.nproc < 1)
		{
			pmi_err_printf(L"invalid size %d\n", pmi_process.nproc);
			return PMI_FAIL;
		}
	}
	smpd_process.nproc = pmi_process.nproc;
#ifdef SINGLE_PROCESS_OPTIMIZATION
/* leave this code #ifdef'd out so we can test rPMI stuff with one process */
	if (pmi_process.nproc == 1)
	{
		pmi_process.local_kvs = PMI_TRUE;
		result = smpd_dbs_init();
		if (result != SMPD_SUCCESS)
		{
			pmi_err_printf(L"unable to initialize the local dbs engine.\n");
			return PMI_FAIL;
		}
		result = smpd_dbs_create(pmi_process.kvs_name);
		if (result != SMPD_SUCCESS)
		{
			pmi_err_printf(L"unable to create the process group kvs\n");
			return PMI_FAIL;
		}
		pmi_process.init_finalized = PMI_INITIALIZED;
		return PMI_SUCCESS;
	}
#endif

	if (GetEnvironmentVariable(L"PMI_CLIQUE", p, MAX_ENV_LENGTH) > 0)
	{
		parse_clique(p);
	}

	/*
	wprintf(L"PMI_ROOT_HOST=%s PMI_ROOT_PORT=%s PMI_RANK=%s PMI_SIZE=%s PMI_KVS=%s PMI_CLIQUE=%s\n",
		MPIU_GetStaticEnv(L"PMI_ROOT_HOST"), MPIU_GetStaticEnv(L"PMI_ROOT_PORT"), MPIU_GetStaticEnv(L"PMI_RANK"), MPIU_GetStaticEnv(L"PMI_SIZE"),
		MPIU_GetStaticEnv(L"PMI_KVS"), MPIU_GetStaticEnv(L"PMI_CLIQUE"));
	fflush(stdout);
	*/

	if (pmi_process.iproc == 0)
	{
		if ((GetEnvironmentVariable(L"PMI_ROOT_LOCAL", p, MAX_ENV_LENGTH) > 0) && (wcscmp(p, L"1") == 0))
		{
			pmi_process.hRootThreadReadyEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
			if (pmi_process.hRootThreadReadyEvent == NULL)
			{
				pmi_err_printf(L"unable to create the root listener synchronization event, error: %d\n", GetLastError());
				return PMI_FAIL;
			}
			pmi_process.hRootThread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)root_smpd, NULL, 0, NULL);
			if (pmi_process.hRootThread == NULL)
			{
				pmi_err_printf(L"unable to create the root listener thread: error %d\n", GetLastError());
				return PMI_FAIL;
			}
			if (WaitForSingleObject(pmi_process.hRootThreadReadyEvent, 60000) != WAIT_OBJECT_0)
			{
				pmi_err_printf(L"the root process thread failed to initialize.\n");
				return PMI_FAIL;
			}
		}
	}

	/* connect to the root */

	result = MPIDU_Sock_create_set(&pmi_process.set);
	if (result != MPI_SUCCESS)
	{
		pmi_err_printf(L"PMI_Init failed: unable to create a sock set, error: %d\n", result);
		return PMI_FAIL;
	}

	result = uPMI_ConnectToHost(pmi_process.root_host, pmi_process.root_port, SMPD_CONNECTING_RPMI);
	if (result != SMPD_SUCCESS)
	{
		pmi_err_printf(L"PMI_Init failed.\n");
		return PMI_FAIL;
	}

	pmi_process.init_finalized = PMI_INITIALIZED;

	swprintf_s(rank_str, 100, L"%d", pmi_process.iproc);
	swprintf_s(size_str, 100, L"%d", pmi_process.nproc);
	result = pmi_create_post_command(L"init", pmi_process.kvs_name, rank_str, size_str);
	if (result != PMI_SUCCESS)
	{
		pmi_err_printf(L"PMI_Init failed: unable to create an init command.\n");
		return PMI_FAIL;
	}

	/* parse the result of the command */
	if (MPIU_Str_get_string_arg(pmi_process.context->read_cmd.cmd, L"result", str, 1024) != MPIU_STR_SUCCESS)
	{
		pmi_err_printf(L"PMI_Init failed: no result string in the result command.\n");
		return PMI_FAIL;
	}
	if (wcscmp(str, SMPD_SUCCESS_STR))
	{
		pmi_err_printf(L"PMI_Init failed: %s\n", str);
		return PMI_FAIL;
	}

	return PMI_SUCCESS;
}

static int rPMI_Finalize()
{
	int result;
	wchar_t rank_str[100];
	wchar_t str[1024];

	if (pmi_process.init_finalized == PMI_FINALIZED)
		return PMI_SUCCESS;

	if (pmi_process.local_kvs)
	{
		smpd_dbs_finalize();
		result = MPIDU_Sock_finalize();
		pmi_process.init_finalized = PMI_FINALIZED;
		return PMI_SUCCESS;
	}

	swprintf_s(rank_str, 100, L"%d", pmi_process.iproc);
	result = pmi_create_post_command(L"finalize", pmi_process.kvs_name, rank_str, NULL);
	if (result != PMI_SUCCESS)
	{
		pmi_err_printf(L"PMI_Finalize failed: unable to create an finalize command.\n");
		return PMI_FAIL;
	}

	/* parse the result of the command */
	if (MPIU_Str_get_string_arg(pmi_process.context->read_cmd.cmd, L"result", str, 1024) != MPIU_STR_SUCCESS)
	{
		pmi_err_printf(L"PMI_Finalize failed: no result string in the result command.\n");
		return PMI_FAIL;
	}
	if (wcscmp(str, SMPD_SUCCESS_STR))
	{
		pmi_err_printf(L"PMI_Finalize failed: %s\n", str);
		return PMI_FAIL;
	}

	if (pmi_process.iproc == 0)
	{
		/* the root process tells the root to exit when all the pmi contexts have exited */
		result = pmi_create_post_command(L"exit_on_done", NULL, NULL, NULL);
		if (result != PMI_SUCCESS)
		{
			pmi_err_printf(L"exit_on_done command failed.\n");
			return PMI_FAIL;
		}
		/*wprintf(L"exit_on_done command returned successfully.\n");fflush(stdout);*/
	}

	/*wprintf(L"entering finalize pmi_barrier.\n");fflush(stdout);*/
	PMI_Barrier();
	/*wprintf(L"after finalize pmi_barrier, posting done command.\n");fflush(stdout);*/

	/* post a done command to close the pmi context */
	result = pmi_create_post_command(L"done", NULL, NULL, NULL);
	if (result != PMI_SUCCESS)
	{
		pmi_err_printf(L"failed.\n");
		return PMI_FAIL;
	}

	if (pmi_process.iproc == 0)
	{
		WaitForSingleObject(pmi_process.hRootThread, INFINITE);
	}

	/*if (pmi_process.sock != MPIDU_SOCK_INVALID_SOCK)*/
	{
		result = MPIDU_Sock_finalize();
		if (result != MPI_SUCCESS)
		{
			/*pmi_err_printf(L"MPIDU_Sock_finalize failed, error: %d\n", result);*/
		}
	}

	pmi_process.init_finalized = PMI_FINALIZED;

	return PMI_SUCCESS;
}

int iPMI_Init(int *spawned)
{
#undef MAX_ENV_LENGTH
#define MAX_ENV_LENGTH 8192
	wchar_t p[MAX_ENV_LENGTH];
	int result;
	wchar_t rank_str[100], size_str[100];
	wchar_t str[1024];

	if (spawned == NULL)
		return PMI_ERR_INVALID_ARG;

	/* don't allow pmi_init to be called more than once */
	if (pmi_process.init_finalized == PMI_INITIALIZED)
		return PMI_SUCCESS;

	/* initialize to defaults */

	result = MPIDU_Sock_init();
	if (result != MPI_SUCCESS)
	{
		pmi_err_printf(L"MPIDU_Sock_init failed,\nsock error: %s\n", get_sock_error_string(result));
		return PMI_FAIL;
	}

	result = smpd_init_process();
	if (result != SMPD_SUCCESS)
	{
		pmi_err_printf(L"unable to initialize the smpd global process structure.\n");
		return PMI_FAIL;
	}

	if (GetEnvironmentVariable(L"PMI_ROOT_HOST", p, MAX_ENV_LENGTH) > 0)
	{
		return rPMI_Init(spawned);
	}

	pmi_process.iproc = 0;
	pmi_process.nproc = 1;


	if (GetEnvironmentVariable(L"PMI_SPAWN", p, MAX_ENV_LENGTH) > 0)
	{
		*spawned = _wtoi(p);
	}
	else
	{
		*spawned = 0;
	}

	if (GetEnvironmentVariable(L"PMI_APPNUM", p, MAX_ENV_LENGTH) > 0)
	{
		pmi_process.appnum = _wtoi(p);
	}
	else
	{
		pmi_process.appnum = 0;
	}

	if (GetEnvironmentVariable(L"PMI_KVS", p, MAX_ENV_LENGTH) > 0)
	{
		MPIU_Wcsncpy(pmi_process.kvs_name, p, PMI_MAX_KVS_NAME_LENGTH);
	}
	else
	{
		pmi_process.local_kvs = PMI_TRUE;
		result = smpd_dbs_init();
		if (result != SMPD_SUCCESS)
		{
			pmi_err_printf(L"unable to initialize the local dbs engine.\n");
			return PMI_FAIL;
		}
		result = smpd_dbs_create(pmi_process.kvs_name, PMI_MAX_KVS_NAME_LENGTH);
		if (result != SMPD_SUCCESS)
		{
			pmi_err_printf(L"unable to create the process group kvs\n");
			return PMI_FAIL;
		}
		MPIU_Wcsncpy(pmi_process.domain_name, smpd_process.domain_name, PMI_MAX_KVS_NAME_LENGTH);
		pmi_process.init_finalized = PMI_INITIALIZED;
		return PMI_SUCCESS;
	}

	if (GetEnvironmentVariable(L"PMI_DOMAIN", p, MAX_ENV_LENGTH) > 0)
	{
		MPIU_Wcsncpy(pmi_process.domain_name, p, PMI_MAX_KVS_NAME_LENGTH);
	}
	else
	{
		MPIU_Wcsncpy(pmi_process.domain_name, L"mpich2", PMI_MAX_KVS_NAME_LENGTH);
	}

	if (GetEnvironmentVariable(L"PMI_RANK", p, MAX_ENV_LENGTH) > 0)
	{
		pmi_process.iproc = _wtoi(p);
		if (pmi_process.iproc < 0)
		{
			pmi_err_printf(L"invalid rank %d, setting to 0\n", pmi_process.iproc);
			pmi_process.iproc = 0;
		}
	}

	if (GetEnvironmentVariable(L"PMI_SIZE", p, MAX_ENV_LENGTH) > 0)
	{
		pmi_process.nproc = _wtoi(p);
		if (pmi_process.nproc < 1)
		{
			pmi_err_printf(L"invalid size %d, setting to 1\n", pmi_process.nproc);
			pmi_process.nproc = 1;
		}
	}

	if (GetEnvironmentVariable(L"PMI_SMPD_ID", p, MAX_ENV_LENGTH) > 0)
	{
		pmi_process.smpd_id = _wtoi(p);
		smpd_process.id = pmi_process.smpd_id;
	}

	if (GetEnvironmentVariable(L"PMI_SMPD_KEY", p, MAX_ENV_LENGTH) > 0)
	{
		pmi_process.smpd_key = _wtoi(p);
	}

	if (GetEnvironmentVariable(L"PMI_SMPD_FD", p, MAX_ENV_LENGTH) > 0)
	{
		result = MPIDU_Sock_create_set(&pmi_process.set);
		if (result != MPI_SUCCESS)
		{
			pmi_err_printf(L"PMI_Init failed: unable to create a sock set, error:\n%s\n",
				get_sock_error_string(result));
			return PMI_FAIL;
		}

		pmi_process.smpd_fd = smpd_decode_handle(p);
		result = MPIDU_Sock_native_to_sock(pmi_process.set, pmi_process.smpd_fd, NULL, &pmi_process.sock);
		if (result != MPI_SUCCESS)
		{
			pmi_err_printf(L"MPIDU_Sock_native_to_sock failed, error %s\n", get_sock_error_string(result));
			return PMI_FAIL;
		}
		result = smpd_create_context(SMPD_CONTEXT_PMI, pmi_process.set, pmi_process.sock, pmi_process.smpd_id, &pmi_process.context);
		if (result != SMPD_SUCCESS)
		{
			pmi_err_printf(L"unable to create a pmi context.\n");
			return PMI_FAIL;
		}
	}
	else
	{
		if (GetEnvironmentVariable(L"PMI_HOST", p, MAX_ENV_LENGTH) > 0)
		{
			MPIU_Wcsncpy(pmi_process.host, p, 100);
			if (GetEnvironmentVariable(L"PMI_PORT", p, MAX_ENV_LENGTH) > 0)
			{
				pmi_process.port = _wtoi(p);

				result = MPIDU_Sock_create_set(&pmi_process.set);
				if (result != MPI_SUCCESS)
				{
					pmi_err_printf(L"PMI_Init failed: unable to create a sock set, error: %d\n", result);
					return PMI_FAIL;
				}

				result = uPMI_ConnectToHost(pmi_process.host, pmi_process.port, SMPD_CONNECTING_PMI);
				if (result != SMPD_SUCCESS)
				{
					pmi_err_printf(L"PMI_Init failed.\n");
					return PMI_FAIL;
				}
			}
			else
			{
				pmi_err_printf(L"No mechanism specified for connecting to the process manager - host %s but no port provided.\n", pmi_process.host);
				return PMI_FAIL;
			}
		}
		else
		{
			pmi_err_printf(L"No mechanism specified for connecting to the process manager.\n");
			return PMI_FAIL;
		}
	}

	if (GetEnvironmentVariable(L"PMI_CLIQUE", p, MAX_ENV_LENGTH) > 0)
	{
		parse_clique(p);
	}

	/*
	wprintf(L"PMI_RANK=%s PMI_SIZE=%s PMI_KVS=%s PMI_SMPD_ID=%s PMI_SMPD_FD=%s PMI_SMPD_KEY=%s\n PMI_SPAWN=%s",
		MPIU_GetStaticEnv(L"PMI_RANK"), MPIU_GetStaticEnv(L"PMI_SIZE"), MPIU_GetStaticEnv(L"PMI_KVS"), MPIU_GetStaticEnv(L"PMI_SMPD_ID"),
		MPIU_GetStaticEnv(L"PMI_SMPD_FD"), MPIU_GetStaticEnv(L"PMI_SMPD_KEY"), MPIU_GetStaticEnv(L"PMI_SPAWN"));
	fflush(stdout);
	*/

	pmi_process.init_finalized = PMI_INITIALIZED;

	swprintf_s(rank_str, 100, L"%d", pmi_process.iproc);
	swprintf_s(size_str, 100, L"%d", pmi_process.nproc);
	result = pmi_create_post_command(L"init", pmi_process.kvs_name, rank_str, size_str);
	if (result != PMI_SUCCESS)
	{
		pmi_err_printf(L"PMI_Init failed: unable to create an init command.\n");
		return PMI_FAIL;
	}

	/* parse the result of the command */
	if (MPIU_Str_get_string_arg(pmi_process.context->read_cmd.cmd, L"result", str, 1024) != MPIU_STR_SUCCESS)
	{
		pmi_err_printf(L"PMI_Init failed: no result string in the result command.\n");
		return PMI_FAIL;
	}
	if (wcscmp(str, SMPD_SUCCESS_STR))
	{
		pmi_err_printf(L"PMI_Init failed: %s\n", str);
		return PMI_FAIL;
	}

	/*
	if (*spawned && pmi_process.iproc == 0)
	{
		wchar_t key[1024], val[8192];
		key[0] = L'\0';
		result = PMI_KVS_Iter_first(pmi_process.kvs_name, key, 1024, val, 8192);
		if (result != PMI_SUCCESS || key[0] == L'\0')
		{
			wprintf(L"No preput values in %s\n", pmi_process.kvs_name);
		}
		while (result == PMI_SUCCESS && key[0] != L'\0')
		{
			wprintf(L"PREPUT key=%s, val=%s\n", key, val);
			result = PMI_KVS_Iter_next(pmi_process.kvs_name, key, 1024, val, 8192);
		}
		fflush(stdout);
	}
	iPMI_Barrier();
	*/

	/*wprintf(L"iPMI_Init returning success.\n");fflush(stdout);*/
	return PMI_SUCCESS;
}

int iPMI_Finalize()
{
	int result;
	wchar_t rank_str[100];
	wchar_t str[1024];

	if (pmi_process.init_finalized == PMI_FINALIZED)
		return PMI_SUCCESS;

	/*
	wprintf(L"PMI_Finalize called.\n");
	fflush(stdout);
	*/

	if (pmi_process.rpmi)
	{
		return rPMI_Finalize();
	}

	if (pmi_process.local_kvs)
	{
		smpd_dbs_finalize();
		result = MPIDU_Sock_finalize();
		pmi_process.init_finalized = PMI_FINALIZED;
		return PMI_SUCCESS;
	}

	swprintf_s(rank_str, 100, L"%d", pmi_process.iproc);
	result = pmi_create_post_command(L"finalize", pmi_process.kvs_name, rank_str, NULL);
	if (result != PMI_SUCCESS)
	{
		pmi_err_printf(L"PMI_Finalize failed: unable to create an finalize command.\n");
		goto fn_fail;
	}

	/* parse the result of the command */
	if (MPIU_Str_get_string_arg(pmi_process.context->read_cmd.cmd, L"result", str, 1024) != MPIU_STR_SUCCESS)
	{
		pmi_err_printf(L"PMI_Finalize failed: no result string in the result command.\n");
		goto fn_fail;
	}
	if (wcscmp(str, SMPD_SUCCESS_STR))
	{
		pmi_err_printf(L"PMI_Finalize failed: %s\n", str);
		goto fn_fail;
	}

	PMI_Barrier();

	/* post the done command and wait for the result */
	result = pmi_create_post_command(L"done", NULL, NULL, NULL);
	if (result != PMI_SUCCESS)
	{
		pmi_err_printf(L"failed.\n");
		goto fn_fail;
	}

	/*if (pmi_process.sock != MPIDU_SOCK_INVALID_SOCK)*/
	{
		result = MPIDU_Sock_finalize();
		if (result != MPI_SUCCESS)
		{
			/*pmi_err_printf(L"MPIDU_Sock_finalize failed,\nsock error: %s\n", get_sock_error_string(result));*/
		}
	}

	pmi_process.init_finalized = PMI_FINALIZED;
	/*wprintf(L"iPMI_Finalize success.\n");fflush(stdout);*/
	return PMI_SUCCESS;

fn_fail:
	/* set the state to finalized so PMI_Abort will not dereference mangled structures due to a failure */
	pmi_process.init_finalized = PMI_FINALIZED;
	return PMI_FAIL;
}

int iPMI_Abort(int exit_code, const char error_msg[])
{
	size_t len, num_converted;
	wchar_t *werror_msg;
	errno_t error;

	len = strlen(error_msg) + 1;
	werror_msg = (wchar_t*)MPIU_Malloc(len * sizeof(wchar_t));
	if (werror_msg == NULL)
	{
		return iPMI_AbortW(exit_code, L"");
	}

	werror_msg[0] = L'\0';
	error = mbstowcs_s(&num_converted, werror_msg, len, error_msg, _TRUNCATE);

	return iPMI_AbortW(exit_code, werror_msg);
}

int iPMI_AbortW(int exit_code, const wchar_t error_msg[])
{
	int result;
	smpd_command_t *cmd_ptr;
	FILE *fout = NULL;

	/* flush any output before aborting */
	/* This doesn't work because it flushes output from the mpich dll but does not flush the main module's output */
	fflush(stdout);
	fflush(stderr);

	if (pmi_process.init_finalized == PMI_FINALIZED)
	{
		fout = get_output_file();
		fwprintf(fout, L"PMI_Abort called after PMI_Finalize, error message:\n%s\n", error_msg);
		fflush(fout);
		close_output_file(fout);
		ExitProcess(exit_code);
	}

	if (pmi_process.local_kvs)
	{
		fout = get_output_file();
		if (smpd_process.verbose_abort_output)
		{
			fwprintf(fout, L"\njob aborted:\n");
			fwprintf(fout, L"process: node: exit code: error message:\n");
			fwprintf(fout, L"0: localhost: %d", exit_code);
			if (error_msg != NULL)
			{
				fwprintf(fout, L": %s", error_msg);
			}
			fwprintf(fout, L"\n");
		}
		else
		{
			if (error_msg != NULL)
			{
				fwprintf(fout, L"%s\n", error_msg);
			}
		}
		fflush(fout);
		close_output_file(fout);
		smpd_dbs_finalize();
		pmi_process.init_finalized = PMI_FINALIZED;
		ExitProcess(exit_code);
	}

	result = smpd_create_command(L"abort_job", pmi_process.smpd_id, 0, SMPD_FALSE, &cmd_ptr);
	if (result != SMPD_SUCCESS)
	{
		pmi_err_printf(L"unable to create an abort command.\n");
		return PMI_FAIL;
	}

	result = smpd_add_command_arg(cmd_ptr, L"name", pmi_process.kvs_name);
	if (result != SMPD_SUCCESS)
	{
		pmi_err_printf(L"unable to add the kvs name('%s') to the abort command.\n", pmi_process.kvs_name);
		return PMI_FAIL;
	}

	result = smpd_add_command_int_arg(cmd_ptr, L"rank", pmi_process.iproc);
	if (result != SMPD_SUCCESS)
	{
		pmi_err_printf(L"unable to add the rank %d to the abort command.\n", pmi_process.iproc);
		return PMI_FAIL;
	}

	result = smpd_add_command_arg(cmd_ptr, L"error", (wchar_t*)error_msg);
	if (result != SMPD_SUCCESS)
	{
		pmi_err_printf(L"unable to add the error message('%s') to the abort command.\n", error_msg);
		return PMI_FAIL;
	}

	result = smpd_add_command_int_arg(cmd_ptr, L"exit_code", exit_code);
	if (result != SMPD_SUCCESS)
	{
		pmi_err_printf(L"unable to add the exit code(%d) to the abort command.\n", exit_code);
		return PMI_FAIL;
	}

	/* post the write of the command */
	result = smpd_post_write_command(pmi_process.context, cmd_ptr);
	if (result != SMPD_SUCCESS)
	{
		pmi_err_printf(L"unable to post a write of the abort command.\n");
		return PMI_FAIL;
	}

	/* and post a read for the result */
	/*
	result = smpd_post_read_command(pmi_process.context);
	if (result != SMPD_SUCCESS)
	{
		pmi_err_printf(L"unable to post a read of the next command on the pmi context.\n");
		return PMI_FAIL;
	}
	*/

	/* let the state machine send the command and receive the result */
	result = smpd_enter_at_state(pmi_process.set, SMPD_WRITING_CMD);
	if (result != SMPD_SUCCESS)
	{
		pmi_err_printf(L"the state machine logic failed to handle the abort command.\n");
		return PMI_FAIL;
	}
	ExitProcess(exit_code);
}

int iPMI_Register_abort_hook(PMI_Abort_callback_t function)
{
	int result;
	smpd_command_t *cmd_ptr;
	wchar_t function_str[20];
	wchar_t str[1024];

	if (pmi_process.init_finalized == PMI_FINALIZED)
		return PMI_ERR_INIT;

	if (pmi_process.local_kvs == PMI_TRUE)
	{
		/* No process manager present so no hook can be registered. */
		return PMI_ERR_NO_MANAGER;
	}

	/* In order for the manager process to create a remote thread in this process it needs the address of
	 * the function to call and a handle to this process.  Pass the function address here and the process id
	 * which the manager can use to get a handle to this process.  Passing NULL for the function pointer
	 * will remove the hook.
	 */
	MPIU_Snprintf(function_str, 20, _TRUNCATE, L"%p", function);

	result = smpd_create_command(L"abort_hook", pmi_process.smpd_id, pmi_process.smpd_id, SMPD_TRUE, &cmd_ptr);
	if (result != SMPD_SUCCESS)
	{
		pmi_err_printf(L"unable to create an abort_hook command.\n");
		return PMI_FAIL;
	}
	result = smpd_add_command_int_arg(cmd_ptr, L"ctx_key", pmi_process.smpd_key);
	if (result != SMPD_SUCCESS)
	{
		pmi_err_printf(L"unable to add the key to the abort_hook command.\n");
		return PMI_FAIL;
	}
	result = smpd_add_command_arg(cmd_ptr, L"address", function_str);
	if (result != SMPD_SUCCESS)
	{
		pmi_err_printf(L"unable to add the function address ('%s') to the abort_hook command.\n", function_str);
		return PMI_FAIL;
	}
	result = smpd_add_command_int_arg(cmd_ptr, L"pid", GetCurrentProcessId());
	if (result != SMPD_SUCCESS)
	{
		pmi_err_printf(L"unable to add the process id (%d) to the abort_hook command.\n", GetCurrentProcessId());
		return PMI_FAIL;
	}

	result = smpd_post_write_command(pmi_process.context, cmd_ptr);
	if (result != SMPD_SUCCESS)
	{
		pmi_err_printf(L"unable to post a write of the abort_hook command.\n");
		return PMI_FAIL;
	}
	result = smpd_post_read_command(pmi_process.context);
	if (result != SMPD_SUCCESS)
	{
		pmi_err_printf(L"unable to post a read of the next command on the pmi context.\n");
		return PMI_FAIL;
	}

	/* let the state machine send the command and receive the result */
	result = smpd_enter_at_state(pmi_process.set, SMPD_WRITING_CMD);
	if (result != SMPD_SUCCESS)
	{
		pmi_err_printf(L"the state machine logic failed to get the result of the abort_hook command.\n");
		return PMI_FAIL;
	}

	/* parse the result of the command */
	if (MPIU_Str_get_string_arg(pmi_process.context->read_cmd.cmd, L"result", str, 1024) != MPIU_STR_SUCCESS)
	{
		pmi_err_printf(L"PMI_Register_abort_hook failed: no result string in the result command.\n");
		return PMI_FAIL;
	}
	if (wcscmp(str, SMPD_SUCCESS_STR))
	{
		pmi_err_printf(L"PMI_Register_abort_hook failed: '%s'\n", str);
		return PMI_FAIL;
	}

	return PMI_SUCCESS;
}

static PMI_Print_queues_fn_t print_queues_fn = NULL;
int iPMI_Register_print_queues_fn(PMI_Print_queues_fn_t function)
{
	if (pmi_process.init_finalized == PMI_FINALIZED)
		return PMI_ERR_INIT;
	print_queues_fn = function;
	return PMI_SUCCESS;
}

int iPMI_Print_queues(wchar_t *str, int max_length)
{
	int i;

	if (pmi_process.init_finalized == PMI_FINALIZED)
		return PMI_ERR_INIT;
	if (str == NULL || max_length < 1)
		return PMI_ERR_INVALID_ARG;
	if (print_queues_fn == NULL)
	{
		str[0] = L'\0';
		return PMI_FAIL;
	}

	print_queues_fn(str, max_length);
	for (i=0; i<max_length; i++)
	{
		if (str[i] == L'\0')
			return PMI_SUCCESS;
	}
	str[max_length-1] = L'\0';
	return PMI_FAIL;
}

static PMI_Print_call_history_fn_t print_call_history_fn = NULL;
int iPMI_Register_print_call_history_fn(PMI_Print_call_history_fn_t function)
{
	if (pmi_process.init_finalized == PMI_FINALIZED)
		return PMI_ERR_INIT;
	print_call_history_fn = function;
	return PMI_SUCCESS;
}

int iPMI_Print_call_history(wchar_t *str, int max_length)
{
	int i;

	if (pmi_process.init_finalized == PMI_FINALIZED)
		return PMI_ERR_INIT;
	if (str == NULL || max_length < 1)
		return PMI_ERR_INVALID_ARG;
	if (print_call_history_fn == NULL)
	{
		str[0] = L'\0';
		return PMI_FAIL;
	}

	print_call_history_fn(str, max_length);
	for (i=0; i<max_length; i++)
	{
		if (str[i] == L'\0')
			return PMI_SUCCESS;
	}
	str[max_length-1] = L'\0';
	return PMI_FAIL;
}

int iPMI_Get_size(int *size)
{
	if (pmi_process.init_finalized == PMI_FINALIZED)
		return PMI_ERR_INIT;
	if (size == NULL)
		return PMI_ERR_INVALID_ARG;

	*size = pmi_process.nproc;

	return PMI_SUCCESS;
}

int iPMI_Get_rank(int *rank)
{
	if (pmi_process.init_finalized == PMI_FINALIZED)
		return PMI_ERR_INIT;
	if (rank == NULL)
		return PMI_ERR_INVALID_ARG;

	*rank = pmi_process.iproc;

	return PMI_SUCCESS;
}

int iPMI_Get_universe_size(int *size)
{
	if (pmi_process.init_finalized == PMI_FINALIZED)
		return PMI_ERR_INIT;
	if (size == NULL)
		return PMI_ERR_INVALID_ARG;

	*size = -1;

	return PMI_SUCCESS;
}

int iPMI_Get_appnum(int *appnum)
{
	if (pmi_process.init_finalized == PMI_FINALIZED)
		return PMI_ERR_INIT;
	if (appnum == NULL)
		return PMI_ERR_INVALID_ARG;

	*appnum = pmi_process.appnum;

	return PMI_SUCCESS;
}

int iPMI_Get_clique_size( int *size )
{
	if (pmi_process.init_finalized == PMI_FINALIZED)
		return PMI_ERR_INIT;
	if (size == NULL)
		return PMI_ERR_INVALID_ARG;

	if (pmi_process.clique_size == 0)
		*size = 1;
	else
		*size = pmi_process.clique_size;
	return PMI_SUCCESS;
}

int iPMI_Get_clique_ranks( int ranks[], int length )
{
	int i;

	if (pmi_process.init_finalized == PMI_FINALIZED)
		return PMI_ERR_INIT;
	if (ranks == NULL)
		return PMI_ERR_INVALID_ARG;
	if (length < pmi_process.clique_size)
		return PMI_ERR_INVALID_LENGTH;

	if (pmi_process.clique_size == 0)
	{
		*ranks = 0;
	}
	else
	{
		for (i=0; i<pmi_process.clique_size; i++)
		{
			ranks[i] = pmi_process.clique_ranks[i];
		}
	}
	return PMI_SUCCESS;
}

int iPMI_Get_id( char id_str[], int length )
{
	return iPMI_KVS_Get_my_name(id_str, length);
}

int iPMI_Get_idW( wchar_t id_str[], int length )
{
	return iPMI_KVS_Get_my_nameW(id_str, length);
}

int iPMI_Get_id_length_max(int *maxlen)
{
	return iPMI_KVS_Get_name_length_max(maxlen);
}

int iPMI_Get_kvs_domain_id(char id_str[], int length)
{
	wchar_t *wstr;
	size_t num_converted;
	int result;
	errno_t error;

	wstr = (wchar_t*)MPIU_Malloc((length+1)*sizeof(wchar_t));
	if (wstr == NULL)
		return PMI_ERR_NOMEM;
	result = iPMI_Get_kvs_domain_idW(wstr, length);
	if (result == PMI_SUCCESS)
	{
		id_str[0] = '\0';
		error = wcstombs_s(&num_converted, id_str, length, wstr, _TRUNCATE);
		if (error != 0)
		{
			/* FIXME: Check for and return a truncation error? */
		}
	}
	MPIU_Free(wstr);
	return result;
}

int iPMI_Get_kvs_domain_idW(wchar_t id_str[], int length)
{
	if (pmi_process.init_finalized == PMI_FINALIZED)
		return PMI_ERR_INIT;
	if (id_str == NULL)
		return PMI_ERR_INVALID_ARG;
	if (length < PMI_MAX_KVS_NAME_LENGTH)
		return PMI_ERR_INVALID_LENGTH;

	MPIU_Wcsncpy(id_str, pmi_process.domain_name, length);

	return PMI_SUCCESS;
}

int iPMI_Barrier()
{
	int result;
	wchar_t count_str[20];
	wchar_t str[1024];
	
	if (pmi_process.init_finalized == PMI_FINALIZED)
		return PMI_ERR_INIT;

	if (pmi_process.nproc == 1)
		return PMI_SUCCESS;

	/*wprintf(L"entering barrier %d, %s\n", pmi_process.nproc, pmi_process.kvs_name);fflush(stdout);*/

	/* encode the size of the barrier */
	_snwprintf_s(count_str, 20, _TRUNCATE, L"%d", pmi_process.nproc);

	/* post the command and wait for the result */
	result = pmi_create_post_command(L"barrier", pmi_process.kvs_name, NULL, count_str);
	if (result != PMI_SUCCESS)
	{
		pmi_err_printf(L"PMI_Barrier failed.\n");
		return PMI_FAIL;
	}

	/* interpret the result */
	if (MPIU_Str_get_string_arg(pmi_process.context->read_cmd.cmd, L"result", str, 1024) != MPIU_STR_SUCCESS)
	{
		pmi_err_printf(L"PMI_Barrier failed: no result string in the result command.\n");
		return PMI_FAIL;
	}
	if (wcscmp(str, DBS_SUCCESS_STR))
	{
		pmi_err_printf(L"PMI_Barrier failed: '%s'\n", str);
		return PMI_FAIL;
	}

	/*wprintf(L"iPMI_Barrier success.\n");fflush(stdout);*/
	return PMI_SUCCESS;
}

int iPMI_KVS_Get_my_name(char kvsname[], int length)
{
	wchar_t *wkvsname;
	size_t num_converted;
	int result;
	errno_t error;

	wkvsname = (wchar_t*)MPIU_Malloc(length * sizeof(wchar_t));
	if (wkvsname == NULL)
		return PMI_ERR_NOMEM;

	result = iPMI_KVS_Get_my_nameW(wkvsname, length);
	if (result == PMI_SUCCESS)
	{
		kvsname[0] = '\0';
		error = wcstombs_s(&num_converted, kvsname, length, wkvsname, _TRUNCATE);
		if (error != 0)
		{
			if (num_converted >= (size_t)length)
			{
				/* truncation error */
				result = PMI_ERR_TRUNCATION;
			}
			else
				result = PMI_FAIL;
		}
	}
	MPIU_Free(wkvsname);
	return result;
}

int iPMI_KVS_Get_my_nameW(wchar_t kvsname[], int length)
{
	if (pmi_process.init_finalized == PMI_FINALIZED)
		return PMI_ERR_INIT;
	if (kvsname == NULL)
		return PMI_ERR_INVALID_ARG;
	if (length < PMI_MAX_KVS_NAME_LENGTH)
		return PMI_ERR_INVALID_LENGTH;

	MPIU_Wcsncpy(kvsname, pmi_process.kvs_name, length);

	/*
	wprintf(L"my kvs name is %s\n", kvsname);fflush(stdout);
	*/

	return PMI_SUCCESS;
}

int iPMI_KVS_Get_name_length_max(int *maxlen)
{
	if (pmi_process.init_finalized == PMI_FINALIZED)
		return PMI_ERR_INIT;
	if (maxlen == NULL)
		return PMI_ERR_INVALID_ARG;
	*maxlen = PMI_MAX_KVS_NAME_LENGTH;
	return PMI_SUCCESS;
}

int iPMI_KVS_Get_key_length_max(int *maxlen)
{
	if (pmi_process.init_finalized == PMI_FINALIZED)
		return PMI_ERR_INIT;
	if (maxlen == NULL)
		return PMI_ERR_INVALID_ARG;
	*maxlen = PMI_MAX_KEY_LEN;
	return PMI_SUCCESS;
}

int iPMI_KVS_Get_value_length_max(int *maxlen)
{
	if (pmi_process.init_finalized == PMI_FINALIZED)
		return PMI_ERR_INIT;
	if (maxlen == NULL)
		return PMI_ERR_INVALID_ARG;
	*maxlen = PMI_MAX_VALUE_LEN;
	return PMI_SUCCESS;
}

int iPMI_KVS_Create(char kvsname[], int length)
{
	wchar_t *wkvsname;
	size_t num_converted;
	int result;
	errno_t error;

	wkvsname = (wchar_t*)MPIU_Malloc(length * sizeof(wchar_t));
	if (wkvsname == NULL)
		return PMI_ERR_NOMEM;

	result = iPMI_KVS_CreateW(wkvsname, length);
	if (result == PMI_SUCCESS)
	{
		kvsname[0] = '\0';
		error = wcstombs_s(&num_converted, kvsname, length, wkvsname, _TRUNCATE);
		if (error != 0)
		{
			if (num_converted >= (size_t)length)
			{
				/* truncation error */
				result = PMI_ERR_TRUNCATION;
			}
			else
				result = PMI_FAIL;
		}
	}
	MPIU_Free(wkvsname);
	return result;
}

int iPMI_KVS_CreateW(wchar_t kvsname[], int length)
{
	int result;
	wchar_t str[1024];

	if (pmi_process.init_finalized == PMI_FINALIZED)
		return PMI_ERR_INIT;
	if (kvsname == NULL)
		return PMI_ERR_INVALID_ARG;
	if (length < PMI_MAX_KVS_NAME_LENGTH)
		return PMI_ERR_INVALID_LENGTH;

	if (pmi_process.local_kvs)
	{
		result = smpd_dbs_create(kvsname, length);
		return (result == SMPD_SUCCESS) ? PMI_SUCCESS : PMI_FAIL;
	}

	result = pmi_create_post_command(L"dbcreate", NULL, NULL, NULL);
	if (result != PMI_SUCCESS)
	{
		pmi_err_printf(L"PMI_KVS_Create failed: unable to create a pmi kvs space.\n");
		return PMI_FAIL;
	}

	/* parse the result of the command */
	if (MPIU_Str_get_string_arg(pmi_process.context->read_cmd.cmd, L"result", str, 1024) != MPIU_STR_SUCCESS)
	{
		pmi_err_printf(L"PMI_KVS_Create failed: no result string in the result command.\n");
		return PMI_FAIL;
	}
	if (wcscmp(str, DBS_SUCCESS_STR))
	{
		pmi_err_printf(L"PMI_KVS_Create failed: %s\n", str);
		return PMI_FAIL;
	}
	if (MPIU_Str_get_string_arg(pmi_process.context->read_cmd.cmd, L"name", str, 1024) != MPIU_STR_SUCCESS)
	{
		pmi_err_printf(L"PMI_KVS_Create failed: no kvs name in the dbcreate result command.\n");
		return PMI_FAIL;
	}
	MPIU_Wcsncpy(kvsname, str, PMI_MAX_KVS_NAME_LENGTH);

	/*wprintf(L"iPMI_KVS_Create success.\n");fflush(stdout);*/
	return PMI_SUCCESS;
}

int iPMI_KVS_Destroy(const char kvsname[])
{
	size_t len, num_converted;
	wchar_t *wstr;
	int result;
	errno_t error;

	len = strlen(kvsname) + 1;
	wstr = (wchar_t*)MPIU_Malloc(len * sizeof(wchar_t));
	if (wstr == NULL)
	{
		return PMI_ERR_NOMEM;
	}

	error = mbstowcs_s(&num_converted, wstr, len, kvsname, _TRUNCATE);
	if (error != 0)
	{
		MPIU_Free(wstr);
		if (num_converted >= len)
		{
			/* truncation error */
			return PMI_ERR_TRUNCATION;
		}
		/* some error */
		return PMI_FAIL;
	}

	result = iPMI_KVS_DestroyW(wstr);
	MPIU_Free(wstr);
	return result;
}

int iPMI_KVS_DestroyW(const wchar_t kvsname[])
{
	int result;
	wchar_t str[1024];

	if (pmi_process.init_finalized == PMI_FINALIZED)
		return PMI_ERR_INIT;
	if (kvsname == NULL)
		return PMI_ERR_INVALID_ARG;

	if (pmi_process.local_kvs)
	{
		result = smpd_dbs_destroy(kvsname);
		return (result == SMPD_SUCCESS) ? PMI_SUCCESS : PMI_FAIL;
	}

	result = pmi_create_post_command(L"dbdestroy", kvsname, NULL, NULL);
	if (result != PMI_SUCCESS)
	{
		pmi_err_printf(L"PMI_KVS_Destroy failed: unable to destroy the pmi kvs space named '%s'.\n", kvsname);
		return PMI_FAIL;
	}

	/* parse the result of the command */
	if (MPIU_Str_get_string_arg(pmi_process.context->read_cmd.cmd, L"result", str, 1024) != MPIU_STR_SUCCESS)
	{
		pmi_err_printf(L"PMI_KVS_Destroy failed: no result string in the result command.\n");
		return PMI_FAIL;
	}
	if (wcscmp(str, DBS_SUCCESS_STR))
	{
		pmi_err_printf(L"PMI_KVS_Destroy failed: %s\n", str);
		return PMI_FAIL;
	}

	return PMI_SUCCESS;
}

int iPMI_KVS_Put(const char kvsname[], const char key[], const char value[])
{
	size_t len, num_converted;
	wchar_t *wkvsname = NULL, *wkey = NULL, *wvalue = NULL;
	int result;
	errno_t error;

	len = strlen(kvsname) + 1;
	wkvsname = (wchar_t*)MPIU_Malloc(len * sizeof(wchar_t));
	if (wkvsname == NULL)
		return PMI_ERR_NOMEM;
	error = mbstowcs_s(&num_converted, wkvsname, len, kvsname, _TRUNCATE);
	if (error != 0)
	{
		if (num_converted >= len)
		{
			/* truncation error */
			result = PMI_ERR_TRUNCATION;
			goto fn_fail;
		}
		result = PMI_FAIL;
		goto fn_fail;
	}

	len = strlen(key) + 1;
	wkey = (wchar_t*)MPIU_Malloc(len * sizeof(wchar_t));
	if (wkey == NULL)
	{
		result = PMI_ERR_NOMEM;
		goto fn_fail;
	}
	error = mbstowcs_s(&num_converted, wkey, len, key, _TRUNCATE);
	if (error != 0)
	{
		if (num_converted >= len)
		{
			/* truncation error */
			result = PMI_ERR_TRUNCATION;
		}
		else
			result = PMI_FAIL;
		goto fn_fail;
	}

	len = strlen(value) + 1;
	wvalue = (wchar_t*)MPIU_Malloc(len * sizeof(wchar_t));
	if (wvalue == NULL)
	{
		result = PMI_ERR_NOMEM;
		goto fn_fail;
	}
	error = mbstowcs_s(&num_converted, wvalue, len, value, _TRUNCATE);
	if (error != 0)
	{
		if (num_converted >= len)
		{
			/* truncation error */
			result = PMI_ERR_TRUNCATION;
		}
		else
			result = PMI_FAIL;
		goto fn_fail;
	}

	result = iPMI_KVS_PutW(wkvsname, wkey, wvalue);

fn_fail:
	if (wkvsname != NULL)
		MPIU_Free(wkvsname);
	if (wkey != NULL)
		MPIU_Free(wkey);
	if (wvalue != NULL)
		MPIU_Free(wvalue);
	return result;
}

int iPMI_KVS_PutW(const wchar_t kvsname[], const wchar_t key[], const wchar_t value[])
{
	int result;
	wchar_t str[1024];

	if (pmi_process.init_finalized == PMI_FINALIZED)
		return PMI_ERR_INIT;
	if (kvsname == NULL)
		return PMI_ERR_INVALID_ARG;
	if (key == NULL)
		return PMI_ERR_INVALID_KEY;
	if (value == NULL)
		return PMI_ERR_INVALID_VAL;

	/*wprintf(L"putting <%s><%s><%s>\n", kvsname, key, value);fflush(stdout);*/

	if (pmi_process.local_kvs)
	{
		result = smpd_dbs_put(kvsname, key, value);
		return (result == SMPD_SUCCESS) ? PMI_SUCCESS : PMI_FAIL;
	}

	result = pmi_create_post_command(L"dbput", kvsname, key, value);
	if (result != PMI_SUCCESS)
	{
		pmi_err_printf(L"PMI_KVS_Put failed: unable to put '%s:%s:%s'\n", kvsname, key, value);
		return PMI_FAIL;
	}

	/* parse the result of the command */
	if (MPIU_Str_get_string_arg(pmi_process.context->read_cmd.cmd, L"result", str, 1024) != MPIU_STR_SUCCESS)
	{
		pmi_err_printf(L"PMI_KVS_Put failed: no result string in the result command.\n");
		return PMI_FAIL;
	}
	if (wcscmp(str, DBS_SUCCESS_STR))
	{
		pmi_err_printf(L"PMI_KVS_Put failed: '%s'\n", str);
		return PMI_FAIL;
	}

	/*wprintf(L"iPMI_KVS_Put success.\n");fflush(stdout);*/
	return PMI_SUCCESS;
}

int iPMI_KVS_Commit(const char kvsname[])
{
	wchar_t *wkvsname;
	int result;

	result = convert_to_wcs(kvsname, &wkvsname);
	if (result == PMI_ERR_NOMEM)
		return result;

	result = iPMI_KVS_CommitW(wkvsname);
	MPIU_Free(wkvsname);
	return result;
}

int iPMI_KVS_CommitW(const wchar_t kvsname[])
{
	if (pmi_process.init_finalized == PMI_FINALIZED)
		return PMI_ERR_INIT;
	if (kvsname == NULL)
		return PMI_ERR_INVALID_ARG;

	if (pmi_process.local_kvs)
	{
		return PMI_SUCCESS;
	}

	/* Make the puts return when the commands are written but not acknowledged.
	   Then have this function wait until all outstanding puts are acknowledged.
	   */

	return PMI_SUCCESS;
}

int iPMI_KVS_Get(const char kvsname[], const char key[], char value[], int length)
{
	wchar_t *wkvsname, *wkey, *wvalue;
	int result;
	size_t num_converted;
	errno_t error;

	wvalue = (wchar_t*)MPIU_Malloc(length * sizeof(wchar_t));
	if (wvalue == NULL)
		return PMI_ERR_NOMEM;

	result = convert_to_wcs(kvsname, &wkvsname);
	if (result != PMI_SUCCESS)
		return result;

	result = convert_to_wcs(key, &wkey);
	if (result != PMI_SUCCESS)
	{
		MPIU_Free(wkvsname);
		return result;
	}

	result = iPMI_KVS_GetW(wkvsname, wkey, wvalue, length);
	if (result == PMI_SUCCESS)
	{
		value[0] = '\0';
		error = wcstombs_s(&num_converted, value, length, wvalue, _TRUNCATE);
		if (error != 0)
		{
			if (num_converted >= (size_t)length)
			{
				/* truncation error */
				result = PMI_ERR_TRUNCATION;
			}
			else
				result = PMI_FAIL;
		}
	}

	MPIU_Free(wkvsname);
	MPIU_Free(wkey);
	MPIU_Free(wvalue);
	return result;
}

int iPMI_KVS_GetW(const wchar_t kvsname[], const wchar_t key[], wchar_t value[], int length)
{
	int result;
	wchar_t str[1024];

	if (pmi_process.init_finalized == PMI_FINALIZED)
		return PMI_ERR_INIT;
	if (kvsname == NULL)
		return PMI_ERR_INVALID_ARG;
	if (key == NULL)
		return PMI_ERR_INVALID_KEY;
	if (value == NULL)
		return PMI_ERR_INVALID_VAL;

	if (pmi_process.local_kvs)
	{
		result = smpd_dbs_get(kvsname, key, value, length);
		return (result == SMPD_SUCCESS) ? PMI_SUCCESS : PMI_FAIL;
	}

	result = pmi_create_post_command(L"dbget", kvsname, key, NULL);
	if (result != PMI_SUCCESS)
	{
		pmi_err_printf(L"PMI_KVS_Get failed: unable to get '%s:%s'\n", kvsname, key);
		return PMI_FAIL;
	}

	/* parse the result of the command */
	if (MPIU_Str_get_string_arg(pmi_process.context->read_cmd.cmd, L"result", str, 1024) != MPIU_STR_SUCCESS)
	{
		pmi_err_printf(L"PMI_KVS_Get failed: no result string in the result command.\n");
		return PMI_FAIL;
	}
	if (wcscmp(str, DBS_SUCCESS_STR))
	{
		/* FIXME: If we are going to use pmi for the publish/lookup interface then gets should be allowed to fail without printing errors */
		pmi_err_printf(L"PMI_KVS_Get failed: '%s'\n", str);
		return PMI_FAIL;
	}
	if (MPIU_Str_get_string_arg(pmi_process.context->read_cmd.cmd, L"value", value, length) != MPIU_STR_SUCCESS)
	{
		pmi_err_printf(L"PMI_KVS_Get failed: no value in the result command for the get: '%s'\n", pmi_process.context->read_cmd.cmd);
		return PMI_FAIL;
	}

	/*
	wprintf(L"iPMI_KVS_Get success.\n");fflush(stdout);
	wprintf(L"get <%s><%s><%s>\n", kvsname, key, value);
	fflush(stdout);
	*/
	return PMI_SUCCESS;
}

int iPMI_KVS_Iter_first(const char kvsname[], char key[], int key_len, char value[], int val_len)
{
	wchar_t *wkvsname, *wkey, *wvalue;
	int result;
	size_t num_converted;
	errno_t error;

	result = convert_to_wcs(kvsname, &wkvsname);
	if (result != PMI_SUCCESS)
		return result;

	wkey = (wchar_t*)MPIU_Malloc(key_len * sizeof(wchar_t));
	if (wkey == NULL)
	{
		MPIU_Free(wkvsname);
		return PMI_ERR_NOMEM;
	}

	wvalue = (wchar_t*)MPIU_Malloc(val_len * sizeof(wchar_t));
	if (wvalue == NULL)
	{
		MPIU_Free(wkvsname);
		MPIU_Free(wkey);
		return PMI_ERR_NOMEM;
	}

	result = iPMI_KVS_Iter_firstW(wkvsname, wkey, key_len, wvalue, val_len);
	if (result == PMI_SUCCESS)
	{
		key[0] = '\0';
		error = wcstombs_s(&num_converted, key, key_len, wkey, _TRUNCATE);
		if (error != 0)
		{
			if (num_converted >= (size_t)key_len)
			{
				/* truncation error */
				result = PMI_ERR_TRUNCATION;
			}
			else
				result = PMI_FAIL;
		}
		value[0] = '\0';
		error = wcstombs_s(&num_converted, value, val_len, wvalue, _TRUNCATE);
		if (error != 0)
		{
			if (num_converted >= (size_t)val_len)
			{
				/* truncation error */
				result = PMI_ERR_TRUNCATION;
			}
			else
				result = PMI_FAIL;
		}
	}

	MPIU_Free(wkvsname);
	MPIU_Free(wkey);
	MPIU_Free(wvalue);

	return result;
}

int iPMI_KVS_Iter_firstW(const wchar_t kvsname[], wchar_t key[], int key_len, wchar_t value[], int val_len)
{
	int result;
	wchar_t str[1024];

	if (pmi_process.init_finalized == PMI_FINALIZED)
		return PMI_ERR_INIT;
	if (kvsname == NULL)
		return PMI_ERR_INVALID_ARG;
	if (key == NULL)
		return PMI_ERR_INVALID_KEY;
	if (key_len < PMI_MAX_KEY_LEN)
		return PMI_ERR_INVALID_KEY_LENGTH;
	if (value == NULL)
		return PMI_ERR_INVALID_VAL;
	if (val_len < PMI_MAX_VALUE_LEN)
		return PMI_ERR_INVALID_VAL_LENGTH;

	if (pmi_process.local_kvs)
	{
		result = smpd_dbs_first(kvsname, key, key_len, value, val_len);
		return (result == SMPD_SUCCESS) ? PMI_SUCCESS : PMI_FAIL;
	}

	result = pmi_create_post_command(L"dbfirst", kvsname, NULL, NULL);
	if (result != PMI_SUCCESS)
	{
		pmi_err_printf(L"PMI_KVS_Iter_first failed: unable to get the first key/value pair from '%s'\n", kvsname);
		return PMI_FAIL;
	}

	/* parse the result of the command */
	if (MPIU_Str_get_string_arg(pmi_process.context->read_cmd.cmd, L"result", str, 1024) != MPIU_STR_SUCCESS)
	{
		pmi_err_printf(L"PMI_KVS_Iter_first failed: no result string in the result command.\n");
		return PMI_FAIL;
	}
	if (wcscmp(str, DBS_SUCCESS_STR))
	{
		pmi_err_printf(L"PMI_KVS_Iter_first failed: %s\n", str);
		return PMI_FAIL;
	}
	if (MPIU_Str_get_string_arg(pmi_process.context->read_cmd.cmd, L"key", str, PMI_MAX_KEY_LEN) != MPIU_STR_SUCCESS)
	{
		pmi_err_printf(L"PMI_KVS_Iter_first failed: no key in the result command for the pmi iter_first: '%s'\n", pmi_process.context->read_cmd.cmd);
		return PMI_FAIL;
	}
	if (wcscmp(str, DBS_END_STR) == 0)
	{
		*key = L'\0';
		*value = L'\0';
		return PMI_SUCCESS;
	}
	wcscpy_s(key, key_len, str);
	if (MPIU_Str_get_string_arg(pmi_process.context->read_cmd.cmd, L"value", value, PMI_MAX_VALUE_LEN) != MPIU_STR_SUCCESS)
	{
		pmi_err_printf(L"PMI_KVS_Iter_first failed: no value in the result command for the pmi iter_first: '%s'\n", pmi_process.context->read_cmd.cmd);
		return PMI_FAIL;
	}

	return PMI_SUCCESS;
}

int iPMI_KVS_Iter_next(const char kvsname[], char key[], int key_len, char value[], int val_len)
{
	wchar_t *wkvsname, *wkey, *wvalue;
	int result;
	size_t num_converted;
	errno_t error;

	result = convert_to_wcs(kvsname, &wkvsname);
	if (result != PMI_SUCCESS)
		return result;

	wkey = (wchar_t*)MPIU_Malloc(key_len * sizeof(wchar_t));
	if (wkey == NULL)
	{
		MPIU_Free(wkvsname);
		return PMI_ERR_NOMEM;
	}

	wvalue = (wchar_t*)MPIU_Malloc(val_len * sizeof(wchar_t));
	if (wvalue == NULL)
	{
		MPIU_Free(wkvsname);
		MPIU_Free(wkey);
		return PMI_ERR_NOMEM;
	}

	result = iPMI_KVS_Iter_nextW(wkvsname, wkey, key_len, wvalue, val_len);
	if (result == PMI_SUCCESS)
	{
		key[0] = '\0';
		error = wcstombs_s(&num_converted, key, key_len, wkey, _TRUNCATE);
		if (error != 0)
		{
			if (num_converted >= (size_t)key_len)
			{
				/* truncation error */
				result = PMI_ERR_TRUNCATION;
			}
			else
				result = PMI_FAIL;
		}
		value[0] = '\0';
		error = wcstombs_s(&num_converted, value, val_len, wvalue, _TRUNCATE);
		if (error != 0)
		{
			if (num_converted >= (size_t)val_len)
			{
				/* truncation error */
				result = PMI_ERR_TRUNCATION;
			}
			else
				result = PMI_FAIL;
		}
	}

	MPIU_Free(wkvsname);
	MPIU_Free(wkey);
	MPIU_Free(wvalue);

	return result;
}

int iPMI_KVS_Iter_nextW(const wchar_t kvsname[], wchar_t key[], int key_len, wchar_t value[], int val_len)
{
	int result;
	wchar_t str[1024];

	if (pmi_process.init_finalized == PMI_FINALIZED)
		return PMI_ERR_INIT;
	if (kvsname == NULL)
		return PMI_ERR_INVALID_ARG;
	if (key == NULL)
		return PMI_ERR_INVALID_KEY;
	if (key_len < PMI_MAX_KEY_LEN)
		return PMI_ERR_INVALID_KEY_LENGTH;
	if (value == NULL)
		return PMI_ERR_INVALID_VAL;
	if (val_len < PMI_MAX_VALUE_LEN)
		return PMI_ERR_INVALID_VAL_LENGTH;

	if (pmi_process.local_kvs)
	{
		result = smpd_dbs_next(kvsname, key, key_len, value, val_len);
		return (result == SMPD_SUCCESS) ? PMI_SUCCESS : PMI_FAIL;
	}

	result = pmi_create_post_command(L"dbnext", kvsname, NULL, NULL);
	if (result != PMI_SUCCESS)
	{
		pmi_err_printf(L"PMI_KVS_Iter_next failed: unable to get the next key/value pair from '%s'\n", kvsname);
		return PMI_FAIL;
	}

	/* parse the result of the command */
	if (MPIU_Str_get_string_arg(pmi_process.context->read_cmd.cmd, L"result", str, 1024) != MPIU_STR_SUCCESS)
	{
		pmi_err_printf(L"PMI_KVS_Iter_next failed: no result string in the result command.\n");
		return PMI_FAIL;
	}
	if (wcscmp(str, DBS_SUCCESS_STR))
	{
		pmi_err_printf(L"PMI_KVS_Iter_next failed: %s\n", str);
		return PMI_FAIL;
	}
	if (MPIU_Str_get_string_arg(pmi_process.context->read_cmd.cmd, L"key", str, PMI_MAX_KEY_LEN) != MPIU_STR_SUCCESS)
	{
		pmi_err_printf(L"PMI_KVS_Iter_next failed: no key in the result command for the pmi iter_next: '%s'\n", pmi_process.context->read_cmd.cmd);
		return PMI_FAIL;
	}
	if (wcscmp(str, DBS_END_STR) == 0)
	{
		*key = L'\0';
		*value = L'\0';
		return PMI_SUCCESS;
	}
	wcscpy_s(key, key_len, str);
	if (MPIU_Str_get_string_arg(pmi_process.context->read_cmd.cmd, L"value", value, PMI_MAX_VALUE_LEN) != MPIU_STR_SUCCESS)
	{
		pmi_err_printf(L"PMI_KVS_Iter_next failed: no value in the result command for the pmi iter_next: '%s'\n", pmi_process.context->read_cmd.cmd);
		return PMI_FAIL;
	}

	return PMI_SUCCESS;
}

int iPMI_Spawn_multiple(int count,
					   const char * cmds[],
					   const char ** argvs[],
					   const int maxprocs[],
					   const int cinfo_keyval_sizes[],
					   const PMI_keyval_t * info_keyval_vectors[],
					   int preput_keyval_size,
					   const PMI_keyval_t preput_keyval_vector[],
					   int errors[],
					   int quiet)
{
	int i, j, k;
	int result = PMI_SUCCESS;
	wchar_t **wcmds = NULL;
	wchar_t ***wargvs = NULL;
	PMI_keyvalW_t **winfo_keyval_vectors = NULL;
	PMI_keyvalW_t *wpreput_keyval_vector = NULL;

	if (pmi_process.init_finalized == PMI_FINALIZED)
		return PMI_ERR_INIT;
	if (count < 1 || cmds == NULL || maxprocs == NULL || preput_keyval_size < 0)
		return PMI_ERR_INVALID_ARG;

	wcmds = (wchar_t**)MPIU_Malloc(count * sizeof(wchar_t*));
	if (wcmds == NULL)
	{
		result = PMI_ERR_NOMEM;
		goto fn_fail;
	}
	for (i=0; i<count; i++) wcmds[i] = NULL;
	for (i=0; i<count; i++)
	{
		if (cmds[i] != NULL)
		{
			result = convert_to_wcs(cmds[i], &wcmds[i]);
			if (result != PMI_SUCCESS)
				goto fn_fail;
		}
	}
	wargvs = (wchar_t***)MPIU_Malloc(count * sizeof(wchar_t**));
	if (wargvs == NULL)
	{
		result = PMI_ERR_NOMEM;
		goto fn_fail;
	}
	for (i=0; i<count; i++) wargvs[i] = NULL;
	for (i=0; i<count; i++)
	{
		if (argvs[i] != NULL)
		{
			const char **iter;
			j=1;
			iter = argvs[i];
			while (*iter)
			{
				j++;
				iter++;
			}
			wargvs[i] = (wchar_t**)MPIU_Malloc(j * sizeof(wchar_t*));
			if (wargvs[i] == NULL)
			{
				result = PMI_ERR_NOMEM;
				goto fn_fail;
			}
			for (k=0; k<j; k++) wargvs[i][k] = NULL;
			for (k=0; k<j; k++)
			{
				if (argvs[i][k] != NULL)
				{
					result = convert_to_wcs(argvs[i][k], &wargvs[i][k]);
					if (result != PMI_SUCCESS)
					{
						goto fn_fail;
					}
				}
			}
		}
	}
	if (info_keyval_vectors != NULL && cinfo_keyval_sizes != NULL)
	{
		winfo_keyval_vectors = (PMI_keyvalW_t**)MPIU_Malloc(count * sizeof(PMI_keyvalW_t*));
		if (winfo_keyval_vectors == NULL)
		{
			result = PMI_ERR_NOMEM;
			goto fn_fail;
		}
		for (i=0; i<count; i++) winfo_keyval_vectors[i] = NULL;
		for (i=0; i<count; i++)
		{
			if (cinfo_keyval_sizes[i] > 0)
			{
				winfo_keyval_vectors[i] = (PMI_keyvalW_t*)MPIU_Malloc(cinfo_keyval_sizes[i] * sizeof(PMI_keyvalW_t));
				if (winfo_keyval_vectors[i] == NULL)
				{
					result = PMI_ERR_NOMEM;
					goto fn_fail;
				}
				for (j=0; j<cinfo_keyval_sizes[i]; j++)
				{
					winfo_keyval_vectors[i][j].key = NULL;
					winfo_keyval_vectors[i][j].val = NULL;
				}
				for (j=0; j<cinfo_keyval_sizes[i]; j++)
				{
					result = convert_to_wcs(info_keyval_vectors[i][j].key, &winfo_keyval_vectors[i][j].key);
					if (result != PMI_SUCCESS)
						goto fn_fail;
					result = convert_to_wcs(info_keyval_vectors[i][j].val, &winfo_keyval_vectors[i][j].val);
					if (result != PMI_SUCCESS)
						goto fn_fail;
				}
			}
		}
	}
	if (preput_keyval_size > 0 && preput_keyval_vector != NULL)
	{
		wpreput_keyval_vector = (PMI_keyvalW_t*)MPIU_Malloc(preput_keyval_size * sizeof(PMI_keyvalW_t));
		if (wpreput_keyval_vector == NULL)
		{
			result = PMI_ERR_NOMEM;
			goto fn_fail;
		}
		for (i=0; i<preput_keyval_size; i++)
		{
			wpreput_keyval_vector[i].key = NULL;
			wpreput_keyval_vector[i].val = NULL;
		}
		for (i=0; i<preput_keyval_size; i++)
		{
			result = convert_to_wcs(preput_keyval_vector[i].key, &wpreput_keyval_vector[i].key);
			if (result != PMI_SUCCESS)
				goto fn_fail;
			result = convert_to_wcs(preput_keyval_vector[i].val, &wpreput_keyval_vector[i].val);
			if (result != PMI_SUCCESS)
				goto fn_fail;
		}
	}

	result = iPMI_Spawn_multipleW(count, wcmds, wargvs, maxprocs, cinfo_keyval_sizes, winfo_keyval_vectors, preput_keyval_size, wpreput_keyval_vector, errors, quiet);

fn_fail:
	if (wcmds != NULL)
	{
		for (i=0; i<count; i++)
		{
			if (wcmds[i] != NULL)
				MPIU_Free(wcmds[i]);
		}
		MPIU_Free(wcmds);
	}
	if (wargvs != NULL)
	{
		for (i=0; i<count; i++)
		{
			if (wargvs[i] != NULL)
			{
				wchar_t **iter = wargvs[i];
				while (*iter != NULL)
				{
					MPIU_Free(*iter);
					iter++;
				}
			}
			MPIU_Free(wargvs[i]);
		}
		MPIU_Free(wargvs);
	}
	if (winfo_keyval_vectors != NULL)
	{
		 for (i=0; i<count; i++)
		 {
			 if (winfo_keyval_vectors[i] != NULL && cinfo_keyval_sizes[i] > 0)
			 {
				 for (j=0; j<cinfo_keyval_sizes[i]; j++)
				 {
					 if (winfo_keyval_vectors[i][j].key != NULL)
					 {
						 MPIU_Free(winfo_keyval_vectors[i][j].key);
					 }
					 if (winfo_keyval_vectors[i][j].val != NULL)
					 {
						 MPIU_Free(winfo_keyval_vectors[i][j].val);
					 }
				 }
				 MPIU_Free(winfo_keyval_vectors[i]);
			 }
		 }
		 MPIU_Free(winfo_keyval_vectors);
	}
	if (wpreput_keyval_vector != NULL && preput_keyval_size > 0)
	{
		for (i=0; i<preput_keyval_size; i++)
		{
			if (wpreput_keyval_vector[i].key != NULL)
			{
				MPIU_Free(wpreput_keyval_vector[i].key);
			}
			if (wpreput_keyval_vector[i].val != NULL)
			{
				MPIU_Free(wpreput_keyval_vector[i].val);
			}
		}
		MPIU_Free(wpreput_keyval_vector);
	}
	return result;
}

int iPMI_Spawn_multipleW(int count,
					   const wchar_t * cmds[],
					   const wchar_t ** argvs[],
					   const int maxprocs[],
					   const int cinfo_keyval_sizes[],
					   const PMI_keyvalW_t * info_keyval_vectors[],
					   int preput_keyval_size,
					   const PMI_keyvalW_t preput_keyval_vector[],
					   int errors[],
					   int quiet)
{
	int result;
	smpd_command_t *cmd_ptr;
	int dest = 0;
	wchar_t buffer[SMPD_MAX_CMD_LENGTH];
	wchar_t keyval_buf[SMPD_MAX_CMD_LENGTH];
	wchar_t key[100];
	wchar_t *iter, *iter2;
	int i, j, maxlen, maxlen2;
	int path_specified = 0;
	wchar_t path[SMPD_MAX_PATH_LENGTH] = L"";
	int *info_keyval_sizes;
	int total_num_processes;

	MPIU_UNREFERENCED_ARG(quiet);

	if (pmi_process.init_finalized == PMI_FINALIZED)
		return PMI_ERR_INIT;
	if (count < 1 || cmds == NULL || maxprocs == NULL || preput_keyval_size < 0)
		return PMI_ERR_INVALID_ARG;

	if (pmi_process.local_kvs)
	{
		return PMI_FAIL;
	}

	/*wprintf(L"creating spawn command.\n");fflush(stdout);*/
	result = smpd_create_command(L"spawn", pmi_process.smpd_id, dest, SMPD_TRUE, &cmd_ptr);
	if (result != SMPD_SUCCESS)
	{
		pmi_err_printf(L"unable to create a spawn command.\n");
		return PMI_FAIL;
	}
	result = smpd_add_command_int_arg(cmd_ptr, L"ctx_key", pmi_process.smpd_key);
	if (result != SMPD_SUCCESS)
	{
		pmi_err_printf(L"unable to add the key to the spawn command.\n");
		return PMI_FAIL;
	}

	/* add the number of commands */
	result = smpd_add_command_int_arg(cmd_ptr, L"ncmds", count);
	if (result != SMPD_SUCCESS)
	{
		pmi_err_printf(L"unable to add the ncmds field to the spawn command.\n");
		return PMI_FAIL;
	}
	/* add the commands and their argv arrays */
	for (i=0; i<count; i++)
	{
		swprintf_s(key, 100, L"cmd%d", i);
		if (wcslen(cmds[i]) > 2)
		{
			if (cmds[i][0] == L'.' && cmds[i][1] == L'/')
			{
				result = smpd_add_command_arg(cmd_ptr, key, (wchar_t*)&cmds[i][2]);
			}
			else
			{
				result = smpd_add_command_arg(cmd_ptr, key, (wchar_t*)cmds[i]);
			}
		}
		else
		{
			result = smpd_add_command_arg(cmd_ptr, key, (wchar_t*)cmds[i]);
		}
		if (result != SMPD_SUCCESS)
		{
			pmi_err_printf(L"unable to add %s(%s) to the spawn command.\n", key, cmds[i]);
			return PMI_FAIL;
		}
		if (argvs)
		{
			buffer[0] = L'\0';
			iter = buffer;
			maxlen = SMPD_MAX_CMD_LENGTH;
			if (argvs[i] != NULL)
			{
				for (j=0; argvs[i][j] != NULL; j++)
				{
					result = MPIU_Str_add_string(&iter, &maxlen, argvs[i][j]);
				}
				if (iter > buffer)
				{
					iter--;
					*iter = L'\0'; /* erase the trailing space */
				}
			}
			swprintf_s(key, 100, L"argv%d", i);
			result = smpd_add_command_arg(cmd_ptr, key, buffer);
			if (result != SMPD_SUCCESS)
			{
				pmi_err_printf(L"unable to add %s(%s) to the spawn command.\n", key, buffer);
				return PMI_FAIL;
			}
		}
	}
	/* add the maxprocs array and count the total number of processes */
	total_num_processes = 0;
	buffer[0] = L'\0';
	for (i=0; i<count; i++)
	{
		total_num_processes += maxprocs[i];
		if (i < count-1)
			swprintf_s(key, 100, L"%d ", maxprocs[i]);
		else
			swprintf_s(key, 100, L"%d", maxprocs[i]);
		wcscat_s(buffer, SMPD_MAX_CMD_LENGTH, key);
	}
	result = smpd_add_command_arg(cmd_ptr, L"maxprocs", buffer);
	if (result != SMPD_SUCCESS)
	{
		pmi_err_printf(L"unable to add maxprocs(%s) to the spawn command.\n", buffer);
		return PMI_FAIL;
	}

	{
		HMODULE hModule;
		wchar_t exe_path[SMPD_MAX_PATH_LENGTH];
		wchar_t *iter;
		int length;

		GetCurrentDirectory(SMPD_MAX_PATH_LENGTH, path);
		hModule = GetModuleHandle(NULL);
		if (GetModuleFileName(hModule, exe_path, SMPD_MAX_PATH_LENGTH)) 
		{
			iter = wcsrchr(exe_path, L'\\');
			if (iter != NULL)
			{
				if (iter == (exe_path + 2) && *(iter-1) == L':')
				{
					/* leave the \ if the path is at the root, like c:\foo.exe */
					iter++;
				}
				*iter = L'\0'; /* erase the file name leaving only the path */
			}
			length = (int)wcslen(path);
			iter = &path[length];
			MPIU_Snprintf(iter, SMPD_MAX_PATH_LENGTH-length, _TRUNCATE, L";%s", exe_path);
		}
	}

	/* create a copy of the sizes so we can change the values locally */
	info_keyval_sizes = (int*)malloc(count * sizeof(int));
	if (info_keyval_sizes == NULL)
	{
		pmi_err_printf(L"unable to allocate an array of kevval sizes.\n");
		return PMI_FAIL;
	}
	for (i=0; i<count; i++)
	{
		info_keyval_sizes[i] = cinfo_keyval_sizes[i];
	}

	/* add the keyvals */
	if (info_keyval_sizes && info_keyval_vectors)
	{
		for (i=0; i<count; i++)
		{
			path_specified = 0;
			buffer[0] = L'\0';
			iter = buffer;
			maxlen = SMPD_MAX_CMD_LENGTH;
			for (j=0; j<info_keyval_sizes[i]; j++)
			{
				keyval_buf[0] = L'\0';
				iter2 = keyval_buf;
				maxlen2 = SMPD_MAX_CMD_LENGTH;
				if (wcscmp(info_keyval_vectors[i][j].key, L"path") == 0)
				{
					size_t val2len;
					wchar_t *val2;
					val2len = sizeof(wchar_t) * (wcslen(info_keyval_vectors[i][j].val) + 1 + wcslen(path) + 1);
					val2 = (wchar_t*)MPIU_Malloc(val2len);
					if (val2 == NULL)
					{
						pmi_err_printf(L"unable to allocate memory for the path key.\n");
						return PMI_FAIL;
					}
					/*wprintf(L"creating path %d: <%s>;<%s>\n", val2len, info_keyval_vectors[i][j].val, path);fflush(stdout);*/
					MPIU_Snprintf(val2, val2len / sizeof(wchar_t), _TRUNCATE, L"%s;%s", info_keyval_vectors[i][j].val, path);
					result = MPIU_Str_add_string_arg(&iter2, &maxlen2, info_keyval_vectors[i][j].key, val2);
					if (result != MPIU_STR_SUCCESS)
					{
						pmi_err_printf(L"unable to add %s=%s to the spawn command.\n", info_keyval_vectors[i][j].key, val2);
						MPIU_Free(val2);
						return PMI_FAIL;
					}
					MPIU_Free(val2);
					path_specified = 1;
				}
				else
				{
					result = MPIU_Str_add_string_arg(&iter2, &maxlen2, info_keyval_vectors[i][j].key, info_keyval_vectors[i][j].val);
					if (result != MPIU_STR_SUCCESS)
					{
						pmi_err_printf(L"unable to add %s=%s to the spawn command.\n", info_keyval_vectors[i][j].key, info_keyval_vectors[i][j].val);
						return PMI_FAIL;
					}
				}
				if (iter2 > keyval_buf)
				{
					iter2--;
					*iter2 = L'\0'; /* remove the trailing space */
				}
				swprintf_s(key, 100, L"%d", j);
				result = MPIU_Str_add_string_arg(&iter, &maxlen, key, keyval_buf);
				if (result != MPIU_STR_SUCCESS)
				{
					pmi_err_printf(L"unable to add %s=%s to the spawn command.\n", key, keyval_buf);
					return PMI_FAIL;
				}
			}
			/* add the current directory as the default path if a path has not been specified */
			if (!path_specified)
			{
				keyval_buf[0] = L'\0';
				iter2 = keyval_buf;
				maxlen2 = SMPD_MAX_CMD_LENGTH;
				result = MPIU_Str_add_string_arg(&iter2, &maxlen2, L"path", path);
				iter2--;
				*iter2 = L'\0';
				swprintf_s(key, 100, L"%d", j);
				result = MPIU_Str_add_string_arg(&iter, &maxlen, key, keyval_buf);
				if (result != MPIU_STR_SUCCESS)
				{
					pmi_err_printf(L"unable to add %s=%s to the spawn command.\n", key, keyval_buf);
					return PMI_FAIL;
				}
				info_keyval_sizes[i]++;
			}
			if (iter != buffer)
			{
				iter--;
				*iter = L'\0'; /* remove the trailing space */
			}
			swprintf_s(key, 100, L"keyvals%d", i);
			result = smpd_add_command_arg(cmd_ptr, key, buffer);
			if (result != SMPD_SUCCESS)
			{
				pmi_err_printf(L"unable to add %s(%s) to the spawn command.\n", key, buffer);
				return PMI_FAIL;
			}
		}
	}
	else
	{
		if (!info_keyval_sizes)
		{
			buffer[0] = L'\0';
			for (i=0; i<count; i++)
			{
				if (i < count-1)
					wcscat_s(buffer, SMPD_MAX_CMD_LENGTH, L"1 ");
				else
					wcscat_s(buffer, SMPD_MAX_CMD_LENGTH, L"1");
			}
			result = smpd_add_command_arg(cmd_ptr, L"nkeyvals", buffer);
			if (result != SMPD_SUCCESS)
			{
				pmi_err_printf(L"unable to add nkeyvals(%s) to the spawn command.\n", buffer);
				return PMI_FAIL;
			}
		}
		for (i=0; i<count; i++)
		{
			buffer[0] = L'\0';
			iter = buffer;
			maxlen = SMPD_MAX_CMD_LENGTH;
			/* add the current directory as the default path if a path has not been specified */
			keyval_buf[0] = L'\0';
			iter2 = keyval_buf;
			maxlen2 = SMPD_MAX_CMD_LENGTH;
			result = MPIU_Str_add_string_arg(&iter2, &maxlen2, L"path", path);
			iter2--;
			*iter2 = L'\0';
			wcscpy_s(key, 100, L"0");
			result = MPIU_Str_add_string_arg(&iter, &maxlen, key, keyval_buf);
			if (result != MPIU_STR_SUCCESS)
			{
				pmi_err_printf(L"unable to add %s=%s to the spawn command.\n", key, keyval_buf);
				return PMI_FAIL;
			}
			swprintf_s(key, 100, L"keyvals%d", i);
			result = smpd_add_command_arg(cmd_ptr, key, buffer);
			if (result != SMPD_SUCCESS)
			{
				pmi_err_printf(L"unable to add %s(%s) to the spawn command.\n", key, buffer);
				return PMI_FAIL;
			}
		}
	}

	/* add the keyval sizes array */
	if (info_keyval_sizes)
	{
		buffer[0] = L'\0';
		for (i=0; i<count; i++)
		{
			if (i < count-1)
				swprintf_s(key, 100, L"%d ", info_keyval_sizes[i] > 0 ? info_keyval_sizes[i] : 1);
			else
				swprintf_s(key, 100, L"%d", info_keyval_sizes[i] > 0 ? info_keyval_sizes[i] : 1);
			wcscat_s(buffer, SMPD_MAX_CMD_LENGTH, key);
		}
		result = smpd_add_command_arg(cmd_ptr, L"nkeyvals", buffer);
		if (result != SMPD_SUCCESS)
		{
			pmi_err_printf(L"unable to add nkeyvals(%s) to the spawn command.\n", buffer);
			return PMI_FAIL;
		}
	}

	free(info_keyval_sizes);

	/* add the pre-put keyvals */
	result = smpd_add_command_int_arg(cmd_ptr, L"npreput", preput_keyval_size);
	if (result != SMPD_SUCCESS)
	{
		pmi_err_printf(L"unable to add npreput=%d to the spawn command.\n", preput_keyval_size);
		return PMI_FAIL;
	}
	if (preput_keyval_size > 0 && preput_keyval_vector)
	{
		buffer[0] = L'\0';
		iter = buffer;
		maxlen = SMPD_MAX_CMD_LENGTH;
		for (i=0; i<preput_keyval_size; i++)
		{
			keyval_buf[0] = L'\0';
			iter2 = keyval_buf;
			maxlen2 = SMPD_MAX_CMD_LENGTH;
			result = MPIU_Str_add_string_arg(&iter2, &maxlen2, preput_keyval_vector[i].key, preput_keyval_vector[i].val);
			if (result != MPIU_STR_SUCCESS)
			{
				pmi_err_printf(L"unable to add %s=%s to the spawn command.\n", preput_keyval_vector[i].key, preput_keyval_vector[i].val);
				return PMI_FAIL;
			}
			if (iter2 > keyval_buf)
			{
				iter2--;
				*iter2 = L'\0'; /* remove the trailing space */
			}
			swprintf_s(key, 100, L"%d", i);
			result = MPIU_Str_add_string_arg(&iter, &maxlen, key, keyval_buf);
			if (result != MPIU_STR_SUCCESS)
			{
				pmi_err_printf(L"unable to add %s=%s to the spawn command.\n", key, keyval_buf);
				return PMI_FAIL;
			}
		}
		result = smpd_add_command_arg(cmd_ptr, L"preput", buffer);
		if (result != SMPD_SUCCESS)
		{
			pmi_err_printf(L"unable to add preput(%s) to the spawn command.\n", buffer);
			return PMI_FAIL;
		}
	}		

	/*wprintf(L"spawn command: <%s>\n", cmd_ptr->cmd);*/

	/* post the write of the command */
	/*
	wprintf(L"posting write of spawn command to %s context, sock %d: '%s'\n",
		smpd_get_context_str(pmi_process.context), MPIDU_Sock_get_sock_id(pmi_process.context->sock), cmd_ptr->cmd);
	fflush(stdout);
	*/
	result = smpd_post_write_command(pmi_process.context, cmd_ptr);
	if (result != SMPD_SUCCESS)
	{
		pmi_err_printf(L"unable to post a write of the spawn command.\n");
		return PMI_FAIL;
	}

	/* post a read for the result*/
	result = smpd_post_read_command(pmi_process.context);
	if (result != SMPD_SUCCESS)
	{
		pmi_err_printf(L"unable to post a read of the next command on the pmi context.\n");
		return PMI_FAIL;
	}

	/* let the state machine send the command and receive the result */
	result = smpd_enter_at_state(pmi_process.set, SMPD_WRITING_CMD);
	if (result != SMPD_SUCCESS)
	{
		/*wprintf(L"PMI_Spawn_multiple returning failure.\n");fflush(stdout);*/
		pmi_err_printf(L"the state machine logic failed to get the result of the spawn command.\n");
		return PMI_FAIL;
	}

	for (i=0; i<total_num_processes; i++)
	{
		errors[i] = PMI_SUCCESS;
	}
	/*wprintf(L"PMI_Spawn_multiple returning success.\n");fflush(stdout);*/
	return PMI_SUCCESS;
}

int iPMI_Parse_option(int num_args, char *args[], int *num_parsed, PMI_keyval_t **keyvalp, int *size)
{
	if (num_args < 1)
		return PMI_ERR_INVALID_NUM_ARGS;
	if (args == NULL)
		return PMI_ERR_INVALID_ARGS;
	if (num_parsed == NULL)
		return PMI_ERR_INVALID_NUM_PARSED;
	if (keyvalp == NULL)
		return PMI_ERR_INVALID_KEYVALP;
	if (size == NULL)
		return PMI_ERR_INVALID_SIZE;
	*num_parsed = 0;
	*keyvalp = NULL;
	*size = 0;
	return PMI_SUCCESS;
}

int iPMI_Args_to_keyval(int *argcp, char *((*argvp)[]), PMI_keyval_t **keyvalp, int *size)
{
	if (argcp == NULL || argvp == NULL || keyvalp == NULL || size == NULL)
		return PMI_ERR_INVALID_ARG;
	return PMI_SUCCESS;
}

int iPMI_Free_keyvals(PMI_keyval_t keyvalp[], int size)
{
	if (keyvalp == NULL || size < 0)
		return PMI_ERR_INVALID_ARG;
	if (size == 0)
		return PMI_SUCCESS;
	/* free stuff */
	return PMI_SUCCESS;
}

static wchar_t * namepub_kvs = NULL;
static int setup_name_service()
{
	int result;
	wchar_t pmi_namepub_kvs[PMI_MAX_KVS_NAME_LENGTH];

	if (namepub_kvs != NULL)
	{
		/* FIXME: Should it be an error to call setup_name_service twice? */
		MPIU_Free(namepub_kvs);
	}

	namepub_kvs = (wchar_t*)MPIU_Malloc(PMI_MAX_KVS_NAME_LENGTH);
	if (!namepub_kvs)
	{
		pmi_err_printf(L"unable to allocate memory for the name publisher kvs.\n");
		return PMI_FAIL;
	}

	if (GetEnvironmentVariable(L"PMI_NAMEPUB_KVS", pmi_namepub_kvs, PMI_MAX_KVS_NAME_LENGTH) > 0)
	{
		MPIU_Wcsncpy(namepub_kvs, pmi_namepub_kvs, PMI_MAX_KVS_NAME_LENGTH);
	}
	else
	{
		/*result = PMI_KVS_Create(namepub_kvs, PMI_MAX_KVS_NAME_LENGTH);*/
		result = iPMI_Get_kvs_domain_idW(namepub_kvs, PMI_MAX_KVS_NAME_LENGTH);
		if (result != PMI_SUCCESS)
		{
			pmi_err_printf(L"unable to get the name publisher kvs name.\n");
			return result;
		}
	}

	/*wprintf(L"namepub kvs: <%s>\n", namepub_kvs);fflush(stdout);*/
	return PMI_SUCCESS;
}

int iPMI_Publish_name( const char service_name[], const char port[] )
{
	wchar_t *wservice_name, *wport;
	int result;

	result = convert_to_wcs(service_name, &wservice_name);
	if (result != PMI_SUCCESS)
		return result;

	result = convert_to_wcs(port, &wport);
	if (result != PMI_SUCCESS)
	{
		MPIU_Free(wservice_name);
		return result;
	}

	result = iPMI_Publish_nameW(wservice_name, wport);
	MPIU_Free(wservice_name);
	MPIU_Free(wport);
	return result;
}

int iPMI_Publish_nameW( const wchar_t service_name[], const wchar_t port[] )
{
	int result;
	if (service_name == NULL || port == NULL)
		return PMI_ERR_INVALID_ARG;
	if (namepub_kvs == NULL)
	{
		result = setup_name_service();
		if (result != PMI_SUCCESS)
			return result;
	}
	/*wprintf(L"publish kvs: <%s>\n", namepub_kvs);fflush(stdout);*/
	result = iPMI_KVS_PutW(namepub_kvs, service_name, port);
	if (result != PMI_SUCCESS)
	{
		pmi_err_printf(L"unable to put the service name and port into the name publisher kvs.\n");
		return result;
	}
	result = iPMI_KVS_CommitW(namepub_kvs);
	if (result != PMI_SUCCESS)
	{
		pmi_err_printf(L"unable to commit the name publisher kvs.\n");
		return result;
	}
	return PMI_SUCCESS;
}

int iPMI_Unpublish_name( const char service_name[] )
{
	wchar_t *wservice_name;
	int result;

	result = convert_to_wcs(service_name, &wservice_name);
	if (result != PMI_SUCCESS)
		return result;

	result = iPMI_Unpublish_nameW(wservice_name);
	MPIU_Free(wservice_name);
	return result;
}

int iPMI_Unpublish_nameW( const wchar_t service_name[] )
{
	int result;
	if (service_name == NULL)
		return PMI_ERR_INVALID_ARG;
	if (namepub_kvs == NULL)
	{
		result = setup_name_service();
		if (result != PMI_SUCCESS)
			return result;
	}
	/*wprintf(L"unpublish kvs: <%s>\n", namepub_kvs);fflush(stdout);*/
	/* This assumes you can put the same key more than once which breaks the PMI specification */
	result = iPMI_KVS_PutW(namepub_kvs, service_name, L"");
	if (result != PMI_SUCCESS)
	{
		pmi_err_printf(L"unable to put the blank service name and port into the name publisher kvs.\n");
		return result;
	}
	result = iPMI_KVS_CommitW(namepub_kvs);
	if (result != PMI_SUCCESS)
	{
		pmi_err_printf(L"unable to commit the name publisher kvs.\n");
		return result;
	}
	return PMI_SUCCESS;
}

int iPMI_Lookup_name( const char service_name[], char port[] )
{
	wchar_t *wservice_name;
	wchar_t wport[MPI_MAX_PORT_NAME];
	int result;
	size_t num_converted;
	errno_t error;

	result = convert_to_wcs(service_name, &wservice_name);
	if (result != PMI_SUCCESS)
		return result;

	result = iPMI_Lookup_nameW(wservice_name, wport);
	if (result == PMI_SUCCESS)
	{
		port[0] = '\0';
		error = wcstombs_s(&num_converted, port, MPI_MAX_PORT_NAME, wport, _TRUNCATE);
		if (error != 0)
		{
			if (num_converted >= MPI_MAX_PORT_NAME)
			{
				/* truncation error */
				result = PMI_ERR_TRUNCATION;
			}
			else
				result = PMI_FAIL;
		}
	}
	return result;
}

int iPMI_Lookup_nameW( const wchar_t service_name[], wchar_t port[] )
{
	int result;
	if (service_name == NULL || port == NULL)
		return PMI_ERR_INVALID_ARG;
	if (namepub_kvs == NULL)
	{
		result = setup_name_service();
		if (result != PMI_SUCCESS)
			return result;
	}
	/*wprintf(L"lookup kvs: <%s>\n", namepub_kvs);fflush(stdout);*/
	silence = 1;
	result = iPMI_KVS_GetW(namepub_kvs, service_name, port, MPI_MAX_PORT_NAME);
	silence = 0;
	if (result != PMI_SUCCESS)
	{
		/* fail silently */
		/*pmi_err_printf(L"unable to get the service name and port from the name publisher kvs.\n");*/
		return result;
	}

	if (port[0] == L'\0')
	{
		return MPI_ERR_NAME;
	}
	return PMI_SUCCESS;
}

int PMIX_Start_root_smpd(int nproc, wchar_t *host, int len, int *port)
{
	DWORD dwLength = len;

	pmi_process.nproc = nproc;

	pmi_process.hRootThreadReadyEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
	if (pmi_process.hRootThreadReadyEvent == NULL)
	{
		pmi_err_printf(L"unable to create the root listener synchronization event, error: %d\n", GetLastError());
		return PMI_FAIL;
	}
	pmi_process.hRootThread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)root_smpd, NULL, 0, NULL);
	if (pmi_process.hRootThread == NULL)
	{
		pmi_err_printf(L"unable to create the root listener thread: error %d\n", GetLastError());
		return PMI_FAIL;
	}
	if (WaitForSingleObject(pmi_process.hRootThreadReadyEvent, 60000) != WAIT_OBJECT_0)
	{
		pmi_err_printf(L"the root process thread failed to initialize.\n");
		return PMI_FAIL;
	}
	/*GetComputerName(host, &dwLength);*/
	GetComputerNameEx(ComputerNameDnsFullyQualified, host, &dwLength);

	*port = pmi_process.root_port;

	return PMI_SUCCESS;
}

int PMIX_Stop_root_smpd()
{
	DWORD result;

	result = WaitForSingleObject(pmi_process.hRootThread, INFINITE);
	if (result != WAIT_OBJECT_0)
	{
		return PMI_FAIL;
	}
	return PMI_SUCCESS;
}

static int root_smpd(void *p)
{
	int result;
	MPIDU_Sock_set_t set;
	MPIDU_Sock_t listener;
	smpd_process_group_t *pg, **pg_iter;
	int i;

	/* unreferenced parameter */
	SMPD_UNREFERENCED_ARG(p);

	smpd_process.id = 1;
	smpd_process.root_smpd = SMPD_FALSE;
	smpd_process.map0to1 = SMPD_TRUE;

	result = MPIDU_Sock_create_set(&set);
	if (result != MPI_SUCCESS)
	{
		pmi_mpi_err_printf(result, L"MPIDU_Sock_create_set failed.\n");
		return PMI_FAIL;
	}
	smpd_process.set = set;
	smpd_dbg_printf(L"created a set for the listener: %d\n", MPIDU_Sock_get_sock_set_id(set));
	result = MPIDU_Sock_listen(set, NULL, &pmi_process.root_port, &listener); 
	if (result != MPI_SUCCESS)
	{
		pmi_mpi_err_printf(result, L"MPIDU_Sock_listen failed.\n");
		return PMI_FAIL;
	}
	smpd_dbg_printf(L"smpd listening on port %d\n", pmi_process.root_port);

	result = smpd_create_context(SMPD_CONTEXT_LISTENER, set, listener, -1, &smpd_process.listener_context);
	if (result != SMPD_SUCCESS)
	{
		pmi_err_printf(L"unable to create a context for the smpd listener.\n");
		return PMI_FAIL;
	}
	result = MPIDU_Sock_set_user_ptr(listener, smpd_process.listener_context);
	if (result != MPI_SUCCESS)
	{
		pmi_mpi_err_printf(result, L"MPIDU_Sock_set_user_ptr failed.\n");
		return PMI_FAIL;
	}
	smpd_process.listener_context->state = SMPD_SMPD_LISTENING;

	smpd_dbs_init();
	smpd_process.have_dbs = SMPD_TRUE;
	if (smpd_process.kvs_name[0] != L'\0')
	{
		result = smpd_dbs_create_name_in(smpd_process.kvs_name);
	}
	else
	{
		result = smpd_dbs_create(smpd_process.kvs_name, SMPD_MAX_DBS_NAME_LEN);
	}
	if (result != SMPD_DBS_SUCCESS)
	{
		pmi_err_printf(L"unable to create a kvs database: name = <%s>.\n", smpd_process.kvs_name);
		return PMI_FAIL;
	}

	/* Set up the process group */
	/* initialize a new process group structure */
	pg = (smpd_process_group_t*)malloc(sizeof(smpd_process_group_t));
	if (pg == NULL)
	{
		pmi_err_printf(L"unable to allocate memory for a process group structure.\n");
		return PMI_FAIL;
	}
	pg->aborted = SMPD_FALSE;
	pg->any_init_received = SMPD_FALSE;
	pg->any_noinit_process_exited = SMPD_FALSE;
	MPIU_Wcsncpy(pg->kvs, smpd_process.kvs_name, SMPD_MAX_DBS_NAME_LEN);
	pg->num_procs = pmi_process.nproc;
	pg->processes = (smpd_exit_process_t*)malloc(pmi_process.nproc * sizeof(smpd_exit_process_t));
	if (pg->processes == NULL)
	{
		pmi_err_printf(L"unable to allocate an array of %d process exit structures.\n", pmi_process.nproc);
		return PMI_FAIL;
	}
	for (i=0; i<pmi_process.nproc; i++)
	{
		pg->processes[i].ctx_key[0] = L'\0';
		pg->processes[i].errmsg = NULL;
		pg->processes[i].exitcode = -1;
		pg->processes[i].exited = SMPD_FALSE;
		pg->processes[i].finalize_called = SMPD_FALSE;
		pg->processes[i].init_called = SMPD_FALSE;
		pg->processes[i].node_id = i+1;
		pg->processes[i].host[0] = L'\0';
		pg->processes[i].suspended = SMPD_FALSE;
		pg->processes[i].suspend_cmd = NULL;
	}
	/* add the process group to the end of the global list */
	pg->next = NULL;
	pg_iter = &smpd_process.pg_list;
	while (*pg_iter != NULL)
		pg_iter = &(*pg_iter)->next;
	*pg_iter = pg;
	/* add to the beginning of the list
	pg->next = smpd_process.pg_list;
	smpd_process.pg_list = pg;
	*/

	SetEvent(pmi_process.hRootThreadReadyEvent);

	result = smpd_enter_at_state(set, SMPD_SMPD_LISTENING);
	if (result != SMPD_SUCCESS)
	{
		pmi_err_printf(L"root_smpd state machine failed.\n");
		return PMI_FAIL;
	}

	result = MPIDU_Sock_destroy_set(set);
	if (result != MPI_SUCCESS)
	{
		pmi_mpi_err_printf(result, L"unable to destroy the set.\n");
	}

	return PMI_SUCCESS;
}
