/*   Copyright (C) 2009 Deino Software. */

#include "ipmi.h"
#ifdef HAVE_MACH_O_DYLD_H
#include <mach-o/dyld.h>
#endif

#define PMILoadLibrary(a,b) a = LoadLibrary( b )
#define PMIGetProcAddress GetProcAddress
#define PMIModule HMODULE

/* global variables */
static ipmi_functions_t fn =
{
	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL
};

int PMI_Init(int *spawned)
{
#define MAX_ENV_STR_LEN 8192
	wchar_t dll_name[MAX_ENV_STR_LEN] = L"";
	PMIModule hModule;

	__try
	{
	if (fn.PMI_Init != NULL)
	{
		/* Init cannot be called more than once? */
		return PMI_FAIL;
	}

	/* Allow the user to choose a PMI dll name or none. */
	/* If no choice is made then the default of DeinoPMI.dll is used. */
	/* If the user defined or default PMI dll cannot be loaded then revert to SMPD. */
	if (GetEnvironmentVariable(L"PMI_WDLL_NAME", dll_name, MAX_ENV_STR_LEN) > 0)
	{
		if (_wcsicmp(dll_name, L"none") == 0)
			dll_name[0] = L'\0';
	}
	else
	{
		wcscpy_s(dll_name, MAX_ENV_STR_LEN, L"DeinoPMI.dll");
	}
	if (dll_name[0] != '\0')
	{
		PMILoadLibrary(hModule, dll_name);
		if (hModule != NULL)
		{
			fn.PMI_Init = (int (*)(int *))PMIGetProcAddress(hModule, "PMI_Init");
			if (fn.PMI_Init == NULL)
			{
				return PMI_FAIL;
			}
			fn.PMI_Initialized = (int (*)(PMI_BOOL *))PMIGetProcAddress(hModule, "PMI_Initialized");
			fn.PMI_Finalize = (int (*)(void))PMIGetProcAddress(hModule, "PMI_Finalize");
			fn.PMI_Get_size = (int (*)(int *))PMIGetProcAddress(hModule, "PMI_Get_size");
			fn.PMI_Get_rank = (int (*)(int *))PMIGetProcAddress(hModule, "PMI_Get_rank");
			fn.PMI_Get_universe_size = (int (*)(int *))PMIGetProcAddress(hModule, "PMI_Get_universe_size");
			fn.PMI_Get_appnum = (int (*)(int *))PMIGetProcAddress(hModule, "PMI_Get_appnum");
			fn.PMI_Get_id = (int (*)(char [], int))PMIGetProcAddress(hModule, "PMI_Get_id");
			fn.PMI_Get_idW = (int (*)(wchar_t [], int))PMIGetProcAddress(hModule, "PMI_Get_idW");
			fn.PMI_Get_kvs_domain_id = (int (*)(char [], int))PMIGetProcAddress(hModule, "PMI_Get_kvs_domain_id");
			fn.PMI_Get_kvs_domain_idW = (int (*)(wchar_t [], int))PMIGetProcAddress(hModule, "PMI_Get_kvs_domain_idW");
			fn.PMI_Get_id_length_max = (int (*)(int *))PMIGetProcAddress(hModule, "PMI_Get_id_length_max");
			fn.PMI_Barrier = (int (*)(void))PMIGetProcAddress(hModule, "PMI_Barrier");
			fn.PMI_Get_clique_size = (int (*)(int *))PMIGetProcAddress(hModule, "PMI_Get_clique_size");
			fn.PMI_Get_clique_ranks = (int (*)(int [], int))PMIGetProcAddress(hModule, "PMI_Get_clique_ranks");
			fn.PMI_Abort = (int (*)(int, const char[]))PMIGetProcAddress(hModule, "PMI_Abort");
			fn.PMI_AbortW = (int (*)(int, const wchar_t[]))PMIGetProcAddress(hModule, "PMI_AbortW");
			fn.PMI_Register_abort_hook = (int (*)(PMI_Abort_callback_t))PMIGetProcAddress(hModule, "PMI_Register_abort_hook");
			fn.PMI_Register_print_queues_fn = (int (*)(PMI_Print_queues_fn_t))PMIGetProcAddress(hModule, "PMI_Register_print_queues_fn");
			fn.PMI_Print_queues = (int (*)(wchar_t *, int ))PMIGetProcAddress(hModule, "PMI_Print_queues");
			fn.PMI_Register_print_call_history_fn = (int (*)(PMI_Print_call_history_fn_t))PMIGetProcAddress(hModule, "PMI_Register_print_call_history_fn");
			fn.PMI_Print_call_history = (int (*)(wchar_t *, int ))PMIGetProcAddress(hModule, "PMI_Print_call_history");
			fn.PMI_KVS_Get_my_name = (int (*)(char [], int))PMIGetProcAddress(hModule, "PMI_KVS_Get_my_name");
			fn.PMI_KVS_Get_my_nameW = (int (*)(wchar_t [], int))PMIGetProcAddress(hModule, "PMI_KVS_Get_my_nameW");
			fn.PMI_KVS_Get_name_length_max = (int (*)(int *))PMIGetProcAddress(hModule, "PMI_KVS_Get_name_length_max");
			fn.PMI_KVS_Get_key_length_max = (int (*)(int *))PMIGetProcAddress(hModule, "PMI_KVS_Get_key_length_max");
			fn.PMI_KVS_Get_value_length_max = (int (*)(int *))PMIGetProcAddress(hModule, "PMI_KVS_Get_value_length_max");
			fn.PMI_KVS_Create = (int (*)(char [], int))PMIGetProcAddress(hModule, "PMI_KVS_Create");
			fn.PMI_KVS_CreateW = (int (*)(wchar_t [], int))PMIGetProcAddress(hModule, "PMI_KVS_CreateW");
			fn.PMI_KVS_Destroy = (int (*)(const char []))PMIGetProcAddress(hModule, "PMI_KVS_Destroy");
			fn.PMI_KVS_DestroyW = (int (*)(const wchar_t []))PMIGetProcAddress(hModule, "PMI_KVS_DestroyW");
			fn.PMI_KVS_Put = (int (*)(const char [], const char [], const char []))PMIGetProcAddress(hModule, "PMI_KVS_Put");
			fn.PMI_KVS_PutW = (int (*)(const wchar_t [], const wchar_t [], const wchar_t []))PMIGetProcAddress(hModule, "PMI_KVS_PutW");
			fn.PMI_KVS_Commit = (int (*)(const char []))PMIGetProcAddress(hModule, "PMI_KVS_Commit");
			fn.PMI_KVS_CommitW = (int (*)(const wchar_t []))PMIGetProcAddress(hModule, "PMI_KVS_CommitW");
			fn.PMI_KVS_Get = (int (*)(const char [], const char [], char [], int))PMIGetProcAddress(hModule, "PMI_KVS_Get");
			fn.PMI_KVS_GetW = (int (*)(const wchar_t [], const wchar_t [], wchar_t [], int))PMIGetProcAddress(hModule, "PMI_KVS_GetW");
			fn.PMI_KVS_Iter_first = (int (*)(const char [], char [], int, char [], int))PMIGetProcAddress(hModule, "PMI_KVS_Iter_first");
			fn.PMI_KVS_Iter_firstW = (int (*)(const wchar_t [], wchar_t [], int, wchar_t [], int))PMIGetProcAddress(hModule, "PMI_KVS_Iter_firstW");
			fn.PMI_KVS_Iter_next = (int (*)(const char [], char [], int, char [], int))PMIGetProcAddress(hModule, "PMI_KVS_Iter_next");
			fn.PMI_KVS_Iter_nextW = (int (*)(const wchar_t [], wchar_t [], int, wchar_t [], int))PMIGetProcAddress(hModule, "PMI_KVS_Iter_nextW");
			fn.PMI_Spawn_multiple = (int (*)(int, const char *[], const char **[], const int [], const int [], const PMI_keyval_t *[], int, const PMI_keyval_t [], int [], int))PMIGetProcAddress(hModule, "PMI_Spawn_multiple");
			fn.PMI_Spawn_multipleW = (int (*)(int, const wchar_t *[], const wchar_t **[], const int [], const int [], const PMI_keyvalW_t *[], int, const PMI_keyvalW_t [], int [], int))PMIGetProcAddress(hModule, "PMI_Spawn_multipleW");
			fn.PMI_Parse_option = (int (*)(int, char *[], int *, PMI_keyval_t **, int *))PMIGetProcAddress(hModule, "PMI_Parse_option");
			fn.PMI_Args_to_keyval = (int (*)(int *, char *((*)[]), PMI_keyval_t **, int *))PMIGetProcAddress(hModule, "PMI_Args_to_keyval");
			fn.PMI_Free_keyvals = (int (*)(PMI_keyval_t [], int))PMIGetProcAddress(hModule, "PMI_Free_keyvals");
			fn.PMI_Publish_name = (int (*)(const char [], const char [] ))PMIGetProcAddress(hModule, "PMI_Publish_name");
			fn.PMI_Publish_nameW = (int (*)(const wchar_t [], const wchar_t [] ))PMIGetProcAddress(hModule, "PMI_Publish_nameW");
			fn.PMI_Unpublish_name = (int (*)( const char [] ))PMIGetProcAddress(hModule, "PMI_Unpublish_name");
			fn.PMI_Unpublish_nameW = (int (*)( const wchar_t [] ))PMIGetProcAddress(hModule, "PMI_Unpublish_nameW");
			fn.PMI_Lookup_name = (int (*)( const char [], char [] ))PMIGetProcAddress(hModule, "PMI_Lookup_name");
			fn.PMI_Lookup_nameW = (int (*)( const wchar_t [], wchar_t [] ))PMIGetProcAddress(hModule, "PMI_Lookup_nameW");
			fn.PMI_Get_last_error_msg = (char * (*)( void ))PMIGetProcAddress(hModule, "PMI_Get_last_error_msg");
			fn.PMI_Get_last_error_msgW = (wchar_t * (*)( void ))PMIGetProcAddress(hModule, "PMI_Get_last_error_msgW");
			fn.PMI_Get_error_string = (int (*)(int , char *, int ))PMIGetProcAddress(hModule, "PMI_Get_error_string");
			fn.PMI_Get_error_stringW = (int (*)(int , wchar_t *, int ))PMIGetProcAddress(hModule, "PMI_Get_error_stringW");
			fn.PMI_Free_error_string = (int (*)(int ))PMIGetProcAddress(hModule, "PMI_Free_error_string");
			return fn.PMI_Init(spawned);
		}
	}

	fn.PMI_Init = iPMI_Init;
	fn.PMI_Initialized = iPMI_Initialized;
	fn.PMI_Finalize = iPMI_Finalize;
	fn.PMI_Get_size = iPMI_Get_size;
	fn.PMI_Get_rank = iPMI_Get_rank;
	fn.PMI_Get_universe_size = iPMI_Get_universe_size;
	fn.PMI_Get_appnum = iPMI_Get_appnum;
	fn.PMI_Get_id = iPMI_Get_id;
	fn.PMI_Get_idW = iPMI_Get_idW;
	fn.PMI_Get_kvs_domain_id = iPMI_Get_kvs_domain_id;
	fn.PMI_Get_kvs_domain_idW = iPMI_Get_kvs_domain_idW;
	fn.PMI_Get_id_length_max = iPMI_Get_id_length_max;
	fn.PMI_Barrier = iPMI_Barrier;
	fn.PMI_Get_clique_size = iPMI_Get_clique_size;
	fn.PMI_Get_clique_ranks = iPMI_Get_clique_ranks;
	fn.PMI_Abort = iPMI_Abort;
	fn.PMI_AbortW = iPMI_AbortW;
	fn.PMI_Register_abort_hook = iPMI_Register_abort_hook;
	fn.PMI_Register_print_queues_fn = iPMI_Register_print_queues_fn;
	fn.PMI_Print_queues = iPMI_Print_queues;
	fn.PMI_Register_print_call_history_fn = iPMI_Register_print_call_history_fn;
	fn.PMI_Print_call_history = iPMI_Print_call_history;
	fn.PMI_KVS_Get_my_name = iPMI_KVS_Get_my_name;
	fn.PMI_KVS_Get_my_nameW = iPMI_KVS_Get_my_nameW;
	fn.PMI_KVS_Get_name_length_max = iPMI_KVS_Get_name_length_max;
	fn.PMI_KVS_Get_key_length_max = iPMI_KVS_Get_key_length_max;
	fn.PMI_KVS_Get_value_length_max = iPMI_KVS_Get_value_length_max;
	fn.PMI_KVS_Create = iPMI_KVS_Create;
	fn.PMI_KVS_CreateW = iPMI_KVS_CreateW;
	fn.PMI_KVS_Destroy = iPMI_KVS_Destroy;
	fn.PMI_KVS_DestroyW = iPMI_KVS_DestroyW;
	fn.PMI_KVS_Put = iPMI_KVS_Put;
	fn.PMI_KVS_PutW = iPMI_KVS_PutW;
	fn.PMI_KVS_Commit = iPMI_KVS_Commit;
	fn.PMI_KVS_CommitW = iPMI_KVS_CommitW;
	fn.PMI_KVS_Get = iPMI_KVS_Get;
	fn.PMI_KVS_GetW = iPMI_KVS_GetW;
	fn.PMI_KVS_Iter_first = iPMI_KVS_Iter_first;
	fn.PMI_KVS_Iter_firstW = iPMI_KVS_Iter_firstW;
	fn.PMI_KVS_Iter_next = iPMI_KVS_Iter_next;
	fn.PMI_KVS_Iter_nextW = iPMI_KVS_Iter_nextW;
	fn.PMI_Spawn_multiple = iPMI_Spawn_multiple;
	fn.PMI_Spawn_multipleW = iPMI_Spawn_multipleW;
	fn.PMI_Parse_option = iPMI_Parse_option;
	fn.PMI_Args_to_keyval = iPMI_Args_to_keyval;
	fn.PMI_Free_keyvals = iPMI_Free_keyvals;
	fn.PMI_Publish_name = iPMI_Publish_name;
	fn.PMI_Publish_nameW = iPMI_Publish_nameW;
	fn.PMI_Unpublish_name = iPMI_Unpublish_name;
	fn.PMI_Unpublish_nameW = iPMI_Unpublish_nameW;
	fn.PMI_Lookup_name = iPMI_Lookup_name;
	fn.PMI_Lookup_nameW = iPMI_Lookup_nameW;
	fn.PMI_Get_last_error_msg = iPMI_Get_last_error_msg;
	fn.PMI_Get_last_error_msgW = iPMI_Get_last_error_msgW;
	fn.PMI_Get_error_string = iPMI_Get_error_string;
	fn.PMI_Get_error_stringW = iPMI_Get_error_stringW;
	fn.PMI_Free_error_string = iPMI_Free_error_string;

	return fn.PMI_Init(spawned);
	}
	__except (EXCEPTION_EXECUTE_HANDLER)
	{
		return PMI_FAIL;
	}
}

int PMI_Finalize()
{
	if (fn.PMI_Finalize == NULL)
		return PMI_FAIL;
	return fn.PMI_Finalize();
}

int PMI_Get_size(int *size)
{
	if (fn.PMI_Get_size == NULL)
		return PMI_FAIL;
	return fn.PMI_Get_size(size);
}

int PMI_Get_rank(int *rank)
{
	if (fn.PMI_Get_rank == NULL)
		return PMI_FAIL;
	return fn.PMI_Get_rank(rank);
}

int PMI_Get_universe_size(int *size)
{
	if (fn.PMI_Get_universe_size == NULL)
		return PMI_FAIL;
	return fn.PMI_Get_universe_size(size);
}

int PMI_Get_appnum(int *appnum)
{
	if (fn.PMI_Get_appnum == NULL)
		return PMI_FAIL;
	return fn.PMI_Get_appnum(appnum);
}

int PMI_Get_clique_size( int *size )
{
	if (fn.PMI_Get_clique_size == NULL)
		return PMI_FAIL;
	return fn.PMI_Get_clique_size(size);
}

int PMI_Get_clique_ranks( int ranks[], int length )
{
	if (fn.PMI_Get_clique_ranks == NULL)
		return PMI_FAIL;
	return fn.PMI_Get_clique_ranks(ranks, length);
}

int PMI_Get_id( char id_str[], int length )
{
	if (fn.PMI_Get_id == NULL)
		return PMI_FAIL;
	return fn.PMI_Get_id(id_str, length);
}

int PMI_Get_idW( wchar_t id_str[], int length )
{
	if (fn.PMI_Get_idW == NULL)
		return PMI_FAIL;
	return fn.PMI_Get_idW(id_str, length);
}

int PMI_Get_kvs_domain_id( char id_str[], int length )
{
	if (fn.PMI_Get_kvs_domain_id == NULL)
		return PMI_FAIL;
	return fn.PMI_Get_kvs_domain_id(id_str, length);
}

int PMI_Get_kvs_domain_idW( wchar_t id_str[], int length )
{
	if (fn.PMI_Get_kvs_domain_idW == NULL)
		return PMI_FAIL;
	return fn.PMI_Get_kvs_domain_idW(id_str, length);
}

int PMI_Get_id_length_max(int *maxlen)
{
	if (fn.PMI_Get_id_length_max == NULL)
		return PMI_FAIL;
	return fn.PMI_Get_id_length_max(maxlen);
}

int PMI_Barrier()
{
	if (fn.PMI_Barrier == NULL)
		return PMI_FAIL;
	return fn.PMI_Barrier();
}

int PMI_Abort(int exit_code, const char error_msg[])
{
	if (fn.PMI_Abort == NULL)
		return PMI_FAIL;
	__try
	{
		return fn.PMI_Abort(exit_code, error_msg);
	}
	__except (EXCEPTION_EXECUTE_HANDLER)
	{
		printf("exception thrown in fn.PMI_Abort\n");fflush(stdout);
		return PMI_FAIL;
	}
}

int PMI_AbortW(int exit_code, const wchar_t error_msg[])
{
	if (fn.PMI_AbortW == NULL)
		return PMI_FAIL;
	__try
	{
		return fn.PMI_AbortW(exit_code, error_msg);
	}
	__except (EXCEPTION_EXECUTE_HANDLER)
	{
		printf("exception thrown in fn.PMI_AbortW\n");fflush(stdout);
		return PMI_FAIL;
	}
}

int PMI_Register_abort_hook(PMI_Abort_callback_t function)
{
	if (fn.PMI_Register_abort_hook == NULL)
		return PMI_FAIL;
	return fn.PMI_Register_abort_hook(function);
}

int PMI_Register_print_queues_fn(PMI_Print_queues_fn_t function)
{
	if (fn.PMI_Register_print_queues_fn == NULL)
		return PMI_FAIL;
	return fn.PMI_Register_print_queues_fn(function);
}

int PMI_Print_queues(wchar_t *str, int max_length)
{
	if (fn.PMI_Print_queues == NULL)
		return PMI_FAIL;
	return fn.PMI_Print_queues(str, max_length);
}

int PMI_Register_print_call_history_fn(PMI_Print_call_history_fn_t function)
{
	if (fn.PMI_Register_print_call_history_fn == NULL)
		return PMI_FAIL;
	return fn.PMI_Register_print_call_history_fn(function);
}

int PMI_Print_call_history(wchar_t *str, int max_length)
{
	if (fn.PMI_Print_call_history == NULL)
		return PMI_FAIL;
	return fn.PMI_Print_call_history(str, max_length);
}

int PMI_KVS_Get_my_name(char kvsname[], int length)
{
	if (fn.PMI_KVS_Get_my_name == NULL)
		return PMI_FAIL;
	return fn.PMI_KVS_Get_my_name(kvsname, length);
}

int PMI_KVS_Get_my_nameW(wchar_t kvsname[], int length)
{
	if (fn.PMI_KVS_Get_my_nameW == NULL)
		return PMI_FAIL;
	return fn.PMI_KVS_Get_my_nameW(kvsname, length);
}

int PMI_KVS_Get_name_length_max(int *maxlen)
{
	if (fn.PMI_KVS_Get_name_length_max == NULL)
		return PMI_FAIL;
	return fn.PMI_KVS_Get_name_length_max(maxlen);
}

int PMI_KVS_Get_key_length_max(int *maxlen)
{
	if (fn.PMI_KVS_Get_key_length_max == NULL)
		return PMI_FAIL;
	return fn.PMI_KVS_Get_key_length_max(maxlen);
}

int PMI_KVS_Get_value_length_max(int *maxlen)
{
	if (fn.PMI_KVS_Get_value_length_max == NULL)
		return PMI_FAIL;
	return fn.PMI_KVS_Get_value_length_max(maxlen);
}

int PMI_KVS_Create(char kvsname[], int length)
{
	if (fn.PMI_KVS_Create == NULL)
		return PMI_FAIL;
	return fn.PMI_KVS_Create(kvsname, length);
}

int PMI_KVS_CreateW(wchar_t kvsname[], int length)
{
	if (fn.PMI_KVS_CreateW == NULL)
		return PMI_FAIL;
	return fn.PMI_KVS_CreateW(kvsname, length);
}

int PMI_KVS_Destroy(const char kvsname[])
{
	if (fn.PMI_KVS_Destroy == NULL)
		return PMI_FAIL;
	return fn.PMI_KVS_Destroy(kvsname);
}

int PMI_KVS_DestroyW(const wchar_t kvsname[])
{
	if (fn.PMI_KVS_DestroyW == NULL)
		return PMI_FAIL;
	return fn.PMI_KVS_DestroyW(kvsname);
}

int PMI_KVS_Put(const char kvsname[], const char key[], const char value[])
{
	if (fn.PMI_KVS_Put == NULL)
		return PMI_FAIL;
	return fn.PMI_KVS_Put(kvsname, key, value);
}

int PMI_KVS_PutW(const wchar_t kvsname[], const wchar_t key[], const wchar_t value[])
{
	if (fn.PMI_KVS_PutW == NULL)
		return PMI_FAIL;
	return fn.PMI_KVS_PutW(kvsname, key, value);
}

int PMI_KVS_Commit(const char kvsname[])
{
	if (fn.PMI_KVS_Commit == NULL)
		return PMI_FAIL;
	return fn.PMI_KVS_Commit(kvsname);
}

int PMI_KVS_CommitW(const wchar_t kvsname[])
{
	if (fn.PMI_KVS_CommitW == NULL)
		return PMI_FAIL;
	return fn.PMI_KVS_CommitW(kvsname);
}

int PMI_KVS_Get(const char kvsname[], const char key[], char value[], int length)
{
	if (fn.PMI_KVS_Get == NULL)
		return PMI_FAIL;
	return fn.PMI_KVS_Get(kvsname, key, value, length);
}

int PMI_KVS_GetW(const wchar_t kvsname[], const wchar_t key[], wchar_t value[], int length)
{
	if (fn.PMI_KVS_GetW == NULL)
		return PMI_FAIL;
	return fn.PMI_KVS_GetW(kvsname, key, value, length);
}

int PMI_KVS_Iter_first(const char kvsname[], char key[], int key_len, char value[], int val_len)
{
	if (fn.PMI_KVS_Iter_first == NULL)
		return PMI_FAIL;
	return fn.PMI_KVS_Iter_first(kvsname, key, key_len, value, val_len);
}

int PMI_KVS_Iter_firstW(const wchar_t kvsname[], wchar_t key[], int key_len, wchar_t value[], int val_len)
{
	if (fn.PMI_KVS_Iter_firstW == NULL)
		return PMI_FAIL;
	return fn.PMI_KVS_Iter_firstW(kvsname, key, key_len, value, val_len);
}

int PMI_KVS_Iter_next(const char kvsname[], char key[], int key_len, char value[], int val_len)
{
	if (fn.PMI_KVS_Iter_next == NULL)
		return PMI_FAIL;
	return fn.PMI_KVS_Iter_next(kvsname, key, key_len, value, val_len);
}

int PMI_KVS_Iter_nextW(const wchar_t kvsname[], wchar_t key[], int key_len, wchar_t value[], int val_len)
{
	if (fn.PMI_KVS_Iter_nextW == NULL)
		return PMI_FAIL;
	return fn.PMI_KVS_Iter_nextW(kvsname, key, key_len, value, val_len);
}

int PMI_Spawn_multiple(int count,
					   const char * cmds[],
					   const char ** argvs[],
					   const int maxprocs[],
					   const int info_keyval_sizes[],
					   const PMI_keyval_t * info_keyval_vectors[],
					   int preput_keyval_size,
					   const PMI_keyval_t preput_keyval_vector[],
					   int errors[],
					   int quiet)
{
	if (fn.PMI_Spawn_multiple == NULL)
		return PMI_FAIL;
	return fn.PMI_Spawn_multiple(count, cmds, argvs, maxprocs,
		info_keyval_sizes, info_keyval_vectors,
		preput_keyval_size, preput_keyval_vector,
		errors, quiet);
}

int PMI_Spawn_multipleW(int count,
					   const wchar_t * cmds[],
					   const wchar_t ** argvs[],
					   const int maxprocs[],
					   const int info_keyval_sizes[],
					   const PMI_keyvalW_t * info_keyval_vectors[],
					   int preput_keyval_size,
					   const PMI_keyvalW_t preput_keyval_vector[],
					   int errors[],
					   int quiet)
{
	if (fn.PMI_Spawn_multipleW == NULL)
		return PMI_FAIL;
	return fn.PMI_Spawn_multipleW(count, cmds, argvs, maxprocs,
		info_keyval_sizes, info_keyval_vectors,
		preput_keyval_size, preput_keyval_vector,
		errors, quiet);
}

int PMI_Parse_option(int num_args, char *args[], int *num_parsed, PMI_keyval_t **keyvalp, int *size)
{
	if (fn.PMI_Parse_option == NULL)
		return PMI_FAIL;
	return fn.PMI_Parse_option(num_args, args, num_parsed, keyvalp, size);
}

int PMI_Args_to_keyval(int *argcp, char *((*argvp)[]), PMI_keyval_t **keyvalp, int *size)
{
	if (fn.PMI_Args_to_keyval == NULL)
		return PMI_FAIL;
	return fn.PMI_Args_to_keyval(argcp, argvp, keyvalp, size);
}

int PMI_Free_keyvals(PMI_keyval_t keyvalp[], int size)
{
	if (fn.PMI_Free_keyvals == NULL)
		return PMI_FAIL;
	return fn.PMI_Free_keyvals(keyvalp, size);
}

int PMI_Publish_name( const char service_name[], const char port[] )
{
	if (fn.PMI_Publish_name == NULL)
		return PMI_FAIL;
	return fn.PMI_Publish_name(service_name, port);
}

int PMI_Publish_nameW( const wchar_t service_name[], const wchar_t port[] )
{
	if (fn.PMI_Publish_nameW == NULL)
		return PMI_FAIL;
	return fn.PMI_Publish_nameW(service_name, port);
}

int PMI_Unpublish_name( const char service_name[] )
{
	if (fn.PMI_Unpublish_name == NULL)
		return PMI_FAIL;
	return fn.PMI_Unpublish_name(service_name);
}

int PMI_Unpublish_nameW( const wchar_t service_name[] )
{
	if (fn.PMI_Unpublish_nameW == NULL)
		return PMI_FAIL;
	return fn.PMI_Unpublish_nameW(service_name);
}

int PMI_Lookup_name( const char service_name[], char port[] )
{
	if (fn.PMI_Lookup_name == NULL)
		return PMI_FAIL;
	return fn.PMI_Lookup_name(service_name, port);
}

int PMI_Lookup_nameW( const wchar_t service_name[], wchar_t port[] )
{
	if (fn.PMI_Lookup_nameW == NULL)
		return PMI_FAIL;
	return fn.PMI_Lookup_nameW(service_name, port);
}

/* Always returning a valid string makes the error printing functions simpler because they don't have to check for NULL */
const char * PMI_Get_last_error_msg( void )
{
	const char *result;
	if (fn.PMI_Get_last_error_msg == NULL)
		return "";
	result = fn.PMI_Get_last_error_msg();
	if (result == NULL)
		return "";
	return result;
}

const wchar_t * PMI_Get_last_error_msgW( void )
{
	const wchar_t *result;
	if (fn.PMI_Get_last_error_msgW == NULL)
		return L"";
	result = fn.PMI_Get_last_error_msgW();
	if (result == NULL)
		return L"";
	return result;
}

int PMI_Get_error_string(int pmi_error, char *str, int max_length)
{
	if (fn.PMI_Get_error_string == NULL)
		return PMI_FAIL;
	return fn.PMI_Get_error_string(pmi_error, str, max_length);
}

int PMI_Get_error_stringW(int pmi_error, wchar_t *str, int max_length)
{
	if (fn.PMI_Get_error_stringW == NULL)
		return PMI_FAIL;
	return fn.PMI_Get_error_stringW(pmi_error, str, max_length);
}

int PMI_Free_error_string(int pmi_error)
{
	if (fn.PMI_Free_error_string == NULL)
		return PMI_FAIL;
	return fn.PMI_Free_error_string(pmi_error);
}
