/* -*- Mode: C; c-basic-offset:4 ; -*- */
/*  $Id: irlog2rlog.c,v 1.16 2005/08/12 22:49:30 ashton Exp $
 *
 *  (C) 2001 by Argonne National Laboratory.
 *	  See COPYRIGHT in top-level directory.
 */

#include "rlog.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <ctype.h> /* isdigit */
//#include "mpichconf.h" /* HAVE_SNPRINTF */
//#include "mpimem.h" /* MPIU_Snprintf */
#define MPIU_Snprintf _snwprintf_s
#define MPIU_Wcsncpy wcsncpy_s
#define MPIU_Malloc malloc
#define MPIU_Free free
#include "mpibase.h" /* MPIU_Error_printf */
//#define MPIU_Error_printf wprintf
//#define MPIU_Msg_printf wprintf
#include <io.h> /* _wunlink */

#ifndef BOOL
#define BOOL int
#endif
#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif
#ifndef min
#define min(a,b) (((a) < (b)) ? (a) : (b))
#endif

#define MAX_RANK 1024*1024

typedef struct RLOG_State_list
{
	RLOG_STATE state;
	struct RLOG_State_list *next;
} RLOG_State_list;

typedef struct RecursionStruct
{
	wchar_t filename[1024];
	int rank;
	int level;
	int num_events;
	FILE *fout;
	struct RecursionStruct *next;
} RecursionStruct;

typedef struct StartArrowStruct
{
	double start_time;
	int src, tag, length;
	struct StartArrowStruct *next;
} StartArrowStruct;

typedef struct EndArrowStruct
{
	int src, tag;
	double timestamp;
	struct EndArrowStruct *next;
} EndArrowStruct;

typedef struct ArrowNode
{
	int rank;
	StartArrowStruct *pStartList;
	EndArrowStruct *pEndList;
	struct ArrowNode *pNext;
} ArrowNode;

ArrowNode *g_pArrowList = NULL;
RecursionStruct *g_pLevel = NULL;
FILE *g_fArrow = NULL;
wchar_t g_pszArrowFilename[1024];
RLOG_State_list *g_pList = NULL;
wchar_t s_error_msg[1024];

static int ReadFileData(char *pBuffer, int length, FILE *fin)
{
	size_t num_read;

	while (length)
	{
		num_read = fread(pBuffer, 1, length, fin);
		if (num_read == (size_t)-1)
		{
			_wcserror_s(s_error_msg, 1024, errno);
			MPIU_Error_printf(L"Error: fread failed - %s\n", s_error_msg);
			return errno;
		}

		/*wprintf(L"fread(%d)", num_read);fflush(stdout);*/

		length -= (int)num_read;
		pBuffer += num_read;
	}
	return 0;
}

static int WriteFileData(const void *pvBuffer, int length, FILE *fout)
{
	size_t num_written;
	const char *pBuffer = pvBuffer;

	while (length)
	{
		num_written = fwrite(pBuffer, 1, length, fout);
		if (num_written == (size_t)-1)
		{
			_wcserror_s(s_error_msg, 1024, errno);
			MPIU_Error_printf(L"Error: fwrite failed - %s\n", s_error_msg);
			return errno;
		}

		/*wprintf(L"fwrite(%d)", num_written);fflush(stdout);*/

		length -= (int)num_written;
		pBuffer += num_written;
	}
	return 0;
}

ArrowNode *GetArrowNode(int rank)
{
	ArrowNode *pNode = g_pArrowList;

	while (pNode)
	{
		if (pNode->rank == rank)
			return pNode;
		pNode = pNode->pNext;
	}

	pNode = (ArrowNode *)MPIU_Malloc(sizeof(ArrowNode));
	pNode->pEndList = NULL;
	pNode->pStartList = NULL;
	pNode->rank = rank;
	pNode->pNext = g_pArrowList;
	g_pArrowList = pNode;

	return pNode;
}

EndArrowStruct *ExtractEndNode(ArrowNode *pNode, int src, int tag)
{
	EndArrowStruct *pIter, *pRet;
	if (pNode->pEndList == NULL)
		return NULL;
	if ((pNode->pEndList->src == src) && (pNode->pEndList->tag == tag))
	{
		pRet = pNode->pEndList;
		pNode->pEndList = pNode->pEndList->next;
		return pRet;
	}
	pIter = pNode->pEndList;
	while (pIter->next != NULL)
	{
		if ((pIter->next->src == src) && (pIter->next->tag == tag))
		{
			pRet = pIter->next;
			pIter->next = pIter->next->next;
			return pRet;
		}
		pIter = pIter->next;
	}
	return NULL;
}

StartArrowStruct *ExtractStartNode(ArrowNode *pNode, int src, int tag)
{
	StartArrowStruct *pIter, *pRet;
	if (pNode->pStartList == NULL)
		return NULL;
	if ((pNode->pStartList->src == src) && (pNode->pStartList->tag == tag))
	{
		pRet = pNode->pStartList;
		pNode->pStartList = pNode->pStartList->next;
		return pRet;
	}
	pIter = pNode->pStartList;
	while (pIter->next != NULL)
	{
		if ((pIter->next->src == src) && (pIter->next->tag == tag))
		{
			pRet = pIter->next;
			pIter->next = pIter->next->next;
			return pRet;
		}
		pIter = pIter->next;
	}
	return NULL;
}

void SaveArrow(RLOG_IARROW *pArrow)
{
	ArrowNode *pNode;
	StartArrowStruct *pStart, *pStartIter;
	EndArrowStruct *pEnd, *pEndIter;
	RLOG_ARROW arrow;
	errno_t error;

	if (g_fArrow == NULL)
	{
		MPIU_Wcsncpy(g_pszArrowFilename, 1024, L"ArrowFile.tmp", _TRUNCATE);
		g_fArrow = NULL;
		error = _wfopen_s(&g_fArrow, g_pszArrowFilename, L"w+b");
		if (error)
		{
			MPIU_Error_printf(L"Unable to open ArrowFile.tmp, error %d\n", error);
			return;
		}
	}

	if (pArrow->sendrecv == RLOG_SENDER)
	{
		pNode = GetArrowNode(pArrow->remote);
		pEnd = ExtractEndNode(pNode, pArrow->rank, pArrow->tag);
		if (pEnd == NULL)
		{
			pStart = (StartArrowStruct *)MPIU_Malloc(sizeof(StartArrowStruct));
			pStart->src = pArrow->rank;
			pStart->tag = pArrow->tag;
			pStart->length = pArrow->length;
			pStart->start_time = pArrow->timestamp;
			pStart->next = NULL;
			if (pNode->pStartList == NULL)
			{
				pNode->pStartList = pStart;
			}
			else
			{
				pStartIter = pNode->pStartList;
				while (pStartIter->next != NULL)
					pStartIter = pStartIter->next;
				pStartIter->next = pStart;
			}
			return;
		}
		arrow.src = pArrow->rank;
		arrow.dest = pArrow->remote;
		arrow.length = pArrow->length;
		arrow.start_time = pEnd->timestamp;
		arrow.end_time = pArrow->timestamp;
		arrow.tag = pArrow->tag;
		arrow.leftright = RLOG_ARROW_LEFT;
		/* fwrite(&arrow, sizeof(RLOG_ARROW), 1, g_fArrow); */
		WriteFileData(&arrow, sizeof(RLOG_ARROW), g_fArrow);
		MPIU_Free(pEnd);
	}
	else
	{
		arrow.dest = pArrow->rank;
		arrow.end_time = pArrow->timestamp;
		arrow.tag = pArrow->tag;
		arrow.length = pArrow->length;

		pNode = GetArrowNode(pArrow->rank);
		pStart = ExtractStartNode(pNode, pArrow->remote, pArrow->tag);
		if (pStart != NULL)
		{
			arrow.src = pStart->src;
			arrow.start_time = pStart->start_time;
			arrow.length = pStart->length; /* the sender length is more accurate than the receiver length */
			arrow.leftright = RLOG_ARROW_RIGHT;
			MPIU_Free(pStart);
			/* fwrite(&arrow, sizeof(RLOG_ARROW), 1, g_fArrow); */
			WriteFileData(&arrow, sizeof(RLOG_ARROW), g_fArrow);
		}
		else
		{
			pEnd = (EndArrowStruct *)MPIU_Malloc(sizeof(EndArrowStruct));
			pEnd->src = pArrow->remote;
			pEnd->tag = pArrow->tag;
			pEnd->timestamp = pArrow->timestamp;
			pEnd->next = NULL;
			if (pNode->pEndList == NULL)
			{
				pNode->pEndList = pEnd;
			}
			else
			{
				pEndIter = pNode->pEndList;
				while (pEndIter->next != NULL)
					pEndIter = pEndIter->next;
				pEndIter->next = pEnd;
			}
		}
	}

	/* fwrite(pArrow, sizeof(RLOG_IARROW), 1, g_fArrow); */
}

RecursionStruct *GetLevel(int rank, int recursion)
{
	errno_t error;
	RecursionStruct *pLevel;

	pLevel = g_pLevel;

	while (pLevel)
	{
		if (pLevel->level == recursion && pLevel->rank == rank)
			return pLevel;
		pLevel = pLevel->next;
	}

	pLevel = (RecursionStruct*)MPIU_Malloc(sizeof(RecursionStruct));
	MPIU_Snprintf(pLevel->filename, 1024, _TRUNCATE, L"irlog.%d.%d.tmp", rank, recursion);
	pLevel->fout = NULL;
	error = _wfopen_s(&pLevel->fout, pLevel->filename, L"w+b");
	if (error)
	{
		MPIU_Error_printf(L"Unable to open %s, error %d\n", pLevel->filename, error);
		return NULL;
	}
	pLevel->rank = rank;
	pLevel->level = recursion;
	pLevel->num_events = 0;
	pLevel->next = g_pLevel;
	g_pLevel = pLevel;

	return pLevel;
}

void SaveEvent(RLOG_EVENT *pEvent)
{
	RecursionStruct *pLevel;

	pLevel = GetLevel(pEvent->rank, pEvent->recursion);
	pLevel->num_events++;
	/* fwrite(pEvent, sizeof(RLOG_EVENT), 1, pLevel->fout); */
	WriteFileData(pEvent, sizeof(RLOG_EVENT), pLevel->fout);
}

void SaveState(RLOG_STATE *pState)
{
	RLOG_State_list *pIter;

	pIter = g_pList;
	while (pIter)
	{
		if (pIter->state.event == pState->event)
		{
			/* replace old with new */
			memcpy(&pIter->state, pState, sizeof(RLOG_STATE));
			return;
		}
		pIter = pIter->next;
	}

	pIter = (RLOG_State_list*)MPIU_Malloc(sizeof(RLOG_State_list));
	memcpy(&pIter->state, pState, sizeof(RLOG_STATE));
	pIter->next = g_pList;
	g_pList = pIter;
}

/*
FILE *OpenRlogFile(wchar_t *filename)
{
	wchar_t out_filename[1024];
	wchar_t *pExt;

	pExt = strstr(filename, L".irlog");
	if (pExt)
	{
		memcpy(out_filename, filename, pExt-filename);
		strcpy(&out_filename[pExt-filename], L".rlog");
		MPIU_Msg_printf(L"out_filename: %s\n", out_filename);
	}
	else
	{
		sprintf(out_filename, L"%s.rlog", filename);
	}

	return fopen(out_filename, L"wb");
}
*/

#define BUFFER_SIZE 16*1024*1024

void AppendFile(FILE *fout, FILE *fin)
{
	long total;
	size_t num_read, num_written;
	char *buffer, *buf;

	buffer = (char*)MPIU_Malloc(sizeof(char) * BUFFER_SIZE);

	total = ftell(fin);
	fseek(fin, 0L, SEEK_SET);

	while (total > 0)
	{
		num_read = fread(buffer, 1, min(BUFFER_SIZE, total), fin);
		if (num_read == 0)
		{
			MPIU_Error_printf(L"failed to read from input file\n");
			return;
		}
		total -= (long)num_read;
		buf = buffer;
		while (num_read > 0)
		{
			num_written = fwrite(buf, 1, num_read, fout);
			if (num_written == 0)
			{
				MPIU_Error_printf(L"failed to write to output file\n");
				return;
			}
			num_read -= num_written;
			buf += num_written;
		}
	}

	MPIU_Free(buffer);
}

int FindMinRank(RecursionStruct *pLevel)
{
	int rank = pLevel->rank;
	while (pLevel)
	{
		if (pLevel->rank < rank)
			rank = pLevel->rank;
		pLevel = pLevel->next;
	}
	return rank;
}

int FindMaxRank(RecursionStruct *pLevel)
{
	int rank = pLevel->rank;
	while (pLevel)
	{
		if (pLevel->rank > rank)
			rank = pLevel->rank;
		pLevel = pLevel->next;
	}
	return rank;
}

RecursionStruct *FindMinLevel(RecursionStruct *pLevel)
{
	RecursionStruct *pRet;
	pRet = pLevel;
	while (pLevel)
	{
		if (pLevel->rank < pRet->rank)
		{
			pRet = pLevel;
		}
		else
		{
			if (pLevel->rank == pRet->rank && pLevel->level < pRet->level)
			{
				pRet = pLevel;
			}
		}
		pLevel = pLevel->next;
	}
	return pRet;
}

int FindMaxRecursion(RecursionStruct *pLevel, int rank)
{
	int nMax = 0;
	while (pLevel)
	{
		if (pLevel->rank == rank)
		{
			nMax = (pLevel->level > nMax) ? pLevel->level : nMax;
		}
		pLevel = pLevel->next;
	}
	return nMax+1; /* return nMax + 1 because the recursion levels begin with zero */
}

void RemoveLevel(int rank)
{
	RecursionStruct *pLevel = g_pLevel, *pTrailer = g_pLevel;
	while (pLevel)
	{
		if (pLevel->rank == rank)
		{
			if (pLevel == pTrailer)
			{
				if (g_pLevel == pLevel)
					g_pLevel = g_pLevel->next;
				pLevel = pLevel->next;
				fclose(pTrailer->fout);
				_wunlink(pTrailer->filename);
				MPIU_Free(pTrailer);
				pTrailer = pLevel;
			}
			else
			{
				pTrailer->next = pLevel->next;
				fclose(pLevel->fout);
				_wunlink(pLevel->filename);
				MPIU_Free(pLevel);
				pLevel = pTrailer->next;
			}
		}
		else
		{
			if (pTrailer != pLevel)
				pTrailer = pTrailer->next;
			pLevel = pLevel->next;
		}
	}
}

BOOL GetNextArrow(IRLOG_IOStruct *pInput)
{
	do
	{
		if (IRLOG_GetNextRecord(pInput))
		{
			IRLOG_CloseInputStruct(&pInput);
			return FALSE;
		}
	} while (pInput->header.type != RLOG_IARROW_TYPE);
	return TRUE;
}

void ReadAllArrows(IRLOG_IOStruct **ppInput, int nNumInputs)
{
	int i, j;
	double dMin;
	int index;

	if (nNumInputs < 1)
		return;

	/* make sure all the inputs are currently on an iarrow record */
	for (i=0; i<nNumInputs; i++)
	{
		if (ppInput[i]->header.type != RLOG_IARROW_TYPE)
		{
			if (!GetNextArrow(ppInput[i]))
			{
				for (j=i; j<nNumInputs-1; j++)
					ppInput[j] = ppInput[j+1];
				nNumInputs--;
				i--;
			}
		}
	}

	/* read iarrow records and save them to the arrow file */
	/* until all the inputs are exhausted */
	while (nNumInputs > 0)
	{
		/* find the earliest record */
		dMin = ppInput[0]->record.iarrow.timestamp;
		index = 0;
		for (i=1; i<nNumInputs; i++)
		{
			if (ppInput[i]->record.iarrow.timestamp < dMin)
			{
				dMin = ppInput[i]->record.iarrow.timestamp;
				index = i;
			}
		}
		/* save the record */
		SaveArrow(&ppInput[index]->record.iarrow);
		/* read the next iarrow record from the same source */
		if (!GetNextArrow(ppInput[index]))
		{
			for (i=index; i<nNumInputs-1; i++)
				ppInput[i] = ppInput[i+1];
			nNumInputs--;
		}
	}
}

BOOL IsNumber(wchar_t *str)
{
	if (str == NULL)
		return FALSE;
	while (*str != L'\0')
	{
		if (!isdigit(*str))
			return FALSE;
		str++;
	}
	return TRUE;
}

static BOOL s_bFreeArgv = FALSE;
void GenerateNewArgv(int *pargc, wchar_t ***pargv, int n)
{
	int argc;
	wchar_t **argv;
	size_t length;
	int i;
	wchar_t *buffer, *str;
	size_t num_chars_left;
	int num_chars;

	length = (sizeof(wchar_t*) * (n+3)) + (wcslen((*pargv)[0]) + 1 + wcslen((*pargv)[1]) + 1 + (15 * n))*sizeof(wchar_t);
	buffer = (wchar_t*)MPIU_Malloc(length);

	argc = n+2;
	argv = (wchar_t**)buffer;
	str = (wchar_t*)((unsigned char *)buffer + (sizeof(wchar_t*) * (n+4)));
	num_chars_left = (length - (sizeof(wchar_t*) * (n+4))) / sizeof(wchar_t);
	argv[0] = str;
	num_chars = swprintf_s(str, num_chars_left, L"%s", (*pargv)[0]);
	str += num_chars;
	num_chars_left -= num_chars + 1; /* Add one for the following str++ */
	*str++ = L'\0';
	argv[1] = str;
	num_chars = swprintf_s(str, num_chars_left, L"%s", (*pargv)[1]);
	str += num_chars;
	num_chars_left -= num_chars + 1; /* Add one for the following str++ */
	*str++ = L'\0';
	for (i=0; i<n; i++)
	{
		argv[i+2] = str;
		num_chars = swprintf_s(str, num_chars_left, L"log%d.irlog", i);
		str += num_chars;
		num_chars_left -= num_chars + 1; /* Add one for the following str++ */
		*str++ = L'\0';
	}
	argv[n+3] = NULL;

	*pargc = argc;
	*pargv = argv;
	s_bFreeArgv = TRUE;
}

int wmain(int argc, wchar_t *argv[])
{
	RLOG_FILE_HEADER header;
	RLOG_State_list *pState;

	int nNumInputs;
	IRLOG_IOStruct *pInput;
	IRLOG_IOStruct **ppInput;

	int nMaxRank = 0;
	int nMinRank = MAX_RANK;
	int nNumStates = 0;
	int type, length;
	int nMaxLevel = 0;
	int nNumLevels = 0;
	int nTotalNumEvents = 0;
	int nNumEvents;
	RecursionStruct *pLevel;
	FILE *fout;
	int i, j;
	int nRank;
	errno_t error;

	if (argc < 3)
	{
		MPIU_Error_printf(L"Usage:\nirlog2rlog out.rlog in0.irlog in1.irlog ...\nirlog2rlog out.rlog n\n");
		return 0;
	}

	if (argc == 3 && IsNumber(argv[2]))
	{
		GenerateNewArgv(&argc, &argv, _wtoi(argv[2]));
	}

	nNumInputs = argc - 2;

	/* open the output rlog file */
	fout = NULL;
	error = _wfopen_s(&fout, argv[1], L"wb");
	if (error)
	{
		MPIU_Error_printf(L"unable to open output file '%s', error %d\n", argv[1], error);
		return -1;
	}

	/* read the arrows from all the files in order */
	ppInput = (IRLOG_IOStruct**)MPIU_Malloc(nNumInputs * sizeof(IRLOG_IOStruct*));
	for (i=0; i<nNumInputs; i++)
	{
		ppInput[i] = IRLOG_CreateInputStruct(argv[i+2]);
		if (ppInput[i] == NULL)
		{
			MPIU_Error_printf(L"Unable to create an input structure for '%s', skipping\n", argv[i+2]);
		}
	}
	for (i=0; i<nNumInputs; i++)
	{
		if (ppInput[i] == NULL)
		{
			for (j=i; j<nNumInputs-1; j++)
				ppInput[j] = ppInput[j+1];
			nNumInputs--;
			i--;
		}
	}
	MPIU_Msg_printf(L"reading the arrows from all the input files.\n");fflush(stdout);
	ReadAllArrows(ppInput, nNumInputs);

	nNumInputs = argc - 2;

	/* read, parse and save all the data from the irlog files */
	for (i=0; i<nNumInputs; i++)
	{
		pInput = IRLOG_CreateInputStruct(argv[i+2]);
		if (pInput == NULL)
		{
			MPIU_Error_printf(L"Unable to create an input structure for '%s', skipping\n", argv[i+2]);
		}
		else
		{
			MPIU_Msg_printf(L"reading irlog file: %s\n", argv[i+2]);fflush(stdout);
			for(;;)
			{
				switch (pInput->header.type)
				{
				case RLOG_STATE_TYPE:
					SaveState(&pInput->record.state);
					break;
				case RLOG_COMM_TYPE:
					nMaxRank = (pInput->record.comm.rank > nMaxRank) ? pInput->record.comm.rank : nMaxRank;
					nMinRank = (pInput->record.comm.rank < nMinRank) ? pInput->record.comm.rank : nMinRank;
					break;
				case RLOG_IARROW_TYPE:
					/* SaveArrow(&pInput->record.iarrow); */
					break;
				case RLOG_EVENT_TYPE:
					SaveEvent(&pInput->record.event);
					break;
				default:
					MPIU_Error_printf(L"Unknown irlog record type: %d\n", pInput->header.type);
					break;
				}
				
				if (IRLOG_GetNextRecord(pInput))
				{
					IRLOG_CloseInputStruct(&pInput);
					break;
				}
			}
		}
	}

	/* set the fields in the header */
	header.nMinRank = FindMinRank(g_pLevel);
	header.nMaxRank = FindMaxRank(g_pLevel);
	if (nMinRank != header.nMinRank)
		MPIU_Error_printf(L"minimum event rank %d does not equal the minimum comm record rank %d\n", header.nMinRank, nMinRank);
	if (nMaxRank != header.nMaxRank)
		MPIU_Error_printf(L"maximum event rank %d does not equal the maximum comm record rank %d\n", header.nMaxRank, nMaxRank);

	/* write header */
	MPIU_Msg_printf(L"writing header.\n");fflush(stdout);
	type = RLOG_HEADER_SECTION;
	length = sizeof(RLOG_FILE_HEADER);
	/* fwrite(&type, sizeof(int), 1, fout); */
	WriteFileData(&type, sizeof(int), fout);
	/* fwrite(&length, sizeof(int), 1, fout);*/
	WriteFileData(&length, sizeof(int), fout);
	/* fwrite(&header, sizeof(RLOG_FILE_HEADER), 1, fout); */
	WriteFileData(&header, sizeof(RLOG_FILE_HEADER), fout);

	/* write states */
	if (g_pList)
	{
		MPIU_Msg_printf(L"writing states.\n");fflush(stdout);
	}
	pState = g_pList;
	while (pState)
	{
		nNumStates++;
		pState = pState->next;
	}
	type = RLOG_STATE_SECTION;
	length = nNumStates * sizeof(RLOG_STATE);
	/* fwrite(&type, sizeof(int), 1, fout); */
	WriteFileData(&type, sizeof(int), fout);
	/* fwrite(&length, sizeof(int), 1, fout); */
	WriteFileData(&length, sizeof(int), fout);
	pState = g_pList;
	while (pState)
	{
		/* fwrite(pState, sizeof(RLOG_STATE), 1, fout); */
		WriteFileData(pState, sizeof(RLOG_STATE), fout);
		pState = pState->next;
	}

	/* write arrows */
	if (g_fArrow)
	{
		MPIU_Msg_printf(L"writing arrows.\n");fflush(stdout);
		type = RLOG_ARROW_SECTION;
		length = ftell(g_fArrow);
		/* fwrite(&type, sizeof(int), 1, fout); */
		WriteFileData(&type, sizeof(int), fout);
		/* fwrite(&length, sizeof(int), 1, fout); */
		WriteFileData(&length, sizeof(int), fout);
		AppendFile(fout, g_fArrow);
	}

	/* write events */
	while (g_pLevel)
	{
		pLevel = FindMinLevel(g_pLevel);
		nNumLevels = FindMaxRecursion(g_pLevel, pLevel->rank);
		nRank = pLevel->rank;

		/* count the events for this rank */
		nNumEvents = 0;
		for (i=0; i<nNumLevels; i++)
		{
			pLevel = GetLevel(pLevel->rank, i);
			nNumEvents += pLevel->num_events;
		}
		/* write an event section for this rank */
		type = RLOG_EVENT_SECTION;
		length = sizeof(int) + sizeof(int) + (nNumLevels * sizeof(int)) + (nNumEvents * sizeof(RLOG_EVENT));
		/* fwrite(&type, sizeof(int), 1, fout); */
		WriteFileData(&type, sizeof(int), fout);
		/* fwrite(&length, sizeof(int), 1, fout); */
		WriteFileData(&length, sizeof(int), fout);
		/* fwrite(&nRank, sizeof(int), 1, fout); */
		WriteFileData(&nRank, sizeof(int), fout);
		/* fwrite(&nNumLevels, sizeof(int), 1, fout); */
		WriteFileData(&nNumLevels, sizeof(int), fout);
		for (i=0; i<nNumLevels; i++)
		{
			pLevel = GetLevel(nRank, i);
			/* fwrite(&pLevel->num_events, sizeof(int), 1, fout); */
			WriteFileData(&pLevel->num_events, sizeof(int), fout);
		}
		for (i=0; i<nNumLevels; i++)
		{
			MPIU_Msg_printf(L"writing event level %d:%d\n", nRank, i);fflush(stdout);
			pLevel = GetLevel(nRank, i);
			AppendFile(fout, pLevel->fout);
		}
		/* remove this rank from the list of levels */
		RemoveLevel(nRank);
	}

	/* free resources */
	while (g_pList)
	{
		pState = g_pList;
		g_pList = g_pList->next;
		MPIU_Free(pState);
	}
	if (g_fArrow)
	{
		fclose(g_fArrow);
		_wunlink(g_pszArrowFilename);
	}

	if (s_bFreeArgv)
		MPIU_Free(argv);

	return 0;
}
