/* -*- Mode: C; c-basic-offset:4 ; -*- */
/*  $Id: printirlog.c,v 1.4 2003/09/30 20:47:02 David Exp $
 *
 *  (C) 2006 by Deino Software.
 *  (C) 2001 by Argonne National Laboratory.
 *	  See COPYRIGHT in top-level directory.
 */
/* Allow the wprintf's that describe the contents of the file.  Error 
   messages use the wprintf routines */
/* style: allow:wprintf:9 sig:0 */

#include "rlog.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

void PrintRecord(IRLOG_IOStruct *pInput)
{
	switch (pInput->header.type)
	{
	case RLOG_EVENT_TYPE:
		wprintf(L"RLOG_EVENT  -");
		wprintf(L" %3d:%04d, start: %10g, end: %10g, %d\n", 
			pInput->record.event.rank,
			pInput->record.event.event, 
			pInput->record.event.start_time, 
			pInput->record.event.end_time,
			pInput->record.event.recursion);
		break;
	case RLOG_IARROW_TYPE:
		wprintf(L"RLOG_IARROW -");
		wprintf(L" %3d %s %3d, tag: %3d, length: %d\n", 
			pInput->record.iarrow.rank,
			(pInput->record.iarrow.sendrecv == RLOG_SENDER) ? L"->" : L"<-",
			pInput->record.iarrow.remote,
			pInput->record.iarrow.tag,
			pInput->record.iarrow.length);
		break;
	case RLOG_COMM_TYPE:
		wprintf(L"RLOG_COMM   -");
		wprintf(L" %3d, newcomm: %d\n",
			pInput->record.comm.rank,
			pInput->record.comm.newcomm);
		break;
	case RLOG_STATE_TYPE:
		wprintf(L"RLOG_STATE  -");
		wprintf(L" id: %3d, %12s : %s\n",
			pInput->record.state.event,
			pInput->record.state.color,
			pInput->record.state.description);
		break;
	case RLOG_ENDLOG_TYPE:
		wprintf(L"RLOG_ENDLOG\n");
		break;
	default:
		wprintf(L"unknown record type %d\n", pInput->header.type);
	}
}

int wmain(int argc, wchar_t *argv[])
{
	int nNumInputs = 0;
	IRLOG_IOStruct *pInput = NULL;
	int rec_type;
	int bAllRecords = 1;
	int bExcludeType = 0;

	if (argc < 2)
	{
		wprintf(L"%s irlogfile [REC_TYPE]\n", argv[0]);
		return -1;
	}

	pInput = IRLOG_CreateInputStruct(argv[1]);
	if (pInput == NULL)
	{
		wprintf(L"Error setting up file '%s'\n", argv[1]);
		return -1;
	}

	if (argc > 2)
	{
		if (wcscmp(argv[2], L"RLOG_EVENT") == 0)
		{
			bAllRecords = 0;
			rec_type = RLOG_EVENT_TYPE;
		}
		if (wcscmp(argv[2], L"RLOG_ARROW") == 0)
		{
			bAllRecords = 0;
			rec_type = RLOG_ARROW_TYPE;
		}
		if (wcscmp(argv[2], L"RLOG_STATE") == 0)
		{
			bAllRecords = 0;
			rec_type = RLOG_STATE_TYPE;
		}
		if (wcscmp(argv[2], L"RLOG_COMM") == 0)
		{
			bAllRecords = 0;
			rec_type = RLOG_COMM_TYPE;
		}
		if (wcscmp(argv[2], L"RLOG_ENDLOG") == 0)
		{
			bAllRecords = 0;
			rec_type = RLOG_ENDLOG_TYPE;
		}
	}
	if (argc > 3)
	{
		if (wcscmp(&argv[3][1], L"exclude") == 0)
			bExcludeType = 1;
	}

	while (1)
	{
		if (bExcludeType)
		{
			if (pInput->header.type != rec_type)
				PrintRecord(pInput);
		}
		else
		{
			if ((bAllRecords) || (pInput->header.type == rec_type))
				PrintRecord(pInput);
		}
		if (IRLOG_GetNextRecord(pInput))
		{
			IRLOG_CloseInputStruct(&pInput);
			break;
		}
	}

	return 0;
}
