/*   Copyright (C) 2009 Deino Software. */

#include "mpiimpl.h"
#include <math.h>

/* global variables */

RLOG_Struct *g_pRLOG = NULL;
RLOG_Struct *g_pRLOGdetailed = NULL;

static int s_RLOG_Initialized = 0;
int MPIU_Timer_init(int rank, int size)
{
	const wchar_t *env;

	if (s_RLOG_Initialized)
	{
		/* MPIU_Timer_init already called. */
		return -1;
	}
	g_pRLOG = RLOG_InitLog(rank, size);
	if (g_pRLOG == NULL)
		return -1;
	env = MPIU_GetStaticEnv(L"RLOG_DETAILED");
	if (env != NULL)
	{
		if (env[0] == L'\0' || wcscmp(env, L"1") == 0 || wcscmp(env, L"on") == 0)
			g_pRLOGdetailed = g_pRLOG;
	}

	RLOG_EnableLogging(g_pRLOG);

	RLOG_SaveFirstTimestamp(g_pRLOG);

	RLOG_LogCommID(g_pRLOG, (int)MPI_COMM_WORLD);

	/* arrow state */
	RLOG_DescribeState(g_pRLOG, RLOG_ARROW_EVENT_ID, L"Arrow", L"255 255 255");

	MPIR_Describe_timer_states();

	s_RLOG_Initialized = 1;
	return MPI_SUCCESS;
}

int MPIU_Timer_finalize()
{
	if (g_pRLOG == NULL)
		return -1;
	if (!s_RLOG_Initialized)
		return 0;

	RLOG_DisableLogging(g_pRLOG);

	/* The code originally included output statements that surrounded the
	   writing of the logfiles with "Writing logfile." and "finished."
	   While sometimes useful, it is often valuable to have 
	   the use of logging make no changes in the observable output 
	   from the program, for example, in testing that the logging 
	   does not impact the correctness of the code or while used 
	   within an environment that expects all stdout/err to come
	   from code that the user has specifically written.

	   To that end, the output lines have been commented out.  Should
	   there be a strong desire to provide this output, it should be
	   made optional, turned on (off by default) with a runtime parameter.
	 */
	/* MPIU_Msg_printf( L"Writing logfile.\n");fflush(stdout); */
	RLOG_FinishLog(g_pRLOG);
	/* MPIU_Msg_printf(L"finished.\n");fflush(stdout); */
	s_RLOG_Initialized = 0;

	return MPI_SUCCESS;
}

/* This routine makes the RLOG_DescribeState call for each name */
#include "state_names.h"
int MPIR_Describe_timer_states( void )
{
	MPIU_State_defs *def = mpich_states;
	
	while (def && def->state >= 0) {
		RLOG_DescribeState( g_pRLOG, def->state, def->funcname, 
							def->color );
		def++;
	}
	return 0;
}
