/*   Copyright (C) 2009 Deino Software. */

#include "mpiimpl.h"
#ifdef HAVE_STDIO_H
#include <stdio.h>
#endif
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#ifdef HAVE_MATH_H
#include <math.h>
#endif
/* ctype is needed for isspace and isascii (isspace is only defined for 
   values on which isascii returns true). */
#include <ctype.h>

#define MPIU_STR_TRUNCATED MPIU_STR_NOMEM

#undef FUNCNAME
#define FUNCNAME encode_buffer
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int encode_buffer(wchar_t *dest, int dest_length, const wchar_t *src, int src_length, int *num_encoded)
{
	int num_used;
	int n = 0;
	wchar_t ch;
	MPIDI_STATE_DECL(MPID_STATE_ENCODE_BUFFER);

	MPIDI_FUNC_ENTER(MPID_STATE_ENCODE_BUFFER);
	if (src_length == 0)
	{
		if (dest_length > 2)
		{
			*dest = MPIU_STR_QUOTE_CHAR;
			dest++;
			*dest = MPIU_STR_QUOTE_CHAR;
			dest++;
			*dest = L'\0';
			*num_encoded = 0;
			MPIDI_FUNC_EXIT(MPID_STATE_ENCODE_BUFFER);
			return MPIU_STR_SUCCESS;
		}
		else
		{
			MPIDI_FUNC_EXIT(MPID_STATE_ENCODE_BUFFER);
			return MPIU_STR_TRUNCATED;
		}
	}
	while (src_length && dest_length)
	{
		ch = *src;
		num_used = MPIU_Snprintf(dest, dest_length, _TRUNCATE, L"%02X", (int)*src);
		if (num_used < 0)
		{
			*num_encoded = n;
			MPIDI_FUNC_EXIT(MPID_STATE_ENCODE_BUFFER);
			return MPIU_STR_TRUNCATED;
		}
		/*MPIU_DBG_MSG_FMT(OTHER,VERBOSE,(MPIU_DBG_FDEST,L" %c = %c%c", ch, dest[0], dest[1]));*/
		dest += num_used;
		dest_length -= num_used;
		src++;
		n++;
		src_length--;
	}
	*num_encoded = n;
	MPIDI_FUNC_EXIT(MPID_STATE_ENCODE_BUFFER);
	return src_length ? MPIU_STR_TRUNCATED : MPIU_STR_SUCCESS;
}

#undef FUNCNAME
#define FUNCNAME decode_buffer
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int decode_buffer(const wchar_t *str, wchar_t *dest, int length, int *num_decoded)
{
	wchar_t hex[3];
	int value;
	int n = 0;
	MPIDI_STATE_DECL(MPID_STATE_DECODE_BUFFER);

	MPIDI_FUNC_ENTER(MPID_STATE_DECODE_BUFFER);

	if (str == NULL || dest == NULL || num_decoded == NULL)
	{
		MPIDI_FUNC_EXIT(MPID_STATE_DECODE_BUFFER);
		return MPIU_STR_FAIL;
	}
	if (length < 1)
	{
		*num_decoded = 0;
		if (*str == L'\0')
		{
			MPIDI_FUNC_EXIT(MPID_STATE_DECODE_BUFFER);
			return MPIU_STR_SUCCESS;
		}
		MPIDI_FUNC_EXIT(MPID_STATE_DECODE_BUFFER);
		return MPIU_STR_TRUNCATED;
	}
	if (*str == MPIU_STR_QUOTE_CHAR)
		str++;
	hex[2] = L'\0';
	while (*str != L'\0' &&
		   *str != MPIU_STR_SEPAR_CHAR &&
		   *str != MPIU_STR_QUOTE_CHAR &&
		   length)
	{
		hex[0] = *str;
		str++;
		hex[1] = *str;
		str++;
		swscanf_s(hex, L"%X", &value);
		*dest = (wchar_t)value;
		/*MPIU_DBG_MSG_FMT(OTHER,VERBOSE,(MPIU_DBG_FDEST,L" %s = %c\n", hex, *dest));*/
		dest++;
		n++;
		length--;
	}
	*num_decoded = n;
	if (length == 0)
	{
		if (*str != L'\0' &&
			*str != MPIU_STR_SEPAR_CHAR &&
			*str != MPIU_STR_QUOTE_CHAR)
		{
			MPIDI_FUNC_EXIT(MPID_STATE_DECODE_BUFFER);
			return MPIU_STR_TRUNCATED;
		}
	}
	MPIDI_FUNC_EXIT(MPID_STATE_DECODE_BUFFER);
	return MPIU_STR_SUCCESS;
}

#undef FUNCNAME
#define FUNCNAME first_token
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static const wchar_t * first_token(const wchar_t *str)
{
	MPIDI_STATE_DECL(MPID_STATE_FIRST_TOKEN);

	MPIDI_FUNC_ENTER(MPID_STATE_FIRST_TOKEN);
	if (str == NULL)
	{
		MPIDI_FUNC_EXIT(MPID_STATE_FIRST_TOKEN);
		return NULL;
	}
	/* isspace is defined only if isascii is true */
	while (/*isascii(*str) && isspace(*str)*/ *str == MPIU_STR_SEPAR_CHAR)
		str++;
	if (*str == L'\0')
	{
		MPIDI_FUNC_EXIT(MPID_STATE_FIRST_TOKEN);
		return NULL;
	}
	MPIDI_FUNC_EXIT(MPID_STATE_FIRST_TOKEN);
	return str;
}

#undef FUNCNAME
#define FUNCNAME next_token
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static const wchar_t * next_token(const wchar_t *str)
{
	MPIDI_STATE_DECL(MPID_STATE_NEXT_TOKEN);

	MPIDI_FUNC_ENTER(MPID_STATE_NEXT_TOKEN);
	if (str == NULL)
	{
		MPIDI_FUNC_EXIT(MPID_STATE_NEXT_TOKEN);
		return NULL;
	}
	str = first_token(str);
	if (str == NULL)
	{
		MPIDI_FUNC_EXIT(MPID_STATE_NEXT_TOKEN);
		return NULL;
	}
	if (*str == MPIU_STR_QUOTE_CHAR)
	{
		/* move over string */
		str++; /* move over the first quote */
		if (*str == L'\0')
		{
			MPIDI_FUNC_EXIT(MPID_STATE_NEXT_TOKEN);
			return NULL;
		}
		while (*str != MPIU_STR_QUOTE_CHAR)
		{
			/* move until the last quote, ignoring escaped quotes */
			if (*str == MPIU_STR_ESCAPE_CHAR)
			{
				str++;
				if (*str == MPIU_STR_QUOTE_CHAR)
					str++;
			}
			else
			{
				str++;
			}
			if (*str == L'\0')
			{
				MPIDI_FUNC_EXIT(MPID_STATE_NEXT_TOKEN);
				return NULL;
			}
		}
		str++; /* move over the last quote */
	}
	else
	{
		if (*str == MPIU_STR_DELIM_CHAR)
		{
			/* move over the DELIM token */
			str++;
		}
		else
		{
			/* move over literal */
			while (/*(isascii(*str) &&
					!isspace(*str)) &&*/
					*str != MPIU_STR_SEPAR_CHAR &&
				   *str != MPIU_STR_DELIM_CHAR &&
				   *str != L'\0')
				str++;
		}
	}
	MPIDI_FUNC_EXIT(MPID_STATE_NEXT_TOKEN);
	return first_token(str);
}

#undef FUNCNAME
#define FUNCNAME compare_token
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int compare_token(const wchar_t *token, const wchar_t *str)
{
	MPIDI_STATE_DECL(MPID_STATE_COMPARE_TOKEN);

	MPIDI_FUNC_ENTER(MPID_STATE_COMPARE_TOKEN);
	if (token == NULL || str == NULL)
	{
		MPIDI_FUNC_EXIT(MPID_STATE_COMPARE_TOKEN);
		return -1;
	}

	if (*token == MPIU_STR_QUOTE_CHAR)
	{
		/* compare quoted strings */
		token++; /* move over the first quote */
		/* compare characters until reaching the end of the string or the end quote character */
		for(;;)
		{
			if (*token == MPIU_STR_ESCAPE_CHAR)
			{
				if (*(token+1) == MPIU_STR_QUOTE_CHAR)
				{
					/* move over the escape character if the next character is a quote character */
					token++;
				}
				if (*token != *str)
					break;
			}
			else
			{
				if (*token != *str || *token == MPIU_STR_QUOTE_CHAR)
					break;
			}
			if (*str == L'\0')
				break;
			token++;
			str++;
		}
		if (*str == L'\0' && *token == MPIU_STR_QUOTE_CHAR)
		{
			MPIDI_FUNC_EXIT(MPID_STATE_COMPARE_TOKEN);
			return 0;
		}
		if (*token == MPIU_STR_QUOTE_CHAR)
		{
			MPIDI_FUNC_EXIT(MPID_STATE_COMPARE_TOKEN);
			return 1;
		}
		if (*str < *token)
		{
			MPIDI_FUNC_EXIT(MPID_STATE_COMPARE_TOKEN);
			return -1;
		}
		MPIDI_FUNC_EXIT(MPID_STATE_COMPARE_TOKEN);
		return 1;
	}

	/* compare DELIM token */
	if (*token == MPIU_STR_DELIM_CHAR)
	{
		if (*str == MPIU_STR_DELIM_CHAR)
		{
			str++;
			if (*str == L'\0')
			{
				MPIDI_FUNC_EXIT(MPID_STATE_COMPARE_TOKEN);
				return 0;
			}
			MPIDI_FUNC_EXIT(MPID_STATE_COMPARE_TOKEN);
			return 1;
		}
		if (*token < *str)
		{
			MPIDI_FUNC_EXIT(MPID_STATE_COMPARE_TOKEN);
			return -1;
		}
		MPIDI_FUNC_EXIT(MPID_STATE_COMPARE_TOKEN);
		return 1;
	}

	/* compare literals */
	while (*token == *str &&
		   *str != L'\0' &&
		   *token != MPIU_STR_DELIM_CHAR && 
		   (/*isascii(*token) && !isspace(*token)*/ *token != MPIU_STR_SEPAR_CHAR) )
	{
		token++;
		str++;
	}
	if ( (*str == L'\0') &&
		 (*token == MPIU_STR_DELIM_CHAR ||
		  (/*isascii(*token) && isspace(*token)*/ *token == MPIU_STR_SEPAR_CHAR) || *token == L'\0') )
	{
		MPIDI_FUNC_EXIT(MPID_STATE_COMPARE_TOKEN);
		return 0;
	}
	if (*token == MPIU_STR_DELIM_CHAR || 
		(/*isascii(*token) && isspace(*token)*/ *token == MPIU_STR_SEPAR_CHAR) || *token < *str)
	{
		MPIDI_FUNC_EXIT(MPID_STATE_COMPARE_TOKEN);
		return -1;
	}
	MPIDI_FUNC_EXIT(MPID_STATE_COMPARE_TOKEN);
	return 1;
}

#undef FUNCNAME
#define FUNCNAME token_copy
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int token_copy(const wchar_t *token, wchar_t *str, int maxlen)
{
	MPIDI_STATE_DECL(MPID_STATE_TOKEN_COPY);

	MPIDI_FUNC_ENTER(MPID_STATE_TOKEN_COPY);
	/* check parameters */
	if (token == NULL || str == NULL)
	{
		MPIDI_FUNC_EXIT(MPID_STATE_TOKEN_COPY);
		return MPIU_STR_FAIL;
	}

	/* check special buffer lengths */
	if (maxlen < 1)
	{
		MPIDI_FUNC_EXIT(MPID_STATE_TOKEN_COPY);
		return MPIU_STR_FAIL;
	}
	if (maxlen == 1)
	{
		*str = L'\0';
		MPIDI_FUNC_EXIT(MPID_STATE_TOKEN_COPY);
		return (str[0] == L'\0') ? MPIU_STR_SUCCESS : MPIU_STR_TRUNCATED;
	}

	/* cosy up to the token */
	token = first_token(token);
	if (token == NULL)
	{
		*str = L'\0';
		MPIDI_FUNC_EXIT(MPID_STATE_TOKEN_COPY);
		return MPIU_STR_SUCCESS;
	}

	if (*token == MPIU_STR_DELIM_CHAR)
	{
		/* copy the special deliminator token */
		str[0] = MPIU_STR_DELIM_CHAR;
		str[1] = L'\0';
		MPIDI_FUNC_EXIT(MPID_STATE_TOKEN_COPY);
		return MPIU_STR_SUCCESS;
	}

	if (*token == MPIU_STR_QUOTE_CHAR)
	{
		/* quoted copy */
		token++; /* move over the first quote */
		do
		{
			if (*token == MPIU_STR_ESCAPE_CHAR)
			{
				if (*(token+1) == MPIU_STR_QUOTE_CHAR)
					token++;
				*str = *token;
			}
			else
			{
				if (*token == MPIU_STR_QUOTE_CHAR)
				{
					*str = L'\0';
					MPIDI_FUNC_EXIT(MPID_STATE_TOKEN_COPY);
					return MPIU_STR_SUCCESS;
				}
				*str = *token;
			}
			str++;
			token++;
			maxlen--;
		} while (maxlen);
		/* we've run out of destination characters so back up and null terminate the string */
		str--;
		*str = L'\0';
		MPIDI_FUNC_EXIT(MPID_STATE_TOKEN_COPY);
		return MPIU_STR_TRUNCATED;
	}

	/* literal copy */
	while (*token != MPIU_STR_DELIM_CHAR && 
		   (/*isascii(*token) && !isspace(*token)*/ *token != MPIU_STR_SEPAR_CHAR) && *token != L'\0' && maxlen)
	{
		*str = *token;
		str++;
		token++;
		maxlen--;
	}
	if (maxlen)
	{
		*str = L'\0';
		MPIDI_FUNC_EXIT(MPID_STATE_TOKEN_COPY);
		return MPIU_STR_SUCCESS;
	}
	str--;
	*str = L'\0';
	MPIDI_FUNC_EXIT(MPID_STATE_TOKEN_COPY);
	return MPIU_STR_TRUNCATED;
}

#undef FUNCNAME
#define FUNCNAME token_hide
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static void token_hide(wchar_t *token)
{
	MPIDI_STATE_DECL(MPID_STATE_TOKEN_HIDE);

	MPIDI_FUNC_ENTER(MPID_STATE_TOKEN_HIDE);
	/* check parameters */
	if (token == NULL)
	{
		MPIDI_FUNC_EXIT(MPID_STATE_TOKEN_HIDE);
		return;
	}

	/* cosy up to the token */
	token = (wchar_t*)first_token(token);
	if (token == NULL)
	{
		MPIDI_FUNC_EXIT(MPID_STATE_TOKEN_HIDE);
		return;
	}

	if (*token == MPIU_STR_DELIM_CHAR)
	{
		*token = MPIU_STR_HIDE_CHAR;
		MPIDI_FUNC_EXIT(MPID_STATE_TOKEN_HIDE);
		return;
	}

	/* quoted */
	if (*token == MPIU_STR_QUOTE_CHAR)
	{
		*token = MPIU_STR_HIDE_CHAR;
		token++; /* move over the first quote */
		while (*token != L'\0')
		{
			if (*token == MPIU_STR_ESCAPE_CHAR)
			{
				if (*(token+1) == MPIU_STR_QUOTE_CHAR)
				{
					*token = MPIU_STR_HIDE_CHAR;
					token++;
				}
				*token = MPIU_STR_HIDE_CHAR;
			}
			else
			{
				if (*token == MPIU_STR_QUOTE_CHAR)
				{
					*token = MPIU_STR_HIDE_CHAR;
					MPIDI_FUNC_EXIT(MPID_STATE_TOKEN_HIDE);
					return;
				}
				*token = MPIU_STR_HIDE_CHAR;
			}
			token++;
		}
		MPIDI_FUNC_EXIT(MPID_STATE_TOKEN_HIDE);
		return;
	}

	/* literal */
	while (*token != MPIU_STR_DELIM_CHAR && 
		   (/*isascii(*token) && !isspace(*token)*/ *token != MPIU_STR_SEPAR_CHAR) && *token != L'\0')
	{
		*token = MPIU_STR_HIDE_CHAR;
		token++;
	}
	MPIDI_FUNC_EXIT(MPID_STATE_TOKEN_HIDE);
}

/*@ MPIU_Str_get_string_arg - Extract an option from a string with a maximum length
  
	Input Parameters:
+   str - Source string
.   key - key
-   maxlen - Maximum total length of 'val'

	Output Parameter:
.   val - output string

	Return value:
	MPIU_STR_SUCCESS, MPIU_STR_NOMEM, MPIU_STR_FAIL

	Notes:
	This routine searches for a "key = value" entry in a string

  Module:
  Utility
  @*/
#undef FUNCNAME
#define FUNCNAME MPIU_Str_get_string_arg
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIU_Str_get_string_arg(const wchar_t *str, const wchar_t *flag, wchar_t *val, int maxlen)
{
	MPIDI_STATE_DECL(MPID_STATE_MPIU_STR_GET_STRING_ARG);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIU_STR_GET_STRING_ARG);
	if (maxlen < 1)
	{
		MPIDI_FUNC_EXIT(MPID_STATE_MPIU_STR_GET_STRING_ARG);
		return MPIU_STR_FAIL;
	}

	/* line up with the first token */
	str = first_token(str);
	if (str == NULL)
	{
		MPIDI_FUNC_EXIT(MPID_STATE_MPIU_STR_GET_STRING_ARG);
		return MPIU_STR_FAIL;
	}

	/* This loop will match the first instance of "flag = value" in the string. */
	do
	{
		/*wprintf(L"str = '%s'\n\n", str);*/
		if (compare_token(str, flag) == 0)
		{
			str = next_token(str);
			if (compare_token(str, MPIU_STR_DELIM_STR) == 0)
			{
				str = next_token(str);
				if (str == NULL)
				{
					MPIDI_FUNC_EXIT(MPID_STATE_MPIU_STR_GET_STRING_ARG);
					return MPIU_STR_FAIL;
				}
				/*wprintf(L"val = '%s'\n\n", str);*/
				{
					int result = token_copy(str, val, maxlen);
					MPIDI_FUNC_EXIT(MPID_STATE_MPIU_STR_GET_STRING_ARG);
					return result;
				}
			}
		}
		else
		{
			str = next_token(str);
		}
	} while (str);
	MPIDI_FUNC_EXIT(MPID_STATE_MPIU_STR_GET_STRING_ARG);
	return MPIU_STR_FAIL;
}

/*@ MPIU_Str_get_binary_arg - Extract an option from a string with a maximum length
  
	Input Parameters:
+   str - Source string
.   key - key
-   maxlen - Maximum total length of 'buffer'

	Output Parameter:
+   buffer - output buffer
-   out_length - output length

	Return value:
	MPIU_STR_SUCCESS, MPIU_STR_NOMEM, MPIU_STR_FAIL

	Notes:
	This routine searches for a "key = value" entry in a string and decodes the value
	back to binary data.  The data must have been encoded with MPIU_Str_add_binary_arg.

  Module:
  Utility
  @*/
#undef FUNCNAME
#define FUNCNAME MPIU_Str_get_binary_arg
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIU_Str_get_binary_arg(const wchar_t *str, const wchar_t *flag, wchar_t *buffer, int maxlen, int *out_length)
{
	MPIDI_STATE_DECL(MPID_STATE_MPIU_STR_GET_BINARY_ARG);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIU_STR_GET_BINARY_ARG);
	if (maxlen < 1)
	{
		MPIDI_FUNC_EXIT(MPID_STATE_MPIU_STR_GET_BINARY_ARG);
		return MPIU_STR_FAIL;
	}

	/* line up with the first token */
	str = first_token(str);
	if (str == NULL)
	{
		MPIDI_FUNC_EXIT(MPID_STATE_MPIU_STR_GET_BINARY_ARG);
		return MPIU_STR_FAIL;
	}

	/* This loop will match the first instance of L"flag = value" in the string. */
	do
	{
		if (compare_token(str, flag) == 0)
		{
			str = next_token(str);
			if (compare_token(str, MPIU_STR_DELIM_STR) == 0)
			{
				str = next_token(str);
				if (str == NULL)
				{
					MPIDI_FUNC_EXIT(MPID_STATE_MPIU_STR_GET_BINARY_ARG);
					return MPIU_STR_FAIL;
				}
				{
					int result = decode_buffer(str, buffer, maxlen, out_length);
					MPIDI_FUNC_EXIT(MPID_STATE_MPIU_STR_GET_BINARY_ARG);
					return result;
				}
			}
		}
		else
		{
			str = next_token(str);
		}
	} while (str);
	MPIDI_FUNC_EXIT(MPID_STATE_MPIU_STR_GET_BINARY_ARG);
	return MPIU_STR_FAIL;
}

/*@ MPIU_Str_hide_string_arg - Over-write the value of an string option with * characters
  
	Input Parameters:
+   str - input string
-   key - key

	Output Parameter:
.   str - The string data is modified if the key is found in the string

	Return value:
	MPIU_STR_SUCCESS, MPIU_STR_FAIL

	Notes:
	This routine covers an option in a string converting "key = value" to "key = *****"

  Module:
  Utility
  @*/
#undef FUNCNAME
#define FUNCNAME MPIU_Str_hide_string_arg
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
MPIU_BOOL MPIU_Str_hide_string_arg(wchar_t *str, const wchar_t *flag)
{
	MPIDI_STATE_DECL(MPID_STATE_MPIU_STR_HIDE_STRING_ARG);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIU_STR_HIDE_STRING_ARG);
	/* line up with the first token */
	str = (wchar_t*)first_token(str);
	if (str == NULL)
	{
		MPIDI_FUNC_EXIT(MPID_STATE_MPIU_STR_HIDE_STRING_ARG);
		return MPIU_TRUE;
	}

	do
	{
		if (compare_token(str, flag) == 0)
		{
			str = (wchar_t*)next_token(str);
			if (compare_token(str, MPIU_STR_DELIM_STR) == 0)
			{
				str = (wchar_t*)next_token(str);
				if (str == NULL)
				{
					MPIDI_FUNC_EXIT(MPID_STATE_MPIU_STR_HIDE_STRING_ARG);
					return MPIU_TRUE;
				}
				token_hide(str);
				MPIDI_FUNC_EXIT(MPID_STATE_MPIU_STR_HIDE_STRING_ARG);
				return MPIU_TRUE;
			}
		}
		else
		{
			str = (wchar_t*)next_token(str);
		}
	} while (str);
	MPIDI_FUNC_EXIT(MPID_STATE_MPIU_STR_HIDE_STRING_ARG);
	return MPIU_FALSE;
}

/*@ MPIU_Str_get_int_arg - Extract an option from a string
  
	Input Parameters:
+   str - Source string
-   key - key

	Output Parameter:
.   val_ptr - pointer to the output integer

	Return value:
	MPIU_STR_SUCCESS, MPIU_STR_NOMEM, MPIU_STR_FAIL

	Notes:
	This routine searches for a "key = value" entry in a string and decodes the value
	back to an int.

  Module:
  Utility
  @*/
#undef FUNCNAME
#define FUNCNAME MPIU_Str_get_int_arg
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIU_Str_get_int_arg(const wchar_t *str, const wchar_t *flag, int *val_ptr)
{
	int result;
	wchar_t int_str[12];
	MPIDI_STATE_DECL(MPID_STATE_MPIU_STR_GET_INT_ARG);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIU_STR_GET_INT_ARG);

	result = MPIU_Str_get_string_arg(str, flag, int_str, 12);
	if (result == MPIU_STR_SUCCESS)
	{
		*val_ptr = _wtoi(int_str);
		MPIDI_FUNC_EXIT(MPID_STATE_MPIU_STR_GET_INT_ARG);
		return MPIU_STR_SUCCESS;
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPIU_STR_GET_INT_ARG);
	return result;
}

/* quoted_printf does not NULL terminate the string if maxlen is reached */
#undef FUNCNAME
#define FUNCNAME quoted_printf
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int quoted_printf(wchar_t *str, int maxlen, const wchar_t *val)
{
	int count = 0;
	MPIDI_STATE_DECL(MPID_STATE_QUOTED_PRINTF);

	MPIDI_FUNC_ENTER(MPID_STATE_QUOTED_PRINTF);
	if (maxlen < 1)
	{
		MPIDI_FUNC_EXIT(MPID_STATE_QUOTED_PRINTF);
		return 0;
	}
	*str = MPIU_STR_QUOTE_CHAR;
	str++;
	maxlen--;
	count++;
	while (maxlen)
	{
		if (*val == L'\0')
			break;
		if (*val == MPIU_STR_QUOTE_CHAR)
		{
			*str = MPIU_STR_ESCAPE_CHAR;
			str++;
			maxlen--;
			count++;
			if (maxlen == 0)
			{
				MPIDI_FUNC_EXIT(MPID_STATE_QUOTED_PRINTF);
				return count;
			}
		}
		*str = *val;
		str++;
		maxlen--;
		count++;
		val++;
	}
	if (maxlen)
	{
		*str = MPIU_STR_QUOTE_CHAR;
		str++;
		maxlen--;
		count++;
		if (maxlen == 0)
		{
			MPIDI_FUNC_EXIT(MPID_STATE_QUOTED_PRINTF);
			return count;
		}
		*str = L'\0';
	}
	MPIDI_FUNC_EXIT(MPID_STATE_QUOTED_PRINTF);
	return count;
}

/*@ MPIU_Str_add_string - Add a string to a string
  
	Input Parameters:
+   str_ptr - pointer to the destination string
.   maxlen_ptr - pointer to the maximum length of '*str_ptr'
-   val - string to add

	Output Parameter:
+   str_ptr - The string pointer is updated to the next available location in the string
-   maxlen_ptr - maxlen is decremented by the amount str_ptr is incremented

	Return value:
	MPIU_STR_SUCCESS, MPIU_STR_NOMEM, MPIU_STR_FAIL

	Notes:
	This routine adds a string to a string in such a way that MPIU_Str_get_string can
	retreive the same string back.  It takes into account spaces and quote characters.
	The string pointer is updated to the start of the next string in the string and maxlen
	is updated accordingly.

  Module:
  Utility
  @*/
#undef FUNCNAME
#define FUNCNAME MPIU_Str_add_string
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIU_Str_add_string(wchar_t **str_ptr, int *maxlen_ptr, const wchar_t *val)
{
	int num_chars;
	wchar_t *str;
	int maxlen;
	MPIDI_STATE_DECL(MPID_STATE_MPIU_STR_ADD_STRING);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIU_STR_ADD_STRING);

	str = *str_ptr;
	maxlen = *maxlen_ptr;

	if (wcschr(val, MPIU_STR_SEPAR_CHAR) ||
		wcschr(val, MPIU_STR_QUOTE_CHAR) ||
		wcschr(val, MPIU_STR_DELIM_CHAR))
	{
		num_chars = quoted_printf(str, maxlen, val);
		if (num_chars == maxlen)
		{
			/* truncation, cleanup string */
			*str = L'\0';
			MPIDI_FUNC_EXIT(MPID_STATE_MPIU_STR_ADD_STRING);
			return -1;
		}
		if (num_chars < maxlen - 1)
		{
			str[num_chars] = MPIU_STR_SEPAR_CHAR;
			str[num_chars+1] = L'\0';
			num_chars++;
		}
		else
		{
			str[num_chars] = L'\0';
		}
	}
	else
	{
		if (*val == L'\0')
		{
			num_chars = MPIU_Snprintf(str, maxlen, _TRUNCATE, MPIU_STR_QUOTE_STR MPIU_STR_QUOTE_STR/*L"\"\""*/);
		}
		else
		{
			num_chars = MPIU_Snprintf(str, maxlen, _TRUNCATE, L"%s%c", val, MPIU_STR_SEPAR_CHAR);
		}
		if (num_chars == maxlen)
		{
			*str = L'\0';
			MPIDI_FUNC_EXIT(MPID_STATE_MPIU_STR_ADD_STRING);
			return -1;
		}
	}
	*str_ptr += num_chars;
	*maxlen_ptr -= num_chars;
	MPIDI_FUNC_EXIT(MPID_STATE_MPIU_STR_ADD_STRING);
	return 0;
}

/*@ MPIU_Str_get_string - Get the next string from a string
  
	Input Parameters:
+   str_ptr - pointer to the destination string
-   maxlen_ptr - pointer to the maximum length of '*str_ptr'

	Output Parameter:
+   str_ptr - location of the next string
-   val - location to store the string

	Return value:
	MPIU_STR_SUCCESS, MPIU_STR_NOMEM, MPIU_STR_FAIL

	Return Value:
	The return value is 0 for success, -1 for insufficient buffer space, and 1 for failure.

	Notes:
	This routine gets a string that was previously added by MPIU_Str_add_string.
	It takes into account spaces and quote characters. The string pointer is updated to the
	start of the next string in the string.

  Module:
  Utility
  @*/
#undef FUNCNAME
#define FUNCNAME MPIU_Str_get_string
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIU_Str_get_string(wchar_t **str_ptr, wchar_t *val, int maxlen)
{
	int result;
	wchar_t *str;
	MPIDI_STATE_DECL(MPID_STATE_MPIU_STR_GET_STRING);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIU_STR_GET_STRING);

	if (str_ptr == NULL)
	{
		MPIDI_FUNC_EXIT(MPID_STATE_MPIU_STR_GET_STRING);
		return -2;
	}

	str = *str_ptr;
	
	if (maxlen < 1)
	{
		MPIDI_FUNC_EXIT(MPID_STATE_MPIU_STR_GET_STRING);
		return 0;
	}

	/* line up with the first token */
	str = (wchar_t*)first_token(str);
	if (str == NULL)
	{
		MPIDI_FUNC_EXIT(MPID_STATE_MPIU_STR_GET_STRING);
		return 0;
	}

	/* copy the token */
	result = token_copy(str, val, maxlen);
	if (result == MPIU_STR_SUCCESS)
	{
		str = (wchar_t*)next_token(str);
		*str_ptr = str;
		MPIDI_FUNC_EXIT(MPID_STATE_MPIU_STR_GET_STRING);
		return 0;
	}
	else if (result == MPIU_STR_TRUNCATED)
	{
		MPIDI_FUNC_EXIT(MPID_STATE_MPIU_STR_GET_STRING);
		return -1;
	}

	/* failure */
	MPIDI_FUNC_EXIT(MPID_STATE_MPIU_STR_GET_STRING);
	return -2;
}

/*@ MPIU_Str_add_string_arg - Add an option to a string with a maximum length
  
	Input Parameters:
+   str_ptr - Pointer to the destination string
.   maxlen_ptr - Pointer to the maximum total length of '*str_ptr'
.   key - key
-   val - input string

	Output Parameter:
+   str_ptr - The string pointer is updated to the next available location in the string
-   maxlen_ptr - maxlen is reduced by the number of characters written

	Return value:
	MPIU_STR_SUCCESS, MPIU_STR_NOMEM, MPIU_STR_FAIL

	Notes:
	This routine adds a string option to a string in the form "key = value".

  Module:
  Utility
  @*/
#undef FUNCNAME
#define FUNCNAME MPIU_Str_add_string_arg
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIU_Str_add_string_arg(wchar_t **str_ptr, int *maxlen_ptr, const wchar_t *flag, const wchar_t *val)
{
	int num_chars;
	wchar_t **orig_str_ptr;
	int orig_maxlen;
	MPIDI_STATE_DECL(MPID_STATE_MPIU_STR_ADD_STRING_ARG);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIU_STR_ADD_STRING_ARG);

	if (maxlen_ptr == NULL)
	{
		MPIDI_FUNC_EXIT(MPID_STATE_MPIU_STR_ADD_STRING_ARG);
		return MPIU_STR_FAIL;
	}

	orig_maxlen = *maxlen_ptr;
	orig_str_ptr = str_ptr;

	if (*maxlen_ptr < 1)
	{
		MPIDI_FUNC_EXIT(MPID_STATE_MPIU_STR_ADD_STRING_ARG);
		return MPIU_STR_FAIL;
	}

	/* add the flag */
/*	 wprintf(L"wcsstr flag\n"); */
	if (wcsstr(flag, MPIU_STR_SEPAR_STR) || wcsstr(flag, MPIU_STR_DELIM_STR) || flag[0] == MPIU_STR_QUOTE_CHAR)
	{
		num_chars = quoted_printf(*str_ptr, *maxlen_ptr, flag);
	}
	else
	{
		num_chars = MPIU_Snprintf(*str_ptr, *maxlen_ptr, _TRUNCATE, L"%s", flag);
	}
	*maxlen_ptr = *maxlen_ptr - num_chars;
	if (*maxlen_ptr < 1)
	{
		MPIU_DBG_MSG_S(OTHER,VERBOSE,
				  L"partial argument added to string: '%s'", *str_ptr);
		**str_ptr = L'\0';
		/*(*str_ptr)[num_chars-1] = L'\0';*/
		MPIDI_FUNC_EXIT(MPID_STATE_MPIU_STR_ADD_STRING_ARG);
		return MPIU_STR_NOMEM;
	}
	*str_ptr = *str_ptr + num_chars;

	/* add the deliminator character */
	**str_ptr = MPIU_STR_DELIM_CHAR;
	*str_ptr = *str_ptr + 1;
	*maxlen_ptr = *maxlen_ptr - 1;

	/* add the value string */
/*	 wprintf(L"wcsstr val\n"); */
	if (wcsstr(val, MPIU_STR_SEPAR_STR) || wcsstr(val, MPIU_STR_DELIM_STR) || val[0] == MPIU_STR_QUOTE_CHAR)
	{
		num_chars = quoted_printf(*str_ptr, *maxlen_ptr, val);
	}
	else
	{
		if (*val == L'\0')
		{
			num_chars = MPIU_Snprintf(*str_ptr, *maxlen_ptr, _TRUNCATE, MPIU_STR_QUOTE_STR MPIU_STR_QUOTE_STR/*L"\"\""*/);
		}
		else
		{
			num_chars = MPIU_Snprintf(*str_ptr, *maxlen_ptr, _TRUNCATE, L"%s", val);
		}
	}
	*str_ptr = *str_ptr + num_chars;
	*maxlen_ptr = *maxlen_ptr - num_chars;
	if (*maxlen_ptr < 2)
	{
		MPIU_DBG_MSG_S(OTHER,VERBOSE,
					   L"partial argument added to string: '%s'", *str_ptr);
		**orig_str_ptr = L'\0';
		/*
		*str_ptr = *str_ptr - 1;
		**str_ptr = L'\0';
		*/
		MPIDI_FUNC_EXIT(MPID_STATE_MPIU_STR_ADD_STRING_ARG);
		return MPIU_STR_NOMEM;
	}
	
	/* add the trailing space */
	**str_ptr = MPIU_STR_SEPAR_CHAR;
	*str_ptr = *str_ptr + 1;
	**str_ptr = L'\0';
	*maxlen_ptr = *maxlen_ptr - 1;

	MPIDI_FUNC_EXIT(MPID_STATE_MPIU_STR_ADD_STRING_ARG);
	return MPIU_STR_SUCCESS;
}

/*@ MPIU_Str_add_int_arg - Add an option to a string with a maximum length
  
	Input Parameters:
+   str_ptr - Pointer to the destination string
.   maxlen_ptr - Pointer to the maximum total length of '*str_ptr'
.   key - key
-   val - input integer

	Output Parameter:
+   str_ptr - The string pointer is updated to the next available location in the string
-   maxlen_ptr - maxlen is reduced by the number of characters written

	Return value:
	MPIU_STR_SUCCESS, MPIU_STR_NOMEM, MPIU_STR_FAIL

	Notes:
	This routine adds an integer option to a string in the form "key = value".

  Module:
  Utility
  @*/
#undef FUNCNAME
#define FUNCNAME MPIU_Str_add_int_arg
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIU_Str_add_int_arg(wchar_t **str_ptr, int *maxlen_ptr, const wchar_t *flag, int val)
{
	wchar_t val_str[20];
	int result;
	MPIDI_STATE_DECL(MPID_STATE_MPIU_STR_ADD_INT_ARG);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIU_STR_ADD_INT_ARG);
	MPIU_Snprintf(val_str, 20, _TRUNCATE, L"%d", val);
	result = MPIU_Str_add_string_arg(str_ptr, maxlen_ptr, flag, val_str);
	MPIDI_FUNC_EXIT(MPID_STATE_MPIU_STR_ADD_INT_ARG);
	return result;
}

/*@ MPIU_Str_add_binary_arg - Add an option to a string with a maximum length
  
	Input Parameters:
+   str_ptr - Pointer to the destination string
.   maxlen_ptr - Pointer to the maximum total length of '*str_ptr'
.   key - key
.   val - input data
-   length - length of the input data

	Output Parameter:
+   str_ptr - The string pointer is updated to the next available location in the string
-   maxlen_ptr - maxlen is reduced by the number of characters written

	Return value:
	MPIU_STR_SUCCESS, MPIU_STR_NOMEM, MPIU_STR_FAIL

	Notes:
	This routine encodes binary data into a string option in the form "key = encoded_value".

  Module:
  Utility
  @*/
#undef FUNCNAME
#define FUNCNAME MPIU_Str_add_binary_arg
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIU_Str_add_binary_arg(wchar_t **str_ptr, int *maxlen_ptr, const wchar_t *flag, const wchar_t *buffer, int length)
{
	int result;
	int num_chars;
	wchar_t **orig_str_ptr;
	int orig_maxlen;
	MPIDI_STATE_DECL(MPID_STATE_MPIU_STR_ADD_BINARY_ARG);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIU_STR_ADD_BINARY_ARG);

	if (maxlen_ptr == NULL)
	{
		MPIDI_FUNC_EXIT(MPID_STATE_MPIU_STR_ADD_BINARY_ARG);
		return MPIU_STR_FAIL;
	}

	orig_maxlen = *maxlen_ptr;
	orig_str_ptr = str_ptr;

	if (*maxlen_ptr < 1)
	{
		MPIDI_FUNC_EXIT(MPID_STATE_MPIU_STR_ADD_BINARY_ARG);
		return MPIU_STR_FAIL;
	}

	/* add the flag */
	if (wcsstr(flag, MPIU_STR_SEPAR_STR) || wcsstr(flag, MPIU_STR_DELIM_STR) || flag[0] == MPIU_STR_QUOTE_CHAR)
	{
		num_chars = quoted_printf(*str_ptr, *maxlen_ptr, flag);
	}
	else
	{
		num_chars = MPIU_Snprintf(*str_ptr, *maxlen_ptr, _TRUNCATE, L"%s", flag);
	}
	*maxlen_ptr = *maxlen_ptr - num_chars;
	if (*maxlen_ptr < 1)
	{
		MPIU_DBG_MSG_S(OTHER,VERBOSE,
					   L"partial argument added to string: '%s'", *str_ptr);
		**str_ptr = L'\0';
		/*(*str_ptr)[num_chars-1] = L'\0';*/
		MPIDI_FUNC_EXIT(MPID_STATE_MPIU_STR_ADD_BINARY_ARG);
		return MPIU_STR_NOMEM;
	}
	*str_ptr = *str_ptr + num_chars;

	/* add the deliminator character */
	**str_ptr = MPIU_STR_DELIM_CHAR;
	*str_ptr = *str_ptr + 1;
	*maxlen_ptr = *maxlen_ptr - 1;

	/* add the value string */
	result = encode_buffer(*str_ptr, *maxlen_ptr, buffer, length, &num_chars);
	if (result != MPIU_STR_SUCCESS)
	{
		**orig_str_ptr = L'\0';
		MPIDI_FUNC_EXIT(MPID_STATE_MPIU_STR_ADD_BINARY_ARG);
		return result;
	}
	num_chars = num_chars * 2; /* the encoding function turns one source character into two destination characters */
	*str_ptr = *str_ptr + num_chars;
	*maxlen_ptr = *maxlen_ptr - num_chars;
	if (*maxlen_ptr < 2)
	{
		MPIU_DBG_MSG_S(OTHER,VERBOSE,
					   L"partial argument added to string: '%s'", *str_ptr);
		/*
		*str_ptr = *str_ptr - 1;
		**str_ptr = L'\0';
		*/
		**orig_str_ptr = L'\0';
		MPIDI_FUNC_EXIT(MPID_STATE_MPIU_STR_ADD_BINARY_ARG);
		return MPIU_STR_NOMEM;
	}
	
	/* add the trailing space */
	**str_ptr = MPIU_STR_SEPAR_CHAR;
	*str_ptr = *str_ptr + 1;
	**str_ptr = L'\0';
	*maxlen_ptr = *maxlen_ptr - 1;

	MPIDI_FUNC_EXIT(MPID_STATE_MPIU_STR_ADD_BINARY_ARG);
	return MPIU_STR_SUCCESS;
}

/*
int test_argstrings()
{
	wchar_t buffer[1024];
	wchar_t buffer2[1024];
	wchar_t key[100];
	wchar_t val[1024];
	int maxlen, maxlen2;
	wchar_t *iter, *iter2;
	int ibuf[3] = { 1, 2, 3 };
	int age;

	iter = buffer;
	maxlen = 1024;
	MPIU_Str_add_string_arg(&iter, &maxlen, L"name", L"David Ashton");
	MPIU_Str_add_string_arg(&iter, &maxlen, L"a b", L"c d");
	MPIU_Str_add_binary_arg(&iter, &maxlen, L"array", ibuf, 12);
	MPIU_Str_add_string_arg(&iter, &maxlen, L"=", L"equals");
	MPIU_Str_add_string_arg(&iter, &maxlen, L"\"help=L", "=\"");
	MPIU_Str_add_int_arg(&iter, &maxlen, L"age", 123);

	wprintf(L"encoded buffer: <%s>\n", buffer);

	MPIU_Str_get_string_arg(buffer, L"name", val, 1024);
	wprintf(L"name = '%s'\n", val);
	MPIU_Str_get_string_arg(buffer, L"a b", val, 1024);
	wprintf(L"a b = '%s'\n", val);
	MPIU_Str_get_string_arg(buffer, L"=", val, 1024);
	wprintf(L"= = '%s'\n", val);
	MPIU_Str_get_string_arg(buffer, L"\"help=", val, 1024);
	wprintf(L"\"help= = L'%s'\n", val);
	MPIU_Str_get_int_arg(buffer, L"age", &age);
	wprintf(L"age = %d\n", age);
	MPIU_Str_get_binary_arg(buffer, L"array", ibuf, 12);
	wprintf(L"ibuf[0] = %d\n", ibuf[0]);
	wprintf(L"ibuf[1] = %d\n", ibuf[1]);
	wprintf(L"ibuf[2] = %d\n", ibuf[2]);

	iter = buffer2;
	maxlen = 1024;
	MPIU_Str_add_string_arg(&iter, &maxlen, L"big arg", buffer);
	MPIU_Str_add_string(&iter, &maxlen, L"Once isn't enough for me");
	MPIU_Str_add_int_arg(&iter, &maxlen, L"size", 100);
	wprintf(L"doubly encoded buffer: <%s>\n", buffer2);

	MPIU_Str_get_string_arg(buffer2, L"big arg", buffer, 1024);
	wprintf(L"big arg = '%s'\n", buffer);
	MPIU_Str_get_string_arg(buffer, L"name", val, 1024);
	wprintf(L"name = '%s'\n", val);
	MPIU_Str_get_string_arg(buffer, L"a b", val, 1024);
	wprintf(L"a b = '%s'\n", val);
	MPIU_Str_get_string_arg(buffer, L"=", val, 1024);
	wprintf(L"= = '%s'\n", val);
	MPIU_Str_get_int_arg(buffer, L"age", &age);
	wprintf(L"age = %d\n", age);
	MPIU_Str_get_binary_arg(buffer, L"array", ibuf, 12);
	wprintf(L"ibuf[0] = %d\n", ibuf[0]);
	wprintf(L"ibuf[1] = %d\n", ibuf[1]);
	wprintf(L"ibuf[2] = %d\n", ibuf[2]);

	MPIU_Str_get_int_arg(buffer2, L"size", &age);
	wprintf(L"size = %d\n", age);
	return 0;
}
*/
