/*   Copyright (C) 2009 Deino Software. */

#include "mpiimpl.h"
#if defined(USE_GETTEXT)
#include <libintl.h>
#endif

#undef FUNCNAME
#define FUNCNAME MPIU_Usage_printf
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIU_Usage_printf(const wchar_t *str, ...)
{
	int n;
	va_list list;
	const wchar_t *format_str;
	MPIDI_STATE_DECL(MPID_STATE_MPIU_USAGE_PRINTF);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIU_USAGE_PRINTF);

	va_start(list, str);
#ifdef USE_GETTEXT
	/* Category is LC_MESSAGES */
	format_str = dgettext( L"mpich", str );
	if (!format_str) format_str = str;
#else
	format_str = str;
#endif
	n = vwprintf(format_str, list);
	va_end(list);

	fflush(stdout);

	MPIDI_FUNC_EXIT(MPID_STATE_MPIU_USAGE_PRINTF);
	return n;
}

#undef FUNCNAME
#define FUNCNAME MPIU_Error_printf
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIU_Error_printf(const wchar_t *str, ...)
{
	int n;
	va_list list;
	const wchar_t *format_str;
	MPIDI_STATE_DECL(MPID_STATE_MPIU_ERROR_PRINTF);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIU_ERROR_PRINTF);

	va_start(list, str);
#ifdef USE_GETTEXT
	/* Category is LC_MESSAGES */
	format_str = dgettext( L"mpich", str );
	if (!format_str) format_str = str;
#else
	format_str = str;
#endif
	n = vfwprintf(stderr, format_str, list);
	va_end(list);

	fflush(stderr);

	MPIDI_FUNC_EXIT(MPID_STATE_MPIU_ERROR_PRINTF);
	return n;
}

#undef FUNCNAME
#define FUNCNAME MPIU_Internal_error_printf
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIU_Internal_error_printf(const wchar_t *str, ...)
{
	int n;
	va_list list;
	const wchar_t *format_str;
	MPIDI_STATE_DECL(MPID_STATE_MPIU_INTERNAL_ERROR_PRINTF);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIU_INTERNAL_ERROR_PRINTF);

	va_start(list, str);
#ifdef USE_GETTEXT
	/* Category is LC_MESSAGES */
	format_str = dgettext( L"mpich", str );
	if (!format_str) format_str = str;
#else
	format_str = str;
#endif
	n = vfwprintf(stderr, format_str, list);
	va_end(list);

	fflush(stderr);

	MPIDI_FUNC_EXIT(MPID_STATE_MPIU_INTERNAL_ERROR_PRINTF);
	return n;
}

/* Like internal_error_printf, but for the system routine name with 
   errno errnum.  Str may be null */
#undef FUNCNAME
#define FUNCNAME MPIU_Internal_sys_error_printf
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIU_Internal_sys_error_printf(const wchar_t *name, int errnum, 
								   const wchar_t *str, ...)
{
	int n = 0;
	va_list list;
	const wchar_t *format_str=0;
	wchar_t error_msg[1024];
	MPIDI_STATE_DECL(MPID_STATE_MPIU_INTERNAL_SYS_ERROR_PRINTF);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIU_INTERNAL_SYS_ERROR_PRINTF);

	/* Prepend information on the system error */
#ifdef HAVE_STRERROR
#ifdef USE_GETTEXT
	/* Category is LC_MESSAGES */
	format_str = dgettext( L"mpich", L"Error in system call %s: %s" );
#endif
	if (!format_str) format_str = L"Error in system call %s: %s\n";

	_wcserror_s(error_msg, 1024, errnum);
	fwprintf( stderr, format_str, name, error_msg );
#else
#ifdef USE_GETTEXT
	/* Category is LC_MESSAGES */
	format_str = dgettext( L"mpich", L"Error in system call %s errno = %d" );
#endif
	if (!format_str) format_str = L"Error in system call %s errno = %d\n";

	fwprintf( stderr, L"Error in %s: errno = %d\n", name, errnum );
#endif

	/* Now add the message that is specific to this use, if any */
	if (str) {
		va_start(list, str);
#ifdef USE_GETTEXT
		/* Category is LC_MESSAGES */
		format_str = dgettext( L"mpich", str );
		if (!format_str) format_str = str;
#else
		format_str = str;
#endif
		n = vfwprintf(stderr, format_str, list);
		va_end(list);
	}

	fflush(stderr);

	MPIDI_FUNC_EXIT(MPID_STATE_MPIU_INTERNAL_SYS_ERROR_PRINTF);
	return n;
}

#undef FUNCNAME
#define FUNCNAME MPIU_Msg_printf
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIU_Msg_printf(const wchar_t *str, ...)
{
	int n;
	va_list list;
	const wchar_t *format_str;
	MPIDI_STATE_DECL(MPID_STATE_MPIU_MSG_PRINTF);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIU_MSG_PRINTF);

	va_start(list, str);
#ifdef USE_GETTEXT
	/* Category is LC_MESSAGES */
	format_str = dgettext( L"mpich", str );
	if (!format_str) format_str = str;
#else
	format_str = str;
#endif
	n = vfwprintf(stdout, format_str, list);
	va_end(list);

	fflush(stderr);

	MPIDI_FUNC_EXIT(MPID_STATE_MPIU_MSG_PRINTF);
	return n;
}
