/*   Copyright (C) 2009 Deino Software. */ 

#include "mpiimpl.h"
#ifdef PROFILE
#include "mpe.h"
#endif
#include <io.h>

#undef FUNCNAME
#define FUNCNAME ADIOI_GEN_WriteStrided_naive
int ADIOI_GEN_WriteStrided_naive(MPID_File_ptr fd, void *buf, int count, MPI_Datatype buftype, int file_ptr_type, MPI_Offset offset, MPI_Status *status)
{
	/* offset is in units of etype relative to the filetype. */
	static const wchar_t FCNAME[] = L"ADIOI_GEN_WriteStrided_naive";
	int mpi_errno = MPI_SUCCESS;
	ADIOI_Flatlist_node *flat_buf, *flat_file;
	/* bwr == buffer write; fwr == file write */
	int bwr_size, fwr_size=0, b_index;
	int bufsize, size, sum, n_etypes_in_filetype, size_in_filetype;
	int n_filetypes, etype_in_filetype;
	MPI_Offset abs_off_in_filetype=0;
	int filetype_size, etype_size, buftype_size, req_len;
	MPI_Aint filetype_extent, buftype_extent; 
	int buf_count, buftype_is_contig, filetype_is_contig;
	MPI_Offset userbuf_off;
	MPI_Offset off, req_off, disp, end_offset=0, start_off;
	MPI_Status status1;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_GEN_WRITESTRIDED_NAIVE);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_GEN_WRITESTRIDED_NAIVE);

	mpi_errno = ADIOI_Datatype_iscontig(buftype, &buftype_is_contig);
	MPIU_ERR_CHK_POP(mpi_errno);
	mpi_errno = ADIOI_Datatype_iscontig(fd->filetype, &filetype_is_contig);
	MPIU_ERR_CHK_POP(mpi_errno);

	NMPI_Type_size(mpi_errno, (fd->filetype, &filetype_size));
	MPIU_ERR_CHK_POP(mpi_errno);
	if ( ! filetype_size )
	{
		goto fn_exit;
	}

	NMPI_Type_extent(mpi_errno, (fd->filetype, &filetype_extent));
	MPIU_ERR_CHK_POP(mpi_errno);
	NMPI_Type_size(mpi_errno, (buftype, &buftype_size));
	MPIU_ERR_CHK_POP(mpi_errno);
	NMPI_Type_extent(mpi_errno, (buftype, &buftype_extent));
	MPIU_ERR_CHK_POP(mpi_errno);
	etype_size = fd->etype_size;

	bufsize = buftype_size * count;

	/* contiguous in buftype and filetype is handled elsewhere */

	if (!buftype_is_contig && filetype_is_contig)
	{
		int b_count;
		/* noncontiguous in memory, contiguous in file. */

		mpi_errno = ADIOI_Flatten_datatype(buftype);
		MPIU_ERR_CHK_POP(mpi_errno);
		flat_buf = ADIOI_Flatlist;
		while (flat_buf->type != buftype)
		{
			flat_buf = flat_buf->next;
		}

		off = (file_ptr_type == ADIO_INDIVIDUAL) ? fd->fp_ind : fd->disp + etype_size * offset;

		start_off = off;
		end_offset = off + bufsize - 1;

		/* if atomicity is true, lock (exclusive) the region to be accessed */
		if ((fd->atomicity) && (fd->file_system != ADIO_PIOFS) && 
			(fd->file_system != ADIO_PVFS))
		{
			ADIOI_WRITE_LOCK(fd, start_off, SEEK_SET, end_offset-start_off+1);
		}

		/* for each region in the buffer, grab the data and put it in place */
		for (b_count=0; b_count < count; b_count++)
		{
			for (b_index=0; b_index < flat_buf->count; b_index++)
			{
				userbuf_off = b_count*buftype_extent + 
					flat_buf->indices[b_index];
				req_off = off;
				req_len = flat_buf->blocklens[b_index];

				mpi_errno = ADIO_WriteContig(fd, (char *) buf + userbuf_off, req_len, MPI_BYTE, ADIO_EXPLICIT_OFFSET, req_off, &status1);
				MPIU_ERR_CHK_POP(mpi_errno);

				/* off is (potentially) used to save the final offset later */
				off += flat_buf->blocklens[b_index];
			}
		}

		if ((fd->atomicity) && (fd->file_system != ADIO_PIOFS) && 
			(fd->file_system != ADIO_PVFS))
		{
			ADIOI_UNLOCK(fd, start_off, SEEK_SET, end_offset-start_off+1);
		}

		if (file_ptr_type == ADIO_INDIVIDUAL) fd->fp_ind = off;

	}
	else
	{
		/* noncontiguous in file */
		int f_index, st_fwr_size, st_index = 0, st_n_filetypes;
		int flag;

		/* First we're going to calculate a set of values for use in all
		* the noncontiguous in file cases:
		* start_off - starting byte position of data in file
		* end_offset - last byte offset to be acessed in the file
		* st_n_filetypes - how far into the file we start in terms of
		*				  whole filetypes
		* st_index - index of block in first filetype that we will be
		*			starting in (?)
		* st_fwr_size - size of the data in the first filetype block
		*			   that we will write (accounts for being part-way
		*			   into writing this block of the filetype
		*
		*/

		/* filetype already flattened in ADIO_Open */
		flat_file = ADIOI_Flatlist;
		while (flat_file->type != fd->filetype) flat_file = flat_file->next;
		disp = fd->disp;

		if (file_ptr_type == ADIO_INDIVIDUAL)
		{
			start_off = fd->fp_ind; /* in bytes */
			n_filetypes = -1;
			flag = 0;
			while (!flag)
			{
				n_filetypes++;
				for (f_index=0; f_index < flat_file->count; f_index++)
				{
					if (disp + flat_file->indices[f_index] + 
						(MPI_Offset) n_filetypes*filetype_extent + 
						flat_file->blocklens[f_index] >= start_off) 
					{
						/* this block contains our starting position */

						st_index = f_index;
						fwr_size = (int) (disp + flat_file->indices[f_index] + 
							(MPI_Offset) n_filetypes*filetype_extent + 
							flat_file->blocklens[f_index] - start_off);
						flag = 1;
						break;
					}
				}
			}
		}
		else
		{
			n_etypes_in_filetype = filetype_size/etype_size;
			n_filetypes = (int) (offset / n_etypes_in_filetype);
			etype_in_filetype = (int) (offset % n_etypes_in_filetype);
			size_in_filetype = etype_in_filetype * etype_size;

			sum = 0;
			for (f_index=0; f_index < flat_file->count; f_index++)
			{
				sum += flat_file->blocklens[f_index];
				if (sum > size_in_filetype)
				{
					st_index = f_index;
					fwr_size = sum - size_in_filetype;
					abs_off_in_filetype = flat_file->indices[f_index] +
						size_in_filetype - 
						(sum - flat_file->blocklens[f_index]);
					break;
				}
			}

			/* abs. offset in bytes in the file */
			start_off = disp + (MPI_Offset) n_filetypes*filetype_extent + abs_off_in_filetype;
		}

		st_fwr_size = fwr_size;
		st_n_filetypes = n_filetypes;

		/* start_off, st_n_filetypes, st_index, and st_fwr_size are 
		* all calculated at this point
		*/

		/* Calculate end_offset, the last byte-offset that will be accessed.
		* e.g., if start_off=0 and 100 bytes to be written, end_offset=99
		*/
		userbuf_off = 0;
		f_index = st_index;
		off = start_off;
		fwr_size = MPIDU_MIN(st_fwr_size, bufsize);
		while (userbuf_off < bufsize)
		{
			userbuf_off += fwr_size;
			end_offset = off + fwr_size - 1;

			if (f_index < (flat_file->count - 1))
			{
				f_index++;
			}
			else
			{
				f_index = 0;
				n_filetypes++;
			}

			off = disp + flat_file->indices[f_index] + (MPI_Offset) n_filetypes*filetype_extent;
			fwr_size = MPIDU_MIN(flat_file->blocklens[f_index], bufsize-(int)userbuf_off);
		}

		/* End of calculations.  At this point the following values have
		* been calculated and are ready for use:
		* - start_off
		* - end_offset
		* - st_n_filetypes
		* - st_index
		* - st_fwr_size
		*/

		/* if atomicity is true, lock (exclusive) the region to be accessed */
		if ((fd->atomicity) && (fd->file_system != ADIO_PIOFS) && 
			(fd->file_system != ADIO_PVFS))
		{
			ADIOI_WRITE_LOCK(fd, start_off, SEEK_SET, end_offset-start_off+1);
		}

		if (buftype_is_contig && !filetype_is_contig)
		{
			/* contiguous in memory, noncontiguous in file. should be the most common case. */

			userbuf_off = 0;
			f_index = st_index;
			off = start_off;
			n_filetypes = st_n_filetypes;
			fwr_size = MPIDU_MIN(st_fwr_size, bufsize);

			/* while there is still space in the buffer, write more data */
			while (userbuf_off < bufsize)
			{
				if (fwr_size)
				{
					/* TYPE_UB and TYPE_LB can result in 
					fwr_size = 0. save system call in such cases */ 
					req_off = off;
					req_len = fwr_size;

					mpi_errno = ADIO_WriteContig(fd, (char *) buf + userbuf_off, req_len, MPI_BYTE, ADIO_EXPLICIT_OFFSET, req_off, &status1);
					MPIU_ERR_CHK_POP(mpi_errno);
				}
				userbuf_off += fwr_size;

				if (off + fwr_size < disp + flat_file->indices[f_index] + flat_file->blocklens[f_index] + (MPI_Offset) n_filetypes*filetype_extent)
				{
					/* important that this value be correct, as it is
					* used to set the offset in the fd near the end of
					* this function.
					*/
					off += fwr_size;
				}
				/* did not reach end of contiguous block in filetype.
				* no more I/O needed. off is incremented by fwr_size.
				*/
				else
				{
					if (f_index < (flat_file->count - 1))
					{
						f_index++;
					}
					else
					{
						f_index = 0;
						n_filetypes++;
					}
					off = disp + flat_file->indices[f_index] + (MPI_Offset) n_filetypes*filetype_extent;
					fwr_size = MPIDU_MIN(flat_file->blocklens[f_index], bufsize-(int)userbuf_off);
				}
			}
		}
		else
		{
			int i, tmp_bufsize = 0;
			/* noncontiguous in memory as well as in file */

			mpi_errno = ADIOI_Flatten_datatype(buftype);
			MPIU_ERR_CHK_POP(mpi_errno);
			flat_buf = ADIOI_Flatlist;
			while (flat_buf->type != buftype)
			{
				flat_buf = flat_buf->next;
			}

			b_index = buf_count = 0;
			i = (int) (flat_buf->indices[0]);
			f_index = st_index;
			off = start_off;
			n_filetypes = st_n_filetypes;
			fwr_size = st_fwr_size;
			bwr_size = flat_buf->blocklens[0];

			/* while we haven't read size * count bytes, keep going */
			while (tmp_bufsize < bufsize)
			{
				int new_bwr_size = bwr_size, new_fwr_size = fwr_size;

				size = MPIDU_MIN(fwr_size, bwr_size);
				if (size)
				{
					req_off = off;
					req_len = size;
					userbuf_off = i;

					mpi_errno = ADIO_WriteContig(fd, (char *) buf + userbuf_off, req_len, MPI_BYTE, ADIO_EXPLICIT_OFFSET, req_off, &status1);
					MPIU_ERR_CHK_POP(mpi_errno);
				}

				if (size == fwr_size)
				{
					/* reached end of contiguous block in file */
					if (f_index < (flat_file->count - 1))
					{
						f_index++;
					}
					else
					{
						f_index = 0;
						n_filetypes++;
					}

					off = disp + flat_file->indices[f_index] + (MPI_Offset) n_filetypes*filetype_extent;

					new_fwr_size = flat_file->blocklens[f_index];
					if (size != bwr_size)
					{
						i += size;
						new_bwr_size -= size;
					}
				}

				if (size == bwr_size)
				{
					/* reached end of contiguous block in memory */

					b_index = (b_index + 1)%flat_buf->count;
					buf_count++;
					i = (int) (buftype_extent*(buf_count/flat_buf->count) + flat_buf->indices[b_index]);
					new_bwr_size = flat_buf->blocklens[b_index];
					if (size != fwr_size)
					{
						off += size;
						new_fwr_size -= size;
					}
				}
				tmp_bufsize += size;
				fwr_size = new_fwr_size;
				bwr_size = new_bwr_size;
			}
		}

		/* unlock the file region if we locked it */
		if ((fd->atomicity) && (fd->file_system != ADIO_PIOFS) && (fd->file_system != ADIO_PVFS))
		{
			ADIOI_UNLOCK(fd, start_off, SEEK_SET, end_offset-start_off+1);
		}

		if (file_ptr_type == ADIO_INDIVIDUAL) fd->fp_ind = off;
	} /* end of (else noncontiguous in file) */

	fd->fp_sys_posn = -1;   /* mark it as invalid. */

#ifdef HAVE_STATUS_SET_BYTES
	MPIR_Status_set_bytes(status, buftype, bufsize);
	/* This is a temporary way of filling in status. The right way is to 
	* keep track of how much data was actually written and placed in buf 
	*/
#endif

	if (!buftype_is_contig)
	{
		mpi_errno = ADIOI_Delete_flattened(buftype);
		MPIU_ERR_CHK_POP(mpi_errno);
	}

fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_GEN_WRITESTRIDED_NAIVE);
	return mpi_errno;

fn_fail:
	/* cleanup? */
	goto fn_exit;
}

/*
#define ADIOI_BUFFERED_WRITE \
{ \
if (req_off >= writebuf_off + writebuf_len) { \
if (writebuf_len) { \
ADIO_WriteContig(fd, writebuf, writebuf_len, MPI_BYTE, \
ADIO_EXPLICIT_OFFSET, writebuf_off, &status1, error_code); \
if (!(fd->atomicity)) ADIOI_UNLOCK(fd, writebuf_off, SEEK_SET, writebuf_len); \
if (*error_code != MPI_SUCCESS) { \
*error_code = MPIO_Err_create_code(*error_code, \
MPIR_ERR_RECOVERABLE, myname, \
__LINE__, MPI_ERR_IO, \
L"**iowswc", 0); \
return; \
} \
} \
writebuf_off = req_off; \
writebuf_len = (int) (MPIDU_MIN(max_bufsize,end_offset-writebuf_off+1));\
if (!(fd->atomicity)) ADIOI_WRITE_LOCK(fd, writebuf_off, SEEK_SET, writebuf_len); \
ADIO_ReadContig(fd, writebuf, writebuf_len, MPI_BYTE, \
ADIO_EXPLICIT_OFFSET, writebuf_off, &status1, error_code); \
if (*error_code != MPI_SUCCESS) { \
*error_code = MPIO_Err_create_code(*error_code, \
MPIR_ERR_RECOVERABLE, myname, \
__LINE__, MPI_ERR_IO, \
L"**iowsrc", 0); \
return; \
} \
} \
write_sz = (int) (MPIDU_MIN(req_len, writebuf_off + writebuf_len - req_off)); \
memcpy(writebuf+req_off-writebuf_off, (char *)buf +userbuf_off, write_sz);\
while (write_sz != req_len) { \
ADIO_WriteContig(fd, writebuf, writebuf_len, MPI_BYTE, \
ADIO_EXPLICIT_OFFSET, writebuf_off, &status1, error_code); \
if (!(fd->atomicity)) ADIOI_UNLOCK(fd, writebuf_off, SEEK_SET, writebuf_len); \
if (*error_code != MPI_SUCCESS) { \
*error_code = MPIO_Err_create_code(*error_code, \
MPIR_ERR_RECOVERABLE, myname, \
__LINE__, MPI_ERR_IO, \
L"**iowswc", 0); \
return; \
} \
req_len -= write_sz; \
userbuf_off += write_sz; \
writebuf_off += writebuf_len; \
writebuf_len = (int) (MPIDU_MIN(max_bufsize,end_offset-writebuf_off+1));\
if (!(fd->atomicity)) ADIOI_WRITE_LOCK(fd, writebuf_off, SEEK_SET, writebuf_len); \
ADIO_ReadContig(fd, writebuf, writebuf_len, MPI_BYTE, \
ADIO_EXPLICIT_OFFSET, writebuf_off, &status1, error_code); \
if (*error_code != MPI_SUCCESS) { \
*error_code = MPIO_Err_create_code(*error_code, \
MPIR_ERR_RECOVERABLE, myname, \
__LINE__, MPI_ERR_IO, \
L"**iowsrc", 0); \
return; \
} \
write_sz = MPIDU_MIN(req_len, writebuf_len); \
memcpy(writebuf, (char *)buf + userbuf_off, write_sz);\
} \
}
*/

#if 0
#undef FUNCNAME
#define FUNCNAME ADIOI_BUFFERED_WRITE
int ADIOI_BUFFERED_WRITE()
{
	static const wchar_t FCNAME[] = L"ADIOI_BUFFERED_WRITE";
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_BUFFERED_WRITE);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_BUFFERED_WRITE);

	if (req_off >= writebuf_off + writebuf_len)
	{
		if (writebuf_len)
		{
			mpi_errno = ADIO_WriteContig(fd, writebuf, writebuf_len, MPI_BYTE, ADIO_EXPLICIT_OFFSET, writebuf_off, &status1);
			if (!(fd->atomicity))
			{
				ADIOI_UNLOCK(fd, writebuf_off, SEEK_SET, writebuf_len);
			}
			MPIU_ERR_CHK_POP(mpi_errno); /*MPI_ERR_IO, L"**iowswc"*/
		}
		writebuf_off = req_off;
		writebuf_len = (int) (MPIDU_MIN(max_bufsize, end_offset-writebuf_off+1));
		if (!(fd->atomicity))
		{
			ADIOI_WRITE_LOCK(fd, writebuf_off, SEEK_SET, writebuf_len);
		}
		mpi_errno = ADIO_ReadContig(fd, writebuf, writebuf_len, MPI_BYTE, ADIO_EXPLICIT_OFFSET, writebuf_off, &status1);
		MPIU_ERR_CHK_POP(mpi_errno); /*MPI_ERR_IO, L"**iowsrc"*/
	}
	write_sz = (int) (MPIDU_MIN(req_len, writebuf_off + writebuf_len - req_off));
	memcpy(writebuf+req_off-writebuf_off, (char *)buf +userbuf_off, write_sz);
	while (write_sz != req_len)
	{
		mpi_errno = ADIO_WriteContig(fd, writebuf, writebuf_len, MPI_BYTE, ADIO_EXPLICIT_OFFSET, writebuf_off, &status1);
		if (!(fd->atomicity))
		{
			ADIOI_UNLOCK(fd, writebuf_off, SEEK_SET, writebuf_len);
		}
		MPIU_ERR_CHK_POP(mpi_errno); /*MPI_ERR_IO, L"**iowswc"*/
		req_len -= write_sz;
		userbuf_off += write_sz;
		writebuf_off += writebuf_len;
		writebuf_len = (int) (MPIDU_MIN(max_bufsize, end_offset-writebuf_off+1));
		if (!(fd->atomicity))
		{
			ADIOI_WRITE_LOCK(fd, writebuf_off, SEEK_SET, writebuf_len);
		}
		mpi_errno = ADIO_ReadContig(fd, writebuf, writebuf_len, MPI_BYTE, ADIO_EXPLICIT_OFFSET, writebuf_off, &status1);
		MPIU_ERR_CHK_POP(mpi_errno); /*MPI_ERR_IO, L"**iowsrc"*/
		write_sz = MPIDU_MIN(req_len, writebuf_len);
		memcpy(writebuf, (char *)buf + userbuf_off, write_sz);
	}

fn_fail:
	return mpi_errno;
}
#endif

/* this macro is used when filetype is contig and buftype is not contig.
it does not do a read-modify-write and does not lock*/
/*
#define ADIOI_BUFFERED_WRITE_WITHOUT_READ \
{ \
if (req_off >= writebuf_off + writebuf_len) { \
ADIO_WriteContig(fd, writebuf, writebuf_len, MPI_BYTE, \
ADIO_EXPLICIT_OFFSET, writebuf_off, &status1, error_code); \
if (*error_code != MPI_SUCCESS) { \
*error_code = MPIO_Err_create_code(*error_code, \
MPIR_ERR_RECOVERABLE, myname, \
__LINE__, MPI_ERR_IO, \
L"**iowswc", 0); \
return; \
} \
writebuf_off = req_off; \
writebuf_len = (int) (MPIDU_MIN(max_bufsize,end_offset-writebuf_off+1));\
} \
write_sz = (int) (MPIDU_MIN(req_len, writebuf_off + writebuf_len - req_off)); \
memcpy(writebuf+req_off-writebuf_off, (char *)buf +userbuf_off, write_sz);\
while (write_sz != req_len) { \
ADIO_WriteContig(fd, writebuf, writebuf_len, MPI_BYTE, \
ADIO_EXPLICIT_OFFSET, writebuf_off, &status1, error_code); \
if (*error_code != MPI_SUCCESS) { \
*error_code = MPIO_Err_create_code(*error_code, \
MPIR_ERR_RECOVERABLE, myname, \
__LINE__, MPI_ERR_IO, \
L"**iowswc", 0); \
return; \
} \
req_len -= write_sz; \
userbuf_off += write_sz; \
writebuf_off += writebuf_len; \
writebuf_len = (int) (MPIDU_MIN(max_bufsize,end_offset-writebuf_off+1));\
write_sz = MPIDU_MIN(req_len, writebuf_len); \
memcpy(writebuf, (char *)buf + userbuf_off, write_sz);\
} \
}
*/

#if 0
#undef FUNCNAME
#define FUNCNAME ADIOI_BUFFERED_WRITE_WITHOUT_READ
int ADIOI_BUFFERED_WRITE_WITHOUT_READ()
{
	static const wchar_t FCNAME[] = L"ADIOI_BUFFERED_WRITE_WITHOUT_READ";
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_BUFFERED_WRITE_WITHOUT_READ);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_BUFFERED_WRITE_WITHOUT_READ);

	if (req_off >= writebuf_off + writebuf_len)
	{
		mpi_errno = ADIO_WriteContig(fd, writebuf, writebuf_len, MPI_BYTE, ADIO_EXPLICIT_OFFSET, writebuf_off, &status1);
		MPIU_ERR_CHK_POP(mpi_errno); /*MPI_ERR_IO, L"**iowswc"*/
		writebuf_off = req_off;
		writebuf_len = (int) (MPIDU_MIN(max_bufsize, end_offset-writebuf_off+1));
	}
	write_sz = (int) (MPIDU_MIN(req_len, writebuf_off + writebuf_len - req_off));
	memcpy(writebuf+req_off-writebuf_off, (char *)buf +userbuf_off, write_sz);
	while (write_sz != req_len)
	{
		mpi_errno = ADIO_WriteContig(fd, writebuf, writebuf_len, MPI_BYTE, ADIO_EXPLICIT_OFFSET, writebuf_off, &status1);
		MPIU_ERR_CHK_POP(mpi_errno); /*MPI_ERR_IO, L"**iowswc"*/
		req_len -= write_sz;
		userbuf_off += write_sz;
		writebuf_off += writebuf_len;
		writebuf_len = (int) (MPIDU_MIN(max_bufsize, end_offset-writebuf_off+1));
		write_sz = MPIDU_MIN(req_len, writebuf_len);
		memcpy(writebuf, (char *)buf + userbuf_off, write_sz);
	}

fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_BUFFERED_WRITE_WITHOUT_READ);
	return mpi_errno;
}
#endif

#undef FUNCNAME
#define FUNCNAME ADIOI_GEN_WriteStrided
int ADIOI_GEN_WriteStrided(MPID_File_ptr fd, void *buf, int count, MPI_Datatype datatype, int file_ptr_type, MPI_Offset offset, MPI_Status *status)
{
	/* offset is in units of etype relative to the filetype. */
	static const wchar_t FCNAME[] = L"ADIOI_GEN_WriteStrided";
	int mpi_errno = MPI_SUCCESS;
	ADIOI_Flatlist_node *flat_buf, *flat_file;
	int i, j, k, bwr_size, fwr_size=0, st_index=0;
	int bufsize, num, size, sum, n_etypes_in_filetype, size_in_filetype;
	int n_filetypes, etype_in_filetype;
	MPI_Offset abs_off_in_filetype=0;
	int filetype_size, etype_size, buftype_size, req_len;
	MPI_Aint filetype_extent, buftype_extent; 
	int buf_count, buftype_is_contig, filetype_is_contig;
	MPI_Offset userbuf_off;
	MPI_Offset off, req_off, disp, end_offset=0, writebuf_off, start_off;
	char *writebuf;
	int flag, st_fwr_size, st_n_filetypes, writebuf_len, write_sz;
	MPI_Status status1;
	int new_bwr_size, new_fwr_size, max_bufsize;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_GEN_WRITESTRIDED);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_GEN_WRITESTRIDED);

	if (fd->hints->ds_write == ADIOI_HINT_DISABLE)
	{
		/* if user has disabled data sieving on reads, use naive
		* approach instead.
		*/
		mpi_errno = ADIOI_GEN_WriteStrided_naive(fd, buf, count, datatype, file_ptr_type, offset, status);
		MPIU_ERR_CHK_POP(mpi_errno);
		goto fn_exit;
	}

	mpi_errno = ADIOI_Datatype_iscontig(datatype, &buftype_is_contig);
	MPIU_ERR_CHK_POP(mpi_errno);
	mpi_errno = ADIOI_Datatype_iscontig(fd->filetype, &filetype_is_contig);
	MPIU_ERR_CHK_POP(mpi_errno);

	NMPI_Type_size(mpi_errno, (fd->filetype, &filetype_size));
	MPIU_ERR_CHK_POP(mpi_errno);
	if ( ! filetype_size )
	{
		goto fn_exit;
	}

	NMPI_Type_extent(mpi_errno, (fd->filetype, &filetype_extent));
	MPIU_ERR_CHK_POP(mpi_errno);
	NMPI_Type_size(mpi_errno, (datatype, &buftype_size));
	MPIU_ERR_CHK_POP(mpi_errno);
	NMPI_Type_extent(mpi_errno, (datatype, &buftype_extent));
	MPIU_ERR_CHK_POP(mpi_errno);
	etype_size = fd->etype_size;

	bufsize = buftype_size * count;

	/* get max_bufsize from the info object. */

	max_bufsize = fd->hints->ind_wr_buffer_size;

	if (!buftype_is_contig && filetype_is_contig)
	{
		/* noncontiguous in memory, contiguous in file. */

		mpi_errno = ADIOI_Flatten_datatype(datatype);
		MPIU_ERR_CHK_POP(mpi_errno);
		flat_buf = ADIOI_Flatlist;
		while (flat_buf->type != datatype)
		{
			flat_buf = flat_buf->next;
		}

		off = (file_ptr_type == ADIO_INDIVIDUAL) ? fd->fp_ind : fd->disp + etype_size * offset;

		start_off = off;
		end_offset = off + bufsize - 1;
		writebuf_off = off;
		writebuf = (char *) MPIU_Malloc(max_bufsize);
		MPIU_ERR_MEM_CHK_POP(mpi_errno, writebuf, max_bufsize, L"writebuf");
		writebuf_len = (int) (MPIDU_MIN(max_bufsize, end_offset-writebuf_off+1));

		/* if atomicity is true, lock the region to be accessed */
		if (fd->atomicity) 
			ADIOI_WRITE_LOCK(fd, start_off, SEEK_SET, end_offset-start_off+1);

		for (j=0; j<count; j++)
		{
			for (i=0; i<flat_buf->count; i++)
			{
				userbuf_off = j*buftype_extent + flat_buf->indices[i];
				req_off = off;
				req_len = flat_buf->blocklens[i];
				/*
				mpi_errno = ADIOI_BUFFERED_WRITE_WITHOUT_READ();
				MPIU_ERR_CHK_POP(mpi_errno);
				*/
				if (req_off >= writebuf_off + writebuf_len)
				{
					mpi_errno = ADIO_WriteContig(fd, writebuf, writebuf_len, MPI_BYTE, ADIO_EXPLICIT_OFFSET, writebuf_off, &status1);
					MPIU_ERR_CHK_POP(mpi_errno); /*MPI_ERR_IO, L"**iowswc"*/
					writebuf_off = req_off;
					writebuf_len = (int) (MPIDU_MIN(max_bufsize, end_offset-writebuf_off+1));
				}
				write_sz = (int) (MPIDU_MIN(req_len, writebuf_off + writebuf_len - req_off));
				memcpy(writebuf+req_off-writebuf_off, (char *)buf +userbuf_off, write_sz);
				while (write_sz != req_len)
				{
					mpi_errno = ADIO_WriteContig(fd, writebuf, writebuf_len, MPI_BYTE, ADIO_EXPLICIT_OFFSET, writebuf_off, &status1);
					MPIU_ERR_CHK_POP(mpi_errno); /*MPI_ERR_IO, L"**iowswc"*/
					req_len -= write_sz;
					userbuf_off += write_sz;
					writebuf_off += writebuf_len;
					writebuf_len = (int) (MPIDU_MIN(max_bufsize, end_offset-writebuf_off+1));
					write_sz = MPIDU_MIN(req_len, writebuf_len);
					memcpy(writebuf, (char *)buf + userbuf_off, write_sz);
				}
				off += flat_buf->blocklens[i];
			}
		}

		/* write the buffer out finally */
		mpi_errno = ADIO_WriteContig(fd, writebuf, writebuf_len, MPI_BYTE, ADIO_EXPLICIT_OFFSET, writebuf_off, &status1);
		if (fd->atomicity)
			ADIOI_UNLOCK(fd, start_off, SEEK_SET, end_offset-start_off+1);
		MPIU_ERR_CHK_POP(mpi_errno);

		MPIU_Free(writebuf); 

		if (file_ptr_type == ADIO_INDIVIDUAL) fd->fp_ind = off;
	}
	else
	{
		/* noncontiguous in file */

		/* filetype already flattened in ADIO_Open */
		flat_file = ADIOI_Flatlist;
		while (flat_file->type != fd->filetype)
		{
			flat_file = flat_file->next;
		}
		disp = fd->disp;

		if (file_ptr_type == ADIO_INDIVIDUAL)
		{
			offset = fd->fp_ind; /* in bytes */
			n_filetypes = -1;
			flag = 0;
			while (!flag)
			{
				n_filetypes++;
				for (i=0; i<flat_file->count; i++)
				{
					if (disp + flat_file->indices[i] + 
						(MPI_Offset) n_filetypes*filetype_extent + flat_file->blocklens[i] >= offset)
					{
						st_index = i;
						fwr_size = (int) (disp + flat_file->indices[i] + (MPI_Offset) n_filetypes*filetype_extent + flat_file->blocklens[i] - offset);
						flag = 1;
						break;
					}
				}
			}
		}
		else
		{
			n_etypes_in_filetype = filetype_size/etype_size;
			n_filetypes = (int) (offset / n_etypes_in_filetype);
			etype_in_filetype = (int) (offset % n_etypes_in_filetype);
			size_in_filetype = etype_in_filetype * etype_size;

			sum = 0;
			for (i=0; i<flat_file->count; i++)
			{
				sum += flat_file->blocklens[i];
				if (sum > size_in_filetype)
				{
					st_index = i;
					fwr_size = sum - size_in_filetype;
					abs_off_in_filetype = flat_file->indices[i] + size_in_filetype - (sum - flat_file->blocklens[i]);
					break;
				}
			}

			/* abs. offset in bytes in the file */
			offset = disp + (MPI_Offset) n_filetypes*filetype_extent + abs_off_in_filetype;
		}

		start_off = offset;

		/* Calculate end_offset, the last byte-offset that will be accessed.
		e.g., if start_offset=0 and 100 bytes to be write, end_offset=99*/

		st_fwr_size = fwr_size;
		st_n_filetypes = n_filetypes;
		i = 0;
		j = st_index;
		off = offset;
		fwr_size = MPIDU_MIN(st_fwr_size, bufsize);
		while (i < bufsize)
		{
			i += fwr_size;
			end_offset = off + fwr_size - 1;

			if (j < (flat_file->count - 1))
			{
				j++;
			}
			else
			{
				j = 0;
				n_filetypes++;
			}

			off = disp + flat_file->indices[j] + (MPI_Offset) n_filetypes*filetype_extent;
			fwr_size = MPIDU_MIN(flat_file->blocklens[j], bufsize-i);
		}

		/* if atomicity is true, lock the region to be accessed */
		if (fd->atomicity) 
			ADIOI_WRITE_LOCK(fd, start_off, SEEK_SET, end_offset-start_off+1);

		writebuf_off = 0;
		writebuf_len = 0;
		writebuf = (char *) MPIU_Malloc(max_bufsize);
		memset(writebuf, -1, max_bufsize);

		if (buftype_is_contig && !filetype_is_contig)
		{
			/* contiguous in memory, noncontiguous in file. should be the most common case. */

			i = 0;
			j = st_index;
			off = offset;
			n_filetypes = st_n_filetypes;
			fwr_size = MPIDU_MIN(st_fwr_size, bufsize);
			while (i < bufsize)
			{
				if (fwr_size)
				{
					/* TYPE_UB and TYPE_LB can result in 
					fwr_size = 0. save system call in such cases */ 
					/* lseek(fd->fd_sys, off, SEEK_SET);
					err = write(fd->fd_sys, ((char *) buf) + i, fwr_size);*/

					req_off = off;
					req_len = fwr_size;
					userbuf_off = i;
					/*
					mpi_errno = ADIOI_BUFFERED_WRITE();
					MPIU_ERR_CHK_POP(mpi_errno);
					*/
					if (req_off >= writebuf_off + writebuf_len)
					{
						if (writebuf_len)
						{
							mpi_errno = ADIO_WriteContig(fd, writebuf, writebuf_len, MPI_BYTE, ADIO_EXPLICIT_OFFSET, writebuf_off, &status1);
							if (!(fd->atomicity))
							{
								ADIOI_UNLOCK(fd, writebuf_off, SEEK_SET, writebuf_len);
							}
							MPIU_ERR_CHK_POP(mpi_errno); /*MPI_ERR_IO, L"**iowswc"*/
						}
						writebuf_off = req_off;
						writebuf_len = (int) (MPIDU_MIN(max_bufsize, end_offset-writebuf_off+1));
						if (!(fd->atomicity))
						{
							ADIOI_WRITE_LOCK(fd, writebuf_off, SEEK_SET, writebuf_len);
						}
						mpi_errno = ADIO_ReadContig(fd, writebuf, writebuf_len, MPI_BYTE, ADIO_EXPLICIT_OFFSET, writebuf_off, &status1);
						MPIU_ERR_CHK_POP(mpi_errno); /*MPI_ERR_IO, L"**iowsrc"*/
					}
					write_sz = (int) (MPIDU_MIN(req_len, writebuf_off + writebuf_len - req_off));
					memcpy(writebuf+req_off-writebuf_off, (char *)buf +userbuf_off, write_sz);
					while (write_sz != req_len)
					{
						mpi_errno = ADIO_WriteContig(fd, writebuf, writebuf_len, MPI_BYTE, ADIO_EXPLICIT_OFFSET, writebuf_off, &status1);
						if (!(fd->atomicity))
						{
							ADIOI_UNLOCK(fd, writebuf_off, SEEK_SET, writebuf_len);
						}
						MPIU_ERR_CHK_POP(mpi_errno); /*MPI_ERR_IO, L"**iowswc"*/
						req_len -= write_sz;
						userbuf_off += write_sz;
						writebuf_off += writebuf_len;
						writebuf_len = (int) (MPIDU_MIN(max_bufsize, end_offset-writebuf_off+1));
						if (!(fd->atomicity))
						{
							ADIOI_WRITE_LOCK(fd, writebuf_off, SEEK_SET, writebuf_len);
						}
						mpi_errno = ADIO_ReadContig(fd, writebuf, writebuf_len, MPI_BYTE, ADIO_EXPLICIT_OFFSET, writebuf_off, &status1);
						MPIU_ERR_CHK_POP(mpi_errno); /*MPI_ERR_IO, L"**iowsrc"*/
						write_sz = MPIDU_MIN(req_len, writebuf_len);
						memcpy(writebuf, (char *)buf + userbuf_off, write_sz);
					}
				}
				i += fwr_size;

				if (off + fwr_size < disp + flat_file->indices[j] +
					flat_file->blocklens[j] + (MPI_Offset) n_filetypes*filetype_extent)
				{
					off += fwr_size;
					/* did not reach end of contiguous block in filetype.
					no more I/O needed. off is incremented by fwr_size. */
				}
				else
				{
					if (j < (flat_file->count - 1)) j++;
					else
					{
						j = 0;
						n_filetypes++;
					}
					off = disp + flat_file->indices[j] + (MPI_Offset) n_filetypes*filetype_extent;
					fwr_size = MPIDU_MIN(flat_file->blocklens[j], bufsize-i);
				}
			}
		}
		else
		{
			/* noncontiguous in memory as well as in file */

			mpi_errno = ADIOI_Flatten_datatype(datatype);
			MPIU_ERR_CHK_POP(mpi_errno);
			flat_buf = ADIOI_Flatlist;
			while (flat_buf->type != datatype)
			{
				flat_buf = flat_buf->next;
			}

			k = num = buf_count = 0;
			i = (int) (flat_buf->indices[0]);
			j = st_index;
			off = offset;
			n_filetypes = st_n_filetypes;
			fwr_size = st_fwr_size;
			bwr_size = flat_buf->blocklens[0];

			while (num < bufsize)
			{
				size = MPIDU_MIN(fwr_size, bwr_size);
				if (size)
				{
					/* lseek(fd->fd_sys, off, SEEK_SET);
					err = write(fd->fd_sys, ((char *) buf) + i, size); */

					req_off = off;
					req_len = size;
					userbuf_off = i;
					/*
					mpi_errno = ADIOI_BUFFERED_WRITE();
					MPIU_ERR_CHK_POP(mpi_errno);
					*/
					if (req_off >= writebuf_off + writebuf_len)
					{
						if (writebuf_len)
						{
							mpi_errno = ADIO_WriteContig(fd, writebuf, writebuf_len, MPI_BYTE, ADIO_EXPLICIT_OFFSET, writebuf_off, &status1);
							if (!(fd->atomicity))
							{
								ADIOI_UNLOCK(fd, writebuf_off, SEEK_SET, writebuf_len);
							}
							MPIU_ERR_CHK_POP(mpi_errno); /*MPI_ERR_IO, L"**iowswc"*/
						}
						writebuf_off = req_off;
						writebuf_len = (int) (MPIDU_MIN(max_bufsize, end_offset-writebuf_off+1));
						if (!(fd->atomicity))
						{
							ADIOI_WRITE_LOCK(fd, writebuf_off, SEEK_SET, writebuf_len);
						}
						mpi_errno = ADIO_ReadContig(fd, writebuf, writebuf_len, MPI_BYTE, ADIO_EXPLICIT_OFFSET, writebuf_off, &status1);
						MPIU_ERR_CHK_POP(mpi_errno); /*MPI_ERR_IO, L"**iowsrc"*/
					}
					write_sz = (int) (MPIDU_MIN(req_len, writebuf_off + writebuf_len - req_off));
					memcpy(writebuf+req_off-writebuf_off, (char *)buf +userbuf_off, write_sz);
					while (write_sz != req_len)
					{
						mpi_errno = ADIO_WriteContig(fd, writebuf, writebuf_len, MPI_BYTE, ADIO_EXPLICIT_OFFSET, writebuf_off, &status1);
						if (!(fd->atomicity))
						{
							ADIOI_UNLOCK(fd, writebuf_off, SEEK_SET, writebuf_len);
						}
						MPIU_ERR_CHK_POP(mpi_errno); /*MPI_ERR_IO, L"**iowswc"*/
						req_len -= write_sz;
						userbuf_off += write_sz;
						writebuf_off += writebuf_len;
						writebuf_len = (int) (MPIDU_MIN(max_bufsize, end_offset-writebuf_off+1));
						if (!(fd->atomicity))
						{
							ADIOI_WRITE_LOCK(fd, writebuf_off, SEEK_SET, writebuf_len);
						}
						mpi_errno = ADIO_ReadContig(fd, writebuf, writebuf_len, MPI_BYTE, ADIO_EXPLICIT_OFFSET, writebuf_off, &status1);
						MPIU_ERR_CHK_POP(mpi_errno); /*MPI_ERR_IO, L"**iowsrc"*/
						write_sz = MPIDU_MIN(req_len, writebuf_len);
						memcpy(writebuf, (char *)buf + userbuf_off, write_sz);
					}
				}

				new_fwr_size = fwr_size;
				new_bwr_size = bwr_size;

				if (size == fwr_size)
				{
					/* reached end of contiguous block in file */
					if (j < (flat_file->count - 1))
					{
						j++;
					}
					else
					{
						j = 0;
						n_filetypes++;
					}

					off = disp + flat_file->indices[j] + (MPI_Offset) n_filetypes*filetype_extent;

					new_fwr_size = flat_file->blocklens[j];
					if (size != bwr_size)
					{
						i += size;
						new_bwr_size -= size;
					}
				}

				if (size == bwr_size)
				{
					/* reached end of contiguous block in memory */

					k = (k + 1)%flat_buf->count;
					buf_count++;
					i = (int) (buftype_extent*(buf_count/flat_buf->count) + flat_buf->indices[k]); 
					new_bwr_size = flat_buf->blocklens[k];
					if (size != fwr_size)
					{
						off += size;
						new_fwr_size -= size;
					}
				}
				num += size;
				fwr_size = new_fwr_size;
				bwr_size = new_bwr_size;
			}
		}

		/* write the buffer out finally */		
		if (writebuf_len)
		{
			mpi_errno = ADIO_WriteContig(fd, writebuf, writebuf_len, MPI_BYTE, ADIO_EXPLICIT_OFFSET, writebuf_off, &status1);
			if (!(fd->atomicity)) 
				ADIOI_UNLOCK(fd, writebuf_off, SEEK_SET, writebuf_len);
			MPIU_ERR_CHK_POP(mpi_errno);
		}
		if (fd->atomicity) 
			ADIOI_UNLOCK(fd, start_off, SEEK_SET, end_offset-start_off+1);

		MPIU_Free(writebuf); 

		if (file_ptr_type == ADIO_INDIVIDUAL) fd->fp_ind = off;
	}

	fd->fp_sys_posn = -1;   /* set it to null. */

#ifdef HAVE_STATUS_SET_BYTES
	MPIR_Status_set_bytes(status, datatype, bufsize);
	/* This is a temporary way of filling in status. The right way is to 
	keep track of how much data was actually written by ADIOI_BUFFERED_WRITE. */
#endif

	if (!buftype_is_contig)
	{
		mpi_errno = ADIOI_Delete_flattened(datatype);
		MPIU_ERR_CHK_POP(mpi_errno);
	}

fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_GEN_WRITESTRIDED);
	return mpi_errno;

fn_fail:
	/* clean up memory */
	goto fn_exit;
}

/* prototypes of functions used for collective writes only. */
static int ADIOI_Exch_and_write(MPID_File_ptr fd, void *buf, MPI_Datatype
								datatype, int nprocs, int myrank, ADIOI_Access
								*others_req, MPI_Offset *offset_list,
								int *len_list, int contig_access_count, MPI_Offset
								min_st_offset, MPI_Offset fd_size,
								MPI_Offset *fd_start, MPI_Offset *fd_end,
								int *buf_idx);
static int ADIOI_W_Exchange_data(MPID_File_ptr fd, void *buf, char *write_buf,
								 ADIOI_Flatlist_node *flat_buf, MPI_Offset 
								 *offset_list, int *len_list, int *send_size, 
								 int *recv_size, MPI_Offset off, int size,
								 int *count, int *start_pos, int *partial_recv, 
								 int *sent_to_proc, int nprocs, 
								 int myrank, int
								 buftype_is_contig, int contig_access_count,
								 MPI_Offset min_st_offset, MPI_Offset fd_size,
								 MPI_Offset *fd_start, MPI_Offset *fd_end, 
								 ADIOI_Access *others_req, 
								 int *send_buf_idx, int *curr_to_proc,
								 int *done_to_proc, int *hole, int iter, 
								 MPI_Aint buftype_extent, int *buf_idx);
static int ADIOI_Fill_send_buffer(MPID_File_ptr fd, void *buf, ADIOI_Flatlist_node
								  *flat_buf, char **send_buf, MPI_Offset 
								  *offset_list, int *len_list, int *send_size, 
								  MPI_Request *requests, int *sent_to_proc, 
								  int nprocs, int myrank, 
								  int contig_access_count, MPI_Offset
								  min_st_offset, MPI_Offset fd_size,
								  MPI_Offset *fd_start, MPI_Offset *fd_end, 
								  int *send_buf_idx, int *curr_to_proc, 
								  int *done_to_proc, int iter, 
								  MPI_Aint buftype_extent);
static int ADIOI_Heap_merge(ADIOI_Access *others_req, int *count, 
							MPI_Offset *srt_off, int *srt_len, int *start_pos,
							int nprocs, int nprocs_recv, int total_elements);


#undef FUNCNAME
#define FUNCNAME ADIOI_GEN_WriteStridedColl
int ADIOI_GEN_WriteStridedColl(MPID_File_ptr fd, void *buf, int count,
							   MPI_Datatype datatype, int file_ptr_type,
							   MPI_Offset offset, MPI_Status *status)
{
	/* Uses a generalized version of the extended two-phase method described
	in "An Extended Two-Phase Method for Accessing Sections of 
	Out-of-Core Arrays", Rajeev Thakur and Alok Choudhary,
	Scientific Programming, (5)4:301--317, Winter 1996. 
	http://www.mcs.anl.gov/home/thakur/ext2ph.ps */

	int mpi_errno = MPI_SUCCESS;
	static const wchar_t FCNAME[] = L"ADIOI_GEN_WriteStridedColl";
	ADIOI_Access *my_req; 
	/* array of nprocs access structures, one for each other process in
	whose file domain this process's request lies */

	ADIOI_Access *others_req;
	/* array of nprocs access structures, one for each other process
	whose request lies in this process's file domain. */

	int i, filetype_is_contig, nprocs, nprocs_for_coll, myrank;
	int contig_access_count=0, interleave_count = 0, buftype_is_contig;
	int *count_my_req_per_proc, count_my_req_procs, count_others_req_procs;
	MPI_Offset orig_fp, start_offset, end_offset, fd_size, min_st_offset, off;
	MPI_Offset *offset_list = NULL, *st_offsets = NULL, *fd_start = NULL,
		*fd_end = NULL, *end_offsets = NULL;
	int *buf_idx = NULL, *len_list = NULL;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_GEN_WRITESTRIDEDCOLL);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_GEN_WRITESTRIDEDCOLL);


#ifdef PROFILE
	MPE_Log_event(13, 0, L"start computation");
#endif

	NMPI_Comm_size(mpi_errno, (fd->comm, &nprocs));
	MPIU_ERR_CHK_POP(mpi_errno);
	NMPI_Comm_rank(mpi_errno, (fd->comm, &myrank));
	MPIU_ERR_CHK_POP(mpi_errno);

	/* the number of processes that actually perform I/O, nprocs_for_coll,
	* is stored in the hints off the MPID_File_ptr structure
	*/
	nprocs_for_coll = fd->hints->cb_nodes;
	orig_fp = fd->fp_ind;

	/* only check for interleaving if cb_write isn't disabled */
	if (fd->hints->cb_write != ADIOI_HINT_DISABLE)
	{
		/* For this process's request, calculate the list of offsets and
		lengths in the file and determine the start and end offsets. */

		/* Note: end_offset points to the last byte-offset that will be accessed.
		e.g., if start_offset=0 and 100 bytes to be read, end_offset=99*/

		mpi_errno = ADIOI_Calc_my_off_len(fd, count, datatype, file_ptr_type, offset,
			&offset_list, &len_list, &start_offset,
			&end_offset, &contig_access_count);
		MPIU_ERR_CHK_POP(mpi_errno);

		/* each process communicates its start and end offsets to other 
		processes. The result is an array each of start and end offsets stored
		in order of process rank. */ 

		st_offsets = (MPI_Offset *) MPIU_Malloc(nprocs*sizeof(MPI_Offset));
		MPIU_ERR_MEM_CHK_POP(mpi_errno, st_offsets, nprocs*sizeof(MPI_Offset), L"st_offsets")
			end_offsets = (MPI_Offset *) MPIU_Malloc(nprocs*sizeof(MPI_Offset));
		MPIU_ERR_MEM_CHK_POP(mpi_errno, end_offsets, nprocs*sizeof(MPI_Offset), L"end_offsets");

		NMPI_Allgather(mpi_errno, (&start_offset, 1, ADIO_OFFSET, st_offsets, 1, ADIO_OFFSET, fd->comm));
		MPIU_ERR_CHK_POP(mpi_errno);
		NMPI_Allgather(mpi_errno, (&end_offset, 1, ADIO_OFFSET, end_offsets, 1, ADIO_OFFSET, fd->comm));
		MPIU_ERR_CHK_POP(mpi_errno);

		/* are the accesses of different processes interleaved? */
		for (i=1; i<nprocs; i++)
		{
			if ((st_offsets[i] < end_offsets[i-1]) && 
				(st_offsets[i] <= end_offsets[i]))
			{
				interleave_count++;
			}
		}
		/* This is a rudimentary check for interleaving, but should suffice
		for the moment. */
	}

	mpi_errno = ADIOI_Datatype_iscontig(datatype, &buftype_is_contig);
	MPIU_ERR_CHK_POP(mpi_errno);

	if (fd->hints->cb_write == ADIOI_HINT_DISABLE || (!interleave_count && (fd->hints->cb_write == ADIOI_HINT_AUTO)))
	{
		/* use independent accesses */
		if (fd->hints->cb_write != ADIOI_HINT_DISABLE)
		{
			MPIU_Free(offset_list);
			MPIU_Free(len_list);
			MPIU_Free(st_offsets);
			MPIU_Free(end_offsets);
		}

		fd->fp_ind = orig_fp;
		mpi_errno = ADIOI_Datatype_iscontig(fd->filetype, &filetype_is_contig);
		MPIU_ERR_CHK_POP(mpi_errno);

		if (buftype_is_contig && filetype_is_contig)
		{
			if (file_ptr_type == ADIO_EXPLICIT_OFFSET)
			{
				off = fd->disp + (fd->etype_size) * offset;
				mpi_errno = ADIO_WriteContig(fd, buf, count, datatype, ADIO_EXPLICIT_OFFSET, off, status);
				MPIU_ERR_CHK_POP(mpi_errno);
			}
			else
			{
				mpi_errno = ADIO_WriteContig(fd, buf, count, datatype, ADIO_INDIVIDUAL, 0, status);
				MPIU_ERR_CHK_POP(mpi_errno);
			}
		}
		else
		{
			mpi_errno = ADIO_WriteStrided(fd, buf, count, datatype, file_ptr_type, offset, status);
			MPIU_ERR_CHK_POP(mpi_errno);
		}

		goto fn_exit;
	}

	/* Divide the I/O workload among "nprocs_for_coll" processes. This is
	done by (logically) dividing the file into file domains (FDs); each
	process may directly access only its own file domain. */

	mpi_errno = ADIOI_Calc_file_domains(st_offsets, end_offsets, nprocs,
		nprocs_for_coll, &min_st_offset,
		&fd_start, &fd_end, &fd_size);
	MPIU_ERR_CHK_POP(mpi_errno);


	/* calculate what portions of the access requests of this process are
	located in what file domains */

	mpi_errno = ADIOI_Calc_my_req(fd, offset_list, len_list, contig_access_count,
		min_st_offset, fd_start, fd_end, fd_size,
		nprocs, &count_my_req_procs, 
		&count_my_req_per_proc, &my_req,
		&buf_idx);
	MPIU_ERR_CHK_POP(mpi_errno);

	/* based on everyone's my_req, calculate what requests of other
	processes lie in this process's file domain.
	count_others_req_procs = number of processes whose requests lie in
	this process's file domain (including this process itself) 
	count_others_req_per_proc[i] indicates how many separate contiguous
	requests of proc. i lie in this process's file domain. */

	mpi_errno = ADIOI_Calc_others_req(fd, count_my_req_procs, 
		count_my_req_per_proc, my_req, 
		nprocs, myrank,
		&count_others_req_procs, &others_req);
	MPIU_ERR_CHK_POP(mpi_errno);

	MPIU_Free(count_my_req_per_proc);
	for (i=0; i < nprocs; i++)
	{
		if (my_req[i].count)
		{
			MPIU_Free(my_req[i].offsets);
			MPIU_Free(my_req[i].lens);
		}
	}
	MPIU_Free(my_req);
	my_req = NULL;

	/* exchange data and write in sizes of no more than coll_bufsize. */
	mpi_errno = ADIOI_Exch_and_write(fd, buf, datatype, nprocs, myrank,
		others_req, offset_list,
		len_list, contig_access_count, min_st_offset,
		fd_size, fd_start, fd_end, buf_idx);
	MPIU_ERR_CHK_POP(mpi_errno);

	if (!buftype_is_contig)
	{
		mpi_errno = ADIOI_Delete_flattened(datatype);
		MPIU_ERR_CHK_POP(mpi_errno);
	}

	/* free all memory allocated for collective I/O */

	for (i=0; i<nprocs; i++)
	{
		if (others_req[i].count)
		{
			MPIU_Free(others_req[i].offsets);
			MPIU_Free(others_req[i].lens);
			MPIU_Free(others_req[i].mem_ptrs);
		}
	}
	MPIU_Free(others_req);

	MPIU_Free(buf_idx);
	MPIU_Free(offset_list);
	MPIU_Free(len_list);
	MPIU_Free(st_offsets);
	MPIU_Free(end_offsets);
	MPIU_Free(fd_start);
	MPIU_Free(fd_end);

#ifdef HAVE_STATUS_SET_BYTES
	if (status)
	{
		int bufsize, size;
		/* Don't set status if it isn't needed */
		NMPI_Type_size(mpi_errno, (datatype, &size));
		MPIU_ERR_CHK_POP(mpi_errno);
		bufsize = size * count;
		MPIR_Status_set_bytes(status, datatype, bufsize);
	}
	/* This is a temporary way of filling in status. The right way is to 
	keep track of how much data was actually written during collective I/O. */
#endif

	fd->fp_sys_posn = -1;   /* set it to null. */

fn_exit:
fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_GEN_WRITESTRIDEDCOLL);
	return mpi_errno;
}



/* If successful, error_code is set to MPI_SUCCESS.  Otherwise an error
* code is created and returned in error_code.
*/
#undef FUNCNAME
#define FUNCNAME ADIOI_Exch_and_write
static int ADIOI_Exch_and_write(MPID_File_ptr fd, void *buf, MPI_Datatype
								datatype, int nprocs, int myrank,
								ADIOI_Access
								*others_req, MPI_Offset *offset_list,
								int *len_list, int contig_access_count,
								MPI_Offset
								min_st_offset, MPI_Offset fd_size,
								MPI_Offset *fd_start, MPI_Offset *fd_end,
								int *buf_idx)
{
	/* Send data to appropriate processes and write in sizes of no more
	than coll_bufsize.	
	The idea is to reduce the amount of extra memory required for
	collective I/O. If all data were written all at once, which is much
	easier, it would require temp space more than the size of user_buf,
	which is often unacceptable. For example, to write a distributed
	array to a file, where each local array is 8Mbytes, requiring
	at least another 8Mbytes of temp space is unacceptable. */

	static const wchar_t FCNAME[] = L"ADIOI_Exch_and_write";
	int mpi_errno = MPI_SUCCESS;
	int hole, i, j, m, size=0, ntimes, max_ntimes, buftype_is_contig;
	MPI_Offset st_loc=-1, end_loc=-1, off, done, req_off;
	char *write_buf=NULL;
	int *curr_offlen_ptr, *count, *send_size, req_len, *recv_size;
	int *partial_recv, *sent_to_proc, *start_pos, flag;
	int *send_buf_idx, *curr_to_proc, *done_to_proc;
	MPI_Status status;
	ADIOI_Flatlist_node *flat_buf=NULL;
	MPI_Aint buftype_extent;
	int info_flag, coll_bufsize;
	wchar_t *value;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_EXCH_AND_WRITE);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_EXCH_AND_WRITE);

	/* calculate the number of writes of size coll_bufsize
	to be done by each process and the max among all processes.
	That gives the no. of communication phases as well. */

	value = (wchar_t *) MPIU_Malloc((MPI_MAX_INFO_VAL+1)*sizeof(wchar_t));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, value, (MPI_MAX_INFO_VAL+1)*sizeof(wchar_t), L"value");
	NMPI_Info_get(mpi_errno, (fd->info, L"cb_buffer_size", MPI_MAX_INFO_VAL, value, &info_flag));
	MPIU_ERR_CHK_POP(mpi_errno);
	coll_bufsize = _wtoi(value);
	MPIU_Free(value);


	for (i=0; i < nprocs; i++)
	{
		if (others_req[i].count)
		{
			st_loc = others_req[i].offsets[0];
			end_loc = others_req[i].offsets[0];
			break;
		}
	}

	for (i=0; i < nprocs; i++)
	{
		for (j=0; j < others_req[i].count; j++)
		{
			st_loc = MPIDU_MIN(st_loc, others_req[i].offsets[j]);
			end_loc = MPIDU_MAX(end_loc, (others_req[i].offsets[j] + others_req[i].lens[j] - 1));
		}
	}

	/* ntimes=ceiling_div(end_loc - st_loc + 1, coll_bufsize)*/

	ntimes = (int) ((end_loc - st_loc + coll_bufsize)/coll_bufsize);

	if ((st_loc==-1) && (end_loc==-1))
	{
		ntimes = 0; /* this process does no writing. */
	}

	NMPI_Allreduce(mpi_errno, (&ntimes, &max_ntimes, 1, MPI_INT, MPI_MAX, fd->comm));
	MPIU_ERR_CHK_POP(mpi_errno);

	if (ntimes)
	{
		write_buf = (char *) MPIU_Malloc(coll_bufsize);
		MPIU_ERR_MEM_CHK_POP(mpi_errno, write_buf, coll_bufsize, L"write_buf");
	}

	curr_offlen_ptr = (int *) MPIU_Calloc(nprocs, sizeof(int)); 
	MPIU_ERR_MEM_CHK_POP(mpi_errno, curr_offlen_ptr, nprocs * sizeof(int), L"curr_offlen_ptr");
	/* its use is explained below. calloc initializes to 0. */

	count = (int *) MPIU_Malloc(nprocs*sizeof(int));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, count, nprocs * sizeof(int), L"count");
	/* to store count of how many off-len pairs per proc are satisfied
	in an iteration. */

	partial_recv = (int *) MPIU_Calloc(nprocs, sizeof(int));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, partial_recv, nprocs * sizeof(int), L"partial_recv");
	/* if only a portion of the last off-len pair is recd. from a process
	in a particular iteration, the length recd. is stored here.
	calloc initializes to 0. */

	send_size = (int *) MPIU_Malloc(nprocs*sizeof(int));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, send_size, nprocs * sizeof(int), L"send_size");
	/* total size of data to be sent to each proc. in an iteration.
	Of size nprocs so that I can use MPI_Alltoall later. */

	recv_size = (int *) MPIU_Malloc(nprocs*sizeof(int));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, recv_size, nprocs * sizeof(int), L"recv_size");
	/* total size of data to be recd. from each proc. in an iteration.*/

	sent_to_proc = (int *) MPIU_Calloc(nprocs, sizeof(int));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, sent_to_proc, nprocs * sizeof(int), L"sent_to_proc");
	/* amount of data sent to each proc so far. Used in ADIOI_Fill_send_buffer. initialized to 0 here. */

	send_buf_idx = (int *) MPIU_Malloc(nprocs*sizeof(int));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, send_buf_idx, nprocs * sizeof(int), L"send_buf_inx");
	curr_to_proc = (int *) MPIU_Malloc(nprocs*sizeof(int));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, curr_to_proc, nprocs * sizeof(int), L"curr_to_proc");
	done_to_proc = (int *) MPIU_Malloc(nprocs*sizeof(int));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, done_to_proc, nprocs * sizeof(int), L"done_to_proc");
	/* Above three are used in ADIOI_Fill_send_buffer*/

	start_pos = (int *) MPIU_Malloc(nprocs*sizeof(int));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, start_pos, nprocs * sizeof(int), L"start_pos");
	/* used to store the starting value of curr_offlen_ptr[i] in this iteration */

	mpi_errno = ADIOI_Datatype_iscontig(datatype, &buftype_is_contig);
	MPIU_ERR_CHK_POP(mpi_errno);
	if (!buftype_is_contig)
	{
		mpi_errno = ADIOI_Flatten_datatype(datatype);
		MPIU_ERR_CHK_POP(mpi_errno);
		flat_buf = ADIOI_Flatlist;
		while (flat_buf->type != datatype)
		{
			flat_buf = flat_buf->next;
		}
	}
	NMPI_Type_extent(mpi_errno, (datatype, &buftype_extent));
	MPIU_ERR_CHK_POP(mpi_errno);


	/* I need to check if there are any outstanding nonblocking writes to
	the file, which could potentially interfere with the writes taking
	place in this collective write call. Since this is not likely to be
	common, let me do the simplest thing possible here: Each process
	completes all pending nonblocking operations before completing. */

	mpi_errno = ADIOI_Complete_async();
	MPIU_ERR_CHK_POP(mpi_errno);
	NMPI_Barrier(mpi_errno, (fd->comm));
	MPIU_ERR_CHK_POP(mpi_errno);

	done = 0;
	off = st_loc;

#ifdef PROFILE
	MPE_Log_event(14, 0, L"end computation");
#endif

	for (m=0; m < ntimes; m++)
	{
		/* go through all others_req and check which will be satisfied
		by the current write */

		/* Note that MPI guarantees that displacements in filetypes are in 
		monotonically nondecreasing order and that, for writes, the
		filetypes cannot specify overlapping regions in the file. This
		simplifies implementation a bit compared to reads. */

		/* off = start offset in the file for the data to be written in 
		this iteration 
		size = size of data written (bytes) corresponding to off
		req_off = off in file for a particular contiguous request 
		minus what was satisfied in previous iteration
		req_size = size corresponding to req_off */

		/* first calculate what should be communicated */

#ifdef PROFILE
		MPE_Log_event(13, 0, L"start computation");
#endif
		for (i=0; i < nprocs; i++)
		{
			count[i] = recv_size[i] = 0;
		}

		size = (int) (MPIDU_MIN(coll_bufsize, end_loc-st_loc+1-done)); 

		for (i=0; i < nprocs; i++)
		{
			if (others_req[i].count)
			{
				start_pos[i] = curr_offlen_ptr[i];
				for (j=curr_offlen_ptr[i]; j<others_req[i].count; j++)
				{
					if (partial_recv[i])
					{
						/* this request may have been partially satisfied in the previous iteration. */
						req_off = others_req[i].offsets[j] + partial_recv[i]; 
						req_len = others_req[i].lens[j] - partial_recv[i];
						partial_recv[i] = 0;
						/* modify the off-len pair to reflect this change */
						others_req[i].offsets[j] = req_off;
						others_req[i].lens[j] = req_len;
					}
					else
					{
						req_off = others_req[i].offsets[j];
						req_len = others_req[i].lens[j];
					}
					if (req_off < off + size)
					{
						count[i]++;
						NMPI_Address(mpi_errno, (write_buf+req_off-off, &(others_req[i].mem_ptrs[j])));
						MPIU_ERR_CHK_POP(mpi_errno);
						recv_size[i] += (int)(MPIDU_MIN(off + (MPI_Offset)size - req_off, req_len));

						if (off+size-req_off < req_len)
						{
							partial_recv[i] = (int) (off + size - req_off);

							/* --BEGIN ERROR HANDLING-- */
							if ((j+1 < others_req[i].count) && (others_req[i].offsets[j+1] < off+size))
							{
								MPIU_ERR_SET(mpi_errno, MPI_ERR_ARG, L"**arg");
								/*L"Filetype specifies overlapping write regions (which is illegal according to the MPI-2 specification)", 0);*/
								/* allow to continue since additional
								* communication might have to occur
								*/
							}
							/* --END ERROR HANDLING-- */
							break;
						}
					}
					else
					{
						break;
					}
				}
				curr_offlen_ptr[i] = j;
			}
		}

#ifdef PROFILE
		MPE_Log_event(14, 0, L"end computation");
		MPE_Log_event(7, 0, L"start communication");
#endif
		mpi_errno = ADIOI_W_Exchange_data(fd, buf, write_buf, flat_buf, offset_list, 
			len_list, send_size, recv_size, off, size, count, 
			start_pos, partial_recv, 
			sent_to_proc, nprocs, myrank, 
			buftype_is_contig, contig_access_count,
			min_st_offset, fd_size, fd_start, fd_end,
			others_req, send_buf_idx, curr_to_proc,
			done_to_proc, &hole, m, buftype_extent, buf_idx);
		MPIU_ERR_CHK_POP(mpi_errno);
#ifdef PROFILE
		MPE_Log_event(8, 0, L"end communication");
#endif

		flag = 0;
		for (i=0; i<nprocs; i++)
		{
			if (count[i]) flag = 1;
		}

		if (flag)
		{
			mpi_errno = ADIO_WriteContig(fd, write_buf, size, MPI_BYTE, ADIO_EXPLICIT_OFFSET, off, &status);
			MPIU_ERR_CHK_POP(mpi_errno);
		}

		off += size;
		done += size;
	}

	for (i=0; i<nprocs; i++) count[i] = recv_size[i] = 0;
#ifdef PROFILE
	MPE_Log_event(7, 0, L"start communication");
#endif
	for (m=ntimes; m<max_ntimes; m++)
	{
		/* nothing to recv, but check for send. */
		mpi_errno = ADIOI_W_Exchange_data(fd, buf, write_buf, flat_buf, offset_list, 
			len_list, send_size, recv_size, off, size, count, 
			start_pos, partial_recv, 
			sent_to_proc, nprocs, myrank, 
			buftype_is_contig, contig_access_count,
			min_st_offset, fd_size, fd_start, fd_end,
			others_req, send_buf_idx, 
			curr_to_proc, done_to_proc, &hole, m, 
			buftype_extent, buf_idx);
		MPIU_ERR_CHK_POP(mpi_errno);
	}
#ifdef PROFILE
	MPE_Log_event(8, 0, L"end communication");
#endif

	if (ntimes)
	{
		MPIU_Free(write_buf);
	}
	MPIU_Free(curr_offlen_ptr);
	MPIU_Free(count);
	MPIU_Free(partial_recv);
	MPIU_Free(send_size);
	MPIU_Free(recv_size);
	MPIU_Free(sent_to_proc);
	MPIU_Free(start_pos);
	MPIU_Free(send_buf_idx);
	MPIU_Free(curr_to_proc);
	MPIU_Free(done_to_proc);

fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_EXCH_AND_WRITE);
	return mpi_errno;
fn_fail:
	/* free allocated memory */
	goto fn_exit;
}


/* Sets error_code to MPI_SUCCESS if successful, or creates an error code in the case of error. */
#undef FUNCNAME
#define FUNCNAME ADIOI_W_Exchange_data
static int ADIOI_W_Exchange_data(MPID_File_ptr fd, void *buf, char *write_buf,
								 ADIOI_Flatlist_node *flat_buf, MPI_Offset 
								 *offset_list, int *len_list, int *send_size, 
								 int *recv_size, MPI_Offset off, int size,
								 int *count, int *start_pos,
								 int *partial_recv,
								 int *sent_to_proc, int nprocs, 
								 int myrank, int
								 buftype_is_contig, int contig_access_count,
								 MPI_Offset min_st_offset,
								 MPI_Offset fd_size,
								 MPI_Offset *fd_start, MPI_Offset *fd_end, 
								 ADIOI_Access *others_req, 
								 int *send_buf_idx, int *curr_to_proc,
								 int *done_to_proc, int *hole, int iter, 
								 MPI_Aint buftype_extent, int *buf_idx)
{
	static const wchar_t FCNAME[] = L"ADIOI_W_Exchange_data";
	int mpi_errno = MPI_SUCCESS;
	int i, j, k, *tmp_len, nprocs_recv, nprocs_send;
	char **send_buf = NULL; 
	MPI_Request *requests, *send_req;
	MPI_Datatype *recv_types;
	MPI_Status *statuses, status;
	int *srt_len, sum;
	MPI_Offset *srt_off;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_W_EXCHANGE_DATA);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_W_EXCHANGE_DATA);

	/* exchange recv_size info so that each process knows how much to send to whom. */

	NMPI_Alltoall(mpi_errno, (recv_size, 1, MPI_INT, send_size, 1, MPI_INT, fd->comm));
	MPIU_ERR_CHK_POP(mpi_errno);

	/* create derived datatypes for recv */

	nprocs_recv = 0;
	for (i=0; i<nprocs; i++)
	{
		if (recv_size[i]) nprocs_recv++;
	}

	recv_types = (MPI_Datatype *)MPIU_Malloc((nprocs_recv+1)*sizeof(MPI_Datatype));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, recv_types, (nprocs_recv+1) * sizeof(MPI_Datatype), L"recv_types");
	/* +1 to avoid a 0-size malloc */

	tmp_len = (int *) MPIU_Malloc(nprocs*sizeof(int));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, tmp_len, nprocs * sizeof(int), L"tmp_len");
	j = 0;
	for (i=0; i<nprocs; i++)
	{
		if (recv_size[i])
		{
			/* take care if the last off-len pair is a partial recv */
			if (partial_recv[i])
			{
				k = start_pos[i] + count[i] - 1;
				tmp_len[i] = others_req[i].lens[k];
				others_req[i].lens[k] = partial_recv[i];
			}
			NMPI_Type_hindexed(mpi_errno, (count[i], 
				&(others_req[i].lens[start_pos[i]]),
				&(others_req[i].mem_ptrs[start_pos[i]]), 
				MPI_BYTE, recv_types+j));
			MPIU_ERR_CHK_POP(mpi_errno);
			/* absolute displacements; use MPI_BOTTOM in recv */
			NMPI_Type_commit(mpi_errno, (recv_types+j));
			MPIU_ERR_CHK_POP(mpi_errno);
			j++;
		}
	}

	/* To avoid a read-modify-write, check if there are holes in the 
	data to be written. For this, merge the (sorted) offset lists
	others_req using a heap-merge. */

	sum = 0;
	for (i=0; i<nprocs; i++) sum += count[i];
	srt_off = (MPI_Offset *) MPIU_Malloc((sum+1)*sizeof(MPI_Offset));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, srt_off, (sum+1)*sizeof(MPI_Offset), L"srt_off");
	srt_len = (int *) MPIU_Malloc((sum+1)*sizeof(int));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, srt_len, (sum+1)*sizeof(int), L"srt_len");
	/* +1 to avoid a 0-size malloc */

	mpi_errno = ADIOI_Heap_merge(others_req, count, srt_off, srt_len, start_pos, nprocs, nprocs_recv, sum);
	MPIU_ERR_CHK_POP(mpi_errno);

	/* for partial recvs, restore original lengths */
	for (i=0; i<nprocs; i++)
	{
		if (partial_recv[i])
		{
			k = start_pos[i] + count[i] - 1;
			others_req[i].lens[k] = tmp_len[i];
		}
	}
	MPIU_Free(tmp_len);
	tmp_len = NULL;

	/* check if there are any holes */
	*hole = 0;
	for (i=0; i<sum-1; i++)
	{
		if (srt_off[i]+srt_len[i] < srt_off[i+1])
		{
			*hole = 1;
			break;
		}
	}

	MPIU_Free(srt_off);
	MPIU_Free(srt_len);
	srt_off = NULL;
	srt_len = NULL;

	if (nprocs_recv)
	{
		if (*hole)
		{
			mpi_errno = ADIO_ReadContig(fd, write_buf, size, MPI_BYTE, ADIO_EXPLICIT_OFFSET, off, &status);
			MPIU_ERR_CHK_POP(mpi_errno);
			/*MPI_ERR_IO, L"**ioRMWrdwr"*/
		}
	}

	nprocs_send = 0;
	for (i=0; i < nprocs; i++)
	{
		if (send_size[i]) nprocs_send++;
	}

	if (fd->atomicity)
	{
		/* bug fix from Wei-keng Liao and Kenin Coloma */
		requests = (MPI_Request *) MPIU_Malloc((nprocs_send+1)*sizeof(MPI_Request));
		MPIU_ERR_MEM_CHK_POP(mpi_errno, requests, (nprocs_send+1) * sizeof(MPI_Request), L"requests");
		send_req = requests;
	}
	else
	{
		requests = (MPI_Request *) MPIU_Malloc((nprocs_send+nprocs_recv+1)*sizeof(MPI_Request));
		MPIU_ERR_MEM_CHK_POP(mpi_errno, requests, (nprocs_send+nprocs_recv+1) * sizeof(MPI_Request), L"requests");
		/* +1 to avoid a 0-size malloc */

		/* post receives */
		j = 0;
		for (i=0; i<nprocs; i++)
		{
			if (recv_size[i])
			{
				NMPI_Irecv(mpi_errno, (MPI_BOTTOM, 1, recv_types[j], i, myrank+i+100*iter, fd->comm, requests+j));
				MPIU_ERR_CHK_POP(mpi_errno);
				j++;
			}
		}
		send_req = requests + nprocs_recv;
	}

	/* post sends. if buftype_is_contig, data can be directly sent from
	user buf at location given by buf_idx. else use send_buf. */

	if (buftype_is_contig)
	{
		j = 0;
		for (i=0; i < nprocs; i++)
		{
			if (send_size[i])
			{
				NMPI_Isend(mpi_errno, (((char *) buf) + buf_idx[i], send_size[i], MPI_BYTE, i,  myrank+i+100*iter, fd->comm, send_req+j));
				MPIU_ERR_CHK_POP(mpi_errno);
				j++;
				buf_idx[i] += send_size[i];
			}
		}
	}
	else if (nprocs_send)
	{
		/* buftype is not contig */
		send_buf = (char **) MPIU_Malloc(nprocs*sizeof(char*));
		MPIU_ERR_MEM_CHK_POP(mpi_errno, send_buf, nprocs * sizeof(char*), L"send_buf");
		for (i=0; i < nprocs; i++)
		{
			if (send_size[i])
			{
				send_buf[i] = (char *) MPIU_Malloc(send_size[i]);
				MPIU_ERR_MEM_CHK_POP(mpi_errno, send_buf[i], send_size[i], L"send_buf[i]");
			}
		}

		mpi_errno = ADIOI_Fill_send_buffer(fd, buf, flat_buf, send_buf,
			offset_list, len_list, send_size, 
			send_req,
			sent_to_proc, nprocs, myrank, 
			contig_access_count,
			min_st_offset, fd_size, fd_start, fd_end, 
			send_buf_idx, curr_to_proc, done_to_proc, iter,
			buftype_extent);
		MPIU_ERR_CHK_POP(mpi_errno);
		/* the send is done in ADIOI_Fill_send_buffer */
	}

	if (fd->atomicity)
	{
		/* bug fix from Wei-keng Liao and Kenin Coloma */
		j = 0;
		for (i=0; i<nprocs; i++)
		{
			MPI_Status wkl_status;
			if (recv_size[i])
			{
				NMPI_Recv(mpi_errno, (MPI_BOTTOM, 1, recv_types[j], i, myrank+i+100*iter, fd->comm, &wkl_status));
				MPIU_ERR_CHK_POP(mpi_errno);
				j++;
			}
		}
	}

	for (i=0; i<nprocs_recv; i++)
	{
		NMPI_Type_free(mpi_errno, (recv_types+i));
		MPIU_ERR_CHK_POP(mpi_errno);
	}
	MPIU_Free(recv_types);

	if (fd->atomicity)
	{
		statuses = (MPI_Status *) MPIU_Malloc((nprocs_send+1) * sizeof(MPI_Status));
		MPIU_ERR_MEM_CHK_POP(mpi_errno, statuses, (nprocs_send+1) * sizeof(MPI_Status), L"statuses");
		/* +1 to avoid a 0-size malloc */
	}
	else
	{
		statuses = (MPI_Status *) MPIU_Malloc((nprocs_send+nprocs_recv+1) * sizeof(MPI_Status));
		MPIU_ERR_MEM_CHK_POP(mpi_errno, statuses, (nprocs_send+nprocs_recv+1) * sizeof(MPI_Status), L"statuses");
		/* +1 to avoid a 0-size malloc */
	}

#ifdef NEEDS_MPI_TEST
	i = 0;
	if (fd->atomicity)
	{
		while (!i)
		{
			mpi_errno = MPI_Testall(nprocs_send, send_req, &i, statuses);
			MPIU_ERR_CHK_POP(mpi_errno);
		}
	}
	else
	{
		while (!i)
		{
			mpi_errno = MPI_Testall(nprocs_send+nprocs_recv, requests, &i, statuses);
			MPIU_ERR_CHK_POP(mpi_errno);
		}
	}
#else
	if (fd->atomicity)
	{
		/* bug fix from Wei-keng Liao and Kenin Coloma */
		NMPI_Waitall(mpi_errno, (nprocs_send, send_req, statuses));
		MPIU_ERR_CHK_POP(mpi_errno);
	}
	else
	{
		NMPI_Waitall(mpi_errno, (nprocs_send+nprocs_recv, requests, statuses));
		MPIU_ERR_CHK_POP(mpi_errno);
	}
#endif

	MPIU_Free(statuses);
	MPIU_Free(requests);
	if (!buftype_is_contig && nprocs_send)
	{
		for (i=0; i < nprocs; i++) 
			if (send_size[i]) MPIU_Free(send_buf[i]);
		MPIU_Free(send_buf);
	}

fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_W_EXCHANGE_DATA);
	return mpi_errno;
fn_fail:
	/* clean up allocated memory? */
	goto fn_exit;
}


#define ADIOI_BUF_INCR \
{ \
	while (buf_incr) { \
	size_in_buf = MPIDU_MIN(buf_incr, flat_buf_sz); \
	user_buf_idx += size_in_buf; \
	flat_buf_sz -= size_in_buf; \
	if (!flat_buf_sz) { \
	if (flat_buf_idx < (flat_buf->count - 1)) flat_buf_idx++; \
			else { \
			flat_buf_idx = 0; \
			n_buftypes++; \
} \
	user_buf_idx = flat_buf->indices[flat_buf_idx] + \
	n_buftypes*buftype_extent; \
	flat_buf_sz = flat_buf->blocklens[flat_buf_idx]; \
	} \
	buf_incr -= size_in_buf; \
	} \
}


#define ADIOI_BUF_COPY_SEND \
{ \
	while (size) { \
	size_in_buf = MPIDU_MIN(size, flat_buf_sz); \
	memcpy(&(send_buf[p][send_buf_idx[p]]), \
	((char *) buf) + user_buf_idx, size_in_buf); \
	send_buf_idx[p] += size_in_buf; \
	user_buf_idx += size_in_buf; \
	flat_buf_sz -= size_in_buf; \
	if (!flat_buf_sz) { \
	if (flat_buf_idx < (flat_buf->count - 1)) flat_buf_idx++; \
			else { \
			flat_buf_idx = 0; \
			n_buftypes++; \
} \
	user_buf_idx = flat_buf->indices[flat_buf_idx] + \
	n_buftypes*buftype_extent; \
	flat_buf_sz = flat_buf->blocklens[flat_buf_idx]; \
	} \
	size -= size_in_buf; \
	buf_incr -= size_in_buf; \
	} \
	ADIOI_BUF_INCR \
}

#undef FUNCNAME
#define FUNCNAME ADIOI_Fill_send_buffer
static int ADIOI_Fill_send_buffer(MPID_File_ptr fd, void *buf, ADIOI_Flatlist_node
								  *flat_buf, char **send_buf, MPI_Offset 
								  *offset_list, int *len_list, int *send_size, 
								  MPI_Request *requests, int *sent_to_proc, 
								  int nprocs, int myrank, 
								  int contig_access_count, 
								  MPI_Offset min_st_offset, MPI_Offset fd_size,
								  MPI_Offset *fd_start, MPI_Offset *fd_end, 
								  int *send_buf_idx, int *curr_to_proc, 
								  int *done_to_proc, int iter,
								  MPI_Aint buftype_extent)
{
	/* this function is only called if buftype is not contig */
	static const wchar_t FCNAME[] = L"ADIOI_Fill_send_buffer";
	int mpi_errno = MPI_SUCCESS;
	int i, p, flat_buf_idx, size;
	int flat_buf_sz, buf_incr, size_in_buf, jj, n_buftypes;
	MPI_Offset off, len, rem_len, user_buf_idx;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_FILL_SEND_BUFFER);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_FILL_SEND_BUFFER);

	/*  curr_to_proc[p] = amount of data sent to proc. p that has already
	been accounted for so far
	done_to_proc[p] = amount of data already sent to proc. p in 
	previous iterations
	user_buf_idx = current location in user buffer 
	send_buf_idx[p] = current location in send_buf of proc. p  */

	for (i=0; i < nprocs; i++)
	{
		send_buf_idx[i] = curr_to_proc[i] = 0;
		done_to_proc[i] = sent_to_proc[i];
	}
	jj = 0;

	user_buf_idx = flat_buf->indices[0];
	flat_buf_idx = 0;
	n_buftypes = 0;
	flat_buf_sz = flat_buf->blocklens[0];

	/* flat_buf_idx = current index into flattened buftype
	flat_buf_sz = size of current contiguous component in 
	flattened buf */

	for (i=0; i<contig_access_count; i++)
	{
		off	 = offset_list[i];
		rem_len = (MPI_Offset) len_list[i];

		/*this request may span the file domains of more than one process*/
		while (rem_len != 0)
		{
			len = rem_len;
			/* NOTE: len value is modified by ADIOI_Calc_aggregator() to be no
			* longer than the single region that processor "p" is responsible
			* for.
			*/
			mpi_errno = ADIOI_Calc_aggregator(fd, off, min_st_offset, &len, fd_size, fd_start, fd_end, &p);
			MPIU_ERR_CHK_POP(mpi_errno);

			if (send_buf_idx[p] < send_size[p])
			{
				if (curr_to_proc[p]+len > done_to_proc[p])
				{
					if (done_to_proc[p] > curr_to_proc[p])
					{
						size = (int)MPIDU_MIN(curr_to_proc[p] + len - done_to_proc[p], send_size[p]-send_buf_idx[p]);
						buf_incr = done_to_proc[p] - curr_to_proc[p];
						ADIOI_BUF_INCR
						buf_incr = (int)(curr_to_proc[p] + len - done_to_proc[p]);
						curr_to_proc[p] = done_to_proc[p] + size;
						ADIOI_BUF_COPY_SEND
					}
					else
					{
						size = (int)MPIDU_MIN(len,send_size[p]-send_buf_idx[p]);
						buf_incr = (int)len;
						curr_to_proc[p] += size;
						ADIOI_BUF_COPY_SEND
					}
					if (send_buf_idx[p] == send_size[p])
					{
						NMPI_Isend(mpi_errno, (send_buf[p], send_size[p], MPI_BYTE, p, myrank+p+100*iter, fd->comm, requests+jj));
						MPIU_ERR_CHK_POP(mpi_errno);
						jj++;
					}
				}
				else
				{
					curr_to_proc[p] += (int)len;
					buf_incr = (int)len;
					ADIOI_BUF_INCR
				}
			}
			else
			{
				buf_incr = (int)len;
				ADIOI_BUF_INCR
			}
			off	 += len;
			rem_len -= len;
		}
	}
	for (i=0; i < nprocs; i++)
	{
		if (send_size[i]) sent_to_proc[i] = curr_to_proc[i];
	}

fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_FILL_SEND_BUFFER);
	return mpi_errno;
}



#undef FUNCNAME
#define FUNCNAME ADIOI_Heap_merge
static int ADIOI_Heap_merge(ADIOI_Access *others_req, int *count, 
							MPI_Offset *srt_off, int *srt_len, int *start_pos,
							int nprocs, int nprocs_recv, int total_elements)
{
	typedef struct
	{
		MPI_Offset *off_list;
		int *len_list;
		int nelem;
	} heap_struct;

	static const wchar_t FCNAME[] = L"ADIOI_Heap_merge";
	int mpi_errno = MPI_SUCCESS;
	heap_struct *a, tmp;
	int i, j, heapsize, l, r, k, smallest;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_HEAP_MERGE);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_HEAP_MERGE);

	a = (heap_struct *) MPIU_Malloc((nprocs_recv+1)*sizeof(heap_struct));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, a, (nprocs_recv+1) * sizeof(heap_struct), L"a");

	j = 0;
	for (i=0; i<nprocs; i++)
	{
		if (count[i])
		{
			a[j].off_list = &(others_req[i].offsets[start_pos[i]]);
			a[j].len_list = &(others_req[i].lens[start_pos[i]]);
			a[j].nelem = count[i];
			j++;
		}
	}

	/* build a heap out of the first element from each list, with
	the smallest element of the heap at the root */

	heapsize = nprocs_recv;
	for (i=heapsize/2 - 1; i>=0; i--)
	{
		/* Heapify(a, i, heapsize); Algorithm from Cormen et al. pg. 143
		modified for a heap with smallest element at root. I have 
		removed the recursion so that there are no function calls.
		Function calls are too expensive. */
		k = i;
		for(;;)
		{
			l = 2*(k+1) - 1;
			r = 2*(k+1);

			if ((l < heapsize) && (*(a[l].off_list) < *(a[k].off_list)))
				smallest = l;
			else
				smallest = k;

			if ((r < heapsize) && (*(a[r].off_list) < *(a[smallest].off_list)))
				smallest = r;

			if (smallest != k)
			{
				tmp.off_list = a[k].off_list;
				tmp.len_list = a[k].len_list;
				tmp.nelem = a[k].nelem;

				a[k].off_list = a[smallest].off_list;
				a[k].len_list = a[smallest].len_list;
				a[k].nelem = a[smallest].nelem;

				a[smallest].off_list = tmp.off_list;
				a[smallest].len_list = tmp.len_list;
				a[smallest].nelem = tmp.nelem;

				k = smallest;
			}
			else
			{
				break;
			}
		}
	}

	for (i=0; i<total_elements; i++)
	{
		/* extract smallest element from heap, i.e. the root */
		srt_off[i] = *(a[0].off_list);
		srt_len[i] = *(a[0].len_list);
		(a[0].nelem)--;

		if (!a[0].nelem)
		{
			a[0].off_list = a[heapsize-1].off_list;
			a[0].len_list = a[heapsize-1].len_list;
			a[0].nelem = a[heapsize-1].nelem;
			heapsize--;
		}
		else
		{
			(a[0].off_list)++;
			(a[0].len_list)++;
		}

		/* Heapify(a, 0, heapsize); */
		k = 0;
		for (;;)
		{
			l = 2*(k+1) - 1;
			r = 2*(k+1);

			if ((l < heapsize) && (*(a[l].off_list) < *(a[k].off_list)))
				smallest = l;
			else
				smallest = k;

			if ((r < heapsize) && (*(a[r].off_list) < *(a[smallest].off_list)))
				smallest = r;

			if (smallest != k)
			{
				tmp.off_list = a[k].off_list;
				tmp.len_list = a[k].len_list;
				tmp.nelem = a[k].nelem;

				a[k].off_list = a[smallest].off_list;
				a[k].len_list = a[smallest].len_list;
				a[k].nelem = a[smallest].nelem;

				a[smallest].off_list = tmp.off_list;
				a[smallest].len_list = tmp.len_list;
				a[smallest].nelem = tmp.nelem;

				k = smallest;
			}
			else
			{
				break;
			}
		}
	}
	MPIU_Free(a);

fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_HEAP_MERGE);
	return mpi_errno;
fn_fail:
	goto fn_exit;
}

#undef FUNCNAME
#define FUNCNAME ADIO_Type_create_subarray
int ADIO_Type_create_subarray(int ndims, int *array_of_sizes, int *array_of_subsizes, int *array_of_starts, int order, MPI_Datatype oldtype, MPI_Datatype *newtype)
{
	static const wchar_t FCNAME[] = L"ADIO_Type_create_subarray";
	int mpi_errno = MPI_SUCCESS;
	MPI_Aint extent, disps[3], size;
	int i, blklens[3];
	MPI_Datatype tmp1, tmp2, types[3];
	MPIDI_STATE_DECL(MPID_STATE_ADIO_TYPE_CREATE_SUBARRAY);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIO_TYPE_CREATE_SUBARRAY);

	NMPI_Type_extent(mpi_errno, (oldtype, &extent));
	MPIU_ERR_CHK_POP(mpi_errno);

	if (order == MPI_ORDER_FORTRAN)
	{
		/* dimension 0 changes fastest */
		if (ndims == 1)
		{
			NMPI_Type_contiguous(mpi_errno, (array_of_subsizes[0], oldtype, &tmp1));
			MPIU_ERR_CHK_POP(mpi_errno);
		}
		else
		{
			NMPI_Type_vector(mpi_errno, (array_of_subsizes[1], array_of_subsizes[0], array_of_sizes[0], oldtype, &tmp1));
			MPIU_ERR_CHK_POP(mpi_errno);

			size = array_of_sizes[0]*extent;
			for (i=2; i<ndims; i++)
			{
				size *= array_of_sizes[i-1];
				NMPI_Type_hvector(mpi_errno, (array_of_subsizes[i], 1, size, tmp1, &tmp2));
				MPIU_ERR_CHK_POP(mpi_errno);
				NMPI_Type_free(mpi_errno, (&tmp1));
				MPIU_ERR_CHK_POP(mpi_errno);
				tmp1 = tmp2;
			}
		}

		/* add displacement and UB */
		disps[1] = array_of_starts[0];
		size = 1;
		for (i=1; i<ndims; i++)
		{
			size *= array_of_sizes[i-1];
			disps[1] += size*array_of_starts[i];
		}  
		/* rest done below for both Fortran and C order */
	}
	else
	{
		/* order == MPI_ORDER_C */
		/* dimension ndims-1 changes fastest */
		if (ndims == 1)
		{
			NMPI_Type_contiguous(mpi_errno, (array_of_subsizes[0], oldtype, &tmp1));
			MPIU_ERR_CHK_POP(mpi_errno);
		}
		else
		{
			NMPI_Type_vector(mpi_errno, (array_of_subsizes[ndims-2], array_of_subsizes[ndims-1], array_of_sizes[ndims-1], oldtype, &tmp1));
			MPIU_ERR_CHK_POP(mpi_errno);

			size = array_of_sizes[ndims-1]*extent;
			for (i=ndims-3; i>=0; i--)
			{
				size *= array_of_sizes[i+1];
				NMPI_Type_hvector(mpi_errno, (array_of_subsizes[i], 1, size, tmp1, &tmp2));
				MPIU_ERR_CHK_POP(mpi_errno);
				NMPI_Type_free(mpi_errno, (&tmp1));
				MPIU_ERR_CHK_POP(mpi_errno);
				tmp1 = tmp2;
			}
		}

		/* add displacement and UB */
		disps[1] = array_of_starts[ndims-1];
		size = 1;
		for (i=ndims-2; i>=0; i--)
		{
			size *= array_of_sizes[i+1];
			disps[1] += size*array_of_starts[i];
		}
	}

	disps[1] *= extent;

	disps[2] = extent;
	for (i=0; i<ndims; i++)
	{
		disps[2] *= array_of_sizes[i];
	}

	disps[0] = 0;
	blklens[0] = blklens[1] = blklens[2] = 1;
	types[0] = MPI_LB;
	types[1] = tmp1;
	types[2] = MPI_UB;

	NMPI_Type_struct(mpi_errno, (3, blklens, disps, types, newtype));
	MPIU_ERR_CHK_POP(mpi_errno);

	NMPI_Type_free(mpi_errno, (&tmp1));
	MPIU_ERR_CHK_POP(mpi_errno);

fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIO_TYPE_CREATE_SUBARRAY);
	return mpi_errno;
fn_fail:
	goto fn_exit;
}

/* this used to be implemented in every file system as an fcntl.  It makes
* deferred open easier if we know ADIO_Fcntl will always need a file to really
* be open. set_view doesn't modify anything related to the open files.
*/
#undef FUNCNAME
#define FUNCNAME ADIO_Set_view
int ADIO_Set_view(MPID_File_ptr fd, MPI_Offset disp, MPI_Datatype etype, MPI_Datatype filetype, MPI_Info info)
{
	static const wchar_t FCNAME[] = L"ADIO_Set_view";
	int mpi_errno = MPI_SUCCESS;
	int combiner, i, j, k, filetype_is_contig;
	MPI_Datatype copy_etype, copy_filetype;
	ADIOI_Flatlist_node *flat_file;
	MPIDI_STATE_DECL(MPID_STATE_ADIO_SET_VIEW);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIO_SET_VIEW);
	/* free copies of old etypes and filetypes and delete flattened version of filetype if necessary */

	NMPI_Type_get_envelope(mpi_errno, (fd->etype, &i, &j, &k, &combiner));
	MPIU_ERR_CHK_POP(mpi_errno);
	if (combiner != MPI_COMBINER_NAMED)
	{
		NMPI_Type_free(mpi_errno, (&(fd->etype)));
		MPIU_ERR_CHK_POP(mpi_errno);
	}

	mpi_errno = ADIOI_Datatype_iscontig(fd->filetype, &filetype_is_contig);
	MPIU_ERR_CHK_POP(mpi_errno);
	if (!filetype_is_contig)
	{
		mpi_errno = ADIOI_Delete_flattened(fd->filetype);
		MPIU_ERR_CHK_POP(mpi_errno);
	}

	NMPI_Type_get_envelope(mpi_errno, (fd->filetype, &i, &j, &k, &combiner));
	MPIU_ERR_CHK_POP(mpi_errno);
	if (combiner != MPI_COMBINER_NAMED)
	{
		NMPI_Type_free(mpi_errno, (&(fd->filetype)));
		MPIU_ERR_CHK_POP(mpi_errno);
	}

	/* set new info */
	mpi_errno = ADIO_SetInfo(fd, info);
	MPIU_ERR_CHK_POP(mpi_errno);

	/* set new etypes and filetypes */

	NMPI_Type_get_envelope(mpi_errno, (etype, &i, &j, &k, &combiner));
	MPIU_ERR_CHK_POP(mpi_errno);
	if (combiner == MPI_COMBINER_NAMED)
	{
		fd->etype = etype;
	}
	else
	{
		NMPI_Type_contiguous(mpi_errno, (1, etype, &copy_etype));
		MPIU_ERR_CHK_POP(mpi_errno);
		NMPI_Type_commit(mpi_errno, (&copy_etype));
		MPIU_ERR_CHK_POP(mpi_errno);
		fd->etype = copy_etype;
	}
	NMPI_Type_get_envelope(mpi_errno, (filetype, &i, &j, &k, &combiner));
	MPIU_ERR_CHK_POP(mpi_errno);
	if (combiner == MPI_COMBINER_NAMED)
	{
		fd->filetype = filetype;
	}
	else
	{
		NMPI_Type_contiguous(mpi_errno, (1, filetype, &copy_filetype));
		MPIU_ERR_CHK_POP(mpi_errno);
		NMPI_Type_commit(mpi_errno, (&copy_filetype));
		MPIU_ERR_CHK_POP(mpi_errno);
		fd->filetype = copy_filetype;
		mpi_errno = ADIOI_Flatten_datatype(fd->filetype);
		MPIU_ERR_CHK_POP(mpi_errno);
		/* this function will not flatten the filetype if it turns out
		to be all contiguous. */
	}

	NMPI_Type_size(mpi_errno, (fd->etype, &(fd->etype_size)));
	MPIU_ERR_CHK_POP(mpi_errno);
	fd->disp = disp;

	/* reset MPI-IO file pointer to point to the first byte that can be accessed in this view. */

	mpi_errno = ADIOI_Datatype_iscontig(fd->filetype, &filetype_is_contig);
	MPIU_ERR_CHK_POP(mpi_errno);
	if (filetype_is_contig)
	{
		fd->fp_ind = disp;
	}
	else
	{
		flat_file = ADIOI_Flatlist;
		while (flat_file->type != fd->filetype)
		{
			flat_file = flat_file->next;
		}
		for (i=0; i<flat_file->count; i++)
		{
			if (flat_file->blocklens[i])
			{
				fd->fp_ind = disp + flat_file->indices[i];
				break;
			}
		}
	}

fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIO_SET_VIEW);
	return mpi_errno;
fn_fail:
	goto fn_exit;
}

/* set the shared file pointer to "offset" etypes relative to the current view */

int ADIOI_NFS_Set_shared_fp(MPID_File_ptr fd, MPI_Offset offset);

#undef FUNCNAME
#define FUNCNAME ADIO_Set_shared_fp
int ADIO_Set_shared_fp(MPID_File_ptr fd, MPI_Offset offset)
{
	static const wchar_t FCNAME[] = L"ADIO_Set_shared_fp";
	int mpi_errno = MPI_SUCCESS;
	MPI_Status status;
	MPI_Comm dupcommself;
	MPIDI_STATE_DECL(MPID_STATE_ADIO_SET_SHARED_FP);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIO_SET_SHARED_FP);

	if (fd->shared_fp_fd == NULL)
	{
		NMPI_Comm_dup(mpi_errno, (MPI_COMM_SELF, &dupcommself));
		MPIU_ERR_CHK_POP(mpi_errno);
		mpi_errno = ADIO_Open(MPI_COMM_SELF, dupcommself, 
			fd->shared_fp_fname, 
			fd->file_system,
			fd->fns,
			ADIO_CREATE | ADIO_RDWR | ADIO_DELETE_ON_CLOSE, 
			0, MPI_BYTE, MPI_BYTE,
			MPI_INFO_NULL, 
			ADIO_PERM_NULL, &fd->shared_fp_fd);
		MPIU_ERR_CHK_POP(mpi_errno);
	}

	ADIOI_WRITE_LOCK(fd->shared_fp_fd, 0, SEEK_SET, sizeof(MPI_Offset));
	mpi_errno = ADIO_WriteContig(fd->shared_fp_fd, &offset, sizeof(MPI_Offset), MPI_BYTE, ADIO_EXPLICIT_OFFSET, 0, &status);
	MPIU_ERR_CHK_POP(mpi_errno);
	ADIOI_UNLOCK(fd->shared_fp_fd, 0, SEEK_SET, sizeof(MPI_Offset));

fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIO_SET_SHARED_FP);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME ADIOI_GEN_SeekIndividual
int ADIOI_GEN_SeekIndividual(MPID_File_ptr fd, MPI_Offset offset, int whence, MPI_Offset *offset_ptr)
{
	/* implemented for whence=SEEK_SET only. SEEK_CUR and SEEK_END must
	be converted to the equivalent with SEEK_SET before calling this 
	routine. */
	/* offset is in units of etype relative to the filetype */
	static const wchar_t FCNAME[] = L"ADIOI_GEN_SeekIndividual";
	int mpi_errno = MPI_SUCCESS;
	MPI_Offset off;
	ADIOI_Flatlist_node *flat_file;

	int i, n_etypes_in_filetype, n_filetypes, etype_in_filetype;
	MPI_Offset abs_off_in_filetype=0;
	int size_in_filetype, sum;
	int filetype_size, etype_size, filetype_is_contig;
	MPI_Aint filetype_extent;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_GEN_SEEKINDIVIDUAL);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_GEN_SEEKINDIVIDUAL);

	MPIU_UNREFERENCED_ARG(whence);

	mpi_errno = ADIOI_Datatype_iscontig(fd->filetype, &filetype_is_contig);
	MPIU_ERR_CHK_POP(mpi_errno);
	etype_size = fd->etype_size;

	if (filetype_is_contig)
	{
		off = fd->disp + etype_size * offset;
	}
	else
	{
		flat_file = ADIOI_Flatlist;
		while (flat_file->type != fd->filetype)
		{
			flat_file = flat_file->next;
		}

		NMPI_Type_extent(mpi_errno, (fd->filetype, &filetype_extent));
		MPIU_ERR_CHK_POP(mpi_errno);
		NMPI_Type_size(mpi_errno, (fd->filetype, &filetype_size));
		MPIU_ERR_CHK_POP(mpi_errno);
		if ( ! filetype_size )
		{
			/* Since offset relative to the filetype size, we can't
			do compute the offset when that result is zero.
			Return zero for the offset for now */
			*offset_ptr = 0;
			goto fn_exit;
		}

		n_etypes_in_filetype = filetype_size/etype_size;
		n_filetypes = (int) (offset / n_etypes_in_filetype);
		etype_in_filetype = (int) (offset % n_etypes_in_filetype);
		size_in_filetype = etype_in_filetype * etype_size;

		sum = 0;
		for (i=0; i<flat_file->count; i++)
		{
			sum += flat_file->blocklens[i];
			if (sum > size_in_filetype)
			{
				abs_off_in_filetype = flat_file->indices[i] + size_in_filetype - (sum - flat_file->blocklens[i]);
				break;
			}
		}

		/* abs. offset in bytes in the file */
		off = fd->disp + (MPI_Offset) n_filetypes * filetype_extent + abs_off_in_filetype;
	}

	/*
	* we used to call lseek here and update both fp_ind and fp_sys_posn, but now
	* we don't seek and only update fp_ind (ROMIO's idea of where we are in the
	* file).  We leave the system file descriptor and fp_sys_posn alone. 
	* The fs-specifc ReadContig and WriteContig will seek to the correct place in
	* the file before reading/writing if the 'offset' parameter doesn't match
	* fp_sys_posn
	*/
	fd->fp_ind = off;

	*offset_ptr = off;

fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_GEN_SEEKINDIVIDUAL);
	return mpi_errno;

fn_fail:
	goto fn_exit;
}

#undef FUNCNAME
#define FUNCNAME ADIOI_GEN_ReadStrided_naive
int ADIOI_GEN_ReadStrided_naive(MPID_File_ptr fd, void *buf, int count, MPI_Datatype buftype, int file_ptr_type, MPI_Offset offset, MPI_Status *status)
{
	/* offset is in units of etype relative to the filetype. */
	static const wchar_t FCNAME[] = L"ADIOI_GEN_ReadStrided_naive";
	int mpi_errno = MPI_SUCCESS;
	ADIOI_Flatlist_node *flat_buf, *flat_file;
	int brd_size, frd_size=0, b_index;
	int bufsize, size, sum, n_etypes_in_filetype, size_in_filetype;
	int n_filetypes, etype_in_filetype;
	MPI_Offset abs_off_in_filetype=0;
	int filetype_size, etype_size, buftype_size, req_len;
	MPI_Aint filetype_extent, buftype_extent; 
	int buf_count, buftype_is_contig, filetype_is_contig;
	MPI_Offset userbuf_off;
	MPI_Offset off, req_off, disp, end_offset=0, start_off;
	MPI_Status status1;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_GEN_READSTRIDED_NAIVE);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_GEN_READSTRIDED_NAIVE);

	mpi_errno = ADIOI_Datatype_iscontig(buftype, &buftype_is_contig);
	MPIU_ERR_CHK_POP(mpi_errno);
	mpi_errno = ADIOI_Datatype_iscontig(fd->filetype, &filetype_is_contig);
	MPIU_ERR_CHK_POP(mpi_errno);

	NMPI_Type_size(mpi_errno, (fd->filetype, &filetype_size));
	MPIU_ERR_CHK_POP(mpi_errno);
	if ( ! filetype_size )
	{
		goto fn_exit;
	}

	NMPI_Type_extent(mpi_errno, (fd->filetype, &filetype_extent));
	MPIU_ERR_CHK_POP(mpi_errno);
	NMPI_Type_size(mpi_errno, (buftype, &buftype_size));
	MPIU_ERR_CHK_POP(mpi_errno);
	NMPI_Type_extent(mpi_errno, (buftype, &buftype_extent));
	MPIU_ERR_CHK_POP(mpi_errno);
	etype_size = fd->etype_size;

	bufsize = buftype_size * count;

	/* contiguous in buftype and filetype is handled elsewhere */

	if (!buftype_is_contig && filetype_is_contig)
	{
		int b_count;
		/* noncontiguous in memory, contiguous in file. */

		mpi_errno = ADIOI_Flatten_datatype(buftype);
		MPIU_ERR_CHK_POP(mpi_errno);
		flat_buf = ADIOI_Flatlist;
		while (flat_buf->type != buftype)
		{
			flat_buf = flat_buf->next;
		}

		off = (file_ptr_type == ADIO_INDIVIDUAL) ? fd->fp_ind : fd->disp + etype_size * offset;

		start_off = off;
		end_offset = off + bufsize - 1;

		/* if atomicity is true, lock (exclusive) the region to be accessed */
		if ((fd->atomicity) && (fd->file_system != ADIO_PIOFS) && 
			(fd->file_system != ADIO_PVFS))
		{
			ADIOI_WRITE_LOCK(fd, start_off, SEEK_SET, end_offset-start_off+1);
		}

		/* for each region in the buffer, grab the data and put it in place */
		for (b_count=0; b_count < count; b_count++)
		{
			for (b_index=0; b_index < flat_buf->count; b_index++)
			{
				userbuf_off = b_count*buftype_extent + flat_buf->indices[b_index];
				req_off = off;
				req_len = flat_buf->blocklens[b_index];

				mpi_errno = ADIO_ReadContig(fd, (char *) buf + userbuf_off, req_len, MPI_BYTE, ADIO_EXPLICIT_OFFSET, req_off, &status1);
				MPIU_ERR_CHK_POP(mpi_errno);

				/* off is (potentially) used to save the final offset later */
				off += flat_buf->blocklens[b_index];
			}
		}

		if ((fd->atomicity) && (fd->file_system != ADIO_PIOFS) && 
			(fd->file_system != ADIO_PVFS))
		{
			ADIOI_UNLOCK(fd, start_off, SEEK_SET, end_offset-start_off+1);
		}

		if (file_ptr_type == ADIO_INDIVIDUAL) fd->fp_ind = off;
	}
	else
	{
		/* noncontiguous in file */
		int f_index, st_frd_size, st_index = 0, st_n_filetypes;
		int flag;

		/* First we're going to calculate a set of values for use in all
		* the noncontiguous in file cases:
		* start_off - starting byte position of data in file
		* end_offset - last byte offset to be acessed in the file
		* st_n_filetypes - how far into the file we start in terms of
		*				  whole filetypes
		* st_index - index of block in first filetype that we will be
		*			starting in (?)
		* st_frd_size - size of the data in the first filetype block
		*			   that we will read (accounts for being part-way
		*			   into reading this block of the filetype
		*
		*/

		/* filetype already flattened in ADIO_Open */
		flat_file = ADIOI_Flatlist;
		while (flat_file->type != fd->filetype)
		{
			flat_file = flat_file->next;
		}
		disp = fd->disp;

		if (file_ptr_type == ADIO_INDIVIDUAL)
		{
			start_off = fd->fp_ind; /* in bytes */
			n_filetypes = -1;
			flag = 0;
			while (!flag)
			{
				n_filetypes++;
				for (f_index=0; f_index < flat_file->count; f_index++)
				{
					if (disp + flat_file->indices[f_index] + (MPI_Offset) n_filetypes*filetype_extent + flat_file->blocklens[f_index] >= start_off)
					{
						/* this block contains our starting position */

						st_index = f_index;
						frd_size = (int) (disp + flat_file->indices[f_index] + 
							(MPI_Offset) n_filetypes*filetype_extent + 
							flat_file->blocklens[f_index] - start_off);
						flag = 1;
						break;
					}
				}
			}
		}
		else
		{
			n_etypes_in_filetype = filetype_size/etype_size;
			n_filetypes = (int) (offset / n_etypes_in_filetype);
			etype_in_filetype = (int) (offset % n_etypes_in_filetype);
			size_in_filetype = etype_in_filetype * etype_size;

			sum = 0;
			for (f_index=0; f_index < flat_file->count; f_index++)
			{
				sum += flat_file->blocklens[f_index];
				if (sum > size_in_filetype)
				{
					st_index = f_index;
					frd_size = sum - size_in_filetype;
					abs_off_in_filetype = flat_file->indices[f_index] + size_in_filetype - (sum - flat_file->blocklens[f_index]);
					break;
				}
			}

			/* abs. offset in bytes in the file */
			start_off = disp + (MPI_Offset) n_filetypes*filetype_extent + abs_off_in_filetype;
		}

		st_frd_size = frd_size;
		st_n_filetypes = n_filetypes;

		/* start_off, st_n_filetypes, st_index, and st_frd_size are all calculated at this point */

		/* Calculate end_offset, the last byte-offset that will be accessed.
		* e.g., if start_off=0 and 100 bytes to be read, end_offset=99
		*/
		userbuf_off = 0;
		f_index = st_index;
		off = start_off;
		frd_size = MPIDU_MIN(st_frd_size, bufsize);
		while (userbuf_off < bufsize)
		{
			userbuf_off += frd_size;
			end_offset = off + frd_size - 1;

			if (f_index < (flat_file->count - 1))
			{
				f_index++;
			}
			else
			{
				f_index = 0;
				n_filetypes++;
			}

			off = disp + flat_file->indices[f_index] + (MPI_Offset) n_filetypes*filetype_extent;
			frd_size = MPIDU_MIN(flat_file->blocklens[f_index], bufsize-(int)userbuf_off);
		}

		/* End of calculations.  At this point the following values have
		* been calculated and are ready for use:
		* - start_off
		* - end_offset
		* - st_n_filetypes
		* - st_index
		* - st_frd_size
		*/

		/* if atomicity is true, lock (exclusive) the region to be accessed */
		if ((fd->atomicity) && (fd->file_system != ADIO_PIOFS) && (fd->file_system != ADIO_PVFS))
		{
			ADIOI_WRITE_LOCK(fd, start_off, SEEK_SET, end_offset-start_off+1);
		}

		if (buftype_is_contig && !filetype_is_contig)
		{
			/* contiguous in memory, noncontiguous in file. should be the most common case. */

			userbuf_off = 0;
			f_index = st_index;
			off = start_off;
			n_filetypes = st_n_filetypes;
			frd_size = MPIDU_MIN(st_frd_size, bufsize);

			/* while there is still space in the buffer, read more data */
			while (userbuf_off < bufsize)
			{
				if (frd_size)
				{
					/* TYPE_UB and TYPE_LB can result in 
					frd_size = 0. save system call in such cases */ 
					req_off = off;
					req_len = frd_size;

					mpi_errno = ADIO_ReadContig(fd, (char *) buf + userbuf_off, req_len, MPI_BYTE, ADIO_EXPLICIT_OFFSET, req_off, &status1);
					MPIU_ERR_CHK_POP(mpi_errno);
				}
				userbuf_off += frd_size;

				if (off + frd_size < disp + flat_file->indices[f_index] + flat_file->blocklens[f_index] + (MPI_Offset) n_filetypes*filetype_extent)
				{
					/* important that this value be correct, as it is
					* used to set the offset in the fd near the end of
					* this function.
					*/
					off += frd_size;
				}
				/* did not reach end of contiguous block in filetype.
				* no more I/O needed. off is incremented by frd_size.
				*/
				else
				{
					if (f_index < (flat_file->count - 1))
					{
						f_index++;
					}
					else
					{
						f_index = 0;
						n_filetypes++;
					}
					off = disp + flat_file->indices[f_index] + (MPI_Offset) n_filetypes*filetype_extent;
					frd_size = MPIDU_MIN(flat_file->blocklens[f_index], bufsize-(int)userbuf_off);
				}
			}
		}
		else
		{
			int i, tmp_bufsize = 0;
			/* noncontiguous in memory as well as in file */

			mpi_errno = ADIOI_Flatten_datatype(buftype);
			MPIU_ERR_CHK_POP(mpi_errno);
			flat_buf = ADIOI_Flatlist;
			while (flat_buf->type != buftype)
			{
				flat_buf = flat_buf->next;
			}

			b_index = buf_count = 0;
			i = (int) (flat_buf->indices[0]);
			f_index = st_index;
			off = start_off;
			n_filetypes = st_n_filetypes;
			frd_size = st_frd_size;
			brd_size = flat_buf->blocklens[0];

			/* while we haven't read size * count bytes, keep going */
			while (tmp_bufsize < bufsize)
			{
				int new_brd_size = brd_size, new_frd_size = frd_size;

				size = MPIDU_MIN(frd_size, brd_size);
				if (size)
				{
					req_off = off;
					req_len = size;
					userbuf_off = i;

					mpi_errno = ADIO_ReadContig(fd, (char *) buf + userbuf_off, req_len, MPI_BYTE, ADIO_EXPLICIT_OFFSET, req_off, &status1);
					MPIU_ERR_CHK_POP(mpi_errno);
				}

				if (size == frd_size)
				{
					/* reached end of contiguous block in file */
					if (f_index < (flat_file->count - 1))
					{
						f_index++;
					}
					else
					{
						f_index = 0;
						n_filetypes++;
					}

					off = disp + flat_file->indices[f_index] + (MPI_Offset) n_filetypes*filetype_extent;

					new_frd_size = flat_file->blocklens[f_index];
					if (size != brd_size)
					{
						i += size;
						new_brd_size -= size;
					}
				}

				if (size == brd_size)
				{
					/* reached end of contiguous block in memory */

					b_index = (b_index + 1)%flat_buf->count;
					buf_count++;
					i = (int) (buftype_extent*(buf_count/flat_buf->count) + flat_buf->indices[b_index]);
					new_brd_size = flat_buf->blocklens[b_index];
					if (size != frd_size)
					{
						off += size;
						new_frd_size -= size;
					}
				}
				tmp_bufsize += size;
				frd_size = new_frd_size;
				brd_size = new_brd_size;
			}
		}

		/* unlock the file region if we locked it */
		if ((fd->atomicity) && (fd->file_system != ADIO_PIOFS) && (fd->file_system != ADIO_PVFS))
		{
			ADIOI_UNLOCK(fd, start_off, SEEK_SET, end_offset-start_off+1);
		}

		if (file_ptr_type == ADIO_INDIVIDUAL) fd->fp_ind = off;
	} /* end of (else noncontiguous in file) */

	fd->fp_sys_posn = -1;   /* mark it as invalid. */

#ifdef HAVE_STATUS_SET_BYTES
	MPIR_Status_set_bytes(status, buftype, bufsize);
	/* This is a temporary way of filling in status. The right way is to 
	* keep track of how much data was actually read and placed in buf 
	*/
#endif

	if (!buftype_is_contig)
	{
		mpi_errno = ADIOI_Delete_flattened(buftype);
		MPIU_ERR_CHK_POP(mpi_errno);
	}

fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_GEN_READSTRIDED_NAIVE);
	return mpi_errno;
fn_fail:
	/* clean up allocated memory? */
	goto fn_exit;
}

/*
#define ADIOI_BUFFERED_READ \
{ \
if (req_off >= readbuf_off + readbuf_len) { \
readbuf_off = req_off; \
readbuf_len = (int) (MPIDU_MIN(max_bufsize, end_offset-readbuf_off+1));\
ADIO_ReadContig(fd, readbuf, readbuf_len, MPI_BYTE, \
ADIO_EXPLICIT_OFFSET, readbuf_off, &status1, error_code); \
if (*error_code != MPI_SUCCESS) return; \
} \
while (req_len > readbuf_off + readbuf_len - req_off) { \
partial_read = (int) (readbuf_off + readbuf_len - req_off); \
tmp_buf = (char *) MPIU_Malloc(partial_read); \
memcpy(tmp_buf, readbuf+readbuf_len-partial_read, partial_read); \
MPIU_Free(readbuf); \
readbuf = (char *) MPIU_Malloc(partial_read + max_bufsize); \
memcpy(readbuf, tmp_buf, partial_read); \
MPIU_Free(tmp_buf); \
readbuf_off += readbuf_len-partial_read; \
readbuf_len = (int) (partial_read + MPIDU_MIN(max_bufsize, \
end_offset-readbuf_off+1)); \
ADIO_ReadContig(fd, readbuf+partial_read, readbuf_len-partial_read, \
MPI_BYTE, ADIO_EXPLICIT_OFFSET, readbuf_off+partial_read, \
&status1, error_code); \
if (*error_code != MPI_SUCCESS) return; \
} \
memcpy((char *)buf + userbuf_off, readbuf+req_off-readbuf_off, req_len); \
}
*/

#if 0
#undef FUNCNAME
#define FUNCNAME ADIOI_BUFFERED_READ
int ADIOI_BUFFERED_READ()
{
	static const wchar_t FCNAME[] = L"ADIOI_BUFFERED_READ";
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_BUFFERED_READ);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_BUFFERED_READ);

	if (req_off >= readbuf_off + readbuf_len)
	{
		readbuf_off = req_off;
		readbuf_len = (int) (MPIDU_MIN(max_bufsize, end_offset-readbuf_off+1));
		mpi_errno = ADIO_ReadContig(fd, readbuf, readbuf_len, MPI_BYTE, ADIO_EXPLICIT_OFFSET, readbuf_off, &status1);
		MPIU_ERR_CHK_POP(mpi_errno);
	}
	while (req_len > readbuf_off + readbuf_len - req_off)
	{
		partial_read = (int) (readbuf_off + readbuf_len - req_off);
		tmp_buf = (char *) MPIU_Malloc(partial_read);
		MPIU_ERR_MEM_CHK_POP(mpi_errno, tmp_buf, partial_read, L"tmp_buf");
		memcpy(tmp_buf, readbuf+readbuf_len-partial_read, partial_read);
		MPIU_Free(readbuf);
		readbuf = (char *) MPIU_Malloc(partial_read + max_bufsize);
		MPIU_ERR_MEM_CHK_POP(mpi_errno, readbuf, partial_read + max_bufsize, L"readbuf");
		memcpy(readbuf, tmp_buf, partial_read);
		MPIU_Free(tmp_buf);
		readbuf_off += readbuf_len-partial_read;
		readbuf_len = (int) (partial_read + MPIDU_MIN(max_bufsize, end_offset-readbuf_off+1));
		mpi_errno = ADIO_ReadContig(fd, readbuf+partial_read, readbuf_len-partial_read, MPI_BYTE, ADIO_EXPLICIT_OFFSET, readbuf_off+partial_read, &status1);
		MPIU_ERR_CHK_POP(mpi_errno);
	}
	memcpy((char *)buf + userbuf_off, readbuf+req_off-readbuf_off, req_len);

fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_BUFFERED_READ);
	return mpi_errno;
}
#endif

#undef FUNCNAME
#define FUNCNAME ADIOI_GEN_ReadStrided
int ADIOI_GEN_ReadStrided(MPID_File_ptr fd, void *buf, int count,
						  MPI_Datatype datatype, int file_ptr_type,
						  MPI_Offset offset, MPI_Status *status)
{
	/* offset is in units of etype relative to the filetype. */
	static const wchar_t FCNAME[] = L"ADIOI_GEN_ReadStrided";
	int mpi_errno = MPI_SUCCESS;
	ADIOI_Flatlist_node *flat_buf, *flat_file;
	int i, j, k, brd_size, frd_size=0, st_index=0;
	int bufsize, num, size, sum, n_etypes_in_filetype, size_in_filetype;
	int n_filetypes, etype_in_filetype;
	MPI_Offset abs_off_in_filetype=0;
	int filetype_size, etype_size, buftype_size, req_len, partial_read;
	MPI_Aint filetype_extent, buftype_extent; 
	int buf_count, buftype_is_contig, filetype_is_contig;
	MPI_Offset userbuf_off;
	MPI_Offset off, req_off, disp, end_offset=0, readbuf_off, start_off;
	char *readbuf, *tmp_buf;
	wchar_t *value;
	int flag, st_frd_size, st_n_filetypes, readbuf_len;
	int new_brd_size, new_frd_size, info_flag, max_bufsize;
	MPI_Status status1;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_GEN_READSTRIDED);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_GEN_READSTRIDED);

	if (fd->hints->ds_read == ADIOI_HINT_DISABLE)
	{
		/* if user has disabled data sieving on reads, use naive approach instead. */
		mpi_errno = ADIOI_GEN_ReadStrided_naive(fd, buf, count, datatype, file_ptr_type, offset, status);
		MPIU_ERR_CHK_POP(mpi_errno);
		goto fn_exit;
	}

	mpi_errno = ADIOI_Datatype_iscontig(datatype, &buftype_is_contig);
	MPIU_ERR_CHK_POP(mpi_errno);
	mpi_errno = ADIOI_Datatype_iscontig(fd->filetype, &filetype_is_contig);
	MPIU_ERR_CHK_POP(mpi_errno);

	NMPI_Type_size(mpi_errno, (fd->filetype, &filetype_size));
	MPIU_ERR_CHK_POP(mpi_errno);
	if ( ! filetype_size )
	{
		goto fn_exit;
	}

	NMPI_Type_extent(mpi_errno, (fd->filetype, &filetype_extent));
	MPIU_ERR_CHK_POP(mpi_errno);
	NMPI_Type_size(mpi_errno, (datatype, &buftype_size));
	MPIU_ERR_CHK_POP(mpi_errno);
	NMPI_Type_extent(mpi_errno, (datatype, &buftype_extent));
	MPIU_ERR_CHK_POP(mpi_errno);
	etype_size = fd->etype_size;

	bufsize = buftype_size * count;

	/* get max_bufsize from the info object. */

	value = (wchar_t *) MPIU_Malloc((MPI_MAX_INFO_VAL+1)*sizeof(wchar_t));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, value, (MPI_MAX_INFO_VAL+1)*sizeof(wchar_t), L"value");
	NMPI_Info_get(mpi_errno, (fd->info, L"ind_rd_buffer_size", MPI_MAX_INFO_VAL, value, &info_flag));
	MPIU_ERR_CHK_POP(mpi_errno);
	max_bufsize = _wtoi(value);
	MPIU_Free(value);
	value = NULL;

	if (!buftype_is_contig && filetype_is_contig)
	{
		/* noncontiguous in memory, contiguous in file. */

		mpi_errno = ADIOI_Flatten_datatype(datatype);
		MPIU_ERR_CHK_POP(mpi_errno);
		flat_buf = ADIOI_Flatlist;
		while (flat_buf->type != datatype)
		{
			flat_buf = flat_buf->next;
		}

		off = (file_ptr_type == ADIO_INDIVIDUAL) ? fd->fp_ind : fd->disp + etype_size * offset;

		start_off = off;
		end_offset = off + bufsize - 1;
		readbuf_off = off;
		readbuf = (char *) MPIU_Malloc(max_bufsize);
		MPIU_ERR_MEM_CHK_POP(mpi_errno, readbuf, max_bufsize, L"readbuf");
		readbuf_len = (int) (MPIDU_MIN(max_bufsize, end_offset-readbuf_off+1));

		/* if atomicity is true, lock (exclusive) the region to be accessed */
		if ((fd->atomicity) && (fd->file_system != ADIO_PIOFS) && (fd->file_system != ADIO_PVFS))
			ADIOI_WRITE_LOCK(fd, start_off, SEEK_SET, end_offset-start_off+1);

		mpi_errno = ADIO_ReadContig(fd, readbuf, readbuf_len, MPI_BYTE, ADIO_EXPLICIT_OFFSET, readbuf_off, &status1);
		MPIU_ERR_CHK_POP(mpi_errno);

		for (j=0; j<count; j++)
		{
			for (i=0; i<flat_buf->count; i++)
			{
				userbuf_off = j*buftype_extent + flat_buf->indices[i];
				req_off = off;
				req_len = flat_buf->blocklens[i];
				/*ADIOI_BUFFERED_READ*/
				if (req_off >= readbuf_off + readbuf_len)
				{
					readbuf_off = req_off;
					readbuf_len = (int) (MPIDU_MIN(max_bufsize, end_offset-readbuf_off+1));
					mpi_errno = ADIO_ReadContig(fd, readbuf, readbuf_len, MPI_BYTE, ADIO_EXPLICIT_OFFSET, readbuf_off, &status1);
					MPIU_ERR_CHK_POP(mpi_errno);
				}
				while (req_len > readbuf_off + readbuf_len - req_off)
				{
					partial_read = (int) (readbuf_off + readbuf_len - req_off);
					tmp_buf = (char *) MPIU_Malloc(partial_read);
					MPIU_ERR_MEM_CHK_POP(mpi_errno, tmp_buf, partial_read, L"tmp_buf");
					memcpy(tmp_buf, readbuf+readbuf_len-partial_read, partial_read);
					MPIU_Free(readbuf);
					readbuf = (char *) MPIU_Malloc(partial_read + max_bufsize);
					MPIU_ERR_MEM_CHK_POP(mpi_errno, readbuf, partial_read + max_bufsize, L"readbuf");
					memcpy(readbuf, tmp_buf, partial_read);
					MPIU_Free(tmp_buf);
					readbuf_off += readbuf_len-partial_read;
					readbuf_len = (int) (partial_read + MPIDU_MIN(max_bufsize, end_offset-readbuf_off+1));
					mpi_errno = ADIO_ReadContig(fd, readbuf+partial_read, readbuf_len-partial_read, MPI_BYTE, ADIO_EXPLICIT_OFFSET, readbuf_off+partial_read, &status1);
					MPIU_ERR_CHK_POP(mpi_errno);
				}
				memcpy((char *)buf + userbuf_off, readbuf+req_off-readbuf_off, req_len);
				off += flat_buf->blocklens[i];
			}
		}

		if ((fd->atomicity) && (fd->file_system != ADIO_PIOFS) && (fd->file_system != ADIO_PVFS))
			ADIOI_UNLOCK(fd, start_off, SEEK_SET, end_offset-start_off+1);

		if (file_ptr_type == ADIO_INDIVIDUAL) fd->fp_ind = off;

		MPIU_Free(readbuf);
		readbuf = NULL;
	}
	else
	{
		/* noncontiguous in file */

		/* filetype already flattened in ADIO_Open */
		flat_file = ADIOI_Flatlist;
		while (flat_file->type != fd->filetype)
		{
			flat_file = flat_file->next;
		}
		disp = fd->disp;

		if (file_ptr_type == ADIO_INDIVIDUAL)
		{
			offset = fd->fp_ind; /* in bytes */
			n_filetypes = -1;
			flag = 0;
			while (!flag)
			{
				n_filetypes++;
				for (i=0; i<flat_file->count; i++)
				{
					if (disp + flat_file->indices[i] + (MPI_Offset) n_filetypes*filetype_extent + flat_file->blocklens[i] >= offset)
					{
						st_index = i;
						frd_size = (int) (disp + flat_file->indices[i] + (MPI_Offset) n_filetypes*filetype_extent + flat_file->blocklens[i] - offset);
						flag = 1;
						break;
					}
				}
			}
		}
		else
		{
			n_etypes_in_filetype = filetype_size/etype_size;
			n_filetypes = (int) (offset / n_etypes_in_filetype);
			etype_in_filetype = (int) (offset % n_etypes_in_filetype);
			size_in_filetype = etype_in_filetype * etype_size;

			sum = 0;
			for (i=0; i<flat_file->count; i++)
			{
				sum += flat_file->blocklens[i];
				if (sum > size_in_filetype)
				{
					st_index = i;
					frd_size = sum - size_in_filetype;
					abs_off_in_filetype = flat_file->indices[i] + size_in_filetype - (sum - flat_file->blocklens[i]);
					break;
				}
			}

			/* abs. offset in bytes in the file */
			offset = disp + (MPI_Offset) n_filetypes*filetype_extent + abs_off_in_filetype;
		}

		start_off = offset;

		/* Calculate end_offset, the last byte-offset that will be accessed.
		e.g., if start_offset=0 and 100 bytes to be read, end_offset=99*/

		st_frd_size = frd_size;
		st_n_filetypes = n_filetypes;
		i = 0;
		j = st_index;
		off = offset;
		frd_size = MPIDU_MIN(st_frd_size, bufsize);
		while (i < bufsize)
		{
			i += frd_size;
			end_offset = off + frd_size - 1;

			if (j < (flat_file->count - 1))
			{
				j++;
			}
			else
			{
				j = 0;
				n_filetypes++;
			}

			off = disp + flat_file->indices[j] + (MPI_Offset) n_filetypes*filetype_extent;
			frd_size = MPIDU_MIN(flat_file->blocklens[j], bufsize-i);
		}

		/* if atomicity is true, lock (exclusive) the region to be accessed */
		if ((fd->atomicity) && (fd->file_system != ADIO_PIOFS) && (fd->file_system != ADIO_PVFS))
			ADIOI_WRITE_LOCK(fd, start_off, SEEK_SET, end_offset-start_off+1);

		readbuf_off = 0;
		readbuf_len = 0;
		readbuf = (char *) MPIU_Malloc(max_bufsize);
		MPIU_ERR_MEM_CHK_POP(mpi_errno, readbuf, max_bufsize, L"readbuf");

		if (buftype_is_contig && !filetype_is_contig)
		{
			/* contiguous in memory, noncontiguous in file. should be the most common case. */

			i = 0;
			j = st_index;
			off = offset;
			n_filetypes = st_n_filetypes;
			frd_size = MPIDU_MIN(st_frd_size, bufsize);
			while (i < bufsize)
			{
				if (frd_size)
				{
					/* TYPE_UB and TYPE_LB can result in frd_size = 0. save system call in such cases */ 
					/* lseek(fd->fd_sys, off, SEEK_SET);
					err = read(fd->fd_sys, ((char *) buf) + i, frd_size);*/

					req_off = off;
					req_len = frd_size;
					userbuf_off = i;
					/*ADIOI_BUFFERED_READ*/
					if (req_off >= readbuf_off + readbuf_len)
					{
						readbuf_off = req_off;
						readbuf_len = (int) (MPIDU_MIN(max_bufsize, end_offset-readbuf_off+1));
						mpi_errno = ADIO_ReadContig(fd, readbuf, readbuf_len, MPI_BYTE, ADIO_EXPLICIT_OFFSET, readbuf_off, &status1);
						MPIU_ERR_CHK_POP(mpi_errno);
					}
					while (req_len > readbuf_off + readbuf_len - req_off)
					{
						partial_read = (int) (readbuf_off + readbuf_len - req_off);
						tmp_buf = (char *) MPIU_Malloc(partial_read);
						MPIU_ERR_MEM_CHK_POP(mpi_errno, tmp_buf, partial_read, L"tmp_buf");
						memcpy(tmp_buf, readbuf+readbuf_len-partial_read, partial_read);
						MPIU_Free(readbuf);
						readbuf = (char *) MPIU_Malloc(partial_read + max_bufsize);
						MPIU_ERR_MEM_CHK_POP(mpi_errno, readbuf, partial_read + max_bufsize, L"readbuf");
						memcpy(readbuf, tmp_buf, partial_read);
						MPIU_Free(tmp_buf);
						readbuf_off += readbuf_len-partial_read;
						readbuf_len = (int) (partial_read + MPIDU_MIN(max_bufsize, end_offset-readbuf_off+1));
						mpi_errno = ADIO_ReadContig(fd, readbuf+partial_read, readbuf_len-partial_read, MPI_BYTE, ADIO_EXPLICIT_OFFSET, readbuf_off+partial_read, &status1);
						MPIU_ERR_CHK_POP(mpi_errno);
					}
					memcpy((char *)buf + userbuf_off, readbuf+req_off-readbuf_off, req_len);
				}
				i += frd_size;

				if (off + frd_size < disp + flat_file->indices[j] + flat_file->blocklens[j] + (MPI_Offset) n_filetypes*filetype_extent)
				{
					off += frd_size;
					/* did not reach end of contiguous block in filetype.
					no more I/O needed. off is incremented by frd_size. */
				}
				else
				{
					if (j < (flat_file->count - 1))
					{
						j++;
					}
					else
					{
						j = 0;
						n_filetypes++;
					}
					off = disp + flat_file->indices[j] + (MPI_Offset) n_filetypes*filetype_extent;
					frd_size = MPIDU_MIN(flat_file->blocklens[j], bufsize-i);
				}
			}
		}
		else
		{
			/* noncontiguous in memory as well as in file */

			ADIOI_Flatten_datatype(datatype);
			flat_buf = ADIOI_Flatlist;
			while (flat_buf->type != datatype)
			{
				flat_buf = flat_buf->next;
			}

			k = num = buf_count = 0;
			i = (int) (flat_buf->indices[0]);
			j = st_index;
			off = offset;
			n_filetypes = st_n_filetypes;
			frd_size = st_frd_size;
			brd_size = flat_buf->blocklens[0];

			while (num < bufsize)
			{
				size = MPIDU_MIN(frd_size, brd_size);
				if (size)
				{
					/* lseek(fd->fd_sys, off, SEEK_SET);
					err = read(fd->fd_sys, ((char *) buf) + i, size); */

					req_off = off;
					req_len = size;
					userbuf_off = i;
					/*ADIOI_BUFFERED_READ*/
					if (req_off >= readbuf_off + readbuf_len)
					{
						readbuf_off = req_off;
						readbuf_len = (int) (MPIDU_MIN(max_bufsize, end_offset-readbuf_off+1));
						mpi_errno = ADIO_ReadContig(fd, readbuf, readbuf_len, MPI_BYTE, ADIO_EXPLICIT_OFFSET, readbuf_off, &status1);
						MPIU_ERR_CHK_POP(mpi_errno);
					}
					while (req_len > readbuf_off + readbuf_len - req_off)
					{
						partial_read = (int) (readbuf_off + readbuf_len - req_off);
						tmp_buf = (char *) MPIU_Malloc(partial_read);
						MPIU_ERR_MEM_CHK_POP(mpi_errno, tmp_buf, partial_read, L"tmp_buf");
						memcpy(tmp_buf, readbuf+readbuf_len-partial_read, partial_read);
						MPIU_Free(readbuf);
						readbuf = (char *) MPIU_Malloc(partial_read + max_bufsize);
						MPIU_ERR_MEM_CHK_POP(mpi_errno, readbuf, partial_read + max_bufsize, L"readbuf");
						memcpy(readbuf, tmp_buf, partial_read);
						MPIU_Free(tmp_buf);
						readbuf_off += readbuf_len-partial_read;
						readbuf_len = (int) (partial_read + MPIDU_MIN(max_bufsize, end_offset-readbuf_off+1));
						mpi_errno = ADIO_ReadContig(fd, readbuf+partial_read, readbuf_len-partial_read, MPI_BYTE, ADIO_EXPLICIT_OFFSET, readbuf_off+partial_read, &status1);
						MPIU_ERR_CHK_POP(mpi_errno);
					}
					memcpy((char *)buf + userbuf_off, readbuf+req_off-readbuf_off, req_len);
				}

				new_frd_size = frd_size;
				new_brd_size = brd_size;

				if (size == frd_size)
				{
					/* reached end of contiguous block in file */
					if (j < (flat_file->count - 1))
					{
						j++;
					}
					else
					{
						j = 0;
						n_filetypes++;
					}

					off = disp + flat_file->indices[j] + (MPI_Offset) n_filetypes*filetype_extent;

					new_frd_size = flat_file->blocklens[j];
					if (size != brd_size)
					{
						i += size;
						new_brd_size -= size;
					}
				}

				if (size == brd_size)
				{
					/* reached end of contiguous block in memory */

					k = (k + 1)%flat_buf->count;
					buf_count++;
					i = (int) (buftype_extent*(buf_count/flat_buf->count) + flat_buf->indices[k]);
					new_brd_size = flat_buf->blocklens[k];
					if (size != frd_size)
					{
						off += size;
						new_frd_size -= size;
					}
				}
				num += size;
				frd_size = new_frd_size;
				brd_size = new_brd_size;
			}
		}

		if ((fd->atomicity) && (fd->file_system != ADIO_PIOFS) && (fd->file_system != ADIO_PVFS))
			ADIOI_UNLOCK(fd, start_off, SEEK_SET, end_offset-start_off+1);

		if (file_ptr_type == ADIO_INDIVIDUAL) fd->fp_ind = off;

		MPIU_Free(readbuf); /* malloced in the buffered_read macro */
		readbuf = NULL;
	}

	fd->fp_sys_posn = -1;   /* set it to null. */

#ifdef HAVE_STATUS_SET_BYTES
	MPIR_Status_set_bytes(status, datatype, bufsize);
	/* This is a temporary way of filling in status. The right way is to 
	keep track of how much data was actually read and placed in buf 
	by ADIOI_BUFFERED_READ. */
#endif

	if (!buftype_is_contig)
	{
		mpi_errno = ADIOI_Delete_flattened(datatype);
		MPIU_ERR_CHK_POP(mpi_errno);
	}

fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_GEN_READSTRIDED);
	return mpi_errno;
fn_fail:
	/* free pointers that are not NULL? */
	goto fn_exit;
}

/* prototypes of functions used for collective reads only. */
static int ADIOI_Read_and_exch(MPID_File_ptr fd, void *buf, MPI_Datatype
							   datatype, int nprocs,
							   int myrank, ADIOI_Access
							   *others_req, MPI_Offset *offset_list,
							   int *len_list, int contig_access_count, 
							   MPI_Offset
							   min_st_offset, MPI_Offset fd_size,
							   MPI_Offset *fd_start, MPI_Offset *fd_end,
							   int *buf_idx);
static int ADIOI_R_Exchange_data(MPID_File_ptr fd, void *buf, ADIOI_Flatlist_node
								 *flat_buf, MPI_Offset *offset_list, int
								 *len_list, int *send_size, int *recv_size,
								 int *count, int *start_pos, 
								 int *partial_send, 
								 int *recd_from_proc, int nprocs, 
								 int myrank, int
								 buftype_is_contig, int contig_access_count,
								 MPI_Offset min_st_offset, 
								 MPI_Offset fd_size,
								 MPI_Offset *fd_start, MPI_Offset *fd_end, 
								 ADIOI_Access *others_req, 
								 int iter, 
								 MPI_Aint buftype_extent, int *buf_idx);
static int ADIOI_Fill_user_buffer(MPID_File_ptr fd, void *buf, ADIOI_Flatlist_node
								  *flat_buf, char **recv_buf, MPI_Offset 
								  *offset_list, int *len_list, 
								  int *recv_size, 
								  MPI_Request *requests, MPI_Status *statuses,
								  int *recd_from_proc, int nprocs,
								  int contig_access_count, 
								  MPI_Offset min_st_offset, 
								  MPI_Offset fd_size, MPI_Offset *fd_start, 
								  MPI_Offset *fd_end,
								  MPI_Aint buftype_extent);


#undef FUNCNAME
#define FUNCNAME ADIOI_GEN_ReadStridedColl
int ADIOI_GEN_ReadStridedColl(MPID_File_ptr fd, void *buf, int count,
							  MPI_Datatype datatype, int file_ptr_type,
							  MPI_Offset offset, MPI_Status *status)
{
	/* Uses a generalized version of the extended two-phase method described
	in "An Extended Two-Phase Method for Accessing Sections of 
	Out-of-Core Arrays", Rajeev Thakur and Alok Choudhary,
	Scientific Programming, (5)4:301--317, Winter 1996. 
	http://www.mcs.anl.gov/home/thakur/ext2ph.ps */

	int mpi_errno = MPI_SUCCESS;
	static const wchar_t FCNAME[] = L"ADIOI_GEN_ReadStridedColl";
	ADIOI_Access *my_req; 
	/* array of nprocs structures, one for each other process in
	whose file domain this process's request lies */
	ADIOI_Access *others_req;
	/* array of nprocs structures, one for each other process
	whose request lies in this process's file domain. */
	int i, filetype_is_contig, nprocs, nprocs_for_coll, myrank;
	int contig_access_count=0, interleave_count = 0, buftype_is_contig;
	int *count_my_req_per_proc, count_my_req_procs, count_others_req_procs;
	MPI_Offset start_offset, end_offset, orig_fp, fd_size, min_st_offset, off;
	MPI_Offset *offset_list = NULL, *st_offsets = NULL, *fd_start = NULL,
		*fd_end = NULL, *end_offsets = NULL;
	int *len_list = NULL, *buf_idx = NULL;
#ifdef HAVE_STATUS_SET_BYTES
	int bufsize, size;
#endif
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_GEN_READSTRIDEDCOLL);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_GEN_READSTRIDEDCOLL);

#ifdef PROFILE
	MPE_Log_event(13, 0, L"start computation");
#endif

	NMPI_Comm_size(mpi_errno, (fd->comm, &nprocs));
	MPIU_ERR_CHK_POP(mpi_errno);
	NMPI_Comm_rank(mpi_errno, (fd->comm, &myrank));
	MPIU_ERR_CHK_POP(mpi_errno);

	/* number of aggregators, cb_nodes, is stored in the hints */
	nprocs_for_coll = fd->hints->cb_nodes;
	orig_fp = fd->fp_ind;

	/* only check for interleaving if cb_read isn't disabled */
	if (fd->hints->cb_read != ADIOI_HINT_DISABLE)
	{
		/* For this process's request, calculate the list of offsets and
		lengths in the file and determine the start and end offsets. */

		/* Note: end_offset points to the last byte-offset that will be accessed.
		e.g., if start_offset=0 and 100 bytes to be read, end_offset=99*/

		mpi_errno = ADIOI_Calc_my_off_len(fd, count, datatype, file_ptr_type, offset,
			&offset_list, &len_list, &start_offset,
			&end_offset, &contig_access_count);
		MPIU_ERR_CHK_POP(mpi_errno);

		/*	for (i=0; i<contig_access_count; i++) {
		FPRINTF(stderr, L"rank %d  off %ld  len %d\n", myrank, offset_list[i], 
		len_list[i]);
		}*/

		/* each process communicates its start and end offsets to other 
		processes. The result is an array each of start and end offsets stored
		in order of process rank. */ 

		st_offsets = (MPI_Offset *) MPIU_Malloc(nprocs*sizeof(MPI_Offset));
		MPIU_ERR_MEM_CHK_POP(mpi_errno, st_offsets, nprocs*sizeof(MPI_Offset), L"st_offsets");
		end_offsets = (MPI_Offset *) MPIU_Malloc(nprocs*sizeof(MPI_Offset));
		MPIU_ERR_MEM_CHK_POP(mpi_errno, end_offsets, nprocs*sizeof(MPI_Offset), L"end_offsets");

		NMPI_Allgather(mpi_errno, (&start_offset, 1, ADIO_OFFSET, st_offsets, 1, ADIO_OFFSET, fd->comm));
		MPIU_ERR_CHK_POP(mpi_errno);
		NMPI_Allgather(mpi_errno, (&end_offset, 1, ADIO_OFFSET, end_offsets, 1, ADIO_OFFSET, fd->comm));
		MPIU_ERR_CHK_POP(mpi_errno);

		/* are the accesses of different processes interleaved? */
		for (i=1; i<nprocs; i++)
		{
			if ((st_offsets[i] < end_offsets[i-1]) && 
				(st_offsets[i] <= end_offsets[i]))
			{
				interleave_count++;
			}
		}
		/* This is a rudimentary check for interleaving, but should suffice
		for the moment. */
	}

	mpi_errno = ADIOI_Datatype_iscontig(datatype, &buftype_is_contig);
	MPIU_ERR_CHK_POP(mpi_errno);

	if (fd->hints->cb_read == ADIOI_HINT_DISABLE || (!interleave_count && (fd->hints->cb_read == ADIOI_HINT_AUTO))) 
	{
		/* don't do aggregation */
		if (fd->hints->cb_read != ADIOI_HINT_DISABLE)
		{
			MPIU_Free(offset_list);
			MPIU_Free(len_list);
			MPIU_Free(st_offsets);
			MPIU_Free(end_offsets);
			offset_list = NULL;
			len_list = NULL;
			st_offsets = NULL;
			end_offsets = NULL;
		}

		fd->fp_ind = orig_fp;
		mpi_errno = ADIOI_Datatype_iscontig(fd->filetype, &filetype_is_contig);
		MPIU_ERR_CHK_POP(mpi_errno);

		if (buftype_is_contig && filetype_is_contig)
		{
			if (file_ptr_type == ADIO_EXPLICIT_OFFSET)
			{
				off = fd->disp + (fd->etype_size) * offset;
				mpi_errno = ADIO_ReadContig(fd, buf, count, datatype, ADIO_EXPLICIT_OFFSET, off, status);
				MPIU_ERR_CHK_POP(mpi_errno);
			}
			else
			{
				mpi_errno = ADIO_ReadContig(fd, buf, count, datatype, ADIO_INDIVIDUAL, 0, status);
				MPIU_ERR_CHK_POP(mpi_errno);
			}
		}
		else
		{
			mpi_errno = ADIO_ReadStrided(fd, buf, count, datatype, file_ptr_type, offset, status);
			MPIU_ERR_CHK_POP(mpi_errno);
		}

		goto fn_exit;
	}

	/* We're going to perform aggregation of I/O.  Here we call
	* ADIOI_Calc_file_domains() to determine what processes will handle I/O
	* to what regions.  We pass nprocs_for_coll into this function; it is
	* used to determine how many processes will perform I/O, which is also
	* the number of regions into which the range of bytes must be divided.
	* These regions are called "file domains", or FDs.
	*
	* When this function returns, fd_start, fd_end, fd_size, and
	* min_st_offset will be filled in.  fd_start holds the starting byte
	* location for each file domain.  fd_end holds the ending byte location.
	* min_st_offset holds the minimum byte location that will be accessed.
	*
	* Both fd_start[] and fd_end[] are indexed by an aggregator number; this
	* needs to be mapped to an actual rank in the communicator later.
	*
	*/
	mpi_errno = ADIOI_Calc_file_domains(st_offsets, end_offsets, nprocs, nprocs_for_coll, &min_st_offset, &fd_start, &fd_end, &fd_size);
	MPIU_ERR_CHK_POP(mpi_errno);

	/* calculate where the portions of the access requests of this process 
	* are located in terms of the file domains.  this could be on the same
	* process or on other processes.  this function fills in:
	* count_my_req_procs - number of processes (including this one) for which
	*	 this process has requests in their file domain
	* count_my_req_per_proc - count of requests for each process, indexed
	*	 by rank of the process
	* my_req[] - array of data structures describing the requests to be
	*	 performed by each process (including self).  indexed by rank.
	* buf_idx[] - array of locations into which data can be directly moved;
	*	 this is only valid for contiguous buffer case
	*/
	mpi_errno = ADIOI_Calc_my_req(fd, offset_list, len_list, contig_access_count,
		min_st_offset, fd_start, fd_end, fd_size,
		nprocs, &count_my_req_procs, 
		&count_my_req_per_proc, &my_req,
		&buf_idx);
	MPIU_ERR_CHK_POP(mpi_errno);

	/* perform a collective communication in order to distribute the
	* data calculated above.  fills in the following:
	* count_others_req_procs - number of processes (including this
	*	 one) which have requests in this process's file domain.
	* count_others_req_per_proc[] - number of separate contiguous
	*	 requests from proc i lie in this process's file domain.
	*/
	mpi_errno = ADIOI_Calc_others_req(fd, count_my_req_procs, 
		count_my_req_per_proc, my_req, 
		nprocs, myrank, &count_others_req_procs, 
		&others_req);
	MPIU_ERR_CHK_POP(mpi_errno);

	/* my_req[] and count_my_req_per_proc aren't needed at this point, so 
	* let's free the memory 
	*/
	MPIU_Free(count_my_req_per_proc);
	count_my_req_per_proc = NULL;
	for (i=0; i<nprocs; i++)
	{
		if (my_req[i].count)
		{
			MPIU_Free(my_req[i].offsets);
			MPIU_Free(my_req[i].lens);
			/*
			my_req[i].offsets = NULL;
			my_req[i].lens = NULL;
			*/
		}
	}
	MPIU_Free(my_req);
	my_req = NULL;


	/* read data in sizes of no more than ADIOI_Coll_bufsize, 
	* communicate, and fill user buf. 
	*/
	mpi_errno = ADIOI_Read_and_exch(fd, buf, datatype, nprocs, myrank,
		others_req, offset_list,
		len_list, contig_access_count, min_st_offset,
		fd_size, fd_start, fd_end, buf_idx);
	MPIU_ERR_CHK_POP(mpi_errno);

	if (!buftype_is_contig)
	{
		mpi_errno = ADIOI_Delete_flattened(datatype);
		MPIU_ERR_CHK_POP(mpi_errno);
	}

	/* free all memory allocated for collective I/O */
	for (i=0; i<nprocs; i++)
	{
		if (others_req[i].count)
		{
			MPIU_Free(others_req[i].offsets);
			MPIU_Free(others_req[i].lens);
			MPIU_Free(others_req[i].mem_ptrs);
		}
	}
	MPIU_Free(others_req);
	others_req = NULL;

	MPIU_Free(buf_idx);
	MPIU_Free(offset_list);
	MPIU_Free(len_list);
	MPIU_Free(st_offsets);
	MPIU_Free(end_offsets);
	MPIU_Free(fd_start);
	MPIU_Free(fd_end);
	buf_idx = NULL;
	offset_list = NULL;
	len_list = NULL;
	st_offsets = NULL;
	end_offsets = NULL;
	fd_start = NULL;
	fd_end = NULL;

#ifdef HAVE_STATUS_SET_BYTES
	NMPI_Type_size(mpi_errno, (datatype, &size));
	MPIU_ERR_CHK_POP(mpi_errno);
	bufsize = size * count;
	MPIR_Status_set_bytes(status, datatype, bufsize);
	/* This is a temporary way of filling in status. The right way is to 
	keep track of how much data was actually read and placed in buf 
	during collective I/O. */
#endif

	fd->fp_sys_posn = -1;   /* set it to null. */

fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_GEN_READSTRIDEDCOLL);
	return mpi_errno;

fn_fail:
	/* Free buffers that are not NULL? */
	goto fn_exit;
}

#undef FUNCNAME
#define FUNCNAME ADIOI_Calc_my_off_len
int ADIOI_Calc_my_off_len(MPID_File_ptr fd, int bufcount, MPI_Datatype
						  datatype, int file_ptr_type, MPI_Offset
						  offset, MPI_Offset **offset_list_ptr, int
						  **len_list_ptr, MPI_Offset *start_offset_ptr,
						  MPI_Offset *end_offset_ptr, int
						  *contig_access_count_ptr)
{
	static const wchar_t FCNAME[] = L"ADIOI_Calc_my_off_len";
	int mpi_errno = MPI_SUCCESS;
	int filetype_size, buftype_size, etype_size;
	int i, j, k, frd_size=0, old_frd_size=0, st_index=0;
	int n_filetypes, etype_in_filetype;
	MPI_Offset abs_off_in_filetype=0;
	int bufsize, sum, n_etypes_in_filetype, size_in_filetype;
	int contig_access_count, *len_list, flag, filetype_is_contig;
	MPI_Aint filetype_extent, filetype_lb;
	ADIOI_Flatlist_node *flat_file;
	MPI_Offset *offset_list, off, end_offset=0, disp;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_CALC_MY_OFF_LEN);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_CALC_MY_OFF_LEN);

	/* For this process's request, calculate the list of offsets and
	lengths in the file and determine the start and end offsets. */

	mpi_errno = ADIOI_Datatype_iscontig(fd->filetype, &filetype_is_contig);
	MPIU_ERR_CHK_POP(mpi_errno);

	NMPI_Type_size(mpi_errno, (fd->filetype, &filetype_size));
	MPIU_ERR_CHK_POP(mpi_errno);
	NMPI_Type_extent(mpi_errno, (fd->filetype, &filetype_extent));
	MPIU_ERR_CHK_POP(mpi_errno);
	NMPI_Type_lb(mpi_errno, (fd->filetype, &filetype_lb));
	MPIU_ERR_CHK_POP(mpi_errno);
	NMPI_Type_size(mpi_errno, (datatype, &buftype_size));
	MPIU_ERR_CHK_POP(mpi_errno);
	etype_size = fd->etype_size;

	if ( ! filetype_size )
	{
		*contig_access_count_ptr = 0;
		*offset_list_ptr = (MPI_Offset *) MPIU_Malloc(2*sizeof(MPI_Offset));
		MPIU_ERR_MEM_CHK_POP(mpi_errno, *offset_list_ptr, 2*sizeof(MPI_Offset), L"offset_list_ptr");
		*len_list_ptr = (int *) MPIU_Malloc(2*sizeof(int));
		MPIU_ERR_MEM_CHK_POP(mpi_errno, *len_list_ptr, 2*sizeof(int), L"len_list_ptr");
		/* 2 is for consistency. everywhere I malloc one more than needed */

		offset_list = *offset_list_ptr;
		len_list = *len_list_ptr;
		offset_list[0] = (file_ptr_type == ADIO_INDIVIDUAL) ? fd->fp_ind : fd->disp + etype_size * offset;
		len_list[0] = 0;
		*start_offset_ptr = offset_list[0];
		*end_offset_ptr = offset_list[0] + len_list[0] - 1;

		goto fn_exit;
	}

	if (filetype_is_contig)
	{
		*contig_access_count_ptr = 1;		
		*offset_list_ptr = (MPI_Offset *) MPIU_Malloc(2*sizeof(MPI_Offset));
		MPIU_ERR_MEM_CHK_POP(mpi_errno, *offset_list_ptr, 2*sizeof(MPI_Offset), L"offset_list_ptr");
		*len_list_ptr = (int *) MPIU_Malloc(2*sizeof(int));
		MPIU_ERR_MEM_CHK_POP(mpi_errno, *len_list_ptr, 2*sizeof(int), L"len_list_ptr");
		/* 2 is for consistency. everywhere I malloc one more than needed */

		offset_list = *offset_list_ptr;
		len_list = *len_list_ptr;
		offset_list[0] = (file_ptr_type == ADIO_INDIVIDUAL) ? fd->fp_ind : fd->disp + etype_size * offset;
		len_list[0] = bufcount * buftype_size;
		*start_offset_ptr = offset_list[0];
		*end_offset_ptr = offset_list[0] + len_list[0] - 1;

		/* update file pointer */
		if (file_ptr_type == ADIO_INDIVIDUAL)
		{
			fd->fp_ind = *end_offset_ptr + 1;
		}
	}
	else
	{
		/* First calculate what size of offset_list and len_list to allocate */

		/* filetype already flattened in ADIO_Open or ADIO_Fcntl */
		flat_file = ADIOI_Flatlist;
		while (flat_file->type != fd->filetype)
		{
			flat_file = flat_file->next;
		}
		disp = fd->disp;

		if (file_ptr_type == ADIO_INDIVIDUAL)
		{
			offset = fd->fp_ind; /* in bytes */
			n_filetypes = -1;
			flag = 0;
			while (!flag)
			{
				n_filetypes++;
				for (i=0; i<flat_file->count; i++)
				{
					if (disp + flat_file->indices[i] + (MPI_Offset) n_filetypes*filetype_extent + flat_file->blocklens[i] >= offset) 
					{
						st_index = i;
						frd_size = (int) (disp + flat_file->indices[i] + (MPI_Offset) n_filetypes*filetype_extent + flat_file->blocklens[i] - offset);
						flag = 1;
						break;
					}
				}
			}
		}
		else
		{
			n_etypes_in_filetype = filetype_size/etype_size;
			n_filetypes = (int) (offset / n_etypes_in_filetype);
			etype_in_filetype = (int) (offset % n_etypes_in_filetype);
			size_in_filetype = etype_in_filetype * etype_size;

			sum = 0;
			for (i=0; i<flat_file->count; i++)
			{
				sum += flat_file->blocklens[i];
				if (sum > size_in_filetype)
				{
					st_index = i;
					frd_size = sum - size_in_filetype;
					abs_off_in_filetype = flat_file->indices[i] + size_in_filetype - (sum - flat_file->blocklens[i]);
					break;
				}
			}

			/* abs. offset in bytes in the file */
			offset = disp + (MPI_Offset) n_filetypes*filetype_extent + abs_off_in_filetype;
		}

		/* calculate how much space to allocate for offset_list, len_list */

		old_frd_size = frd_size;
		contig_access_count = i = 0;
		j = st_index;
		bufsize = buftype_size * bufcount;
		frd_size = MPIDU_MIN(frd_size, bufsize);
		while (i < bufsize)
		{
			if (frd_size) contig_access_count++;
			i += frd_size;
			j = (j + 1) % flat_file->count;
			frd_size = MPIDU_MIN(flat_file->blocklens[j], bufsize-i);
		}

		/* allocate space for offset_list and len_list */

		*offset_list_ptr = (MPI_Offset *)MPIU_Malloc((contig_access_count+1)*sizeof(MPI_Offset));
		MPIU_ERR_MEM_CHK_POP(mpi_errno, *offset_list_ptr, (contig_access_count+1)*sizeof(MPI_Offset), L"offset_list_ptr");
		*len_list_ptr = (int *) MPIU_Malloc((contig_access_count+1)*sizeof(int));
		MPIU_ERR_MEM_CHK_POP(mpi_errno, *len_list_ptr, (contig_access_count+1)*sizeof(int), L"len_list_ptr");
		/* +1 to avoid a 0-size malloc */

		offset_list = *offset_list_ptr;
		len_list = *len_list_ptr;

		/* find start offset, end offset, and fill in offset_list and len_list */

		*start_offset_ptr = offset; /* calculated above */

		i = k = 0;
		j = st_index;
		off = offset;
		frd_size = MPIDU_MIN(old_frd_size, bufsize);
		while (i < bufsize)
		{
			if (frd_size)
			{
				offset_list[k] = off;
				len_list[k] = frd_size;
				k++;
			}
			i += frd_size;
			end_offset = off + frd_size - 1;

			/* Note: end_offset points to the last byte-offset that will be accessed.
			e.g., if start_offset=0 and 100 bytes to be read, end_offset=99*/

			if (off + frd_size < disp + flat_file->indices[j] + flat_file->blocklens[j] + (MPI_Offset) n_filetypes*filetype_extent)
			{
				off += frd_size;
				/* did not reach end of contiguous block in filetype.
				* no more I/O needed. off is incremented by frd_size. 
				*/
			}
			else
			{
				if (j < (flat_file->count - 1)) j++;
				else
				{
					/* hit end of flattened filetype; 
					* start at beginning again 
					*/
					j = 0;
					n_filetypes++;
				}
				off = disp + flat_file->indices[j] + (MPI_Offset) n_filetypes*filetype_extent;
				frd_size = MPIDU_MIN(flat_file->blocklens[j], bufsize-i);
			}
		}

		/* update file pointer */
		if (file_ptr_type == ADIO_INDIVIDUAL) fd->fp_ind = off;

		*contig_access_count_ptr = contig_access_count;
		*end_offset_ptr = end_offset;
	}

fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_CALC_MY_OFF_LEN);
	return mpi_errno;
fn_fail:
	goto fn_exit;
}

#undef FUNCNAME
#define FUNCNAME ADIOI_Read_and_exch
static int ADIOI_Read_and_exch(MPID_File_ptr fd, void *buf, MPI_Datatype
							   datatype, int nprocs,
							   int myrank, ADIOI_Access
							   *others_req, MPI_Offset *offset_list,
							   int *len_list, int contig_access_count, MPI_Offset
							   min_st_offset, MPI_Offset fd_size,
							   MPI_Offset *fd_start, MPI_Offset *fd_end,
							   int *buf_idx)
{
	/* Read in sizes of no more than coll_bufsize, an info parameter.
	Send data to appropriate processes. 
	Place recd. data in user buf.
	The idea is to reduce the amount of extra memory required for
	collective I/O. If all data were read all at once, which is much
	easier, it would require temp space more than the size of user_buf,
	which is often unacceptable. For example, to read a distributed
	array from a file, where each local array is 8Mbytes, requiring
	at least another 8Mbytes of temp space is unacceptable. */

	static const wchar_t FCNAME[] = L"ADIOI_Read_and_exch";
	int mpi_errno = MPI_SUCCESS;
	int i, j, m, size, ntimes, max_ntimes, buftype_is_contig;
	MPI_Offset st_loc=-1, end_loc=-1, off, done, real_off, req_off;
	char *read_buf = NULL, *tmp_buf;
	int *curr_offlen_ptr, *count, *send_size, *recv_size;
	int *partial_send, *recd_from_proc, *start_pos, for_next_iter;
	int real_size, req_len, flag, for_curr_iter, rank;
	MPI_Status status;
	ADIOI_Flatlist_node *flat_buf=NULL;
	MPI_Aint buftype_extent;
	int coll_bufsize;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_READ_AND_EXCH);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_READ_AND_EXCH);

	/* calculate the number of reads of size coll_bufsize
	to be done by each process and the max among all processes.
	That gives the no. of communication phases as well.
	coll_bufsize is obtained from the hints object. */

	coll_bufsize = fd->hints->cb_buffer_size;

	/* grab some initial values for st_loc and end_loc */
	for (i=0; i < nprocs; i++)
	{
		if (others_req[i].count)
		{
			st_loc = others_req[i].offsets[0];
			end_loc = others_req[i].offsets[0];
			break;
		}
	}

	/* now find the real values */
	for (i=0; i < nprocs; i++)
	{
		for (j=0; j<others_req[i].count; j++)
		{
			st_loc = MPIDU_MIN(st_loc, others_req[i].offsets[j]);
			end_loc = MPIDU_MAX(end_loc, (others_req[i].offsets[j] + others_req[i].lens[j] - 1));
		}
	}

	/* calculate ntimes, the number of times this process must perform I/O
	* operations in order to complete all the requests it has received.
	* the need for multiple I/O operations comes from the restriction that
	* we only use coll_bufsize bytes of memory for internal buffering.
	*/
	if ((st_loc==-1) && (end_loc==-1))
	{
		/* this process does no I/O. */
		ntimes = 0;
	}
	else
	{
		/* ntimes=ceiling_div(end_loc - st_loc + 1, coll_bufsize)*/
		ntimes = (int) ((end_loc - st_loc + coll_bufsize)/coll_bufsize);
	}

	NMPI_Allreduce(mpi_errno, (&ntimes, &max_ntimes, 1, MPI_INT, MPI_MAX, fd->comm));
	MPIU_ERR_CHK_POP(mpi_errno);

	if (ntimes)
	{
		read_buf = (char *) MPIU_Malloc(coll_bufsize);
		MPIU_ERR_MEM_CHK_POP(mpi_errno, read_buf, coll_bufsize, L"read_buf");
	}

	curr_offlen_ptr = (int *) MPIU_Calloc(nprocs, sizeof(int));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, curr_offlen_ptr, nprocs * sizeof(int), L"curr_offlen_ptr");
	/* its use is explained below. calloc initializes to 0. */

	count = (int *) MPIU_Malloc(nprocs * sizeof(int));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, count, nprocs * sizeof(int), L"count");
	/* to store count of how many off-len pairs per proc are satisfied
	in an iteration. */

	partial_send = (int *) MPIU_Calloc(nprocs, sizeof(int));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, partial_send, nprocs * sizeof(int), L"partial_send");
	/* if only a portion of the last off-len pair is sent to a process 
	in a particular iteration, the length sent is stored here.
	calloc initializes to 0. */

	send_size = (int *) MPIU_Malloc(nprocs * sizeof(int));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, send_size, nprocs * sizeof(int), L"send_size");
	/* total size of data to be sent to each proc. in an iteration */

	recv_size = (int *) MPIU_Malloc(nprocs * sizeof(int));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, recv_size, nprocs * sizeof(int), L"recv_size");
	/* total size of data to be recd. from each proc. in an iteration.
	Of size nprocs so that I can use MPI_Alltoall later. */

	recd_from_proc = (int *) MPIU_Calloc(nprocs, sizeof(int));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, recd_from_proc, nprocs * sizeof(int), L"recd_from_proc");
	/* amount of data recd. so far from each proc. Used in
	ADIOI_Fill_user_buffer. initialized to 0 here. */

	start_pos = (int *) MPIU_Malloc(nprocs * sizeof(int));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, start_pos, nprocs * sizeof(int), L"start_pos");
	/* used to store the starting value of curr_offlen_ptr[i] in 
	this iteration */

	mpi_errno = ADIOI_Datatype_iscontig(datatype, &buftype_is_contig);
	MPIU_ERR_CHK_POP(mpi_errno);
	if (!buftype_is_contig)
	{
		mpi_errno = ADIOI_Flatten_datatype(datatype);
		MPIU_ERR_CHK_POP(mpi_errno);
		flat_buf = ADIOI_Flatlist;
		while (flat_buf->type != datatype)
		{
			flat_buf = flat_buf->next;
		}
	}
	NMPI_Type_extent(mpi_errno, (datatype, &buftype_extent));
	MPIU_ERR_CHK_POP(mpi_errno);

	done = 0;
	off = st_loc;
	for_curr_iter = for_next_iter = 0;

	NMPI_Comm_rank(mpi_errno, (fd->comm, &rank));
	MPIU_ERR_CHK_POP(mpi_errno);

#ifdef PROFILE
	MPE_Log_event(14, 0, L"end computation");
#endif

	for (m=0; m<ntimes; m++)
	{
		/* read buf of size coll_bufsize (or less) */
		/* go through all others_req and check if any are satisfied
		by the current read */

		/* since MPI guarantees that displacements in filetypes are in 
		monotonically nondecreasing order, I can maintain a pointer
		(curr_offlen_ptr) to 
		current off-len pair for each process in others_req and scan
		further only from there. There is still a problem of filetypes
		such as:  (1, 2, 3 are not process nos. They are just numbers for
		three chunks of data, specified by a filetype.)

		1  -------!--
		2	-----!----
		3	   --!-----

		where ! indicates where the current read_size limitation cuts 
		through the filetype.  I resolve this by reading up to !, but
		filling the communication buffer only for 1. I copy the portion
		left over for 2 into a tmp_buf for use in the next
		iteration. i.e., 2 and 3 will be satisfied in the next
		iteration. This simplifies filling in the user's buf at the
		other end, as only one off-len pair with incomplete data
		will be sent. I also don't need to send the individual
		offsets and lens along with the data, as the data is being
		sent in a particular order. */ 

		/* off = start offset in the file for the data actually read in 
		this iteration 
		size = size of data read corresponding to off
		real_off = off minus whatever data was retained in memory from
		previous iteration for cases like 2, 3 illustrated above
		real_size = size plus the extra corresponding to real_off
		req_off = off in file for a particular contiguous request 
		minus what was satisfied in previous iteration
		req_size = size corresponding to req_off */

#ifdef PROFILE
		MPE_Log_event(13, 0, L"start computation");
#endif
		size = (int) (MPIDU_MIN(coll_bufsize, end_loc-st_loc+1-done)); 
		real_off = off - for_curr_iter;
		real_size = size + for_curr_iter;

		for (i=0; i<nprocs; i++)
		{
			count[i] = send_size[i] = 0;
		}
		for_next_iter = 0;

		for (i=0; i<nprocs; i++)
		{
			/* FPRINTF(stderr, L"rank %d, i %d, others_count %d\n", rank, i, others_req[i].count); */
			if (others_req[i].count)
			{
				start_pos[i] = curr_offlen_ptr[i];
				for (j=curr_offlen_ptr[i]; j<others_req[i].count; j++)
				{
					if (partial_send[i])
					{
						/* this request may have been partially
						satisfied in the previous iteration. */
						req_off = others_req[i].offsets[j] + partial_send[i]; 
						req_len = others_req[i].lens[j] - partial_send[i];
						partial_send[i] = 0;
						/* modify the off-len pair to reflect this change */
						others_req[i].offsets[j] = req_off;
						others_req[i].lens[j] = req_len;
					}
					else
					{
						req_off = others_req[i].offsets[j];
						req_len = others_req[i].lens[j];
					}
					if (req_off < real_off + real_size)
					{
						count[i]++;
						NMPI_Address(mpi_errno, (read_buf+req_off-real_off, &(others_req[i].mem_ptrs[j])));
						MPIU_ERR_CHK_POP(mpi_errno);
						send_size[i] += (int)(MPIDU_MIN(real_off + (MPI_Offset)real_size - req_off, req_len));

						if (real_off+real_size-req_off < req_len)
						{
							partial_send[i] = (int) (real_off+real_size - req_off);
							if ((j+1 < others_req[i].count) && (others_req[i].offsets[j+1] < real_off+real_size))
							{
								/* this is the case illustrated in the figure above. */
								for_next_iter = (int) (MPIDU_MAX(for_next_iter, real_off + real_size - others_req[i].offsets[j+1]));
								/* max because it must cover requests from different processes */
							}
							break;
						}
					}
					else
					{
						break;
					}
				}
				curr_offlen_ptr[i] = j;
			}
		}

		flag = 0;
		for (i=0; i<nprocs; i++)
		{
			if (count[i]) flag = 1;
		}

#ifdef PROFILE
		MPE_Log_event(14, 0, L"end computation");
#endif
		if (flag)
		{
			mpi_errno = ADIO_ReadContig(fd, read_buf+for_curr_iter, size, MPI_BYTE, ADIO_EXPLICIT_OFFSET, off, &status);
			MPIU_ERR_CHK_POP(mpi_errno);
		}

		for_curr_iter = for_next_iter;

#ifdef PROFILE
		MPE_Log_event(7, 0, L"start communication");
#endif
		mpi_errno = ADIOI_R_Exchange_data(fd, buf, flat_buf, offset_list, len_list,
			send_size, recv_size, count, 
			start_pos, partial_send, recd_from_proc, nprocs,
			myrank, 
			buftype_is_contig, contig_access_count,
			min_st_offset, fd_size, fd_start, fd_end,
			others_req, 
			m, buftype_extent, buf_idx);
		MPIU_ERR_CHK_POP(mpi_errno);

#ifdef PROFILE
		MPE_Log_event(8, 0, L"end communication");
#endif

		if (for_next_iter)
		{
			tmp_buf = (char *) MPIU_Malloc(for_next_iter);
			MPIU_ERR_MEM_CHK_POP(mpi_errno, tmp_buf, for_next_iter, L"tmp_buf");
			memcpy(tmp_buf, read_buf+real_size-for_next_iter, for_next_iter);
			MPIU_Free(read_buf);
			read_buf = (char *) MPIU_Malloc(for_next_iter+coll_bufsize);
			MPIU_ERR_MEM_CHK_POP(mpi_errno, read_buf, for_next_iter+coll_bufsize, L"read_buf");
			memcpy(read_buf, tmp_buf, for_next_iter);
			MPIU_Free(tmp_buf);
		}

		off += size;
		done += size;
	}

	for (i=0; i<nprocs; i++)
	{
		count[i] = send_size[i] = 0;
	}
#ifdef PROFILE
	MPE_Log_event(7, 0, L"start communication");
#endif
	for (m=ntimes; m<max_ntimes; m++)
	{
		/* nothing to send, but check for recv. */
		mpi_errno = ADIOI_R_Exchange_data(fd, buf, flat_buf, offset_list, len_list,
			send_size, recv_size, count, 
			start_pos, partial_send, recd_from_proc, nprocs,
			myrank, 
			buftype_is_contig, contig_access_count,
			min_st_offset, fd_size, fd_start, fd_end,
			others_req, m,
			buftype_extent, buf_idx);
		MPIU_ERR_CHK_POP(mpi_errno);
	}
#ifdef PROFILE
	MPE_Log_event(8, 0, L"end communication");
#endif

	if (ntimes)
	{
		MPIU_Free(read_buf);
	}
	MPIU_Free(curr_offlen_ptr);
	MPIU_Free(count);
	MPIU_Free(partial_send);
	MPIU_Free(send_size);
	MPIU_Free(recv_size);
	MPIU_Free(recd_from_proc);
	MPIU_Free(start_pos);

fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_READ_AND_EXCH);
	return mpi_errno;
fn_fail:
	/* Free pointers that are not NULL? */
	goto fn_exit;
}

#undef FUNCNAME
#define FUNCNAME ADIOI_R_Exchange_data
static int ADIOI_R_Exchange_data(MPID_File_ptr fd, void *buf, ADIOI_Flatlist_node
								 *flat_buf, MPI_Offset *offset_list, int
								 *len_list, int *send_size, int *recv_size,
								 int *count, int *start_pos, int *partial_send, 
								 int *recd_from_proc, int nprocs, 
								 int myrank, int
								 buftype_is_contig, int contig_access_count,
								 MPI_Offset min_st_offset, MPI_Offset fd_size,
								 MPI_Offset *fd_start, MPI_Offset *fd_end, 
								 ADIOI_Access *others_req, 
								 int iter, MPI_Aint buftype_extent, int *buf_idx)
{
	static const wchar_t FCNAME[] = L"ADIOI_R_Exchange_data";
	int mpi_errno = MPI_SUCCESS;
	int i, j, k=0, tmp=0, nprocs_recv, nprocs_send;
	char **recv_buf = NULL; 
	MPI_Request *requests;
	MPI_Datatype send_type;
	MPI_Status *statuses;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_R_EXCHANGE_DATA);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_R_EXCHANGE_DATA);

	/* exchange send_size info so that each process knows how much to
	receive from whom and how much memory to allocate. */

	NMPI_Alltoall(mpi_errno, (send_size, 1, MPI_INT, recv_size, 1, MPI_INT, fd->comm));
	MPIU_ERR_CHK_POP(mpi_errno);

	nprocs_recv = 0;
	for (i=0; i < nprocs; i++)
	{
		if (recv_size[i]) nprocs_recv++;
	}

	nprocs_send = 0;
	for (i=0; i<nprocs; i++)
	{
		if (send_size[i]) nprocs_send++;
	}

	requests = (MPI_Request *) MPIU_Malloc((nprocs_send+nprocs_recv+1)*sizeof(MPI_Request));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, requests, (nprocs_send+nprocs_recv+1)*sizeof(MPI_Request), L"requests");
	/* +1 to avoid a 0-size malloc */

	/* post recvs. if buftype_is_contig, data can be directly recd. into
	user buf at location given by buf_idx. else use recv_buf. */

	if (buftype_is_contig)
	{
		j = 0;
		for (i=0; i < nprocs; i++)
		{
			if (recv_size[i])
			{
				NMPI_Irecv(mpi_errno, (((char *) buf) + buf_idx[i], recv_size[i], 
					MPI_BYTE, i, myrank+i+100*iter, fd->comm, requests+j));
				MPIU_ERR_CHK_POP(mpi_errno);
				j++;
				buf_idx[i] += recv_size[i];
			}
		}
	}
	else
	{
		/* allocate memory for recv_buf and post receives */
		recv_buf = (char **) MPIU_Malloc(nprocs * sizeof(char*));
		MPIU_ERR_MEM_CHK_POP(mpi_errno, recv_buf, nprocs * sizeof(char*), L"recv_buf");
		for (i=0; i < nprocs; i++)
		{
			if (recv_size[i])
			{
				recv_buf[i] = (char *) MPIU_Malloc(recv_size[i]);
				MPIU_ERR_MEM_CHK_POP(mpi_errno, recv_buf[i], recv_size[i], L"recv_buf");
			}
		}

		j = 0;
		for (i=0; i < nprocs; i++)
		{
			if (recv_size[i])
			{
				NMPI_Irecv(mpi_errno, (recv_buf[i], recv_size[i], MPI_BYTE, i, 
					myrank+i+100*iter, fd->comm, requests+j));
				MPIU_ERR_CHK_POP(mpi_errno);
				j++;
				/* FPRINTF(stderr, L"node %d, recv_size %d, tag %d \n", 
				myrank, recv_size[i], myrank+i+100*iter); */
			}
		}
	}

	/* create derived datatypes and send data */

	j = 0;
	for (i=0; i<nprocs; i++)
	{
		if (send_size[i])
		{
			/* take care if the last off-len pair is a partial send */
			if (partial_send[i])
			{
				k = start_pos[i] + count[i] - 1;
				tmp = others_req[i].lens[k];
				others_req[i].lens[k] = partial_send[i];
			}
			NMPI_Type_hindexed(mpi_errno, (count[i], &(others_req[i].lens[start_pos[i]]), &(others_req[i].mem_ptrs[start_pos[i]]), MPI_BYTE, &send_type));
			MPIU_ERR_CHK_POP(mpi_errno);
			/* absolute displacement; use MPI_BOTTOM in send */
			NMPI_Type_commit(mpi_errno, (&send_type));
			MPIU_ERR_CHK_POP(mpi_errno);
			NMPI_Isend(mpi_errno, (MPI_BOTTOM, 1, send_type, i, myrank+i+100*iter, fd->comm, requests+nprocs_recv+j));
			MPIU_ERR_CHK_POP(mpi_errno);
			NMPI_Type_free(mpi_errno, (&send_type));
			MPIU_ERR_CHK_POP(mpi_errno);
			if (partial_send[i])
			{
				others_req[i].lens[k] = tmp;
			}
			j++;
		}
	}

	statuses = (MPI_Status *) MPIU_Malloc((nprocs_send+nprocs_recv+1) * sizeof(MPI_Status));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, statuses, (nprocs_send+nprocs_recv+1) * sizeof(MPI_Status), L"statuses");
	/* +1 to avoid a 0-size malloc */

	/* wait on the receives */
	if (nprocs_recv)
	{
#ifdef NEEDS_MPI_TEST
		j = 0;
		while (!j)
		{
			mpi_errno = MPI_Testall(nprocs_recv, requests, &j, statuses);
			MPIU_ERR_CHK_POP(mpi_errno);
		}
#else
		NMPI_Waitall(mpi_errno, (nprocs_recv, requests, statuses));
		MPIU_ERR_CHK_POP(mpi_errno);
#endif

		/* if noncontiguous, to the copies from the recv buffers */
		if (!buftype_is_contig)
		{
			mpi_errno = ADIOI_Fill_user_buffer(fd, buf, flat_buf, recv_buf,
				offset_list, len_list, recv_size, 
				requests, statuses, recd_from_proc, 
				nprocs, contig_access_count,
				min_st_offset, fd_size, fd_start, fd_end,
				buftype_extent);
			MPIU_ERR_CHK_POP(mpi_errno);
		}
	}

	/* wait on the sends*/
	NMPI_Waitall(mpi_errno, (nprocs_send, requests+nprocs_recv, statuses+nprocs_recv));
	MPIU_ERR_CHK_POP(mpi_errno);

	MPIU_Free(statuses);
	MPIU_Free(requests);
	statuses = NULL;
	requests = NULL;

	if (!buftype_is_contig)
	{
		for (i=0; i < nprocs; i++) 
			if (recv_size[i]) MPIU_Free(recv_buf[i]);
		MPIU_Free(recv_buf);
		recv_buf = NULL;
	}

fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_R_EXCHANGE_DATA);
	return mpi_errno;

fn_fail:
	/* free buffers that are not NULL? */
	goto fn_exit;
}

#define ADIOI_BUF_COPY_RECV \
{ \
	while (size) { \
	size_in_buf = MPIDU_MIN(size, flat_buf_sz); \
	memcpy(((char *) buf) + user_buf_idx, \
	&(recv_buf[p][recv_buf_idx[p]]), size_in_buf); \
	recv_buf_idx[p] += size_in_buf; \
	user_buf_idx += size_in_buf; \
	flat_buf_sz -= size_in_buf; \
	if (!flat_buf_sz) { \
	if (flat_buf_idx < (flat_buf->count - 1)) flat_buf_idx++; \
			else { \
			flat_buf_idx = 0; \
			n_buftypes++; \
} \
	user_buf_idx = flat_buf->indices[flat_buf_idx] + \
	n_buftypes*buftype_extent; \
	flat_buf_sz = flat_buf->blocklens[flat_buf_idx]; \
	} \
	size -= size_in_buf; \
	buf_incr -= size_in_buf; \
	} \
	ADIOI_BUF_INCR \
}


#undef FUNCNAME
#define FUNCNAME ADIOI_Fill_user_buffer
static int ADIOI_Fill_user_buffer(MPID_File_ptr fd, void *buf, ADIOI_Flatlist_node
								  *flat_buf, char **recv_buf, MPI_Offset 
								  *offset_list, int *len_list, 
								  int *recv_size, 
								  MPI_Request *requests, MPI_Status *statuses,
								  int *recd_from_proc, int nprocs,
								  int contig_access_count, 
								  MPI_Offset min_st_offset, 
								  MPI_Offset fd_size, MPI_Offset *fd_start, 
								  MPI_Offset *fd_end,
								  MPI_Aint buftype_extent)
{
	/* this function is only called if buftype is not contig */
	static const wchar_t FCNAME[] = L"ADIOI_Fill_user_buffer";
	int mpi_errno = MPI_SUCCESS;
	int i, p, flat_buf_idx, size, buf_incr;
	int flat_buf_sz, size_in_buf, n_buftypes;
	MPI_Offset off, len, rem_len, user_buf_idx;
	int *curr_from_proc, *done_from_proc, *recv_buf_idx;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_FILL_USER_BUFFER);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_FILL_USER_BUFFER);

	MPIU_UNREFERENCED_ARG(requests);
	MPIU_UNREFERENCED_ARG(statuses);

	/*  curr_from_proc[p] = amount of data recd from proc. p that has already
	been accounted for so far
	done_from_proc[p] = amount of data already recd from proc. p and 
	filled into user buffer in previous iterations
	user_buf_idx = current location in user buffer 
	recv_buf_idx[p] = current location in recv_buf of proc. p  */
	curr_from_proc = (int *) MPIU_Malloc(nprocs * sizeof(int));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, curr_from_proc, nprocs * sizeof(int), L"curr_from_proc");
	done_from_proc = (int *) MPIU_Malloc(nprocs * sizeof(int));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, done_from_proc, nprocs * sizeof(int), L"done_from_proc");
	recv_buf_idx   = (int *) MPIU_Malloc(nprocs * sizeof(int));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, recv_buf_idx, nprocs * sizeof(int), L"recv_buf_idx");

	for (i=0; i < nprocs; i++)
	{
		recv_buf_idx[i] = curr_from_proc[i] = 0;
		done_from_proc[i] = recd_from_proc[i];
	}

	user_buf_idx = flat_buf->indices[0];
	flat_buf_idx = 0;
	n_buftypes = 0;
	flat_buf_sz = flat_buf->blocklens[0];

	/* flat_buf_idx = current index into flattened buftype
	flat_buf_sz = size of current contiguous component in 
	flattened buf */

	for (i=0; i<contig_access_count; i++)
	{
		off	 = offset_list[i];
		rem_len = (MPI_Offset) len_list[i];

		/* this request may span the file domains of more than one process */
		while (rem_len != 0)
		{
			len = rem_len;
			/* NOTE: len value is modified by ADIOI_Calc_aggregator() to be no
			* longer than the single region that processor "p" is responsible
			* for.
			*/
			mpi_errno = ADIOI_Calc_aggregator(fd, off, min_st_offset, &len, fd_size, fd_start, fd_end, &p);
			MPIU_ERR_CHK_POP(mpi_errno);

			if (recv_buf_idx[p] < recv_size[p])
			{
				if (curr_from_proc[p]+len > done_from_proc[p])
				{
					if (done_from_proc[p] > curr_from_proc[p])
					{
						size = (int)MPIDU_MIN(curr_from_proc[p] + len - done_from_proc[p], recv_size[p]-recv_buf_idx[p]);
						buf_incr = done_from_proc[p] - curr_from_proc[p];
						ADIOI_BUF_INCR
						buf_incr = (int)(curr_from_proc[p]+len-done_from_proc[p]);
						curr_from_proc[p] = done_from_proc[p] + size;
						ADIOI_BUF_COPY_RECV
					}
					else
					{
						size = (int)MPIDU_MIN(len,recv_size[p]-recv_buf_idx[p]);
						buf_incr = (int)len;
						curr_from_proc[p] += size;
						ADIOI_BUF_COPY_RECV
					}
				}
				else
				{
					curr_from_proc[p] += (int)len;
					buf_incr = (int)len;
					ADIOI_BUF_INCR
				}
			}
			else
			{
				buf_incr = (int)len;
				ADIOI_BUF_INCR
			}
			off	 += len;
			rem_len -= len;
		}
	}
	for (i=0; i < nprocs; i++)
	{
		if (recv_size[i]) recd_from_proc[i] = curr_from_proc[i];
	}

	MPIU_Free(curr_from_proc);
	MPIU_Free(done_from_proc);
	MPIU_Free(recv_buf_idx);

fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_FILL_USER_BUFFER);
	return mpi_errno;
}

/* this used to be implemented in every file system as an fcntl, but the code
* is identical for all file systems without a real "preallocate" system call.
* This naive approach will get the job done, but not in a terribly efficient
* manner.
*/
#undef FUNCNAME
#define FUNCNAME ADIOI_GEN_Prealloc
int ADIOI_GEN_Prealloc(MPID_File_ptr fd, MPI_Offset diskspace)
{
	static const wchar_t FCNAME[] = L"ADIOI_GEN_Prealloc";
	int mpi_errno = MPI_SUCCESS;
	MPI_Offset curr_fsize, alloc_size, size, done;
	int len;
	MPI_Status status;
	int i, ntimes;
	char *buf = NULL;
	ADIO_Fcntl_t *fcntl_struct = NULL;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_GEN_PREALLOC);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_GEN_PREALLOC);

	/* will be called by one process only */
	/* On file systems with no preallocation function, we have to 
	explicitly write 
	to allocate space. Since there could be holes in the file, 
	we need to read up to the current file size, write it back, 
	and then write beyond that depending on how much 
	preallocation is needed.
	read/write in sizes of no more than ADIOI_PREALLOC_BUFSZ */

	/*curr_fsize = fd->fp_ind; */
	fcntl_struct = (ADIO_Fcntl_t *) MPIU_Malloc(sizeof(ADIO_Fcntl_t));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, fcntl_struct, sizeof(ADIO_Fcntl_t), L"fcntl_struct");

	mpi_errno = ADIO_Fcntl(fd, ADIO_FCNTL_GET_FSIZE, fcntl_struct);
	MPIU_ERR_CHK_POP(mpi_errno);

	curr_fsize = fcntl_struct->fsize; /* don't rely on fd->fp_ind: might be
									  working on a pre-existing file */
	alloc_size = diskspace;

	size = MPIDU_MIN(curr_fsize, alloc_size);

	ntimes = (int)((size + ADIOI_PREALLOC_BUFSZ - 1)/ADIOI_PREALLOC_BUFSZ);
	buf = (char *) MPIU_Malloc(ADIOI_PREALLOC_BUFSZ);
	MPIU_ERR_MEM_CHK_POP(mpi_errno, buf, ADIOI_PREALLOC_BUFSZ, L"buf");
	done = 0;

	for (i=0; i<ntimes; i++)
	{
		len = (int)MPIDU_MIN(size-done, ADIOI_PREALLOC_BUFSZ);
		mpi_errno = ADIO_ReadContig(fd, buf, len, MPI_BYTE, ADIO_EXPLICIT_OFFSET, done, &status);
		if (mpi_errno != MPI_SUCCESS)
		{
			MPIU_ERR_SET(mpi_errno, MPI_ERR_IO, L"**iopreallocrdwr");
			goto fn_fail;
		}
		mpi_errno = ADIO_WriteContig(fd, buf, len, MPI_BYTE, ADIO_EXPLICIT_OFFSET, done, &status);
		MPIU_ERR_CHK_POP(mpi_errno);
		done += len;
	}

	if (alloc_size > curr_fsize)
	{
		memset(buf, 0, ADIOI_PREALLOC_BUFSZ); 
		size = alloc_size - curr_fsize;
		ntimes = (int)((size + ADIOI_PREALLOC_BUFSZ - 1)/ADIOI_PREALLOC_BUFSZ);
		for (i=0; i<ntimes; i++)
		{
			len = (int)MPIDU_MIN(alloc_size-done, ADIOI_PREALLOC_BUFSZ);
			mpi_errno = ADIO_WriteContig(fd, buf, len, MPI_BYTE, ADIO_EXPLICIT_OFFSET, done, &status);
			MPIU_ERR_CHK_POP(mpi_errno);
			done += len;  
		}
	}
	MPIU_Free(fcntl_struct);
	MPIU_Free(buf);

fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_GEN_PREALLOC);
	return mpi_errno;

fn_fail:
	if (fcntl_struct != NULL)
	{
		MPIU_Free(fcntl_struct);
	}
	if (buf != NULL)
	{
		MPIU_Free(buf);
	}
	goto fn_exit;
}

static int is_aggregator(int rank, MPID_File_ptr fd);
static int uses_generic_read(MPID_File_ptr fd);
static int uses_generic_write(MPID_File_ptr fd);

#undef FUNCNAME
#define FUNCNAME ADIO_Open
int ADIO_Open(MPI_Comm orig_comm,
			  MPI_Comm comm, wchar_t *filename, int file_system,
			  ADIOI_Fns *ops,
			  int access_mode, MPI_Offset disp, MPI_Datatype etype, 
			  MPI_Datatype filetype,
			  MPI_Info info, int perm, MPID_File_ptr *mpi_fh_ptr)
{
	static const wchar_t FCNAME[] = L"ADIO_Open";
	int mpi_errno = MPI_SUCCESS, tmp_errno;
	MPID_File_ptr fd = NULL;
	ADIO_cb_name_array array;
	int orig_amode_excl, orig_amode_wronly, err, rank, procs;
	wchar_t *value = NULL;
	int rank_ct, max_error_code;
	int *tmp_ranklist;
	MPI_Comm aggregator_comm = MPI_COMM_NULL; /* just for deferred opens */
	MPID_Errhandler *errhan_ptr = NULL;
	MPIDI_STATE_DECL(MPID_STATE_ADIO_OPEN);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIO_OPEN);

	/* obtain MPI_File handle */
	fd = (MPID_File *)MPIU_Handle_obj_alloc( &MPID_File_mem );
	MPIU_ERR_CHKANDJUMP(!fd, mpi_errno, MPI_ERR_OTHER, L"**nomem");

	fd->cookie = ADIOI_FILE_COOKIE;
	fd->fp_ind = disp;
	fd->fp_sys_posn = 0;
	fd->comm = comm;	   /* dup'ed in MPI_File_open */
	fd->filename = MPIU_Strdup(filename);
	fd->file_system = file_system;
	fd->fs_ptr = NULL;

	fd->fns = ops;

	fd->disp = disp;
	fd->split_coll_count = 0;
	fd->shared_fp_fd = NULL;
	fd->atomicity = 0;
	fd->etype = etype;		  /* MPI_BYTE by default */
	fd->filetype = filetype;	/* MPI_BYTE by default */
	fd->etype_size = 1;  /* default etype is MPI_BYTE */

	fd->perm = perm;

	fd->async_count = 0;

	MPID_Errhandler_get_ptr( MPI_ERRORS_RETURN, errhan_ptr );
	fd->errhandler = errhan_ptr;

	/* create and initialize info object */
	fd->hints = (ADIOI_Hints *)MPIU_Malloc(sizeof(struct ADIOI_Hints));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, fd->hints, sizeof(struct ADIOI_Hints), L"hints");

	fd->hints->cb_config_list = NULL;
	fd->hints->ranklist = NULL;
	fd->hints->initialized = 0;
	fd->info = MPI_INFO_NULL;
	mpi_errno = ADIO_SetInfo(fd, info);
	MPIU_ERR_CHK_POP(mpi_errno);

	/* deferred open: 
	* we can only do this optimization if 'fd->hints->deferred_open' is set
	* (which means the user hinted 'no_indep_rw' and collective buffering).
	* Furthermore, we only do this if our collective read/write routines use
	* our generic function, and not an fs-specific routine (we can defer opens
	* only if we use our aggreagation code). */
	if (fd->hints->deferred_open && !(uses_generic_read(fd) && uses_generic_write(fd)))
	{
		fd->hints->deferred_open = 0;
	}
	if (fd->file_system == ADIO_PVFS2)
	{
		/* disable deferred open on PVFS2 so that scalable broadcast will
		* always use the propper communicator */
		fd->hints->deferred_open = 0;
	}

	/* gather the processor name array if we don't already have it */

	/* this has to be done here so that we can cache the name array in both
	* the dup'd communicator (in case we want it later) and the original
	* communicator
	*/
	mpi_errno = ADIOI_cb_gather_name_array(orig_comm, comm, &array);
	MPIU_ERR_CHK_POP(mpi_errno);

	/* parse the cb_config_list and create a rank map on rank 0 */
	NMPI_Comm_rank(mpi_errno, (comm, &rank));
	MPIU_ERR_CHK_POP(mpi_errno);
	if (rank == 0)
	{
		NMPI_Comm_size(mpi_errno, (comm, &procs));
		MPIU_ERR_CHK_POP(mpi_errno);
		tmp_ranklist = (int *) MPIU_Malloc(sizeof(int) * procs);
		MPIU_ERR_MEM_CHK_POP(mpi_errno, tmp_ranklist, sizeof(int) * procs, L"tmp_ranklist");

		mpi_errno = ADIOI_cb_config_list_parse(fd->hints->cb_config_list, array, tmp_ranklist, fd->hints->cb_nodes, &rank_ct);
		MPIU_ERR_CHK_POP(mpi_errno);

		/* store the ranklist using the minimum amount of memory */
		if (rank_ct > 0)
		{
			fd->hints->ranklist = (int *) MPIU_Malloc(sizeof(int) * rank_ct);
			MPIU_ERR_MEM_CHK_POP(mpi_errno, fd->hints->ranklist, sizeof(int) * rank_ct, L"ranklist");
			memcpy(fd->hints->ranklist, tmp_ranklist, sizeof(int) * rank_ct);
		}
		MPIU_Free(tmp_ranklist);
		fd->hints->cb_nodes = rank_ct;
		/* TEMPORARY -- REMOVE WHEN NO LONGER UPDATING INFO FOR FS-INDEP. */
		value = (wchar_t *) MPIU_Malloc((MPI_MAX_INFO_VAL+1)*sizeof(wchar_t));
		MPIU_ERR_MEM_CHK_POP(mpi_errno, value, (MPI_MAX_INFO_VAL+1) * sizeof(wchar_t), L"value");
		MPIU_Snprintf(value, MPI_MAX_INFO_VAL+1, _TRUNCATE, L"%d", rank_ct);
		NMPI_Info_set(mpi_errno, (fd->info, L"cb_nodes", value));
		MPIU_ERR_CHK_POP(mpi_errno);
		MPIU_Free(value);
		value = NULL;
	}

	mpi_errno = ADIOI_cb_bcast_rank_map(fd);
	MPIU_ERR_CHK_POP(mpi_errno);

	if (fd->hints->cb_nodes <= 0)
	{
		MPIU_ERR_SET(mpi_errno, MPI_ERR_IO, L"**ioagnomatch");
		fd = NULL;
		goto fn_fail;
	}

	/* deferred open: if we are an aggregator, create a new communicator.
	* we'll use this aggregator communicator for opens and closes.
	* otherwise, we have a NULL communicator until we try to do independent
	* IO */
	fd->agg_comm = MPI_COMM_NULL;
	fd->is_open = 0;
	if (fd->hints->deferred_open)
	{
		/* MPI_Comm_split will create a communication group of aggregators.
		* for non-aggregators it will return MPI_COMM_NULL .  we rely on
		* fd->agg_comm == MPI_COMM_NULL for non-aggregators in several
		* tests in the code  */
		if (is_aggregator(rank, fd))
		{
			NMPI_Comm_split(mpi_errno, (fd->comm, 1, 0, &aggregator_comm));
			MPIU_ERR_CHK_POP(mpi_errno);
			fd->agg_comm = aggregator_comm;
		}
		else
		{
			NMPI_Comm_split(mpi_errno, (fd->comm, MPI_UNDEFINED, 0, &aggregator_comm));
			MPIU_ERR_CHK_POP(mpi_errno);
			fd->agg_comm = aggregator_comm;
		}
	}

	orig_amode_excl = access_mode;

	/* optimization: by having just one process create a file, close it, then
	* have all N processes open it, we can possibly avoid contention for write
	* locks on a directory for some file systems. 
	*
	* we used to special-case EXCL|CREATE, since when N processes are trying
	* to create a file exclusively, only 1 will succeed and the rest will
	* (spuriously) fail.   Since we are now carrying out the CREATE on one
	* process anyway, the EXCL case falls out and we don't need to explicitly
	* worry about it, other than turning off both the EXCL and CREATE flags 
	*/
	/* pvfs2 handles opens specially, so it is actually more efficent for that
	* file system if we skip this optimization */
	if (access_mode & ADIO_CREATE && fd->file_system != ADIO_PVFS2)
	{
		if (rank == fd->hints->ranklist[0])
		{
			/* remove delete_on_close flag if set */
			if (access_mode & ADIO_DELETE_ON_CLOSE)
				fd->access_mode = access_mode ^ ADIO_DELETE_ON_CLOSE;
			else
				fd->access_mode = access_mode;
			mpi_errno = (*(fd->fns->ADIOI_xxx_Open))(fd);
			MPIU_ERR_CHK_POP(mpi_errno);
			err = mpi_errno;
			NMPI_Bcast(mpi_errno, (&err, 1, MPI_INT, fd->hints->ranklist[0], fd->comm));
			MPIU_ERR_CHK_POP(mpi_errno);
			MPIU_ERR_CHK_POP(err);
			/* if no error, close the file and reopen normally below */
			mpi_errno = (*(fd->fns->ADIOI_xxx_Close))(fd);
			MPIU_ERR_CHK_POP(mpi_errno);
			fd->access_mode = access_mode; /* back to original */
		}
		else
		{
			err = mpi_errno;
			NMPI_Bcast(mpi_errno, (&err, 1, MPI_INT, fd->hints->ranklist[0], fd->comm));
			MPIU_ERR_CHK_POP(mpi_errno);
			MPIU_ERR_CHK_POP(err);
		}

		if (mpi_errno != MPI_SUCCESS)
		{
			goto fn_exit;
		}
		else
		{
			/* turn off CREAT (and EXCL if set) for real multi-processor open */
			access_mode ^= ADIO_CREATE;
			if (access_mode & ADIO_EXCL)
				access_mode ^= ADIO_EXCL;
		}
	}

	/* if we are doing deferred open, non-aggregators should return now */
	if (fd->hints->deferred_open )
	{
		if (fd->agg_comm == MPI_COMM_NULL)
		{
			/* we might have turned off EXCL for the aggregators.
			* restore access_mode that non-aggregators get the right
			* value from get_amode */
			fd->access_mode = orig_amode_excl;
			goto fn_exit;
		}
	}

	/* For writing with data sieving, a read-modify-write is needed. If 
	the file is opened for write_only, the read will fail. Therefore,
	if write_only, open the file as read_write, but record it as write_only
	in fd, so that get_amode returns the right answer. */

	orig_amode_wronly = access_mode;
	if (access_mode & ADIO_WRONLY)
	{
		access_mode = access_mode ^ ADIO_WRONLY;
		access_mode = access_mode | ADIO_RDWR;
	}
	fd->access_mode = access_mode;

	mpi_errno = (*(fd->fns->ADIOI_xxx_Open))(fd);
	/*MPIU_ERR_CHK_POP(mpi_errno);*/
	/* if error, may be it was due to the change in amode above. 
	therefore, reopen with access mode provided by the user.*/ 
	fd->access_mode = orig_amode_wronly;  
	if (mpi_errno != MPI_SUCCESS)
	{
		mpi_errno = (*(fd->fns->ADIOI_xxx_Open))(fd);
		/*MPIU_ERR_CHK_POP(mpi_errno);*/ /* Let this error get sent to the other processes */
	}

	/* if we turned off EXCL earlier, then we should turn it back on */
	if (fd->access_mode != orig_amode_excl) fd->access_mode = orig_amode_excl;

	/* for deferred open: this process has opened the file (because if we are
	* not an aggregaor and we are doing deferred open, we returned earlier)*/
	fd->is_open = 1;

fn_exit:
	NMPI_Allreduce(tmp_errno, (&mpi_errno, &max_error_code, 1, MPI_INT, MPI_MAX, comm));
	MPIU_ERR_CHK_POP(tmp_errno);
	if (max_error_code != MPI_SUCCESS)
	{
		/* If the file was successfully opened, close it */
		if (mpi_errno == MPI_SUCCESS)
		{
			/* in the deferred open case, only those who have actually
			opened the file should close it */
			if (fd->hints->deferred_open)
			{
				if (fd->agg_comm != MPI_COMM_NULL)
				{
					mpi_errno = (*(fd->fns->ADIOI_xxx_Close))(fd);
				}
			}
			else
			{
				mpi_errno = (*(fd->fns->ADIOI_xxx_Close))(fd);
			}
		}

		if (fd->filename)
		{
			MPIU_Free(fd->filename);
			fd->filename = NULL;
		}
		if (fd->hints->ranklist)
		{
			MPIU_Free(fd->hints->ranklist);
			fd->hints->ranklist = NULL;
		}
		if (fd->hints->cb_config_list)
		{
			MPIU_Free(fd->hints->cb_config_list);
			fd->hints->cb_config_list = NULL;
		}
		if (fd->hints)
		{
			MPIU_Free(fd->hints);
			fd->hints = NULL;
		}
		if (fd->info != MPI_INFO_NULL)
		{
			NMPI_Info_free(mpi_errno, (&(fd->info)));
		}
		MPIU_Handle_obj_free(&MPID_File_mem, fd);
		fd = NULL;
		if (mpi_errno == MPI_SUCCESS)
		{
			MPIU_ERR_SET(mpi_errno, MPI_ERR_IO, L"**oremote_fail");
		}
	}

fn_exit2:
	if (mpi_fh_ptr != NULL)
	{
		*mpi_fh_ptr = fd;
	}
	MPIDI_FUNC_EXIT(MPID_STATE_ADIO_OPEN);
	return mpi_errno;

fn_fail:
	if (value != NULL)
	{
		MPIU_Free(value);
	}
	goto fn_exit2;
}

/* a simple linear search. possible enancement: add a my_cb_nodes_index member
* ( index into cb_nodes, else -1 if not aggregator ) for faster lookups 
*
* fd->hints->cb_nodes is the number of aggregators
* fd->hints->ranklist[] is an array of the ranks of aggregators
*
* might want to move this to adio/common/cb_config_list.c 
*/
#undef FUNCNAME
#define FUNCNAME is_aggregator
int is_aggregator(int rank, MPID_File_ptr fd )
{
	int i;
	MPIDI_STATE_DECL(MPID_STATE_IS_AGGREGATOR);

	MPIDI_FUNC_ENTER(MPID_STATE_IS_AGGREGATOR);

	for (i=0; i< fd->hints->cb_nodes; i++ )
	{
		if ( rank == fd->hints->ranklist[i] )
		{
			MPIDI_FUNC_EXIT(MPID_STATE_IS_AGGREGATOR);
			return 1;
		}
	}
	MPIDI_FUNC_EXIT(MPID_STATE_IS_AGGREGATOR);
	return 0;
}

/* 
* we special-case TESTFS because all it does is wrap logging info around GEN 
*/
#undef FUNCNAME
#define FUNCNAME uses_generic_read
static int uses_generic_read(MPID_File_ptr fd)
{
	ADIOI_Fns *fns = fd->fns;
	MPIDI_STATE_DECL(MPID_STATE_USES_GENERIC_READ);

	MPIDI_FUNC_ENTER(MPID_STATE_USES_GENERIC_READ);
	if (fns->ADIOI_xxx_ReadStridedColl == ADIOI_GEN_ReadStridedColl || 
		fd->file_system == ADIO_TESTFS )
	{
		MPIDI_FUNC_EXIT(MPID_STATE_USES_GENERIC_READ);
		return 1;
	}
	MPIDI_FUNC_EXIT(MPID_STATE_USES_GENERIC_READ);
	return 0;
}

#undef FUNCNAME
#define FUNCNAME uses_generic_write
static int uses_generic_write(MPID_File_ptr fd)
{
	ADIOI_Fns *fns = fd->fns;
	MPIDI_STATE_DECL(MPID_STATE_USES_GENERIC_WRITE);

	MPIDI_FUNC_ENTER(MPID_STATE_USES_GENERIC_WRITE);
	if (fns->ADIOI_xxx_WriteStridedColl == ADIOI_GEN_WriteStridedColl ||
		fd->file_system == ADIO_TESTFS )
	{
		MPIDI_FUNC_EXIT(MPID_STATE_USES_GENERIC_WRITE);
		return 1;
	}
	MPIDI_FUNC_EXIT(MPID_STATE_USES_GENERIC_WRITE);
	return 0;
}

#undef FUNCNAME
#define FUNCNAME ADIOI_GEN_Delete
int ADIOI_GEN_Delete(wchar_t *filename)
{
	int mpi_errno = MPI_SUCCESS;
	int err;
	static const wchar_t FCNAME[] = L"ADIOI_GEN_Delete";
	wchar_t error_msg[1024];
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_GEN_DELETE);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_GEN_DELETE);

	err = _wunlink(filename);
	if (err == -1)
	{
		_wcserror_s(error_msg, 1024, errno);
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE,
			FCNAME, __LINE__, MPI_ERR_IO, L"**io",
			L"**io %s", error_msg);
	}
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_GEN_DELETE);
	return mpi_errno;
}

int MPIOI_Type_block(int *array_of_gsizes, int dim, int ndims, int nprocs,
					 int rank, int darg, int order, MPI_Aint orig_extent,
					 MPI_Datatype type_old, MPI_Datatype *type_new,
					 MPI_Aint *st_offset);
int MPIOI_Type_cyclic(int *array_of_gsizes, int dim, int ndims, int nprocs,
					  int rank, int darg, int order, MPI_Aint orig_extent,
					  MPI_Datatype type_old, MPI_Datatype *type_new,
					  MPI_Aint *st_offset);


#undef FUNCNAME
#define FUNCNAME ADIO_Type_create_darray
int ADIO_Type_create_darray(int size, int rank, int ndims, 
							int *array_of_gsizes, int *array_of_distribs, 
							int *array_of_dargs, int *array_of_psizes, 
							int order, MPI_Datatype oldtype, 
							MPI_Datatype *newtype) 
{
	static const wchar_t FCNAME[] = L"ADIO_Type_create_darray";
	int mpi_errno = MPI_SUCCESS;
	MPI_Datatype type_old, type_new=MPI_DATATYPE_NULL, types[3];
	int procs, tmp_rank, i, tmp_size, blklens[3], *coords;
	MPI_Aint *st_offsets, orig_extent, disps[3];
	MPIDI_STATE_DECL(MPID_STATE_ADIO_TYPE_CREATE_DARRAY);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIO_TYPE_CREATE_DARRAY);

	NMPI_Type_extent(mpi_errno, (oldtype, &orig_extent));
	MPIU_ERR_CHK_POP(mpi_errno);

	/* calculate position in Cartesian grid as MPI would (row-major
	ordering) */
	coords = (int *) MPIU_Malloc(ndims*sizeof(int));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, coords, ndims*sizeof(int), L"coords");
	procs = size;
	tmp_rank = rank;
	for (i=0; i<ndims; i++)
	{
		procs = procs/array_of_psizes[i];
		coords[i] = tmp_rank/procs;
		tmp_rank = tmp_rank % procs;
	}

	st_offsets = (MPI_Aint *) MPIU_Malloc(ndims*sizeof(MPI_Aint));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, st_offsets, ndims*sizeof(MPI_Aint), L"st_offsets");
	type_old = oldtype;

	if (order == MPI_ORDER_FORTRAN)
	{
		/* dimension 0 changes fastest */
		for (i=0; i<ndims; i++)
		{
			switch(array_of_distribs[i])
			{
			case MPI_DISTRIBUTE_BLOCK:
				mpi_errno = MPIOI_Type_block(array_of_gsizes, i, ndims,
					array_of_psizes[i],
					coords[i], array_of_dargs[i],
					order, orig_extent, 
					type_old, &type_new,
					st_offsets+i);
				MPIU_ERR_CHK_POP(mpi_errno);
				break;
			case MPI_DISTRIBUTE_CYCLIC:
				mpi_errno = MPIOI_Type_cyclic(array_of_gsizes, i, ndims, 
					array_of_psizes[i], coords[i],
					array_of_dargs[i], order,
					orig_extent, type_old,
					&type_new, st_offsets+i);
				MPIU_ERR_CHK_POP(mpi_errno);
				break;
			case MPI_DISTRIBUTE_NONE:
				/* treat it as a block distribution on 1 process */
				mpi_errno = MPIOI_Type_block(array_of_gsizes, i, ndims, 1, 0, 
					MPI_DISTRIBUTE_DFLT_DARG, order,
					orig_extent, 
					type_old, &type_new,
					st_offsets+i);
				MPIU_ERR_CHK_POP(mpi_errno);
				break;
			}
			if (i)
			{
				NMPI_Type_free(mpi_errno, (&type_old));
				MPIU_ERR_CHK_POP(mpi_errno);
			}
			type_old = type_new;
		}

		/* add displacement and UB */
		disps[1] = st_offsets[0];
		tmp_size = 1;
		for (i=1; i<ndims; i++)
		{
			tmp_size *= array_of_gsizes[i-1];
			disps[1] += tmp_size*st_offsets[i];
		}
		/* rest done below for both Fortran and C order */
	}
	else
	{
		/* order == MPI_ORDER_C */
		/* dimension ndims-1 changes fastest */
		for (i=ndims-1; i>=0; i--)
		{
			switch(array_of_distribs[i])
			{
			case MPI_DISTRIBUTE_BLOCK:
				mpi_errno = MPIOI_Type_block(array_of_gsizes, i, ndims, array_of_psizes[i],
					coords[i], array_of_dargs[i], order,
					orig_extent, type_old, &type_new,
					st_offsets+i);
				MPIU_ERR_CHK_POP(mpi_errno);
				break;
			case MPI_DISTRIBUTE_CYCLIC:
				mpi_errno = MPIOI_Type_cyclic(array_of_gsizes, i, ndims, 
					array_of_psizes[i], coords[i],
					array_of_dargs[i], order, 
					orig_extent, type_old, &type_new,
					st_offsets+i);
				MPIU_ERR_CHK_POP(mpi_errno);
				break;
			case MPI_DISTRIBUTE_NONE:
				/* treat it as a block distribution on 1 process */
				mpi_errno = MPIOI_Type_block(array_of_gsizes, i, ndims, array_of_psizes[i],
					coords[i], MPI_DISTRIBUTE_DFLT_DARG, order, orig_extent, 
					type_old, &type_new, st_offsets+i); 
				MPIU_ERR_CHK_POP(mpi_errno);
				break;
			}
			if (i != ndims-1)
			{
				NMPI_Type_free(mpi_errno, (&type_old));
				MPIU_ERR_CHK_POP(mpi_errno);
			}
			type_old = type_new;
		}

		/* add displacement and UB */
		disps[1] = st_offsets[ndims-1];
		tmp_size = 1;
		for (i=ndims-2; i>=0; i--)
		{
			tmp_size *= array_of_gsizes[i+1];
			disps[1] += tmp_size*st_offsets[i];
		}
	}

	disps[1] *= orig_extent;

	disps[2] = orig_extent;
	for (i=0; i<ndims; i++) disps[2] *= array_of_gsizes[i];

	disps[0] = 0;
	blklens[0] = blklens[1] = blklens[2] = 1;
	types[0] = MPI_LB;
	types[1] = type_new;
	types[2] = MPI_UB;

	NMPI_Type_struct(mpi_errno, (3, blklens, disps, types, newtype));
	MPIU_ERR_CHK_POP(mpi_errno);

	NMPI_Type_free(mpi_errno, (&type_new));
	MPIU_ERR_CHK_POP(mpi_errno);
	MPIU_Free(st_offsets);
	MPIU_Free(coords);

fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIO_TYPE_CREATE_DARRAY);
	return mpi_errno;
}


/* Returns MPI_SUCCESS on success, an MPI error code on failure.  Code above
* needs to call MPIO_Err_return_xxx.
*/
#undef FUNCNAME
#define FUNCNAME MPIOI_Type_block
int MPIOI_Type_block(int *array_of_gsizes, int dim, int ndims, int nprocs,
					 int rank, int darg, int order, MPI_Aint orig_extent,
					 MPI_Datatype type_old, MPI_Datatype *type_new,
					 MPI_Aint *st_offset) 
{
	/* nprocs = no. of processes in dimension dim of grid
	rank = coordinate of this process in dimension dim */
	static const wchar_t FCNAME[] = L"MPIOI_Type_block";
	int mpi_errno = MPI_SUCCESS;
	int blksize, global_size, mysize, i, j;
	MPI_Aint stride;
	MPIDI_STATE_DECL(MPID_STATE_MPIOI_TYPE_BLOCK);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIOI_TYPE_BLOCK);

	global_size = array_of_gsizes[dim];

	if (darg == MPI_DISTRIBUTE_DFLT_DARG)
	{
		blksize = (global_size + nprocs - 1)/nprocs;
	}
	else
	{
		blksize = darg;

		/* --BEGIN ERROR HANDLING-- */
		if (blksize <= 0)
		{
			MPIU_ERR_SET2(mpi_errno, MPI_ERR_ARG, L"**argnonpos", L"**argnonpos %s %d", L"blksize", blksize);
			goto fn_fail;
		}

		if (blksize * nprocs < global_size)
		{
			MPIU_ERR_SET2(mpi_errno, MPI_ERR_ARG, L"**arg", L"**arg %s %d", L"blksize", blksize);
			goto fn_fail;
		}
		/* --END ERROR HANDLING-- */
	}

	j = global_size - blksize*rank;
	mysize = MPIDU_MIN(blksize, j);
	if (mysize < 0) mysize = 0;

	stride = orig_extent;
	if (order == MPI_ORDER_FORTRAN) 
	{
		if (dim == 0)
		{
			NMPI_Type_contiguous(mpi_errno, (mysize, type_old, type_new));
			MPIU_ERR_CHK_POP(mpi_errno);
		}
		else
		{
			for (i=0; i<dim; i++) stride *= array_of_gsizes[i];
			NMPI_Type_hvector(mpi_errno, (mysize, 1, stride, type_old, type_new));
			MPIU_ERR_CHK_POP(mpi_errno);
		}
	}
	else
	{
		if (dim == ndims-1)
		{
			NMPI_Type_contiguous(mpi_errno, (mysize, type_old, type_new));
			MPIU_ERR_CHK_POP(mpi_errno);
		}
		else
		{
			for (i=ndims-1; i>dim; i--) stride *= array_of_gsizes[i];
			NMPI_Type_hvector(mpi_errno, (mysize, 1, stride, type_old, type_new));
			MPIU_ERR_CHK_POP(mpi_errno);
		}
	}

	*st_offset = blksize * rank;
	/* in terms of no. of elements of type oldtype in this dimension */
	if (mysize == 0) *st_offset = 0;

fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIOI_TYPE_BLOCK);
	return mpi_errno;
}


/* Returns MPI_SUCCESS on success, an MPI error code on failure.  Code above
* needs to call MPIO_Err_return_xxx.
*/
#undef FUNCNAME
#define FUNCNAME MPIOI_Type_cyclic
int MPIOI_Type_cyclic(int *array_of_gsizes, int dim, int ndims, int nprocs,
					  int rank, int darg, int order, MPI_Aint orig_extent,
					  MPI_Datatype type_old, MPI_Datatype *type_new,
					  MPI_Aint *st_offset) 
{
	/* nprocs = no. of processes in dimension dim of grid
	rank = coordinate of this process in dimension dim */
	static const wchar_t FCNAME[] = L"MPIOI_Type_cyclic";
	int mpi_errno = MPI_SUCCESS;
	int blksize, i, blklens[3], st_index, end_index, local_size, rem, count;
	MPI_Aint stride, disps[3];
	MPI_Datatype type_tmp, types[3];
	MPIDI_STATE_DECL(MPID_STATE_MPIOI_TYPE_CYCLIC);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIOI_TYPE_CYCLIC);

	if (darg == MPI_DISTRIBUTE_DFLT_DARG) blksize = 1;
	else blksize = darg;

	/* --BEGIN ERROR HANDLING-- */
	if (blksize <= 0)
	{
		MPIU_ERR_SET2(mpi_errno, MPI_ERR_ARG, L"**argnonpos", L"**argnonpos %s %d", L"blksize", blksize);
		goto fn_fail;
	}
	/* --END ERROR HANDLING-- */

	st_index = rank*blksize;
	end_index = array_of_gsizes[dim] - 1;

	if (end_index < st_index)
	{
		local_size = 0;
	}
	else
	{
		local_size = ((end_index - st_index + 1)/(nprocs*blksize))*blksize;
		rem = (end_index - st_index + 1) % (nprocs*blksize);
		local_size += MPIDU_MIN(rem, blksize);
	}

	count = local_size/blksize;
	rem = local_size % blksize;

	stride = nprocs * blksize * orig_extent;
	if (order == MPI_ORDER_FORTRAN)
	{
		for (i=0; i<dim; i++)
		{
			stride *= array_of_gsizes[i];
		}
	}
	else
	{
		for (i=ndims-1; i>dim; i--)
		{
			stride *= array_of_gsizes[i];
		}
	}

	NMPI_Type_hvector(mpi_errno, (count, blksize, stride, type_old, type_new));
	MPIU_ERR_CHK_POP(mpi_errno);

	if (rem)
	{
		/* if the last block is of size less than blksize, include it separately using MPI_Type_struct */

		types[0] = *type_new;
		types[1] = type_old;
		disps[0] = 0;
		disps[1] = count*stride;
		blklens[0] = 1;
		blklens[1] = rem;

		NMPI_Type_struct(mpi_errno, (2, blklens, disps, types, &type_tmp));
		MPIU_ERR_CHK_POP(mpi_errno);

		NMPI_Type_free(mpi_errno, (type_new));
		MPIU_ERR_CHK_POP(mpi_errno);
		*type_new = type_tmp;
	}

	/* In the first iteration, we need to set the displacement in that dimension correctly. */ 
	if ( ((order == MPI_ORDER_FORTRAN) && (dim == 0)) || ((order == MPI_ORDER_C) && (dim == ndims-1)) )
	{
		types[0] = MPI_LB;
		disps[0] = 0;
		types[1] = *type_new;
		disps[1] = rank * blksize * orig_extent;
		types[2] = MPI_UB;
		disps[2] = orig_extent * array_of_gsizes[dim];
		blklens[0] = blklens[1] = blklens[2] = 1;
		NMPI_Type_struct(mpi_errno, (3, blklens, disps, types, &type_tmp));
		MPIU_ERR_CHK_POP(mpi_errno);
		NMPI_Type_free(mpi_errno, (type_new));
		*type_new = type_tmp;

		*st_offset = 0;  /* set it to 0 because it is taken care of in the struct above */
	}
	else
	{
		*st_offset = rank * blksize; 
		/* st_offset is in terms of no. of elements of type oldtype in this dimension */ 
	}

	if (local_size == 0) *st_offset = 0;

fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIOI_TYPE_CYCLIC);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME ADIO_Close
int ADIO_Close(MPID_File_ptr fd)
{
	int mpi_errno = MPI_SUCCESS;
	int i, j, k, combiner, is_contig;
	/*int myrank;*/
	static const wchar_t FCNAME[] = L"ADIO_Close";
	wchar_t error_msg[1024];
	MPIDI_STATE_DECL(MPID_STATE_ADIO_CLOSE);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIO_CLOSE);

	if (fd->async_count)
	{
		_wcserror_s(error_msg, 1024, errno);
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_IO, L"**io", L"**io %s", error_msg);
		MPIDI_FUNC_EXIT(MPID_STATE_ADIO_CLOSE);
		return mpi_errno;
	}

	/* because of deferred open, this warants a bit of explaining.  First, if
	* we've done aggregation (fd->agg_comm has a non-nulll communicator ),
	* then close the file.  Then, if any process left has done independent
	* i/o, close the file.  Otherwise, we'll skip the fs-specific close and
	* just say everything is a-ok.
	*
	* XXX: is it ok for those processes with a "real" communicator and those
	* with "MPI_COMM_SELF" to both call ADIOI_xxx_Close at the same time ?
	* everyone who ever opened the file will close it. Is order important? Is
	* timing important?
	*/
	if (fd->agg_comm != MPI_COMM_NULL)
	{
		mpi_errno = (*(fd->fns->ADIOI_xxx_Close))(fd);
	}
	else
	{
		if (fd->is_open)
		{
			mpi_errno = (*(fd->fns->ADIOI_xxx_Close))(fd);
		}
	}

#if 0 /* Windows implementes DELETE_ON_CLOSE so there is no need to delete the file here.  It was deleted by the close call above. */
	if (fd->access_mode & ADIO_DELETE_ON_CLOSE)
	{
		/* if we are doing aggregation and deferred open, then it's possible
		* that rank 0 does not have access to the file. make sure only an
		* aggregator deletes the file.*/
		NMPI_Comm_rank(mpi_errno, (fd->comm, &myrank));
		MPIU_ERR_CHK_POP(mpi_errno);
		if (myrank == fd->hints->ranklist[0])
		{
			mpi_errno = ADIO_Delete(fd->filename);
			MPIU_ERR_CHK_POP(mpi_errno);
		}
		NMPI_Barrier(mpi_errno, (fd->comm));
		MPIU_ERR_CHK_POP(mpi_errno);
	}
#endif

	MPIU_Free(fd->hints->ranklist);
	MPIU_Free(fd->hints->cb_config_list);
	MPIU_Free(fd->hints);
	NMPI_Comm_free(mpi_errno, (&(fd->comm)));
	MPIU_ERR_CHK_POP(mpi_errno);
	/* deferred open: if we created an aggregator communicator, free it */
	if (fd->agg_comm != MPI_COMM_NULL)
	{
		NMPI_Comm_free(mpi_errno, (&(fd->agg_comm)));
		MPIU_ERR_CHK_POP(mpi_errno);
	}
	MPIU_Free(fd->filename); 

	NMPI_Type_get_envelope(mpi_errno, (fd->etype, &i, &j, &k, &combiner));
	MPIU_ERR_CHK_POP(mpi_errno);
	if (combiner != MPI_COMBINER_NAMED)
	{
		NMPI_Type_free(mpi_errno, (&(fd->etype)));
		MPIU_ERR_CHK_POP(mpi_errno);
	}

	mpi_errno = ADIOI_Datatype_iscontig(fd->filetype, &is_contig);
	MPIU_ERR_CHK_POP(mpi_errno);
	if (!is_contig)
	{
		mpi_errno = ADIOI_Delete_flattened(fd->filetype);
		MPIU_ERR_CHK_POP(mpi_errno);
	}

	NMPI_Type_get_envelope(mpi_errno, (fd->filetype, &i, &j, &k, &combiner));
	MPIU_ERR_CHK_POP(mpi_errno);
	if (combiner != MPI_COMBINER_NAMED)
	{
		NMPI_Type_free(mpi_errno, (&(fd->filetype)));
		MPIU_ERR_CHK_POP(mpi_errno);
	}

	NMPI_Info_free(mpi_errno, (&(fd->info)));
	MPIU_ERR_CHK_POP(mpi_errno);

	/* memory for fd is freed in MPI_File_close */

fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIO_CLOSE);
	return mpi_errno;
}

#undef AGG_DEBUG

/* This file contains four functions:
*
* ADIOI_Calc_aggregator()
* ADIOI_Calc_file_domains()
* ADIOI_Calc_my_req()
* ADIOI_Calc_others_req()
*
* The last three of these were originally in ad_read_coll.c, but they are
* also shared with ad_write_coll.c.  I felt that they were better kept with
* the rest of the shared aggregation code.  
*/

/* Discussion of values available from above:
*
* MPI_Offset st_offsets[0..nprocs-1]
* MPI_Offset end_offsets[0..nprocs-1]
*	These contain a list of start and end offsets for each process in 
*	the communicator.  For example, an access at loc 10, size 10 would
*	have a start offset of 10 and end offset of 19.
* int nprocs
*	number of processors in the collective I/O communicator
* MPI_Offset min_st_offset
* MPI_Offset fd_start[0..nprocs_for_coll-1]
*	starting location of "file domain"; region that a given process will
*	perform aggregation for (i.e. actually do I/O)
* MPI_Offset fd_end[0..nprocs_for_coll-1]
*	start + size - 1 roughly, but it can be less, or 0, in the case of 
*	uneven distributions
*/

/* ADIOI_Calc_aggregator()
*
* The intention here is to implement a function which provides basically 
* the same functionality as in Rajeev's original version of 
* ADIOI_Calc_my_req().  He used a ceiling division approach to assign the 
* file domains, and we use the same approach here when calculating the
* location of an offset/len in a specific file domain.  Further we assume
* this same distribution when calculating the rank_index, which is later
*  used to map to a specific process rank in charge of the file domain.
*
* A better (i.e. more general) approach would be to use the list of file
* domains only.  This would be slower in the case where the
* original ceiling division was used, but it would allow for arbitrary
* distributions of regions to aggregators.  We'd need to know the 
* nprocs_for_coll in that case though, which we don't have now.
*
* Note a significant difference between this function and Rajeev's old code:
* this code doesn't necessarily return a rank in the range
* 0..nprocs_for_coll; instead you get something in 0..nprocs.  This is a
* result of the rank mapping; any set of ranks in the communicator could be
* used now.
*
* Returns an integer representing a rank in the collective I/O communicator.
*
* The "len" parameter is also modified to indicate the amount of data
* actually available in this file domain.
*/
#undef FUNCNAME
#define FUNCNAME ADIOI_Calc_aggregator
int ADIOI_Calc_aggregator(MPID_File_ptr fd,
						  MPI_Offset off, 
						  MPI_Offset min_off, 
						  MPI_Offset *len, 
						  MPI_Offset fd_size,
						  MPI_Offset *fd_start,
						  MPI_Offset *fd_end,
						  int *rank_ptr)
{
	static const wchar_t FCNAME[] = L"ADIOI_Calc_aggregator";
	int mpi_errno = MPI_SUCCESS;
	int rank_index;
	MPI_Offset avail_bytes;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_CALC_AGGREGATOR);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_CALC_AGGREGATOR);

	MPIU_UNREFERENCED_ARG(fd_start);

#ifdef AGG_DEBUG
#if 0
	FPRINTF(stdout, L"off = %lld, min_off = %lld, len = %lld, fd_size = %lld\n", off, min_off, *len, fd_size);
#endif
#endif

	/* get an index into our array of aggregators */
	rank_index = (int) ((off - min_off + fd_size)/ fd_size - 1);

	/* we index into fd_end with rank_index, and fd_end was allocated to be no
	* bigger than fd->hins->cb_nodes.   If we ever violate that, we're
	* overrunning arrays.  Obviously, we should never ever hit this abort */
	if (rank_index >= fd->hints->cb_nodes)
	{
		MPIU_ERR_POP(mpi_errno);
	}

	/* remember here that even in Rajeev's original code it was the case that
	* different aggregators could end up with different amounts of data to
	* aggregate.  here we use fd_end[] to make sure that we know how much
	* data this aggregator is working with.  
	*
	* the +1 is to take into account the end vs. length issue.
	*/
	avail_bytes = fd_end[rank_index] + 1 - off;
	if (avail_bytes < *len)
	{
		/* this file domain only has part of the requested contig. region */
		*len = avail_bytes;
	}

	/* map our index to a rank */
	/* NOTE: FOR NOW WE DON'T HAVE A MAPPING...JUST DO 0..NPROCS_FOR_COLL */
	*rank_ptr = fd->hints->ranklist[rank_index];

fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_CALC_AGGREGATOR);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME ADIOI_Calc_file_domains
int ADIOI_Calc_file_domains(MPI_Offset *st_offsets, MPI_Offset
							*end_offsets, int nprocs, int nprocs_for_coll,
							MPI_Offset *min_st_offset_ptr,
							MPI_Offset **fd_start_ptr, MPI_Offset 
							**fd_end_ptr, MPI_Offset *fd_size_ptr)
{
	/* Divide the I/O workload among "nprocs_for_coll" processes. This is
	done by (logically) dividing the file into file domains (FDs); each
	process may directly access only its own file domain. */
	static const wchar_t FCNAME[] = L"ADIOI_Calc_file_domains";
	int mpi_errno = MPI_SUCCESS;
	MPI_Offset min_st_offset, max_end_offset, *fd_start, *fd_end, fd_size;
	int i;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_CALC_FILE_DOMAINS);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_CALC_FILE_DOMAINS);

#ifdef AGG_DEBUG
	FPRINTF(stderr, L"ADIOI_Calc_file_domains: %d aggregator(s)\n", nprocs_for_coll);
#endif

	/* find min of start offsets and max of end offsets of all processes */

	min_st_offset = st_offsets[0];
	max_end_offset = end_offsets[0];

	for (i=1; i<nprocs; i++)
	{
		min_st_offset = MPIDU_MIN(min_st_offset, st_offsets[i]);
		max_end_offset = MPIDU_MAX(max_end_offset, end_offsets[i]);
	}

	/* determine the "file domain (FD)" of each process, i.e., the portion of
	the file that will be L"owned" by each process */

	/* partition the total file access range equally among nprocs_for_coll processes */ 
	fd_size = ((max_end_offset - min_st_offset + 1) + nprocs_for_coll - 1)/nprocs_for_coll; 
	/* ceiling division as in HPF block distribution */

	*fd_start_ptr = (MPI_Offset *) MPIU_Malloc(nprocs_for_coll*sizeof(MPI_Offset));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, *fd_start_ptr, nprocs_for_coll*sizeof(MPI_Offset), L"fd_start_ptr");
	*fd_end_ptr = (MPI_Offset *) MPIU_Malloc(nprocs_for_coll*sizeof(MPI_Offset));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, *fd_end_ptr, nprocs_for_coll*sizeof(MPI_Offset), L"fd_end_ptr");

	fd_start = *fd_start_ptr;
	fd_end = *fd_end_ptr;

	fd_start[0] = min_st_offset;
	fd_end[0] = min_st_offset + fd_size - 1;

	for (i=1; i<nprocs_for_coll; i++)
	{
		fd_start[i] = fd_end[i-1] + 1;
		fd_end[i] = fd_start[i] + fd_size - 1;
	}

	/* take care of cases in which the total file access range is not
	divisible by the number of processes. In such cases, the last
	process, or the last few processes, may have unequal load (even 0).
	For example, a range of 97 divided among 16 processes.
	Note that the division is ceiling division. */

	for (i=0; i<nprocs_for_coll; i++)
	{
		if (fd_start[i] > max_end_offset)
			fd_start[i] = fd_end[i] = -1;
		if (fd_end[i] > max_end_offset)
			fd_end[i] = max_end_offset;
	}

	*fd_size_ptr = fd_size;
	*min_st_offset_ptr = min_st_offset;

fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_CALC_FILE_DOMAINS);
	return mpi_errno;
}


/* ADIOI_Calc_my_req() - calculate what portions of the access requests
* of this process are located in the file domains of various processes
* (including this one)
*/
#undef FUNCNAME
#define FUNCNAME ADIOI_Calc_my_req
int ADIOI_Calc_my_req(MPID_File_ptr fd, MPI_Offset *offset_list, int *len_list, 
					  int contig_access_count, MPI_Offset 
					  min_st_offset, MPI_Offset *fd_start,
					  MPI_Offset *fd_end, MPI_Offset fd_size,
					  int nprocs,
					  int *count_my_req_procs_ptr,
					  int **count_my_req_per_proc_ptr,
					  ADIOI_Access **my_req_ptr,
					  int **buf_idx_ptr)
{
	static const wchar_t FCNAME[] = L"ADIOI_Calc_my_req";
	int mpi_errno = MPI_SUCCESS;
	int *count_my_req_per_proc, count_my_req_procs, *buf_idx;
	int i, l, proc;
	MPI_Offset fd_len, rem_len, curr_idx, off;
	ADIOI_Access *my_req;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_CALC_MY_REQ);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_CALC_MY_REQ);

	*count_my_req_per_proc_ptr = (int *) MPIU_Calloc(nprocs,sizeof(int)); 
	count_my_req_per_proc = *count_my_req_per_proc_ptr;
	/* count_my_req_per_proc[i] gives the no. of contig. requests of this
	process in process i's file domain. calloc initializes to zero.
	I'm allocating memory of size nprocs, so that I can do an 
	MPI_Alltoall later on.*/

	buf_idx = (int *) MPIU_Malloc(nprocs*sizeof(int));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, buf_idx, nprocs*sizeof(int), L"buf_idx");

	/* buf_idx is relevant only if buftype_is_contig.
	buf_idx[i] gives the index into user_buf where data received
	from proc. i should be placed. This allows receives to be done
	without extra buffer. This can't be done if buftype is not contig. */

	/* initialize buf_idx to -1 */
	for (i=0; i < nprocs; i++) buf_idx[i] = -1;

	/* one pass just to calculate how much space to allocate for my_req;
	* contig_access_count was calculated way back in ADIOI_Calc_my_off_len()
	*/
	for (i=0; i < contig_access_count; i++)
	{
		/* short circuit offset/len processing if len == 0 
		* 		(zero-byte  read/write */
		if (len_list[i] == 0) 
			continue;
		off = offset_list[i];
		fd_len = len_list[i];
		/* note: we set fd_len to be the total size of the access.  then
		* ADIOI_Calc_aggregator() will modify the value to return the 
		* amount that was available from the file domain that holds the
		* first part of the access.
		*/
		mpi_errno = ADIOI_Calc_aggregator(fd, off, min_st_offset, &fd_len, fd_size, fd_start, fd_end, &proc);
		MPIU_ERR_CHK_POP(mpi_errno);
		count_my_req_per_proc[proc]++;

		/* figure out how much data is remaining in the access (i.e. wasn't 
		* part of the file domain that had the starting byte); we'll take 
		* care of this data (if there is any) in the while loop below.
		*/
		rem_len = len_list[i] - fd_len;

		while (rem_len != 0)
		{
			off += fd_len; /* point to first remaining byte */
			fd_len = rem_len; /* save remaining size, pass to calc */
			mpi_errno = ADIOI_Calc_aggregator(fd, off, min_st_offset, &fd_len, fd_size, fd_start, fd_end, &proc);
			MPIU_ERR_CHK_POP(mpi_errno);

			count_my_req_per_proc[proc]++;
			rem_len -= fd_len; /* reduce remaining length by amount from fd */
		}
	}

	/* now allocate space for my_req, offset, and len */

	*my_req_ptr = (ADIOI_Access *)MPIU_Malloc(nprocs*sizeof(ADIOI_Access)); 
	my_req = *my_req_ptr;

	count_my_req_procs = 0;
	for (i=0; i < nprocs; i++)
	{
		if (count_my_req_per_proc[i])
		{
			my_req[i].offsets = (MPI_Offset *)MPIU_Malloc(count_my_req_per_proc[i] * sizeof(MPI_Offset));
			my_req[i].lens = (int *)MPIU_Malloc(count_my_req_per_proc[i] * sizeof(int));
			count_my_req_procs++;
		}
		my_req[i].count = 0;  /* will be incremented where needed later */
	}

	/* now fill in my_req */
	curr_idx = 0;
	for (i=0; i<contig_access_count; i++)
	{
		/* short circuit offset/len processing if len == 0 (zero-byte  read/write */
		if (len_list[i] == 0)
			continue;
		off = offset_list[i];
		fd_len = len_list[i];
		mpi_errno = ADIOI_Calc_aggregator(fd, off, min_st_offset, &fd_len, fd_size, fd_start, fd_end, &proc);
		MPIU_ERR_CHK_POP(mpi_errno);

		/* for each separate contiguous access from this process */
		if (buf_idx[proc] == -1) buf_idx[proc] = (int) curr_idx;

		l = my_req[proc].count;
		curr_idx += (int) fd_len; /* NOTE: Why is curr_idx an int?  Fix? */

		rem_len = len_list[i] - fd_len;

		/* store the proc, offset, and len information in an array
		* of structures, my_req. Each structure contains the 
		* offsets and lengths located in that process's FD, 
		* and the associated count. 
		*/
		my_req[proc].offsets[l] = off;
		my_req[proc].lens[l] = (int) fd_len;
		my_req[proc].count++;

		while (rem_len != 0)
		{
			off += fd_len;
			fd_len = rem_len;
			mpi_errno = ADIOI_Calc_aggregator(fd, off, min_st_offset, &fd_len, fd_size, fd_start, fd_end, &proc);
			MPIU_ERR_CHK_POP(mpi_errno);

			if (buf_idx[proc] == -1) buf_idx[proc] = (int) curr_idx;

			l = my_req[proc].count;
			curr_idx += fd_len;
			rem_len -= fd_len;

			my_req[proc].offsets[l] = off;
			my_req[proc].lens[l] = (int) fd_len;
			my_req[proc].count++;
		}
	}

#ifdef AGG_DEBUG
	for (i=0; i<nprocs; i++)
	{
		if (count_my_req_per_proc[i] > 0)
		{
			FPRINTF(stdout, L"data needed from %d (count = %d):\n", i, 
				my_req[i].count);
			for (l=0; l < my_req[i].count; l++)
			{
				FPRINTF(stdout, L"   off[%d] = %lld, len[%d] = %d\n", l,
					my_req[i].offsets[l], l, my_req[i].lens[l]);
			}
		}
	}
#if 0
	for (i=0; i<nprocs; i++)
	{
		FPRINTF(stdout, L"buf_idx[%d] = 0x%x\n", i, buf_idx[i]);
	}
#endif
#endif

	*count_my_req_procs_ptr = count_my_req_procs;
	*buf_idx_ptr = buf_idx;

fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_CALC_MY_REQ);
	return mpi_errno;
}



#undef FUNCNAME
#define FUNCNAME ADIOI_Calc_others_req
int ADIOI_Calc_others_req(MPID_File_ptr fd, int count_my_req_procs, 
						  int *count_my_req_per_proc,
						  ADIOI_Access *my_req, 
						  int nprocs, int myrank,
						  int *count_others_req_procs_ptr,
						  ADIOI_Access **others_req_ptr)  
{
	/* determine what requests of other processes lie in this process's file domain */

	/* count_others_req_procs = number of processes whose requests lie in
	this process's file domain (including this process itself) 
	count_others_req_per_proc[i] indicates how many separate contiguous
	requests of proc. i lie in this process's file domain. */

	static const wchar_t FCNAME[] = L"ADIOI_Calc_others_req";
	int mpi_errno = MPI_SUCCESS;
	int *count_others_req_per_proc, count_others_req_procs;
	int i, j;
	MPI_Request *send_requests, *recv_requests;
	MPI_Status *statuses;
	ADIOI_Access *others_req;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_CALC_OTHERS_REQ);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_CALC_OTHERS_REQ);

	/* first find out how much to send/recv and from/to whom */

	count_others_req_per_proc = (int *) MPIU_Malloc(nprocs*sizeof(int));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, count_others_req_per_proc, nprocs*sizeof(int), L"count_others_req_ptr_proc");

	NMPI_Alltoall(mpi_errno, (count_my_req_per_proc, 1, MPI_INT,
		count_others_req_per_proc, 1, MPI_INT, fd->comm));
	MPIU_ERR_CHK_POP(mpi_errno);

	*others_req_ptr = (ADIOI_Access *) MPIU_Malloc(nprocs*sizeof(ADIOI_Access));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, *others_req_ptr, nprocs*sizeof(ADIOI_Access), L"others_req_ptr");
	others_req = *others_req_ptr;

	count_others_req_procs = 0;
	for (i=0; i<nprocs; i++)
	{
		if (count_others_req_per_proc[i])
		{
			others_req[i].count = count_others_req_per_proc[i];
			others_req[i].offsets = (MPI_Offset *) MPIU_Malloc(count_others_req_per_proc[i]*sizeof(MPI_Offset));
			MPIU_ERR_MEM_CHK_POP(mpi_errno, others_req[i].offsets, count_others_req_per_proc[i]*sizeof(MPI_Offset), L"offsets");
			others_req[i].lens = (int *) MPIU_Malloc(count_others_req_per_proc[i]*sizeof(int));
			MPIU_ERR_MEM_CHK_POP(mpi_errno, others_req[i].lens, count_others_req_per_proc[i]*sizeof(int), L"lens");
			others_req[i].mem_ptrs = (MPI_Aint *) MPIU_Malloc(count_others_req_per_proc[i]*sizeof(MPI_Aint));
			MPIU_ERR_MEM_CHK_POP(mpi_errno, others_req[i].mem_ptrs, count_others_req_per_proc[i]*sizeof(MPI_Aint), L"mem_ptrs");
			count_others_req_procs++;
		}
		else
		{
			others_req[i].count = 0;
		}
	}

	/* now send the calculated offsets and lengths to respective processes */

	send_requests = (MPI_Request *) MPIU_Malloc(2*(count_my_req_procs+1)*sizeof(MPI_Request)); 
	MPIU_ERR_MEM_CHK_POP(mpi_errno, send_requests, 2*(count_my_req_procs+1)*sizeof(MPI_Request), L"send_requests");
	recv_requests = (MPI_Request *) MPIU_Malloc(2*(count_others_req_procs+1)*sizeof(MPI_Request));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, recv_requests, 2*(count_others_req_procs+1)*sizeof(MPI_Request), L"recv_requests");
	/* +1 to avoid a 0-size malloc */

	j = 0;
	for (i=0; i<nprocs; i++)
	{
		if (others_req[i].count)
		{
			NMPI_Irecv(mpi_errno, (others_req[i].offsets, others_req[i].count, ADIO_OFFSET, i, i+myrank, fd->comm, &recv_requests[j]));
			MPIU_ERR_CHK_POP(mpi_errno);
			j++;
			NMPI_Irecv(mpi_errno, (others_req[i].lens, others_req[i].count, MPI_INT, i, i+myrank+1, fd->comm, &recv_requests[j]));
			MPIU_ERR_CHK_POP(mpi_errno);
			j++;
		}
	}

	j = 0;
	for (i=0; i < nprocs; i++)
	{
		if (my_req[i].count)
		{
			NMPI_Isend(mpi_errno, (my_req[i].offsets, my_req[i].count, ADIO_OFFSET, i, i+myrank, fd->comm, &send_requests[j]));
			MPIU_ERR_CHK_POP(mpi_errno);
			j++;
			NMPI_Isend(mpi_errno, (my_req[i].lens, my_req[i].count, MPI_INT, i, i+myrank+1, fd->comm, &send_requests[j]));
			MPIU_ERR_CHK_POP(mpi_errno);
			j++;
		}
	}

	statuses = (MPI_Status *) MPIU_Malloc((1 + 2* MPIDU_MAX(count_my_req_procs,count_others_req_procs)) * sizeof(MPI_Status));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, statuses, (1 + 2* MPIDU_MAX(count_my_req_procs,count_others_req_procs)) * sizeof(MPI_Status), L"statuses");
	/* +1 to avoid a 0-size malloc */

	NMPI_Waitall(mpi_errno, (2*count_my_req_procs, send_requests, statuses));
	MPIU_ERR_CHK_POP(mpi_errno);
	NMPI_Waitall(mpi_errno, (2*count_others_req_procs, recv_requests, statuses));
	MPIU_ERR_CHK_POP(mpi_errno);

	MPIU_Free(send_requests);
	MPIU_Free(recv_requests);			
	MPIU_Free(statuses);
	MPIU_Free(count_others_req_per_proc);

	*count_others_req_procs_ptr = count_others_req_procs;

fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_CALC_OTHERS_REQ);
	return mpi_errno;
}

/* Generic implementation of IwriteContig calls the blocking WriteContig immediately. */
#undef FUNCNAME
#define FUNCNAME ADIOI_FAKE_IwriteContig
int ADIOI_FAKE_IwriteContig(MPID_File_ptr fd, void *buf, int count, MPI_Datatype datatype, int file_ptr_type, MPI_Offset offset, ADIO_Request *request)
{
	static const wchar_t FCNAME[] = L"ADIOI_FAKE_IwriteContig";
	int mpi_errno = MPI_SUCCESS;
	MPI_Status status;
	int len, typesize;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_FAKE_IWRITECONTIG);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_FAKE_IWRITECONTIG);

	mpi_errno = ADIOI_Malloc_request(request);
	MPIU_ERR_MEM_CHK_POP(mpi_errno, *request, sizeof(ADIO_Request_t), L"request");
	(*request)->optype = ADIOI_WRITE;
	(*request)->fd = fd;
	(*request)->queued = 0;
	(*request)->datatype = datatype;

	NMPI_Type_size(mpi_errno, (datatype, &typesize));
	MPIU_ERR_CHK_POP(mpi_errno);
	len = count * typesize;

	/* Call the blocking function.  It will create an error code if necessary. */
	mpi_errno = ADIO_WriteContig(fd, buf, len, MPI_BYTE, file_ptr_type, offset, &status);
	MPIU_ERR_CHK_POP(mpi_errno);

	fd->async_count++;

#ifdef HAVE_STATUS_SET_BYTES
	NMPI_Get_elements(mpi_errno, (&status, MPI_BYTE, &len));
	MPIU_ERR_CHK_POP(mpi_errno);
	(*request)->nbytes = len;
#endif

fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_FAKE_IWRITECONTIG);
	return mpi_errno;
}


/* Generic implementation of IwriteStrided calls the blocking WriteStrided immediately. */
#undef FUNCNAME
#define FUNCNAME ADIOI_FAKE_IwriteStrided
int ADIOI_FAKE_IwriteStrided(MPID_File_ptr fd, void *buf, int count, 
							 MPI_Datatype datatype, int file_ptr_type,
							 MPI_Offset offset, ADIO_Request *request)
{
	static const wchar_t FCNAME[] = L"ADIOI_FAKE_IwriteStrided";
	int mpi_errno = MPI_SUCCESS;
	MPI_Status status;
#ifdef HAVE_STATUS_SET_BYTES
	int typesize;
#endif
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_FAKE_IWRITESTRIDED);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_FAKE_IWRITESTRIDED);

	mpi_errno = ADIOI_Malloc_request(request);
	MPIU_ERR_MEM_CHK_POP(mpi_errno, *request, sizeof(ADIO_Request_t), L"request");
	(*request)->optype = ADIOI_WRITE;
	(*request)->fd = fd;
	(*request)->datatype = datatype;
	(*request)->queued = 0;
	(*request)->handle = 0;

	/* Call the blocking function.  It will create an error code if necessary. */
	mpi_errno = ADIO_WriteStrided(fd, buf, count, datatype, file_ptr_type, offset, &status);
	MPIU_ERR_CHK_POP(mpi_errno);

	fd->async_count++;

#ifdef HAVE_STATUS_SET_BYTES
	NMPI_Type_size(mpi_errno, (datatype, &typesize));
	MPIU_ERR_CHK_POP(mpi_errno);
	(*request)->nbytes = count * typesize;
#endif

fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_FAKE_IWRITESTRIDED);
	return mpi_errno;
}

/* Generic implementation of IreadContig calls the blocking ReadContig immediately. */
#undef FUNCNAME
#define FUNCNAME ADIOI_FAKE_IreadContig
int ADIOI_FAKE_IreadContig(MPID_File_ptr fd, void *buf, int count, 
						   MPI_Datatype datatype, int file_ptr_type,
						   MPI_Offset offset, ADIO_Request *request)  
{
	static const wchar_t FCNAME[] = L"ADIOI_FAKE_IreadContig";
	int mpi_errno = MPI_SUCCESS;
	MPI_Status status;
	int len, typesize;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_FAKE_IREADCONTIG);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_FAKE_IREADCONTIG);

	mpi_errno = ADIOI_Malloc_request(request);
	MPIU_ERR_MEM_CHK_POP(mpi_errno, *request, sizeof(ADIO_Request_t), L"request");
	(*request)->optype = ADIOI_READ;
	(*request)->fd = fd;
	(*request)->queued = 0;
	(*request)->datatype = datatype;

	NMPI_Type_size(mpi_errno, (datatype, &typesize));
	MPIU_ERR_CHK_POP(mpi_errno);
	len = count * typesize;

	/* Call the blocking function.  It will create an error code if necessary. */
	mpi_errno = ADIO_ReadContig(fd, buf, len, MPI_BYTE, file_ptr_type, offset, &status);
	MPIU_ERR_CHK_POP(mpi_errno);

	fd->async_count++;

#ifdef HAVE_STATUS_SET_BYTES
	NMPI_Get_elements(mpi_errno, (&status, MPI_BYTE, &len));
	MPIU_ERR_CHK_POP(mpi_errno);
	(*request)->nbytes = len;
#endif

fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_FAKE_IREADCONTIG);
	return mpi_errno;
}


/* Generic implementation of IreadStrided calls the blocking ReadStrided immediately. */
#undef FUNCNAME
#define FUNCNAME ADIOI_FAKE_IreadStrided
int ADIOI_FAKE_IreadStrided(MPID_File_ptr fd, void *buf, int count, 
							MPI_Datatype datatype, int file_ptr_type,
							MPI_Offset offset, ADIO_Request *request)
{
	static const wchar_t FCNAME[] = L"ADIOI_FAKE_IreadStrided";
	int mpi_errno = MPI_SUCCESS;
	MPI_Status status;
#ifdef HAVE_STATUS_SET_BYTES
	int typesize;
#endif
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_FAKE_IREADSTRIDED);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_FAKE_IREADSTRIDED);

	mpi_errno = ADIOI_Malloc_request(request);
	MPIU_ERR_MEM_CHK_POP(mpi_errno, *request, sizeof(ADIO_Request_t), L"request");
	(*request)->optype = ADIOI_READ;
	(*request)->fd = fd;
	(*request)->datatype = datatype;
	(*request)->queued = 0;
	(*request)->handle = 0;

	/* Call the blocking function.  It will create an error code if necessary. */
	mpi_errno = ADIO_ReadStrided(fd, buf, count, datatype, file_ptr_type, offset, &status);
	MPIU_ERR_CHK_POP(mpi_errno);

	fd->async_count++;

#ifdef HAVE_STATUS_SET_BYTES
	NMPI_Type_size(mpi_errno, (datatype, &typesize));
	MPIU_ERR_CHK_POP(mpi_errno);
	(*request)->nbytes = count * typesize;
#endif

fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_FAKE_IREADSTRIDED);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME ADIO_ImmediateOpen
int ADIO_ImmediateOpen(MPID_File_ptr fd)
{
	static const wchar_t FCNAME[] = L"ADIO_ImmediateOpen";
	int mpi_errno = MPI_SUCCESS;
	/*int nprocs, myrank; */
	MPIDI_STATE_DECL(MPID_STATE_ADIO_IMMEDIATEOPEN);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIO_IMMEDIATEOPEN);

	mpi_errno = (*(fd->fns->ADIOI_xxx_Open))(fd);
	MPIU_ERR_CHK_POP(mpi_errno);
	fd->is_open = 1;

	/* DEBUG: remove following lines
	NMPI_Comm_size(mpi_errno, (fd->comm, &nprocs));
	MPIU_ERR_CHK_POP(mpi_errno);
	NMPI_Comm_rank(mpi_errno, (fd->comm, &myrank));
	MPIU_ERR_CHK_POP(mpi_errno);
	FPRINTF(stdout, L"[%d/%d] DEBUG: ADIOI_ImmediateOpen called on %s\n", myrank, nprocs, fd->filename);
	*/

fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIO_IMMEDIATEOPEN);
	return mpi_errno;
} 

ADIOI_Flatlist_node *ADIOI_Flatlist = NULL;
ADIOI_Async_node *ADIOI_Async_list_head = NULL, *ADIOI_Async_list_tail = NULL;
/* list of outstanding asynchronous requests */
ADIOI_Async_node *ADIOI_Async_avail_head = NULL,
*ADIOI_Async_avail_tail = NULL;
/* list of available (already malloced) nodes for above async list */
ADIOI_Malloc_async *ADIOI_Malloc_async_head = NULL,
*ADIOI_Malloc_async_tail = NULL;
/* list of malloced areas for async_list, which must be freed in ADIO_End */

int ADIO_Init_keyval = MPI_KEYVAL_INVALID;

#undef FUNCNAME
#define FUNCNAME ADIO_Init
int ADIO_Init(int *argc, wchar_t ***argv)
{
	static const wchar_t FCNAME[] = L"ADIO_Init";
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_ADIO_INIT);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIO_INIT);

	MPIU_UNREFERENCED_ARG(argc);
	MPIU_UNREFERENCED_ARG(argv);

	/* initialize the linked list containing flattened datatypes */
	ADIOI_Flatlist = (ADIOI_Flatlist_node *) MPIU_Malloc(sizeof(ADIOI_Flatlist_node));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, ADIOI_Flatlist, sizeof(ADIOI_Flatlist_node), L"ADIOI_Flatlist");
	ADIOI_Flatlist->type = MPI_DATATYPE_NULL;
	ADIOI_Flatlist->next = NULL;
	ADIOI_Flatlist->blocklens = NULL;
	ADIOI_Flatlist->indices = NULL;

fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIO_INIT);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME ADIOI_GEN_SetInfo
int ADIOI_GEN_SetInfo(MPID_File_ptr fd, MPI_Info users_info)
{
	/* if fd->info is null, create a new info object. 
	Initialize fd->info to default values.
	Initialize fd->hints to default values.
	Examine the info object passed by the user. If it contains values that
	ROMIO understands, override the default. */
	static const wchar_t FCNAME[] = L"ADIOI_GEN_SetInfo";
	int mpi_errno = MPI_SUCCESS;
	MPI_Info info;
	wchar_t *value;
	int flag, intval, tmp_val, nprocs=0, nprocs_is_valid = 0, len;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_GEN_SETINFO);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_GEN_SETINFO);

	if (fd->info == MPI_INFO_NULL)
	{
		NMPI_Info_create(mpi_errno, (&(fd->info)));
		MPIU_ERR_CHK_POP(mpi_errno);
	}
	info = fd->info;

	/* Note that fd->hints is allocated at file open time; thus it is
	* not necessary to allocate it, or check for allocation, here.
	*/

	value = (wchar_t *) MPIU_Malloc((MPI_MAX_INFO_VAL+1)*sizeof(wchar_t));
	MPIU_ERR_MEM_CHK_POP(mpi_errno, value, (MPI_MAX_INFO_VAL+1)*sizeof(wchar_t), L"value");

	/* initialize info and hints to default values if they haven't been
	* previously initialized
	*/
	if (!fd->hints->initialized)
	{
		/* buffer size for collective I/O */
		NMPI_Info_set(mpi_errno, (info, L"cb_buffer_size", ADIOI_CB_BUFFER_SIZE_DFLT));
		MPIU_ERR_CHK_POP(mpi_errno);
		fd->hints->cb_buffer_size = _wtoi(ADIOI_CB_BUFFER_SIZE_DFLT);

		/* default is to let romio automatically decide when to use
		* collective buffering
		*/
		NMPI_Info_set(mpi_errno, (info, L"romio_cb_read", L"automatic"));
		MPIU_ERR_CHK_POP(mpi_errno);
		fd->hints->cb_read = ADIOI_HINT_AUTO;
		NMPI_Info_set(mpi_errno, (info, L"romio_cb_write", L"automatic"));
		MPIU_ERR_CHK_POP(mpi_errno);
		fd->hints->cb_write = ADIOI_HINT_AUTO;

		fd->hints->cb_config_list = NULL;

		/* number of processes that perform I/O in collective I/O */
		NMPI_Comm_size(mpi_errno, (fd->comm, &nprocs));
		MPIU_ERR_CHK_POP(mpi_errno);
		nprocs_is_valid = 1;
		MPIU_Snprintf(value, MPI_MAX_INFO_VAL+1, _TRUNCATE, L"%d", nprocs);
		NMPI_Info_set(mpi_errno, (info, L"cb_nodes", value));
		MPIU_ERR_CHK_POP(mpi_errno);
		fd->hints->cb_nodes = nprocs;

		/* hint indicating that no indep. I/O will be performed on this file */
		NMPI_Info_set(mpi_errno, (info, L"romio_no_indep_rw", L"false"));
		MPIU_ERR_CHK_POP(mpi_errno);
		fd->hints->no_indep_rw = 0;
		/* deferred_open derrived from no_indep_rw and cb_{read,write} */
		fd->hints->deferred_open = 0;

		/* buffer size for data sieving in independent reads */
		NMPI_Info_set(mpi_errno, (info, L"ind_rd_buffer_size", ADIOI_IND_RD_BUFFER_SIZE_DFLT));
		MPIU_ERR_CHK_POP(mpi_errno);
		fd->hints->ind_rd_buffer_size = _wtoi(ADIOI_IND_RD_BUFFER_SIZE_DFLT);

		/* buffer size for data sieving in independent writes */
		NMPI_Info_set(mpi_errno, (info, L"ind_wr_buffer_size", ADIOI_IND_WR_BUFFER_SIZE_DFLT));
		MPIU_ERR_CHK_POP(mpi_errno);
		fd->hints->ind_wr_buffer_size = _wtoi(ADIOI_IND_WR_BUFFER_SIZE_DFLT);

		/* default is to let romio automatically decide when to use data
		* sieving
		*/
		NMPI_Info_set(mpi_errno, (info, L"romio_ds_read", L"automatic"));
		MPIU_ERR_CHK_POP(mpi_errno);
		fd->hints->ds_read = ADIOI_HINT_AUTO;
		NMPI_Info_set(mpi_errno, (info, L"romio_ds_write", L"automatic"));
		MPIU_ERR_CHK_POP(mpi_errno);
		fd->hints->ds_write = ADIOI_HINT_AUTO;

		fd->hints->initialized = 1;
	}

	/* add in user's info if supplied */
	if (users_info != MPI_INFO_NULL)
	{
		NMPI_Info_get(mpi_errno, (users_info, L"cb_buffer_size", MPI_MAX_INFO_VAL, value, &flag));
		MPIU_ERR_CHK_POP(mpi_errno);
		if (flag && ((intval=_wtoi(value)) > 0))
		{
			tmp_val = intval;

			NMPI_Bcast(mpi_errno, (&tmp_val, 1, MPI_INT, 0, fd->comm));
			MPIU_ERR_CHK_POP(mpi_errno);
			/* --BEGIN ERROR HANDLING-- */
			if (tmp_val != intval)
			{
				MPIU_ERR_SET2(mpi_errno, MPI_ERR_NOT_SAME, L"**notsame", L"**notsame %s %s", L"cb_buffer_size", L"MPI_File_set_info");
				goto fn_fail;
			}
			/* --END ERROR HANDLING-- */

			NMPI_Info_set(mpi_errno, (info, L"cb_buffer_size", value));
			MPIU_ERR_CHK_POP(mpi_errno);
			fd->hints->cb_buffer_size = intval;
		}

		/* new hints for enabling/disabling coll. buffering on reads/writes */
		NMPI_Info_get(mpi_errno, (users_info, L"romio_cb_read", MPI_MAX_INFO_VAL, value, &flag));
		MPIU_ERR_CHK_POP(mpi_errno);
		if (flag)
		{
			if (!wcscmp(value, L"enable") || !wcscmp(value, L"ENABLE"))
			{
				NMPI_Info_set(mpi_errno, (info, L"romio_cb_read", value));
				MPIU_ERR_CHK_POP(mpi_errno);
				fd->hints->cb_read = ADIOI_HINT_ENABLE;
			}
			else if (!wcscmp(value, L"disable") || !wcscmp(value, L"DISABLE"))
			{
				/* romio_cb_read overrides no_indep_rw */
				NMPI_Info_set(mpi_errno, (info, L"romio_cb_read", value));
				MPIU_ERR_CHK_POP(mpi_errno);
				NMPI_Info_set(mpi_errno, (info, L"romio_no_indep_rw", L"false"));
				MPIU_ERR_CHK_POP(mpi_errno);
				fd->hints->cb_read = ADIOI_HINT_DISABLE;
				fd->hints->no_indep_rw = ADIOI_HINT_DISABLE;
			}
			else if (!wcscmp(value, L"automatic") || !wcscmp(value, L"AUTOMATIC"))
			{
				NMPI_Info_set(mpi_errno, (info, L"romio_cb_read", value));
				MPIU_ERR_CHK_POP(mpi_errno);
				fd->hints->cb_read = ADIOI_HINT_AUTO;
			}

			tmp_val = fd->hints->cb_read;

			NMPI_Bcast(mpi_errno, (&tmp_val, 1, MPI_INT, 0, fd->comm));
			MPIU_ERR_CHK_POP(mpi_errno);
			/* --BEGIN ERROR HANDLING-- */
			if (tmp_val != fd->hints->cb_read)
			{
				MPIU_ERR_SET2(mpi_errno, MPI_ERR_NOT_SAME, L"**notsame", L"**notsame %s %s", L"romio_cb_read", L"MPI_File_set_info");
				goto fn_fail;
			}
			/* --END ERROR HANDLING-- */
		}
		NMPI_Info_get(mpi_errno, (users_info, L"romio_cb_write", MPI_MAX_INFO_VAL, value, &flag));
		MPIU_ERR_CHK_POP(mpi_errno);

		if (flag)
		{
			if (!wcscmp(value, L"enable") || !wcscmp(value, L"ENABLE"))
			{
				NMPI_Info_set(mpi_errno, (info, L"romio_cb_write", value));
				MPIU_ERR_CHK_POP(mpi_errno);
				fd->hints->cb_write = ADIOI_HINT_ENABLE;
			}
			else if (!wcscmp(value, L"disable") || !wcscmp(value, L"DISABLE"))
			{
				/* romio_cb_write overrides no_indep_rw, too */
				NMPI_Info_set(mpi_errno, (info, L"romio_cb_write", value));
				MPIU_ERR_CHK_POP(mpi_errno);
				NMPI_Info_set(mpi_errno, (info, L"romio_no_indep_rw", L"false"));
				MPIU_ERR_CHK_POP(mpi_errno);
				fd->hints->cb_write = ADIOI_HINT_DISABLE;
				fd->hints->no_indep_rw = ADIOI_HINT_DISABLE;
			}
			else if (!wcscmp(value, L"automatic") || !wcscmp(value, L"AUTOMATIC"))
			{
				NMPI_Info_set(mpi_errno, (info, L"romio_cb_write", value));
				MPIU_ERR_CHK_POP(mpi_errno);
				fd->hints->cb_write = ADIOI_HINT_AUTO;
			}

			tmp_val = fd->hints->cb_write;

			NMPI_Bcast(mpi_errno, (&tmp_val, 1, MPI_INT, 0, fd->comm));
			MPIU_ERR_CHK_POP(mpi_errno);
			/* --BEGIN ERROR HANDLING-- */
			if (tmp_val != fd->hints->cb_write)
			{
				MPIU_ERR_SET2(mpi_errno, MPI_ERR_NOT_SAME, L"**notsame", L"**notsame %s %s", L"romio_cb_write", L"MPI_File_set_info");
				goto fn_fail;
			}
			/* --END ERROR HANDLING-- */
		}

		/* new hint for specifying no indep. read/write will be performed */
		NMPI_Info_get(mpi_errno, (users_info, L"romio_no_indep_rw", MPI_MAX_INFO_VAL, value, &flag));
		MPIU_ERR_CHK_POP(mpi_errno);

		if (flag)
		{
			if (!wcscmp(value, L"true") || !wcscmp(value, L"TRUE"))
			{
				/* if 'no_indep_rw' set, also hint that we will do
				* collective buffering: if we aren't doing independent io,
				* then we have to do collective  */
				NMPI_Info_set(mpi_errno, (info, L"romio_no_indep_rw", value));
				MPIU_ERR_CHK_POP(mpi_errno);
				NMPI_Info_set(mpi_errno, (info, L"romio_cb_write", L"enable"));
				MPIU_ERR_CHK_POP(mpi_errno);
				NMPI_Info_set(mpi_errno, (info, L"romio_cb_read", L"enable"));
				MPIU_ERR_CHK_POP(mpi_errno);
				fd->hints->no_indep_rw = 1;
				fd->hints->cb_read = 1;
				fd->hints->cb_write = 1;
				tmp_val = 1;
			}
			else if (!wcscmp(value, L"false") || !wcscmp(value, L"FALSE"))
			{
				NMPI_Info_set(mpi_errno, (info, L"romio_no_indep_rw", value));
				MPIU_ERR_CHK_POP(mpi_errno);
				fd->hints->no_indep_rw = 0;
				tmp_val = 0;
			}
			else
			{
				/* default is above */
				tmp_val = 0;
			}

			NMPI_Bcast(mpi_errno, (&tmp_val, 1, MPI_INT, 0, fd->comm));
			MPIU_ERR_CHK_POP(mpi_errno);
			/* --BEGIN ERROR HANDLING-- */
			if (tmp_val != fd->hints->no_indep_rw)
			{
				MPIU_ERR_SET2(mpi_errno, MPI_ERR_NOT_SAME, L"**notsame", L"**notsame %s %s", L"romio_no_indep_rw", L"MPI_File_set_info");
				goto fn_fail;
			}
			/* --END ERROR HANDLING-- */
		}
		/* new hints for enabling/disabling data sieving on reads/writes */
		NMPI_Info_get(mpi_errno, (users_info, L"romio_ds_read", MPI_MAX_INFO_VAL, value, &flag));
		MPIU_ERR_CHK_POP(mpi_errno);

		if (flag)
		{
			if (!wcscmp(value, L"enable") || !wcscmp(value, L"ENABLE"))
			{
				NMPI_Info_set(mpi_errno, (info, L"romio_ds_read", value));
				MPIU_ERR_CHK_POP(mpi_errno);
				fd->hints->ds_read = ADIOI_HINT_ENABLE;
			}
			else if (!wcscmp(value, L"disable") || !wcscmp(value, L"DISABLE"))
			{
				NMPI_Info_set(mpi_errno, (info, L"romio_ds_read", value));
				MPIU_ERR_CHK_POP(mpi_errno);
				fd->hints->ds_read = ADIOI_HINT_DISABLE;
			}
			else if (!wcscmp(value, L"automatic") || !wcscmp(value, L"AUTOMATIC"))
			{
				NMPI_Info_set(mpi_errno, (info, L"romio_ds_read", value));
				MPIU_ERR_CHK_POP(mpi_errno);
				fd->hints->ds_read = ADIOI_HINT_AUTO;
			}
			/* otherwise ignore */
		}
		NMPI_Info_get(mpi_errno, (users_info, L"romio_ds_write", MPI_MAX_INFO_VAL, value, &flag));
		MPIU_ERR_CHK_POP(mpi_errno);

		if (flag)
		{
			if (!wcscmp(value, L"enable") || !wcscmp(value, L"ENABLE"))
			{
				NMPI_Info_set(mpi_errno, (info, L"romio_ds_write", value));
				MPIU_ERR_CHK_POP(mpi_errno);
				fd->hints->ds_write = ADIOI_HINT_ENABLE;
			}
			else if (!wcscmp(value, L"disable") || !wcscmp(value, L"DISABLE"))
			{
				NMPI_Info_set(mpi_errno, (info, L"romio_ds_write", value));
				MPIU_ERR_CHK_POP(mpi_errno);
				fd->hints->ds_write = ADIOI_HINT_DISABLE;
			}
			else if (!wcscmp(value, L"automatic") || !wcscmp(value, L"AUTOMATIC"))
			{
				NMPI_Info_set(mpi_errno, (info, L"romio_ds_write", value));
				MPIU_ERR_CHK_POP(mpi_errno);
				fd->hints->ds_write = ADIOI_HINT_AUTO;
			}
			/* otherwise ignore */
		}

		NMPI_Info_get(mpi_errno, (users_info, L"cb_nodes", MPI_MAX_INFO_VAL, value, &flag));
		MPIU_ERR_CHK_POP(mpi_errno);

		if (flag && ((intval=_wtoi(value)) > 0))
		{
			tmp_val = intval;

			NMPI_Bcast(mpi_errno, (&tmp_val, 1, MPI_INT, 0, fd->comm));
			MPIU_ERR_CHK_POP(mpi_errno);
			/* --BEGIN ERROR HANDLING-- */
			if (tmp_val != intval)
			{
				MPIU_ERR_SET2(mpi_errno, MPI_ERR_NOT_SAME, L"**notsame", L"**notsame %s %s", L"cb_nodes", L"MPI_File_set_info");
				goto fn_fail;
			}
			/* --END ERROR HANDLING-- */

			if (!nprocs_is_valid)
			{
				/* if hints were already initialized, we might not
				* have already gotten this?
				*/
				NMPI_Comm_size(mpi_errno, (fd->comm, &nprocs));
				MPIU_ERR_CHK_POP(mpi_errno);
				nprocs_is_valid = 1;
			}
			if (intval < nprocs)
			{
				NMPI_Info_set(mpi_errno, (info, L"cb_nodes", value));
				MPIU_ERR_CHK_POP(mpi_errno);
				fd->hints->cb_nodes = intval;
			}
		}

		NMPI_Info_get(mpi_errno, (users_info, L"ind_wr_buffer_size", MPI_MAX_INFO_VAL, value, &flag));
		MPIU_ERR_CHK_POP(mpi_errno);

		if (flag && ((intval = _wtoi(value)) > 0))
		{
			NMPI_Info_set(mpi_errno, (info, L"ind_wr_buffer_size", value));
			MPIU_ERR_CHK_POP(mpi_errno);
			fd->hints->ind_wr_buffer_size = intval;
		}

		NMPI_Info_get(mpi_errno, (users_info, L"ind_rd_buffer_size", MPI_MAX_INFO_VAL, value, &flag));
		MPIU_ERR_CHK_POP(mpi_errno);

		if (flag && ((intval = _wtoi(value)) > 0))
		{
			NMPI_Info_set(mpi_errno, (info, L"ind_rd_buffer_size", value));
			MPIU_ERR_CHK_POP(mpi_errno);
			fd->hints->ind_rd_buffer_size = intval;
		}

		NMPI_Info_get(mpi_errno, (users_info, L"cb_config_list", MPI_MAX_INFO_VAL, value, &flag));
		MPIU_ERR_CHK_POP(mpi_errno);

		if (flag)
		{
			if (fd->hints->cb_config_list == NULL)
			{
				/* only set cb_config_list if it isn't already set.
				* Note that since we set it below, this ensures that
				* the cb_config_list hint will be set at file open time
				* either by the user or to the default
				*/
				NMPI_Info_set(mpi_errno, (info, L"cb_config_list", value));
				MPIU_ERR_CHK_POP(mpi_errno);
				len = (int)wcslen(value)+1;
				fd->hints->cb_config_list = MPIU_Malloc(len * sizeof(wchar_t));
				MPIU_ERR_MEM_CHK_POP(mpi_errno, fd->hints->cb_config_list, len * sizeof(wchar_t), L"cb_config_list");
				MPIU_Wcsncpy(fd->hints->cb_config_list, value, len);
			}
			/* if it has been set already, we ignore it the second time. 
			* otherwise we would get an error if someone used the same
			* info value with a cb_config_list value in it in a couple
			* of calls, which would be irritating. */
		}
	}

	/* handle cb_config_list default value here; avoids an extra
	* free/alloc and insures it is always set
	*/
	if (fd->hints->cb_config_list == NULL)
	{
		NMPI_Info_set(mpi_errno, (info, L"cb_config_list", ADIOI_CB_CONFIG_LIST_DFLT));
		MPIU_ERR_CHK_POP(mpi_errno);
		len = (int)wcslen(ADIOI_CB_CONFIG_LIST_DFLT)+1;
		fd->hints->cb_config_list = MPIU_Malloc(len * sizeof(wchar_t));
		MPIU_ERR_MEM_CHK_POP(mpi_errno, fd->hints->cb_config_list, len * sizeof(wchar_t), L"cb_config_list");
		MPIU_Wcsncpy(fd->hints->cb_config_list, ADIOI_CB_CONFIG_LIST_DFLT, len);
	}
	/* deferred_open won't be set by callers, but if the user doesn't
	* explicitly disable collecitve buffering (two-phase) and does hint that
	* io w/o independent io is going on, we'll set this internal hint as a
	* convenience */
	if ( ( (fd->hints->cb_read != ADIOI_HINT_DISABLE) &&
		(fd->hints->cb_write != ADIOI_HINT_DISABLE) &&
		fd->hints->no_indep_rw ) )
	{
		fd->hints->deferred_open = 1;
	}
	else
	{
		/* setting romio_no_indep_rw enable and romio_cb_{read,write}
		* disable at the same time doesn't make sense. honor
		* romio_cb_{read,write} and force the no_indep_rw hint to
		* 'disable' */
		NMPI_Info_set(mpi_errno, (info, L"romio_no_indep_rw", L"false"));
		MPIU_ERR_CHK_POP(mpi_errno);
		fd->hints->no_indep_rw = 0;
		fd->hints->deferred_open = 0;
	}

	if ((fd->file_system == ADIO_PIOFS) || (fd->file_system == ADIO_PVFS))
	{
		/* no data sieving for writes in PIOFS and PVFS, because they do not support file locking */
		NMPI_Info_get(mpi_errno, (info, L"ind_wr_buffer_size", MPI_MAX_INFO_VAL, value, &flag));
		MPIU_ERR_CHK_POP(mpi_errno);

		if (flag)
		{
			/* get rid of this value if it is set */
			NMPI_Info_delete(mpi_errno, (info, L"ind_wr_buffer_size"));
			MPIU_ERR_CHK_POP(mpi_errno);
		}
		/* note: leave ind_wr_buffer_size alone; used for other cases as well. */
		NMPI_Info_set(mpi_errno, (info, L"romio_ds_write", L"disable"));
		MPIU_ERR_CHK_POP(mpi_errno);
		fd->hints->ds_write = ADIOI_HINT_DISABLE;
	}

	MPIU_Free(value);

fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_GEN_SETINFO);
	return mpi_errno;
}

/* returns the current location of the shared_fp in terms of the
no. of etypes relative to the current view, and also increments the
shared_fp by the number of etypes to be accessed (incr) in the read
or write following this function. */

#undef FUNCNAME
#define FUNCNAME ADIO_Get_shared_fp
int ADIO_Get_shared_fp(MPID_File_ptr fd, int incr, MPI_Offset *shared_fp)
{
	static const wchar_t FCNAME[] = L"ADIO_Get_shared_fp";
	int mpi_errno = MPI_SUCCESS;
	MPI_Status status;
	MPI_Offset new_fp;
	MPI_Comm dupcommself;
	int locked = 0;
	MPIDI_STATE_DECL(MPID_STATE_ADIO_GET_SHARED_FP);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIO_GET_SHARED_FP);

	if (fd->shared_fp_fd == NULL)
	{
		NMPI_Comm_dup(mpi_errno, (MPI_COMM_SELF, &dupcommself));
		MPIU_ERR_CHK_POP(mpi_errno);
		mpi_errno = ADIO_Open(MPI_COMM_SELF, dupcommself, 
			fd->shared_fp_fname, 
			fd->file_system,
			fd->fns,
			ADIO_CREATE | ADIO_RDWR | ADIO_DELETE_ON_CLOSE, 
			0, MPI_BYTE, MPI_BYTE,
			MPI_INFO_NULL, 
			ADIO_PERM_NULL, &fd->shared_fp_fd);
		MPIU_ERR_CHK_POP(mpi_errno);
		*shared_fp = 0;
		ADIOI_WRITE_LOCK(fd->shared_fp_fd, 0, SEEK_SET, sizeof(MPI_Offset));
		locked = 1;
		mpi_errno = ADIO_ReadContig(fd->shared_fp_fd, shared_fp, sizeof(MPI_Offset), 
			MPI_BYTE, ADIO_EXPLICIT_OFFSET, 0, &status);
		MPIU_ERR_CHK_POP(mpi_errno);
		/* if the file is empty, the above function may return error
		(reading beyond end of file). In that case, shared_fp = 0, 
		set above, is the correct value. */
	}
	else
	{
		ADIOI_WRITE_LOCK(fd->shared_fp_fd, 0, SEEK_SET, sizeof(MPI_Offset));
		locked = 1;
		mpi_errno = ADIO_ReadContig(fd->shared_fp_fd, shared_fp, sizeof(MPI_Offset), MPI_BYTE, ADIO_EXPLICIT_OFFSET, 0, &status);
		MPIU_ERR_CHK_POP(mpi_errno);
	}

	new_fp = *shared_fp + incr;

	mpi_errno = ADIO_WriteContig(fd->shared_fp_fd, &new_fp, sizeof(MPI_Offset), MPI_BYTE, ADIO_EXPLICIT_OFFSET, 0, &status);
	MPIU_ERR_CHK_POP(mpi_errno);
	ADIOI_UNLOCK(fd->shared_fp_fd, 0, SEEK_SET, sizeof(MPI_Offset));

fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIO_GET_SHARED_FP);
	return mpi_errno;

fn_fail:
	if (locked)
	{
		ADIOI_UNLOCK(fd->shared_fp_fd, 0, SEEK_SET, sizeof(MPI_Offset));
	}
	goto fn_exit;
}

static int ADIO_FileSysType_prefix(wchar_t *filename, int *fstype);
static int ADIO_FileSysType_fncall(wchar_t *filename, int *fstype);

/*
ADIO_FileSysType_fncall - determines the file system type for a given file 
using a system-dependent function call

Input Parameters:
. filename - pointer to file name character array

Output Parameters:
. fstype - location in which to store file system type (ADIO_XXX)
. error_code - location in which to store error code

MPI_SUCCESS is stored in the location pointed to by error_code on success.

This function is used by MPI_File_open() and MPI_File_delete() to determine 
file system type.  Most other functions use the type which is stored when the 
file is opened.
*/
#undef FUNCNAME
#define FUNCNAME ADIO_FileSysType_fncall
static int ADIO_FileSysType_fncall(wchar_t *filename, int *fstype)
{
	static const wchar_t FCNAME[] = L"ADIO_FileSysType_fncall";
	MPIDI_STATE_DECL(MPID_STATE_ADIO_FILESYSTYPE_FNCALL);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIO_FILESYSTYPE_FNCALL);
	MPIU_UNREFERENCED_ARG(filename);
	*fstype = ADIO_NTFS; /* only supported FS on Windows */
	MPIDI_FUNC_EXIT(MPID_STATE_ADIO_FILESYSTYPE_FNCALL);
	return MPI_SUCCESS;
}

/*
ADIO_FileSysType_prefix - determines file system type for a file using 
a prefix on the file name.  upper layer should have already determined
that a prefix is present.

Input Parameters:
. filename - path to file, including prefix (xxx:)

Output Parameters:
. fstype - pointer to integer in which to store file system type (ADIO_XXX)
. error_code - pointer to integer in which to store error code

Returns MPI_SUCCESS in error_code on success.  Filename not having a prefix
is considered an error. Except for on Windows systems where the default is NTFS.

*/
#undef FUNCNAME
#define FUNCNAME ADIO_FileSysType_prefix
static int ADIO_FileSysType_prefix(wchar_t *filename, int *fstype)
{
	int mpi_errno = MPI_SUCCESS;
	static const wchar_t FCNAME[] = L"ADIO_ResolveFileSysType_prefix";
	MPIDI_STATE_DECL(MPID_STATE_ADIO_FILESYSTYPE_PREFIX);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIO_FILESYSTYPE_PREFIX);

	if (!wcsncmp(filename, L"pfs:", 4) || !wcsncmp(filename, L"PFS:", 4))
	{
		*fstype = ADIO_PFS;
	}
	else if (!wcsncmp(filename, L"piofs:", 6) || !wcsncmp(filename, L"PIOFS:", 6))
	{
		*fstype = ADIO_PIOFS;
	}
	else if (!wcsncmp(filename, L"ufs:", 4) || !wcsncmp(filename, L"UFS:", 4))
	{
		*fstype = ADIO_UFS;
	}
	else if (!wcsncmp(filename, L"nfs:", 4) || !wcsncmp(filename, L"NFS:", 4))
	{
		*fstype = ADIO_NFS;
	}
	else if (!wcsncmp(filename, L"panfs:", 6) || !wcsncmp(filename, L"PANFS:", 6))
	{
		*fstype = ADIO_PANFS;
	}
	else if (!wcsncmp(filename, L"hfs:", 4) || !wcsncmp(filename, L"HFS:", 4))
	{
		*fstype = ADIO_HFS;
	}
	else if (!wcsncmp(filename, L"xfs:", 4) || !wcsncmp(filename, L"XFS:", 4))
	{
		*fstype = ADIO_XFS;
	}
	else if (!wcsncmp(filename, L"sfs:", 4) || !wcsncmp(filename, L"SFS:", 4))
	{
		*fstype = ADIO_SFS;
	}
	else if (!wcsncmp(filename, L"pvfs:", 5) || !wcsncmp(filename, L"PVFS:", 5))
	{
		*fstype = ADIO_PVFS;
	}
	else if (!wcsncmp(filename, L"pvfs2:", 6)||!wcsncmp(filename, L"PVFS2:", 6))
	{
		*fstype = ADIO_PVFS2;
	}
	else if (!wcsncmp(filename, L"testfs:", 7) || !wcsncmp(filename, L"TESTFS:", 7))
	{
		*fstype = ADIO_TESTFS;
	}
	else if (!wcsncmp(filename, L"ftp:", 4) || !wcsncmp(filename, L"gsiftp:", 7))
	{
		*fstype = ADIO_GRIDFTP;
	}
	else
	{
#ifdef ROMIO_NTFS
		*fstype = ADIO_NTFS;
#else
		*fstype = 0;
		/* --BEGIN ERROR HANDLING-- */
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE,
			FCNAME, __LINE__, MPI_ERR_NO_SUCH_FILE,
			L"**filename", L"**filename %s", filename);
		/* --END ERROR HANDLING-- */
#endif
	}
	MPIDI_FUNC_EXIT(MPID_STATE_ADIO_FILESYSTYPE_PREFIX);
	return mpi_errno;
}

/*@
ADIO_ResolveFileType - determines file system type and operations from
file name string; this is a collective call

Input Parameters:
. comm - communicator across which collective open is performed
. filename - name of file (string)

Output Parameters:
. fstype - (pointer to) int holding file system type
. ops - (address of) pointer to table of valid file operations
. error_code - (pointer to) int holding error code

Notes:
This code used to be in MPI_File_open(), but it has been moved into here in 
order to clean things up.  The goal is to separate all this "did we compile
for this fs type" code from the MPI layer and also to introduce the ADIOI_Fns
tables in a reasonable way. -- Rob, 06/06/2001
@*/
#undef FUNCNAME
#define FUNCNAME ADIO_ResolveFileType
int ADIO_ResolveFileType(MPI_Comm comm, wchar_t *filename, int *fstype, ADIOI_Fns **ops)
{
	int mpi_errno = MPI_SUCCESS;
	int file_system, min_code;
	wchar_t *tmp;
	static const wchar_t FCNAME[] = L"ADIO_ResolveFileType";
	MPIDI_STATE_DECL(MPID_STATE_ADIO_RESOLVEFILETYPE);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIO_RESOLVEFILETYPE);

	file_system = -1;
	tmp = wcschr(filename, L':');
	if (!tmp)
	{
		/* no prefix; use system-dependent function call to determine type */
		mpi_errno = ADIO_FileSysType_fncall(filename, &file_system);
		MPIU_ERR_CHK_POP(mpi_errno);

		/* ensure that everyone came up with the same file system type */
		NMPI_Allreduce(mpi_errno, (&file_system, &min_code, 1, MPI_INT, MPI_MIN, comm));
		MPIU_ERR_CHK_POP(mpi_errno);
		if (min_code == ADIO_NFS) file_system = ADIO_NFS;
	}
	else
	{
		/* prefix specified; just match via prefix and assume everyone got the same thing.
		*
		* perhaps we should have this code go through the allreduce as well?
		*/
		mpi_errno = ADIO_FileSysType_prefix(filename, &file_system);
		MPIU_ERR_CHK_POP(mpi_errno);
	}

	/* verify that we support this file system type and set ops pointer */
	if (file_system == ADIO_PFS)
	{
#ifndef ROMIO_PFS
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_IO, L"**iofstypeunsupported", 0);
		MPIDI_FUNC_EXIT(MPID_STATE_ADIO_RESOLVEFILETYPE);
		return mpi_errno;
#else
		*ops = &ADIO_PFS_operations;
#endif
	}
	if (file_system == ADIO_PIOFS)
	{
#ifndef ROMIO_PIOFS
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_IO, L"**iofstypeunsupported", 0);
		MPIDI_FUNC_EXIT(MPID_STATE_ADIO_RESOLVEFILETYPE);
		return mpi_errno;
#else
		*ops = &ADIO_PIOFS_operations;
#endif
	}
	if (file_system == ADIO_UFS)
	{
#ifndef ROMIO_UFS
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_IO, L"**iofstypeunsupported", 0);
		MPIDI_FUNC_EXIT(MPID_STATE_ADIO_RESOLVEFILETYPE);
		return mpi_errno;
#else
		*ops = &ADIO_UFS_operations;
#endif
	}
	if (file_system == ADIO_NFS)
	{
#ifndef ROMIO_NFS
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_IO, L"**iofstypeunsupported", 0);
		MPIDI_FUNC_EXIT(MPID_STATE_ADIO_RESOLVEFILETYPE);
		return mpi_errno;
#else
		*ops = &ADIO_NFS_operations;
#endif
	}
	if (file_system == ADIO_PANFS)
	{
#ifndef ROMIO_PANFS
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_IO, L"**iofstypeunsupported", 0);
		MPIDI_FUNC_EXIT(MPID_STATE_ADIO_RESOLVEFILETYPE);
		return mpi_errno;
#else
		*ops = &ADIO_PANFS_operations;
#endif
	}
	if (file_system == ADIO_HFS)
	{
#ifndef ROMIO_HFS
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_IO, L"**iofstypeunsupported", 0);
		MPIDI_FUNC_EXIT(MPID_STATE_ADIO_RESOLVEFILETYPE);
		return mpi_errno;
#else
		*ops = &ADIO_HFS_operations;
#endif
	}
	if (file_system == ADIO_XFS)
	{
#ifndef ROMIO_XFS
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_IO, L"**iofstypeunsupported", 0);
		MPIDI_FUNC_EXIT(MPID_STATE_ADIO_RESOLVEFILETYPE);
		return mpi_errno;
#else
		*ops = &ADIO_XFS_operations;
#endif
	}
	if (file_system == ADIO_SFS)
	{
#ifndef ROMIO_SFS
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_IO, L"**iofstypeunsupported", 0);
		MPIDI_FUNC_EXIT(MPID_STATE_ADIO_RESOLVEFILETYPE);
		return mpi_errno;
#else
		*ops = &ADIO_SFS_operations;
#endif
	}
	if (file_system == ADIO_PVFS)
	{
#ifndef ROMIO_PVFS
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_IO, L"**iofstypeunsupported", 0);
		MPIDI_FUNC_EXIT(MPID_STATE_ADIO_RESOLVEFILETYPE);
		return mpi_errno;
#else
		*ops = &ADIO_PVFS_operations;
#endif
	}
	if (file_system == ADIO_PVFS2)
	{
#ifndef ROMIO_PVFS2
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_IO, L"**iofstypeunsupported", 0);
		MPIDI_FUNC_EXIT(MPID_STATE_ADIO_RESOLVEFILETYPE);
		return mpi_errno;
#else
		*ops = &ADIO_PVFS2_operations;
#endif
	}
	if (file_system == ADIO_NTFS)
	{
#ifndef ROMIO_NTFS
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_IO, L"**iofstypeunsupported", 0);
		MPIDI_FUNC_EXIT(MPID_STATE_ADIO_RESOLVEFILETYPE);
		return mpi_errno;
#else
		*ops = &ADIO_NTFS_operations;
#endif
	}
	if (file_system == ADIO_TESTFS)
	{
#ifndef ROMIO_TESTFS
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_IO, L"**iofstypeunsupported", 0);
		MPIDI_FUNC_EXIT(MPID_STATE_ADIO_RESOLVEFILETYPE);
		return mpi_errno;
#else
		*ops = &ADIO_TESTFS_operations;
#endif
	}
	if (file_system == ADIO_GRIDFTP)
	{
#ifndef ROMIO_GRIDFTP
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_IO, L"**iofstypeunsupported", 0);
		MPIDI_FUNC_EXIT(MPID_STATE_ADIO_RESOLVEFILETYPE);
		return mpi_errno;
#else
		*ops = &ADIO_GRIDFTP_operations;
#endif
	}
	*fstype = file_system;

fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIO_RESOLVEFILETYPE);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME ADIO_End
int ADIO_End()
{
	static const wchar_t FCNAME[] = L"ADIO_End";
	int mpi_errno = MPI_SUCCESS;
	ADIOI_Flatlist_node *curr, *next;
	ADIOI_Malloc_async *tmp;
	ADIOI_Malloc_req *tmp1;
	MPIDI_STATE_DECL(MPID_STATE_ADIO_END);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIO_END);

	/*	FPRINTF(stderr, L"reached end\n"); */

	/* delete the flattened datatype list */
	curr = ADIOI_Flatlist;
	while (curr)
	{
		if (curr->blocklens) MPIU_Free(curr->blocklens);
		if (curr->indices) MPIU_Free(curr->indices);
		next = curr->next;
		MPIU_Free(curr);
		curr = next;
	}
	ADIOI_Flatlist = NULL;

	/* --BEGIN ERROR HANDLING-- */
	if (ADIOI_Async_list_head)
	{
		MPIU_ERR_SET(mpi_errno, MPI_ERR_IO, L"**ioasyncwaiting");
		goto fn_fail;
	}
	/* --END ERROR HANDLING-- */

	/* free list of available ADIOI_Async_nodes. */
	while (ADIOI_Malloc_async_head)
	{
		MPIU_Free(ADIOI_Malloc_async_head->ptr);
		tmp = ADIOI_Malloc_async_head;
		ADIOI_Malloc_async_head = ADIOI_Malloc_async_head->next;
		MPIU_Free(tmp);
	}
	ADIOI_Async_avail_head = ADIOI_Async_avail_tail = NULL;
	ADIOI_Malloc_async_head = ADIOI_Malloc_async_tail = NULL;

	/* free all available request objects */
	while (ADIOI_Malloc_req_head)
	{
		MPIU_Free(ADIOI_Malloc_req_head->ptr);
		tmp1 = ADIOI_Malloc_req_head;
		ADIOI_Malloc_req_head = ADIOI_Malloc_req_head->next;
		MPIU_Free(tmp1);
	}
	ADIOI_Malloc_req_head = ADIOI_Malloc_req_tail = NULL;

fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIO_END);
	return mpi_errno;
}

/* This is the delete callback function associated with ADIO_Init_keyval when MPI_COMM_WORLD is freed */

#undef FUNCNAME
#define FUNCNAME ADIOI_End_call
int ADIOI_End_call(MPI_Comm comm, int keyval, void *attribute_val, void *extra_state)
{
	static const wchar_t FCNAME[] = L"ADIOI_End_call";
	int mpi_errno;
	MPIDI_STATE_DECL(MPID_STATE_ADIOI_END_CALL);

	MPIDI_FUNC_ENTER(MPID_STATE_ADIOI_END_CALL);

	MPIU_UNREFERENCED_ARG(comm);
	MPIU_UNREFERENCED_ARG(keyval);
	MPIU_UNREFERENCED_ARG(attribute_val);
	MPIU_UNREFERENCED_ARG(extra_state);

	mpi_errno = ADIO_End();
	MPIU_ERR_CHK_POP(mpi_errno);

fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_ADIOI_END_CALL);
	return mpi_errno;
}
