/*  (C) 2009 by Deino Software. */

#include "mpidi_ch3_impl.h"
#include "mpidimpl.h"
#include "pmi.h"
#include "mpidu_sock.h"
#include "ch3usock.h"

/* Definitions */

/* Define the name of the kvs key used to provide the port name to the children */
#define PARENT_PORT_KVSKEY					L"PARENT_ROOT_PORT_NAME"
#define MPIR_MAX_SPAWN_ERROR_CODES			 256
/* Manage the connection information that is exported to other processes */
#define MPIDI_CH3I_HOST_DESCRIPTION_KEY		L"description"
#define MPIDI_CH3I_PORT_KEY					L"port"
#define MPIDI_CH3I_IFNAME_KEY				L"ifname"
#define BLOCK_ALLOC_SIZE					 4096
/* Request allocation and deallocation */
#define MPID_REQUEST_PREALLOC			     8
#define MPIDI_PASSIVE_TARGET_DONE_TAG	348297
#define MPIDI_PASSIVE_TARGET_RMA_TAG	563924
#define MPIDI_COPY_BUFFER_SZ			 16384

/* Structures */

typedef struct MPIDI_Datatype_stream
{
	unsigned char *buffer;
	unsigned char *cur_pos;
	unsigned char *end;
} MPIDI_Datatype_stream;

typedef struct pg_translation
{
	int pg_index;	/* index of a process group (index in pg_node) */
	int pg_rank;	/* rank in that process group */
} pg_translation;

typedef struct pg_node
{
	int				 index;		/* Internal index of process group (see pg_translation) */
	const wchar_t	*pg_id;
	wchar_t			*str;		/* String describing connection info for pg */
	int				 lenStr;	/* Length of this string */
	struct pg_node	*next;
} pg_node;

typedef struct MPIDI_CH3I_Acceptq_s
{
	struct MPIDI_VC				*vc;
	int							 port_name_tag;
	struct MPIDI_CH3I_Acceptq_s *next;
}
MPIDI_CH3I_Acceptq_t;

typedef struct Win_create_pkt_t
{
	MPIDI_Message_pointer_t addr;
	int pointer_size;
	int disp_unit;
	MPI_Win win;
} Win_create_pkt_t;

/* Prototypes */

static int getNumProcessors( void );
static int getNodeRootRank( int, int * );
static int connection_post_recv_pkt(MPIDI_CH3I_Connection_t * conn);
static int connection_post_send_pkt(MPIDI_CH3I_Connection_t * conn);
static int connection_post_send_pkt_and_pgid(MPIDI_CH3I_Connection_t * conn);
static int connection_post_sendq_req(MPIDI_CH3I_Connection_t * conn);
static void connection_destroy(MPIDI_CH3I_Connection_t * conn);
static int create_derived_datatype(MPID_Request * rreq, MPID_Datatype ** dtp);
static int do_accumulate_op(MPID_Request * rreq);
static int do_simple_accumulate(MPIDI_PT_single_op *single_op);
static int do_simple_get(MPID_Win *win_ptr, MPIDI_Win_lock_queue *lock_queue);
/* These functions help implement the connect/accept algorithm */
static int ExtractLocalPGInfo( MPID_Comm *, pg_translation [], pg_node **, int * );
static int ReceivePGAndDistribute( MPID_Comm *, MPID_Comm *, int, int *, int, MPIDI_PG_t *[] );
static int SendPGtoPeerAndFree( MPID_Comm *, int *, pg_node * );
static int FreeNewVC( MPIDI_VC_t *new_vc );
static int SetupNewIntercomm( MPID_Comm *comm_ptr, int remote_comm_size, pg_translation remote_translation[], MPIDI_PG_t **remote_pg, MPID_Comm *intercomm );
static int MPIDI_CH3I_Initialize_tmp_comm(MPID_Comm **comm_pptr, MPIDI_VC_t *vc_ptr, int is_low_group, int context_id_offset);
static int MPIDI_CH3I_Send_rma_msg(MPIDI_RMA_ops * rma_op, MPID_Win * win_ptr, MPI_Win source_win_handle, MPI_Win target_win_handle, MPIDI_RMA_dtype_info * dtype_info, void ** dataloop, MPID_Request ** request);
static int MPIDI_CH3I_Recv_rma_msg(MPIDI_RMA_ops * rma_op, MPID_Win * win_ptr, MPI_Win source_win_handle, MPI_Win target_win_handle, MPIDI_RMA_dtype_info * dtype_info, void ** dataloop, MPID_Request ** request); 
static int MPIDI_CH3I_Do_passive_target_rma(MPID_Win *win_ptr, int *wait_for_rma_done_pkt);
static int MPIDI_CH3I_Send_lock_put_or_acc(MPID_Win *win_ptr);
static int MPIDI_CH3I_Send_lock_get(MPID_Win *win_ptr);

/* Global Variables */

/* Required for (socket version) upcall to Connect_to_root (see FIXME) */
extern MPIDU_Sock_set_t MPIDI_CH3I_sock_set;

/* Count the number of outstanding close requests */
volatile int MPIDI_Outstanding_close_ops = 0;

MPID_Request MPID_Request_direct[MPID_REQUEST_PREALLOC] = {{0}};
MPIU_Object_alloc_t MPID_Request_mem = { 0, 0, 0, 0, MPID_REQUEST, sizeof(MPID_Request), MPID_Request_direct, MPID_REQUEST_PREALLOC };

static int MPIR_Spawn_error_class = MPI_SUCCESS;
static int MPIR_Spawn_error_code[MPIR_MAX_SPAWN_ERROR_CODES];
static int MPIR_Cur_spawn_error_code = 0;

/* FIXME: We need a finalize handler to perform MPIU_Free(parent_port_name) if it is allocated */
static wchar_t *parent_port_name = NULL; /* Name of parent port if this process was spawned (and is root of comm world) or null */

static int MPIDI_CH3I_listener_port = 0;
static MPIDI_CH3I_Connection_t * MPIDI_CH3I_listener_conn = NULL;

static int entered_flag = 0;
static int entered_count = 0;

static MPIDI_CH3I_Acceptq_t * acceptq_head=0;

static MPID_Request * recvq_posted_head = 0;
static MPID_Request * recvq_posted_tail = 0;
static MPID_Request * recvq_unexpected_head = 0;
static MPID_Request * recvq_unexpected_tail = 0;

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_Shm_connect
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3I_Shm_connect(MPIDI_VC_t *vc, const wchar_t *business_card, int *flag)
{
	int mpi_errno;
	wchar_t hostname[256];
	wchar_t queue_name[100];
	MPIDI_CH3I_BootstrapQ queue;
	MPIDI_CH3I_Shmem_queue_info shm_info;
	int i;
	wchar_t pid_str[20];
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_SHM_CONNECT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_SHM_CONNECT);

	/* get the host and queue from the business card */
	mpi_errno = MPIU_Str_get_string_arg(business_card, MPIDI_CH3I_SHM_HOST_KEY, hostname, 256);
	if (mpi_errno != MPIU_STR_SUCCESS)
	{
		/*wprintf(L"getstringarg(%s, %s) failed.\n", MPIDI_CH3I_SHM_HOST_KEY, business_card);fflush(stdout);*/
		*flag = FALSE;
		/* If there is no shm host key, assume that we can't use shared memory */
		/*mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**argstr_shmhost", 0);*/
		mpi_errno = 0;
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_CONNECT);
		return mpi_errno;
	}
	mpi_errno = MPIU_Str_get_string_arg(business_card, MPIDI_CH3I_SHM_QUEUE_KEY, queue_name, 100);
	if (mpi_errno != MPIU_STR_SUCCESS)
	{
		*flag = FALSE;
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**argstr_shmq", 0);
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_CONNECT);
		return mpi_errno;
	}

	mpi_errno = MPIU_Str_get_string_arg(business_card, MPIDI_CH3I_SHM_PID_KEY, pid_str, 20);
	if (mpi_errno != MPIU_STR_SUCCESS)
	{
		*flag = FALSE;
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**argstr_shmpid", 0);
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_CONNECT);
		return mpi_errno;
	}

	/* compare this host's name with the business card host name */
	if (wcscmp(MPIDI_Process.my_pg->ch.shm_hostname, hostname) != 0)
	{
		*flag = FALSE;
		/*MPIU_DBG_PRINTF((L"%s != %s\n", MPIDI_Process.my_pg->ch.shm_hostname, hostname));*/
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_CONNECT);
		return MPI_SUCCESS;
	}

	*flag = TRUE;
	/*MPIU_DBG_PRINTF((L"%s == %s\n", MPIDI_Process.my_pg->ch.shm_hostname, hostname));*/

	MPIU_DBG_PRINTF((L"attaching to queue: %s\n", queue_name));
	mpi_errno = MPIDI_CH3I_BootstrapQ_attach(queue_name, &queue);
	if (mpi_errno != MPI_SUCCESS)
	{
		*flag = FALSE;
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_CONNECT);
		return MPI_SUCCESS;
	}

	/* create the write queue */
	mpi_errno = MPIDI_CH3I_SHM_Get_memblock(sizeof(MPIDI_CH3I_SHM_Queue_t), &vc->ch.shm_write_queue_info);
	if (mpi_errno != MPI_SUCCESS)
	{
		*flag = FALSE;
		mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**shmconnect_getmem", 0);
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_CONNECT);
		return mpi_errno;
	}
	/*wprintf(L"rank %d sending queue(%s)\n", MPIR_Process.comm_world->rank, vc->ch.shm_write_queue_info.name);*/

	vc->ch.write_shmq = vc->ch.shm_write_queue_info.addr;
	vc->ch.write_shmq->head_index = 0;
	vc->ch.write_shmq->tail_index = 0;
	MPIDI_DBG_PRINTF((60, FCNAME, L"write_shmq head = 0"));
	MPIDI_DBG_PRINTF((60, FCNAME, L"write_shmq tail = 0"));
	for (i=0; i<MPIDI_CH3I_NUM_PACKETS; i++)
	{
		vc->ch.write_shmq->packet[i].offset = 0;
		vc->ch.write_shmq->packet[i].avail = MPIDI_CH3I_PKT_EMPTY;
	}

	/* send the queue connection information */
	/*MPIU_DBG_PRINTF((L"write_shmq: %p, name - %s\n", vc->ch.write_shmq, vc->ch.shm_write_queue_info.key));*/
	shm_info.info = vc->ch.shm_write_queue_info;
	/*
	wprintf(L"comm_world rank %d\nvc->pg_rank %d\nmy_pg_rank %d\n\npg_id:\n<%s>\n",
		MPIR_Process.comm_world->rank,
		vc->pg_rank,
		MPIDI_Process.my_pg_rank,
		vc->pg->id);
	fflush(stdout);
	*/
	MPIU_Wcsncpy(shm_info.pg_id, MPIDI_Process.my_pg->id, 100);
	shm_info.pg_rank = MPIDI_Process.my_pg_rank;
	shm_info.pid = (int)GetProcessId(GetCurrentProcess());
	MPIU_DBG_PRINTF((L"MPIDI_CH3I_Shm_connect: sending bootstrap queue info from rank %d to msg queue %s\n", MPIR_Process.comm_world->rank, queue_name));
	mpi_errno = MPIDI_CH3I_BootstrapQ_send_msg(queue, &shm_info, sizeof(shm_info));
	if (mpi_errno != MPI_SUCCESS)
	{
		MPIDI_CH3I_SHM_Unlink_memblock(&vc->ch.shm_write_queue_info);
		MPIDI_CH3I_SHM_Release_memblock(&vc->ch.shm_write_queue_info);
		*flag = FALSE;
		mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**boot_send", 0);
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_CONNECT);
		return mpi_errno;
	}

	/* MPIU_Free the queue resource */
	/*MPIU_DBG_PRINTF((L"detaching from queue: %s\n", queue_name));*/
	mpi_errno = MPIDI_CH3I_BootstrapQ_detach(queue);
	if (mpi_errno != MPI_SUCCESS)
	{
		MPIDI_CH3I_SHM_Unlink_memblock(&vc->ch.shm_write_queue_info);
		MPIDI_CH3I_SHM_Release_memblock(&vc->ch.shm_write_queue_info);
		*flag = FALSE;
		mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**boot_detach", 0);
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_CONNECT);
		return mpi_errno;
	}

	/*MPIU_DBG_PRINTF((L"Opening process[%d]: %d\n", i, pSharedProcess[i].nPid));*/
	vc->ch.hSharedProcessHandle =
		OpenProcess(STANDARD_RIGHTS_REQUIRED | PROCESS_VM_READ | PROCESS_VM_WRITE | PROCESS_VM_OPERATION, 
		FALSE, _wtoi(pid_str));
	if (vc->ch.hSharedProcessHandle == NULL)
	{
		int err = GetLastError();
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**OpenProcess", L"**OpenProcess %d %d", i, err);
		goto fn_fail;
	}

	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_CONNECT);
	return MPI_SUCCESS;
 fn_fail:
	MPIDI_CH3I_SHM_Unlink_memblock(&vc->ch.shm_write_queue_info);
	MPIDI_CH3I_SHM_Release_memblock(&vc->ch.shm_write_queue_info);
	*flag = FALSE;
	return mpi_errno;
}


#undef FUNCNAME
#define FUNCNAME  MPIDI_CH3I_Connect_to_root_sshm
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3I_Connect_to_root_sshm(const wchar_t * port_name, MPIDI_VC_t ** new_vc)
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_VC_t * vc;
	MPIU_CHKPMEM_DECL(1);
	int port_name_tag;
	int connected;
	MPIDI_CH3_Pkt_t pkt;
	int num_written;
	const wchar_t *cached_pg_id, *dummy_id = L"";
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_CONNECT_TO_ROOT_SSHM);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_CONNECT_TO_ROOT_SSHM);

	mpi_errno = MPIDI_GetTagFromPort(port_name, &port_name_tag);
	if (mpi_errno != MPIU_STR_SUCCESS)
	{
		MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, L"**argstr_port_name_tag");
	}
	MPIU_DBG_MSG_D(CH3_CONNECT,VERBOSE,L"port tag %d",port_name_tag);

	if (*new_vc != NULL)
	{
		MPIU_DBG_MSG(CH3_CONNECT,VERBOSE,L"using old connection");
		vc = *new_vc;
	}
	else
	{
		MPIU_CHKPMEM_MALLOC(vc,MPIDI_VC_t *,sizeof(MPIDI_VC_t),mpi_errno,L"vc");
		/* FIXME - where does this vc get freed? */

		*new_vc = vc;

		MPIDI_VC_Init(vc, NULL, 0);
		MPIDI_CH3.VC_Init(vc);
	}

	vc->ch.state = MPIDI_CH3I_VC_STATE_CONNECTING;

	connected = FALSE;
	/* Make it so that the pg_id is not matched on the other side by sending a dummy value */
	cached_pg_id = MPIDI_Process.my_pg->id;
	MPIDI_Process.my_pg->id = dummy_id;
	mpi_errno = MPIDI_CH3I_Shm_connect(vc, port_name, &connected);
	MPIDI_Process.my_pg->id = cached_pg_id;
	if (mpi_errno != MPI_SUCCESS)
	{
		/* FIXME: Do we want to free the vc instead? Or put this into the fail block? */
		vc->ch.state = MPIDI_CH3I_VC_STATE_FAILED;
		MPIU_ERR_POP(mpi_errno);
	}
	if (!connected)
	{
		/* FIXME: Non conforming (not internationalizable) error message. why not just MPIU_ERR_POP? */
		mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", L"**fail %s", L"unable to establish a shared memory queue connection");
		goto fn_fail;
	}

	MPIDI_CH3I_SHM_Add_to_writer_list(vc);
	vc->ch.state = MPIDI_CH3I_VC_STATE_CONNECTED;
	vc->ch.shm_reading_pkt = TRUE;
	vc->ch.send_active = MPIDI_CH3I_SendQ_head(vc); /* MT */

	MPIDI_Pkt_init(&pkt, MPIDI_CH3I_PKT_SC_CONN_ACCEPT);
	pkt.sc_conn_accept.port_name_tag = port_name_tag;

	mpi_errno = MPIDI_CH3I_SHM_write(vc, &pkt, sizeof(MPIDI_CH3_Pkt_t), &num_written);
	if (mpi_errno != MPI_SUCCESS || num_written != sizeof(MPIDI_CH3_Pkt_t))
	{
		MPIU_ERR_POP(mpi_errno);
	}

 fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_CONNECT_TO_ROOT_SSHM);
	return mpi_errno;
 fn_fail:
	MPIU_CHKPMEM_REAP();
	goto fn_exit;
}

#ifdef USE_DBG_LOGGING
#undef FUNCNAME
#define FUNCNAME  MPIDI_CH3_VC_GetStateString
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
const char * MPIDI_CH3_VC_GetStateString( int state )
{
	const wchar_t *name = L"unknown";
	static wchar_t asdigits[20];
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_VC_GETSTATESTRING);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_VC_GETSTATESTRING);
	
	switch (state)
	{
	case MPIDI_CH3I_VC_STATE_UNCONNECTED: name = L"CH3I_VC_STATE_UNCONNECTED"; break;
	case MPIDI_CH3I_VC_STATE_CONNECTING:  name = L"CH3I_VC_STATE_CONNECTING"; break;
	case MPIDI_CH3I_VC_STATE_CONNECTED:   name = L"CH3I_VC_STATE_CONNECTED"; break;
	case MPIDI_CH3I_VC_STATE_FAILED:	  name = L"CH3I_VC_STATE_FAILED"; break;
	default:
		MPIU_Snprintf( asdigits, sizeof(asdigits)/sizeof(wchar_t), _TRUNCATE, L"%d", state );
		asdigits[20-1] = 0;
		name = (const wchar_t *)asdigits;
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_VC_GETSTATESTRING);
	return name;
}
#endif

/*  MPIDI_CH3U_Finalize_sshm - does scalable shared memory specific channel finalization
 */
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3U_Finalize_sshm
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3U_Finalize_sshm()
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_PG_t * pg;
	MPIDI_PG_t * pg_next;
	int inuse;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3U_FINALIZE_SSHM);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3U_FINALIZE_SSHM);

	/* Free resources allocated in CH3_Init() */
	while (MPIDI_CH3I_Process.shm_reading_list)
	{
		MPIDI_CH3I_SHM_Release_memblock(&MPIDI_CH3I_Process.shm_reading_list->ch.shm_read_queue_info);
		MPIDI_CH3I_Process.shm_reading_list = MPIDI_CH3I_Process.shm_reading_list->ch.shm_next_reader;
	}
	while (MPIDI_CH3I_Process.shm_writing_list)
	{
		MPIDI_CH3I_SHM_Release_memblock(&MPIDI_CH3I_Process.shm_writing_list->ch.shm_write_queue_info);
		MPIDI_CH3I_Process.shm_writing_list = MPIDI_CH3I_Process.shm_writing_list->ch.shm_next_writer;
	}

	/* Used to unlink this within Init but now done in finalize in case someone spawned needs to attach to this bootstrapQ. */
	mpi_errno = MPIDI_CH3I_BootstrapQ_unlink(MPIDI_Process.my_pg->ch.bootstrapQ);
	if (mpi_errno != MPI_SUCCESS)
	{
		MPIU_ERR_SET(mpi_errno,MPI_ERR_OTHER, L"**boot_unlink");
	}
	
	mpi_errno = MPIDI_CH3I_BootstrapQ_destroy(MPIDI_Process.my_pg->ch.bootstrapQ);
	if (mpi_errno != MPI_SUCCESS)
	{
		MPIU_ERR_SET(mpi_errno,MPI_ERR_OTHER, L"**finalize_boot");
	}
	
	/* Added for dynamic processes in ssm.  needed because the vct's can't be freed
	 * earlier since the vc's themselves are still needed here to walk though and
	 * free their member fields.
	 */
	MPIDI_PG_Iterate_reset();
	MPIDI_PG_Get_next(&pg);
	MPIDI_PG_Get_next(&pg_next);
	while(pg_next)
	{
		/* the last one is the original and handled in mpid_finalize.c  */
		MPIDI_PG_Release_ref(pg, &inuse);
		if (inuse == 0)
		{
			MPIDI_PG_Destroy(pg);
		}
		pg = pg_next;
		MPIDI_PG_Get_next(&pg_next);		
	}

	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3U_FINALIZE_SSHM);
	return mpi_errno;
}


/*  MPIDI_CH3U_Get_business_card_sshm - does sshm specific portion of getting a business card
 *
 *	 bc_val_p	 - business card value buffer pointer, updated to the next 
 *					available location or freed if published.
 *	 val_max_sz_p - ptr to maximum value buffer size reduced by the number 
 *					of characters written
 */

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3U_Get_business_card_sshm
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3U_Get_business_card_sshm(wchar_t **bc_val_p, int *val_max_sz_p)
{
	wchar_t queue_name[100];
	int mpi_errno;
	MPIDI_PG_t * pg = MPIDI_Process.my_pg;
	wchar_t pid_str[20];
	DWORD pid;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3U_GET_BUSINESS_CARD_SSHM);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3U_GET_BUSINESS_CARD_SSHM);

	/* must ensure shm_hostname is set prior to this upcall */
/*	 wprintf(L"before first MPIU_Str_add_string_arg\n"); */
	mpi_errno = MPIU_Str_add_string_arg(bc_val_p, val_max_sz_p, MPIDI_CH3I_SHM_HOST_KEY, pg->ch.shm_hostname);
	/* --BEGIN ERROR HANDLING-- */
	if (mpi_errno != MPIU_STR_SUCCESS)
	{
		if (mpi_errno == MPIU_STR_NOMEM)
		{
			MPIU_ERR_SET(mpi_errno,MPI_ERR_OTHER, L"**buscard_len");
		}
		else
		{
			MPIU_ERR_SET(mpi_errno,MPI_ERR_OTHER, L"**buscard");
		}
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3U_GET_BUSINESS_CARD_SSHM);
		return mpi_errno;
	}
	/* --END ERROR HANDLING-- */

/*	 wprintf(L"before MPIDI_CH3I_BootstrapQ_tostring\n"); */
	/* brad: must ensure that KVS was already appropriately updated (i.e. bootstrapQ was set) */
	queue_name[0] = L'\0';
	mpi_errno = MPIDI_CH3I_BootstrapQ_tostring(pg->ch.bootstrapQ, queue_name, sizeof(queue_name)/sizeof(wchar_t));
	/* --BEGIN ERROR HANDLING-- */
	if (mpi_errno != 0)
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**pmi_kvs_get", L"**pmi_kvs_get %d", mpi_errno);
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3U_GET_BUSINESS_CARD_SSHM);
		return mpi_errno;
	}
	/* --END ERROR HANDLING-- */
	
	mpi_errno = MPIU_Str_add_string_arg(bc_val_p, val_max_sz_p, MPIDI_CH3I_SHM_QUEUE_KEY, queue_name);
	/* --BEGIN ERROR HANDLING-- */
	if (mpi_errno != MPIU_STR_SUCCESS)
	{
		if (mpi_errno == MPIU_STR_NOMEM)
		{
			MPIU_ERR_SET(mpi_errno,MPI_ERR_OTHER, L"**buscard_len");
		}
		else
		{
			MPIU_ERR_SET(mpi_errno,MPI_ERR_OTHER, L"**buscard");
		}
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3U_GET_BUSINESS_CARD_SSHM);
		return mpi_errno;
	}
	/* --END ERROR HANDLING-- */

	pid = GetCurrentProcessId();
	MPIU_Snprintf(pid_str, 20, _TRUNCATE, L"%d", pid);
	mpi_errno = MPIU_Str_add_string_arg(bc_val_p, val_max_sz_p, MPIDI_CH3I_SHM_PID_KEY, pid_str);
	/* --BEGIN ERROR HANDLING-- */
	if (mpi_errno != MPIU_STR_SUCCESS)
	{
		if (mpi_errno == MPIU_STR_NOMEM)
		{
			MPIU_ERR_SET(mpi_errno,MPI_ERR_OTHER, L"**buscard_len");
		}
		else
		{
			MPIU_ERR_SET(mpi_errno,MPI_ERR_OTHER, L"**buscard");
		}
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3U_GET_BUSINESS_CARD_SSHM);
		return mpi_errno;
	}
	/* --END ERROR HANDLING-- */

	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3U_GET_BUSINESS_CARD_SSHM);
	return MPI_SUCCESS;
}

/*  MPIDI_CH3U_Init_sshm - does scalable shared memory specific channel 
 *  initialization
 *	 bc_val	 - business card value buffer pointer, updated to the next 
 *				  available location or freed if published.
 *	 val_max_sz - maximum value buffer size reduced by the number of 
 *				  characters written
 *							   
 */

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3U_Init_sshm
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3U_Init_sshm(int has_parent, MPIDI_PG_t *pg_p, int pg_rank, wchar_t **bc_val_p, int *val_max_sz_p)
{
	int mpi_errno = MPI_SUCCESS;
	int pmi_errno;
	int pg_size;
	int p;
	int key_max_sz;
	int val_max_sz;
	wchar_t * key = NULL;
	wchar_t * val = NULL;
	wchar_t *kvsname = NULL;
	MPIU_CHKLMEM_DECL(2);
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3U_INIT_SSHM);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3U_INIT_SSHM);

	MPIU_UNREFERENCED_ARG(has_parent);

	srand((unsigned int)GetProcessId(GetCurrentProcess())); /* needed by MPIDI_CH3U_Generate_shm_string */

	MPIDI_CH3I_Process.shm_reading_list = NULL;
	MPIDI_CH3I_Process.shm_writing_list = NULL;
	MPIDI_CH3I_Process.num_cpus = -1;

	/* Get the kvsname associated with MPI_COMM_WORLD */
	MPIDI_PG_GetConnKVSname( &kvsname );

	pmi_errno = PMI_KVS_Get_key_length_max(&key_max_sz);
	if (pmi_errno != PMI_SUCCESS)
	{
		MPIU_ERR_SETANDJUMP2(mpi_errno,MPI_ERR_OTHER, L"**pmi_kvs_get_key_length_max", L"**pmi_kvs_get_key_length_max %d %s", pmi_errno, PMI_Get_last_error_msgW());
	}

	MPIU_CHKLMEM_MALLOC(key,wchar_t *,key_max_sz*sizeof(wchar_t),mpi_errno,L"key");
	
	pmi_errno = PMI_KVS_Get_value_length_max(&val_max_sz);
	if (pmi_errno != PMI_SUCCESS)
	{
		MPIU_ERR_SETANDJUMP2(mpi_errno,MPI_ERR_OTHER, L"**pmi_kvs_get_value_length_max", L"**pmi_kvs_get_value_length_max %d %s", pmi_errno, PMI_Get_last_error_msgW());
	}

	MPIU_CHKLMEM_MALLOC(val,wchar_t *,val_max_sz*sizeof(wchar_t),mpi_errno,L"val");

	/* set the global variable defaults */
	pg_p->ch.nShmEagerLimit = sshmMPIDI_SHM_EAGER_LIMIT;
	pg_p->ch.nShmRndvLimit = sshmMPIDI_SHM_RNDV_LIMIT;
	pg_p->ch.nShmWaitSpinCount = MPIDI_CH3I_SPIN_COUNT_DEFAULT;
	pg_p->ch.nShmWaitYieldCount = MPIDI_CH3I_YIELD_COUNT_DEFAULT;

	/* Figure out how many processors are available and set the spin count 
	   accordingly */
	/* If there were topology information available we could calculate a 
	   multi-cpu number */
	{
		int ncpus = getNumProcessors();
		/* if you know the number of processors, calculate the spin count 
		   relative to that number */
		if (ncpus == 1)
			pg_p->ch.nShmWaitSpinCount = 1;
		/* FIXME: Why is this commented out? */
		/*
		else if (ncpus  < num_procs_per_node)
			pg->ch.nShmWaitSpinCount = ( MPIDI_CH3I_SPIN_COUNT_DEFAULT * ncpus) / num_procs_per_node;
		*/
		if (ncpus > 0)
			MPIDI_CH3I_Process.num_cpus = ncpus;
	}

	pmi_errno = PMI_Get_size(&pg_size);
	if (pmi_errno != 0)
	{
		MPIU_ERR_SETANDJUMP2(mpi_errno,MPI_ERR_OTHER, L"**pmi_get_size", L"**pmi_get_size %d %s", pmi_errno, PMI_Get_last_error_msgW());
	}

	/* FIXME: Is this initialization correct (it should have been done
	   when the pg was created. If not, then the process by which channels
	   and PG's are initialized needs to be documented and explained! 
	   An alternate approach to having this done after the PG is created
	   is to define a method initialization routine that is called
	   after the pg is created but before the final, channel-specific 
	   virtual connection initialization is done. */
	/* Initialize the VC table associated with this process group (and thus COMM_WORLD) */
	for (p = 0; p < pg_size; p++)
	{
		MPIDI_CH3.VC_Init( &pg_p->vct[p] );
	}

	/* brad : do the shared memory specific setup items so we can later do the
	 *		 shared memory aspects of the business card
	 */

	MPID_Get_processor_name( pg_p->ch.shm_hostname, sizeof(pg_p->ch.shm_hostname)/sizeof(wchar_t), 0);

	MPIU_UNREFERENCED_ARG(pg_rank);

	mpi_errno = MPIDI_CH3I_BootstrapQ_create(&pg_p->ch.bootstrapQ);
	if (mpi_errno != MPI_SUCCESS)
	{
		MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, L"**boot_create");
	}

	/* the pg needs to be set for sshm channels.  for all channels this is done in mpid_init.c */
	MPIDI_Process.my_pg = pg_p;

	/* get the sshm part of the business card  */
	mpi_errno = MPIDI_CH3U_Get_business_card_sshm(bc_val_p, val_max_sz_p);
	if (mpi_errno != MPI_SUCCESS)
	{
		MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, L"**init_buscard");
	}

 fn_exit:
	MPIU_CHKLMEM_FREEALL();
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3U_INIT_SSHM);
	return mpi_errno;
 fn_fail:
	/* --BEGIN ERROR HANDLING-- */
	if (pg_p != NULL)
	{
		MPIDI_PG_Destroy( pg_p );
	}

	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* This routine initializes shm-specific elements of the VC */
#undef FUNCNAME
#define FUNCNAME MPIDI_VC_InitShm
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_VC_InitShm( MPIDI_VC_t *vc ) 
{
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_VC_INITSHM);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_VC_INITSHM);
	vc->ch.recv_active		= NULL;
	vc->ch.send_active		= NULL;
	vc->ch.req				= NULL;
	vc->ch.read_shmq		  = NULL;
	vc->ch.write_shmq		 = NULL;
	vc->ch.shm				= NULL;
	vc->ch.shm_state		  = 0;
	vc->ch.shm_next_reader	= NULL;
	vc->ch.shm_next_writer	= NULL;
	vc->ch.shm_read_connected = 0;
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_VC_INITSHM);
	return 0;
}

/* Return the number of processors, or one if the number cannot be determined */
#undef FUNCNAME
#define FUNCNAME getNumProcessors
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int getNumProcessors( void )
{
	SYSTEM_INFO info;
	int num_cpus;
	MPIDI_STATE_DECL(MPID_STATE_GETNUMPROCESSORS);

	MPIDI_FUNC_ENTER(MPID_STATE_GETNUMPROCESSORS);
	GetSystemInfo(&info);
	num_cpus = info.dwNumberOfProcessors;
	MPIDI_FUNC_EXIT(MPID_STATE_GETNUMPROCESSORS);
	return (num_cpus > 0) ? num_cpus : 1;
}

/* Return the lowest rank on the same node as this process */
#undef FUNCNAME
#define FUNCNAME getNodeRootRank
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int getNodeRootRank(int pg_rank, int *root_rank)
{
	int *ranks = NULL, num_ranks, min_rank;
	int pmi_errno, i;
	int mpi_errno = MPI_SUCCESS;
	MPIU_CHKLMEM_DECL(1);
	MPIDI_STATE_DECL(MPID_STATE_GETNODEROOTRANK);

	MPIDI_FUNC_ENTER(MPID_STATE_GETNODEROOTRANK);

	MPIU_UNREFERENCED_ARG(pg_rank);

	pmi_errno = PMI_Get_clique_size(&num_ranks);
	if (pmi_errno != PMI_SUCCESS)
	{
		MPIU_ERR_SETANDJUMP2(mpi_errno,MPI_ERR_OTHER, L"**pmi_get_clique_size", L"**pmi_get_clique_size %d %s", pmi_errno, PMI_Get_last_error_msgW());
	}

	if (num_ranks < 1)
	{
		/* PMI_Get_clique_size returned an invalid size */
		MPIU_ERR_SETANDJUMP1(mpi_errno, MPI_ERR_OTHER, L"**pmi_invalid_clique_size", L"**pmi_invalid_clique_size %d", num_ranks);
	}

	MPIU_CHKLMEM_MALLOC(ranks, int *, sizeof(int) * num_ranks, mpi_errno, L"an array of ranks");

	pmi_errno = PMI_Get_clique_ranks(ranks, num_ranks);
	if (pmi_errno != PMI_SUCCESS)
	{
		MPIU_ERR_SETANDJUMP2(mpi_errno,MPI_ERR_OTHER, L"**pmi_get_clique_ranks", L"**pmi_get_clique_ranks %d %s", pmi_errno, PMI_Get_last_error_msgW());
	}

	min_rank = ranks[0];
	/* If the list is not sorted then find the lowest rank */
	for (i=1; i<num_ranks; i++)
	{
		min_rank = MPIDU_MIN(min_rank, ranks[i]);
	}

	*root_rank = min_rank;
fn_exit:
	MPIU_CHKLMEM_FREEALL();
	MPIDI_FUNC_EXIT(MPID_STATE_GETNODEROOTRANK);
	return mpi_errno;
fn_fail:
	goto fn_exit;
}

/* Turn a SINGLE MPI_Info into an array of PMI_keyvals (return the pointer to the array of PMI keyvals) */
/*
#undef FUNCNAME
#define FUNCNAME mpi_to_pmi_keyvals
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int  mpi_to_pmi_keyvals( MPID_Info *info_ptr, PMI_keyvalW_t **kv_ptr, 
								int *nkeys_ptr )
{
	wchar_t key[MPI_MAX_INFO_KEY];
	PMI_keyvalW_t *kv = 0;
	int		  i, nkeys = 0, vallen, flag, mpi_errno=MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPI_TO_PMI_KEYVALS);

	MPIDI_FUNC_ENTER(MPID_STATE_MPI_TO_PMI_KEYVALS);

	if (!info_ptr || info_ptr->handle == MPI_INFO_NULL) {
		goto fn_exit;
	}

	NMPI_Info_get_nkeys(mpi_errno, ( info_ptr->handle, &nkeys ));
	if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }
	if (nkeys == 0) {
		goto fn_exit;
	}
	kv = (PMI_keyvalW_t *)MPIU_Malloc( nkeys * sizeof(PMI_keyvalW_t) );
	if (!kv) { MPIU_ERR_POP(mpi_errno); }

	for (i=0; i<nkeys; i++) {
		NMPI_Info_get_nthkey(mpi_errno, ( info_ptr->handle, i, key ));
		if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }
		NMPI_Info_get_valuelen(mpi_errno, ( info_ptr->handle, key, &vallen, 
											&flag ));
		if (!flag || mpi_errno) { MPIU_ERR_POP(mpi_errno); }
		kv[i].key = MPIU_Strdup(key);
		kv[i].val = MPIU_Malloc( (vallen + 1) * sizeof(wchar_t) );
		if (!kv[i].key || !kv[i].val) { 
			MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER,L"**nomem" );
		}
		NMPI_Info_get(mpi_errno, ( info_ptr->handle, key, vallen+1,
								   kv[i].val, &flag ));
		if (!flag || mpi_errno) { MPIU_ERR_POP(mpi_errno); }
		MPIU_DBG_PRINTF((L"key: <%s>, value: <%s>\n", kv[i].key, kv[i].val));
	}

 fn_fail:
 fn_exit:
	*kv_ptr	= kv;
	*nkeys_ptr = nkeys;
	MPIDI_FUNC_EXIT(MPID_STATE_MPI_TO_PMI_KEYVALS);
	return mpi_errno;
}
*/

#undef FUNCNAME
#define FUNCNAME free_pmi_keyvals
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static void free_pmi_keyvals(PMI_keyvalW_t **kv, int size, int *counts)
{
	int i,j;
	MPIDI_STATE_DECL(MPID_STATE_FREE_PMI_KEYVALS);

	MPIDI_FUNC_ENTER(MPID_STATE_FREE_PMI_KEYVALS);

	for (i=0; i<size; i++)
	{
		for (j=0; j<counts[i]; j++)
		{
			if (kv[i][j].key != NULL)
				MPIU_Free(kv[i][j].key);
			if (kv[i][j].val != NULL)
				MPIU_Free(kv[i][j].val);
		}
		if (kv[i] != NULL)
		{
			MPIU_Free(kv[i]);
		}
	}
	if (kv != NULL)
		MPIU_Free(kv);
	MPIDI_FUNC_EXIT(MPID_STATE_FREE_PMI_KEYVALS);
}

/*
 * MPIDI_CH3_Comm_spawn_multiple()
 */
#undef FUNCNAME
#define FUNCNAME MPIDI_Comm_spawn_multiple
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_Comm_spawn_multiple(int count, wchar_t **commands, 
								  wchar_t ***argvs, int *maxprocs, 
								  MPID_Info **info_ptrs, int root,
								  MPID_Comm *comm_ptr, MPID_Comm
								  **intercomm, int *errcodes) 
{
	wchar_t port_name[MPI_MAX_PORT_NAME];
	int *info_keyval_sizes, i, mpi_errno=MPI_SUCCESS;
	PMI_keyvalW_t **info_keyval_vectors, preput_keyval_vector;
	int icount, j;
	MPID_Info *iter;
	wchar_t key[MPI_MAX_INFO_KEY];
	int vallen, flag;
	int *pmi_errcodes;
	int total_num_processes;
	int errors_found = 0;
	int quiet = 0;
	int spawn_errno = MPI_SUCCESS;
	wchar_t pmi_error_str[MPI_MAX_ERROR_STRING];
	MPIU_THREADPRIV_DECL;

	MPIDI_STATE_DECL(MPID_STATE_MPIDI_COMM_SPAWN_MULTIPLE);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_COMM_SPAWN_MULTIPLE);

	MPIU_THREADPRIV_GET;
	MPIR_Nest_incr();

	if (comm_ptr->rank == root)
	{
		if (MPIR_Spawn_error_class == MPI_SUCCESS)
		{
			/* Create an error class and error codes for spawn errors */
			NMPI_Add_error_class(mpi_errno, (&MPIR_Spawn_error_class));
			MPIU_ERR_CHK_POP(mpi_errno);
			NMPI_Add_error_string(mpi_errno, (MPIR_Spawn_error_class, L"The process manager was unable to spawn the process"));
			MPIU_ERR_CHK_POP(mpi_errno);
			for (i=0; i<100; i++)
			{
				NMPI_Add_error_code(mpi_errno, (MPIR_Spawn_error_class, &MPIR_Spawn_error_code[i]));
				MPIU_ERR_CHK_POP(mpi_errno);
			}
		}

		/* convert the infos into PMI keyvals */
		info_keyval_sizes = (int *) MPIU_Malloc(count * sizeof(int));
		info_keyval_vectors = (PMI_keyvalW_t**) MPIU_Malloc(count * sizeof(PMI_keyvalW_t*));

		for (i=0; i<count; i++)
		{
			if (info_ptrs != NULL)
			{
				icount = 0;
				if (info_ptrs[i] != NULL)
				{
					NMPI_Info_get_nkeys(mpi_errno, (info_ptrs[i]->handle, &icount));
					MPIU_ERR_CHK_POP(mpi_errno);
				}
				info_keyval_sizes[i] = icount;
				if (icount > 0)
				{
					info_keyval_vectors[i] = (PMI_keyvalW_t*) MPIU_Malloc(icount * sizeof(PMI_keyvalW_t));
				}
				else
				{
					info_keyval_vectors[i] = NULL;
				}
				iter = info_ptrs[i];
				for (j=0; j<icount; j++)
				{
					NMPI_Info_get_nthkey(mpi_errno, (info_ptrs[i]->handle, j, key));
					MPIU_ERR_CHK_POP(mpi_errno);
					NMPI_Info_get_valuelen(mpi_errno, (info_ptrs[i]->handle, key, &vallen, &flag));
					MPIU_ERR_CHK_POP(mpi_errno);
					if (!flag)
					{
						MPIU_ERR_POP(mpi_errno);
					}

					info_keyval_vectors[i][j].key = MPIU_Strdup(key);
					info_keyval_vectors[i][j].val = MPIU_Malloc((vallen + 1)* sizeof(wchar_t));
					NMPI_Info_get(mpi_errno, (info_ptrs[i]->handle, key, vallen+1, info_keyval_vectors[i][j].val, &flag));
					MPIU_ERR_CHK_POP(mpi_errno);
					if (!flag)
					{
						MPIU_ERR_POP(mpi_errno);
					}
					MPIU_DBG_PRINTF((L"key: <%s>, value: <%s>\n", info_keyval_vectors[i][j].key, info_keyval_vectors[i][j].val));
				}
			}
			else
			{
				info_keyval_sizes[i] = 0;
				info_keyval_vectors[i] = NULL;
			}
		}

		/* create an array for the pmi error codes */
		total_num_processes = 0;
		for (i=0; i<count; i++)
		{
			total_num_processes += maxprocs[i];
		}
		pmi_errcodes = (int*)MPIU_Malloc(sizeof(int) * total_num_processes);
		/* --BEGIN ERROR HANDLING-- */
		if (pmi_errcodes == NULL)
		{
			free_pmi_keyvals(info_keyval_vectors, count, info_keyval_sizes);
			MPIU_Free(info_keyval_sizes);
			mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
			goto fn_fail;
		}
		/* --END ERROR HANDLING-- */

		/* initialize them to 0 */
		for (i=0; i<total_num_processes; i++)
			pmi_errcodes[i] = 0;

		/* Open a port for the spawned processes to connect to */
		/* FIXME: info may be needed for port name */
		mpi_errno = MPID_Open_port(NULL, port_name);
		/* --BEGIN ERROR HANDLING-- */
		if (mpi_errno != MPI_SUCCESS)
		{
			free_pmi_keyvals(info_keyval_vectors, count, info_keyval_sizes);
			MPIU_Free(info_keyval_sizes);
			MPIU_Free(pmi_errcodes);
			MPIU_ERR_POP(mpi_errno);
		}
		/* --END ERROR HANDLING-- */

		preput_keyval_vector.key = PARENT_PORT_KVSKEY;
		preput_keyval_vector.val = port_name;

		/* If the user will receive the errors in the output array then don't print errors.
		 * This allows users to handle spawn errors without muddying the output with error messages.
		 * This only affects errors during PMI_Spawn_multipleW.  Errors after startup will still be 
		 * printed to stdout. */
		if (comm_ptr->errhandler != NULL && comm_ptr->errhandler->handle == MPI_ERRORS_RETURN)
			quiet = 1;

		/* Spawn the processes */
		mpi_errno = PMI_Spawn_multipleW(count, (const wchar_t **)
									   commands, 
									   (const wchar_t ***) argvs,
									   maxprocs, info_keyval_sizes,
									   (const PMI_keyvalW_t **)
									   info_keyval_vectors, 1, 
									   &preput_keyval_vector,
									   pmi_errcodes, quiet);

		/* --BEGIN ERROR HANDLING-- */
		if (mpi_errno != PMI_SUCCESS)
		{
			free_pmi_keyvals(info_keyval_vectors, count, info_keyval_sizes);
			MPIU_Free(info_keyval_sizes);
			MPIU_Free(pmi_errcodes);
			mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**pmi_spawn_multiple", L"**pmi_spawn_multiple %d %s", mpi_errno, PMI_Get_last_error_msgW());
			goto fn_fail;
		}

		/* --END ERROR HANDLING-- */
		if (errcodes != MPI_ERRCODES_IGNORE)
		{
			for (i=0; i<total_num_processes; i++)
			{
				if (pmi_errcodes[i] != 0)
				{
					errors_found = 1;
					/* FIXME: Should we bypass this error creation code when !quiet since the user will not get access to the errcodes array? */
					if (PMI_Get_error_stringW(pmi_errcodes[i], pmi_error_str, MPI_MAX_ERROR_STRING) == PMI_SUCCESS)
					{
						errcodes[i] = MPIR_Spawn_error_code[MPIR_Cur_spawn_error_code++];
						if (MPIR_Cur_spawn_error_code == MPIR_MAX_SPAWN_ERROR_CODES)
							MPIR_Cur_spawn_error_code = 0;
						NMPI_Add_error_string(mpi_errno, (errcodes[i], pmi_error_str));
						MPIU_ERR_CHK_POP(mpi_errno);
						if (!quiet)
						{
							/* Without MPI_ERRORS_RETURN the process is going to abort.  Save the last error number so one
							   error will be printed out.  The user will not get access to the error code array. */
							spawn_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", L"**fail %s", pmi_error_str);
							/*wprintf(L"PMI error %d string: <%s>\n", pmi_errcodes[i], pmi_error_str);fflush(stdout);*/
							/*spawn_errno = errcodes[i];*/ /* This doesn't work because the error stack code can't handle dynamic error codes */
						}
						//errcodes[i] = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", L"**fail %s", pmi_error_str);
					}
					else
						errcodes[i] = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", NULL);
					PMI_Free_error_string(pmi_errcodes[i]);
				}
				else
					errcodes[i] = MPI_SUCCESS;
			}
		}
		else
		{
			for (i=0; i<total_num_processes; i++)
			{
				if (pmi_errcodes[i] != 0)
				{
					errors_found = 1;
					if (!quiet)
					{
						spawn_errno = pmi_errcodes[i];
					}
				}
			}
			if (spawn_errno != MPI_SUCCESS)
			{
				if (PMI_Get_error_stringW(spawn_errno, pmi_error_str, MPI_MAX_ERROR_STRING) == PMI_SUCCESS)
				{
					spawn_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", L"**fail %s", pmi_error_str);
					/* This doesn't work because the error stack code can't handle dynamic error codes
					spawn_errno = MPIR_Spawn_error_code[MPIR_Cur_spawn_error_code++];
					if (MPIR_Cur_spawn_error_code == MPIR_MAX_SPAWN_ERROR_CODES)
						MPIR_Cur_spawn_error_code = 0;
					NMPI_Add_error_string(mpi_errno, (spawn_errno, pmi_error_str));
					MPIU_ERR_CHK_POP(mpi_errno);
					*/
				}
				else
				{
					spawn_errno = MPI_SUCCESS; /* Ignore PMI error number that can't be converted into an MPI error code */
				}
			}
		}
		
		free_pmi_keyvals(info_keyval_vectors, count, info_keyval_sizes);
		MPIU_Free(info_keyval_sizes);
		MPIU_Free(pmi_errcodes);
	}

	MPIR_Bcast(&errors_found, 1, MPI_INT, root, comm_ptr);
	if (!errors_found)
	{
		mpi_errno = MPID_Comm_accept(port_name, NULL, root, comm_ptr, intercomm); 
		MPIU_ERR_CHK_POP(mpi_errno);
	}
	else
	{
		mpi_errno = MPIR_Err_create_code(spawn_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", NULL);
	}

 fn_exit:
	MPIR_Nest_decr();
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_COMM_SPAWN_MULTIPLE);
	return mpi_errno;
 fn_fail:
	goto fn_exit;
}

/* FIXME: What does this function do?  Who calls it?  Can we assume that
   it is called only dynamic process operations (specifically spawn) 
   are supported?  Do we need the concept of a port? For example,
   could a channel that supported only shared memory call this (it doesn't
   look like it right now, so this could go into util/sock, perhaps?
   
   It might make more sense to have this function provided as a function 
   pointer as part of the channel init setup, particularly since this
   function appears to access channel-specific storage (MPIDI_CH3_Process) */

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_Get_parent_port
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3_Get_parent_port(wchar_t ** parent_port)
{
	int mpi_errno = MPI_SUCCESS;
	wchar_t val[MPIDI_MAX_KVS_VALUE_LEN];
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_GET_PARENT_PORT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_GET_PARENT_PORT);

	if (parent_port_name == NULL)
	{
		wchar_t *kvsname = NULL;
		/* We can always use PMI_KVS_GetW on our own process group */
		mpi_errno = MPIDI_PG_GetConnKVSname( &kvsname );
		MPIU_ERR_CHK_POP(mpi_errno);
		mpi_errno = PMI_KVS_GetW(kvsname, PARENT_PORT_KVSKEY, val, MPIDI_MAX_KVS_VALUE_LEN);
		if (mpi_errno != PMI_SUCCESS)
		{
			MPIU_ERR_SETANDJUMP2(mpi_errno,MPI_ERR_OTHER, L"**pmi_kvs_get", L"**pmi_kvs_get %d %s", mpi_errno, PMI_Get_last_error_msgW());
		}

		parent_port_name = MPIU_Strdup(val);
		if (parent_port_name == NULL)
		{
			MPIU_ERR_POP(mpi_errno);
		}
	}

	*parent_port = parent_port_name;

 fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_GET_PARENT_PORT);
	return mpi_errno;
 fn_fail:
	goto fn_exit;
}

void MPIDI_CH3_FreeParentPort(void)
{
	if (parent_port_name)
	{
		MPIU_Free( parent_port_name );
		parent_port_name = NULL;
	}
}

/* Partial description: 
   This file contains the routines that are used to create socket connections,
   including the routines used to encode/decode the description of a connection
   into/out of the "business card".

   ToDo: change the "host description" to an "interface address" so that
   socket connections are targeted at particularly interfaces, not
   compute nodes, and that the address is in ready-to-use IP address format, 
   and does not require a gethostbyname lookup.  - Partially done
 */

/*
 * Routines for establishing a listener socket on the socket set that
 * is used for all communication.  These should be called from the
 * channel init and finalize routines.
 */

#undef FUNCNAME
#define FUNCNAME MPIDU_CH3I_SetupListener
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDU_CH3I_SetupListener( MPIDU_Sock_set_t sock_set )
{
	int mpi_errno = MPI_SUCCESS;
	MPIDU_Sock_t sock;
	MPIDI_STATE_DECL(MPID_STATE_MPIDU_CH3I_SETUPLISTENER);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDU_CH3I_SETUPLISTENER);

	mpi_errno = MPIDI_CH3I_Connection_alloc(&MPIDI_CH3I_listener_conn);
	MPIU_ERR_CHK_POP(mpi_errno);

	MPIU_DBG_MSG(CH3_CONNECT,TYPICAL,L"Setting listener connect state to CONN_STATE_LISTENING");
	MPIDI_CH3I_listener_conn->sock		  = NULL;
	MPIDI_CH3I_listener_conn->vc		  = NULL;
	MPIDI_CH3I_listener_conn->state		  = CONN_STATE_LISTENING;
	MPIDI_CH3I_listener_conn->send_active = NULL;
	MPIDI_CH3I_listener_conn->recv_active = NULL;
	
	mpi_errno = MPIDU_Sock_listen(sock_set, MPIDI_CH3I_listener_conn, &MPIDI_CH3I_listener_port, &sock);
	MPIU_ERR_CHK_POP(mpi_errno);

	MPIDI_CH3I_listener_conn->sock = sock;

fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDU_CH3I_SETUPLISTENER);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME MPIDU_CH3I_ShutdownListener
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDU_CH3I_ShutdownListener( void )
{
	int mpi_errno;
	MPID_Progress_state progress_state;
	MPIDI_STATE_DECL(MPID_STATE_MPIDU_CH3I_SHUTDOWNLISTENER);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDU_CH3I_SHUTDOWNLISTENER);

	mpi_errno = MPIDU_Sock_post_close(MPIDI_CH3I_listener_conn->sock);
	MPIU_ERR_CHK_POP(mpi_errno);

	MPID_Progress_start(&progress_state);
	while(MPIDI_CH3I_listener_conn != NULL)
	{
		mpi_errno = MPID_Progress_wait(&progress_state);
		MPIU_ERR_CHK_POP(mpi_errno);
		
	}
	MPID_Progress_end(&progress_state);

fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDU_CH3I_SHUTDOWNLISTENER);
	return mpi_errno;
}

/* Allocates a connection and the pg_id field for a connection only.
   Does not initialize any connection fields other than pg_id.
   Called by routines that create connections, used in this
   file and in ch3_progress*.c in various channels.
*/
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_Connection_alloc
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3I_Connection_alloc(MPIDI_CH3I_Connection_t ** connp)
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_CH3I_Connection_t * conn = NULL;
	int id_sz;
	MPIU_CHKPMEM_DECL(2);
	MPIDI_STATE_DECL(MPID_STATE_CONNECTION_ALLOC);

	MPIDI_FUNC_ENTER(MPID_STATE_CONNECTION_ALLOC);

	MPIU_CHKPMEM_MALLOC(conn, MPIDI_CH3I_Connection_t*, sizeof(MPIDI_CH3I_Connection_t), mpi_errno, L"conn");

	/* FIXME: This size is unchanging, so get it only once (at most); 
	   we might prefer for connections to simply point at the single process
	   group to which the remote process belong */
	mpi_errno = PMI_Get_id_length_max(&id_sz);
	if (mpi_errno != PMI_SUCCESS)
	{
		MPIU_ERR_SETANDJUMP2(mpi_errno,MPI_ERR_OTHER, L"**pmi_get_id_length_max", L"**pmi_get_id_length_max %d %s", mpi_errno, PMI_Get_last_error_msgW());
	}
	MPIU_CHKPMEM_MALLOC(conn->pg_id, wchar_t*, (id_sz + 1)*sizeof(wchar_t), mpi_errno, L"conn->pg_id");
	conn->pg_id[0] = 0;		   /* Be careful about pg_id in case a later error */
	*connp = conn;

  fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_CONNECTION_ALLOC);
	return mpi_errno;
  fn_fail:
	MPIU_CHKPMEM_REAP();
	goto fn_exit;
}

/* FIXME: Describe the algorithm for the connection logic */
#undef FUNCNAME
#define FUNCNAME  MPIDI_CH3I_Connect_to_root_sock
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3I_Connect_to_root_sock(const wchar_t * port_name, MPIDI_VC_t ** new_vc)
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_VC_t * vc;
	wchar_t host_description[MAX_HOST_DESCRIPTION_LEN];
	int port, port_name_tag;
	MPIDU_Sock_ifaddr_t ifaddr;
	int hasIfaddr = 0;
	MPIDI_CH3I_Connection_t * conn;
	MPIU_CHKPMEM_DECL(1);
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_CONNECT_TO_ROOT_SOCK);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_CONNECT_TO_ROOT_SOCK);

	MPIU_DBG_MSG_S(CH3_CONNECT,VERBOSE,L"Connect to root with portstring %s", port_name );

	/*wprintf(L"MPIDI_CH3I_Connect_to_root_sock\n");fflush(stdout);*/
	mpi_errno = MPIDU_Sock_get_conninfo_from_bc( port_name, host_description, MAX_HOST_DESCRIPTION_LEN, &port, &ifaddr, &hasIfaddr );
	if (mpi_errno)
	{
		MPIU_ERR_POP(mpi_errno);
	}
	mpi_errno = MPIDI_GetTagFromPort(port_name, &port_name_tag);
	if (mpi_errno != MPIU_STR_SUCCESS)
	{
		MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, L"**argstr_port_name_tag");
	}

	MPIU_DBG_MSG_D(CH3_CONNECT,VERBOSE,L"port tag %d",port_name_tag);

	MPIU_CHKPMEM_MALLOC(vc,MPIDI_VC_t *,sizeof(MPIDI_VC_t),mpi_errno,L"vc");
	/* FIXME - where does this vc get freed? */

	*new_vc = vc;

	/* FIXME: Note that these *must* always happen together, so they should be in a single routine */
	/* FIXME: There may need to be a THIRD routine here, to ensure that the
	   channel is initialized for this pair of process groups (this process
	   and the remote process to which the vc will connect). */
	MPIDI_VC_Init(vc, NULL, 0);
	MPIDI_CH3.VC_Init( vc );

	mpi_errno = MPIDI_CH3I_Connection_alloc(&conn);
	MPIU_ERR_CHK_POP(mpi_errno);

	/* conn->pg_id is not used for this conection */

	/* FIXME: To avoid this global (MPIDI_CH3I_sock_set) which is 
	   used only in ch3_progress.c and ch3_progress_connect.c in the channels,
	   this should be a call into the channel, asking it to setup the
	   socket for a connection and return the connection.  That will
	   keep the socket set out of the general ch3 code, even if this
	   is the socket utility functions. */
	MPIU_DBG_MSG_FMT(CH3_CONNECT,VERBOSE,(MPIU_DBG_FDEST, L"posting connect to host %s, port %d", host_description, port ));
	mpi_errno = MPIDU_Sock_post_connect(MPIDI_CH3I_sock_set, conn, host_description, port, &conn->sock);
	if (mpi_errno == MPI_SUCCESS)
	{
		vc->ch.sock = conn->sock;
		vc->ch.conn = conn;
		vc->ch.state = MPIDI_CH3I_VC_STATE_CONNECTING;
		conn->vc = vc;
		MPIU_DBG_CONNSTATECHANGE(vc, conn, CONN_STATE_CONNECT_ACCEPT);
		conn->state = CONN_STATE_CONNECT_ACCEPT;
		conn->send_active = NULL;
		conn->recv_active = NULL;

		/* place the port name tag in the pkt that will eventually be sent to the other side */
		conn->pkt.sc_conn_accept.port_name_tag = port_name_tag;
	}
	/* --BEGIN ERROR HANDLING-- */
	else
	{
		if (MPIR_ERR_GET_CLASS(mpi_errno) == MPIDU_SOCK_ERR_BAD_HOST)
		{ 
			mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|sock|badhost", L"**ch3|sock|badhost %s %d %s", conn->pg_id, conn->vc->pg_rank, port_name);
		}
		else if (MPIR_ERR_GET_CLASS(mpi_errno) == MPIDU_SOCK_ERR_CONN_FAILED)
		{ 
			mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|sock|connrefused", L"**ch3|sock|connrefused %s %d %s", conn->pg_id, conn->vc->pg_rank, port_name);
		}
		else
		{
			MPIU_ERR_POP(mpi_errno);
		}
		vc->ch.state = MPIDI_CH3I_VC_STATE_FAILED;
		MPIU_Free(conn);
		goto fn_fail;
	}
	/* --END ERROR HANDLING-- */

 fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_CONNECT_TO_ROOT_SOCK);
	return mpi_errno;
 fn_fail:
	MPIU_CHKPMEM_REAP();
	goto fn_exit;
}

/* ------------------------------------------------------------------------- */
/* Business card management.  These routines insert or extract connection
   information when using sockets from the business card */
/* ------------------------------------------------------------------------- */

/* FIXME: These are small routines; we may want to bring them together 
   into a more specific post-connection-for-sock */

/* The host_description should be of length MAX_HOST_DESCRIPTION_LEN */

#undef FUNCNAME
#define FUNCNAME MPIDU_Sock_get_conninfo_from_bc
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDU_Sock_get_conninfo_from_bc( const wchar_t *bc, 
									 wchar_t *host_description, int maxlen,
									 int *port, MPIDU_Sock_ifaddr_t *ifaddr, 
									 int *hasIfaddr )
{
	int mpi_errno = MPI_SUCCESS;
	int str_errno;
	MPIDI_STATE_DECL(MPID_STATE_MPIDU_SOCK_GET_CONNINFO_FROM_BC);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDU_SOCK_GET_CONNINFO_FROM_BC);

	/*wprintf(L"[%s]getting %s from <%s>\n", MPIU_GetStaticEnv(L"PMI_RANK"), MPIDI_CH3I_HOST_DESCRIPTION_KEY, bc);fflush(stdout);*/
	str_errno = MPIU_Str_get_string_arg(bc, MPIDI_CH3I_HOST_DESCRIPTION_KEY, host_description, maxlen);
	if (str_errno != MPIU_STR_SUCCESS)
	{
		/* --BEGIN ERROR HANDLING */
		if (str_errno == MPIU_STR_FAIL)
		{
			MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER,L"**argstr_missinghost");
		}
		else
		{
			/* MPIU_STR_TRUNCATED or MPIU_STR_NONEM */
			MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, L"**argstr_hostd");
		}
		/* --END ERROR HANDLING-- */
	}
	/*wprintf(L"got <%s>\n", host_description);fflush(stdout);*/
	/*wprintf(L"getting %s from <%s>\n", MPIDI_CH3I_PORT_KEY, bc);fflush(stdout);*/
	str_errno = MPIU_Str_get_int_arg(bc, MPIDI_CH3I_PORT_KEY, port);
	if (str_errno != MPIU_STR_SUCCESS)
	{
		/* --BEGIN ERROR HANDLING */
		if (str_errno == MPIU_STR_FAIL)
		{
			MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, L"**argstr_missingport");
		}
		else
		{
			/* MPIU_STR_TRUNCATED or MPIU_STR_NONEM */
			MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, L"**argstr_port");
		}
		/* --END ERROR HANDLING-- */
	}
	/*wprintf(L"got %d\n", *port);fflush(stdout);*/
	/* ifname is optional */
	/* FIXME: This is a hack to allow Windows to continue to use
	   the host description string instead of the interface address
	   bytes when posting a socket connection.  This should be fixed 
	   by changing the Sock_post_connect to only accept interface
	   address.  Note also that Windows does not have the inet_pton 
	   routine; the Windows version of this routine will need to 
	   be identified or written.  See also channels/sock/ch3_progress.c and
	   channels/ssm/ch3_progress_connect.c */
	*hasIfaddr = 0;
	MPIU_UNREFERENCED_ARG(ifaddr);
	
 fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDU_SOCK_GET_CONNINFO_FROM_BC);
	return mpi_errno;
 fn_fail:
	goto fn_exit;
}


/*  MPIDI_CH3U_Get_business_card_sock - does socket specific portion of 
 *  setting up a business card
 *  
 *  Parameters:
 *	 bc_val_p	 - business card value buffer pointer, updated to the next 
 *					available location or freed if published.
 *	 val_max_sz_p - ptr to maximum value buffer size reduced by the number 
 *					of characters written
 *							   
 */

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3U_Get_business_card_sock
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3U_Get_business_card_sock(wchar_t **bc_val_p, int *val_max_sz_p)
{
	int mpi_errno = MPI_SUCCESS;
	int port;
	wchar_t host_description[MAX_HOST_DESCRIPTION_LEN];
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3U_GET_BUSINESS_CARD_SOCK);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3U_GET_BUSINESS_CARD_SOCK);

	mpi_errno = MPIDU_Sock_get_host_description(host_description, MAX_HOST_DESCRIPTION_LEN);
	if (mpi_errno != MPI_SUCCESS)
	{
		MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, L"**init_description");
	}

	port = MPIDI_CH3I_listener_port;
	mpi_errno = MPIU_Str_add_int_arg(bc_val_p, val_max_sz_p, MPIDI_CH3I_PORT_KEY, port);
	/* --BEGIN ERROR HANDLING-- */
	if (mpi_errno != MPIU_STR_SUCCESS)
	{
		if (mpi_errno == MPIU_STR_NOMEM)
		{
			MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, L"**buscard_len");
		}
		else
		{
			MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, L"**buscard");
		}
	}
	/* --END ERROR HANDLING-- */
	
	mpi_errno = MPIU_Str_add_string_arg(bc_val_p, val_max_sz_p, MPIDI_CH3I_HOST_DESCRIPTION_KEY, host_description);
	/* --BEGIN ERROR HANDLING-- */
	if (mpi_errno != MPIU_STR_SUCCESS)
	{
		if (mpi_errno == MPIU_STR_NOMEM)
		{
			MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, L"**buscard_len");
		}
		else
		{
			MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, L"**buscard");
		}
		MPIU_UNREACHABLE_CODE();
		/*return mpi_errno;*/ /* Unreachable code */
	}
	/* --END ERROR HANDLING-- */

 fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3U_GET_BUSINESS_CARD_SOCK);
	return mpi_errno;
 fn_fail:
	goto fn_exit;
}

/* ------------------------------------------------------------------------- */
/* Below will be/is the code that is used to create a connection and
 * to handle changes to the state of a connection.  
 */
/* ------------------------------------------------------------------------- */

/* This routine is called in response to an MPIDU_SOCK_OP_ACCEPT event 
   in ch3_progress */
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_Sockconn_handle_accept_event
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3_Sockconn_handle_accept_event( void )
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_CH3I_Connection_t * conn;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_SOCKCONN_HANDLE_ACCEPT_EVENT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_SOCKCONN_HANDLE_ACCEPT_EVENT);
	
	mpi_errno = MPIDI_CH3I_Connection_alloc(&conn);
	MPIU_ERR_CHK_POP(mpi_errno);
	mpi_errno = MPIDU_Sock_accept(MPIDI_CH3I_listener_conn->sock, MPIDI_CH3I_sock_set, conn, &conn->sock);
	if (mpi_errno != MPI_SUCCESS)
	{
		MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, L"**ch3|sock|accept");
	}
	
	conn->vc = NULL;
	MPIU_DBG_CONNSTATECHANGE(conn->vc, conn, CONN_STATE_OPEN_LRECV_PKT);
	conn->state = CONN_STATE_OPEN_LRECV_PKT;
	conn->send_active = NULL;
	conn->recv_active = NULL;
	
	mpi_errno = connection_post_recv_pkt(conn);
	MPIU_ERR_CHK_POP(mpi_errno);

 fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_SOCKCONN_HANDLE_ACCEPT_EVENT);

	return mpi_errno;
 fn_fail:
	goto fn_exit;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_Sockconn_handle_connect_event
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3_Sockconn_handle_connect_event( MPIDI_CH3I_Connection_t *conn, int event_error )
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_SOCKCONN_HANDLE_CONNECT_EVENT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_SOCKCONN_HANDLE_CONNECT_EVENT);
	
	/* --BEGIN ERROR HANDLING-- */
	if (event_error != MPI_SUCCESS)
	{
		/* If the connection fails, conn->vc etc is probably invalid,
		   so we can only report that the connection failed */
		mpi_errno = event_error;
		MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER,L"**ch3|sock|connfailed" );
	}
	/* --END ERROR HANDLING-- */

	if (conn->state == CONN_STATE_CONNECTING)
	{
		MPIU_DBG_CONNSTATECHANGE(conn->vc, conn, CONN_STATE_OPEN_CSEND);
		conn->state = CONN_STATE_OPEN_CSEND;
		MPIDI_Pkt_init(&conn->pkt, MPIDI_CH3I_PKT_SC_OPEN_REQ);
		conn->pkt.sc_open_req.pg_id_len = (int) (wcslen(MPIDI_Process.my_pg->id) + 1)*sizeof(wchar_t);
		conn->pkt.sc_open_req.pg_rank = MPIR_Process.comm_world->rank;
		
		mpi_errno = connection_post_send_pkt_and_pgid(conn);
		if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }
	}
	else
	{
		/* CONN_STATE_CONNECT_ACCEPT */
		int port_name_tag;

		MPIU_Assert(conn->state == CONN_STATE_CONNECT_ACCEPT);
		MPIU_DBG_CONNSTATECHANGE(conn->vc, conn, CONN_STATE_OPEN_CSEND);
		conn->state = CONN_STATE_OPEN_CSEND;
		
		/* pkt contains port name tag. In memory debugging mode, 
		   MPIDI_Pkt_init resets the packet contents. Therefore,
		   save the port name tag and then add it back. */
		port_name_tag = conn->pkt.sc_conn_accept.port_name_tag;
		
		MPIDI_Pkt_init(&conn->pkt, MPIDI_CH3I_PKT_SC_CONN_ACCEPT);
		
		conn->pkt.sc_conn_accept.port_name_tag = port_name_tag;
		
		mpi_errno = connection_post_send_pkt(conn);
		if (mpi_errno != MPI_SUCCESS)
		{
			MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_INTERN, L"**ch3|sock|scconnaccept");
		}
	}

 fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_SOCKCONN_HANDLE_CONNECT_EVENT);
	return mpi_errno;
 fn_fail:
	goto fn_exit;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_Sockconn_handle_close_event
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3_Sockconn_handle_close_event( MPIDI_CH3I_Connection_t * conn )
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_SOCKCONN_HANDLE_CLOSE_EVENT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_SOCKCONN_HANDLE_CLOSE_EVENT);
				
	/* If the conn pointer is NULL then the close was intentional */
	/* FIXME: What does the above comment mean? */
	if (conn != NULL)
	{
		if (conn->state == CONN_STATE_CLOSING)
		{
			MPIU_Assert(conn->send_active == NULL);
			MPIU_Assert(conn->recv_active == NULL);
			if (conn->vc != NULL)
			{
				MPIU_DBG_VCCHSTATECHANGE(conn->vc, VC_STATE_UNCONNECTED);
				conn->vc->ch.state = MPIDI_CH3I_VC_STATE_UNCONNECTED;
				conn->vc->ch.sock  = MPIDU_SOCK_INVALID_SOCK;
				/* FIXME: Make sure that this is the correct state for the vc */
				/* conn->vc->state	= MPIDI_VC_STATE_INACTIVE; */
				/* Handle_connection takes care of updating the state on the VC */
				mpi_errno = MPIDI_CH3U_Handle_connection(conn->vc, MPIDI_VC_EVENT_TERMINATED);
				MPIU_ERR_CHK_POP(mpi_errno);
			}
		}
		else
		{
			MPIU_Assert(conn->state == CONN_STATE_LISTENING);
			MPIDI_CH3I_listener_conn = NULL;
			MPIDI_CH3I_listener_port = 0;
			
			MPIDI_CH3_Progress_signal_completion();
			/* FIXME: Why is this commented out? */
			/* MPIDI_CH3I_progress_completion_count++; */
		}
				
		conn->sock = MPIDU_SOCK_INVALID_SOCK;
		MPIU_DBG_CONNSTATECHANGE(conn->vc, conn, CONN_STATE_CLOSED);
		conn->state = CONN_STATE_CLOSED;
		if (conn->vc)
		{
			/* This step is important; without this, test
			   disconnect_reconnect fails because the vc->ch.conn 
			   connection will continue to be used, even though
			   the memory has been freed */
			if (conn->vc->ch.conn == conn) conn->vc->ch.conn = 0;
			/* FIXME: If this isn't the associated connection, 
			   there may be a problem */
		}
		connection_destroy(conn); 
	}
 fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_SOCKCONN_HANDLE_CLOSE_EVENT);
	return mpi_errno;
 fn_fail:
	goto fn_exit;
}

/* Cycle through the connection setup states */
/* FIXME: separate out the accept and connect sides to make it easier to follow the logic */
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_Sockconn_handle_conn_event
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3_Sockconn_handle_conn_event( MPIDI_CH3I_Connection_t * conn )
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_SOCKCONN_HANDLE_CONN_EVENT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_SOCKCONN_HANDLE_CONN_EVENT);

	/* FIXME: Is there an assumption about conn->state? */

	if (conn->pkt.type == MPIDI_CH3I_PKT_SC_OPEN_REQ)
	{
		/* Answer to fixme: it appears from the control flow that this is
		   the required state) */
		MPIU_Assert( conn->state == CONN_STATE_OPEN_LRECV_PKT);
		MPIU_DBG_CONNSTATECHANGE(conn->vc, conn, CONN_STATE_OPEN_LRECV_DATA);
		conn->state = CONN_STATE_OPEN_LRECV_DATA;
		mpi_errno = MPIDU_Sock_post_read(conn->sock, conn->pg_id, conn->pkt.sc_open_req.pg_id_len, conn->pkt.sc_open_req.pg_id_len, NULL);   
		MPIU_ERR_CHK_POP(mpi_errno);
	}
	else if (conn->pkt.type == MPIDI_CH3I_PKT_SC_CONN_ACCEPT)
	{
		MPIDI_VC_t *vc; 
		int port_name_tag;
		
		vc = (MPIDI_VC_t *) MPIU_Malloc(sizeof(MPIDI_VC_t));
		/* --BEGIN ERROR HANDLING-- */
		if (vc == NULL)
		{
			mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", NULL);
			goto fn_fail;
		}
		/* --END ERROR HANDLING-- */
		/* FIXME - where does this vc get freed? */

		/* FIXME: There should be a single VC init function; this should
		   invoke a channel-specific function to initialize channel-specific items */
		MPIDI_VC_Init(vc, NULL, 0);
		vc->ch.sendq_head = NULL;
		vc->ch.sendq_tail = NULL;
		MPIU_DBG_VCCHSTATECHANGE(vc,VC_STATE_CONNECTING);
		vc->ch.state = MPIDI_CH3I_VC_STATE_CONNECTING;
		vc->ch.sock = conn->sock;
		vc->ch.conn = conn;
		conn->vc = vc;
		port_name_tag = conn->pkt.sc_conn_accept.port_name_tag;

		MPIDI_Pkt_init(&conn->pkt, MPIDI_CH3I_PKT_SC_OPEN_RESP);
		conn->pkt.sc_open_resp.ack = TRUE;
		
		/* FIXME: Possible ambiguous state (two ways to get to OPEN_LSEND) */
		MPIU_DBG_CONNSTATECHANGE(conn->vc, conn, CONN_STATE_OPEN_LSEND);
		conn->state = CONN_STATE_OPEN_LSEND;
		mpi_errno = connection_post_send_pkt(conn);
		if (mpi_errno != MPI_SUCCESS)
		{
			MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_INTERN, L"**ch3|sock|scconnaccept");
		}
		
		/* ENQUEUE vc */
		MPIDI_CH3I_Acceptq_enqueue(vc, port_name_tag);

	}
	else if (conn->pkt.type == MPIDI_CH3I_PKT_SC_OPEN_RESP)
	{
		/* FIXME: is this the correct assert? */
		MPIU_Assert( conn->state == CONN_STATE_OPEN_CRECV );
		if (conn->pkt.sc_open_resp.ack)
		{
			MPIU_DBG_CONNSTATECHANGE(conn->vc, conn, CONN_STATE_CONNECTED);
			conn->state = CONN_STATE_CONNECTED;
			conn->vc->ch.state = MPIDI_CH3I_VC_STATE_CONNECTED;
			MPIU_Assert(conn->vc->ch.conn == conn);
			MPIU_Assert(conn->vc->ch.sock == conn->sock);
			
			mpi_errno = connection_post_recv_pkt(conn);
			MPIU_ERR_CHK_POP(mpi_errno);
			mpi_errno = connection_post_sendq_req(conn);
			if (mpi_errno != MPI_SUCCESS)
			{
				MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_INTERN, L"**ch3|sock|scopenresp");
			}
		}
		else
		{
			/* FIXME: Should conn->vc be freed? Who allocated? Why not? */
			/* FIXME: Should probably reduce ref count on conn->vc */
			/* FIXME: What happens to the state of the associated VC? 
			   Why isn't it changed?  Is there an assert here, 
			   such as conn->vc->conn != conn (there is another connection 
			   chosen for the vc)? */
			/* MPIU_Assert( conn->vc->ch.conn != conn ); */
			/* Set the candidate vc for this connection to NULL (we
			   are discarding this connection because (I think) we
			   are performing a head-to-head connection, and this
			   connection is being rejected in favor of the connection
			   from the other side. */
			if (conn->vc->ch.conn == conn) conn->vc->ch.conn = NULL;
			conn->vc = NULL;
			MPIU_DBG_CONNSTATECHANGE_MSG(conn->vc, conn, CONN_STATE_CLOSING, L"because ack on OPEN_CRECV was false");
			conn->state = CONN_STATE_CLOSING;
			/* FIXME: What does post close do here? */
			MPIDU_Sock_post_close(conn->sock);
		}
	}
	/* --BEGIN ERROR HANDLING-- */
	else
	{
		MPIU_DBG_STMT(CH3_CONNECT,VERBOSE,MPIDI_DBG_Print_packet(&conn->pkt));
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_INTERN, L"**ch3|sock|badpacket", L"**ch3|sock|badpacket %d", conn->pkt.type);
		goto fn_fail;
	}
	/* --END ERROR HANDLING-- */


 fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_SOCKCONN_HANDLE_CONN_EVENT);
	return mpi_errno;
 fn_fail:
	goto fn_exit;
}

/* FIXME: This should really be combined with handle_conn_event */
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_Sockconn_handle_connopen_event
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3_Sockconn_handle_connopen_event( MPIDI_CH3I_Connection_t * conn )
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_PG_t * pg;
	int pg_rank;
	MPIDI_VC_t * vc;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_SOCKCONN_HANDLE_CONNOPEN_EVENT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_SOCKCONN_HANDLE_CONNOPEN_EVENT);

	/* Look up pg based on conn->pg_id */
	mpi_errno = MPIDI_PG_Find(conn->pg_id, &pg);
	if (pg == NULL)
	{
		MPIU_ERR_SETANDJUMP1(mpi_errno,MPI_ERR_OTHER, L"**pglookup", L"**pglookup %s", conn->pg_id);
	}
	
	pg_rank = conn->pkt.sc_open_req.pg_rank;
	MPIDI_PG_Get_vc(pg, pg_rank, &vc);
	MPIU_Assert(vc->pg_rank == pg_rank);
	
	if (vc->ch.conn == NULL)
	{
		/* no head-to-head connects, accept the connection */
		MPIU_DBG_VCCHSTATECHANGE(vc, VC_STATE_CONNECTING);
		vc->ch.state = MPIDI_CH3I_VC_STATE_CONNECTING;
		vc->ch.sock = conn->sock;
		vc->ch.conn = conn;
		conn->vc = vc;
		
		MPIDI_Pkt_init(&conn->pkt, MPIDI_CH3I_PKT_SC_OPEN_RESP);
		conn->pkt.sc_open_resp.ack = TRUE;
	}
	else
	{
		/* head to head situation */
		if (pg == MPIDI_Process.my_pg)
		{
			/* the other process is in the same comm_world; just compare the ranks */
			if (MPIR_Process.comm_world->rank < pg_rank)
			{
				/* accept connection */
				MPIU_DBG_VCCHSTATECHANGE(vc, VC_STATE_CONNECTING);
				vc->ch.state = MPIDI_CH3I_VC_STATE_CONNECTING;
				vc->ch.sock = conn->sock;
				vc->ch.conn = conn;
				conn->vc = vc;
				
				MPIDI_Pkt_init(&conn->pkt, MPIDI_CH3I_PKT_SC_OPEN_RESP);
				conn->pkt.sc_open_resp.ack = TRUE;
			}
			else
			{
				/* refuse connection */
				MPIU_DBG_MSG_FMT(CH3_CONNECT, TYPICAL, (MPIU_DBG_FDEST,
				L"vc=%p,conn=%p:Refuse head-to-head connection (my process group)",vc,conn));
				MPIDI_Pkt_init(&conn->pkt, MPIDI_CH3I_PKT_SC_OPEN_RESP);
				conn->pkt.sc_open_resp.ack = FALSE;
			}
		}
		else
		{
			/* the two processes are in different comm_worlds; compare their unique pg_ids. */
			if (wcscmp(MPIDI_Process.my_pg->id, pg->id) < 0)
			{
				/* accept connection */
				MPIU_DBG_VCCHSTATECHANGE(vc, VC_STATE_CONNECTING);
				vc->ch.state = MPIDI_CH3I_VC_STATE_CONNECTING;
				vc->ch.sock = conn->sock;
				vc->ch.conn = conn;
				conn->vc = vc;
				
				MPIDI_Pkt_init(&conn->pkt, MPIDI_CH3I_PKT_SC_OPEN_RESP);
				conn->pkt.sc_open_resp.ack = TRUE;
			}
			else
			{
				/* refuse connection */
				MPIU_DBG_MSG_FMT(CH3_CONNECT, TYPICAL, (MPIU_DBG_FDEST, L"vc=%p,conn=%p:Refuse head-to-head connection (two process groups)",vc,conn));
				MPIDI_Pkt_init(&conn->pkt, MPIDI_CH3I_PKT_SC_OPEN_RESP);
				conn->pkt.sc_open_resp.ack = FALSE;
			}
		}
	}
	
	MPIU_DBG_CONNSTATECHANGE(conn->vc, conn, CONN_STATE_OPEN_LSEND);
	conn->state = CONN_STATE_OPEN_LSEND;
	mpi_errno = connection_post_send_pkt(conn);
	if (mpi_errno != MPI_SUCCESS)
	{
		MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_INTERN, L"**ch3|sock|open_lrecv_data");
	}

 fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_SOCKCONN_HANDLE_CONNOPEN_EVENT);
	return mpi_errno;
 fn_fail:
	goto fn_exit;
}

/* FIXME: This routine is called when?  What is valid in conn? */
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_Sockconn_handle_connwrite
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3_Sockconn_handle_connwrite( MPIDI_CH3I_Connection_t * conn )
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_SOCKCONN_HANDLE_CONNWRITE);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_SOCKCONN_HANDLE_CONNWRITE);

	if (conn->state == CONN_STATE_OPEN_CSEND)
	{
		/* finished sending open request packet */
		/* post receive for open response packet */
		MPIU_DBG_CONNSTATECHANGE(conn->vc, conn, CONN_STATE_OPEN_CRECV);
		conn->state = CONN_STATE_OPEN_CRECV;
		mpi_errno = connection_post_recv_pkt(conn);
		MPIU_ERR_CHK_POP(mpi_errno);
	}
	else if (conn->state == CONN_STATE_OPEN_LSEND)
	{
		/* finished sending open response packet */
		if (conn->pkt.sc_open_resp.ack == TRUE)
		{
			/* post receive for packet header */
			MPIU_DBG_CONNSTATECHANGE(conn->vc, conn, CONN_STATE_CONNECTED);
			conn->state = CONN_STATE_CONNECTED;
			MPIU_DBG_VCCHSTATECHANGE(conn->vc, VC_STATE_CONNECTED);
			conn->vc->ch.state = MPIDI_CH3I_VC_STATE_CONNECTED;
			mpi_errno = connection_post_recv_pkt(conn);
			MPIU_ERR_CHK_POP(mpi_errno);
			
			mpi_errno = connection_post_sendq_req(conn);
			if (mpi_errno != MPI_SUCCESS)
			{
				MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_INTERN, L"**ch3|sock|openlsend");
			}
		}
		else
		{
			/* head-to-head connections - close this connection */
			MPIU_DBG_CONNSTATECHANGE(conn->vc, conn, CONN_STATE_CLOSING);
			/* FIXME: the connect side of this sets conn->vc to NULL. Why is
			   this different? The code that checks CONN_STATE_CLOSING uses
			   conn == NULL to identify intentional close, which this 
			   appears to be. */
			conn->state = CONN_STATE_CLOSING;
			mpi_errno = MPIDU_Sock_post_close(conn->sock);
			if (mpi_errno != MPI_SUCCESS)
			{
				MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, L"**sock_post_close");
			}
		}
	}

 fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_SOCKCONN_HANDLE_CONNWRITE);
	return mpi_errno;
 fn_fail:
	goto fn_exit;
}

/* ----------------------------------------------------------------------- */
/* FIXME: What does this do? */
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_VC_post_sockconnect
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3I_VC_post_sockconnect(MPIDI_VC_t * vc)
{
	int mpi_errno = MPI_SUCCESS;
	wchar_t val[MPIDI_MAX_KVS_VALUE_LEN];
	wchar_t host_description[MAX_HOST_DESCRIPTION_LEN];
	int port;
	MPIDU_Sock_ifaddr_t ifaddr;
	int hasIfaddr = 0;
	MPIDI_CH3I_Connection_t * conn = 0;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_VC_POST_SOCKCONNECT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_VC_POST_SOCKCONNECT);
	
	MPIU_Assert(vc->ch.state == MPIDI_CH3I_VC_STATE_UNCONNECTED);
	
	MPIU_DBG_VCCHSTATECHANGE(vc, VC_STATE_CONNECTING);
	vc->ch.state = MPIDI_CH3I_VC_STATE_CONNECTING;

	mpi_errno = MPIDI_PG_GetConnString( vc->pg, vc->pg_rank, val, MPIDI_MAX_KVS_VALUE_LEN);
	MPIU_ERR_CHK_POP(mpi_errno);
	/*wprintf(L"MPIDI_CH3I_VC_post_sockconnect\n");fflush(stdout);*/
	mpi_errno = MPIDU_Sock_get_conninfo_from_bc( val, host_description, MAX_HOST_DESCRIPTION_LEN, &port, &ifaddr, &hasIfaddr );
	if (mpi_errno)
	{
		MPIU_ERR_POP(mpi_errno);
	}

	mpi_errno = MPIDI_CH3I_Connection_alloc(&conn);
	if (mpi_errno == MPI_SUCCESS)
	{
		/* FIXME: This is a hack to allow Windows to continue to use
		   the host description string instead of the interface address
		   bytes when posting a socket connection.  This should be fixed 
		   by changing the Sock_post_connect to only accept interface
		   address.  See also channels/ssm/ch3_progress_connect.c */
		{
			/*wprintf(L"posting connect to <%s> on port %d\n", host_description, port);fflush(stdout);*/
			mpi_errno = MPIDU_Sock_post_connect(MPIDI_CH3I_sock_set, conn, host_description, port, &conn->sock);
		}
		if (mpi_errno == MPI_SUCCESS)
		{
			MPIU_DBG_CONNSTATECHANGE(vc, conn, CONN_STATE_CONNECTING);
			vc->ch.sock = conn->sock;
			vc->ch.conn = conn;
			conn->vc = vc;
			conn->state = CONN_STATE_CONNECTING;
			conn->send_active = NULL;
			conn->recv_active = NULL;
		}
		/* --BEGIN ERROR HANDLING-- */
		else
		{
			MPIU_DBG_VCCHSTATECHANGE(vc, VC_STATE_FAILED);
			vc->ch.state = MPIDI_CH3I_VC_STATE_FAILED;
			mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|sock|postconnect", L"**ch3|sock|postconnect %d %d %s", MPIR_Process.comm_world->rank, vc->pg_rank, val);
			goto fn_fail;
		}
		/* --END ERROR HANDLING-- */
	}
	else
	{
		MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, L"**ch3|sock|connalloc");
	}

  fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_VC_POST_SOCKCONNECT);
	return mpi_errno;
 fn_fail:
	/* --BEGIN ERROR HANDLING-- */
	if (conn)
	{
		connection_destroy(conn);
	}
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}
/* end MPIDI_CH3I_VC_post_sockconnect() */


/* FIXME: What does this do? */
/* Guess: Setup a wait-to-read on the socket that was set after the accept 
   was handled */
/* Wrong guess.  */
#undef FUNCNAME
#define FUNCNAME connection_post_recv_pkt
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int connection_post_recv_pkt(MPIDI_CH3I_Connection_t * conn)
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_CONNECTION_POST_RECV_PKT);

	MPIDI_FUNC_ENTER(MPID_STATE_CONNECTION_POST_RECV_PKT);

	mpi_errno = MPIDU_Sock_post_read(conn->sock, &conn->pkt, sizeof(conn->pkt), sizeof(conn->pkt), NULL);
	MPIU_ERR_CHK_POP(mpi_errno);

 fn_fail:	
	MPIDI_FUNC_EXIT(MPID_STATE_CONNECTION_POST_RECV_PKT);
	return mpi_errno;
}


#undef FUNCNAME
#define FUNCNAME connection_post_send_pkt
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int connection_post_send_pkt(MPIDI_CH3I_Connection_t * conn)
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_CONNECTION_POST_SEND_PKT);

	MPIDI_FUNC_ENTER(MPID_STATE_CONNECTION_POST_SEND_PKT);
 
	MPIU_DBG_PKT(conn,&conn->pkt,"connect");
	mpi_errno = MPIDU_Sock_post_write(conn->sock, &conn->pkt, sizeof(conn->pkt), sizeof(conn->pkt), NULL);
	MPIU_ERR_CHK_POP(mpi_errno);
	
 fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_CONNECTION_POST_SEND_PKT);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME connection_post_send_pkt_and_pgid
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int connection_post_send_pkt_and_pgid(MPIDI_CH3I_Connection_t * conn)
{
	int mpi_errno;
	MPIDI_STATE_DECL(MPID_STATE_CONNECTION_POST_SEND_PKT_AND_PGID);

	MPIDI_FUNC_ENTER(MPID_STATE_CONNECTION_POST_SEND_PKT_AND_PGID);
	
	conn->iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST) &conn->pkt;
	conn->iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST) sizeof(conn->pkt);

	conn->iov[1].MPID_IOV_BUF = (MPID_IOV_BUF_CAST) MPIDI_Process.my_pg->id;
	conn->iov[1].MPID_IOV_LEN = (MPID_IOV_LEN_CAST) ((wcslen(MPIDI_Process.my_pg->id) + 1)*sizeof(wchar_t));

	MPIU_DBG_PKT(conn,&conn->pkt,"connect-pgid");
	mpi_errno = MPIDU_Sock_post_writev(conn->sock, conn->iov, 2, NULL);
	MPIU_ERR_CHK_POP(mpi_errno);
	
 fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_CONNECTION_POST_SEND_PKT_AND_PGID);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME connection_post_sendq_req
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int connection_post_sendq_req(MPIDI_CH3I_Connection_t * conn)
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_CONNECTION_POST_SENDQ_REQ);

	MPIDI_FUNC_ENTER(MPID_STATE_CONNECTION_POST_SENDQ_REQ);
	/* post send of next request on the send queue */
	conn->send_active = MPIDI_CH3I_SendQ_head(conn->vc); /* MT */
	if (conn->send_active != NULL)
	{
		MPIU_DBG_MSG_P(CH3,TYPICAL,"conn=%p: Posting message from connection send queue", conn );
		mpi_errno = MPIDU_Sock_post_writev(conn->sock, conn->send_active->dev.iov, conn->send_active->dev.iov_count, NULL);
		MPIU_ERR_CHK_POP(mpi_errno);
	}
	
 fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_CONNECTION_POST_SENDQ_REQ);
	return mpi_errno;
}


/* This routine frees all of the memory associated with a connection.
   It is named destroy instead of free because routines with name "free" 
   should have MPI semantics - free means to 
   decrement reference count and free if reference count is zero */
#undef FUNCNAME
#define FUNCNAME connection_destroy
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static void connection_destroy(MPIDI_CH3I_Connection_t * conn)
{
	MPIDI_STATE_DECL(MPID_STATE_CONNECTION_DESTROY);

	MPIDI_FUNC_ENTER(MPID_STATE_CONNECTION_DESTROY);

	MPIU_Free(conn->pg_id);
	MPIU_Free(conn);
	
	MPIDI_FUNC_EXIT(MPID_STATE_CONNECTION_DESTROY);
}


#ifdef USE_DBG_LOGGING
const wchar_t * MPIDI_CH3_VC_GetStateString( int state )
{
	const wchar_t *name = L"unknown";
	static wchar_t asdigits[20];
	
	switch (state) {
	case MPIDI_CH3I_VC_STATE_UNCONNECTED: name = L"CH3I_VC_STATE_UNCONNECTED"; break;
	case MPIDI_CH3I_VC_STATE_CONNECTING:  name = L"CH3I_VC_STATE_CONNECTING"; break;
	case MPIDI_CH3I_VC_STATE_CONNECTED:   name = L"CH3I_VC_STATE_CONNECTED"; break;
	case MPIDI_CH3I_VC_STATE_FAILED:	  name = L"CH3I_VC_STATE_FAILED"; break;
	default:
		MPIU_Snprintf( asdigits, sizeof(asdigits)/sizeof(wchar_t), _TRUNCATE, L"%d", state );
		asdigits[20-1] = 0;
		name = (const wchar_t *)asdigits;
	}
	return name;
}
#endif

static void PrintNamedDatatype(MPI_Datatype datatype)
{
	switch (datatype)
	{
		case MPI_CHAR:
			printf("MPI_CHAR\n");
			break;
		case MPI_SIGNED_CHAR:
			printf("MPI_SIGNED_CHAR\n");
			break;
		case MPI_UNSIGNED_CHAR:
			printf("MPI_UNSIGNED_CHAR\n");
			break;
		case MPI_BYTE:
			printf("MPI_BYTE\n");
			break;
		case MPI_WCHAR:
			printf("MPI_WCHAR\n");
			break;
		case MPI_SHORT:
			printf("MPI_SHORT\n");
			break;
		case MPI_UNSIGNED_SHORT:
			printf("MPI_UNSIGNED_SHORT\n");
			break;
		case MPI_INT:
			printf("MPI_INT\n");
			break;
		case MPI_UNSIGNED:
			printf("MPI_UNSIGNED\n");
			break;
		case MPI_LONG:
			printf("MPI_LONG\n");
			break;
		case MPI_UNSIGNED_LONG:
			printf("MPI_UNSIGNED_LONG\n");
			break;
		case MPI_FLOAT:
			printf("MPI_FLOAT\n");
			break;
		case MPI_DOUBLE:
			printf("MPI_DOUBLE\n");
			break;
		case MPI_LONG_DOUBLE:
			printf("MPI_LONG_DOUBLE\n");
			break;
		case MPI_LONG_LONG_INT:
			printf("MPI_LONG_LONG_INT\n");
			break;
		case MPI_UNSIGNED_LONG_LONG:
			printf("MPI_UNSIGNED_LONG_LONG\n");
			break;
			/*
		case MPI_LONG_LONG:
			printf("MPI_LONG_LONG\n");
			break;
			*/
		case MPI_PACKED:
			printf("MPI_PACKED\n");
			break;
		case MPI_LB:
			printf("MPI_LB\n");
			break;
		case MPI_UB:
			printf("MPI_UB\n");
			break;
		case MPI_FLOAT_INT:
			printf("MPI_FLOAT_INT\n");
			break;
		case MPI_DOUBLE_INT:
			printf("MPI_DOUBLE_INT\n");
			break;
		case MPI_LONG_INT:
			printf("MPI_LONG_INT\n");
			break;
		case MPI_SHORT_INT:
			printf("MPI_SHORT_INT\n");
			break;
		case MPI_2INT:
			printf("MPI_2INT\n");
			break;
		case MPI_LONG_DOUBLE_INT:
			printf("MPI_LONG_DOUBLE_INT\n");
			break;
		case MPI_COMPLEX:
			printf("MPI_COMPLEX\n");
			break;
		case MPI_DOUBLE_COMPLEX:
			printf("MPI_DOUBLE_COMPLEX\n");
			break;
		case MPI_LOGICAL:
			printf("MPI_LOGICAL\n");
			break;
		case MPI_REAL:
			printf("MPI_REAL\n");
			break;
		case MPI_DOUBLE_PRECISION:
			printf("MPI_DOUBLE_PRECISION\n");
			break;
		case MPI_INTEGER:
			printf("MPI_INTEGER\n");
			break;
		case MPI_2INTEGER:
			printf("MPI_2INTEGER\n");
			break;
		case MPI_2COMPLEX:
			printf("MPI_2COMPLEX\n");
			break;
		case MPI_2DOUBLE_COMPLEX:
			printf("MPI_2DOUBLE_COMPLEX\n");
			break;
		case MPI_2REAL:
			printf("MPI_2REAL\n");
			break;
		case MPI_2DOUBLE_PRECISION:
			printf("MPI_2DOUBLE_PRECISION\n");
			break;
		case MPI_CHARACTER:
			printf("MPI_CHARACTER\n");
			break;
		case MPI_REAL4:
			printf("MPI_REAL4\n");
			break;
		case MPI_REAL8:
			printf("MPI_REAL8\n");
			break;
		case MPI_REAL16:
			printf("MPI_REAL16\n");
			break;
		case MPI_COMPLEX8:
			printf("MPI_COMPLEX8\n");
			break;
		case MPI_COMPLEX16:
			printf("MPI_COMPLEX16\n");
			break;
		case MPI_COMPLEX32:
			printf("MPI_COMPLEX32\n");
			break;
		case MPI_INTEGER1:
			printf("MPI_INTEGER1\n");
			break;
		case MPI_INTEGER2:
			printf("MPI_INTEGER2\n");
			break;
		case MPI_INTEGER4:
			printf("MPI_INTEGER4\n");
			break;
		case MPI_INTEGER8:
			printf("MPI_INTEGER8\n");
			break;
		case MPI_INTEGER16:
			printf("MPI_INTEGER16\n");
			break;
		default:
			printf("unknown named datatype\n");
			break;
	}
}

static int PrintDatatype(MPI_Datatype datatype)
{
	int mpi_errno;
	int *array_of_ints, twoints[2];
	MPI_Aint *array_of_adds, a, twoaints[2];
	MPI_Datatype *array_of_dtypes;
	int num_ints, num_adds, num_dtypes, combiner;
	int i, count;

	NMPI_Type_get_envelope(mpi_errno, (datatype, &num_ints, &num_adds, &num_dtypes, &combiner));

	switch (combiner)
	{
	case MPI_COMBINER_NAMED:
		PrintNamedDatatype(datatype);
		fflush(stdout);
		return 0;
	case MPI_COMBINER_DUP:
		printf("Dupped:\n");
		NMPI_Type_get_contents(mpi_errno, (datatype, 0, 0, 1, NULL, NULL, &datatype));
		if (PrintDatatype(datatype))
			NMPI_Type_free(mpi_errno, (&datatype));
		break;
	case MPI_COMBINER_CONTIGUOUS:
		printf("Contiguous:\n");
		NMPI_Type_get_contents(mpi_errno, (datatype, 1, 0, 1, &i, NULL, &datatype));
		printf("count %d, type:\n", i);
		if (PrintDatatype(datatype))
			NMPI_Type_free(mpi_errno, (&datatype));
		break;
	case MPI_COMBINER_VECTOR:
		printf("Vector:\n");
		array_of_ints = (int*)MPIU_Malloc(3 * sizeof(int));
		NMPI_Type_get_contents(mpi_errno, (datatype, 3, 0, 1, array_of_ints, NULL, &datatype));
		printf("count %d, blocklength %d, stride %d, type:\n", array_of_ints[0], array_of_ints[1], array_of_ints[2], &datatype);
		if (PrintDatatype(datatype))
			NMPI_Type_free(mpi_errno, (&datatype));
		MPIU_Free(array_of_ints);
		break;
	case MPI_COMBINER_HVECTOR_INTEGER:
	case MPI_COMBINER_HVECTOR:
		printf("HVector:\n");
		array_of_ints = (int*)MPIU_Malloc(2*sizeof(int));
		NMPI_Type_get_contents(mpi_errno, (datatype, 2, 1, 1, array_of_ints, &a, &datatype));
		printf("count %d, blocklength %d, stride %ld, type:\n", array_of_ints[0], array_of_ints[1], a);
		if (PrintDatatype(datatype))
			NMPI_Type_free(mpi_errno, (&datatype));
		MPIU_Free(array_of_ints);
		break;
	case MPI_COMBINER_INDEXED:
		printf("Indexed:\n");
		array_of_ints = (int*)MPIU_Malloc(num_ints * sizeof(int));
		NMPI_Type_get_contents(mpi_errno, (datatype, num_ints, 0, 1, array_of_ints, NULL, &datatype));
		printf("count %d\n", array_of_ints[0]);
		for (i=1; i<=array_of_ints[0]; i++)
		{
			printf("blocklength %d\n", array_of_ints[i]);
		}
		for (i=array_of_ints[0]+1; i<=array_of_ints[2*array_of_ints[0]]; i++)
		{
			printf("displacement %d\n", array_of_ints[i]);
		}
		printf("type:\n");
		if (PrintDatatype(datatype))
			NMPI_Type_free(mpi_errno, (&datatype));
		MPIU_Free(array_of_ints);
		break;
	case MPI_COMBINER_HINDEXED_INTEGER:
	case MPI_COMBINER_HINDEXED:
		printf("HIndexed:\n");
		array_of_ints = (int*)MPIU_Malloc(num_ints * sizeof(int));
		array_of_adds = (MPI_Aint*)MPIU_Malloc(num_adds * sizeof(MPI_Aint));
		NMPI_Type_get_contents(mpi_errno, (datatype, num_ints, num_adds, 1, array_of_ints, array_of_adds, &datatype));
		printf("count %d\n", array_of_ints[0]);
		for (i=1; i<=array_of_ints[0]; i++)
		{
			printf("blocklength %d\n", array_of_ints[i]);
		}
		for (i=0; i<array_of_ints[0]; i++)
		{
			printf("displacement %ld\n", array_of_adds[i]);
		}
		printf("type:\n");
		if (PrintDatatype(datatype))
			NMPI_Type_free(mpi_errno, (&datatype));
		MPIU_Free(array_of_ints);
		MPIU_Free(array_of_adds);
		break;
	case MPI_COMBINER_INDEXED_BLOCK:
		printf("Indexed block:\n");
		array_of_ints = (int*)MPIU_Malloc(num_ints * sizeof(int));
		NMPI_Type_get_contents(mpi_errno, (datatype, num_ints, 0, 1, array_of_ints, NULL, &datatype));
		printf("count %d\n", array_of_ints[0]);
		printf("blocklength %d\n", array_of_ints[1]);
		for (i=2; i<=array_of_ints[0]+1; i++)
		{
			printf("displacement %d\n", array_of_ints[i]);
		}
		if (PrintDatatype(datatype))
			NMPI_Type_free(mpi_errno, (&datatype));
		MPIU_Free(array_of_ints);
		break;
	case MPI_COMBINER_STRUCT_INTEGER:
	case MPI_COMBINER_STRUCT:
		printf("Struct:\n");
		array_of_ints = (int*)MPIU_Malloc(num_ints * sizeof(int));
		array_of_adds = (MPI_Aint*)MPIU_Malloc(num_adds * sizeof(MPI_Aint));
		array_of_dtypes = (MPI_Datatype*)MPIU_Malloc(num_dtypes * sizeof(MPI_Datatype));
		NMPI_Type_get_contents(mpi_errno, (datatype, num_ints, num_adds, num_dtypes, array_of_ints, array_of_adds, array_of_dtypes));
		printf("%d datatypes\n", array_of_ints[0]);
		for (i=0; i<array_of_ints[0]; i++)
		{
			printf("blocklength %d, displacement %ld, type:\n", array_of_ints[i+1], array_of_adds[i]);
			if (PrintDatatype(array_of_dtypes[i]))
			{
				NMPI_Type_free(mpi_errno, (&array_of_dtypes[i]));
			}
		}
		MPIU_Free(array_of_ints);
		MPIU_Free(array_of_adds);
		MPIU_Free(array_of_dtypes);
		break;
	case MPI_COMBINER_SUBARRAY:
		printf("Subarray:\n");
		array_of_ints = (int*)MPIU_Malloc(num_ints * sizeof(int));
		NMPI_Type_get_contents(mpi_errno, (datatype, num_ints, 0, 1, array_of_ints, NULL, &datatype));
		count = array_of_ints[0];
		printf("ndims %d\n", count);
		for (i=1; i<=count; i++)
		{
			printf("size[%d] = %d\n", i-1, array_of_ints[i]);
		}
		for (i=count+1; i<=2*count; i++)
		{
			printf("subsize[%d] = %d\n", i-count-1, array_of_ints[i]);
		}
		for (i=2*count+1; i<=3*count; i++)
		{
			printf("start[%d] = %d\n", i-(2*count)-1, array_of_ints[i]);
		}
		printf("order = %d\n", array_of_ints[3*count+1]);
		if (PrintDatatype(datatype))
			NMPI_Type_free(mpi_errno, (&datatype));
		MPIU_Free(array_of_ints);
		break;
	case MPI_COMBINER_DARRAY:
		printf("Darray:\n");
		array_of_ints = (int*)MPIU_Malloc(num_ints * sizeof(int));
		NMPI_Type_get_contents(mpi_errno, (datatype, num_ints, 0, 1, array_of_ints, NULL, &datatype));
		printf("size %d, rank %d, ndims %d\n", array_of_ints[0], array_of_ints[1], array_of_ints[2]);
		count = array_of_ints[2];
		for (i=3; i<=count+2; i++)
		{
			printf("gsize[%d] = %d\n", i-3, array_of_ints[i]);
		}
		for (i=count+3; i<=(2*count)+2; i++)
		{
			printf("distrib[%d] = %d\n", i-count-3, array_of_ints[i]);
		}
		for (i=2*count+3; i<=3*count+2; i++)
		{
			printf("dargs[%d] = %d\n", i-(2*count)-3, array_of_ints[i]);
		}
		for (i=3*count+3; i<=4*count+2; i++)
		{
			printf("psize[%d] = %d\n", i-(3*count)-3, array_of_ints[i]);
		}
		printf("order = %d\n", array_of_ints[4*count+3]);
		if (PrintDatatype(datatype))
			NMPI_Type_free(mpi_errno, (&datatype));
		MPIU_Free(array_of_ints);
		break;
	case MPI_COMBINER_F90_REAL:
		printf("F90_real:\n");
		NMPI_Type_get_contents(mpi_errno, (datatype, 2, 0, 0, twoints, NULL, NULL));
		printf("p = %d, r = %d\n", twoints[0], twoints[1]);
		break;
	case MPI_COMBINER_F90_COMPLEX:
		printf("F90_complex:\n");
		NMPI_Type_get_contents(mpi_errno, (datatype, 2, 0, 0, twoints, NULL, NULL));
		printf("p = %d, r = %d\n", twoints[0], twoints[1]);
		break;
	case MPI_COMBINER_F90_INTEGER:
		printf("F90_integer:\n");
		NMPI_Type_get_contents(mpi_errno, (datatype, 1, 0, 0, &i, NULL, NULL));
		printf("r = %d\n", i);
		break;
	case MPI_COMBINER_RESIZED:
		printf("Resized:\n");
		NMPI_Type_get_contents(mpi_errno, (datatype, 0, 2, 1, NULL, twoaints, &datatype));
		printf("lb = %ld, extent = %ld, type:\n", twoaints[0] ,twoaints[1]);
		if (PrintDatatype(datatype))
			NMPI_Type_free(mpi_errno, (&datatype));
		break;
	default:
		break;
	}
	fflush(stdout);
	return 1;
}

static int AllocStream(MPIDI_Datatype_stream **stream)
{
	*stream = MPIU_Malloc(sizeof(MPIDI_Datatype_stream));
	if (*stream == NULL)
		return -1;
	(*stream)->buffer = NULL;
	(*stream)->cur_pos = NULL;
	(*stream)->end = NULL;
	return MPI_SUCCESS;
}

static int FreeStream(MPIDI_Datatype_stream **stream)
{
	if ((*stream)->buffer != NULL)
		MPIU_Free((*stream)->buffer);
	MPIU_Free(*stream);
	*stream = NULL;
	return MPI_SUCCESS;
}

static int GrowStream(MPIDI_Datatype_stream *stream)
{
	unsigned char *new_buffer;

	if (stream->buffer == NULL)
	{
		stream->buffer = MPIU_Malloc(BLOCK_ALLOC_SIZE);
		if (stream->buffer == NULL)
			return -1;
		stream->cur_pos = stream->buffer;
		stream->end = stream->cur_pos + BLOCK_ALLOC_SIZE;
		return MPI_SUCCESS;
	}

	/* Allocate a new buffer */
	new_buffer = MPIU_Malloc((stream->end - stream->buffer) + BLOCK_ALLOC_SIZE);
	/* Copy The old buffer into the new buffer */
	memcpy(new_buffer, stream->buffer, stream->cur_pos - stream->buffer);
	/* Set the end and cur_pos fields relative to the new buffer */
	stream->end = new_buffer + (stream->end - stream->buffer) + BLOCK_ALLOC_SIZE;
	stream->cur_pos = new_buffer + (stream->cur_pos - stream->buffer);
	/* Free the old buffer and save the new buffer */
	MPIU_Free(stream->buffer);
	stream->buffer = new_buffer;

	return MPI_SUCCESS;
}

static int GrowStreamSize(MPIDI_Datatype_stream *stream, int min_size)
{
	unsigned char *new_buffer;
	size_t size = max(BLOCK_ALLOC_SIZE, min_size);

	if (stream->buffer == NULL)
	{
		stream->buffer = MPIU_Malloc(size);
		if (stream->buffer == NULL)
			return -1;
		stream->cur_pos = stream->buffer;
		stream->end = stream->cur_pos + size;
		return MPI_SUCCESS;
	}

	/* Allocate a new buffer */
	new_buffer = MPIU_Malloc((stream->end - stream->buffer) + size);
	/* Copy The old buffer into the new buffer */
	memcpy(new_buffer, stream->buffer, stream->cur_pos - stream->buffer);
	/* Set the end and cur_pos fields relative to the new buffer */
	stream->end = new_buffer + (stream->end - stream->buffer) + size;
	stream->cur_pos = new_buffer + (stream->cur_pos - stream->buffer);
	/* Free the old buffer and save the new buffer */
	MPIU_Free(stream->buffer);
	stream->buffer = new_buffer;

	return MPI_SUCCESS;
}

static int SerializeInt(MPIDI_Datatype_stream *stream, int i)
{
	if (stream->end - stream->cur_pos < sizeof(int))
	{
		GrowStream(stream);
	}
	memcpy(stream->cur_pos, &i, sizeof(int));
	stream->cur_pos += sizeof(int);
	return MPI_SUCCESS;
}

static int SerializeInts(MPIDI_Datatype_stream *stream, int *a, int count)
{
	if ((stream->end - stream->cur_pos) < (count * (int)sizeof(int)))
	{
		GrowStreamSize(stream, count * sizeof(int));
	}

	memcpy(stream->cur_pos, a, count * sizeof(int));
	stream->cur_pos += count * sizeof(int);

	return MPI_SUCCESS;
}

static int SerializeAint(MPIDI_Datatype_stream *stream, MPI_Aint i)
{
	int64_t i64;

	i64 = i;
	if (stream->end - stream->cur_pos < sizeof(int64_t))
	{
		GrowStream(stream);
	}
	memcpy(stream->cur_pos, &i64, sizeof(int64_t));
	stream->cur_pos += sizeof(int64_t);
	return MPI_SUCCESS;
}

static int SerializeAints(MPIDI_Datatype_stream *stream, MPI_Aint *a, int count)
{
	int i;
	for (i=0; i<count; i++)
		SerializeAint(stream, a[i]);
	return MPI_SUCCESS;
}

static int RemoveSerializedBuffer(MPIDI_Datatype_stream *stream, void **buffer, int *length)
{
	*buffer = stream->buffer;
	*length = (int)(stream->cur_pos - stream->buffer);
	stream->buffer = NULL;
	stream->cur_pos = NULL;
	stream->end = NULL;
	return MPI_SUCCESS;
}

static int DeSerializeInt(void **stream, int *int_ptr)
{
	memcpy(int_ptr, *stream, sizeof(int));
	*stream = (unsigned char *)(*stream) + sizeof(int);
	return MPI_SUCCESS;
}

static int DeSerializeInts(void **stream, int *a, int count)
{
	memcpy(a, *stream, count * sizeof(int));
	*stream = (unsigned char *)(*stream) + (count * sizeof(int));
	return MPI_SUCCESS;
}

static int DeSerializeAint(void **stream, MPI_Aint *aint_ptr)
{
	int64_t stream_aint;

	memcpy(&stream_aint, *stream, sizeof(int64_t));
	*aint_ptr = (MPI_Aint)stream_aint;
	*stream = (unsigned char *)(*stream) + sizeof(int64_t);
	return MPI_SUCCESS;
}

static int DeSerializeAints(void **stream, MPI_Aint *a, int count)
{
	int i;
	for (i=0; i<count; i++)
	{
		DeSerializeAint(stream, &a[i]);
	}
	return MPI_SUCCESS;
}

static int SerializeDatatypeToStream(MPIDI_Datatype_stream *stream, MPI_Datatype datatype)
{
	int mpi_errno;
	int *array_of_ints, twoints[2];
	MPI_Aint *array_of_adds, a, twoaints[2];
	MPI_Datatype *array_of_dtypes;
	int num_ints, num_adds, num_dtypes, combiner;
	int i;

	NMPI_Type_get_envelope(mpi_errno, (datatype, &num_ints, &num_adds, &num_dtypes, &combiner));

	switch (combiner)
	{
	case MPI_COMBINER_NAMED:
		SerializeInt(stream, MPI_COMBINER_NAMED);
		SerializeInt(stream, datatype);
		/*
		PrintNamedDatatype(datatype);
		fflush(stdout);
		*/
		return 0;
	case MPI_COMBINER_DUP:
		SerializeInt(stream, MPI_COMBINER_DUP);
		/*printf("Dupped:\n");*/
		NMPI_Type_get_contents(mpi_errno, (datatype, 0, 0, 1, NULL, NULL, &datatype));
		if (SerializeDatatypeToStream(stream, datatype))
			NMPI_Type_free(mpi_errno, (&datatype));
		break;
	case MPI_COMBINER_CONTIGUOUS:
		SerializeInt(stream, MPI_COMBINER_CONTIGUOUS);
		/*printf("Contiguous:\n");*/
		NMPI_Type_get_contents(mpi_errno, (datatype, 1, 0, 1, &i, NULL, &datatype));
		SerializeInt(stream, i);
		/*printf("count %d, type:\n", i);*/
		if (SerializeDatatypeToStream(stream, datatype))
			NMPI_Type_free(mpi_errno, (&datatype));
		break;
	case MPI_COMBINER_VECTOR:
		SerializeInt(stream, MPI_COMBINER_VECTOR);
		/*printf("Vector:\n");*/
		array_of_ints = (int*)MPIU_Malloc(3 * sizeof(int));
		NMPI_Type_get_contents(mpi_errno, (datatype, 3, 0, 1, array_of_ints, NULL, &datatype));
		/*printf("count %d, blocklength %d, stride %d, type:\n", array_of_ints[0], array_of_ints[1], array_of_ints[2], &datatype);*/
		SerializeInts(stream, array_of_ints, 3);
		if (SerializeDatatypeToStream(stream, datatype))
			NMPI_Type_free(mpi_errno, (&datatype));
		MPIU_Free(array_of_ints);
		break;
	case MPI_COMBINER_HVECTOR_INTEGER:
	case MPI_COMBINER_HVECTOR:
		SerializeInt(stream, MPI_COMBINER_HVECTOR);
		/*printf("HVector:\n");*/
		array_of_ints = (int*)MPIU_Malloc(2*sizeof(int));
		NMPI_Type_get_contents(mpi_errno, (datatype, 2, 1, 1, array_of_ints, &a, &datatype));
		SerializeInts(stream, array_of_ints, 2);
		SerializeAint(stream, a);
		/*printf("count %d, blocklength %d, stride %ld, type:\n", array_of_ints[0], array_of_ints[1], a);*/
		if (SerializeDatatypeToStream(stream, datatype))
			NMPI_Type_free(mpi_errno, (&datatype));
		MPIU_Free(array_of_ints);
		break;
	case MPI_COMBINER_INDEXED:
		SerializeInt(stream, MPI_COMBINER_INDEXED);
		/*printf("Indexed:\n");*/
		array_of_ints = (int*)MPIU_Malloc(num_ints * sizeof(int));
		NMPI_Type_get_contents(mpi_errno, (datatype, num_ints, 0, 1, array_of_ints, NULL, &datatype));
		SerializeInts(stream, array_of_ints, num_ints);
		/*
		printf("count %d\n", array_of_ints[0]);
		for (i=1; i<=array_of_ints[0]; i++)
		{
			printf("blocklength %d\n", array_of_ints[i]);
		}
		for (i=array_of_ints[0]+1; i<=array_of_ints[2*array_of_ints[0]]; i++)
		{
			printf("displacement %d\n", array_of_ints[i]);
		}
		printf("type:\n");
		*/
		if (SerializeDatatypeToStream(stream, datatype))
			NMPI_Type_free(mpi_errno, (&datatype));
		MPIU_Free(array_of_ints);
		break;
	case MPI_COMBINER_HINDEXED_INTEGER:
	case MPI_COMBINER_HINDEXED:
		SerializeInt(stream, MPI_COMBINER_HINDEXED);
		/*printf("HIndexed:\n");*/
		array_of_ints = (int*)MPIU_Malloc(num_ints * sizeof(int));
		array_of_adds = (MPI_Aint*)MPIU_Malloc(num_adds * sizeof(MPI_Aint));
		NMPI_Type_get_contents(mpi_errno, (datatype, num_ints, num_adds, 1, array_of_ints, array_of_adds, &datatype));
		SerializeInts(stream, array_of_ints, num_ints);
		SerializeAints(stream, array_of_adds, num_adds);
		/*
		printf("count %d\n", array_of_ints[0]);
		for (i=1; i<=array_of_ints[0]; i++)
		{
			printf("blocklength %d\n", array_of_ints[i]);
		}
		for (i=0; i<array_of_ints[0]; i++)
		{
			printf("displacement %ld\n", array_of_adds[i]);
		}
		printf("type:\n");
		*/
		if (SerializeDatatypeToStream(stream, datatype))
			NMPI_Type_free(mpi_errno, (&datatype));
		MPIU_Free(array_of_ints);
		MPIU_Free(array_of_adds);
		break;
	case MPI_COMBINER_INDEXED_BLOCK:
		SerializeInt(stream, MPI_COMBINER_INDEXED_BLOCK);
		/*printf("Indexed block:\n");*/
		array_of_ints = (int*)MPIU_Malloc(num_ints * sizeof(int));
		NMPI_Type_get_contents(mpi_errno, (datatype, num_ints, 0, 1, array_of_ints, NULL, &datatype));
		SerializeInts(stream, array_of_ints, num_ints);
		/*
		printf("count %d\n", array_of_ints[0]);
		printf("blocklength %d\n", array_of_ints[1]);
		for (i=2; i<=array_of_ints[0]+1; i++)
		{
			printf("displacement %d\n", array_of_ints[i]);
		}
		*/
		if (SerializeDatatypeToStream(stream, datatype))
			NMPI_Type_free(mpi_errno, (&datatype));
		MPIU_Free(array_of_ints);
		break;
	case MPI_COMBINER_STRUCT_INTEGER:
	case MPI_COMBINER_STRUCT:
		SerializeInt(stream, MPI_COMBINER_STRUCT);
		/*printf("Struct:\n");*/
		array_of_ints = (int*)MPIU_Malloc(num_ints * sizeof(int));
		array_of_adds = (MPI_Aint*)MPIU_Malloc(num_adds * sizeof(MPI_Aint));
		array_of_dtypes = (MPI_Datatype*)MPIU_Malloc(num_dtypes * sizeof(MPI_Datatype));
		NMPI_Type_get_contents(mpi_errno, (datatype, num_ints, num_adds, num_dtypes, array_of_ints, array_of_adds, array_of_dtypes));
		SerializeInts(stream, array_of_ints, num_ints);
		SerializeAints(stream, array_of_adds, num_adds);
		/*printf("%d datatypes\n", array_of_ints[0]);*/
		for (i=0; i<num_dtypes; i++)
		{
			/*printf("blocklength %d, displacement %ld, type:\n", array_of_ints[i+1], array_of_adds[i]);*/
			if (SerializeDatatypeToStream(stream, array_of_dtypes[i]))
			{
				NMPI_Type_free(mpi_errno, (&array_of_dtypes[i]));
			}
		}
		MPIU_Free(array_of_ints);
		MPIU_Free(array_of_adds);
		MPIU_Free(array_of_dtypes);
		break;
	case MPI_COMBINER_SUBARRAY:
		SerializeInt(stream, MPI_COMBINER_SUBARRAY);
		/*printf("Subarray:\n");*/
		array_of_ints = (int*)MPIU_Malloc(num_ints * sizeof(int));
		NMPI_Type_get_contents(mpi_errno, (datatype, num_ints, 0, 1, array_of_ints, NULL, &datatype));
		SerializeInts(stream, array_of_ints, num_ints);
		/*
		count = array_of_ints[0];
		printf("ndims %d\n", count);
		for (i=1; i<=count; i++)
		{
			printf("size[%d] = %d\n", i-1, array_of_ints[i]);
		}
		for (i=count+1; i<=2*count; i++)
		{
			printf("subsize[%d] = %d\n", i-count-1, array_of_ints[i]);
		}
		for (i=2*count+1; i<=3*count; i++)
		{
			printf("start[%d] = %d\n", i-(2*count)-1, array_of_ints[i]);
		}
		printf("order = %d\n", array_of_ints[3*count+1]);
		*/
		if (SerializeDatatypeToStream(stream, datatype))
			NMPI_Type_free(mpi_errno, (&datatype));
		MPIU_Free(array_of_ints);
		break;
	case MPI_COMBINER_DARRAY:
		SerializeInt(stream, MPI_COMBINER_DARRAY);
		/*printf("Darray:\n");*/
		array_of_ints = (int*)MPIU_Malloc(num_ints * sizeof(int));
		NMPI_Type_get_contents(mpi_errno, (datatype, num_ints, 0, 1, array_of_ints, NULL, &datatype));
		SerializeInts(stream, array_of_ints, num_ints);
		/*
		printf("size %d, rank %d, ndims %d\n", array_of_ints[0], array_of_ints[1], array_of_ints[2]);
		count = array_of_ints[2];
		for (i=3; i<=count+2; i++)
		{
			printf("gsize[%d] = %d\n", i-3, array_of_ints[i]);
		}
		for (i=count+3; i<=(2*count)+2; i++)
		{
			printf("distrib[%d] = %d\n", i-count-3, array_of_ints[i]);
		}
		for (i=2*count+3; i<=3*count+2; i++)
		{
			printf("dargs[%d] = %d\n", i-(2*count)-3, array_of_ints[i]);
		}
		for (i=3*count+3; i<=4*count+2; i++)
		{
			printf("psize[%d] = %d\n", i-(3*count)-3, array_of_ints[i]);
		}
		printf("order = %d\n", array_of_ints[4*count+3]);
		*/
		if (SerializeDatatypeToStream(stream, datatype))
			NMPI_Type_free(mpi_errno, (&datatype));
		MPIU_Free(array_of_ints);
		break;
	case MPI_COMBINER_F90_REAL:
		SerializeInt(stream, MPI_COMBINER_F90_REAL);
		/*printf("F90_real:\n");*/
		NMPI_Type_get_contents(mpi_errno, (datatype, 2, 0, 0, twoints, NULL, NULL));
		SerializeInts(stream, twoints, 2);
		/*printf("p = %d, r = %d\n", twoints[0], twoints[1]);*/
		break;
	case MPI_COMBINER_F90_COMPLEX:
		SerializeInt(stream, MPI_COMBINER_F90_COMPLEX);
		/*printf("F90_complex:\n");*/
		NMPI_Type_get_contents(mpi_errno, (datatype, 2, 0, 0, twoints, NULL, NULL));
		SerializeInts(stream, twoints, 2);
		/*printf("p = %d, r = %d\n", twoints[0], twoints[1]);*/
		break;
	case MPI_COMBINER_F90_INTEGER:
		SerializeInt(stream, MPI_COMBINER_F90_INTEGER);
		/*printf("F90_integer:\n");*/
		NMPI_Type_get_contents(mpi_errno, (datatype, 1, 0, 0, &i, NULL, NULL));
		SerializeInt(stream, i);
		/*printf("r = %d\n", i);*/
		break;
	case MPI_COMBINER_RESIZED:
		SerializeInt(stream, MPI_COMBINER_RESIZED);
		/*printf("Resized:\n");*/
		NMPI_Type_get_contents(mpi_errno, (datatype, 0, 2, 1, NULL, twoaints, &datatype));
		SerializeAints(stream, twoaints, 2);
		/*printf("lb = %ld, extent = %ld, type:\n", twoaints[0] ,twoaints[1]);*/
		if (SerializeDatatypeToStream(stream, datatype))
			NMPI_Type_free(mpi_errno, (&datatype));
		break;
	default:
		break;
	}
	/*fflush(stdout);*/
	return 1;
}

int SerializeDatatype(MPI_Datatype datatype, void **buffer, int *length)
{
	MPIDI_Datatype_stream *stream;

	/*printf("[%d]inside SerializeDatatype\n", MPIR_Process.comm_world->rank);fflush(stdout);*/
	AllocStream(&stream);
	if (stream == NULL)
	{
		return -1;
	}
	SerializeDatatypeToStream(stream, datatype);
	RemoveSerializedBuffer(stream, buffer, length);
	FreeStream(&stream);
	/*printf("[%d]returning from SerializeDatatype\n", MPIR_Process.comm_world->rank);fflush(stdout);*/
	return 0;
}

#undef FUNCNAME
#define FUNCNAME DeSerializeDatatypeFromStream
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int DeSerializeDatatypeFromStream(void **stream, MPI_Datatype *datatype_ptr)
{
	int mpi_errno = MPI_SUCCESS;
	int *array_of_ints, twoints[2];
	MPI_Aint *array_of_adds, a, twoaints[2];
	MPI_Datatype *array_of_dtypes;
	int combiner;
	int i, count;
#if 0 /* If F90 support enabled, use these variables */
	int p, r;
#endif
	int blen, *blens, *displs;
	MPI_Aint *adispls;
	int *sizes, *subsize, *start, order;
	int size, rank, ndims;
	int *gsize, *distrib, *dargs, *psize;

	/*printf("reading the combiner:\n");fflush(stdout);*/
	DeSerializeInt(stream, &combiner);
	/*printf("the combiner = %d\n", combiner);fflush(stdout);*/

	switch (combiner)
	{
	case MPI_COMBINER_NAMED:
		DeSerializeInt(stream, datatype_ptr);
		/*PrintNamedDatatype(*datatype_ptr);*/
		fflush(stdout);
		return 0;
	case MPI_COMBINER_DUP:
		/*printf("Dupped:\n");*/
		DeSerializeDatatypeFromStream(stream, datatype_ptr);
		NMPI_Type_dup(mpi_errno, (*datatype_ptr, datatype_ptr));
		MPIU_ERR_CHK_POP(mpi_errno);
		break;
	case MPI_COMBINER_CONTIGUOUS:
		/*printf("Contiguous:\n");*/
		DeSerializeInt(stream, &count);
		/*printf("count %d, type:\n", count);*/
		DeSerializeDatatypeFromStream(stream, datatype_ptr);
		NMPI_Type_contiguous(mpi_errno, (count, *datatype_ptr, datatype_ptr));
		MPIU_ERR_CHK_POP(mpi_errno);
		break;
	case MPI_COMBINER_VECTOR:
		/*printf("Vector:\n");*/
		array_of_ints = (int*)MPIU_Malloc(3 * sizeof(int));
		DeSerializeInts(stream, array_of_ints, 3);
		/*printf("count %d, blocklength %d, stride %d, type:\n", array_of_ints[0], array_of_ints[1], array_of_ints[2]);*/
		DeSerializeDatatypeFromStream(stream, datatype_ptr);
		NMPI_Type_vector(mpi_errno, (array_of_ints[0], array_of_ints[1], array_of_ints[2], *datatype_ptr, datatype_ptr));
		MPIU_ERR_CHK_POP(mpi_errno);
		MPIU_Free(array_of_ints);
		break;
	case MPI_COMBINER_HVECTOR_INTEGER:
	case MPI_COMBINER_HVECTOR:
		/*printf("HVector:\n");*/
		DeSerializeInts(stream, twoints, 2);
		DeSerializeAint(stream, &a);
		/*printf("count %d, blocklength %d, stride %ld, type:\n", twoints[0], twoints[1], a);*/
		DeSerializeDatatypeFromStream(stream, datatype_ptr);
		NMPI_Type_hvector(mpi_errno, (twoints[0], twoints[1], a, *datatype_ptr, datatype_ptr));
		MPIU_ERR_CHK_POP(mpi_errno);
		break;
	case MPI_COMBINER_INDEXED:
		/*printf("Indexed:\n");*/
		DeSerializeInt(stream, &count);
		blens = (int*)MPIU_Malloc(count * sizeof(int));
		displs = (int*)MPIU_Malloc(count * sizeof(int));
		DeSerializeInts(stream, blens, count);
		DeSerializeInts(stream, displs, count);
		/*
		printf("count %d\n", count);
		for (i=0; i<count; i++)
		{
			if (i == 10)
				printf("...\n");
			else if (i < 10)
				printf("blocklength %d\n", blens[i]);
		}
		for (i=0; i<count; i++)
		{
			if (i == 10)
				printf("...\n");
			else if (i < 10)
				printf("displacement %d\n", displs[i]);
		}
		*/
		DeSerializeDatatypeFromStream(stream, datatype_ptr);
		NMPI_Type_indexed(mpi_errno, (count, blens, displs, *datatype_ptr, datatype_ptr));
		MPIU_ERR_CHK_POP(mpi_errno);
		MPIU_Free(blens);
		MPIU_Free(displs);
		break;
	case MPI_COMBINER_HINDEXED_INTEGER:
	case MPI_COMBINER_HINDEXED:
		/*printf("HIndexed:\n");fflush(stdout);*/
		DeSerializeInt(stream, &count);
		blens = (int*)MPIU_Malloc(count * sizeof(int));
		adispls = (MPI_Aint*)MPIU_Malloc(count * sizeof(MPI_Aint));
		DeSerializeInts(stream, blens, count);
		DeSerializeAints(stream, adispls, count);
		/*
		printf("count %d\n", count);
		for (i=0; i<count; i++)
		{
			if (i == 10)
				printf("...\n");
			else if (i < 10)
				printf("blocklength %d\n", blens[i]);
		}
		for (i=0; i<count; i++)
		{
			if (i == 10)
				printf("...\n");
			else if (i < 10)
				printf("displacement %ld\n", adispls[i]);
		}
		fflush(stdout);
		*/
		DeSerializeDatatypeFromStream(stream, datatype_ptr);
		NMPI_Type_hindexed(mpi_errno, (count, blens, adispls, *datatype_ptr, datatype_ptr));
		MPIU_ERR_CHK_POP(mpi_errno);
		MPIU_Free(blens);
		MPIU_Free(adispls);
		break;
	case MPI_COMBINER_INDEXED_BLOCK:
		/*printf("Indexed block:\n");fflush(stdout);*/
		DeSerializeInt(stream, &count);
		DeSerializeInt(stream, &blen);
		blens = (int*)MPIU_Malloc(count * sizeof(int));
		if (blens == NULL)
			return -1;
		DeSerializeInts(stream, blens, count);
		/*
		printf("count %d\n", count);
		printf("blocklength %d\n", blen);
		for (i=0; i<count; i++)
		{
			if (i == 10)
				printf("...\n");
			else if (i < 10)
				printf("displacement %d\n", blens[i]);
		}
		fflush(stdout);
		*/
		DeSerializeDatatypeFromStream(stream, datatype_ptr);
		NMPI_Type_create_indexed_block(mpi_errno, (count, blen, blens, *datatype_ptr, datatype_ptr));
		MPIU_ERR_CHK_POP(mpi_errno);
		MPIU_Free(blens);
		break;
	case MPI_COMBINER_STRUCT_INTEGER:
	case MPI_COMBINER_STRUCT:
		/*printf("Struct:\n");fflush(stdout);*/
		DeSerializeInt(stream, &count);
		array_of_ints = (int*)MPIU_Malloc(count * sizeof(int));
		array_of_adds = (MPI_Aint*)MPIU_Malloc(count * sizeof(MPI_Aint));
		array_of_dtypes = (MPI_Datatype*)MPIU_Malloc(count * sizeof(MPI_Datatype));
		DeSerializeInts(stream, array_of_ints, count);
		DeSerializeAints(stream, array_of_adds, count);
		/*printf("%d datatypes\n", count);fflush(stdout);*/
		for (i=0; i<count; i++)
		{
			/*printf("blocklength %d, displacement %ld, type:\n", array_of_ints[i], array_of_adds[i]);fflush(stdout);*/
			DeSerializeDatatypeFromStream(stream, &array_of_dtypes[i]);
		}
		NMPI_Type_struct(mpi_errno, (count, array_of_ints, array_of_adds, array_of_dtypes, datatype_ptr));
		MPIU_ERR_CHK_POP(mpi_errno);
		MPIU_Free(array_of_ints);
		MPIU_Free(array_of_adds);
		MPIU_Free(array_of_dtypes);
		break;
	case MPI_COMBINER_SUBARRAY:
		/*printf("Subarray:\n");fflush(stdout);*/
		DeSerializeInt(stream, &count);
		sizes = (int*)MPIU_Malloc(count * sizeof(int));
		if (sizes == NULL)
			return -1;
		subsize = (int*)MPIU_Malloc(count * sizeof(int));
		if (subsize == NULL)
			return -1;
		start = (int*)MPIU_Malloc(count * sizeof(int));
		if (start == NULL)
			return -1;
		DeSerializeInts(stream, sizes, count);
		DeSerializeInts(stream, subsize, count);
		DeSerializeInts(stream, start, count);
		DeSerializeInt(stream, &order);
		/*
		printf("ndims %d\n", count);
		for (i=0; i<count; i++)
		{
			if (i == 10)
				printf("...\n");
			else if (i < 10)
				printf("size[%d] = %d\n", i, sizes[i]);
		}
		for (i=0; i<count; i++)
		{
			if (i == 10)
				printf("...\n");
			else if (i < 10)
				printf("subsize[%d] = %d\n", i, subsize[i]);
		}
		for (i=0; i<count; i++)
		{
			if (i == 10)
				printf("...\n");
			else if (i < 10)
				printf("start[%d] = %d\n", i, start[i]);
		}
		printf("order = %d\n", order);
		fflush(stdout);
		*/
		DeSerializeDatatypeFromStream(stream, datatype_ptr);
		NMPI_Type_create_subarray(mpi_errno, (count, sizes, subsize, start, order, *datatype_ptr, datatype_ptr));
		MPIU_ERR_CHK_POP(mpi_errno);
		MPIU_Free(sizes);
		MPIU_Free(subsize);
		MPIU_Free(start);
		break;
	case MPI_COMBINER_DARRAY:
		/*printf("Darray:\n");fflush(stdout);*/
		DeSerializeInt(stream, &size);
		DeSerializeInt(stream, &rank);
		DeSerializeInt(stream, &ndims);
		gsize = (int*)MPIU_Malloc(ndims * sizeof(int));
		if (gsize == NULL)
			return -1;
		distrib = (int*)MPIU_Malloc(ndims * sizeof(int));
		if (distrib == NULL)
			return -1;
		dargs = (int*)MPIU_Malloc(ndims * sizeof(int));
		if (dargs == NULL)
			return -1;
		psize = (int*)MPIU_Malloc(ndims * sizeof(int));
		if (psize == NULL)
			return -1;
		DeSerializeInts(stream, gsize, ndims);
		DeSerializeInts(stream, distrib, ndims);
		DeSerializeInts(stream, dargs, ndims);
		DeSerializeInts(stream, psize, ndims);
		DeSerializeInt(stream, &order);
		/*
		printf("size %d, rank %d, ndims %d\n", size, rank, ndims);
		for (i=0; i<ndims; i++)
		{
			if (i == 10)
				printf("...\n");
			else if (i < 10)
				printf("gsize[%d] = %d\n", i, gsize[i]);
		}
		for (i=0; i<ndims; i++)
		{
			if (i == 10)
				printf("...\n");
			else if (i < 10)
				printf("distrib[%d] = %d\n", i, distrib[i]);
		}
		for (i=0; i<ndims; i++)
		{
			if (i == 10)
				printf("...\n");
			else if (i < 10)
				printf("dargs[%d] = %d\n", i, dargs[i]);
		}
		for (i=0; i<ndims; i++)
		{
			if (i == 10)
				printf("...\n");
			else if (i < 10)
				printf("psize[%d] = %d\n", i, psize[i]);
		}
		printf("order = %d\n", order);
		fflush(stdout);
		*/
		DeSerializeDatatypeFromStream(stream, datatype_ptr);
		NMPI_Type_create_darray(mpi_errno, (size, rank, ndims, gsize, distrib, dargs, psize, order, *datatype_ptr, datatype_ptr));
		MPIU_ERR_CHK_POP(mpi_errno);
		MPIU_Free(gsize);
		MPIU_Free(distrib);
		MPIU_Free(dargs);
		MPIU_Free(psize);
		break;
#if 0 /* If F90 support is enabled then add these functions */
	case MPI_COMBINER_F90_REAL:
		/*printf("F90_real:\n");*/
		DeSerializeInt(stream, &p);
		DeSerializeInt(stream, &r);
		/*printf("p = %d, r = %d\n", twoints[0], twoints[1]);*/
		NMPI_Type_create_f90_real(mpi_errno, (p, r, datatype_ptr));
		MPIU_ERR_CHK_POP(mpi_errno);
		break;
	case MPI_COMBINER_F90_COMPLEX:
		/*printf("F90_complex:\n");*/
		DeSerializeInt(stream, &p);
		DeSerializeInt(stream, &r);
		/*printf("p = %d, r = %d\n", twoints[0], twoints[1]);*/
		NMPI_Type_create_f90_complex(mpi_errno, (p, r, datatype_ptr));
		MPIU_ERR_CHK_POP(mpi_errno);
		break;
	case MPI_COMBINER_F90_INTEGER:
		/*printf("F90_integer:\n");*/
		DeSerializeInt(stream, &r);
		/*printf("r = %d\n", r);*/
		NMPI_Type_create_f90_integer(mpi_errno, (r, datatype_ptr));
		MPIU_ERR_CHK_POP(mpi_errno);
		break;
#endif
	case MPI_COMBINER_RESIZED:
		/*printf("Resized:\n");*/
		DeSerializeAints(stream, twoaints, 2);
		/*printf("lb = %ld, extent = %ld, type:\n", twoaints[0] ,twoaints[1]);*/
		DeSerializeDatatypeFromStream(stream, datatype_ptr);
		NMPI_Type_create_resized(mpi_errno, (*datatype_ptr, twoaints[0], twoaints[1], datatype_ptr));
		MPIU_ERR_CHK_POP(mpi_errno);
		break;
	default:
		/*printf("unknown combiner: %d\n", combiner);fflush(stdout);*/
		mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_RECOVERABLE, L"DeSerializeDatatypeFromStream", __LINE__, MPI_ERR_OTHER, L"**fail", NULL);
		break;
	}
	/*fflush(stdout);*/
fn_exit:
	return mpi_errno;
fn_fail:
	goto fn_exit;
}

int DeSerializeDatatype(void *buffer, MPI_Datatype *datatype_ptr)
{
	int mpi_errno;
	/*printf("[%d]inside DeSerializeDatatype\n", MPIR_Process.comm_world->rank);fflush(stdout);*/
	mpi_errno = DeSerializeDatatypeFromStream(&buffer, datatype_ptr);
	/*
	printf("[%d]printing DeSerialized datatype:\n", MPIR_Process.comm_world->rank);fflush(stdout);
	PrintDatatype(*datatype_ptr);
	*/
	/*printf("[%d]returning from DeSerializeDatatype\n", MPIR_Process.comm_world->rank);fflush(stdout);*/
	return mpi_errno;
}

/* FIXME: What is this routine for?
   It appears to be used only in ch3_progress, ch3_progress_connect, or
   ch3_progress_sock files.  Is this a general operation, or does it 
   belong in util/sock ? It appears to be used in multiple channels, 
   but probably belongs in mpid_vc, along with the vc exit code that 
   is currently in MPID_Finalize */

/* FIXME: The only event is event_terminated.  Should this have 
   a different name/expected function? */

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3U_Handle_connection
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
/*@
  MPIDI_CH3U_Handle_connection - handle connection event

  Input Parameters:
+ vc - virtual connection
. event - connection event

  NOTE:
  At present this function is only used for connection termination
@*/
int MPIDI_CH3U_Handle_connection(MPIDI_VC_t * vc, MPIDI_VC_Event_t event)
{
	int inuse;
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3U_HANDLE_CONNECTION);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3U_HANDLE_CONNECTION);

	switch (event)
	{
		case MPIDI_VC_EVENT_TERMINATED:
		{
			switch (vc->state)
			{
				case MPIDI_VC_STATE_CLOSE_ACKED:
				{
					MPIU_DBG_VCSTATECHANGE(vc, VC_STATE_INACTIVE);
					vc->state = MPIDI_VC_STATE_INACTIVE;
					/* FIXME: Decrement the reference count?  Who increments? */
					/* FIXME: The reference count is often already 0.  But
					   not always */
					/* MPIU_Object_set_ref(vc, 0); ??? */

					/*
					 * FIXME: The VC used in connect accept has a NULL process group
					 */
					if (vc->pg != NULL && vc->ref_count == 0)
					{ 
						/* FIXME: Who increments the reference count that
						   this is decrementing? */
						/* When the reference count for a vc becomes zero, 
						   decrement the reference count
						   of the associated process group.  */
						/* FIXME: This should be done when the reference 
						   count of the vc is first decremented */
						MPIDI_PG_Release_ref(vc->pg, &inuse);
						if (inuse == 0)
						{
							MPIDI_PG_Destroy(vc->pg);
						}
					}

					/* MT: this is not thread safe */
					MPIDI_Outstanding_close_ops -= 1;
					/*wprintf(L"close_ops = %d\n", MPIDI_Outstanding_close_ops);fflush(stdout);*/
					MPIU_DBG_MSG_D(CH3_CONNECT,VERBOSE, L"outstanding close operations = %d", MPIDI_Outstanding_close_ops);
			
					if (MPIDI_Outstanding_close_ops == 0)
					{
						MPIDI_CH3_Progress_signal_completion();
					}

					break;
				}

				default:
				{
					MPIU_DBG_MSG_D(CH3_CONNECT,TYPICAL,L"Unhandled connection state %d when closing connection",vc->state);
					mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_INTERN, L"**ch3|unhandled_connection_state", L"**ch3|unhandled_connection_state %p %d", vc, event);
					break;
				}
			}

			break;
		}
	
		default:
		{
			break;
		}
	}
		
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3U_HANDLE_CONNECTION);

	return mpi_errno;
}

/* This function simply tells the CH3 device to use the defaults for the MPI-2 RMA functions */
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_RMAFnsInit
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3U_RMAFnsInit( MPIDI_RMAFns *a ) 
{ 
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_RMAFNSINIT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_RMAFNSINIT);
	MPIU_UNREFERENCED_ARG(a);
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_RMAFNSINIT);
	return MPI_SUCCESS;
}

MPIU_BOOL MPIDI_CH3U_VC_NeedsSendClose( MPIDI_VC_t *vc )
{
	if (vc->state == MPIDI_VC_STATE_ACTIVE || vc->state == MPIDI_VC_STATE_REMOTE_CLOSE)
		return MPIU_TRUE;
	return MPIU_FALSE;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_VC_SendClose
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
/*@
  MPIDI_CH3U_VC_SendClose - Initiate a close on a virtual connection
  
  Input Parameters:
+ vc - Virtual connection to close
- i  - rank of virtual connection within a process group (used for debugging)

  Notes:
  The current state of this connection must be either 'MPIDI_VC_STATE_ACTIVE' 
  or 'MPIDI_VC_STATE_REMOTE_CLOSE'.  
  @*/
int MPIDI_CH3U_VC_SendClose( MPIDI_VC_t *vc, int rank )
{
	MPIDI_CH3_Pkt_t upkt;
	MPIDI_CH3_Pkt_close_t * close_pkt = &upkt.close;
	MPID_Request * sreq;
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3U_VC_SENDCLOSE);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3U_VC_SENDCLOSE);

	MPIU_UNREFERENCED_ARG(rank);

	MPIU_Assert( vc->state == MPIDI_VC_STATE_ACTIVE || 
				 vc->state == MPIDI_VC_STATE_REMOTE_CLOSE );

	MPIDI_Pkt_init(close_pkt, MPIDI_CH3_PKT_CLOSE);
	close_pkt->ack = (vc->state == MPIDI_VC_STATE_ACTIVE) ? FALSE : TRUE;
	
	/* MT: this is not thread safe */
	MPIDI_Outstanding_close_ops += 1;
	/*wprintf(L"close_ops = %d\n", MPIDI_Outstanding_close_ops);fflush(stdout);*/
	MPIU_DBG_MSG_FMT(CH3_CONNECT,VERBOSE,(MPIU_DBG_FDEST, L"sending close(%s) on vc (pg=%p) %p to rank %d, ops = %d", close_pkt->ack ? L"TRUE" : L"FALSE", vc->pg, vc, rank, MPIDI_Outstanding_close_ops));
					

	/*
	 * A close packet acknowledging this close request could be
	 * received during iStartMsg, therefore the state must
	 * be changed before the close packet is sent.
	 */
	if (vc->state == MPIDI_VC_STATE_ACTIVE)
	{
		MPIU_DBG_VCSTATECHANGE(vc, VC_STATE_LOCAL_CLOSE);
		vc->state = MPIDI_VC_STATE_LOCAL_CLOSE;
	}
	else
	{
		MPIU_Assert( vc->state == MPIDI_VC_STATE_REMOTE_CLOSE );
		MPIU_DBG_VCSTATECHANGE(vc, VC_STATE_CLOSE_ACKED);
		vc->state = MPIDI_VC_STATE_CLOSE_ACKED;
	}
				
	mpi_errno = MPIDI_CH3.iStartMsg(vc, close_pkt, sizeof(*close_pkt), &sreq);
	if (mpi_errno != MPI_SUCCESS)
	{
		MPIU_ERR_SET(mpi_errno,MPI_ERR_OTHER, L"**ch3|send_close_ack");
	}
	
	if (sreq != NULL)
	{
		MPID_Request_release(sreq);
	}

	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3U_VC_SENDCLOSE);
	return mpi_errno;
}

/* Here is the matching code that processes a close packet when it is received */
int MPIDI_CH3_PktHandler_Close( MPIDI_VC_t *vc, MPIDI_CH3_Pkt_t *pkt, MPID_Request **rreqp )
{
	MPIDI_CH3_Pkt_close_t * close_pkt = &pkt->close;
	int mpi_errno = MPI_SUCCESS;
			
	if (vc->state == MPIDI_VC_STATE_LOCAL_CLOSE)
	{
		MPIDI_CH3_Pkt_t upkt;
		MPIDI_CH3_Pkt_close_t * resp_pkt = &upkt.close;
		MPID_Request * resp_sreq;
		
		MPIDI_Pkt_init(resp_pkt, MPIDI_CH3_PKT_CLOSE);
		resp_pkt->ack = TRUE;
		
		MPIU_DBG_MSG_D(CH3_DISCONNECT,TYPICAL,L"sending close(TRUE) to %d", vc->pg_rank);
		mpi_errno = MPIDI_CH3.iStartMsg(vc, resp_pkt, sizeof(*resp_pkt), &resp_sreq);
		if (mpi_errno != MPI_SUCCESS)
		{
			MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, L"**ch3|send_close_ack");
		}
		
		if (resp_sreq != NULL)
		{
			MPID_Request_release(resp_sreq);
			wprintf( L"Panic on send close\n" ); fflush(stdout);
		}
	}
	
	if (close_pkt->ack == FALSE)
	{
		if (vc->state == MPIDI_VC_STATE_LOCAL_CLOSE)
		{
			MPIU_DBG_MSG_D(CH3_DISCONNECT,TYPICAL, L"received close(FALSE) from %d, moving to CLOSE_ACKED.", vc->pg_rank);
			MPIU_DBG_VCSTATECHANGE(vc,VC_STATE_CLOSE_ACKED);
			vc->state = MPIDI_VC_STATE_CLOSE_ACKED;
		}
#if 0
		else if (vc->state == MPIDI_VC_STATE_CLOSE_ACKED) {
			/* FIXME: This situation has been seen with the ssm device,
			   when running on a single process.  It may indicate that the
			   close protocol, for shared memory connections, can
			   occasionally reach this state when both sides start
			   closing the connection.  We will act as if this is
			   duplicate information can be ignored (rather than triggering
			   the Assert in the next case) */
			MPIU_DBG_MSG(CH3_DISCONNECT,TYPICAL, L"Saw CLOSE_ACKED while already in that state");
			vc->state = MPIDI_VC_STATE_REMOTE_CLOSE;
			/* We need this terminate to decrement the outstanding closes */
			/* For example, with sockets, Connection_terminate will close the socket */
			mpi_errno = MPIDI_CH3.Connection_terminate(vc);
		}
#endif
		else /* (vc->state == MPIDI_VC_STATE_ACTIVE) */
		{
			/* FIXME: Debugging */
			if (vc->state != MPIDI_VC_STATE_ACTIVE)
			{
				wprintf( L"Unexpected state %d in vc %p\n", vc->state, vc);
				fflush(stdout);
			}
			MPIU_DBG_MSG_D(CH3_DISCONNECT,TYPICAL, L"received close(FALSE) from %d, moving to REMOTE_CLOSE.", vc->pg_rank);
			MPIU_Assert(vc->state == MPIDI_VC_STATE_ACTIVE);
			MPIU_DBG_VCSTATECHANGE(vc,VC_STATE_REMOTE_CLOSE);
			vc->state = MPIDI_VC_STATE_REMOTE_CLOSE;
		}
	}
	else
	{
		MPIU_DBG_MSG_D(CH3_DISCONNECT,TYPICAL, L"received close(TRUE) from %d, moving to CLOSE_ACKED.", vc->pg_rank);
		MPIU_Assert (vc->state == MPIDI_VC_STATE_LOCAL_CLOSE || vc->state == MPIDI_VC_STATE_CLOSE_ACKED);
		MPIU_DBG_VCSTATECHANGE(vc,VC_STATE_CLOSE_ACKED);

		vc->state = MPIDI_VC_STATE_CLOSE_ACKED;
		/* For example, with sockets, Connection_terminate will close the socket */
		mpi_errno = MPIDI_CH3.Connection_terminate(vc);
	}
	
	*rreqp = NULL;

 fn_fail:
	return mpi_errno;
}

#ifdef MPICH_DBG_OUTPUT
int MPIDI_CH3_PktPrint_Close( FILE *fp, MPIDI_CH3_Pkt_t *pkt )
{
	MPIU_DBG_PRINTF((L" type ......... MPIDI_CH3_PKT_CLOSE\n"));
	MPIU_DBG_PRINTF((L" ack ......... %s\n", pkt->close.ack ? L"TRUE" : L"FALSE"));
	return MPI_SUCCESS;
}
#endif

/* 
 * This routine can be called to progress until all pending close operations
 * (initiated in the SendClose routine above) are completed.  It is 
 * used in MPID_Finalize and MPID_Comm_disconnect.
 */
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3U_VC_WaitForClose
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
/*@
  MPIDI_CH3U_VC_WaitForClose - Wait for all virtual connections to close
  @*/
int MPIDI_CH3U_VC_WaitForClose( void )
{
	MPID_Progress_state progress_state;
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3U_VC_WAITFORCLOSE);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3U_VC_WAITFORCLOSE);

	MPID_Progress_start(&progress_state);
	while(MPIDI_Outstanding_close_ops > 0)
	{
		MPIU_DBG_MSG_D(CH3_CONNECT,VERBOSE,L"Waiting for %d close operations", MPIDI_Outstanding_close_ops);
		mpi_errno = MPID_Progress_wait(&progress_state);
		/* --BEGIN ERROR HANDLING-- */
		if (mpi_errno != MPI_SUCCESS)
		{
			MPID_Progress_end(&progress_state);
			MPIU_ERR_SET(mpi_errno,MPI_ERR_OTHER,L"**ch3|close_progress");
			break;
		}
		/* --END ERROR HANDLING-- */
	}
	MPID_Progress_end(&progress_state);

	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3U_VC_WAITFORCLOSE);
	return mpi_errno;
}

#define set_request_info(rreq_, pkt_, msg_type_)				\
{																\
	(rreq_)->status.MPI_SOURCE = (pkt_)->match.rank;				\
	(rreq_)->status.MPI_TAG = (pkt_)->match.tag;				\
	(rreq_)->status.count = (int)(pkt_)->data_sz;				\
	(rreq_)->dev.sender_req_id = (pkt_)->sender_req_id;				\
	(rreq_)->dev.recv_data_sz = (MPIDI_msg_sz_t)(pkt_)->data_sz;\
	MPIDI_Request_set_msg_type((rreq_), (msg_type_));				\
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3U_Handle_recv_rndv_pkt
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3U_Handle_recv_rndv_pkt(MPIDI_VC_t * vc, MPIDI_CH3_Pkt_t * pkt, MPID_Request ** rreqp, int *foundp)
{
	int mpi_errno = MPI_SUCCESS;
	MPID_Request *rreq;
	MPIDI_CH3_Pkt_rndv_req_to_send_t * rts_pkt = &pkt->rndv_req_to_send;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3U_HANDLE_RECV_RNDV_PKT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3U_HANDLE_RECV_RNDV_PKT);

	MPIU_UNREFERENCED_ARG(vc);

	rreq = MPIDI_CH3U_Recvq_FDP_or_AEU(&rts_pkt->match, foundp);
	if (rreq == NULL)
	{
		MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, L"**nomemreq");
	}

	set_request_info(rreq, rts_pkt, MPIDI_REQUEST_RNDV_MSG);

	if (!*foundp)
	{
		MPIU_DBG_MSG(CH3_OTHER,VERBOSE,L"unexpected request allocated");

		/*
		 * An MPID_Probe() may be waiting for the request we just inserted, 
		 * so we need to tell the progress engine to exit.
		 *
		 * FIXME: This will cause MPID_Progress_wait() to return to the MPI 
		 * layer each time an unexpected RTS packet is
		 * received.  MPID_Probe() should atomically increment a counter and 
		 * MPIDI_CH3_Progress_signal_completion()
		 * should only be called if that counter is greater than zero.
		 */
		MPIDI_CH3_Progress_signal_completion();
	}

	/* return the request */
	*rreqp = rreq;

 fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3U_HANDLE_RECV_RNDV_PKT);
	return mpi_errno;
}

/*
 * MPIDI_CH3U_Handle_recv_pkt()
 *
 * NOTE: Multiple threads may NOT simultaneously call this routine with the same VC.  This constraint eliminates the need to
 * lock the VC.  If simultaneous upcalls are a possible, the calling routine is responsible for serializing the calls.
 */

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3U_Handle_recv_pkt
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3U_Handle_recv_pkt(MPIDI_VC_t * vc, MPIDI_CH3_Pkt_t * pkt, MPID_Request ** rreqp)
{
	int type_size;
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3U_HANDLE_ORDERED_RECV_PKT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3U_HANDLE_ORDERED_RECV_PKT);

	MPIU_DBG_STMT(CH3_OTHER,VERBOSE,MPIDI_DBG_Print_packet(pkt));

	switch(pkt->type)
	{
		/* FIXME: This is not optimized for short messages, which 
		   should have the data in the same packet when the data is
		   particularly short (e.g., one 8 byte long word) */
		case MPIDI_CH3_PKT_EAGER_SEND:
		{
			MPIDI_CH3_Pkt_eager_send_t * eager_pkt = &pkt->eager_send;
			MPID_Request * rreq;
			int found;

			MPIU_DBG_MSG_FMT(CH3_OTHER,VERBOSE,(MPIU_DBG_FDEST, L"received eager send pkt, sreq=0x%08x, rank=%d, tag=%d, context=%d", eager_pkt->sender_req_id, eager_pkt->match.rank, eager_pkt->match.tag, eager_pkt->match.context_id));
			
			rreq = MPIDI_CH3U_Recvq_FDP_or_AEU(&eager_pkt->match, &found);
			if (rreq == NULL)
			{
				MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, L"**nomemreq");
			}
			
			set_request_info(rreq, eager_pkt, MPIDI_REQUEST_EAGER_MSG);
			*rreqp = rreq;
			/* FIXME: What is the logic here?  On an eager receive, the data
			   should be available already, and we should be optimizing
			   for short messages */
			mpi_errno = MPIDI_CH3U_Post_data_receive(found, rreqp);
			if (mpi_errno != MPI_SUCCESS)
			{
				MPIU_ERR_SETANDJUMP1(mpi_errno,MPI_ERR_OTHER, L"**ch3|postrecv", L"**ch3|postrecv %s", L"MPIDI_CH3_PKT_EAGER_SEND");
			}
			break;
		}
		
		case MPIDI_CH3_PKT_READY_SEND:
		{
			MPIDI_CH3_Pkt_ready_send_t * ready_pkt = &pkt->ready_send;
			MPID_Request * rreq;
			int found;
			
			MPIU_DBG_MSG_FMT(CH3_OTHER,VERBOSE,(MPIU_DBG_FDEST, L"received ready send pkt, sreq=0x%08x, rank=%d, tag=%d, context=%d", ready_pkt->sender_req_id, ready_pkt->match.rank, ready_pkt->match.tag, ready_pkt->match.context_id));
			
			rreq = MPIDI_CH3U_Recvq_FDP_or_AEU(&ready_pkt->match, &found);
			if (rreq == NULL)
			{
				MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, L"**nomemreq");
			}
			
			set_request_info(rreq, ready_pkt, MPIDI_REQUEST_EAGER_MSG);
			*rreqp = rreq;
			if (found)
			{
				mpi_errno = MPIDI_CH3U_Post_data_receive(TRUE, rreqp);
				if (mpi_errno != MPI_SUCCESS)
				{
					MPIU_ERR_SETANDJUMP1(mpi_errno,MPI_ERR_OTHER, L"**ch3|postrecv", L"**ch3|postrecv %s", L"MPIDI_CH3_PKT_READY_SEND");
				}
			}
			else
			{
				/* FIXME: an error packet should be sent back to the sender 
				   indicating that the ready-send failed.  On the send
				   side, the error handler for the communicator can be invoked
				   even if the ready-send request has already
				   completed. */

				/* We need to consume any outstanding associated data and 
				   mark the request with an error. */

				rreq->status.MPI_ERROR = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**rsendnomatch", L"**rsendnomatch %d %d", ready_pkt->match.rank, ready_pkt->match.tag);
				rreq->status.count = 0;
				if (rreq->dev.recv_data_sz > 0)
				{
					 /* force read of extra data */
					rreq->dev.segment_first = 0;
					rreq->dev.segment_size = 0;
					mpi_errno = MPIDI_CH3U_Request_load_recv_iov(rreq);
					if (mpi_errno != MPI_SUCCESS)
					{
						MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, L"**ch3|loadrecviov");
					}
				}
				else
				{
					/* mark data transfer as complete and decrement CC */
					MPIDI_CH3U_Request_complete(rreq);
					*rreqp = NULL;
				}
			}
			break;
		}
		
		case MPIDI_CH3_PKT_EAGER_SYNC_SEND:
		{
			MPIDI_CH3_Pkt_eager_send_t * es_pkt = &pkt->eager_send;
			MPID_Request * rreq;
			int found;

			MPIU_DBG_MSG_FMT(CH3_OTHER,VERBOSE,(MPIU_DBG_FDEST, L"received eager sync send pkt, sreq=0x%08x, rank=%d, tag=%d, context=%d", es_pkt->sender_req_id, es_pkt->match.rank, es_pkt->match.tag, es_pkt->match.context_id));
			
			rreq = MPIDI_CH3U_Recvq_FDP_or_AEU(&es_pkt->match, &found);
			if (rreq == NULL)
			{
				MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, L"**nomemreq");
			}
			
			set_request_info(rreq, es_pkt, MPIDI_REQUEST_EAGER_MSG);
			*rreqp = rreq;
			mpi_errno = MPIDI_CH3U_Post_data_receive(found, rreqp);
			if (mpi_errno != MPI_SUCCESS)
			{
				MPIU_ERR_SETANDJUMP1(mpi_errno,MPI_ERR_OTHER, L"**ch3|postrecv", L"**ch3|postrecv %s", L"MPIDI_CH3_PKT_EAGER_SYNC_SEND");
			}
			
			if (found)
			{
				MPIDI_CH3_Pkt_t upkt;
				MPIDI_CH3_Pkt_eager_sync_ack_t * const esa_pkt = &upkt.eager_sync_ack;
				MPID_Request * esa_req;
					
				MPIU_DBG_MSG(CH3_OTHER,VERBOSE,L"sending eager sync ack");
						
				MPIDI_Pkt_init(esa_pkt, MPIDI_CH3_PKT_EAGER_SYNC_ACK);
				esa_pkt->sender_req_id = rreq->dev.sender_req_id;
				mpi_errno = MPIDI_CH3.iStartMsg(vc, esa_pkt, sizeof(*esa_pkt), &esa_req);
				if (mpi_errno != MPI_SUCCESS)
				{
					MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, L"**ch3|syncack");
				}
				if (esa_req != NULL)
				{
					MPID_Request_release(esa_req);
				}
			}
			else
			{
				MPIDI_Request_set_sync_send_flag(rreq, TRUE);
			}
			
			break;
		}

		case MPIDI_CH3_PKT_EAGER_SYNC_ACK:
		{
			MPIDI_CH3_Pkt_eager_sync_ack_t * esa_pkt = &pkt->eager_sync_ack;
			MPID_Request * sreq;
			
			MPIU_DBG_MSG_P(CH3_OTHER,VERBOSE, L"received eager sync ack pkt, sreq=0x%08x", esa_pkt->sender_req_id);
			
			MPID_Request_get_ptr(esa_pkt->sender_req_id, sreq);
			/* decrement CC (but don't mark data transfer as complete since the transfer could still be in progress) */
			MPIDI_CH3U_Request_complete(sreq);  /* brad : seen this segfault in ssm dynamic process...? */

			*rreqp = NULL;
			break;
		}
		
		case MPIDI_CH3_PKT_RNDV_REQ_TO_SEND:
		{
			MPID_Request * rreq;
			int found;
			MPIDI_CH3_Pkt_rndv_req_to_send_t * rts_pkt = &pkt->rndv_req_to_send;

			MPIU_DBG_MSG_FMT(CH3_OTHER,VERBOSE,(MPIU_DBG_FDEST, L"received rndv RTS pkt, sreq=0x%08x, rank=%d, tag=%d, context=%d, data_sz=%d", rts_pkt->sender_req_id, rts_pkt->match.rank, rts_pkt->match.tag, rts_pkt->match.context_id, rts_pkt->data_sz));

			rreq = MPIDI_CH3U_Recvq_FDP_or_AEU(&rts_pkt->match, &found);
			if (rreq == NULL)
			{
				MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, L"**nomemreq");
			}

			set_request_info(rreq, rts_pkt, MPIDI_REQUEST_RNDV_MSG);

			if (found)
			{
				MPID_Request * cts_req;
				MPIDI_CH3_Pkt_t upkt;
				MPIDI_CH3_Pkt_rndv_clr_to_send_t * cts_pkt = &upkt.rndv_clr_to_send;

				MPIU_DBG_MSG(CH3_OTHER,VERBOSE,L"posted request found");

				/* FIXME: What if the receive user buffer is not big enough to hold the data about to be cleared for sending? */

				MPIU_DBG_MSG(CH3_OTHER,VERBOSE,L"sending rndv CTS packet");
				MPIDI_Pkt_init(cts_pkt, MPIDI_CH3_PKT_RNDV_CLR_TO_SEND);
				cts_pkt->sender_req_id = rts_pkt->sender_req_id;
				cts_pkt->receiver_req_id = rreq->handle;
				mpi_errno = MPIDI_CH3.iStartMsg(vc, cts_pkt, sizeof(*cts_pkt), &cts_req);
				if (mpi_errno != MPI_SUCCESS)
				{
					MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, L"**ch3|ctspkt");
				}
				if (cts_req != NULL)
				{
					MPID_Request_release(cts_req);
				}
			}
			else
			{
				MPIU_DBG_MSG(CH3_OTHER,VERBOSE,L"unexpected request allocated");

				/*
				* A MPID_Probe() may be waiting for the request we just 
				* inserted, so we need to tell the progress engine to exit.
				*
				* FIXME: This will cause MPID_Progress_wait() to return to the
				* MPI layer each time an unexpected RTS packet is
				* received.  MPID_Probe() should atomically increment a
				* counter and MPIDI_CH3_Progress_signal_completion()
				* should only be called if that counter is greater than zero.
				*/
				MPIDI_CH3_Progress_signal_completion();
			}

			*rreqp = NULL;
			break;
		}
		
		case MPIDI_CH3_PKT_RNDV_CLR_TO_SEND:
		{
			MPIDI_CH3_Pkt_rndv_clr_to_send_t * cts_pkt = &pkt->rndv_clr_to_send;
			MPID_Request * sreq;
			MPID_Request * rts_sreq;
			MPIDI_CH3_Pkt_t upkt;
			MPIDI_CH3_Pkt_rndv_send_t * rs_pkt = &upkt.rndv_send;
			int dt_contig;
			MPI_Aint dt_true_lb;
			MPIDI_msg_sz_t data_sz;
			MPID_Datatype * dt_ptr;
			MPID_IOV iov[MPID_IOV_LIMIT];
			int iov_n;
			int mpi_errno = MPI_SUCCESS;

			MPIU_DBG_MSG(CH3_OTHER,VERBOSE,L"received rndv CTS pkt");

			MPID_Request_get_ptr(cts_pkt->sender_req_id, sreq);
			MPIU_DBG_PRINTF((L"received cts, count=%d\n", sreq->dev.user_count));

			/* Release the RTS request if one exists.  
			   MPID_Request_fetch_and_clear_rts_sreq() needs to be atomic to 
			   prevent
			   cancel send from cancelling the wrong (future) request.  
			   If MPID_Request_fetch_and_clear_rts_sreq() returns a NULL
			   rts_sreq, then MPID_Cancel_send() is responsible for releasing 
			   the RTS request object. */
			MPIDI_Request_fetch_and_clear_rts_sreq(sreq, &rts_sreq);
			if (rts_sreq != NULL)
			{
				MPID_Request_release(rts_sreq);
			}
			
			MPIDI_Pkt_init(rs_pkt, MPIDI_CH3_PKT_RNDV_SEND);
			rs_pkt->receiver_req_id = cts_pkt->receiver_req_id;
			iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)rs_pkt;
			iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)sizeof(*rs_pkt);

			MPIDI_Datatype_get_info(sreq->dev.user_count, sreq->dev.datatype, dt_contig, data_sz, dt_ptr, dt_true_lb);
		
			if (dt_contig) 
			{
				MPIU_DBG_MSG_FMT(CH3_OTHER,VERBOSE,(MPIU_DBG_FDEST, L"sending contiguous rndv data, data_sz=" MPIDI_MSG_SZ_FMT, data_sz));
				
				sreq->dev.ca = MPIDI_CH3_CA_COMPLETE;
				
				iov[1].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)((char *)sreq->dev.user_buf + dt_true_lb);
				iov[1].MPID_IOV_LEN = (MPID_IOV_LEN_CAST) data_sz;
				iov_n = 2;
			}
			else
			{
				MPID_Segment_init(sreq->dev.user_buf, sreq->dev.user_count, sreq->dev.datatype, &sreq->dev.segment, 0);
				iov_n = MPID_IOV_LIMIT - 1;
				sreq->dev.segment_first = 0;
				sreq->dev.segment_size = data_sz;
				mpi_errno = MPIDI_CH3U_Request_load_send_iov(sreq, &iov[1], &iov_n);
				if (mpi_errno != MPI_SUCCESS)
				{
					MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, L"**ch3|loadsendiov");
				}
				iov_n += 1;
			}
			
			mpi_errno = MPIDI_CH3.iSendv(vc, sreq, iov, iov_n);
			if (mpi_errno != MPI_SUCCESS)
			{
				MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, L"**ch3|senddata");
			}
			
			*rreqp = NULL;
			break;
		}
		
		case MPIDI_CH3_PKT_RNDV_SEND:
		{
			MPIDI_CH3_Pkt_rndv_send_t * rs_pkt = &pkt->rndv_send;
					
			MPIU_DBG_MSG(CH3_OTHER,VERBOSE,L"received rndv send (data) pkt");
			MPID_Request_get_ptr(rs_pkt->receiver_req_id, *rreqp);
			mpi_errno = MPIDI_CH3U_Post_data_receive(TRUE, rreqp);
			if (mpi_errno != MPI_SUCCESS)
			{
				MPIU_ERR_SETANDJUMP1(mpi_errno,MPI_ERR_OTHER, L"**ch3|postrecv", L"**ch3|postrecv %s", L"MPIDI_CH3_PKT_RNDV_SEND");
			}
			break;
		}
		
		case MPIDI_CH3_PKT_CANCEL_SEND_REQ:
		{
			MPIDI_CH3_Pkt_cancel_send_req_t * req_pkt = &pkt->cancel_send_req;
			MPID_Request * rreq;
			int ack;
			MPIDI_CH3_Pkt_t upkt;
			MPIDI_CH3_Pkt_cancel_send_resp_t * resp_pkt = &upkt.cancel_send_resp;
			MPID_Request * resp_sreq;

			MPIU_DBG_MSG_FMT(CH3_OTHER,VERBOSE,(MPIU_DBG_FDEST, L"received cancel send req pkt, sreq=0x%08x, rank=%d, tag=%d, context=%d", req_pkt->sender_req_id, req_pkt->match.rank, req_pkt->match.tag, req_pkt->match.context_id));
			
			rreq = MPIDI_CH3U_Recvq_FDU(req_pkt->sender_req_id, &req_pkt->match);
			if (rreq != NULL)
			{
				MPIU_DBG_MSG(CH3_OTHER,VERBOSE,L"message cancelled");
				if (MPIDI_Request_get_msg_type(rreq) == MPIDI_REQUEST_EAGER_MSG && rreq->dev.recv_data_sz > 0)
				{
					MPIU_Free(rreq->dev.tmpbuf);
				}
				MPID_Request_release(rreq);
				ack = TRUE;
			}
			else
			{
				MPIU_DBG_MSG(CH3_OTHER,TYPICAL,L"unable to cancel message");
				ack = FALSE;
			}
			
			MPIDI_Pkt_init(resp_pkt, MPIDI_CH3_PKT_CANCEL_SEND_RESP);
			resp_pkt->sender_req_id = req_pkt->sender_req_id;
			resp_pkt->ack = ack;
			mpi_errno = MPIDI_CH3.iStartMsg(vc, resp_pkt, sizeof(*resp_pkt), &resp_sreq);
			if (mpi_errno != MPI_SUCCESS)
			{
				MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, L"**ch3|cancelresp");
			}
			if (resp_sreq != NULL)
			{
				MPID_Request_release(resp_sreq);
			}

			*rreqp = NULL;
			break;
		}
		
		case MPIDI_CH3_PKT_CANCEL_SEND_RESP:
		{
			MPIDI_CH3_Pkt_cancel_send_resp_t * resp_pkt = &pkt->cancel_send_resp;
			MPID_Request * sreq;

			MPIU_DBG_MSG_FMT(CH3_OTHER,VERBOSE,(MPIU_DBG_FDEST, L"received cancel send resp pkt, sreq=0x%08x, ack=%d", resp_pkt->sender_req_id, resp_pkt->ack));
			
			MPID_Request_get_ptr(resp_pkt->sender_req_id, sreq);
			
			if (resp_pkt->ack)
			{
				sreq->status.cancelled = TRUE;
				
				if (MPIDI_Request_get_msg_type(sreq) == MPIDI_REQUEST_RNDV_MSG ||
					MPIDI_Request_get_type(sreq) == MPIDI_REQUEST_TYPE_SSEND)
				{
					int cc;
					
					/* decrement the CC one additional time for the CTS/sync ack that is never going to arrive */
					MPIDI_CH3U_Request_decrement_cc(sreq, &cc);
				}
				
				MPIU_DBG_MSG(CH3_OTHER,VERBOSE,L"message cancelled");
			}
			else
			{
				MPIU_DBG_MSG(CH3_OTHER,VERBOSE,L"unable to cancel message");
			}

			MPIDI_CH3U_Request_complete(sreq);

			*rreqp = NULL;
			break;
		}
		
		case MPIDI_CH3_PKT_PUT:
		{
			MPIDI_CH3_Pkt_put_t * put_pkt = &pkt->put;
			MPID_Request *req;
			int predefined;

			MPIU_DBG_MSG(CH3_OTHER,VERBOSE,L"received put pkt");

			if (put_pkt->count == 0)
			{
				MPID_Win *win_ptr;

				/* it's a 0-byte message sent just to decrement the
				   completion counter. This happens only in
				   post/start/complete/wait sync model; therefore, no need
				   to check lock queue. */
				if (put_pkt->target_win_handle != MPI_WIN_NULL)
				{
					MPID_Win_get_ptr(put_pkt->target_win_handle, win_ptr);
					/* FIXME: MT: this has to be done atomically */
					win_ptr->my_counter -= 1;
				}
				MPIDI_CH3_Progress_signal_completion();		
				*rreqp = NULL;
			}
			else
			{
				req = MPID_Request_create();
				MPIU_Object_set_ref(req, 1);
				
				req->dev.user_buf = MPIDI_Message_pointer_t_to_local_ptr(put_pkt->addr);
				req->dev.user_count = put_pkt->count;
				req->dev.target_win_handle = put_pkt->target_win_handle;
				req->dev.source_win_handle = put_pkt->source_win_handle;

				MPIDI_CH3I_DATATYPE_IS_PREDEFINED(put_pkt->datatype, predefined);
				if (predefined)
				{
					MPIDI_Request_set_type(req, MPIDI_REQUEST_TYPE_PUT_RESP);
					req->dev.datatype = put_pkt->datatype;
				
					MPID_Datatype_get_size_macro(put_pkt->datatype, type_size);
					req->dev.recv_data_sz = type_size * put_pkt->count;

					*rreqp = req;
					mpi_errno = MPIDI_CH3U_Post_data_receive(TRUE, rreqp);
				}
				else
				{
					/* derived datatype */
					MPIDI_Request_set_type(req, MPIDI_REQUEST_TYPE_PUT_RESP_DERIVED_DT);
					req->dev.datatype = MPI_DATATYPE_NULL;

					if (put_pkt->dataloop_size < 0)
					{
						req->dev.dtype_info = NULL;
						req->dev.dataloop = NULL;
						req->dev.serialized_datatype_buffer = MPIU_Malloc(-put_pkt->dataloop_size);
						if (! req->dev.serialized_datatype_buffer)
						{
							MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER,L"**nomem");
						}

						req->dev.iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)req->dev.serialized_datatype_buffer;
						req->dev.iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)-put_pkt->dataloop_size;
						req->dev.iov_count = 1;
					}
					else
					{
						req->dev.dtype_info = (MPIDI_RMA_dtype_info *)MPIU_Malloc(sizeof(MPIDI_RMA_dtype_info));
						if (! req->dev.dtype_info)
						{
							MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER,L"**nomem");
						}

						req->dev.dataloop = MPIU_Malloc(put_pkt->dataloop_size);
						if (! req->dev.dataloop)
						{
							MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER,L"**nomem");
						}

						req->dev.iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)req->dev.dtype_info;
						req->dev.iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)sizeof(MPIDI_RMA_dtype_info);
						req->dev.iov[1].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)req->dev.dataloop;
						req->dev.iov[1].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)put_pkt->dataloop_size;
						req->dev.iov_count = 2;
					}

					req->dev.ca = MPIDI_CH3_CA_COMPLETE;

					*rreqp = req;
				}

				/* --BEGIN ERROR HANDLING-- */
				if (mpi_errno != MPI_SUCCESS)
				{
					mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|postrecv", L"**ch3|postrecv %s", L"MPIDI_CH3_PKT_PUT");
				}
				/* --END ERROR HANDLING-- */

			}
			break;
		}
		
		case MPIDI_CH3_PKT_ACCUMULATE:
		{
			MPIDI_CH3_Pkt_accum_t * accum_pkt = &pkt->accum;
			MPID_Request *req;
			MPI_Aint true_lb, true_extent, extent;
			void *tmp_buf;
			int predefined;

			MPIU_DBG_MSG(CH3_OTHER,VERBOSE,L"received accumulate pkt");

			req = MPID_Request_create();
			MPIU_Object_set_ref(req, 1);
			*rreqp = req;

			req->dev.user_count = accum_pkt->count;
			req->dev.op = accum_pkt->op;
			req->dev.real_user_buf = MPIDI_Message_pointer_t_to_local_ptr(accum_pkt->addr);
			req->dev.target_win_handle = accum_pkt->target_win_handle;
			req->dev.source_win_handle = accum_pkt->source_win_handle;

			MPIDI_CH3I_DATATYPE_IS_PREDEFINED(accum_pkt->datatype, predefined);
			if (predefined)
			{
				MPIU_THREADPRIV_DECL;
				MPIU_THREADPRIV_GET;
				MPIDI_Request_set_type(req, MPIDI_REQUEST_TYPE_ACCUM_RESP);
				req->dev.datatype = accum_pkt->datatype;

				NMPI_Type_get_true_extent(mpi_errno, (accum_pkt->datatype, &true_lb, &true_extent));
				/* --BEGIN ERROR HANDLING-- */
				if (mpi_errno)
				{
					MPIR_Err_create_code(mpi_errno , MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", 0);
					return mpi_errno;
				}
				/* --END ERROR HANDLING-- */

				MPID_Datatype_get_extent_macro(accum_pkt->datatype, extent); 
				tmp_buf = MPIU_Malloc(accum_pkt->count * (MPIR_MAX(extent,true_extent)));
				/* --BEGIN ERROR HANDLING-- */
				if (!tmp_buf)
				{
					mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0 );
					return mpi_errno;
				}
				/* --END ERROR HANDLING-- */

				/* adjust for potential negative lower bound in datatype */
				tmp_buf = (void *)((char*)tmp_buf - true_lb);

				req->dev.user_buf = tmp_buf;

				MPID_Datatype_get_size_macro(accum_pkt->datatype, type_size);
				req->dev.recv_data_sz = type_size * accum_pkt->count;
				
				mpi_errno = MPIDI_CH3U_Post_data_receive(TRUE, rreqp);
			}
			else
			{
				MPIDI_Request_set_type(req, MPIDI_REQUEST_TYPE_ACCUM_RESP_DERIVED_DT);
				req->dev.datatype = MPI_DATATYPE_NULL;

				if (accum_pkt->dataloop_size < 0)
				{
					req->dev.dtype_info = NULL;
					req->dev.dataloop = NULL;
					req->dev.serialized_datatype_buffer = MPIU_Malloc(-accum_pkt->dataloop_size);
					if (! req->dev.serialized_datatype_buffer)
					{
						MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER,L"**nomem" );
					}

					req->dev.iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)req->dev.serialized_datatype_buffer;
					req->dev.iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)-accum_pkt->dataloop_size;
					req->dev.iov_count = 1;
					req->dev.ca = MPIDI_CH3_CA_COMPLETE;
				}
				else
				{
					req->dev.dtype_info = (MPIDI_RMA_dtype_info *)MPIU_Malloc(sizeof(MPIDI_RMA_dtype_info));
					if (! req->dev.dtype_info)
					{
						MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER,L"**nomem" );
					}

					req->dev.dataloop = MPIU_Malloc(accum_pkt->dataloop_size);
					if (! req->dev.dataloop)
					{
						MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER,L"**nomem" );
					}

					req->dev.iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)req->dev.dtype_info;
					req->dev.iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)sizeof(MPIDI_RMA_dtype_info);
					req->dev.iov[1].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)req->dev.dataloop;
					req->dev.iov[1].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)accum_pkt->dataloop_size;
					req->dev.iov_count = 2;
					req->dev.ca = MPIDI_CH3_CA_COMPLETE;
				}
			}

			if (mpi_errno != MPI_SUCCESS)
			{
				MPIU_ERR_SETANDJUMP1(mpi_errno,MPI_ERR_OTHER,L"**ch3|postrecv", L"**ch3|postrecv %s", L"MPIDI_CH3_PKT_ACCUMULATE");
			}
			break;
		}

		case MPIDI_CH3_PKT_GET:
		{
			MPIDI_CH3_Pkt_get_t * get_pkt = &pkt->get;
			MPID_Request *req;
			MPID_IOV iov[MPID_IOV_LIMIT];
			int predefined;

			MPIU_DBG_MSG(CH3_OTHER,VERBOSE,L"received get pkt");

			req = MPID_Request_create();
			req->dev.target_win_handle = get_pkt->target_win_handle;
			req->dev.source_win_handle = get_pkt->source_win_handle;
			req->dev.ca = MPIDI_CH3_CA_COMPLETE;

			MPIDI_CH3I_DATATYPE_IS_PREDEFINED(get_pkt->datatype, predefined);
			if (predefined)
			{
				/* basic datatype. send the data. */
				MPIDI_CH3_Pkt_t upkt;
				MPIDI_CH3_Pkt_get_resp_t * get_resp_pkt = &upkt.get_resp;

				MPIDI_Request_set_type(req, MPIDI_REQUEST_TYPE_GET_RESP); 
				req->kind = MPID_REQUEST_SEND;

				MPIDI_Pkt_init(get_resp_pkt, MPIDI_CH3_PKT_GET_RESP);
				get_resp_pkt->request_handle = get_pkt->request_handle;
				
				iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST) get_resp_pkt;
				iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)sizeof(*get_resp_pkt);

				iov[1].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)MPIDI_Message_pointer_t_to_local_ptr(get_pkt->addr);
				MPID_Datatype_get_size_macro(get_pkt->datatype, type_size);
				iov[1].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)(get_pkt->count * type_size);
			
				mpi_errno = MPIDI_CH3.iSendv(vc, req, iov, 2);
				/* --BEGIN ERROR HANDLING-- */
				if (mpi_errno != MPI_SUCCESS)
				{
					MPIU_Object_set_ref(req, 0);
					MPIDI_CH3_Request_destroy(req);
					mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|rmamsg", 0);
					return mpi_errno;
				}
				/* --END ERROR HANDLING-- */

				*rreqp = NULL;
			}
			else
			{
				/* derived datatype. first get the dtype_info and dataloop. */

				MPIDI_Request_set_type(req, MPIDI_REQUEST_TYPE_GET_RESP_DERIVED_DT);

				req->dev.user_buf = MPIDI_Message_pointer_t_to_local_ptr(get_pkt->addr);
				req->dev.user_count = get_pkt->count;
				req->dev.datatype = MPI_DATATYPE_NULL;
				req->dev.request_handle = get_pkt->request_handle;

				if (get_pkt->dataloop_size < 0)
				{
					req->dev.dtype_info = NULL;
					req->dev.dataloop = NULL;
					req->dev.serialized_datatype_buffer = MPIU_Malloc(-get_pkt->dataloop_size);
					if (! req->dev.serialized_datatype_buffer)
					{
						MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER,L"**nomem" );
					}

					req->dev.iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)req->dev.serialized_datatype_buffer;
					req->dev.iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)-get_pkt->dataloop_size;
					req->dev.iov_count = 1;
				}
				else
				{
					req->dev.dtype_info = (MPIDI_RMA_dtype_info *)MPIU_Malloc(sizeof(MPIDI_RMA_dtype_info));
					if (! req->dev.dtype_info)
					{
						MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER,L"**nomem" );
					}

					req->dev.dataloop = MPIU_Malloc(get_pkt->dataloop_size);
					if (! req->dev.dataloop)
					{
						MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER,L"**nomem" );
					}

					req->dev.iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)req->dev.dtype_info;
					req->dev.iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)sizeof(MPIDI_RMA_dtype_info);
					req->dev.iov[1].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)req->dev.dataloop;
					req->dev.iov[1].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)get_pkt->dataloop_size;
					req->dev.iov_count = 2;
				}

				*rreqp = req;
			}

			break;
		}

		case MPIDI_CH3_PKT_GET_RESP:
		{
			MPIDI_CH3_Pkt_get_resp_t * get_resp_pkt = &pkt->get_resp;
			MPID_Request *req;

			MPIU_DBG_MSG(CH3_OTHER,VERBOSE,L"received get response pkt");

			MPID_Request_get_ptr(get_resp_pkt->request_handle, req);

			MPID_Datatype_get_size_macro(req->dev.datatype, type_size);
			req->dev.recv_data_sz = type_size * req->dev.user_count;

			*rreqp = req;
			mpi_errno = MPIDI_CH3U_Post_data_receive(TRUE, rreqp);
			/* --BEGIN ERROR HANDLING-- */
			if (mpi_errno != MPI_SUCCESS)
			{
				mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|postrecv", L"**ch3|postrecv %s", L"MPIDI_CH3_PKT_GET_RESP");
			}
			/* --END ERROR HANDLING-- */

			break;
		}

		case MPIDI_CH3_PKT_LOCK:
		{
			MPIDI_CH3_Pkt_lock_t * lock_pkt = &pkt->lock;
			MPID_Win *win_ptr;

			MPIU_DBG_MSG(CH3_OTHER,VERBOSE,L"received lock pkt");

			MPID_Win_get_ptr(lock_pkt->target_win_handle, win_ptr);

			if (MPIDI_CH3I_Try_acquire_win_lock(win_ptr, lock_pkt->lock_type) == 1)
			{
				/* send lock granted packet. */
				mpi_errno = MPIDI_CH3I_Send_lock_granted_pkt(vc, lock_pkt->source_win_handle);
			}
			else
			{
				/* queue the lock information */
				MPIDI_Win_lock_queue *curr_ptr, *prev_ptr, *new_ptr;

				/* FIXME: MT: This may need to be done atomically. */

				curr_ptr = (MPIDI_Win_lock_queue *) win_ptr->lock_queue;
				prev_ptr = curr_ptr;
				while (curr_ptr != NULL)
				{
					prev_ptr = curr_ptr;
					curr_ptr = curr_ptr->next;
				}

				new_ptr = (MPIDI_Win_lock_queue *) MPIU_Malloc(sizeof(MPIDI_Win_lock_queue));
				if (!new_ptr)
				{
					MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER,L"**nomem" );
				}
				if (prev_ptr != NULL)
					prev_ptr->next = new_ptr;
				else 
					win_ptr->lock_queue = new_ptr;
		
				new_ptr->next = NULL;  
				new_ptr->lock_type = lock_pkt->lock_type;
				new_ptr->source_win_handle = lock_pkt->source_win_handle;
				new_ptr->vc = vc;
				new_ptr->pt_single_op = NULL;
			}

			*rreqp = NULL;

			break;
		}

		case MPIDI_CH3_PKT_LOCK_GRANTED:
		{
			MPIDI_CH3_Pkt_lock_granted_t * lock_granted_pkt = &pkt->lock_granted;
			MPID_Win *win_ptr;
			MPIU_DBG_MSG(CH3_OTHER,VERBOSE,L"received lock granted pkt");

			MPID_Win_get_ptr(lock_granted_pkt->source_win_handle, win_ptr);
			/* set the lock_granted flag in the window */
			win_ptr->lock_granted = 1;

			*rreqp = NULL;
			MPIDI_CH3_Progress_signal_completion();		

			break;
		}

		case MPIDI_CH3_PKT_PT_RMA_DONE:
		{
			MPIDI_CH3_Pkt_pt_rma_done_t * pt_rma_done_pkt = &pkt->pt_rma_done;
			MPID_Win *win_ptr;

			MPIU_DBG_MSG(CH3_OTHER,VERBOSE,L"received shared lock ops done pkt");

			MPID_Win_get_ptr(pt_rma_done_pkt->source_win_handle, win_ptr);
			/* reset the lock_granted flag in the window */
			win_ptr->lock_granted = 0;

			*rreqp = NULL;
			MPIDI_CH3_Progress_signal_completion();		

			break;
		}

		case MPIDI_CH3_PKT_LOCK_PUT_UNLOCK:
		{
			MPIDI_CH3_Pkt_lock_put_unlock_t * lock_put_unlock_pkt = &pkt->lock_put_unlock;
			MPID_Win *win_ptr;
			MPID_Request *req;

			MPIU_DBG_MSG(CH3_OTHER,VERBOSE,L"received lock_put_unlock pkt");

			req = MPID_Request_create();
			MPIU_Object_set_ref(req, 1);
				
			req->dev.datatype = lock_put_unlock_pkt->datatype;
			MPID_Datatype_get_size_macro(lock_put_unlock_pkt->datatype, type_size);
			req->dev.recv_data_sz = type_size * lock_put_unlock_pkt->count;
			req->dev.user_count = lock_put_unlock_pkt->count;
			req->dev.target_win_handle = lock_put_unlock_pkt->target_win_handle;
			
			MPID_Win_get_ptr(lock_put_unlock_pkt->target_win_handle, win_ptr);

			if (MPIDI_CH3I_Try_acquire_win_lock(win_ptr, lock_put_unlock_pkt->lock_type) == 1)
			{
				/* do the put. for this optimization, only basic datatypes supported. */
				MPIDI_Request_set_type(req, MPIDI_REQUEST_TYPE_PUT_RESP);
				req->dev.user_buf = MPIDI_Message_pointer_t_to_local_ptr(lock_put_unlock_pkt->addr);
				req->dev.source_win_handle = lock_put_unlock_pkt->source_win_handle;
				req->dev.single_op_opt = 1;
			}
			else
			{
				/* queue the information */
				MPIDI_Win_lock_queue *curr_ptr, *prev_ptr, *new_ptr;

				new_ptr = (MPIDI_Win_lock_queue *) MPIU_Malloc(sizeof(MPIDI_Win_lock_queue));
				if (!new_ptr)
				{
					MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER,L"**nomem" );
				}

				new_ptr->pt_single_op = (MPIDI_PT_single_op *) MPIU_Malloc(sizeof(MPIDI_PT_single_op));
				if (new_ptr->pt_single_op == NULL)
				{
					MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER,L"**nomem" );
				}

				/* FIXME: MT: The queuing may need to be done atomically. */

				curr_ptr = (MPIDI_Win_lock_queue *) win_ptr->lock_queue;
				prev_ptr = curr_ptr;
				while (curr_ptr != NULL)
				{
					prev_ptr = curr_ptr;
					curr_ptr = curr_ptr->next;
				}

				if (prev_ptr != NULL)
					prev_ptr->next = new_ptr;
				else 
					win_ptr->lock_queue = new_ptr;
		
				new_ptr->next = NULL;  
				new_ptr->lock_type = lock_put_unlock_pkt->lock_type;
				new_ptr->source_win_handle = lock_put_unlock_pkt->source_win_handle;
				new_ptr->vc = vc;
				
				new_ptr->pt_single_op->type = MPIDI_RMA_PUT;
				new_ptr->pt_single_op->addr = MPIDI_Message_pointer_t_to_local_ptr(lock_put_unlock_pkt->addr);
				new_ptr->pt_single_op->count = lock_put_unlock_pkt->count;
				new_ptr->pt_single_op->datatype = lock_put_unlock_pkt->datatype;
				/* allocate memory to receive the data */
				new_ptr->pt_single_op->data = MPIU_Malloc(req->dev.recv_data_sz);
				if (new_ptr->pt_single_op->data == NULL)
				{
					MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER,L"**nomem" );
				}

				new_ptr->pt_single_op->data_recd = 0;

				MPIDI_Request_set_type(req, MPIDI_REQUEST_TYPE_PT_SINGLE_PUT);
				req->dev.user_buf = new_ptr->pt_single_op->data;
				req->dev.lock_queue_entry = new_ptr;
			}

			*rreqp = req;
			mpi_errno = MPIDI_CH3U_Post_data_receive(TRUE, rreqp);
			/* --BEGIN ERROR HANDLING-- */
			if (mpi_errno != MPI_SUCCESS)
			{
				mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|postrecv", L"**ch3|postrecv %s", L"MPIDI_CH3_PKT_LOCK_PUT_UNLOCK");
			}
			/* --END ERROR HANDLING-- */

			break;
		}

		case MPIDI_CH3_PKT_LOCK_ACCUM_UNLOCK:
		{
			MPIDI_CH3_Pkt_lock_accum_unlock_t * lock_accum_unlock_pkt = &pkt->lock_accum_unlock;
			MPID_Request *req;
			MPID_Win *win_ptr;
			MPIDI_Win_lock_queue *curr_ptr, *prev_ptr, *new_ptr;

			MPIU_DBG_MSG(CH3_OTHER,VERBOSE,L"received lock_accum_unlock pkt");

			/* no need to acquire the lock here because we need to receive the data into a temporary buffer first */

			req = MPID_Request_create();
			MPIU_Object_set_ref(req, 1);
				
			req->dev.datatype = lock_accum_unlock_pkt->datatype;
			MPID_Datatype_get_size_macro(lock_accum_unlock_pkt->datatype, type_size);
			req->dev.recv_data_sz = type_size * lock_accum_unlock_pkt->count;
			req->dev.user_count = lock_accum_unlock_pkt->count;
			req->dev.target_win_handle = lock_accum_unlock_pkt->target_win_handle;
			
			/* queue the information */

			new_ptr = (MPIDI_Win_lock_queue *) MPIU_Malloc(sizeof(MPIDI_Win_lock_queue));
			if (!new_ptr)
			{
				MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER,L"**nomem" );
			}

			new_ptr->pt_single_op = (MPIDI_PT_single_op *) MPIU_Malloc(sizeof(MPIDI_PT_single_op));
			if (new_ptr->pt_single_op == NULL)
			{
				MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER,L"**nomem" );
			}

			MPID_Win_get_ptr(lock_accum_unlock_pkt->target_win_handle, win_ptr);

			/* FIXME: MT: The queuing may need to be done atomically. */
			
			curr_ptr = (MPIDI_Win_lock_queue *) win_ptr->lock_queue;
			prev_ptr = curr_ptr;
			while (curr_ptr != NULL)
			{
				prev_ptr = curr_ptr;
				curr_ptr = curr_ptr->next;
			}

			if (prev_ptr != NULL)
				prev_ptr->next = new_ptr;
			else 
				win_ptr->lock_queue = new_ptr;
		
			new_ptr->next = NULL;  
			new_ptr->lock_type = lock_accum_unlock_pkt->lock_type;
			new_ptr->source_win_handle = lock_accum_unlock_pkt->source_win_handle;
			new_ptr->vc = vc;
				
			new_ptr->pt_single_op->type = MPIDI_RMA_ACCUMULATE;
			new_ptr->pt_single_op->addr = MPIDI_Message_pointer_t_to_local_ptr(lock_accum_unlock_pkt->addr);
			new_ptr->pt_single_op->count = lock_accum_unlock_pkt->count;
			new_ptr->pt_single_op->datatype = lock_accum_unlock_pkt->datatype;
			new_ptr->pt_single_op->op = lock_accum_unlock_pkt->op;
			/* allocate memory to receive the data */
			new_ptr->pt_single_op->data = MPIU_Malloc(req->dev.recv_data_sz);
			if (new_ptr->pt_single_op->data == NULL)
			{
				MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER,L"**nomem" );
			}
			
			new_ptr->pt_single_op->data_recd = 0;
			
			MPIDI_Request_set_type(req, MPIDI_REQUEST_TYPE_PT_SINGLE_ACCUM);
			req->dev.user_buf = new_ptr->pt_single_op->data;
			req->dev.lock_queue_entry = new_ptr;

			*rreqp = req;
			mpi_errno = MPIDI_CH3U_Post_data_receive(TRUE, rreqp);
			/* --BEGIN ERROR HANDLING-- */
			if (mpi_errno != MPI_SUCCESS)
			{
				mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|postrecv", L"**ch3|postrecv %s", L"MPIDI_CH3_PKT_LOCK_ACCUM_UNLOCK");
			}
			/* --END ERROR HANDLING-- */

			break;
		}

		case MPIDI_CH3_PKT_LOCK_GET_UNLOCK:
		{
			MPIDI_CH3_Pkt_lock_get_unlock_t * lock_get_unlock_pkt = &pkt->lock_get_unlock;
			MPID_Win *win_ptr;

			MPIU_DBG_MSG(CH3_OTHER,VERBOSE,L"received lock_get_unlock pkt");

			MPID_Win_get_ptr(lock_get_unlock_pkt->target_win_handle, win_ptr);

			if (MPIDI_CH3I_Try_acquire_win_lock(win_ptr, lock_get_unlock_pkt->lock_type) == 1)
			{
				/* do the get. for this optimization, only basic datatypes supported. */
				MPIDI_CH3_Pkt_t upkt;
				MPIDI_CH3_Pkt_get_resp_t * get_resp_pkt = &upkt.get_resp;
				MPID_Request *req;
				MPID_IOV iov[MPID_IOV_LIMIT];

				req = MPID_Request_create();
				req->dev.target_win_handle = lock_get_unlock_pkt->target_win_handle;
				req->dev.source_win_handle = lock_get_unlock_pkt->source_win_handle;
				req->dev.single_op_opt = 1;
				req->dev.ca = MPIDI_CH3_CA_COMPLETE;

				MPIDI_Request_set_type(req, MPIDI_REQUEST_TYPE_GET_RESP); 
				req->kind = MPID_REQUEST_SEND;

				MPIDI_Pkt_init(get_resp_pkt, MPIDI_CH3_PKT_GET_RESP);
				get_resp_pkt->request_handle = lock_get_unlock_pkt->request_handle;
				
				iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST) get_resp_pkt;
				iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)sizeof(*get_resp_pkt);

				iov[1].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)MPIDI_Message_pointer_t_to_local_ptr(lock_get_unlock_pkt->addr);
				MPID_Datatype_get_size_macro(lock_get_unlock_pkt->datatype, type_size);
				iov[1].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)(lock_get_unlock_pkt->count * type_size);
			
				mpi_errno = MPIDI_CH3.iSendv(vc, req, iov, 2);
				/* --BEGIN ERROR HANDLING-- */
				if (mpi_errno != MPI_SUCCESS)
				{
					MPIU_Object_set_ref(req, 0);
					MPIDI_CH3_Request_destroy(req);
					mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|rmamsg", 0);
					return mpi_errno;
				}
				/* --END ERROR HANDLING-- */
			}
			else
			{
				/* queue the information */
				MPIDI_Win_lock_queue *curr_ptr, *prev_ptr, *new_ptr;

				/* FIXME: MT: This may need to be done atomically. */

				curr_ptr = (MPIDI_Win_lock_queue *) win_ptr->lock_queue;
				prev_ptr = curr_ptr;
				while (curr_ptr != NULL)
				{
					prev_ptr = curr_ptr;
					curr_ptr = curr_ptr->next;
				}

				new_ptr = (MPIDI_Win_lock_queue *) MPIU_Malloc(sizeof(MPIDI_Win_lock_queue));
				if (!new_ptr)
				{
					MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER,L"**nomem" );
				}
				new_ptr->pt_single_op = (MPIDI_PT_single_op *) MPIU_Malloc(sizeof(MPIDI_PT_single_op));
				if (new_ptr->pt_single_op == NULL)
				{
					MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER,L"**nomem" );
				}

				if (prev_ptr != NULL)
					prev_ptr->next = new_ptr;
				else 
					win_ptr->lock_queue = new_ptr;
		
				new_ptr->next = NULL;  
				new_ptr->lock_type = lock_get_unlock_pkt->lock_type;
				new_ptr->source_win_handle = lock_get_unlock_pkt->source_win_handle;
				new_ptr->vc = vc;

				new_ptr->pt_single_op->type = MPIDI_RMA_GET;
				new_ptr->pt_single_op->addr = MPIDI_Message_pointer_t_to_local_ptr(lock_get_unlock_pkt->addr);
				new_ptr->pt_single_op->count = lock_get_unlock_pkt->count;
				new_ptr->pt_single_op->datatype = lock_get_unlock_pkt->datatype;
				new_ptr->pt_single_op->data = NULL;
				new_ptr->pt_single_op->request_handle = lock_get_unlock_pkt->request_handle;
				new_ptr->pt_single_op->data_recd = 1;
			}

			*rreqp = NULL;

			break;
		}

		case MPIDI_CH3_PKT_CLOSE:
		{
			MPIDI_CH3_Pkt_close_t * close_pkt = &pkt->close;
			
			if (vc->state == MPIDI_VC_STATE_LOCAL_CLOSE)
			{
				MPIDI_CH3_Pkt_t upkt;
				MPIDI_CH3_Pkt_close_t * resp_pkt = &upkt.close;
				MPID_Request * resp_sreq;

				MPIDI_Pkt_init(resp_pkt, MPIDI_CH3_PKT_CLOSE);
				resp_pkt->ack = TRUE;

				MPIU_DBG_MSG_D(CH3_OTHER,VERBOSE,L"sending close(TRUE) to %d", vc->pg_rank);
				mpi_errno = MPIDI_CH3.iStartMsg(vc, resp_pkt, sizeof(*resp_pkt), &resp_sreq);
				if (mpi_errno != MPI_SUCCESS)
				{
					MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, L"**ch3|send_close_ack");
				}

				if (resp_sreq != NULL)
				{
					MPID_Request_release(resp_sreq);
				}
			}

			if (close_pkt->ack == FALSE)
			{
				if (vc->state == MPIDI_VC_STATE_LOCAL_CLOSE)
				{
					MPIU_DBG_MSG_D(CH3_CONNECT,VERBOSE, L"received close(FALSE) from %d, moving to CLOSE_ACKED.", vc->pg_rank);
					MPIU_DBG_VCSTATECHANGE(vc, VC_STATE_CLOSE_ACKED);
					vc->state = MPIDI_VC_STATE_CLOSE_ACKED;
				}
				else /* (vc->state == MPIDI_VC_STATE_ACTIVE) */
				{
					MPIU_Assert(vc->state == MPIDI_VC_STATE_ACTIVE);
					MPIU_DBG_MSG_D(CH3_CONNECT,VERBOSE, L"received close(FALSE) from %d, moving to REMOTE_CLOSE.", vc->pg_rank);
					MPIU_DBG_VCSTATECHANGE(vc, VC_STATE_REMOTE_CLOSE);
					vc->state = MPIDI_VC_STATE_REMOTE_CLOSE;
				}
			}
			else
			{
				MPIU_DBG_MSG_D(CH3_CONNECT,VERBOSE, L"received close(TRUE) from %d, moving to CLOSE_ACKED.", vc->pg_rank);
				MPIU_Assert (vc->state == MPIDI_VC_STATE_LOCAL_CLOSE || vc->state == MPIDI_VC_STATE_CLOSE_ACKED);
				MPIU_DBG_VCSTATECHANGE(vc, VC_STATE_CLOSE_ACKED);

				vc->state = MPIDI_VC_STATE_CLOSE_ACKED;
				mpi_errno = MPIDI_CH3.Connection_terminate(vc);
			}

			*rreqp = NULL;
			break;
		}
		
		case MPIDI_CH3_PKT_FLOW_CNTL_UPDATE:
		{
			/* --BEGIN ERROR HANDLING-- */
			MPIU_DBG_MSG(CH3_OTHER,VERBOSE,L"received flow control update pkt");
			mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_INTERN, L"**ch3|flowcntlpkt", 0);
			*rreqp = NULL;
			break;
			/* --END ERROR HANDLING-- */
		}
		
		default:
		{
			/* --BEGIN ERROR HANDLING-- */
			*rreqp = NULL;
			MPIU_ERR_SETANDJUMP1(mpi_errno,MPI_ERR_INTERN, L"**ch3|unknownpkt", L"**ch3|unknownpkt %d", pkt->type);
			break;
			/* --END ERROR HANDLING-- */
		}
	}

  fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3U_HANDLE_ORDERED_RECV_PKT);
	return mpi_errno;
}

/* FIXME: What does this routine do */
/* This function is used in conjunction with MPIDI_CH3_iStartRndvTransfer */
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3U_Post_data_receive
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3U_Post_data_receive(int found, MPID_Request ** rreqp)
{
	int dt_contig;
	MPI_Aint dt_true_lb;
	MPIDI_msg_sz_t userbuf_sz;
	MPID_Datatype * dt_ptr;
	MPIDI_msg_sz_t data_sz;
	MPID_Request * rreq = *rreqp;
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3U_POST_DATA_RECEIVE);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3U_POST_DATA_RECEIVE);

	if (rreq->dev.recv_data_sz == 0)
	{
		MPIU_DBG_MSG_S(CH3_OTHER,VERBOSE, L"null message, %s, decrementing completion counter", (found ? L"posted request found" : L"unexpected request allocated"));
		/* mark data transfer as complete and decrment CC */
		MPIDI_CH3U_Request_complete(rreq);
		*rreqp = NULL;
		goto fn_exit;
	}
		
	if (found)
	{
		MPIU_DBG_MSG(CH3_OTHER,VERBOSE,L"posted request found");
		
		MPIDI_Datatype_get_info(rreq->dev.user_count, rreq->dev.datatype, dt_contig, userbuf_sz, dt_ptr, dt_true_lb);
				
		if (rreq->dev.recv_data_sz <= userbuf_sz)
		{
			data_sz = rreq->dev.recv_data_sz;
		}
		else
		{
			MPIU_DBG_MSG_FMT(CH3_OTHER,VERBOSE,(MPIU_DBG_FDEST, L"receive buffer too small; message truncated, msg_sz=" MPIDI_MSG_SZ_FMT L", userbuf_sz=" MPIDI_MSG_SZ_FMT, rreq->dev.recv_data_sz, userbuf_sz));
			rreq->status.MPI_ERROR = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_TRUNCATE, L"**truncate", L"**truncate %d %d %d %d", rreq->status.MPI_SOURCE, rreq->status.MPI_TAG, rreq->dev.recv_data_sz, userbuf_sz );
			rreq->status.count = (int)userbuf_sz;
			data_sz = userbuf_sz;
		}

		if (dt_contig && data_sz == rreq->dev.recv_data_sz)
		{
			/* user buffer is contiguous and large enough to store the entire message */
			/* FIXME: So why don't we move it *now* ? */
			MPIU_DBG_MSG(CH3_OTHER,VERBOSE,L"IOV loaded for contiguous read");
			rreq->dev.iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)((char*)(rreq->dev.user_buf) + dt_true_lb);
			rreq->dev.iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST) data_sz;
			rreq->dev.iov_count = 1;
			rreq->dev.ca = MPIDI_CH3_CA_COMPLETE;
		}
		else
		{
			/* user buffer is not contiguous or is too small to hold the entire message */
			int mpi_errno;
					
			MPIU_DBG_MSG(CH3_OTHER,VERBOSE,L"IOV loaded for non-contiguous read");
			MPID_Segment_init(rreq->dev.user_buf, rreq->dev.user_count, rreq->dev.datatype, &rreq->dev.segment, 0);
			rreq->dev.segment_first = 0;
			rreq->dev.segment_size = data_sz;
			mpi_errno = MPIDI_CH3U_Request_load_recv_iov(rreq);
			/* --BEGIN ERROR HANDLING-- */
			if (mpi_errno != MPI_SUCCESS)
			{
				mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|loadrecviov", 0);
				goto fn_exit;
			}
			/* --END ERROR HANDLING-- */
		}
	}
	else /* if (!found) */
	{
		/* TODO: to improve performance, allocate temporary buffer from a specialized buffer pool. */
		MPIU_DBG_MSG(CH3_OTHER,VERBOSE,L"unexpected request allocated");
				
		rreq->dev.tmpbuf = MPIU_Malloc(rreq->dev.recv_data_sz);
		MPIU_ERR_MEM_CHK_POP(mpi_errno, rreq->dev.tmpbuf, rreq->dev.recv_data_sz, "rreq->dev.tmpbuf");
		rreq->dev.tmpbuf_sz = rreq->dev.recv_data_sz;
				
		rreq->dev.iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)rreq->dev.tmpbuf;
		rreq->dev.iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)rreq->dev.recv_data_sz;
		rreq->dev.iov_count = 1;
		rreq->dev.ca = MPIDI_CH3_CA_UNPACK_UEBUF_AND_COMPLETE;
		rreq->dev.recv_pending_count = 2;
	}

fn_exit:
fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3U_POST_DATA_RECEIVE);
	return mpi_errno;
}


/* Check if requested lock can be granted. If it can, set 
   win_ptr->current_lock_type to the new lock type and return 1. Else return 0.

   FIXME: MT: This function must be atomic because two threads could be trying 
   to do the same thing, e.g., the main thread in MPI_Win_lock(source=target) 
   and another thread in the progress engine.
 */
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_Try_acquire_win_lock
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3I_Try_acquire_win_lock(MPID_Win *win_ptr, int requested_lock)
{
	int existing_lock;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_TRY_ACQUIRE_WIN_LOCK);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_TRY_ACQUIRE_WIN_LOCK);

	existing_lock = win_ptr->current_lock_type;

	/* Locking Rules:
	   
	Requested		  Existing			 Action
	--------		   --------			 ------
	Shared			 Exclusive			Queue it
	Shared			 NoLock/Shared		Grant it
	Exclusive		  NoLock			   Grant it
	Exclusive		  Exclusive/Shared	 Queue it
	*/

	if ( ( (requested_lock == MPI_LOCK_SHARED) && 
		   ((existing_lock == MPID_LOCK_NONE) ||
			(existing_lock == MPI_LOCK_SHARED) ) )
		 || 
		 ( (requested_lock == MPI_LOCK_EXCLUSIVE) &&
		   (existing_lock == MPID_LOCK_NONE) ) )
	{

		/* grant lock.  set new lock type on window */
		win_ptr->current_lock_type = requested_lock;

		/* if shared lock, incr. ref. count */
		if (requested_lock == MPI_LOCK_SHARED)
			win_ptr->shared_lock_ref_cnt++;

		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_TRY_ACQUIRE_WIN_LOCK);
		return 1;
	}

	/* do not grant lock */
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_TRY_ACQUIRE_WIN_LOCK);
	return 0;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_Send_lock_granted_pkt
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3I_Send_lock_granted_pkt(MPIDI_VC_t *vc, MPI_Win source_win_handle)
{
	MPIDI_CH3_Pkt_t upkt;
	MPIDI_CH3_Pkt_lock_granted_t *lock_granted_pkt = &upkt.lock_granted;
	MPID_Request *req;
	int mpi_errno;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_SEND_LOCK_GRANTED_PKT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_SEND_LOCK_GRANTED_PKT);

	/* send lock granted packet */
	MPIDI_Pkt_init(lock_granted_pkt, MPIDI_CH3_PKT_LOCK_GRANTED);
	lock_granted_pkt->source_win_handle = source_win_handle;
		
	mpi_errno = MPIDI_CH3.iStartMsg(vc, lock_granted_pkt, sizeof(*lock_granted_pkt), &req);
	/* --BEGIN ERROR HANDLING-- */
	if (mpi_errno != MPI_SUCCESS)
	{
		mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|rmamsg", 0);
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SEND_LOCK_GRANTED_PKT);
		return mpi_errno;
	}
	/* --END ERROR HANDLING-- */
	if (req != NULL)
	{
		MPID_Request_release(req);
	}

	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SEND_LOCK_GRANTED_PKT);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3U_Handle_recv_req
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3U_Handle_recv_req(MPIDI_VC_t * vc, MPID_Request * rreq, int * complete)
{
	static int in_routine = FALSE;
	MPID_Win *win_ptr;
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3U_HANDLE_RECV_REQ);
	MPIDI_STATE_DECL(MPID_STATE_CH3_CA_COMPLETE);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3U_HANDLE_RECV_REQ);

	MPIU_Assert(in_routine == FALSE);
	in_routine = TRUE;
	
	switch(rreq->dev.ca)
	{
		case MPIDI_CH3_CA_COMPLETE:
		{
			MPIDI_FUNC_ENTER(MPID_STATE_CH3_CA_COMPLETE)
			/* FIXME: put ONC operations into their own completion action */
			
			if (MPIDI_Request_get_type(rreq) == MPIDI_REQUEST_TYPE_RECV)
			{
				/* mark data transfer as complete and decrement CC */
				MPIDI_CH3U_Request_complete(rreq);
				*complete = TRUE;
			}
			else if ((MPIDI_Request_get_type(rreq) == MPIDI_REQUEST_TYPE_PUT_RESP) ||
					 (MPIDI_Request_get_type(rreq) == MPIDI_REQUEST_TYPE_ACCUM_RESP))
			{
				if (MPIDI_Request_get_type(rreq) == MPIDI_REQUEST_TYPE_ACCUM_RESP)
				{
					/* accumulate data from tmp_buf into user_buf */
					mpi_errno = do_accumulate_op(rreq);
					if (mpi_errno)
					{
						MPIU_ERR_POP(mpi_errno);
					}
				}

				MPID_Win_get_ptr(rreq->dev.target_win_handle, win_ptr);
				
				/* if passive target RMA, increment counter */
				if (win_ptr->current_lock_type != MPID_LOCK_NONE)
					win_ptr->my_pt_rma_puts_accs++;

				if (rreq->dev.source_win_handle != MPI_WIN_NULL)
				{
					/* Last RMA operation from source. If active
					   target RMA, decrement window counter. If
					   passive target RMA, release lock on window and
					   grant next lock in the lock queue if there is
					   any. If it's a shared lock or a lock-put-unlock
					   type of optimization, we also need to send an
					   ack to the source. */ 

					if (win_ptr->current_lock_type == MPID_LOCK_NONE)
					{
						/* FIXME: MT: this has to be done atomically */
						win_ptr->my_counter -= 1;
					}
					else
					{
						if ((win_ptr->current_lock_type == MPI_LOCK_SHARED) || (rreq->dev.single_op_opt == 1))
						{
							mpi_errno = MPIDI_CH3I_Send_pt_rma_done_pkt(vc, rreq->dev.source_win_handle);
							if (mpi_errno)
							{
								MPIU_ERR_POP(mpi_errno);
							}
						}
						mpi_errno = MPIDI_CH3I_Release_lock(win_ptr);
					}
				}

				/* mark data transfer as complete and decrement CC */
				MPIDI_CH3U_Request_complete(rreq);
				*complete = TRUE;
			}
			else if (MPIDI_Request_get_type(rreq) == MPIDI_REQUEST_TYPE_PUT_RESP_DERIVED_DT)
			{
				MPID_Datatype *new_dtp;
				
				/* create derived datatype */
				create_derived_datatype(rreq, &new_dtp);

				/* update request to get the data */
				MPIDI_Request_set_type(rreq, MPIDI_REQUEST_TYPE_PUT_RESP);
				rreq->dev.datatype = new_dtp->handle;
				rreq->dev.recv_data_sz = new_dtp->size * rreq->dev.user_count;
				
				rreq->dev.datatype_ptr = new_dtp;
				/* this will cause the datatype to be freed when the request is freed. free dtype_info here. */
				if (rreq->dev.dtype_info != NULL)
				{
					MPIU_Free(rreq->dev.dtype_info);
				}

				MPID_Segment_init(rreq->dev.user_buf, rreq->dev.user_count, rreq->dev.datatype, &rreq->dev.segment, 0);
				rreq->dev.segment_first = 0;
				rreq->dev.segment_size = rreq->dev.recv_data_sz;

				mpi_errno = MPIDI_CH3U_Request_load_recv_iov(rreq);
				if (mpi_errno != MPI_SUCCESS)
				{
					MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, L"**ch3|loadrecviov");
				}

				*complete = FALSE;
			}
			else if (MPIDI_Request_get_type(rreq) == MPIDI_REQUEST_TYPE_ACCUM_RESP_DERIVED_DT)
			{
				MPID_Datatype *new_dtp;
				MPI_Aint true_lb, true_extent, extent;
				void *tmp_buf;
				MPIU_THREADPRIV_DECL;

				MPIU_THREADPRIV_GET;
			   
				/* create derived datatype */
				create_derived_datatype(rreq, &new_dtp);

				/* update new request to get the data */
				MPIDI_Request_set_type(rreq, MPIDI_REQUEST_TYPE_ACCUM_RESP);

				/* first need to allocate tmp_buf to recv the data into */

				NMPI_Type_get_true_extent(mpi_errno, (new_dtp->handle, &true_lb, &true_extent));
				if (mpi_errno)
				{
					MPIU_ERR_POP(mpi_errno);
				}

				MPID_Datatype_get_extent_macro(new_dtp->handle, extent); 

				tmp_buf = MPIU_Malloc(rreq->dev.user_count * (MPIR_MAX(extent,true_extent)));  
				/* --BEGIN ERROR HANDLING-- */
				if (!tmp_buf)
				{
					mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0 );
					goto fn_fail;
				}
				/* --END ERROR HANDLING-- */

				/* adjust for potential negative lower bound in datatype */
				tmp_buf = (void *)((char*)tmp_buf - true_lb);

				rreq->dev.user_buf = tmp_buf;
				rreq->dev.datatype = new_dtp->handle;
				rreq->dev.recv_data_sz = new_dtp->size * rreq->dev.user_count; 
				rreq->dev.datatype_ptr = new_dtp;
				/* this will cause the datatype to be freed when the request is freed. free dtype_info here. */
				if (rreq->dev.dtype_info != NULL)
				{
					MPIU_Free(rreq->dev.dtype_info);
				}

				MPID_Segment_init(rreq->dev.user_buf, rreq->dev.user_count, rreq->dev.datatype, &rreq->dev.segment, 0);
				rreq->dev.segment_first = 0;
				rreq->dev.segment_size = rreq->dev.recv_data_sz;

				mpi_errno = MPIDI_CH3U_Request_load_recv_iov(rreq);
				if (mpi_errno != MPI_SUCCESS)
				{
					MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, L"**ch3|loadrecviov");
				}

				*complete = FALSE;
			}
			else if (MPIDI_Request_get_type(rreq) == MPIDI_REQUEST_TYPE_GET_RESP_DERIVED_DT)
			{
				MPID_Datatype *new_dtp;
				MPIDI_CH3_Pkt_t upkt;
				MPIDI_CH3_Pkt_get_resp_t * get_resp_pkt = &upkt.get_resp;
				MPID_IOV iov[MPID_IOV_LIMIT];
				MPID_Request * sreq;
				int iov_n;
				
				/* create derived datatype */
				create_derived_datatype(rreq, &new_dtp);
				if (rreq->dev.dtype_info != NULL)
				{
					MPIU_Free(rreq->dev.dtype_info);
				}

				/* create request for sending data */
				sreq = MPID_Request_create();
				if (sreq == NULL)
				{
					/* --BEGIN ERROR HANDLING-- */
					mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
					goto fn_exit;
					/* --END ERROR HANDLING-- */
				}
				sreq->kind = MPID_REQUEST_SEND;
				MPIDI_Request_set_type(sreq, MPIDI_REQUEST_TYPE_GET_RESP);
				sreq->dev.user_buf = rreq->dev.user_buf;
				sreq->dev.user_count = rreq->dev.user_count;
				sreq->dev.datatype = new_dtp->handle;
				sreq->dev.datatype_ptr = new_dtp;
				sreq->dev.target_win_handle = rreq->dev.target_win_handle;
				sreq->dev.source_win_handle = rreq->dev.source_win_handle;
				
				MPIDI_Pkt_init(get_resp_pkt, MPIDI_CH3_PKT_GET_RESP);
				get_resp_pkt->request_handle = rreq->dev.request_handle;
				
				iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST) get_resp_pkt;
				iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)sizeof(*get_resp_pkt);

				MPID_Segment_init(sreq->dev.user_buf, sreq->dev.user_count, sreq->dev.datatype, &sreq->dev.segment, 0);
				sreq->dev.segment_first = 0;
				sreq->dev.segment_size = new_dtp->size * sreq->dev.user_count;

				iov_n = MPID_IOV_LIMIT - 1;
				mpi_errno = MPIDI_CH3U_Request_load_send_iov(sreq, &iov[1], &iov_n);
				if (mpi_errno == MPI_SUCCESS)
				{
					iov_n += 1;
				
					mpi_errno = MPIDI_CH3.iSendv(vc, sreq, iov, iov_n);
					/* --BEGIN ERROR HANDLING-- */
					if (mpi_errno != MPI_SUCCESS)
					{
						MPIU_Object_set_ref(sreq, 0);
						MPIDI_CH3_Request_destroy(sreq);
						sreq = NULL;
						mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|rmamsg", 0);
						goto fn_exit;
					}
					/* --END ERROR HANDLING-- */
				}

				/* mark receive data transfer as complete and decrement CC in receive request */
				MPIDI_CH3U_Request_complete(rreq);
				*complete = TRUE;
			}
			else if ((MPIDI_Request_get_type(rreq) == MPIDI_REQUEST_TYPE_PT_SINGLE_PUT) ||
					 (MPIDI_Request_get_type(rreq) == MPIDI_REQUEST_TYPE_PT_SINGLE_ACCUM))
			{
				/* received all the data for single lock-put(accum)-unlock 
				   optimization where the lock was not acquired in 
				   ch3u_handle_recv_pkt. Try to acquire the lock and do the 
				   operation. */

				MPID_Win *win_ptr;
				MPIDI_Win_lock_queue *lock_queue_entry, *curr_ptr, **curr_ptr_ptr;

				MPID_Win_get_ptr(rreq->dev.target_win_handle, win_ptr);

				lock_queue_entry = rreq->dev.lock_queue_entry;
				
				if (MPIDI_CH3I_Try_acquire_win_lock(win_ptr, lock_queue_entry->lock_type) == 1)
				{
					if (MPIDI_Request_get_type(rreq) == MPIDI_REQUEST_TYPE_PT_SINGLE_PUT)
					{
						/* copy the data over */
						mpi_errno = MPIR_Localcopy(rreq->dev.user_buf,
												   rreq->dev.user_count,
												   rreq->dev.datatype,
										 lock_queue_entry->pt_single_op->addr,
										 lock_queue_entry->pt_single_op->count,
									 lock_queue_entry->pt_single_op->datatype);
					}
					else
					{
						mpi_errno = do_simple_accumulate(lock_queue_entry->pt_single_op);
					}

					if (mpi_errno)
					{
						MPIU_ERR_POP(mpi_errno);
					}

					/* increment counter */
					win_ptr->my_pt_rma_puts_accs++;

					/* send done packet */
					mpi_errno = MPIDI_CH3I_Send_pt_rma_done_pkt(vc, lock_queue_entry->source_win_handle);
					if (mpi_errno)
					{
						MPIU_ERR_POP(mpi_errno);
					}

					/* free lock_queue_entry including data buffer and remove it from the queue. */
					curr_ptr = (MPIDI_Win_lock_queue *) win_ptr->lock_queue;
					curr_ptr_ptr = (MPIDI_Win_lock_queue **) &(win_ptr->lock_queue);
					while (curr_ptr != lock_queue_entry)
					{
						curr_ptr_ptr = &(curr_ptr->next);
						curr_ptr = curr_ptr->next;
					}					
					*curr_ptr_ptr = curr_ptr->next;

					MPIU_Free(lock_queue_entry->pt_single_op->data);
					MPIU_Free(lock_queue_entry->pt_single_op);
					MPIU_Free(lock_queue_entry);

					/* Release lock and grant next lock if there is one. */
					mpi_errno = MPIDI_CH3I_Release_lock(win_ptr);
				}
				else
				{
					/* could not acquire lock. mark data recd as 1 */
					lock_queue_entry->pt_single_op->data_recd = 1;
				}

				/* mark data transfer as complete and decrement CC */
				MPIDI_CH3U_Request_complete(rreq);
				*complete = TRUE;
			}
			/* --BEGIN ERROR HANDLING-- */
			else
			{
				/* We shouldn't reach this code because the only other request types are sends */
				MPIU_Assert(MPIDI_Request_get_type(rreq) == MPIDI_REQUEST_TYPE_RECV);
				MPIDI_CH3U_Request_complete(rreq);
				*complete = TRUE;
			}
			/* --END ERROR HANDLING-- */
			
			MPIDI_FUNC_EXIT(MPID_STATE_CH3_CA_COMPLETE)
			break;
		}
		
		case MPIDI_CH3_CA_UNPACK_UEBUF_AND_COMPLETE:
		{
			int recv_pending;
			
			MPIDI_Request_recv_pending(rreq, &recv_pending);
			if (!recv_pending)
			{ 
				if (rreq->dev.recv_data_sz > 0)
				{
					MPIDI_CH3U_Request_unpack_uebuf(rreq);
					MPIU_Free(rreq->dev.tmpbuf);
				}
			}
			else
			{
				/* The receive has not been posted yet.  MPID_{Recv/Irecv}() is responsible for unpacking the buffer. */
			}
			
			/* mark data transfer as complete and decrement CC */
			MPIDI_CH3U_Request_complete(rreq);
			*complete = TRUE;
			
			break;
		}
		
		case MPIDI_CH3_CA_UNPACK_SRBUF_AND_COMPLETE:
		{
			MPIDI_CH3U_Request_unpack_srbuf(rreq);

			if ((MPIDI_Request_get_type(rreq) == MPIDI_REQUEST_TYPE_PUT_RESP) ||
			   (MPIDI_Request_get_type(rreq) == MPIDI_REQUEST_TYPE_ACCUM_RESP))

			{
				if (MPIDI_Request_get_type(rreq) == MPIDI_REQUEST_TYPE_ACCUM_RESP)
				{
					/* accumulate data from tmp_buf into user_buf */
					mpi_errno = do_accumulate_op(rreq);
					if (mpi_errno)
					{
						MPIU_ERR_POP(mpi_errno);
					}
				}

				MPID_Win_get_ptr(rreq->dev.target_win_handle, win_ptr);
				
				/* if passive target RMA, increment counter */
				if (win_ptr->current_lock_type != MPID_LOCK_NONE)
					win_ptr->my_pt_rma_puts_accs++;
				
				if (rreq->dev.source_win_handle != MPI_WIN_NULL)
				{
					/* Last RMA operation from source. If active
					   target RMA, decrement window counter. If
					   passive target RMA, release lock on window and
					   grant next lock in the lock queue if there is
					   any. If it's a shared lock or a lock-put-unlock
					   type of optimization, we also need to send an
					   ack to the source. */ 
					
					if (win_ptr->current_lock_type == MPID_LOCK_NONE)
					{
						/* FIXME: MT: this has to be done atomically */
						win_ptr->my_counter -= 1;
					}
					else
					{
						if ((win_ptr->current_lock_type == MPI_LOCK_SHARED) || (rreq->dev.single_op_opt == 1))
						{
							mpi_errno = MPIDI_CH3I_Send_pt_rma_done_pkt(vc, rreq->dev.source_win_handle);
							if (mpi_errno)
							{
								MPIU_ERR_POP(mpi_errno);
							}
						}
						mpi_errno = MPIDI_CH3I_Release_lock(win_ptr);
					}
				}
			}

			/* mark data transfer as complete and decrement CC */
			MPIDI_CH3U_Request_complete(rreq);
			*complete = TRUE;
			
			break;
		}
		
		case MPIDI_CH3_CA_UNPACK_SRBUF_AND_RELOAD_IOV:
		{
			MPIDI_CH3U_Request_unpack_srbuf(rreq);
			mpi_errno = MPIDI_CH3U_Request_load_recv_iov(rreq);
			/* --BEGIN ERROR HANDLING-- */
			if (mpi_errno != MPI_SUCCESS)
			{
				mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|loadrecviov", L"**ch3|loadrecviov %s", L"MPIDI_CH3_CA_UNPACK_SRBUF_AND_RELOAD_IOV");
				goto fn_fail;
			}
			/* --END ERROR HANDLING-- */
			*complete = FALSE;
			break;
		}
		
		case MPIDI_CH3_CA_RELOAD_IOV:
		{
			mpi_errno = MPIDI_CH3U_Request_load_recv_iov(rreq);
			/* --BEGIN ERROR HANDLING-- */
			if (mpi_errno != MPI_SUCCESS)
			{
				mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|loadrecviov", L"**ch3|loadrecviov %s", L"MPIDI_CH3_CA_RELOAD_IOV");
				goto fn_fail;
			}
			/* --END ERROR HANDLING-- */
			*complete = FALSE;
			break;
		}

		/* --BEGIN ERROR HANDLING-- */
		default:
		{
			*complete = TRUE;
			mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_INTERN, L"**ch3|badca", L"**ch3|badca %d", rreq->dev.ca);
			break;
		}
		/* --END ERROR HANDLING-- */
	}

  fn_exit:
	in_routine = FALSE;
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3U_HANDLE_RECV_REQ);
	return mpi_errno;
 fn_fail:
	goto fn_exit;
}



#undef FUNCNAME
#define FUNCNAME create_derived_datatype
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int create_derived_datatype(MPID_Request *req, MPID_Datatype **dtp)
{
	MPIDI_RMA_dtype_info *dtype_info;
	void *dataloop;
	MPID_Datatype *new_dtp;
	int mpi_errno=MPI_SUCCESS;
	MPI_Aint ptrdiff;
	MPI_Datatype datatype;
	MPIDI_STATE_DECL(MPID_STATE_CREATE_DERIVED_DATATYPE);

	MPIDI_FUNC_ENTER(MPID_STATE_CREATE_DERIVED_DATATYPE);

	if (req->dev.serialized_datatype_buffer != NULL)
	{
		mpi_errno = DeSerializeDatatype(req->dev.serialized_datatype_buffer, &datatype);
		if (mpi_errno != MPI_SUCCESS)
		{
			mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", 0);
			MPIDI_FUNC_EXIT(MPID_STATE_CREATE_DERIVED_DATATYPE);
			return mpi_errno;
		}
		NMPI_Type_commit(mpi_errno, (&datatype));
		if (mpi_errno != MPI_SUCCESS)
		{
			mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", 0);
			MPIDI_FUNC_EXIT(MPID_STATE_CREATE_DERIVED_DATATYPE);
			return mpi_errno;
		}
		MPID_Datatype_get_ptr(datatype, (*dtp));
		MPIU_Free(req->dev.serialized_datatype_buffer);
		req->dev.serialized_datatype_buffer = NULL;
	}
	else
	{
		dtype_info = req->dev.dtype_info;
		dataloop = req->dev.dataloop;

		/* allocate new datatype object and handle */
		new_dtp = (MPID_Datatype *) MPIU_Handle_obj_alloc(&MPID_Datatype_mem);
		/* --BEGIN ERROR HANDLING-- */
		if (!new_dtp)
		{
			mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
			MPIDI_FUNC_EXIT(MPID_STATE_CREATE_DERIVED_DATATYPE);
			return mpi_errno;
		}
		/* --END ERROR HANDLING-- */

		*dtp = new_dtp;

		/* Note: handle is filled in by MPIU_Handle_obj_alloc() */
		MPIU_Object_set_ref(new_dtp, 1);
		new_dtp->is_permanent = 0;
		new_dtp->is_committed = 1;
		new_dtp->attributes   = 0;
		new_dtp->cache_id	 = 0;
		new_dtp->name[0]	  = 0;
		new_dtp->is_contig = dtype_info->is_contig;
		new_dtp->n_contig_blocks = dtype_info->n_contig_blocks; 
		new_dtp->size = dtype_info->size;
		new_dtp->extent = dtype_info->extent;
		new_dtp->dataloop_size = dtype_info->dataloop_size;
		new_dtp->dataloop_depth = dtype_info->dataloop_depth; 
		new_dtp->eltype = dtype_info->eltype;
		/* set dataloop pointer */
		new_dtp->dataloop = req->dev.dataloop;

		new_dtp->ub = dtype_info->ub;
		new_dtp->lb = dtype_info->lb;
		new_dtp->true_ub = dtype_info->true_ub;
		new_dtp->true_lb = dtype_info->true_lb;
		new_dtp->has_sticky_ub = dtype_info->has_sticky_ub;
		new_dtp->has_sticky_lb = dtype_info->has_sticky_lb;
		/* update pointers in dataloop */
		ptrdiff = (MPI_Aint)((char *) (new_dtp->dataloop) - (char *)(dtype_info->dataloop));

		MPID_Dataloop_update(new_dtp->dataloop, ptrdiff);

		new_dtp->contents = NULL;
	}

	MPIDI_FUNC_EXIT(MPID_STATE_CREATE_DERIVED_DATATYPE);
	return mpi_errno;
}


#undef FUNCNAME
#define FUNCNAME do_accumulate_op
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int do_accumulate_op(MPID_Request *rreq)
{
	int mpi_errno = MPI_SUCCESS, predefined;
	MPI_Aint true_lb, true_extent;
	MPI_User_function *uop;
	MPIU_THREADPRIV_DECL;
	MPIDI_STATE_DECL(MPID_STATE_DO_ACCUMULATE_OP);

	MPIDI_FUNC_ENTER(MPID_STATE_DO_ACCUMULATE_OP);

	MPIU_THREADPRIV_GET;

	if (rreq->dev.op == MPI_REPLACE)
	{
		/* simply copy the data */
		mpi_errno = MPIR_Localcopy(rreq->dev.user_buf, rreq->dev.user_count,
								   rreq->dev.datatype,
								   rreq->dev.real_user_buf,
								   rreq->dev.user_count,
								   rreq->dev.datatype);
		/* --BEGIN ERROR HANDLING-- */
		if (mpi_errno)
		{
			mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", 0);
			MPIDI_FUNC_EXIT(MPID_STATE_DO_ACCUMULATE_OP);
			return mpi_errno;
		}
		/* --END ERROR HANDLING-- */
		goto fn_exit;
	}

	if (HANDLE_GET_KIND(rreq->dev.op) == HANDLE_KIND_BUILTIN)
	{
		/* get the function by indexing into the op table */
		uop = MPIR_Op_table[(rreq->dev.op)%16 - 1];
	}
	else
	{
		/* --BEGIN ERROR HANDLING-- */
		mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OP, L"**opnotpredefined", L"**opnotpredefined %d", rreq->dev.op );
		MPIDI_FUNC_EXIT(MPID_STATE_DO_ACCUMULATE_OP);
		return mpi_errno;
		/* --END ERROR HANDLING-- */
	}
	
	MPIDI_CH3I_DATATYPE_IS_PREDEFINED(rreq->dev.datatype, predefined);
	if (predefined)
	{
		(*uop)(rreq->dev.user_buf, rreq->dev.real_user_buf, &(rreq->dev.user_count), &(rreq->dev.datatype));
	}
	else
	{
		/* derived datatype */
		MPID_Segment *segp;
		DLOOP_VECTOR *dloop_vec;
		MPI_Aint first, last;
		int vec_len, i, type_size, count;
		MPI_Datatype type;
		MPID_Datatype *dtp;
		
		segp = MPID_Segment_alloc();
		/* --BEGIN ERROR HANDLING-- */
		if (!segp)
		{
			mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0 );
			MPIDI_FUNC_EXIT(MPID_STATE_DO_ACCUMULATE_OP);
			return mpi_errno;
		}
		/* --END ERROR HANDLING-- */
		MPID_Segment_init(NULL, rreq->dev.user_count, rreq->dev.datatype, segp, 0);
		first = 0;
		last  = SEGMENT_IGNORE_LAST;
		
		MPID_Datatype_get_ptr(rreq->dev.datatype, dtp);
		vec_len = dtp->n_contig_blocks * rreq->dev.user_count + 1; 
		/* +1 needed because Rob says so */
		dloop_vec = (DLOOP_VECTOR *)MPIU_Malloc(vec_len * sizeof(DLOOP_VECTOR));
		/* --BEGIN ERROR HANDLING-- */
		if (!dloop_vec)
		{
			mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0 );
			MPIDI_FUNC_EXIT(MPID_STATE_DO_ACCUMULATE_OP);
			return mpi_errno;
		}
		/* --END ERROR HANDLING-- */
		
		MPID_Segment_pack_vector(segp, first, &last, dloop_vec, &vec_len);
		
		type = dtp->eltype;
		MPID_Datatype_get_size_macro(type, type_size);
		for (i=0; i<vec_len; i++)
		{
			count = (dloop_vec[i].DLOOP_VECTOR_LEN)/type_size;
			(*uop)((char *)rreq->dev.user_buf + MPIU_PtrToAint(dloop_vec[i].DLOOP_VECTOR_BUF),
				   (char *)rreq->dev.real_user_buf + MPIU_PtrToAint(dloop_vec[i].DLOOP_VECTOR_BUF),
				   &count, &type);
		}
		
		MPID_Segment_free(segp);
		MPIU_Free(dloop_vec);
	}

 fn_exit:
	/* free the temporary buffer */
	NMPI_Type_get_true_extent(mpi_errno, (rreq->dev.datatype, &true_lb, &true_extent));
	/* --BEGIN ERROR HANDLING-- */
	if (mpi_errno)
	{
		mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", 0);
		MPIDI_FUNC_EXIT(MPID_STATE_DO_ACCUMULATE_OP);
		return mpi_errno;
	}
	/* --END ERROR HANDLING-- */
	
	MPIU_Free((char *) rreq->dev.user_buf + true_lb);

	MPIDI_FUNC_EXIT(MPID_STATE_DO_ACCUMULATE_OP);
	return mpi_errno;
}

/* Release the current lock on the window and grant the next lock in the queue if any */
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_Release_lock
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3I_Release_lock(MPID_Win *win_ptr)
{
	MPIDI_Win_lock_queue *lock_queue, **lock_queue_ptr;
	int requested_lock, mpi_errno = MPI_SUCCESS, temp_entered_count;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_RELEASE_LOCK);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_RELEASE_LOCK);

	if (win_ptr->current_lock_type == MPI_LOCK_SHARED)
	{
		/* decr ref cnt */
		/* FIXME: MT: Must be done atomically */
		win_ptr->shared_lock_ref_cnt--;
	}

	/* If shared lock ref count is 0 (which is also true if the lock is an
	   exclusive lock), release the lock. */
	if (win_ptr->shared_lock_ref_cnt == 0)
	{

		/* This function needs to be reentrant even in the single-threaded case
		   because when going through the lock queue, the do_simple_get called in the 
		   lock-get-unlock case may itself cause a request to complete, and this function
		   may again get called in the completion action in ch3u_handle_send_req.c. To
		   handle this possibility, we use an entered_flag. If the flag is not 0, we simply
		   increment the entered_count and return. The loop through the lock queue is repeated 
		   if the entered_count has changed while we are in the loop.
		 */
		if (entered_flag != 0)
		{
			entered_count++;
			goto fn_exit;
		}
		else
		{
			entered_flag = 1;
			temp_entered_count = entered_count;
		}

		do
		{
			if (temp_entered_count != entered_count) temp_entered_count++;

			/* FIXME: MT: The setting of the lock type must be done atomically */
			win_ptr->current_lock_type = MPID_LOCK_NONE;
			
			/* If there is a lock queue, try to satisfy as many lock requests as 
			   possible. If the first one is a shared lock, grant it and grant all 
			   other shared locks. If the first one is an exclusive lock, grant 
			   only that one. */
			
			/* FIXME: MT: All queue accesses need to be made atomic */
			lock_queue = (MPIDI_Win_lock_queue *) win_ptr->lock_queue;
			lock_queue_ptr = (MPIDI_Win_lock_queue **) &(win_ptr->lock_queue);
			while (lock_queue)
			{
				/* if it is not a lock-op-unlock type case or if it is a 
				   lock-op-unlock type case but all the data has been received, 
				   try to acquire the lock */
				if ((lock_queue->pt_single_op == NULL) || 
					(lock_queue->pt_single_op->data_recd == 1))
				{
					
					requested_lock = lock_queue->lock_type;
					if (MPIDI_CH3I_Try_acquire_win_lock(win_ptr, requested_lock) == 1)
					{
						if (lock_queue->pt_single_op != NULL)
						{
							/* single op. do it here */
							MPIDI_PT_single_op * single_op;
							
							single_op = lock_queue->pt_single_op;
							if (single_op->type == MPIDI_RMA_PUT)
							{
								mpi_errno = MPIR_Localcopy(single_op->data,
														   single_op->count,
														   single_op->datatype,
														   single_op->addr,
														   single_op->count,
														   single_op->datatype);
							}   
							else if (single_op->type == MPIDI_RMA_ACCUMULATE)
							{
								mpi_errno = do_simple_accumulate(single_op);
							}
							else if (single_op->type == MPIDI_RMA_GET)
							{
								mpi_errno = do_simple_get(win_ptr, lock_queue);
							}
							
							/* --BEGIN ERROR HANDLING-- */
							if (mpi_errno != MPI_SUCCESS) goto fn_exit;
							/* --END ERROR HANDLING-- */
							
							/* if put or accumulate, send rma done packet and release lock. */
							if (single_op->type != MPIDI_RMA_GET)
							{
								/* increment counter */
								win_ptr->my_pt_rma_puts_accs++;
								
								mpi_errno = MPIDI_CH3I_Send_pt_rma_done_pkt(lock_queue->vc, lock_queue->source_win_handle);
								/* --BEGIN ERROR HANDLING-- */
								if (mpi_errno != MPI_SUCCESS) goto fn_exit;
								/* --END ERROR HANDLING-- */
								
								/* release the lock */
								if (win_ptr->current_lock_type == MPI_LOCK_SHARED)
								{
									/* decr ref cnt */
									/* FIXME: MT: Must be done atomically */
									win_ptr->shared_lock_ref_cnt--;
								}
								
								/* If shared lock ref count is 0 
								   (which is also true if the lock is an
								   exclusive lock), release the lock. */
								if (win_ptr->shared_lock_ref_cnt == 0)
								{
									/* FIXME: MT: The setting of the lock type 
									   must be done atomically */
									win_ptr->current_lock_type = MPID_LOCK_NONE;
								}
								
								/* dequeue entry from lock queue */
								MPIU_Free(single_op->data);
								MPIU_Free(single_op);
								*lock_queue_ptr = lock_queue->next;
								MPIU_Free(lock_queue);
								lock_queue = *lock_queue_ptr;
							}
							else
							{
								/* it's a get. The operation is not complete. It 
								   will be completed in ch3u_handle_send_req.c. 
								   Free the single_op structure. If it's an 
								   exclusive lock, break. Otherwise continue to the
								   next operation. */
								
								MPIU_Free(single_op);
								*lock_queue_ptr = lock_queue->next;
								MPIU_Free(lock_queue);
								lock_queue = *lock_queue_ptr;
								
								if (requested_lock == MPI_LOCK_EXCLUSIVE)
									break;
							}
						}
						else
						{
							/* send lock granted packet. */
							mpi_errno = MPIDI_CH3I_Send_lock_granted_pkt(lock_queue->vc, lock_queue->source_win_handle);
							
							/* dequeue entry from lock queue */
							*lock_queue_ptr = lock_queue->next;
							MPIU_Free(lock_queue);
							lock_queue = *lock_queue_ptr;
							
							/* if the granted lock is exclusive, no need to continue */
							if (requested_lock == MPI_LOCK_EXCLUSIVE)
								break;
						}
					}
				}
				else
				{
					lock_queue_ptr = &(lock_queue->next);
					lock_queue = lock_queue->next;
				}
			}
		} while (temp_entered_count != entered_count);
		entered_count = entered_flag = 0;
	}

 fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_RELEASE_LOCK);
	return mpi_errno;
}


#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_Send_pt_rma_done_pkt
#undef FCNAME 
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3I_Send_pt_rma_done_pkt(MPIDI_VC_t *vc, MPI_Win source_win_handle)
{
	MPIDI_CH3_Pkt_t upkt;
	MPIDI_CH3_Pkt_pt_rma_done_t *pt_rma_done_pkt = &upkt.pt_rma_done;
	MPID_Request *req;
	int mpi_errno=MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_SEND_PT_RMA_DONE_PKT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_SEND_PT_RMA_DONE_PKT);

	MPIDI_Pkt_init(pt_rma_done_pkt, MPIDI_CH3_PKT_PT_RMA_DONE);
	pt_rma_done_pkt->source_win_handle = source_win_handle;

	mpi_errno = MPIDI_CH3.iStartMsg(vc, pt_rma_done_pkt, sizeof(*pt_rma_done_pkt), &req);
	/* --BEGIN ERROR HANDLING-- */
	if (mpi_errno != MPI_SUCCESS)
	{
		mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|rmamsg", 0);
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SEND_PT_RMA_DONE_PKT);
		return mpi_errno;
	}
	/* --END ERROR HANDLING-- */
	if (req != NULL)
	{
		MPID_Request_release(req);
	}

	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SEND_PT_RMA_DONE_PKT);
	return mpi_errno;
}


#undef FUNCNAME
#define FUNCNAME do_simple_accumulate
#undef FCNAME 
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int do_simple_accumulate(MPIDI_PT_single_op *single_op)
{
	int mpi_errno = MPI_SUCCESS;
	MPI_User_function *uop;
	MPIDI_STATE_DECL(MPID_STATE_DO_SIMPLE_ACCUMULATE);

	MPIDI_FUNC_ENTER(MPID_STATE_DO_SIMPLE_ACCUMULATE);

	if (single_op->op == MPI_REPLACE)
	{
		/* simply copy the data */
		mpi_errno = MPIR_Localcopy(single_op->data, single_op->count,
								   single_op->datatype, single_op->addr,
								   single_op->count, single_op->datatype);
		/* --BEGIN ERROR HANDLING-- */
		if (mpi_errno)
		{
			mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", 0);
			MPIDI_FUNC_EXIT(MPID_STATE_DO_SIMPLE_ACCUMULATE);
			return mpi_errno;
		}
		/* --END ERROR HANDLING-- */
		goto fn_exit;
	}

	if (HANDLE_GET_KIND(single_op->op) == HANDLE_KIND_BUILTIN)
	{
		/* get the function by indexing into the op table */
		uop = MPIR_Op_table[(single_op->op)%16 - 1];
	}
	else
	{
		/* --BEGIN ERROR HANDLING-- */
		mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OP, L"**opnotpredefined", L"**opnotpredefined %d", single_op->op );
		MPIDI_FUNC_EXIT(MPID_STATE_DO_SIMPLE_ACCUMULATE);
		return mpi_errno;
		/* --END ERROR HANDLING-- */
	}
	
	/* only basic datatypes supported for this optimization. */
	(*uop)(single_op->data, single_op->addr,
		   &(single_op->count), &(single_op->datatype));

 fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_DO_SIMPLE_ACCUMULATE);
	return mpi_errno;
}



#undef FUNCNAME
#define FUNCNAME do_simple_get
#undef FCNAME 
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int do_simple_get(MPID_Win *win_ptr, MPIDI_Win_lock_queue *lock_queue)
{
	MPIDI_CH3_Pkt_t upkt;
	MPIDI_CH3_Pkt_get_resp_t * get_resp_pkt = &upkt.get_resp;
	MPID_Request *req;
	MPID_IOV iov[MPID_IOV_LIMIT];
	int type_size, mpi_errno=MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_DO_SIMPLE_GET);

	MPIDI_FUNC_ENTER(MPID_STATE_DO_SIMPLE_GET);

	req = MPID_Request_create();
	if (req == NULL)
	{
		/* --BEGIN ERROR HANDLING-- */
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
		MPIDI_FUNC_EXIT(MPID_STATE_DO_SIMPLE_GET);
		return mpi_errno;
		/* --END ERROR HANDLING-- */
	}
	req->dev.target_win_handle = win_ptr->handle;
	req->dev.source_win_handle = lock_queue->source_win_handle;
	req->dev.single_op_opt = 1;
	req->dev.ca = MPIDI_CH3_CA_COMPLETE;
	
	MPIDI_Request_set_type(req, MPIDI_REQUEST_TYPE_GET_RESP); 
	req->kind = MPID_REQUEST_SEND;
	
	MPIDI_Pkt_init(get_resp_pkt, MPIDI_CH3_PKT_GET_RESP);
	get_resp_pkt->request_handle = lock_queue->pt_single_op->request_handle;
	
	iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST) get_resp_pkt;
	iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)sizeof(*get_resp_pkt);
	
	iov[1].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)lock_queue->pt_single_op->addr;
	MPID_Datatype_get_size_macro(lock_queue->pt_single_op->datatype, type_size);
	iov[1].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)(lock_queue->pt_single_op->count * type_size);
	
	mpi_errno = MPIDI_CH3.iSendv(lock_queue->vc, req, iov, 2);
	/* --BEGIN ERROR HANDLING-- */
	if (mpi_errno != MPI_SUCCESS)
	{
		MPIU_Object_set_ref(req, 0);
		MPIDI_CH3_Request_destroy(req);
		mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|rmamsg", 0);
	}
	/* --END ERROR HANDLING-- */

	MPIDI_FUNC_EXIT(MPID_STATE_DO_SIMPLE_GET);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3U_Handle_send_req
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3U_Handle_send_req(MPIDI_VC_t * vc, MPID_Request * sreq, int *complete)
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3U_HANDLE_SEND_REQ);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3U_HANDLE_SEND_REQ);

	MPIU_UNREFERENCED_ARG(vc);

	switch(sreq->dev.ca)
	{
		case MPIDI_CH3_CA_COMPLETE:
		{
			if (MPIDI_Request_get_type(sreq) == MPIDI_REQUEST_TYPE_GET_RESP)
			{ 
				if (sreq->dev.source_win_handle != MPI_WIN_NULL)
				{
					MPID_Win *win_ptr;
					/* Last RMA operation (get) from source. If active target RMA,
					   decrement window counter. If passive target RMA, 
					   release lock on window and grant next lock in the 
					   lock queue if there is any; no need to send rma done 
					   packet since the last operation is a get. */

					MPID_Win_get_ptr(sreq->dev.target_win_handle, win_ptr);
					if (win_ptr->current_lock_type == MPID_LOCK_NONE)
					{
						/* FIXME: MT: this has to be done atomically */
						win_ptr->my_counter -= 1;
					}
					else
					{
						mpi_errno = MPIDI_CH3I_Release_lock(win_ptr);
					}
				}
			}

			/* mark data transfer as complete and decrement CC */
			MPIDI_CH3U_Request_complete(sreq);
			*complete = TRUE;

			break;
		}
		
		case MPIDI_CH3_CA_RELOAD_IOV:
		{
			sreq->dev.iov_count = MPID_IOV_LIMIT;
			mpi_errno = MPIDI_CH3U_Request_load_send_iov(sreq, sreq->dev.iov, &sreq->dev.iov_count);
			/* --BEGIN ERROR HANDLING-- */
			if (mpi_errno != MPI_SUCCESS)
			{
				mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|loadsendiov", 0);
				goto fn_exit;
			}
			/* --END ERROR HANDLING-- */
			
			*complete = FALSE;
			break;
		}
		/* --BEGIN ERROR HANDLING-- */
		default:
		{
			*complete = FALSE;
			mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_INTERN, L"**ch3|badca", L"**ch3|badca %d", sreq->dev.ca);
			break;
		}
		/* --END ERROR HANDLING-- */
	}

  fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3U_HANDLE_SEND_REQ);
	return mpi_errno;
}

/*  MPIDI_CH3U_Init_sock - does socket specific channel initialization
 *	 publish_bc_p - if non-NULL, will be a pointer to the original position 
 *					of the bc_val and should
 *					do KVS Put/Commit/Barrier on business card before 
 *					returning
 *	 bc_key_p	 - business card key buffer pointer.  
 *	 bc_val_p	 - business card value buffer pointer, updated to the next
 *					available location or freed if published.
 *	 val_max_sz_p - ptr to maximum value buffer size reduced by the number 
 *					of characters written
 *							   
 */

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3U_Init_sock
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3U_Init_sock(int has_parent, MPIDI_PG_t *pg_p, int pg_rank, wchar_t **bc_val_p, int *val_max_sz_p)
{
	int mpi_errno = MPI_SUCCESS;
	int pmi_errno;
	int pg_size;
	int p;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3U_INIT_SOCK);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3U_INIT_SOCK);

	/* FIXME: Why are these unused? */
	MPIU_UNREFERENCED_ARG(has_parent);
	MPIU_UNREFERENCED_ARG(pg_rank);

	/*
	 * Initialize the VCs associated with this process group (and thus 
	 * MPI_COMM_WORLD)
	 */

	/* FIXME: Get the size from the process group */
	pmi_errno = PMI_Get_size(&pg_size);
	if (pmi_errno != 0)
	{
		MPIU_ERR_SETANDJUMP2(mpi_errno,MPI_ERR_OTHER, L"**pmi_get_size", L"**pmi_get_size %d %s", pmi_errno, PMI_Get_last_error_msgW());
	}

	/* FIXME: This should probably be the same as MPIDI_VC_InitSock.  If
	   not, why not? */
	/* FIXME: Note that MPIDI_CH3_VC_Init sets state, sendq_head and tail.
	   so this should be MPIDI_CH3_VC_Init( &pg_p->vct[p] );
	   followed by MPIDI_VC_InitSock( ditto );  
	   In fact, there should be a single VC_Init call here */
	for (p = 0; p < pg_size; p++)
	{
		pg_p->vct[p].ch.sendq_head = NULL;
		pg_p->vct[p].ch.sendq_tail = NULL;
		pg_p->vct[p].ch.state = MPIDI_CH3I_VC_STATE_UNCONNECTED;
		pg_p->vct[p].ch.sock = MPIDU_SOCK_INVALID_SOCK;
		pg_p->vct[p].ch.conn = NULL;
	}	

	mpi_errno = MPIDI_CH3U_Get_business_card_sock(bc_val_p, val_max_sz_p);
	if (mpi_errno != MPI_SUCCESS)
	{
		MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, L"**init_buscard");
	}

 fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3U_INIT_SOCK);
	return mpi_errno;
	
 fn_fail:
	/* FIXME: This doesn't belong here, since the pg is not created in this routine */
	/* --BEGIN ERROR HANDLING-- */
	if (pg_p != NULL) 
	{
		MPIDI_PG_Destroy(pg_p);
	}

	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

/* This routine initializes Sock-specific elements of the VC */
#undef FUNCNAME
#define FUNCNAME MPIDI_VC_InitSock
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_VC_InitSock( MPIDI_VC_t *vc ) 
{
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_VC_INITSOCK);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_VC_INITSOCK);
	vc->ch.sock			   = MPIDU_SOCK_INVALID_SOCK;
	vc->ch.conn			   = NULL;
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_VC_INITSOCK);
	return MPI_SUCCESS;
}

#ifdef USE_DBG_LOGGING
const char * MPIDI_Conn_GetStateString(int state) 
{
	const char *name = "unknown";

	switch (state) {
	case CONN_STATE_UNCONNECTED:	 name = "CONN_STATE_UNCONNECTED"; break;
	case CONN_STATE_LISTENING:	   name = "CONN_STATE_LISTENING"; break;
	case CONN_STATE_CONNECTING:	  name = "CONN_STATE_CONNECTING"; break;
	case CONN_STATE_CONNECT_ACCEPT:  name = "CONN_STATE_CONNECT_ACCEPT"; break; 
	case CONN_STATE_OPEN_CSEND:	  name = "CONN_STATE_OPEN_CSEND"; break;
	case CONN_STATE_OPEN_CRECV:	  name = "CONN_STATE_OPEN_CRECV"; break;
	case CONN_STATE_OPEN_LRECV_PKT:  name = "CONN_STATE_OPEN_LRECV_PKT"; break;
	case CONN_STATE_OPEN_LRECV_DATA: name = "CONN_STATE_OPEN_LRECV_DATA"; break;
	case CONN_STATE_OPEN_LSEND:	  name = "CONN_STATE_OPEN_LSEND"; break;
	case CONN_STATE_CONNECTED:	   name = "CONN_STATE_CONNECTED"; break;
	case CONN_STATE_CLOSING:		 name = "CONN_STATE_CLOSING"; break;
	case CONN_STATE_CLOSED:		  name = "CONN_STATE_CLOSED"; break;
	case CONN_STATE_FAILED:		  name = "CONN_STATE_FAILE"; break;
	}

	return name;
}

#endif

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_Complete_unidirectional_connection
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3_Complete_unidirectional_connection_connect( MPIDI_VC_t *connect_vc )
{
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_COMPLETE_UNIDIRECTIONAL_CONNECTION);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_COMPLETE_UNIDIRECTIONAL_CONNECTION);
	MPIU_UNREFERENCED_ARG(connect_vc);
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_COMPLETE_UNIDIRECTIONAL_CONNECTION);
	return MPI_SUCCESS;
}

/* This routine is the "other side" of MPIDI_CH3_Complete_unidirectional_connection */
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_Complete_unidirectional_connection2
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3_Complete_unidirectional_connection_accept( MPIDI_VC_t *new_vc )
{
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_COMPLETE_UNIDIRECTIONAL_CONNECTION2);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_COMPLETE_UNIDIRECTIONAL_CONNECTION2);
	MPIU_UNREFERENCED_ARG(new_vc);
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_COMPLETE_UNIDIRECTIONAL_CONNECTION2);
	return MPI_SUCCESS;
}

/* This routine is used to free any channel-specific resources that were allocated when building a connection */
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_Cleanup_after_connection
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3_Cleanup_after_connection( MPIDI_VC_t *new_vc )
{
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_CLEANUP_AFTER_CONNECTION);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_CLEANUP_AFTER_CONNECTION);
	MPIU_UNREFERENCED_ARG(new_vc);
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_CLEANUP_AFTER_CONNECTION);
	return MPI_SUCCESS;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_Create_inter_root_communicator_connect
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int MPIDI_Create_inter_root_communicator_connect(const wchar_t *port_name, MPID_Comm **comm_pptr, MPIDI_VC_t **vc_pptr)
{
	int mpi_errno = MPI_SUCCESS;
	MPID_Comm *tmp_comm;
	MPIDI_VC_t *connect_vc = NULL;
	int port_name_tag;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CREATE_INTER_ROOT_COMMUNICATOR_CONNECT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CREATE_INTER_ROOT_COMMUNICATOR_CONNECT);

	/* Connect to the root on the other side. Create a
	   temporary intercommunicator between the two roots so that
	   we can use MPI functions to communicate data between them. */

	mpi_errno = MPIDI_CH3.Connect_to_root(port_name, &connect_vc);
	MPIU_ERR_CHK_POP(mpi_errno);

    /* extract the tag from the port_name */
    mpi_errno = MPIDI_GetTagFromPort( port_name, &port_name_tag);
    if (mpi_errno != MPIU_STR_SUCCESS)
	{
		MPIU_ERR_POP(mpi_errno);
    }

	mpi_errno = MPIDI_CH3I_Initialize_tmp_comm(&tmp_comm, connect_vc, 1, port_name_tag);
	MPIU_ERR_CHK_POP(mpi_errno);

	/* If the VC creates non-duplex connections then the acceptor will
	 * need to connect back to form the other half of the connection. */
	/* FIXME: A hook should not be such a specific function; instead,
	   it should invoke a function pointer defined in the channel 
	   interface structure */
	mpi_errno = MPIDI_CH3.Complete_unidirectional_connection_connect( connect_vc );

	*comm_pptr = tmp_comm;
	*vc_pptr = connect_vc;

 fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CREATE_INTER_ROOT_COMMUNICATOR_CONNECT);
	return mpi_errno;
 fn_fail:
	goto fn_exit;
}


/*
   MPIDI_Comm_connect()

   Algorithm: First create a connection (vc) between this root and the
   root on the accept side. Using this vc, create a temporary
   intercomm between the two roots. Use MPI functions to communicate
   the other information needed to create the real intercommunicator
   between the processes on the two sides. Then free the
   intercommunicator between the roots. Most of the complexity is
   because there can be multiple process groups on each side. 
*/ 

#undef FUNCNAME
#define FUNCNAME MPIDI_Comm_connect
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_Comm_connect(const wchar_t *port_name, MPID_Info *info, int root, MPID_Comm *comm_ptr, MPID_Comm **newcomm)
{
	int mpi_errno=MPI_SUCCESS;
	int j, i, rank, recv_ints[3], send_ints[2], context_id;
	int remote_comm_size=0;
	MPID_Comm *tmp_comm = NULL, *intercomm;
	MPIDI_VC_t *new_vc = NULL;
	int sendtag=100, recvtag=100, n_remote_pgs;
	int n_local_pgs=1, local_comm_size;
	pg_translation *local_translation = NULL, *remote_translation = NULL;
	pg_node *pg_list = NULL;
	MPIDI_PG_t **remote_pg = NULL;
	MPIU_CHKLMEM_DECL(3);
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_COMM_CONNECT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_COMM_CONNECT);

	MPIU_UNREFERENCED_ARG(info);

    /* Create the new intercommunicator here. We need to send the
       context id to the other side. */
    /* FIXME: If we fail to connect, someone needs to free this newcomm */
    mpi_errno = MPIR_Comm_create(newcomm);
	MPIU_ERR_CHK_POP(mpi_errno);
    mpi_errno = MPIR_Get_contextid( comm_ptr, &(*newcomm)->recvcontext_id );
    MPIU_ERR_CHK_POP(mpi_errno);
    /* FIXME why is this commented out? */
    /* (*newcomm)->context_id = (*newcomm)->recvcontext_id; */
	/*wprintf(L"connect context id: %d\n", (*newcomm)->recvcontext_id);fflush(stdout);*/

	/*wprintf(L"comm_connect: <%s>\n", port_name);fflush(stdout);*/
	rank = comm_ptr->rank;
	local_comm_size = comm_ptr->local_size;

	if (rank == root)
	{
		/* Establish a communicator to communicate with the root on the other side. */
		mpi_errno = MPIDI_Create_inter_root_communicator_connect(port_name, &tmp_comm, &new_vc);
		MPIU_ERR_CHK_POP(mpi_errno);

		/* Make an array to translate local ranks to process group index and rank */
		MPIU_CHKLMEM_MALLOC(local_translation,pg_translation*, local_comm_size*sizeof(pg_translation), mpi_errno,L"local_translation");

		/* Make a list of the local communicator's process groups and encode 
		   them in strings to be sent to the other side.
		   The encoded string for each process group contains the process 
		   group id, size and all its KVS values */
		mpi_errno = ExtractLocalPGInfo( comm_ptr, local_translation, &pg_list, &n_local_pgs );
		MPIU_ERR_CHK_POP(mpi_errno);

		/* Send the remote root: n_local_pgs, local_comm_size,
		   Recv from the remote root: n_remote_pgs, remote_comm_size,
		   context_id for newcomm */

		send_ints[0] = n_local_pgs;
		send_ints[1] = local_comm_size;
        send_ints[2] = (*newcomm)->recvcontext_id;

		MPIU_DBG_MSG_FMT(CH3_CONNECT,VERBOSE,(MPIU_DBG_FDEST, L"sending two ints, %d and %d, and receiving 3 ints", send_ints[0], send_ints[1]));
		mpi_errno = MPIC_Sendrecv(send_ints, 3, MPI_INT, 0, sendtag++, recv_ints, 3, MPI_INT, 0, recvtag++, tmp_comm->handle, MPI_STATUS_IGNORE);
		MPIU_ERR_CHK_POP(mpi_errno);
	}

	/* broadcast the received info to local processes */
	MPIU_DBG_MSG(CH3_CONNECT,VERBOSE,L"broadcasting the received 3 ints");
	/*wprintf(L"connect broadcasting the received 3 ints from root %d, my rank %d, context_id %d\n", root, comm_ptr->rank, recv_ints[2]);fflush(stdout);*/
	mpi_errno = MPIR_Bcast(recv_ints, 3, MPI_INT, root, comm_ptr);
	MPIU_ERR_CHK_POP(mpi_errno);

	n_remote_pgs = recv_ints[0];
	remote_comm_size = recv_ints[1];
	context_id = recv_ints[2];
	MPIU_CHKLMEM_MALLOC(remote_pg,MPIDI_PG_t**, n_remote_pgs * sizeof(MPIDI_PG_t*), mpi_errno,L"remote_pg");
	MPIU_CHKLMEM_MALLOC(remote_translation,pg_translation*, remote_comm_size * sizeof(pg_translation), mpi_errno,L"remote_translation");
	MPIU_DBG_MSG(CH3_CONNECT,VERBOSE,L"allocated remote process groups");

	/* Exchange the process groups and their corresponding KVSes */
	if (rank == root)
	{
		mpi_errno = SendPGtoPeerAndFree( tmp_comm, &sendtag, pg_list );
		mpi_errno = ReceivePGAndDistribute( tmp_comm, comm_ptr, root, &recvtag, n_remote_pgs, remote_pg );
		/* Receive the translations from remote process rank to process group index */
		MPIU_DBG_MSG_FMT(CH3_CONNECT,VERBOSE,(MPIU_DBG_FDEST, L"sending %d ints, receiving %d ints", local_comm_size * 2, remote_comm_size * 2));
		mpi_errno = MPIC_Sendrecv(local_translation, local_comm_size * 2, 
								  MPI_INT, 0, sendtag++,
								  remote_translation, remote_comm_size * 2, 
								  MPI_INT, 0, recvtag++, tmp_comm->handle, 
								  MPI_STATUS_IGNORE);
		MPIU_ERR_CHK_POP(mpi_errno);

#ifdef MPICH_DBG_OUTPUT
		MPIU_DBG_PRINTF((L"[%d]connect:Received remote_translation:\n", rank));
		for (i=0; i<remote_comm_size; i++)
		{
			MPIU_DBG_PRINTF((L" remote_translation[%d].pg_index = %d\n remote_translation[%d].pg_rank = %d\n", i, remote_translation[i].pg_index, i, remote_translation[i].pg_rank));
		}
#endif
	}
	else
	{
		mpi_errno = ReceivePGAndDistribute( tmp_comm, comm_ptr, root, &recvtag, n_remote_pgs, remote_pg );
		MPIU_ERR_CHK_POP(mpi_errno);
	}

	/* Broadcast out the remote rank translation array */
	MPIU_DBG_MSG(CH3_CONNECT,VERBOSE,L"Broadcasting remote translation");
	mpi_errno = MPIR_Bcast(remote_translation, remote_comm_size * 2, MPI_INT, root, comm_ptr);
	MPIU_ERR_CHK_POP(mpi_errno);
#ifdef MPICH_DBG_OUTPUT
	MPIU_DBG_PRINTF((L"[%d]connect:Received remote_translation after broadcast:\n", rank));
	for (i=0; i<remote_comm_size; i++)
	{
		MPIU_DBG_PRINTF((L" remote_translation[%d].pg_index = %d\n remote_translation[%d].pg_rank = %d\n", i, remote_translation[i].pg_index, i, remote_translation[i].pg_rank));
	}
#endif

	intercomm = *newcomm;
	intercomm->context_id = (int16_t)context_id;
	/*wprintf(L"connect remote context id: %d\n", context_id);fflush(stdout);*/
	intercomm->is_low_group = 1;

	mpi_errno = SetupNewIntercomm( comm_ptr, remote_comm_size, remote_translation, remote_pg, intercomm );
	MPIU_ERR_CHK_POP(mpi_errno);

	/* synchronize with remote root */
	if (rank == root)
	{
		MPIU_DBG_MSG(CH3_CONNECT,VERBOSE,L"sync with peer");
		mpi_errno = MPIC_Sendrecv(&i, 0, MPI_INT, 0, sendtag++, &j, 0, MPI_INT, 0, recvtag++, tmp_comm->handle, MPI_STATUS_IGNORE);
		MPIU_ERR_CHK_POP(mpi_errno);

		/* All communication with remote root done. Release the communicator. */
		MPIR_Comm_release(tmp_comm, 0);
	}

	/*wprintf(L"connect:barrier\n");fflush(stdout);*/
	mpi_errno = MPIR_Barrier(comm_ptr);
	MPIU_ERR_CHK_POP(mpi_errno);

	/* Free new_vc. It was explicitly allocated in MPIDI_CH3_Connect_to_root.*/
	if (rank == root)
	{
		mpi_errno = FreeNewVC( new_vc );
		MPIU_ERR_CHK_POP(mpi_errno);
	}

 fn_exit: 
	MPIU_DBG_MSG(CH3_CONNECT,VERBOSE,L"Exiting ch3u_comm_connect");
	MPIU_CHKLMEM_FREEALL();
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_COMM_CONNECT);
	return mpi_errno;
 fn_fail:
	goto fn_exit;
}

/*
 * Extract all of the process groups from the given communicator and 
 * form a list (returned in pg_list) of those process groups.
 * Also returned is an array (local_translation) that contains tuples mapping
 * rank in process group to rank in that communicator (local translation
 * must be allocated before this routine is called).  The number of 
 * distinct process groups is returned in n_local_pgs_p .
 *
 * This allows an intercomm_create to exchange the full description of
 * all of the process groups that have made up the communicator that
 * will define the "remote group".  
 */
#undef FUNCNAME
#define FUNCNAME ExtractLocalPGInfo
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int ExtractLocalPGInfo( MPID_Comm *comm_p, pg_translation local_translation[], pg_node **pg_list_p, int *n_local_pgs_p )
{
	pg_node		*pg_list = 0, *pg_iter, *pg_trailer;
	int			i, cur_index = 0, local_comm_size, mpi_errno = 0;
	MPIU_CHKPMEM_DECL(1);
	MPIDI_STATE_DECL(MPID_STATE_EXTRACTLOCALPGINFO);

	MPIDI_FUNC_ENTER(MPID_STATE_EXTRACTLOCALPGINFO);
	local_comm_size = comm_p->local_size;

	/* Make a list of the local communicator's process groups and encode 
	   them in strings to be sent to the other side.
	   The encoded string for each process group contains the process 
	   group id, size and all its KVS values */
	
	cur_index = 0;
	MPIU_CHKPMEM_MALLOC(pg_list,pg_node*,sizeof(pg_node),mpi_errno, L"pg_list");
	
	pg_list->pg_id = MPIU_Strdup(comm_p->vcr[0]->pg->id);
	pg_list->index = cur_index++;
	pg_list->next = NULL;
	MPIU_Assert( comm_p->vcr[0]->pg->ref_count);
	mpi_errno = MPIDI_PG_To_string(comm_p->vcr[0]->pg, &pg_list->str, &pg_list->lenStr );
	MPIU_ERR_CHK_POP(mpi_errno);
	MPIU_DBG_STMT(CH3_CONNECT,VERBOSE,MPIDI_PrintConnStr(__FILE__,__LINE__,L"PG as string is", pg_list->str ));
	local_translation[0].pg_index = 0;
	local_translation[0].pg_rank = comm_p->vcr[0]->pg_rank;
	pg_iter = pg_list;
	for (i=1; i<local_comm_size; i++)
	{
		pg_iter = pg_list;
		pg_trailer = pg_list;
		while (pg_iter != NULL)
		{
			/* Check to ensure pg is (probably) valid */
			MPIU_Assert(comm_p->vcr[i]->pg->ref_count != 0);
			if (MPIDI_PG_Id_compare(comm_p->vcr[i]->pg->id, pg_iter->pg_id))
			{
				local_translation[i].pg_index = pg_iter->index;
				local_translation[i].pg_rank  = comm_p->vcr[i]->pg_rank;
				break;
			}
			if (pg_trailer != pg_iter)
				pg_trailer = pg_trailer->next;
			pg_iter = pg_iter->next;
		}
		if (pg_iter == NULL)
		{
			/* We use MPIU_Malloc directly because we do not know in 
			   advance how many nodes we may allocate */
			pg_iter = (pg_node*)MPIU_Malloc(sizeof(pg_node));
			if (!pg_iter)
			{
				MPIU_ERR_POP(mpi_errno);
			}
			pg_iter->pg_id = MPIU_Strdup(comm_p->vcr[i]->pg->id);
			pg_iter->index = cur_index++;
			pg_iter->next = NULL;
			mpi_errno = MPIDI_PG_To_string(comm_p->vcr[i]->pg, &pg_iter->str, &pg_iter->lenStr );
			MPIU_ERR_CHK_POP(mpi_errno);
			local_translation[i].pg_index = pg_iter->index;
			local_translation[i].pg_rank = comm_p->vcr[i]->pg_rank;
			pg_trailer->next = pg_iter;
		}
	}

	*n_local_pgs_p = cur_index;
	*pg_list_p	 = pg_list;
	
#ifdef MPICH_DBG_OUTPUT
	pg_iter = pg_list;
	while (pg_iter != NULL)
	{
		MPIU_DBG_PRINTF((L"connect:PG: '%s'\n<%s>\n", pg_iter->pg_id, pg_iter->str));
		pg_iter = pg_iter->next;
	}
#endif


 fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_EXTRACTLOCALPGINFO);
	return mpi_errno;
 fn_fail:
	MPIU_CHKPMEM_REAP();
	goto fn_exit;
}


/* The root process in comm_ptr receives strings describing the
   process groups and then distributes them to the other processes
   in comm_ptr.
   See SendPGToPeer for the routine that sends the descriptions */
#undef FUNCNAME
#define FUNCNAME ReceivePGAndDistribute
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int ReceivePGAndDistribute( MPID_Comm *tmp_comm, MPID_Comm *comm_ptr, int root, int *recvtag_p, int n_remote_pgs, MPIDI_PG_t *remote_pg[] )
{
	wchar_t *pg_str = 0;
	int  i, j, flag;
	int  rank = comm_ptr->rank;
	int  mpi_errno = 0;
	int  recvtag = *recvtag_p;
	MPIDI_STATE_DECL(MPID_STATE_RECEIVEPGANDDISTRIBUTE);

	MPIDI_FUNC_ENTER(MPID_STATE_RECEIVEPGANDDISTRIBUTE);

	for (i=0; i<n_remote_pgs; i++)
	{
		if (rank == root)
		{
			/* First, receive the pg description from the partner */
			mpi_errno = MPIC_Recv(&j, 1, MPI_INT, 0, recvtag++, tmp_comm->handle, MPI_STATUS_IGNORE);
			*recvtag_p = recvtag;
			MPIU_ERR_CHK_POP(mpi_errno);
			pg_str = (wchar_t*)MPIU_Malloc(j * sizeof(wchar_t));
			if (pg_str == NULL)
			{
				MPIU_ERR_POP(mpi_errno);
			}
			mpi_errno = MPIC_Recv(pg_str, j, MPI_WCHAR, 0, recvtag++, tmp_comm->handle, MPI_STATUS_IGNORE);
			*recvtag_p = recvtag;
			MPIU_ERR_CHK_POP(mpi_errno);
		}

		/* Broadcast the size and data to the local communicator */
		/*wprintf(L"accept:broadcasting 1 int\n");fflush(stdout);*/
		mpi_errno = MPIR_Bcast(&j, 1, MPI_INT, root, comm_ptr);
		MPIU_ERR_CHK_POP(mpi_errno);

		if (rank != root)
		{
			/* The root has already allocated this string */
			pg_str = (wchar_t*)MPIU_Malloc(j*sizeof(wchar_t));
			if (pg_str == NULL)
			{
				MPIU_ERR_POP(mpi_errno);
			}
		}
		/*wprintf(L"accept:broadcasting string of length %d\n", j);fflush(stdout);*/
		mpi_errno = MPIR_Bcast(pg_str, j, MPI_WCHAR, root, comm_ptr);
		MPIU_ERR_CHK_POP(mpi_errno);
		/* Then reconstruct the received process group.  This step also initializes the created process group */

		MPIU_DBG_STMT(CH3_CONNECT,VERBOSE,MPIDI_PrintConnStr(__FILE__,__LINE__,L"Creating pg from string", pg_str ));
		mpi_errno = MPIDI_PG_Create_from_string(pg_str, &remote_pg[i], &flag);
		MPIU_ERR_CHK_POP(mpi_errno);
		
		MPIU_Free(pg_str);
		if (flag)
		{
			/* extra pg ref needed for shared memory modules because the shm_XXXXing_list's
			 * need to be walked though in the later stages of finalize to free queue_info's.
			 */
			if (MPIDI_CH3.PG_Init != NULL)
				MPIDI_CH3.PG_Init(remote_pg[i]);
		}
	}
 fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_RECEIVEPGANDDISTRIBUTE);
	return mpi_errno;
 fn_fail:
	goto fn_exit;
}

/* Used internally to broadcast process groups belonging to peercomm to
 all processes in comm.  The process with rank root in comm is the 
 process in peercomm from which the process groups are taken. This routine 
 is collective over comm_p . */
#undef FUNCNAME
#define FUNCNAME MPID_PG_BCast
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPID_PG_BCast( MPID_Comm *peercomm_p, MPID_Comm *comm_p, int root )
{
	int n_local_pgs=0, mpi_errno = 0;
	pg_translation *local_translation = 0;
	pg_node *pg_list, *pg_next, *pg_head = 0;
	int rank, i, peer_comm_size;
	MPIU_CHKLMEM_DECL(1);
	MPIDI_STATE_DECL(MPID_STATE_MPID_PG_BCAST);

	MPIDI_FUNC_ENTER(MPID_STATE_MPID_PG_BCAST);

	peer_comm_size = comm_p->local_size;
	rank			= comm_p->rank;

	MPIU_CHKLMEM_MALLOC(local_translation,pg_translation*, peer_comm_size*sizeof(pg_translation), mpi_errno,L"local_translation");
	
	if (rank == root)
	{
		/* Get the process groups known to the *peercomm* */
		ExtractLocalPGInfo( peercomm_p, local_translation, &pg_head, &n_local_pgs );
	}

	/* Now, broadcast the number of local pgs */
	NMPI_Bcast(mpi_errno, ( &n_local_pgs, 1, MPI_INT, root, comm_p->handle ));

	pg_list = pg_head;
	for (i=0; i<n_local_pgs; i++)
	{
		int len, flag;
		wchar_t *pg_str=0;
		MPIDI_PG_t *pgptr;

		if (rank == root)
		{
			if (!pg_list)
			{
				/* FIXME: Error, the pg_list is broken */
				wprintf( L"Unexpected end of pg_list\n" ); fflush(stdout);
				break;
			}
			
			pg_str  = pg_list->str;
			len	 = pg_list->lenStr + 1;
			pg_list = pg_list->next;
		}
		NMPI_Bcast(mpi_errno, ( &len, 1, MPI_INT, root, comm_p->handle ));
		if (rank != root)
		{
			pg_str = (wchar_t *)MPIU_Malloc(len*sizeof(wchar_t));
		}
		NMPI_Bcast(mpi_errno, ( pg_str, len, MPI_WCHAR, root, comm_p->handle ));
		if (rank != root)
		{
			/* flag is true if the pg was created, false if it
			   already existed. This step also initializes the created process group  */
			MPIDI_PG_Create_from_string( pg_str, &pgptr, &flag );
			if (flag)
			{
				/*wprintf( L"[%d]Added pg named %s to list\n", rank, (wchar_t *)pgptr->id ); fflush(stdout); */
			}
			MPIU_Free( pg_str );
		}
	}

	/* Free pg_list */
	pg_list = pg_head;
#if 1   /* used to try and detect an error */
	/* FIXME: We should use the PG destroy function for this, and ensure that
	   the PG fields are valid for that function */
	while (pg_list)
	{
		pg_next = pg_list->next;
		MPIU_Free( pg_list->str );
		if (pg_list->pg_id )
		{
			MPIU_Free( pg_list->pg_id );
		}
		MPIU_Free( pg_list );
		pg_list = pg_next;
	}
#endif

 fn_exit:
	MPIU_CHKLMEM_FREEALL();
	MPIDI_FUNC_EXIT(MPID_STATE_MPID_PG_BCAST);
	return mpi_errno;
 fn_fail:
	goto fn_exit;
}

/* Sends the process group information to the peer and frees the pg_list */
#undef FUNCNAME
#define FUNCNAME SendPGtoPeerAndFree
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int SendPGtoPeerAndFree( MPID_Comm *tmp_comm, int *sendtag_p, pg_node *pg_list )
{
	int mpi_errno = 0;
	int sendtag = *sendtag_p, i;
	pg_node *pg_iter;
	MPIDI_STATE_DECL(MPID_STATE_SENDPGTOPEERANDFREE);

	MPIDI_FUNC_ENTER(MPID_STATE_SENDPGTOPEERANDFREE);

	while (pg_list != NULL)
	{
		pg_iter = pg_list;
		i = pg_iter->lenStr + 1;
		/*wprintf(L"connect:sending 1 int: %d\n", i);fflush(stdout);*/
		mpi_errno = MPIC_Send(&i, 1, MPI_INT, 0, sendtag++, tmp_comm->handle);
		*sendtag_p = sendtag;
		MPIU_ERR_CHK_POP(mpi_errno);
		
		/* wprintf(L"connect:sending string length %d\n", i);fflush(stdout); */
		mpi_errno = MPIC_Send(pg_iter->str, i, MPI_WCHAR, 0, sendtag++, tmp_comm->handle);
		*sendtag_p = sendtag;
		MPIU_ERR_CHK_POP(mpi_errno);
		
		pg_list = pg_list->next;
		MPIU_Free(pg_iter->str);
		MPIU_Free(pg_iter->pg_id);
		MPIU_Free(pg_iter);
	}

 fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_SENDPGTOPEERANDFREE);
	return mpi_errno;
 fn_fail:
	goto fn_exit;
}

/* ---------------------------------------------------------------------- */
/* Creates a communicator for the purpose of communicating with one other 
   process (the root of the other group).  It also returns the virtual
   connection */
#undef FUNCNAME
#define FUNCNAME MPIDI_Create_inter_root_communicator_accept
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int MPIDI_Create_inter_root_communicator_accept(const wchar_t *port_name, MPID_Comm **comm_pptr, MPIDI_VC_t **vc_pptr)
{
	int mpi_errno = MPI_SUCCESS;
	MPID_Comm *tmp_comm;
	MPIDI_VC_t *new_vc = NULL;
	MPID_Progress_state progress_state;
	int port_name_tag;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CREATE_INTER_ROOT_COMMUNICATOR_ACCEPT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CREATE_INTER_ROOT_COMMUNICATOR_ACCEPT);

	/* FIXME: This code should parallel the MPIDI_CH3_Connect_to_root
	   code used in the MPIDI_Create_inter_root_communicator_connect */
	/* extract the tag from the port_name */
	mpi_errno = MPIDI_GetTagFromPort( port_name, &port_name_tag);
	if (mpi_errno != MPIU_STR_SUCCESS)
	{
		MPIU_ERR_POP(mpi_errno);
	}

	/* FIXME: Describe the algorithm used here, and what routine 
	   is user on the other side of this connection */
	/* dequeue the accept queue to see if a connection with the
	   root on the connect side has been formed in the progress
	   engine (the connection is returned in the form of a vc). If
	   not, poke the progress engine. */

	MPID_Progress_start(&progress_state);
	for(;;)
	{
		MPIDI_CH3I_Acceptq_dequeue(&new_vc, port_name_tag);
		if (new_vc != NULL)
		{
			break;
		}

		mpi_errno = MPID_Progress_wait(&progress_state);
		/* --BEGIN ERROR HANDLING-- */
		if (mpi_errno)
		{
			MPID_Progress_end(&progress_state);
			MPIU_ERR_POP(mpi_errno);
		}
		/* --END ERROR HANDLING-- */
	}
	MPID_Progress_end(&progress_state);

	mpi_errno = MPIDI_CH3I_Initialize_tmp_comm(&tmp_comm, new_vc, 0, port_name_tag);

	/* If the VC creates non-duplex connections then the acceptor will
	 * need to connect back to form the other half of the connection. */
	mpi_errno = MPIDI_CH3.Complete_unidirectional_connection_accept( new_vc );

	*comm_pptr = tmp_comm;
	*vc_pptr = new_vc;

fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CREATE_INTER_ROOT_COMMUNICATOR_ACCEPT);
	return mpi_errno;

fn_fail:
	goto fn_exit;
}

/*
 * MPIDI_Comm_accept()

   Algorithm: First dequeue the vc from the accept queue (it was
   enqueued by the progress engine in response to a connect request
   from the root process that is attempting the connection on 
   the connect side). Use this vc to create an
   intercommunicator between this root and the root on the connect
   side. Use this intercomm. to communicate the other information
   needed to create the real intercommunicator between the processes
   on the two sides. Then free the intercommunicator between the
   roots. Most of the complexity is because there can be multiple
   process groups on each side.

 */
#undef FUNCNAME
#define FUNCNAME MPIDI_Comm_accept
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_Comm_accept(const wchar_t *port_name, MPID_Info *info, int root, MPID_Comm *comm_ptr, MPID_Comm **newcomm)
{
	int mpi_errno=MPI_SUCCESS;
	int i, j, rank, recv_ints[2], send_ints[3], context_id;
	int remote_comm_size=0;
	MPID_Comm *tmp_comm = NULL, *intercomm;
	MPIDI_VC_t *new_vc = NULL;
	int sendtag=100, recvtag=100, local_comm_size;
	int n_local_pgs=1, n_remote_pgs;
	pg_translation *local_translation = NULL, *remote_translation = NULL;
	pg_node *pg_list = NULL;
	MPIDI_PG_t **remote_pg = NULL;
	MPIU_CHKLMEM_DECL(3);
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_COMM_ACCEPT);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_COMM_ACCEPT);

	MPIU_UNREFERENCED_ARG(info);

	/*wprintf(L"comm_accept: <%s>\n", port_name);fflush(stdout);*/
	/* Create the new intercommunicator here. We need to send the
	   context id to the other side. */
	/* FIXME: There is a danger that the context id won't be unique
	   on the other side of this connection */
	mpi_errno = MPIR_Comm_create(newcomm);
	MPIU_ERR_CHK_POP(mpi_errno);

	mpi_errno = MPIR_Get_contextid( comm_ptr, &(*newcomm)->recvcontext_id );
	MPIU_ERR_CHK_POP(mpi_errno);
	/*wprintf(L"accept context id: %d\n", (*newcomm)->recvcontext_id);fflush(stdout); */

	/*
	if ((*newcomm)->context_id == 0)
	{
		MPIU_ERR_SETANDJUMP(mpi_errno, MPI_ERR_OTHER, L"**toomanycomm" );
	}
	*/
	
	rank = comm_ptr->rank;
	local_comm_size = comm_ptr->local_size;
	
	if (rank == root)
	{
		/* Establish a communicator to communicate with the root on the 
		   other side. */
		mpi_errno = MPIDI_Create_inter_root_communicator_accept(port_name, &tmp_comm, &new_vc);
		MPIU_ERR_CHK_POP(mpi_errno);

		/* Make an array to translate local ranks to process group index and 
		   rank */
		MPIU_CHKLMEM_MALLOC(local_translation,pg_translation*, local_comm_size*sizeof(pg_translation), mpi_errno,L"local_translation");

		/* Make a list of the local communicator's process groups and encode 
		   them in strings to be sent to the other side.
		   The encoded string for each process group contains the process 
		   group id, size and all its KVS values */
		mpi_errno = ExtractLocalPGInfo( comm_ptr, local_translation, &pg_list, &n_local_pgs );
		/* Send the remote root: n_local_pgs, local_comm_size, context_id for newcomm.
		   Recv from the remote root: n_remote_pgs, remote_comm_size */

		send_ints[0] = n_local_pgs;
		send_ints[1] = local_comm_size;
		send_ints[2] = (*newcomm)->recvcontext_id;

		/*wprintf(L"accept:sending 3 ints, %d, %d, %d, and receiving 2 ints\n", send_ints[0], send_ints[1], send_ints[2]);fflush(stdout);*/
		mpi_errno = MPIC_Sendrecv(send_ints, 3, MPI_INT, 0, sendtag++, recv_ints, 3, MPI_INT, 0, recvtag++, tmp_comm->handle, MPI_STATUS_IGNORE);
		MPIU_ERR_CHK_POP(mpi_errno);
	}

	/* broadcast the received info to local processes */
	/*wprintf(L"accept:broadcasting 2 ints - %d and %d\n", recv_ints[0], recv_ints[1]);fflush(stdout);*/
	/*wprintf(L"accept broadcasting the received 3 ints from root %d, my rank %d, context_id %d\n", root, comm_ptr->rank, recv_ints[2]);fflush(stdout);*/
	mpi_errno = MPIR_Bcast(recv_ints, 3, MPI_INT, root, comm_ptr);
	MPIU_ERR_CHK_POP(mpi_errno);

	n_remote_pgs		= recv_ints[0];
	remote_comm_size	= recv_ints[1];
	context_id			= recv_ints[2];
	MPIU_CHKLMEM_MALLOC(remote_pg,MPIDI_PG_t**, n_remote_pgs * sizeof(MPIDI_PG_t*), mpi_errno,L"remote_pg");
	MPIU_CHKLMEM_MALLOC(remote_translation,pg_translation*, remote_comm_size * sizeof(pg_translation), mpi_errno, L"remote_translation");
	MPIU_DBG_PRINTF((L"[%d]accept:remote process groups: %d\nremote comm size: %d\n", rank, n_remote_pgs, remote_comm_size));

	/* Exchange the process groups and their corresponding KVSes */
	if (rank == root)
	{
		/* The root receives the PG from the peer (in tmp_comm) and distributes them to the processes in comm_ptr */
		mpi_errno = ReceivePGAndDistribute( tmp_comm, comm_ptr, root, &recvtag, n_remote_pgs, remote_pg );
		
		mpi_errno = SendPGtoPeerAndFree( tmp_comm, &sendtag, pg_list );

		/* Receive the translations from remote process rank to process group index */
		/*wprintf(L"accept:sending %d ints and receiving %d ints\n", local_comm_size * 2, remote_comm_size * 2);fflush(stdout);*/
		mpi_errno = MPIC_Sendrecv(local_translation, local_comm_size * 2, 
								  MPI_INT, 0, sendtag++,
								  remote_translation, remote_comm_size * 2, 
								  MPI_INT, 0, recvtag++, tmp_comm->handle, 
								  MPI_STATUS_IGNORE);
#ifdef MPICH_DBG_OUTPUT
		MPIU_DBG_PRINTF((L"[%d]accept:Received remote_translation:\n", rank));
		for (i=0; i<remote_comm_size; i++)
		{
			MPIU_DBG_PRINTF((L" remote_translation[%d].pg_index = %d\n remote_translation[%d].pg_rank = %d\n", i, remote_translation[i].pg_index, i, remote_translation[i].pg_rank));
		}
#endif
	}
	else
	{
		mpi_errno = ReceivePGAndDistribute( tmp_comm, comm_ptr, root, &recvtag, n_remote_pgs, remote_pg );
		MPIU_ERR_CHK_POP(mpi_errno);
	}

	/* Broadcast out the remote rank translation array */
	MPIU_DBG_MSG(CH3_CONNECT,VERBOSE,L"Broadcast remote_translation");
	mpi_errno = MPIR_Bcast(remote_translation, remote_comm_size * 2, MPI_INT, root, comm_ptr);
	MPIU_ERR_CHK_POP(mpi_errno);
#ifdef MPICH_DBG_OUTPUT
	MPIU_DBG_PRINTF((L"[%d]accept:Received remote_translation after broadcast:\n", rank));
	for (i=0; i<remote_comm_size; i++)
	{
		MPIU_DBG_PRINTF((L" remote_translation[%d].pg_index = %d\n remote_translation[%d].pg_rank = %d\n", i, remote_translation[i].pg_index, i, remote_translation[i].pg_rank));
	}
#endif

	/* Now fill in newcomm */
	intercomm = *newcomm;
	/*wprintf(L"accept remote context id: %d\n", context_id);fflush(stdout);*/
    intercomm->context_id = (MPIR_Context_id_t)context_id;
	intercomm->is_low_group = 0;

	mpi_errno = SetupNewIntercomm( comm_ptr, remote_comm_size, remote_translation, remote_pg, intercomm );
	MPIU_ERR_CHK_POP(mpi_errno);

	/* synchronize with remote root */
	if (rank == root)
	{
		MPIU_DBG_MSG(CH3_CONNECT,VERBOSE,L"sync with peer");
		mpi_errno = MPIC_Sendrecv(&i, 0, MPI_INT, 0, sendtag++, &j, 0, MPI_INT, 0, recvtag++, tmp_comm->handle, MPI_STATUS_IGNORE);
		MPIU_ERR_CHK_POP(mpi_errno);

		/* All communication with remote root done. Release the communicator. */
		MPIR_Comm_release(tmp_comm, 0);
	}

	MPIU_DBG_MSG(CH3_CONNECT,VERBOSE,L"Barrier");
	mpi_errno = MPIR_Barrier(comm_ptr);
	MPIU_ERR_CHK_POP(mpi_errno);

	/* Free new_vc once the connection is completed. It was explicitly allocated in ch3_progress.c and returned by MPIDI_CH3I_Acceptq_dequeue. */
	if (rank == root)
	{
		mpi_errno = FreeNewVC( new_vc );
		MPIU_ERR_CHK_POP(mpi_errno);
	}

fn_exit:
	MPIU_CHKLMEM_FREEALL();
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_COMM_ACCEPT);
	return mpi_errno;

fn_fail:
	goto fn_exit;
}

/* This is a utility routine used to initialize temporary communicators used in connect/accept operations */
#undef FUNCNAME
#define FUNCNAME  MPIDI_CH3I_Initialize_tmp_comm
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int MPIDI_CH3I_Initialize_tmp_comm(MPID_Comm **comm_pptr, MPIDI_VC_t *vc_ptr, int is_low_group, int context_id_offset)
{
	int mpi_errno = MPI_SUCCESS;
	MPID_Comm *tmp_comm, *commself_ptr;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_INITIALIZE_TMP_COMM);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_INITIALIZE_TMP_COMM);

	MPID_Comm_get_ptr( MPI_COMM_SELF, commself_ptr );
	/* WDG-old code allocated a context id that was then discarded */
	mpi_errno = MPIR_Comm_create(&tmp_comm);
	MPIU_ERR_CHK_POP(mpi_errno);
	/* fill in all the fields of tmp_comm. */

    tmp_comm->context_id     = (MPIR_Context_id_t)((MPIR_MAX_CONTEXT_MASK * MPIR_CONTEXT_INT_BITS) + context_id_offset);
    tmp_comm->recvcontext_id = tmp_comm->context_id;
	tmp_comm->remote_size = 1;

	/* Fill in new intercomm */
	tmp_comm->local_size   = 1;
	tmp_comm->rank		 = 0;
	tmp_comm->comm_kind	= MPID_INTERCOMM;
	tmp_comm->local_comm   = NULL;
	tmp_comm->is_low_group = is_low_group;

	/* No pg structure needed since vc has already been set up (connection has been established). */

	/* Point local vcr, vcrt at those of commself_ptr */
	tmp_comm->local_vcrt = commself_ptr->vcrt;
	MPID_VCRT_Add_ref(commself_ptr->vcrt);
	tmp_comm->local_vcr  = commself_ptr->vcr;

	/* No pg needed since connection has already been formed. 
	   FIXME - ensure that the comm_release code does not try to
	   free an unallocated pg */

	/* Set up VC reference table */
	mpi_errno = MPID_VCRT_Create(tmp_comm->remote_size, &tmp_comm->vcrt);
	if (mpi_errno != MPI_SUCCESS)
	{
		MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, L"**init_vcrt");
	}
	mpi_errno = MPID_VCRT_Get_ptr(tmp_comm->vcrt, &tmp_comm->vcr);
	if (mpi_errno != MPI_SUCCESS)
	{
		MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, L"**init_getptr");
	}
	
	MPID_VCR_Dup(vc_ptr, tmp_comm->vcr);

	*comm_pptr = tmp_comm;

fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_INITIALIZE_TMP_COMM);
	return mpi_errno;
fn_fail:
	goto fn_exit;
}

/* This routine initializes the new intercomm, setting up the
   VCRT and other common structures.  The is_low_group and context_id
   fields are NOT set because they differ in the use of this 
   routine in Comm_accept and Comm_connect.  The virtual connections
   are initialized from a collection of process groups. 

   Input parameters:
+  comm_ptr - communicator that gives the group for the "local" group on the
   new intercommnicator
.  remote_comm_size - size of remote group
.  remote_translation - array that specifies the process group and rank in 
   that group for each of the processes to include in the remote group of the
   new intercommunicator
-  remote_pg - array of remote process groups

   Input/Output Parameter:
.  intercomm - New intercommunicator.  The intercommunicator must already
   have been allocated; this routine initializes many of the fields

   Note:
   This routine performance a barrier over 'comm_ptr'.  Why?
*/
#undef FUNCNAME
#define FUNCNAME SetupNewIntercomm
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int SetupNewIntercomm( MPID_Comm *comm_ptr, int remote_comm_size, 
							  pg_translation remote_translation[],
							  MPIDI_PG_t **remote_pg, 
							  MPID_Comm *intercomm )
{
	int mpi_errno = MPI_SUCCESS, i;
	MPIDI_STATE_DECL(MPID_STATE_SETUPNEWINTERCOMM);

	MPIDI_FUNC_ENTER(MPID_STATE_SETUPNEWINTERCOMM);

	/* FIXME: How much of this could/should be common with the
	   upper level (src/mpi/comm/ *.c) code? For best robustness, 
	   this should use the same routine (not copy/paste code) as
	   in the upper level code. */
	intercomm->attributes   = NULL;
	intercomm->remote_size  = remote_comm_size;
	intercomm->local_size   = comm_ptr->local_size;
	intercomm->rank			= comm_ptr->rank;
	intercomm->local_group  = NULL;
	intercomm->remote_group = NULL;
	intercomm->comm_kind	= MPID_INTERCOMM;
	intercomm->local_comm   = NULL;
	intercomm->coll_fns		= NULL;

	/* Point local vcr, vcrt at those of incoming intracommunicator */
	intercomm->local_vcrt = comm_ptr->vcrt;
	MPID_VCRT_Add_ref(comm_ptr->vcrt);
	intercomm->local_vcr  = comm_ptr->vcr;

	/* Set up VC reference table */
	mpi_errno = MPID_VCRT_Create(intercomm->remote_size, &intercomm->vcrt);
	if (mpi_errno != MPI_SUCCESS)
	{
		MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, L"**init_vcrt");
	}
	mpi_errno = MPID_VCRT_Get_ptr(intercomm->vcrt, &intercomm->vcr);
	if (mpi_errno != MPI_SUCCESS)
	{
		MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, L"**init_getptr");
	}
	for (i=0; i < intercomm->remote_size; i++)
	{
		MPIDI_PG_Dup_vcr(remote_pg[remote_translation[i].pg_index], remote_translation[i].pg_rank, &intercomm->vcr[i]);
	}

	MPIU_DBG_MSG(CH3_CONNECT,VERBOSE,L"Barrier");
	mpi_errno = MPIR_Barrier(comm_ptr);
	MPIU_ERR_CHK_POP(mpi_errno);

 fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_SETUPNEWINTERCOMM);
	return mpi_errno;

 fn_fail:
	goto fn_exit;
}

/* Free new_vc. It was explicitly allocated in MPIDI_CH3_Connect_to_root. */
#undef FUNCNAME
#define FUNCNAME FreeNewVC
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int FreeNewVC( MPIDI_VC_t *new_vc )
{
	MPID_Progress_state progress_state;
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_FREENEWVC);

	MPIDI_FUNC_ENTER(MPID_STATE_FREENEWVC);
	
	if (new_vc->state != MPIDI_VC_STATE_INACTIVE)
	{
		/* If the new_vc isn't done, run the progress engine until
		   the state of the new vc is complete */
		MPID_Progress_start(&progress_state);
		while (new_vc->state != MPIDI_VC_STATE_INACTIVE)
		{
			mpi_errno = MPID_Progress_wait(&progress_state);
			/* --BEGIN ERROR HANDLING-- */
			if (mpi_errno != MPI_SUCCESS)
			{
				MPID_Progress_end(&progress_state);
				MPIU_ERR_POP(mpi_errno);
			}
			/* --END ERROR HANDLING-- */
		}
		MPID_Progress_end(&progress_state);
	}

	mpi_errno = MPIDI_CH3.Cleanup_after_connection( new_vc );

	MPIU_Free(new_vc);
 fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_FREENEWVC);
	return mpi_errno;
}

/* FIXME: What is an Accept queue and who uses it?  
   Is this part of the connect/accept support?  
   These routines appear to be called by channel progress routines; 
   perhaps this belongs in util/sock (note the use of a port_name_tag in the 
   dequeue code, though this could be any string).

   Are the locks required?  If this is only called within the progress
   engine, then the progress engine locks should be sufficient.  If a
   finer grain lock model is used, it needs to be very carefully 
   designed and documented.
*/

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_Acceptq_enqueue
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3I_Acceptq_enqueue(MPIDI_VC_t * vc, int port_name_tag)
{
	int mpi_errno=MPI_SUCCESS;
	MPIDI_CH3I_Acceptq_t *q_item;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_ACCEPTQ_ENQUEUE);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_ACCEPTQ_ENQUEUE);

	/* FIXME: Use CHKPMEM */
	q_item = (MPIDI_CH3I_Acceptq_t *)
		MPIU_Malloc(sizeof(MPIDI_CH3I_Acceptq_t)); 
	/* --BEGIN ERROR HANDLING-- */
	if (q_item == NULL)
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
		goto fn_exit;
	}
	/* --END ERROR HANDLING-- */

	q_item->vc = vc;
	q_item->port_name_tag = port_name_tag;

	MPIU_DBG_MSG_P(CH3_CONNECT,TYPICAL,L"vc=%p:Enqueuing accept connection",vc);
	q_item->next = acceptq_head;
	acceptq_head = q_item;
	
 fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_ACCEPTQ_ENQUEUE);
	return mpi_errno;
}


/* Attempt to dequeue a vc from the accept queue. If the queue is
   empty or the port_name_tag doesn't match, return a NULL vc. */
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_Acceptq_dequeue
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3I_Acceptq_dequeue(MPIDI_VC_t ** vc, int port_name_tag)
{
	int mpi_errno=MPI_SUCCESS;
	MPIDI_CH3I_Acceptq_t *q_item, *prev;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_ACCEPTQ_DEQUEUE);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_ACCEPTQ_DEQUEUE);

	*vc = NULL;
	q_item = acceptq_head;
	prev = q_item;
	while (q_item != NULL)
	{
		if (q_item->port_name_tag == port_name_tag)
		{
			*vc = q_item->vc;

			if ( q_item == acceptq_head )
				acceptq_head = q_item->next;
			else
				prev->next = q_item->next;

			MPIU_Free(q_item);
			break;;
		}
		else
		{
			prev = q_item;
			q_item = q_item->next;
		}
	}

	MPIU_DBG_MSG_FMT(CH3_CONNECT,TYPICAL, (MPIU_DBG_FDEST,L"vc=%p:Dequeuing accept connection with tag %d", *vc,port_name_tag));

	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_ACCEPTQ_DEQUEUE);
	return mpi_errno;
}

/* FIXME: 
 * Recvq_lock/unlock removed because it is not needed for the SINGLE_CS
 * approach and we might want a different, non-lock-based approach in 
 * a finer-grained thread-sync version.  For example, 
 * some routines can be implemented in a lock-free
 * fashion (because the user is required to guarantee non-conflicting 
 * accesses, such as doing a probe and a receive that matches in different 
 * threads).  
 *
 * There are a lot of routines here.  Do we really need them all?
 * 
 * The search criteria can be implemented in a single 64 bit compare on
 * systems with efficient 64-bit operations.  The rank and contextid can also
 * in many cases be combined into a single 32-bit word for the comparison
 * (in which case the message info should be stored in the queue in a 
 * naturally aligned, 64-bit word.
 * 
 */

static const wchar_t * GetDTypeString(MPI_Datatype d)
{
	static wchar_t default_str[64];
	int num_integers, num_addresses, num_datatypes, combiner = 0;
	wchar_t *str;
	MPIDI_STATE_DECL(MPID_STATE_GETDTYPESTRING);

	MPIDI_FUNC_ENTER(MPID_STATE_GETDTYPESTRING);

	if (d == MPI_DATATYPE_NULL)
	{
		MPIDI_FUNC_EXIT(MPID_STATE_GETDTYPESTRING);
		return L"MPI_DATATYPE_NULL";
	}

	if (d == 0)
	{
		MPIU_Wcsncpy(default_str, L"dtype=0x0", 64);
		MPIDI_FUNC_EXIT(MPID_STATE_GETDTYPESTRING);
		return default_str;
	}

	MPID_Type_get_envelope(d, &num_integers, &num_addresses, &num_datatypes, &combiner);
	if (combiner == MPI_COMBINER_NAMED)
	{
		str = MPIDU_Datatype_builtin_to_string(d);
		if (str == NULL)
		{
			MPIU_Snprintf(default_str, 64, _TRUNCATE, L"dtype=0x%08x", d);
			MPIDI_FUNC_EXIT(MPID_STATE_GETDTYPESTRING);
			return default_str;
		}
		MPIDI_FUNC_EXIT(MPID_STATE_GETDTYPESTRING);
		return str;
	}
	
	/* default is not thread safe */
	str = MPIDU_Datatype_combiner_to_string(combiner);
	if (str == NULL)
	{
		MPIU_Snprintf(default_str, 64, _TRUNCATE, L"dtype=USER<0x%08x>", d);
		MPIDI_FUNC_EXIT(MPID_STATE_GETDTYPESTRING);
		return default_str;
	}
	MPIU_Snprintf(default_str, 64, _TRUNCATE, L"dtype=USER<%s>", str);
	MPIDI_FUNC_EXIT(MPID_STATE_GETDTYPESTRING);
	return default_str;
}

static const wchar_t * GetContextString(int context_id)
{
	if (context_id & MPID_CONTEXT_INTRA_COLL)
		return L"(Collective)";
	if (context_id & MPID_CONTEXT_INTRA_FILE)
		return L"(File)";
	if (context_id & MPID_CONTEXT_INTRA_WIN)
		return L"(Win)";
	return L"";
}

int MPIU_Queues_to_string(wchar_t *str, int max_length)
{
	int mpi_errno = MPI_SUCCESS;
	MPID_Request *iter;
	size_t len;
	wchar_t *the_end;

	/* Make sure we return either an empty string or in the case of truncation that the string is NULL terminated */
	str[0] = L'\0';
	the_end = &str[max_length-1];

	if (recvq_posted_head != NULL || recvq_unexpected_head != NULL)
	{
		MPIU_Snprintf(str, max_length, _TRUNCATE, L"Rank %d queues:", MPIR_Process.comm_world->rank);
		len = wcslen(str);
		str = str + len;
		max_length = max_length - (int)len;
	}

	iter = recvq_posted_head;
	if (iter != NULL)
	{
		MPIU_Snprintf(str, max_length, _TRUNCATE, L"\n Posted receive queue:");
		len = wcslen(str);
		str = str + len;
		max_length = max_length - (int)len;
	}
	while (iter != NULL)
	{
		if (iter->dev.match.rank == MPI_ANY_SOURCE)
		{
			MPIU_Snprintf(str, max_length, _TRUNCATE, L"\n  rank=MPI_ANY_SOURCE, ");
		}
		else
		{
			MPIU_Snprintf(str, max_length, _TRUNCATE, L"\n  rank=%d, ", iter->dev.match.rank);
		}
		len = wcslen(str); str = str + len; max_length = max_length - (int)len;
		if (iter->dev.match.tag == MPI_ANY_TAG)
		{
			MPIU_Snprintf(str, max_length, _TRUNCATE, L"tag=MPI_ANY_TAG, ");
		}
		else
		{
			MPIU_Snprintf(str, max_length, _TRUNCATE, L"tag=%d, ", iter->dev.match.tag);
		}
		len = wcslen(str); str = str + len; max_length = max_length - (int)len;
		if (iter->dev.match.context_id == (MPIR_Process.comm_world->context_id + MPID_CONTEXT_INTRA_PT2PT))
		{
			MPIU_Snprintf(str, max_length, _TRUNCATE, L"context_id=MPI_COMM_WORLD");
		}
		else
		{
			MPIU_Snprintf(str, max_length, _TRUNCATE, L"context_id=%d%s", iter->dev.match.context_id, GetContextString(iter->dev.match.context_id));
		}
		len = wcslen(str); str = str + len; max_length = max_length - (int)len;
		MPIU_Snprintf(str, max_length, _TRUNCATE, L", count=%d, dtype=%s", iter->dev.user_count, GetDTypeString(iter->dev.datatype));
		len = wcslen(str); str = str + len; max_length = max_length - (int)len;
		iter = iter->dev.next;
		if (iter == recvq_posted_head)
			break; /* Extra check for loops */
	}

	iter = recvq_unexpected_head;
	if (iter != NULL)
	{
		MPIU_Snprintf(str, max_length, _TRUNCATE, L"\n Received but unmatched queue:");
		len = wcslen(str);
		str = str + len;
		max_length = max_length - (int)len;
	}
	while (iter != NULL)
	{
		MPIU_Snprintf(str, max_length, _TRUNCATE, L"\n  rank=%d, tag=%d, ", iter->dev.match.rank, iter->dev.match.tag);
		len = wcslen(str);
		str = str + len;
		max_length = max_length - (int)len;
		if (iter->dev.match.context_id == (MPIR_Process.comm_world->context_id + MPID_CONTEXT_INTRA_PT2PT))
		{
			MPIU_Snprintf(str, max_length, _TRUNCATE, L"context_id=MPI_COMM_WORLD");
		}
		else
		{
			MPIU_Snprintf(str, max_length, _TRUNCATE, L"context_id=%d%s", iter->dev.match.context_id, GetContextString(iter->dev.match.context_id));
		}
		len = wcslen(str); str = str + len; max_length = max_length - (int)len;
		MPIU_Snprintf(str, max_length, _TRUNCATE, L", length=%d", iter->dev.recv_data_sz);
		len = wcslen(str); str = str + len; max_length = max_length - (int)len;
		iter = iter->dev.next;
		if (iter == recvq_unexpected_head)
			break; /* Extra check for loops */
	}

	/* In case of truncation, make sure the string is NULL terminated. */
	*the_end = L'\0';

	return mpi_errno;
}

/*
 * MPIDI_CH3U_Recvq_FU()
 *
 * Search for a matching request in the unexpected receive queue.  Return 
 * true if one is found, false otherwise.  If the status arguement is
 * not MPI_STATUS_IGNORE, return information about the request in that
 * parameter.  This routine is used by mpid_probe and mpid_iprobe.
 */
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3U_Recvq_FU
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3U_Recvq_FU(int source, int tag, int context_id, MPI_Status *s)
{
	MPID_Request * rreq;
	int found = 0;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3U_RECVQ_FU);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3U_RECVQ_FU);

	if (tag != MPI_ANY_TAG && source != MPI_ANY_SOURCE)
	{
		rreq = recvq_unexpected_head;
		while(rreq != NULL)
		{
			if (rreq->dev.match.context_id == context_id && 
				rreq->dev.match.rank == source && rreq->dev.match.tag == tag)
			{
				break;
			}
			
			rreq = rreq->dev.next;
		}
	}
	else
	{
		MPIDI_Message_match match;
		MPIDI_Message_match mask;

		match.context_id = (int16_t)context_id;
		mask.context_id = ~0;
		if (tag == MPI_ANY_TAG)
		{
			match.tag = 0;
			mask.tag = 0;
		}
		else
		{
			match.tag = tag;
			mask.tag = ~0;
		}
		if (source == MPI_ANY_SOURCE)
		{
			match.rank = 0;
			mask.rank = 0;
		}
		else
		{
			match.rank = (int16_t)source;
			mask.rank = ~0;
		}
		
		rreq = recvq_unexpected_head;
		while (rreq != NULL)
		{
			if (rreq->dev.match.context_id == match.context_id && 
				(rreq->dev.match.rank & mask.rank) == match.rank &&
				(rreq->dev.match.tag & mask.tag) == match.tag)
			{
				break;
			}
			
			rreq = rreq->dev.next;
		}
	}

	/* Save the information about the request before releasing the queue */
	if (rreq)
	{
		if (s != MPI_STATUS_IGNORE)
		{
			/* Avoid setting L"extra" fields like MPI_ERROR */
			s->MPI_SOURCE = rreq->status.MPI_SOURCE;
			s->MPI_TAG	= rreq->status.MPI_TAG;
			s->count	  = rreq->status.count;
			s->cancelled  = rreq->status.cancelled;
		}
		found = 1;
	}

	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3U_RECVQ_FU);
	return found;
}

/*
 * MPIDI_CH3U_Recvq_FDU()
 *
 * Find a request in the unexpected queue and dequeue it; otherwise return NULL.
 */
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3U_Recvq_FDU
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
MPID_Request * MPIDI_CH3U_Recvq_FDU(MPI_Request sreq_id, MPIDI_Message_match * match)
{
	MPID_Request * rreq;
	MPID_Request * prev_rreq;
	MPID_Request * cur_rreq;
	MPID_Request * matching_prev_rreq;
	MPID_Request * matching_cur_rreq;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3U_RECVQ_FDU);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3U_RECVQ_FDU);

	matching_prev_rreq = NULL;
	matching_cur_rreq = NULL;
	prev_rreq = NULL;
	
	cur_rreq = recvq_unexpected_head;
	while(cur_rreq != NULL)
	{
		if (cur_rreq->dev.sender_req_id == sreq_id && 
			cur_rreq->dev.match.context_id == match->context_id &&
			cur_rreq->dev.match.rank == match->rank && 
			cur_rreq->dev.match.tag == match->tag)
		{
			matching_prev_rreq = prev_rreq;
			matching_cur_rreq = cur_rreq;
		}
		
		prev_rreq = cur_rreq;
		cur_rreq = cur_rreq->dev.next;
	}
	
	if (matching_cur_rreq != NULL)
	{
		if (matching_prev_rreq != NULL)
		{
			matching_prev_rreq->dev.next = matching_cur_rreq->dev.next;
		}
		else
		{
			recvq_unexpected_head = matching_cur_rreq->dev.next;
		}
		
		if (matching_cur_rreq->dev.next == NULL)
		{
			recvq_unexpected_tail = matching_prev_rreq;
		}

		rreq = matching_cur_rreq;
	}
	else
	{
		rreq = NULL;
	}

	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3U_RECVQ_FDU);
	return rreq;
}


/*
 * MPIDI_CH3U_Recvq_FDU_or_AEP()
 *
 * Atomically find a request in the unexpected queue and dequeue it, or 
 * allocate a new request and enqueue it in the posted queue
 */
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3U_Recvq_FDU_or_AEP
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
MPID_Request * MPIDI_CH3U_Recvq_FDU_or_AEP(int source, int tag, int context_id, int * foundp)
{
	int found;
	MPID_Request * rreq, *prev_rreq;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3U_RECVQ_FDU_OR_AEP);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3U_RECVQ_FDU_OR_AEP);

	/* Optimize this loop for an empty unexpected receive queue */
	rreq = recvq_unexpected_head;
	if (rreq)
	{
		prev_rreq = NULL;
		if (tag != MPI_ANY_TAG && source != MPI_ANY_SOURCE)
		{
			do
			{
				if (rreq->dev.match.context_id == context_id && 
					rreq->dev.match.rank == source && 
					rreq->dev.match.tag == tag)
				{
					if (prev_rreq != NULL)
					{
						prev_rreq->dev.next = rreq->dev.next;
					}
					else
					{
						recvq_unexpected_head = rreq->dev.next;
					}
					if (rreq->dev.next == NULL)
					{
						recvq_unexpected_tail = prev_rreq;
					}
					found = TRUE;
					goto lock_exit;
				} 
				
				prev_rreq = rreq;
				rreq	  = rreq->dev.next;
			} while (rreq);
		}
		else
		{
			MPIDI_Message_match match;
			MPIDI_Message_match mask;
			
			match.context_id = (int16_t)context_id;
			mask.context_id = ~0;
			if (tag == MPI_ANY_TAG)
			{
				match.tag = 0;
				mask.tag = 0;
			}
			else
			{
				match.tag = tag;
				mask.tag = ~0;
			}
			if (source == MPI_ANY_SOURCE)
			{
				match.rank = 0;
				mask.rank = 0;
			}
			else
			{
				match.rank = (int16_t)source;
				mask.rank = ~0;
			}
			
			do
			{
				if (rreq->dev.match.context_id == match.context_id && 
					(rreq->dev.match.rank & mask.rank) == match.rank &&
					(rreq->dev.match.tag & mask.tag) == match.tag) {
					if (prev_rreq != NULL)
					{
						prev_rreq->dev.next = rreq->dev.next;
					}
					else
					{
						recvq_unexpected_head = rreq->dev.next;
					}
					if (rreq->dev.next == NULL)
					{
						recvq_unexpected_tail = prev_rreq;
					}
					found = TRUE;
					goto lock_exit;
				}
				
				prev_rreq = rreq;
				rreq = rreq->dev.next;
			} while (rreq);
		}
	}
	
	/* A matching request was not found in the unexpected queue, so we 
	   need to allocate a new request and add it to the posted queue */
	rreq = MPID_Request_create();
	if (rreq != NULL)
	{
		MPIU_Object_set_ref(rreq, 2);
		rreq->kind = MPID_REQUEST_RECV;
		rreq->dev.match.tag = tag;
		rreq->dev.match.rank = (int16_t)source;
		rreq->dev.match.context_id = (int16_t)context_id;
		rreq->dev.next = NULL;
		
		if (recvq_posted_tail != NULL)
		{
			recvq_posted_tail->dev.next = rreq;
		}
		else
		{
			recvq_posted_head = rreq;
		}
		recvq_posted_tail = rreq;
	}
	
	found = FALSE;

  lock_exit:

	*foundp = found;
	
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3U_RECVQ_FDU_OR_AEP);
	return rreq;
}


/*
 * MPIDI_CH3U_Recvq_DP()
 *
 * Given an existing request, dequeue that request from the posted queue, or return NULL if the request was not in the posted
 * queued
 */
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3U_Recvq_DP
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3U_Recvq_DP(MPID_Request * rreq)
{
	int found;
	MPID_Request * cur_rreq;
	MPID_Request * prev_rreq;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3U_RECVQ_DP);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3U_RECVQ_DP);
	
	found = FALSE;
	prev_rreq = NULL;
	
	cur_rreq = recvq_posted_head;
	while (cur_rreq != NULL)
	{
		if (cur_rreq == rreq)
		{
			if (prev_rreq != NULL)
			{
				prev_rreq->dev.next = cur_rreq->dev.next;
			}
			else
			{
				recvq_posted_head = cur_rreq->dev.next;
			}
			if (cur_rreq->dev.next == NULL)
			{
				recvq_posted_tail = prev_rreq;
			}
			found = TRUE;
			break;
		}
		
		prev_rreq = cur_rreq;
		cur_rreq = cur_rreq->dev.next;
	}

	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3U_RECVQ_DP);
	return found;
}

/*
 * MPIDI_CH3U_Recvq_FDP_or_AEU()
 *
 * Locate a request in the posted queue and dequeue it, or allocate a new 
 * request and enqueue it in the unexpected queue
 */
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3U_Recvq_FDP_or_AEU
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
MPID_Request * MPIDI_CH3U_Recvq_FDP_or_AEU(MPIDI_Message_match * match, int * foundp)
{
	int found;
	MPID_Request * rreq;
	MPID_Request * prev_rreq;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3U_RECVQ_FDP_OR_AEU);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3U_RECVQ_FDP_OR_AEU);
	
	prev_rreq = NULL;

	rreq = recvq_posted_head;
	while (rreq != NULL)
	{
		if ((rreq->dev.match.context_id == match->context_id) &&
			(rreq->dev.match.rank == match->rank || 
			 rreq->dev.match.rank == MPI_ANY_SOURCE) &&
			(rreq->dev.match.tag == match->tag || 
			 rreq->dev.match.tag == MPI_ANY_TAG))
		{
			if (prev_rreq != NULL)
			{
				prev_rreq->dev.next = rreq->dev.next;
			}
			else
			{
				recvq_posted_head = rreq->dev.next;
			}
			if (rreq->dev.next == NULL)
			{
				recvq_posted_tail = prev_rreq;
			}
			found = TRUE;
			goto lock_exit;
		}
		
		prev_rreq = rreq;
		rreq = rreq->dev.next;
	}

	/* A matching request was not found in the posted queue, so we 
	   need to allocate a new request and add it to the unexpected queue */
	rreq = MPID_Request_create();
	if (rreq != NULL)
	{
		MPIU_Object_set_ref(rreq, 2);
		rreq->kind = MPID_REQUEST_RECV;
		rreq->dev.match = *match;
		rreq->dev.next = NULL;
		
		if (recvq_unexpected_tail != NULL)
		{
			recvq_unexpected_tail->dev.next = rreq;
		}
		else
		{
			recvq_unexpected_head = rreq;
		}
		recvq_unexpected_tail = rreq;
	}
	
	found = FALSE;

  lock_exit:

	*foundp = found;

	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3U_RECVQ_FDP_OR_AEU);
	return rreq;
}

/* This file contains two types of routines associated with requests: 
 * Routines to allocate and free requests
 * Routines to manage iovs on requests 
 *
 * Note that there are a number of macros that also manage requests defined
 * in mpidimpl.h ; these are intended to optimize request creation for
 * specific types of requests.  See the comments in mpidimpl.h for more
 * details.
 */

#undef FUNCNAME
#define FUNCNAME MPID_Request_create
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
MPID_Request * MPID_Request_create()
{
	MPID_Request * req;
	MPIDI_STATE_DECL(MPID_STATE_MPID_REQUEST_CREATE);

	MPIDI_FUNC_ENTER(MPID_STATE_MPID_REQUEST_CREATE);
	
	req = MPIU_Handle_obj_alloc(&MPID_Request_mem);
	if (req != NULL)
	{
		MPIU_DBG_MSG_P(CH3_OTHER,VERBOSE, L"allocated request, handle=0x%08x", req->handle);
#	   if defined(HAVE_ERROR_CHECKING)
		{
			int mpi_errno;
			
			if (HANDLE_GET_MPI_KIND(req->handle) != MPID_REQUEST)
			{
				mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_INTERN, L"**invalid_handle", L"**invalid_handle %d %p", req->handle, req);
				MPID_Abort(MPIR_Process.comm_world, mpi_errno, -1, NULL);
			}
		}
#	   endif

		/* FIXME: This makes request creation expensive.  We need to trim
		   this to the basics, with additional setup for special-purpose 
		   requests (think base class and inheritance).  For example, do we 
		   *really* want to set the kind to UNDEFINED? And should the RMA 
		   values be set only for RMA requests? */
		MPIU_Object_set_ref(req, 1);
		req->kind					= MPID_REQUEST_UNDEFINED;
		req->cc						= 1;
		req->cc_ptr					= &req->cc;
		/* FIXME: status fields meaningful only for receive, and even then should not need to be set. */
		req->status.MPI_SOURCE		= MPI_UNDEFINED;
		req->status.MPI_TAG			= MPI_UNDEFINED;
		req->status.MPI_ERROR		= MPI_SUCCESS;
		req->status.count			= 0;
		req->status.cancelled		= FALSE;
		req->comm					= NULL;
		req->dev.datatype_ptr		= NULL;
		/* Masks and flags for channel device state in an MPID_Request */
		req->dev.state				= 0;
		req->dev.cancel_pending		= FALSE;
		/* FIXME: RMA ops shouldn't need to be set except when creating a request for RMA operations */
		req->dev.target_win_handle	= MPI_WIN_NULL;
		req->dev.source_win_handle	= MPI_WIN_NULL;
		req->dev.single_op_opt		= 0;
		req->dev.lock_queue_entry	= NULL;
		req->dev.dtype_info			= NULL;
		req->dev.dataloop			= NULL;
		req->dev.serialized_datatype_buffer = NULL;
		req->dev.rdma_iov_count		= 0;
		req->dev.rdma_iov_offset	= 0;
	}
	else
	{
		/* FIXME: This fails to fail if debugging is turned off */
		MPIU_DBG_MSG(CH3_OTHER,TYPICAL,L"unable to allocate a request");
	}
	
	MPIDI_FUNC_EXIT(MPID_STATE_MPID_REQUEST_CREATE);
	return req;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3_Request_destroy
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
void MPIDI_CH3_Request_destroy(MPID_Request * req)
{
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3_REQUEST_DESTROY);
	
	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3_REQUEST_DESTROY);
	MPIU_DBG_MSG_P(CH3_OTHER,VERBOSE, L"freeing request, handle=0x%08x", req->handle);
	
#   if defined(HAVE_ERROR_CHECKING)
	{
		if (HANDLE_GET_MPI_KIND(req->handle) != MPID_REQUEST)
		{
			int mpi_errno;

			mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**invalid_handle", L"**invalid_handle %d %p", req->handle, req);
			MPID_Abort(MPIR_Process.comm_world, mpi_errno, -1, NULL);
		}
		if (req->ref_count != 0)
		{
			int mpi_errno;
			
			mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**invalid_refcount", L"**invalid_refcount %d %p %d", req->handle, req, req->ref_count);
			MPID_Abort(MPIR_Process.comm_world, mpi_errno, -1, NULL);
		}
	}
#   endif
	
	if (req->comm != NULL)
	{
		MPIR_Comm_release(req->comm, 0);
	}

	if (req->dev.datatype_ptr != NULL)
	{
		MPID_Datatype_release(req->dev.datatype_ptr);
	}

	if (MPIDI_Request_get_srbuf_flag(req))
	{
		MPIDI_CH3U_SRBuf_free(req);
	}

	MPIU_Handle_obj_free(&MPID_Request_mem, req);
	
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3_REQUEST_DESTROY);
}


/* ------------------------------------------------------------------------- */
/* Here are the routines to manipulate the iovs in the requests			  */
/* ------------------------------------------------------------------------- */



/*
 * MPIDI_CH3U_Request_load_send_iov()
 *
 * Fill the provided IOV with the next (or remaining) portion of data described by the segment contained in the request structure.
 * If the density of IOV is not sufficient, pack the data into a send/receive buffer and point the IOV at the buffer.
 */
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3U_Request_load_send_iov
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3U_Request_load_send_iov(MPID_Request * const sreq, MPID_IOV * const iov, int * const iov_n)
{
	MPIDI_msg_sz_t last;
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3U_REQUEST_LOAD_SEND_IOV);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3U_REQUEST_LOAD_SEND_IOV);
	last = sreq->dev.segment_size;
	MPIU_DBG_MSG_FMT(CH3_OTHER,VERBOSE,(MPIU_DBG_FDEST, L"pre-pv: first=" MPIDI_MSG_SZ_FMT L", last=" MPIDI_MSG_SZ_FMT L", iov_n=%d", sreq->dev.segment_first, last, *iov_n));
	MPIU_Assert(sreq->dev.segment_first < last);
	MPIU_Assert(last > 0);
	MPIU_Assert(*iov_n > 0 && *iov_n <= MPID_IOV_LIMIT);
	MPID_Segment_pack_vector(&sreq->dev.segment, sreq->dev.segment_first, &last, iov, iov_n);
	MPIU_DBG_MSG_FMT(CH3_OTHER,VERBOSE,(MPIU_DBG_FDEST, L"post-pv: first=" MPIDI_MSG_SZ_FMT L", last=" MPIDI_MSG_SZ_FMT L", iov_n=%d", sreq->dev.segment_first, last, *iov_n));
	MPIU_Assert(*iov_n > 0 && *iov_n <= MPID_IOV_LIMIT);
	
	if (last == sreq->dev.segment_size)
	{
		MPIU_DBG_MSG(CH3_OTHER,VERBOSE,L"remaining data loaded into IOV");
		sreq->dev.ca = MPIDI_CH3_CA_COMPLETE;
	}
	else if ((last - sreq->dev.segment_first) / *iov_n >= MPIDI_IOV_DENSITY_MIN)
	{
		MPIU_DBG_MSG(CH3_OTHER,VERBOSE,L"more data loaded into IOV");
		sreq->dev.segment_first = last;
		sreq->dev.ca = MPIDI_CH3_CA_RELOAD_IOV;
	}
	else
	{
		MPIDI_msg_sz_t data_sz;
		
		MPIU_DBG_MSG(CH3_OTHER,VERBOSE,L"low density.  using SRBuf.");
			
		data_sz = sreq->dev.segment_size - sreq->dev.segment_first;
		if (!MPIDI_Request_get_srbuf_flag(sreq))
		{
			MPIDI_CH3U_SRBuf_alloc(sreq, data_sz);
			/* --BEGIN ERROR HANDLING-- */
			if (sreq->dev.tmpbuf_sz == 0)
			{
				MPIU_DBG_MSG(CH3_OTHER,TYPICAL,L"SRBuf allocation failure");
				mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
				sreq->status.MPI_ERROR = mpi_errno;
				goto fn_exit;
			}
			/* --END ERROR HANDLING-- */
		}
					
		last = (data_sz <= sreq->dev.tmpbuf_sz) ? sreq->dev.segment_size : sreq->dev.segment_first + sreq->dev.tmpbuf_sz;
		MPIU_DBG_MSG_FMT(CH3_OTHER,VERBOSE,(MPIU_DBG_FDEST, L"pre-pack: first=" MPIDI_MSG_SZ_FMT L", last=" MPIDI_MSG_SZ_FMT, sreq->dev.segment_first, last));
		MPID_Segment_pack(&sreq->dev.segment, sreq->dev.segment_first, &last, sreq->dev.tmpbuf);
		MPIU_DBG_MSG_FMT(CH3_OTHER,VERBOSE,(MPIU_DBG_FDEST, L"post-pack: first=" MPIDI_MSG_SZ_FMT L", last=" MPIDI_MSG_SZ_FMT, sreq->dev.segment_first, last));
		iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)sreq->dev.tmpbuf;
		iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)(last - sreq->dev.segment_first);
		*iov_n = 1;
		if (last == sreq->dev.segment_size)
		{
			MPIU_DBG_MSG(CH3_OTHER,VERBOSE,L"remaining data packed into SRBuf");
			sreq->dev.ca = MPIDI_CH3_CA_COMPLETE;
		}
		else 
		{
			MPIU_DBG_MSG(CH3_OTHER,VERBOSE,L"more data packed into SRBuf");
			sreq->dev.segment_first = last;
			sreq->dev.ca = MPIDI_CH3_CA_RELOAD_IOV;
		}
	}
	
  fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3U_REQUEST_LOAD_SEND_IOV);
	return mpi_errno;
}

/*
 * MPIDI_CH3U_Request_load_recv_iov()
 *
 * Fill the request's IOV with the next (or remaining) portion of data described by the segment (also contained in the request
 * structure).  If the density of IOV is not sufficient, allocate a send/receive buffer and point the IOV at the buffer.
 */
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3U_Request_load_recv_iov
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3U_Request_load_recv_iov(MPID_Request * const rreq)
{
	MPIDI_msg_sz_t last;
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3U_REQUEST_LOAD_RECV_IOV);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3U_REQUEST_LOAD_RECV_IOV);
	if (rreq->dev.segment_first < rreq->dev.segment_size)
	{
		/* still reading data that needs to go into the user buffer */
		
		if (MPIDI_Request_get_srbuf_flag(rreq))
		{
			MPIDI_msg_sz_t data_sz;
			MPIDI_msg_sz_t tmpbuf_sz;

			/* Once a SRBuf is in use, we continue to use it since a small amount of data may already be present at the beginning
			   of the buffer.  This data is left over from the previous unpack, most like a result of alignment issues.  NOTE: we
			   could force the use of the SRBuf only when (rreq->dev.tmpbuf_off > 0)... */
			
			data_sz = rreq->dev.segment_size - rreq->dev.segment_first - rreq->dev.tmpbuf_off;
			MPIU_Assert(data_sz > 0);
			tmpbuf_sz = rreq->dev.tmpbuf_sz - rreq->dev.tmpbuf_off;
			if (data_sz > tmpbuf_sz)
			{
				data_sz = tmpbuf_sz;
			}
			rreq->dev.iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)((char *) rreq->dev.tmpbuf + rreq->dev.tmpbuf_off);
			rreq->dev.iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST) data_sz;
			rreq->dev.iov_count = 1;
			MPIU_Assert(rreq->dev.segment_first + data_sz + rreq->dev.tmpbuf_off <= rreq->dev.recv_data_sz);
			if (rreq->dev.segment_first + data_sz + rreq->dev.tmpbuf_off == rreq->dev.recv_data_sz)
			{
				MPIU_DBG_MSG(CH3_OTHER,VERBOSE, L"updating rreq to read the remaining data into the SRBuf");
				rreq->dev.ca = MPIDI_CH3_CA_UNPACK_SRBUF_AND_COMPLETE;
			}
			else
			{
				MPIU_DBG_MSG(CH3_OTHER,VERBOSE, L"updating rreq to read more data into the SRBuf");
				rreq->dev.ca = MPIDI_CH3_CA_UNPACK_SRBUF_AND_RELOAD_IOV;
			}
			goto fn_exit;
		}
		
		last = rreq->dev.segment_size;
		rreq->dev.iov_count = MPID_IOV_LIMIT;
		MPIU_DBG_MSG_FMT(CH3_OTHER,VERBOSE,(MPIU_DBG_FDEST, L"pre-upv: first=" MPIDI_MSG_SZ_FMT L", last=" MPIDI_MSG_SZ_FMT L", iov_n=%d", rreq->dev.segment_first, last, rreq->dev.iov_count));
		MPIU_Assert(rreq->dev.segment_first < last);
		MPIU_Assert(last > 0);
		MPID_Segment_unpack_vector(&rreq->dev.segment, rreq->dev.segment_first, &last, rreq->dev.iov, &rreq->dev.iov_count);
		MPIU_DBG_MSG_FMT(CH3_OTHER,VERBOSE,(MPIU_DBG_FDEST, L"post-upv: first=" MPIDI_MSG_SZ_FMT L", last=" MPIDI_MSG_SZ_FMT L", iov_n=%d", rreq->dev.segment_first, last, rreq->dev.iov_count));
		MPIU_Assert(rreq->dev.iov_count > 0 && rreq->dev.iov_count <= MPID_IOV_LIMIT);

		/* --BEGIN ERROR HANDLING-- */
		if (rreq->dev.iov_count == 0)
		{
			/* If the data can't be unpacked, the we have a mis-match between the datatype and the amount of data received.  Adjust
			   the segment info so that the remaining data is received and thrown away. */
			rreq->status.MPI_ERROR = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_TYPE, L"**dtypemismatch", 0);
			rreq->status.count = (int)rreq->dev.segment_first;
			rreq->dev.segment_size = rreq->dev.segment_first;
			mpi_errno = MPIDI_CH3U_Request_load_recv_iov(rreq);
			goto fn_exit;
		}
		/* --END ERROR HANDLING-- */

		if (last == rreq->dev.recv_data_sz)
		{
			MPIU_DBG_MSG(CH3_OTHER,VERBOSE, L"updating rreq to read the remaining data directly into the user buffer");
			rreq->dev.ca = MPIDI_CH3_CA_COMPLETE;
		}
		else if (last == rreq->dev.segment_size || (last - rreq->dev.segment_first) / rreq->dev.iov_count >= MPIDI_IOV_DENSITY_MIN)
		{
			MPIU_DBG_MSG(CH3_OTHER,VERBOSE, L"updating rreq to read more data directly into the user buffer");
			rreq->dev.segment_first = last;
			rreq->dev.ca = MPIDI_CH3_CA_RELOAD_IOV;
		}
		else
		{
			/* Too little data would have been received using an IOV.  We will start receiving data into a SRBuf and unpacking it later. */
			MPIU_Assert(MPIDI_Request_get_srbuf_flag(rreq) == FALSE);
			
			MPIDI_CH3U_SRBuf_alloc(rreq, rreq->dev.segment_size - rreq->dev.segment_first);
			rreq->dev.tmpbuf_off = 0;
			/* --BEGIN ERROR HANDLING-- */
			if (rreq->dev.tmpbuf_sz == 0)
			{
				/* FIXME - we should drain the data off the pipe here, but we don't have a buffer to drain it into.  should this be a fatal error? */
				MPIU_DBG_MSG(CH3_OTHER,VERBOSE,L"SRBuf allocation failure");
				mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
				rreq->status.MPI_ERROR = mpi_errno;
				goto fn_exit;
			}
			/* --END ERROR HANDLING-- */

			/* fill in the IOV using a recursive call */
			mpi_errno = MPIDI_CH3U_Request_load_recv_iov(rreq);
		}
	}
	else
	{
		/* receive and toss any extra data that does not fit in the user's buffer */
		MPIDI_msg_sz_t data_sz;

		data_sz = rreq->dev.recv_data_sz - rreq->dev.segment_first;
		if (!MPIDI_Request_get_srbuf_flag(rreq))
		{
			MPIDI_CH3U_SRBuf_alloc(rreq, data_sz);
			/* --BEGIN ERROR HANDLING-- */
			if (rreq->dev.tmpbuf_sz == 0)
			{
				MPIU_DBG_MSG(CH3_OTHER,TYPICAL,L"SRBuf allocation failure");
				mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
				rreq->status.MPI_ERROR = mpi_errno;
				goto fn_exit;
			}
			/* --END ERROR HANDLING-- */
		}

		if (data_sz <= rreq->dev.tmpbuf_sz)
		{
			MPIU_DBG_MSG(CH3_OTHER,VERBOSE, L"updating rreq to read overflow data into the SRBuf and complete");
			rreq->dev.iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST) data_sz;
			rreq->dev.ca = MPIDI_CH3_CA_COMPLETE;
		}
		else
		{
			MPIU_DBG_MSG(CH3_OTHER,VERBOSE, L"updating rreq to read overflow data into the SRBuf and reload IOV");
			rreq->dev.iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST) rreq->dev.tmpbuf_sz;
			rreq->dev.segment_first += rreq->dev.tmpbuf_sz;
			rreq->dev.ca = MPIDI_CH3_CA_RELOAD_IOV;
		}

		rreq->dev.iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)rreq->dev.tmpbuf;
		rreq->dev.iov_count = 1;
	}
	
  fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3U_REQUEST_LOAD_RECV_IOV);
	return mpi_errno;
}

/*
 * MPIDI_CH3U_Request_unpack_srbuf
 *
 * Unpack data from a send/receive buffer into the user buffer.
 */
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3U_Request_unpack_srbuf
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3U_Request_unpack_srbuf(MPID_Request * rreq)
{
	MPIDI_msg_sz_t last;
	int tmpbuf_last;
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3U_REQUEST_UNPACK_SRBUF);
	
	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3U_REQUEST_UNPACK_SRBUF);

	tmpbuf_last = (int)(rreq->dev.segment_first + rreq->dev.tmpbuf_sz);
	if (rreq->dev.segment_size < tmpbuf_last)
	{
		tmpbuf_last = (int)rreq->dev.segment_size;
	}
	last = tmpbuf_last;
	MPID_Segment_unpack(&rreq->dev.segment, rreq->dev.segment_first, &last, rreq->dev.tmpbuf);
	if (last == 0 || last == rreq->dev.segment_first)
	{
		/* --BEGIN ERROR HANDLING-- */
		/* If no data can be unpacked, then we have a datatype processing problem.  Adjust the segment info so that the remaining
		   data is received and thrown away. */
		rreq->status.count = (int)rreq->dev.segment_first;
		rreq->dev.segment_size = rreq->dev.segment_first;
		rreq->dev.segment_first += tmpbuf_last;
		rreq->status.MPI_ERROR = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_TYPE, L"**dtypemismatch", 0);
		/* --END ERROR HANDLING-- */
	}
	else if (tmpbuf_last == rreq->dev.segment_size)
	{
		/* --BEGIN ERROR HANDLING-- */
		if (last != tmpbuf_last)
		{
			/* received data was not entirely consumed by unpack() because too few bytes remained to fill the next basic datatype.
			   Note: the segment_first field is set to segment_last so that if this is a truncated message, extra data will be read
			   off the pipe. */
			rreq->status.count = (int)last;
			rreq->dev.segment_size = last;
			rreq->dev.segment_first = tmpbuf_last;
			rreq->status.MPI_ERROR = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_TYPE, L"**dtypemismatch", 0);
		}
		/* --END ERROR HANDLING-- */
	}
	else
	{
		rreq->dev.tmpbuf_off = (int)(tmpbuf_last - last);
		if (rreq->dev.tmpbuf_off > 0)
		{
			/* move any remaining data to the beginning of the buffer.  Note: memmove() is used since the data regions could overlap. */
			memmove(rreq->dev.tmpbuf, (char *) rreq->dev.tmpbuf + (last - rreq->dev.segment_first), rreq->dev.tmpbuf_off);
		}
		rreq->dev.segment_first = last;
	}

	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3U_REQUEST_UNPACK_SRBUF);
	return mpi_errno;
}

/*
 * MPIDI_CH3U_Request_unpack_uebuf
 *
 * Copy/unpack data from an "unexpected eager buffer" into the user buffer.
 */
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3U_Request_unpack_uebuf
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3U_Request_unpack_uebuf(MPID_Request * rreq)
{
	int dt_contig;
	MPI_Aint dt_true_lb;
	MPIDI_msg_sz_t userbuf_sz;
	MPID_Datatype * dt_ptr;
	MPIDI_msg_sz_t unpack_sz;
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3U_REQUEST_UNPACK_UEBUF);
	MPIDI_STATE_DECL(MPID_STATE_MEMCPY);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3U_REQUEST_UNPACK_UEBUF);

	MPIDI_Datatype_get_info(rreq->dev.user_count, rreq->dev.datatype, dt_contig, userbuf_sz, dt_ptr, dt_true_lb);
	
	if (rreq->dev.recv_data_sz <= userbuf_sz)
	{
		unpack_sz = rreq->dev.recv_data_sz;
	}
	else
	{
		/* --BEGIN ERROR HANDLING-- */
		MPIU_DBG_MSG_FMT(CH3_OTHER,VERBOSE,(MPIU_DBG_FDEST, L"receive buffer overflow; message truncated, msg_sz=" MPIDI_MSG_SZ_FMT L", buf_sz=" MPIDI_MSG_SZ_FMT, rreq->dev.recv_data_sz, userbuf_sz));
		unpack_sz = userbuf_sz;
		rreq->status.count = (int)userbuf_sz;
		rreq->status.MPI_ERROR = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_TRUNCATE, L"**truncate", L"**truncate %d %d", rreq->dev.recv_data_sz, userbuf_sz);
		/* --END ERROR HANDLING-- */
	}

	if (unpack_sz > 0)
	{
		if (dt_contig)
		{
			/* TODO - check that amount of data is consistent with datatype.  In other words, if we were to use Segment_unpack()
			   would last = unpack?  If not we should return an error (unless configured with --enable-fast) */
			MPIDI_FUNC_ENTER(MPID_STATE_MEMCPY);
			memcpy((char *)rreq->dev.user_buf + dt_true_lb, rreq->dev.tmpbuf, unpack_sz);
			MPIDI_FUNC_EXIT(MPID_STATE_MEMCPY);
		}
		else
		{
			MPID_Segment seg;
			MPIDI_msg_sz_t last;

			MPID_Segment_init(rreq->dev.user_buf, rreq->dev.user_count, rreq->dev.datatype, &seg, 0);
			last = unpack_sz;
			MPID_Segment_unpack(&seg, 0, &last, rreq->dev.tmpbuf);
			if (last != unpack_sz)
			{
				/* --BEGIN ERROR HANDLING-- */
				/* received data was not entirely consumed by unpack() because too few bytes remained to fill the next basic datatype */
				rreq->status.count = (int)last;
				rreq->status.MPI_ERROR = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_TYPE, L"**dtypemismatch", 0);
				/* --END ERROR HANDLING-- */
			}
		}
	}

	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3U_REQUEST_UNPACK_UEBUF);
	return mpi_errno;
}

/* 
 * Export the function to set a request as completed for use by
 * the generalized request functions in mpich2/src/pt2pt/greq_complete.c
 */
void MPID_Request_set_completed( MPID_Request *req )
{
	MPID_REQUEST_SET_COMPLETED(req);
}

#undef FUNCNAME
#define FUNCNAME MPIDI_Win_create
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_Win_create(void *base, MPI_Aint size, int disp_unit, MPID_Info *info, MPID_Comm *comm_ptr, MPID_Win **win_ptr, MPIDI_RMAFns *RMAFns)
{
	int mpi_errno=MPI_SUCCESS, i, comm_size, rank;
	Win_create_pkt_t *tmp_buf;
	MPIU_CHKLMEM_DECL(1);
	MPIU_CHKPMEM_DECL(5);
	MPIU_THREADPRIV_DECL;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_WIN_CREATE);
	
	MPIDI_RMA_FUNC_ENTER(MPID_STATE_MPIDI_WIN_CREATE);

	/* FIXME: There should be no unreferenced args */
	MPIU_UNREFERENCED_ARG(info);
	MPIU_UNREFERENCED_ARG(RMAFns);

	MPIU_THREADPRIV_GET;

	comm_size = comm_ptr->local_size;
	rank = comm_ptr->rank;
	
	*win_ptr = (MPID_Win *)MPIU_Handle_obj_alloc( &MPID_Win_mem );
	MPIU_ERR_CHKANDJUMP(!(*win_ptr),mpi_errno,MPI_ERR_OTHER,L"**nomem");

	(*win_ptr)->fence_cnt = 0;
	(*win_ptr)->base = base;
	(*win_ptr)->pointer_size = SIZEOF_VOID_P;
	(*win_ptr)->size = size;
	(*win_ptr)->disp_unit = disp_unit;
	(*win_ptr)->start_group_ptr = NULL; 
	(*win_ptr)->start_assert = 0; 
	(*win_ptr)->attributes = NULL;
	(*win_ptr)->rma_ops_list = NULL;
	(*win_ptr)->lock_granted = 0;
	(*win_ptr)->current_lock_type = MPID_LOCK_NONE;
	(*win_ptr)->shared_lock_ref_cnt = 0;
	(*win_ptr)->lock_queue = NULL;
	(*win_ptr)->my_counter = 0;
	(*win_ptr)->my_pt_rma_puts_accs = 0;
	
	NMPI_Comm_dup(mpi_errno, (comm_ptr->handle, &((*win_ptr)->comm)));
	if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }

	/* allocate memory for the base addresses, disp_units, and completion counters of all processes */
	MPIU_CHKPMEM_MALLOC((*win_ptr)->base_addrs, MPIDI_Message_pointer_t *, comm_size*sizeof(MPIDI_Message_pointer_t), mpi_errno, L"(*win_ptr)->base_addrs");
	MPIU_CHKPMEM_MALLOC((*win_ptr)->pointer_sizes, int *, comm_size*sizeof(int), mpi_errno, L"(*win_ptr)->pointer_sizes");
	MPIU_CHKPMEM_MALLOC((*win_ptr)->disp_units, int *, comm_size*sizeof(int), mpi_errno, L"(*win_ptr)->disp_units");
	MPIU_CHKPMEM_MALLOC((*win_ptr)->all_win_handles, MPI_Win *, comm_size*sizeof(MPI_Win), mpi_errno, L"(*win_ptr)->all_win_handles");
	MPIU_CHKPMEM_MALLOC((*win_ptr)->pt_rma_puts_accs, int *, comm_size*sizeof(int), mpi_errno, L"(*win_ptr)->pt_rma_puts_accs");
	for (i=0; i<comm_size; i++)
		(*win_ptr)->pt_rma_puts_accs[i] = 0;
	
	/* get the addresses of the windows, window objects, and completion counters of all processes.  allocate temp. buffer for communication */
	MPIU_CHKLMEM_MALLOC(tmp_buf, Win_create_pkt_t *, comm_size*sizeof(Win_create_pkt_t), mpi_errno, L"tmp_buf");
	
	/* FIXME: This needs to be fixed for heterogeneous systems */
	tmp_buf[rank].addr = MPIDI_Local_ptr_to_message_pointer_t(base);
	tmp_buf[rank].pointer_size = SIZEOF_VOID_P;
	tmp_buf[rank].disp_unit = disp_unit;
	tmp_buf[rank].win = (MPI_Win) (*win_ptr)->handle;
	
	NMPI_Allgather(mpi_errno, (MPI_IN_PLACE, 0, MPI_DATATYPE_NULL, tmp_buf, sizeof(Win_create_pkt_t), MPI_BYTE, comm_ptr->handle));
	if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }
	
	for (i=0; i<comm_size; i++)
	{
		(*win_ptr)->base_addrs[i] = tmp_buf[i].addr;
		(*win_ptr)->pointer_sizes[i] = tmp_buf[i].pointer_size;
		(*win_ptr)->disp_units[i] = tmp_buf[i].disp_unit;
		(*win_ptr)->all_win_handles[i] = tmp_buf[i].win;
	}
		
 fn_exit:
	MPIU_CHKLMEM_FREEALL();
	MPIDI_RMA_FUNC_EXIT(MPID_STATE_MPIDI_WIN_CREATE);
	return mpi_errno;
	/* --BEGIN ERROR HANDLING-- */
 fn_fail:
	MPIU_CHKPMEM_REAP();
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

#undef FUNCNAME
#define FUNCNAME MPIDI_Win_free
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_Win_free(MPID_Win **win_ptr)
{
	int mpi_errno=MPI_SUCCESS, total_pt_rma_puts_accs, i, *recvcnts, comm_size;
	MPID_Comm *comm_ptr;
	MPIU_CHKLMEM_DECL(1);
	MPIU_THREADPRIV_DECL;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_WIN_FREE);
		
	MPIDI_RMA_FUNC_ENTER(MPID_STATE_MPIDI_WIN_FREE);
		
	MPIU_THREADPRIV_GET;

	/* set up the recvcnts array for the reduce scatter to check if all passive target rma operations are done */
	MPID_Comm_get_ptr( (*win_ptr)->comm, comm_ptr );
	comm_size = comm_ptr->local_size;
		
	MPIU_CHKLMEM_MALLOC(recvcnts, int *, comm_size*sizeof(int), mpi_errno, L"recvcnts");
	for (i=0; i<comm_size; i++)  recvcnts[i] = 1;
		
	NMPI_Reduce_scatter(mpi_errno, ((*win_ptr)->pt_rma_puts_accs, &total_pt_rma_puts_accs, recvcnts, MPI_INT, MPI_SUM, (*win_ptr)->comm));
	if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }

	if (total_pt_rma_puts_accs != (*win_ptr)->my_pt_rma_puts_accs)
	{
		MPID_Progress_state progress_state;
			
		/* poke the progress engine until the two are equal */
		MPID_Progress_start(&progress_state);
		while (total_pt_rma_puts_accs != (*win_ptr)->my_pt_rma_puts_accs)
		{
			mpi_errno = MPID_Progress_wait(&progress_state);
			/* --BEGIN ERROR HANDLING-- */
			if (mpi_errno != MPI_SUCCESS)
			{
				MPID_Progress_end(&progress_state);
				mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", L"**fail %s", L"making progress on the rma messages failed");
				goto fn_exit;
			}
			/* --END ERROR HANDLING-- */
		}
		MPID_Progress_end(&progress_state);
	}

	NMPI_Comm_free(mpi_errno, (&((*win_ptr)->comm)));
		
	MPIU_Free((*win_ptr)->base_addrs);
	MPIU_Free((*win_ptr)->pointer_sizes);
	MPIU_Free((*win_ptr)->disp_units);
	MPIU_Free((*win_ptr)->all_win_handles);
	MPIU_Free((*win_ptr)->pt_rma_puts_accs);
		
	/* check whether refcount needs to be decremented here as in group_free */
	MPIU_Handle_obj_free( &MPID_Win_mem, *win_ptr );
		
 fn_exit:
	MPIU_CHKLMEM_FREEALL();
	MPIDI_RMA_FUNC_EXIT(MPID_STATE_MPIDI_WIN_FREE);
	return mpi_errno;
	/* --BEGIN ERROR HANDLING-- */
 fn_fail:
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

#undef FUNCNAME
#define FUNCNAME MPIDI_Put
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_Put(void *origin_addr, int origin_count, MPI_Datatype
			origin_datatype, int target_rank, MPI_Aint target_disp,
			int target_count, MPI_Datatype target_datatype, MPID_Win *win_ptr)
{
	int mpi_errno = MPI_SUCCESS;
	int dt_contig, rank, predefined;
	MPIDI_RMA_ops *curr_ptr, *prev_ptr, *new_ptr;
	MPID_Datatype *dtp;
	MPI_Aint dt_true_lb;
	MPIDI_msg_sz_t data_sz;
	MPIU_CHKPMEM_DECL(1);
	MPIU_THREADPRIV_DECL;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_PUT);
		
	MPIDI_RMA_FUNC_ENTER(MPID_STATE_MPIDI_PUT);

	MPIU_THREADPRIV_GET;
	MPIDI_Datatype_get_info(origin_count, origin_datatype, dt_contig, data_sz, dtp,dt_true_lb); 
	
	if ((data_sz == 0) || (target_rank == MPI_PROC_NULL))
	{
		goto fn_exit;
	}

	/* FIXME: It makes sense to save the rank (and size) of the
	   communicator in the window structure to speed up these operations,
	   or to save a pointer to the communicator structure, rather than
	   just the handle 
	*/
	NMPI_Comm_rank(mpi_errno, (win_ptr->comm, &rank));

	/* If the put is a local operation, do it here */
	if (target_rank == rank)
	{
		mpi_errno = MPIR_Localcopy(origin_addr, origin_count, origin_datatype,
								   (char *) win_ptr->base + win_ptr->disp_unit *
								   target_disp, target_count, target_datatype); 
	}
	else
	{
		/* queue it up */
		curr_ptr = win_ptr->rma_ops_list;
		prev_ptr = curr_ptr;
		while (curr_ptr != NULL)
		{
			prev_ptr = curr_ptr;
			curr_ptr = curr_ptr->next;
		}
		
		MPIU_CHKPMEM_MALLOC(new_ptr, MPIDI_RMA_ops *, sizeof(MPIDI_RMA_ops), mpi_errno, L"RMA operation entry");
		if (prev_ptr != NULL)
			prev_ptr->next = new_ptr;
		else 
			win_ptr->rma_ops_list = new_ptr;
		
		new_ptr->next = NULL;  
		new_ptr->type = MPIDI_RMA_PUT;
		new_ptr->origin_addr = origin_addr;
		new_ptr->origin_count = origin_count;
		new_ptr->origin_datatype = origin_datatype;
		new_ptr->target_rank = target_rank;
		new_ptr->target_disp = target_disp;
		new_ptr->target_count = target_count;
		new_ptr->target_datatype = target_datatype;
		
		/* if source or target datatypes are derived, increment their reference counts */ 
		MPIDI_CH3I_DATATYPE_IS_PREDEFINED(origin_datatype, predefined);
		if (!predefined)
		{
			MPID_Datatype_get_ptr(origin_datatype, dtp);
			MPID_Datatype_add_ref(dtp);
		}
		MPIDI_CH3I_DATATYPE_IS_PREDEFINED(target_datatype, predefined);
		if (!predefined)
		{
			MPID_Datatype_get_ptr(target_datatype, dtp);
			MPID_Datatype_add_ref(dtp);
		}
	}

  fn_exit:
	MPIDI_RMA_FUNC_EXIT(MPID_STATE_MPIDI_PUT);	
	return mpi_errno;

	/* --BEGIN ERROR HANDLING-- */
  fn_fail:
	MPIU_CHKPMEM_REAP();
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}



#undef FUNCNAME
#define FUNCNAME MPIDI_Get
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_Get(void *origin_addr, int origin_count, MPI_Datatype
			origin_datatype, int target_rank, MPI_Aint target_disp,
			int target_count, MPI_Datatype target_datatype, MPID_Win *win_ptr)
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_msg_sz_t data_sz;
	int dt_contig, rank, predefined;
	MPI_Aint dt_true_lb;
	MPIDI_RMA_ops *curr_ptr, *prev_ptr, *new_ptr;
	MPID_Datatype *dtp;
	MPIU_CHKPMEM_DECL(1);
	MPIU_THREADPRIV_DECL;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_GET);
		
	MPIDI_RMA_FUNC_ENTER(MPID_STATE_MPIDI_GET);

	MPIU_THREADPRIV_GET;
	MPIDI_Datatype_get_info(origin_count, origin_datatype, dt_contig, data_sz, dtp, dt_true_lb); 

	if ((data_sz == 0) || (target_rank == MPI_PROC_NULL))
	{
		goto fn_exit;
	}

	/* FIXME: It makes sense to save the rank (and size) of the
	   communicator in the window structure to speed up these operations */
	NMPI_Comm_rank(mpi_errno, (win_ptr->comm, &rank));

	/* If the get is a local operation, do it here */
	if (target_rank == rank)
	{
		mpi_errno = MPIR_Localcopy((char *) win_ptr->base +
								   win_ptr->disp_unit * target_disp,
								   target_count, target_datatype,
								   origin_addr, origin_count,
								   origin_datatype);  
	}
	else
	{
		/* queue it up */
		curr_ptr = win_ptr->rma_ops_list;
		prev_ptr = curr_ptr;
		while (curr_ptr != NULL)
		{
			prev_ptr = curr_ptr;
			curr_ptr = curr_ptr->next;
		}
		
		MPIU_CHKPMEM_MALLOC(new_ptr, MPIDI_RMA_ops *, sizeof(MPIDI_RMA_ops), mpi_errno, L"RMA operation entry");
		if (prev_ptr != NULL)
		{
			prev_ptr->next = new_ptr;
		}
		else
		{
			win_ptr->rma_ops_list = new_ptr;
		}
			
		new_ptr->next = NULL;  
		new_ptr->type = MPIDI_RMA_GET;
		new_ptr->origin_addr = origin_addr;
		new_ptr->origin_count = origin_count;
		new_ptr->origin_datatype = origin_datatype;
		new_ptr->target_rank = target_rank;
		new_ptr->target_disp = target_disp;
		new_ptr->target_count = target_count;
		new_ptr->target_datatype = target_datatype;
		
		/* if source or target datatypes are derived, increment their reference counts */ 
		MPIDI_CH3I_DATATYPE_IS_PREDEFINED(origin_datatype, predefined);
		if (!predefined)
		{
			MPID_Datatype_get_ptr(origin_datatype, dtp);
			MPID_Datatype_add_ref(dtp);
		}
		MPIDI_CH3I_DATATYPE_IS_PREDEFINED(target_datatype, predefined);
		if (!predefined)
		{
			MPID_Datatype_get_ptr(target_datatype, dtp);
			MPID_Datatype_add_ref(dtp);
		}
	}

  fn_exit:
	MPIDI_RMA_FUNC_EXIT(MPID_STATE_MPIDI_GET);
	return mpi_errno;

	/* --BEGIN ERROR HANDLING-- */
  fn_fail:
	MPIU_CHKPMEM_REAP();
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}



#undef FUNCNAME
#define FUNCNAME MPIDI_Accumulate
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_Accumulate(void *origin_addr, int origin_count, MPI_Datatype
					origin_datatype, int target_rank, MPI_Aint target_disp,
					int target_count, MPI_Datatype target_datatype, MPI_Op op,
					MPID_Win *win_ptr)
{
	int mpi_errno=MPI_SUCCESS;
	MPIDI_msg_sz_t data_sz;
	int dt_contig, rank, origin_predefined, target_predefined;
	MPI_Aint dt_true_lb;
	MPIDI_RMA_ops *curr_ptr, *prev_ptr, *new_ptr;
	MPID_Datatype *dtp;
	MPIU_CHKLMEM_DECL(2);
	MPIU_CHKPMEM_DECL(1);
	MPIU_THREADPRIV_DECL;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_ACCUMULATE);
	
	MPIDI_RMA_FUNC_ENTER(MPID_STATE_MPIDI_ACCUMULATE);

	MPIU_THREADPRIV_GET;
	MPIDI_Datatype_get_info(origin_count, origin_datatype, dt_contig, data_sz, dtp, dt_true_lb);  
	
	if ((data_sz == 0) || (target_rank == MPI_PROC_NULL))
	{
		goto fn_exit;
	}

	/* FIXME: It makes sense to save the rank (and size) of the
	   communicator in the window structure to speed up these operations,
	   or to save a pointer to the communicator structure, rather than
	   just the handle 
	*/
	NMPI_Comm_rank(mpi_errno, (win_ptr->comm, &rank));
	
	MPIDI_CH3I_DATATYPE_IS_PREDEFINED(origin_datatype, origin_predefined);
	MPIDI_CH3I_DATATYPE_IS_PREDEFINED(target_datatype, target_predefined);

	if (target_rank == rank)
	{
		MPI_User_function *uop;
		
		if (op == MPI_REPLACE)
		{
			mpi_errno = MPIR_Localcopy(origin_addr, origin_count, origin_datatype,
									   (char *) win_ptr->base + win_ptr->disp_unit *
									   target_disp, target_count, target_datatype); 
			goto fn_exit;
		}
		
		MPIU_ERR_CHKANDJUMP1((HANDLE_GET_KIND(op) != HANDLE_KIND_BUILTIN), mpi_errno, MPI_ERR_OP, L"**opnotpredefined", L"**opnotpredefined %d", op );
		
		/* get the function by indexing into the op table */
		uop = MPIR_Op_table[(op)%16 - 1];
		
		if (origin_predefined && target_predefined)
		{	
			(*uop)(origin_addr, (char *) win_ptr->base + win_ptr->disp_unit * target_disp, &target_count, &target_datatype);
		}
		else
		{
			/* derived datatype */
			
			MPID_Segment *segp;
			DLOOP_VECTOR *dloop_vec;
			MPI_Aint first, last;
			int vec_len, i, type_size, count;
			MPI_Datatype type;
			MPID_Datatype *dtp;
			MPI_Aint true_lb, true_extent, extent;
			void *tmp_buf=NULL, *source_buf, *target_buf;
			
			if (origin_datatype != target_datatype)
			{
				/* first copy the data into a temporary buffer with
				   the same datatype as the target. Then do the
				   accumulate operation. */
				
				NMPI_Type_get_true_extent(mpi_errno, (target_datatype, &true_lb, &true_extent));
				MPIU_ERR_CHKANDJUMP((mpi_errno), mpi_errno, MPI_ERR_OTHER, L"**fail");
				
				MPID_Datatype_get_extent_macro(target_datatype, extent); 
				
				MPIU_CHKLMEM_MALLOC(tmp_buf, void *, target_count * (MPIR_MAX(extent,true_extent)), mpi_errno, L"temporary buffer");
				/* adjust for potential negative lower bound in datatype */
				tmp_buf = (void *)((char*)tmp_buf - true_lb);
				
				mpi_errno = MPIR_Localcopy(origin_addr, origin_count, origin_datatype, tmp_buf, target_count, target_datatype);  
				if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }
			}
			
			segp = MPID_Segment_alloc();
			MPIU_ERR_CHKANDJUMP((!segp), mpi_errno, MPI_ERR_OTHER, L"**nomem"); 
			MPID_Segment_init(NULL, target_count, target_datatype, segp, 0);
			first = 0;
			last  = SEGMENT_IGNORE_LAST;
			
			MPID_Datatype_get_ptr(target_datatype, dtp);
			vec_len = dtp->n_contig_blocks * target_count + 1; 
			/* +1 needed because Rob says so */
			MPIU_CHKLMEM_MALLOC(dloop_vec, DLOOP_VECTOR *, vec_len * sizeof(DLOOP_VECTOR), mpi_errno, L"dloop vector");
			
			MPID_Segment_pack_vector(segp, first, &last, dloop_vec, &vec_len);
			
			source_buf = (tmp_buf != NULL) ? tmp_buf : origin_addr;
			target_buf = (char *) win_ptr->base + win_ptr->disp_unit * target_disp;
			type = dtp->eltype;
			type_size = MPID_Datatype_get_basic_size(type);
			for (i=0; i<vec_len; i++)
			{
				count = (dloop_vec[i].DLOOP_VECTOR_LEN)/type_size;
				(*uop)((char *)source_buf + MPIU_PtrToAint(dloop_vec[i].DLOOP_VECTOR_BUF),
					   (char *)target_buf + MPIU_PtrToAint(dloop_vec[i].DLOOP_VECTOR_BUF),
					   &count, &type);
			}
			
			MPID_Segment_free(segp);
		}
	}
	else
	{
		/* queue it up */
		curr_ptr = win_ptr->rma_ops_list;
		prev_ptr = curr_ptr;
		while (curr_ptr != NULL)
		{
			prev_ptr = curr_ptr;
			curr_ptr = curr_ptr->next;
		}
		
		MPIU_CHKPMEM_MALLOC(new_ptr, MPIDI_RMA_ops *, sizeof(MPIDI_RMA_ops), mpi_errno, L"RMA operation entry");
		if (prev_ptr != NULL)
		{
			prev_ptr->next = new_ptr;
		}
		else
		{
			win_ptr->rma_ops_list = new_ptr;
		}
		
		new_ptr->next = NULL;  
		new_ptr->type = MPIDI_RMA_ACCUMULATE;
		new_ptr->origin_addr = origin_addr;
		new_ptr->origin_count = origin_count;
		new_ptr->origin_datatype = origin_datatype;
		new_ptr->target_rank = target_rank;
		new_ptr->target_disp = target_disp;
		new_ptr->target_count = target_count;
		new_ptr->target_datatype = target_datatype;
		new_ptr->op = op;
		
		/* if source or target datatypes are derived, increment their reference counts */ 
		if (!origin_predefined)
		{
			MPID_Datatype_get_ptr(origin_datatype, dtp);
			MPID_Datatype_add_ref(dtp);
		}
		if (!target_predefined)
		{
			MPID_Datatype_get_ptr(target_datatype, dtp);
			MPID_Datatype_add_ref(dtp);
		}
	}

 fn_exit:
	MPIU_CHKLMEM_FREEALL();
	MPIDI_RMA_FUNC_EXIT(MPID_STATE_MPIDI_ACCUMULATE);
	return mpi_errno;

	/* --BEGIN ERROR HANDLING-- */
  fn_fail:
	MPIU_CHKPMEM_REAP();
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}


#undef FUNCNAME
#define FUNCNAME MPIDI_Alloc_mem
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
void *MPIDI_Alloc_mem( size_t size, MPID_Info *info_ptr )
{
	void *ap;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_ALLOC_MEM);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_ALLOC_MEM);

	MPIU_UNREFERENCED_ARG(info_ptr);
	ap = MPIU_Malloc(size);
	
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_ALLOC_MEM);
	return ap;
}


#undef FUNCNAME
#define FUNCNAME MPIDI_Free_mem
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_Free_mem( void *ptr )
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_FREE_MEM);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_FREE_MEM);

	MPIU_Free(ptr);
	
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_FREE_MEM);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_Win_fence
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_Win_fence(int assert, MPID_Win *win_ptr)
{
	int mpi_errno = MPI_SUCCESS;
	int comm_size, done, *recvcnts;
	int *rma_target_proc, *nops_to_proc, i, total_op_count, *curr_ops_cnt;
	MPIDI_RMA_ops *curr_ptr, *next_ptr;
	MPID_Comm *comm_ptr;
	MPID_Request **requests=NULL; /* array of requests */
	MPI_Win source_win_handle, target_win_handle;
	MPIDI_RMA_dtype_info *dtype_infos=NULL;
	void **dataloops=NULL;	/* to store dataloops for each datatype */
	MPID_Progress_state progress_state;
	MPIU_CHKLMEM_DECL(7);
	MPIU_THREADPRIV_DECL;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_WIN_FENCE);

	MPIDI_RMA_FUNC_ENTER(MPID_STATE_MPIDI_WIN_FENCE);

	MPIU_THREADPRIV_GET;
	/* In case this process was previously the target of passive target rma
	 * operations, we need to take care of the following...
	 * Since we allow MPI_Win_unlock to return without a done ack from
	 * the target in the case of multiple rma ops and exclusive lock,
	 * we need to check whether there is a lock on the window, and if
	 * there is a lock, poke the progress engine until the operartions
	 * have completed and the lock is released. */
	if (win_ptr->current_lock_type != MPID_LOCK_NONE)
	{
		MPID_Progress_start(&progress_state);
		while (win_ptr->current_lock_type != MPID_LOCK_NONE)
		{
			/* poke the progress engine */
			mpi_errno = MPID_Progress_wait(&progress_state);
			/* --BEGIN ERROR HANDLING-- */
			if (mpi_errno != MPI_SUCCESS)
			{
				MPID_Progress_end(&progress_state);
				mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", L"**fail %s", L"making progress on the rma messages failed");
				goto fn_exit;
			}
			/* --END ERROR HANDLING-- */
		}
		MPID_Progress_end(&progress_state);
	}
	
	if (assert & MPI_MODE_NOPRECEDE)
	{
		win_ptr->fence_cnt = (assert & MPI_MODE_NOSUCCEED) ? 0 : 1;
		goto fn_exit;
	}
	
	if ((win_ptr->fence_cnt == 0) && ((assert & MPI_MODE_NOSUCCEED) != 1))
	{
		/* win_ptr->fence_cnt == 0 means either this is the very first
		   call to fence or the preceding fence had the
		   MPI_MODE_NOSUCCEED assert. 
		   Do nothing except increment the count. */
		win_ptr->fence_cnt = 1;
	}
	else
	{
		/* This is the second or later fence. Do all the preceding RMA ops. */
		
		MPID_Comm_get_ptr( win_ptr->comm, comm_ptr );
		
		/* First inform every process whether it is a target of RMA ops from this process */
		comm_size = comm_ptr->local_size;

		MPIU_CHKLMEM_MALLOC(rma_target_proc, int *, comm_size*sizeof(int), mpi_errno, L"rma_target_proc");
		for (i=0; i<comm_size; i++) rma_target_proc[i] = 0;
		
		/* keep track of no. of ops to each proc. Needed for knowing
		   whether or not to decrement the completion counter. The
		   completion counter is decremented only on the last
		   operation. */
		MPIU_CHKLMEM_MALLOC(nops_to_proc, int *, comm_size*sizeof(int), mpi_errno, L"nops_to_proc");
		for (i=0; i<comm_size; i++) nops_to_proc[i] = 0;

		/* set rma_target_proc[i] to 1 if rank i is a target of RMA ops from this process */
		total_op_count = 0;
		curr_ptr = win_ptr->rma_ops_list;
		while (curr_ptr != NULL)
		{
			total_op_count++;
			rma_target_proc[curr_ptr->target_rank] = 1;
			nops_to_proc[curr_ptr->target_rank]++;
			curr_ptr = curr_ptr->next;
		}
		
		MPIU_CHKLMEM_MALLOC(curr_ops_cnt, int *, comm_size*sizeof(int), mpi_errno, L"curr_ops_cnt");
		for (i=0; i<comm_size; i++) curr_ops_cnt[i] = 0;
		
		if (total_op_count != 0)
		{
			MPIU_CHKLMEM_MALLOC(requests, MPID_Request **, total_op_count*sizeof(MPID_Request*), mpi_errno, L"requests");
			MPIU_CHKLMEM_MALLOC(dtype_infos, MPIDI_RMA_dtype_info *, total_op_count*sizeof(MPIDI_RMA_dtype_info), mpi_errno, L"dtype_infos");
			MPIU_CHKLMEM_MALLOC(dataloops, void **, total_op_count*sizeof(void*), mpi_errno, L"dataloops");
			for (i=0; i<total_op_count; i++)
				dataloops[i] = NULL;
		}
		
		/* do a reduce_scatter (with MPI_SUM) on rma_target_proc. As a result,
		   each process knows how many other processes will be doing
		   RMA ops on its window */  
			
		/* first initialize the completion counter. */
		win_ptr->my_counter = comm_size;
			
		/* set up the recvcnts array for reduce scatter */
		MPIU_CHKLMEM_MALLOC(recvcnts, int *, comm_size*sizeof(int), mpi_errno, L"recvcnts");
		for (i=0; i<comm_size; i++) recvcnts[i] = 1;

		NMPI_Reduce_scatter(mpi_errno, (MPI_IN_PLACE, rma_target_proc, recvcnts, MPI_INT, MPI_SUM, win_ptr->comm));
		/* result is stored in rma_target_proc[0] */
		if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }

		/* Set the completion counter */
		/* FIXME: MT: this needs to be done atomically because other
		   procs have the address and could decrement it. */
		win_ptr->my_counter = win_ptr->my_counter - comm_size + rma_target_proc[0];  
			
		i = 0;
		curr_ptr = win_ptr->rma_ops_list;
		while (curr_ptr != NULL)
		{
			/* The completion counter at the target is decremented only on 
			   the last RMA operation. We indicate the last operation by 
			   passing the source_win_handle only on the last operation. 
			   Otherwise, we pass NULL */
			if (curr_ops_cnt[curr_ptr->target_rank] == nops_to_proc[curr_ptr->target_rank] - 1) 
				source_win_handle = win_ptr->handle;
			else 
				source_win_handle = MPI_WIN_NULL;
			
			target_win_handle = win_ptr->all_win_handles[curr_ptr->target_rank];
			
			switch (curr_ptr->type)
			{
			case (MPIDI_RMA_PUT):
			case (MPIDI_RMA_ACCUMULATE):
				mpi_errno = MPIDI_CH3I_Send_rma_msg(curr_ptr, win_ptr, source_win_handle, target_win_handle, &dtype_infos[i], &dataloops[i], &requests[i]);
				if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }
				break;
			case (MPIDI_RMA_GET):
				mpi_errno = MPIDI_CH3I_Recv_rma_msg(curr_ptr, win_ptr, source_win_handle, target_win_handle, &dtype_infos[i], &dataloops[i], &requests[i]);
				if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }
				break;
			default:
				/* --BEGIN ERROR HANDLING-- */
				mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", L"**fail %s", L"invalid RMA operation");
				goto fn_exit;
				/* --END ERROR HANDLING-- */
			}
			i++;
			curr_ops_cnt[curr_ptr->target_rank]++;
			curr_ptr = curr_ptr->next;
		}
		
			
		if (total_op_count)
		{ 
			done = 1;
			MPID_Progress_start(&progress_state);
			while (total_op_count)
			{
				for (i=0; i<total_op_count; i++)
				{
					if (requests[i] != NULL)
					{
						if (*(requests[i]->cc_ptr) != 0)
						{
							done = 0;
							break;
						}
						else
						{
							mpi_errno = requests[i]->status.MPI_ERROR;
							/* --BEGIN ERROR HANDLING-- */
							if (mpi_errno != MPI_SUCCESS)
							{
								MPID_Progress_end(&progress_state);
								mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", L"**fail %s", L"rma message operation failed");
								goto fn_exit;
							}
							/* --END ERROR HANDLING-- */
							/* if origin datatype was a derived datatype, it will get freed when the request gets freed. */ 
							MPID_Request_release(requests[i]);
							requests[i] = NULL;
						}
					}
				}
					
				if (done)
				{
					break;
				}
					
				mpi_errno = MPID_Progress_wait(&progress_state);
				/* --BEGIN ERROR HANDLING-- */
				if (mpi_errno != MPI_SUCCESS)
				{
					MPID_Progress_end(&progress_state);
					mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", L"**fail %s", L"making progress on the rma messages failed");
					goto fn_exit;
				}
				/* --END ERROR HANDLING-- */
				
				done = 1;
			} 
			MPID_Progress_end(&progress_state);
		}
			
		if (total_op_count != 0)
		{
			for (i=0; i<total_op_count; i++)
			{
				if (dataloops[i] != NULL)
				{
					MPIU_Free(dataloops[i]); /* allocated in send_rma_msg or recv_rma_msg */
				}
			}
		}
		
		/* free MPIDI_RMA_ops_list */
		curr_ptr = win_ptr->rma_ops_list;
		while (curr_ptr != NULL)
		{
			next_ptr = curr_ptr->next;
			MPIU_Free(curr_ptr);
			curr_ptr = next_ptr;
		}
		win_ptr->rma_ops_list = NULL;
		
		/* wait for all operations from other processes to finish */
		if (win_ptr->my_counter)
		{
			MPID_Progress_start(&progress_state);
			while (win_ptr->my_counter)
			{
				mpi_errno = MPID_Progress_wait(&progress_state);
				/* --BEGIN ERROR HANDLING-- */
				if (mpi_errno != MPI_SUCCESS)
				{
					MPID_Progress_end(&progress_state);
					mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", L"**fail %s", L"making progress on the rma messages failed");
					goto fn_exit;
				}
				/* --END ERROR HANDLING-- */
			}
			MPID_Progress_end(&progress_state);
		} 
		
		if (assert & MPI_MODE_NOSUCCEED)
		{
			win_ptr->fence_cnt = 0;
		}
	}

 fn_exit:
	MPIU_CHKLMEM_FREEALL();
	MPIDI_RMA_FUNC_EXIT(MPID_STATE_MPIDI_WIN_FENCE);
	return mpi_errno;
	/* --BEGIN ERROR HANDLING-- */
 fn_fail:
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}


#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_Send_rma_msg
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int MPIDI_CH3I_Send_rma_msg(MPIDI_RMA_ops *rma_op, MPID_Win *win_ptr,
							MPI_Win source_win_handle, MPI_Win target_win_handle, 
							MPIDI_RMA_dtype_info *dtype_info, 
							void **dataloop, MPID_Request **request) 
{
	void *serialized_datatype_buffer = NULL;
	int serialized_buffer_length = 0;
	MPIDI_CH3_Pkt_t upkt;
	MPIDI_CH3_Pkt_put_t *put_pkt = &upkt.put;
	MPIDI_CH3_Pkt_accum_t *accum_pkt = &upkt.accum;
	MPID_IOV iov[MPID_IOV_LIMIT];
	int mpi_errno=MPI_SUCCESS, predefined;
	int origin_dt_derived, target_dt_derived, origin_type_size, iovcnt, iov_n; 
	MPIDI_VC_t * vc;
	MPID_Comm *comm_ptr;
	MPID_Datatype *target_dtp=NULL, *origin_dtp=NULL;
	MPIU_CHKPMEM_DECL(1);
	MPIU_CHKLMEM_DECL(1);
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_SEND_RMA_MSG);
	MPIDI_STATE_DECL(MPID_STATE_MEMCPY);

	MPIDI_RMA_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_SEND_RMA_MSG);

	if (rma_op->type == MPIDI_RMA_PUT)
	{
		MPIDI_Pkt_init(put_pkt, MPIDI_CH3_PKT_PUT);
		put_pkt->addr = ((MPIDI_Message_uchar_pointer_t) win_ptr->base_addrs[rma_op->target_rank]) +
			win_ptr->disp_units[rma_op->target_rank] * rma_op->target_disp;

		put_pkt->count = rma_op->target_count;
		put_pkt->datatype = rma_op->target_datatype;
		put_pkt->dataloop_size = 0;
		put_pkt->target_win_handle = target_win_handle;
		put_pkt->source_win_handle = source_win_handle;

		iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST) put_pkt;
		iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)sizeof(*put_pkt);
	}
	else
	{
		MPIDI_Pkt_init(accum_pkt, MPIDI_CH3_PKT_ACCUMULATE);
		accum_pkt->addr = ((MPIDI_Message_uchar_pointer_t) win_ptr->base_addrs[rma_op->target_rank]) +
			win_ptr->disp_units[rma_op->target_rank] * rma_op->target_disp;
		accum_pkt->count = rma_op->target_count;
		accum_pkt->datatype = rma_op->target_datatype;
		accum_pkt->dataloop_size = 0;
		accum_pkt->op = rma_op->op;
		accum_pkt->target_win_handle = target_win_handle;
		accum_pkt->source_win_handle = source_win_handle;

		iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST) accum_pkt;
		iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)sizeof(*accum_pkt);
	}

/*	wprintf(L"send pkt: type %d, addr %d, count %d, base %d\n", rma_pkt->type,
		   rma_pkt->addr, rma_pkt->count, win_ptr->base_addrs[rma_op->target_rank]);
	fflush(stdout);
*/

	MPID_Comm_get_ptr(win_ptr->comm, comm_ptr);
	MPIDI_Comm_get_vc(comm_ptr, rma_op->target_rank, &vc);

	MPIDI_CH3I_DATATYPE_IS_PREDEFINED(rma_op->origin_datatype, predefined);
	if (!predefined)
	{
		origin_dt_derived = 1;
		MPID_Datatype_get_ptr(rma_op->origin_datatype, origin_dtp);
	}
	else
	{
		origin_dt_derived = 0;
	}

	MPIDI_CH3I_DATATYPE_IS_PREDEFINED(rma_op->target_datatype, predefined);
	if (!predefined)
	{
		target_dt_derived = 1;
		MPID_Datatype_get_ptr(rma_op->target_datatype, target_dtp);
	}
	else
	{
		target_dt_derived = 0;
	}

	if (target_dt_derived && (win_ptr->pointer_sizes[rma_op->target_rank] != win_ptr->pointer_size))
	{
#if 0
		/* Transform the target datatype to match the pointer size of the target. */
		MPID_Datatype *tmp_datatype;

		/* FIXME: This doesn't work because the dataloop inside the MPID_Datatype structure is allocated by the ConvertDatatype function */
		/* So when the tmp_datatype structure is freed the dataloop memory region is orphaned. */
		MPIU_CHKLMEM_MALLOC(tmp_datatype, MPID_Datatype *, sizeof(MPID_Datatype), mpi_errno, L"datatype");
		MPIDI_FUNC_ENTER(MPID_STATE_MEMCPY);
		memcpy(tmp_datatype, target_dtp, sizeof(MPID_Datatype));
		MPIDI_FUNC_EXIT(MPID_STATE_MEMCPY);
		target_dtp = tmp_datatype;
		mpi_errno = ConvertDatatype(win_ptr->pointer_size, win_ptr->pointer_sizes[rma_op->target_rank], target_dtp);
		MPIU_ERR_CHK_POP(mpi_errno);
#endif

		/*MPI_Datatype datatype;*/
		SerializeDatatype(rma_op->target_datatype, &serialized_datatype_buffer, &serialized_buffer_length);
		/*
		wprintf(L"serialized length: %d\n", length);fflush(stdout);
		DeSerializeDatatype(serialized_datatype_buffer, &datatype);
		wprintf(L"printing the de-serialized datatype:\n");fflush(stdout);
		PrintDatatype(datatype);
		*/
	}

	if (target_dt_derived)
	{
		/* derived datatype on target. fill derived datatype info */
		if (serialized_datatype_buffer != NULL)
		{
			/* Use a negative length to indicate a serialized datatype is being sent */
			if (rma_op->type == MPIDI_RMA_PUT)
			{
				put_pkt->dataloop_size = -serialized_buffer_length;
			}
			else
			{
				accum_pkt->dataloop_size = -serialized_buffer_length;
			}
			*dataloop = NULL;
		}
		else
		{
			dtype_info->is_contig = target_dtp->is_contig;
			dtype_info->n_contig_blocks = target_dtp->n_contig_blocks;
			dtype_info->size = target_dtp->size;
			dtype_info->extent = target_dtp->extent;
			dtype_info->dataloop_size = target_dtp->dataloop_size;
			dtype_info->dataloop_depth = target_dtp->dataloop_depth;
			dtype_info->eltype = target_dtp->eltype;
			dtype_info->dataloop = target_dtp->dataloop;
			dtype_info->ub = target_dtp->ub;
			dtype_info->lb = target_dtp->lb;
			dtype_info->true_ub = target_dtp->true_ub;
			dtype_info->true_lb = target_dtp->true_lb;
			dtype_info->has_sticky_ub = target_dtp->has_sticky_ub;
			dtype_info->has_sticky_lb = target_dtp->has_sticky_lb;

			MPIU_CHKPMEM_MALLOC(*dataloop, void *, target_dtp->dataloop_size, mpi_errno, L"dataloop");

			MPIDI_FUNC_ENTER(MPID_STATE_MEMCPY);
			memcpy(*dataloop, target_dtp->dataloop, target_dtp->dataloop_size);
			MPIDI_FUNC_EXIT(MPID_STATE_MEMCPY);

			if (rma_op->type == MPIDI_RMA_PUT)
			{
				put_pkt->dataloop_size = target_dtp->dataloop_size;
			}
			else
			{
				accum_pkt->dataloop_size = target_dtp->dataloop_size;
			}
		}
	}

	MPID_Datatype_get_size_macro(rma_op->origin_datatype, origin_type_size);

	if (!origin_dt_derived)
	{
		/* basic datatype on origin */
		if (!target_dt_derived)
		{
			/* basic datatype on target */
			iov[1].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)rma_op->origin_addr;
			iov[1].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)(rma_op->origin_count * origin_type_size);
			iovcnt = 2;
		}
		else
		{
			/* derived datatype on target */
			if (serialized_datatype_buffer != NULL)
			{
				iov[1].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)serialized_datatype_buffer;
				iov[1].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)serialized_buffer_length;

				iov[2].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)rma_op->origin_addr;
				iov[2].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)(rma_op->origin_count * origin_type_size);
				iovcnt = 3;
			}
			else
			{
				iov[1].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)dtype_info;
				iov[1].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)sizeof(*dtype_info);
				iov[2].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)*dataloop;
				iov[2].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)target_dtp->dataloop_size;

				iov[3].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)rma_op->origin_addr;
				iov[3].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)(rma_op->origin_count * origin_type_size);
				iovcnt = 4;
			}
		}

		mpi_errno = MPIDI_CH3.iStartMsgv(vc, iov, iovcnt, request);
		/* --BEGIN ERROR HANDLING-- */
		if (mpi_errno != MPI_SUCCESS)
		{
			mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|rmamsg", 0);
			MPIDI_RMA_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SEND_RMA_MSG);
			return mpi_errno;
		}
		/* --END ERROR HANDLING-- */
	}
	else
	{
		/* derived datatype on origin */

		if (!target_dt_derived)
		{
			/* basic datatype on target */
			iovcnt = 1;
		}
		else
		{
			/* derived datatype on target */
			if (serialized_datatype_buffer != NULL)
			{
				iov[1].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)serialized_datatype_buffer;
				iov[1].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)serialized_buffer_length;
				iovcnt = 2;
			}
			else
			{
				iov[1].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)dtype_info;
				iov[1].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)sizeof(*dtype_info);
				iov[2].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)*dataloop;
				iov[2].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)target_dtp->dataloop_size;
				iovcnt = 3;
			}
		}

		*request = MPID_Request_create();
		if (*request == NULL)
		{
			/* --BEGIN ERROR HANDLING-- */
			mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
			MPIDI_RMA_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SEND_RMA_MSG);
			return mpi_errno;
			/* --END ERROR HANDLING-- */
		}

		MPIU_Object_set_ref(*request, 2);
		(*request)->kind = MPID_REQUEST_SEND;
			
		(*request)->dev.datatype_ptr = origin_dtp;
		/* this will cause the datatype to be freed when the request is freed. */ 

		MPID_Segment_init(rma_op->origin_addr, rma_op->origin_count, rma_op->origin_datatype, &((*request)->dev.segment), 0);
		(*request)->dev.segment_first = 0;
		(*request)->dev.segment_size = rma_op->origin_count * origin_type_size;
			
		iov_n = MPID_IOV_LIMIT - iovcnt;
		mpi_errno = MPIDI_CH3U_Request_load_send_iov(*request, &iov[iovcnt], &iov_n); 
		if (mpi_errno == MPI_SUCCESS)
		{
			iov_n += iovcnt;
			
			mpi_errno = MPIDI_CH3.iSendv(vc, *request, iov, iov_n);
			/* --BEGIN ERROR HANDLING-- */
			if (mpi_errno != MPI_SUCCESS)
			{
				MPID_Datatype_release((*request)->dev.datatype_ptr);
				MPIU_Object_set_ref(*request, 0);
				MPIDI_CH3_Request_destroy(*request);
				*request = NULL;
				mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|rmamsg", 0);
				MPIDI_RMA_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SEND_RMA_MSG);
				return mpi_errno;
			}
			/* --END ERROR HANDLING-- */
		}
		else
		{
			/* --BEGIN ERROR HANDLING-- */
			MPID_Datatype_release((*request)->dev.datatype_ptr);
			MPIU_Object_set_ref(*request, 0);
			MPIDI_CH3_Request_destroy(*request);
			*request = NULL;
			mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|loadsendiov", 0);
			MPIDI_RMA_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SEND_RMA_MSG);
			return mpi_errno;
			/* --END ERROR HANDLING-- */
		}
	}

	if (target_dt_derived)
	{
		MPID_Datatype_release(target_dtp);
	}

 fn_exit:
	MPIU_CHKLMEM_FREEALL();
	MPIDI_RMA_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SEND_RMA_MSG);
	return mpi_errno;
	/* --BEGIN ERROR HANDLING-- */
 fn_fail:
	MPIU_CHKPMEM_REAP();
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}



#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_Recv_rma_msg
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int MPIDI_CH3I_Recv_rma_msg(MPIDI_RMA_ops *rma_op, MPID_Win *win_ptr,
							MPI_Win source_win_handle, MPI_Win target_win_handle, 
							MPIDI_RMA_dtype_info *dtype_info, void **dataloop, 
							MPID_Request **request) 
{
	void *serialized_datatype_buffer = NULL;
	int serialized_buffer_length = 0;
	MPIDI_CH3_Pkt_t upkt;
	MPIDI_CH3_Pkt_get_t *get_pkt = &upkt.get;
	int mpi_errno, predefined;
	MPIDI_VC_t * vc;
	MPID_Comm *comm_ptr;
	MPID_Request *req = NULL;
	MPID_Datatype *dtp;
	MPID_IOV iov[MPID_IOV_LIMIT];
	MPIU_CHKPMEM_DECL(1);
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_RECV_RMA_MSG);
	MPIDI_STATE_DECL(MPID_STATE_MEMCPY);

	MPIDI_RMA_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_RECV_RMA_MSG);

	/* create a request, store the origin buf, cnt, datatype in it,
	   and pass a handle to it in the get packet. When the get
	   response comes from the target, it will contain the request
	   handle. */  
	req = MPID_Request_create();
	if (req == NULL)
	{
		/* --BEGIN ERROR HANDLING-- */
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
		MPIDI_RMA_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_RECV_RMA_MSG);
		return mpi_errno;
		/* --END ERROR HANDLING-- */
	}

	*request = req;

	MPIU_Object_set_ref(req, 2);

	req->dev.user_buf = rma_op->origin_addr;
	req->dev.user_count = rma_op->origin_count;
	req->dev.datatype = rma_op->origin_datatype;
	req->dev.target_win_handle = MPI_WIN_NULL;
	req->dev.source_win_handle = source_win_handle;
	MPIDI_CH3I_DATATYPE_IS_PREDEFINED(req->dev.datatype, predefined);
	if (!predefined)
	{
		MPID_Datatype_get_ptr(req->dev.datatype, dtp);
		req->dev.datatype_ptr = dtp;
		/* this will cause the datatype to be freed when the request is freed. */  
	}

	MPIDI_Pkt_init(get_pkt, MPIDI_CH3_PKT_GET);
	get_pkt->addr = ((MPIDI_Message_uchar_pointer_t) win_ptr->base_addrs[rma_op->target_rank]) +
		win_ptr->disp_units[rma_op->target_rank] * rma_op->target_disp;
	get_pkt->count = rma_op->target_count;
	get_pkt->datatype = rma_op->target_datatype;
	get_pkt->request_handle = req->handle;
	get_pkt->target_win_handle = target_win_handle;
	get_pkt->source_win_handle = source_win_handle;

/*	wprintf(L"send pkt: type %d, addr %d, count %d, base %d\n", rma_pkt->type,
		   rma_pkt->addr, rma_pkt->count, win_ptr->base_addrs[rma_op->target_rank]);
	fflush(stdout);
*/
			
	MPID_Comm_get_ptr(win_ptr->comm, comm_ptr);
	MPIDI_Comm_get_vc(comm_ptr, rma_op->target_rank, &vc);

	MPIDI_CH3I_DATATYPE_IS_PREDEFINED(rma_op->target_datatype, predefined);
	if (predefined)
	{
		/* basic datatype on target. simply send the get_pkt. */
		mpi_errno = MPIDI_CH3.iStartMsg(vc, get_pkt, sizeof(*get_pkt), &req);
	}
	else
	{
		iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)get_pkt;
		iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)sizeof(*get_pkt);

		/* derived datatype on target. fill derived datatype info and
		   send it along with get_pkt. */

		if (win_ptr->pointer_sizes[rma_op->target_rank] != win_ptr->pointer_size)
		{
			*dataloop = NULL;
			SerializeDatatype(rma_op->target_datatype, &serialized_datatype_buffer, &serialized_buffer_length);

			/* Use a negative length to indicate a serialized datatype is being sent */
			get_pkt->dataloop_size = -serialized_buffer_length;
			iov[1].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)serialized_datatype_buffer;
			iov[1].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)serialized_buffer_length;

			mpi_errno = MPIDI_CH3.iStartMsgv(vc, iov, 2, &req);
		}
		else
		{
			MPID_Datatype_get_ptr(rma_op->target_datatype, dtp);
			dtype_info->is_contig = dtp->is_contig;
			dtype_info->n_contig_blocks = dtp->n_contig_blocks;
			dtype_info->size = dtp->size;
			dtype_info->extent = dtp->extent;
			dtype_info->dataloop_size = dtp->dataloop_size;
			dtype_info->dataloop_depth = dtp->dataloop_depth;
			dtype_info->eltype = dtp->eltype;
			dtype_info->dataloop = dtp->dataloop;
			dtype_info->ub = dtp->ub;
			dtype_info->lb = dtp->lb;
			dtype_info->true_ub = dtp->true_ub;
			dtype_info->true_lb = dtp->true_lb;
			dtype_info->has_sticky_ub = dtp->has_sticky_ub;
			dtype_info->has_sticky_lb = dtp->has_sticky_lb;

			MPIU_CHKPMEM_MALLOC(*dataloop, void *, dtp->dataloop_size, mpi_errno, L"dataloop");

			MPIDI_FUNC_ENTER(MPID_STATE_MEMCPY);
			memcpy(*dataloop, dtp->dataloop, dtp->dataloop_size);
			MPIDI_FUNC_EXIT(MPID_STATE_MEMCPY);

			get_pkt->dataloop_size = dtp->dataloop_size;

			iov[1].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)dtype_info;
			iov[1].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)sizeof(*dtype_info);
			iov[2].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)*dataloop;
			iov[2].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)dtp->dataloop_size;

			mpi_errno = MPIDI_CH3.iStartMsgv(vc, iov, 3, &req);

			/* release the target datatype */
			MPID_Datatype_release(dtp);
		}
	}

	/* --BEGIN ERROR HANDLING-- */
	if (mpi_errno != MPI_SUCCESS)
	{
		mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|rmamsg", 0);
		MPIDI_RMA_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_RECV_RMA_MSG);
		return mpi_errno;
	}
	/* --END ERROR HANDLING-- */

	/* release the request returned by iStartMsg or iStartMsgv */
	if (req != NULL)
	{
		MPID_Request_release(req);
	}

 fn_exit:
	MPIDI_RMA_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_RECV_RMA_MSG);
	return mpi_errno;

	/* --BEGIN ERROR HANDLING-- */
 fn_fail:
	MPIU_CHKPMEM_REAP();
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}


#undef FUNCNAME
#define FUNCNAME MPIDI_Win_post
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_Win_post(MPID_Group *group_ptr, int assert, MPID_Win *win_ptr)
{
	int mpi_errno=MPI_SUCCESS;
	MPI_Group win_grp, post_grp;
	int i, post_grp_size, *ranks_in_post_grp, *ranks_in_win_grp, dst, rank;
	MPIU_CHKLMEM_DECL(2);
	MPIU_THREADPRIV_DECL;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_WIN_POST);

	MPIDI_RMA_FUNC_ENTER(MPID_STATE_MPIDI_WIN_POST);

	MPIU_THREADPRIV_GET;
	/* Reset the fence counter so that in case the user has switched from 
	   fence to 
	   post-wait synchronization, he cannot use the previous fence to mark 
	   the beginning of a fence epoch.  */
	win_ptr->fence_cnt = 0;

	/* In case this process was previously the target of passive target rma
	 * operations, we need to take care of the following...
	 * Since we allow MPI_Win_unlock to return without a done ack from
	 * the target in the case of multiple rma ops and exclusive lock,
	 * we need to check whether there is a lock on the window, and if
	 * there is a lock, poke the progress engine until the operations
	 * have completed and the lock is therefore released. */
	if (win_ptr->current_lock_type != MPID_LOCK_NONE)
	{
		MPID_Progress_state progress_state;
		
		/* poke the progress engine */
		MPID_Progress_start(&progress_state);
		while (win_ptr->current_lock_type != MPID_LOCK_NONE)
		{
			mpi_errno = MPID_Progress_wait(&progress_state);
			/* --BEGIN ERROR HANDLING-- */
			if (mpi_errno != MPI_SUCCESS)
			{
				MPID_Progress_end(&progress_state);
				mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", L"**fail %s", L"making progress on the rma messages failed");
				goto fn_exit;
			}
			/* --END ERROR HANDLING-- */
		}
		MPID_Progress_end(&progress_state);
	}
		
	post_grp_size = group_ptr->size;
		
	/* initialize the completion counter */
	win_ptr->my_counter = post_grp_size;
		
	if ((assert & MPI_MODE_NOCHECK) == 0)
	{
		/* NOCHECK not specified. We need to notify the source processes that Post has been called. */  
		
		/* We need to translate the ranks of the processes in post_group to ranks in win_ptr->comm, so that we can do communication */

		MPIU_CHKLMEM_MALLOC(ranks_in_post_grp, int *, post_grp_size * sizeof(int), mpi_errno, L"ranks_in_post_grp");
		MPIU_CHKLMEM_MALLOC(ranks_in_win_grp, int *, post_grp_size * sizeof(int), mpi_errno, L"ranks_in_win_grp");

		for (i=0; i<post_grp_size; i++)
		{
			ranks_in_post_grp[i] = i;
		}

		NMPI_Comm_group(mpi_errno, (win_ptr->comm, &win_grp));
		if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }
		
		post_grp = group_ptr->handle;

		NMPI_Group_translate_ranks(mpi_errno, (post_grp, post_grp_size, ranks_in_post_grp, win_grp, ranks_in_win_grp));
		if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }
		
		NMPI_Comm_rank(mpi_errno, (win_ptr->comm, &rank));
		
		/* Send a 0-byte message to the source processes */
		for (i=0; i<post_grp_size; i++)
		{
			dst = ranks_in_win_grp[i];
			
			if (dst != rank)
			{
				NMPI_Send(mpi_errno, (&i, 0, MPI_INT, dst, 100, win_ptr->comm));
				if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }
			}
		}
		
		NMPI_Group_free(mpi_errno, (&win_grp));
		if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }
	}	

 fn_exit:
	MPIU_CHKLMEM_FREEALL();
	MPIDI_RMA_FUNC_EXIT(MPID_STATE_MPIDI_WIN_POST);
	return mpi_errno;
	/* --BEGIN ERROR HANDLING-- */
 fn_fail:
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}



#undef FUNCNAME
#define FUNCNAME MPIDI_Win_start
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_Win_start(MPID_Group *group_ptr, int assert, MPID_Win *win_ptr)
{
	int mpi_errno=MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_WIN_START);

	MPIDI_RMA_FUNC_ENTER(MPID_STATE_MPIDI_WIN_START);

	/* Reset the fence counter so that in case the user has switched from fence to 
	   start-complete synchronization, he cannot use the previous fence to mark the 
	   beginning of a fence epoch.  */
	win_ptr->fence_cnt = 0;

	/* In case this process was previously the target of passive target rma
	 * operations, we need to take care of the following...
	 * Since we allow MPI_Win_unlock to return without a done ack from
	 * the target in the case of multiple rma ops and exclusive lock,
	 * we need to check whether there is a lock on the window, and if
	 * there is a lock, poke the progress engine until the operations
	 * have completed and the lock is therefore released. */
	if (win_ptr->current_lock_type != MPID_LOCK_NONE)
	{
		MPID_Progress_state progress_state;
		
		/* poke the progress engine */
		MPID_Progress_start(&progress_state);
		while (win_ptr->current_lock_type != MPID_LOCK_NONE)
		{
			mpi_errno = MPID_Progress_wait(&progress_state);
			/* --BEGIN ERROR HANDLING-- */
			if (mpi_errno != MPI_SUCCESS)
			{
				MPID_Progress_end(&progress_state);
				mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", L"**fail %s", L"making progress on the rma messages failed");
				goto fn_exit;
			}
			/* --END ERROR HANDLING-- */
		}
		MPID_Progress_end(&progress_state);
	}
	
	win_ptr->start_group_ptr = group_ptr;
	MPIU_Object_add_ref( group_ptr );
	win_ptr->start_assert = assert;
	
 fn_exit:
	MPIDI_RMA_FUNC_EXIT(MPID_STATE_MPIDI_WIN_START);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_Win_complete
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_Win_complete(MPID_Win *win_ptr)
{
	int mpi_errno = MPI_SUCCESS;
	int comm_size, *nops_to_proc, src, new_total_op_count;
	int i, j, dst, done, total_op_count, *curr_ops_cnt;
	MPIDI_RMA_ops *curr_ptr, *next_ptr;
	MPID_Comm *comm_ptr;
	MPID_Request **requests; /* array of requests */
	MPI_Win source_win_handle, target_win_handle;
	MPIDI_RMA_dtype_info *dtype_infos=NULL;
	void **dataloops=NULL;	/* to store dataloops for each datatype */
	MPI_Group win_grp, start_grp;
	int start_grp_size, *ranks_in_start_grp, *ranks_in_win_grp, rank;
	MPIU_CHKLMEM_DECL(7);
	MPIU_THREADPRIV_DECL;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_WIN_COMPLETE);

	MPIDI_RMA_FUNC_ENTER(MPID_STATE_MPIDI_WIN_COMPLETE);

	MPIU_THREADPRIV_GET;
	MPID_Comm_get_ptr( win_ptr->comm, comm_ptr );
	comm_size = comm_ptr->local_size;
		
	/* Translate the ranks of the processes in start_group to ranks in win_ptr->comm */
	
	start_grp_size = win_ptr->start_group_ptr->size;
		
	MPIU_CHKLMEM_MALLOC(ranks_in_start_grp, int *, start_grp_size*sizeof(int), mpi_errno, L"ranks_in_start_grp");
	MPIU_CHKLMEM_MALLOC(ranks_in_win_grp, int *, start_grp_size*sizeof(int), mpi_errno, L"ranks_in_win_grp");

	for (i=0; i<start_grp_size; i++)
	{
		ranks_in_start_grp[i] = i;
	}

	NMPI_Comm_group(mpi_errno, (win_ptr->comm, &win_grp));
	if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }

	start_grp = win_ptr->start_group_ptr->handle;

	NMPI_Group_translate_ranks(mpi_errno, (start_grp, start_grp_size, ranks_in_start_grp, win_grp, ranks_in_win_grp));
	if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }
		
		
	NMPI_Comm_rank(mpi_errno, (win_ptr->comm, &rank));
	/* If MPI_MODE_NOCHECK was not specified, we need to check if
	   Win_post was called on the target processes. Wait for a 0-byte sync
	   message from each target process */
	if ((win_ptr->start_assert & MPI_MODE_NOCHECK) == 0)
	{
		for (i=0; i<start_grp_size; i++)
		{
			src = ranks_in_win_grp[i];
			if (src != rank)
			{
				NMPI_Recv(mpi_errno, (NULL, 0, MPI_INT, src, 100, win_ptr->comm, MPI_STATUS_IGNORE));
				if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }
			}
		}
	}
		
	/* keep track of no. of ops to each proc. Needed for knowing
	   whether or not to decrement the completion counter. The
	   completion counter is decremented only on the last
	   operation. */
		
	MPIU_CHKLMEM_MALLOC(nops_to_proc, int *, comm_size*sizeof(int), mpi_errno, L"nops_to_proc");
	for (i=0; i<comm_size; i++) nops_to_proc[i] = 0;

	total_op_count = 0;
	curr_ptr = win_ptr->rma_ops_list;
	while (curr_ptr != NULL)
	{
		nops_to_proc[curr_ptr->target_rank]++;
		total_op_count++;
		curr_ptr = curr_ptr->next;
	}
	
	MPIU_CHKLMEM_MALLOC(requests, MPID_Request **, (total_op_count+start_grp_size) * sizeof(MPID_Request*), mpi_errno, L"requests");
	/* We allocate a few extra requests because if there are no RMA
	   ops to a target process, we need to send a 0-byte message just
	   to decrement the completion counter. */
		
	MPIU_CHKLMEM_MALLOC(curr_ops_cnt, int *, comm_size*sizeof(int), mpi_errno, L"curr_ops_cnt");
	for (i=0; i<comm_size; i++) curr_ops_cnt[i] = 0;
	
	if (total_op_count != 0)
	{
		MPIU_CHKLMEM_MALLOC(dtype_infos, MPIDI_RMA_dtype_info *, total_op_count*sizeof(MPIDI_RMA_dtype_info), mpi_errno, L"dtype_infos");
		MPIU_CHKLMEM_MALLOC(dataloops, void **, total_op_count*sizeof(void*), mpi_errno, L"dataloops");
		for (i=0; i<total_op_count; i++) dataloops[i] = NULL;
	}
		
	i = 0;
	curr_ptr = win_ptr->rma_ops_list;
	while (curr_ptr != NULL)
	{
		/* The completion counter at the target is decremented only on 
		   the last RMA operation. We indicate the last operation by 
		   passing the source_win_handle only on the last operation. 
		   Otherwise, we pass NULL */
		if (curr_ops_cnt[curr_ptr->target_rank] == nops_to_proc[curr_ptr->target_rank] - 1) 
			source_win_handle = win_ptr->handle;
		else 
			source_win_handle = MPI_WIN_NULL;
		
		target_win_handle = win_ptr->all_win_handles[curr_ptr->target_rank];
		
		switch (curr_ptr->type)
		{
		case (MPIDI_RMA_PUT):
		case (MPIDI_RMA_ACCUMULATE):
			mpi_errno = MPIDI_CH3I_Send_rma_msg(curr_ptr, win_ptr, source_win_handle, target_win_handle, &dtype_infos[i], &dataloops[i], &requests[i]); 
			if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }
			break;
		case (MPIDI_RMA_GET):
			mpi_errno = MPIDI_CH3I_Recv_rma_msg(curr_ptr, win_ptr, source_win_handle, target_win_handle, &dtype_infos[i], &dataloops[i], &requests[i]);
			if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }
			break;
		default:
			/* --BEGIN ERROR HANDLING-- */
			mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", L"**fail %s", L"invalid RMA operation");
			goto fn_exit;
			/* --END ERROR HANDLING-- */
		}
		i++;
		curr_ops_cnt[curr_ptr->target_rank]++;
		curr_ptr = curr_ptr->next;
	}
		
	/* If the start_group included some processes that did not end up
	   becoming targets of  RMA operations from this process, we need
	   to send a dummy message to those processes just to decrement
	   the completion counter */
		
	j = i;
	new_total_op_count = total_op_count;
	for (i=0; i<start_grp_size; i++)
	{
		dst = ranks_in_win_grp[i];
		if (dst == rank)
		{
			/* FIXME: MT: this has to be done atomically */
			win_ptr->my_counter -= 1;
		}
		else if (nops_to_proc[dst] == 0)
		{
			MPIDI_CH3_Pkt_t upkt;
			MPIDI_CH3_Pkt_put_t *put_pkt = &upkt.put;
			MPIDI_VC_t * vc;
			
			MPIDI_Pkt_init(put_pkt, MPIDI_CH3_PKT_PUT);
			put_pkt->addr = MPIDI_Local_ptr_to_message_pointer_t(NULL);
			put_pkt->count = 0;
			put_pkt->datatype = MPI_INT;
			put_pkt->target_win_handle = win_ptr->all_win_handles[dst];
			put_pkt->source_win_handle = win_ptr->handle;
			
			MPIDI_Comm_get_vc(comm_ptr, dst, &vc);
			
			mpi_errno = MPIDI_CH3.iStartMsg(vc, put_pkt, sizeof(*put_pkt), &requests[j]);
			/* --BEGIN ERROR HANDLING-- */
			if (mpi_errno != MPI_SUCCESS)
			{
				mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|rmamsg", 0);
				goto fn_exit;
			}
			/* --END ERROR HANDLING-- */
			j++;
			new_total_op_count++;
		}
	}
		
	if (new_total_op_count)
	{
		MPID_Progress_state progress_state;
		
		done = 1;
		MPID_Progress_start(&progress_state);
		while (new_total_op_count)
		{
			for (i=0; i<new_total_op_count; i++)
			{
				if (requests[i] != NULL)
				{
					if (*(requests[i]->cc_ptr) != 0)
					{
						done = 0;
						break;
					}
					else
					{
						mpi_errno = requests[i]->status.MPI_ERROR;
						/* --BEGIN ERROR HANDLING-- */
						if (mpi_errno != MPI_SUCCESS)
						{
							MPID_Progress_end(&progress_state);
							mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", 0);
							goto fn_exit;
						}
						/* --END ERROR HANDLING-- */
						MPID_Request_release(requests[i]);
						requests[i] = NULL;
					}
				}
			}
				
			if (done)
			{
				break;
			}
			
			mpi_errno = MPID_Progress_wait(&progress_state);
			done = 1;
		} 
		MPID_Progress_end(&progress_state);
	}
		
	if (total_op_count != 0)
	{
		for (i=0; i<total_op_count; i++)
		{
			if (dataloops[i] != NULL)
			{
				MPIU_Free(dataloops[i]);
			}
		}
	}
		
	/* free MPIDI_RMA_ops_list */
	curr_ptr = win_ptr->rma_ops_list;
	while (curr_ptr != NULL)
	{
		next_ptr = curr_ptr->next;
		MPIU_Free(curr_ptr);
		curr_ptr = next_ptr;
	}
	win_ptr->rma_ops_list = NULL;
	
	NMPI_Group_free(mpi_errno, (&win_grp));
	if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }

	/* free the group stored in window */
	MPIR_Group_release(win_ptr->start_group_ptr);
	win_ptr->start_group_ptr = NULL; 
	
 fn_exit:
	MPIU_CHKLMEM_FREEALL();
	MPIDI_RMA_FUNC_EXIT(MPID_STATE_MPIDI_WIN_COMPLETE);
	return mpi_errno;
	/* --BEGIN ERROR HANDLING-- */
 fn_fail:
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}



#undef FUNCNAME
#define FUNCNAME MPIDI_Win_wait
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_Win_wait(MPID_Win *win_ptr)
{
	int mpi_errno=MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_WIN_WAIT);

	MPIDI_RMA_FUNC_ENTER(MPID_STATE_MPIDI_WIN_WAIT);

	/* wait for all operations from other processes to finish */
	if (win_ptr->my_counter)
	{
		MPID_Progress_state progress_state;
		
		MPID_Progress_start(&progress_state);
		while (win_ptr->my_counter)
		{
			mpi_errno = MPID_Progress_wait(&progress_state);
			/* --BEGIN ERROR HANDLING-- */
			if (mpi_errno != MPI_SUCCESS)
			{
				MPID_Progress_end(&progress_state);
				MPIDI_RMA_FUNC_EXIT(MPID_STATE_MPIDI_WIN_WAIT);
				return mpi_errno;
			}
			/* --END ERROR HANDLING-- */
		}
		MPID_Progress_end(&progress_state);
	} 

	MPIDI_RMA_FUNC_EXIT(MPID_STATE_MPIDI_WIN_WAIT);
	return mpi_errno;
}



#undef FUNCNAME
#define FUNCNAME MPIDI_Win_lock
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_Win_lock(int lock_type, int dest, int assert, MPID_Win *win_ptr)
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_RMA_ops *new_ptr;
	MPID_Comm *comm_ptr;
	MPIU_CHKPMEM_DECL(1);
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_WIN_LOCK);

	MPIDI_RMA_FUNC_ENTER(MPID_STATE_MPIDI_WIN_LOCK);

	MPIU_UNREFERENCED_ARG(assert);

	/* Reset the fence counter so that in case the user has switched from fence to 
	   lock-unlock synchronization, he cannot use the previous fence to mark the beginning 
	   of a fence epoch.  */
	win_ptr->fence_cnt = 0;

	if (dest == MPI_PROC_NULL) goto fn_exit;
		
	MPID_Comm_get_ptr( win_ptr->comm, comm_ptr );
	
	if (dest == comm_ptr->rank)
	{
		/* The target is this process itself. We must block until the lock is acquired. */

		/* poke the progress engine until lock is granted */
		if (MPIDI_CH3I_Try_acquire_win_lock(win_ptr, lock_type) == 0)
		{
			MPID_Progress_state progress_state;
			
			MPID_Progress_start(&progress_state);
			while (MPIDI_CH3I_Try_acquire_win_lock(win_ptr, lock_type) == 0) 
			{
				mpi_errno = MPID_Progress_wait(&progress_state);
				/* --BEGIN ERROR HANDLING-- */
				if (mpi_errno != MPI_SUCCESS)
				{
					MPID_Progress_end(&progress_state);
					mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", L"**fail %s", L"making progress on rma messages failed");
					goto fn_exit;
				}
				/* --END ERROR HANDLING-- */
			}
			MPID_Progress_end(&progress_state);
		}
		/* local lock acquired. local puts, gets, accumulates will be done directly without queueing. */
	}
	else
	{
		/* target is some other process. add the lock request to rma_ops_list */
			
		MPIU_CHKPMEM_MALLOC(new_ptr, MPIDI_RMA_ops *, sizeof(MPIDI_RMA_ops), mpi_errno, L"RMA operation entry");

		win_ptr->rma_ops_list = new_ptr;

		new_ptr->next = NULL;  
		new_ptr->type = MPIDI_RMA_LOCK;
		new_ptr->target_rank = dest;
		new_ptr->lock_type = lock_type;
	}

 fn_exit:
	MPIDI_RMA_FUNC_EXIT(MPID_STATE_MPIDI_WIN_LOCK);
	return mpi_errno;
	/* --BEGIN ERROR HANDLING-- */
 fn_fail:
	MPIU_CHKPMEM_REAP();
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}

#undef FUNCNAME
#define FUNCNAME MPIDI_Win_unlock
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_Win_unlock(int dest, MPID_Win *win_ptr)
{
	int mpi_errno=MPI_SUCCESS;
	int single_op_opt, type_size;
	MPIDI_RMA_ops *rma_op, *curr_op;
	MPID_Comm *comm_ptr;
	MPID_Request *req=NULL; 
	MPIDI_CH3_Pkt_t upkt;
	MPIDI_CH3_Pkt_lock_t *lock_pkt = &upkt.lock;
	MPIDI_VC_t * vc;
	int wait_for_rma_done_pkt = 0, predefined;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_WIN_UNLOCK);
	MPIDI_RMA_FUNC_ENTER(MPID_STATE_MPIDI_WIN_UNLOCK);

	if (dest == MPI_PROC_NULL) goto fn_exit;
		
	MPID_Comm_get_ptr( win_ptr->comm, comm_ptr );
		
	if (dest == comm_ptr->rank)
	{
		/* local lock. release the lock on the window, grant the next one in the queue, and return. */
		mpi_errno = MPIDI_CH3I_Release_lock(win_ptr);
		MPIU_ERR_CHK_POP(mpi_errno);
		mpi_errno = MPID_Progress_poke();
		goto fn_exit;
	}
		
	rma_op = win_ptr->rma_ops_list;
	
	if ( (rma_op == NULL) || (rma_op->type != MPIDI_RMA_LOCK) )
	{
		/* win_lock was not called. return error */
		mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**rmasync", 0 );
		goto fn_exit;
	}
		
	if (rma_op->target_rank != dest)
	{
		/* The target rank is different from the one passed to win_lock! */
		mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**winunlockrank", L"**winunlockrank %d %d", dest, rma_op->target_rank);
		goto fn_exit;
	}
		
	if (rma_op->next == NULL)
	{
		/* only win_lock called, no put/get/acc. Do nothing and return. */
		MPIU_Free(rma_op);
		win_ptr->rma_ops_list = NULL;
		goto fn_exit;
	}
		
	single_op_opt = 0;
		
	if (rma_op->next->next == NULL)
	{
		/* Single put, get, or accumulate between the lock and unlock. If it
		 * is of small size and predefined datatype at the target, we
		 * do an optimization where the lock and the RMA operation are
		 * sent in a single packet. Otherwise, we send a separate lock
		 * request first. */
		
		curr_op = rma_op->next;
		
		MPID_Datatype_get_size_macro(curr_op->origin_datatype, type_size);
		
		MPIDI_CH3I_DATATYPE_IS_PREDEFINED(curr_op->target_datatype, predefined);

		if ( predefined && (type_size * curr_op->origin_count <= MPIDI_CH3I_Eager_max_msg_size) )
		{
			single_op_opt = 1;
			/* Set the lock granted flag to 1 */
			win_ptr->lock_granted = 1;
			if (curr_op->type == MPIDI_RMA_GET)
			{
				mpi_errno = MPIDI_CH3I_Send_lock_get(win_ptr);
				wait_for_rma_done_pkt = 0;
			}
			else
			{
				mpi_errno = MPIDI_CH3I_Send_lock_put_or_acc(win_ptr);
				wait_for_rma_done_pkt = 1;
			}
			if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }
		}
	}
		
	if (single_op_opt == 0)
	{
		/* Send a lock packet over to the target. wait for the lock_granted reply. then do all the RMA ops. */ 

		MPIDI_Pkt_init(lock_pkt, MPIDI_CH3_PKT_LOCK);
		lock_pkt->target_win_handle = win_ptr->all_win_handles[dest];
		lock_pkt->source_win_handle = win_ptr->handle;
		lock_pkt->lock_type = rma_op->lock_type;
		
		MPIDI_Comm_get_vc(comm_ptr, dest, &vc);
		
		/* Set the lock granted flag to 0 */
		win_ptr->lock_granted = 0;
		
		mpi_errno = MPIDI_CH3.iStartMsg(vc, lock_pkt, sizeof(*lock_pkt), &req);
		/* --BEGIN ERROR HANDLING-- */
		if (mpi_errno != MPI_SUCCESS)
		{
			mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", L"**fail %s", L"sending the rma message failed");
			goto fn_exit;
		}
		/* --END ERROR HANDLING-- */
		
		/* release the request returned by iStartMsg */
		if (req != NULL)
		{
			MPID_Request_release(req);
		}
		
		/* After the target grants the lock, it sends a lock_granted
		 * packet. This packet is received in ch3u_handle_recv_pkt.c.
		 * The handler for the packet sets the win_ptr->lock_granted flag to 1. */
		
		/* poke the progress engine until lock_granted flag is set to 1 */
		if (win_ptr->lock_granted == 0)
		{
			MPID_Progress_state progress_state;
			
			MPID_Progress_start(&progress_state);
			while (win_ptr->lock_granted == 0)
			{
				mpi_errno = MPID_Progress_wait(&progress_state);
				/* --BEGIN ERROR HANDLING-- */
				if (mpi_errno != MPI_SUCCESS)
				{
					MPID_Progress_end(&progress_state);
					mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", L"**fail %s", L"making progress on the rma messages failed");
					goto fn_exit;
				}
				/* --END ERROR HANDLING-- */
			}
			MPID_Progress_end(&progress_state);
		}
		
		/* Now do all the RMA operations */
		mpi_errno = MPIDI_CH3I_Do_passive_target_rma(win_ptr, &wait_for_rma_done_pkt);
		if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }
	}
		
	/* If the lock is a shared lock or we have done the single op
	   optimization, we need to wait until the target informs us that
	   all operations are done on the target. */ 
	if (wait_for_rma_done_pkt == 1)
	{
		/* wait until the "pt rma done" packet is received from the target. This packet resets the win_ptr->lock_granted flag back to 0. */

		/* poke the progress engine until lock_granted flag is reset to 0 */
		if (win_ptr->lock_granted != 0)
		{
			MPID_Progress_state progress_state;
			
			MPID_Progress_start(&progress_state);
			while (win_ptr->lock_granted != 0)
			{
				mpi_errno = MPID_Progress_wait(&progress_state);
				/* --BEGIN ERROR HANDLING-- */
				if (mpi_errno != MPI_SUCCESS)
				{
					MPID_Progress_end(&progress_state);
					mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", L"**fail %s", L"making progress on the rma messages failed");
					goto fn_exit;
				}
				/* --END ERROR HANDLING-- */
			}
			MPID_Progress_end(&progress_state);
		}
	}
	else
		win_ptr->lock_granted = 0; 
	
 fn_exit:
	MPIDI_RMA_FUNC_EXIT(MPID_STATE_MPIDI_WIN_UNLOCK);
	return mpi_errno;
	/* --BEGIN ERROR HANDLING-- */
 fn_fail:
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}


#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_Do_passive_target_rma
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int MPIDI_CH3I_Do_passive_target_rma(MPID_Win *win_ptr, int *wait_for_rma_done_pkt)
{
	int mpi_errno = MPI_SUCCESS, comm_size, done, i, nops;
	MPIDI_RMA_ops *curr_ptr, *next_ptr, **curr_ptr_ptr, *tmp_ptr;
	MPID_Comm *comm_ptr;
	MPID_Request **requests=NULL; /* array of requests */
	MPIDI_RMA_dtype_info *dtype_infos=NULL;
	void **dataloops=NULL;	/* to store dataloops for each datatype */
	MPI_Win source_win_handle, target_win_handle;
	MPIU_CHKLMEM_DECL(3);
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_DO_PASSIVE_TARGET_RMA);

	MPIDI_RMA_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_DO_PASSIVE_TARGET_RMA);

	if (win_ptr->rma_ops_list->lock_type == MPI_LOCK_EXCLUSIVE)
	{
		/* exclusive lock. no need to wait for rma done pkt at the end */
		*wait_for_rma_done_pkt = 0;
	}
	else
	{
		/* shared lock. check if any of the rma ops is a get. If so, move it 
		   to the end of the list and do it last, in which case an rma done 
		   pkt is not needed. If there is no get, rma done pkt is needed */

		/* First check whether the last operation is a get. Skip the first op, which is a lock. */

		curr_ptr = win_ptr->rma_ops_list->next;
		while (curr_ptr->next != NULL) 
			curr_ptr = curr_ptr->next;
	
		if (curr_ptr->type == MPIDI_RMA_GET)
		{
			/* last operation is a get. no need to wait for rma done pkt */
			*wait_for_rma_done_pkt = 0;
		}
		else
		{
			/* go through the list and move the first get operation (if there is one) to the end */

			curr_ptr = win_ptr->rma_ops_list->next;
			curr_ptr_ptr = &(win_ptr->rma_ops_list->next);
			
			*wait_for_rma_done_pkt = 1;
			
			while (curr_ptr != NULL)
			{
				if (curr_ptr->type == MPIDI_RMA_GET)
				{
					*wait_for_rma_done_pkt = 0;
					*curr_ptr_ptr = curr_ptr->next;
					tmp_ptr = curr_ptr;
					while (curr_ptr->next != NULL)
						curr_ptr = curr_ptr->next;
					curr_ptr->next = tmp_ptr;
					tmp_ptr->next = NULL;
					break;
				}
				else
				{
					curr_ptr_ptr = &(curr_ptr->next);
					curr_ptr = curr_ptr->next;
				}
			}
		}
	}

	MPID_Comm_get_ptr( win_ptr->comm, comm_ptr );
	comm_size = comm_ptr->local_size;

	/* Ignore the first op in the list because it is a win_lock and do the rest */

	curr_ptr = win_ptr->rma_ops_list->next;
	nops = 0;
	while (curr_ptr != NULL)
	{
		nops++;
		curr_ptr = curr_ptr->next;
	}

	MPIU_CHKLMEM_MALLOC(requests, MPID_Request **, nops*sizeof(MPID_Request*), mpi_errno, L"requests");
	MPIU_CHKLMEM_MALLOC(dtype_infos, MPIDI_RMA_dtype_info *, nops*sizeof(MPIDI_RMA_dtype_info), mpi_errno, L"dtype_infos");
	MPIU_CHKLMEM_MALLOC(dataloops, void **, nops*sizeof(void*), mpi_errno, L"dataloops");

	for (i=0; i<nops; i++)
	{
		dataloops[i] = NULL;
	}
	
	i = 0;
	curr_ptr = win_ptr->rma_ops_list->next;
	target_win_handle = win_ptr->all_win_handles[curr_ptr->target_rank];
	while (curr_ptr != NULL)
	{
		/* To indicate the last RMA operation, we pass the source_win_handle only on the last operation. Otherwise, we pass MPI_WIN_NULL. */
		if (i == nops - 1)
			source_win_handle = win_ptr->handle;
		else 
			source_win_handle = MPI_WIN_NULL;
		
		switch (curr_ptr->type)
		{
		case (MPIDI_RMA_PUT):  /* same as accumulate */
		case (MPIDI_RMA_ACCUMULATE):
			win_ptr->pt_rma_puts_accs[curr_ptr->target_rank]++;
			mpi_errno = MPIDI_CH3I_Send_rma_msg(curr_ptr, win_ptr,
						 source_win_handle, target_win_handle, &dtype_infos[i],
												&dataloops[i], &requests[i]);
			if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }
			break;
		case (MPIDI_RMA_GET):
			mpi_errno = MPIDI_CH3I_Recv_rma_msg(curr_ptr, win_ptr,
						 source_win_handle, target_win_handle, &dtype_infos[i],
												&dataloops[i], &requests[i]);
			if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }
			break;
		default:
			/* --BEGIN ERROR HANDLING-- */
			mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", L"**fail %s", L"invalid RMA operation");
			goto fn_exit;
			/* --END ERROR HANDLING-- */
		}
		i++;
		curr_ptr = curr_ptr->next;
	}
	
	if (nops)
	{
		MPID_Progress_state progress_state;
		
		done = 1;
		MPID_Progress_start(&progress_state);
		while (nops)
		{
			for (i=0; i<nops; i++)
			{
				if (requests[i] != NULL)
				{
					if (*(requests[i]->cc_ptr) != 0)
					{
						done = 0;
						break;
					}
					else
					{
						mpi_errno = requests[i]->status.MPI_ERROR;
						/* --BEGIN ERROR HANDLING-- */
						if (mpi_errno != MPI_SUCCESS)
						{
							MPID_Progress_end(&progress_state);
							mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", L"**fail %s", L"rma message operation failed");
							goto fn_exit;
						}
						/* --END ERROR HANDLING-- */
						/* if origin datatype was a derived datatype, it will get freed when the request gets freed. */ 
						MPID_Request_release(requests[i]);
						requests[i] = NULL;
					}
				}
			}
		
			if (done) 
			{
				break;
			}
		
			mpi_errno = MPID_Progress_wait(&progress_state);
			/* --BEGIN ERROR HANDLING-- */
			if (mpi_errno != MPI_SUCCESS)
			{
				MPID_Progress_end(&progress_state);
				mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", L"**fail %s", L"making progress on the rma messages failed");
				goto fn_exit;
			}
			/* --END ERROR HANDLING-- */
			done = 1;
		}
		MPID_Progress_end(&progress_state);
	} 
	
	for (i=0; i<nops; i++)
	{
		if (dataloops[i] != NULL)
		{
			MPIU_Free(dataloops[i]);
		}
	}
	
	/* free MPIDI_RMA_ops_list */
	curr_ptr = win_ptr->rma_ops_list;
	while (curr_ptr != NULL)
	{
		next_ptr = curr_ptr->next;
		MPIU_Free(curr_ptr);
		curr_ptr = next_ptr;
	}
	win_ptr->rma_ops_list = NULL;

 fn_exit:
	MPIU_CHKLMEM_FREEALL();
	MPIDI_RMA_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_DO_PASSIVE_TARGET_RMA);
	return mpi_errno;
	/* --BEGIN ERROR HANDLING-- */
 fn_fail:
	goto fn_exit;
	/* --END ERROR HANDLING-- */
}


#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_Send_lock_put_or_acc
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int MPIDI_CH3I_Send_lock_put_or_acc(MPID_Win *win_ptr)
{
	int mpi_errno=MPI_SUCCESS, lock_type, origin_dt_derived, iov_n, iovcnt;
	MPIDI_RMA_ops *rma_op;
	MPID_Request *request=NULL;
	MPIDI_VC_t * vc;
	MPID_IOV iov[MPID_IOV_LIMIT];
	MPID_Comm *comm_ptr;
	MPID_Datatype *origin_dtp=NULL;
	int origin_type_size, predefined;
	MPIDI_CH3_Pkt_t upkt;
	MPIDI_CH3_Pkt_lock_put_unlock_t *lock_put_unlock_pkt = &upkt.lock_put_unlock;
	MPIDI_CH3_Pkt_lock_accum_unlock_t *lock_accum_unlock_pkt = &upkt.lock_accum_unlock;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_SEND_LOCK_PUT_OR_ACC);

	MPIDI_RMA_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_SEND_LOCK_PUT_OR_ACC);

	lock_type = win_ptr->rma_ops_list->lock_type;

	rma_op = win_ptr->rma_ops_list->next;

	win_ptr->pt_rma_puts_accs[rma_op->target_rank]++;

	if (rma_op->type == MPIDI_RMA_PUT)
	{
		MPIDI_Pkt_init(lock_put_unlock_pkt, MPIDI_CH3_PKT_LOCK_PUT_UNLOCK);
		lock_put_unlock_pkt->target_win_handle = win_ptr->all_win_handles[rma_op->target_rank];
		lock_put_unlock_pkt->source_win_handle = win_ptr->handle;
		lock_put_unlock_pkt->lock_type = lock_type;
 
		lock_put_unlock_pkt->addr = 
			((MPIDI_Message_uchar_pointer_t) win_ptr->base_addrs[rma_op->target_rank]) +
			win_ptr->disp_units[rma_op->target_rank] * rma_op->target_disp;
		
		lock_put_unlock_pkt->count = rma_op->target_count;
		lock_put_unlock_pkt->datatype = rma_op->target_datatype;

		iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST) lock_put_unlock_pkt;
		iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)sizeof(*lock_put_unlock_pkt);
	}
	else if (rma_op->type == MPIDI_RMA_ACCUMULATE)
	{
		MPIDI_Pkt_init(lock_accum_unlock_pkt, MPIDI_CH3_PKT_LOCK_ACCUM_UNLOCK);
		lock_accum_unlock_pkt->target_win_handle = win_ptr->all_win_handles[rma_op->target_rank];
		lock_accum_unlock_pkt->source_win_handle = win_ptr->handle;
		lock_accum_unlock_pkt->lock_type = lock_type;

		lock_accum_unlock_pkt->addr = 
			((MPIDI_Message_uchar_pointer_t) win_ptr->base_addrs[rma_op->target_rank]) +
			win_ptr->disp_units[rma_op->target_rank] * rma_op->target_disp;
		
		lock_accum_unlock_pkt->count = rma_op->target_count;
		lock_accum_unlock_pkt->datatype = rma_op->target_datatype;
		lock_accum_unlock_pkt->op = rma_op->op;

		iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST) lock_accum_unlock_pkt;
		iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)sizeof(*lock_accum_unlock_pkt);
	}

	MPID_Comm_get_ptr(win_ptr->comm, comm_ptr);
	MPIDI_Comm_get_vc(comm_ptr, rma_op->target_rank, &vc);

	MPIDI_CH3I_DATATYPE_IS_PREDEFINED(rma_op->origin_datatype, predefined);
	if (!predefined)
	{
		origin_dt_derived = 1;
		MPID_Datatype_get_ptr(rma_op->origin_datatype, origin_dtp);
	}
	else
	{
		origin_dt_derived = 0;
	}

	MPID_Datatype_get_size_macro(rma_op->origin_datatype, origin_type_size);

	if (!origin_dt_derived)
	{
		/* basic datatype on origin */

		iov[1].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)rma_op->origin_addr;
		iov[1].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)(rma_op->origin_count * origin_type_size);
		iovcnt = 2;

		mpi_errno = MPIDI_CH3.iStartMsgv(vc, iov, iovcnt, &request);
		/* --BEGIN ERROR HANDLING-- */
		if (mpi_errno != MPI_SUCCESS)
		{
			mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|rmamsg", 0);
			goto fn_exit;
		}
		/* --END ERROR HANDLING-- */
	}
	else
	{
		/* derived datatype on origin */

		iovcnt = 1;

		request = MPID_Request_create();
		if (request == NULL)
		{
			/* --BEGIN ERROR HANDLING-- */
			mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
			goto fn_exit;
			/* --END ERROR HANDLING-- */
		}

		MPIU_Object_set_ref(request, 2);
		request->kind = MPID_REQUEST_SEND;
			
		request->dev.datatype_ptr = origin_dtp;
		/* this will cause the datatype to be freed when the request is freed. */ 

		MPID_Segment_init(rma_op->origin_addr, rma_op->origin_count, rma_op->origin_datatype, &(request->dev.segment), 0);
		request->dev.segment_first = 0;
		request->dev.segment_size = rma_op->origin_count * origin_type_size;
			
		iov_n = MPID_IOV_LIMIT - iovcnt;
		mpi_errno = MPIDI_CH3U_Request_load_send_iov(request, &iov[iovcnt], &iov_n); 
		if (mpi_errno == MPI_SUCCESS)
		{
			iov_n += iovcnt;
			
			mpi_errno = MPIDI_CH3.iSendv(vc, request, iov, iov_n);
			/* --BEGIN ERROR HANDLING-- */
			if (mpi_errno != MPI_SUCCESS)
			{
				MPID_Datatype_release(request->dev.datatype_ptr);
				MPIU_Object_set_ref(request, 0);
				MPIDI_CH3_Request_destroy(request);
				mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|rmamsg", 0);
				goto fn_exit;
			}
			/* --END ERROR HANDLING-- */
		}
		/* --BEGIN ERROR HANDLING-- */
		else
		{
			MPID_Datatype_release(request->dev.datatype_ptr);
			MPIU_Object_set_ref(request, 0);
			MPIDI_CH3_Request_destroy(request);
			mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|loadsendiov", 0);
			goto fn_exit;
		}
		/* --END ERROR HANDLING-- */
	}

	if (request != NULL)
	{
		if (*(request->cc_ptr) != 0)
		{
			MPID_Progress_state progress_state;
			
			MPID_Progress_start(&progress_state);
			while (*(request->cc_ptr) != 0)
			{
				mpi_errno = MPID_Progress_wait(&progress_state);
				/* --BEGIN ERROR HANDLING-- */
				if (mpi_errno != MPI_SUCCESS)
				{
					MPID_Progress_end(&progress_state);
					mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", L"**fail %s", L"rma message operation failed");
					goto fn_exit;
				}
				/* --END ERROR HANDLING-- */
			}
			MPID_Progress_end(&progress_state);
		}
		
		mpi_errno = request->status.MPI_ERROR;
		/* --BEGIN ERROR HANDLING-- */
		if (mpi_errno != MPI_SUCCESS)
		{
			mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", L"**fail %s", L"rma message operation failed");
			goto fn_exit;
		}
		/* --END ERROR HANDLING-- */
				
		/* if origin datatype was a derived datatype, it will get freed when the request gets freed. */ 
		MPID_Request_release(request);
	}

	/* free MPIDI_RMA_ops_list */
	MPIU_Free(win_ptr->rma_ops_list->next);
	MPIU_Free(win_ptr->rma_ops_list);
	win_ptr->rma_ops_list = NULL;

 fn_exit:
	MPIDI_RMA_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SEND_LOCK_PUT_OR_ACC);
	return mpi_errno;
}


#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_Send_lock_get
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static int MPIDI_CH3I_Send_lock_get(MPID_Win *win_ptr)
{
	int mpi_errno=MPI_SUCCESS, lock_type, predefined;
	MPIDI_RMA_ops *rma_op;
	MPID_Request *rreq=NULL, *sreq=NULL;
	MPIDI_VC_t * vc;
	MPID_IOV iov[MPID_IOV_LIMIT];
	MPID_Comm *comm_ptr;
	MPID_Datatype *dtp;
	MPIDI_CH3_Pkt_t upkt;
	MPIDI_CH3_Pkt_lock_get_unlock_t *lock_get_unlock_pkt = &upkt.lock_get_unlock;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_SEND_LOCK_GET);

	MPIDI_RMA_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_SEND_LOCK_GET);

	lock_type = win_ptr->rma_ops_list->lock_type;

	rma_op = win_ptr->rma_ops_list->next;

	/* create a request, store the origin buf, cnt, datatype in it,
	   and pass a handle to it in the get packet. When the get
	   response comes from the target, it will contain the request
	   handle. */  
	rreq = MPID_Request_create();
	if (rreq == NULL)
	{
		/* --BEGIN ERROR HANDLING-- */
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
		goto fn_exit;
		/* --END ERROR HANDLING-- */
	}

	MPIU_Object_set_ref(rreq, 2);

	rreq->dev.user_buf = rma_op->origin_addr;
	rreq->dev.user_count = rma_op->origin_count;
	rreq->dev.datatype = rma_op->origin_datatype;
	rreq->dev.target_win_handle = MPI_WIN_NULL;
	rreq->dev.source_win_handle = win_ptr->handle;

	MPIDI_CH3I_DATATYPE_IS_PREDEFINED(rreq->dev.datatype, predefined);
	if (!predefined)
	{
		MPID_Datatype_get_ptr(rreq->dev.datatype, dtp);
		rreq->dev.datatype_ptr = dtp;
		/* this will cause the datatype to be freed when the request is freed. */  
	}

	MPIDI_Pkt_init(lock_get_unlock_pkt, MPIDI_CH3_PKT_LOCK_GET_UNLOCK);
	lock_get_unlock_pkt->target_win_handle = win_ptr->all_win_handles[rma_op->target_rank];
	lock_get_unlock_pkt->source_win_handle = win_ptr->handle;
	lock_get_unlock_pkt->lock_type = lock_type;
 
	lock_get_unlock_pkt->addr = 
		((MPIDI_Message_uchar_pointer_t) win_ptr->base_addrs[rma_op->target_rank]) +
		win_ptr->disp_units[rma_op->target_rank] * rma_op->target_disp;
		
	lock_get_unlock_pkt->count = rma_op->target_count;
	lock_get_unlock_pkt->datatype = rma_op->target_datatype;
	lock_get_unlock_pkt->request_handle = rreq->handle;

	iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST) lock_get_unlock_pkt;
	iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)sizeof(*lock_get_unlock_pkt);

	MPID_Comm_get_ptr(win_ptr->comm, comm_ptr);
	MPIDI_Comm_get_vc(comm_ptr, rma_op->target_rank, &vc);

	mpi_errno = MPIDI_CH3.iStartMsg(vc, lock_get_unlock_pkt, sizeof(*lock_get_unlock_pkt), &sreq);
	if (mpi_errno != MPI_SUCCESS)
	{
	 /* --BEGIN ERROR HANDLING-- */
		mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|rmamsg", 0);
		goto fn_exit;
	/* --END ERROR HANDLING-- */
	}

	/* release the request returned by iStartMsg */
	if (sreq != NULL)
	{
		MPID_Request_release(sreq);
	}

	/* now wait for the data to arrive */
	if (*(rreq->cc_ptr) != 0)
	{
		MPID_Progress_state progress_state;
		
		MPID_Progress_start(&progress_state);
		while (*(rreq->cc_ptr) != 0)
		{
			mpi_errno = MPID_Progress_wait(&progress_state);
			/* --BEGIN ERROR HANDLING-- */
			if (mpi_errno != MPI_SUCCESS)
			{
				MPID_Progress_end(&progress_state);
				mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", L"**fail %s", L"rma message operation failed");
				goto fn_exit;
			}
			/* --END ERROR HANDLING-- */
		}
		MPID_Progress_end(&progress_state);
	}
	
	mpi_errno = rreq->status.MPI_ERROR;
	/* --BEGIN ERROR HANDLING-- */
	if (mpi_errno != MPI_SUCCESS)
	{
		mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", L"**fail %s", L"rma message operation failed");
		goto fn_exit;
	}
	/* --END ERROR HANDLING-- */
			
	/* if origin datatype was a derived datatype, it will get freed when the rreq gets freed. */ 
	MPID_Request_release(rreq);

	/* free MPIDI_RMA_ops_list */
	MPIU_Free(win_ptr->rma_ops_list->next);
	MPIU_Free(win_ptr->rma_ops_list);
	win_ptr->rma_ops_list = NULL;

 fn_exit:
	MPIDI_RMA_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SEND_LOCK_GET);
	return mpi_errno;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3U_Buffer_copy
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
void MPIDI_CH3U_Buffer_copy(
	const void * const sbuf, int scount, MPI_Datatype sdt, int * smpi_errno,
	void * const rbuf, int rcount, MPI_Datatype rdt, MPIDI_msg_sz_t * rsz,
	int * rmpi_errno)
{
	int sdt_contig;
	int rdt_contig;
	MPI_Aint sdt_true_lb, rdt_true_lb;
	MPIDI_msg_sz_t sdata_sz;
	MPIDI_msg_sz_t rdata_sz;
	MPID_Datatype * sdt_ptr;
	MPID_Datatype * rdt_ptr;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3U_BUFFER_COPY);
	MPIDI_STATE_DECL(MPID_STATE_MEMCPY);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3U_BUFFER_COPY);
	*smpi_errno = MPI_SUCCESS;
	*rmpi_errno = MPI_SUCCESS;

	MPIDI_Datatype_get_info(scount, sdt, sdt_contig, sdata_sz, sdt_ptr, sdt_true_lb);
	MPIDI_Datatype_get_info(rcount, rdt, rdt_contig, rdata_sz, rdt_ptr, rdt_true_lb);

	/* --BEGIN ERROR HANDLING-- */
	if (sdata_sz > rdata_sz)
	{
		MPIU_DBG_MSG_FMT(CH3_OTHER,TYPICAL,(MPIU_DBG_FDEST, L"message truncated, sdata_sz=" MPIDI_MSG_SZ_FMT L" rdata_sz=" MPIDI_MSG_SZ_FMT, sdata_sz, rdata_sz));
		sdata_sz = rdata_sz;
		*rmpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_TRUNCATE, L"**truncate", L"**truncate %d %d", sdata_sz, rdata_sz );
	}
	/* --END ERROR HANDLING-- */
	
	if (sdata_sz == 0)
	{
		*rsz = 0;
		goto fn_exit;
	}
	
	if (sdt_contig && rdt_contig)
	{
		MPIDI_FUNC_ENTER(MPID_STATE_MEMCPY);
		memcpy((char *)rbuf + rdt_true_lb, (const char *)sbuf + sdt_true_lb, sdata_sz);
		MPIDI_FUNC_EXIT(MPID_STATE_MEMCPY);
		*rsz = sdata_sz;
	}
	else if (sdt_contig)
	{
		MPID_Segment seg;
		MPIDI_msg_sz_t last;

		MPID_Segment_init(rbuf, rcount, rdt, &seg, 0);
		last = sdata_sz;
		MPIU_DBG_MSG_FMT(CH3_OTHER,VERBOSE,(MPIU_DBG_FDEST, L"pre-unpack last=" MPIDI_MSG_SZ_FMT, last ));
		MPID_Segment_unpack(&seg, 0, &last, (char*)sbuf + sdt_true_lb);
		MPIU_DBG_MSG_FMT(CH3_OTHER,VERBOSE,(MPIU_DBG_FDEST, L"pre-unpack last=" MPIDI_MSG_SZ_FMT, last ));
		/* --BEGIN ERROR HANDLING-- */
		if (last != sdata_sz)
		{
			*rmpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_TYPE, L"**dtypemismatch", 0);
		}
		/* --END ERROR HANDLING-- */

		*rsz = last;
	}
	else if (rdt_contig)
	{
		MPID_Segment seg;
		MPIDI_msg_sz_t last;

		MPID_Segment_init(sbuf, scount, sdt, &seg, 0);
		last = sdata_sz;
		MPIU_DBG_MSG_FMT(CH3_OTHER,VERBOSE,(MPIU_DBG_FDEST, L"pre-pack last=" MPIDI_MSG_SZ_FMT, last ));
		MPID_Segment_pack(&seg, 0, &last, (char*)rbuf + rdt_true_lb);
		MPIU_DBG_MSG_FMT(CH3_OTHER,VERBOSE,(MPIU_DBG_FDEST, L"post-pack last=" MPIDI_MSG_SZ_FMT, last ));
		/* --BEGIN ERROR HANDLING-- */
		if (last != sdata_sz)
		{
			*rmpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_TYPE, L"**dtypemismatch", 0);
		}
		/* --END ERROR HANDLING-- */

		*rsz = last;
	}
	else
	{
		char * buf;
		MPIDI_msg_sz_t buf_off;
		MPID_Segment sseg;
		MPIDI_msg_sz_t sfirst;
		MPID_Segment rseg;
		MPIDI_msg_sz_t rfirst;

		buf = MPIU_Malloc(MPIDI_COPY_BUFFER_SZ);
		/* --BEGIN ERROR HANDLING-- */
		if (buf == NULL)
		{
			MPIU_DBG_MSG(CH3_OTHER,TYPICAL,L"SRBuf allocation failure");
			*smpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
			*rmpi_errno = *smpi_errno;
			*rsz = 0;
			goto fn_exit;
		}
		/* --END ERROR HANDLING-- */

		MPID_Segment_init(sbuf, scount, sdt, &sseg, 0);
		MPID_Segment_init(rbuf, rcount, rdt, &rseg, 0);

		sfirst = 0;
		rfirst = 0;
		buf_off = 0;
		
		for(;;)
		{
			MPIDI_msg_sz_t last;
			char * buf_end;

			if (sdata_sz - sfirst > MPIDI_COPY_BUFFER_SZ - buf_off)
			{
				last = sfirst + (MPIDI_COPY_BUFFER_SZ - buf_off);
			}
			else
			{
				last = sdata_sz;
			}
			
			MPIU_DBG_MSG_FMT(CH3_OTHER,VERBOSE,(MPIU_DBG_FDEST, L"pre-pack first=" MPIDI_MSG_SZ_FMT L", last=" MPIDI_MSG_SZ_FMT, sfirst, last ));
			MPID_Segment_pack(&sseg, sfirst, &last, buf + buf_off);
			MPIU_DBG_MSG_FMT(CH3_OTHER,VERBOSE,(MPIU_DBG_FDEST, L"post-pack first=" MPIDI_MSG_SZ_FMT L", last=" MPIDI_MSG_SZ_FMT, sfirst, last ));
			/* --BEGIN ERROR HANDLING-- */
			MPIU_Assert(last > sfirst);
			/* --END ERROR HANDLING-- */
			
			buf_end = buf + buf_off + (last - sfirst);
			sfirst = last;
			
			MPIU_DBG_MSG_FMT(CH3_OTHER,VERBOSE,(MPIU_DBG_FDEST, L"pre-unpack first=" MPIDI_MSG_SZ_FMT L", last=" MPIDI_MSG_SZ_FMT, rfirst, last ));
			MPID_Segment_unpack(&rseg, rfirst, &last, buf);
			MPIU_DBG_MSG_FMT(CH3_OTHER,VERBOSE,(MPIU_DBG_FDEST, L"post-unpack first=" MPIDI_MSG_SZ_FMT L", last=" MPIDI_MSG_SZ_FMT, rfirst, last ));
			/* --BEGIN ERROR HANDLING-- */
			MPIU_Assert(last > rfirst);
			/* --END ERROR HANDLING-- */

			rfirst = last;

			if (rfirst == sdata_sz)
			{
				/* successful completion */
				break;
			}

			/* --BEGIN ERROR HANDLING-- */
			if (sfirst == sdata_sz)
			{
				/* datatype mismatch -- remaining bytes could not be unpacked */
				*rmpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_TYPE, L"**dtypemismatch", 0);
				break;
			}
			/* --END ERROR HANDLING-- */

			buf_off = sfirst - rfirst;
			if (buf_off > 0)
			{
				MPIU_DBG_MSG_FMT(CH3_OTHER, VERBOSE, (MPIU_DBG_FDEST, L"moved " MPIDI_MSG_SZ_FMT L" bytes to the beginning of the tmp buffer", buf_off));
				memmove(buf, buf_end - buf_off, buf_off);
			}
		}

		*rsz = rfirst;
		MPIU_Free(buf);
	}

  fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3U_BUFFER_COPY);
}

/*
 * MPIDI_CH3I_Request_adjust_iov()
 *
 * Adjust the iovec in the request by the supplied number of bytes.  If the iovec has been consumed, return true; otherwise return
 * false.
 */
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3U_Request_adjust_iov
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
inline int MPIDI_CH3I_Request_adjust_iov(MPID_Request * req, MPIDI_msg_sz_t nb)
{
	int offset = req->ch.iov_offset;
	const int count = req->dev.iov_count;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_REQUEST_ADJUST_IOV);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_REQUEST_ADJUST_IOV);
	
	while (offset < count)
	{
		if (req->dev.iov[offset].MPID_IOV_LEN <= (MPID_IOV_LEN_CAST)nb)
		{
			nb -= (MPIDI_msg_sz_t)req->dev.iov[offset].MPID_IOV_LEN;
			offset++;
		}
		else
		{
			req->dev.iov[offset].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)((char *) req->dev.iov[offset].MPID_IOV_BUF + nb);
			req->dev.iov[offset].MPID_IOV_LEN -= (MPID_IOV_LEN_CAST)nb;
			req->ch.iov_offset = offset;
			MPIDI_DBG_PRINTF((60, FCNAME, L"adjust_iov returning FALSE"));
			MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_REQUEST_ADJUST_IOV);
			return FALSE;
		}
	}
	
	req->ch.iov_offset = 0;

	MPIDI_DBG_PRINTF((60, FCNAME, L"adjust_iov returning TRUE"));
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_REQUEST_ADJUST_IOV);
	return TRUE;
}

/*@
   MPIDI_CH3I_SHM_Get_memblock - allocate and get the address and size of a shared memory block

   Parameters:
+  int size - size
-  MPIDI_CH3I_Shmem_block_request_result* pOutput - output

   Notes:
@*/
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_SHM_Get_memblock
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3I_SHM_Get_memblock(int size, MPIDI_CH3I_Shmem_block_request_result *pOutput)
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_SHM_GET_MEM);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_SHM_GET_MEM);

	if (size == 0 || size > MPIDU_MAX_SHM_BLOCK_SIZE )
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**arg", 0);
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_GET_MEM);
		return mpi_errno;
	}

	/* Create the shared memory object */
	MPIDI_CH3U_Generate_shm_string(pOutput->key);
	pOutput->id = CreateFileMapping( INVALID_HANDLE_VALUE, NULL, PAGE_READWRITE, 0, size, pOutput->key);
	if (pOutput->id == NULL) 
	{
		pOutput->error = GetLastError();
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**CreateFileMapping", L"**CreateFileMapping %d", pOutput->error); /*L"Error in CreateFileMapping, %d", pOutput->error);*/
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_GET_MEM);
		return mpi_errno;
	}

	pOutput->addr = NULL;
	pOutput->addr = MapViewOfFileEx( pOutput->id, FILE_MAP_WRITE, 0, 0, size, NULL );
	if (pOutput->addr == NULL)
	{
		pOutput->error = GetLastError();
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**MapViewOfFileEx", L"**MapViewOfFileEx %d", pOutput->error);
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_GET_MEM);
		return mpi_errno;
	}

	pOutput->size = size;
	pOutput->error = MPI_SUCCESS;

	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_GET_MEM);
	return MPI_SUCCESS;
}


/*@
   MPIDI_CH3I_SHM_Unlink_memblock - 

   Notes:
@*/
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_SHM_Unlink_memblock
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3I_SHM_Unlink_memblock(MPIDI_CH3I_Shmem_block_request_result *p)
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_SHM_UNLINK_MEM);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_SHM_UNLINK_MEM);

	/* Windows requires no special handling */
	MPIU_UNREFERENCED_ARG(p);

	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_UNLINK_MEM);
	return mpi_errno;
}

/*@
   MPIDI_CH3I_SHM_Release_memblock - 

   Notes:
@*/
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_SHM_Release_memblock
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3I_SHM_Release_memblock(MPIDI_CH3I_Shmem_block_request_result *p)
{
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_SHM_RELEASE_MEM);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_SHM_RELEASE_MEM);

	UnmapViewOfFile(p->addr);
	CloseHandle(p->id);

	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_RELEASE_MEM);
	return MPI_SUCCESS;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_SHM_Remove_vc_read_references
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
static void MPIDI_CH3I_SHM_Remove_vc_read_references(MPIDI_VC_t *vc)
{
	MPIDI_VC_t *iter, *trailer;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_SHM_REMOVE_VC_READ_REFERENCES);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_SHM_REMOVE_VC_READ_REFERENCES);

	/* remove vc from the reading list */
	iter = trailer = MPIDI_CH3I_Process.shm_reading_list;
	while (iter != NULL)
	{
		if (iter == vc)
		{
			/* Mark the connection as NOT read connected so it won't be mistaken for active */
			/* Since shm connections are uni-directional the following three states are considered active:
			 * MPIDI_VC_STATE_ACTIVE + ch.shm_read_connected = 0 - active in the write direction
			 * MPIDI_VC_STATE_INACTIVE + ch.shm_read_connected = 1 - active in the read direction
			 * MPIDI_VC_STATE_ACTIVE + ch.shm_read_connected = 1 - active in both directions
			 */
			vc->ch.shm_read_connected = 0;
			if (trailer != iter)
			{
				/* remove the vc from the list */
				trailer->ch.shm_next_reader = iter->ch.shm_next_reader;
			}
			else
			{
				/* remove the vc from the head of the list */
				MPIDI_CH3I_Process.shm_reading_list = MPIDI_CH3I_Process.shm_reading_list->ch.shm_next_reader;
			}
		}
		if (trailer != iter)
			trailer = trailer->ch.shm_next_reader;
		iter = iter->ch.shm_next_reader;
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_REMOVE_VC_READ_REFERENCES);
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_SHM_Add_to_reader_list
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
void MPIDI_CH3I_SHM_Add_to_reader_list(MPIDI_VC_t *vc)
{
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_SHM_ADD_TO_READER_LIST);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_SHM_ADD_TO_READER_LIST);
	MPIDI_CH3I_SHM_Remove_vc_read_references(vc);
	/*wprintf(L"[%s%d]adding read vc%d.state = %s (%s)\n", MPIU_DBG_parent_str, MPIR_Process.comm_world->rank, vc->pg_rank, VCState(vc), (vc->pg && vc->pg_rank >= 0) ? vc->pg->id : L"?");fflush(stdout);*/
	vc->ch.shm_next_reader = MPIDI_CH3I_Process.shm_reading_list;
	MPIDI_CH3I_Process.shm_reading_list = vc;
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_ADD_TO_READER_LIST);
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_SHM_Remove_vc_write_references
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
void MPIDI_CH3I_SHM_Remove_vc_write_references(MPIDI_VC_t *vc)
{
	MPIDI_VC_t *iter, *trailer;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_SHM_REMOVE_VC_WRITE_REFERENCES);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_SHM_REMOVE_VC_WRITE_REFERENCES);

	/* remove the vc from the writing list */
	iter = trailer = MPIDI_CH3I_Process.shm_writing_list;
	while (iter != NULL)
	{
		if (iter == vc)
		{
			/*wprintf(L"[%s%d]removing write vc%d.state = %s (%s)\n", MPIU_DBG_parent_str, MPIR_Process.comm_world->rank, vc->pg_rank, VCState(vc), (vc->pg && vc->pg_rank >= 0) ? vc->pg->id : L"?");fflush(stdout);*/
			if (trailer != iter)
			{
				/* remove the vc from the list */
				trailer->ch.shm_next_writer = iter->ch.shm_next_writer;
			}
			else
			{
				/* remove the vc from the head of the list */
				MPIDI_CH3I_Process.shm_writing_list = MPIDI_CH3I_Process.shm_writing_list->ch.shm_next_writer;
			}
		}
		if (trailer != iter)
			trailer = trailer->ch.shm_next_writer;
		iter = iter->ch.shm_next_writer;
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_REMOVE_VC_WRITE_REFERENCES);
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_SHM_Remove_vc_references
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
void MPIDI_CH3I_SHM_Remove_vc_references(MPIDI_VC_t *vc)
{
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_SHM_REMOVE_VC_REFERENCES);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_SHM_REMOVE_VC_REFERENCES);
	MPIDI_CH3I_SHM_Remove_vc_read_references(vc);
	MPIDI_CH3I_SHM_Remove_vc_write_references(vc);
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_REMOVE_VC_REFERENCES);
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_SHM_Add_to_writer_list
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
void MPIDI_CH3I_SHM_Add_to_writer_list(MPIDI_VC_t *vc)
{
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_SHM_ADD_TO_WRITER_LIST);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_SHM_ADD_TO_WRITER_LIST);
	MPIDI_CH3I_SHM_Remove_vc_write_references(vc);
	/*wprintf(L"[%s%d]adding write vc%d.state = %s (%s)\n", MPIU_DBG_parent_str, MPIR_Process.comm_world->rank, vc->pg_rank, VCState(vc), (vc->pg && vc->pg_rank >= 0) ? vc->pg->id : L"?");fflush(stdout);*/
	vc->ch.shm_next_writer = MPIDI_CH3I_Process.shm_writing_list;
	MPIDI_CH3I_Process.shm_writing_list = vc;
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_ADD_TO_WRITER_LIST);
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_SHM_write
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3I_SHM_write(MPIDI_VC_t * vc, void *buf, int len, int *num_bytes_ptr)
{
	int total = 0;
	int length;
	int index;
	MPIDI_CH3I_SHM_Queue_t * writeq;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_SHM_WRITE);
	MPIDI_STATE_DECL(MPID_STATE_MEMCPY);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_SHM_WRITE);
	MPIDI_DBG_PRINTF((60, FCNAME, L"entering"));

	writeq = vc->ch.write_shmq;
	index = writeq->tail_index;
	if (writeq->packet[index].avail == MPIDI_CH3I_PKT_FILLED)
	{
		*num_bytes_ptr = total;
		MPIDI_DBG_PRINTF((60, FCNAME, L"exiting"));
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_WRITE);
		return MPI_SUCCESS;
	}
	
	while (len)
	{
		length = min(len, MPIDI_CH3I_PACKET_SIZE);
		/*writeq->packet[index].offset = 0; the reader guarantees this is reset to zero */
		writeq->packet[index].num_bytes = length;
		MPIDI_FUNC_ENTER(MPID_STATE_MEMCPY);
		memcpy(writeq->packet[index].data, buf, length);
		MPIDI_FUNC_EXIT(MPID_STATE_MEMCPY);
		MPID_WRITE_BARRIER();
		writeq->packet[index].avail = MPIDI_CH3I_PKT_FILLED;
		buf = (char *) buf + length;
		total += length;
		len -= length;

		index = (index + 1) % MPIDI_CH3I_NUM_PACKETS;
		if (writeq->packet[index].avail == MPIDI_CH3I_PKT_FILLED)
		{
			writeq->tail_index = index;
			*num_bytes_ptr = total;
			MPIDI_DBG_PRINTF((60, FCNAME, L"exiting"));
			MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_WRITE);
			return MPI_SUCCESS;
		}
	}

	writeq->tail_index = index;
	*num_bytes_ptr = total;
	MPIDI_DBG_PRINTF((60, FCNAME, L"exiting"));
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_WRITE);
	return MPI_SUCCESS;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_SHM_write_progress
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3I_SHM_write_progress(MPIDI_VC_t * vc)
{
	int mpi_errno;
	int nb;
	int total = 0;
	int complete;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_SHM_WRITE_PROGRESS);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_SHM_WRITE_PROGRESS);
	
	MPIDI_DBG_PRINTF((60, FCNAME, L"entering"));
	while (vc->ch.send_active != NULL)
	{
		MPID_Request * req = vc->ch.send_active;

#ifdef MPICH_DBG_OUTPUT
		if (req->ch.iov_offset >= req->dev.iov_count)
		{
			mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**iov_offset", L"**iov_offset %d %d", req->ch.iov_offset, req->dev.iov_count);
			MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_WRITE_PROGRESS);
			return mpi_errno;
		}
#endif
		/* Check here or inside shm_writev?
		if (vc->ch.write_shmq->packet[vc->ch.write_shmq->tail_index].avail == MPIDI_CH3I_PKT_EMPTY)
			mpi_errno = MPIDI_CH3I_SHM_writev(vc, req->dev.iov + req->ch.iov_offset, req->dev.iov_count - req->ch.iov_offset, &nb);
		else
			nb = 0;
		*/
		mpi_errno = MPIDI_CH3I_SHM_writev(vc, req->dev.iov + req->ch.iov_offset, req->dev.iov_count - req->ch.iov_offset, &nb);
		if (mpi_errno != MPI_SUCCESS)
		{
			mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", 0);
			MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_WRITE_PROGRESS);
			return mpi_errno;
		}

		if (nb > 0)
		{
			MPIDI_DBG_PRINTF((60, FCNAME, L"shm_writev returned %d", nb));
			total += nb;
			if (MPIDI_CH3I_Request_adjust_iov(req, nb))
			{
				/* Write operation complete */
				mpi_errno = MPIDI_CH3U_Handle_send_req(vc, req, &complete);
				if (mpi_errno != MPI_SUCCESS)
				{
					mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", 0);
					MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_WRITE_PROGRESS);
					return mpi_errno;
				}
				if (complete)
				{
					MPIDI_CH3I_SendQ_dequeue(vc);
				}
				vc->ch.send_active = MPIDI_CH3I_SendQ_head(vc);
			}
			else
			{
				MPIDI_DBG_PRINTF((65, FCNAME, L"iovec updated by %d bytes but not complete", nb));
				MPIU_Assert(req->ch.iov_offset < req->dev.iov_count);
				break;
			}
		}
		else
		{
#ifdef MPICH_DBG_OUTPUT
			if (nb != 0)
			{
				MPIDI_DBG_PRINTF((65, FCNAME, L"shm_writev returned %d bytes", nb));
			}
#endif
			break;
		}
	}

	MPIDI_DBG_PRINTF((60, FCNAME, L"exiting"));

	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_WRITE_PROGRESS);
	return total > 0 ? MPI_SUCCESS : SHM_WAIT_TIMEOUT;
}


#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_Handle_shm_read
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3I_Handle_shm_read(MPIDI_VC_t *vc, int nb)
{
	int mpi_errno = MPI_SUCCESS;
	MPID_Request * req;
	int complete;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_HANDLE_SHM_READ);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_HANDLE_SHM_READ);

	req = vc->ch.recv_active;
	if (req == NULL)
	{
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_HANDLE_SHM_READ);
		return MPI_SUCCESS;
	}

	if (nb > 0)
	{
		if (MPIDI_CH3I_Request_adjust_iov(req, nb))
		{
			/* Read operation complete */
			mpi_errno = MPIDI_CH3U_Handle_recv_req(vc, req, &complete);
			if (mpi_errno != MPI_SUCCESS)
			{
				mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", 0);
			}
			if (complete)
			{
				vc->ch.recv_active = NULL;
				vc->ch.shm_reading_pkt = TRUE;
			}
			else
			{
				vc->ch.recv_active = req;
				mpi_errno = MPIDI_CH3I_SHM_post_readv(vc, req->dev.iov, req->dev.iov_count, NULL);
			}
		}
		else
		{
#ifdef MPICH_DBG_OUTPUT
			if (req->ch.iov_offset >= req->dev.iov_count)
			{
				mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**iov_offset", L"**iov_offset %d %d", req->ch.iov_offset, req->dev.iov_count);
				MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_HANDLE_SHM_READ);
				return mpi_errno;
			}
#endif
			MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_HANDLE_SHM_READ);
			return MPI_SUCCESS;
		}
	}
	else
	{
		MPIDI_DBG_PRINTF((65, FCNAME, L"Read args were iov=%x, count=%d", req->dev.iov + req->ch.iov_offset, req->dev.iov_count - req->ch.iov_offset));
	}
	
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_HANDLE_SHM_READ);
	return MPI_SUCCESS;
}


#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_SHM_write
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_SHM_writev
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3I_SHM_writev(MPIDI_VC_t *vc, MPID_IOV *iov, int n, int *num_bytes_ptr)
{
#ifdef MPICH_DBG_OUTPUT
	int mpi_errno;
#endif
	int i;
	unsigned int total = 0;
	unsigned int num_bytes=0;
	unsigned int cur_avail, dest_avail;
	char *cur_pos, *dest_pos;
	int index;
	MPIDI_CH3I_SHM_Queue_t * writeq;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_SHM_WRITEV);
	MPIDI_STATE_DECL(MPID_STATE_MEMCPY);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_SHM_WRITEV);

	MPIDI_DBG_PRINTF((60, FCNAME, L"entering"));
	writeq = vc->ch.write_shmq;
	index = writeq->tail_index;
	if (writeq->packet[index].avail == MPIDI_CH3I_PKT_FILLED)
	{
		*num_bytes_ptr = 0;
		MPIDI_DBG_PRINTF((60, FCNAME, L"exiting"));
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_WRITEV);
		return MPI_SUCCESS;
	}

	MPIU_DBG_PRINTF((L"writing to write_shmq %p\n", writeq));
	if (n == 2 && (iov[0].MPID_IOV_LEN + iov[1].MPID_IOV_LEN) < MPIDI_CH3I_PACKET_SIZE)
	{
		MPIDI_DBG_PRINTF((60, FCNAME, L"writing %d bytes to write_shmq %08p packet[%d]", iov[0].MPID_IOV_LEN + iov[1].MPID_IOV_LEN, writeq, index));
		MPIDI_FUNC_ENTER(MPID_STATE_MEMCPY);
		memcpy(writeq->packet[index].data, iov[0].MPID_IOV_BUF, iov[0].MPID_IOV_LEN);
		MPIDI_FUNC_EXIT(MPID_STATE_MEMCPY);
		MPIDI_FUNC_ENTER(MPID_STATE_MEMCPY);
		memcpy(&writeq->packet[index].data[iov[0].MPID_IOV_LEN], iov[1].MPID_IOV_BUF, iov[1].MPID_IOV_LEN);
		MPIDI_FUNC_EXIT(MPID_STATE_MEMCPY);
		writeq->packet[index].num_bytes = (int)(iov[0].MPID_IOV_LEN + iov[1].MPID_IOV_LEN);
		total = writeq->packet[index].num_bytes;
		MPID_WRITE_BARRIER();
		writeq->packet[index].avail = MPIDI_CH3I_PKT_FILLED;
#ifdef MPICH_DBG_OUTPUT
		/*MPIU_Assert(index == writeq->tail_index);*/
		if (index != writeq->tail_index)
		{
			mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**shmq_index", L"**shmq_index %d %d", index, writeq->tail_index);
			MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_WRITEV);
			return mpi_errno;
		}
#endif
		writeq->tail_index = (writeq->tail_index + 1) % MPIDI_CH3I_NUM_PACKETS;
		MPIDI_DBG_PRINTF((60, FCNAME, L"write_shmq tail = %d", writeq->tail_index));
		*num_bytes_ptr = total;
		MPIDI_DBG_PRINTF((60, FCNAME, L"exiting"));
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_WRITEV);
		return MPI_SUCCESS;
	}

	dest_pos = writeq->packet[index].data;
	dest_avail = MPIDI_CH3I_PACKET_SIZE;
	writeq->packet[index].num_bytes = 0;
	for (i=0; i<n; i++)
	{
		if (iov[i].MPID_IOV_LEN <= (MPID_IOV_LEN_CAST)dest_avail)
		{
			total += (unsigned int)iov[i].MPID_IOV_LEN;
			writeq->packet[index].num_bytes += (int)iov[i].MPID_IOV_LEN;
			dest_avail -= (unsigned int)iov[i].MPID_IOV_LEN;
			MPIDI_DBG_PRINTF((60, FCNAME, L"writing %d bytes to write_shmq %08p packet[%d]", iov[i].MPID_IOV_LEN, writeq, index));
			MPIDI_FUNC_ENTER(MPID_STATE_MEMCPY);
			memcpy(dest_pos, iov[i].MPID_IOV_BUF, iov[i].MPID_IOV_LEN);
			MPIDI_FUNC_EXIT(MPID_STATE_MEMCPY);
			dest_pos += iov[i].MPID_IOV_LEN;
		}
		else
		{
			total += dest_avail;
			writeq->packet[index].num_bytes = MPIDI_CH3I_PACKET_SIZE;
			MPIDI_DBG_PRINTF((60, FCNAME, L"writing %d bytes to write_shmq %08p packet[%d]", dest_avail, writeq, index));
			MPIDI_FUNC_ENTER(MPID_STATE_MEMCPY);
			memcpy(dest_pos, iov[i].MPID_IOV_BUF, dest_avail);
			MPIDI_FUNC_EXIT(MPID_STATE_MEMCPY);
			MPID_WRITE_BARRIER();
			writeq->packet[index].avail = MPIDI_CH3I_PKT_FILLED;
			cur_pos = (char *) iov[i].MPID_IOV_BUF + dest_avail;
			cur_avail = (unsigned int)iov[i].MPID_IOV_LEN - dest_avail;
			while (cur_avail)
			{
				index = writeq->tail_index = (writeq->tail_index + 1) % MPIDI_CH3I_NUM_PACKETS;
				MPIDI_DBG_PRINTF((60, FCNAME, L"write_shmq tail = %d", writeq->tail_index));
				if (writeq->packet[index].avail == MPIDI_CH3I_PKT_FILLED)
				{
					*num_bytes_ptr = total;
					MPIDI_DBG_PRINTF((60, FCNAME, L"exiting"));
					MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_WRITEV);
					return MPI_SUCCESS;
				}
				num_bytes = min(cur_avail, MPIDI_CH3I_PACKET_SIZE);
				writeq->packet[index].num_bytes = num_bytes;
				MPIDI_DBG_PRINTF((60, FCNAME, L"writing %d bytes to write_shmq %08p packet[%d]", num_bytes, writeq, index));
				MPIDI_FUNC_ENTER(MPID_STATE_MEMCPY);
				memcpy(writeq->packet[index].data, cur_pos, num_bytes);
				MPIDI_FUNC_EXIT(MPID_STATE_MEMCPY);
				total += num_bytes;
				cur_pos += num_bytes;
				cur_avail -= num_bytes;
				if (cur_avail)
				{
					MPID_WRITE_BARRIER();
					writeq->packet[index].avail = MPIDI_CH3I_PKT_FILLED;
				}
			}
			dest_pos = writeq->packet[index].data + num_bytes;
			dest_avail = MPIDI_CH3I_PACKET_SIZE - num_bytes;
		}
		if (dest_avail == 0)
		{
			MPID_WRITE_BARRIER();
			writeq->packet[index].avail = MPIDI_CH3I_PKT_FILLED;
			index = writeq->tail_index = (writeq->tail_index + 1) % MPIDI_CH3I_NUM_PACKETS;
			MPIDI_DBG_PRINTF((60, FCNAME, L"write_shmq tail = %d", writeq->tail_index));
			if (writeq->packet[index].avail == MPIDI_CH3I_PKT_FILLED)
			{
				*num_bytes_ptr = total;
				MPIDI_DBG_PRINTF((60, FCNAME, L"exiting"));
				MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_WRITEV);
				return MPI_SUCCESS;
			}
			dest_pos = writeq->packet[index].data;
			dest_avail = MPIDI_CH3I_PACKET_SIZE;
			writeq->packet[index].num_bytes = 0;
		}
	}
	if (dest_avail < MPIDI_CH3I_PACKET_SIZE)
	{
		MPID_WRITE_BARRIER();
		writeq->packet[index].avail = MPIDI_CH3I_PKT_FILLED;
		writeq->tail_index = (writeq->tail_index + 1) % MPIDI_CH3I_NUM_PACKETS;
		MPIDI_DBG_PRINTF((60, FCNAME, L"write_shmq tail = %d", writeq->tail_index));
	}

	*num_bytes_ptr = total;
	MPIDI_DBG_PRINTF((60, FCNAME, L"exiting"));
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_WRITEV);
	return MPI_SUCCESS;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_SHM_rdma_writev
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3I_SHM_rdma_writev(MPIDI_VC_t *vc, MPID_Request *sreq)
{
	int mpi_errno = MPI_SUCCESS;
	int i;
	char *rbuf, *sbuf;
	int rbuf_len, riov_offset;
	int sbuf_len;
	int len;
	SIZE_T num_written;
	MPID_IOV *send_iov, *recv_iov;
	int send_count, recv_count;
	int complete;
	MPIDI_CH3_Pkt_t pkt;
	MPIDI_CH3_Pkt_rdma_reload_t * reload_pkt = &pkt.reload;
	MPID_Request * reload_sreq;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_SHM_RDMA_WRITEV);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_SHM_RDMA_WRITEV);

	/* save the receiver's request to send back with the reload packet */
	reload_pkt->rreq = sreq->dev.rdma_request;
	reload_pkt->sreq = sreq->handle;

	for (;;)
	{
		send_iov = sreq->dev.iov;
		send_count = sreq->dev.iov_count;
		recv_iov = sreq->dev.rdma_iov;
		recv_count = sreq->dev.rdma_iov_count;

		/*
		wprintf(L"shm_rdma: writing %d send buffers into %d recv buffers.\n", send_count, recv_count);fflush(stdout);
		for (i=0; i<send_count; i++)
		{
			wprintf(L"shm_rdma: send buf[%d] = %p, len = %d\n", i, send_iov[i].MPID_IOV_BUF, send_iov[i].MPID_IOV_LEN);
		}
		for (i=0; i<recv_count; i++)
		{
			wprintf(L"shm_rdma: recv buf[%d] = %p, len = %d\n", i, recv_iov[i].MPID_IOV_BUF, recv_iov[i].MPID_IOV_LEN);
		}
		fflush(stdout);
		*/
		rbuf = recv_iov[0].MPID_IOV_BUF;
		rbuf_len = (int)recv_iov[0].MPID_IOV_LEN;
		riov_offset = 0;
		for (i=sreq->ch.iov_offset; i<send_count; i++)
		{
			sbuf = send_iov[i].MPID_IOV_BUF;
			sbuf_len = (int)send_iov[i].MPID_IOV_LEN;
			while (sbuf_len)
			{
				len = MPIDU_MIN(sbuf_len, rbuf_len);
				/*wprintf(L"writing %d bytes to remote process.\n", len);fflush(stdout);*/
				if (!WriteProcessMemory(vc->ch.hSharedProcessHandle, rbuf, sbuf, len, &num_written))
				{
					mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", L"**fail %s %d", L"WriteProcessMemory failed", GetLastError());
					MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_RDMA_WRITEV);
					return mpi_errno;
				}
				if (num_written == -1)
				{
					mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", L"**fail %s", L"WriteProcessMemory returned -1 bytes written");
					MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_RDMA_WRITEV);
					return mpi_errno;
				}
				/*wprintf(L"wrote %d bytes to remote process\n", num_written);fflush(stdout);*/
				if (num_written < (SIZE_T)rbuf_len)
				{
					rbuf = rbuf + num_written;
					rbuf_len = rbuf_len - (int)num_written;
				}
				else
				{
					riov_offset = riov_offset + 1;
					if (riov_offset < recv_count)
					{
						rbuf = recv_iov[riov_offset].MPID_IOV_BUF;
						rbuf_len = (int)recv_iov[riov_offset].MPID_IOV_LEN;
					}
				}
				sbuf = sbuf + num_written;
				sbuf_len = sbuf_len - (int)num_written;
				if (riov_offset == recv_count)
				{
					if ( (i != (send_count - 1)) || (sbuf_len != 0) )
					{
						/* the recv iov needs to be reloaded */
						if (sbuf_len != 0)
						{
							sreq->dev.iov[i].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)sbuf;
							sreq->dev.iov[i].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)sbuf_len;
						}
						sreq->ch.iov_offset = i;

						/* send the reload packet to the receiver */
						MPIDI_Pkt_init(reload_pkt, MPIDI_CH3_PKT_RELOAD);
						reload_pkt->send_recv = MPIDI_CH3_PKT_RELOAD_RECV;
						mpi_errno = MPIDI_CH3.iStartMsg(vc, reload_pkt, sizeof(*reload_pkt), &reload_sreq);
						/* --BEGIN ERROR HANDLING-- */
						if (mpi_errno != MPI_SUCCESS)
						{
							mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", 0);
							MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_RDMA_WRITEV);
							return mpi_errno;
						}
						/* --END ERROR HANDLING-- */
						if (reload_sreq != NULL)
						{
							/* The sender doesn't need to know when the packet has been sent.  So release the request immediately */
							MPID_Request_release(reload_sreq);
						}
						MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_RDMA_WRITEV);
						return MPI_SUCCESS;
					}
				}
			}
		}

		/* update the sender's request */
		mpi_errno = MPIDI_CH3U_Handle_send_req(vc, sreq, &complete);
		if (mpi_errno != MPI_SUCCESS)
		{
			mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", L"**fail %s", L"unable to update request after rdma write");
			MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_RDMA_WRITEV);
			return mpi_errno;
		}

		if (complete || (riov_offset == recv_count))
		{
			/* send the reload/done packet to the receiver */
			MPIDI_Pkt_init(reload_pkt, MPIDI_CH3_PKT_RELOAD);
			reload_pkt->send_recv = MPIDI_CH3_PKT_RELOAD_RECV;
			mpi_errno = MPIDI_CH3.iStartMsg(vc, reload_pkt, sizeof(*reload_pkt), &reload_sreq);
			/* --BEGIN ERROR HANDLING-- */
			if (mpi_errno != MPI_SUCCESS)
			{
				mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", 0);
				MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_RDMA_WRITEV);
				return mpi_errno;
			}
			/* --END ERROR HANDLING-- */
			if (reload_sreq != NULL)
			{
				/* The sender doesn't need to know when the packet has been sent.  So release the request immediately */
				MPID_Request_release(reload_sreq);
			}
			MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_RDMA_WRITEV);
			return MPI_SUCCESS;
		}
	}

	MPIU_UNREACHABLE_CODE();
	/*
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_RDMA_WRITEV);
	return mpi_errno;
	*/
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_SHM_rdma_readv
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3I_SHM_rdma_readv(MPIDI_VC_t *vc, MPID_Request *rreq)
{
	int mpi_errno = MPI_SUCCESS;
	int i;
	char *rbuf, *sbuf;
	int rbuf_len;
	int sbuf_len, siov_offset;
	int len;
	SIZE_T num_read;
	MPID_IOV *send_iov, *recv_iov;
	int send_count, recv_count;
	int complete;
	MPIDI_CH3_Pkt_t pkt;
	MPIDI_CH3_Pkt_rdma_reload_t * reload_pkt = &pkt.reload;
	MPID_Request * reload_rreq;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_SHM_RDMA_READV);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_SHM_RDMA_READV);

	/* save the sender's request to send back with the reload packet */
	/*reload_pkt->req = rreq->dev.rdma_request;*/
	reload_pkt->sreq = rreq->dev.sender_req_id;
	reload_pkt->rreq = rreq->handle;

	for (;;)
	{
		recv_iov = rreq->dev.iov;
		recv_count = rreq->dev.iov_count;
		send_iov = rreq->dev.rdma_iov;
		send_count = rreq->dev.rdma_iov_count;
		siov_offset = rreq->dev.rdma_iov_offset;

		/*
		wprintf(L"shm_rdma: reading %d send buffers into %d recv buffers.\n", send_count, recv_count);fflush(stdout);
		for (i=siov_offset; i<send_count; i++)
		{
			wprintf(L"shm_rdma: send buf[%d] = %p, len = %d\n", i, send_iov[i].MPID_IOV_BUF, send_iov[i].MPID_IOV_LEN);
		}
		for (i=0; i<recv_count; i++)
		{
			wprintf(L"shm_rdma: recv buf[%d] = %p, len = %d\n", i, recv_iov[i].MPID_IOV_BUF, recv_iov[i].MPID_IOV_LEN);
		}
		fflush(stdout);
		*/
		sbuf = send_iov[siov_offset].MPID_IOV_BUF;
		sbuf_len = (int)send_iov[siov_offset].MPID_IOV_LEN;
		for (i=rreq->ch.iov_offset; i<recv_count; i++)
		{
			rbuf = recv_iov[i].MPID_IOV_BUF;
			rbuf_len = (int)recv_iov[i].MPID_IOV_LEN;
			while (rbuf_len)
			{
				len = MPIDU_MIN(rbuf_len, sbuf_len);
				/*wprintf(L"reading %d bytes from the remote process.\n", len);fflush(stdout);*/
				if (!ReadProcessMemory(vc->ch.hSharedProcessHandle, sbuf, rbuf, len, &num_read))
				{
					/*wprintf(L"ReadProcessMemory failed, error %d\n", GetLastError());*/
					mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", L"**fail %s %d", L"ReadProcessMemory failed", GetLastError());
					MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_RDMA_READV);
					return mpi_errno;
				}
				if (num_read == -1)
				{
					/*wprintf(L"ReadProcessMemory read -1 bytes.\n");fflush(stdout);*/
					mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", L"**fail %s", L"ReadProcessMemory returned -1 bytes written");
					MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_RDMA_READV);
					return mpi_errno;
				}
				/*wprintf(L"read %d bytes from the remote process\n", num_read);fflush(stdout);*/
				if (num_read < (SIZE_T)sbuf_len)
				{
					sbuf = sbuf + num_read;
					sbuf_len = sbuf_len - (int)num_read;
				}
				else
				{
					siov_offset = siov_offset + 1;
					if (siov_offset < send_count)
					{
						sbuf = send_iov[siov_offset].MPID_IOV_BUF;
						sbuf_len = (int)send_iov[siov_offset].MPID_IOV_LEN;
					}
					else
					{
						sbuf_len = 0;
					}
				}
				rbuf = rbuf + num_read;
				rbuf_len = rbuf_len - (int)num_read;
				if (siov_offset == send_count)
				{
					if ( (i != (recv_count - 1)) || (rbuf_len != 0) )
					{
						/* the send iov needs to be reloaded */
						if (rbuf_len != 0)
						{
							rreq->dev.iov[i].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)rbuf;
							rreq->dev.iov[i].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)rbuf_len;
						}
						rreq->ch.iov_offset = i;

						/* send the reload packet to the sender */
						/*wprintf(L"sending reload packet to the sender.\n");fflush(stdout);*/
						MPIDI_Pkt_init(reload_pkt, MPIDI_CH3_PKT_RELOAD);
						reload_pkt->send_recv = MPIDI_CH3_PKT_RELOAD_SEND;
						mpi_errno = MPIDI_CH3.iStartMsg(vc, reload_pkt, sizeof(*reload_pkt), &reload_rreq);
						/* --BEGIN ERROR HANDLING-- */
						if (mpi_errno != MPI_SUCCESS)
						{
							mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", 0);
							MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_RDMA_READV);
							return mpi_errno;
						}
						/* --END ERROR HANDLING-- */
						if (reload_rreq != NULL)
						{
							/* The sender doesn't need to know when the packet has been sent.  So release the request immediately */
							MPID_Request_release(reload_rreq);
						}
						MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_RDMA_READV);
						return MPI_SUCCESS;
					}
				}
			}
		}

		/* update the sender's request */
		mpi_errno = MPIDI_CH3U_Handle_recv_req(vc, rreq, &complete);
		if (mpi_errno != MPI_SUCCESS)
		{
			mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", L"**fail %s", L"unable to update request after rdma read");
			MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_RDMA_READV);
			return mpi_errno;
		}

		if (complete || (siov_offset == send_count))
		{
			/*wprintf(L"sending reload send packet.\n");fflush(stdout);*/
			/* send the reload/done packet to the sender */
			MPIDI_Pkt_init(reload_pkt, MPIDI_CH3_PKT_RELOAD);
			reload_pkt->send_recv = MPIDI_CH3_PKT_RELOAD_SEND;
			mpi_errno = MPIDI_CH3.iStartMsg(vc, reload_pkt, sizeof(*reload_pkt), &reload_rreq);
			/* --BEGIN ERROR HANDLING-- */
			if (mpi_errno != MPI_SUCCESS)
			{
				mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", 0);
				MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_RDMA_READV);
				return mpi_errno;
			}
			/* --END ERROR HANDLING-- */
			if (reload_rreq != NULL)
			{
				/* The sender doesn't need to know when the packet has been sent.  So release the request immediately */
				MPID_Request_release(reload_rreq);
			}
			MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_RDMA_READV);
			return MPI_SUCCESS;
		}

		rreq->dev.rdma_iov_offset = siov_offset;
		rreq->dev.rdma_iov[siov_offset].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)sbuf;
		rreq->dev.rdma_iov[siov_offset].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)sbuf_len;
	}

	MPIU_UNREACHABLE_CODE();
	/*
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_RDMA_READV);
	return mpi_errno;
	*/
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_SHM_read_progress
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3I_SHM_read_progress(MPIDI_VC_t *recv_vc_ptr, int millisecond_timeout, MPIDI_VC_t **vc_pptr, int *num_bytes_ptr)
{
	int mpi_errno = MPI_SUCCESS;
	void *mem_ptr;
	char *iter_ptr;
	int num_bytes;
	unsigned int offset;
	MPIDI_CH3I_SHM_Packet_t *pkt_ptr;
	MPIDI_CH3I_SHM_Queue_t *shm_ptr;
	register int index, working;
	MPID_Request *sreq, *rreq;
	int complete;
	int i;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_SHM_READ_PROGRESS);
	MPIDI_STATE_DECL(MPID_STATE_MEMCPY);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_SHM_READ_PROGRESS);

	for (;;) 
	{
		working = FALSE;

		while (recv_vc_ptr)
		{
			shm_ptr = recv_vc_ptr->ch.read_shmq;
			if (shm_ptr == NULL)
			{
				recv_vc_ptr = recv_vc_ptr->ch.shm_next_reader;
				continue;
			}
			index = shm_ptr->head_index;
			pkt_ptr = &shm_ptr->packet[index];

			/* if the packet at the head index is available, the queue is empty */
			if (pkt_ptr->avail == MPIDI_CH3I_PKT_EMPTY)
			{
				recv_vc_ptr = recv_vc_ptr->ch.shm_next_reader;
				continue;
			}
			MPID_READ_BARRIER(); /* no loads after this line can occur before the avail flag has been read */

			working = TRUE;
			MPIU_DBG_PRINTF((L"MPIDI_CH3I_SHM_read_progress: reading from queue %p\n", shm_ptr));

			mem_ptr = (void*)(pkt_ptr->data + pkt_ptr->offset);
			num_bytes = pkt_ptr->num_bytes;

			if (recv_vc_ptr->ch.shm_reading_pkt)
			{
				MPIDI_DBG_PRINTF((60, FCNAME, L"reading header(%d bytes) from read_shmq %08p packet[%d]", sizeof(MPIDI_CH3_Pkt_t), shm_ptr, index));
				if (((MPIDI_CH3_Pkt_t*)mem_ptr)->type > MPIDI_CH3I_PKT_SC_CLOSE/*MPIDI_CH3_PKT_END_CH3*/)
				{
					if (((MPIDI_CH3_Pkt_t*)mem_ptr)->type == MPIDI_CH3_PKT_RTS_IOV)
					{
						/*wprintf(L"received rts packet.\n");fflush(stdout);*/
						rreq = MPID_Request_create();
						if (rreq == NULL)
						{
							mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
							MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_READ_PROGRESS);
							return mpi_errno;
						}
						MPIU_Object_set_ref(rreq, 1);
						rreq->kind = MPIDI_CH3I_RTS_IOV_READ_REQUEST;
						rreq->dev.rdma_request = ((MPIDI_CH3_Pkt_rdma_rts_iov_t*)mem_ptr)->sreq;
						rreq->dev.rdma_iov_count = ((MPIDI_CH3_Pkt_rdma_rts_iov_t*)mem_ptr)->iov_len;
						rreq->dev.iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)&rreq->dev.rdma_iov;
						rreq->dev.iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)(rreq->dev.rdma_iov_count * sizeof(MPID_IOV));
						rreq->dev.iov[1].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)&rreq->ch.pkt;
						rreq->dev.iov[1].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)sizeof(MPIDI_CH3_Pkt_t);
						rreq->dev.iov_count = 2;
						rreq->ch.req = NULL;
						recv_vc_ptr->ch.recv_active = rreq;
					}
					else if (((MPIDI_CH3_Pkt_t*)mem_ptr)->type == MPIDI_CH3_PKT_CTS_IOV)
					{
						/*wprintf(L"received cts packet.\n");fflush(stdout);*/
						MPID_Request_get_ptr(((MPIDI_CH3_Pkt_rdma_cts_iov_t*)mem_ptr)->sreq, sreq);
						sreq->dev.rdma_request = ((MPIDI_CH3_Pkt_rdma_cts_iov_t*)mem_ptr)->rreq;
						sreq->dev.rdma_iov_count = ((MPIDI_CH3_Pkt_rdma_cts_iov_t*)mem_ptr)->iov_len;
						rreq = MPID_Request_create();
						if (rreq == NULL)
						{
							mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
							MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_READ_PROGRESS);
							return mpi_errno;
						}
						MPIU_Object_set_ref(rreq, 1);
						rreq->kind = MPIDI_CH3I_IOV_WRITE_REQUEST;
						rreq->dev.iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)&sreq->dev.rdma_iov;
						rreq->dev.iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)(sreq->dev.rdma_iov_count * sizeof(MPID_IOV));
						rreq->dev.iov_count = 1;
						rreq->ch.req = sreq;
						recv_vc_ptr->ch.recv_active = rreq;
						/*MPIDI_CH3I_SHM_post_read(recv_vc_ptr, &sreq->ch.rdma_iov, sreq->ch.rdma_iov_count * sizeof(MPID_IOV), NULL);*/
					}
					else if (((MPIDI_CH3_Pkt_t*)mem_ptr)->type == MPIDI_CH3_PKT_IOV)
					{
						if ( ((MPIDI_CH3_Pkt_rdma_iov_t*)mem_ptr)->send_recv == MPIDI_CH3_PKT_RELOAD_SEND )
						{
							/*wprintf(L"received sender's iov packet, posting a read of %d iovs.\n", ((MPIDI_CH3_Pkt_rdma_iov_t*)mem_ptr)->iov_len);fflush(stdout);*/
							MPID_Request_get_ptr(((MPIDI_CH3_Pkt_rdma_iov_t*)mem_ptr)->req, sreq);
							sreq->dev.rdma_iov_count = ((MPIDI_CH3_Pkt_rdma_iov_t*)mem_ptr)->iov_len;
							rreq = MPID_Request_create();
							if (rreq == NULL)
							{
								mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
								MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_READ_PROGRESS);
								return mpi_errno;
							}
							MPIU_Object_set_ref(rreq, 1);
							rreq->kind = MPIDI_CH3I_IOV_READ_REQUEST;
							rreq->dev.iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)&sreq->dev.rdma_iov;
							rreq->dev.iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)(sreq->dev.rdma_iov_count * sizeof(MPID_IOV));
							rreq->dev.iov_count = 1;
							rreq->ch.req = sreq;
							recv_vc_ptr->ch.recv_active = rreq;
						}
						else if ( ((MPIDI_CH3_Pkt_rdma_iov_t*)mem_ptr)->send_recv == MPIDI_CH3_PKT_RELOAD_RECV )
						{
							/*wprintf(L"received receiver's iov packet, posting a read of %d iovs.\n", ((MPIDI_CH3_Pkt_rdma_iov_t*)mem_ptr)->iov_len);fflush(stdout);*/
							MPID_Request_get_ptr(((MPIDI_CH3_Pkt_rdma_iov_t*)mem_ptr)->req, rreq);
							rreq->dev.rdma_iov_count = ((MPIDI_CH3_Pkt_rdma_iov_t*)mem_ptr)->iov_len;
							sreq = MPID_Request_create();
							if (sreq == NULL)
							{
								mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
								MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_READ_PROGRESS);
								return mpi_errno;
							}
							MPIU_Object_set_ref(sreq, 1);
							sreq->kind = MPIDI_CH3I_IOV_WRITE_REQUEST;
							sreq->dev.iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)&rreq->dev.rdma_iov;
							sreq->dev.iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)(rreq->dev.rdma_iov_count * sizeof(MPID_IOV));
							sreq->dev.iov_count = 1;
							sreq->ch.req = rreq;
							recv_vc_ptr->ch.recv_active = sreq;
						}
						else
						{
							mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", L"**fail %s", L"received invalid MPIDI_CH3_PKT_IOV packet");
							MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_READ_PROGRESS);
							return mpi_errno;
						}
					}
					else if (((MPIDI_CH3_Pkt_t*)mem_ptr)->type == MPIDI_CH3_PKT_RELOAD)
					{
						if (((MPIDI_CH3_Pkt_rdma_reload_t*)mem_ptr)->send_recv == MPIDI_CH3_PKT_RELOAD_SEND)
						{
							/*wprintf(L"received reload send packet.\n");fflush(stdout);*/
							MPID_Request_get_ptr(((MPIDI_CH3_Pkt_rdma_reload_t*)mem_ptr)->sreq, sreq);
							mpi_errno = MPIDI_CH3U_Handle_send_req(recv_vc_ptr, sreq, &complete);
							if (mpi_errno != MPI_SUCCESS)
							{
								mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", L"**fail %s", L"unable to update send request after receiving a reload packet");
								MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_READ_PROGRESS);
								return mpi_errno;
							}
							if (!complete)
							{
								/* send a new iov */
								MPID_Request * rts_sreq;
								MPIDI_CH3_Pkt_t pkt;

								/*wprintf(L"sending reloaded send iov of length %d\n", sreq->dev.iov_count);fflush(stdout);*/
								MPIDI_Pkt_init(&pkt.iov, MPIDI_CH3_PKT_IOV);
								pkt.iov.send_recv = MPIDI_CH3_PKT_RELOAD_SEND;
								pkt.iov.req = ((MPIDI_CH3_Pkt_rdma_reload_t*)mem_ptr)->rreq;
								pkt.iov.iov_len = sreq->dev.iov_count;

								sreq->dev.rdma_iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)&pkt;
								sreq->dev.rdma_iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)sizeof(MPIDI_CH3_Pkt_t);
								sreq->dev.rdma_iov[1].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)sreq->dev.iov;
								sreq->dev.rdma_iov[1].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)(sreq->dev.iov_count * sizeof(MPID_IOV));

								mpi_errno = MPIDI_CH3.iStartMsgv(recv_vc_ptr, sreq->dev.rdma_iov, 2, &rts_sreq);
								/* --BEGIN ERROR HANDLING-- */
								if (mpi_errno != MPI_SUCCESS)
								{
									MPIU_Object_set_ref(sreq, 0);
									MPIDI_CH3_Request_destroy(sreq);
									sreq = NULL;
									mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|rtspkt", 0);
									MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_READ_PROGRESS);
									return mpi_errno;
								}
								/* --END ERROR HANDLING-- */
								if (rts_sreq != NULL)
								{
									/* The sender doesn't need to know when the message has been sent.  So release the request immediately */
									MPID_Request_release(rts_sreq);
								}
							}
						}
						else if (((MPIDI_CH3_Pkt_rdma_reload_t*)mem_ptr)->send_recv == MPIDI_CH3_PKT_RELOAD_RECV)
						{
							/*wprintf(L"received reload recv packet.\n");fflush(stdout);*/
							MPID_Request_get_ptr(((MPIDI_CH3_Pkt_rdma_reload_t*)mem_ptr)->rreq, rreq);
							mpi_errno = MPIDI_CH3U_Handle_recv_req(recv_vc_ptr, rreq, &complete);
							if (mpi_errno != MPI_SUCCESS)
							{
								mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", L"**fail %s", L"unable to update request after receiving a reload packet");
								MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_READ_PROGRESS);
								return mpi_errno;
							}
							if (!complete)
							{
								/* send a new iov */
								MPID_Request * cts_sreq;
								MPIDI_CH3_Pkt_t pkt;

								/*wprintf(L"sending reloaded recv iov of length %d\n", rreq->dev.iov_count);fflush(stdout);*/
								MPIDI_Pkt_init(&pkt.iov, MPIDI_CH3_PKT_IOV);
								pkt.iov.send_recv = MPIDI_CH3_PKT_RELOAD_RECV;
								pkt.iov.req = ((MPIDI_CH3_Pkt_rdma_reload_t*)mem_ptr)->sreq;
								pkt.iov.iov_len = rreq->dev.iov_count;

								rreq->dev.rdma_iov[0].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)&pkt;
								rreq->dev.rdma_iov[0].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)sizeof(MPIDI_CH3_Pkt_t);
								rreq->dev.rdma_iov[1].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)rreq->dev.iov;
								rreq->dev.rdma_iov[1].MPID_IOV_LEN = (MPID_IOV_LEN_CAST)(rreq->dev.iov_count * sizeof(MPID_IOV));

								mpi_errno = MPIDI_CH3.iStartMsgv(recv_vc_ptr, rreq->dev.rdma_iov, 2, &cts_sreq);
								/* --BEGIN ERROR HANDLING-- */
								if (mpi_errno != MPI_SUCCESS)
								{
									/* This destruction probably isn't correct. */
									/* I think it needs to save the error in the request, complete the request and return */
									MPIU_Object_set_ref(rreq, 0);
									MPIDI_CH3_Request_destroy(rreq);
									rreq = NULL;
									mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|ctspkt", 0);
									MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_READ_PROGRESS);
									return mpi_errno;
								}
								/* --END ERROR HANDLING-- */
								if (cts_sreq != NULL)
								{
									/* The sender doesn't need to know when the message has been sent.  So release the request immediately */
									MPID_Request_release(cts_sreq);
								}
							}
						}
						else
						{
							mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", 0);
							MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_READ_PROGRESS);
							return mpi_errno;
						}

						recv_vc_ptr->ch.recv_active = NULL;
						recv_vc_ptr->ch.shm_reading_pkt = TRUE;
						if (num_bytes > sizeof(MPIDI_CH3_Pkt_t))
						{
							pkt_ptr->offset += sizeof(MPIDI_CH3_Pkt_t);
							num_bytes -= sizeof(MPIDI_CH3_Pkt_t);
							pkt_ptr->num_bytes = num_bytes;
							mem_ptr = (char*)mem_ptr + sizeof(MPIDI_CH3_Pkt_t);
						}
						else
						{
							pkt_ptr->offset = 0;
							MPID_READ_WRITE_BARRIER(); /* the writing of the flag cannot occur before the reading of the last piece of data */
							pkt_ptr->avail = MPIDI_CH3I_PKT_EMPTY;
							shm_ptr->head_index = (index + 1) % MPIDI_CH3I_NUM_PACKETS;
						}
						/* return from the wait */
						*num_bytes_ptr = 0;
						*vc_pptr = recv_vc_ptr;
						MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_READ_PROGRESS);
						return MPI_SUCCESS;
					}
					else
					{
						mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", L"**fail %s", L"shared memory read progress unable to handle unknown rdma packet");
						MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_READ_PROGRESS);
						return mpi_errno;
					}
				}
				else
				{
#if 0
				/*This is the sshm code. I think it is for detecting connect/accept attempts*/
				if (((MPIDI_CH3_Pkt_t*)mem_ptr)->type == MPIDI_CH3I_PKT_SC_CONN_ACCEPT)
				{
					/*recv_vc_ptr->ch.port_name_tag = ((MPIDI_CH3_Pkt_t*)mem_ptr)->sc_conn_accept.port_name_tag;*/
					MPIDI_CH3I_Acceptq_enqueue(recv_vc_ptr, ((MPIDI_CH3_Pkt_t*)mem_ptr)->sc_conn_accept.port_name_tag);
					MPIDI_CH3I_progress_completion_count++;
					MPIDI_CH3I_SHM_Remove_vc_references(recv_vc_ptr);
					recv_vc_ptr = NULL;
					if (num_bytes > sizeof(MPIDI_CH3_Pkt_t))
					{
						pkt_ptr->offset += sizeof(MPIDI_CH3_Pkt_t);
						num_bytes -= sizeof(MPIDI_CH3_Pkt_t);
						pkt_ptr->num_bytes = num_bytes;
						mem_ptr = (char*)mem_ptr + sizeof(MPIDI_CH3_Pkt_t);
					}
					else
					{
						pkt_ptr->offset = 0;
						MPID_READ_WRITE_BARRIER(); /* the writing of the flag cannot occur before the reading of the last piece of data */
						pkt_ptr->avail = MPIDI_CH3I_PKT_EMPTY;
						shm_ptr->head_index = (index + 1) % MPIDI_CH3I_NUM_PACKETS;
						MPIDI_DBG_PRINTF((60, FCNAME, L"read_shmq head = %d", shm_ptr->head_index));
					}
					continue;
				}
				else
				{
#endif
				mpi_errno = MPIDI_CH3U_Handle_recv_pkt(recv_vc_ptr, (MPIDI_CH3_Pkt_t*)mem_ptr, &recv_vc_ptr->ch.recv_active);
				if (mpi_errno != MPI_SUCCESS)
				{
					mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", L"**fail %s", L"shared memory read progress unable to handle incoming packet");
					MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_READ_PROGRESS);
					return mpi_errno;
				}
#if 0
				}
#endif
				}

				if (recv_vc_ptr->ch.recv_active == NULL)
				{
					recv_vc_ptr->ch.shm_reading_pkt = TRUE;
				}
				else
				{
					mpi_errno = MPIDI_CH3I_SHM_post_readv(recv_vc_ptr, recv_vc_ptr->ch.recv_active->dev.iov, recv_vc_ptr->ch.recv_active->dev.iov_count, NULL);
				}
				if (num_bytes > sizeof(MPIDI_CH3_Pkt_t))
				{
					pkt_ptr->offset += sizeof(MPIDI_CH3_Pkt_t);
					num_bytes -= sizeof(MPIDI_CH3_Pkt_t);
					pkt_ptr->num_bytes = num_bytes;
					mem_ptr = (char*)mem_ptr + sizeof(MPIDI_CH3_Pkt_t);
				}
				else
				{
					pkt_ptr->offset = 0;
					MPID_READ_WRITE_BARRIER(); /* the writing of the flag cannot occur before the reading of the last piece of data */
					pkt_ptr->avail = MPIDI_CH3I_PKT_EMPTY;
					shm_ptr->head_index = (index + 1) % MPIDI_CH3I_NUM_PACKETS;
					MPIDI_DBG_PRINTF((60, FCNAME, L"read_shmq head = %d", shm_ptr->head_index));
					
					recv_vc_ptr = recv_vc_ptr->ch.shm_next_reader;
					continue;
				}
				if (recv_vc_ptr->ch.recv_active == NULL)
				{
					recv_vc_ptr = recv_vc_ptr->ch.shm_next_reader;
					continue;
				}
			}

			MPIDI_DBG_PRINTF((60, FCNAME, L"read %d bytes", num_bytes));
			/*MPIDI_DBG_PRINTF((60, FCNAME, L"shm_wait(recv finished %d bytes)", num_bytes));*/
			if (!(recv_vc_ptr->ch.shm_state & SHM_READING_BIT))
			{
				recv_vc_ptr = recv_vc_ptr->ch.shm_next_reader;
				continue;
			}
			MPIDI_DBG_PRINTF((60, FCNAME, L"read update, total = %d + %d = %d", recv_vc_ptr->ch.read.total, num_bytes, recv_vc_ptr->ch.read.total + num_bytes));
			if (recv_vc_ptr->ch.read.use_iov)
			{
				iter_ptr = mem_ptr;
				while (num_bytes && recv_vc_ptr->ch.read.iovlen > 0)
				{
					if ((int)recv_vc_ptr->ch.read.iov[recv_vc_ptr->ch.read.index].MPID_IOV_LEN <= num_bytes)
					{
						/* copy the received data */
						MPIDI_DBG_PRINTF((60, FCNAME, L"reading %d bytes from read_shmq %08p packet[%d]", recv_vc_ptr->ch.read.iov[recv_vc_ptr->ch.read.index].MPID_IOV_LEN, shm_ptr, index));
						MPIDI_FUNC_ENTER(MPID_STATE_MEMCPY);
						memcpy(recv_vc_ptr->ch.read.iov[recv_vc_ptr->ch.read.index].MPID_IOV_BUF, iter_ptr,
							recv_vc_ptr->ch.read.iov[recv_vc_ptr->ch.read.index].MPID_IOV_LEN);
						MPIDI_FUNC_EXIT(MPID_STATE_MEMCPY);
						iter_ptr += recv_vc_ptr->ch.read.iov[recv_vc_ptr->ch.read.index].MPID_IOV_LEN;
						/* update the iov */
						num_bytes -= (int)recv_vc_ptr->ch.read.iov[recv_vc_ptr->ch.read.index].MPID_IOV_LEN;
						recv_vc_ptr->ch.read.index++;
						recv_vc_ptr->ch.read.iovlen--;
					}
					else
					{
						/* copy the received data */
						MPIDI_DBG_PRINTF((60, FCNAME, L"reading %d bytes from read_shmq %08p packet[%d]", num_bytes, shm_ptr, index));
						MPIDI_FUNC_ENTER(MPID_STATE_MEMCPY);
						memcpy(recv_vc_ptr->ch.read.iov[recv_vc_ptr->ch.read.index].MPID_IOV_BUF, iter_ptr, num_bytes);
						MPIDI_FUNC_EXIT(MPID_STATE_MEMCPY);
						iter_ptr += num_bytes;
						/* update the iov */
						recv_vc_ptr->ch.read.iov[recv_vc_ptr->ch.read.index].MPID_IOV_LEN -= (MPID_IOV_LEN_CAST)num_bytes;
						recv_vc_ptr->ch.read.iov[recv_vc_ptr->ch.read.index].MPID_IOV_BUF = (MPID_IOV_BUF_CAST)(
							(char*)(recv_vc_ptr->ch.read.iov[recv_vc_ptr->ch.read.index].MPID_IOV_BUF) + num_bytes);
						num_bytes = 0;
					}
				}
				offset = (unsigned int)((unsigned char*)iter_ptr - (unsigned char*)mem_ptr);
				recv_vc_ptr->ch.read.total += offset;
				if (num_bytes == 0)
				{
					/* put the shmem buffer back in the queue */
					pkt_ptr->offset = 0;
					MPID_READ_WRITE_BARRIER(); /* the writing of the flag cannot occur before the reading of the last piece of data */
					pkt_ptr->avail = MPIDI_CH3I_PKT_EMPTY;
#ifdef MPICH_DBG_OUTPUT
					/*MPIU_Assert(&shm_ptr->packet[index] == pkt_ptr);*/
					if (&shm_ptr->packet[index] != pkt_ptr)
					{
						mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**pkt_ptr", L"**pkt_ptr %p %p", &shm_ptr->packet[index], pkt_ptr);
						MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_READ_PROGRESS);
						return mpi_errno;
					}
#endif
					shm_ptr->head_index = (index + 1) % MPIDI_CH3I_NUM_PACKETS;
					MPIDI_DBG_PRINTF((60, FCNAME, L"read_shmq head = %d", shm_ptr->head_index));
				}
				else
				{
					/* update the head of the shmem queue */
					pkt_ptr->offset += (pkt_ptr->num_bytes - num_bytes);
					pkt_ptr->num_bytes = num_bytes;
				}
				if (recv_vc_ptr->ch.read.iovlen == 0)
				{
					if (recv_vc_ptr->ch.recv_active->kind < MPID_LAST_REQUEST_KIND)
					{
						recv_vc_ptr->ch.shm_state &= ~SHM_READING_BIT;
						*num_bytes_ptr = recv_vc_ptr->ch.read.total;
						*vc_pptr = recv_vc_ptr;
						MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_READ_PROGRESS);
						return MPI_SUCCESS;
					}
					else if (recv_vc_ptr->ch.recv_active->kind == MPIDI_CH3I_RTS_IOV_READ_REQUEST)
					{
						int found;
						/*wprintf(L"received rts iov_read.\n");fflush(stdout);*/

						mpi_errno = MPIDI_CH3U_Handle_recv_rndv_pkt(recv_vc_ptr, &recv_vc_ptr->ch.recv_active->ch.pkt, &rreq, &found);
						/* --BEGIN ERROR HANDLING-- */
						if (mpi_errno != MPI_SUCCESS)
						{
							mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", L"**fail %s", L"shared memory read progress unable to handle incoming rts(get) packet");
							MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_READ_PROGRESS);
							return mpi_errno;
						}
						/* --END ERROR HANDLING-- */

						for (i=0; i<recv_vc_ptr->ch.recv_active->dev.rdma_iov_count; i++)
						{
							rreq->dev.rdma_iov[i].MPID_IOV_BUF = recv_vc_ptr->ch.recv_active->dev.rdma_iov[i].MPID_IOV_BUF;
							rreq->dev.rdma_iov[i].MPID_IOV_LEN = recv_vc_ptr->ch.recv_active->dev.rdma_iov[i].MPID_IOV_LEN;
						}
						rreq->dev.rdma_iov_count = recv_vc_ptr->ch.recv_active->dev.rdma_iov_count;

						if (found)
						{
							mpi_errno = MPIDI_CH3U_Post_data_receive(TRUE, &rreq);
							/* --BEGIN ERROR HANDLING-- */
							if (mpi_errno != MPI_SUCCESS)
							{
								mpi_errno = MPIR_Err_create_code (mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|postrecv", L"**ch3|postrecv %s", L"MPIDI_CH3_PKT_RNDV_REQ_TO_SEND");
								MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_READ_PROGRESS);
								return mpi_errno;
							}
							/* --END ERROR HANDLING-- */
							mpi_errno = MPIDI_CH3.iStartRndvTransfer(recv_vc_ptr, rreq);
							/* --BEGIN ERROR HANDLING-- */
							if (mpi_errno != MPI_SUCCESS)
							{
								mpi_errno = MPIR_Err_create_code (mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**ch3|ctspkt", 0);
								MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_READ_PROGRESS);
								return mpi_errno;
							}
							/* --END ERROR HANDLING-- */
						}

						rreq = recv_vc_ptr->ch.recv_active;
						/* free the request used to receive the rts packet and iov data */
						MPIU_Object_set_ref(rreq, 0);
						MPIDI_CH3_Request_destroy(rreq);

						recv_vc_ptr->ch.recv_active = NULL;
						recv_vc_ptr->ch.shm_reading_pkt = TRUE;
					}
					else if (recv_vc_ptr->ch.recv_active->kind == MPIDI_CH3I_IOV_READ_REQUEST)
					{
						/*wprintf(L"received iov_read.\n");fflush(stdout);*/
						rreq = recv_vc_ptr->ch.recv_active;

						mpi_errno = MPIDI_CH3.iStartRndvTransfer(recv_vc_ptr, rreq->ch.req);
						/* --BEGIN ERROR HANDLING-- */
						if (mpi_errno != MPI_SUCCESS)
						{
							mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", L"**fail %s", L"shared memory read progress unable to handle incoming rts(get) iov");
							MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_READ_PROGRESS);
							return mpi_errno;
						}
						/* --END ERROR HANDLING-- */

						recv_vc_ptr->ch.recv_active = NULL;
						recv_vc_ptr->ch.shm_reading_pkt = TRUE;

						/* free the request used to receive the iov data */
						MPIU_Object_set_ref(rreq, 0);
						MPIDI_CH3_Request_destroy(rreq);
					}
					else if (recv_vc_ptr->ch.recv_active->kind == MPIDI_CH3I_IOV_WRITE_REQUEST)
					{
						/*wprintf(L"received iov_write.\n");fflush(stdout);*/
						mpi_errno = MPIDI_CH3I_SHM_rdma_writev(recv_vc_ptr, recv_vc_ptr->ch.recv_active->ch.req);
						/* --BEGIN ERROR HANDLING-- */
						if (mpi_errno != MPI_SUCCESS)
						{
							mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", 0);
							MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_READ_PROGRESS);
							return mpi_errno;
						}
						/* --END ERROR HANDLING-- */

						/* return from the wait */
						MPID_Request_release(recv_vc_ptr->ch.recv_active);
						recv_vc_ptr->ch.recv_active = NULL;
						recv_vc_ptr->ch.shm_reading_pkt = TRUE;
						*num_bytes_ptr = 0;
						*vc_pptr = recv_vc_ptr;
						MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_READ_PROGRESS);
						return MPI_SUCCESS;
					}
					else
					{
						mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**fail", L"**fail %s %d", L"invalid request type", recv_vc_ptr->ch.recv_active->kind);
						MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_READ_PROGRESS);
						return mpi_errno;
					}
				}
			}
			else
			{
				if ((unsigned int)num_bytes > recv_vc_ptr->ch.read.bufflen)
				{
					/* copy the received data */
					MPIDI_DBG_PRINTF((60, FCNAME, L"reading %d bytes from read_shmq %08p packet[%d]", recv_vc_ptr->ch.read.bufflen, shm_ptr, index));
					MPIDI_FUNC_ENTER(MPID_STATE_MEMCPY);
					memcpy(recv_vc_ptr->ch.read.buffer, mem_ptr, recv_vc_ptr->ch.read.bufflen);
					MPIDI_FUNC_EXIT(MPID_STATE_MEMCPY);
					recv_vc_ptr->ch.read.total = recv_vc_ptr->ch.read.bufflen;
					pkt_ptr->offset += recv_vc_ptr->ch.read.bufflen;
					pkt_ptr->num_bytes = num_bytes - recv_vc_ptr->ch.read.bufflen;
					recv_vc_ptr->ch.read.bufflen = 0;
				}
				else
				{
					/* copy the received data */
					MPIDI_DBG_PRINTF((60, FCNAME, L"reading %d bytes from read_shmq %08p packet[%d]", num_bytes, shm_ptr, index));
					MPIDI_FUNC_ENTER(MPID_STATE_MEMCPY);
					memcpy(recv_vc_ptr->ch.read.buffer, mem_ptr, num_bytes);
					MPIDI_FUNC_EXIT(MPID_STATE_MEMCPY);
					recv_vc_ptr->ch.read.total += num_bytes;
					/* advance the user pointer */
					recv_vc_ptr->ch.read.buffer = (char*)(recv_vc_ptr->ch.read.buffer) + num_bytes;
					recv_vc_ptr->ch.read.bufflen -= num_bytes;
					/* put the shmem buffer back in the queue */
					pkt_ptr->offset = 0;
					MPID_READ_WRITE_BARRIER(); /* the writing of the flag cannot occur before the reading of the last piece of data */
					pkt_ptr->avail = MPIDI_CH3I_PKT_EMPTY;
#ifdef MPICH_DBG_OUTPUT
					/*MPIU_Assert(&shm_ptr->packet[index] == pkt_ptr);*/
					if (&shm_ptr->packet[index] != pkt_ptr)
					{
						mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**pkt_ptr", L"**pkt_ptr %p %p", &shm_ptr->packet[index], pkt_ptr);
						MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_READ_PROGRESS);
						return mpi_errno;
					}
#endif
					shm_ptr->head_index = (index + 1) % MPIDI_CH3I_NUM_PACKETS;
					MPIDI_DBG_PRINTF((60, FCNAME, L"read_shmq head = %d", shm_ptr->head_index));
				}
				if (recv_vc_ptr->ch.read.bufflen == 0)
				{
					recv_vc_ptr->ch.shm_state &= ~SHM_READING_BIT;
					*num_bytes_ptr = recv_vc_ptr->ch.read.total;
					*vc_pptr = recv_vc_ptr;
					MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_READ_PROGRESS);
					return MPI_SUCCESS;
				}
			}
			recv_vc_ptr = recv_vc_ptr->ch.shm_next_reader;
		}

		if (millisecond_timeout == 0 && !working)
		{
			/* FIXME: This is wrong because SHM_WAIT_TIMEOUT might conflict with a real mpi error code */
			MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_READ_PROGRESS);
			return SHM_WAIT_TIMEOUT;
		}
	}

	MPIU_UNREACHABLE_CODE();
	/*
	mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**notimpl", 0);
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_READ_PROGRESS);
	return mpi_errno;
	*/
}

/* non-blocking functions */

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_SHM_post_read
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3I_SHM_post_read(MPIDI_VC_t *vc, void *buf, int len, int (*rfn)(int, void*))
{
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_SHM_POST_READ);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_SHM_POST_READ);

	MPIU_UNREFERENCED_ARG(rfn);

	MPIDI_DBG_PRINTF((60, FCNAME, L"posting read of %d bytes", len));
	/*wprintf(L"...\n");fflush(stdout);*/
	vc->ch.read.total = 0;
	vc->ch.read.buffer = buf;
	vc->ch.read.bufflen = len;
	vc->ch.read.use_iov = FALSE;
	vc->ch.shm_state |= SHM_READING_BIT;
	vc->ch.shm_reading_pkt = FALSE;
#ifdef USE_SHM_UNEX
	if (vc->ch.unex_list)
		shmi_read_unex(vc);
#endif
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_POST_READ);
	return MPI_SUCCESS;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_SHM_post_readv
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3I_SHM_post_readv(MPIDI_VC_t *vc, MPID_IOV *iov, int n, int (*rfn)(int, void*))
{
#ifdef MPICH_DBG_OUTPUT
	int i;
#endif
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_SHM_POST_READV);
#ifdef USE_SHM_IOV_COPY
	MPIDI_STATE_DECL(MPID_STATE_MEMCPY);
#endif

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_SHM_POST_READV);

	MPIU_UNREFERENCED_ARG(rfn);

	/* FIXME */
	/* Remove this stripping code after the updated segment code no longer
	   produces iov's with empty buffers */
	/* strip any trailing empty buffers */
	/*
	while (n && iov[n-1].MPID_IOV_LEN == 0)
		n--;
	*/
	MPIU_Assert(iov[n-1].MPID_IOV_LEN > 0);

	MPIDI_DBG_PRINTF((60, FCNAME, L"posting read of iov[%d]", n));
	/*wprintf(L".x.x.\n");fflush(stdout);*/
#ifdef MPICH_DBG_OUTPUT
	if (n > 1)
	{
		for (i=0; i<n; i++)
		{
			MPIU_DBG_PRINTF((L"iov[%d].len = %d\n", i, iov[i].MPID_IOV_LEN));
		}
	}
#endif
	vc->ch.read.total = 0;
#ifdef USE_SHM_IOV_COPY
	/* This isn't necessary if we require the iov to be valid for the duration of the operation */
	MPIDI_FUNC_ENTER(MPID_STATE_MEMCPY);
	memcpy(vc->ch.read.iov, iov, sizeof(MPID_IOV) * n);
	MPIDI_FUNC_EXIT(MPID_STATE_MEMCPY);
#else
	vc->ch.read.iov = iov;
#endif
	vc->ch.read.iovlen = n;
	vc->ch.read.index = 0;
	vc->ch.read.use_iov = TRUE;
	vc->ch.shm_state |= SHM_READING_BIT;
	vc->ch.shm_reading_pkt = FALSE;
#ifdef USE_SHM_UNEX
	if (vc->ch.unex_list)
		shmi_readv_unex(vc);
#endif
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_POST_READV);
	return MPI_SUCCESS;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3U_Generate_shm_string
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
void MPIDI_CH3U_Generate_shm_string(wchar_t *str)
{
	UUID guid;
	MPIDI_STATE_DECL(MPID_STATE_GENERATE_SHM_STRING);

	MPIDI_FUNC_ENTER(MPID_STATE_GENERATE_SHM_STRING);
	UuidCreate(&guid);
	MPIU_Snprintf(str, 40, _TRUNCATE, L"%08lX-%04X-%04x-%02X%02X-%02X%02X%02X%02X%02X%02X",
		guid.Data1, guid.Data2, guid.Data3,
		guid.Data4[0], guid.Data4[1], guid.Data4[2], guid.Data4[3],
		guid.Data4[4], guid.Data4[5], guid.Data4[6], guid.Data4[7]);
	MPIU_DBG_PRINTF((L"GUID = %s\n", str));
	MPIDI_FUNC_EXIT(MPID_STATE_GENERATE_SHM_STRING);
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_SHM_Attach_to_memblock
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
/*
   MPIDI_CH3I_SHM_Attach_to_memblock - attach to an existing shmem queue
*/
int MPIDI_CH3I_SHM_Attach_to_memblock(MPIDI_CH3I_Shmem_block_request_result *pInput, MPIDI_CH3I_Shmem_block_request_result *pOutput)
{
	int mpi_errno;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_SHM_ATTACH_TO_MEM);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_SHM_ATTACH_TO_MEM);

	/* Create the shared memory object */
	MPIU_DBG_PRINTF((L"MPIDI_CH3I_SHM_Attach_to_memblock: Creating file mapping of size %d named %s\n", pInput->size, pInput->key));
	pOutput->id = CreateFileMapping( INVALID_HANDLE_VALUE, NULL, PAGE_READWRITE, 0, pInput->size, pInput->key);
	if (pOutput->id == NULL) 
	{
		pOutput->error = GetLastError();
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**CreateFileMapping", L"**CreateFileMapping %d", pOutput->error); /*L"Error in CreateFileMapping, %d", pOutput->error);*/
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_ATTACH_TO_MEM);
		return mpi_errno;
	}

	pOutput->addr = NULL;
	pOutput->addr = MapViewOfFileEx( pOutput->id, FILE_MAP_WRITE, 0, 0, pInput->size, NULL );
	if (pOutput->addr == NULL)
	{
		pOutput->error = GetLastError();
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**MapViewOfFileEx", L"**MapViewOfFileEx %d", pOutput->error);
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_ATTACH_TO_MEM);
		return mpi_errno;
	}

	pOutput->error = MPI_SUCCESS;

	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_ATTACH_TO_MEM);
	return MPI_SUCCESS;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_SHM_Get_memblock_named
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
/*@
   MPIDI_CH3I_SHM_Get_memblock_named - allocate and get the address and size of a shared 
   memory block

   Parameters:
+  int size - size
-  MPIDI_CH3I_Shmem_block_request_result* pOutput - output
@*/
int MPIDI_CH3I_SHM_Get_memblock_named(int size, MPIDI_CH3I_Shmem_block_request_result *pOutput)
{
	int mpi_errno = MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_SHM_GET_MEM_NAMED);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_SHM_GET_MEM_NAMED);

	if (size == 0 || size > MPIDU_MAX_SHM_BLOCK_SIZE )
	{
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**arg", 0);
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_GET_MEM_NAMED);
		return mpi_errno;
	}

	/* Create the shared memory object */
	pOutput->id = CreateFileMapping( INVALID_HANDLE_VALUE, NULL, PAGE_READWRITE, 0, size, pOutput->name);
	if (pOutput->id == NULL) 
	{
		pOutput->error = GetLastError();
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**CreateFileMapping", L"**CreateFileMapping %d", pOutput->error); /*L"Error in CreateFileMapping, %d", pOutput->error);*/
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_GET_MEM_NAMED);
		return mpi_errno;
	}

	/*wprintf(L"[%d] mmapping the shared memory object\n", MPIR_Process.comm_world->rank);fflush(stdout);*/
	pOutput->addr = NULL;
	pOutput->addr = MapViewOfFileEx( pOutput->id, FILE_MAP_WRITE, 0, 0, size, NULL );
	if (pOutput->addr == NULL)
	{
		pOutput->error = GetLastError();
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**MapViewOfFileEx", L"**MapViewOfFileEx %d", pOutput->error);
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_GET_MEM_NAMED);
		return mpi_errno;
	}

	pOutput->size = size;
	pOutput->error = MPI_SUCCESS;

	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_GET_MEM_NAMED);
	return MPI_SUCCESS;
}

/*@
   MPIDI_CH3I_SHM_Unlink_and_detach_mem - 

   Notes: This function unlinks and detaches a shared memory segment. It is called in
		  MPIDI_CH3I_Free_mem (for rma).
		 brad : it is also called in MIPDI_CH3_Finalize and MPIDI_CH3_Win_free 
@*/
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_SHM_Unlink_and_detach_mem
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3I_SHM_Unlink_and_detach_mem(MPIDI_CH3I_Shmem_block_request_result *p)
{
	int mpi_errno=MPI_SUCCESS;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_SHM_UNLINK_AND_DETACH_MEM);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_SHM_UNLINK_AND_DETACH_MEM);

	mpi_errno = MPIDI_CH3I_SHM_Unlink_memblock(p);
	MPIU_ERR_CHK_POP(mpi_errno);

	mpi_errno = MPIDI_CH3I_SHM_Release_memblock(p);

 fn_fail:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_UNLINK_AND_DETACH_MEM);
	return mpi_errno;
}


/*@
   MPIDI_CH3I_SHM_Attach - attach to an existing shared memory without calling unlink

   Parameters:
+  MPIDI_CH3I_Shmem_block_request_result* pInput - input

   Notes: It is used in MPID_CH3I_Win_create

@*/
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_SHM_Attach_notunlink_mem
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3I_SHM_Attach_notunlink_mem(MPIDI_CH3I_Shmem_block_request_result *pInput)
{
	int mpi_errno;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_SHM_ATTACH_NOTUNLINK_MEM);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_SHM_ATTACH_NOTUNLINK_MEM);

	/* Create the shared memory object */
	MPIU_DBG_PRINTF((L"MPIDI_CH3I_SHM_Attach: Creating file mapping of size %d named %s\n", pInput->size, pInput->key));
	pInput->id = CreateFileMapping( INVALID_HANDLE_VALUE, NULL, PAGE_READWRITE, 0, pInput->size, pInput->key );
	if (pInput->id == NULL) 
	{
		pInput->error = GetLastError();
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**CreateFileMapping", L"**CreateFileMapping %d", pInput->error); /*L"Error in CreateFileMapping, %d", pOutput->error);*/
		goto fn_exit;
	}

	pInput->addr = NULL;
	pInput->addr = MapViewOfFileEx( pInput->id, FILE_MAP_WRITE, 0, 0, pInput->size, NULL );
	if (pInput->addr == NULL)
	{
		pInput->error = GetLastError();
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**MapViewOfFileEx", L"**MapViewOfFileEx %d", pInput->error);
		goto fn_exit;
	}

	pInput->error = MPI_SUCCESS;

 fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_ATTACH_NOTUNLINK_MEM);
	return MPI_SUCCESS;
}

/* shmem functions */

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_SHM_read
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3I_SHM_read(MPIDI_VC_t * recv_vc_ptr, void *buf, int len, int *num_bytes_ptr)
{
	int mpi_errno = MPI_SUCCESS;
	void *mem_ptr;
	int num_bytes;
	MPIDI_CH3I_SHM_Packet_t *pkt_ptr;
	MPIDI_CH3I_SHM_Queue_t *shm_ptr;
	register int index;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_SHM_READ);
	MPIDI_STATE_DECL(MPID_STATE_MEMCPY);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_SHM_READ);

	shm_ptr = recv_vc_ptr->ch.read_shmq;
	if (shm_ptr == NULL)
	{
		*num_bytes_ptr = 0;
		goto fn_exit;
	}
	index = shm_ptr->head_index;
	pkt_ptr = &shm_ptr->packet[index];

	/* if the packet at the head index is available, the queue is empty */
	if (pkt_ptr->avail == MPIDI_CH3I_PKT_EMPTY)
	{
		*num_bytes_ptr = 0;
		goto fn_exit;
	}
	MPID_READ_BARRIER(); /* no loads after this line can occur before the avail flag has been read */

	MPIU_DBG_PRINTF((L"MPIDI_CH3I_SHM_read_progress: reading from queue %p\n", shm_ptr));

	mem_ptr = (void*)(pkt_ptr->data + pkt_ptr->offset);
	num_bytes = pkt_ptr->num_bytes;

	MPIDI_DBG_PRINTF((60, FCNAME, L"read %d bytes", num_bytes));
	if (num_bytes > len)
	{
		/* copy the received data */
		MPIDI_DBG_PRINTF((60, FCNAME, L"reading %d bytes from read_shmq %08p packet[%d]", recv_vc_ptr->ch.read.bufflen, shm_ptr, index));
		MPIDI_FUNC_ENTER(MPID_STATE_MEMCPY);
		memcpy(buf, mem_ptr, len);
		MPIDI_FUNC_EXIT(MPID_STATE_MEMCPY);
		*num_bytes_ptr = len;
		pkt_ptr->offset += len;
		pkt_ptr->num_bytes = num_bytes - len;
	}
	else
	{
		/* copy the received data */
		MPIDI_DBG_PRINTF((60, FCNAME, L"reading %d bytes from read_shmq %08p packet[%d]", num_bytes, shm_ptr, index));
		MPIDI_FUNC_ENTER(MPID_STATE_MEMCPY);
		memcpy(buf, mem_ptr, num_bytes);
		MPIDI_FUNC_EXIT(MPID_STATE_MEMCPY);
		*num_bytes_ptr = num_bytes;
		/* put the shmem buffer back in the queue */
		pkt_ptr->offset = 0;
		MPID_READ_WRITE_BARRIER(); /* the writing of the flag cannot occur before the reading of the last piece of data */
		pkt_ptr->avail = MPIDI_CH3I_PKT_EMPTY;
#ifdef MPICH_DBG_OUTPUT
		/*MPIU_Assert(&shm_ptr->packet[index] == pkt_ptr);*/
		if (&shm_ptr->packet[index] != pkt_ptr)
		{
			mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**pkt_ptr", L"**pkt_ptr %p %p", &shm_ptr->packet[index], pkt_ptr);
			goto fn_exit;
		}
#endif
		shm_ptr->head_index = (index + 1) % MPIDI_CH3I_NUM_PACKETS;
		MPIDI_DBG_PRINTF((60, FCNAME, L"read_shmq head = %d", shm_ptr->head_index));
	}

fn_exit:
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_READ);
	return mpi_errno;
}

/*@
   MPIDI_CH3I_SHM_Get_mem - allocate and get address and size of memory shared by all processes. 

   Parameters:
+  int nTotalSize
.  int nRank
-  int nNproc

   Notes:
	Set the global variables pg->ch.addr, pg->ch.size, pg->ch.id
@*/
#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_SHM_Get_mem
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPIDI_CH3I_SHM_Get_mem(MPIDI_PG_t *pg, int nTotalSize, int nRank, int nNproc, BOOL bUseShm)
{
	int mpi_errno;
	int shp_offset;
#ifdef USE_SYNCHRONIZE_SHMAPPING
	HANDLE hSyncEvent1, hSyncEvent2;
#endif
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_SHM_GET_MEM);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_SHM_GET_MEM);

	/* add room at the end of the shard memory region for the shared process information */
	shp_offset = nTotalSize;
	nTotalSize += nNproc * sizeof(MPIDI_CH3I_Shared_process_t);

#ifdef USE_SYNCHRONIZE_SHMAPPING
	hSyncEvent1 = CreateEvent(NULL, TRUE, FALSE, L"mpich2shmsyncevent1");
	hSyncEvent2 = CreateEvent(NULL, TRUE, FALSE, L"mpich2shmsyncevent2");
#else
	MPIU_UNREFERENCED_ARG(nRank);
#endif

	if (nTotalSize < 1)
	{
		/*MPIDI_err_printf(L"MPIDI_CH3I_SHM_Get_mem", L"unable to allocate %d bytes of shared memory: must be greater than zero.\n", nTotalSize);*/
		mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**nomem", 0);
		pg->ch.addr = NULL;
		MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_GET_MEM);
		return mpi_errno;
	}

	if (bUseShm)
	{
		/* Create the shared memory object */
		pg->ch.id = CreateFileMapping(
			INVALID_HANDLE_VALUE,
			NULL,
			PAGE_READWRITE,
			0, 
			nTotalSize,
			pg->ch.key);
		if (pg->ch.id == NULL) 
		{
			mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**CreateFileMapping", L"**CreateFileMapping %d", GetLastError()); /*L"Error in CreateFileMapping, %d\n", GetLastError());*/
			pg->ch.addr = NULL;
			MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_GET_MEM);
			return mpi_errno;
		}

		/* Get the shmem pointer */
#ifdef USE_SYNCHRONIZE_SHMAPPING
		if (nRank == 0)
		{
			ResetEvent(hSyncEvent2);
			SetEvent(hSyncEvent1);
			WaitForSingleObject(hSyncEvent2, INFINITE);
		}
		else
		{
			WaitForSingleObject(hSyncEvent1, INFINITE);
			ResetEvent(hSyncEvent1);
			SetEvent(hSyncEvent2);
		}
#endif
		pg->ch.addr = NULL;
		MPIU_DBG_PRINTF((L"[%d] mapping shared memory\n", nRank));
		pg->ch.addr = MapViewOfFileEx(
			pg->ch.id,
			FILE_MAP_WRITE,
			0, 0,
			nTotalSize,
			NULL
			);
		MPIU_DBG_PRINTF((L"."));
		if (pg->ch.addr == NULL)
		{
			/*MPIDI_err_printf(L"MPIDI_CH3I_SHM_Get_mem", L"Error in MapViewOfFileEx, %d\n", GetLastError());*/
			mpi_errno = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_FATAL, FCNAME, __LINE__, MPI_ERR_OTHER, L"**MapViewOfFileEx", 0);
			pg->ch.addr = NULL;
			return mpi_errno;
		}
		MPIU_DBG_PRINTF((L"\n[%d] finished mapping shared memory: addr:%x\n", nRank, pg->ch.addr));

		pg->ch.pSHP = (MPIDI_CH3I_Shared_process_t*)((char*)pg->ch.addr + shp_offset);
	}
	else
	{
		pg->ch.addr = MPIU_Malloc(nTotalSize);
	}

	MPIU_DBG_PRINTF((L"[%d] made it: shm address: %x\n", nRank, pg->ch.addr));
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_GET_MEM);
	return MPI_SUCCESS;
}

#undef FUNCNAME
#define FUNCNAME MPIDI_CH3I_SHM_Release_mem
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
/*@
   MPIDI_CH3I_SHM_Release_mem - 

   Notes:
@*/
int MPIDI_CH3I_SHM_Release_mem(MPIDI_PG_t *pg, BOOL bUseShm)
{
	int i;
	MPIDI_STATE_DECL(MPID_STATE_MPIDI_CH3I_SHM_RELEASE_MEM);

	MPIDI_FUNC_ENTER(MPID_STATE_MPIDI_CH3I_SHM_RELEASE_MEM);
	
	if (bUseShm)
	{
		UnmapViewOfFile(pg->ch.addr);
		pg->ch.addr = NULL;
		CloseHandle(pg->ch.id);
		pg->ch.id = NULL;
		for (i=0; i<MPIDI_PG_Get_size(pg); i++)
			CloseHandle(pg->ch.pSharedProcessHandles[i]);
		MPIU_Free(pg->ch.pSharedProcessHandles);
		pg->ch.pSharedProcessHandles = NULL;
	}
	else
	{
		MPIU_Free(pg->ch.addr);
	}
	MPIDI_FUNC_EXIT(MPID_STATE_MPIDI_CH3I_SHM_RELEASE_MEM);
	return MPI_SUCCESS;
}

